//============================================================================//
//
//  Software Name : MEGADOCK
//
//  Class Name : Parameter
//
//  Contact address : Tokyo Institute of Technology, AKIYAMA Lab.
//
//============================================================================//

#include "parameter.h"

#include <string>
using namespace std;

//============================================================================//
void Parameter::initialize(int argc,char *argv[])
//============================================================================//
{
    default_param();

    for( int i = 0 ; i < 3 ; i++ ) { //Receptor PDB, Ligand PDB, Outfile
        _IO_flag[i] = 0;
    }

    optind = 1;
    int ch;

    std::string pdb_ext_r_b("_r_b.pdb"); // for input support
    std::string pdb_ext_l_b("_l_b.pdb");
    std::string pdb_ext_r_u("_r_u.pdb");
    std::string pdb_ext_l_u("_l_u.pdb");
    std::string detail_ext(".detail");
    std::string csv_ext(".csv");



    while( (ch = getopt(argc, argv,
                        /*
                            Line 1:input files
                            Line 2:output
                            Line 3:parameters
                            Line 4:calculation options
                        //*/
                        "\
    R:L:B:U:\
    o:ON:t:E:\
    a:b:e:d:F:v:\
    Dhy:z:r:f:i:jkl:mnPS:G:T:"
                       )) != -1 ) {
///    while( (ch = getopt(argc, argv, "OhzF:L:r:N:R:o:t:v:f:e:d:a:b:B:U:")) != -1 ) {
        switch( ch ) {
            //-----------------------------------------------------------------
            // input PDB file -------------------------------------------------

        case 'R':
            _RecPDB_file  = optarg;
            _IO_flag[0]   = 1;
            break;
        case 'L':
            _LigPDB_file  = optarg;
            _IO_flag[1]   = 1;
            break;
        case 'B': // for input support
            _RecPDB_file  = optarg+pdb_ext_r_b;
            _IO_flag[0]   = 1;
            _LigPDB_file  = optarg+pdb_ext_l_b;
            _IO_flag[1]   = 1;
            break;
        case 'U': // for input support
            _RecPDB_file  = optarg+pdb_ext_r_u;
            _IO_flag[0]   = 1;
            _LigPDB_file  = optarg+pdb_ext_l_u;
            _IO_flag[1]   = 1;
            break;

            //-----------------------------------------------------------------
            // output options -------------------------------------------------

        case 'o':
            _RLOut_file = optarg;
            _RLOut_file_detail = optarg+detail_ext;
            _RLOut_file_csv = optarg+csv_ext;
            if(_RLOut_file.length() > 4) {
                if(_RLOut_file.substr(_RLOut_file.length()-4)==".out") {
                    _RLOut_file_detail = _RLOut_file.substr(0,_RLOut_file.length()-4)+detail_ext;;
                    _RLOut_file_csv = _RLOut_file.substr(0,_RLOut_file.length()-4)+csv_ext;;
                }
            }
            _IO_flag[2]   = 1;
            break;
        case 'O':
            detail_output_flag = 1;
            break;
        case 'N':
            _Num_output   = atoi(optarg);
            _Num_output_flag  = 1;
            cout << "#Number of output = " << _Num_output << endl;
            break;
        case 't':
            _Num_sort = atoi(optarg);
            cout << "#Set number of scores per one angle = " << _Num_sort << endl;
            break;
        case 'E':
            calc_time_log_output_flag = atoi(optarg);
            break;
        case 'i':
            calc_id = optarg;
            break;

            //-----------------------------------------------------------------
            // setting parameters ---------------------------------------------

        case 'e':
            _Elec_ratio   = atof(optarg);
            cout << "#Set electric term ratio = "<< _Elec_ratio << endl;
            break;
        case 'd':
            _ACE_ratio = atof(optarg);
            cout << "#Set ACE term ratio = " << _ACE_ratio << endl;
            break;
        case 'a':
            _rPSC_param_rec_core = atof(optarg); // Receptor core
            break;
        case 'b':
            _rPSC_param_lig_core = atof(optarg); // Ligand core
            break;
        case 'F':
            _Num_fft  = atoi(optarg);
            _Num_grid = _Num_fft / 2;
            _Num_fft_flag = 1;
            cout << "#Number of FFT N  = " << _Num_fft << endl;
            break;
            //*
        case 'P':
            _Old_voxel_flag = 1;
            cout << "Using old voxel (MEGADOCK ver.2.5.3)"   << endl;
            break;
            //*/
        case 'v':
            _Grid_size    = atof(optarg);
            cout << "#Set voxel size   = " << _Grid_size << endl;
            break;

            //-----------------------------------------------------------------
            // setting calculation mode ---------------------------------------
        case 'T':
            _Num_thread_limit = atoi(optarg);
            if(_Num_thread_limit<1) {
                printf("Please set a positive number (Maximum number of cores).\n");
                exit(1);
            }
            break;
        case 'G':
            _Num_GPU_limit = atoi(optarg);
            if(_Num_GPU_limit<0) {
                printf("Please set a non-negative number (Maximum number of GPUs).\n");
                exit(1);
            }
            break;
        case 'f':
            _Score_func   = atoi(optarg);
            cout << "#Set score Function = "   << _Score_func << endl;
            assert( 1 <= _Score_func && _Score_func <= 3 );
            if(_Score_func == 1) {
                _Elec_ratio   = 0.0;
                _ACE_ratio    = 0.0;
            } else if(_Score_func == 2) {
                _ACE_ratio    = 0.0;
            }
            _Score_func = 3; //because rpsc() is not prepared
            /*
            if(_Score_func != 3) {
                printf("rpsc() is not prepared in receptor.cpp & ligand.cpp.\n");
                exit(1);
            }
            //*/
            break;
        case 'r':
            _Rotation_angle_set = atoi(optarg);
            if(_Rotation_angle_set == 54000) {
                //cout << "#Set 54,000 rotational angles (6 degree)"<< endl;
            } else if(_Rotation_angle_set == 1) {
                //cout << "#Set 1 rotational angle"<< endl;
            } else if(_Rotation_angle_set == 3) {
                //cout << "#Set 3 rotational angle"<< endl;
            } else if(_Rotation_angle_set == 24) {
                //cout << "#Set 24 rotational angles"<< endl;
            } else if(_Rotation_angle_set == 360) {
                //cout << "#Set 360 rotational angles"<< endl;
            }
            break;
        case 'D':
            _Rotation_angle_set = 54000;
            cout << "#Set 54,000 rotational angles (6 degree)"<< endl;
            break;

            //-----------------------------------------------------------------
            // help -----------------------------------------------------------
        case 'j':
            tem_flag1 = 1;
            break;
        case 'k':
            tem_flag3 = 1;
            break;
        case 'm':
            tem_flag2 = 1;
            break;
        case 'n':
            tem_flag4 = 1;
            break;
        case 'S':
            lig_elec_serial_flag = atoi(optarg);
            break;
        case 'l':
            fft_base_set = atoi(optarg);
            break;

        case 'h':
            usage ();
            break;
        default:
            usage ();
        }
    }

    pdb_step();
    parameter_set();

    return;
}

//============================================================================//
void Parameter::usage()
//============================================================================//
{
    /*
    std::cerr << "megadock [-R receptor_pdb] [-L ligand_pdb] [-o output_file]\n";
    std::cerr << "         [-N num_output] [-t num_select_1angle] [-F num_fft]\n";
    std::cerr << "         [-v voxel_size] [-D (6 deg. search))] [-h (usage)]\n";
    std::cerr << "         [-a rPSC_receptor_core] [-b rPSC_ligand_core]\n";
    std::cerr << "         [-e electro_ratio] [-d ACE_ratio] [-f function_num(1/2/3)]\n" << std::endl;
    */
    /*
    std::cerr << "megadock [-R receptor_pdb] [-L ligand_pdb] [-o output_file]\n";
    std::cerr << "         [-N num_output] [-t num_select_1angle] [-F num_fft]\n";
    std::cerr << "         [-v voxel_size] [-D (6 deg. search))] [-h (usage)]\n" << std::endl;
    */

    std::cerr << "\n Usage: megadock -R [receptor.pdb] -L [ligand.pdb] [more options]\n";
    std::cerr << "\n Available options:\n";
    std::cerr << "  -o filename    : set the output filename (default to \"$R-$L.out\")\n";
    std::cerr << "  -N integer     : set the number of output predictions (default to 2000)\n";
    std::cerr << "  -t integer     : set the number of predictions per each rotation (default to 1)\n";
    std::cerr << "  -F integer     : set the number of FFT point (default to none)\n";
    std::cerr << "  -v float       : set the voxel size (default to 1.2)\n";
    std::cerr << "  -r integer     : set the number of rotational sampling angles \n";
    std::cerr << "                   (54000: 54000 angles, 1: 1 angles, 24: 24 angles, default to 3600 angles)\n";
    std::cerr << "  -e float       : set the electrostatics term ratio (default to 1.0)\n";
    std::cerr << "  -d float       : set the hydrophobic term ratio (default to 1.0)\n";
    std::cerr << "  -a float       : set the rPSC receptor core penalty (default to -45.0)\n";
    std::cerr << "  -b float       : set the rPSC ligand core penalty (default to 1.0)\n";
    std::cerr << "  -f 1/2/3       : set function \n";
    std::cerr << "                   (1: rPSC, 2: rPSC+Elec, 3: rPSC+Elec+Hydrophobic, default to 3)\n";
    std::cerr << "  -h             : show this message\n" << std::endl;

    exit (-1);
}

//============================================================================//
void Parameter::default_param()
//============================================================================//
{
    // Set default parameters and flag initialization
    _Score_func   = 3;        // rPSC + CHARMM + RecACE (2: rPSC+CHARMM, 1: rPSC)
    _Grid_size    = 1.2;
    _Num_grid     = 0;
    _Num_fft      = 0;
    _Num_fft_flag = 0;
    _Num_sort     = 1;
    _Elec_ratio   = 1.0;
    _ACE_ratio    = 1.0;
    _Num_output   = 2000;
    _Num_output_flag = 0;
    _Num_atom_max = 0;
    _Num_thread_limit = 9999999;
    _Num_GPU_limit = 9999999;

    _rPSC_param_rec_core  = -45.0;
    _rPSC_param_lig_core  = 1.0;

    detail_output_flag = 0; // output detailed result
    calc_time_log_output_flag = 0;
    calc_id = "undefined";

    _Rotation_angle_set  = 3600; //3600 rotational angles

    lig_elec_serial_flag = 0;
    fft_base_set = 11;

    tem_flag1 = 0;
    tem_flag2 = 0;
    tem_flag3 = 0;
    tem_flag4 = 0;
    
    ligand_max_edge = 0.0;

    _Old_voxel_flag = 0;
    return;
}

//============================================================================//
void Parameter::pdb_step()
//============================================================================//
{
    if( !_IO_flag[0] ) {
        cerr << "[ERROR] Receptor PDB file is not assignment!!" << endl;
        usage();
        exit(1);
    }

    if( !_IO_flag[1] ) {
        cerr << "[ERROR] Ligand PDB file is not assignment!!" << endl;
        usage();
        exit(1);
    }

    if( !_IO_flag[2] ) {
        string  rfile = _RecPDB_file;
        string  lfile = _LigPDB_file;
        string  ofile;
        int     ipr;
        int     ipl;

        while(1) {
            ipr   = rfile.rfind("/");

            if( ipr == (int) string::npos ) {
                break;
            } else {
                rfile = rfile.substr(ipr+1);
            }
        }

        ipr   = rfile.rfind(".");
        rfile = rfile.substr(0,ipr);
        rfile = rfile + "-";

        while(1) {
            ipl   = lfile.rfind("/");

            if( ipl == (int) string::npos ) {
                break;
            } else {
                lfile = lfile.substr(ipl+1);
            }
        }

        ipl   = lfile.rfind(".");
        lfile = lfile.substr(0,ipl);

        ofile = rfile + lfile + ".out";
        _RLOut_file = ofile;
        ofile = rfile + lfile + ".detail";
        _RLOut_file_detail = ofile;
        ofile = rfile + lfile + ".csv";
        _RLOut_file_csv = ofile;
    }

    //cout << "#Receptor = " << _RecPDB_file << endl;
    //cout << "#Ligand   = " << _LigPDB_file << endl;
    cout << "#Output file = " << _RLOut_file << endl;

    return;
}

//============================================================================//
void Parameter::parameter_set()
//============================================================================//
{
    if( _Rotation_angle_set == 3600 ) {
        dangle_rot3600();
    } else if( _Rotation_angle_set == 54000 ) {
        dangle_rot54000();
    } else if( _Rotation_angle_set == 1 ) {
        dangle_rot1();
    } else if( _Rotation_angle_set == 3 ) {
        dangle_rot3();
    } else if( _Rotation_angle_set == 24 ) {
        dangle_rot24();
    } else if( _Rotation_angle_set == 360 ) {
        dangle_rot360();
    } else {
        cerr << "[ERROR] Not support Rotation_angle_mode = " << _Rotation_angle_set << endl;
        exit(1);
    }

    if( _Num_rot_angles*_Num_sort < _Num_output ) {
        cerr << "#[WARNING] Number of output file was changed from " << _Num_output;
        _Num_output = _Num_rot_angles * _Num_sort;
        cerr << " to " << _Num_output << " (" << _Num_rot_angles << " times " << _Num_sort << ")\n";
    }

    _Grid_space_rec = 8.0;
    _Grid_space_lig = 3.0;

    if( _Score_func == 1 ) {
        charmm_radius();
    } else if( _Score_func == 2 ) {
        charmm_radius();
        charmm_charge();
    } else if( _Score_func == 3 ) {
        charmm_radius();
        charmm_charge();
        ace_value();
    }

    else {
        cerr << "[ERROR] Score function is mismatch!!" << endl;
        exit(1);
    }

    return;
}

//============================================================================//
float Parameter::atom_radius(const string &atype)
//============================================================================//
{
    map<string,float>::iterator itr;
    float defaultr = 1.8;     // Carbon radius

    itr = _Charmmr.find(atype);

    if( itr != _Charmmr.end() ) {
        return itr->second;
    } else {
        return defaultr;
    }
}

//============================================================================//
float Parameter::atom_charge(const string &atype)
//============================================================================//
{
    map<string,float>::iterator itr;
    const float defaultr = 0.0;       // neutral

    itr = _Charmmc.find(atype);

    if( itr != _Charmmc.end() ) {
        return itr->second;
    } else {
        return defaultr;
    }
}

//============================================================================//
float Parameter::atom_ace(const string &atype)
//============================================================================//
{
    map<string,float>::iterator itr;
    const float defaultr = 0.0;           // No Atomic Contact Energy

    itr = _ACE.find(atype);

    if( itr != _ACE.end() ) {
        return itr->second;
    } else {
        return defaultr;
    }
}

//============================================================================//
void Parameter::charmm_radius()
//============================================================================//
{
    _Charmmr.insert(map<string,float>::value_type("N ALA",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ALA",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ALA",1.94));
    _Charmmr.insert(map<string,float>::value_type("C ALA",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ALA",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ALA",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ALA",1.48));
    _Charmmr.insert(map<string,float>::value_type("N ARG",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ARG",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ARG",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG ARG",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD ARG",1.99));
    _Charmmr.insert(map<string,float>::value_type("NE ARG",1.63));
    _Charmmr.insert(map<string,float>::value_type("CZ ARG",1.67));
    _Charmmr.insert(map<string,float>::value_type("NH1 ARG",1.63));
    _Charmmr.insert(map<string,float>::value_type("NH2 ARG",1.63));
    _Charmmr.insert(map<string,float>::value_type("C ARG",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ARG",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ARG",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ARG",1.48));
    _Charmmr.insert(map<string,float>::value_type("N ARGN",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ARGN",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ARGN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG ARGN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD ARGN",1.99));
    _Charmmr.insert(map<string,float>::value_type("NE ARGN",1.63));
    _Charmmr.insert(map<string,float>::value_type("CZ ARGN",1.67));
    _Charmmr.insert(map<string,float>::value_type("NH1 ARGN",1.63));
    _Charmmr.insert(map<string,float>::value_type("NH2 ARGN",1.63));
    _Charmmr.insert(map<string,float>::value_type("C ARGN",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ARGN",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ARGN",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ARGN",1.48));
    _Charmmr.insert(map<string,float>::value_type("N ASN",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ASN",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ASN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG ASN",1.67));
    _Charmmr.insert(map<string,float>::value_type("OD1 ASN",1.38));
    _Charmmr.insert(map<string,float>::value_type("ND2 ASN",1.63));
    _Charmmr.insert(map<string,float>::value_type("C ASN",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ASN",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ASN",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ASN",1.48));
    _Charmmr.insert(map<string,float>::value_type("N ASP",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ASP",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ASP",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG ASP",1.67));
    _Charmmr.insert(map<string,float>::value_type("OD1 ASP",1.48));
    _Charmmr.insert(map<string,float>::value_type("OD2 ASP",1.48));
    _Charmmr.insert(map<string,float>::value_type("C ASP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ASP",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ASP",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ASP",1.48));
    _Charmmr.insert(map<string,float>::value_type("N ASPH",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ASPH",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ASPH",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG ASPH",1.67));
    _Charmmr.insert(map<string,float>::value_type("OD1 ASPH",1.36));
    _Charmmr.insert(map<string,float>::value_type("OD2 ASPH",1.38));
    _Charmmr.insert(map<string,float>::value_type("C ASPH",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ASPH",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ASPH",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ASPH",1.48));
    _Charmmr.insert(map<string,float>::value_type("N CYS",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA CYS",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB CYS",1.99));
    _Charmmr.insert(map<string,float>::value_type("SG CYS",1.69));
    _Charmmr.insert(map<string,float>::value_type("C CYS",1.67));
    _Charmmr.insert(map<string,float>::value_type("O CYS",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT CYS",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 CYS",1.48));
    _Charmmr.insert(map<string,float>::value_type("N GLN",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA GLN",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB GLN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG GLN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD GLN",1.67));
    _Charmmr.insert(map<string,float>::value_type("OE1 GLN",1.38));
    _Charmmr.insert(map<string,float>::value_type("NE2 GLN",1.63));
    _Charmmr.insert(map<string,float>::value_type("C GLN",1.67));
    _Charmmr.insert(map<string,float>::value_type("O GLN",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT GLN",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 GLN",1.48));
    _Charmmr.insert(map<string,float>::value_type("N GLU",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA GLU",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB GLU",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG GLU",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD GLU",1.67));
    _Charmmr.insert(map<string,float>::value_type("OE1 GLU",1.48));
    _Charmmr.insert(map<string,float>::value_type("OE2 GLU",1.48));
    _Charmmr.insert(map<string,float>::value_type("C GLU",1.67));
    _Charmmr.insert(map<string,float>::value_type("O GLU",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT GLU",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 GLU",1.48));
    _Charmmr.insert(map<string,float>::value_type("N GLUH",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA GLUH",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB GLUH",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG GLUH",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD GLUH",1.67));
    _Charmmr.insert(map<string,float>::value_type("OE1 GLUH",1.36));
    _Charmmr.insert(map<string,float>::value_type("OE2 GLUH",1.38));
    _Charmmr.insert(map<string,float>::value_type("C GLUH",1.67));
    _Charmmr.insert(map<string,float>::value_type("O GLUH",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT GLUH",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 GLUH",1.48));
    _Charmmr.insert(map<string,float>::value_type("N GLY",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA GLY",1.99));
    _Charmmr.insert(map<string,float>::value_type("C GLY",1.67));
    _Charmmr.insert(map<string,float>::value_type("O GLY",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT GLY",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 GLY",1.48));
    _Charmmr.insert(map<string,float>::value_type("N HIS",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA HIS",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB HIS",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG HIS",1.82));
    _Charmmr.insert(map<string,float>::value_type("ND1 HIS",1.54));
    _Charmmr.insert(map<string,float>::value_type("CD2 HIS",1.87));
    _Charmmr.insert(map<string,float>::value_type("NE2 HIS",1.54));
    _Charmmr.insert(map<string,float>::value_type("CE1 HIS",1.87));
    _Charmmr.insert(map<string,float>::value_type("C HIS",1.67));
    _Charmmr.insert(map<string,float>::value_type("O HIS",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT HIS",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 HIS",1.48));
    _Charmmr.insert(map<string,float>::value_type("N ILE",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA ILE",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB ILE",2.03));
    _Charmmr.insert(map<string,float>::value_type("CG2 ILE",1.94));
    _Charmmr.insert(map<string,float>::value_type("CG1 ILE",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD1 ILE",1.94));
    _Charmmr.insert(map<string,float>::value_type("C ILE",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ILE",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT ILE",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 ILE",1.48));
    _Charmmr.insert(map<string,float>::value_type("N LEU",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA LEU",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB LEU",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG LEU",2.03));
    _Charmmr.insert(map<string,float>::value_type("CD1 LEU",1.94));
    _Charmmr.insert(map<string,float>::value_type("CD2 LEU",1.94));
    _Charmmr.insert(map<string,float>::value_type("C LEU",1.67));
    _Charmmr.insert(map<string,float>::value_type("O LEU",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT LEU",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 LEU",1.48));
    _Charmmr.insert(map<string,float>::value_type("N LYS",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA LYS",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB LYS",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG LYS",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD LYS",1.99));
    _Charmmr.insert(map<string,float>::value_type("CE LYS",1.99));
    _Charmmr.insert(map<string,float>::value_type("NZ LYS",1.47));
    _Charmmr.insert(map<string,float>::value_type("C LYS",1.67));
    _Charmmr.insert(map<string,float>::value_type("O LYS",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT LYS",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 LYS",1.48));
    _Charmmr.insert(map<string,float>::value_type("N LYSN",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA LYSN",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB LYSN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG LYSN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD LYSN",1.99));
    _Charmmr.insert(map<string,float>::value_type("CE LYSN",1.99));
    _Charmmr.insert(map<string,float>::value_type("NZ LYSN",1.47));
    _Charmmr.insert(map<string,float>::value_type("C LYSN",1.67));
    _Charmmr.insert(map<string,float>::value_type("O LYSN",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT LYSN",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 LYSN",1.48));
    _Charmmr.insert(map<string,float>::value_type("N MET",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA MET",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB MET",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG MET",1.99));
    _Charmmr.insert(map<string,float>::value_type("SD MET",1.76));
    _Charmmr.insert(map<string,float>::value_type("CE MET",1.94));
    _Charmmr.insert(map<string,float>::value_type("C MET",1.67));
    _Charmmr.insert(map<string,float>::value_type("O MET",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT MET",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 MET",1.48));
    _Charmmr.insert(map<string,float>::value_type("N PHE",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA PHE",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB PHE",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG PHE",1.82));
    _Charmmr.insert(map<string,float>::value_type("CD1 PHE",1.78));
    _Charmmr.insert(map<string,float>::value_type("CD2 PHE",1.78));
    _Charmmr.insert(map<string,float>::value_type("CE1 PHE",1.78));
    _Charmmr.insert(map<string,float>::value_type("CE2 PHE",1.78));
    _Charmmr.insert(map<string,float>::value_type("CZ PHE",1.78));
    _Charmmr.insert(map<string,float>::value_type("C PHE",1.67));
    _Charmmr.insert(map<string,float>::value_type("O PHE",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT PHE",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 PHE",1.48));
    _Charmmr.insert(map<string,float>::value_type("N PRO",1.63));
    _Charmmr.insert(map<string,float>::value_type("CD PRO",1.99));
    _Charmmr.insert(map<string,float>::value_type("CA PRO",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB PRO",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG PRO",1.99));
    _Charmmr.insert(map<string,float>::value_type("C PRO",1.67));
    _Charmmr.insert(map<string,float>::value_type("O PRO",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT PRO",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 PRO",1.48));
    _Charmmr.insert(map<string,float>::value_type("N SER",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA SER",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB SER",1.99));
    _Charmmr.insert(map<string,float>::value_type("OG SER",1.38));
    _Charmmr.insert(map<string,float>::value_type("C SER",1.67));
    _Charmmr.insert(map<string,float>::value_type("O SER",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT SER",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 SER",1.48));
    _Charmmr.insert(map<string,float>::value_type("N THR",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA THR",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB THR",2.03));
    _Charmmr.insert(map<string,float>::value_type("OG1 THR",1.38));
    _Charmmr.insert(map<string,float>::value_type("CG2 THR",1.94));
    _Charmmr.insert(map<string,float>::value_type("C THR",1.67));
    _Charmmr.insert(map<string,float>::value_type("O THR",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT THR",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 THR",1.48));
    _Charmmr.insert(map<string,float>::value_type("N TRP",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA TRP",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB TRP",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG TRP",1.82));
    _Charmmr.insert(map<string,float>::value_type("CD2 TRP",1.82));
    _Charmmr.insert(map<string,float>::value_type("CE2 TRP",1.82));
    _Charmmr.insert(map<string,float>::value_type("CE3 TRP",1.78));
    _Charmmr.insert(map<string,float>::value_type("CD1 TRP",1.87));
    _Charmmr.insert(map<string,float>::value_type("NE1 TRP",1.54));
    _Charmmr.insert(map<string,float>::value_type("CZ2 TRP",1.78));
    _Charmmr.insert(map<string,float>::value_type("CZ3 TRP",1.78));
    _Charmmr.insert(map<string,float>::value_type("CH2 TRP",1.78));
    _Charmmr.insert(map<string,float>::value_type("C TRP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O TRP",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT TRP",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 TRP",1.48));
    _Charmmr.insert(map<string,float>::value_type("N TYR",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA TYR",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB TYR",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG TYR",1.82));
    _Charmmr.insert(map<string,float>::value_type("CD1 TYR",1.78));
    _Charmmr.insert(map<string,float>::value_type("CE1 TYR",1.78));
    _Charmmr.insert(map<string,float>::value_type("CD2 TYR",1.78));
    _Charmmr.insert(map<string,float>::value_type("CE2 TYR",1.78));
    _Charmmr.insert(map<string,float>::value_type("CZ TYR",1.82));
    _Charmmr.insert(map<string,float>::value_type("OH TYR",1.38));
    _Charmmr.insert(map<string,float>::value_type("C TYR",1.67));
    _Charmmr.insert(map<string,float>::value_type("O TYR",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT TYR",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 TYR",1.48));
    _Charmmr.insert(map<string,float>::value_type("N VAL",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA VAL",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB VAL",2.03));
    _Charmmr.insert(map<string,float>::value_type("CG1 VAL",1.94));
    _Charmmr.insert(map<string,float>::value_type("CG2 VAL",1.94));
    _Charmmr.insert(map<string,float>::value_type("C VAL",1.67));
    _Charmmr.insert(map<string,float>::value_type("O VAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("OXT VAL",1.48));
    _Charmmr.insert(map<string,float>::value_type("OT2 VAL",1.48));
    _Charmmr.insert(map<string,float>::value_type("CM ME",1.94));
    _Charmmr.insert(map<string,float>::value_type("C CH3",1.94));
    _Charmmr.insert(map<string,float>::value_type("C ACE",1.67));
    _Charmmr.insert(map<string,float>::value_type("O ACE",1.38));
    _Charmmr.insert(map<string,float>::value_type("CH3 ACE",1.94));
    _Charmmr.insert(map<string,float>::value_type("N 5HP",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA 5HP",2.03));
    _Charmmr.insert(map<string,float>::value_type("C 5HP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O 5HP",1.38));
    _Charmmr.insert(map<string,float>::value_type("CB 5HP",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG 5HP",1.99));
    _Charmmr.insert(map<string,float>::value_type("CD 5HP",1.99));
    _Charmmr.insert(map<string,float>::value_type("OD 5HP",1.38));
    _Charmmr.insert(map<string,float>::value_type("N MSE",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA MSE",2.03));
    _Charmmr.insert(map<string,float>::value_type("C MSE",1.67));
    _Charmmr.insert(map<string,float>::value_type("O MSE",1.38));
    _Charmmr.insert(map<string,float>::value_type("CB MSE",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG MSE",1.99));
    _Charmmr.insert(map<string,float>::value_type("SE MSE",1.76));
    _Charmmr.insert(map<string,float>::value_type("CE MSE",1.94));
    _Charmmr.insert(map<string,float>::value_type("N SOC",1.63));
    _Charmmr.insert(map<string,float>::value_type("CA SOC",2.03));
    _Charmmr.insert(map<string,float>::value_type("C SOC",1.67));
    _Charmmr.insert(map<string,float>::value_type("O SOC",1.38));
    _Charmmr.insert(map<string,float>::value_type("CB SOC",1.99));
    _Charmmr.insert(map<string,float>::value_type("SE SOC",1.69));
    _Charmmr.insert(map<string,float>::value_type("OD1 SOC",1.48));
    _Charmmr.insert(map<string,float>::value_type("OD2 SOC",1.48));
    _Charmmr.insert(map<string,float>::value_type("N SAH",1.47));
    _Charmmr.insert(map<string,float>::value_type("CA SAH",2.03));
    _Charmmr.insert(map<string,float>::value_type("CB SAH",1.99));
    _Charmmr.insert(map<string,float>::value_type("CG SAH",1.99));
    _Charmmr.insert(map<string,float>::value_type("SD SAH",1.76));
    _Charmmr.insert(map<string,float>::value_type("C SAH",1.67));
    _Charmmr.insert(map<string,float>::value_type("O SAH",1.48));
    _Charmmr.insert(map<string,float>::value_type("OXT SAH",1.48));
    _Charmmr.insert(map<string,float>::value_type("C5* SAH",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* SAH",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* SAH",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* SAH",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* SAH",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* SAH",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* SAH",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* SAH",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 SAH",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 SAH",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 SAH",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 SAH",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 SAH",1.67));
    _Charmmr.insert(map<string,float>::value_type("N6 SAH",1.63));
    _Charmmr.insert(map<string,float>::value_type("N1 SAH",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 SAH",1.82));
    _Charmmr.insert(map<string,float>::value_type("N3 SAH",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 SAH",1.82));
    _Charmmr.insert(map<string,float>::value_type("PG ATP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1G ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2G ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3G ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3B ATP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PB ATP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1B ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2B ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3A ATP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA ATP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1A ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A ATP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5* ATP",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* ATP",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* ATP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* ATP",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* ATP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* ATP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* ATP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* ATP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* ATP",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 ATP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 ATP",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 ATP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 ATP",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 ATP",1.67));
    _Charmmr.insert(map<string,float>::value_type("N6 ATP",1.63));
    _Charmmr.insert(map<string,float>::value_type("N1 ATP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 ATP",1.82));
    _Charmmr.insert(map<string,float>::value_type("N3 ATP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 ATP",1.82));
    _Charmmr.insert(map<string,float>::value_type("PC GCP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1C GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2C GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3C GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("C3B GCP",1.67));
    _Charmmr.insert(map<string,float>::value_type("PB GCP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1B GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2B GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3A GCP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA GCP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1A GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A GCP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5* GCP",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* GCP",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* GCP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* GCP",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* GCP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* GCP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* GCP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* GCP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* GCP",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 GCP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 GCP",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 GCP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 GCP",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 GCP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O6 GCP",1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 GCP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 GCP",1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 GCP",1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 GCP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 GCP",1.82));
    _Charmmr.insert(map<string,float>::value_type("PB GDP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1B GDP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2B GDP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3B GDP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3A GDP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA GDP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1A GDP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A GDP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5* GDP",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* GDP",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* GDP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* GDP",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* GDP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* GDP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* GDP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* GDP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* GDP",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 GDP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 GDP",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 GDP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 GDP",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 GDP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O6 GDP",1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 GDP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 GDP",1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 GDP",1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 GDP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 GDP",1.82));
    _Charmmr.insert(map<string,float>::value_type("PG GNP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1G GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2G GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3G GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("N3B GNP",1.63));
    _Charmmr.insert(map<string,float>::value_type("PB GNP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1B GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2B GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3A GNP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA GNP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1A GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A GNP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5* GNP",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* GNP",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* GNP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* GNP",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* GNP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* GNP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* GNP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* GNP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* GNP",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 GNP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 GNP",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 GNP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 GNP",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 GNP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O6 GNP",1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 GNP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 GNP",1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 GNP",1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 GNP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 GNP",1.82));
    _Charmmr.insert(map<string,float>::value_type("PG GSP",1.63));
    _Charmmr.insert(map<string,float>::value_type("S1G GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2G GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3G GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3B GSP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PB GSP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1B GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2B GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3A GSP",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA GSP",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1A GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A GSP",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5* GSP",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* GSP",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* GSP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* GSP",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* GSP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* GSP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* GSP",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* GSP",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* GSP",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 GSP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 GSP",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 GSP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 GSP",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 GSP",1.67));
    _Charmmr.insert(map<string,float>::value_type("O6 GSP",1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 GSP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 GSP",1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 GSP",1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 GSP",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 GSP",1.82));
    _Charmmr.insert(map<string,float>::value_type("PG GTN",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1G GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2G GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3G GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("N3B GTN",1.63));
    _Charmmr.insert(map<string,float>::value_type("PB GTN",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1B GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2B GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3A GTN",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA GTN",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1A GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A GTN",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5* GTN",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* GTN",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4* GTN",2.02));
    _Charmmr.insert(map<string,float>::value_type("O4* GTN",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3* GTN",2.02));
    _Charmmr.insert(map<string,float>::value_type("O3* GTN",1.38));
    _Charmmr.insert(map<string,float>::value_type("C2* GTN",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2* GTN",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1* GTN",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 GTN",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 GTN",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7 GTN",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 GTN",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 GTN",1.67));
    _Charmmr.insert(map<string,float>::value_type("O6 GTN",1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 GTN",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 GTN",1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 GTN",1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 GTN",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 GTN",1.82));
    _Charmmr.insert(map<string,float>::value_type("C1T COT",1.99));
    _Charmmr.insert(map<string,float>::value_type("CGT COT",1.82));
    _Charmmr.insert(map<string,float>::value_type("CD2T COT",1.82));
    _Charmmr.insert(map<string,float>::value_type("CE2T COT",1.82));
    _Charmmr.insert(map<string,float>::value_type("CE3T COT",1.78));
    _Charmmr.insert(map<string,float>::value_type("CD1T COT",1.87));
    _Charmmr.insert(map<string,float>::value_type("NE1T COT",1.54));
    _Charmmr.insert(map<string,float>::value_type("CZ2T COT",1.78));
    _Charmmr.insert(map<string,float>::value_type("CZ3T COT",1.78));
    _Charmmr.insert(map<string,float>::value_type("CH2T COT",1.78));
    _Charmmr.insert(map<string,float>::value_type("S COT",1.76));
    _Charmmr.insert(map<string,float>::value_type("C6T COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("C4T COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("O5T COT",1.38));
    _Charmmr.insert(map<string,float>::value_type("N3T COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("C2T COT",1.99));
    _Charmmr.insert(map<string,float>::value_type("OPB3 COT",1.43));
    _Charmmr.insert(map<string,float>::value_type("PB COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("OPB1 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("OPB2 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("OPA3 COT",1.43));
    _Charmmr.insert(map<string,float>::value_type("O3*A COT",1.43));
    _Charmmr.insert(map<string,float>::value_type("PA COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("OPA2 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("OPA1 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("O5*A COT",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5*A COT",2.00));
    _Charmmr.insert(map<string,float>::value_type("C4*A COT",2.02));
    _Charmmr.insert(map<string,float>::value_type("C3*A COT",2.02));
    _Charmmr.insert(map<string,float>::value_type("O2*A COT",1.38));
    _Charmmr.insert(map<string,float>::value_type("O4*A COT",2.02));
    _Charmmr.insert(map<string,float>::value_type("PC COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("OPC1 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("OPC2 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("OPC3 COT",1.48));
    _Charmmr.insert(map<string,float>::value_type("C2*A COT",2.02));
    _Charmmr.insert(map<string,float>::value_type("C1*A COT",2.02));
    _Charmmr.insert(map<string,float>::value_type("N9A COT",1.54));
    _Charmmr.insert(map<string,float>::value_type("C8A COT",1.88));
    _Charmmr.insert(map<string,float>::value_type("N7A COT",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5A COT",1.82));
    _Charmmr.insert(map<string,float>::value_type("C6A COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("N6A COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("N1A COT",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2A COT",1.82));
    _Charmmr.insert(map<string,float>::value_type("N3A COT",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4A COT",1.82));
    _Charmmr.insert(map<string,float>::value_type("N3P COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("C4P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("O5P COT",1.38));
    _Charmmr.insert(map<string,float>::value_type("C6P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("C7P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("N8P COT",1.63));
    _Charmmr.insert(map<string,float>::value_type("C9P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("O10P COT",1.38));
    _Charmmr.insert(map<string,float>::value_type("C11P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("O12P COT",1.38));
    _Charmmr.insert(map<string,float>::value_type("C13P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("C16P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("C14P COT",1.94));
    _Charmmr.insert(map<string,float>::value_type("C15P COT",1.94));
    _Charmmr.insert(map<string,float>::value_type("C2P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("C1P COT",1.67));
    _Charmmr.insert(map<string,float>::value_type("AP FAD",1.63));
    _Charmmr.insert(map<string,float>::value_type("AO1 FAD",1.48));
    _Charmmr.insert(map<string,float>::value_type("AO2 FAD",1.48));
    _Charmmr.insert(map<string,float>::value_type("AO5* FAD",1.43));
    _Charmmr.insert(map<string,float>::value_type("AC5* FAD",2.00));
    _Charmmr.insert(map<string,float>::value_type("AC4* FAD",2.02));
    _Charmmr.insert(map<string,float>::value_type("AO4* FAD",2.02));
    _Charmmr.insert(map<string,float>::value_type("AC3* FAD",2.02));
    _Charmmr.insert(map<string,float>::value_type("AO3* FAD",1.38));
    _Charmmr.insert(map<string,float>::value_type("AC2* FAD",2.02));
    _Charmmr.insert(map<string,float>::value_type("AO2* FAD",1.38));
    _Charmmr.insert(map<string,float>::value_type("AC1* FAD",2.02));
    _Charmmr.insert(map<string,float>::value_type("AN9 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("AC8 FAD",1.88));
    _Charmmr.insert(map<string,float>::value_type("AN7 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("AC5 FAD",1.82));
    _Charmmr.insert(map<string,float>::value_type("AC6 FAD",1.67));
    _Charmmr.insert(map<string,float>::value_type("AN6 FAD",1.63));
    _Charmmr.insert(map<string,float>::value_type("AN1 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("AC2 FAD",1.82));
    _Charmmr.insert(map<string,float>::value_type("AN3 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("AC4 FAD",1.82));
    _Charmmr.insert(map<string,float>::value_type("N1 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 FAD",1.67));
    _Charmmr.insert(map<string,float>::value_type("O2 FAD",1.38));
    _Charmmr.insert(map<string,float>::value_type("N3 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 FAD",1.67));
    _Charmmr.insert(map<string,float>::value_type("O4 FAD",1.38));
    _Charmmr.insert(map<string,float>::value_type("C4A FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("N5 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("C5A FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("C6 FAD",1.78));
    _Charmmr.insert(map<string,float>::value_type("C7 FAD",1.78));
    _Charmmr.insert(map<string,float>::value_type("C7M FAD",1.94));
    _Charmmr.insert(map<string,float>::value_type("C8 FAD",1.78));
    _Charmmr.insert(map<string,float>::value_type("C8M FAD",1.94));
    _Charmmr.insert(map<string,float>::value_type("C9 FAD",1.78));
    _Charmmr.insert(map<string,float>::value_type("C9A FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("N10 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("C10 FAD",1.54));
    _Charmmr.insert(map<string,float>::value_type("C1* FAD",2.03));
    _Charmmr.insert(map<string,float>::value_type("C2* FAD",2.03));
    _Charmmr.insert(map<string,float>::value_type("O2* FAD",1.43));
    _Charmmr.insert(map<string,float>::value_type("C3* FAD",2.03));
    _Charmmr.insert(map<string,float>::value_type("O3* FAD",1.43));
    _Charmmr.insert(map<string,float>::value_type("C4* FAD",2.03));
    _Charmmr.insert(map<string,float>::value_type("O4* FAD",1.43));
    _Charmmr.insert(map<string,float>::value_type("C5* FAD",2.03));
    _Charmmr.insert(map<string,float>::value_type("O5* FAD",1.43));
    _Charmmr.insert(map<string,float>::value_type("P FAD",1.63));
    _Charmmr.insert(map<string,float>::value_type("O1P FAD",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2P FAD",1.48));
    _Charmmr.insert(map<string,float>::value_type("O3P FAD",1.43));
    _Charmmr.insert(map<string,float>::value_type("C1 FUC",1.78));
    _Charmmr.insert(map<string,float>::value_type("C2 FUC",1.78));
    _Charmmr.insert(map<string,float>::value_type("C3 FUC",1.78));
    _Charmmr.insert(map<string,float>::value_type("C4 FUC",1.78));
    _Charmmr.insert(map<string,float>::value_type("C5 FUC",1.78));
    _Charmmr.insert(map<string,float>::value_type("C6 FUC",1.94));
    _Charmmr.insert(map<string,float>::value_type("O1 FUC",1.38));
    _Charmmr.insert(map<string,float>::value_type("O2 FUC",1.38));
    _Charmmr.insert(map<string,float>::value_type("O3 FUC",1.38));
    _Charmmr.insert(map<string,float>::value_type("O4 FUC",1.38));
    _Charmmr.insert(map<string,float>::value_type("O5 FUC",1.43));
    _Charmmr.insert(map<string,float>::value_type("C1 GAL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C2 GAL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C3 GAL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C4 GAL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C5 GAL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C6 GAL",1.78));
    _Charmmr.insert(map<string,float>::value_type("O1 GAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("O2 GAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("O3 GAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("O4 GAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("O5 GAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("O6 GAL",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1 MAN",1.78));
    _Charmmr.insert(map<string,float>::value_type("C2 MAN",1.78));
    _Charmmr.insert(map<string,float>::value_type("C3 MAN",1.78));
    _Charmmr.insert(map<string,float>::value_type("C4 MAN",1.78));
    _Charmmr.insert(map<string,float>::value_type("C5 MAN",1.78));
    _Charmmr.insert(map<string,float>::value_type("C6 MAN",1.78));
    _Charmmr.insert(map<string,float>::value_type("O1 MAN",1.38));
    _Charmmr.insert(map<string,float>::value_type("O2 MAN",1.38));
    _Charmmr.insert(map<string,float>::value_type("O3 MAN",1.38));
    _Charmmr.insert(map<string,float>::value_type("O4 MAN",1.38));
    _Charmmr.insert(map<string,float>::value_type("O5 MAN",1.38));
    _Charmmr.insert(map<string,float>::value_type("O6 MAN",1.38));
    _Charmmr.insert(map<string,float>::value_type("C1 NAG",1.78));
    _Charmmr.insert(map<string,float>::value_type("C2 NAG",1.78));
    _Charmmr.insert(map<string,float>::value_type("C3 NAG",1.78));
    _Charmmr.insert(map<string,float>::value_type("C4 NAG",1.78));
    _Charmmr.insert(map<string,float>::value_type("C5 NAG",1.78));
    _Charmmr.insert(map<string,float>::value_type("C6 NAG",1.78));
    _Charmmr.insert(map<string,float>::value_type("C7 NAG",2.03));
    _Charmmr.insert(map<string,float>::value_type("C8 NAG",1.94));
    _Charmmr.insert(map<string,float>::value_type("N2 NAG",1.63));
    _Charmmr.insert(map<string,float>::value_type("O3 NAG",1.38));
    _Charmmr.insert(map<string,float>::value_type("O4 NAG",1.38));
    _Charmmr.insert(map<string,float>::value_type("O5 NAG",1.38));
    _Charmmr.insert(map<string,float>::value_type("O6 NAG",1.38));
    _Charmmr.insert(map<string,float>::value_type("O7 NAG",1.38));
    _Charmmr.insert(map<string,float>::value_type("FE HEM",2.13));
    _Charmmr.insert(map<string,float>::value_type("CHA HEM",2.03));
    _Charmmr.insert(map<string,float>::value_type("CHB HEM",2.03));
    _Charmmr.insert(map<string,float>::value_type("CHC HEM",2.03));
    _Charmmr.insert(map<string,float>::value_type("CHD HEM",2.03));
    _Charmmr.insert(map<string,float>::value_type("N A HEM",1.54));
    _Charmmr.insert(map<string,float>::value_type("C1A HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C2A HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C3A HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C4A HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("CMA HEM",1.94));
    _Charmmr.insert(map<string,float>::value_type("CAA HEM",1.99));
    _Charmmr.insert(map<string,float>::value_type("CBA HEM",1.99));
    _Charmmr.insert(map<string,float>::value_type("CGA HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("O1A HEM",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2A HEM",1.48));
    _Charmmr.insert(map<string,float>::value_type("N B HEM",1.54));
    _Charmmr.insert(map<string,float>::value_type("C1B HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C2B HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C3B HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C4B HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("CMB HEM",1.94));
    _Charmmr.insert(map<string,float>::value_type("CAB HEM",2.03));
    _Charmmr.insert(map<string,float>::value_type("CBB HEM",1.94));
    _Charmmr.insert(map<string,float>::value_type("N C HEM",1.54));
    _Charmmr.insert(map<string,float>::value_type("C1C HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C2C HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C3C HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C4C HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("CMC HEM",1.94));
    _Charmmr.insert(map<string,float>::value_type("CAC HEM",2.03));
    _Charmmr.insert(map<string,float>::value_type("CBC HEM",1.99));
    _Charmmr.insert(map<string,float>::value_type("N D HEM",1.54));
    _Charmmr.insert(map<string,float>::value_type("C1D HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C2D HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C3D HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("C4D HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("CMD HEM",1.94));
    _Charmmr.insert(map<string,float>::value_type("CAD HEM",1.99));
    _Charmmr.insert(map<string,float>::value_type("CBD HEM",1.99));
    _Charmmr.insert(map<string,float>::value_type("CGD HEM",1.67));
    _Charmmr.insert(map<string,float>::value_type("O1D HEM",1.48));
    _Charmmr.insert(map<string,float>::value_type("O2D HEM",1.48));
    _Charmmr.insert(map<string,float>::value_type("C1 RTL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C2 RTL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C3 RTL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C4 RTL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C5 RTL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C6 RTL",1.78));
    _Charmmr.insert(map<string,float>::value_type("C7 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C8 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C9 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C10 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C11 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C12 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C13 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C14 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C15 RTL",2.03));
    _Charmmr.insert(map<string,float>::value_type("C16 RTL",1.94));
    _Charmmr.insert(map<string,float>::value_type("C17 RTL",1.94));
    _Charmmr.insert(map<string,float>::value_type("C18 RTL",1.94));
    _Charmmr.insert(map<string,float>::value_type("C19 RTL",1.94));
    _Charmmr.insert(map<string,float>::value_type("C20 RTL",1.94));
    _Charmmr.insert(map<string,float>::value_type("OR RTL",1.38));
    _Charmmr.insert(map<string,float>::value_type("P A", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 A", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 A", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP3 A", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' A", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' A", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' A", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' A", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' A", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' A", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' A", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O2' A", 1.42));
    _Charmmr.insert(map<string,float>::value_type("C1' A", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("N7 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C6 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("N6 A", 1.63));
    _Charmmr.insert(map<string,float>::value_type("N1 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 A", 1.82));
    _Charmmr.insert(map<string,float>::value_type("N3 A", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 A", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P G", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 G", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 G", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP3 G", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' G", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' G", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' G", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' G", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' G", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' G", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' G", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O2' G", 1.42));
    _Charmmr.insert(map<string,float>::value_type("C1' G", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("N7 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C6 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("O6 G", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 G", 1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 G", 1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 G", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 G", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P C", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 C", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 C", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP3 C", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' C", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' C", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' C", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' C", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' C", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' C", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' C", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O2' C", 1.42));
    _Charmmr.insert(map<string,float>::value_type("C1' C", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N1 C", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 C", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O2 C", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N3 C", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 C", 1.82));
    _Charmmr.insert(map<string,float>::value_type("N4 C", 1.63));
    _Charmmr.insert(map<string,float>::value_type("C5 C", 1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 C", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P U", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 U", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 U", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP3 U", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' U", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' U", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' U", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' U", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' U", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' U", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' U", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O2' U", 1.42));
    _Charmmr.insert(map<string,float>::value_type("C1' U", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N1 U", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 U", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O2 U", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N3 U", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 U", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O4 U", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C5 U", 1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 U", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P DA", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 DA", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 DA", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' DA", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' DA", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' DA", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' DA", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' DA", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' DA", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' DA", 2.02));
    _Charmmr.insert(map<string,float>::value_type("C1' DA", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("N7 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C6 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("N6 DA", 1.63));
    _Charmmr.insert(map<string,float>::value_type("N1 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 DA", 1.82));
    _Charmmr.insert(map<string,float>::value_type("N3 DA", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 DA", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P DG", 1.700));
    _Charmmr.insert(map<string,float>::value_type("OP1 DG", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 DG", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' DG", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' DG", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' DG", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' DG", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' DG", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' DG", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' DG", 2.02));
    _Charmmr.insert(map<string,float>::value_type("C1' DG", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N9 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C8 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("N7 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C5 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C6 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("O6 DG", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N1 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 DG", 1.82));
    _Charmmr.insert(map<string,float>::value_type("N2 DG", 1.63));
    _Charmmr.insert(map<string,float>::value_type("N3 DG", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 DG", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P DC", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 DC", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 DC", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' DC", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' DC", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' DC", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' DC", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' DC", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' DC", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' DC", 2.02));
    _Charmmr.insert(map<string,float>::value_type("C1' DC", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N1 DC", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 DC", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O2 DC", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N3 DC", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 DC", 1.82));
    _Charmmr.insert(map<string,float>::value_type("N4 DC", 1.63));
    _Charmmr.insert(map<string,float>::value_type("C5 DC", 1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 DC", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P DT", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 DT", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 DT", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' DT", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' DT", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' DT", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' DT", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' DT", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' DT", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' DT", 2.02));
    _Charmmr.insert(map<string,float>::value_type("C1' DT", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N1 DT", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 DT", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O2 DT", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N3 DT", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 DT", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O4 DT", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C5 DT", 1.82));
    _Charmmr.insert(map<string,float>::value_type("C5M DT", 1.94));
    _Charmmr.insert(map<string,float>::value_type("C6 DT", 1.82));
    _Charmmr.insert(map<string,float>::value_type("P DU", 1.70));
    _Charmmr.insert(map<string,float>::value_type("OP1 DU", 1.48));
    _Charmmr.insert(map<string,float>::value_type("OP2 DU", 1.48));
    _Charmmr.insert(map<string,float>::value_type("O5' DU", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C5' DU", 2.00));
    _Charmmr.insert(map<string,float>::value_type("C4' DU", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O4' DU", 1.43));
    _Charmmr.insert(map<string,float>::value_type("C3' DU", 2.02));
    _Charmmr.insert(map<string,float>::value_type("O3' DU", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C2' DU", 2.02));
    _Charmmr.insert(map<string,float>::value_type("C1' DU", 2.02));
    _Charmmr.insert(map<string,float>::value_type("N1 DU", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C2 DU", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O2 DU", 1.38));
    _Charmmr.insert(map<string,float>::value_type("N3 DU", 1.54));
    _Charmmr.insert(map<string,float>::value_type("C4 DU", 1.67));
    _Charmmr.insert(map<string,float>::value_type("O4 DU", 1.38));
    _Charmmr.insert(map<string,float>::value_type("C5 DU", 1.82));
    _Charmmr.insert(map<string,float>::value_type("C6 DU", 1.82));

    return;
}

//============================================================================//
void Parameter::charmm_charge()
//============================================================================//
{
    _Charmmc.insert(map<string,float>::value_type("N ALA",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ALA", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ALA", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C ALA", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ALA",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ALA",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ALA",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N ARG",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ARG", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ARG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG ARG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD ARG", 0.10));
    _Charmmc.insert(map<string,float>::value_type("NE ARG",-0.10));
    _Charmmc.insert(map<string,float>::value_type("CZ ARG", 0.50));
    _Charmmc.insert(map<string,float>::value_type("NH1 ARG", 0.25));
    _Charmmc.insert(map<string,float>::value_type("NH2 ARG", 0.25));
    _Charmmc.insert(map<string,float>::value_type("C ARG", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ARG",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ARG",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ARG",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N ARGN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ARGN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ARGN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG ARGN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD ARGN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("NE ARGN",-0.10));
    _Charmmc.insert(map<string,float>::value_type("CZ ARGN",-0.20));
    _Charmmc.insert(map<string,float>::value_type("NH1 ARGN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("NH2 ARGN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C ARGN", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ARGN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ARGN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ARGN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N ASN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ASN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ASN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG ASN", 0.55));
    _Charmmc.insert(map<string,float>::value_type("OD1 ASN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("ND2 ASN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C ASN", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ASN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ASN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ASN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N ASP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ASP", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ASP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG ASP", 0.14));
    _Charmmc.insert(map<string,float>::value_type("OD1 ASP",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OD2 ASP",-0.57));
    _Charmmc.insert(map<string,float>::value_type("C ASP", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ASP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ASP",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ASP",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N ASPH",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ASPH", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ASPH", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG ASPH", 0.70));
    _Charmmc.insert(map<string,float>::value_type("OD1 ASPH",-0.35));
    _Charmmc.insert(map<string,float>::value_type("OD2 ASPH",-0.35));
    _Charmmc.insert(map<string,float>::value_type("C ASPH", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ASPH",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ASPH",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ASPH",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N CYS",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA CYS", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB CYS", 0.19));
    _Charmmc.insert(map<string,float>::value_type("SG CYS",-0.19));
    _Charmmc.insert(map<string,float>::value_type("C CYS", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O CYS",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT CYS",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 CYS",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N GLN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA GLN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB GLN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG GLN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD GLN", 0.55));
    _Charmmc.insert(map<string,float>::value_type("OE1 GLN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("NE2 GLN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C GLN", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O GLN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT GLN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 GLN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N GLU",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA GLU", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB GLU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG GLU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD GLU", 0.14));
    _Charmmc.insert(map<string,float>::value_type("OE1 GLU",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OE2 GLU",-0.57));
    _Charmmc.insert(map<string,float>::value_type("C GLU", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O GLU",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT GLU",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 GLU",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N GLUH",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA GLUH", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB GLUH", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG GLUH", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD GLUH", 0.70));
    _Charmmc.insert(map<string,float>::value_type("OE1 GLUH",-0.35));
    _Charmmc.insert(map<string,float>::value_type("OE2 GLUH",-0.35));
    _Charmmc.insert(map<string,float>::value_type("C GLUH", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O GLUH",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT GLUH",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 GLUH",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N GLY",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA GLY", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C GLY", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O GLY",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT GLY",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 GLY",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N HIS",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA HIS", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB HIS", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG HIS", 0.10));
    _Charmmc.insert(map<string,float>::value_type("ND1 HIS",-0.10));
    _Charmmc.insert(map<string,float>::value_type("CD2 HIS", 0.10));
    _Charmmc.insert(map<string,float>::value_type("NE2 HIS",-0.40));
    _Charmmc.insert(map<string,float>::value_type("CE1 HIS", 0.30));
    _Charmmc.insert(map<string,float>::value_type("C HIS", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O HIS",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT HIS",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 HIS",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N ILE",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA ILE", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB ILE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG2 ILE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG1 ILE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD1 ILE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C ILE", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ILE",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT ILE",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 ILE",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N LEU",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA LEU", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB LEU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG LEU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD1 LEU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD2 LEU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C LEU", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O LEU",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT LEU",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 LEU",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N LYS",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA LYS", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB LYS", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG LYS", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD LYS", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CE LYS", 0.25));
    _Charmmc.insert(map<string,float>::value_type("NZ LYS", 0.75));
    _Charmmc.insert(map<string,float>::value_type("C LYS", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O LYS",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT LYS",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 LYS",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N LYSN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA LYSN", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB LYSN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG LYSN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD LYSN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CE LYSN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("NZ LYSN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C LYSN", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O LYSN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT LYSN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 LYSN",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N MET",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA MET", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB MET", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG MET", 0.06));
    _Charmmc.insert(map<string,float>::value_type("SD MET",-0.12));
    _Charmmc.insert(map<string,float>::value_type("CE MET", 0.06));
    _Charmmc.insert(map<string,float>::value_type("C MET", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O MET",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT MET",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 MET",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N PHE",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA PHE", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD1 PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD2 PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CE1 PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CE2 PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CZ PHE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C PHE", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O PHE",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT PHE",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 PHE",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N PRO",-0.25));
    _Charmmc.insert(map<string,float>::value_type("CD PRO", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CA PRO", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB PRO", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG PRO", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C PRO", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O PRO",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT PRO",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 PRO",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N SER",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA SER", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB SER", 0.25));
    _Charmmc.insert(map<string,float>::value_type("OG SER",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C SER", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O SER",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT SER",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 SER",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N THR",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA THR", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB THR", 0.25));
    _Charmmc.insert(map<string,float>::value_type("OG1 THR",-0.25));
    _Charmmc.insert(map<string,float>::value_type("CG2 THR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C THR", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O THR",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT THR",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 THR",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N TRP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA TRP", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB TRP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG TRP",-0.03));
    _Charmmc.insert(map<string,float>::value_type("CD2 TRP", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CE2 TRP",-0.04));
    _Charmmc.insert(map<string,float>::value_type("CE3 TRP",-0.03));
    _Charmmc.insert(map<string,float>::value_type("CD1 TRP", 0.06));
    _Charmmc.insert(map<string,float>::value_type("NE1 TRP",-0.06));
    _Charmmc.insert(map<string,float>::value_type("CZ2 TRP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CZ3 TRP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CH2 TRP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C TRP", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O TRP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT TRP",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 TRP",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N TYR",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA TYR", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB TYR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG TYR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD1 TYR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CE1 TYR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD2 TYR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CE2 TYR", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CZ TYR", 0.25));
    _Charmmc.insert(map<string,float>::value_type("OH TYR",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C TYR", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O TYR",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT TYR",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 TYR",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N VAL",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA VAL", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB VAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG1 VAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG2 VAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C VAL", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O VAL",-0.55));
    _Charmmc.insert(map<string,float>::value_type("OXT VAL",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OT2 VAL",-0.57));
    _Charmmc.insert(map<string,float>::value_type("CM ME", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C CH3", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C ACE", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O ACE",-0.55));
    _Charmmc.insert(map<string,float>::value_type("CH3 ACE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("N 5HP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("CA 5HP", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C 5HP", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O 5HP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("CB 5HP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG 5HP", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CD 5HP", 0.10));
    _Charmmc.insert(map<string,float>::value_type("OD 5HP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("N MSE",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA MSE", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C MSE", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O MSE",-0.55));
    _Charmmc.insert(map<string,float>::value_type("CB MSE", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG MSE", 0.06));
    _Charmmc.insert(map<string,float>::value_type("SE MSE",-0.12));
    _Charmmc.insert(map<string,float>::value_type("CE MSE", 0.06));
    _Charmmc.insert(map<string,float>::value_type("N SOC",-0.15));
    _Charmmc.insert(map<string,float>::value_type("CA SOC", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C SOC", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O SOC",-0.55));
    _Charmmc.insert(map<string,float>::value_type("CB SOC", 0.19));
    _Charmmc.insert(map<string,float>::value_type("SE SOC",-0.19));
    _Charmmc.insert(map<string,float>::value_type("OD1 SOC",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OD2 SOC",-0.57));
    _Charmmc.insert(map<string,float>::value_type("N SAH", 0.75));
    _Charmmc.insert(map<string,float>::value_type("CA SAH", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CB SAH", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CG SAH", 0.06));
    _Charmmc.insert(map<string,float>::value_type("SD SAH",-0.12));
    _Charmmc.insert(map<string,float>::value_type("C SAH", 0.14));
    _Charmmc.insert(map<string,float>::value_type("O SAH",-0.57));
    _Charmmc.insert(map<string,float>::value_type("OXT SAH",-0.57));
    _Charmmc.insert(map<string,float>::value_type("C5* SAH", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* SAH", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* SAH", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* SAH", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* SAH",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* SAH", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* SAH",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* SAH", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 SAH",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 SAH", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 SAH",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 SAH",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 SAH", 0.69));
    _Charmmc.insert(map<string,float>::value_type("N6 SAH",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N1 SAH",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 SAH", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N3 SAH",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 SAH", 0.39));
    _Charmmc.insert(map<string,float>::value_type("PG ATP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1G ATP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O2G ATP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3G ATP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3B ATP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PB ATP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1B ATP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2B ATP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O3A ATP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA ATP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1A ATP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2A ATP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O5* ATP",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5* ATP", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* ATP", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* ATP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* ATP", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* ATP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* ATP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* ATP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* ATP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 ATP",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 ATP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 ATP",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 ATP",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 ATP", 0.69));
    _Charmmc.insert(map<string,float>::value_type("N6 ATP",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N1 ATP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 ATP", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N3 ATP",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 ATP", 0.39));
    _Charmmc.insert(map<string,float>::value_type("PC GCP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1C GCP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O2C GCP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3C GCP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("C3B GCP", 0.50));
    _Charmmc.insert(map<string,float>::value_type("PB GCP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1B GCP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2B GCP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O3A GCP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA GCP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1A GCP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2A GCP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O5* GCP",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5* GCP", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* GCP", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* GCP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* GCP", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* GCP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* GCP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* GCP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* GCP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 GCP",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 GCP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 GCP",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 GCP",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 GCP", 0.69));
    _Charmmc.insert(map<string,float>::value_type("O6 GCP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N1 GCP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 GCP", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N2 GCP",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N3 GCP",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 GCP", 0.39));
    _Charmmc.insert(map<string,float>::value_type("PB GDP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1B GDP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O2B GDP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3B GDP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3A GDP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA GDP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1A GDP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2A GDP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O5* GDP",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5* GDP", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* GDP", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* GDP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* GDP", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* GDP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* GDP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* GDP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* GDP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 GDP",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 GDP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 GDP",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 GDP",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 GDP", 0.69));
    _Charmmc.insert(map<string,float>::value_type("O6 GDP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N1 GDP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 GDP", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N2 GDP",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N3 GDP",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 GDP", 0.39));
    _Charmmc.insert(map<string,float>::value_type("PG GNP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1G GNP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O2G GNP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3G GNP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("N3B GNP",-0.10));
    _Charmmc.insert(map<string,float>::value_type("PB GNP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1B GNP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2B GNP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O3A GNP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA GNP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1A GNP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2A GNP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O5* GNP",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5* GNP", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* GNP", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* GNP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* GNP", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* GNP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* GNP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* GNP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* GNP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 GNP",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 GNP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 GNP",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 GNP",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 GNP", 0.69));
    _Charmmc.insert(map<string,float>::value_type("O6 GNP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N1 GNP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 GNP", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N2 GNP",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N3 GNP",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 GNP", 0.39));
    _Charmmc.insert(map<string,float>::value_type("PG GSP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("S1G GSP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O2G GSP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3G GSP",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3B GSP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PB GSP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1B GSP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2B GSP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O3A GSP",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA GSP",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1A GSP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2A GSP",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O5* GSP",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5* GSP", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* GSP", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* GSP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* GSP", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* GSP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* GSP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* GSP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* GSP", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 GSP",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 GSP", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 GSP",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 GSP",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 GSP", 0.69));
    _Charmmc.insert(map<string,float>::value_type("O6 GSP",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N1 GSP",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 GSP", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N2 GSP",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N3 GSP",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 GSP", 0.39));
    _Charmmc.insert(map<string,float>::value_type("PG GTN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1G GTN",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O2G GTN",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O3G GTN",-0.62));
    _Charmmc.insert(map<string,float>::value_type("N3B GTN",-0.10));
    _Charmmc.insert(map<string,float>::value_type("PB GTN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1B GTN",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2B GTN",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O3A GTN",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA GTN",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1A GTN",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2A GTN",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O5* GTN",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5* GTN", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4* GTN", 0.11));
    _Charmmc.insert(map<string,float>::value_type("O4* GTN", 0.21));
    _Charmmc.insert(map<string,float>::value_type("C3* GTN", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O3* GTN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2* GTN", 0.33));
    _Charmmc.insert(map<string,float>::value_type("O2* GTN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1* GTN", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9 GTN",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8 GTN", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7 GTN",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5 GTN",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 GTN", 0.69));
    _Charmmc.insert(map<string,float>::value_type("O6 GTN",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N1 GTN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2 GTN", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N2 GTN",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N3 GTN",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4 GTN", 0.39));
    _Charmmc.insert(map<string,float>::value_type("C1T COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CGT COT",-0.03));
    _Charmmc.insert(map<string,float>::value_type("CD2T COT", 0.10));
    _Charmmc.insert(map<string,float>::value_type("CE2T COT",-0.04));
    _Charmmc.insert(map<string,float>::value_type("CE3T COT",-0.03));
    _Charmmc.insert(map<string,float>::value_type("CD1T COT", 0.06));
    _Charmmc.insert(map<string,float>::value_type("NE1T COT",-0.06));
    _Charmmc.insert(map<string,float>::value_type("CZ2T COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CZ3T COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("CH2T COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("S COT",-0.12));
    _Charmmc.insert(map<string,float>::value_type("C6T COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("C4T COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O5T COT",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N3T COT",-0.15));
    _Charmmc.insert(map<string,float>::value_type("C2T COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("OPB3 COT",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PB COT",-0.15));
    _Charmmc.insert(map<string,float>::value_type("OPB1 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("OPB2 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("OPA3 COT",-0.67));
    _Charmmc.insert(map<string,float>::value_type("O3*A COT",-0.67));
    _Charmmc.insert(map<string,float>::value_type("PA COT",-0.15));
    _Charmmc.insert(map<string,float>::value_type("OPA2 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("OPA1 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("O5*A COT",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C5*A COT", 0.01));
    _Charmmc.insert(map<string,float>::value_type("C4*A COT", 0.11));
    _Charmmc.insert(map<string,float>::value_type("C3*A COT", 0.26));
    _Charmmc.insert(map<string,float>::value_type("O2*A COT",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O4*A COT", 0.21));
    _Charmmc.insert(map<string,float>::value_type("PC COT",-0.15));
    _Charmmc.insert(map<string,float>::value_type("OPC1 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("OPC2 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("OPC3 COT",-0.62));
    _Charmmc.insert(map<string,float>::value_type("C2*A COT", 0.33));
    _Charmmc.insert(map<string,float>::value_type("C1*A COT", 0.21));
    _Charmmc.insert(map<string,float>::value_type("N9A COT",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C8A COT", 0.33));
    _Charmmc.insert(map<string,float>::value_type("N7A COT",-0.54));
    _Charmmc.insert(map<string,float>::value_type("C5A COT",-0.06));
    _Charmmc.insert(map<string,float>::value_type("C6A COT", 0.69));
    _Charmmc.insert(map<string,float>::value_type("N6A COT",-0.12));
    _Charmmc.insert(map<string,float>::value_type("N1A COT",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C2A COT", 0.88));
    _Charmmc.insert(map<string,float>::value_type("N3A COT",-0.71));
    _Charmmc.insert(map<string,float>::value_type("C4A COT", 0.39));
    _Charmmc.insert(map<string,float>::value_type("N3P COT",-0.15));
    _Charmmc.insert(map<string,float>::value_type("C4P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O5P COT",-0.55));
    _Charmmc.insert(map<string,float>::value_type("C6P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("C7P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("N8P COT",-0.15));
    _Charmmc.insert(map<string,float>::value_type("C9P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O10P COT",-0.55));
    _Charmmc.insert(map<string,float>::value_type("C11P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("O12P COT",-0.55));
    _Charmmc.insert(map<string,float>::value_type("C13P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("C16P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("C14P COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C15P COT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("C1P COT", 0.60));
    _Charmmc.insert(map<string,float>::value_type("AP FAD",-0.15));
    _Charmmc.insert(map<string,float>::value_type("AO1 FAD",-0.72));
    _Charmmc.insert(map<string,float>::value_type("AO2 FAD",-0.72));
    _Charmmc.insert(map<string,float>::value_type("AO5* FAD",-0.46));
    _Charmmc.insert(map<string,float>::value_type("AC5* FAD", 0.01));
    _Charmmc.insert(map<string,float>::value_type("AC4* FAD", 0.11));
    _Charmmc.insert(map<string,float>::value_type("AO4* FAD", 0.21));
    _Charmmc.insert(map<string,float>::value_type("AC3* FAD", 0.26));
    _Charmmc.insert(map<string,float>::value_type("AO3* FAD",-0.25));
    _Charmmc.insert(map<string,float>::value_type("AC2* FAD", 0.33));
    _Charmmc.insert(map<string,float>::value_type("AO2* FAD",-0.25));
    _Charmmc.insert(map<string,float>::value_type("AC1* FAD", 0.21));
    _Charmmc.insert(map<string,float>::value_type("AN9 FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("AC8 FAD", 0.33));
    _Charmmc.insert(map<string,float>::value_type("AN7 FAD",-0.54));
    _Charmmc.insert(map<string,float>::value_type("AC5 FAD",-0.06));
    _Charmmc.insert(map<string,float>::value_type("AC6 FAD", 0.69));
    _Charmmc.insert(map<string,float>::value_type("AN6 FAD",-0.12));
    _Charmmc.insert(map<string,float>::value_type("AN1 FAD",-0.25));
    _Charmmc.insert(map<string,float>::value_type("AC2 FAD", 0.88));
    _Charmmc.insert(map<string,float>::value_type("AN3 FAD",-0.71));
    _Charmmc.insert(map<string,float>::value_type("AC4 FAD", 0.39));
    _Charmmc.insert(map<string,float>::value_type("N1 FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C2 FAD", 0.55));
    _Charmmc.insert(map<string,float>::value_type("O2 FAD",-0.55));
    _Charmmc.insert(map<string,float>::value_type("N3 FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C4 FAD", 0.55));
    _Charmmc.insert(map<string,float>::value_type("O4 FAD",-0.55));
    _Charmmc.insert(map<string,float>::value_type("C4A FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("N5 FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C5A FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C6 FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C7 FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C7M FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C8 FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C8M FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C9 FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C9A FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("N10 FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C10 FAD",-0.11));
    _Charmmc.insert(map<string,float>::value_type("C1* FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2* FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O2* FAD",-0.37));
    _Charmmc.insert(map<string,float>::value_type("C3* FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O3* FAD",-0.37));
    _Charmmc.insert(map<string,float>::value_type("C4* FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O4* FAD",-0.37));
    _Charmmc.insert(map<string,float>::value_type("C5* FAD", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O5* FAD",-0.67));
    _Charmmc.insert(map<string,float>::value_type("P FAD",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O1P FAD",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O2P FAD",-0.72));
    _Charmmc.insert(map<string,float>::value_type("O3P FAD",-0.67));
    _Charmmc.insert(map<string,float>::value_type("C1 FUC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2 FUC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C3 FUC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C4 FUC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C5 FUC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 FUC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O1 FUC",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O2 FUC",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O3 FUC",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O4 FUC",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O5 FUC",-0.46));
    _Charmmc.insert(map<string,float>::value_type("C1 GAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2 GAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C3 GAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C4 GAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C5 GAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 GAL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O1 GAL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O2 GAL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O3 GAL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O4 GAL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O5 GAL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O6 GAL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1 MAN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2 MAN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C3 MAN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C4 MAN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C5 MAN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 MAN", 0.00));
    _Charmmc.insert(map<string,float>::value_type("O1 MAN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O2 MAN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O3 MAN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O4 MAN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O5 MAN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O6 MAN",-0.25));
    _Charmmc.insert(map<string,float>::value_type("C1 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C3 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C4 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C5 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C7 NAG", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C8 NAG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("N2 NAG",-0.15));
    _Charmmc.insert(map<string,float>::value_type("O3 NAG",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O4 NAG",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O5 NAG",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O6 NAG",-0.25));
    _Charmmc.insert(map<string,float>::value_type("O7 NAG",-0.55));
    _Charmmc.insert(map<string,float>::value_type("FE HEM", 0.24));
    _Charmmc.insert(map<string,float>::value_type("CHA HEM", 0.04));
    _Charmmc.insert(map<string,float>::value_type("CHB HEM", 0.04));
    _Charmmc.insert(map<string,float>::value_type("CHC HEM", 0.04));
    _Charmmc.insert(map<string,float>::value_type("CHD HEM", 0.04));
    _Charmmc.insert(map<string,float>::value_type("N A HEM",-0.18));
    _Charmmc.insert(map<string,float>::value_type("C1A HEM", 0.03));
    _Charmmc.insert(map<string,float>::value_type("C2A HEM",-0.02));
    _Charmmc.insert(map<string,float>::value_type("C3A HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("C4A HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("CMA HEM",-0.04));
    _Charmmc.insert(map<string,float>::value_type("CAA HEM", 0.04));
    _Charmmc.insert(map<string,float>::value_type("CBA HEM",-0.10));
    _Charmmc.insert(map<string,float>::value_type("CGA HEM", 0.30));
    _Charmmc.insert(map<string,float>::value_type("O1A HEM",-0.50));
    _Charmmc.insert(map<string,float>::value_type("O2A HEM",-0.50));
    _Charmmc.insert(map<string,float>::value_type("N B HEM",-0.18));
    _Charmmc.insert(map<string,float>::value_type("C1B HEM", 0.03));
    _Charmmc.insert(map<string,float>::value_type("C2B HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("C3B HEM",-0.05));
    _Charmmc.insert(map<string,float>::value_type("C4B HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("CMB HEM",-0.04));
    _Charmmc.insert(map<string,float>::value_type("CAB HEM", 0.03));
    _Charmmc.insert(map<string,float>::value_type("CBB HEM",-0.04));
    _Charmmc.insert(map<string,float>::value_type("N C HEM",-0.18));
    _Charmmc.insert(map<string,float>::value_type("C1C HEM", 0.03));
    _Charmmc.insert(map<string,float>::value_type("C2C HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("C3C HEM",-0.05));
    _Charmmc.insert(map<string,float>::value_type("C4C HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("CMC HEM",-0.04));
    _Charmmc.insert(map<string,float>::value_type("CAC HEM", 0.03));
    _Charmmc.insert(map<string,float>::value_type("CBC HEM",-0.10));
    _Charmmc.insert(map<string,float>::value_type("N D HEM",-0.18));
    _Charmmc.insert(map<string,float>::value_type("C1D HEM", 0.03));
    _Charmmc.insert(map<string,float>::value_type("C2D HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("C3D HEM",-0.02));
    _Charmmc.insert(map<string,float>::value_type("C4D HEM", 0.02));
    _Charmmc.insert(map<string,float>::value_type("CMD HEM",-0.04));
    _Charmmc.insert(map<string,float>::value_type("CAD HEM", 0.04));
    _Charmmc.insert(map<string,float>::value_type("CBD HEM",-0.10));
    _Charmmc.insert(map<string,float>::value_type("CGD HEM", 0.30));
    _Charmmc.insert(map<string,float>::value_type("O1D HEM",-0.50));
    _Charmmc.insert(map<string,float>::value_type("O2D HEM",-0.50));
    _Charmmc.insert(map<string,float>::value_type("C1 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C2 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C3 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C4 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C5 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C7 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C8 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C9 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C10 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C11 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C12 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C13 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C14 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C15 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C16 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C17 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C18 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C19 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C20 RTL", 0.00));
    _Charmmc.insert(map<string,float>::value_type("OR RTL",-0.25));
    _Charmmc.insert(map<string,float>::value_type("P A", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 A", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 A", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP3 A", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' A", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' A", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' A", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' A", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' A", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' A", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' A", 0.23));
    _Charmmc.insert(map<string,float>::value_type("O2' A", -0.23));
    _Charmmc.insert(map<string,float>::value_type("C1' A", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N9 A", -0.05));
    _Charmmc.insert(map<string,float>::value_type("C8 A", 0.46));
    _Charmmc.insert(map<string,float>::value_type("N7 A", -0.71));
    _Charmmc.insert(map<string,float>::value_type("C5 A", 0.28));
    _Charmmc.insert(map<string,float>::value_type("C6 A", 0.46));
    _Charmmc.insert(map<string,float>::value_type("N6 A", -0.01));
    _Charmmc.insert(map<string,float>::value_type("N1 A", -0.74));
    _Charmmc.insert(map<string,float>::value_type("C2 A", 0.63));
    _Charmmc.insert(map<string,float>::value_type("N3 A", -0.75));
    _Charmmc.insert(map<string,float>::value_type("C4 A", 0.43));
    _Charmmc.insert(map<string,float>::value_type("P G", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 G", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 G", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP3 G", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' G", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' G", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' G", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' G", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' G", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' G", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' G", 0.23));
    _Charmmc.insert(map<string,float>::value_type("O2' G", -0.23));
    _Charmmc.insert(map<string,float>::value_type("C1' G", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N9 G", -0.02));
    _Charmmc.insert(map<string,float>::value_type("C8 G", 0.41));
    _Charmmc.insert(map<string,float>::value_type("N7 G", -0.60));
    _Charmmc.insert(map<string,float>::value_type("C5 G", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 G", 0.54));
    _Charmmc.insert(map<string,float>::value_type("O6 G", -0.51));
    _Charmmc.insert(map<string,float>::value_type("N1 G", -0.08));
    _Charmmc.insert(map<string,float>::value_type("C2 G", 0.75));
    _Charmmc.insert(map<string,float>::value_type("N2 G", -0.01));
    _Charmmc.insert(map<string,float>::value_type("N3 G", -0.74));
    _Charmmc.insert(map<string,float>::value_type("C4 G", 0.26));
    _Charmmc.insert(map<string,float>::value_type("P C", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 C", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 C", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP3 C", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' C", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' C", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' C", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' C", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' C", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' C", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' C", 0.23));
    _Charmmc.insert(map<string,float>::value_type("O2' C", -0.23));
    _Charmmc.insert(map<string,float>::value_type("C1' C", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N1 C", -0.13));
    _Charmmc.insert(map<string,float>::value_type("C2 C", 0.52));
    _Charmmc.insert(map<string,float>::value_type("O2 C", -0.49));
    _Charmmc.insert(map<string,float>::value_type("N3 C", -0.66));
    _Charmmc.insert(map<string,float>::value_type("C4 C", 0.65));
    _Charmmc.insert(map<string,float>::value_type("N4 C", -0.05));
    _Charmmc.insert(map<string,float>::value_type("C5 C", -0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 C", 0.22));
    _Charmmc.insert(map<string,float>::value_type("P U", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 U", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 U", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP3 U", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' U", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' U", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' U", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' U", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' U", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' U", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' U", 0.23));
    _Charmmc.insert(map<string,float>::value_type("O2' U", -0.23));
    _Charmmc.insert(map<string,float>::value_type("C1' U", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N1 U", -0.34));
    _Charmmc.insert(map<string,float>::value_type("C2 U", 0.55));
    _Charmmc.insert(map<string,float>::value_type("O2 U", -0.45));
    _Charmmc.insert(map<string,float>::value_type("N3 U", -0.10));
    _Charmmc.insert(map<string,float>::value_type("C4 U", 0.53));
    _Charmmc.insert(map<string,float>::value_type("O4 U", -0.48));
    _Charmmc.insert(map<string,float>::value_type("C5 U", -0.05));
    _Charmmc.insert(map<string,float>::value_type("C6 U", 0.34));
    _Charmmc.insert(map<string,float>::value_type("P DA", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 DA", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 DA", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' DA", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' DA", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' DA", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' DA", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' DA", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' DA", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' DA", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C1' DA", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N9 DA", -0.08));
    _Charmmc.insert(map<string,float>::value_type("C8 DA", 0.46));
    _Charmmc.insert(map<string,float>::value_type("N7 DA", -0.71));
    _Charmmc.insert(map<string,float>::value_type("C5 DA", 0.28));
    _Charmmc.insert(map<string,float>::value_type("C6 DA", 0.46));
    _Charmmc.insert(map<string,float>::value_type("N6 DA", -0.01));
    _Charmmc.insert(map<string,float>::value_type("N1 DA", -0.74));
    _Charmmc.insert(map<string,float>::value_type("C2 DA", 0.63));
    _Charmmc.insert(map<string,float>::value_type("N3 DA", -0.75));
    _Charmmc.insert(map<string,float>::value_type("C4 DA", 0.43));
    _Charmmc.insert(map<string,float>::value_type("P DG", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 DG", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 DG", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' DG", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' DG", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' DG", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' DG", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' DG", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' DG", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' DG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C1' DG", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N9 DG", -0.02));
    _Charmmc.insert(map<string,float>::value_type("C8 DG", 0.41));
    _Charmmc.insert(map<string,float>::value_type("N7 DG", -0.60));
    _Charmmc.insert(map<string,float>::value_type("C5 DG", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C6 DG", 0.54));
    _Charmmc.insert(map<string,float>::value_type("O6 DG", -0.51));
    _Charmmc.insert(map<string,float>::value_type("N1 DG", -0.08));
    _Charmmc.insert(map<string,float>::value_type("C2 DG", 0.75));
    _Charmmc.insert(map<string,float>::value_type("N2 DG", -0.01));
    _Charmmc.insert(map<string,float>::value_type("N3 DG", -0.74));
    _Charmmc.insert(map<string,float>::value_type("C4 DG", 0.26));
    _Charmmc.insert(map<string,float>::value_type("P DC", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 DC", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 DC", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' DC", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' DC", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' DC", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' DC", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' DC", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' DC", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' DC", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C1' DC", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N1 DC", -0.13));
    _Charmmc.insert(map<string,float>::value_type("C2 DC", 0.52));
    _Charmmc.insert(map<string,float>::value_type("O2 DC", -0.49));
    _Charmmc.insert(map<string,float>::value_type("N3 DC", -0.66));
    _Charmmc.insert(map<string,float>::value_type("C4 DC", 0.65));
    _Charmmc.insert(map<string,float>::value_type("N4 DC", -0.05));
    _Charmmc.insert(map<string,float>::value_type("C5 DC", -0.06));
    _Charmmc.insert(map<string,float>::value_type("C6 DC", 0.22));
    _Charmmc.insert(map<string,float>::value_type("P DT", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 DT", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 DT", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' DT", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' DT", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' DT", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' DT", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' DT", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' DT", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' DT", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C1' DT", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N1 DT", -0.34));
    _Charmmc.insert(map<string,float>::value_type("C2 DT", 0.51));
    _Charmmc.insert(map<string,float>::value_type("O2 DT", -0.41));
    _Charmmc.insert(map<string,float>::value_type("N3 DT", -0.10));
    _Charmmc.insert(map<string,float>::value_type("C4 DT", 0.50));
    _Charmmc.insert(map<string,float>::value_type("O4 DT", -0.45));
    _Charmmc.insert(map<string,float>::value_type("C5 DT", -0.15));
    _Charmmc.insert(map<string,float>::value_type("C5M DT", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C6 DT", 0.34));
    _Charmmc.insert(map<string,float>::value_type("P DU", 1.50));
    _Charmmc.insert(map<string,float>::value_type("OP1 DU", -0.78));
    _Charmmc.insert(map<string,float>::value_type("OP2 DU", -0.78));
    _Charmmc.insert(map<string,float>::value_type("O5' DU", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C5' DU", 0.10));
    _Charmmc.insert(map<string,float>::value_type("C4' DU", 0.25));
    _Charmmc.insert(map<string,float>::value_type("O4' DU", -0.50));
    _Charmmc.insert(map<string,float>::value_type("C3' DU", 0.10));
    _Charmmc.insert(map<string,float>::value_type("O3' DU", -0.57));
    _Charmmc.insert(map<string,float>::value_type("C2' DU", 0.00));
    _Charmmc.insert(map<string,float>::value_type("C1' DU", 0.25));
    _Charmmc.insert(map<string,float>::value_type("N1 DU", -0.34));
    _Charmmc.insert(map<string,float>::value_type("C2 DU", 0.51));
    _Charmmc.insert(map<string,float>::value_type("O2 DU", -0.41));
    _Charmmc.insert(map<string,float>::value_type("N3 DU", -0.10));
    _Charmmc.insert(map<string,float>::value_type("C4 DU", 0.50));
    _Charmmc.insert(map<string,float>::value_type("O4 DU", -0.45));
    _Charmmc.insert(map<string,float>::value_type("C5 DU", -0.15));
    _Charmmc.insert(map<string,float>::value_type("C6 DU", 0.34));

    return;
}

//============================================================================//
void Parameter::ace_value()
//============================================================================//
{
    _ACE.insert(map<string,float>::value_type("N NTER", -0.495));
    _ACE.insert(map<string,float>::value_type("CA NTER", -0.553));
    _ACE.insert(map<string,float>::value_type("C CTER", -0.464));
    _ACE.insert(map<string,float>::value_type("N ALA", -0.495));
    _ACE.insert(map<string,float>::value_type("CA ALA", -0.553));
    _ACE.insert(map<string,float>::value_type("C ALA", -0.464));
    _ACE.insert(map<string,float>::value_type("O ALA", -0.079));
    _ACE.insert(map<string,float>::value_type("H ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("HN ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("1H ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("2H ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("3H ALA", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 ALA", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 ALA", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT ALA", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 ALA", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 ALA", 0.933));
    _ACE.insert(map<string,float>::value_type("CB ALA", -0.353));
    _ACE.insert(map<string,float>::value_type("N ARG", -0.495));
    _ACE.insert(map<string,float>::value_type("CA ARG", -0.553));
    _ACE.insert(map<string,float>::value_type("C ARG", -0.464));
    _ACE.insert(map<string,float>::value_type("O ARG", -0.079));
    _ACE.insert(map<string,float>::value_type("H ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HN ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("1H ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("2H ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("3H ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 ARG", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 ARG", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT ARG", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 ARG", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 ARG", 0.933));
    _ACE.insert(map<string,float>::value_type("CB ARG", -0.353));
    _ACE.insert(map<string,float>::value_type("CG ARG", -0.432));
    _ACE.insert(map<string,float>::value_type("CD ARG", 0.606));
    _ACE.insert(map<string,float>::value_type("CZ ARG", 0.726));
    _ACE.insert(map<string,float>::value_type("NE ARG", 0.606));
    _ACE.insert(map<string,float>::value_type("NH1 ARG", 0.726));
    _ACE.insert(map<string,float>::value_type("NH2 ARG", 0.726));
    _ACE.insert(map<string,float>::value_type("HE ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HH11 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HH12 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HH21 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("HH22 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("1HH1 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("2HH1 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("1HH2 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("2HH2 ARG", 0.0));
    _ACE.insert(map<string,float>::value_type("N ASN", -0.495));
    _ACE.insert(map<string,float>::value_type("CA ASN", -0.553));
    _ACE.insert(map<string,float>::value_type("C ASN", -0.464));
    _ACE.insert(map<string,float>::value_type("O ASN", -0.079));
    _ACE.insert(map<string,float>::value_type("H ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("HN ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("1H ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("2H ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("3H ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 ASN", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 ASN", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT ASN", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 ASN", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 ASN", 0.933));
    _ACE.insert(map<string,float>::value_type("CB ASN", -0.353));
    _ACE.insert(map<string,float>::value_type("CG ASN", 0.693));
    _ACE.insert(map<string,float>::value_type("OD1 ASN", 0.693));
    _ACE.insert(map<string,float>::value_type("ND2 ASN", 0.693));
    _ACE.insert(map<string,float>::value_type("HD21 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("HD22 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("1HD2 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("2HD2 ASN", 0.0));
    _ACE.insert(map<string,float>::value_type("N ASP", -0.495));
    _ACE.insert(map<string,float>::value_type("CA ASP", -0.553));
    _ACE.insert(map<string,float>::value_type("C ASP", -0.464));
    _ACE.insert(map<string,float>::value_type("O ASP", -0.079));
    _ACE.insert(map<string,float>::value_type("H ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("HN ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("1H ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("2H ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("3H ASP", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("CB ASP", -0.353));
    _ACE.insert(map<string,float>::value_type("CG ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("OD1 ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("OD2 ASP", 0.933));
    _ACE.insert(map<string,float>::value_type("N CYS", -0.495));
    _ACE.insert(map<string,float>::value_type("CA CYS", -0.553));
    _ACE.insert(map<string,float>::value_type("C CYS", -0.464));
    _ACE.insert(map<string,float>::value_type("O CYS", -0.079));
    _ACE.insert(map<string,float>::value_type("H CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HN CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("1H CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("2H CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("3H CYS", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 CYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 CYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT CYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 CYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 CYS", 0.933));
    _ACE.insert(map<string,float>::value_type("CB CYS", -0.353));
    _ACE.insert(map<string,float>::value_type("SG CYS", -1.827));
    _ACE.insert(map<string,float>::value_type("N GLN", -0.495));
    _ACE.insert(map<string,float>::value_type("CA GLN", -0.553));
    _ACE.insert(map<string,float>::value_type("C GLN", -0.464));
    _ACE.insert(map<string,float>::value_type("O GLN", -0.079));
    _ACE.insert(map<string,float>::value_type("H GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("HN GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("1H GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("2H GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("3H GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 GLN", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 GLN", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT GLN", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 GLN", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 GLN", 0.933));
    _ACE.insert(map<string,float>::value_type("CB GLN", -0.353));
    _ACE.insert(map<string,float>::value_type("CG GLN", -0.432));
    _ACE.insert(map<string,float>::value_type("CD GLN", 0.693));
    _ACE.insert(map<string,float>::value_type("OE1 GLN", 0.693));
    _ACE.insert(map<string,float>::value_type("NE2 GLN", 0.693));
    _ACE.insert(map<string,float>::value_type("HE21 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("HE22 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("1HE2 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("2HE2 GLN", 0.0));
    _ACE.insert(map<string,float>::value_type("N GLU", -0.495));
    _ACE.insert(map<string,float>::value_type("CA GLU", -0.553));
    _ACE.insert(map<string,float>::value_type("C GLU", -0.464));
    _ACE.insert(map<string,float>::value_type("O GLU", -0.079));
    _ACE.insert(map<string,float>::value_type("H GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("HN GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("1H GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("2H GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("3H GLU", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("CB GLU", -0.353));
    _ACE.insert(map<string,float>::value_type("CG GLU", -0.432));
    _ACE.insert(map<string,float>::value_type("CD GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("OE1 GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("OE2 GLU", 0.933));
    _ACE.insert(map<string,float>::value_type("N GLY", -0.495));
    _ACE.insert(map<string,float>::value_type("CA GLY", 0.008));
    _ACE.insert(map<string,float>::value_type("C GLY", -0.464));
    _ACE.insert(map<string,float>::value_type("O GLY", -0.079));
    _ACE.insert(map<string,float>::value_type("H GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("HN GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("1H GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("2H GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("3H GLY", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 GLY", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 GLY", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT GLY", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 GLY", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 GLY", 0.933));
    _ACE.insert(map<string,float>::value_type("N HIS", -0.495));
    _ACE.insert(map<string,float>::value_type("CA HIS", -0.553));
    _ACE.insert(map<string,float>::value_type("C HIS", -0.464));
    _ACE.insert(map<string,float>::value_type("O HIS", -0.079));
    _ACE.insert(map<string,float>::value_type("H HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("HN HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("1H HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("2H HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("3H HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 HIS", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 HIS", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT HIS", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 HIS", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 HIS", 0.933));
    _ACE.insert(map<string,float>::value_type("CB HIS", -0.353));
    _ACE.insert(map<string,float>::value_type("CG HIS", 0.061));
    _ACE.insert(map<string,float>::value_type("CD2 HIS", 0.061));
    _ACE.insert(map<string,float>::value_type("CE1 HIS", 0.061));
    _ACE.insert(map<string,float>::value_type("ND1 HIS", 0.061));
    _ACE.insert(map<string,float>::value_type("NE2 HIS", 0.061));
    _ACE.insert(map<string,float>::value_type("HD1 HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("HE2 HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("HD2 HIS", 0.0));
    _ACE.insert(map<string,float>::value_type("N ILE", -0.495));
    _ACE.insert(map<string,float>::value_type("CA ILE", -0.553));
    _ACE.insert(map<string,float>::value_type("C ILE", -0.464));
    _ACE.insert(map<string,float>::value_type("O ILE", -0.079));
    _ACE.insert(map<string,float>::value_type("H ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("HN ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("1H ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("2H ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("3H ILE", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 ILE", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 ILE", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT ILE", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 ILE", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 ILE", 0.933));
    _ACE.insert(map<string,float>::value_type("CB ILE", -0.353));
    _ACE.insert(map<string,float>::value_type("CG1 ILE", -0.432));
    _ACE.insert(map<string,float>::value_type("CG2 ILE", -0.987));
    _ACE.insert(map<string,float>::value_type("CD ILE", -0.987));
    _ACE.insert(map<string,float>::value_type("CD1 ILE", -0.987));
    _ACE.insert(map<string,float>::value_type("N LEU", -0.495));
    _ACE.insert(map<string,float>::value_type("CA LEU", -0.553));
    _ACE.insert(map<string,float>::value_type("C LEU", -0.464));
    _ACE.insert(map<string,float>::value_type("O LEU", -0.079));
    _ACE.insert(map<string,float>::value_type("H LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("HN LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("1H LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("2H LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("3H LEU", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 LEU", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 LEU", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT LEU", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 LEU", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 LEU", 0.933));
    _ACE.insert(map<string,float>::value_type("CB LEU", -0.353));
    _ACE.insert(map<string,float>::value_type("CG LEU", -0.432));
    _ACE.insert(map<string,float>::value_type("CD1 LEU", -0.987));
    _ACE.insert(map<string,float>::value_type("CD2 LEU", -0.987));
    _ACE.insert(map<string,float>::value_type("N LYS", -0.495));
    _ACE.insert(map<string,float>::value_type("CA LYS", -0.553));
    _ACE.insert(map<string,float>::value_type("C LYS", -0.464));
    _ACE.insert(map<string,float>::value_type("O LYS", -0.079));
    _ACE.insert(map<string,float>::value_type("H LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HN LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("1H LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("2H LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("3H LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 LYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 LYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT LYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 LYS", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 LYS", 0.933));
    _ACE.insert(map<string,float>::value_type("CB LYS", -0.353));
    _ACE.insert(map<string,float>::value_type("CG LYS", -0.432));
    _ACE.insert(map<string,float>::value_type("CD LYS", 1.046));
    _ACE.insert(map<string,float>::value_type("CE LYS", 1.334));
    _ACE.insert(map<string,float>::value_type("NZ LYS", 1.334));
    _ACE.insert(map<string,float>::value_type("HZ1 LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HZ2 LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("HZ3 LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("1HZ LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("2HZ LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("3HZ LYS", 0.0));
    _ACE.insert(map<string,float>::value_type("N MET", -0.495));
    _ACE.insert(map<string,float>::value_type("CA MET", -0.553));
    _ACE.insert(map<string,float>::value_type("C MET", -0.464));
    _ACE.insert(map<string,float>::value_type("O MET", -0.079));
    _ACE.insert(map<string,float>::value_type("H MET", 0.0));
    _ACE.insert(map<string,float>::value_type("HN MET", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 MET", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 MET", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 MET", 0.0));
    _ACE.insert(map<string,float>::value_type("1H MET", 0.0));
    _ACE.insert(map<string,float>::value_type("2H MET", 0.0));
    _ACE.insert(map<string,float>::value_type("3H MET", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 MET", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 MET", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT MET", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 MET", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 MET", 0.933));
    _ACE.insert(map<string,float>::value_type("CB MET", -0.353));
    _ACE.insert(map<string,float>::value_type("CG MET", -0.432));
    _ACE.insert(map<string,float>::value_type("CE MET", -0.987));
    _ACE.insert(map<string,float>::value_type("SD MET", -0.432));
    _ACE.insert(map<string,float>::value_type("N PHE", -0.495));
    _ACE.insert(map<string,float>::value_type("CA PHE", -0.553));
    _ACE.insert(map<string,float>::value_type("C PHE", -0.464));
    _ACE.insert(map<string,float>::value_type("O PHE", -0.079));
    _ACE.insert(map<string,float>::value_type("H PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("HN PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("1H PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("2H PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("3H PHE", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 PHE", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 PHE", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT PHE", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 PHE", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 PHE", 0.933));
    _ACE.insert(map<string,float>::value_type("CB PHE", -0.353));
    _ACE.insert(map<string,float>::value_type("CG PHE", -0.432));
    _ACE.insert(map<string,float>::value_type("CD1 PHE", -0.432));
    _ACE.insert(map<string,float>::value_type("CD2 PHE", -0.432));
    _ACE.insert(map<string,float>::value_type("CE1 PHE", -0.432));
    _ACE.insert(map<string,float>::value_type("CE2 PHE", -0.432));
    _ACE.insert(map<string,float>::value_type("CZ PHE", -0.432));
    _ACE.insert(map<string,float>::value_type("N PRO", -0.495));
    _ACE.insert(map<string,float>::value_type("CA PRO", -0.553));
    _ACE.insert(map<string,float>::value_type("C PRO", -0.464));
    _ACE.insert(map<string,float>::value_type("O PRO", -0.079));
    _ACE.insert(map<string,float>::value_type("H PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("HN PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("1H PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("2H PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("3H PRO", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 PRO", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 PRO", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT PRO", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 PRO", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 PRO", 0.933));
    _ACE.insert(map<string,float>::value_type("CB PRO", -0.353));
    _ACE.insert(map<string,float>::value_type("CG PRO", -0.353));
    _ACE.insert(map<string,float>::value_type("CD PRO", -0.353));
    _ACE.insert(map<string,float>::value_type("N SER", -0.495));
    _ACE.insert(map<string,float>::value_type("CA SER", -0.553));
    _ACE.insert(map<string,float>::value_type("C SER", -0.464));
    _ACE.insert(map<string,float>::value_type("O SER", -0.079));
    _ACE.insert(map<string,float>::value_type("H SER", 0.0));
    _ACE.insert(map<string,float>::value_type("HN SER", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 SER", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 SER", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 SER", 0.0));
    _ACE.insert(map<string,float>::value_type("1H SER", 0.0));
    _ACE.insert(map<string,float>::value_type("2H SER", 0.0));
    _ACE.insert(map<string,float>::value_type("3H SER", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 SER", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 SER", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT SER", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 SER", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 SER", 0.933));
    _ACE.insert(map<string,float>::value_type("CB SER", 0.232));
    _ACE.insert(map<string,float>::value_type("OG SER", 0.232));
    _ACE.insert(map<string,float>::value_type("HG SER", 0.0));
    _ACE.insert(map<string,float>::value_type("N SEP", -0.495));
    _ACE.insert(map<string,float>::value_type("CA SEP", -0.553));
    _ACE.insert(map<string,float>::value_type("C SEP", -0.464));
    _ACE.insert(map<string,float>::value_type("O SEP", -0.079));
    _ACE.insert(map<string,float>::value_type("H SEP", 0.0));
    _ACE.insert(map<string,float>::value_type("HN SEP", 0.0));
    _ACE.insert(map<string,float>::value_type("CB SEP", 0.232));
    _ACE.insert(map<string,float>::value_type("OG SEP", 0.232));
    _ACE.insert(map<string,float>::value_type("PO4 SEP", -0.495));
    _ACE.insert(map<string,float>::value_type("O1P SEP", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P SEP", 0.933));
    _ACE.insert(map<string,float>::value_type("O3P SEP", 0.933));
    _ACE.insert(map<string,float>::value_type("HT SEP", 0.0));
    _ACE.insert(map<string,float>::value_type("N THR", -0.495));
    _ACE.insert(map<string,float>::value_type("CA THR", -0.553));
    _ACE.insert(map<string,float>::value_type("C THR", -0.464));
    _ACE.insert(map<string,float>::value_type("O THR", -0.079));
    _ACE.insert(map<string,float>::value_type("H THR", 0.0));
    _ACE.insert(map<string,float>::value_type("HN THR", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 THR", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 THR", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 THR", 0.0));
    _ACE.insert(map<string,float>::value_type("1H THR", 0.0));
    _ACE.insert(map<string,float>::value_type("2H THR", 0.0));
    _ACE.insert(map<string,float>::value_type("3H THR", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 THR", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 THR", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT THR", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 THR", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 THR", 0.933));
    _ACE.insert(map<string,float>::value_type("CB THR", -0.353));
    _ACE.insert(map<string,float>::value_type("CG2 THR", -0.432));
    _ACE.insert(map<string,float>::value_type("OG1 THR", 0.232));
    _ACE.insert(map<string,float>::value_type("HG1 THR", 0.0));
    _ACE.insert(map<string,float>::value_type("N TRP", -0.495));
    _ACE.insert(map<string,float>::value_type("CA TRP", -0.553));
    _ACE.insert(map<string,float>::value_type("C TRP", -0.464));
    _ACE.insert(map<string,float>::value_type("O TRP", -0.079));
    _ACE.insert(map<string,float>::value_type("H TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("HN TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("1H TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("2H TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("3H TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 TRP", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 TRP", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT TRP", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 TRP", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 TRP", 0.933));
    _ACE.insert(map<string,float>::value_type("CB TRP", -0.353));
    _ACE.insert(map<string,float>::value_type("CG TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("CD2 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("CE2 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("CD1 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("CE3 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("NE1 TRP", 0.061));
    _ACE.insert(map<string,float>::value_type("CH2 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("CZ2 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("CZ3 TRP", -0.432));
    _ACE.insert(map<string,float>::value_type("HE1 TRP", 0.0));
    _ACE.insert(map<string,float>::value_type("N TYR", -0.495));
    _ACE.insert(map<string,float>::value_type("CA TYR", -0.553));
    _ACE.insert(map<string,float>::value_type("C TYR", -0.464));
    _ACE.insert(map<string,float>::value_type("O TYR", -0.079));
    _ACE.insert(map<string,float>::value_type("H TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("HN TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("1H TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("2H TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("3H TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 TYR", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 TYR", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT TYR", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 TYR", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 TYR", 0.933));
    _ACE.insert(map<string,float>::value_type("CB TYR", -0.353));
    _ACE.insert(map<string,float>::value_type("CG TYR", -0.432));
    _ACE.insert(map<string,float>::value_type("CZ TYR", -0.289));
    _ACE.insert(map<string,float>::value_type("CD1 TYR", -0.432));
    _ACE.insert(map<string,float>::value_type("CD2 TYR", -0.432));
    _ACE.insert(map<string,float>::value_type("CE1 TYR", -0.289));
    _ACE.insert(map<string,float>::value_type("CE2 TYR", -0.289));
    _ACE.insert(map<string,float>::value_type("OH TYR", 0.232));
    _ACE.insert(map<string,float>::value_type("HH TYR", 0.0));
    _ACE.insert(map<string,float>::value_type("N VAL", -0.495));
    _ACE.insert(map<string,float>::value_type("CA VAL", -0.553));
    _ACE.insert(map<string,float>::value_type("C VAL", -0.464));
    _ACE.insert(map<string,float>::value_type("O VAL", -0.079));
    _ACE.insert(map<string,float>::value_type("H VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("HN VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("1H VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("2H VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("3H VAL", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 VAL", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 VAL", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT VAL", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 VAL", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 VAL", 0.933));
    _ACE.insert(map<string,float>::value_type("CB VAL", -0.353));
    _ACE.insert(map<string,float>::value_type("CG1 VAL", -0.987));
    _ACE.insert(map<string,float>::value_type("CG2 VAL", -0.987));
    _ACE.insert(map<string,float>::value_type("P T", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P T", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 T", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P T", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 T", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* T", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' T", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* T", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' T", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* T", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' T", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* T", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' T", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* T", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' T", -0.353));
    _ACE.insert(map<string,float>::value_type("N1 T", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 T", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 T", 0.061));
    _ACE.insert(map<string,float>::value_type("O2 T", -0.079));
    _ACE.insert(map<string,float>::value_type("N3 T", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 T", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 T", 0.061));
    _ACE.insert(map<string,float>::value_type("O4 T", -0.079));
    _ACE.insert(map<string,float>::value_type("C5 T", 0.061));
    _ACE.insert(map<string,float>::value_type("C5M T", -0.987));
    _ACE.insert(map<string,float>::value_type("C7 T", -0.987));
    _ACE.insert(map<string,float>::value_type("C2* T", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' T", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* T", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' T", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* T", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' T", 0.232));
    _ACE.insert(map<string,float>::value_type("P A", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P A", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 A", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P A", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 A", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* A", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' A", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* A", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' A", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* A", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' A", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* A", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' A", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* A", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' A", -0.353));
    _ACE.insert(map<string,float>::value_type("N9 A", 0.061));
    _ACE.insert(map<string,float>::value_type("C4 A", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 A", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 A", 0.061));
    _ACE.insert(map<string,float>::value_type("N1 A", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 A", 0.061));
    _ACE.insert(map<string,float>::value_type("N6 A", 1.334));
    _ACE.insert(map<string,float>::value_type("1H6 A", 0.0));
    _ACE.insert(map<string,float>::value_type("H61 A", 0.0));
    _ACE.insert(map<string,float>::value_type("2H6 A", 0.0));
    _ACE.insert(map<string,float>::value_type("H62 A", 0.0));
    _ACE.insert(map<string,float>::value_type("C5 A", 0.061));
    _ACE.insert(map<string,float>::value_type("N7 A", 0.061));
    _ACE.insert(map<string,float>::value_type("C8 A", 0.061));
    _ACE.insert(map<string,float>::value_type("C2* A", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' A", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* A", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' A", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* A", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' A", 0.232));
    _ACE.insert(map<string,float>::value_type("P G", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P G", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 G", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P G", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 G", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* G", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' G", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* G", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' G", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* G", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' G", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* G", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' G", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* G", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' G", -0.353));
    _ACE.insert(map<string,float>::value_type("N9 G", 0.061));
    _ACE.insert(map<string,float>::value_type("C4 G", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 G", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 G", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 G", 1.334));
    _ACE.insert(map<string,float>::value_type("1H2 G", 0.0));
    _ACE.insert(map<string,float>::value_type("H21 G", 0.0));
    _ACE.insert(map<string,float>::value_type("2H2 G", 0.0));
    _ACE.insert(map<string,float>::value_type("H22 G", 0.0));
    _ACE.insert(map<string,float>::value_type("N1 G", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 G", 0.0));
    _ACE.insert(map<string,float>::value_type("C6 G", 0.061));
    _ACE.insert(map<string,float>::value_type("O6 G", -0.079));
    _ACE.insert(map<string,float>::value_type("C5 G", 0.061));
    _ACE.insert(map<string,float>::value_type("N7 G", 0.061));
    _ACE.insert(map<string,float>::value_type("C8 G", 0.061));
    _ACE.insert(map<string,float>::value_type("C2* G", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' G", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* G", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' G", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* G", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' G", 0.232));
    _ACE.insert(map<string,float>::value_type("P C", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P C", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 C", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P C", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 C", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* C", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' C", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* C", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' C", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* C", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' C", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* C", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' C", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* C", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' C", -0.353));
    _ACE.insert(map<string,float>::value_type("N1 C", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 C", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 C", 0.061));
    _ACE.insert(map<string,float>::value_type("O2 C", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 C", 0.061));
    _ACE.insert(map<string,float>::value_type("C4 C", 0.061));
    _ACE.insert(map<string,float>::value_type("N4 C", 1.334));
    _ACE.insert(map<string,float>::value_type("1H4 C", 0.0));
    _ACE.insert(map<string,float>::value_type("H41 C", 0.0));
    _ACE.insert(map<string,float>::value_type("2H4 C", 0.0));
    _ACE.insert(map<string,float>::value_type("H42 C", 0.0));
    _ACE.insert(map<string,float>::value_type("C5 C", 0.061));
    _ACE.insert(map<string,float>::value_type("C2* C", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' C", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* C", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' C", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* C", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' C", 0.232));
    _ACE.insert(map<string,float>::value_type("P DT", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P DT", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 DT", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P DT", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 DT", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* DT", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' DT", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' DT", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* DT", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' DT", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' DT", -0.353));
    _ACE.insert(map<string,float>::value_type("N1 DT", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 DT", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 DT", 0.061));
    _ACE.insert(map<string,float>::value_type("O2 DT", -0.079));
    _ACE.insert(map<string,float>::value_type("N3 DT", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 DT", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 DT", 0.061));
    _ACE.insert(map<string,float>::value_type("O4 DT", -0.079));
    _ACE.insert(map<string,float>::value_type("C5 DT", 0.061));
    _ACE.insert(map<string,float>::value_type("C5M DT", -0.987));
    _ACE.insert(map<string,float>::value_type("C7 DT", -0.987));
    _ACE.insert(map<string,float>::value_type("C2* DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* DT", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' DT", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* DT", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' DT", 0.232));
    _ACE.insert(map<string,float>::value_type("P DA", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P DA", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 DA", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P DA", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 DA", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* DA", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' DA", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' DA", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* DA", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' DA", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' DA", -0.353));
    _ACE.insert(map<string,float>::value_type("N9 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("C4 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("N1 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("N6 DA", 1.334));
    _ACE.insert(map<string,float>::value_type("1H6 DA", 0.0));
    _ACE.insert(map<string,float>::value_type("H61 DA", 0.0));
    _ACE.insert(map<string,float>::value_type("2H6 DA", 0.0));
    _ACE.insert(map<string,float>::value_type("H62 DA", 0.0));
    _ACE.insert(map<string,float>::value_type("C5 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("N7 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("C8 DA", 0.061));
    _ACE.insert(map<string,float>::value_type("C2* DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* DA", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' DA", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* DA", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' DA", 0.232));
    _ACE.insert(map<string,float>::value_type("P DG", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P DG", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 DG", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P DG", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 DG", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* DG", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' DG", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' DG", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* DG", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' DG", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' DG", -0.353));
    _ACE.insert(map<string,float>::value_type("N9 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("C4 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 DG", 1.334));
    _ACE.insert(map<string,float>::value_type("1H2 DG", 0.0));
    _ACE.insert(map<string,float>::value_type("H21 DG", 0.0));
    _ACE.insert(map<string,float>::value_type("2H2 DG", 0.0));
    _ACE.insert(map<string,float>::value_type("H22 DG", 0.0));
    _ACE.insert(map<string,float>::value_type("N1 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 DG", 0.0));
    _ACE.insert(map<string,float>::value_type("C6 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("O6 DG", -0.079));
    _ACE.insert(map<string,float>::value_type("C5 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("N7 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("C8 DG", 0.061));
    _ACE.insert(map<string,float>::value_type("C2* DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* DG", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' DG", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* DG", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' DG", 0.232));
    _ACE.insert(map<string,float>::value_type("P DC", 0.933));
    _ACE.insert(map<string,float>::value_type("O1P DC", 0.933));
    _ACE.insert(map<string,float>::value_type("OP1 DC", 0.933));
    _ACE.insert(map<string,float>::value_type("O2P DC", 0.933));
    _ACE.insert(map<string,float>::value_type("OP2 DC", 0.933));
    _ACE.insert(map<string,float>::value_type("O5* DC", -0.079));
    _ACE.insert(map<string,float>::value_type("O5' DC", -0.079));
    _ACE.insert(map<string,float>::value_type("C5* DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C5' DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C4* DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C4' DC", -0.353));
    _ACE.insert(map<string,float>::value_type("O4* DC", -0.079));
    _ACE.insert(map<string,float>::value_type("O4' DC", -0.079));
    _ACE.insert(map<string,float>::value_type("C1* DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C1' DC", -0.353));
    _ACE.insert(map<string,float>::value_type("N1 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("C2 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("O2 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("C4 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("N4 DC", 1.334));
    _ACE.insert(map<string,float>::value_type("1H4 DC", 0.0));
    _ACE.insert(map<string,float>::value_type("H41 DC", 0.0));
    _ACE.insert(map<string,float>::value_type("2H4 DC", 0.0));
    _ACE.insert(map<string,float>::value_type("H42 DC", 0.0));
    _ACE.insert(map<string,float>::value_type("C5 DC", 0.061));
    _ACE.insert(map<string,float>::value_type("C2* DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C2' DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C3* DC", -0.353));
    _ACE.insert(map<string,float>::value_type("C3' DC", -0.353));
    _ACE.insert(map<string,float>::value_type("O3* DC", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' DC", 0.232));
    _ACE.insert(map<string,float>::value_type("O 5HP", -0.079));
    _ACE.insert(map<string,float>::value_type("N 5HP", -0.495));
    _ACE.insert(map<string,float>::value_type("CB 5HP", -0.353));
    _ACE.insert(map<string,float>::value_type("OD 5HP", 0.232));
    _ACE.insert(map<string,float>::value_type("CD 5HP", -0.353));
    _ACE.insert(map<string,float>::value_type("C 5HP", -0.464));
    _ACE.insert(map<string,float>::value_type("CG 5HP", -0.353));
    _ACE.insert(map<string,float>::value_type("CA 5HP", -0.553));
    _ACE.insert(map<string,float>::value_type("O ACE", -0.079));
    _ACE.insert(map<string,float>::value_type("CH3 ACE", -0.987));
    _ACE.insert(map<string,float>::value_type("C ACE", -0.464));
    _ACE.insert(map<string,float>::value_type("N MSE", -0.495));
    _ACE.insert(map<string,float>::value_type("CA MSE", -0.553));
    _ACE.insert(map<string,float>::value_type("C MSE", -0.464));
    _ACE.insert(map<string,float>::value_type("O MSE", -0.079));
    _ACE.insert(map<string,float>::value_type("H MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("HN MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT1 MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT2 MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("HT3 MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("1H MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("2H MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("3H MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("OCT1 MSE", 0.933));
    _ACE.insert(map<string,float>::value_type("OCT2 MSE", 0.933));
    _ACE.insert(map<string,float>::value_type("OXT MSE", 0.933));
    _ACE.insert(map<string,float>::value_type("OT1 MSE", 0.933));
    _ACE.insert(map<string,float>::value_type("OT2 MSE", 0.933));
    _ACE.insert(map<string,float>::value_type("SE MSE", -0.432));
    _ACE.insert(map<string,float>::value_type("O MSE", -0.079));
    _ACE.insert(map<string,float>::value_type("N MSE", -0.495));
    _ACE.insert(map<string,float>::value_type("CB MSE", -0.353));
    _ACE.insert(map<string,float>::value_type("CE MSE", -0.987));
    _ACE.insert(map<string,float>::value_type("H MSE", 0.0));
    _ACE.insert(map<string,float>::value_type("C MSE", -0.464));
    _ACE.insert(map<string,float>::value_type("CG MSE", -0.432));
    _ACE.insert(map<string,float>::value_type("CA MSE", -0.553));
    _ACE.insert(map<string,float>::value_type("O5* GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("N9 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 GNP", 0.726));
    _ACE.insert(map<string,float>::value_type("PG GNP", -0.495));
    _ACE.insert(map<string,float>::value_type("C1* GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2G GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("C2* GNP", -0.289));
    _ACE.insert(map<string,float>::value_type("C2' GNP", -0.289));
    _ACE.insert(map<string,float>::value_type("PA GNP", -0.495));
    _ACE.insert(map<string,float>::value_type("C3* GNP", -0.289));
    _ACE.insert(map<string,float>::value_type("C3' GNP", -0.289));
    _ACE.insert(map<string,float>::value_type("C8 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 GNP", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("C5 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 GNP", 0.0));
    _ACE.insert(map<string,float>::value_type("H2 GNP", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("HNA GNP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2B GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("C4* GNP", -0.289));
    _ACE.insert(map<string,float>::value_type("C4' GNP", -0.289));
    _ACE.insert(map<string,float>::value_type("O3A GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("N3B GNP", 0.606));
    _ACE.insert(map<string,float>::value_type("O2* GNP", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' GNP", 0.232));
    _ACE.insert(map<string,float>::value_type("O4* GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("O4' GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("O3G GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("HNB GNP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2A GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("N1 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("O1B GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("PB GNP", -0.495));
    _ACE.insert(map<string,float>::value_type("N7 GNP", 0.061));
    _ACE.insert(map<string,float>::value_type("O3* GNP", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' GNP", 0.232));
    _ACE.insert(map<string,float>::value_type("C5* GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' GNP", -0.432));
    _ACE.insert(map<string,float>::value_type("O1G GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("O1A GNP", 0.933));
    _ACE.insert(map<string,float>::value_type("O6 GNP", -0.079));
    _ACE.insert(map<string,float>::value_type("AC2 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("O5* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("N3 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AO3* FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("AO3' FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("AC2* FAD", -0.289));
    _ACE.insert(map<string,float>::value_type("AC2' FAD", -0.289));
    _ACE.insert(map<string,float>::value_type("C1* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AC5* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AC5' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C2* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C2' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("N5 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("C3* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C3' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AC4 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("C8 FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("H3 FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 FAD", 0.693));
    _ACE.insert(map<string,float>::value_type("C6 FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AO5* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AO5' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AN1 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("C4A FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("H2 FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("C9A FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AH6B FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("AC4* FAD", -0.289));
    _ACE.insert(map<string,float>::value_type("AC4' FAD", -0.289));
    _ACE.insert(map<string,float>::value_type("C4 FAD", 0.693));
    _ACE.insert(map<string,float>::value_type("AC8 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("C10 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("O1P FAD", 0.933));
    _ACE.insert(map<string,float>::value_type("AO4* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AO4' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C4* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C4' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C7M FAD", -0.987));
    _ACE.insert(map<string,float>::value_type("O4 FAD", 0.693));
    _ACE.insert(map<string,float>::value_type("O2P FAD", 0.933));
    _ACE.insert(map<string,float>::value_type("C9 FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("O2* FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("N10 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AN6 FAD", 0.726));
    _ACE.insert(map<string,float>::value_type("AH6A FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("O4* FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("O4' FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("AO1 FAD", 0.933));
    _ACE.insert(map<string,float>::value_type("AC3* FAD", -0.289));
    _ACE.insert(map<string,float>::value_type("AC3' FAD", -0.289));
    _ACE.insert(map<string,float>::value_type("AO2 FAD", 0.933));
    _ACE.insert(map<string,float>::value_type("AC5 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AC1* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AC1' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("N1 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("C5A FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AP FAD", -0.495));
    _ACE.insert(map<string,float>::value_type("O2 FAD", 0.693));
    _ACE.insert(map<string,float>::value_type("P FAD", -0.495));
    _ACE.insert(map<string,float>::value_type("C7 FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("O3P FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("AC6 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AH2 FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("C8M FAD", -0.987));
    _ACE.insert(map<string,float>::value_type("AN9 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("H4 FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("AN7 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("AO2* FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("AO2' FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("O3* FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' FAD", 0.232));
    _ACE.insert(map<string,float>::value_type("C5* FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' FAD", -0.432));
    _ACE.insert(map<string,float>::value_type("HN FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("AH3 FAD", 0.0));
    _ACE.insert(map<string,float>::value_type("AN3 FAD", 0.061));
    _ACE.insert(map<string,float>::value_type("O5* GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("N9 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 GSP", 0.726));
    _ACE.insert(map<string,float>::value_type("PG GSP", -0.495));
    _ACE.insert(map<string,float>::value_type("O3B GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1* GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2G GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("C2* GSP", -0.289));
    _ACE.insert(map<string,float>::value_type("C2' GSP", -0.289));
    _ACE.insert(map<string,float>::value_type("PA GSP", -0.495));
    _ACE.insert(map<string,float>::value_type("C3* GSP", -0.289));
    _ACE.insert(map<string,float>::value_type("C3' GSP", -0.289));
    _ACE.insert(map<string,float>::value_type("C8 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 GSP", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("C5 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 GSP", 0.0));
    _ACE.insert(map<string,float>::value_type("H2 GSP", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("HNA GSP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2B GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("S1G GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("C4* GSP", -0.289));
    _ACE.insert(map<string,float>::value_type("C4' GSP", -0.289));
    _ACE.insert(map<string,float>::value_type("O3A GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2* GSP", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' GSP", 0.232));
    _ACE.insert(map<string,float>::value_type("O4* GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("O4' GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("O3G GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("HNB GSP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2A GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("N1 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("O1B GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("PB GSP", -0.495));
    _ACE.insert(map<string,float>::value_type("N7 GSP", 0.061));
    _ACE.insert(map<string,float>::value_type("O3* GSP", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' GSP", 0.232));
    _ACE.insert(map<string,float>::value_type("C5* GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' GSP", -0.432));
    _ACE.insert(map<string,float>::value_type("O1A GSP", 0.933));
    _ACE.insert(map<string,float>::value_type("O6 GSP", -0.079));
    _ACE.insert(map<string,float>::value_type("O5* GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("N9 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 GTN", 0.726));
    _ACE.insert(map<string,float>::value_type("PG GTN", -0.495));
    _ACE.insert(map<string,float>::value_type("C1* GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("O2G GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("C2* GTN", -0.289));
    _ACE.insert(map<string,float>::value_type("C2' GTN", -0.289));
    _ACE.insert(map<string,float>::value_type("PA GTN", -0.495));
    _ACE.insert(map<string,float>::value_type("C3* GTN", -0.289));
    _ACE.insert(map<string,float>::value_type("C3' GTN", -0.289));
    _ACE.insert(map<string,float>::value_type("C8 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 GTN", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("C5 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 GTN", 0.0));
    _ACE.insert(map<string,float>::value_type("H2 GTN", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("HNA GTN", 0.0));
    _ACE.insert(map<string,float>::value_type("O2B GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("C4* GTN", -0.289));
    _ACE.insert(map<string,float>::value_type("C4' GTN", -0.289));
    _ACE.insert(map<string,float>::value_type("O3A GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("N3B GTN", 0.606));
    _ACE.insert(map<string,float>::value_type("O2* GTN", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' GTN", 0.232));
    _ACE.insert(map<string,float>::value_type("O4* GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("O4' GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("O3G GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("HNB GTN", 0.0));
    _ACE.insert(map<string,float>::value_type("O2A GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("N1 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("O1B GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("PB GTN", -0.495));
    _ACE.insert(map<string,float>::value_type("N7 GTN", 0.061));
    _ACE.insert(map<string,float>::value_type("O3* GTN", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' GTN", 0.232));
    _ACE.insert(map<string,float>::value_type("C5* GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' GTN", -0.432));
    _ACE.insert(map<string,float>::value_type("O1G GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("O1A GTN", 0.933));
    _ACE.insert(map<string,float>::value_type("O6 GTN", -0.079));
    _ACE.insert(map<string,float>::value_type("O5* ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("N9 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("PG ATP", -0.495));
    _ACE.insert(map<string,float>::value_type("O3B ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1* ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2G ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("C2* ATP", -0.289));
    _ACE.insert(map<string,float>::value_type("C2' ATP", -0.289));
    _ACE.insert(map<string,float>::value_type("H6B ATP", 0.0));
    _ACE.insert(map<string,float>::value_type("PA ATP", -0.495));
    _ACE.insert(map<string,float>::value_type("C3* ATP", -0.289));
    _ACE.insert(map<string,float>::value_type("C3' ATP", -0.289));
    _ACE.insert(map<string,float>::value_type("C8 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 ATP", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("C5 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("H2 ATP", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("O2B ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("C4* ATP", -0.289));
    _ACE.insert(map<string,float>::value_type("C4' ATP", -0.289));
    _ACE.insert(map<string,float>::value_type("O3A ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2* ATP", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' ATP", 0.232));
    _ACE.insert(map<string,float>::value_type("O4* ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("O4' ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("O3G ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("O2A ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("O1B ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("N1 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("N6 ATP", 0.726));
    _ACE.insert(map<string,float>::value_type("PB ATP", -0.495));
    _ACE.insert(map<string,float>::value_type("N7 ATP", 0.061));
    _ACE.insert(map<string,float>::value_type("H6A ATP", 0.0));
    _ACE.insert(map<string,float>::value_type("C5* ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' ATP", -0.432));
    _ACE.insert(map<string,float>::value_type("O3* ATP", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' ATP", 0.232));
    _ACE.insert(map<string,float>::value_type("O1G ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("O1A ATP", 0.933));
    _ACE.insert(map<string,float>::value_type("C SOC", -0.464));
    _ACE.insert(map<string,float>::value_type("CA SOC", -0.553));
    _ACE.insert(map<string,float>::value_type("O SOC", -0.079));
    _ACE.insert(map<string,float>::value_type("N SOC", -0.495));
    _ACE.insert(map<string,float>::value_type("H SOC", 0.0));
    _ACE.insert(map<string,float>::value_type("HN SOC", 0.0));
    _ACE.insert(map<string,float>::value_type("CB SOC", -0.353));
    _ACE.insert(map<string,float>::value_type("SE SOC", -1.827));
    _ACE.insert(map<string,float>::value_type("OD1 SOC", 0.933));
    _ACE.insert(map<string,float>::value_type("OD2 SOC", 0.933));
    _ACE.insert(map<string,float>::value_type("CBC HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("C2D HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("NC HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("O2D HEM", 0.933));
    _ACE.insert(map<string,float>::value_type("CBA HEM", -0.353));
    _ACE.insert(map<string,float>::value_type("CHB HEM", -0.553));
    _ACE.insert(map<string,float>::value_type("CGA HEM", 0.933));
    _ACE.insert(map<string,float>::value_type("CHC HEM", -0.553));
    _ACE.insert(map<string,float>::value_type("C3D HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CBB HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("CAB HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("NA HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CHA HEM", -0.553));
    _ACE.insert(map<string,float>::value_type("CGD HEM", 0.933));
    _ACE.insert(map<string,float>::value_type("C2A HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C4A HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C4B HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C1D HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C3A HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("O1D HEM", 0.933));
    _ACE.insert(map<string,float>::value_type("C4C HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C2B HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CAA HEM", -0.353));
    _ACE.insert(map<string,float>::value_type("C1C HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CAD HEM", -0.353));
    _ACE.insert(map<string,float>::value_type("CAC HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("C2C HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("ND HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CHD HEM", -0.553));
    _ACE.insert(map<string,float>::value_type("CBD HEM", -0.353));
    _ACE.insert(map<string,float>::value_type("CMD HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("C1B HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("NB HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("FE HEM", 0.726));
    _ACE.insert(map<string,float>::value_type("O2A HEM", 0.933));
    _ACE.insert(map<string,float>::value_type("C3C HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C3B HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CMB HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("C4D HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("C1A HEM", 0.061));
    _ACE.insert(map<string,float>::value_type("CMA HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("O1A HEM", 0.933));
    _ACE.insert(map<string,float>::value_type("CMC HEM", -0.987));
    _ACE.insert(map<string,float>::value_type("C9 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C12 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C19 RTL", -0.987));
    _ACE.insert(map<string,float>::value_type("C11 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C18 RTL", -0.987));
    _ACE.insert(map<string,float>::value_type("C1 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C14 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C8 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C2 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C16 RTL", -0.987));
    _ACE.insert(map<string,float>::value_type("C6 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C5 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C7 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C4 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("HAA RTL", 0.0));
    _ACE.insert(map<string,float>::value_type("C10 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C13 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C20 RTL", -0.987));
    _ACE.insert(map<string,float>::value_type("C3 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("C17 RTL", -0.987));
    _ACE.insert(map<string,float>::value_type("C15 RTL", -0.432));
    _ACE.insert(map<string,float>::value_type("OR RTL", 0.232));
    _ACE.insert(map<string,float>::value_type("O5* GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("N9 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 GDP", 0.726));
    _ACE.insert(map<string,float>::value_type("O3B GDP", 0.933));
    _ACE.insert(map<string,float>::value_type("C1* GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("C2* GDP", -0.289));
    _ACE.insert(map<string,float>::value_type("C2' GDP", -0.289));
    _ACE.insert(map<string,float>::value_type("PA GDP", -0.495));
    _ACE.insert(map<string,float>::value_type("C3* GDP", -0.289));
    _ACE.insert(map<string,float>::value_type("C3' GDP", -0.289));
    _ACE.insert(map<string,float>::value_type("C8 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 GDP", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("C5 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 GDP", 0.0));
    _ACE.insert(map<string,float>::value_type("H2 GDP", 0.0));
    _ACE.insert(map<string,float>::value_type("C4 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("HNA GDP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2B GDP", 0.933));
    _ACE.insert(map<string,float>::value_type("C4* GDP", -0.289));
    _ACE.insert(map<string,float>::value_type("C4' GDP", -0.289));
    _ACE.insert(map<string,float>::value_type("O3A GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2* GDP", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' GDP", 0.232));
    _ACE.insert(map<string,float>::value_type("O4* GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("O4' GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2A GDP", 0.933));
    _ACE.insert(map<string,float>::value_type("HNB GDP", 0.0));
    _ACE.insert(map<string,float>::value_type("N1 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("O1B GDP", 0.933));
    _ACE.insert(map<string,float>::value_type("PB GDP", -0.495));
    _ACE.insert(map<string,float>::value_type("N7 GDP", 0.061));
    _ACE.insert(map<string,float>::value_type("C5* GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' GDP", -0.432));
    _ACE.insert(map<string,float>::value_type("O3* GDP", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' GDP", 0.232));
    _ACE.insert(map<string,float>::value_type("O1A GDP", 0.933));
    _ACE.insert(map<string,float>::value_type("O6 GDP", -0.079));
    _ACE.insert(map<string,float>::value_type("O5* GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("O5' GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("N9 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("N3 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("N2 GCP", 0.726));
    _ACE.insert(map<string,float>::value_type("C1* GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("C1' GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("C2* GCP", -0.289));
    _ACE.insert(map<string,float>::value_type("C2' GCP", -0.289));
    _ACE.insert(map<string,float>::value_type("PA GCP", -0.495));
    _ACE.insert(map<string,float>::value_type("C3* GCP", -0.289));
    _ACE.insert(map<string,float>::value_type("C3' GCP", -0.289));
    _ACE.insert(map<string,float>::value_type("C8 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("H3 GCP", 0.0));
    _ACE.insert(map<string,float>::value_type("C2 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("C6 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("C5 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("H1 GCP", 0.0));
    _ACE.insert(map<string,float>::value_type("O3C GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("H2 GCP", 0.0));
    _ACE.insert(map<string,float>::value_type("PC GCP", -0.495));
    _ACE.insert(map<string,float>::value_type("C4 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("HNA GCP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2B GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("C4* GCP", -0.289));
    _ACE.insert(map<string,float>::value_type("C4' GCP", -0.289));
    _ACE.insert(map<string,float>::value_type("O3A GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("O1C GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("O2* GCP", 0.232));
    _ACE.insert(map<string,float>::value_type("O2' GCP", 0.232));
    _ACE.insert(map<string,float>::value_type("O4* GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("O4' GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("O2C GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("HNB GCP", 0.0));
    _ACE.insert(map<string,float>::value_type("O2A GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("N1 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("O1B GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("C3B GCP", 0.726));
    _ACE.insert(map<string,float>::value_type("PB GCP", -0.495));
    _ACE.insert(map<string,float>::value_type("N7 GCP", 0.061));
    _ACE.insert(map<string,float>::value_type("O3* GCP", 0.232));
    _ACE.insert(map<string,float>::value_type("O3' GCP", 0.232));
    _ACE.insert(map<string,float>::value_type("C5* GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("C5' GCP", -0.432));
    _ACE.insert(map<string,float>::value_type("O1A GCP", 0.933));
    _ACE.insert(map<string,float>::value_type("O6 GCP", -0.079));

    return;
}

//============================================================================//
void Parameter::dangle_rot1()
//============================================================================//
{
    float Zangle[] = {
        0,0,0
    };

    _Num_rot_angles  = sizeof(Zangle)/sizeof(Zangle[0])/3;
    _Zangle   = new float[_Num_rot_angles*3];

    for( int i = 0 ; i < _Num_rot_angles*3 ; i++ ) {
        _Zangle[i] = Zangle[i];
    }


    return;
}

//============================================================================//
void Parameter::dangle_rot3()
//============================================================================//
{
    float Zangle[] = {
        0,0,0,
        1,1,1,
        2,2,2
    };

    _Num_rot_angles  = sizeof(Zangle)/sizeof(Zangle[0])/3;
    _Zangle   = new float[_Num_rot_angles*3];

    for( int i = 0 ; i < _Num_rot_angles*3 ; i++ ) {
        _Zangle[i] = Zangle[i];
    }

    return;
}


//============================================================================//
void Parameter::dangle_rot24()
//============================================================================//
{
    float Zangle[] = {
        2.617994,0.080730,-2.622036,
        -3.141593,1.450119,-2.739050,
        -2.356194,1.134497,-1.109631,
        0.261799,1.184361,0.625864,
        -0.785398,2.478122,2.609777,
        0.785398,1.617114,-1.572037,
        -2.617994,1.045978,0.343692,
        -2.356194,0.866955,-0.931351,
        -1.308997,1.766067,2.153843,
        -2.879793,1.195355,1.200319,
        2.879793,1.301460,-0.509251,
        0.785398,2.174483,2.653697,
        -2.617994,1.504028,0.636812,
        -0.523599,0.864993,2.278281,
        2.879793,1.325504,-1.671296,
        -1.047198,0.700570,1.559208,
        2.094395,1.741644,-2.695914,
        -0.785398,1.904882,-1.478089,
        -0.785398,2.587733,0.232235,
        -1.047198,1.242974,-2.977882,
        1.308997,1.870545,-2.976234,
        2.617994,1.135834,-1.431087,
        2.879793,1.658653,-2.409879,
        -2.617994,2.041297,-2.664597
    };

    _Num_rot_angles  = sizeof(Zangle)/sizeof(Zangle[0])/3;
    _Zangle   = new float[_Num_rot_angles*3];

    for( int i = 0 ; i < _Num_rot_angles*3 ; i++ ) {
        _Zangle[i] = Zangle[i];
    }

    return;
}



//============================================================================//
void Parameter::dangle_rot360()
//============================================================================//
{
    float Zangle[] = {
        2.617994,0.080730,-2.622036,
        -3.141593,0.992645,-1.723759,
        -3.141593,1.014500,0.906702,
        -3.141593,1.025536,1.576859,
        -3.141593,1.045978,0.343692,
        -3.141593,1.050534,3.042579,
        -3.141593,1.113094,-0.272241,
        -3.141593,1.128190,-2.288622,
        -3.141593,1.134497,-1.109631,
        -3.141593,1.134796,2.434290,
        -3.141593,1.135813,-2.658866,
        -3.141593,1.135834,-1.431087,
        -3.141593,1.168678,2.108595,
        -3.141593,1.184361,0.625864,
        -3.141593,1.194549,-0.801900,
        -3.141593,1.195355,1.200319,
        -3.141593,1.214854,-1.966664,
        -3.141593,1.218453,2.777284,
        -3.141593,1.242974,-2.977882,
        -3.141593,1.247695,1.809890,
        -3.141593,1.256302,0.062705,
        -3.141593,1.301460,-0.509251,
        -3.141593,1.310100,1.488307,
        -3.141593,1.325504,-1.671296,
        -3.141593,1.335669,0.897240,
        -3.141593,1.367556,0.358405,
        -3.141593,1.374601,-2.456031,
        -3.141593,1.391910,3.030322,
        -3.141593,1.413347,-1.339188,
        -3.141593,1.413430,2.550734,
        -3.141593,1.417623,-0.202254,
        -3.141593,1.433411,-1.035986,
        -3.141593,1.440580,2.260026,
        -3.141593,1.448123,-2.163292,
        -3.141593,1.450119,-2.739050,
        -3.141593,1.488617,1.159546,
        -3.141593,1.504028,0.636812,
        -3.141593,1.506530,1.988858,
        -3.141593,1.510910,-0.740472,
        -3.141593,1.545873,-1.877082,
        -3.141593,1.554776,0.100351,
        -3.141593,1.555207,1.708892,
        -3.141593,1.569311,-3.003978,
        -3.141593,1.574060,2.783565,
        -3.141593,1.599406,1.426070,
        -3.141593,1.606104,-0.451151,
        -3.141593,1.617114,-1.572037,
        -3.141593,1.658653,-2.409879,
        -3.141593,1.662363,0.915448,
        -3.141593,1.683268,0.374177,
        -3.141593,1.696674,-1.253233,
        -3.141593,1.699788,2.452890,
        -3.141593,1.721184,-0.153948,
        -3.141593,1.732874,-0.951685,
        -3.141593,1.736875,3.023612,
        -3.141593,1.741644,-2.695914,
        -3.141593,1.756530,-2.105156,
        -3.141593,1.766067,2.153843,
        -3.141593,1.809167,1.213774,
        -3.141593,1.812176,0.656819,
        -3.141593,1.824694,1.850876,
        -3.141593,1.836682,-0.676006,
        -3.141593,1.846342,-1.796954,
        -3.141593,1.869911,2.701025,
        -3.141593,1.870545,-2.976234,
        -3.141593,1.872854,1.543456,
        -3.141593,1.875281,0.107086,
        -3.141593,1.904882,-1.478089,
        -3.141593,1.909857,-0.381669,
        -3.141593,1.948085,-2.363031,
        -3.141593,1.977582,0.969896,
        -3.141593,1.982574,-1.115518,
        -3.141593,1.998705,2.370472,
        -3.141593,2.001437,0.381215,
        -3.141593,2.033014,2.977192,
        -3.141593,2.041297,-2.664597,
        -3.141593,2.063651,-2.044600,
        -3.141593,2.073035,2.027405,
        -3.141593,2.089241,-0.114193,
        -3.141593,2.104282,-0.813972,
        -3.141593,2.108018,1.311430,
        -3.141593,2.115563,0.685730,
        -3.141593,2.138223,1.678323,
        -3.141593,2.153910,-1.714959,
        -3.141593,2.164094,-2.991358,
        -3.141593,2.174483,2.653697,
        -3.141593,2.180315,-1.357379,
        -3.141593,2.187132,-0.479875,
        -3.141593,2.247462,0.191623,
        -3.141593,2.251330,-2.330739,
        -3.141593,2.295480,1.008277,
        -3.141593,2.299159,2.264663,
        -3.141593,2.343981,-1.040456,
        -3.141593,2.349581,-2.702517,
        -3.141593,2.356369,3.005869,
        -3.141593,2.379497,0.561089,
        -3.141593,2.393226,1.857173,
        -3.141593,2.396274,-1.959260,
        -3.141593,2.404280,-0.174044,
        -3.141593,2.407836,1.410646,
        -3.141593,2.455966,-0.641220,
        -3.141593,2.456751,-1.503343,
        -3.141593,2.478122,2.609777,
        -3.141593,2.579248,-2.364801,
        -3.141593,2.587733,0.232235,
        -3.141593,2.588535,-2.976617,
        -3.141593,2.592629,0.909713,
        -3.141593,2.637879,2.121598,
        -3.141593,2.664082,-1.071402,
        -3.141593,2.704512,1.488683,
        -3.141593,2.719179,-1.780644,
        -3.141593,2.724429,-0.348782,
        -3.141593,2.760637,2.781135,
        -3.141593,2.863997,0.549740,
        -3.141593,2.881233,-2.582574,
        -3.141593,2.979456,-1.009304,
        -3.141593,2.984978,1.954446,
        -2.879793,0.080730,-2.622036,
        -2.879793,0.203748,0.590255,
        -2.879793,0.290380,-0.796960,
        -2.879793,0.297065,2.160177,
        -2.879793,0.370812,-2.547742,
        -2.879793,0.379747,-1.713059,
        -2.879793,0.418037,1.360082,
        -2.879793,0.419640,2.974122,
        -2.879793,0.458961,-0.107937,
        -2.879793,0.496451,0.593859,
        -2.879793,0.571433,-1.188419,
        -2.879793,0.591903,2.010408,
        -2.879793,0.615174,-0.666022,
        -2.879793,0.618092,2.539118,
        -2.879793,0.619654,-2.216909,
        -2.879793,0.640596,-2.797608,
        -2.879793,0.674054,1.013609,
        -2.879793,0.685216,-1.712745,
        -2.879793,0.700570,1.559208,
        -2.879793,0.709924,0.218797,
        -2.879793,0.733948,3.021411,
        -2.879793,0.762244,-0.273349,
        -2.879793,0.831641,0.616218,
        -2.879793,0.850607,-1.335383,
        -2.879793,0.864993,2.278281,
        -2.879793,0.866395,-2.492918,
        -2.879793,0.866955,-0.931351,
        -2.879793,0.899350,-2.069907,
        -2.879793,0.906588,1.252948,
        -2.879793,0.910508,1.903534,
        -2.879793,0.914445,2.687814,
        -2.879793,0.929350,-2.910676,
        -2.879793,0.960688,-0.577532,
        -2.879793,0.961715,0.014085,
        -2.879793,0.992645,-1.723759,
        -2.879793,1.014500,0.906702,
        -2.879793,1.025536,1.576859,
        -2.879793,1.045978,0.343692,
        -2.879793,1.050534,3.042579,
        -2.879793,1.113094,-0.272241,
        -2.879793,1.128190,-2.288622,
        -2.879793,1.134497,-1.109631,
        -2.879793,1.134796,2.434290,
        -2.879793,1.135813,-2.658866,
        -2.879793,1.135834,-1.431087,
        -2.879793,1.168678,2.108595,
        -2.879793,1.184361,0.625864,
        -2.879793,1.194549,-0.801900,
        -2.879793,1.195355,1.200319,
        -2.879793,1.214854,-1.966664,
        -2.879793,1.218453,2.777284,
        -2.879793,1.242974,-2.977882,
        -2.879793,1.247695,1.809890,
        -2.879793,1.256302,0.062705,
        -2.879793,1.301460,-0.509251,
        -2.879793,1.310100,1.488307,
        -2.879793,1.325504,-1.671296,
        -2.879793,1.335669,0.897240,
        -2.879793,1.367556,0.358405,
        -2.879793,1.374601,-2.456031,
        -2.879793,1.391910,3.030322,
        -2.879793,1.413347,-1.339188,
        -2.879793,1.413430,2.550734,
        -2.879793,1.417623,-0.202254,
        -2.879793,1.433411,-1.035986,
        -2.879793,1.440580,2.260026,
        -2.879793,1.448123,-2.163292,
        -2.879793,1.450119,-2.739050,
        -2.879793,1.488617,1.159546,
        -2.879793,1.504028,0.636812,
        -2.879793,1.506530,1.988858,
        -2.879793,1.510910,-0.740472,
        -2.879793,1.545873,-1.877082,
        -2.879793,1.554776,0.100351,
        -2.879793,1.555207,1.708892,
        -2.879793,1.569311,-3.003978,
        -2.879793,1.574060,2.783565,
        -2.879793,1.599406,1.426070,
        -2.879793,1.606104,-0.451151,
        -2.879793,1.617114,-1.572037,
        -2.879793,1.658653,-2.409879,
        -2.879793,1.662363,0.915448,
        -2.879793,1.683268,0.374177,
        -2.879793,1.696674,-1.253233,
        -2.879793,1.699788,2.452890,
        -2.879793,1.721184,-0.153948,
        -2.879793,1.732874,-0.951685,
        -2.879793,1.736875,3.023612,
        -2.879793,1.741644,-2.695914,
        -2.879793,1.756530,-2.105156,
        -2.879793,1.766067,2.153843,
        -2.879793,1.809167,1.213774,
        -2.879793,1.812176,0.656819,
        -2.879793,1.824694,1.850876,
        -2.879793,1.836682,-0.676006,
        -2.879793,1.846342,-1.796954,
        -2.879793,1.869911,2.701025,
        -2.879793,1.870545,-2.976234,
        -2.879793,1.872854,1.543456,
        -2.879793,1.875281,0.107086,
        -2.879793,1.904882,-1.478089,
        -2.879793,1.909857,-0.381669,
        -2.879793,1.948085,-2.363031,
        -2.879793,1.977582,0.969896,
        -2.879793,1.982574,-1.115518,
        -2.879793,1.998705,2.370472,
        -2.879793,2.001437,0.381215,
        -2.879793,2.033014,2.977192,
        -2.879793,2.041297,-2.664597,
        -2.879793,2.063651,-2.044600,
        -2.879793,2.073035,2.027405,
        -2.879793,2.089241,-0.114193,
        -2.879793,2.104282,-0.813972,
        -2.879793,2.108018,1.311430,
        -2.879793,2.115563,0.685730,
        -2.879793,2.138223,1.678323,
        -2.879793,2.153910,-1.714959,
        -2.879793,2.164094,-2.991358,
        -2.879793,2.174483,2.653697,
        -2.879793,2.180315,-1.357379,
        -2.879793,2.187132,-0.479875,
        -2.879793,2.247462,0.191623,
        -2.879793,2.251330,-2.330739,
        -2.879793,2.295480,1.008277,
        -2.879793,2.299159,2.264663,
        -2.879793,2.343981,-1.040456,
        -2.879793,2.349581,-2.702517,
        -2.879793,2.356369,3.005869,
        -2.879793,2.379497,0.561089,
        -2.879793,2.393226,1.857173,
        -2.879793,2.396274,-1.959260,
        -2.879793,2.404280,-0.174044,
        -2.879793,2.407836,1.410646,
        -2.879793,2.455966,-0.641220,
        -2.879793,2.456751,-1.503343,
        -2.879793,2.478122,2.609777,
        -2.879793,2.579248,-2.364801,
        -2.879793,2.587733,0.232235,
        -2.879793,2.588535,-2.976617,
        -2.879793,2.592629,0.909713,
        -2.879793,2.637879,2.121598,
        -2.879793,2.664082,-1.071402,
        -2.879793,2.704512,1.488683,
        -2.879793,2.719179,-1.780644,
        -2.879793,2.724429,-0.348782,
        -2.879793,2.760637,2.781135,
        -2.879793,2.863997,0.549740,
        -2.879793,2.881233,-2.582574,
        -2.879793,2.979456,-1.009304,
        -2.879793,2.984978,1.954446,
        -2.617994,0.080730,-2.622036,
        -2.617994,0.203748,0.590255,
        -2.617994,0.290380,-0.796960,
        -2.617994,0.297065,2.160177,
        -2.617994,0.370812,-2.547742,
        -2.617994,0.379747,-1.713059,
        -2.617994,0.418037,1.360082,
        -2.617994,0.419640,2.974122,
        -2.617994,0.458961,-0.107937,
        -2.617994,0.496451,0.593859,
        -2.617994,0.571433,-1.188419,
        -2.617994,0.591903,2.010408,
        -2.617994,0.615174,-0.666022,
        -2.617994,0.618092,2.539118,
        -2.617994,0.619654,-2.216909,
        -2.617994,0.640596,-2.797608,
        -2.617994,0.674054,1.013609,
        -2.617994,0.685216,-1.712745,
        -2.617994,0.700570,1.559208,
        -2.617994,0.709924,0.218797,
        -2.617994,0.733948,3.021411,
        -2.617994,0.762244,-0.273349,
        -2.617994,0.831641,0.616218,
        -2.617994,0.850607,-1.335383,
        -2.617994,0.864993,2.278281,
        -2.617994,0.866395,-2.492918,
        -2.617994,0.866955,-0.931351,
        -2.617994,0.899350,-2.069907,
        -2.617994,0.906588,1.252948,
        -2.617994,0.910508,1.903534,
        -2.617994,0.914445,2.687814,
        -2.617994,0.929350,-2.910676,
        -2.617994,0.960688,-0.577532,
        -2.617994,0.961715,0.014085,
        -2.617994,0.992645,-1.723759,
        -2.617994,1.014500,0.906702,
        -2.617994,1.025536,1.576859,
        -2.617994,1.045978,0.343692,
        -2.617994,1.050534,3.042579,
        -2.617994,1.113094,-0.272241,
        -2.617994,1.128190,-2.288622,
        -2.617994,1.134497,-1.109631,
        -2.617994,1.134796,2.434290,
        -2.617994,1.135813,-2.658866,
        -2.617994,1.135834,-1.431087,
        -2.617994,1.168678,2.108595,
        -2.617994,1.184361,0.625864,
        -2.617994,1.194549,-0.801900,
        -2.617994,1.195355,1.200319,
        -2.617994,1.214854,-1.966664,
        -2.617994,1.218453,2.777284,
        -2.617994,1.242974,-2.977882,
        -2.617994,1.247695,1.809890,
        -2.617994,1.256302,0.062705,
        -2.617994,1.301460,-0.509251,
        -2.617994,1.310100,1.488307,
        -2.617994,1.325504,-1.671296,
        -2.617994,1.335669,0.897240,
        -2.617994,1.367556,0.358405,
        -2.617994,1.374601,-2.456031,
        -2.617994,1.391910,3.030322,
        -2.617994,1.413347,-1.339188,
        -2.617994,1.413430,2.550734,
        -2.617994,1.417623,-0.202254,
        -2.617994,1.433411,-1.035986,
        -2.617994,1.440580,2.260026,
        -2.617994,1.448123,-2.163292,
        -2.617994,1.450119,-2.739050,
        -2.617994,1.488617,1.159546,
        -2.617994,1.504028,0.636812,
        -2.617994,1.506530,1.988858,
        -2.617994,1.510910,-0.740472,
        -2.617994,1.545873,-1.877082,
        -2.617994,1.554776,0.100351,
        -2.617994,1.555207,1.708892,
        -2.617994,1.569311,-3.003978,
        -2.617994,1.574060,2.783565,
        -2.617994,1.599406,1.426070,
        -2.617994,1.606104,-0.451151,
        -2.617994,1.617114,-1.572037,
        -2.617994,1.658653,-2.409879,
        -2.617994,1.662363,0.915448,
        -2.617994,1.683268,0.374177,
        -2.617994,1.696674,-1.253233,
        -2.617994,1.699788,2.452890,
        -2.617994,1.721184,-0.153948,
        -2.617994,1.732874,-0.951685,
        -2.617994,1.736875,3.023612,
        -2.617994,1.741644,-2.695914,
        -2.617994,1.756530,-2.105156,
        -2.617994,1.766067,2.153843,
        -2.617994,1.809167,1.213774,
        -2.617994,1.812176,0.656819
    };

    _Num_rot_angles  = sizeof(Zangle)/sizeof(Zangle[0])/3;
    _Zangle   = new float[_Num_rot_angles*3];

    for( int i = 0 ; i < _Num_rot_angles*3 ; i++ ) {
        _Zangle[i] = Zangle[i];
    }

    return;
}


//============================================================================//
void Parameter::dangle_rot3600()
//============================================================================//
{
    float Zangle[] = {
        -3.141593,0.080730,-2.622036,
        -3.141593,0.203748,0.590255,
        -3.141593,0.290380,-0.796960,
        -3.141593,0.297065,2.160177,
        -3.141593,0.370812,-2.547742,
        -3.141593,0.379747,-1.713059,
        -3.141593,0.418037,1.360082,
        -3.141593,0.419640,2.974122,
        -3.141593,0.458961,-0.107937,
        -3.141593,0.496451,0.593859,
        -3.141593,0.571433,-1.188419,
        -3.141593,0.591903,2.010408,
        -3.141593,0.615174,-0.666022,
        -3.141593,0.618092,2.539118,
        -3.141593,0.619654,-2.216909,
        -3.141593,0.640596,-2.797608,
        -3.141593,0.674054,1.013609,
        -3.141593,0.685216,-1.712745,
        -3.141593,0.700570,1.559208,
        -3.141593,0.709924,0.218797,
        -3.141593,0.733948,3.021411,
        -3.141593,0.762244,-0.273349,
        -3.141593,0.831641,0.616218,
        -3.141593,0.850607,-1.335383,
        -3.141593,0.864993,2.278281,
        -3.141593,0.866395,-2.492918,
        -3.141593,0.866955,-0.931351,
        -3.141593,0.899350,-2.069907,
        -3.141593,0.906588,1.252948,
        -3.141593,0.910508,1.903534,
        -3.141593,0.914445,2.687814,
        -3.141593,0.929350,-2.910676,
        -3.141593,0.960688,-0.577532,
        -3.141593,0.961715,0.014085,
        -3.141593,0.992645,-1.723759,
        -3.141593,1.014500,0.906702,
        -3.141593,1.025536,1.576859,
        -3.141593,1.045978,0.343692,
        -3.141593,1.050534,3.042579,
        -3.141593,1.113094,-0.272241,
        -3.141593,1.128190,-2.288622,
        -3.141593,1.134497,-1.109631,
        -3.141593,1.134796,2.434290,
        -3.141593,1.135813,-2.658866,
        -3.141593,1.135834,-1.431087,
        -3.141593,1.168678,2.108595,
        -3.141593,1.184361,0.625864,
        -3.141593,1.194549,-0.801900,
        -3.141593,1.195355,1.200319,
        -3.141593,1.214854,-1.966664,
        -3.141593,1.218453,2.777284,
        -3.141593,1.242974,-2.977882,
        -3.141593,1.247695,1.809890,
        -3.141593,1.256302,0.062705,
        -3.141593,1.301460,-0.509251,
        -3.141593,1.310100,1.488307,
        -3.141593,1.325504,-1.671296,
        -3.141593,1.335669,0.897240,
        -3.141593,1.367556,0.358405,
        -3.141593,1.374601,-2.456031,
        -3.141593,1.391910,3.030322,
        -3.141593,1.413347,-1.339188,
        -3.141593,1.413430,2.550734,
        -3.141593,1.417623,-0.202254,
        -3.141593,1.433411,-1.035986,
        -3.141593,1.440580,2.260026,
        -3.141593,1.448123,-2.163292,
        -3.141593,1.450119,-2.739050,
        -3.141593,1.488617,1.159546,
        -3.141593,1.504028,0.636812,
        -3.141593,1.506530,1.988858,
        -3.141593,1.510910,-0.740472,
        -3.141593,1.545873,-1.877082,
        -3.141593,1.554776,0.100351,
        -3.141593,1.555207,1.708892,
        -3.141593,1.569311,-3.003978,
        -3.141593,1.574060,2.783565,
        -3.141593,1.599406,1.426070,
        -3.141593,1.606104,-0.451151,
        -3.141593,1.617114,-1.572037,
        -3.141593,1.658653,-2.409879,
        -3.141593,1.662363,0.915448,
        -3.141593,1.683268,0.374177,
        -3.141593,1.696674,-1.253233,
        -3.141593,1.699788,2.452890,
        -3.141593,1.721184,-0.153948,
        -3.141593,1.732874,-0.951685,
        -3.141593,1.736875,3.023612,
        -3.141593,1.741644,-2.695914,
        -3.141593,1.756530,-2.105156,
        -3.141593,1.766067,2.153843,
        -3.141593,1.809167,1.213774,
        -3.141593,1.812176,0.656819,
        -3.141593,1.824694,1.850876,
        -3.141593,1.836682,-0.676006,
        -3.141593,1.846342,-1.796954,
        -3.141593,1.869911,2.701025,
        -3.141593,1.870545,-2.976234,
        -3.141593,1.872854,1.543456,
        -3.141593,1.875281,0.107086,
        -3.141593,1.904882,-1.478089,
        -3.141593,1.909857,-0.381669,
        -3.141593,1.948085,-2.363031,
        -3.141593,1.977582,0.969896,
        -3.141593,1.982574,-1.115518,
        -3.141593,1.998705,2.370472,
        -3.141593,2.001437,0.381215,
        -3.141593,2.033014,2.977192,
        -3.141593,2.041297,-2.664597,
        -3.141593,2.063651,-2.044600,
        -3.141593,2.073035,2.027405,
        -3.141593,2.089241,-0.114193,
        -3.141593,2.104282,-0.813972,
        -3.141593,2.108018,1.311430,
        -3.141593,2.115563,0.685730,
        -3.141593,2.138223,1.678323,
        -3.141593,2.153910,-1.714959,
        -3.141593,2.164094,-2.991358,
        -3.141593,2.174483,2.653697,
        -3.141593,2.180315,-1.357379,
        -3.141593,2.187132,-0.479875,
        -3.141593,2.247462,0.191623,
        -3.141593,2.251330,-2.330739,
        -3.141593,2.295480,1.008277,
        -3.141593,2.299159,2.264663,
        -3.141593,2.343981,-1.040456,
        -3.141593,2.349581,-2.702517,
        -3.141593,2.356369,3.005869,
        -3.141593,2.379497,0.561089,
        -3.141593,2.393226,1.857173,
        -3.141593,2.396274,-1.959260,
        -3.141593,2.404280,-0.174044,
        -3.141593,2.407836,1.410646,
        -3.141593,2.455966,-0.641220,
        -3.141593,2.456751,-1.503343,
        -3.141593,2.478122,2.609777,
        -3.141593,2.579248,-2.364801,
        -3.141593,2.587733,0.232235,
        -3.141593,2.588535,-2.976617,
        -3.141593,2.592629,0.909713,
        -3.141593,2.637879,2.121598,
        -3.141593,2.664082,-1.071402,
        -3.141593,2.704512,1.488683,
        -3.141593,2.719179,-1.780644,
        -3.141593,2.724429,-0.348782,
        -3.141593,2.760637,2.781135,
        -3.141593,2.863997,0.549740,
        -3.141593,2.881233,-2.582574,
        -3.141593,2.979456,-1.009304,
        -3.141593,2.984978,1.954446,
        -2.879793,0.080730,-2.622036,
        -2.879793,0.203748,0.590255,
        -2.879793,0.290380,-0.796960,
        -2.879793,0.297065,2.160177,
        -2.879793,0.370812,-2.547742,
        -2.879793,0.379747,-1.713059,
        -2.879793,0.418037,1.360082,
        -2.879793,0.419640,2.974122,
        -2.879793,0.458961,-0.107937,
        -2.879793,0.496451,0.593859,
        -2.879793,0.571433,-1.188419,
        -2.879793,0.591903,2.010408,
        -2.879793,0.615174,-0.666022,
        -2.879793,0.618092,2.539118,
        -2.879793,0.619654,-2.216909,
        -2.879793,0.640596,-2.797608,
        -2.879793,0.674054,1.013609,
        -2.879793,0.685216,-1.712745,
        -2.879793,0.700570,1.559208,
        -2.879793,0.709924,0.218797,
        -2.879793,0.733948,3.021411,
        -2.879793,0.762244,-0.273349,
        -2.879793,0.831641,0.616218,
        -2.879793,0.850607,-1.335383,
        -2.879793,0.864993,2.278281,
        -2.879793,0.866395,-2.492918,
        -2.879793,0.866955,-0.931351,
        -2.879793,0.899350,-2.069907,
        -2.879793,0.906588,1.252948,
        -2.879793,0.910508,1.903534,
        -2.879793,0.914445,2.687814,
        -2.879793,0.929350,-2.910676,
        -2.879793,0.960688,-0.577532,
        -2.879793,0.961715,0.014085,
        -2.879793,0.992645,-1.723759,
        -2.879793,1.014500,0.906702,
        -2.879793,1.025536,1.576859,
        -2.879793,1.045978,0.343692,
        -2.879793,1.050534,3.042579,
        -2.879793,1.113094,-0.272241,
        -2.879793,1.128190,-2.288622,
        -2.879793,1.134497,-1.109631,
        -2.879793,1.134796,2.434290,
        -2.879793,1.135813,-2.658866,
        -2.879793,1.135834,-1.431087,
        -2.879793,1.168678,2.108595,
        -2.879793,1.184361,0.625864,
        -2.879793,1.194549,-0.801900,
        -2.879793,1.195355,1.200319,
        -2.879793,1.214854,-1.966664,
        -2.879793,1.218453,2.777284,
        -2.879793,1.242974,-2.977882,
        -2.879793,1.247695,1.809890,
        -2.879793,1.256302,0.062705,
        -2.879793,1.301460,-0.509251,
        -2.879793,1.310100,1.488307,
        -2.879793,1.325504,-1.671296,
        -2.879793,1.335669,0.897240,
        -2.879793,1.367556,0.358405,
        -2.879793,1.374601,-2.456031,
        -2.879793,1.391910,3.030322,
        -2.879793,1.413347,-1.339188,
        -2.879793,1.413430,2.550734,
        -2.879793,1.417623,-0.202254,
        -2.879793,1.433411,-1.035986,
        -2.879793,1.440580,2.260026,
        -2.879793,1.448123,-2.163292,
        -2.879793,1.450119,-2.739050,
        -2.879793,1.488617,1.159546,
        -2.879793,1.504028,0.636812,
        -2.879793,1.506530,1.988858,
        -2.879793,1.510910,-0.740472,
        -2.879793,1.545873,-1.877082,
        -2.879793,1.554776,0.100351,
        -2.879793,1.555207,1.708892,
        -2.879793,1.569311,-3.003978,
        -2.879793,1.574060,2.783565,
        -2.879793,1.599406,1.426070,
        -2.879793,1.606104,-0.451151,
        -2.879793,1.617114,-1.572037,
        -2.879793,1.658653,-2.409879,
        -2.879793,1.662363,0.915448,
        -2.879793,1.683268,0.374177,
        -2.879793,1.696674,-1.253233,
        -2.879793,1.699788,2.452890,
        -2.879793,1.721184,-0.153948,
        -2.879793,1.732874,-0.951685,
        -2.879793,1.736875,3.023612,
        -2.879793,1.741644,-2.695914,
        -2.879793,1.756530,-2.105156,
        -2.879793,1.766067,2.153843,
        -2.879793,1.809167,1.213774,
        -2.879793,1.812176,0.656819,
        -2.879793,1.824694,1.850876,
        -2.879793,1.836682,-0.676006,
        -2.879793,1.846342,-1.796954,
        -2.879793,1.869911,2.701025,
        -2.879793,1.870545,-2.976234,
        -2.879793,1.872854,1.543456,
        -2.879793,1.875281,0.107086,
        -2.879793,1.904882,-1.478089,
        -2.879793,1.909857,-0.381669,
        -2.879793,1.948085,-2.363031,
        -2.879793,1.977582,0.969896,
        -2.879793,1.982574,-1.115518,
        -2.879793,1.998705,2.370472,
        -2.879793,2.001437,0.381215,
        -2.879793,2.033014,2.977192,
        -2.879793,2.041297,-2.664597,
        -2.879793,2.063651,-2.044600,
        -2.879793,2.073035,2.027405,
        -2.879793,2.089241,-0.114193,
        -2.879793,2.104282,-0.813972,
        -2.879793,2.108018,1.311430,
        -2.879793,2.115563,0.685730,
        -2.879793,2.138223,1.678323,
        -2.879793,2.153910,-1.714959,
        -2.879793,2.164094,-2.991358,
        -2.879793,2.174483,2.653697,
        -2.879793,2.180315,-1.357379,
        -2.879793,2.187132,-0.479875,
        -2.879793,2.247462,0.191623,
        -2.879793,2.251330,-2.330739,
        -2.879793,2.295480,1.008277,
        -2.879793,2.299159,2.264663,
        -2.879793,2.343981,-1.040456,
        -2.879793,2.349581,-2.702517,
        -2.879793,2.356369,3.005869,
        -2.879793,2.379497,0.561089,
        -2.879793,2.393226,1.857173,
        -2.879793,2.396274,-1.959260,
        -2.879793,2.404280,-0.174044,
        -2.879793,2.407836,1.410646,
        -2.879793,2.455966,-0.641220,
        -2.879793,2.456751,-1.503343,
        -2.879793,2.478122,2.609777,
        -2.879793,2.579248,-2.364801,
        -2.879793,2.587733,0.232235,
        -2.879793,2.588535,-2.976617,
        -2.879793,2.592629,0.909713,
        -2.879793,2.637879,2.121598,
        -2.879793,2.664082,-1.071402,
        -2.879793,2.704512,1.488683,
        -2.879793,2.719179,-1.780644,
        -2.879793,2.724429,-0.348782,
        -2.879793,2.760637,2.781135,
        -2.879793,2.863997,0.549740,
        -2.879793,2.881233,-2.582574,
        -2.879793,2.979456,-1.009304,
        -2.879793,2.984978,1.954446,
        -2.617994,0.080730,-2.622036,
        -2.617994,0.203748,0.590255,
        -2.617994,0.290380,-0.796960,
        -2.617994,0.297065,2.160177,
        -2.617994,0.370812,-2.547742,
        -2.617994,0.379747,-1.713059,
        -2.617994,0.418037,1.360082,
        -2.617994,0.419640,2.974122,
        -2.617994,0.458961,-0.107937,
        -2.617994,0.496451,0.593859,
        -2.617994,0.571433,-1.188419,
        -2.617994,0.591903,2.010408,
        -2.617994,0.615174,-0.666022,
        -2.617994,0.618092,2.539118,
        -2.617994,0.619654,-2.216909,
        -2.617994,0.640596,-2.797608,
        -2.617994,0.674054,1.013609,
        -2.617994,0.685216,-1.712745,
        -2.617994,0.700570,1.559208,
        -2.617994,0.709924,0.218797,
        -2.617994,0.733948,3.021411,
        -2.617994,0.762244,-0.273349,
        -2.617994,0.831641,0.616218,
        -2.617994,0.850607,-1.335383,
        -2.617994,0.864993,2.278281,
        -2.617994,0.866395,-2.492918,
        -2.617994,0.866955,-0.931351,
        -2.617994,0.899350,-2.069907,
        -2.617994,0.906588,1.252948,
        -2.617994,0.910508,1.903534,
        -2.617994,0.914445,2.687814,
        -2.617994,0.929350,-2.910676,
        -2.617994,0.960688,-0.577532,
        -2.617994,0.961715,0.014085,
        -2.617994,0.992645,-1.723759,
        -2.617994,1.014500,0.906702,
        -2.617994,1.025536,1.576859,
        -2.617994,1.045978,0.343692,
        -2.617994,1.050534,3.042579,
        -2.617994,1.113094,-0.272241,
        -2.617994,1.128190,-2.288622,
        -2.617994,1.134497,-1.109631,
        -2.617994,1.134796,2.434290,
        -2.617994,1.135813,-2.658866,
        -2.617994,1.135834,-1.431087,
        -2.617994,1.168678,2.108595,
        -2.617994,1.184361,0.625864,
        -2.617994,1.194549,-0.801900,
        -2.617994,1.195355,1.200319,
        -2.617994,1.214854,-1.966664,
        -2.617994,1.218453,2.777284,
        -2.617994,1.242974,-2.977882,
        -2.617994,1.247695,1.809890,
        -2.617994,1.256302,0.062705,
        -2.617994,1.301460,-0.509251,
        -2.617994,1.310100,1.488307,
        -2.617994,1.325504,-1.671296,
        -2.617994,1.335669,0.897240,
        -2.617994,1.367556,0.358405,
        -2.617994,1.374601,-2.456031,
        -2.617994,1.391910,3.030322,
        -2.617994,1.413347,-1.339188,
        -2.617994,1.413430,2.550734,
        -2.617994,1.417623,-0.202254,
        -2.617994,1.433411,-1.035986,
        -2.617994,1.440580,2.260026,
        -2.617994,1.448123,-2.163292,
        -2.617994,1.450119,-2.739050,
        -2.617994,1.488617,1.159546,
        -2.617994,1.504028,0.636812,
        -2.617994,1.506530,1.988858,
        -2.617994,1.510910,-0.740472,
        -2.617994,1.545873,-1.877082,
        -2.617994,1.554776,0.100351,
        -2.617994,1.555207,1.708892,
        -2.617994,1.569311,-3.003978,
        -2.617994,1.574060,2.783565,
        -2.617994,1.599406,1.426070,
        -2.617994,1.606104,-0.451151,
        -2.617994,1.617114,-1.572037,
        -2.617994,1.658653,-2.409879,
        -2.617994,1.662363,0.915448,
        -2.617994,1.683268,0.374177,
        -2.617994,1.696674,-1.253233,
        -2.617994,1.699788,2.452890,
        -2.617994,1.721184,-0.153948,
        -2.617994,1.732874,-0.951685,
        -2.617994,1.736875,3.023612,
        -2.617994,1.741644,-2.695914,
        -2.617994,1.756530,-2.105156,
        -2.617994,1.766067,2.153843,
        -2.617994,1.809167,1.213774,
        -2.617994,1.812176,0.656819,
        -2.617994,1.824694,1.850876,
        -2.617994,1.836682,-0.676006,
        -2.617994,1.846342,-1.796954,
        -2.617994,1.869911,2.701025,
        -2.617994,1.870545,-2.976234,
        -2.617994,1.872854,1.543456,
        -2.617994,1.875281,0.107086,
        -2.617994,1.904882,-1.478089,
        -2.617994,1.909857,-0.381669,
        -2.617994,1.948085,-2.363031,
        -2.617994,1.977582,0.969896,
        -2.617994,1.982574,-1.115518,
        -2.617994,1.998705,2.370472,
        -2.617994,2.001437,0.381215,
        -2.617994,2.033014,2.977192,
        -2.617994,2.041297,-2.664597,
        -2.617994,2.063651,-2.044600,
        -2.617994,2.073035,2.027405,
        -2.617994,2.089241,-0.114193,
        -2.617994,2.104282,-0.813972,
        -2.617994,2.108018,1.311430,
        -2.617994,2.115563,0.685730,
        -2.617994,2.138223,1.678323,
        -2.617994,2.153910,-1.714959,
        -2.617994,2.164094,-2.991358,
        -2.617994,2.174483,2.653697,
        -2.617994,2.180315,-1.357379,
        -2.617994,2.187132,-0.479875,
        -2.617994,2.247462,0.191623,
        -2.617994,2.251330,-2.330739,
        -2.617994,2.295480,1.008277,
        -2.617994,2.299159,2.264663,
        -2.617994,2.343981,-1.040456,
        -2.617994,2.349581,-2.702517,
        -2.617994,2.356369,3.005869,
        -2.617994,2.379497,0.561089,
        -2.617994,2.393226,1.857173,
        -2.617994,2.396274,-1.959260,
        -2.617994,2.404280,-0.174044,
        -2.617994,2.407836,1.410646,
        -2.617994,2.455966,-0.641220,
        -2.617994,2.456751,-1.503343,
        -2.617994,2.478122,2.609777,
        -2.617994,2.579248,-2.364801,
        -2.617994,2.587733,0.232235,
        -2.617994,2.588535,-2.976617,
        -2.617994,2.592629,0.909713,
        -2.617994,2.637879,2.121598,
        -2.617994,2.664082,-1.071402,
        -2.617994,2.704512,1.488683,
        -2.617994,2.719179,-1.780644,
        -2.617994,2.724429,-0.348782,
        -2.617994,2.760637,2.781135,
        -2.617994,2.863997,0.549740,
        -2.617994,2.881233,-2.582574,
        -2.617994,2.979456,-1.009304,
        -2.617994,2.984978,1.954446,
        -2.356194,0.080730,-2.622036,
        -2.356194,0.203748,0.590255,
        -2.356194,0.290380,-0.796960,
        -2.356194,0.297065,2.160177,
        -2.356194,0.370812,-2.547742,
        -2.356194,0.379747,-1.713059,
        -2.356194,0.418037,1.360082,
        -2.356194,0.419640,2.974122,
        -2.356194,0.458961,-0.107937,
        -2.356194,0.496451,0.593859,
        -2.356194,0.571433,-1.188419,
        -2.356194,0.591903,2.010408,
        -2.356194,0.615174,-0.666022,
        -2.356194,0.618092,2.539118,
        -2.356194,0.619654,-2.216909,
        -2.356194,0.640596,-2.797608,
        -2.356194,0.674054,1.013609,
        -2.356194,0.685216,-1.712745,
        -2.356194,0.700570,1.559208,
        -2.356194,0.709924,0.218797,
        -2.356194,0.733948,3.021411,
        -2.356194,0.762244,-0.273349,
        -2.356194,0.831641,0.616218,
        -2.356194,0.850607,-1.335383,
        -2.356194,0.864993,2.278281,
        -2.356194,0.866395,-2.492918,
        -2.356194,0.866955,-0.931351,
        -2.356194,0.899350,-2.069907,
        -2.356194,0.906588,1.252948,
        -2.356194,0.910508,1.903534,
        -2.356194,0.914445,2.687814,
        -2.356194,0.929350,-2.910676,
        -2.356194,0.960688,-0.577532,
        -2.356194,0.961715,0.014085,
        -2.356194,0.992645,-1.723759,
        -2.356194,1.014500,0.906702,
        -2.356194,1.025536,1.576859,
        -2.356194,1.045978,0.343692,
        -2.356194,1.050534,3.042579,
        -2.356194,1.113094,-0.272241,
        -2.356194,1.128190,-2.288622,
        -2.356194,1.134497,-1.109631,
        -2.356194,1.134796,2.434290,
        -2.356194,1.135813,-2.658866,
        -2.356194,1.135834,-1.431087,
        -2.356194,1.168678,2.108595,
        -2.356194,1.184361,0.625864,
        -2.356194,1.194549,-0.801900,
        -2.356194,1.195355,1.200319,
        -2.356194,1.214854,-1.966664,
        -2.356194,1.218453,2.777284,
        -2.356194,1.242974,-2.977882,
        -2.356194,1.247695,1.809890,
        -2.356194,1.256302,0.062705,
        -2.356194,1.301460,-0.509251,
        -2.356194,1.310100,1.488307,
        -2.356194,1.325504,-1.671296,
        -2.356194,1.335669,0.897240,
        -2.356194,1.367556,0.358405,
        -2.356194,1.374601,-2.456031,
        -2.356194,1.391910,3.030322,
        -2.356194,1.413347,-1.339188,
        -2.356194,1.413430,2.550734,
        -2.356194,1.417623,-0.202254,
        -2.356194,1.433411,-1.035986,
        -2.356194,1.440580,2.260026,
        -2.356194,1.448123,-2.163292,
        -2.356194,1.450119,-2.739050,
        -2.356194,1.488617,1.159546,
        -2.356194,1.504028,0.636812,
        -2.356194,1.506530,1.988858,
        -2.356194,1.510910,-0.740472,
        -2.356194,1.545873,-1.877082,
        -2.356194,1.554776,0.100351,
        -2.356194,1.555207,1.708892,
        -2.356194,1.569311,-3.003978,
        -2.356194,1.574060,2.783565,
        -2.356194,1.599406,1.426070,
        -2.356194,1.606104,-0.451151,
        -2.356194,1.617114,-1.572037,
        -2.356194,1.658653,-2.409879,
        -2.356194,1.662363,0.915448,
        -2.356194,1.683268,0.374177,
        -2.356194,1.696674,-1.253233,
        -2.356194,1.699788,2.452890,
        -2.356194,1.721184,-0.153948,
        -2.356194,1.732874,-0.951685,
        -2.356194,1.736875,3.023612,
        -2.356194,1.741644,-2.695914,
        -2.356194,1.756530,-2.105156,
        -2.356194,1.766067,2.153843,
        -2.356194,1.809167,1.213774,
        -2.356194,1.812176,0.656819,
        -2.356194,1.824694,1.850876,
        -2.356194,1.836682,-0.676006,
        -2.356194,1.846342,-1.796954,
        -2.356194,1.869911,2.701025,
        -2.356194,1.870545,-2.976234,
        -2.356194,1.872854,1.543456,
        -2.356194,1.875281,0.107086,
        -2.356194,1.904882,-1.478089,
        -2.356194,1.909857,-0.381669,
        -2.356194,1.948085,-2.363031,
        -2.356194,1.977582,0.969896,
        -2.356194,1.982574,-1.115518,
        -2.356194,1.998705,2.370472,
        -2.356194,2.001437,0.381215,
        -2.356194,2.033014,2.977192,
        -2.356194,2.041297,-2.664597,
        -2.356194,2.063651,-2.044600,
        -2.356194,2.073035,2.027405,
        -2.356194,2.089241,-0.114193,
        -2.356194,2.104282,-0.813972,
        -2.356194,2.108018,1.311430,
        -2.356194,2.115563,0.685730,
        -2.356194,2.138223,1.678323,
        -2.356194,2.153910,-1.714959,
        -2.356194,2.164094,-2.991358,
        -2.356194,2.174483,2.653697,
        -2.356194,2.180315,-1.357379,
        -2.356194,2.187132,-0.479875,
        -2.356194,2.247462,0.191623,
        -2.356194,2.251330,-2.330739,
        -2.356194,2.295480,1.008277,
        -2.356194,2.299159,2.264663,
        -2.356194,2.343981,-1.040456,
        -2.356194,2.349581,-2.702517,
        -2.356194,2.356369,3.005869,
        -2.356194,2.379497,0.561089,
        -2.356194,2.393226,1.857173,
        -2.356194,2.396274,-1.959260,
        -2.356194,2.404280,-0.174044,
        -2.356194,2.407836,1.410646,
        -2.356194,2.455966,-0.641220,
        -2.356194,2.456751,-1.503343,
        -2.356194,2.478122,2.609777,
        -2.356194,2.579248,-2.364801,
        -2.356194,2.587733,0.232235,
        -2.356194,2.588535,-2.976617,
        -2.356194,2.592629,0.909713,
        -2.356194,2.637879,2.121598,
        -2.356194,2.664082,-1.071402,
        -2.356194,2.704512,1.488683,
        -2.356194,2.719179,-1.780644,
        -2.356194,2.724429,-0.348782,
        -2.356194,2.760637,2.781135,
        -2.356194,2.863997,0.549740,
        -2.356194,2.881233,-2.582574,
        -2.356194,2.979456,-1.009304,
        -2.356194,2.984978,1.954446,
        -2.094395,0.080730,-2.622036,
        -2.094395,0.203748,0.590255,
        -2.094395,0.290380,-0.796960,
        -2.094395,0.297065,2.160177,
        -2.094395,0.370812,-2.547742,
        -2.094395,0.379747,-1.713059,
        -2.094395,0.418037,1.360082,
        -2.094395,0.419640,2.974122,
        -2.094395,0.458961,-0.107937,
        -2.094395,0.496451,0.593859,
        -2.094395,0.571433,-1.188419,
        -2.094395,0.591903,2.010408,
        -2.094395,0.615174,-0.666022,
        -2.094395,0.618092,2.539118,
        -2.094395,0.619654,-2.216909,
        -2.094395,0.640596,-2.797608,
        -2.094395,0.674054,1.013609,
        -2.094395,0.685216,-1.712745,
        -2.094395,0.700570,1.559208,
        -2.094395,0.709924,0.218797,
        -2.094395,0.733948,3.021411,
        -2.094395,0.762244,-0.273349,
        -2.094395,0.831641,0.616218,
        -2.094395,0.850607,-1.335383,
        -2.094395,0.864993,2.278281,
        -2.094395,0.866395,-2.492918,
        -2.094395,0.866955,-0.931351,
        -2.094395,0.899350,-2.069907,
        -2.094395,0.906588,1.252948,
        -2.094395,0.910508,1.903534,
        -2.094395,0.914445,2.687814,
        -2.094395,0.929350,-2.910676,
        -2.094395,0.960688,-0.577532,
        -2.094395,0.961715,0.014085,
        -2.094395,0.992645,-1.723759,
        -2.094395,1.014500,0.906702,
        -2.094395,1.025536,1.576859,
        -2.094395,1.045978,0.343692,
        -2.094395,1.050534,3.042579,
        -2.094395,1.113094,-0.272241,
        -2.094395,1.128190,-2.288622,
        -2.094395,1.134497,-1.109631,
        -2.094395,1.134796,2.434290,
        -2.094395,1.135813,-2.658866,
        -2.094395,1.135834,-1.431087,
        -2.094395,1.168678,2.108595,
        -2.094395,1.184361,0.625864,
        -2.094395,1.194549,-0.801900,
        -2.094395,1.195355,1.200319,
        -2.094395,1.214854,-1.966664,
        -2.094395,1.218453,2.777284,
        -2.094395,1.242974,-2.977882,
        -2.094395,1.247695,1.809890,
        -2.094395,1.256302,0.062705,
        -2.094395,1.301460,-0.509251,
        -2.094395,1.310100,1.488307,
        -2.094395,1.325504,-1.671296,
        -2.094395,1.335669,0.897240,
        -2.094395,1.367556,0.358405,
        -2.094395,1.374601,-2.456031,
        -2.094395,1.391910,3.030322,
        -2.094395,1.413347,-1.339188,
        -2.094395,1.413430,2.550734,
        -2.094395,1.417623,-0.202254,
        -2.094395,1.433411,-1.035986,
        -2.094395,1.440580,2.260026,
        -2.094395,1.448123,-2.163292,
        -2.094395,1.450119,-2.739050,
        -2.094395,1.488617,1.159546,
        -2.094395,1.504028,0.636812,
        -2.094395,1.506530,1.988858,
        -2.094395,1.510910,-0.740472,
        -2.094395,1.545873,-1.877082,
        -2.094395,1.554776,0.100351,
        -2.094395,1.555207,1.708892,
        -2.094395,1.569311,-3.003978,
        -2.094395,1.574060,2.783565,
        -2.094395,1.599406,1.426070,
        -2.094395,1.606104,-0.451151,
        -2.094395,1.617114,-1.572037,
        -2.094395,1.658653,-2.409879,
        -2.094395,1.662363,0.915448,
        -2.094395,1.683268,0.374177,
        -2.094395,1.696674,-1.253233,
        -2.094395,1.699788,2.452890,
        -2.094395,1.721184,-0.153948,
        -2.094395,1.732874,-0.951685,
        -2.094395,1.736875,3.023612,
        -2.094395,1.741644,-2.695914,
        -2.094395,1.756530,-2.105156,
        -2.094395,1.766067,2.153843,
        -2.094395,1.809167,1.213774,
        -2.094395,1.812176,0.656819,
        -2.094395,1.824694,1.850876,
        -2.094395,1.836682,-0.676006,
        -2.094395,1.846342,-1.796954,
        -2.094395,1.869911,2.701025,
        -2.094395,1.870545,-2.976234,
        -2.094395,1.872854,1.543456,
        -2.094395,1.875281,0.107086,
        -2.094395,1.904882,-1.478089,
        -2.094395,1.909857,-0.381669,
        -2.094395,1.948085,-2.363031,
        -2.094395,1.977582,0.969896,
        -2.094395,1.982574,-1.115518,
        -2.094395,1.998705,2.370472,
        -2.094395,2.001437,0.381215,
        -2.094395,2.033014,2.977192,
        -2.094395,2.041297,-2.664597,
        -2.094395,2.063651,-2.044600,
        -2.094395,2.073035,2.027405,
        -2.094395,2.089241,-0.114193,
        -2.094395,2.104282,-0.813972,
        -2.094395,2.108018,1.311430,
        -2.094395,2.115563,0.685730,
        -2.094395,2.138223,1.678323,
        -2.094395,2.153910,-1.714959,
        -2.094395,2.164094,-2.991358,
        -2.094395,2.174483,2.653697,
        -2.094395,2.180315,-1.357379,
        -2.094395,2.187132,-0.479875,
        -2.094395,2.247462,0.191623,
        -2.094395,2.251330,-2.330739,
        -2.094395,2.295480,1.008277,
        -2.094395,2.299159,2.264663,
        -2.094395,2.343981,-1.040456,
        -2.094395,2.349581,-2.702517,
        -2.094395,2.356369,3.005869,
        -2.094395,2.379497,0.561089,
        -2.094395,2.393226,1.857173,
        -2.094395,2.396274,-1.959260,
        -2.094395,2.404280,-0.174044,
        -2.094395,2.407836,1.410646,
        -2.094395,2.455966,-0.641220,
        -2.094395,2.456751,-1.503343,
        -2.094395,2.478122,2.609777,
        -2.094395,2.579248,-2.364801,
        -2.094395,2.587733,0.232235,
        -2.094395,2.588535,-2.976617,
        -2.094395,2.592629,0.909713,
        -2.094395,2.637879,2.121598,
        -2.094395,2.664082,-1.071402,
        -2.094395,2.704512,1.488683,
        -2.094395,2.719179,-1.780644,
        -2.094395,2.724429,-0.348782,
        -2.094395,2.760637,2.781135,
        -2.094395,2.863997,0.549740,
        -2.094395,2.881233,-2.582574,
        -2.094395,2.979456,-1.009304,
        -2.094395,2.984978,1.954446,
        -1.832596,0.080730,-2.622036,
        -1.832596,0.203748,0.590255,
        -1.832596,0.290380,-0.796960,
        -1.832596,0.297065,2.160177,
        -1.832596,0.370812,-2.547742,
        -1.832596,0.379747,-1.713059,
        -1.832596,0.418037,1.360082,
        -1.832596,0.419640,2.974122,
        -1.832596,0.458961,-0.107937,
        -1.832596,0.496451,0.593859,
        -1.832596,0.571433,-1.188419,
        -1.832596,0.591903,2.010408,
        -1.832596,0.615174,-0.666022,
        -1.832596,0.618092,2.539118,
        -1.832596,0.619654,-2.216909,
        -1.832596,0.640596,-2.797608,
        -1.832596,0.674054,1.013609,
        -1.832596,0.685216,-1.712745,
        -1.832596,0.700570,1.559208,
        -1.832596,0.709924,0.218797,
        -1.832596,0.733948,3.021411,
        -1.832596,0.762244,-0.273349,
        -1.832596,0.831641,0.616218,
        -1.832596,0.850607,-1.335383,
        -1.832596,0.864993,2.278281,
        -1.832596,0.866395,-2.492918,
        -1.832596,0.866955,-0.931351,
        -1.832596,0.899350,-2.069907,
        -1.832596,0.906588,1.252948,
        -1.832596,0.910508,1.903534,
        -1.832596,0.914445,2.687814,
        -1.832596,0.929350,-2.910676,
        -1.832596,0.960688,-0.577532,
        -1.832596,0.961715,0.014085,
        -1.832596,0.992645,-1.723759,
        -1.832596,1.014500,0.906702,
        -1.832596,1.025536,1.576859,
        -1.832596,1.045978,0.343692,
        -1.832596,1.050534,3.042579,
        -1.832596,1.113094,-0.272241,
        -1.832596,1.128190,-2.288622,
        -1.832596,1.134497,-1.109631,
        -1.832596,1.134796,2.434290,
        -1.832596,1.135813,-2.658866,
        -1.832596,1.135834,-1.431087,
        -1.832596,1.168678,2.108595,
        -1.832596,1.184361,0.625864,
        -1.832596,1.194549,-0.801900,
        -1.832596,1.195355,1.200319,
        -1.832596,1.214854,-1.966664,
        -1.832596,1.218453,2.777284,
        -1.832596,1.242974,-2.977882,
        -1.832596,1.247695,1.809890,
        -1.832596,1.256302,0.062705,
        -1.832596,1.301460,-0.509251,
        -1.832596,1.310100,1.488307,
        -1.832596,1.325504,-1.671296,
        -1.832596,1.335669,0.897240,
        -1.832596,1.367556,0.358405,
        -1.832596,1.374601,-2.456031,
        -1.832596,1.391910,3.030322,
        -1.832596,1.413347,-1.339188,
        -1.832596,1.413430,2.550734,
        -1.832596,1.417623,-0.202254,
        -1.832596,1.433411,-1.035986,
        -1.832596,1.440580,2.260026,
        -1.832596,1.448123,-2.163292,
        -1.832596,1.450119,-2.739050,
        -1.832596,1.488617,1.159546,
        -1.832596,1.504028,0.636812,
        -1.832596,1.506530,1.988858,
        -1.832596,1.510910,-0.740472,
        -1.832596,1.545873,-1.877082,
        -1.832596,1.554776,0.100351,
        -1.832596,1.555207,1.708892,
        -1.832596,1.569311,-3.003978,
        -1.832596,1.574060,2.783565,
        -1.832596,1.599406,1.426070,
        -1.832596,1.606104,-0.451151,
        -1.832596,1.617114,-1.572037,
        -1.832596,1.658653,-2.409879,
        -1.832596,1.662363,0.915448,
        -1.832596,1.683268,0.374177,
        -1.832596,1.696674,-1.253233,
        -1.832596,1.699788,2.452890,
        -1.832596,1.721184,-0.153948,
        -1.832596,1.732874,-0.951685,
        -1.832596,1.736875,3.023612,
        -1.832596,1.741644,-2.695914,
        -1.832596,1.756530,-2.105156,
        -1.832596,1.766067,2.153843,
        -1.832596,1.809167,1.213774,
        -1.832596,1.812176,0.656819,
        -1.832596,1.824694,1.850876,
        -1.832596,1.836682,-0.676006,
        -1.832596,1.846342,-1.796954,
        -1.832596,1.869911,2.701025,
        -1.832596,1.870545,-2.976234,
        -1.832596,1.872854,1.543456,
        -1.832596,1.875281,0.107086,
        -1.832596,1.904882,-1.478089,
        -1.832596,1.909857,-0.381669,
        -1.832596,1.948085,-2.363031,
        -1.832596,1.977582,0.969896,
        -1.832596,1.982574,-1.115518,
        -1.832596,1.998705,2.370472,
        -1.832596,2.001437,0.381215,
        -1.832596,2.033014,2.977192,
        -1.832596,2.041297,-2.664597,
        -1.832596,2.063651,-2.044600,
        -1.832596,2.073035,2.027405,
        -1.832596,2.089241,-0.114193,
        -1.832596,2.104282,-0.813972,
        -1.832596,2.108018,1.311430,
        -1.832596,2.115563,0.685730,
        -1.832596,2.138223,1.678323,
        -1.832596,2.153910,-1.714959,
        -1.832596,2.164094,-2.991358,
        -1.832596,2.174483,2.653697,
        -1.832596,2.180315,-1.357379,
        -1.832596,2.187132,-0.479875,
        -1.832596,2.247462,0.191623,
        -1.832596,2.251330,-2.330739,
        -1.832596,2.295480,1.008277,
        -1.832596,2.299159,2.264663,
        -1.832596,2.343981,-1.040456,
        -1.832596,2.349581,-2.702517,
        -1.832596,2.356369,3.005869,
        -1.832596,2.379497,0.561089,
        -1.832596,2.393226,1.857173,
        -1.832596,2.396274,-1.959260,
        -1.832596,2.404280,-0.174044,
        -1.832596,2.407836,1.410646,
        -1.832596,2.455966,-0.641220,
        -1.832596,2.456751,-1.503343,
        -1.832596,2.478122,2.609777,
        -1.832596,2.579248,-2.364801,
        -1.832596,2.587733,0.232235,
        -1.832596,2.588535,-2.976617,
        -1.832596,2.592629,0.909713,
        -1.832596,2.637879,2.121598,
        -1.832596,2.664082,-1.071402,
        -1.832596,2.704512,1.488683,
        -1.832596,2.719179,-1.780644,
        -1.832596,2.724429,-0.348782,
        -1.832596,2.760637,2.781135,
        -1.832596,2.863997,0.549740,
        -1.832596,2.881233,-2.582574,
        -1.832596,2.979456,-1.009304,
        -1.832596,2.984978,1.954446,
        -1.570796,0.080730,-2.622036,
        -1.570796,0.203748,0.590255,
        -1.570796,0.290380,-0.796960,
        -1.570796,0.297065,2.160177,
        -1.570796,0.370812,-2.547742,
        -1.570796,0.379747,-1.713059,
        -1.570796,0.418037,1.360082,
        -1.570796,0.419640,2.974122,
        -1.570796,0.458961,-0.107937,
        -1.570796,0.496451,0.593859,
        -1.570796,0.571433,-1.188419,
        -1.570796,0.591903,2.010408,
        -1.570796,0.615174,-0.666022,
        -1.570796,0.618092,2.539118,
        -1.570796,0.619654,-2.216909,
        -1.570796,0.640596,-2.797608,
        -1.570796,0.674054,1.013609,
        -1.570796,0.685216,-1.712745,
        -1.570796,0.700570,1.559208,
        -1.570796,0.709924,0.218797,
        -1.570796,0.733948,3.021411,
        -1.570796,0.762244,-0.273349,
        -1.570796,0.831641,0.616218,
        -1.570796,0.850607,-1.335383,
        -1.570796,0.864993,2.278281,
        -1.570796,0.866395,-2.492918,
        -1.570796,0.866955,-0.931351,
        -1.570796,0.899350,-2.069907,
        -1.570796,0.906588,1.252948,
        -1.570796,0.910508,1.903534,
        -1.570796,0.914445,2.687814,
        -1.570796,0.929350,-2.910676,
        -1.570796,0.960688,-0.577532,
        -1.570796,0.961715,0.014085,
        -1.570796,0.992645,-1.723759,
        -1.570796,1.014500,0.906702,
        -1.570796,1.025536,1.576859,
        -1.570796,1.045978,0.343692,
        -1.570796,1.050534,3.042579,
        -1.570796,1.113094,-0.272241,
        -1.570796,1.128190,-2.288622,
        -1.570796,1.134497,-1.109631,
        -1.570796,1.134796,2.434290,
        -1.570796,1.135813,-2.658866,
        -1.570796,1.135834,-1.431087,
        -1.570796,1.168678,2.108595,
        -1.570796,1.184361,0.625864,
        -1.570796,1.194549,-0.801900,
        -1.570796,1.195355,1.200319,
        -1.570796,1.214854,-1.966664,
        -1.570796,1.218453,2.777284,
        -1.570796,1.242974,-2.977882,
        -1.570796,1.247695,1.809890,
        -1.570796,1.256302,0.062705,
        -1.570796,1.301460,-0.509251,
        -1.570796,1.310100,1.488307,
        -1.570796,1.325504,-1.671296,
        -1.570796,1.335669,0.897240,
        -1.570796,1.367556,0.358405,
        -1.570796,1.374601,-2.456031,
        -1.570796,1.391910,3.030322,
        -1.570796,1.413347,-1.339188,
        -1.570796,1.413430,2.550734,
        -1.570796,1.417623,-0.202254,
        -1.570796,1.433411,-1.035986,
        -1.570796,1.440580,2.260026,
        -1.570796,1.448123,-2.163292,
        -1.570796,1.450119,-2.739050,
        -1.570796,1.488617,1.159546,
        -1.570796,1.504028,0.636812,
        -1.570796,1.506530,1.988858,
        -1.570796,1.510910,-0.740472,
        -1.570796,1.545873,-1.877082,
        -1.570796,1.554776,0.100351,
        -1.570796,1.555207,1.708892,
        -1.570796,1.569311,-3.003978,
        -1.570796,1.574060,2.783565,
        -1.570796,1.599406,1.426070,
        -1.570796,1.606104,-0.451151,
        -1.570796,1.617114,-1.572037,
        -1.570796,1.658653,-2.409879,
        -1.570796,1.662363,0.915448,
        -1.570796,1.683268,0.374177,
        -1.570796,1.696674,-1.253233,
        -1.570796,1.699788,2.452890,
        -1.570796,1.721184,-0.153948,
        -1.570796,1.732874,-0.951685,
        -1.570796,1.736875,3.023612,
        -1.570796,1.741644,-2.695914,
        -1.570796,1.756530,-2.105156,
        -1.570796,1.766067,2.153843,
        -1.570796,1.809167,1.213774,
        -1.570796,1.812176,0.656819,
        -1.570796,1.824694,1.850876,
        -1.570796,1.836682,-0.676006,
        -1.570796,1.846342,-1.796954,
        -1.570796,1.869911,2.701025,
        -1.570796,1.870545,-2.976234,
        -1.570796,1.872854,1.543456,
        -1.570796,1.875281,0.107086,
        -1.570796,1.904882,-1.478089,
        -1.570796,1.909857,-0.381669,
        -1.570796,1.948085,-2.363031,
        -1.570796,1.977582,0.969896,
        -1.570796,1.982574,-1.115518,
        -1.570796,1.998705,2.370472,
        -1.570796,2.001437,0.381215,
        -1.570796,2.033014,2.977192,
        -1.570796,2.041297,-2.664597,
        -1.570796,2.063651,-2.044600,
        -1.570796,2.073035,2.027405,
        -1.570796,2.089241,-0.114193,
        -1.570796,2.104282,-0.813972,
        -1.570796,2.108018,1.311430,
        -1.570796,2.115563,0.685730,
        -1.570796,2.138223,1.678323,
        -1.570796,2.153910,-1.714959,
        -1.570796,2.164094,-2.991358,
        -1.570796,2.174483,2.653697,
        -1.570796,2.180315,-1.357379,
        -1.570796,2.187132,-0.479875,
        -1.570796,2.247462,0.191623,
        -1.570796,2.251330,-2.330739,
        -1.570796,2.295480,1.008277,
        -1.570796,2.299159,2.264663,
        -1.570796,2.343981,-1.040456,
        -1.570796,2.349581,-2.702517,
        -1.570796,2.356369,3.005869,
        -1.570796,2.379497,0.561089,
        -1.570796,2.393226,1.857173,
        -1.570796,2.396274,-1.959260,
        -1.570796,2.404280,-0.174044,
        -1.570796,2.407836,1.410646,
        -1.570796,2.455966,-0.641220,
        -1.570796,2.456751,-1.503343,
        -1.570796,2.478122,2.609777,
        -1.570796,2.579248,-2.364801,
        -1.570796,2.587733,0.232235,
        -1.570796,2.588535,-2.976617,
        -1.570796,2.592629,0.909713,
        -1.570796,2.637879,2.121598,
        -1.570796,2.664082,-1.071402,
        -1.570796,2.704512,1.488683,
        -1.570796,2.719179,-1.780644,
        -1.570796,2.724429,-0.348782,
        -1.570796,2.760637,2.781135,
        -1.570796,2.863997,0.549740,
        -1.570796,2.881233,-2.582574,
        -1.570796,2.979456,-1.009304,
        -1.570796,2.984978,1.954446,
        -1.308997,0.080730,-2.622036,
        -1.308997,0.203748,0.590255,
        -1.308997,0.290380,-0.796960,
        -1.308997,0.297065,2.160177,
        -1.308997,0.370812,-2.547742,
        -1.308997,0.379747,-1.713059,
        -1.308997,0.418037,1.360082,
        -1.308997,0.419640,2.974122,
        -1.308997,0.458961,-0.107937,
        -1.308997,0.496451,0.593859,
        -1.308997,0.571433,-1.188419,
        -1.308997,0.591903,2.010408,
        -1.308997,0.615174,-0.666022,
        -1.308997,0.618092,2.539118,
        -1.308997,0.619654,-2.216909,
        -1.308997,0.640596,-2.797608,
        -1.308997,0.674054,1.013609,
        -1.308997,0.685216,-1.712745,
        -1.308997,0.700570,1.559208,
        -1.308997,0.709924,0.218797,
        -1.308997,0.733948,3.021411,
        -1.308997,0.762244,-0.273349,
        -1.308997,0.831641,0.616218,
        -1.308997,0.850607,-1.335383,
        -1.308997,0.864993,2.278281,
        -1.308997,0.866395,-2.492918,
        -1.308997,0.866955,-0.931351,
        -1.308997,0.899350,-2.069907,
        -1.308997,0.906588,1.252948,
        -1.308997,0.910508,1.903534,
        -1.308997,0.914445,2.687814,
        -1.308997,0.929350,-2.910676,
        -1.308997,0.960688,-0.577532,
        -1.308997,0.961715,0.014085,
        -1.308997,0.992645,-1.723759,
        -1.308997,1.014500,0.906702,
        -1.308997,1.025536,1.576859,
        -1.308997,1.045978,0.343692,
        -1.308997,1.050534,3.042579,
        -1.308997,1.113094,-0.272241,
        -1.308997,1.128190,-2.288622,
        -1.308997,1.134497,-1.109631,
        -1.308997,1.134796,2.434290,
        -1.308997,1.135813,-2.658866,
        -1.308997,1.135834,-1.431087,
        -1.308997,1.168678,2.108595,
        -1.308997,1.184361,0.625864,
        -1.308997,1.194549,-0.801900,
        -1.308997,1.195355,1.200319,
        -1.308997,1.214854,-1.966664,
        -1.308997,1.218453,2.777284,
        -1.308997,1.242974,-2.977882,
        -1.308997,1.247695,1.809890,
        -1.308997,1.256302,0.062705,
        -1.308997,1.301460,-0.509251,
        -1.308997,1.310100,1.488307,
        -1.308997,1.325504,-1.671296,
        -1.308997,1.335669,0.897240,
        -1.308997,1.367556,0.358405,
        -1.308997,1.374601,-2.456031,
        -1.308997,1.391910,3.030322,
        -1.308997,1.413347,-1.339188,
        -1.308997,1.413430,2.550734,
        -1.308997,1.417623,-0.202254,
        -1.308997,1.433411,-1.035986,
        -1.308997,1.440580,2.260026,
        -1.308997,1.448123,-2.163292,
        -1.308997,1.450119,-2.739050,
        -1.308997,1.488617,1.159546,
        -1.308997,1.504028,0.636812,
        -1.308997,1.506530,1.988858,
        -1.308997,1.510910,-0.740472,
        -1.308997,1.545873,-1.877082,
        -1.308997,1.554776,0.100351,
        -1.308997,1.555207,1.708892,
        -1.308997,1.569311,-3.003978,
        -1.308997,1.574060,2.783565,
        -1.308997,1.599406,1.426070,
        -1.308997,1.606104,-0.451151,
        -1.308997,1.617114,-1.572037,
        -1.308997,1.658653,-2.409879,
        -1.308997,1.662363,0.915448,
        -1.308997,1.683268,0.374177,
        -1.308997,1.696674,-1.253233,
        -1.308997,1.699788,2.452890,
        -1.308997,1.721184,-0.153948,
        -1.308997,1.732874,-0.951685,
        -1.308997,1.736875,3.023612,
        -1.308997,1.741644,-2.695914,
        -1.308997,1.756530,-2.105156,
        -1.308997,1.766067,2.153843,
        -1.308997,1.809167,1.213774,
        -1.308997,1.812176,0.656819,
        -1.308997,1.824694,1.850876,
        -1.308997,1.836682,-0.676006,
        -1.308997,1.846342,-1.796954,
        -1.308997,1.869911,2.701025,
        -1.308997,1.870545,-2.976234,
        -1.308997,1.872854,1.543456,
        -1.308997,1.875281,0.107086,
        -1.308997,1.904882,-1.478089,
        -1.308997,1.909857,-0.381669,
        -1.308997,1.948085,-2.363031,
        -1.308997,1.977582,0.969896,
        -1.308997,1.982574,-1.115518,
        -1.308997,1.998705,2.370472,
        -1.308997,2.001437,0.381215,
        -1.308997,2.033014,2.977192,
        -1.308997,2.041297,-2.664597,
        -1.308997,2.063651,-2.044600,
        -1.308997,2.073035,2.027405,
        -1.308997,2.089241,-0.114193,
        -1.308997,2.104282,-0.813972,
        -1.308997,2.108018,1.311430,
        -1.308997,2.115563,0.685730,
        -1.308997,2.138223,1.678323,
        -1.308997,2.153910,-1.714959,
        -1.308997,2.164094,-2.991358,
        -1.308997,2.174483,2.653697,
        -1.308997,2.180315,-1.357379,
        -1.308997,2.187132,-0.479875,
        -1.308997,2.247462,0.191623,
        -1.308997,2.251330,-2.330739,
        -1.308997,2.295480,1.008277,
        -1.308997,2.299159,2.264663,
        -1.308997,2.343981,-1.040456,
        -1.308997,2.349581,-2.702517,
        -1.308997,2.356369,3.005869,
        -1.308997,2.379497,0.561089,
        -1.308997,2.393226,1.857173,
        -1.308997,2.396274,-1.959260,
        -1.308997,2.404280,-0.174044,
        -1.308997,2.407836,1.410646,
        -1.308997,2.455966,-0.641220,
        -1.308997,2.456751,-1.503343,
        -1.308997,2.478122,2.609777,
        -1.308997,2.579248,-2.364801,
        -1.308997,2.587733,0.232235,
        -1.308997,2.588535,-2.976617,
        -1.308997,2.592629,0.909713,
        -1.308997,2.637879,2.121598,
        -1.308997,2.664082,-1.071402,
        -1.308997,2.704512,1.488683,
        -1.308997,2.719179,-1.780644,
        -1.308997,2.724429,-0.348782,
        -1.308997,2.760637,2.781135,
        -1.308997,2.863997,0.549740,
        -1.308997,2.881233,-2.582574,
        -1.308997,2.979456,-1.009304,
        -1.308997,2.984978,1.954446,
        -1.047198,0.080730,-2.622036,
        -1.047198,0.203748,0.590255,
        -1.047198,0.290380,-0.796960,
        -1.047198,0.297065,2.160177,
        -1.047198,0.370812,-2.547742,
        -1.047198,0.379747,-1.713059,
        -1.047198,0.418037,1.360082,
        -1.047198,0.419640,2.974122,
        -1.047198,0.458961,-0.107937,
        -1.047198,0.496451,0.593859,
        -1.047198,0.571433,-1.188419,
        -1.047198,0.591903,2.010408,
        -1.047198,0.615174,-0.666022,
        -1.047198,0.618092,2.539118,
        -1.047198,0.619654,-2.216909,
        -1.047198,0.640596,-2.797608,
        -1.047198,0.674054,1.013609,
        -1.047198,0.685216,-1.712745,
        -1.047198,0.700570,1.559208,
        -1.047198,0.709924,0.218797,
        -1.047198,0.733948,3.021411,
        -1.047198,0.762244,-0.273349,
        -1.047198,0.831641,0.616218,
        -1.047198,0.850607,-1.335383,
        -1.047198,0.864993,2.278281,
        -1.047198,0.866395,-2.492918,
        -1.047198,0.866955,-0.931351,
        -1.047198,0.899350,-2.069907,
        -1.047198,0.906588,1.252948,
        -1.047198,0.910508,1.903534,
        -1.047198,0.914445,2.687814,
        -1.047198,0.929350,-2.910676,
        -1.047198,0.960688,-0.577532,
        -1.047198,0.961715,0.014085,
        -1.047198,0.992645,-1.723759,
        -1.047198,1.014500,0.906702,
        -1.047198,1.025536,1.576859,
        -1.047198,1.045978,0.343692,
        -1.047198,1.050534,3.042579,
        -1.047198,1.113094,-0.272241,
        -1.047198,1.128190,-2.288622,
        -1.047198,1.134497,-1.109631,
        -1.047198,1.134796,2.434290,
        -1.047198,1.135813,-2.658866,
        -1.047198,1.135834,-1.431087,
        -1.047198,1.168678,2.108595,
        -1.047198,1.184361,0.625864,
        -1.047198,1.194549,-0.801900,
        -1.047198,1.195355,1.200319,
        -1.047198,1.214854,-1.966664,
        -1.047198,1.218453,2.777284,
        -1.047198,1.242974,-2.977882,
        -1.047198,1.247695,1.809890,
        -1.047198,1.256302,0.062705,
        -1.047198,1.301460,-0.509251,
        -1.047198,1.310100,1.488307,
        -1.047198,1.325504,-1.671296,
        -1.047198,1.335669,0.897240,
        -1.047198,1.367556,0.358405,
        -1.047198,1.374601,-2.456031,
        -1.047198,1.391910,3.030322,
        -1.047198,1.413347,-1.339188,
        -1.047198,1.413430,2.550734,
        -1.047198,1.417623,-0.202254,
        -1.047198,1.433411,-1.035986,
        -1.047198,1.440580,2.260026,
        -1.047198,1.448123,-2.163292,
        -1.047198,1.450119,-2.739050,
        -1.047198,1.488617,1.159546,
        -1.047198,1.504028,0.636812,
        -1.047198,1.506530,1.988858,
        -1.047198,1.510910,-0.740472,
        -1.047198,1.545873,-1.877082,
        -1.047198,1.554776,0.100351,
        -1.047198,1.555207,1.708892,
        -1.047198,1.569311,-3.003978,
        -1.047198,1.574060,2.783565,
        -1.047198,1.599406,1.426070,
        -1.047198,1.606104,-0.451151,
        -1.047198,1.617114,-1.572037,
        -1.047198,1.658653,-2.409879,
        -1.047198,1.662363,0.915448,
        -1.047198,1.683268,0.374177,
        -1.047198,1.696674,-1.253233,
        -1.047198,1.699788,2.452890,
        -1.047198,1.721184,-0.153948,
        -1.047198,1.732874,-0.951685,
        -1.047198,1.736875,3.023612,
        -1.047198,1.741644,-2.695914,
        -1.047198,1.756530,-2.105156,
        -1.047198,1.766067,2.153843,
        -1.047198,1.809167,1.213774,
        -1.047198,1.812176,0.656819,
        -1.047198,1.824694,1.850876,
        -1.047198,1.836682,-0.676006,
        -1.047198,1.846342,-1.796954,
        -1.047198,1.869911,2.701025,
        -1.047198,1.870545,-2.976234,
        -1.047198,1.872854,1.543456,
        -1.047198,1.875281,0.107086,
        -1.047198,1.904882,-1.478089,
        -1.047198,1.909857,-0.381669,
        -1.047198,1.948085,-2.363031,
        -1.047198,1.977582,0.969896,
        -1.047198,1.982574,-1.115518,
        -1.047198,1.998705,2.370472,
        -1.047198,2.001437,0.381215,
        -1.047198,2.033014,2.977192,
        -1.047198,2.041297,-2.664597,
        -1.047198,2.063651,-2.044600,
        -1.047198,2.073035,2.027405,
        -1.047198,2.089241,-0.114193,
        -1.047198,2.104282,-0.813972,
        -1.047198,2.108018,1.311430,
        -1.047198,2.115563,0.685730,
        -1.047198,2.138223,1.678323,
        -1.047198,2.153910,-1.714959,
        -1.047198,2.164094,-2.991358,
        -1.047198,2.174483,2.653697,
        -1.047198,2.180315,-1.357379,
        -1.047198,2.187132,-0.479875,
        -1.047198,2.247462,0.191623,
        -1.047198,2.251330,-2.330739,
        -1.047198,2.295480,1.008277,
        -1.047198,2.299159,2.264663,
        -1.047198,2.343981,-1.040456,
        -1.047198,2.349581,-2.702517,
        -1.047198,2.356369,3.005869,
        -1.047198,2.379497,0.561089,
        -1.047198,2.393226,1.857173,
        -1.047198,2.396274,-1.959260,
        -1.047198,2.404280,-0.174044,
        -1.047198,2.407836,1.410646,
        -1.047198,2.455966,-0.641220,
        -1.047198,2.456751,-1.503343,
        -1.047198,2.478122,2.609777,
        -1.047198,2.579248,-2.364801,
        -1.047198,2.587733,0.232235,
        -1.047198,2.588535,-2.976617,
        -1.047198,2.592629,0.909713,
        -1.047198,2.637879,2.121598,
        -1.047198,2.664082,-1.071402,
        -1.047198,2.704512,1.488683,
        -1.047198,2.719179,-1.780644,
        -1.047198,2.724429,-0.348782,
        -1.047198,2.760637,2.781135,
        -1.047198,2.863997,0.549740,
        -1.047198,2.881233,-2.582574,
        -1.047198,2.979456,-1.009304,
        -1.047198,2.984978,1.954446,
        -0.785398,0.080730,-2.622036,
        -0.785398,0.203748,0.590255,
        -0.785398,0.290380,-0.796960,
        -0.785398,0.297065,2.160177,
        -0.785398,0.370812,-2.547742,
        -0.785398,0.379747,-1.713059,
        -0.785398,0.418037,1.360082,
        -0.785398,0.419640,2.974122,
        -0.785398,0.458961,-0.107937,
        -0.785398,0.496451,0.593859,
        -0.785398,0.571433,-1.188419,
        -0.785398,0.591903,2.010408,
        -0.785398,0.615174,-0.666022,
        -0.785398,0.618092,2.539118,
        -0.785398,0.619654,-2.216909,
        -0.785398,0.640596,-2.797608,
        -0.785398,0.674054,1.013609,
        -0.785398,0.685216,-1.712745,
        -0.785398,0.700570,1.559208,
        -0.785398,0.709924,0.218797,
        -0.785398,0.733948,3.021411,
        -0.785398,0.762244,-0.273349,
        -0.785398,0.831641,0.616218,
        -0.785398,0.850607,-1.335383,
        -0.785398,0.864993,2.278281,
        -0.785398,0.866395,-2.492918,
        -0.785398,0.866955,-0.931351,
        -0.785398,0.899350,-2.069907,
        -0.785398,0.906588,1.252948,
        -0.785398,0.910508,1.903534,
        -0.785398,0.914445,2.687814,
        -0.785398,0.929350,-2.910676,
        -0.785398,0.960688,-0.577532,
        -0.785398,0.961715,0.014085,
        -0.785398,0.992645,-1.723759,
        -0.785398,1.014500,0.906702,
        -0.785398,1.025536,1.576859,
        -0.785398,1.045978,0.343692,
        -0.785398,1.050534,3.042579,
        -0.785398,1.113094,-0.272241,
        -0.785398,1.128190,-2.288622,
        -0.785398,1.134497,-1.109631,
        -0.785398,1.134796,2.434290,
        -0.785398,1.135813,-2.658866,
        -0.785398,1.135834,-1.431087,
        -0.785398,1.168678,2.108595,
        -0.785398,1.184361,0.625864,
        -0.785398,1.194549,-0.801900,
        -0.785398,1.195355,1.200319,
        -0.785398,1.214854,-1.966664,
        -0.785398,1.218453,2.777284,
        -0.785398,1.242974,-2.977882,
        -0.785398,1.247695,1.809890,
        -0.785398,1.256302,0.062705,
        -0.785398,1.301460,-0.509251,
        -0.785398,1.310100,1.488307,
        -0.785398,1.325504,-1.671296,
        -0.785398,1.335669,0.897240,
        -0.785398,1.367556,0.358405,
        -0.785398,1.374601,-2.456031,
        -0.785398,1.391910,3.030322,
        -0.785398,1.413347,-1.339188,
        -0.785398,1.413430,2.550734,
        -0.785398,1.417623,-0.202254,
        -0.785398,1.433411,-1.035986,
        -0.785398,1.440580,2.260026,
        -0.785398,1.448123,-2.163292,
        -0.785398,1.450119,-2.739050,
        -0.785398,1.488617,1.159546,
        -0.785398,1.504028,0.636812,
        -0.785398,1.506530,1.988858,
        -0.785398,1.510910,-0.740472,
        -0.785398,1.545873,-1.877082,
        -0.785398,1.554776,0.100351,
        -0.785398,1.555207,1.708892,
        -0.785398,1.569311,-3.003978,
        -0.785398,1.574060,2.783565,
        -0.785398,1.599406,1.426070,
        -0.785398,1.606104,-0.451151,
        -0.785398,1.617114,-1.572037,
        -0.785398,1.658653,-2.409879,
        -0.785398,1.662363,0.915448,
        -0.785398,1.683268,0.374177,
        -0.785398,1.696674,-1.253233,
        -0.785398,1.699788,2.452890,
        -0.785398,1.721184,-0.153948,
        -0.785398,1.732874,-0.951685,
        -0.785398,1.736875,3.023612,
        -0.785398,1.741644,-2.695914,
        -0.785398,1.756530,-2.105156,
        -0.785398,1.766067,2.153843,
        -0.785398,1.809167,1.213774,
        -0.785398,1.812176,0.656819,
        -0.785398,1.824694,1.850876,
        -0.785398,1.836682,-0.676006,
        -0.785398,1.846342,-1.796954,
        -0.785398,1.869911,2.701025,
        -0.785398,1.870545,-2.976234,
        -0.785398,1.872854,1.543456,
        -0.785398,1.875281,0.107086,
        -0.785398,1.904882,-1.478089,
        -0.785398,1.909857,-0.381669,
        -0.785398,1.948085,-2.363031,
        -0.785398,1.977582,0.969896,
        -0.785398,1.982574,-1.115518,
        -0.785398,1.998705,2.370472,
        -0.785398,2.001437,0.381215,
        -0.785398,2.033014,2.977192,
        -0.785398,2.041297,-2.664597,
        -0.785398,2.063651,-2.044600,
        -0.785398,2.073035,2.027405,
        -0.785398,2.089241,-0.114193,
        -0.785398,2.104282,-0.813972,
        -0.785398,2.108018,1.311430,
        -0.785398,2.115563,0.685730,
        -0.785398,2.138223,1.678323,
        -0.785398,2.153910,-1.714959,
        -0.785398,2.164094,-2.991358,
        -0.785398,2.174483,2.653697,
        -0.785398,2.180315,-1.357379,
        -0.785398,2.187132,-0.479875,
        -0.785398,2.247462,0.191623,
        -0.785398,2.251330,-2.330739,
        -0.785398,2.295480,1.008277,
        -0.785398,2.299159,2.264663,
        -0.785398,2.343981,-1.040456,
        -0.785398,2.349581,-2.702517,
        -0.785398,2.356369,3.005869,
        -0.785398,2.379497,0.561089,
        -0.785398,2.393226,1.857173,
        -0.785398,2.396274,-1.959260,
        -0.785398,2.404280,-0.174044,
        -0.785398,2.407836,1.410646,
        -0.785398,2.455966,-0.641220,
        -0.785398,2.456751,-1.503343,
        -0.785398,2.478122,2.609777,
        -0.785398,2.579248,-2.364801,
        -0.785398,2.587733,0.232235,
        -0.785398,2.588535,-2.976617,
        -0.785398,2.592629,0.909713,
        -0.785398,2.637879,2.121598,
        -0.785398,2.664082,-1.071402,
        -0.785398,2.704512,1.488683,
        -0.785398,2.719179,-1.780644,
        -0.785398,2.724429,-0.348782,
        -0.785398,2.760637,2.781135,
        -0.785398,2.863997,0.549740,
        -0.785398,2.881233,-2.582574,
        -0.785398,2.979456,-1.009304,
        -0.785398,2.984978,1.954446,
        -0.523599,0.080730,-2.622036,
        -0.523599,0.203748,0.590255,
        -0.523599,0.290380,-0.796960,
        -0.523599,0.297065,2.160177,
        -0.523599,0.370812,-2.547742,
        -0.523599,0.379747,-1.713059,
        -0.523599,0.418037,1.360082,
        -0.523599,0.419640,2.974122,
        -0.523599,0.458961,-0.107937,
        -0.523599,0.496451,0.593859,
        -0.523599,0.571433,-1.188419,
        -0.523599,0.591903,2.010408,
        -0.523599,0.615174,-0.666022,
        -0.523599,0.618092,2.539118,
        -0.523599,0.619654,-2.216909,
        -0.523599,0.640596,-2.797608,
        -0.523599,0.674054,1.013609,
        -0.523599,0.685216,-1.712745,
        -0.523599,0.700570,1.559208,
        -0.523599,0.709924,0.218797,
        -0.523599,0.733948,3.021411,
        -0.523599,0.762244,-0.273349,
        -0.523599,0.831641,0.616218,
        -0.523599,0.850607,-1.335383,
        -0.523599,0.864993,2.278281,
        -0.523599,0.866395,-2.492918,
        -0.523599,0.866955,-0.931351,
        -0.523599,0.899350,-2.069907,
        -0.523599,0.906588,1.252948,
        -0.523599,0.910508,1.903534,
        -0.523599,0.914445,2.687814,
        -0.523599,0.929350,-2.910676,
        -0.523599,0.960688,-0.577532,
        -0.523599,0.961715,0.014085,
        -0.523599,0.992645,-1.723759,
        -0.523599,1.014500,0.906702,
        -0.523599,1.025536,1.576859,
        -0.523599,1.045978,0.343692,
        -0.523599,1.050534,3.042579,
        -0.523599,1.113094,-0.272241,
        -0.523599,1.128190,-2.288622,
        -0.523599,1.134497,-1.109631,
        -0.523599,1.134796,2.434290,
        -0.523599,1.135813,-2.658866,
        -0.523599,1.135834,-1.431087,
        -0.523599,1.168678,2.108595,
        -0.523599,1.184361,0.625864,
        -0.523599,1.194549,-0.801900,
        -0.523599,1.195355,1.200319,
        -0.523599,1.214854,-1.966664,
        -0.523599,1.218453,2.777284,
        -0.523599,1.242974,-2.977882,
        -0.523599,1.247695,1.809890,
        -0.523599,1.256302,0.062705,
        -0.523599,1.301460,-0.509251,
        -0.523599,1.310100,1.488307,
        -0.523599,1.325504,-1.671296,
        -0.523599,1.335669,0.897240,
        -0.523599,1.367556,0.358405,
        -0.523599,1.374601,-2.456031,
        -0.523599,1.391910,3.030322,
        -0.523599,1.413347,-1.339188,
        -0.523599,1.413430,2.550734,
        -0.523599,1.417623,-0.202254,
        -0.523599,1.433411,-1.035986,
        -0.523599,1.440580,2.260026,
        -0.523599,1.448123,-2.163292,
        -0.523599,1.450119,-2.739050,
        -0.523599,1.488617,1.159546,
        -0.523599,1.504028,0.636812,
        -0.523599,1.506530,1.988858,
        -0.523599,1.510910,-0.740472,
        -0.523599,1.545873,-1.877082,
        -0.523599,1.554776,0.100351,
        -0.523599,1.555207,1.708892,
        -0.523599,1.569311,-3.003978,
        -0.523599,1.574060,2.783565,
        -0.523599,1.599406,1.426070,
        -0.523599,1.606104,-0.451151,
        -0.523599,1.617114,-1.572037,
        -0.523599,1.658653,-2.409879,
        -0.523599,1.662363,0.915448,
        -0.523599,1.683268,0.374177,
        -0.523599,1.696674,-1.253233,
        -0.523599,1.699788,2.452890,
        -0.523599,1.721184,-0.153948,
        -0.523599,1.732874,-0.951685,
        -0.523599,1.736875,3.023612,
        -0.523599,1.741644,-2.695914,
        -0.523599,1.756530,-2.105156,
        -0.523599,1.766067,2.153843,
        -0.523599,1.809167,1.213774,
        -0.523599,1.812176,0.656819,
        -0.523599,1.824694,1.850876,
        -0.523599,1.836682,-0.676006,
        -0.523599,1.846342,-1.796954,
        -0.523599,1.869911,2.701025,
        -0.523599,1.870545,-2.976234,
        -0.523599,1.872854,1.543456,
        -0.523599,1.875281,0.107086,
        -0.523599,1.904882,-1.478089,
        -0.523599,1.909857,-0.381669,
        -0.523599,1.948085,-2.363031,
        -0.523599,1.977582,0.969896,
        -0.523599,1.982574,-1.115518,
        -0.523599,1.998705,2.370472,
        -0.523599,2.001437,0.381215,
        -0.523599,2.033014,2.977192,
        -0.523599,2.041297,-2.664597,
        -0.523599,2.063651,-2.044600,
        -0.523599,2.073035,2.027405,
        -0.523599,2.089241,-0.114193,
        -0.523599,2.104282,-0.813972,
        -0.523599,2.108018,1.311430,
        -0.523599,2.115563,0.685730,
        -0.523599,2.138223,1.678323,
        -0.523599,2.153910,-1.714959,
        -0.523599,2.164094,-2.991358,
        -0.523599,2.174483,2.653697,
        -0.523599,2.180315,-1.357379,
        -0.523599,2.187132,-0.479875,
        -0.523599,2.247462,0.191623,
        -0.523599,2.251330,-2.330739,
        -0.523599,2.295480,1.008277,
        -0.523599,2.299159,2.264663,
        -0.523599,2.343981,-1.040456,
        -0.523599,2.349581,-2.702517,
        -0.523599,2.356369,3.005869,
        -0.523599,2.379497,0.561089,
        -0.523599,2.393226,1.857173,
        -0.523599,2.396274,-1.959260,
        -0.523599,2.404280,-0.174044,
        -0.523599,2.407836,1.410646,
        -0.523599,2.455966,-0.641220,
        -0.523599,2.456751,-1.503343,
        -0.523599,2.478122,2.609777,
        -0.523599,2.579248,-2.364801,
        -0.523599,2.587733,0.232235,
        -0.523599,2.588535,-2.976617,
        -0.523599,2.592629,0.909713,
        -0.523599,2.637879,2.121598,
        -0.523599,2.664082,-1.071402,
        -0.523599,2.704512,1.488683,
        -0.523599,2.719179,-1.780644,
        -0.523599,2.724429,-0.348782,
        -0.523599,2.760637,2.781135,
        -0.523599,2.863997,0.549740,
        -0.523599,2.881233,-2.582574,
        -0.523599,2.979456,-1.009304,
        -0.523599,2.984978,1.954446,
        -0.261799,0.080730,-2.622036,
        -0.261799,0.203748,0.590255,
        -0.261799,0.290380,-0.796960,
        -0.261799,0.297065,2.160177,
        -0.261799,0.370812,-2.547742,
        -0.261799,0.379747,-1.713059,
        -0.261799,0.418037,1.360082,
        -0.261799,0.419640,2.974122,
        -0.261799,0.458961,-0.107937,
        -0.261799,0.496451,0.593859,
        -0.261799,0.571433,-1.188419,
        -0.261799,0.591903,2.010408,
        -0.261799,0.615174,-0.666022,
        -0.261799,0.618092,2.539118,
        -0.261799,0.619654,-2.216909,
        -0.261799,0.640596,-2.797608,
        -0.261799,0.674054,1.013609,
        -0.261799,0.685216,-1.712745,
        -0.261799,0.700570,1.559208,
        -0.261799,0.709924,0.218797,
        -0.261799,0.733948,3.021411,
        -0.261799,0.762244,-0.273349,
        -0.261799,0.831641,0.616218,
        -0.261799,0.850607,-1.335383,
        -0.261799,0.864993,2.278281,
        -0.261799,0.866395,-2.492918,
        -0.261799,0.866955,-0.931351,
        -0.261799,0.899350,-2.069907,
        -0.261799,0.906588,1.252948,
        -0.261799,0.910508,1.903534,
        -0.261799,0.914445,2.687814,
        -0.261799,0.929350,-2.910676,
        -0.261799,0.960688,-0.577532,
        -0.261799,0.961715,0.014085,
        -0.261799,0.992645,-1.723759,
        -0.261799,1.014500,0.906702,
        -0.261799,1.025536,1.576859,
        -0.261799,1.045978,0.343692,
        -0.261799,1.050534,3.042579,
        -0.261799,1.113094,-0.272241,
        -0.261799,1.128190,-2.288622,
        -0.261799,1.134497,-1.109631,
        -0.261799,1.134796,2.434290,
        -0.261799,1.135813,-2.658866,
        -0.261799,1.135834,-1.431087,
        -0.261799,1.168678,2.108595,
        -0.261799,1.184361,0.625864,
        -0.261799,1.194549,-0.801900,
        -0.261799,1.195355,1.200319,
        -0.261799,1.214854,-1.966664,
        -0.261799,1.218453,2.777284,
        -0.261799,1.242974,-2.977882,
        -0.261799,1.247695,1.809890,
        -0.261799,1.256302,0.062705,
        -0.261799,1.301460,-0.509251,
        -0.261799,1.310100,1.488307,
        -0.261799,1.325504,-1.671296,
        -0.261799,1.335669,0.897240,
        -0.261799,1.367556,0.358405,
        -0.261799,1.374601,-2.456031,
        -0.261799,1.391910,3.030322,
        -0.261799,1.413347,-1.339188,
        -0.261799,1.413430,2.550734,
        -0.261799,1.417623,-0.202254,
        -0.261799,1.433411,-1.035986,
        -0.261799,1.440580,2.260026,
        -0.261799,1.448123,-2.163292,
        -0.261799,1.450119,-2.739050,
        -0.261799,1.488617,1.159546,
        -0.261799,1.504028,0.636812,
        -0.261799,1.506530,1.988858,
        -0.261799,1.510910,-0.740472,
        -0.261799,1.545873,-1.877082,
        -0.261799,1.554776,0.100351,
        -0.261799,1.555207,1.708892,
        -0.261799,1.569311,-3.003978,
        -0.261799,1.574060,2.783565,
        -0.261799,1.599406,1.426070,
        -0.261799,1.606104,-0.451151,
        -0.261799,1.617114,-1.572037,
        -0.261799,1.658653,-2.409879,
        -0.261799,1.662363,0.915448,
        -0.261799,1.683268,0.374177,
        -0.261799,1.696674,-1.253233,
        -0.261799,1.699788,2.452890,
        -0.261799,1.721184,-0.153948,
        -0.261799,1.732874,-0.951685,
        -0.261799,1.736875,3.023612,
        -0.261799,1.741644,-2.695914,
        -0.261799,1.756530,-2.105156,
        -0.261799,1.766067,2.153843,
        -0.261799,1.809167,1.213774,
        -0.261799,1.812176,0.656819,
        -0.261799,1.824694,1.850876,
        -0.261799,1.836682,-0.676006,
        -0.261799,1.846342,-1.796954,
        -0.261799,1.869911,2.701025,
        -0.261799,1.870545,-2.976234,
        -0.261799,1.872854,1.543456,
        -0.261799,1.875281,0.107086,
        -0.261799,1.904882,-1.478089,
        -0.261799,1.909857,-0.381669,
        -0.261799,1.948085,-2.363031,
        -0.261799,1.977582,0.969896,
        -0.261799,1.982574,-1.115518,
        -0.261799,1.998705,2.370472,
        -0.261799,2.001437,0.381215,
        -0.261799,2.033014,2.977192,
        -0.261799,2.041297,-2.664597,
        -0.261799,2.063651,-2.044600,
        -0.261799,2.073035,2.027405,
        -0.261799,2.089241,-0.114193,
        -0.261799,2.104282,-0.813972,
        -0.261799,2.108018,1.311430,
        -0.261799,2.115563,0.685730,
        -0.261799,2.138223,1.678323,
        -0.261799,2.153910,-1.714959,
        -0.261799,2.164094,-2.991358,
        -0.261799,2.174483,2.653697,
        -0.261799,2.180315,-1.357379,
        -0.261799,2.187132,-0.479875,
        -0.261799,2.247462,0.191623,
        -0.261799,2.251330,-2.330739,
        -0.261799,2.295480,1.008277,
        -0.261799,2.299159,2.264663,
        -0.261799,2.343981,-1.040456,
        -0.261799,2.349581,-2.702517,
        -0.261799,2.356369,3.005869,
        -0.261799,2.379497,0.561089,
        -0.261799,2.393226,1.857173,
        -0.261799,2.396274,-1.959260,
        -0.261799,2.404280,-0.174044,
        -0.261799,2.407836,1.410646,
        -0.261799,2.455966,-0.641220,
        -0.261799,2.456751,-1.503343,
        -0.261799,2.478122,2.609777,
        -0.261799,2.579248,-2.364801,
        -0.261799,2.587733,0.232235,
        -0.261799,2.588535,-2.976617,
        -0.261799,2.592629,0.909713,
        -0.261799,2.637879,2.121598,
        -0.261799,2.664082,-1.071402,
        -0.261799,2.704512,1.488683,
        -0.261799,2.719179,-1.780644,
        -0.261799,2.724429,-0.348782,
        -0.261799,2.760637,2.781135,
        -0.261799,2.863997,0.549740,
        -0.261799,2.881233,-2.582574,
        -0.261799,2.979456,-1.009304,
        -0.261799,2.984978,1.954446,
        0.000000,0.080730,-2.622036,
        0.000000,0.203748,0.590255,
        0.000000,0.290380,-0.796960,
        0.000000,0.297065,2.160177,
        0.000000,0.370812,-2.547742,
        0.000000,0.379747,-1.713059,
        0.000000,0.418037,1.360082,
        0.000000,0.419640,2.974122,
        0.000000,0.458961,-0.107937,
        0.000000,0.496451,0.593859,
        0.000000,0.571433,-1.188419,
        0.000000,0.591903,2.010408,
        0.000000,0.615174,-0.666022,
        0.000000,0.618092,2.539118,
        0.000000,0.619654,-2.216909,
        0.000000,0.640596,-2.797608,
        0.000000,0.674054,1.013609,
        0.000000,0.685216,-1.712745,
        0.000000,0.700570,1.559208,
        0.000000,0.709924,0.218797,
        0.000000,0.733948,3.021411,
        0.000000,0.762244,-0.273349,
        0.000000,0.831641,0.616218,
        0.000000,0.850607,-1.335383,
        0.000000,0.864993,2.278281,
        0.000000,0.866395,-2.492918,
        0.000000,0.866955,-0.931351,
        0.000000,0.899350,-2.069907,
        0.000000,0.906588,1.252948,
        0.000000,0.910508,1.903534,
        0.000000,0.914445,2.687814,
        0.000000,0.929350,-2.910676,
        0.000000,0.960688,-0.577532,
        0.000000,0.961715,0.014085,
        0.000000,0.992645,-1.723759,
        0.000000,1.014500,0.906702,
        0.000000,1.025536,1.576859,
        0.000000,1.045978,0.343692,
        0.000000,1.050534,3.042579,
        0.000000,1.113094,-0.272241,
        0.000000,1.128190,-2.288622,
        0.000000,1.134497,-1.109631,
        0.000000,1.134796,2.434290,
        0.000000,1.135813,-2.658866,
        0.000000,1.135834,-1.431087,
        0.000000,1.168678,2.108595,
        0.000000,1.184361,0.625864,
        0.000000,1.194549,-0.801900,
        0.000000,1.195355,1.200319,
        0.000000,1.214854,-1.966664,
        0.000000,1.218453,2.777284,
        0.000000,1.242974,-2.977882,
        0.000000,1.247695,1.809890,
        0.000000,1.256302,0.062705,
        0.000000,1.301460,-0.509251,
        0.000000,1.310100,1.488307,
        0.000000,1.325504,-1.671296,
        0.000000,1.335669,0.897240,
        0.000000,1.367556,0.358405,
        0.000000,1.374601,-2.456031,
        0.000000,1.391910,3.030322,
        0.000000,1.413347,-1.339188,
        0.000000,1.413430,2.550734,
        0.000000,1.417623,-0.202254,
        0.000000,1.433411,-1.035986,
        0.000000,1.440580,2.260026,
        0.000000,1.448123,-2.163292,
        0.000000,1.450119,-2.739050,
        0.000000,1.488617,1.159546,
        0.000000,1.504028,0.636812,
        0.000000,1.506530,1.988858,
        0.000000,1.510910,-0.740472,
        0.000000,1.545873,-1.877082,
        0.000000,1.554776,0.100351,
        0.000000,1.555207,1.708892,
        0.000000,1.569311,-3.003978,
        0.000000,1.574060,2.783565,
        0.000000,1.599406,1.426070,
        0.000000,1.606104,-0.451151,
        0.000000,1.617114,-1.572037,
        0.000000,1.658653,-2.409879,
        0.000000,1.662363,0.915448,
        0.000000,1.683268,0.374177,
        0.000000,1.696674,-1.253233,
        0.000000,1.699788,2.452890,
        0.000000,1.721184,-0.153948,
        0.000000,1.732874,-0.951685,
        0.000000,1.736875,3.023612,
        0.000000,1.741644,-2.695914,
        0.000000,1.756530,-2.105156,
        0.000000,1.766067,2.153843,
        0.000000,1.809167,1.213774,
        0.000000,1.812176,0.656819,
        0.000000,1.824694,1.850876,
        0.000000,1.836682,-0.676006,
        0.000000,1.846342,-1.796954,
        0.000000,1.869911,2.701025,
        0.000000,1.870545,-2.976234,
        0.000000,1.872854,1.543456,
        0.000000,1.875281,0.107086,
        0.000000,1.904882,-1.478089,
        0.000000,1.909857,-0.381669,
        0.000000,1.948085,-2.363031,
        0.000000,1.977582,0.969896,
        0.000000,1.982574,-1.115518,
        0.000000,1.998705,2.370472,
        0.000000,2.001437,0.381215,
        0.000000,2.033014,2.977192,
        0.000000,2.041297,-2.664597,
        0.000000,2.063651,-2.044600,
        0.000000,2.073035,2.027405,
        0.000000,2.089241,-0.114193,
        0.000000,2.104282,-0.813972,
        0.000000,2.108018,1.311430,
        0.000000,2.115563,0.685730,
        0.000000,2.138223,1.678323,
        0.000000,2.153910,-1.714959,
        0.000000,2.164094,-2.991358,
        0.000000,2.174483,2.653697,
        0.000000,2.180315,-1.357379,
        0.000000,2.187132,-0.479875,
        0.000000,2.247462,0.191623,
        0.000000,2.251330,-2.330739,
        0.000000,2.295480,1.008277,
        0.000000,2.299159,2.264663,
        0.000000,2.343981,-1.040456,
        0.000000,2.349581,-2.702517,
        0.000000,2.356369,3.005869,
        0.000000,2.379497,0.561089,
        0.000000,2.393226,1.857173,
        0.000000,2.396274,-1.959260,
        0.000000,2.404280,-0.174044,
        0.000000,2.407836,1.410646,
        0.000000,2.455966,-0.641220,
        0.000000,2.456751,-1.503343,
        0.000000,2.478122,2.609777,
        0.000000,2.579248,-2.364801,
        0.000000,2.587733,0.232235,
        0.000000,2.588535,-2.976617,
        0.000000,2.592629,0.909713,
        0.000000,2.637879,2.121598,
        0.000000,2.664082,-1.071402,
        0.000000,2.704512,1.488683,
        0.000000,2.719179,-1.780644,
        0.000000,2.724429,-0.348782,
        0.000000,2.760637,2.781135,
        0.000000,2.863997,0.549740,
        0.000000,2.881233,-2.582574,
        0.000000,2.979456,-1.009304,
        0.000000,2.984978,1.954446,
        0.261799,0.080730,-2.622036,
        0.261799,0.203748,0.590255,
        0.261799,0.290380,-0.796960,
        0.261799,0.297065,2.160177,
        0.261799,0.370812,-2.547742,
        0.261799,0.379747,-1.713059,
        0.261799,0.418037,1.360082,
        0.261799,0.419640,2.974122,
        0.261799,0.458961,-0.107937,
        0.261799,0.496451,0.593859,
        0.261799,0.571433,-1.188419,
        0.261799,0.591903,2.010408,
        0.261799,0.615174,-0.666022,
        0.261799,0.618092,2.539118,
        0.261799,0.619654,-2.216909,
        0.261799,0.640596,-2.797608,
        0.261799,0.674054,1.013609,
        0.261799,0.685216,-1.712745,
        0.261799,0.700570,1.559208,
        0.261799,0.709924,0.218797,
        0.261799,0.733948,3.021411,
        0.261799,0.762244,-0.273349,
        0.261799,0.831641,0.616218,
        0.261799,0.850607,-1.335383,
        0.261799,0.864993,2.278281,
        0.261799,0.866395,-2.492918,
        0.261799,0.866955,-0.931351,
        0.261799,0.899350,-2.069907,
        0.261799,0.906588,1.252948,
        0.261799,0.910508,1.903534,
        0.261799,0.914445,2.687814,
        0.261799,0.929350,-2.910676,
        0.261799,0.960688,-0.577532,
        0.261799,0.961715,0.014085,
        0.261799,0.992645,-1.723759,
        0.261799,1.014500,0.906702,
        0.261799,1.025536,1.576859,
        0.261799,1.045978,0.343692,
        0.261799,1.050534,3.042579,
        0.261799,1.113094,-0.272241,
        0.261799,1.128190,-2.288622,
        0.261799,1.134497,-1.109631,
        0.261799,1.134796,2.434290,
        0.261799,1.135813,-2.658866,
        0.261799,1.135834,-1.431087,
        0.261799,1.168678,2.108595,
        0.261799,1.184361,0.625864,
        0.261799,1.194549,-0.801900,
        0.261799,1.195355,1.200319,
        0.261799,1.214854,-1.966664,
        0.261799,1.218453,2.777284,
        0.261799,1.242974,-2.977882,
        0.261799,1.247695,1.809890,
        0.261799,1.256302,0.062705,
        0.261799,1.301460,-0.509251,
        0.261799,1.310100,1.488307,
        0.261799,1.325504,-1.671296,
        0.261799,1.335669,0.897240,
        0.261799,1.367556,0.358405,
        0.261799,1.374601,-2.456031,
        0.261799,1.391910,3.030322,
        0.261799,1.413347,-1.339188,
        0.261799,1.413430,2.550734,
        0.261799,1.417623,-0.202254,
        0.261799,1.433411,-1.035986,
        0.261799,1.440580,2.260026,
        0.261799,1.448123,-2.163292,
        0.261799,1.450119,-2.739050,
        0.261799,1.488617,1.159546,
        0.261799,1.504028,0.636812,
        0.261799,1.506530,1.988858,
        0.261799,1.510910,-0.740472,
        0.261799,1.545873,-1.877082,
        0.261799,1.554776,0.100351,
        0.261799,1.555207,1.708892,
        0.261799,1.569311,-3.003978,
        0.261799,1.574060,2.783565,
        0.261799,1.599406,1.426070,
        0.261799,1.606104,-0.451151,
        0.261799,1.617114,-1.572037,
        0.261799,1.658653,-2.409879,
        0.261799,1.662363,0.915448,
        0.261799,1.683268,0.374177,
        0.261799,1.696674,-1.253233,
        0.261799,1.699788,2.452890,
        0.261799,1.721184,-0.153948,
        0.261799,1.732874,-0.951685,
        0.261799,1.736875,3.023612,
        0.261799,1.741644,-2.695914,
        0.261799,1.756530,-2.105156,
        0.261799,1.766067,2.153843,
        0.261799,1.809167,1.213774,
        0.261799,1.812176,0.656819,
        0.261799,1.824694,1.850876,
        0.261799,1.836682,-0.676006,
        0.261799,1.846342,-1.796954,
        0.261799,1.869911,2.701025,
        0.261799,1.870545,-2.976234,
        0.261799,1.872854,1.543456,
        0.261799,1.875281,0.107086,
        0.261799,1.904882,-1.478089,
        0.261799,1.909857,-0.381669,
        0.261799,1.948085,-2.363031,
        0.261799,1.977582,0.969896,
        0.261799,1.982574,-1.115518,
        0.261799,1.998705,2.370472,
        0.261799,2.001437,0.381215,
        0.261799,2.033014,2.977192,
        0.261799,2.041297,-2.664597,
        0.261799,2.063651,-2.044600,
        0.261799,2.073035,2.027405,
        0.261799,2.089241,-0.114193,
        0.261799,2.104282,-0.813972,
        0.261799,2.108018,1.311430,
        0.261799,2.115563,0.685730,
        0.261799,2.138223,1.678323,
        0.261799,2.153910,-1.714959,
        0.261799,2.164094,-2.991358,
        0.261799,2.174483,2.653697,
        0.261799,2.180315,-1.357379,
        0.261799,2.187132,-0.479875,
        0.261799,2.247462,0.191623,
        0.261799,2.251330,-2.330739,
        0.261799,2.295480,1.008277,
        0.261799,2.299159,2.264663,
        0.261799,2.343981,-1.040456,
        0.261799,2.349581,-2.702517,
        0.261799,2.356369,3.005869,
        0.261799,2.379497,0.561089,
        0.261799,2.393226,1.857173,
        0.261799,2.396274,-1.959260,
        0.261799,2.404280,-0.174044,
        0.261799,2.407836,1.410646,
        0.261799,2.455966,-0.641220,
        0.261799,2.456751,-1.503343,
        0.261799,2.478122,2.609777,
        0.261799,2.579248,-2.364801,
        0.261799,2.587733,0.232235,
        0.261799,2.588535,-2.976617,
        0.261799,2.592629,0.909713,
        0.261799,2.637879,2.121598,
        0.261799,2.664082,-1.071402,
        0.261799,2.704512,1.488683,
        0.261799,2.719179,-1.780644,
        0.261799,2.724429,-0.348782,
        0.261799,2.760637,2.781135,
        0.261799,2.863997,0.549740,
        0.261799,2.881233,-2.582574,
        0.261799,2.979456,-1.009304,
        0.261799,2.984978,1.954446,
        0.523599,0.080730,-2.622036,
        0.523599,0.203748,0.590255,
        0.523599,0.290380,-0.796960,
        0.523599,0.297065,2.160177,
        0.523599,0.370812,-2.547742,
        0.523599,0.379747,-1.713059,
        0.523599,0.418037,1.360082,
        0.523599,0.419640,2.974122,
        0.523599,0.458961,-0.107937,
        0.523599,0.496451,0.593859,
        0.523599,0.571433,-1.188419,
        0.523599,0.591903,2.010408,
        0.523599,0.615174,-0.666022,
        0.523599,0.618092,2.539118,
        0.523599,0.619654,-2.216909,
        0.523599,0.640596,-2.797608,
        0.523599,0.674054,1.013609,
        0.523599,0.685216,-1.712745,
        0.523599,0.700570,1.559208,
        0.523599,0.709924,0.218797,
        0.523599,0.733948,3.021411,
        0.523599,0.762244,-0.273349,
        0.523599,0.831641,0.616218,
        0.523599,0.850607,-1.335383,
        0.523599,0.864993,2.278281,
        0.523599,0.866395,-2.492918,
        0.523599,0.866955,-0.931351,
        0.523599,0.899350,-2.069907,
        0.523599,0.906588,1.252948,
        0.523599,0.910508,1.903534,
        0.523599,0.914445,2.687814,
        0.523599,0.929350,-2.910676,
        0.523599,0.960688,-0.577532,
        0.523599,0.961715,0.014085,
        0.523599,0.992645,-1.723759,
        0.523599,1.014500,0.906702,
        0.523599,1.025536,1.576859,
        0.523599,1.045978,0.343692,
        0.523599,1.050534,3.042579,
        0.523599,1.113094,-0.272241,
        0.523599,1.128190,-2.288622,
        0.523599,1.134497,-1.109631,
        0.523599,1.134796,2.434290,
        0.523599,1.135813,-2.658866,
        0.523599,1.135834,-1.431087,
        0.523599,1.168678,2.108595,
        0.523599,1.184361,0.625864,
        0.523599,1.194549,-0.801900,
        0.523599,1.195355,1.200319,
        0.523599,1.214854,-1.966664,
        0.523599,1.218453,2.777284,
        0.523599,1.242974,-2.977882,
        0.523599,1.247695,1.809890,
        0.523599,1.256302,0.062705,
        0.523599,1.301460,-0.509251,
        0.523599,1.310100,1.488307,
        0.523599,1.325504,-1.671296,
        0.523599,1.335669,0.897240,
        0.523599,1.367556,0.358405,
        0.523599,1.374601,-2.456031,
        0.523599,1.391910,3.030322,
        0.523599,1.413347,-1.339188,
        0.523599,1.413430,2.550734,
        0.523599,1.417623,-0.202254,
        0.523599,1.433411,-1.035986,
        0.523599,1.440580,2.260026,
        0.523599,1.448123,-2.163292,
        0.523599,1.450119,-2.739050,
        0.523599,1.488617,1.159546,
        0.523599,1.504028,0.636812,
        0.523599,1.506530,1.988858,
        0.523599,1.510910,-0.740472,
        0.523599,1.545873,-1.877082,
        0.523599,1.554776,0.100351,
        0.523599,1.555207,1.708892,
        0.523599,1.569311,-3.003978,
        0.523599,1.574060,2.783565,
        0.523599,1.599406,1.426070,
        0.523599,1.606104,-0.451151,
        0.523599,1.617114,-1.572037,
        0.523599,1.658653,-2.409879,
        0.523599,1.662363,0.915448,
        0.523599,1.683268,0.374177,
        0.523599,1.696674,-1.253233,
        0.523599,1.699788,2.452890,
        0.523599,1.721184,-0.153948,
        0.523599,1.732874,-0.951685,
        0.523599,1.736875,3.023612,
        0.523599,1.741644,-2.695914,
        0.523599,1.756530,-2.105156,
        0.523599,1.766067,2.153843,
        0.523599,1.809167,1.213774,
        0.523599,1.812176,0.656819,
        0.523599,1.824694,1.850876,
        0.523599,1.836682,-0.676006,
        0.523599,1.846342,-1.796954,
        0.523599,1.869911,2.701025,
        0.523599,1.870545,-2.976234,
        0.523599,1.872854,1.543456,
        0.523599,1.875281,0.107086,
        0.523599,1.904882,-1.478089,
        0.523599,1.909857,-0.381669,
        0.523599,1.948085,-2.363031,
        0.523599,1.977582,0.969896,
        0.523599,1.982574,-1.115518,
        0.523599,1.998705,2.370472,
        0.523599,2.001437,0.381215,
        0.523599,2.033014,2.977192,
        0.523599,2.041297,-2.664597,
        0.523599,2.063651,-2.044600,
        0.523599,2.073035,2.027405,
        0.523599,2.089241,-0.114193,
        0.523599,2.104282,-0.813972,
        0.523599,2.108018,1.311430,
        0.523599,2.115563,0.685730,
        0.523599,2.138223,1.678323,
        0.523599,2.153910,-1.714959,
        0.523599,2.164094,-2.991358,
        0.523599,2.174483,2.653697,
        0.523599,2.180315,-1.357379,
        0.523599,2.187132,-0.479875,
        0.523599,2.247462,0.191623,
        0.523599,2.251330,-2.330739,
        0.523599,2.295480,1.008277,
        0.523599,2.299159,2.264663,
        0.523599,2.343981,-1.040456,
        0.523599,2.349581,-2.702517,
        0.523599,2.356369,3.005869,
        0.523599,2.379497,0.561089,
        0.523599,2.393226,1.857173,
        0.523599,2.396274,-1.959260,
        0.523599,2.404280,-0.174044,
        0.523599,2.407836,1.410646,
        0.523599,2.455966,-0.641220,
        0.523599,2.456751,-1.503343,
        0.523599,2.478122,2.609777,
        0.523599,2.579248,-2.364801,
        0.523599,2.587733,0.232235,
        0.523599,2.588535,-2.976617,
        0.523599,2.592629,0.909713,
        0.523599,2.637879,2.121598,
        0.523599,2.664082,-1.071402,
        0.523599,2.704512,1.488683,
        0.523599,2.719179,-1.780644,
        0.523599,2.724429,-0.348782,
        0.523599,2.760637,2.781135,
        0.523599,2.863997,0.549740,
        0.523599,2.881233,-2.582574,
        0.523599,2.979456,-1.009304,
        0.523599,2.984978,1.954446,
        0.785398,0.080730,-2.622036,
        0.785398,0.203748,0.590255,
        0.785398,0.290380,-0.796960,
        0.785398,0.297065,2.160177,
        0.785398,0.370812,-2.547742,
        0.785398,0.379747,-1.713059,
        0.785398,0.418037,1.360082,
        0.785398,0.419640,2.974122,
        0.785398,0.458961,-0.107937,
        0.785398,0.496451,0.593859,
        0.785398,0.571433,-1.188419,
        0.785398,0.591903,2.010408,
        0.785398,0.615174,-0.666022,
        0.785398,0.618092,2.539118,
        0.785398,0.619654,-2.216909,
        0.785398,0.640596,-2.797608,
        0.785398,0.674054,1.013609,
        0.785398,0.685216,-1.712745,
        0.785398,0.700570,1.559208,
        0.785398,0.709924,0.218797,
        0.785398,0.733948,3.021411,
        0.785398,0.762244,-0.273349,
        0.785398,0.831641,0.616218,
        0.785398,0.850607,-1.335383,
        0.785398,0.864993,2.278281,
        0.785398,0.866395,-2.492918,
        0.785398,0.866955,-0.931351,
        0.785398,0.899350,-2.069907,
        0.785398,0.906588,1.252948,
        0.785398,0.910508,1.903534,
        0.785398,0.914445,2.687814,
        0.785398,0.929350,-2.910676,
        0.785398,0.960688,-0.577532,
        0.785398,0.961715,0.014085,
        0.785398,0.992645,-1.723759,
        0.785398,1.014500,0.906702,
        0.785398,1.025536,1.576859,
        0.785398,1.045978,0.343692,
        0.785398,1.050534,3.042579,
        0.785398,1.113094,-0.272241,
        0.785398,1.128190,-2.288622,
        0.785398,1.134497,-1.109631,
        0.785398,1.134796,2.434290,
        0.785398,1.135813,-2.658866,
        0.785398,1.135834,-1.431087,
        0.785398,1.168678,2.108595,
        0.785398,1.184361,0.625864,
        0.785398,1.194549,-0.801900,
        0.785398,1.195355,1.200319,
        0.785398,1.214854,-1.966664,
        0.785398,1.218453,2.777284,
        0.785398,1.242974,-2.977882,
        0.785398,1.247695,1.809890,
        0.785398,1.256302,0.062705,
        0.785398,1.301460,-0.509251,
        0.785398,1.310100,1.488307,
        0.785398,1.325504,-1.671296,
        0.785398,1.335669,0.897240,
        0.785398,1.367556,0.358405,
        0.785398,1.374601,-2.456031,
        0.785398,1.391910,3.030322,
        0.785398,1.413347,-1.339188,
        0.785398,1.413430,2.550734,
        0.785398,1.417623,-0.202254,
        0.785398,1.433411,-1.035986,
        0.785398,1.440580,2.260026,
        0.785398,1.448123,-2.163292,
        0.785398,1.450119,-2.739050,
        0.785398,1.488617,1.159546,
        0.785398,1.504028,0.636812,
        0.785398,1.506530,1.988858,
        0.785398,1.510910,-0.740472,
        0.785398,1.545873,-1.877082,
        0.785398,1.554776,0.100351,
        0.785398,1.555207,1.708892,
        0.785398,1.569311,-3.003978,
        0.785398,1.574060,2.783565,
        0.785398,1.599406,1.426070,
        0.785398,1.606104,-0.451151,
        0.785398,1.617114,-1.572037,
        0.785398,1.658653,-2.409879,
        0.785398,1.662363,0.915448,
        0.785398,1.683268,0.374177,
        0.785398,1.696674,-1.253233,
        0.785398,1.699788,2.452890,
        0.785398,1.721184,-0.153948,
        0.785398,1.732874,-0.951685,
        0.785398,1.736875,3.023612,
        0.785398,1.741644,-2.695914,
        0.785398,1.756530,-2.105156,
        0.785398,1.766067,2.153843,
        0.785398,1.809167,1.213774,
        0.785398,1.812176,0.656819,
        0.785398,1.824694,1.850876,
        0.785398,1.836682,-0.676006,
        0.785398,1.846342,-1.796954,
        0.785398,1.869911,2.701025,
        0.785398,1.870545,-2.976234,
        0.785398,1.872854,1.543456,
        0.785398,1.875281,0.107086,
        0.785398,1.904882,-1.478089,
        0.785398,1.909857,-0.381669,
        0.785398,1.948085,-2.363031,
        0.785398,1.977582,0.969896,
        0.785398,1.982574,-1.115518,
        0.785398,1.998705,2.370472,
        0.785398,2.001437,0.381215,
        0.785398,2.033014,2.977192,
        0.785398,2.041297,-2.664597,
        0.785398,2.063651,-2.044600,
        0.785398,2.073035,2.027405,
        0.785398,2.089241,-0.114193,
        0.785398,2.104282,-0.813972,
        0.785398,2.108018,1.311430,
        0.785398,2.115563,0.685730,
        0.785398,2.138223,1.678323,
        0.785398,2.153910,-1.714959,
        0.785398,2.164094,-2.991358,
        0.785398,2.174483,2.653697,
        0.785398,2.180315,-1.357379,
        0.785398,2.187132,-0.479875,
        0.785398,2.247462,0.191623,
        0.785398,2.251330,-2.330739,
        0.785398,2.295480,1.008277,
        0.785398,2.299159,2.264663,
        0.785398,2.343981,-1.040456,
        0.785398,2.349581,-2.702517,
        0.785398,2.356369,3.005869,
        0.785398,2.379497,0.561089,
        0.785398,2.393226,1.857173,
        0.785398,2.396274,-1.959260,
        0.785398,2.404280,-0.174044,
        0.785398,2.407836,1.410646,
        0.785398,2.455966,-0.641220,
        0.785398,2.456751,-1.503343,
        0.785398,2.478122,2.609777,
        0.785398,2.579248,-2.364801,
        0.785398,2.587733,0.232235,
        0.785398,2.588535,-2.976617,
        0.785398,2.592629,0.909713,
        0.785398,2.637879,2.121598,
        0.785398,2.664082,-1.071402,
        0.785398,2.704512,1.488683,
        0.785398,2.719179,-1.780644,
        0.785398,2.724429,-0.348782,
        0.785398,2.760637,2.781135,
        0.785398,2.863997,0.549740,
        0.785398,2.881233,-2.582574,
        0.785398,2.979456,-1.009304,
        0.785398,2.984978,1.954446,
        1.047198,0.080730,-2.622036,
        1.047198,0.203748,0.590255,
        1.047198,0.290380,-0.796960,
        1.047198,0.297065,2.160177,
        1.047198,0.370812,-2.547742,
        1.047198,0.379747,-1.713059,
        1.047198,0.418037,1.360082,
        1.047198,0.419640,2.974122,
        1.047198,0.458961,-0.107937,
        1.047198,0.496451,0.593859,
        1.047198,0.571433,-1.188419,
        1.047198,0.591903,2.010408,
        1.047198,0.615174,-0.666022,
        1.047198,0.618092,2.539118,
        1.047198,0.619654,-2.216909,
        1.047198,0.640596,-2.797608,
        1.047198,0.674054,1.013609,
        1.047198,0.685216,-1.712745,
        1.047198,0.700570,1.559208,
        1.047198,0.709924,0.218797,
        1.047198,0.733948,3.021411,
        1.047198,0.762244,-0.273349,
        1.047198,0.831641,0.616218,
        1.047198,0.850607,-1.335383,
        1.047198,0.864993,2.278281,
        1.047198,0.866395,-2.492918,
        1.047198,0.866955,-0.931351,
        1.047198,0.899350,-2.069907,
        1.047198,0.906588,1.252948,
        1.047198,0.910508,1.903534,
        1.047198,0.914445,2.687814,
        1.047198,0.929350,-2.910676,
        1.047198,0.960688,-0.577532,
        1.047198,0.961715,0.014085,
        1.047198,0.992645,-1.723759,
        1.047198,1.014500,0.906702,
        1.047198,1.025536,1.576859,
        1.047198,1.045978,0.343692,
        1.047198,1.050534,3.042579,
        1.047198,1.113094,-0.272241,
        1.047198,1.128190,-2.288622,
        1.047198,1.134497,-1.109631,
        1.047198,1.134796,2.434290,
        1.047198,1.135813,-2.658866,
        1.047198,1.135834,-1.431087,
        1.047198,1.168678,2.108595,
        1.047198,1.184361,0.625864,
        1.047198,1.194549,-0.801900,
        1.047198,1.195355,1.200319,
        1.047198,1.214854,-1.966664,
        1.047198,1.218453,2.777284,
        1.047198,1.242974,-2.977882,
        1.047198,1.247695,1.809890,
        1.047198,1.256302,0.062705,
        1.047198,1.301460,-0.509251,
        1.047198,1.310100,1.488307,
        1.047198,1.325504,-1.671296,
        1.047198,1.335669,0.897240,
        1.047198,1.367556,0.358405,
        1.047198,1.374601,-2.456031,
        1.047198,1.391910,3.030322,
        1.047198,1.413347,-1.339188,
        1.047198,1.413430,2.550734,
        1.047198,1.417623,-0.202254,
        1.047198,1.433411,-1.035986,
        1.047198,1.440580,2.260026,
        1.047198,1.448123,-2.163292,
        1.047198,1.450119,-2.739050,
        1.047198,1.488617,1.159546,
        1.047198,1.504028,0.636812,
        1.047198,1.506530,1.988858,
        1.047198,1.510910,-0.740472,
        1.047198,1.545873,-1.877082,
        1.047198,1.554776,0.100351,
        1.047198,1.555207,1.708892,
        1.047198,1.569311,-3.003978,
        1.047198,1.574060,2.783565,
        1.047198,1.599406,1.426070,
        1.047198,1.606104,-0.451151,
        1.047198,1.617114,-1.572037,
        1.047198,1.658653,-2.409879,
        1.047198,1.662363,0.915448,
        1.047198,1.683268,0.374177,
        1.047198,1.696674,-1.253233,
        1.047198,1.699788,2.452890,
        1.047198,1.721184,-0.153948,
        1.047198,1.732874,-0.951685,
        1.047198,1.736875,3.023612,
        1.047198,1.741644,-2.695914,
        1.047198,1.756530,-2.105156,
        1.047198,1.766067,2.153843,
        1.047198,1.809167,1.213774,
        1.047198,1.812176,0.656819,
        1.047198,1.824694,1.850876,
        1.047198,1.836682,-0.676006,
        1.047198,1.846342,-1.796954,
        1.047198,1.869911,2.701025,
        1.047198,1.870545,-2.976234,
        1.047198,1.872854,1.543456,
        1.047198,1.875281,0.107086,
        1.047198,1.904882,-1.478089,
        1.047198,1.909857,-0.381669,
        1.047198,1.948085,-2.363031,
        1.047198,1.977582,0.969896,
        1.047198,1.982574,-1.115518,
        1.047198,1.998705,2.370472,
        1.047198,2.001437,0.381215,
        1.047198,2.033014,2.977192,
        1.047198,2.041297,-2.664597,
        1.047198,2.063651,-2.044600,
        1.047198,2.073035,2.027405,
        1.047198,2.089241,-0.114193,
        1.047198,2.104282,-0.813972,
        1.047198,2.108018,1.311430,
        1.047198,2.115563,0.685730,
        1.047198,2.138223,1.678323,
        1.047198,2.153910,-1.714959,
        1.047198,2.164094,-2.991358,
        1.047198,2.174483,2.653697,
        1.047198,2.180315,-1.357379,
        1.047198,2.187132,-0.479875,
        1.047198,2.247462,0.191623,
        1.047198,2.251330,-2.330739,
        1.047198,2.295480,1.008277,
        1.047198,2.299159,2.264663,
        1.047198,2.343981,-1.040456,
        1.047198,2.349581,-2.702517,
        1.047198,2.356369,3.005869,
        1.047198,2.379497,0.561089,
        1.047198,2.393226,1.857173,
        1.047198,2.396274,-1.959260,
        1.047198,2.404280,-0.174044,
        1.047198,2.407836,1.410646,
        1.047198,2.455966,-0.641220,
        1.047198,2.456751,-1.503343,
        1.047198,2.478122,2.609777,
        1.047198,2.579248,-2.364801,
        1.047198,2.587733,0.232235,
        1.047198,2.588535,-2.976617,
        1.047198,2.592629,0.909713,
        1.047198,2.637879,2.121598,
        1.047198,2.664082,-1.071402,
        1.047198,2.704512,1.488683,
        1.047198,2.719179,-1.780644,
        1.047198,2.724429,-0.348782,
        1.047198,2.760637,2.781135,
        1.047198,2.863997,0.549740,
        1.047198,2.881233,-2.582574,
        1.047198,2.979456,-1.009304,
        1.047198,2.984978,1.954446,
        1.308997,0.080730,-2.622036,
        1.308997,0.203748,0.590255,
        1.308997,0.290380,-0.796960,
        1.308997,0.297065,2.160177,
        1.308997,0.370812,-2.547742,
        1.308997,0.379747,-1.713059,
        1.308997,0.418037,1.360082,
        1.308997,0.419640,2.974122,
        1.308997,0.458961,-0.107937,
        1.308997,0.496451,0.593859,
        1.308997,0.571433,-1.188419,
        1.308997,0.591903,2.010408,
        1.308997,0.615174,-0.666022,
        1.308997,0.618092,2.539118,
        1.308997,0.619654,-2.216909,
        1.308997,0.640596,-2.797608,
        1.308997,0.674054,1.013609,
        1.308997,0.685216,-1.712745,
        1.308997,0.700570,1.559208,
        1.308997,0.709924,0.218797,
        1.308997,0.733948,3.021411,
        1.308997,0.762244,-0.273349,
        1.308997,0.831641,0.616218,
        1.308997,0.850607,-1.335383,
        1.308997,0.864993,2.278281,
        1.308997,0.866395,-2.492918,
        1.308997,0.866955,-0.931351,
        1.308997,0.899350,-2.069907,
        1.308997,0.906588,1.252948,
        1.308997,0.910508,1.903534,
        1.308997,0.914445,2.687814,
        1.308997,0.929350,-2.910676,
        1.308997,0.960688,-0.577532,
        1.308997,0.961715,0.014085,
        1.308997,0.992645,-1.723759,
        1.308997,1.014500,0.906702,
        1.308997,1.025536,1.576859,
        1.308997,1.045978,0.343692,
        1.308997,1.050534,3.042579,
        1.308997,1.113094,-0.272241,
        1.308997,1.128190,-2.288622,
        1.308997,1.134497,-1.109631,
        1.308997,1.134796,2.434290,
        1.308997,1.135813,-2.658866,
        1.308997,1.135834,-1.431087,
        1.308997,1.168678,2.108595,
        1.308997,1.184361,0.625864,
        1.308997,1.194549,-0.801900,
        1.308997,1.195355,1.200319,
        1.308997,1.214854,-1.966664,
        1.308997,1.218453,2.777284,
        1.308997,1.242974,-2.977882,
        1.308997,1.247695,1.809890,
        1.308997,1.256302,0.062705,
        1.308997,1.301460,-0.509251,
        1.308997,1.310100,1.488307,
        1.308997,1.325504,-1.671296,
        1.308997,1.335669,0.897240,
        1.308997,1.367556,0.358405,
        1.308997,1.374601,-2.456031,
        1.308997,1.391910,3.030322,
        1.308997,1.413347,-1.339188,
        1.308997,1.413430,2.550734,
        1.308997,1.417623,-0.202254,
        1.308997,1.433411,-1.035986,
        1.308997,1.440580,2.260026,
        1.308997,1.448123,-2.163292,
        1.308997,1.450119,-2.739050,
        1.308997,1.488617,1.159546,
        1.308997,1.504028,0.636812,
        1.308997,1.506530,1.988858,
        1.308997,1.510910,-0.740472,
        1.308997,1.545873,-1.877082,
        1.308997,1.554776,0.100351,
        1.308997,1.555207,1.708892,
        1.308997,1.569311,-3.003978,
        1.308997,1.574060,2.783565,
        1.308997,1.599406,1.426070,
        1.308997,1.606104,-0.451151,
        1.308997,1.617114,-1.572037,
        1.308997,1.658653,-2.409879,
        1.308997,1.662363,0.915448,
        1.308997,1.683268,0.374177,
        1.308997,1.696674,-1.253233,
        1.308997,1.699788,2.452890,
        1.308997,1.721184,-0.153948,
        1.308997,1.732874,-0.951685,
        1.308997,1.736875,3.023612,
        1.308997,1.741644,-2.695914,
        1.308997,1.756530,-2.105156,
        1.308997,1.766067,2.153843,
        1.308997,1.809167,1.213774,
        1.308997,1.812176,0.656819,
        1.308997,1.824694,1.850876,
        1.308997,1.836682,-0.676006,
        1.308997,1.846342,-1.796954,
        1.308997,1.869911,2.701025,
        1.308997,1.870545,-2.976234,
        1.308997,1.872854,1.543456,
        1.308997,1.875281,0.107086,
        1.308997,1.904882,-1.478089,
        1.308997,1.909857,-0.381669,
        1.308997,1.948085,-2.363031,
        1.308997,1.977582,0.969896,
        1.308997,1.982574,-1.115518,
        1.308997,1.998705,2.370472,
        1.308997,2.001437,0.381215,
        1.308997,2.033014,2.977192,
        1.308997,2.041297,-2.664597,
        1.308997,2.063651,-2.044600,
        1.308997,2.073035,2.027405,
        1.308997,2.089241,-0.114193,
        1.308997,2.104282,-0.813972,
        1.308997,2.108018,1.311430,
        1.308997,2.115563,0.685730,
        1.308997,2.138223,1.678323,
        1.308997,2.153910,-1.714959,
        1.308997,2.164094,-2.991358,
        1.308997,2.174483,2.653697,
        1.308997,2.180315,-1.357379,
        1.308997,2.187132,-0.479875,
        1.308997,2.247462,0.191623,
        1.308997,2.251330,-2.330739,
        1.308997,2.295480,1.008277,
        1.308997,2.299159,2.264663,
        1.308997,2.343981,-1.040456,
        1.308997,2.349581,-2.702517,
        1.308997,2.356369,3.005869,
        1.308997,2.379497,0.561089,
        1.308997,2.393226,1.857173,
        1.308997,2.396274,-1.959260,
        1.308997,2.404280,-0.174044,
        1.308997,2.407836,1.410646,
        1.308997,2.455966,-0.641220,
        1.308997,2.456751,-1.503343,
        1.308997,2.478122,2.609777,
        1.308997,2.579248,-2.364801,
        1.308997,2.587733,0.232235,
        1.308997,2.588535,-2.976617,
        1.308997,2.592629,0.909713,
        1.308997,2.637879,2.121598,
        1.308997,2.664082,-1.071402,
        1.308997,2.704512,1.488683,
        1.308997,2.719179,-1.780644,
        1.308997,2.724429,-0.348782,
        1.308997,2.760637,2.781135,
        1.308997,2.863997,0.549740,
        1.308997,2.881233,-2.582574,
        1.308997,2.979456,-1.009304,
        1.308997,2.984978,1.954446,
        1.570796,0.080730,-2.622036,
        1.570796,0.203748,0.590255,
        1.570796,0.290380,-0.796960,
        1.570796,0.297065,2.160177,
        1.570796,0.370812,-2.547742,
        1.570796,0.379747,-1.713059,
        1.570796,0.418037,1.360082,
        1.570796,0.419640,2.974122,
        1.570796,0.458961,-0.107937,
        1.570796,0.496451,0.593859,
        1.570796,0.571433,-1.188419,
        1.570796,0.591903,2.010408,
        1.570796,0.615174,-0.666022,
        1.570796,0.618092,2.539118,
        1.570796,0.619654,-2.216909,
        1.570796,0.640596,-2.797608,
        1.570796,0.674054,1.013609,
        1.570796,0.685216,-1.712745,
        1.570796,0.700570,1.559208,
        1.570796,0.709924,0.218797,
        1.570796,0.733948,3.021411,
        1.570796,0.762244,-0.273349,
        1.570796,0.831641,0.616218,
        1.570796,0.850607,-1.335383,
        1.570796,0.864993,2.278281,
        1.570796,0.866395,-2.492918,
        1.570796,0.866955,-0.931351,
        1.570796,0.899350,-2.069907,
        1.570796,0.906588,1.252948,
        1.570796,0.910508,1.903534,
        1.570796,0.914445,2.687814,
        1.570796,0.929350,-2.910676,
        1.570796,0.960688,-0.577532,
        1.570796,0.961715,0.014085,
        1.570796,0.992645,-1.723759,
        1.570796,1.014500,0.906702,
        1.570796,1.025536,1.576859,
        1.570796,1.045978,0.343692,
        1.570796,1.050534,3.042579,
        1.570796,1.113094,-0.272241,
        1.570796,1.128190,-2.288622,
        1.570796,1.134497,-1.109631,
        1.570796,1.134796,2.434290,
        1.570796,1.135813,-2.658866,
        1.570796,1.135834,-1.431087,
        1.570796,1.168678,2.108595,
        1.570796,1.184361,0.625864,
        1.570796,1.194549,-0.801900,
        1.570796,1.195355,1.200319,
        1.570796,1.214854,-1.966664,
        1.570796,1.218453,2.777284,
        1.570796,1.242974,-2.977882,
        1.570796,1.247695,1.809890,
        1.570796,1.256302,0.062705,
        1.570796,1.301460,-0.509251,
        1.570796,1.310100,1.488307,
        1.570796,1.325504,-1.671296,
        1.570796,1.335669,0.897240,
        1.570796,1.367556,0.358405,
        1.570796,1.374601,-2.456031,
        1.570796,1.391910,3.030322,
        1.570796,1.413347,-1.339188,
        1.570796,1.413430,2.550734,
        1.570796,1.417623,-0.202254,
        1.570796,1.433411,-1.035986,
        1.570796,1.440580,2.260026,
        1.570796,1.448123,-2.163292,
        1.570796,1.450119,-2.739050,
        1.570796,1.488617,1.159546,
        1.570796,1.504028,0.636812,
        1.570796,1.506530,1.988858,
        1.570796,1.510910,-0.740472,
        1.570796,1.545873,-1.877082,
        1.570796,1.554776,0.100351,
        1.570796,1.555207,1.708892,
        1.570796,1.569311,-3.003978,
        1.570796,1.574060,2.783565,
        1.570796,1.599406,1.426070,
        1.570796,1.606104,-0.451151,
        1.570796,1.617114,-1.572037,
        1.570796,1.658653,-2.409879,
        1.570796,1.662363,0.915448,
        1.570796,1.683268,0.374177,
        1.570796,1.696674,-1.253233,
        1.570796,1.699788,2.452890,
        1.570796,1.721184,-0.153948,
        1.570796,1.732874,-0.951685,
        1.570796,1.736875,3.023612,
        1.570796,1.741644,-2.695914,
        1.570796,1.756530,-2.105156,
        1.570796,1.766067,2.153843,
        1.570796,1.809167,1.213774,
        1.570796,1.812176,0.656819,
        1.570796,1.824694,1.850876,
        1.570796,1.836682,-0.676006,
        1.570796,1.846342,-1.796954,
        1.570796,1.869911,2.701025,
        1.570796,1.870545,-2.976234,
        1.570796,1.872854,1.543456,
        1.570796,1.875281,0.107086,
        1.570796,1.904882,-1.478089,
        1.570796,1.909857,-0.381669,
        1.570796,1.948085,-2.363031,
        1.570796,1.977582,0.969896,
        1.570796,1.982574,-1.115518,
        1.570796,1.998705,2.370472,
        1.570796,2.001437,0.381215,
        1.570796,2.033014,2.977192,
        1.570796,2.041297,-2.664597,
        1.570796,2.063651,-2.044600,
        1.570796,2.073035,2.027405,
        1.570796,2.089241,-0.114193,
        1.570796,2.104282,-0.813972,
        1.570796,2.108018,1.311430,
        1.570796,2.115563,0.685730,
        1.570796,2.138223,1.678323,
        1.570796,2.153910,-1.714959,
        1.570796,2.164094,-2.991358,
        1.570796,2.174483,2.653697,
        1.570796,2.180315,-1.357379,
        1.570796,2.187132,-0.479875,
        1.570796,2.247462,0.191623,
        1.570796,2.251330,-2.330739,
        1.570796,2.295480,1.008277,
        1.570796,2.299159,2.264663,
        1.570796,2.343981,-1.040456,
        1.570796,2.349581,-2.702517,
        1.570796,2.356369,3.005869,
        1.570796,2.379497,0.561089,
        1.570796,2.393226,1.857173,
        1.570796,2.396274,-1.959260,
        1.570796,2.404280,-0.174044,
        1.570796,2.407836,1.410646,
        1.570796,2.455966,-0.641220,
        1.570796,2.456751,-1.503343,
        1.570796,2.478122,2.609777,
        1.570796,2.579248,-2.364801,
        1.570796,2.587733,0.232235,
        1.570796,2.588535,-2.976617,
        1.570796,2.592629,0.909713,
        1.570796,2.637879,2.121598,
        1.570796,2.664082,-1.071402,
        1.570796,2.704512,1.488683,
        1.570796,2.719179,-1.780644,
        1.570796,2.724429,-0.348782,
        1.570796,2.760637,2.781135,
        1.570796,2.863997,0.549740,
        1.570796,2.881233,-2.582574,
        1.570796,2.979456,-1.009304,
        1.570796,2.984978,1.954446,
        1.832596,0.080730,-2.622036,
        1.832596,0.203748,0.590255,
        1.832596,0.290380,-0.796960,
        1.832596,0.297065,2.160177,
        1.832596,0.370812,-2.547742,
        1.832596,0.379747,-1.713059,
        1.832596,0.418037,1.360082,
        1.832596,0.419640,2.974122,
        1.832596,0.458961,-0.107937,
        1.832596,0.496451,0.593859,
        1.832596,0.571433,-1.188419,
        1.832596,0.591903,2.010408,
        1.832596,0.615174,-0.666022,
        1.832596,0.618092,2.539118,
        1.832596,0.619654,-2.216909,
        1.832596,0.640596,-2.797608,
        1.832596,0.674054,1.013609,
        1.832596,0.685216,-1.712745,
        1.832596,0.700570,1.559208,
        1.832596,0.709924,0.218797,
        1.832596,0.733948,3.021411,
        1.832596,0.762244,-0.273349,
        1.832596,0.831641,0.616218,
        1.832596,0.850607,-1.335383,
        1.832596,0.864993,2.278281,
        1.832596,0.866395,-2.492918,
        1.832596,0.866955,-0.931351,
        1.832596,0.899350,-2.069907,
        1.832596,0.906588,1.252948,
        1.832596,0.910508,1.903534,
        1.832596,0.914445,2.687814,
        1.832596,0.929350,-2.910676,
        1.832596,0.960688,-0.577532,
        1.832596,0.961715,0.014085,
        1.832596,0.992645,-1.723759,
        1.832596,1.014500,0.906702,
        1.832596,1.025536,1.576859,
        1.832596,1.045978,0.343692,
        1.832596,1.050534,3.042579,
        1.832596,1.113094,-0.272241,
        1.832596,1.128190,-2.288622,
        1.832596,1.134497,-1.109631,
        1.832596,1.134796,2.434290,
        1.832596,1.135813,-2.658866,
        1.832596,1.135834,-1.431087,
        1.832596,1.168678,2.108595,
        1.832596,1.184361,0.625864,
        1.832596,1.194549,-0.801900,
        1.832596,1.195355,1.200319,
        1.832596,1.214854,-1.966664,
        1.832596,1.218453,2.777284,
        1.832596,1.242974,-2.977882,
        1.832596,1.247695,1.809890,
        1.832596,1.256302,0.062705,
        1.832596,1.301460,-0.509251,
        1.832596,1.310100,1.488307,
        1.832596,1.325504,-1.671296,
        1.832596,1.335669,0.897240,
        1.832596,1.367556,0.358405,
        1.832596,1.374601,-2.456031,
        1.832596,1.391910,3.030322,
        1.832596,1.413347,-1.339188,
        1.832596,1.413430,2.550734,
        1.832596,1.417623,-0.202254,
        1.832596,1.433411,-1.035986,
        1.832596,1.440580,2.260026,
        1.832596,1.448123,-2.163292,
        1.832596,1.450119,-2.739050,
        1.832596,1.488617,1.159546,
        1.832596,1.504028,0.636812,
        1.832596,1.506530,1.988858,
        1.832596,1.510910,-0.740472,
        1.832596,1.545873,-1.877082,
        1.832596,1.554776,0.100351,
        1.832596,1.555207,1.708892,
        1.832596,1.569311,-3.003978,
        1.832596,1.574060,2.783565,
        1.832596,1.599406,1.426070,
        1.832596,1.606104,-0.451151,
        1.832596,1.617114,-1.572037,
        1.832596,1.658653,-2.409879,
        1.832596,1.662363,0.915448,
        1.832596,1.683268,0.374177,
        1.832596,1.696674,-1.253233,
        1.832596,1.699788,2.452890,
        1.832596,1.721184,-0.153948,
        1.832596,1.732874,-0.951685,
        1.832596,1.736875,3.023612,
        1.832596,1.741644,-2.695914,
        1.832596,1.756530,-2.105156,
        1.832596,1.766067,2.153843,
        1.832596,1.809167,1.213774,
        1.832596,1.812176,0.656819,
        1.832596,1.824694,1.850876,
        1.832596,1.836682,-0.676006,
        1.832596,1.846342,-1.796954,
        1.832596,1.869911,2.701025,
        1.832596,1.870545,-2.976234,
        1.832596,1.872854,1.543456,
        1.832596,1.875281,0.107086,
        1.832596,1.904882,-1.478089,
        1.832596,1.909857,-0.381669,
        1.832596,1.948085,-2.363031,
        1.832596,1.977582,0.969896,
        1.832596,1.982574,-1.115518,
        1.832596,1.998705,2.370472,
        1.832596,2.001437,0.381215,
        1.832596,2.033014,2.977192,
        1.832596,2.041297,-2.664597,
        1.832596,2.063651,-2.044600,
        1.832596,2.073035,2.027405,
        1.832596,2.089241,-0.114193,
        1.832596,2.104282,-0.813972,
        1.832596,2.108018,1.311430,
        1.832596,2.115563,0.685730,
        1.832596,2.138223,1.678323,
        1.832596,2.153910,-1.714959,
        1.832596,2.164094,-2.991358,
        1.832596,2.174483,2.653697,
        1.832596,2.180315,-1.357379,
        1.832596,2.187132,-0.479875,
        1.832596,2.247462,0.191623,
        1.832596,2.251330,-2.330739,
        1.832596,2.295480,1.008277,
        1.832596,2.299159,2.264663,
        1.832596,2.343981,-1.040456,
        1.832596,2.349581,-2.702517,
        1.832596,2.356369,3.005869,
        1.832596,2.379497,0.561089,
        1.832596,2.393226,1.857173,
        1.832596,2.396274,-1.959260,
        1.832596,2.404280,-0.174044,
        1.832596,2.407836,1.410646,
        1.832596,2.455966,-0.641220,
        1.832596,2.456751,-1.503343,
        1.832596,2.478122,2.609777,
        1.832596,2.579248,-2.364801,
        1.832596,2.587733,0.232235,
        1.832596,2.588535,-2.976617,
        1.832596,2.592629,0.909713,
        1.832596,2.637879,2.121598,
        1.832596,2.664082,-1.071402,
        1.832596,2.704512,1.488683,
        1.832596,2.719179,-1.780644,
        1.832596,2.724429,-0.348782,
        1.832596,2.760637,2.781135,
        1.832596,2.863997,0.549740,
        1.832596,2.881233,-2.582574,
        1.832596,2.979456,-1.009304,
        1.832596,2.984978,1.954446,
        2.094395,0.080730,-2.622036,
        2.094395,0.203748,0.590255,
        2.094395,0.290380,-0.796960,
        2.094395,0.297065,2.160177,
        2.094395,0.370812,-2.547742,
        2.094395,0.379747,-1.713059,
        2.094395,0.418037,1.360082,
        2.094395,0.419640,2.974122,
        2.094395,0.458961,-0.107937,
        2.094395,0.496451,0.593859,
        2.094395,0.571433,-1.188419,
        2.094395,0.591903,2.010408,
        2.094395,0.615174,-0.666022,
        2.094395,0.618092,2.539118,
        2.094395,0.619654,-2.216909,
        2.094395,0.640596,-2.797608,
        2.094395,0.674054,1.013609,
        2.094395,0.685216,-1.712745,
        2.094395,0.700570,1.559208,
        2.094395,0.709924,0.218797,
        2.094395,0.733948,3.021411,
        2.094395,0.762244,-0.273349,
        2.094395,0.831641,0.616218,
        2.094395,0.850607,-1.335383,
        2.094395,0.864993,2.278281,
        2.094395,0.866395,-2.492918,
        2.094395,0.866955,-0.931351,
        2.094395,0.899350,-2.069907,
        2.094395,0.906588,1.252948,
        2.094395,0.910508,1.903534,
        2.094395,0.914445,2.687814,
        2.094395,0.929350,-2.910676,
        2.094395,0.960688,-0.577532,
        2.094395,0.961715,0.014085,
        2.094395,0.992645,-1.723759,
        2.094395,1.014500,0.906702,
        2.094395,1.025536,1.576859,
        2.094395,1.045978,0.343692,
        2.094395,1.050534,3.042579,
        2.094395,1.113094,-0.272241,
        2.094395,1.128190,-2.288622,
        2.094395,1.134497,-1.109631,
        2.094395,1.134796,2.434290,
        2.094395,1.135813,-2.658866,
        2.094395,1.135834,-1.431087,
        2.094395,1.168678,2.108595,
        2.094395,1.184361,0.625864,
        2.094395,1.194549,-0.801900,
        2.094395,1.195355,1.200319,
        2.094395,1.214854,-1.966664,
        2.094395,1.218453,2.777284,
        2.094395,1.242974,-2.977882,
        2.094395,1.247695,1.809890,
        2.094395,1.256302,0.062705,
        2.094395,1.301460,-0.509251,
        2.094395,1.310100,1.488307,
        2.094395,1.325504,-1.671296,
        2.094395,1.335669,0.897240,
        2.094395,1.367556,0.358405,
        2.094395,1.374601,-2.456031,
        2.094395,1.391910,3.030322,
        2.094395,1.413347,-1.339188,
        2.094395,1.413430,2.550734,
        2.094395,1.417623,-0.202254,
        2.094395,1.433411,-1.035986,
        2.094395,1.440580,2.260026,
        2.094395,1.448123,-2.163292,
        2.094395,1.450119,-2.739050,
        2.094395,1.488617,1.159546,
        2.094395,1.504028,0.636812,
        2.094395,1.506530,1.988858,
        2.094395,1.510910,-0.740472,
        2.094395,1.545873,-1.877082,
        2.094395,1.554776,0.100351,
        2.094395,1.555207,1.708892,
        2.094395,1.569311,-3.003978,
        2.094395,1.574060,2.783565,
        2.094395,1.599406,1.426070,
        2.094395,1.606104,-0.451151,
        2.094395,1.617114,-1.572037,
        2.094395,1.658653,-2.409879,
        2.094395,1.662363,0.915448,
        2.094395,1.683268,0.374177,
        2.094395,1.696674,-1.253233,
        2.094395,1.699788,2.452890,
        2.094395,1.721184,-0.153948,
        2.094395,1.732874,-0.951685,
        2.094395,1.736875,3.023612,
        2.094395,1.741644,-2.695914,
        2.094395,1.756530,-2.105156,
        2.094395,1.766067,2.153843,
        2.094395,1.809167,1.213774,
        2.094395,1.812176,0.656819,
        2.094395,1.824694,1.850876,
        2.094395,1.836682,-0.676006,
        2.094395,1.846342,-1.796954,
        2.094395,1.869911,2.701025,
        2.094395,1.870545,-2.976234,
        2.094395,1.872854,1.543456,
        2.094395,1.875281,0.107086,
        2.094395,1.904882,-1.478089,
        2.094395,1.909857,-0.381669,
        2.094395,1.948085,-2.363031,
        2.094395,1.977582,0.969896,
        2.094395,1.982574,-1.115518,
        2.094395,1.998705,2.370472,
        2.094395,2.001437,0.381215,
        2.094395,2.033014,2.977192,
        2.094395,2.041297,-2.664597,
        2.094395,2.063651,-2.044600,
        2.094395,2.073035,2.027405,
        2.094395,2.089241,-0.114193,
        2.094395,2.104282,-0.813972,
        2.094395,2.108018,1.311430,
        2.094395,2.115563,0.685730,
        2.094395,2.138223,1.678323,
        2.094395,2.153910,-1.714959,
        2.094395,2.164094,-2.991358,
        2.094395,2.174483,2.653697,
        2.094395,2.180315,-1.357379,
        2.094395,2.187132,-0.479875,
        2.094395,2.247462,0.191623,
        2.094395,2.251330,-2.330739,
        2.094395,2.295480,1.008277,
        2.094395,2.299159,2.264663,
        2.094395,2.343981,-1.040456,
        2.094395,2.349581,-2.702517,
        2.094395,2.356369,3.005869,
        2.094395,2.379497,0.561089,
        2.094395,2.393226,1.857173,
        2.094395,2.396274,-1.959260,
        2.094395,2.404280,-0.174044,
        2.094395,2.407836,1.410646,
        2.094395,2.455966,-0.641220,
        2.094395,2.456751,-1.503343,
        2.094395,2.478122,2.609777,
        2.094395,2.579248,-2.364801,
        2.094395,2.587733,0.232235,
        2.094395,2.588535,-2.976617,
        2.094395,2.592629,0.909713,
        2.094395,2.637879,2.121598,
        2.094395,2.664082,-1.071402,
        2.094395,2.704512,1.488683,
        2.094395,2.719179,-1.780644,
        2.094395,2.724429,-0.348782,
        2.094395,2.760637,2.781135,
        2.094395,2.863997,0.549740,
        2.094395,2.881233,-2.582574,
        2.094395,2.979456,-1.009304,
        2.094395,2.984978,1.954446,
        2.356194,0.080730,-2.622036,
        2.356194,0.203748,0.590255,
        2.356194,0.290380,-0.796960,
        2.356194,0.297065,2.160177,
        2.356194,0.370812,-2.547742,
        2.356194,0.379747,-1.713059,
        2.356194,0.418037,1.360082,
        2.356194,0.419640,2.974122,
        2.356194,0.458961,-0.107937,
        2.356194,0.496451,0.593859,
        2.356194,0.571433,-1.188419,
        2.356194,0.591903,2.010408,
        2.356194,0.615174,-0.666022,
        2.356194,0.618092,2.539118,
        2.356194,0.619654,-2.216909,
        2.356194,0.640596,-2.797608,
        2.356194,0.674054,1.013609,
        2.356194,0.685216,-1.712745,
        2.356194,0.700570,1.559208,
        2.356194,0.709924,0.218797,
        2.356194,0.733948,3.021411,
        2.356194,0.762244,-0.273349,
        2.356194,0.831641,0.616218,
        2.356194,0.850607,-1.335383,
        2.356194,0.864993,2.278281,
        2.356194,0.866395,-2.492918,
        2.356194,0.866955,-0.931351,
        2.356194,0.899350,-2.069907,
        2.356194,0.906588,1.252948,
        2.356194,0.910508,1.903534,
        2.356194,0.914445,2.687814,
        2.356194,0.929350,-2.910676,
        2.356194,0.960688,-0.577532,
        2.356194,0.961715,0.014085,
        2.356194,0.992645,-1.723759,
        2.356194,1.014500,0.906702,
        2.356194,1.025536,1.576859,
        2.356194,1.045978,0.343692,
        2.356194,1.050534,3.042579,
        2.356194,1.113094,-0.272241,
        2.356194,1.128190,-2.288622,
        2.356194,1.134497,-1.109631,
        2.356194,1.134796,2.434290,
        2.356194,1.135813,-2.658866,
        2.356194,1.135834,-1.431087,
        2.356194,1.168678,2.108595,
        2.356194,1.184361,0.625864,
        2.356194,1.194549,-0.801900,
        2.356194,1.195355,1.200319,
        2.356194,1.214854,-1.966664,
        2.356194,1.218453,2.777284,
        2.356194,1.242974,-2.977882,
        2.356194,1.247695,1.809890,
        2.356194,1.256302,0.062705,
        2.356194,1.301460,-0.509251,
        2.356194,1.310100,1.488307,
        2.356194,1.325504,-1.671296,
        2.356194,1.335669,0.897240,
        2.356194,1.367556,0.358405,
        2.356194,1.374601,-2.456031,
        2.356194,1.391910,3.030322,
        2.356194,1.413347,-1.339188,
        2.356194,1.413430,2.550734,
        2.356194,1.417623,-0.202254,
        2.356194,1.433411,-1.035986,
        2.356194,1.440580,2.260026,
        2.356194,1.448123,-2.163292,
        2.356194,1.450119,-2.739050,
        2.356194,1.488617,1.159546,
        2.356194,1.504028,0.636812,
        2.356194,1.506530,1.988858,
        2.356194,1.510910,-0.740472,
        2.356194,1.545873,-1.877082,
        2.356194,1.554776,0.100351,
        2.356194,1.555207,1.708892,
        2.356194,1.569311,-3.003978,
        2.356194,1.574060,2.783565,
        2.356194,1.599406,1.426070,
        2.356194,1.606104,-0.451151,
        2.356194,1.617114,-1.572037,
        2.356194,1.658653,-2.409879,
        2.356194,1.662363,0.915448,
        2.356194,1.683268,0.374177,
        2.356194,1.696674,-1.253233,
        2.356194,1.699788,2.452890,
        2.356194,1.721184,-0.153948,
        2.356194,1.732874,-0.951685,
        2.356194,1.736875,3.023612,
        2.356194,1.741644,-2.695914,
        2.356194,1.756530,-2.105156,
        2.356194,1.766067,2.153843,
        2.356194,1.809167,1.213774,
        2.356194,1.812176,0.656819,
        2.356194,1.824694,1.850876,
        2.356194,1.836682,-0.676006,
        2.356194,1.846342,-1.796954,
        2.356194,1.869911,2.701025,
        2.356194,1.870545,-2.976234,
        2.356194,1.872854,1.543456,
        2.356194,1.875281,0.107086,
        2.356194,1.904882,-1.478089,
        2.356194,1.909857,-0.381669,
        2.356194,1.948085,-2.363031,
        2.356194,1.977582,0.969896,
        2.356194,1.982574,-1.115518,
        2.356194,1.998705,2.370472,
        2.356194,2.001437,0.381215,
        2.356194,2.033014,2.977192,
        2.356194,2.041297,-2.664597,
        2.356194,2.063651,-2.044600,
        2.356194,2.073035,2.027405,
        2.356194,2.089241,-0.114193,
        2.356194,2.104282,-0.813972,
        2.356194,2.108018,1.311430,
        2.356194,2.115563,0.685730,
        2.356194,2.138223,1.678323,
        2.356194,2.153910,-1.714959,
        2.356194,2.164094,-2.991358,
        2.356194,2.174483,2.653697,
        2.356194,2.180315,-1.357379,
        2.356194,2.187132,-0.479875,
        2.356194,2.247462,0.191623,
        2.356194,2.251330,-2.330739,
        2.356194,2.295480,1.008277,
        2.356194,2.299159,2.264663,
        2.356194,2.343981,-1.040456,
        2.356194,2.349581,-2.702517,
        2.356194,2.356369,3.005869,
        2.356194,2.379497,0.561089,
        2.356194,2.393226,1.857173,
        2.356194,2.396274,-1.959260,
        2.356194,2.404280,-0.174044,
        2.356194,2.407836,1.410646,
        2.356194,2.455966,-0.641220,
        2.356194,2.456751,-1.503343,
        2.356194,2.478122,2.609777,
        2.356194,2.579248,-2.364801,
        2.356194,2.587733,0.232235,
        2.356194,2.588535,-2.976617,
        2.356194,2.592629,0.909713,
        2.356194,2.637879,2.121598,
        2.356194,2.664082,-1.071402,
        2.356194,2.704512,1.488683,
        2.356194,2.719179,-1.780644,
        2.356194,2.724429,-0.348782,
        2.356194,2.760637,2.781135,
        2.356194,2.863997,0.549740,
        2.356194,2.881233,-2.582574,
        2.356194,2.979456,-1.009304,
        2.356194,2.984978,1.954446,
        2.617994,0.080730,-2.622036,
        2.617994,0.203748,0.590255,
        2.617994,0.290380,-0.796960,
        2.617994,0.297065,2.160177,
        2.617994,0.370812,-2.547742,
        2.617994,0.379747,-1.713059,
        2.617994,0.418037,1.360082,
        2.617994,0.419640,2.974122,
        2.617994,0.458961,-0.107937,
        2.617994,0.496451,0.593859,
        2.617994,0.571433,-1.188419,
        2.617994,0.591903,2.010408,
        2.617994,0.615174,-0.666022,
        2.617994,0.618092,2.539118,
        2.617994,0.619654,-2.216909,
        2.617994,0.640596,-2.797608,
        2.617994,0.674054,1.013609,
        2.617994,0.685216,-1.712745,
        2.617994,0.700570,1.559208,
        2.617994,0.709924,0.218797,
        2.617994,0.733948,3.021411,
        2.617994,0.762244,-0.273349,
        2.617994,0.831641,0.616218,
        2.617994,0.850607,-1.335383,
        2.617994,0.864993,2.278281,
        2.617994,0.866395,-2.492918,
        2.617994,0.866955,-0.931351,
        2.617994,0.899350,-2.069907,
        2.617994,0.906588,1.252948,
        2.617994,0.910508,1.903534,
        2.617994,0.914445,2.687814,
        2.617994,0.929350,-2.910676,
        2.617994,0.960688,-0.577532,
        2.617994,0.961715,0.014085,
        2.617994,0.992645,-1.723759,
        2.617994,1.014500,0.906702,
        2.617994,1.025536,1.576859,
        2.617994,1.045978,0.343692,
        2.617994,1.050534,3.042579,
        2.617994,1.113094,-0.272241,
        2.617994,1.128190,-2.288622,
        2.617994,1.134497,-1.109631,
        2.617994,1.134796,2.434290,
        2.617994,1.135813,-2.658866,
        2.617994,1.135834,-1.431087,
        2.617994,1.168678,2.108595,
        2.617994,1.184361,0.625864,
        2.617994,1.194549,-0.801900,
        2.617994,1.195355,1.200319,
        2.617994,1.214854,-1.966664,
        2.617994,1.218453,2.777284,
        2.617994,1.242974,-2.977882,
        2.617994,1.247695,1.809890,
        2.617994,1.256302,0.062705,
        2.617994,1.301460,-0.509251,
        2.617994,1.310100,1.488307,
        2.617994,1.325504,-1.671296,
        2.617994,1.335669,0.897240,
        2.617994,1.367556,0.358405,
        2.617994,1.374601,-2.456031,
        2.617994,1.391910,3.030322,
        2.617994,1.413347,-1.339188,
        2.617994,1.413430,2.550734,
        2.617994,1.417623,-0.202254,
        2.617994,1.433411,-1.035986,
        2.617994,1.440580,2.260026,
        2.617994,1.448123,-2.163292,
        2.617994,1.450119,-2.739050,
        2.617994,1.488617,1.159546,
        2.617994,1.504028,0.636812,
        2.617994,1.506530,1.988858,
        2.617994,1.510910,-0.740472,
        2.617994,1.545873,-1.877082,
        2.617994,1.554776,0.100351,
        2.617994,1.555207,1.708892,
        2.617994,1.569311,-3.003978,
        2.617994,1.574060,2.783565,
        2.617994,1.599406,1.426070,
        2.617994,1.606104,-0.451151,
        2.617994,1.617114,-1.572037,
        2.617994,1.658653,-2.409879,
        2.617994,1.662363,0.915448,
        2.617994,1.683268,0.374177,
        2.617994,1.696674,-1.253233,
        2.617994,1.699788,2.452890,
        2.617994,1.721184,-0.153948,
        2.617994,1.732874,-0.951685,
        2.617994,1.736875,3.023612,
        2.617994,1.741644,-2.695914,
        2.617994,1.756530,-2.105156,
        2.617994,1.766067,2.153843,
        2.617994,1.809167,1.213774,
        2.617994,1.812176,0.656819,
        2.617994,1.824694,1.850876,
        2.617994,1.836682,-0.676006,
        2.617994,1.846342,-1.796954,
        2.617994,1.869911,2.701025,
        2.617994,1.870545,-2.976234,
        2.617994,1.872854,1.543456,
        2.617994,1.875281,0.107086,
        2.617994,1.904882,-1.478089,
        2.617994,1.909857,-0.381669,
        2.617994,1.948085,-2.363031,
        2.617994,1.977582,0.969896,
        2.617994,1.982574,-1.115518,
        2.617994,1.998705,2.370472,
        2.617994,2.001437,0.381215,
        2.617994,2.033014,2.977192,
        2.617994,2.041297,-2.664597,
        2.617994,2.063651,-2.044600,
        2.617994,2.073035,2.027405,
        2.617994,2.089241,-0.114193,
        2.617994,2.104282,-0.813972,
        2.617994,2.108018,1.311430,
        2.617994,2.115563,0.685730,
        2.617994,2.138223,1.678323,
        2.617994,2.153910,-1.714959,
        2.617994,2.164094,-2.991358,
        2.617994,2.174483,2.653697,
        2.617994,2.180315,-1.357379,
        2.617994,2.187132,-0.479875,
        2.617994,2.247462,0.191623,
        2.617994,2.251330,-2.330739,
        2.617994,2.295480,1.008277,
        2.617994,2.299159,2.264663,
        2.617994,2.343981,-1.040456,
        2.617994,2.349581,-2.702517,
        2.617994,2.356369,3.005869,
        2.617994,2.379497,0.561089,
        2.617994,2.393226,1.857173,
        2.617994,2.396274,-1.959260,
        2.617994,2.404280,-0.174044,
        2.617994,2.407836,1.410646,
        2.617994,2.455966,-0.641220,
        2.617994,2.456751,-1.503343,
        2.617994,2.478122,2.609777,
        2.617994,2.579248,-2.364801,
        2.617994,2.587733,0.232235,
        2.617994,2.588535,-2.976617,
        2.617994,2.592629,0.909713,
        2.617994,2.637879,2.121598,
        2.617994,2.664082,-1.071402,
        2.617994,2.704512,1.488683,
        2.617994,2.719179,-1.780644,
        2.617994,2.724429,-0.348782,
        2.617994,2.760637,2.781135,
        2.617994,2.863997,0.549740,
        2.617994,2.881233,-2.582574,
        2.617994,2.979456,-1.009304,
        2.617994,2.984978,1.954446,
        2.879793,0.080730,-2.622036,
        2.879793,0.203748,0.590255,
        2.879793,0.290380,-0.796960,
        2.879793,0.297065,2.160177,
        2.879793,0.370812,-2.547742,
        2.879793,0.379747,-1.713059,
        2.879793,0.418037,1.360082,
        2.879793,0.419640,2.974122,
        2.879793,0.458961,-0.107937,
        2.879793,0.496451,0.593859,
        2.879793,0.571433,-1.188419,
        2.879793,0.591903,2.010408,
        2.879793,0.615174,-0.666022,
        2.879793,0.618092,2.539118,
        2.879793,0.619654,-2.216909,
        2.879793,0.640596,-2.797608,
        2.879793,0.674054,1.013609,
        2.879793,0.685216,-1.712745,
        2.879793,0.700570,1.559208,
        2.879793,0.709924,0.218797,
        2.879793,0.733948,3.021411,
        2.879793,0.762244,-0.273349,
        2.879793,0.831641,0.616218,
        2.879793,0.850607,-1.335383,
        2.879793,0.864993,2.278281,
        2.879793,0.866395,-2.492918,
        2.879793,0.866955,-0.931351,
        2.879793,0.899350,-2.069907,
        2.879793,0.906588,1.252948,
        2.879793,0.910508,1.903534,
        2.879793,0.914445,2.687814,
        2.879793,0.929350,-2.910676,
        2.879793,0.960688,-0.577532,
        2.879793,0.961715,0.014085,
        2.879793,0.992645,-1.723759,
        2.879793,1.014500,0.906702,
        2.879793,1.025536,1.576859,
        2.879793,1.045978,0.343692,
        2.879793,1.050534,3.042579,
        2.879793,1.113094,-0.272241,
        2.879793,1.128190,-2.288622,
        2.879793,1.134497,-1.109631,
        2.879793,1.134796,2.434290,
        2.879793,1.135813,-2.658866,
        2.879793,1.135834,-1.431087,
        2.879793,1.168678,2.108595,
        2.879793,1.184361,0.625864,
        2.879793,1.194549,-0.801900,
        2.879793,1.195355,1.200319,
        2.879793,1.214854,-1.966664,
        2.879793,1.218453,2.777284,
        2.879793,1.242974,-2.977882,
        2.879793,1.247695,1.809890,
        2.879793,1.256302,0.062705,
        2.879793,1.301460,-0.509251,
        2.879793,1.310100,1.488307,
        2.879793,1.325504,-1.671296,
        2.879793,1.335669,0.897240,
        2.879793,1.367556,0.358405,
        2.879793,1.374601,-2.456031,
        2.879793,1.391910,3.030322,
        2.879793,1.413347,-1.339188,
        2.879793,1.413430,2.550734,
        2.879793,1.417623,-0.202254,
        2.879793,1.433411,-1.035986,
        2.879793,1.440580,2.260026,
        2.879793,1.448123,-2.163292,
        2.879793,1.450119,-2.739050,
        2.879793,1.488617,1.159546,
        2.879793,1.504028,0.636812,
        2.879793,1.506530,1.988858,
        2.879793,1.510910,-0.740472,
        2.879793,1.545873,-1.877082,
        2.879793,1.554776,0.100351,
        2.879793,1.555207,1.708892,
        2.879793,1.569311,-3.003978,
        2.879793,1.574060,2.783565,
        2.879793,1.599406,1.426070,
        2.879793,1.606104,-0.451151,
        2.879793,1.617114,-1.572037,
        2.879793,1.658653,-2.409879,
        2.879793,1.662363,0.915448,
        2.879793,1.683268,0.374177,
        2.879793,1.696674,-1.253233,
        2.879793,1.699788,2.452890,
        2.879793,1.721184,-0.153948,
        2.879793,1.732874,-0.951685,
        2.879793,1.736875,3.023612,
        2.879793,1.741644,-2.695914,
        2.879793,1.756530,-2.105156,
        2.879793,1.766067,2.153843,
        2.879793,1.809167,1.213774,
        2.879793,1.812176,0.656819,
        2.879793,1.824694,1.850876,
        2.879793,1.836682,-0.676006,
        2.879793,1.846342,-1.796954,
        2.879793,1.869911,2.701025,
        2.879793,1.870545,-2.976234,
        2.879793,1.872854,1.543456,
        2.879793,1.875281,0.107086,
        2.879793,1.904882,-1.478089,
        2.879793,1.909857,-0.381669,
        2.879793,1.948085,-2.363031,
        2.879793,1.977582,0.969896,
        2.879793,1.982574,-1.115518,
        2.879793,1.998705,2.370472,
        2.879793,2.001437,0.381215,
        2.879793,2.033014,2.977192,
        2.879793,2.041297,-2.664597,
        2.879793,2.063651,-2.044600,
        2.879793,2.073035,2.027405,
        2.879793,2.089241,-0.114193,
        2.879793,2.104282,-0.813972,
        2.879793,2.108018,1.311430,
        2.879793,2.115563,0.685730,
        2.879793,2.138223,1.678323,
        2.879793,2.153910,-1.714959,
        2.879793,2.164094,-2.991358,
        2.879793,2.174483,2.653697,
        2.879793,2.180315,-1.357379,
        2.879793,2.187132,-0.479875,
        2.879793,2.247462,0.191623,
        2.879793,2.251330,-2.330739,
        2.879793,2.295480,1.008277,
        2.879793,2.299159,2.264663,
        2.879793,2.343981,-1.040456,
        2.879793,2.349581,-2.702517,
        2.879793,2.356369,3.005869,
        2.879793,2.379497,0.561089,
        2.879793,2.393226,1.857173,
        2.879793,2.396274,-1.959260,
        2.879793,2.404280,-0.174044,
        2.879793,2.407836,1.410646,
        2.879793,2.455966,-0.641220,
        2.879793,2.456751,-1.503343,
        2.879793,2.478122,2.609777,
        2.879793,2.579248,-2.364801,
        2.879793,2.587733,0.232235,
        2.879793,2.588535,-2.976617,
        2.879793,2.592629,0.909713,
        2.879793,2.637879,2.121598,
        2.879793,2.664082,-1.071402,
        2.879793,2.704512,1.488683,
        2.879793,2.719179,-1.780644,
        2.879793,2.724429,-0.348782,
        2.879793,2.760637,2.781135,
        2.879793,2.863997,0.549740,
        2.879793,2.881233,-2.582574,
        2.879793,2.979456,-1.009304,
        2.879793,2.984978,1.954446
    };
    //*/

    _Num_rot_angles  = sizeof(Zangle)/sizeof(Zangle[0])/3;
    _Zangle   = new float[_Num_rot_angles*3];

    for( int i = 0 ; i < _Num_rot_angles*3 ; i++ ) {
        _Zangle[i] = Zangle[i];
    }

    return;
}

//============================================================================//
void Parameter::dangle_rot54000()
//============================================================================//
{
    //* comment out for faster compliation
    float Zangle[] = {
        -3.141593,0.053884,-2.408394,
        -3.141593,0.077645,1.453433,
        -3.141593,0.088408,-0.311160,
        -3.141593,0.142314,2.590387,
        -3.141593,0.147206,-1.385164,
        -3.141593,0.163506,0.573203,
        -3.141593,0.165485,-2.875396,
        -3.141593,0.182564,-2.125821,
        -3.141593,0.190569,1.929141,
        -3.141593,0.201102,1.249082,
        -3.141593,0.201397,-0.716646,
        -3.141593,0.210568,-0.069559,
        -3.141593,0.251672,2.848900,
        -3.141593,0.266262,-1.678068,
        -3.141593,0.269108,-1.215563,
        -3.141593,0.273777,-2.506756,
        -3.141593,0.279854,0.814581,
        -3.141593,0.281664,-2.969147,
        -3.141593,0.283625,2.405603,
        -3.141593,0.284058,0.357350,
        -3.141593,0.294107,1.593880,
        -3.141593,0.302491,-0.400410,
        -3.141593,0.308059,2.021491,
        -3.141593,0.312219,-2.072057,
        -3.141593,0.321905,-0.829121,
        -3.141593,0.329984,1.201621,
        -3.141593,0.339714,-0.013712,
        -3.141593,0.368784,2.974974,
        -3.141593,0.376428,-1.495532,
        -3.141593,0.378103,2.646436,
        -3.141593,0.385088,0.590355,
        -3.141593,0.386248,-2.690943,
        -3.141593,0.389851,-2.340031,
        -3.141593,0.391502,-1.171374,
        -3.141593,0.398249,-1.812694,
        -3.141593,0.400542,-2.995729,
        -3.141593,0.404273,0.918163,
        -3.141593,0.405617,1.773533,
        -3.141593,0.409009,-0.570243,
        -3.141593,0.410700,0.270433,
        -3.141593,0.415893,1.464890,
        -3.141593,0.420680,2.088164,
        -3.141593,0.424766,-0.258729,
        -3.141593,0.426507,2.367872,
        -3.141593,0.441292,-0.890727,
        -3.141593,0.451839,-2.078193,
        -3.141593,0.459270,1.191562,
        -3.141593,0.470622,0.014002,
        -3.141593,0.481292,-2.529098,
        -3.141593,0.482460,2.849843,
        -3.141593,0.492205,-1.397963,
        -3.141593,0.493314,3.117264,
        -3.141593,0.497870,0.727828,
        -3.141593,0.499783,-1.654379,
        -3.141593,0.502651,0.463333,
        -3.141593,0.505463,2.594557,
        -3.141593,0.506104,-2.874910,
        -3.141593,0.513579,-1.147354,
        -3.141593,0.518569,1.644237,
        -3.141593,0.519001,1.900488,
        -3.141593,0.519277,-0.673809,
        -3.141593,0.524769,-0.421955,
        -3.141593,0.525564,-2.288838,
        -3.141593,0.532541,0.978737,
        -3.141593,0.534610,2.154808,
        -3.141593,0.535020,-1.891846,
        -3.141593,0.540208,1.400497,
        -3.141593,0.542006,0.228337,
        -3.141593,0.550762,-0.180771,
        -3.141593,0.560531,2.378474,
        -3.141593,0.561629,-0.924576,
        -3.141593,0.567465,-2.679392,
        -3.141593,0.588527,1.181755,
        -3.141593,0.591643,-2.098364,
        -3.141593,0.594963,2.997575,
        -3.141593,0.599643,-3.045235,
        -3.141593,0.599962,2.773556,
        -3.141593,0.602620,0.029270,
        -3.141593,0.605711,0.598454,
        -3.141593,0.607708,-1.541026,
        -3.141593,0.609360,-2.464379,
        -3.141593,0.610738,-1.326786,
        -3.141593,0.616968,0.820489,
        -3.141593,0.626379,1.773610,
        -3.141593,0.628626,-1.751180,
        -3.141593,0.630671,0.391949,
        -3.141593,0.632622,-0.737642,
        -3.141593,0.632708,2.566722,
        -3.141593,0.632797,1.989858,
        -3.141593,0.634455,-0.534790,
        -3.141593,0.636205,1.561170,
        -3.141593,0.637681,-1.123423,
        -3.141593,0.641441,-2.842109,
        -3.141593,0.644807,-0.327139,
        -3.141593,0.652440,2.197129,
        -3.141593,0.660897,-2.278593,
        -3.141593,0.664008,1.360091,
        -3.141593,0.670850,-1.945257,
        -3.141593,0.671396,1.004859,
        -3.141593,0.673985,0.203855,
        -3.141593,0.680326,-0.133370,
        -3.141593,0.685083,-0.942363,
        -3.141593,0.688204,-2.632204,
        -3.141593,0.689777,2.385011,
        -3.141593,0.700065,3.115430,
        -3.141593,0.704080,2.911636,
        -3.141593,0.712883,1.174216,
        -3.141593,0.715701,0.699906,
        -3.141593,0.718851,-1.442215,
        -3.141593,0.721019,2.721536,
        -3.141593,0.725300,-2.975134,
        -3.141593,0.727136,-1.633288,
        -3.141593,0.727399,-2.120567,
        -3.141593,0.730751,0.522488,
        -3.141593,0.732892,-1.254330,
        -3.141593,0.734333,-2.446184,
        -3.141593,0.735125,0.037270,
        -3.141593,0.737893,1.866463,
        -3.141593,0.740650,1.681075,
        -3.141593,0.745971,-0.781286,
        -3.141593,0.749432,-0.442559,
        -3.141593,0.750102,2.050636,
        -3.141593,0.751046,0.863821,
        -3.141593,0.757018,-0.608833,
        -3.141593,0.757046,-1.816019,
        -3.141593,0.757389,1.501184,
        -3.141593,0.759800,2.546181,
        -3.141593,0.759859,0.351923,
        -3.141593,0.761141,-2.781102,
        -3.141593,0.769035,-1.078574,
        -3.141593,0.770166,-0.266604,
        -3.141593,0.772218,2.227380,
        -3.141593,0.788382,1.331051,
        -3.141593,0.790269,-2.282409,
        -3.141593,0.803410,-1.984567,
        -3.141593,0.803813,0.191520,
        -3.141593,0.806832,3.022155,
        -3.141593,0.807476,-2.597675,
        -3.141593,0.812512,-0.107335,
        -3.141593,0.814161,1.010949,
        -3.141593,0.817732,-3.084486,
        -3.141593,0.817846,2.390688,
        -3.141593,0.819323,2.845960,
        -3.141593,0.819430,-0.915017,
        -3.141593,0.829198,-1.532491,
        -3.141593,0.829531,0.665695,
        -3.141593,0.829816,1.167147,
        -3.141593,0.833613,-1.357492,
        -3.141593,0.842945,-2.907318,
        -3.141593,0.844883,2.679708,
        -3.141593,0.848189,-1.705003,
        -3.141593,0.849888,1.767901,
        -3.141593,0.851126,0.493370,
        -3.141593,0.852911,1.933394,
        -3.141593,0.857516,-2.432863,
        -3.141593,0.858261,-1.191729,
        -3.141593,0.858279,-2.138895,
        -3.141593,0.859723,-0.734928,
        -3.141593,0.859754,1.603292,
        -3.141593,0.861655,-0.394821,
        -3.141593,0.864300,0.043549,
        -3.141593,0.866940,0.851229,
        -3.141593,0.869748,2.095969,
        -3.141593,0.877352,-0.574668,
        -3.141593,0.879254,-2.733338,
        -3.141593,0.882554,1.443757,
        -3.141593,0.884060,0.334199,
        -3.141593,0.885626,-1.866061,
        -3.141593,0.889834,2.529957,
        -3.141593,0.893140,2.251698,
        -3.141593,0.897463,-0.240822,
        -3.141593,0.898758,-1.031948,
        -3.141593,0.913918,1.287618,
        -3.141593,0.916555,-2.287173,
        -3.141593,0.918305,3.107845,
        -3.141593,0.919522,2.948277,
        -3.141593,0.927138,-2.569686,
        -3.141593,0.929031,0.185383,
        -3.141593,0.933766,-3.015917,
        -3.141593,0.934873,-2.010822,
        -3.141593,0.935636,-1.445855,
        -3.141593,0.936702,-0.866461,
        -3.141593,0.937395,0.729108,
        -3.141593,0.937910,2.787575,
        -3.141593,0.939144,-0.092962,
        -3.141593,0.941718,1.125085,
        -3.141593,0.942873,-1.607680,
        -3.141593,0.943701,0.980328,
        -3.141593,0.947774,2.395256,
        -3.141593,0.952433,0.584347,
        -3.141593,0.954456,-1.287293,
        -3.141593,0.958731,-2.852552,
        -3.141593,0.961838,1.839284,
        -3.141593,0.961898,-0.469425,
        -3.141593,0.965426,1.686615,
        -3.141593,0.970108,1.990771,
        -3.141593,0.972820,-1.762780,
        -3.141593,0.973500,-0.711881,
        -3.141593,0.975253,2.641082,
        -3.141593,0.977824,0.441928,
        -3.141593,0.981485,-2.424766,
        -3.141593,0.981877,1.537100,
        -3.141593,0.982609,-1.136920,
        -3.141593,0.984265,-2.156073,
        -3.141593,0.989150,0.045493,
        -3.141593,0.993052,2.137478,
        -3.141593,0.997766,-2.693871,
        -3.141593,0.999847,-0.334738,
        -3.141593,1.001920,0.848054,
        -3.141593,1.006146,1.389734,
        -3.141593,1.012180,0.304066,
        -3.141593,1.016938,2.270674,
        -3.141593,1.017371,-0.979790,
        -3.141593,1.019495,-1.894487,
        -3.141593,1.023796,-0.583691,
        -3.141593,1.026525,2.506945,
        -3.141593,1.027004,-0.203653,
        -3.141593,1.028027,3.036573,
        -3.141593,1.031435,1.241021,
        -3.141593,1.032316,2.884221,
        -3.141593,1.035863,-3.110475,
        -3.141593,1.042085,-2.292950,
        -3.141593,1.043272,-1.522986,
        -3.141593,1.046265,-2.958847,
        -3.141593,1.049272,-1.376344,
        -3.141593,1.050039,-2.552086,
        -3.141593,1.053682,0.689863,
        -3.141593,1.055458,0.170394,
        -3.141593,1.056622,-0.831866,
        -3.141593,1.057515,1.097998,
        -3.141593,1.058706,-2.036438,
        -3.141593,1.061389,-1.672248,
        -3.141593,1.063758,2.743093,
        -3.141593,1.070157,-0.073422,
        -3.141593,1.072154,1.761506,
        -3.141593,1.074029,1.909193,
        -3.141593,1.074204,0.545276,
        -3.141593,1.074714,-2.805147,
        -3.141593,1.076262,0.966104,
        -3.141593,1.077642,-1.227073,
        -3.141593,1.084151,1.618525,
        -3.141593,1.085085,2.377673,
        -3.141593,1.086822,-0.441112,
        -3.141593,1.087142,2.058815,
        -3.141593,1.098383,-0.696430,
        -3.141593,1.099138,-1.084033,
        -3.141593,1.103208,1.477773,
        -3.141593,1.103783,0.408254,
        -3.141593,1.106275,-2.420800,
        -3.141593,1.108079,2.612943,
        -3.141593,1.109345,-2.172718,
        -3.141593,1.115957,-1.787597,
        -3.141593,1.118930,0.045983,
        -3.141593,1.119777,0.834702,
        -3.141593,1.121943,-2.669042,
        -3.141593,1.124930,2.219917,
        -3.141593,1.125217,1.336294,
        -3.141593,1.130958,2.972931,
        -3.141593,1.131400,-0.306629,
        -3.141593,1.137135,-0.936514,
        -3.141593,1.139125,-1.924528,
        -3.141593,1.139877,0.276382,
        -3.141593,1.142003,-3.058022,
        -3.141593,1.147363,-0.558448,
        -3.141593,1.150348,-0.176418,
        -3.141593,1.151198,1.197016,
        -3.141593,1.153930,-1.597104,
        -3.141593,1.155389,2.837148,
        -3.141593,1.155458,-1.460623,
        -3.141593,1.157331,-2.906362,
        -3.141593,1.158594,2.487618,
        -3.141593,1.162103,-1.322553,
        -3.141593,1.165407,3.092861,
        -3.141593,1.168025,0.717075,
        -3.141593,1.168406,-2.301519,
        -3.141593,1.174782,-2.538755,
        -3.141593,1.177998,1.061256,
        -3.141593,1.178898,1.827305,
        -3.141593,1.181331,-2.060330,
        -3.141593,1.182540,-0.798952,
        -3.141593,1.184569,0.147517,
        -3.141593,1.184734,1.968919,
        -3.141593,1.184804,0.593386,
        -3.141593,1.188278,1.690808,
        -3.141593,1.194360,2.110440,
        -3.141593,1.194776,2.710455,
        -3.141593,1.197745,-1.173705,
        -3.141593,1.199898,-2.777257,
        -3.141593,1.201164,2.355931,
        -3.141593,1.203760,1.556404,
        -3.141593,1.205845,-0.415993,
        -3.141593,1.206585,0.933648,
        -3.141593,1.206950,-0.037248,
        -3.141593,1.209341,0.471162,
        -3.141593,1.213483,-1.039417,
        -3.141593,1.218146,-0.672643,
        -3.141593,1.221608,1.421669,
        -3.141593,1.226043,-1.693533,
        -3.141593,1.232870,-1.822773,
        -3.141593,1.234375,-2.421194,
        -3.141593,1.234531,-2.190148,
        -3.141593,1.239414,0.348959,
        -3.141593,1.240183,2.587255,
        -3.141593,1.242312,2.941234,
        -3.141593,1.244075,1.286644,
        -3.141593,1.245961,-2.997461,
        -3.141593,1.246495,-2.646872,
        -3.141593,1.247736,2.242810,
        -3.141593,1.253604,0.812222,
        -3.141593,1.257816,-0.894317,
        -3.141593,1.258247,-1.953634,
        -3.141593,1.258993,-0.284099,
        -3.141593,1.262380,-0.157609,
        -3.141593,1.262791,-1.549229,
        -3.141593,1.263772,-0.533073,
        -3.141593,1.266731,-1.412461,
        -3.141593,1.268591,-3.122146,
        -3.141593,1.269614,-1.275246,
        -3.141593,1.270745,1.152533,
        -3.141593,1.273606,0.224545,
        -3.141593,1.279716,2.812101,
        -3.141593,1.283786,2.462124,
        -3.141593,1.284798,0.078209,
        -3.141593,1.285402,-2.882049,
        -3.141593,1.286936,1.886053,
        -3.141593,1.289408,0.683122,
        -3.141593,1.294736,1.756184,
        -3.141593,1.296978,-2.311511,
        -3.141593,1.298546,2.020601,
        -3.141593,1.300924,1.022478,
        -3.141593,1.303305,3.043152,
        -3.141593,1.304210,-2.084225,
        -3.141593,1.305241,-2.530249,
        -3.141593,1.307513,1.628003,
        -3.141593,1.310722,0.558267,
        -3.141593,1.314142,2.139329,
        -3.141593,1.316087,-0.777971,
        -3.141593,1.317907,-1.121023,
        -3.141593,1.321019,-2.749776,
        -3.141593,1.321609,2.687717,
        -3.141593,1.322568,1.499415,
        -3.141593,1.322972,-0.395699,
        -3.141593,1.325286,-0.994192,
        -3.141593,1.329146,-0.648513,
        -3.141593,1.329795,-0.042570,
        -3.141593,1.331651,2.344497,
        -3.141593,1.336522,0.435098,
        -3.141593,1.339410,0.900087,
        -3.141593,1.339550,1.370353,
        -3.141593,1.340441,-1.731580,
        -3.141593,1.351855,-1.855234,
        -3.141593,1.360161,-3.018404,
        -3.141593,1.361031,-1.613971,
        -3.141593,1.361196,2.914780,
        -3.141593,1.361998,-2.207254,
        -3.141593,1.362302,1.239384,
        -3.141593,1.365228,2.563808,
        -3.141593,1.365452,0.310522,
        -3.141593,1.367028,-2.420041,
        -3.141593,1.373422,-1.484281,
        -3.141593,1.374245,-1.226417,
        -3.141593,1.375020,-0.170393,
        -3.141593,1.376560,-1.355481,
        -3.141593,1.377070,-1.980984,
        -3.141593,1.377578,-2.636339,
        -3.141593,1.378096,0.774626,
        -3.141593,1.379086,-0.508192,
        -3.141593,1.382562,0.177141,
        -3.141593,1.388849,2.233847,
        -3.141593,1.389268,-0.886349,
        -3.141593,1.391324,1.109396,
        -3.141593,1.391699,-0.288725,
        -3.141593,1.392869,3.136903,
        -3.141593,1.393612,1.952016,
        -3.141593,1.395545,-2.850807,
        -3.141593,1.401926,1.822930,
        -3.141593,1.402149,2.788898,
        -3.141593,1.409552,0.647980,
        -3.141593,1.410840,0.056219,
        -3.141593,1.411944,2.444377,
        -3.141593,1.413639,1.697982,
        -3.141593,1.426479,1.573427,
        -3.141593,1.426725,2.102242,
        -3.141593,1.427868,0.984637,
        -3.141593,1.429580,-2.107232,
        -3.141593,1.431375,-2.313269,
        -3.141593,1.431907,3.015410,
        -3.141593,1.432325,-1.029054,
        -3.141593,1.433341,0.524268,
        -3.141593,1.435181,-0.739771,
        -3.141593,1.437208,-2.529085,
        -3.141593,1.439041,-0.617966,
        -3.141593,1.441473,1.448544,
        -3.141593,1.444527,2.664744,
        -3.141593,1.450513,-0.400625,
        -3.141593,1.451907,-0.066823,
        -3.141593,1.452754,-2.738631,
        -3.141593,1.454061,-2.950416,
        -3.141593,1.456736,1.323154,
        -3.141593,1.458370,0.400433,
        -3.141593,1.459314,-1.765003,
        -3.141593,1.459902,-1.145835,
        -3.141593,1.464952,2.329245,
        -3.141593,1.466050,0.862104,
        -3.141593,1.471330,-1.886160,
        -3.141593,1.475119,-1.648255,
        -3.141593,1.480878,1.194576,
        -3.141593,1.480942,0.273235,
        -3.141593,1.481086,2.890269,
        -3.141593,1.484118,-1.529156,
        -3.141593,1.484741,-1.279219,
        -3.141593,1.487681,-1.406069,
        -3.141593,1.487886,-3.066037,
        -3.141593,1.490555,2.543284,
        -3.141593,1.492424,-0.190932,
        -3.141593,1.495002,-0.839810,
        -3.141593,1.495297,2.007188,
        -3.141593,1.496414,-2.009267,
        -3.141593,1.498261,-2.425270,
        -3.141593,1.498907,-2.207327,
        -3.141593,1.500204,0.737638,
        -3.141593,1.502469,0.151156,
        -3.141593,1.507448,-0.514782,
        -3.141593,1.509287,2.204647,
        -3.141593,1.510518,1.884920,
        -3.141593,1.511948,-2.632360,
        -3.141593,1.515235,1.068565,
        -3.141593,1.522237,3.102018,
        -3.141593,1.522373,1.764991,
        -3.141593,1.522416,-0.954956,
        -3.141593,1.523102,2.765920,
        -3.141593,1.524811,-2.844718,
        -3.141593,1.527261,-0.304916,
        -3.141593,1.529788,0.613208,
        -3.141593,1.533648,1.644808,
        -3.141593,1.535018,0.028869,
        -3.141593,1.541082,2.424608,
        -3.141593,1.546010,1.523462,
        -3.141593,1.551135,-0.642994,
        -3.141593,1.553895,0.489607,
        -3.141593,1.555041,0.947813,
        -3.141593,1.559671,-1.074934,
        -3.141593,1.560345,2.985899,
        -3.141593,1.560487,-2.323400,
        -3.141593,1.561127,1.401111,
        -3.141593,1.567193,2.643172,
        -3.141593,1.568892,2.098362,
        -3.141593,1.571702,-2.098435,
        -3.141593,1.572701,-2.528730,
        -3.141593,1.573572,1.278593,
        -3.141593,1.574849,-0.091505,
        -3.141593,1.576334,0.365236,
        -3.141593,1.577029,-2.962888,
        -3.141593,1.577688,-1.814941,
        -3.141593,1.579542,-0.755917,
        -3.141593,1.579624,-1.199759,
        -3.141593,1.581515,-0.419414,
        -3.141593,1.584581,-2.738702,
        -3.141593,1.586948,-1.696296,
        -3.141593,1.588901,2.301496,
        -3.141593,1.589936,0.825020,
        -3.141593,1.590018,-1.941652,
        -3.141593,1.596968,-1.328358,
        -3.141593,1.597178,-1.575607,
        -3.141593,1.600470,-1.453196,
        -3.141593,1.601408,0.242409,
        -3.141593,1.602259,1.152656,
        -3.141593,1.603064,2.865774,
        -3.141593,1.612085,1.968802,
        -3.141593,1.612720,-3.080498,
        -3.141593,1.616061,2.521531,
        -3.141593,1.616900,-0.207457,
        -3.141593,1.620988,-0.875920,
        -3.141593,1.622114,0.701509,
        -3.141593,1.622749,0.121426,
        -3.141593,1.622993,-2.221629,
        -3.141593,1.629346,-0.542669,
        -3.141593,1.630007,1.844552,
        -3.141593,1.634747,-2.426198,
        -3.141593,1.642346,1.722617,
        -3.141593,1.643925,1.034629,
        -3.141593,1.644848,2.743774,
        -3.141593,1.645258,2.188109,
        -3.141593,1.645417,-2.634725,
        -3.141593,1.647395,3.085437,
        -3.141593,1.648538,-2.855044,
        -3.141593,1.649932,0.576808,
        -3.141593,1.652224,-0.996105,
        -3.141593,1.652871,1.600681,
        -3.141593,1.657997,-0.325127,
        -3.141593,1.661936,0.001037,
        -3.141593,1.665187,1.477402,
        -3.141593,1.665496,2.397872,
        -3.141593,1.673267,-0.665453,
        -3.141593,1.673557,0.452526,
        -3.141593,1.679277,-1.122421,
        -3.141593,1.679696,0.912482,
        -3.141593,1.682420,1.353922,
        -3.141593,1.685031,-2.009038,
        -3.141593,1.685196,2.967046,
        -3.141593,1.690107,2.620910,
        -3.141593,1.690453,-2.123076,
        -3.141593,1.691693,-1.871572,
        -3.141593,1.691817,2.061688,
        -3.141593,1.692644,1.237670,
        -3.141593,1.696386,0.328123,
        -3.141593,1.697162,-1.249026,
        -3.141593,1.697488,-2.322785,
        -3.141593,1.698726,-2.973576,
        -3.141593,1.699172,-1.749717,
        -3.141593,1.704478,-0.106295,
        -3.141593,1.707244,-0.444770,
        -3.141593,1.707500,-2.531497,
        -3.141593,1.707714,-0.787321,
        -3.141593,1.708430,-1.627552,
        -3.141593,1.710583,-1.376890,
        -3.141593,1.712415,-2.747921,
        -3.141593,1.713125,-1.503073,
        -3.141593,1.713452,0.789419,
        -3.141593,1.720612,2.279888,
        -3.141593,1.724852,0.202831,
        -3.141593,1.724868,2.846246,
        -3.141593,1.728940,1.931535,
        -3.141593,1.732838,1.127584,
        -3.141593,1.735719,-3.092485,
        -3.141593,1.739676,2.496017,
        -3.141593,1.739971,-0.228876,
        -3.141593,1.743328,0.663347,
        -3.141593,1.744927,-0.911892,
        -3.141593,1.745952,0.084733,
        -3.141593,1.748077,1.804354,
        -3.141593,1.752959,-0.567548,
        -3.141593,1.760169,1.678132,
        -3.141593,1.763887,-2.218450,
        -3.141593,1.766737,2.722752,
        -3.141593,1.767999,1.004737,
        -3.141593,1.769149,1.550298,
        -3.141593,1.770119,-2.427175,
        -3.141593,1.770460,2.154208,
        -3.141593,1.770514,3.070982,
        -3.141593,1.770735,0.537245,
        -3.141593,1.772673,-2.864760,
        -3.141593,1.773527,-1.039582,
        -3.141593,1.776419,-2.642421,
        -3.141593,1.780769,1.416330,
        -3.141593,1.786634,-0.346247,
        -3.141593,1.788829,-1.947553,
        -3.141593,1.793737,0.410270,
        -3.141593,1.794553,2.374618,
        -3.141593,1.794699,-0.694061,
        -3.141593,1.797469,-1.169105,
        -3.141593,1.803705,0.881076,
        -3.141593,1.804124,1.264249,
        -3.141593,1.807825,-2.090455,
        -3.141593,1.808822,-1.817375,
        -3.141593,1.809020,2.949741,
        -3.141593,1.812466,2.597595,
        -3.141593,1.813159,2.025395,
        -3.141593,1.813816,-1.298435,
        -3.141593,1.815016,0.278951,
        -3.141593,1.816510,-0.008631,
        -3.141593,1.817772,-1.689625,
        -3.141593,1.821360,-2.984352,
        -3.141593,1.822312,-0.127211,
        -3.141593,1.823386,-1.429807,
        -3.141593,1.823834,-1.560483,
        -3.141593,1.832060,-0.821627,
        -3.141593,1.832690,-0.468440,
        -3.141593,1.833945,-2.320278,
        -3.141593,1.834787,0.751968,
        -3.141593,1.840579,-2.536445,
        -3.141593,1.841033,-2.756932,
        -3.141593,1.846350,1.893983,
        -3.141593,1.846644,1.113766,
        -3.141593,1.846833,2.826543,
        -3.141593,1.847325,2.249051,
        -3.141593,1.854119,0.134161,
        -3.141593,1.858863,-3.106022,
        -3.141593,1.865046,1.763384,
        -3.141593,1.865619,-0.952036,
        -3.141593,1.865912,0.624453,
        -3.141593,1.866694,2.473686,
        -3.141593,1.867524,-0.240083,
        -3.141593,1.875561,1.632831,
        -3.141593,1.876049,-0.594378,
        -3.141593,1.880509,1.500204,
        -3.141593,1.883504,1.360457,
        -3.141593,1.886971,-2.196665,
        -3.141593,1.888979,0.982922,
        -3.141593,1.889815,2.701570,
        -3.141593,1.892144,-2.005340,
        -3.141593,1.892348,0.496522,
        -3.141593,1.893281,-1.083976,
        -3.141593,1.893828,3.053453,
        -3.141593,1.895117,2.121107,
        -3.141593,1.898741,-2.875735,
        -3.141593,1.904429,-2.428029,
        -3.141593,1.908408,-2.648903,
        -3.141593,1.911828,-0.364489,
        -3.141593,1.912224,0.366565,
        -3.141593,1.914854,1.214468,
        -3.141593,1.915467,-1.883143,
        -3.141593,1.915763,-0.724679,
        -3.141593,1.915790,-1.216366,
        -3.141593,1.922256,2.347806,
        -3.141593,1.922616,0.844633,
        -3.141593,1.925185,0.231170,
        -3.141593,1.927474,-1.754616,
        -3.141593,1.928342,-1.351174,
        -3.141593,1.933328,-1.623328,
        -3.141593,1.933878,0.034150,
        -3.141593,1.934585,-1.487979,
        -3.141593,1.934802,2.927634,
        -3.141593,1.934862,1.989776,
        -3.141593,1.936327,-0.099350,
        -3.141593,1.938088,2.578507,
        -3.141593,1.945677,-2.999805,
        -3.141593,1.954572,-0.859208,
        -3.141593,1.957868,-0.490784,
        -3.141593,1.958989,0.712772,
        -3.141593,1.960212,-2.307561,
        -3.141593,1.964774,1.855507,
        -3.141593,1.964971,-2.100533,
        -3.141593,1.969112,1.088273,
        -3.141593,1.971438,-2.765415,
        -3.141593,1.972312,2.805134,
        -3.141593,1.974950,2.220596,
        -3.141593,1.975135,-2.539127,
        -3.141593,1.982107,1.720829,
        -3.141593,1.983195,-3.127984,
        -3.141593,1.983318,-0.247182,
        -3.141593,1.986720,-0.993220,
        -3.141593,1.990381,1.586005,
        -3.141593,1.990470,0.580283,
        -3.141593,1.990523,1.450326,
        -3.141593,1.993909,1.312897,
        -3.141593,1.994694,2.452813,
        -3.141593,1.999226,-0.620717,
        -3.141593,2.005145,0.941012,
        -3.141593,2.007301,0.142337,
        -3.141593,2.014149,0.444138,
        -3.141593,2.014338,-1.127392,
        -3.141593,2.018131,3.021778,
        -3.141593,2.019676,-1.980944,
        -3.141593,2.019912,2.686612,
        -3.141593,2.020545,2.088672,
        -3.141593,2.026074,0.300945,
        -3.141593,2.026196,-2.889057,
        -3.141593,2.031959,-1.268178,
        -3.141593,2.032917,-2.199917,
        -3.141593,2.033572,-1.842884,
        -3.141593,2.034712,-2.419505,
        -3.141593,2.035818,-0.379479,
        -3.141593,2.036004,-0.757062,
        -3.141593,2.040501,-1.411488,
        -3.141593,2.042874,-1.701367,
        -3.141593,2.043060,-2.653746,
        -3.141593,2.043729,-1.557938,
        -3.141593,2.048538,0.806623,
        -3.141593,2.049006,1.199930,
        -3.141593,2.049502,-0.137989,
        -3.141593,2.052181,2.326959,
        -3.141593,2.052794,-0.008526,
        -3.141593,2.057608,1.951726,
        -3.141593,2.065568,2.567730,
        -3.141593,2.070371,2.879808,
        -3.141593,2.071256,-3.018961,
        -3.141593,2.079746,-0.897699,
        -3.141593,2.082784,1.042403,
        -3.141593,2.083752,-0.511086,
        -3.141593,2.084401,1.810502,
        -3.141593,2.085731,0.671081,
        -3.141593,2.097496,1.373172,
        -3.141593,2.099185,1.667006,
        -3.141593,2.100482,-2.301084,
        -3.141593,2.103118,2.194745,
        -3.141593,2.103223,1.521685,
        -3.141593,2.103695,-2.777707,
        -3.141593,2.107523,3.129581,
        -3.141593,2.108277,-2.095907,
        -3.141593,2.108891,-2.532166,
        -3.141593,2.111003,-0.268632,
        -3.141593,2.112276,-1.027246,
        -3.141593,2.114380,0.531457,
        -3.141593,2.119166,0.100552,
        -3.141593,2.119564,2.754818,
        -3.141593,2.122474,-0.648314,
        -3.141593,2.125183,2.445598,
        -3.141593,2.129501,0.238290,
        -3.141593,2.131178,0.386663,
        -3.141593,2.131223,-1.174311,
        -3.141593,2.134269,-1.948042,
        -3.141593,2.134684,0.907065,
        -3.141593,2.137615,2.988806,
        -3.141593,2.145635,-1.327708,
        -3.141593,2.146467,2.055081,
        -3.141593,2.146582,-1.793788,
        -3.141593,2.149210,-1.482321,
        -3.141593,2.152901,-1.636788,
        -3.141593,2.156337,1.151558,
        -3.141593,2.156474,-2.912065,
        -3.141593,2.157194,-0.792563,
        -3.141593,2.166724,-0.401638,
        -3.141593,2.166980,-0.134879,
        -3.141593,2.173644,-2.403429,
        -3.141593,2.175711,-2.655523,
        -3.141593,2.177132,2.600419,
        -3.141593,2.178879,0.767722,
        -3.141593,2.180752,1.908485,
        -3.141593,2.181338,2.313721,
        -3.141593,2.185580,1.287673,
        -3.141593,2.195400,-2.217546,
        -3.141593,2.200418,-3.056442,
        -3.141593,2.201829,-0.001735,
        -3.141593,2.203804,1.756176,
        -3.141593,2.211602,1.443930,
        -3.141593,2.212686,-0.544809,
        -3.141593,2.213937,0.621892,
        -3.141593,2.214075,1.600691,
        -3.141593,2.216256,1.020029,
        -3.141593,2.219060,2.886631,
        -3.141593,2.223375,-0.906185,
        -3.141593,2.226002,-1.060248,
        -3.141593,2.226413,-2.064034,
        -3.141593,2.230554,2.171334,
        -3.141593,2.232413,2.735777,
        -3.141593,2.233709,3.070830,
        -3.141593,2.234383,-2.792065,
        -3.141593,2.237183,0.468799,
        -3.141593,2.237695,-0.271388,
        -3.141593,2.239419,0.151693,
        -3.141593,2.243347,-2.521049,
        -3.141593,2.244035,-1.229456,
        -3.141593,2.244355,0.309899,
        -3.141593,2.245175,2.460236,
        -3.141593,2.245628,-1.899216,
        -3.141593,2.245973,-0.702234,
        -3.141593,2.255601,-1.728131,
        -3.141593,2.256702,-1.401367,
        -3.141593,2.261466,-1.557500,
        -3.141593,2.267390,0.878499,
        -3.141593,2.272099,2.018594,
        -3.141593,2.278618,1.170386,
        -3.141593,2.283304,-2.345584,
        -3.141593,2.285175,-2.943090,
        -3.141593,2.293000,-0.124451,
        -3.141593,2.294058,-0.421070,
        -3.141593,2.303283,1.855938,
        -3.141593,2.304968,2.608978,
        -3.141593,2.305347,2.311346,
        -3.141593,2.307188,-2.656134,
        -3.141593,2.307301,1.336200,
        -3.141593,2.309655,0.726020,
        -3.141593,2.317634,-2.185225,
        -3.141593,2.321510,1.685664,
        -3.141593,2.324376,-3.111006,
        -3.141593,2.324731,1.511186,
        -3.141593,2.329587,0.035498,
        -3.141593,2.330158,2.977264,
        -3.141593,2.335096,-0.588468,
        -3.141593,2.336046,-1.115799,
        -3.141593,2.338060,2.803613,
        -3.141593,2.338083,-0.943400,
        -3.141593,2.340210,0.562145,
        -3.141593,2.341931,1.023926,
        -3.141593,2.343265,-0.779061,
        -3.141593,2.343434,-2.013692,
        -3.141593,2.353029,-1.310529,
        -3.141593,2.354868,0.211543,
        -3.141593,2.355120,0.388593,
        -3.141593,2.355165,-1.833173,
        -3.141593,2.355462,2.151441,
        -3.141593,2.362948,-2.487477,
        -3.141593,2.363101,-2.810456,
        -3.141593,2.363394,-0.272918,
        -3.141593,2.364509,-1.644533,
        -3.141593,2.370465,2.466824,
        -3.141593,2.386110,-1.481351,
        -3.141593,2.394960,0.862133,
        -3.141593,2.395716,1.977184,
        -3.141593,2.396339,1.200900,
        -3.141593,2.407941,-2.316753,
        -3.141593,2.409450,-2.988432,
        -3.141593,2.417202,-0.103363,
        -3.141593,2.417386,-0.443097,
        -3.141593,2.417950,2.666662,
        -3.141593,2.423190,1.789783,
        -3.141593,2.423692,1.392696,
        -3.141593,2.431122,3.086379,
        -3.141593,2.431406,-2.647820,
        -3.141593,2.431584,2.307631,
        -3.141593,2.434090,1.592818,
        -3.141593,2.435993,0.683844,
        -3.141593,2.440898,-2.133300,
        -3.141593,2.442222,-1.195837,
        -3.141593,2.443112,2.883761,
        -3.141593,2.449191,-0.633651,
        -3.141593,2.450821,0.084741,
        -3.141593,2.453499,-1.008861,
        -3.141593,2.460010,-1.938011,
        -3.141593,2.460199,-0.822008,
        -3.141593,2.460902,0.490079,
        -3.141593,2.461774,-1.734913,
        -3.141593,2.463986,1.034407,
        -3.141593,2.467647,0.287667,
        -3.141593,2.479849,2.126925,
        -3.141593,2.485907,-2.835471,
        -3.141593,2.485945,-1.374853,
        -3.141593,2.487747,2.505912,
        -3.141593,2.487957,-0.274279,
        -3.141593,2.489741,-2.468428,
        -3.141593,2.512538,1.239900,
        -3.141593,2.517528,1.923289,
        -3.141593,2.518454,-1.565852,
        -3.141593,2.519348,0.843503,
        -3.141593,2.522766,-3.059231,
        -3.141593,2.528686,2.737228,
        -3.141593,2.534798,-2.266671,
        -3.141593,2.536533,1.466274,
        -3.141593,2.538504,1.699812,
        -3.141593,2.539516,-0.476942,
        -3.141593,2.539532,-0.074548,
        -3.141593,2.543197,2.990252,
        -3.141593,2.554022,2.320619,
        -3.141593,2.556919,-2.655005,
        -3.141593,2.558146,0.627191,
        -3.141593,2.558223,-1.164189,
        -3.141593,2.563452,-2.044028,
        -3.141593,2.568462,0.150655,
        -3.141593,2.569130,-0.700837,
        -3.141593,2.570050,-1.804492,
        -3.141593,2.572710,-0.934665,
        -3.141593,2.575994,0.391718,
        -3.141593,2.584774,1.046703,
        -3.141593,2.601432,2.102535,
        -3.141593,2.603559,2.555114,
        -3.141593,2.604150,-2.885208,
        -3.141593,2.608630,-1.378011,
        -3.141593,2.611256,-0.279110,
        -3.141593,2.617344,-2.439480,
        -3.141593,2.627402,1.293552,
        -3.141593,2.635396,1.848850,
        -3.141593,2.635686,2.834950,
        -3.141593,2.637087,3.132730,
        -3.141593,2.642029,0.813274,
        -3.141593,2.645260,1.570203,
        -3.141593,2.646182,-1.614036,
        -3.141593,2.660113,-2.186513,
        -3.141593,2.660483,-0.034608,
        -3.141593,2.661713,-0.529511,
        -3.141593,2.672720,-1.110422,
        -3.141593,2.676626,2.335353,
        -3.141593,2.676638,0.540364,
        -3.141593,2.676780,-1.901158,
        -3.141593,2.681411,-2.667965,
        -3.141593,2.682152,0.245751,
        -3.141593,2.682859,-0.817071,
        -3.141593,2.705107,1.062142,
        -3.141593,2.717705,2.625995,
        -3.141593,2.721001,2.060097,
        -3.141593,2.722319,-2.954069,
        -3.141593,2.729840,-1.368237,
        -3.141593,2.734349,-0.289298,
        -3.141593,2.738409,2.975671,
        -3.141593,2.740366,1.374341,
        -3.141593,2.744633,-2.391311,
        -3.141593,2.747398,1.727166,
        -3.141593,2.763799,0.760372,
        -3.141593,2.768238,-1.679243,
        -3.141593,2.779245,-2.046268,
        -3.141593,2.780208,-0.626453,
        -3.141593,2.780471,0.027732,
        -3.141593,2.784696,-1.013128,
        -3.141593,2.789817,0.395431,
        -3.141593,2.801154,2.353143,
        -3.141593,2.807091,-2.687477,
        -3.141593,2.825762,1.085035,
        -3.141593,2.830209,2.734943,
        -3.141593,2.833517,-3.085508,
        -3.141593,2.838301,1.980061,
        -3.141593,2.849871,1.512012,
        -3.141593,2.850089,-1.343899,
        -3.141593,2.856905,-0.318441,
        -3.141593,2.870445,-2.291001,
        -3.141593,2.885017,0.655562,
        -3.141593,2.886428,-1.793217,
        -3.141593,2.890337,-0.825633,
        -3.141593,2.898932,0.139973,
        -3.141593,2.931927,-2.728768,
        -3.141593,2.932096,2.377790,
        -3.141593,2.937409,2.947245,
        -3.141593,2.948048,1.133127,
        -3.141593,2.950896,1.797889,
        -3.141593,2.970019,-1.282382,
        -3.141593,2.978333,-0.412090,
        -3.141593,2.996437,-2.089121,
        -3.141593,3.011274,0.434110,
        -3.141593,3.050410,3.138492,
        -3.141593,3.064492,1.647259,
        -3.141593,3.088868,-0.989501,
        -3.036873,0.053884,-2.408394,
        -3.036873,0.077645,1.453433,
        -3.036873,0.088408,-0.311160,
        -3.036873,0.142314,2.590387,
        -3.036873,0.147206,-1.385164,
        -3.036873,0.163506,0.573203,
        -3.036873,0.165485,-2.875396,
        -3.036873,0.182564,-2.125821,
        -3.036873,0.190569,1.929141,
        -3.036873,0.201102,1.249082,
        -3.036873,0.201397,-0.716646,
        -3.036873,0.210568,-0.069559,
        -3.036873,0.251672,2.848900,
        -3.036873,0.266262,-1.678068,
        -3.036873,0.269108,-1.215563,
        -3.036873,0.273777,-2.506756,
        -3.036873,0.279854,0.814581,
        -3.036873,0.281664,-2.969147,
        -3.036873,0.283625,2.405603,
        -3.036873,0.284058,0.357350,
        -3.036873,0.294107,1.593880,
        -3.036873,0.302491,-0.400410,
        -3.036873,0.308059,2.021491,
        -3.036873,0.312219,-2.072057,
        -3.036873,0.321905,-0.829121,
        -3.036873,0.329984,1.201621,
        -3.036873,0.339714,-0.013712,
        -3.036873,0.368784,2.974974,
        -3.036873,0.376428,-1.495532,
        -3.036873,0.378103,2.646436,
        -3.036873,0.385088,0.590355,
        -3.036873,0.386248,-2.690943,
        -3.036873,0.389851,-2.340031,
        -3.036873,0.391502,-1.171374,
        -3.036873,0.398249,-1.812694,
        -3.036873,0.400542,-2.995729,
        -3.036873,0.404273,0.918163,
        -3.036873,0.405617,1.773533,
        -3.036873,0.409009,-0.570243,
        -3.036873,0.410700,0.270433,
        -3.036873,0.415893,1.464890,
        -3.036873,0.420680,2.088164,
        -3.036873,0.424766,-0.258729,
        -3.036873,0.426507,2.367872,
        -3.036873,0.441292,-0.890727,
        -3.036873,0.451839,-2.078193,
        -3.036873,0.459270,1.191562,
        -3.036873,0.470622,0.014002,
        -3.036873,0.481292,-2.529098,
        -3.036873,0.482460,2.849843,
        -3.036873,0.492205,-1.397963,
        -3.036873,0.493314,3.117264,
        -3.036873,0.497870,0.727828,
        -3.036873,0.499783,-1.654379,
        -3.036873,0.502651,0.463333,
        -3.036873,0.505463,2.594557,
        -3.036873,0.506104,-2.874910,
        -3.036873,0.513579,-1.147354,
        -3.036873,0.518569,1.644237,
        -3.036873,0.519001,1.900488,
        -3.036873,0.519277,-0.673809,
        -3.036873,0.524769,-0.421955,
        -3.036873,0.525564,-2.288838,
        -3.036873,0.532541,0.978737,
        -3.036873,0.534610,2.154808,
        -3.036873,0.535020,-1.891846,
        -3.036873,0.540208,1.400497,
        -3.036873,0.542006,0.228337,
        -3.036873,0.550762,-0.180771,
        -3.036873,0.560531,2.378474,
        -3.036873,0.561629,-0.924576,
        -3.036873,0.567465,-2.679392,
        -3.036873,0.588527,1.181755,
        -3.036873,0.591643,-2.098364,
        -3.036873,0.594963,2.997575,
        -3.036873,0.599643,-3.045235,
        -3.036873,0.599962,2.773556,
        -3.036873,0.602620,0.029270,
        -3.036873,0.605711,0.598454,
        -3.036873,0.607708,-1.541026,
        -3.036873,0.609360,-2.464379,
        -3.036873,0.610738,-1.326786,
        -3.036873,0.616968,0.820489,
        -3.036873,0.626379,1.773610,
        -3.036873,0.628626,-1.751180,
        -3.036873,0.630671,0.391949,
        -3.036873,0.632622,-0.737642,
        -3.036873,0.632708,2.566722,
        -3.036873,0.632797,1.989858,
        -3.036873,0.634455,-0.534790,
        -3.036873,0.636205,1.561170,
        -3.036873,0.637681,-1.123423,
        -3.036873,0.641441,-2.842109,
        -3.036873,0.644807,-0.327139,
        -3.036873,0.652440,2.197129,
        -3.036873,0.660897,-2.278593,
        -3.036873,0.664008,1.360091,
        -3.036873,0.670850,-1.945257,
        -3.036873,0.671396,1.004859,
        -3.036873,0.673985,0.203855,
        -3.036873,0.680326,-0.133370,
        -3.036873,0.685083,-0.942363,
        -3.036873,0.688204,-2.632204,
        -3.036873,0.689777,2.385011,
        -3.036873,0.700065,3.115430,
        -3.036873,0.704080,2.911636,
        -3.036873,0.712883,1.174216,
        -3.036873,0.715701,0.699906,
        -3.036873,0.718851,-1.442215,
        -3.036873,0.721019,2.721536,
        -3.036873,0.725300,-2.975134,
        -3.036873,0.727136,-1.633288,
        -3.036873,0.727399,-2.120567,
        -3.036873,0.730751,0.522488,
        -3.036873,0.732892,-1.254330,
        -3.036873,0.734333,-2.446184,
        -3.036873,0.735125,0.037270,
        -3.036873,0.737893,1.866463,
        -3.036873,0.740650,1.681075,
        -3.036873,0.745971,-0.781286,
        -3.036873,0.749432,-0.442559,
        -3.036873,0.750102,2.050636,
        -3.036873,0.751046,0.863821,
        -3.036873,0.757018,-0.608833,
        -3.036873,0.757046,-1.816019,
        -3.036873,0.757389,1.501184,
        -3.036873,0.759800,2.546181,
        -3.036873,0.759859,0.351923,
        -3.036873,0.761141,-2.781102,
        -3.036873,0.769035,-1.078574,
        -3.036873,0.770166,-0.266604,
        -3.036873,0.772218,2.227380,
        -3.036873,0.788382,1.331051,
        -3.036873,0.790269,-2.282409,
        -3.036873,0.803410,-1.984567,
        -3.036873,0.803813,0.191520,
        -3.036873,0.806832,3.022155,
        -3.036873,0.807476,-2.597675,
        -3.036873,0.812512,-0.107335,
        -3.036873,0.814161,1.010949,
        -3.036873,0.817732,-3.084486,
        -3.036873,0.817846,2.390688,
        -3.036873,0.819323,2.845960,
        -3.036873,0.819430,-0.915017,
        -3.036873,0.829198,-1.532491,
        -3.036873,0.829531,0.665695,
        -3.036873,0.829816,1.167147,
        -3.036873,0.833613,-1.357492,
        -3.036873,0.842945,-2.907318,
        -3.036873,0.844883,2.679708,
        -3.036873,0.848189,-1.705003,
        -3.036873,0.849888,1.767901,
        -3.036873,0.851126,0.493370,
        -3.036873,0.852911,1.933394,
        -3.036873,0.857516,-2.432863,
        -3.036873,0.858261,-1.191729,
        -3.036873,0.858279,-2.138895,
        -3.036873,0.859723,-0.734928,
        -3.036873,0.859754,1.603292,
        -3.036873,0.861655,-0.394821,
        -3.036873,0.864300,0.043549,
        -3.036873,0.866940,0.851229,
        -3.036873,0.869748,2.095969,
        -3.036873,0.877352,-0.574668,
        -3.036873,0.879254,-2.733338,
        -3.036873,0.882554,1.443757,
        -3.036873,0.884060,0.334199,
        -3.036873,0.885626,-1.866061,
        -3.036873,0.889834,2.529957,
        -3.036873,0.893140,2.251698,
        -3.036873,0.897463,-0.240822,
        -3.036873,0.898758,-1.031948,
        -3.036873,0.913918,1.287618,
        -3.036873,0.916555,-2.287173,
        -3.036873,0.918305,3.107845,
        -3.036873,0.919522,2.948277,
        -3.036873,0.927138,-2.569686,
        -3.036873,0.929031,0.185383,
        -3.036873,0.933766,-3.015917,
        -3.036873,0.934873,-2.010822,
        -3.036873,0.935636,-1.445855,
        -3.036873,0.936702,-0.866461,
        -3.036873,0.937395,0.729108,
        -3.036873,0.937910,2.787575,
        -3.036873,0.939144,-0.092962,
        -3.036873,0.941718,1.125085,
        -3.036873,0.942873,-1.607680,
        -3.036873,0.943701,0.980328,
        -3.036873,0.947774,2.395256,
        -3.036873,0.952433,0.584347,
        -3.036873,0.954456,-1.287293,
        -3.036873,0.958731,-2.852552,
        -3.036873,0.961838,1.839284,
        -3.036873,0.961898,-0.469425,
        -3.036873,0.965426,1.686615,
        -3.036873,0.970108,1.990771,
        -3.036873,0.972820,-1.762780,
        -3.036873,0.973500,-0.711881,
        -3.036873,0.975253,2.641082,
        -3.036873,0.977824,0.441928,
        -3.036873,0.981485,-2.424766,
        -3.036873,0.981877,1.537100,
        -3.036873,0.982609,-1.136920,
        -3.036873,0.984265,-2.156073,
        -3.036873,0.989150,0.045493,
        -3.036873,0.993052,2.137478,
        -3.036873,0.997766,-2.693871,
        -3.036873,0.999847,-0.334738,
        -3.036873,1.001920,0.848054,
        -3.036873,1.006146,1.389734,
        -3.036873,1.012180,0.304066,
        -3.036873,1.016938,2.270674,
        -3.036873,1.017371,-0.979790,
        -3.036873,1.019495,-1.894487,
        -3.036873,1.023796,-0.583691,
        -3.036873,1.026525,2.506945,
        -3.036873,1.027004,-0.203653,
        -3.036873,1.028027,3.036573,
        -3.036873,1.031435,1.241021,
        -3.036873,1.032316,2.884221,
        -3.036873,1.035863,-3.110475,
        -3.036873,1.042085,-2.292950,
        -3.036873,1.043272,-1.522986,
        -3.036873,1.046265,-2.958847,
        -3.036873,1.049272,-1.376344,
        -3.036873,1.050039,-2.552086,
        -3.036873,1.053682,0.689863,
        -3.036873,1.055458,0.170394,
        -3.036873,1.056622,-0.831866,
        -3.036873,1.057515,1.097998,
        -3.036873,1.058706,-2.036438,
        -3.036873,1.061389,-1.672248,
        -3.036873,1.063758,2.743093,
        -3.036873,1.070157,-0.073422,
        -3.036873,1.072154,1.761506,
        -3.036873,1.074029,1.909193,
        -3.036873,1.074204,0.545276,
        -3.036873,1.074714,-2.805147,
        -3.036873,1.076262,0.966104,
        -3.036873,1.077642,-1.227073,
        -3.036873,1.084151,1.618525,
        -3.036873,1.085085,2.377673,
        -3.036873,1.086822,-0.441112,
        -3.036873,1.087142,2.058815,
        -3.036873,1.098383,-0.696430,
        -3.036873,1.099138,-1.084033,
        -3.036873,1.103208,1.477773,
        -3.036873,1.103783,0.408254,
        -3.036873,1.106275,-2.420800,
        -3.036873,1.108079,2.612943,
        -3.036873,1.109345,-2.172718,
        -3.036873,1.115957,-1.787597,
        -3.036873,1.118930,0.045983,
        -3.036873,1.119777,0.834702,
        -3.036873,1.121943,-2.669042,
        -3.036873,1.124930,2.219917,
        -3.036873,1.125217,1.336294,
        -3.036873,1.130958,2.972931,
        -3.036873,1.131400,-0.306629,
        -3.036873,1.137135,-0.936514,
        -3.036873,1.139125,-1.924528,
        -3.036873,1.139877,0.276382,
        -3.036873,1.142003,-3.058022,
        -3.036873,1.147363,-0.558448,
        -3.036873,1.150348,-0.176418,
        -3.036873,1.151198,1.197016,
        -3.036873,1.153930,-1.597104,
        -3.036873,1.155389,2.837148,
        -3.036873,1.155458,-1.460623,
        -3.036873,1.157331,-2.906362,
        -3.036873,1.158594,2.487618,
        -3.036873,1.162103,-1.322553,
        -3.036873,1.165407,3.092861,
        -3.036873,1.168025,0.717075,
        -3.036873,1.168406,-2.301519,
        -3.036873,1.174782,-2.538755,
        -3.036873,1.177998,1.061256,
        -3.036873,1.178898,1.827305,
        -3.036873,1.181331,-2.060330,
        -3.036873,1.182540,-0.798952,
        -3.036873,1.184569,0.147517,
        -3.036873,1.184734,1.968919,
        -3.036873,1.184804,0.593386,
        -3.036873,1.188278,1.690808,
        -3.036873,1.194360,2.110440,
        -3.036873,1.194776,2.710455,
        -3.036873,1.197745,-1.173705,
        -3.036873,1.199898,-2.777257,
        -3.036873,1.201164,2.355931,
        -3.036873,1.203760,1.556404,
        -3.036873,1.205845,-0.415993,
        -3.036873,1.206585,0.933648,
        -3.036873,1.206950,-0.037248,
        -3.036873,1.209341,0.471162,
        -3.036873,1.213483,-1.039417,
        -3.036873,1.218146,-0.672643,
        -3.036873,1.221608,1.421669,
        -3.036873,1.226043,-1.693533,
        -3.036873,1.232870,-1.822773,
        -3.036873,1.234375,-2.421194,
        -3.036873,1.234531,-2.190148,
        -3.036873,1.239414,0.348959,
        -3.036873,1.240183,2.587255,
        -3.036873,1.242312,2.941234,
        -3.036873,1.244075,1.286644,
        -3.036873,1.245961,-2.997461,
        -3.036873,1.246495,-2.646872,
        -3.036873,1.247736,2.242810,
        -3.036873,1.253604,0.812222,
        -3.036873,1.257816,-0.894317,
        -3.036873,1.258247,-1.953634,
        -3.036873,1.258993,-0.284099,
        -3.036873,1.262380,-0.157609,
        -3.036873,1.262791,-1.549229,
        -3.036873,1.263772,-0.533073,
        -3.036873,1.266731,-1.412461,
        -3.036873,1.268591,-3.122146,
        -3.036873,1.269614,-1.275246,
        -3.036873,1.270745,1.152533,
        -3.036873,1.273606,0.224545,
        -3.036873,1.279716,2.812101,
        -3.036873,1.283786,2.462124,
        -3.036873,1.284798,0.078209,
        -3.036873,1.285402,-2.882049,
        -3.036873,1.286936,1.886053,
        -3.036873,1.289408,0.683122,
        -3.036873,1.294736,1.756184,
        -3.036873,1.296978,-2.311511,
        -3.036873,1.298546,2.020601,
        -3.036873,1.300924,1.022478,
        -3.036873,1.303305,3.043152,
        -3.036873,1.304210,-2.084225,
        -3.036873,1.305241,-2.530249,
        -3.036873,1.307513,1.628003,
        -3.036873,1.310722,0.558267,
        -3.036873,1.314142,2.139329,
        -3.036873,1.316087,-0.777971,
        -3.036873,1.317907,-1.121023,
        -3.036873,1.321019,-2.749776,
        -3.036873,1.321609,2.687717,
        -3.036873,1.322568,1.499415,
        -3.036873,1.322972,-0.395699,
        -3.036873,1.325286,-0.994192,
        -3.036873,1.329146,-0.648513,
        -3.036873,1.329795,-0.042570,
        -3.036873,1.331651,2.344497,
        -3.036873,1.336522,0.435098,
        -3.036873,1.339410,0.900087,
        -3.036873,1.339550,1.370353,
        -3.036873,1.340441,-1.731580,
        -3.036873,1.351855,-1.855234,
        -3.036873,1.360161,-3.018404,
        -3.036873,1.361031,-1.613971,
        -3.036873,1.361196,2.914780,
        -3.036873,1.361998,-2.207254,
        -3.036873,1.362302,1.239384,
        -3.036873,1.365228,2.563808,
        -3.036873,1.365452,0.310522,
        -3.036873,1.367028,-2.420041,
        -3.036873,1.373422,-1.484281,
        -3.036873,1.374245,-1.226417,
        -3.036873,1.375020,-0.170393,
        -3.036873,1.376560,-1.355481,
        -3.036873,1.377070,-1.980984,
        -3.036873,1.377578,-2.636339,
        -3.036873,1.378096,0.774626,
        -3.036873,1.379086,-0.508192,
        -3.036873,1.382562,0.177141,
        -3.036873,1.388849,2.233847,
        -3.036873,1.389268,-0.886349,
        -3.036873,1.391324,1.109396,
        -3.036873,1.391699,-0.288725,
        -3.036873,1.392869,3.136903,
        -3.036873,1.393612,1.952016,
        -3.036873,1.395545,-2.850807,
        -3.036873,1.401926,1.822930,
        -3.036873,1.402149,2.788898,
        -3.036873,1.409552,0.647980,
        -3.036873,1.410840,0.056219,
        -3.036873,1.411944,2.444377,
        -3.036873,1.413639,1.697982,
        -3.036873,1.426479,1.573427,
        -3.036873,1.426725,2.102242,
        -3.036873,1.427868,0.984637,
        -3.036873,1.429580,-2.107232,
        -3.036873,1.431375,-2.313269,
        -3.036873,1.431907,3.015410,
        -3.036873,1.432325,-1.029054,
        -3.036873,1.433341,0.524268,
        -3.036873,1.435181,-0.739771,
        -3.036873,1.437208,-2.529085,
        -3.036873,1.439041,-0.617966,
        -3.036873,1.441473,1.448544,
        -3.036873,1.444527,2.664744,
        -3.036873,1.450513,-0.400625,
        -3.036873,1.451907,-0.066823,
        -3.036873,1.452754,-2.738631,
        -3.036873,1.454061,-2.950416,
        -3.036873,1.456736,1.323154,
        -3.036873,1.458370,0.400433,
        -3.036873,1.459314,-1.765003,
        -3.036873,1.459902,-1.145835,
        -3.036873,1.464952,2.329245,
        -3.036873,1.466050,0.862104,
        -3.036873,1.471330,-1.886160,
        -3.036873,1.475119,-1.648255,
        -3.036873,1.480878,1.194576,
        -3.036873,1.480942,0.273235,
        -3.036873,1.481086,2.890269,
        -3.036873,1.484118,-1.529156,
        -3.036873,1.484741,-1.279219,
        -3.036873,1.487681,-1.406069,
        -3.036873,1.487886,-3.066037,
        -3.036873,1.490555,2.543284,
        -3.036873,1.492424,-0.190932,
        -3.036873,1.495002,-0.839810,
        -3.036873,1.495297,2.007188,
        -3.036873,1.496414,-2.009267,
        -3.036873,1.498261,-2.425270,
        -3.036873,1.498907,-2.207327,
        -3.036873,1.500204,0.737638,
        -3.036873,1.502469,0.151156,
        -3.036873,1.507448,-0.514782,
        -3.036873,1.509287,2.204647,
        -3.036873,1.510518,1.884920,
        -3.036873,1.511948,-2.632360,
        -3.036873,1.515235,1.068565,
        -3.036873,1.522237,3.102018,
        -3.036873,1.522373,1.764991,
        -3.036873,1.522416,-0.954956,
        -3.036873,1.523102,2.765920,
        -3.036873,1.524811,-2.844718,
        -3.036873,1.527261,-0.304916,
        -3.036873,1.529788,0.613208,
        -3.036873,1.533648,1.644808,
        -3.036873,1.535018,0.028869,
        -3.036873,1.541082,2.424608,
        -3.036873,1.546010,1.523462,
        -3.036873,1.551135,-0.642994,
        -3.036873,1.553895,0.489607,
        -3.036873,1.555041,0.947813,
        -3.036873,1.559671,-1.074934,
        -3.036873,1.560345,2.985899,
        -3.036873,1.560487,-2.323400,
        -3.036873,1.561127,1.401111,
        -3.036873,1.567193,2.643172,
        -3.036873,1.568892,2.098362,
        -3.036873,1.571702,-2.098435,
        -3.036873,1.572701,-2.528730,
        -3.036873,1.573572,1.278593,
        -3.036873,1.574849,-0.091505,
        -3.036873,1.576334,0.365236,
        -3.036873,1.577029,-2.962888,
        -3.036873,1.577688,-1.814941,
        -3.036873,1.579542,-0.755917,
        -3.036873,1.579624,-1.199759,
        -3.036873,1.581515,-0.419414,
        -3.036873,1.584581,-2.738702,
        -3.036873,1.586948,-1.696296,
        -3.036873,1.588901,2.301496,
        -3.036873,1.589936,0.825020,
        -3.036873,1.590018,-1.941652,
        -3.036873,1.596968,-1.328358,
        -3.036873,1.597178,-1.575607,
        -3.036873,1.600470,-1.453196,
        -3.036873,1.601408,0.242409,
        -3.036873,1.602259,1.152656,
        -3.036873,1.603064,2.865774,
        -3.036873,1.612085,1.968802,
        -3.036873,1.612720,-3.080498,
        -3.036873,1.616061,2.521531,
        -3.036873,1.616900,-0.207457,
        -3.036873,1.620988,-0.875920,
        -3.036873,1.622114,0.701509,
        -3.036873,1.622749,0.121426,
        -3.036873,1.622993,-2.221629,
        -3.036873,1.629346,-0.542669,
        -3.036873,1.630007,1.844552,
        -3.036873,1.634747,-2.426198,
        -3.036873,1.642346,1.722617,
        -3.036873,1.643925,1.034629,
        -3.036873,1.644848,2.743774,
        -3.036873,1.645258,2.188109,
        -3.036873,1.645417,-2.634725,
        -3.036873,1.647395,3.085437,
        -3.036873,1.648538,-2.855044,
        -3.036873,1.649932,0.576808,
        -3.036873,1.652224,-0.996105,
        -3.036873,1.652871,1.600681,
        -3.036873,1.657997,-0.325127,
        -3.036873,1.661936,0.001037,
        -3.036873,1.665187,1.477402,
        -3.036873,1.665496,2.397872,
        -3.036873,1.673267,-0.665453,
        -3.036873,1.673557,0.452526,
        -3.036873,1.679277,-1.122421,
        -3.036873,1.679696,0.912482,
        -3.036873,1.682420,1.353922,
        -3.036873,1.685031,-2.009038,
        -3.036873,1.685196,2.967046,
        -3.036873,1.690107,2.620910,
        -3.036873,1.690453,-2.123076,
        -3.036873,1.691693,-1.871572,
        -3.036873,1.691817,2.061688,
        -3.036873,1.692644,1.237670,
        -3.036873,1.696386,0.328123,
        -3.036873,1.697162,-1.249026,
        -3.036873,1.697488,-2.322785,
        -3.036873,1.698726,-2.973576,
        -3.036873,1.699172,-1.749717,
        -3.036873,1.704478,-0.106295,
        -3.036873,1.707244,-0.444770,
        -3.036873,1.707500,-2.531497,
        -3.036873,1.707714,-0.787321,
        -3.036873,1.708430,-1.627552,
        -3.036873,1.710583,-1.376890,
        -3.036873,1.712415,-2.747921,
        -3.036873,1.713125,-1.503073,
        -3.036873,1.713452,0.789419,
        -3.036873,1.720612,2.279888,
        -3.036873,1.724852,0.202831,
        -3.036873,1.724868,2.846246,
        -3.036873,1.728940,1.931535,
        -3.036873,1.732838,1.127584,
        -3.036873,1.735719,-3.092485,
        -3.036873,1.739676,2.496017,
        -3.036873,1.739971,-0.228876,
        -3.036873,1.743328,0.663347,
        -3.036873,1.744927,-0.911892,
        -3.036873,1.745952,0.084733,
        -3.036873,1.748077,1.804354,
        -3.036873,1.752959,-0.567548,
        -3.036873,1.760169,1.678132,
        -3.036873,1.763887,-2.218450,
        -3.036873,1.766737,2.722752,
        -3.036873,1.767999,1.004737,
        -3.036873,1.769149,1.550298,
        -3.036873,1.770119,-2.427175,
        -3.036873,1.770460,2.154208,
        -3.036873,1.770514,3.070982,
        -3.036873,1.770735,0.537245,
        -3.036873,1.772673,-2.864760,
        -3.036873,1.773527,-1.039582,
        -3.036873,1.776419,-2.642421,
        -3.036873,1.780769,1.416330,
        -3.036873,1.786634,-0.346247,
        -3.036873,1.788829,-1.947553,
        -3.036873,1.793737,0.410270,
        -3.036873,1.794553,2.374618,
        -3.036873,1.794699,-0.694061,
        -3.036873,1.797469,-1.169105,
        -3.036873,1.803705,0.881076,
        -3.036873,1.804124,1.264249,
        -3.036873,1.807825,-2.090455,
        -3.036873,1.808822,-1.817375,
        -3.036873,1.809020,2.949741,
        -3.036873,1.812466,2.597595,
        -3.036873,1.813159,2.025395,
        -3.036873,1.813816,-1.298435,
        -3.036873,1.815016,0.278951,
        -3.036873,1.816510,-0.008631,
        -3.036873,1.817772,-1.689625,
        -3.036873,1.821360,-2.984352,
        -3.036873,1.822312,-0.127211,
        -3.036873,1.823386,-1.429807,
        -3.036873,1.823834,-1.560483,
        -3.036873,1.832060,-0.821627,
        -3.036873,1.832690,-0.468440,
        -3.036873,1.833945,-2.320278,
        -3.036873,1.834787,0.751968,
        -3.036873,1.840579,-2.536445,
        -3.036873,1.841033,-2.756932,
        -3.036873,1.846350,1.893983,
        -3.036873,1.846644,1.113766,
        -3.036873,1.846833,2.826543,
        -3.036873,1.847325,2.249051,
        -3.036873,1.854119,0.134161,
        -3.036873,1.858863,-3.106022,
        -3.036873,1.865046,1.763384,
        -3.036873,1.865619,-0.952036,
        -3.036873,1.865912,0.624453,
        -3.036873,1.866694,2.473686,
        -3.036873,1.867524,-0.240083,
        -3.036873,1.875561,1.632831,
        -3.036873,1.876049,-0.594378,
        -3.036873,1.880509,1.500204,
        -3.036873,1.883504,1.360457,
        -3.036873,1.886971,-2.196665,
        -3.036873,1.888979,0.982922,
        -3.036873,1.889815,2.701570,
        -3.036873,1.892144,-2.005340,
        -3.036873,1.892348,0.496522,
        -3.036873,1.893281,-1.083976,
        -3.036873,1.893828,3.053453,
        -3.036873,1.895117,2.121107,
        -3.036873,1.898741,-2.875735,
        -3.036873,1.904429,-2.428029,
        -3.036873,1.908408,-2.648903,
        -3.036873,1.911828,-0.364489,
        -3.036873,1.912224,0.366565,
        -3.036873,1.914854,1.214468,
        -3.036873,1.915467,-1.883143,
        -3.036873,1.915763,-0.724679,
        -3.036873,1.915790,-1.216366,
        -3.036873,1.922256,2.347806,
        -3.036873,1.922616,0.844633,
        -3.036873,1.925185,0.231170,
        -3.036873,1.927474,-1.754616,
        -3.036873,1.928342,-1.351174,
        -3.036873,1.933328,-1.623328,
        -3.036873,1.933878,0.034150,
        -3.036873,1.934585,-1.487979,
        -3.036873,1.934802,2.927634,
        -3.036873,1.934862,1.989776,
        -3.036873,1.936327,-0.099350,
        -3.036873,1.938088,2.578507,
        -3.036873,1.945677,-2.999805,
        -3.036873,1.954572,-0.859208,
        -3.036873,1.957868,-0.490784,
        -3.036873,1.958989,0.712772,
        -3.036873,1.960212,-2.307561,
        -3.036873,1.964774,1.855507,
        -3.036873,1.964971,-2.100533,
        -3.036873,1.969112,1.088273,
        -3.036873,1.971438,-2.765415,
        -3.036873,1.972312,2.805134,
        -3.036873,1.974950,2.220596,
        -3.036873,1.975135,-2.539127,
        -3.036873,1.982107,1.720829,
        -3.036873,1.983195,-3.127984,
        -3.036873,1.983318,-0.247182,
        -3.036873,1.986720,-0.993220,
        -3.036873,1.990381,1.586005,
        -3.036873,1.990470,0.580283,
        -3.036873,1.990523,1.450326,
        -3.036873,1.993909,1.312897,
        -3.036873,1.994694,2.452813,
        -3.036873,1.999226,-0.620717,
        -3.036873,2.005145,0.941012,
        -3.036873,2.007301,0.142337,
        -3.036873,2.014149,0.444138,
        -3.036873,2.014338,-1.127392,
        -3.036873,2.018131,3.021778,
        -3.036873,2.019676,-1.980944,
        -3.036873,2.019912,2.686612,
        -3.036873,2.020545,2.088672,
        -3.036873,2.026074,0.300945,
        -3.036873,2.026196,-2.889057,
        -3.036873,2.031959,-1.268178,
        -3.036873,2.032917,-2.199917,
        -3.036873,2.033572,-1.842884,
        -3.036873,2.034712,-2.419505,
        -3.036873,2.035818,-0.379479,
        -3.036873,2.036004,-0.757062,
        -3.036873,2.040501,-1.411488,
        -3.036873,2.042874,-1.701367,
        -3.036873,2.043060,-2.653746,
        -3.036873,2.043729,-1.557938,
        -3.036873,2.048538,0.806623,
        -3.036873,2.049006,1.199930,
        -3.036873,2.049502,-0.137989,
        -3.036873,2.052181,2.326959,
        -3.036873,2.052794,-0.008526,
        -3.036873,2.057608,1.951726,
        -3.036873,2.065568,2.567730,
        -3.036873,2.070371,2.879808,
        -3.036873,2.071256,-3.018961,
        -3.036873,2.079746,-0.897699,
        -3.036873,2.082784,1.042403,
        -3.036873,2.083752,-0.511086,
        -3.036873,2.084401,1.810502,
        -3.036873,2.085731,0.671081,
        -3.036873,2.097496,1.373172,
        -3.036873,2.099185,1.667006,
        -3.036873,2.100482,-2.301084,
        -3.036873,2.103118,2.194745,
        -3.036873,2.103223,1.521685,
        -3.036873,2.103695,-2.777707,
        -3.036873,2.107523,3.129581,
        -3.036873,2.108277,-2.095907,
        -3.036873,2.108891,-2.532166,
        -3.036873,2.111003,-0.268632,
        -3.036873,2.112276,-1.027246,
        -3.036873,2.114380,0.531457,
        -3.036873,2.119166,0.100552,
        -3.036873,2.119564,2.754818,
        -3.036873,2.122474,-0.648314,
        -3.036873,2.125183,2.445598,
        -3.036873,2.129501,0.238290,
        -3.036873,2.131178,0.386663,
        -3.036873,2.131223,-1.174311,
        -3.036873,2.134269,-1.948042,
        -3.036873,2.134684,0.907065,
        -3.036873,2.137615,2.988806,
        -3.036873,2.145635,-1.327708,
        -3.036873,2.146467,2.055081,
        -3.036873,2.146582,-1.793788,
        -3.036873,2.149210,-1.482321,
        -3.036873,2.152901,-1.636788,
        -3.036873,2.156337,1.151558,
        -3.036873,2.156474,-2.912065,
        -3.036873,2.157194,-0.792563,
        -3.036873,2.166724,-0.401638,
        -3.036873,2.166980,-0.134879,
        -3.036873,2.173644,-2.403429,
        -3.036873,2.175711,-2.655523,
        -3.036873,2.177132,2.600419,
        -3.036873,2.178879,0.767722,
        -3.036873,2.180752,1.908485,
        -3.036873,2.181338,2.313721,
        -3.036873,2.185580,1.287673,
        -3.036873,2.195400,-2.217546,
        -3.036873,2.200418,-3.056442,
        -3.036873,2.201829,-0.001735,
        -3.036873,2.203804,1.756176,
        -3.036873,2.211602,1.443930,
        -3.036873,2.212686,-0.544809,
        -3.036873,2.213937,0.621892,
        -3.036873,2.214075,1.600691,
        -3.036873,2.216256,1.020029,
        -3.036873,2.219060,2.886631,
        -3.036873,2.223375,-0.906185,
        -3.036873,2.226002,-1.060248,
        -3.036873,2.226413,-2.064034,
        -3.036873,2.230554,2.171334,
        -3.036873,2.232413,2.735777,
        -3.036873,2.233709,3.070830,
        -3.036873,2.234383,-2.792065,
        -3.036873,2.237183,0.468799,
        -3.036873,2.237695,-0.271388,
        -3.036873,2.239419,0.151693,
        -3.036873,2.243347,-2.521049,
        -3.036873,2.244035,-1.229456,
        -3.036873,2.244355,0.309899,
        -3.036873,2.245175,2.460236,
        -3.036873,2.245628,-1.899216,
        -3.036873,2.245973,-0.702234,
        -3.036873,2.255601,-1.728131,
        -3.036873,2.256702,-1.401367,
        -3.036873,2.261466,-1.557500,
        -3.036873,2.267390,0.878499,
        -3.036873,2.272099,2.018594,
        -3.036873,2.278618,1.170386,
        -3.036873,2.283304,-2.345584,
        -3.036873,2.285175,-2.943090,
        -3.036873,2.293000,-0.124451,
        -3.036873,2.294058,-0.421070,
        -3.036873,2.303283,1.855938,
        -3.036873,2.304968,2.608978,
        -3.036873,2.305347,2.311346,
        -3.036873,2.307188,-2.656134,
        -3.036873,2.307301,1.336200,
        -3.036873,2.309655,0.726020,
        -3.036873,2.317634,-2.185225,
        -3.036873,2.321510,1.685664,
        -3.036873,2.324376,-3.111006,
        -3.036873,2.324731,1.511186,
        -3.036873,2.329587,0.035498,
        -3.036873,2.330158,2.977264,
        -3.036873,2.335096,-0.588468,
        -3.036873,2.336046,-1.115799,
        -3.036873,2.338060,2.803613,
        -3.036873,2.338083,-0.943400,
        -3.036873,2.340210,0.562145,
        -3.036873,2.341931,1.023926,
        -3.036873,2.343265,-0.779061,
        -3.036873,2.343434,-2.013692,
        -3.036873,2.353029,-1.310529,
        -3.036873,2.354868,0.211543,
        -3.036873,2.355120,0.388593,
        -3.036873,2.355165,-1.833173,
        -3.036873,2.355462,2.151441,
        -3.036873,2.362948,-2.487477,
        -3.036873,2.363101,-2.810456,
        -3.036873,2.363394,-0.272918,
        -3.036873,2.364509,-1.644533,
        -3.036873,2.370465,2.466824,
        -3.036873,2.386110,-1.481351,
        -3.036873,2.394960,0.862133,
        -3.036873,2.395716,1.977184,
        -3.036873,2.396339,1.200900,
        -3.036873,2.407941,-2.316753,
        -3.036873,2.409450,-2.988432,
        -3.036873,2.417202,-0.103363,
        -3.036873,2.417386,-0.443097,
        -3.036873,2.417950,2.666662,
        -3.036873,2.423190,1.789783,
        -3.036873,2.423692,1.392696,
        -3.036873,2.431122,3.086379,
        -3.036873,2.431406,-2.647820,
        -3.036873,2.431584,2.307631,
        -3.036873,2.434090,1.592818,
        -3.036873,2.435993,0.683844,
        -3.036873,2.440898,-2.133300,
        -3.036873,2.442222,-1.195837,
        -3.036873,2.443112,2.883761,
        -3.036873,2.449191,-0.633651,
        -3.036873,2.450821,0.084741,
        -3.036873,2.453499,-1.008861,
        -3.036873,2.460010,-1.938011,
        -3.036873,2.460199,-0.822008,
        -3.036873,2.460902,0.490079,
        -3.036873,2.461774,-1.734913,
        -3.036873,2.463986,1.034407,
        -3.036873,2.467647,0.287667,
        -3.036873,2.479849,2.126925,
        -3.036873,2.485907,-2.835471,
        -3.036873,2.485945,-1.374853,
        -3.036873,2.487747,2.505912,
        -3.036873,2.487957,-0.274279,
        -3.036873,2.489741,-2.468428,
        -3.036873,2.512538,1.239900,
        -3.036873,2.517528,1.923289,
        -3.036873,2.518454,-1.565852,
        -3.036873,2.519348,0.843503,
        -3.036873,2.522766,-3.059231,
        -3.036873,2.528686,2.737228,
        -3.036873,2.534798,-2.266671,
        -3.036873,2.536533,1.466274,
        -3.036873,2.538504,1.699812,
        -3.036873,2.539516,-0.476942,
        -3.036873,2.539532,-0.074548,
        -3.036873,2.543197,2.990252,
        -3.036873,2.554022,2.320619,
        -3.036873,2.556919,-2.655005,
        -3.036873,2.558146,0.627191,
        -3.036873,2.558223,-1.164189,
        -3.036873,2.563452,-2.044028,
        -3.036873,2.568462,0.150655,
        -3.036873,2.569130,-0.700837,
        -3.036873,2.570050,-1.804492,
        -3.036873,2.572710,-0.934665,
        -3.036873,2.575994,0.391718,
        -3.036873,2.584774,1.046703,
        -3.036873,2.601432,2.102535,
        -3.036873,2.603559,2.555114,
        -3.036873,2.604150,-2.885208,
        -3.036873,2.608630,-1.378011,
        -3.036873,2.611256,-0.279110,
        -3.036873,2.617344,-2.439480,
        -3.036873,2.627402,1.293552,
        -3.036873,2.635396,1.848850,
        -3.036873,2.635686,2.834950,
        -3.036873,2.637087,3.132730,
        -3.036873,2.642029,0.813274,
        -3.036873,2.645260,1.570203,
        -3.036873,2.646182,-1.614036,
        -3.036873,2.660113,-2.186513,
        -3.036873,2.660483,-0.034608,
        -3.036873,2.661713,-0.529511,
        -3.036873,2.672720,-1.110422,
        -3.036873,2.676626,2.335353,
        -3.036873,2.676638,0.540364,
        -3.036873,2.676780,-1.901158,
        -3.036873,2.681411,-2.667965,
        -3.036873,2.682152,0.245751,
        -3.036873,2.682859,-0.817071,
        -3.036873,2.705107,1.062142,
        -3.036873,2.717705,2.625995,
        -3.036873,2.721001,2.060097,
        -3.036873,2.722319,-2.954069,
        -3.036873,2.729840,-1.368237,
        -3.036873,2.734349,-0.289298,
        -3.036873,2.738409,2.975671,
        -3.036873,2.740366,1.374341,
        -3.036873,2.744633,-2.391311,
        -3.036873,2.747398,1.727166,
        -3.036873,2.763799,0.760372,
        -3.036873,2.768238,-1.679243,
        -3.036873,2.779245,-2.046268,
        -3.036873,2.780208,-0.626453,
        -3.036873,2.780471,0.027732,
        -3.036873,2.784696,-1.013128,
        -3.036873,2.789817,0.395431,
        -3.036873,2.801154,2.353143,
        -3.036873,2.807091,-2.687477,
        -3.036873,2.825762,1.085035,
        -3.036873,2.830209,2.734943,
        -3.036873,2.833517,-3.085508,
        -3.036873,2.838301,1.980061,
        -3.036873,2.849871,1.512012,
        -3.036873,2.850089,-1.343899,
        -3.036873,2.856905,-0.318441,
        -3.036873,2.870445,-2.291001,
        -3.036873,2.885017,0.655562,
        -3.036873,2.886428,-1.793217,
        -3.036873,2.890337,-0.825633,
        -3.036873,2.898932,0.139973,
        -3.036873,2.931927,-2.728768,
        -3.036873,2.932096,2.377790,
        -3.036873,2.937409,2.947245,
        -3.036873,2.948048,1.133127,
        -3.036873,2.950896,1.797889,
        -3.036873,2.970019,-1.282382,
        -3.036873,2.978333,-0.412090,
        -3.036873,2.996437,-2.089121,
        -3.036873,3.011274,0.434110,
        -3.036873,3.050410,3.138492,
        -3.036873,3.064492,1.647259,
        -3.036873,3.088868,-0.989501,
        -2.932153,0.053884,-2.408394,
        -2.932153,0.077645,1.453433,
        -2.932153,0.088408,-0.311160,
        -2.932153,0.142314,2.590387,
        -2.932153,0.147206,-1.385164,
        -2.932153,0.163506,0.573203,
        -2.932153,0.165485,-2.875396,
        -2.932153,0.182564,-2.125821,
        -2.932153,0.190569,1.929141,
        -2.932153,0.201102,1.249082,
        -2.932153,0.201397,-0.716646,
        -2.932153,0.210568,-0.069559,
        -2.932153,0.251672,2.848900,
        -2.932153,0.266262,-1.678068,
        -2.932153,0.269108,-1.215563,
        -2.932153,0.273777,-2.506756,
        -2.932153,0.279854,0.814581,
        -2.932153,0.281664,-2.969147,
        -2.932153,0.283625,2.405603,
        -2.932153,0.284058,0.357350,
        -2.932153,0.294107,1.593880,
        -2.932153,0.302491,-0.400410,
        -2.932153,0.308059,2.021491,
        -2.932153,0.312219,-2.072057,
        -2.932153,0.321905,-0.829121,
        -2.932153,0.329984,1.201621,
        -2.932153,0.339714,-0.013712,
        -2.932153,0.368784,2.974974,
        -2.932153,0.376428,-1.495532,
        -2.932153,0.378103,2.646436,
        -2.932153,0.385088,0.590355,
        -2.932153,0.386248,-2.690943,
        -2.932153,0.389851,-2.340031,
        -2.932153,0.391502,-1.171374,
        -2.932153,0.398249,-1.812694,
        -2.932153,0.400542,-2.995729,
        -2.932153,0.404273,0.918163,
        -2.932153,0.405617,1.773533,
        -2.932153,0.409009,-0.570243,
        -2.932153,0.410700,0.270433,
        -2.932153,0.415893,1.464890,
        -2.932153,0.420680,2.088164,
        -2.932153,0.424766,-0.258729,
        -2.932153,0.426507,2.367872,
        -2.932153,0.441292,-0.890727,
        -2.932153,0.451839,-2.078193,
        -2.932153,0.459270,1.191562,
        -2.932153,0.470622,0.014002,
        -2.932153,0.481292,-2.529098,
        -2.932153,0.482460,2.849843,
        -2.932153,0.492205,-1.397963,
        -2.932153,0.493314,3.117264,
        -2.932153,0.497870,0.727828,
        -2.932153,0.499783,-1.654379,
        -2.932153,0.502651,0.463333,
        -2.932153,0.505463,2.594557,
        -2.932153,0.506104,-2.874910,
        -2.932153,0.513579,-1.147354,
        -2.932153,0.518569,1.644237,
        -2.932153,0.519001,1.900488,
        -2.932153,0.519277,-0.673809,
        -2.932153,0.524769,-0.421955,
        -2.932153,0.525564,-2.288838,
        -2.932153,0.532541,0.978737,
        -2.932153,0.534610,2.154808,
        -2.932153,0.535020,-1.891846,
        -2.932153,0.540208,1.400497,
        -2.932153,0.542006,0.228337,
        -2.932153,0.550762,-0.180771,
        -2.932153,0.560531,2.378474,
        -2.932153,0.561629,-0.924576,
        -2.932153,0.567465,-2.679392,
        -2.932153,0.588527,1.181755,
        -2.932153,0.591643,-2.098364,
        -2.932153,0.594963,2.997575,
        -2.932153,0.599643,-3.045235,
        -2.932153,0.599962,2.773556,
        -2.932153,0.602620,0.029270,
        -2.932153,0.605711,0.598454,
        -2.932153,0.607708,-1.541026,
        -2.932153,0.609360,-2.464379,
        -2.932153,0.610738,-1.326786,
        -2.932153,0.616968,0.820489,
        -2.932153,0.626379,1.773610,
        -2.932153,0.628626,-1.751180,
        -2.932153,0.630671,0.391949,
        -2.932153,0.632622,-0.737642,
        -2.932153,0.632708,2.566722,
        -2.932153,0.632797,1.989858,
        -2.932153,0.634455,-0.534790,
        -2.932153,0.636205,1.561170,
        -2.932153,0.637681,-1.123423,
        -2.932153,0.641441,-2.842109,
        -2.932153,0.644807,-0.327139,
        -2.932153,0.652440,2.197129,
        -2.932153,0.660897,-2.278593,
        -2.932153,0.664008,1.360091,
        -2.932153,0.670850,-1.945257,
        -2.932153,0.671396,1.004859,
        -2.932153,0.673985,0.203855,
        -2.932153,0.680326,-0.133370,
        -2.932153,0.685083,-0.942363,
        -2.932153,0.688204,-2.632204,
        -2.932153,0.689777,2.385011,
        -2.932153,0.700065,3.115430,
        -2.932153,0.704080,2.911636,
        -2.932153,0.712883,1.174216,
        -2.932153,0.715701,0.699906,
        -2.932153,0.718851,-1.442215,
        -2.932153,0.721019,2.721536,
        -2.932153,0.725300,-2.975134,
        -2.932153,0.727136,-1.633288,
        -2.932153,0.727399,-2.120567,
        -2.932153,0.730751,0.522488,
        -2.932153,0.732892,-1.254330,
        -2.932153,0.734333,-2.446184,
        -2.932153,0.735125,0.037270,
        -2.932153,0.737893,1.866463,
        -2.932153,0.740650,1.681075,
        -2.932153,0.745971,-0.781286,
        -2.932153,0.749432,-0.442559,
        -2.932153,0.750102,2.050636,
        -2.932153,0.751046,0.863821,
        -2.932153,0.757018,-0.608833,
        -2.932153,0.757046,-1.816019,
        -2.932153,0.757389,1.501184,
        -2.932153,0.759800,2.546181,
        -2.932153,0.759859,0.351923,
        -2.932153,0.761141,-2.781102,
        -2.932153,0.769035,-1.078574,
        -2.932153,0.770166,-0.266604,
        -2.932153,0.772218,2.227380,
        -2.932153,0.788382,1.331051,
        -2.932153,0.790269,-2.282409,
        -2.932153,0.803410,-1.984567,
        -2.932153,0.803813,0.191520,
        -2.932153,0.806832,3.022155,
        -2.932153,0.807476,-2.597675,
        -2.932153,0.812512,-0.107335,
        -2.932153,0.814161,1.010949,
        -2.932153,0.817732,-3.084486,
        -2.932153,0.817846,2.390688,
        -2.932153,0.819323,2.845960,
        -2.932153,0.819430,-0.915017,
        -2.932153,0.829198,-1.532491,
        -2.932153,0.829531,0.665695,
        -2.932153,0.829816,1.167147,
        -2.932153,0.833613,-1.357492,
        -2.932153,0.842945,-2.907318,
        -2.932153,0.844883,2.679708,
        -2.932153,0.848189,-1.705003,
        -2.932153,0.849888,1.767901,
        -2.932153,0.851126,0.493370,
        -2.932153,0.852911,1.933394,
        -2.932153,0.857516,-2.432863,
        -2.932153,0.858261,-1.191729,
        -2.932153,0.858279,-2.138895,
        -2.932153,0.859723,-0.734928,
        -2.932153,0.859754,1.603292,
        -2.932153,0.861655,-0.394821,
        -2.932153,0.864300,0.043549,
        -2.932153,0.866940,0.851229,
        -2.932153,0.869748,2.095969,
        -2.932153,0.877352,-0.574668,
        -2.932153,0.879254,-2.733338,
        -2.932153,0.882554,1.443757,
        -2.932153,0.884060,0.334199,
        -2.932153,0.885626,-1.866061,
        -2.932153,0.889834,2.529957,
        -2.932153,0.893140,2.251698,
        -2.932153,0.897463,-0.240822,
        -2.932153,0.898758,-1.031948,
        -2.932153,0.913918,1.287618,
        -2.932153,0.916555,-2.287173,
        -2.932153,0.918305,3.107845,
        -2.932153,0.919522,2.948277,
        -2.932153,0.927138,-2.569686,
        -2.932153,0.929031,0.185383,
        -2.932153,0.933766,-3.015917,
        -2.932153,0.934873,-2.010822,
        -2.932153,0.935636,-1.445855,
        -2.932153,0.936702,-0.866461,
        -2.932153,0.937395,0.729108,
        -2.932153,0.937910,2.787575,
        -2.932153,0.939144,-0.092962,
        -2.932153,0.941718,1.125085,
        -2.932153,0.942873,-1.607680,
        -2.932153,0.943701,0.980328,
        -2.932153,0.947774,2.395256,
        -2.932153,0.952433,0.584347,
        -2.932153,0.954456,-1.287293,
        -2.932153,0.958731,-2.852552,
        -2.932153,0.961838,1.839284,
        -2.932153,0.961898,-0.469425,
        -2.932153,0.965426,1.686615,
        -2.932153,0.970108,1.990771,
        -2.932153,0.972820,-1.762780,
        -2.932153,0.973500,-0.711881,
        -2.932153,0.975253,2.641082,
        -2.932153,0.977824,0.441928,
        -2.932153,0.981485,-2.424766,
        -2.932153,0.981877,1.537100,
        -2.932153,0.982609,-1.136920,
        -2.932153,0.984265,-2.156073,
        -2.932153,0.989150,0.045493,
        -2.932153,0.993052,2.137478,
        -2.932153,0.997766,-2.693871,
        -2.932153,0.999847,-0.334738,
        -2.932153,1.001920,0.848054,
        -2.932153,1.006146,1.389734,
        -2.932153,1.012180,0.304066,
        -2.932153,1.016938,2.270674,
        -2.932153,1.017371,-0.979790,
        -2.932153,1.019495,-1.894487,
        -2.932153,1.023796,-0.583691,
        -2.932153,1.026525,2.506945,
        -2.932153,1.027004,-0.203653,
        -2.932153,1.028027,3.036573,
        -2.932153,1.031435,1.241021,
        -2.932153,1.032316,2.884221,
        -2.932153,1.035863,-3.110475,
        -2.932153,1.042085,-2.292950,
        -2.932153,1.043272,-1.522986,
        -2.932153,1.046265,-2.958847,
        -2.932153,1.049272,-1.376344,
        -2.932153,1.050039,-2.552086,
        -2.932153,1.053682,0.689863,
        -2.932153,1.055458,0.170394,
        -2.932153,1.056622,-0.831866,
        -2.932153,1.057515,1.097998,
        -2.932153,1.058706,-2.036438,
        -2.932153,1.061389,-1.672248,
        -2.932153,1.063758,2.743093,
        -2.932153,1.070157,-0.073422,
        -2.932153,1.072154,1.761506,
        -2.932153,1.074029,1.909193,
        -2.932153,1.074204,0.545276,
        -2.932153,1.074714,-2.805147,
        -2.932153,1.076262,0.966104,
        -2.932153,1.077642,-1.227073,
        -2.932153,1.084151,1.618525,
        -2.932153,1.085085,2.377673,
        -2.932153,1.086822,-0.441112,
        -2.932153,1.087142,2.058815,
        -2.932153,1.098383,-0.696430,
        -2.932153,1.099138,-1.084033,
        -2.932153,1.103208,1.477773,
        -2.932153,1.103783,0.408254,
        -2.932153,1.106275,-2.420800,
        -2.932153,1.108079,2.612943,
        -2.932153,1.109345,-2.172718,
        -2.932153,1.115957,-1.787597,
        -2.932153,1.118930,0.045983,
        -2.932153,1.119777,0.834702,
        -2.932153,1.121943,-2.669042,
        -2.932153,1.124930,2.219917,
        -2.932153,1.125217,1.336294,
        -2.932153,1.130958,2.972931,
        -2.932153,1.131400,-0.306629,
        -2.932153,1.137135,-0.936514,
        -2.932153,1.139125,-1.924528,
        -2.932153,1.139877,0.276382,
        -2.932153,1.142003,-3.058022,
        -2.932153,1.147363,-0.558448,
        -2.932153,1.150348,-0.176418,
        -2.932153,1.151198,1.197016,
        -2.932153,1.153930,-1.597104,
        -2.932153,1.155389,2.837148,
        -2.932153,1.155458,-1.460623,
        -2.932153,1.157331,-2.906362,
        -2.932153,1.158594,2.487618,
        -2.932153,1.162103,-1.322553,
        -2.932153,1.165407,3.092861,
        -2.932153,1.168025,0.717075,
        -2.932153,1.168406,-2.301519,
        -2.932153,1.174782,-2.538755,
        -2.932153,1.177998,1.061256,
        -2.932153,1.178898,1.827305,
        -2.932153,1.181331,-2.060330,
        -2.932153,1.182540,-0.798952,
        -2.932153,1.184569,0.147517,
        -2.932153,1.184734,1.968919,
        -2.932153,1.184804,0.593386,
        -2.932153,1.188278,1.690808,
        -2.932153,1.194360,2.110440,
        -2.932153,1.194776,2.710455,
        -2.932153,1.197745,-1.173705,
        -2.932153,1.199898,-2.777257,
        -2.932153,1.201164,2.355931,
        -2.932153,1.203760,1.556404,
        -2.932153,1.205845,-0.415993,
        -2.932153,1.206585,0.933648,
        -2.932153,1.206950,-0.037248,
        -2.932153,1.209341,0.471162,
        -2.932153,1.213483,-1.039417,
        -2.932153,1.218146,-0.672643,
        -2.932153,1.221608,1.421669,
        -2.932153,1.226043,-1.693533,
        -2.932153,1.232870,-1.822773,
        -2.932153,1.234375,-2.421194,
        -2.932153,1.234531,-2.190148,
        -2.932153,1.239414,0.348959,
        -2.932153,1.240183,2.587255,
        -2.932153,1.242312,2.941234,
        -2.932153,1.244075,1.286644,
        -2.932153,1.245961,-2.997461,
        -2.932153,1.246495,-2.646872,
        -2.932153,1.247736,2.242810,
        -2.932153,1.253604,0.812222,
        -2.932153,1.257816,-0.894317,
        -2.932153,1.258247,-1.953634,
        -2.932153,1.258993,-0.284099,
        -2.932153,1.262380,-0.157609,
        -2.932153,1.262791,-1.549229,
        -2.932153,1.263772,-0.533073,
        -2.932153,1.266731,-1.412461,
        -2.932153,1.268591,-3.122146,
        -2.932153,1.269614,-1.275246,
        -2.932153,1.270745,1.152533,
        -2.932153,1.273606,0.224545,
        -2.932153,1.279716,2.812101,
        -2.932153,1.283786,2.462124,
        -2.932153,1.284798,0.078209,
        -2.932153,1.285402,-2.882049,
        -2.932153,1.286936,1.886053,
        -2.932153,1.289408,0.683122,
        -2.932153,1.294736,1.756184,
        -2.932153,1.296978,-2.311511,
        -2.932153,1.298546,2.020601,
        -2.932153,1.300924,1.022478,
        -2.932153,1.303305,3.043152,
        -2.932153,1.304210,-2.084225,
        -2.932153,1.305241,-2.530249,
        -2.932153,1.307513,1.628003,
        -2.932153,1.310722,0.558267,
        -2.932153,1.314142,2.139329,
        -2.932153,1.316087,-0.777971,
        -2.932153,1.317907,-1.121023,
        -2.932153,1.321019,-2.749776,
        -2.932153,1.321609,2.687717,
        -2.932153,1.322568,1.499415,
        -2.932153,1.322972,-0.395699,
        -2.932153,1.325286,-0.994192,
        -2.932153,1.329146,-0.648513,
        -2.932153,1.329795,-0.042570,
        -2.932153,1.331651,2.344497,
        -2.932153,1.336522,0.435098,
        -2.932153,1.339410,0.900087,
        -2.932153,1.339550,1.370353,
        -2.932153,1.340441,-1.731580,
        -2.932153,1.351855,-1.855234,
        -2.932153,1.360161,-3.018404,
        -2.932153,1.361031,-1.613971,
        -2.932153,1.361196,2.914780,
        -2.932153,1.361998,-2.207254,
        -2.932153,1.362302,1.239384,
        -2.932153,1.365228,2.563808,
        -2.932153,1.365452,0.310522,
        -2.932153,1.367028,-2.420041,
        -2.932153,1.373422,-1.484281,
        -2.932153,1.374245,-1.226417,
        -2.932153,1.375020,-0.170393,
        -2.932153,1.376560,-1.355481,
        -2.932153,1.377070,-1.980984,
        -2.932153,1.377578,-2.636339,
        -2.932153,1.378096,0.774626,
        -2.932153,1.379086,-0.508192,
        -2.932153,1.382562,0.177141,
        -2.932153,1.388849,2.233847,
        -2.932153,1.389268,-0.886349,
        -2.932153,1.391324,1.109396,
        -2.932153,1.391699,-0.288725,
        -2.932153,1.392869,3.136903,
        -2.932153,1.393612,1.952016,
        -2.932153,1.395545,-2.850807,
        -2.932153,1.401926,1.822930,
        -2.932153,1.402149,2.788898,
        -2.932153,1.409552,0.647980,
        -2.932153,1.410840,0.056219,
        -2.932153,1.411944,2.444377,
        -2.932153,1.413639,1.697982,
        -2.932153,1.426479,1.573427,
        -2.932153,1.426725,2.102242,
        -2.932153,1.427868,0.984637,
        -2.932153,1.429580,-2.107232,
        -2.932153,1.431375,-2.313269,
        -2.932153,1.431907,3.015410,
        -2.932153,1.432325,-1.029054,
        -2.932153,1.433341,0.524268,
        -2.932153,1.435181,-0.739771,
        -2.932153,1.437208,-2.529085,
        -2.932153,1.439041,-0.617966,
        -2.932153,1.441473,1.448544,
        -2.932153,1.444527,2.664744,
        -2.932153,1.450513,-0.400625,
        -2.932153,1.451907,-0.066823,
        -2.932153,1.452754,-2.738631,
        -2.932153,1.454061,-2.950416,
        -2.932153,1.456736,1.323154,
        -2.932153,1.458370,0.400433,
        -2.932153,1.459314,-1.765003,
        -2.932153,1.459902,-1.145835,
        -2.932153,1.464952,2.329245,
        -2.932153,1.466050,0.862104,
        -2.932153,1.471330,-1.886160,
        -2.932153,1.475119,-1.648255,
        -2.932153,1.480878,1.194576,
        -2.932153,1.480942,0.273235,
        -2.932153,1.481086,2.890269,
        -2.932153,1.484118,-1.529156,
        -2.932153,1.484741,-1.279219,
        -2.932153,1.487681,-1.406069,
        -2.932153,1.487886,-3.066037,
        -2.932153,1.490555,2.543284,
        -2.932153,1.492424,-0.190932,
        -2.932153,1.495002,-0.839810,
        -2.932153,1.495297,2.007188,
        -2.932153,1.496414,-2.009267,
        -2.932153,1.498261,-2.425270,
        -2.932153,1.498907,-2.207327,
        -2.932153,1.500204,0.737638,
        -2.932153,1.502469,0.151156,
        -2.932153,1.507448,-0.514782,
        -2.932153,1.509287,2.204647,
        -2.932153,1.510518,1.884920,
        -2.932153,1.511948,-2.632360,
        -2.932153,1.515235,1.068565,
        -2.932153,1.522237,3.102018,
        -2.932153,1.522373,1.764991,
        -2.932153,1.522416,-0.954956,
        -2.932153,1.523102,2.765920,
        -2.932153,1.524811,-2.844718,
        -2.932153,1.527261,-0.304916,
        -2.932153,1.529788,0.613208,
        -2.932153,1.533648,1.644808,
        -2.932153,1.535018,0.028869,
        -2.932153,1.541082,2.424608,
        -2.932153,1.546010,1.523462,
        -2.932153,1.551135,-0.642994,
        -2.932153,1.553895,0.489607,
        -2.932153,1.555041,0.947813,
        -2.932153,1.559671,-1.074934,
        -2.932153,1.560345,2.985899,
        -2.932153,1.560487,-2.323400,
        -2.932153,1.561127,1.401111,
        -2.932153,1.567193,2.643172,
        -2.932153,1.568892,2.098362,
        -2.932153,1.571702,-2.098435,
        -2.932153,1.572701,-2.528730,
        -2.932153,1.573572,1.278593,
        -2.932153,1.574849,-0.091505,
        -2.932153,1.576334,0.365236,
        -2.932153,1.577029,-2.962888,
        -2.932153,1.577688,-1.814941,
        -2.932153,1.579542,-0.755917,
        -2.932153,1.579624,-1.199759,
        -2.932153,1.581515,-0.419414,
        -2.932153,1.584581,-2.738702,
        -2.932153,1.586948,-1.696296,
        -2.932153,1.588901,2.301496,
        -2.932153,1.589936,0.825020,
        -2.932153,1.590018,-1.941652,
        -2.932153,1.596968,-1.328358,
        -2.932153,1.597178,-1.575607,
        -2.932153,1.600470,-1.453196,
        -2.932153,1.601408,0.242409,
        -2.932153,1.602259,1.152656,
        -2.932153,1.603064,2.865774,
        -2.932153,1.612085,1.968802,
        -2.932153,1.612720,-3.080498,
        -2.932153,1.616061,2.521531,
        -2.932153,1.616900,-0.207457,
        -2.932153,1.620988,-0.875920,
        -2.932153,1.622114,0.701509,
        -2.932153,1.622749,0.121426,
        -2.932153,1.622993,-2.221629,
        -2.932153,1.629346,-0.542669,
        -2.932153,1.630007,1.844552,
        -2.932153,1.634747,-2.426198,
        -2.932153,1.642346,1.722617,
        -2.932153,1.643925,1.034629,
        -2.932153,1.644848,2.743774,
        -2.932153,1.645258,2.188109,
        -2.932153,1.645417,-2.634725,
        -2.932153,1.647395,3.085437,
        -2.932153,1.648538,-2.855044,
        -2.932153,1.649932,0.576808,
        -2.932153,1.652224,-0.996105,
        -2.932153,1.652871,1.600681,
        -2.932153,1.657997,-0.325127,
        -2.932153,1.661936,0.001037,
        -2.932153,1.665187,1.477402,
        -2.932153,1.665496,2.397872,
        -2.932153,1.673267,-0.665453,
        -2.932153,1.673557,0.452526,
        -2.932153,1.679277,-1.122421,
        -2.932153,1.679696,0.912482,
        -2.932153,1.682420,1.353922,
        -2.932153,1.685031,-2.009038,
        -2.932153,1.685196,2.967046,
        -2.932153,1.690107,2.620910,
        -2.932153,1.690453,-2.123076,
        -2.932153,1.691693,-1.871572,
        -2.932153,1.691817,2.061688,
        -2.932153,1.692644,1.237670,
        -2.932153,1.696386,0.328123,
        -2.932153,1.697162,-1.249026,
        -2.932153,1.697488,-2.322785,
        -2.932153,1.698726,-2.973576,
        -2.932153,1.699172,-1.749717,
        -2.932153,1.704478,-0.106295,
        -2.932153,1.707244,-0.444770,
        -2.932153,1.707500,-2.531497,
        -2.932153,1.707714,-0.787321,
        -2.932153,1.708430,-1.627552,
        -2.932153,1.710583,-1.376890,
        -2.932153,1.712415,-2.747921,
        -2.932153,1.713125,-1.503073,
        -2.932153,1.713452,0.789419,
        -2.932153,1.720612,2.279888,
        -2.932153,1.724852,0.202831,
        -2.932153,1.724868,2.846246,
        -2.932153,1.728940,1.931535,
        -2.932153,1.732838,1.127584,
        -2.932153,1.735719,-3.092485,
        -2.932153,1.739676,2.496017,
        -2.932153,1.739971,-0.228876,
        -2.932153,1.743328,0.663347,
        -2.932153,1.744927,-0.911892,
        -2.932153,1.745952,0.084733,
        -2.932153,1.748077,1.804354,
        -2.932153,1.752959,-0.567548,
        -2.932153,1.760169,1.678132,
        -2.932153,1.763887,-2.218450,
        -2.932153,1.766737,2.722752,
        -2.932153,1.767999,1.004737,
        -2.932153,1.769149,1.550298,
        -2.932153,1.770119,-2.427175,
        -2.932153,1.770460,2.154208,
        -2.932153,1.770514,3.070982,
        -2.932153,1.770735,0.537245,
        -2.932153,1.772673,-2.864760,
        -2.932153,1.773527,-1.039582,
        -2.932153,1.776419,-2.642421,
        -2.932153,1.780769,1.416330,
        -2.932153,1.786634,-0.346247,
        -2.932153,1.788829,-1.947553,
        -2.932153,1.793737,0.410270,
        -2.932153,1.794553,2.374618,
        -2.932153,1.794699,-0.694061,
        -2.932153,1.797469,-1.169105,
        -2.932153,1.803705,0.881076,
        -2.932153,1.804124,1.264249,
        -2.932153,1.807825,-2.090455,
        -2.932153,1.808822,-1.817375,
        -2.932153,1.809020,2.949741,
        -2.932153,1.812466,2.597595,
        -2.932153,1.813159,2.025395,
        -2.932153,1.813816,-1.298435,
        -2.932153,1.815016,0.278951,
        -2.932153,1.816510,-0.008631,
        -2.932153,1.817772,-1.689625,
        -2.932153,1.821360,-2.984352,
        -2.932153,1.822312,-0.127211,
        -2.932153,1.823386,-1.429807,
        -2.932153,1.823834,-1.560483,
        -2.932153,1.832060,-0.821627,
        -2.932153,1.832690,-0.468440,
        -2.932153,1.833945,-2.320278,
        -2.932153,1.834787,0.751968,
        -2.932153,1.840579,-2.536445,
        -2.932153,1.841033,-2.756932,
        -2.932153,1.846350,1.893983,
        -2.932153,1.846644,1.113766,
        -2.932153,1.846833,2.826543,
        -2.932153,1.847325,2.249051,
        -2.932153,1.854119,0.134161,
        -2.932153,1.858863,-3.106022,
        -2.932153,1.865046,1.763384,
        -2.932153,1.865619,-0.952036,
        -2.932153,1.865912,0.624453,
        -2.932153,1.866694,2.473686,
        -2.932153,1.867524,-0.240083,
        -2.932153,1.875561,1.632831,
        -2.932153,1.876049,-0.594378,
        -2.932153,1.880509,1.500204,
        -2.932153,1.883504,1.360457,
        -2.932153,1.886971,-2.196665,
        -2.932153,1.888979,0.982922,
        -2.932153,1.889815,2.701570,
        -2.932153,1.892144,-2.005340,
        -2.932153,1.892348,0.496522,
        -2.932153,1.893281,-1.083976,
        -2.932153,1.893828,3.053453,
        -2.932153,1.895117,2.121107,
        -2.932153,1.898741,-2.875735,
        -2.932153,1.904429,-2.428029,
        -2.932153,1.908408,-2.648903,
        -2.932153,1.911828,-0.364489,
        -2.932153,1.912224,0.366565,
        -2.932153,1.914854,1.214468,
        -2.932153,1.915467,-1.883143,
        -2.932153,1.915763,-0.724679,
        -2.932153,1.915790,-1.216366,
        -2.932153,1.922256,2.347806,
        -2.932153,1.922616,0.844633,
        -2.932153,1.925185,0.231170,
        -2.932153,1.927474,-1.754616,
        -2.932153,1.928342,-1.351174,
        -2.932153,1.933328,-1.623328,
        -2.932153,1.933878,0.034150,
        -2.932153,1.934585,-1.487979,
        -2.932153,1.934802,2.927634,
        -2.932153,1.934862,1.989776,
        -2.932153,1.936327,-0.099350,
        -2.932153,1.938088,2.578507,
        -2.932153,1.945677,-2.999805,
        -2.932153,1.954572,-0.859208,
        -2.932153,1.957868,-0.490784,
        -2.932153,1.958989,0.712772,
        -2.932153,1.960212,-2.307561,
        -2.932153,1.964774,1.855507,
        -2.932153,1.964971,-2.100533,
        -2.932153,1.969112,1.088273,
        -2.932153,1.971438,-2.765415,
        -2.932153,1.972312,2.805134,
        -2.932153,1.974950,2.220596,
        -2.932153,1.975135,-2.539127,
        -2.932153,1.982107,1.720829,
        -2.932153,1.983195,-3.127984,
        -2.932153,1.983318,-0.247182,
        -2.932153,1.986720,-0.993220,
        -2.932153,1.990381,1.586005,
        -2.932153,1.990470,0.580283,
        -2.932153,1.990523,1.450326,
        -2.932153,1.993909,1.312897,
        -2.932153,1.994694,2.452813,
        -2.932153,1.999226,-0.620717,
        -2.932153,2.005145,0.941012,
        -2.932153,2.007301,0.142337,
        -2.932153,2.014149,0.444138,
        -2.932153,2.014338,-1.127392,
        -2.932153,2.018131,3.021778,
        -2.932153,2.019676,-1.980944,
        -2.932153,2.019912,2.686612,
        -2.932153,2.020545,2.088672,
        -2.932153,2.026074,0.300945,
        -2.932153,2.026196,-2.889057,
        -2.932153,2.031959,-1.268178,
        -2.932153,2.032917,-2.199917,
        -2.932153,2.033572,-1.842884,
        -2.932153,2.034712,-2.419505,
        -2.932153,2.035818,-0.379479,
        -2.932153,2.036004,-0.757062,
        -2.932153,2.040501,-1.411488,
        -2.932153,2.042874,-1.701367,
        -2.932153,2.043060,-2.653746,
        -2.932153,2.043729,-1.557938,
        -2.932153,2.048538,0.806623,
        -2.932153,2.049006,1.199930,
        -2.932153,2.049502,-0.137989,
        -2.932153,2.052181,2.326959,
        -2.932153,2.052794,-0.008526,
        -2.932153,2.057608,1.951726,
        -2.932153,2.065568,2.567730,
        -2.932153,2.070371,2.879808,
        -2.932153,2.071256,-3.018961,
        -2.932153,2.079746,-0.897699,
        -2.932153,2.082784,1.042403,
        -2.932153,2.083752,-0.511086,
        -2.932153,2.084401,1.810502,
        -2.932153,2.085731,0.671081,
        -2.932153,2.097496,1.373172,
        -2.932153,2.099185,1.667006,
        -2.932153,2.100482,-2.301084,
        -2.932153,2.103118,2.194745,
        -2.932153,2.103223,1.521685,
        -2.932153,2.103695,-2.777707,
        -2.932153,2.107523,3.129581,
        -2.932153,2.108277,-2.095907,
        -2.932153,2.108891,-2.532166,
        -2.932153,2.111003,-0.268632,
        -2.932153,2.112276,-1.027246,
        -2.932153,2.114380,0.531457,
        -2.932153,2.119166,0.100552,
        -2.932153,2.119564,2.754818,
        -2.932153,2.122474,-0.648314,
        -2.932153,2.125183,2.445598,
        -2.932153,2.129501,0.238290,
        -2.932153,2.131178,0.386663,
        -2.932153,2.131223,-1.174311,
        -2.932153,2.134269,-1.948042,
        -2.932153,2.134684,0.907065,
        -2.932153,2.137615,2.988806,
        -2.932153,2.145635,-1.327708,
        -2.932153,2.146467,2.055081,
        -2.932153,2.146582,-1.793788,
        -2.932153,2.149210,-1.482321,
        -2.932153,2.152901,-1.636788,
        -2.932153,2.156337,1.151558,
        -2.932153,2.156474,-2.912065,
        -2.932153,2.157194,-0.792563,
        -2.932153,2.166724,-0.401638,
        -2.932153,2.166980,-0.134879,
        -2.932153,2.173644,-2.403429,
        -2.932153,2.175711,-2.655523,
        -2.932153,2.177132,2.600419,
        -2.932153,2.178879,0.767722,
        -2.932153,2.180752,1.908485,
        -2.932153,2.181338,2.313721,
        -2.932153,2.185580,1.287673,
        -2.932153,2.195400,-2.217546,
        -2.932153,2.200418,-3.056442,
        -2.932153,2.201829,-0.001735,
        -2.932153,2.203804,1.756176,
        -2.932153,2.211602,1.443930,
        -2.932153,2.212686,-0.544809,
        -2.932153,2.213937,0.621892,
        -2.932153,2.214075,1.600691,
        -2.932153,2.216256,1.020029,
        -2.932153,2.219060,2.886631,
        -2.932153,2.223375,-0.906185,
        -2.932153,2.226002,-1.060248,
        -2.932153,2.226413,-2.064034,
        -2.932153,2.230554,2.171334,
        -2.932153,2.232413,2.735777,
        -2.932153,2.233709,3.070830,
        -2.932153,2.234383,-2.792065,
        -2.932153,2.237183,0.468799,
        -2.932153,2.237695,-0.271388,
        -2.932153,2.239419,0.151693,
        -2.932153,2.243347,-2.521049,
        -2.932153,2.244035,-1.229456,
        -2.932153,2.244355,0.309899,
        -2.932153,2.245175,2.460236,
        -2.932153,2.245628,-1.899216,
        -2.932153,2.245973,-0.702234,
        -2.932153,2.255601,-1.728131,
        -2.932153,2.256702,-1.401367,
        -2.932153,2.261466,-1.557500,
        -2.932153,2.267390,0.878499,
        -2.932153,2.272099,2.018594,
        -2.932153,2.278618,1.170386,
        -2.932153,2.283304,-2.345584,
        -2.932153,2.285175,-2.943090,
        -2.932153,2.293000,-0.124451,
        -2.932153,2.294058,-0.421070,
        -2.932153,2.303283,1.855938,
        -2.932153,2.304968,2.608978,
        -2.932153,2.305347,2.311346,
        -2.932153,2.307188,-2.656134,
        -2.932153,2.307301,1.336200,
        -2.932153,2.309655,0.726020,
        -2.932153,2.317634,-2.185225,
        -2.932153,2.321510,1.685664,
        -2.932153,2.324376,-3.111006,
        -2.932153,2.324731,1.511186,
        -2.932153,2.329587,0.035498,
        -2.932153,2.330158,2.977264,
        -2.932153,2.335096,-0.588468,
        -2.932153,2.336046,-1.115799,
        -2.932153,2.338060,2.803613,
        -2.932153,2.338083,-0.943400,
        -2.932153,2.340210,0.562145,
        -2.932153,2.341931,1.023926,
        -2.932153,2.343265,-0.779061,
        -2.932153,2.343434,-2.013692,
        -2.932153,2.353029,-1.310529,
        -2.932153,2.354868,0.211543,
        -2.932153,2.355120,0.388593,
        -2.932153,2.355165,-1.833173,
        -2.932153,2.355462,2.151441,
        -2.932153,2.362948,-2.487477,
        -2.932153,2.363101,-2.810456,
        -2.932153,2.363394,-0.272918,
        -2.932153,2.364509,-1.644533,
        -2.932153,2.370465,2.466824,
        -2.932153,2.386110,-1.481351,
        -2.932153,2.394960,0.862133,
        -2.932153,2.395716,1.977184,
        -2.932153,2.396339,1.200900,
        -2.932153,2.407941,-2.316753,
        -2.932153,2.409450,-2.988432,
        -2.932153,2.417202,-0.103363,
        -2.932153,2.417386,-0.443097,
        -2.932153,2.417950,2.666662,
        -2.932153,2.423190,1.789783,
        -2.932153,2.423692,1.392696,
        -2.932153,2.431122,3.086379,
        -2.932153,2.431406,-2.647820,
        -2.932153,2.431584,2.307631,
        -2.932153,2.434090,1.592818,
        -2.932153,2.435993,0.683844,
        -2.932153,2.440898,-2.133300,
        -2.932153,2.442222,-1.195837,
        -2.932153,2.443112,2.883761,
        -2.932153,2.449191,-0.633651,
        -2.932153,2.450821,0.084741,
        -2.932153,2.453499,-1.008861,
        -2.932153,2.460010,-1.938011,
        -2.932153,2.460199,-0.822008,
        -2.932153,2.460902,0.490079,
        -2.932153,2.461774,-1.734913,
        -2.932153,2.463986,1.034407,
        -2.932153,2.467647,0.287667,
        -2.932153,2.479849,2.126925,
        -2.932153,2.485907,-2.835471,
        -2.932153,2.485945,-1.374853,
        -2.932153,2.487747,2.505912,
        -2.932153,2.487957,-0.274279,
        -2.932153,2.489741,-2.468428,
        -2.932153,2.512538,1.239900,
        -2.932153,2.517528,1.923289,
        -2.932153,2.518454,-1.565852,
        -2.932153,2.519348,0.843503,
        -2.932153,2.522766,-3.059231,
        -2.932153,2.528686,2.737228,
        -2.932153,2.534798,-2.266671,
        -2.932153,2.536533,1.466274,
        -2.932153,2.538504,1.699812,
        -2.932153,2.539516,-0.476942,
        -2.932153,2.539532,-0.074548,
        -2.932153,2.543197,2.990252,
        -2.932153,2.554022,2.320619,
        -2.932153,2.556919,-2.655005,
        -2.932153,2.558146,0.627191,
        -2.932153,2.558223,-1.164189,
        -2.932153,2.563452,-2.044028,
        -2.932153,2.568462,0.150655,
        -2.932153,2.569130,-0.700837,
        -2.932153,2.570050,-1.804492,
        -2.932153,2.572710,-0.934665,
        -2.932153,2.575994,0.391718,
        -2.932153,2.584774,1.046703,
        -2.932153,2.601432,2.102535,
        -2.932153,2.603559,2.555114,
        -2.932153,2.604150,-2.885208,
        -2.932153,2.608630,-1.378011,
        -2.932153,2.611256,-0.279110,
        -2.932153,2.617344,-2.439480,
        -2.932153,2.627402,1.293552,
        -2.932153,2.635396,1.848850,
        -2.932153,2.635686,2.834950,
        -2.932153,2.637087,3.132730,
        -2.932153,2.642029,0.813274,
        -2.932153,2.645260,1.570203,
        -2.932153,2.646182,-1.614036,
        -2.932153,2.660113,-2.186513,
        -2.932153,2.660483,-0.034608,
        -2.932153,2.661713,-0.529511,
        -2.932153,2.672720,-1.110422,
        -2.932153,2.676626,2.335353,
        -2.932153,2.676638,0.540364,
        -2.932153,2.676780,-1.901158,
        -2.932153,2.681411,-2.667965,
        -2.932153,2.682152,0.245751,
        -2.932153,2.682859,-0.817071,
        -2.932153,2.705107,1.062142,
        -2.932153,2.717705,2.625995,
        -2.932153,2.721001,2.060097,
        -2.932153,2.722319,-2.954069,
        -2.932153,2.729840,-1.368237,
        -2.932153,2.734349,-0.289298,
        -2.932153,2.738409,2.975671,
        -2.932153,2.740366,1.374341,
        -2.932153,2.744633,-2.391311,
        -2.932153,2.747398,1.727166,
        -2.932153,2.763799,0.760372,
        -2.932153,2.768238,-1.679243,
        -2.932153,2.779245,-2.046268,
        -2.932153,2.780208,-0.626453,
        -2.932153,2.780471,0.027732,
        -2.932153,2.784696,-1.013128,
        -2.932153,2.789817,0.395431,
        -2.932153,2.801154,2.353143,
        -2.932153,2.807091,-2.687477,
        -2.932153,2.825762,1.085035,
        -2.932153,2.830209,2.734943,
        -2.932153,2.833517,-3.085508,
        -2.932153,2.838301,1.980061,
        -2.932153,2.849871,1.512012,
        -2.932153,2.850089,-1.343899,
        -2.932153,2.856905,-0.318441,
        -2.932153,2.870445,-2.291001,
        -2.932153,2.885017,0.655562,
        -2.932153,2.886428,-1.793217,
        -2.932153,2.890337,-0.825633,
        -2.932153,2.898932,0.139973,
        -2.932153,2.931927,-2.728768,
        -2.932153,2.932096,2.377790,
        -2.932153,2.937409,2.947245,
        -2.932153,2.948048,1.133127,
        -2.932153,2.950896,1.797889,
        -2.932153,2.970019,-1.282382,
        -2.932153,2.978333,-0.412090,
        -2.932153,2.996437,-2.089121,
        -2.932153,3.011274,0.434110,
        -2.932153,3.050410,3.138492,
        -2.932153,3.064492,1.647259,
        -2.932153,3.088868,-0.989501,
        -2.827433,0.053884,-2.408394,
        -2.827433,0.077645,1.453433,
        -2.827433,0.088408,-0.311160,
        -2.827433,0.142314,2.590387,
        -2.827433,0.147206,-1.385164,
        -2.827433,0.163506,0.573203,
        -2.827433,0.165485,-2.875396,
        -2.827433,0.182564,-2.125821,
        -2.827433,0.190569,1.929141,
        -2.827433,0.201102,1.249082,
        -2.827433,0.201397,-0.716646,
        -2.827433,0.210568,-0.069559,
        -2.827433,0.251672,2.848900,
        -2.827433,0.266262,-1.678068,
        -2.827433,0.269108,-1.215563,
        -2.827433,0.273777,-2.506756,
        -2.827433,0.279854,0.814581,
        -2.827433,0.281664,-2.969147,
        -2.827433,0.283625,2.405603,
        -2.827433,0.284058,0.357350,
        -2.827433,0.294107,1.593880,
        -2.827433,0.302491,-0.400410,
        -2.827433,0.308059,2.021491,
        -2.827433,0.312219,-2.072057,
        -2.827433,0.321905,-0.829121,
        -2.827433,0.329984,1.201621,
        -2.827433,0.339714,-0.013712,
        -2.827433,0.368784,2.974974,
        -2.827433,0.376428,-1.495532,
        -2.827433,0.378103,2.646436,
        -2.827433,0.385088,0.590355,
        -2.827433,0.386248,-2.690943,
        -2.827433,0.389851,-2.340031,
        -2.827433,0.391502,-1.171374,
        -2.827433,0.398249,-1.812694,
        -2.827433,0.400542,-2.995729,
        -2.827433,0.404273,0.918163,
        -2.827433,0.405617,1.773533,
        -2.827433,0.409009,-0.570243,
        -2.827433,0.410700,0.270433,
        -2.827433,0.415893,1.464890,
        -2.827433,0.420680,2.088164,
        -2.827433,0.424766,-0.258729,
        -2.827433,0.426507,2.367872,
        -2.827433,0.441292,-0.890727,
        -2.827433,0.451839,-2.078193,
        -2.827433,0.459270,1.191562,
        -2.827433,0.470622,0.014002,
        -2.827433,0.481292,-2.529098,
        -2.827433,0.482460,2.849843,
        -2.827433,0.492205,-1.397963,
        -2.827433,0.493314,3.117264,
        -2.827433,0.497870,0.727828,
        -2.827433,0.499783,-1.654379,
        -2.827433,0.502651,0.463333,
        -2.827433,0.505463,2.594557,
        -2.827433,0.506104,-2.874910,
        -2.827433,0.513579,-1.147354,
        -2.827433,0.518569,1.644237,
        -2.827433,0.519001,1.900488,
        -2.827433,0.519277,-0.673809,
        -2.827433,0.524769,-0.421955,
        -2.827433,0.525564,-2.288838,
        -2.827433,0.532541,0.978737,
        -2.827433,0.534610,2.154808,
        -2.827433,0.535020,-1.891846,
        -2.827433,0.540208,1.400497,
        -2.827433,0.542006,0.228337,
        -2.827433,0.550762,-0.180771,
        -2.827433,0.560531,2.378474,
        -2.827433,0.561629,-0.924576,
        -2.827433,0.567465,-2.679392,
        -2.827433,0.588527,1.181755,
        -2.827433,0.591643,-2.098364,
        -2.827433,0.594963,2.997575,
        -2.827433,0.599643,-3.045235,
        -2.827433,0.599962,2.773556,
        -2.827433,0.602620,0.029270,
        -2.827433,0.605711,0.598454,
        -2.827433,0.607708,-1.541026,
        -2.827433,0.609360,-2.464379,
        -2.827433,0.610738,-1.326786,
        -2.827433,0.616968,0.820489,
        -2.827433,0.626379,1.773610,
        -2.827433,0.628626,-1.751180,
        -2.827433,0.630671,0.391949,
        -2.827433,0.632622,-0.737642,
        -2.827433,0.632708,2.566722,
        -2.827433,0.632797,1.989858,
        -2.827433,0.634455,-0.534790,
        -2.827433,0.636205,1.561170,
        -2.827433,0.637681,-1.123423,
        -2.827433,0.641441,-2.842109,
        -2.827433,0.644807,-0.327139,
        -2.827433,0.652440,2.197129,
        -2.827433,0.660897,-2.278593,
        -2.827433,0.664008,1.360091,
        -2.827433,0.670850,-1.945257,
        -2.827433,0.671396,1.004859,
        -2.827433,0.673985,0.203855,
        -2.827433,0.680326,-0.133370,
        -2.827433,0.685083,-0.942363,
        -2.827433,0.688204,-2.632204,
        -2.827433,0.689777,2.385011,
        -2.827433,0.700065,3.115430,
        -2.827433,0.704080,2.911636,
        -2.827433,0.712883,1.174216,
        -2.827433,0.715701,0.699906,
        -2.827433,0.718851,-1.442215,
        -2.827433,0.721019,2.721536,
        -2.827433,0.725300,-2.975134,
        -2.827433,0.727136,-1.633288,
        -2.827433,0.727399,-2.120567,
        -2.827433,0.730751,0.522488,
        -2.827433,0.732892,-1.254330,
        -2.827433,0.734333,-2.446184,
        -2.827433,0.735125,0.037270,
        -2.827433,0.737893,1.866463,
        -2.827433,0.740650,1.681075,
        -2.827433,0.745971,-0.781286,
        -2.827433,0.749432,-0.442559,
        -2.827433,0.750102,2.050636,
        -2.827433,0.751046,0.863821,
        -2.827433,0.757018,-0.608833,
        -2.827433,0.757046,-1.816019,
        -2.827433,0.757389,1.501184,
        -2.827433,0.759800,2.546181,
        -2.827433,0.759859,0.351923,
        -2.827433,0.761141,-2.781102,
        -2.827433,0.769035,-1.078574,
        -2.827433,0.770166,-0.266604,
        -2.827433,0.772218,2.227380,
        -2.827433,0.788382,1.331051,
        -2.827433,0.790269,-2.282409,
        -2.827433,0.803410,-1.984567,
        -2.827433,0.803813,0.191520,
        -2.827433,0.806832,3.022155,
        -2.827433,0.807476,-2.597675,
        -2.827433,0.812512,-0.107335,
        -2.827433,0.814161,1.010949,
        -2.827433,0.817732,-3.084486,
        -2.827433,0.817846,2.390688,
        -2.827433,0.819323,2.845960,
        -2.827433,0.819430,-0.915017,
        -2.827433,0.829198,-1.532491,
        -2.827433,0.829531,0.665695,
        -2.827433,0.829816,1.167147,
        -2.827433,0.833613,-1.357492,
        -2.827433,0.842945,-2.907318,
        -2.827433,0.844883,2.679708,
        -2.827433,0.848189,-1.705003,
        -2.827433,0.849888,1.767901,
        -2.827433,0.851126,0.493370,
        -2.827433,0.852911,1.933394,
        -2.827433,0.857516,-2.432863,
        -2.827433,0.858261,-1.191729,
        -2.827433,0.858279,-2.138895,
        -2.827433,0.859723,-0.734928,
        -2.827433,0.859754,1.603292,
        -2.827433,0.861655,-0.394821,
        -2.827433,0.864300,0.043549,
        -2.827433,0.866940,0.851229,
        -2.827433,0.869748,2.095969,
        -2.827433,0.877352,-0.574668,
        -2.827433,0.879254,-2.733338,
        -2.827433,0.882554,1.443757,
        -2.827433,0.884060,0.334199,
        -2.827433,0.885626,-1.866061,
        -2.827433,0.889834,2.529957,
        -2.827433,0.893140,2.251698,
        -2.827433,0.897463,-0.240822,
        -2.827433,0.898758,-1.031948,
        -2.827433,0.913918,1.287618,
        -2.827433,0.916555,-2.287173,
        -2.827433,0.918305,3.107845,
        -2.827433,0.919522,2.948277,
        -2.827433,0.927138,-2.569686,
        -2.827433,0.929031,0.185383,
        -2.827433,0.933766,-3.015917,
        -2.827433,0.934873,-2.010822,
        -2.827433,0.935636,-1.445855,
        -2.827433,0.936702,-0.866461,
        -2.827433,0.937395,0.729108,
        -2.827433,0.937910,2.787575,
        -2.827433,0.939144,-0.092962,
        -2.827433,0.941718,1.125085,
        -2.827433,0.942873,-1.607680,
        -2.827433,0.943701,0.980328,
        -2.827433,0.947774,2.395256,
        -2.827433,0.952433,0.584347,
        -2.827433,0.954456,-1.287293,
        -2.827433,0.958731,-2.852552,
        -2.827433,0.961838,1.839284,
        -2.827433,0.961898,-0.469425,
        -2.827433,0.965426,1.686615,
        -2.827433,0.970108,1.990771,
        -2.827433,0.972820,-1.762780,
        -2.827433,0.973500,-0.711881,
        -2.827433,0.975253,2.641082,
        -2.827433,0.977824,0.441928,
        -2.827433,0.981485,-2.424766,
        -2.827433,0.981877,1.537100,
        -2.827433,0.982609,-1.136920,
        -2.827433,0.984265,-2.156073,
        -2.827433,0.989150,0.045493,
        -2.827433,0.993052,2.137478,
        -2.827433,0.997766,-2.693871,
        -2.827433,0.999847,-0.334738,
        -2.827433,1.001920,0.848054,
        -2.827433,1.006146,1.389734,
        -2.827433,1.012180,0.304066,
        -2.827433,1.016938,2.270674,
        -2.827433,1.017371,-0.979790,
        -2.827433,1.019495,-1.894487,
        -2.827433,1.023796,-0.583691,
        -2.827433,1.026525,2.506945,
        -2.827433,1.027004,-0.203653,
        -2.827433,1.028027,3.036573,
        -2.827433,1.031435,1.241021,
        -2.827433,1.032316,2.884221,
        -2.827433,1.035863,-3.110475,
        -2.827433,1.042085,-2.292950,
        -2.827433,1.043272,-1.522986,
        -2.827433,1.046265,-2.958847,
        -2.827433,1.049272,-1.376344,
        -2.827433,1.050039,-2.552086,
        -2.827433,1.053682,0.689863,
        -2.827433,1.055458,0.170394,
        -2.827433,1.056622,-0.831866,
        -2.827433,1.057515,1.097998,
        -2.827433,1.058706,-2.036438,
        -2.827433,1.061389,-1.672248,
        -2.827433,1.063758,2.743093,
        -2.827433,1.070157,-0.073422,
        -2.827433,1.072154,1.761506,
        -2.827433,1.074029,1.909193,
        -2.827433,1.074204,0.545276,
        -2.827433,1.074714,-2.805147,
        -2.827433,1.076262,0.966104,
        -2.827433,1.077642,-1.227073,
        -2.827433,1.084151,1.618525,
        -2.827433,1.085085,2.377673,
        -2.827433,1.086822,-0.441112,
        -2.827433,1.087142,2.058815,
        -2.827433,1.098383,-0.696430,
        -2.827433,1.099138,-1.084033,
        -2.827433,1.103208,1.477773,
        -2.827433,1.103783,0.408254,
        -2.827433,1.106275,-2.420800,
        -2.827433,1.108079,2.612943,
        -2.827433,1.109345,-2.172718,
        -2.827433,1.115957,-1.787597,
        -2.827433,1.118930,0.045983,
        -2.827433,1.119777,0.834702,
        -2.827433,1.121943,-2.669042,
        -2.827433,1.124930,2.219917,
        -2.827433,1.125217,1.336294,
        -2.827433,1.130958,2.972931,
        -2.827433,1.131400,-0.306629,
        -2.827433,1.137135,-0.936514,
        -2.827433,1.139125,-1.924528,
        -2.827433,1.139877,0.276382,
        -2.827433,1.142003,-3.058022,
        -2.827433,1.147363,-0.558448,
        -2.827433,1.150348,-0.176418,
        -2.827433,1.151198,1.197016,
        -2.827433,1.153930,-1.597104,
        -2.827433,1.155389,2.837148,
        -2.827433,1.155458,-1.460623,
        -2.827433,1.157331,-2.906362,
        -2.827433,1.158594,2.487618,
        -2.827433,1.162103,-1.322553,
        -2.827433,1.165407,3.092861,
        -2.827433,1.168025,0.717075,
        -2.827433,1.168406,-2.301519,
        -2.827433,1.174782,-2.538755,
        -2.827433,1.177998,1.061256,
        -2.827433,1.178898,1.827305,
        -2.827433,1.181331,-2.060330,
        -2.827433,1.182540,-0.798952,
        -2.827433,1.184569,0.147517,
        -2.827433,1.184734,1.968919,
        -2.827433,1.184804,0.593386,
        -2.827433,1.188278,1.690808,
        -2.827433,1.194360,2.110440,
        -2.827433,1.194776,2.710455,
        -2.827433,1.197745,-1.173705,
        -2.827433,1.199898,-2.777257,
        -2.827433,1.201164,2.355931,
        -2.827433,1.203760,1.556404,
        -2.827433,1.205845,-0.415993,
        -2.827433,1.206585,0.933648,
        -2.827433,1.206950,-0.037248,
        -2.827433,1.209341,0.471162,
        -2.827433,1.213483,-1.039417,
        -2.827433,1.218146,-0.672643,
        -2.827433,1.221608,1.421669,
        -2.827433,1.226043,-1.693533,
        -2.827433,1.232870,-1.822773,
        -2.827433,1.234375,-2.421194,
        -2.827433,1.234531,-2.190148,
        -2.827433,1.239414,0.348959,
        -2.827433,1.240183,2.587255,
        -2.827433,1.242312,2.941234,
        -2.827433,1.244075,1.286644,
        -2.827433,1.245961,-2.997461,
        -2.827433,1.246495,-2.646872,
        -2.827433,1.247736,2.242810,
        -2.827433,1.253604,0.812222,
        -2.827433,1.257816,-0.894317,
        -2.827433,1.258247,-1.953634,
        -2.827433,1.258993,-0.284099,
        -2.827433,1.262380,-0.157609,
        -2.827433,1.262791,-1.549229,
        -2.827433,1.263772,-0.533073,
        -2.827433,1.266731,-1.412461,
        -2.827433,1.268591,-3.122146,
        -2.827433,1.269614,-1.275246,
        -2.827433,1.270745,1.152533,
        -2.827433,1.273606,0.224545,
        -2.827433,1.279716,2.812101,
        -2.827433,1.283786,2.462124,
        -2.827433,1.284798,0.078209,
        -2.827433,1.285402,-2.882049,
        -2.827433,1.286936,1.886053,
        -2.827433,1.289408,0.683122,
        -2.827433,1.294736,1.756184,
        -2.827433,1.296978,-2.311511,
        -2.827433,1.298546,2.020601,
        -2.827433,1.300924,1.022478,
        -2.827433,1.303305,3.043152,
        -2.827433,1.304210,-2.084225,
        -2.827433,1.305241,-2.530249,
        -2.827433,1.307513,1.628003,
        -2.827433,1.310722,0.558267,
        -2.827433,1.314142,2.139329,
        -2.827433,1.316087,-0.777971,
        -2.827433,1.317907,-1.121023,
        -2.827433,1.321019,-2.749776,
        -2.827433,1.321609,2.687717,
        -2.827433,1.322568,1.499415,
        -2.827433,1.322972,-0.395699,
        -2.827433,1.325286,-0.994192,
        -2.827433,1.329146,-0.648513,
        -2.827433,1.329795,-0.042570,
        -2.827433,1.331651,2.344497,
        -2.827433,1.336522,0.435098,
        -2.827433,1.339410,0.900087,
        -2.827433,1.339550,1.370353,
        -2.827433,1.340441,-1.731580,
        -2.827433,1.351855,-1.855234,
        -2.827433,1.360161,-3.018404,
        -2.827433,1.361031,-1.613971,
        -2.827433,1.361196,2.914780,
        -2.827433,1.361998,-2.207254,
        -2.827433,1.362302,1.239384,
        -2.827433,1.365228,2.563808,
        -2.827433,1.365452,0.310522,
        -2.827433,1.367028,-2.420041,
        -2.827433,1.373422,-1.484281,
        -2.827433,1.374245,-1.226417,
        -2.827433,1.375020,-0.170393,
        -2.827433,1.376560,-1.355481,
        -2.827433,1.377070,-1.980984,
        -2.827433,1.377578,-2.636339,
        -2.827433,1.378096,0.774626,
        -2.827433,1.379086,-0.508192,
        -2.827433,1.382562,0.177141,
        -2.827433,1.388849,2.233847,
        -2.827433,1.389268,-0.886349,
        -2.827433,1.391324,1.109396,
        -2.827433,1.391699,-0.288725,
        -2.827433,1.392869,3.136903,
        -2.827433,1.393612,1.952016,
        -2.827433,1.395545,-2.850807,
        -2.827433,1.401926,1.822930,
        -2.827433,1.402149,2.788898,
        -2.827433,1.409552,0.647980,
        -2.827433,1.410840,0.056219,
        -2.827433,1.411944,2.444377,
        -2.827433,1.413639,1.697982,
        -2.827433,1.426479,1.573427,
        -2.827433,1.426725,2.102242,
        -2.827433,1.427868,0.984637,
        -2.827433,1.429580,-2.107232,
        -2.827433,1.431375,-2.313269,
        -2.827433,1.431907,3.015410,
        -2.827433,1.432325,-1.029054,
        -2.827433,1.433341,0.524268,
        -2.827433,1.435181,-0.739771,
        -2.827433,1.437208,-2.529085,
        -2.827433,1.439041,-0.617966,
        -2.827433,1.441473,1.448544,
        -2.827433,1.444527,2.664744,
        -2.827433,1.450513,-0.400625,
        -2.827433,1.451907,-0.066823,
        -2.827433,1.452754,-2.738631,
        -2.827433,1.454061,-2.950416,
        -2.827433,1.456736,1.323154,
        -2.827433,1.458370,0.400433,
        -2.827433,1.459314,-1.765003,
        -2.827433,1.459902,-1.145835,
        -2.827433,1.464952,2.329245,
        -2.827433,1.466050,0.862104,
        -2.827433,1.471330,-1.886160,
        -2.827433,1.475119,-1.648255,
        -2.827433,1.480878,1.194576,
        -2.827433,1.480942,0.273235,
        -2.827433,1.481086,2.890269,
        -2.827433,1.484118,-1.529156,
        -2.827433,1.484741,-1.279219,
        -2.827433,1.487681,-1.406069,
        -2.827433,1.487886,-3.066037,
        -2.827433,1.490555,2.543284,
        -2.827433,1.492424,-0.190932,
        -2.827433,1.495002,-0.839810,
        -2.827433,1.495297,2.007188,
        -2.827433,1.496414,-2.009267,
        -2.827433,1.498261,-2.425270,
        -2.827433,1.498907,-2.207327,
        -2.827433,1.500204,0.737638,
        -2.827433,1.502469,0.151156,
        -2.827433,1.507448,-0.514782,
        -2.827433,1.509287,2.204647,
        -2.827433,1.510518,1.884920,
        -2.827433,1.511948,-2.632360,
        -2.827433,1.515235,1.068565,
        -2.827433,1.522237,3.102018,
        -2.827433,1.522373,1.764991,
        -2.827433,1.522416,-0.954956,
        -2.827433,1.523102,2.765920,
        -2.827433,1.524811,-2.844718,
        -2.827433,1.527261,-0.304916,
        -2.827433,1.529788,0.613208,
        -2.827433,1.533648,1.644808,
        -2.827433,1.535018,0.028869,
        -2.827433,1.541082,2.424608,
        -2.827433,1.546010,1.523462,
        -2.827433,1.551135,-0.642994,
        -2.827433,1.553895,0.489607,
        -2.827433,1.555041,0.947813,
        -2.827433,1.559671,-1.074934,
        -2.827433,1.560345,2.985899,
        -2.827433,1.560487,-2.323400,
        -2.827433,1.561127,1.401111,
        -2.827433,1.567193,2.643172,
        -2.827433,1.568892,2.098362,
        -2.827433,1.571702,-2.098435,
        -2.827433,1.572701,-2.528730,
        -2.827433,1.573572,1.278593,
        -2.827433,1.574849,-0.091505,
        -2.827433,1.576334,0.365236,
        -2.827433,1.577029,-2.962888,
        -2.827433,1.577688,-1.814941,
        -2.827433,1.579542,-0.755917,
        -2.827433,1.579624,-1.199759,
        -2.827433,1.581515,-0.419414,
        -2.827433,1.584581,-2.738702,
        -2.827433,1.586948,-1.696296,
        -2.827433,1.588901,2.301496,
        -2.827433,1.589936,0.825020,
        -2.827433,1.590018,-1.941652,
        -2.827433,1.596968,-1.328358,
        -2.827433,1.597178,-1.575607,
        -2.827433,1.600470,-1.453196,
        -2.827433,1.601408,0.242409,
        -2.827433,1.602259,1.152656,
        -2.827433,1.603064,2.865774,
        -2.827433,1.612085,1.968802,
        -2.827433,1.612720,-3.080498,
        -2.827433,1.616061,2.521531,
        -2.827433,1.616900,-0.207457,
        -2.827433,1.620988,-0.875920,
        -2.827433,1.622114,0.701509,
        -2.827433,1.622749,0.121426,
        -2.827433,1.622993,-2.221629,
        -2.827433,1.629346,-0.542669,
        -2.827433,1.630007,1.844552,
        -2.827433,1.634747,-2.426198,
        -2.827433,1.642346,1.722617,
        -2.827433,1.643925,1.034629,
        -2.827433,1.644848,2.743774,
        -2.827433,1.645258,2.188109,
        -2.827433,1.645417,-2.634725,
        -2.827433,1.647395,3.085437,
        -2.827433,1.648538,-2.855044,
        -2.827433,1.649932,0.576808,
        -2.827433,1.652224,-0.996105,
        -2.827433,1.652871,1.600681,
        -2.827433,1.657997,-0.325127,
        -2.827433,1.661936,0.001037,
        -2.827433,1.665187,1.477402,
        -2.827433,1.665496,2.397872,
        -2.827433,1.673267,-0.665453,
        -2.827433,1.673557,0.452526,
        -2.827433,1.679277,-1.122421,
        -2.827433,1.679696,0.912482,
        -2.827433,1.682420,1.353922,
        -2.827433,1.685031,-2.009038,
        -2.827433,1.685196,2.967046,
        -2.827433,1.690107,2.620910,
        -2.827433,1.690453,-2.123076,
        -2.827433,1.691693,-1.871572,
        -2.827433,1.691817,2.061688,
        -2.827433,1.692644,1.237670,
        -2.827433,1.696386,0.328123,
        -2.827433,1.697162,-1.249026,
        -2.827433,1.697488,-2.322785,
        -2.827433,1.698726,-2.973576,
        -2.827433,1.699172,-1.749717,
        -2.827433,1.704478,-0.106295,
        -2.827433,1.707244,-0.444770,
        -2.827433,1.707500,-2.531497,
        -2.827433,1.707714,-0.787321,
        -2.827433,1.708430,-1.627552,
        -2.827433,1.710583,-1.376890,
        -2.827433,1.712415,-2.747921,
        -2.827433,1.713125,-1.503073,
        -2.827433,1.713452,0.789419,
        -2.827433,1.720612,2.279888,
        -2.827433,1.724852,0.202831,
        -2.827433,1.724868,2.846246,
        -2.827433,1.728940,1.931535,
        -2.827433,1.732838,1.127584,
        -2.827433,1.735719,-3.092485,
        -2.827433,1.739676,2.496017,
        -2.827433,1.739971,-0.228876,
        -2.827433,1.743328,0.663347,
        -2.827433,1.744927,-0.911892,
        -2.827433,1.745952,0.084733,
        -2.827433,1.748077,1.804354,
        -2.827433,1.752959,-0.567548,
        -2.827433,1.760169,1.678132,
        -2.827433,1.763887,-2.218450,
        -2.827433,1.766737,2.722752,
        -2.827433,1.767999,1.004737,
        -2.827433,1.769149,1.550298,
        -2.827433,1.770119,-2.427175,
        -2.827433,1.770460,2.154208,
        -2.827433,1.770514,3.070982,
        -2.827433,1.770735,0.537245,
        -2.827433,1.772673,-2.864760,
        -2.827433,1.773527,-1.039582,
        -2.827433,1.776419,-2.642421,
        -2.827433,1.780769,1.416330,
        -2.827433,1.786634,-0.346247,
        -2.827433,1.788829,-1.947553,
        -2.827433,1.793737,0.410270,
        -2.827433,1.794553,2.374618,
        -2.827433,1.794699,-0.694061,
        -2.827433,1.797469,-1.169105,
        -2.827433,1.803705,0.881076,
        -2.827433,1.804124,1.264249,
        -2.827433,1.807825,-2.090455,
        -2.827433,1.808822,-1.817375,
        -2.827433,1.809020,2.949741,
        -2.827433,1.812466,2.597595,
        -2.827433,1.813159,2.025395,
        -2.827433,1.813816,-1.298435,
        -2.827433,1.815016,0.278951,
        -2.827433,1.816510,-0.008631,
        -2.827433,1.817772,-1.689625,
        -2.827433,1.821360,-2.984352,
        -2.827433,1.822312,-0.127211,
        -2.827433,1.823386,-1.429807,
        -2.827433,1.823834,-1.560483,
        -2.827433,1.832060,-0.821627,
        -2.827433,1.832690,-0.468440,
        -2.827433,1.833945,-2.320278,
        -2.827433,1.834787,0.751968,
        -2.827433,1.840579,-2.536445,
        -2.827433,1.841033,-2.756932,
        -2.827433,1.846350,1.893983,
        -2.827433,1.846644,1.113766,
        -2.827433,1.846833,2.826543,
        -2.827433,1.847325,2.249051,
        -2.827433,1.854119,0.134161,
        -2.827433,1.858863,-3.106022,
        -2.827433,1.865046,1.763384,
        -2.827433,1.865619,-0.952036,
        -2.827433,1.865912,0.624453,
        -2.827433,1.866694,2.473686,
        -2.827433,1.867524,-0.240083,
        -2.827433,1.875561,1.632831,
        -2.827433,1.876049,-0.594378,
        -2.827433,1.880509,1.500204,
        -2.827433,1.883504,1.360457,
        -2.827433,1.886971,-2.196665,
        -2.827433,1.888979,0.982922,
        -2.827433,1.889815,2.701570,
        -2.827433,1.892144,-2.005340,
        -2.827433,1.892348,0.496522,
        -2.827433,1.893281,-1.083976,
        -2.827433,1.893828,3.053453,
        -2.827433,1.895117,2.121107,
        -2.827433,1.898741,-2.875735,
        -2.827433,1.904429,-2.428029,
        -2.827433,1.908408,-2.648903,
        -2.827433,1.911828,-0.364489,
        -2.827433,1.912224,0.366565,
        -2.827433,1.914854,1.214468,
        -2.827433,1.915467,-1.883143,
        -2.827433,1.915763,-0.724679,
        -2.827433,1.915790,-1.216366,
        -2.827433,1.922256,2.347806,
        -2.827433,1.922616,0.844633,
        -2.827433,1.925185,0.231170,
        -2.827433,1.927474,-1.754616,
        -2.827433,1.928342,-1.351174,
        -2.827433,1.933328,-1.623328,
        -2.827433,1.933878,0.034150,
        -2.827433,1.934585,-1.487979,
        -2.827433,1.934802,2.927634,
        -2.827433,1.934862,1.989776,
        -2.827433,1.936327,-0.099350,
        -2.827433,1.938088,2.578507,
        -2.827433,1.945677,-2.999805,
        -2.827433,1.954572,-0.859208,
        -2.827433,1.957868,-0.490784,
        -2.827433,1.958989,0.712772,
        -2.827433,1.960212,-2.307561,
        -2.827433,1.964774,1.855507,
        -2.827433,1.964971,-2.100533,
        -2.827433,1.969112,1.088273,
        -2.827433,1.971438,-2.765415,
        -2.827433,1.972312,2.805134,
        -2.827433,1.974950,2.220596,
        -2.827433,1.975135,-2.539127,
        -2.827433,1.982107,1.720829,
        -2.827433,1.983195,-3.127984,
        -2.827433,1.983318,-0.247182,
        -2.827433,1.986720,-0.993220,
        -2.827433,1.990381,1.586005,
        -2.827433,1.990470,0.580283,
        -2.827433,1.990523,1.450326,
        -2.827433,1.993909,1.312897,
        -2.827433,1.994694,2.452813,
        -2.827433,1.999226,-0.620717,
        -2.827433,2.005145,0.941012,
        -2.827433,2.007301,0.142337,
        -2.827433,2.014149,0.444138,
        -2.827433,2.014338,-1.127392,
        -2.827433,2.018131,3.021778,
        -2.827433,2.019676,-1.980944,
        -2.827433,2.019912,2.686612,
        -2.827433,2.020545,2.088672,
        -2.827433,2.026074,0.300945,
        -2.827433,2.026196,-2.889057,
        -2.827433,2.031959,-1.268178,
        -2.827433,2.032917,-2.199917,
        -2.827433,2.033572,-1.842884,
        -2.827433,2.034712,-2.419505,
        -2.827433,2.035818,-0.379479,
        -2.827433,2.036004,-0.757062,
        -2.827433,2.040501,-1.411488,
        -2.827433,2.042874,-1.701367,
        -2.827433,2.043060,-2.653746,
        -2.827433,2.043729,-1.557938,
        -2.827433,2.048538,0.806623,
        -2.827433,2.049006,1.199930,
        -2.827433,2.049502,-0.137989,
        -2.827433,2.052181,2.326959,
        -2.827433,2.052794,-0.008526,
        -2.827433,2.057608,1.951726,
        -2.827433,2.065568,2.567730,
        -2.827433,2.070371,2.879808,
        -2.827433,2.071256,-3.018961,
        -2.827433,2.079746,-0.897699,
        -2.827433,2.082784,1.042403,
        -2.827433,2.083752,-0.511086,
        -2.827433,2.084401,1.810502,
        -2.827433,2.085731,0.671081,
        -2.827433,2.097496,1.373172,
        -2.827433,2.099185,1.667006,
        -2.827433,2.100482,-2.301084,
        -2.827433,2.103118,2.194745,
        -2.827433,2.103223,1.521685,
        -2.827433,2.103695,-2.777707,
        -2.827433,2.107523,3.129581,
        -2.827433,2.108277,-2.095907,
        -2.827433,2.108891,-2.532166,
        -2.827433,2.111003,-0.268632,
        -2.827433,2.112276,-1.027246,
        -2.827433,2.114380,0.531457,
        -2.827433,2.119166,0.100552,
        -2.827433,2.119564,2.754818,
        -2.827433,2.122474,-0.648314,
        -2.827433,2.125183,2.445598,
        -2.827433,2.129501,0.238290,
        -2.827433,2.131178,0.386663,
        -2.827433,2.131223,-1.174311,
        -2.827433,2.134269,-1.948042,
        -2.827433,2.134684,0.907065,
        -2.827433,2.137615,2.988806,
        -2.827433,2.145635,-1.327708,
        -2.827433,2.146467,2.055081,
        -2.827433,2.146582,-1.793788,
        -2.827433,2.149210,-1.482321,
        -2.827433,2.152901,-1.636788,
        -2.827433,2.156337,1.151558,
        -2.827433,2.156474,-2.912065,
        -2.827433,2.157194,-0.792563,
        -2.827433,2.166724,-0.401638,
        -2.827433,2.166980,-0.134879,
        -2.827433,2.173644,-2.403429,
        -2.827433,2.175711,-2.655523,
        -2.827433,2.177132,2.600419,
        -2.827433,2.178879,0.767722,
        -2.827433,2.180752,1.908485,
        -2.827433,2.181338,2.313721,
        -2.827433,2.185580,1.287673,
        -2.827433,2.195400,-2.217546,
        -2.827433,2.200418,-3.056442,
        -2.827433,2.201829,-0.001735,
        -2.827433,2.203804,1.756176,
        -2.827433,2.211602,1.443930,
        -2.827433,2.212686,-0.544809,
        -2.827433,2.213937,0.621892,
        -2.827433,2.214075,1.600691,
        -2.827433,2.216256,1.020029,
        -2.827433,2.219060,2.886631,
        -2.827433,2.223375,-0.906185,
        -2.827433,2.226002,-1.060248,
        -2.827433,2.226413,-2.064034,
        -2.827433,2.230554,2.171334,
        -2.827433,2.232413,2.735777,
        -2.827433,2.233709,3.070830,
        -2.827433,2.234383,-2.792065,
        -2.827433,2.237183,0.468799,
        -2.827433,2.237695,-0.271388,
        -2.827433,2.239419,0.151693,
        -2.827433,2.243347,-2.521049,
        -2.827433,2.244035,-1.229456,
        -2.827433,2.244355,0.309899,
        -2.827433,2.245175,2.460236,
        -2.827433,2.245628,-1.899216,
        -2.827433,2.245973,-0.702234,
        -2.827433,2.255601,-1.728131,
        -2.827433,2.256702,-1.401367,
        -2.827433,2.261466,-1.557500,
        -2.827433,2.267390,0.878499,
        -2.827433,2.272099,2.018594,
        -2.827433,2.278618,1.170386,
        -2.827433,2.283304,-2.345584,
        -2.827433,2.285175,-2.943090,
        -2.827433,2.293000,-0.124451,
        -2.827433,2.294058,-0.421070,
        -2.827433,2.303283,1.855938,
        -2.827433,2.304968,2.608978,
        -2.827433,2.305347,2.311346,
        -2.827433,2.307188,-2.656134,
        -2.827433,2.307301,1.336200,
        -2.827433,2.309655,0.726020,
        -2.827433,2.317634,-2.185225,
        -2.827433,2.321510,1.685664,
        -2.827433,2.324376,-3.111006,
        -2.827433,2.324731,1.511186,
        -2.827433,2.329587,0.035498,
        -2.827433,2.330158,2.977264,
        -2.827433,2.335096,-0.588468,
        -2.827433,2.336046,-1.115799,
        -2.827433,2.338060,2.803613,
        -2.827433,2.338083,-0.943400,
        -2.827433,2.340210,0.562145,
        -2.827433,2.341931,1.023926,
        -2.827433,2.343265,-0.779061,
        -2.827433,2.343434,-2.013692,
        -2.827433,2.353029,-1.310529,
        -2.827433,2.354868,0.211543,
        -2.827433,2.355120,0.388593,
        -2.827433,2.355165,-1.833173,
        -2.827433,2.355462,2.151441,
        -2.827433,2.362948,-2.487477,
        -2.827433,2.363101,-2.810456,
        -2.827433,2.363394,-0.272918,
        -2.827433,2.364509,-1.644533,
        -2.827433,2.370465,2.466824,
        -2.827433,2.386110,-1.481351,
        -2.827433,2.394960,0.862133,
        -2.827433,2.395716,1.977184,
        -2.827433,2.396339,1.200900,
        -2.827433,2.407941,-2.316753,
        -2.827433,2.409450,-2.988432,
        -2.827433,2.417202,-0.103363,
        -2.827433,2.417386,-0.443097,
        -2.827433,2.417950,2.666662,
        -2.827433,2.423190,1.789783,
        -2.827433,2.423692,1.392696,
        -2.827433,2.431122,3.086379,
        -2.827433,2.431406,-2.647820,
        -2.827433,2.431584,2.307631,
        -2.827433,2.434090,1.592818,
        -2.827433,2.435993,0.683844,
        -2.827433,2.440898,-2.133300,
        -2.827433,2.442222,-1.195837,
        -2.827433,2.443112,2.883761,
        -2.827433,2.449191,-0.633651,
        -2.827433,2.450821,0.084741,
        -2.827433,2.453499,-1.008861,
        -2.827433,2.460010,-1.938011,
        -2.827433,2.460199,-0.822008,
        -2.827433,2.460902,0.490079,
        -2.827433,2.461774,-1.734913,
        -2.827433,2.463986,1.034407,
        -2.827433,2.467647,0.287667,
        -2.827433,2.479849,2.126925,
        -2.827433,2.485907,-2.835471,
        -2.827433,2.485945,-1.374853,
        -2.827433,2.487747,2.505912,
        -2.827433,2.487957,-0.274279,
        -2.827433,2.489741,-2.468428,
        -2.827433,2.512538,1.239900,
        -2.827433,2.517528,1.923289,
        -2.827433,2.518454,-1.565852,
        -2.827433,2.519348,0.843503,
        -2.827433,2.522766,-3.059231,
        -2.827433,2.528686,2.737228,
        -2.827433,2.534798,-2.266671,
        -2.827433,2.536533,1.466274,
        -2.827433,2.538504,1.699812,
        -2.827433,2.539516,-0.476942,
        -2.827433,2.539532,-0.074548,
        -2.827433,2.543197,2.990252,
        -2.827433,2.554022,2.320619,
        -2.827433,2.556919,-2.655005,
        -2.827433,2.558146,0.627191,
        -2.827433,2.558223,-1.164189,
        -2.827433,2.563452,-2.044028,
        -2.827433,2.568462,0.150655,
        -2.827433,2.569130,-0.700837,
        -2.827433,2.570050,-1.804492,
        -2.827433,2.572710,-0.934665,
        -2.827433,2.575994,0.391718,
        -2.827433,2.584774,1.046703,
        -2.827433,2.601432,2.102535,
        -2.827433,2.603559,2.555114,
        -2.827433,2.604150,-2.885208,
        -2.827433,2.608630,-1.378011,
        -2.827433,2.611256,-0.279110,
        -2.827433,2.617344,-2.439480,
        -2.827433,2.627402,1.293552,
        -2.827433,2.635396,1.848850,
        -2.827433,2.635686,2.834950,
        -2.827433,2.637087,3.132730,
        -2.827433,2.642029,0.813274,
        -2.827433,2.645260,1.570203,
        -2.827433,2.646182,-1.614036,
        -2.827433,2.660113,-2.186513,
        -2.827433,2.660483,-0.034608,
        -2.827433,2.661713,-0.529511,
        -2.827433,2.672720,-1.110422,
        -2.827433,2.676626,2.335353,
        -2.827433,2.676638,0.540364,
        -2.827433,2.676780,-1.901158,
        -2.827433,2.681411,-2.667965,
        -2.827433,2.682152,0.245751,
        -2.827433,2.682859,-0.817071,
        -2.827433,2.705107,1.062142,
        -2.827433,2.717705,2.625995,
        -2.827433,2.721001,2.060097,
        -2.827433,2.722319,-2.954069,
        -2.827433,2.729840,-1.368237,
        -2.827433,2.734349,-0.289298,
        -2.827433,2.738409,2.975671,
        -2.827433,2.740366,1.374341,
        -2.827433,2.744633,-2.391311,
        -2.827433,2.747398,1.727166,
        -2.827433,2.763799,0.760372,
        -2.827433,2.768238,-1.679243,
        -2.827433,2.779245,-2.046268,
        -2.827433,2.780208,-0.626453,
        -2.827433,2.780471,0.027732,
        -2.827433,2.784696,-1.013128,
        -2.827433,2.789817,0.395431,
        -2.827433,2.801154,2.353143,
        -2.827433,2.807091,-2.687477,
        -2.827433,2.825762,1.085035,
        -2.827433,2.830209,2.734943,
        -2.827433,2.833517,-3.085508,
        -2.827433,2.838301,1.980061,
        -2.827433,2.849871,1.512012,
        -2.827433,2.850089,-1.343899,
        -2.827433,2.856905,-0.318441,
        -2.827433,2.870445,-2.291001,
        -2.827433,2.885017,0.655562,
        -2.827433,2.886428,-1.793217,
        -2.827433,2.890337,-0.825633,
        -2.827433,2.898932,0.139973,
        -2.827433,2.931927,-2.728768,
        -2.827433,2.932096,2.377790,
        -2.827433,2.937409,2.947245,
        -2.827433,2.948048,1.133127,
        -2.827433,2.950896,1.797889,
        -2.827433,2.970019,-1.282382,
        -2.827433,2.978333,-0.412090,
        -2.827433,2.996437,-2.089121,
        -2.827433,3.011274,0.434110,
        -2.827433,3.050410,3.138492,
        -2.827433,3.064492,1.647259,
        -2.827433,3.088868,-0.989501,
        -2.722714,0.053884,-2.408394,
        -2.722714,0.077645,1.453433,
        -2.722714,0.088408,-0.311160,
        -2.722714,0.142314,2.590387,
        -2.722714,0.147206,-1.385164,
        -2.722714,0.163506,0.573203,
        -2.722714,0.165485,-2.875396,
        -2.722714,0.182564,-2.125821,
        -2.722714,0.190569,1.929141,
        -2.722714,0.201102,1.249082,
        -2.722714,0.201397,-0.716646,
        -2.722714,0.210568,-0.069559,
        -2.722714,0.251672,2.848900,
        -2.722714,0.266262,-1.678068,
        -2.722714,0.269108,-1.215563,
        -2.722714,0.273777,-2.506756,
        -2.722714,0.279854,0.814581,
        -2.722714,0.281664,-2.969147,
        -2.722714,0.283625,2.405603,
        -2.722714,0.284058,0.357350,
        -2.722714,0.294107,1.593880,
        -2.722714,0.302491,-0.400410,
        -2.722714,0.308059,2.021491,
        -2.722714,0.312219,-2.072057,
        -2.722714,0.321905,-0.829121,
        -2.722714,0.329984,1.201621,
        -2.722714,0.339714,-0.013712,
        -2.722714,0.368784,2.974974,
        -2.722714,0.376428,-1.495532,
        -2.722714,0.378103,2.646436,
        -2.722714,0.385088,0.590355,
        -2.722714,0.386248,-2.690943,
        -2.722714,0.389851,-2.340031,
        -2.722714,0.391502,-1.171374,
        -2.722714,0.398249,-1.812694,
        -2.722714,0.400542,-2.995729,
        -2.722714,0.404273,0.918163,
        -2.722714,0.405617,1.773533,
        -2.722714,0.409009,-0.570243,
        -2.722714,0.410700,0.270433,
        -2.722714,0.415893,1.464890,
        -2.722714,0.420680,2.088164,
        -2.722714,0.424766,-0.258729,
        -2.722714,0.426507,2.367872,
        -2.722714,0.441292,-0.890727,
        -2.722714,0.451839,-2.078193,
        -2.722714,0.459270,1.191562,
        -2.722714,0.470622,0.014002,
        -2.722714,0.481292,-2.529098,
        -2.722714,0.482460,2.849843,
        -2.722714,0.492205,-1.397963,
        -2.722714,0.493314,3.117264,
        -2.722714,0.497870,0.727828,
        -2.722714,0.499783,-1.654379,
        -2.722714,0.502651,0.463333,
        -2.722714,0.505463,2.594557,
        -2.722714,0.506104,-2.874910,
        -2.722714,0.513579,-1.147354,
        -2.722714,0.518569,1.644237,
        -2.722714,0.519001,1.900488,
        -2.722714,0.519277,-0.673809,
        -2.722714,0.524769,-0.421955,
        -2.722714,0.525564,-2.288838,
        -2.722714,0.532541,0.978737,
        -2.722714,0.534610,2.154808,
        -2.722714,0.535020,-1.891846,
        -2.722714,0.540208,1.400497,
        -2.722714,0.542006,0.228337,
        -2.722714,0.550762,-0.180771,
        -2.722714,0.560531,2.378474,
        -2.722714,0.561629,-0.924576,
        -2.722714,0.567465,-2.679392,
        -2.722714,0.588527,1.181755,
        -2.722714,0.591643,-2.098364,
        -2.722714,0.594963,2.997575,
        -2.722714,0.599643,-3.045235,
        -2.722714,0.599962,2.773556,
        -2.722714,0.602620,0.029270,
        -2.722714,0.605711,0.598454,
        -2.722714,0.607708,-1.541026,
        -2.722714,0.609360,-2.464379,
        -2.722714,0.610738,-1.326786,
        -2.722714,0.616968,0.820489,
        -2.722714,0.626379,1.773610,
        -2.722714,0.628626,-1.751180,
        -2.722714,0.630671,0.391949,
        -2.722714,0.632622,-0.737642,
        -2.722714,0.632708,2.566722,
        -2.722714,0.632797,1.989858,
        -2.722714,0.634455,-0.534790,
        -2.722714,0.636205,1.561170,
        -2.722714,0.637681,-1.123423,
        -2.722714,0.641441,-2.842109,
        -2.722714,0.644807,-0.327139,
        -2.722714,0.652440,2.197129,
        -2.722714,0.660897,-2.278593,
        -2.722714,0.664008,1.360091,
        -2.722714,0.670850,-1.945257,
        -2.722714,0.671396,1.004859,
        -2.722714,0.673985,0.203855,
        -2.722714,0.680326,-0.133370,
        -2.722714,0.685083,-0.942363,
        -2.722714,0.688204,-2.632204,
        -2.722714,0.689777,2.385011,
        -2.722714,0.700065,3.115430,
        -2.722714,0.704080,2.911636,
        -2.722714,0.712883,1.174216,
        -2.722714,0.715701,0.699906,
        -2.722714,0.718851,-1.442215,
        -2.722714,0.721019,2.721536,
        -2.722714,0.725300,-2.975134,
        -2.722714,0.727136,-1.633288,
        -2.722714,0.727399,-2.120567,
        -2.722714,0.730751,0.522488,
        -2.722714,0.732892,-1.254330,
        -2.722714,0.734333,-2.446184,
        -2.722714,0.735125,0.037270,
        -2.722714,0.737893,1.866463,
        -2.722714,0.740650,1.681075,
        -2.722714,0.745971,-0.781286,
        -2.722714,0.749432,-0.442559,
        -2.722714,0.750102,2.050636,
        -2.722714,0.751046,0.863821,
        -2.722714,0.757018,-0.608833,
        -2.722714,0.757046,-1.816019,
        -2.722714,0.757389,1.501184,
        -2.722714,0.759800,2.546181,
        -2.722714,0.759859,0.351923,
        -2.722714,0.761141,-2.781102,
        -2.722714,0.769035,-1.078574,
        -2.722714,0.770166,-0.266604,
        -2.722714,0.772218,2.227380,
        -2.722714,0.788382,1.331051,
        -2.722714,0.790269,-2.282409,
        -2.722714,0.803410,-1.984567,
        -2.722714,0.803813,0.191520,
        -2.722714,0.806832,3.022155,
        -2.722714,0.807476,-2.597675,
        -2.722714,0.812512,-0.107335,
        -2.722714,0.814161,1.010949,
        -2.722714,0.817732,-3.084486,
        -2.722714,0.817846,2.390688,
        -2.722714,0.819323,2.845960,
        -2.722714,0.819430,-0.915017,
        -2.722714,0.829198,-1.532491,
        -2.722714,0.829531,0.665695,
        -2.722714,0.829816,1.167147,
        -2.722714,0.833613,-1.357492,
        -2.722714,0.842945,-2.907318,
        -2.722714,0.844883,2.679708,
        -2.722714,0.848189,-1.705003,
        -2.722714,0.849888,1.767901,
        -2.722714,0.851126,0.493370,
        -2.722714,0.852911,1.933394,
        -2.722714,0.857516,-2.432863,
        -2.722714,0.858261,-1.191729,
        -2.722714,0.858279,-2.138895,
        -2.722714,0.859723,-0.734928,
        -2.722714,0.859754,1.603292,
        -2.722714,0.861655,-0.394821,
        -2.722714,0.864300,0.043549,
        -2.722714,0.866940,0.851229,
        -2.722714,0.869748,2.095969,
        -2.722714,0.877352,-0.574668,
        -2.722714,0.879254,-2.733338,
        -2.722714,0.882554,1.443757,
        -2.722714,0.884060,0.334199,
        -2.722714,0.885626,-1.866061,
        -2.722714,0.889834,2.529957,
        -2.722714,0.893140,2.251698,
        -2.722714,0.897463,-0.240822,
        -2.722714,0.898758,-1.031948,
        -2.722714,0.913918,1.287618,
        -2.722714,0.916555,-2.287173,
        -2.722714,0.918305,3.107845,
        -2.722714,0.919522,2.948277,
        -2.722714,0.927138,-2.569686,
        -2.722714,0.929031,0.185383,
        -2.722714,0.933766,-3.015917,
        -2.722714,0.934873,-2.010822,
        -2.722714,0.935636,-1.445855,
        -2.722714,0.936702,-0.866461,
        -2.722714,0.937395,0.729108,
        -2.722714,0.937910,2.787575,
        -2.722714,0.939144,-0.092962,
        -2.722714,0.941718,1.125085,
        -2.722714,0.942873,-1.607680,
        -2.722714,0.943701,0.980328,
        -2.722714,0.947774,2.395256,
        -2.722714,0.952433,0.584347,
        -2.722714,0.954456,-1.287293,
        -2.722714,0.958731,-2.852552,
        -2.722714,0.961838,1.839284,
        -2.722714,0.961898,-0.469425,
        -2.722714,0.965426,1.686615,
        -2.722714,0.970108,1.990771,
        -2.722714,0.972820,-1.762780,
        -2.722714,0.973500,-0.711881,
        -2.722714,0.975253,2.641082,
        -2.722714,0.977824,0.441928,
        -2.722714,0.981485,-2.424766,
        -2.722714,0.981877,1.537100,
        -2.722714,0.982609,-1.136920,
        -2.722714,0.984265,-2.156073,
        -2.722714,0.989150,0.045493,
        -2.722714,0.993052,2.137478,
        -2.722714,0.997766,-2.693871,
        -2.722714,0.999847,-0.334738,
        -2.722714,1.001920,0.848054,
        -2.722714,1.006146,1.389734,
        -2.722714,1.012180,0.304066,
        -2.722714,1.016938,2.270674,
        -2.722714,1.017371,-0.979790,
        -2.722714,1.019495,-1.894487,
        -2.722714,1.023796,-0.583691,
        -2.722714,1.026525,2.506945,
        -2.722714,1.027004,-0.203653,
        -2.722714,1.028027,3.036573,
        -2.722714,1.031435,1.241021,
        -2.722714,1.032316,2.884221,
        -2.722714,1.035863,-3.110475,
        -2.722714,1.042085,-2.292950,
        -2.722714,1.043272,-1.522986,
        -2.722714,1.046265,-2.958847,
        -2.722714,1.049272,-1.376344,
        -2.722714,1.050039,-2.552086,
        -2.722714,1.053682,0.689863,
        -2.722714,1.055458,0.170394,
        -2.722714,1.056622,-0.831866,
        -2.722714,1.057515,1.097998,
        -2.722714,1.058706,-2.036438,
        -2.722714,1.061389,-1.672248,
        -2.722714,1.063758,2.743093,
        -2.722714,1.070157,-0.073422,
        -2.722714,1.072154,1.761506,
        -2.722714,1.074029,1.909193,
        -2.722714,1.074204,0.545276,
        -2.722714,1.074714,-2.805147,
        -2.722714,1.076262,0.966104,
        -2.722714,1.077642,-1.227073,
        -2.722714,1.084151,1.618525,
        -2.722714,1.085085,2.377673,
        -2.722714,1.086822,-0.441112,
        -2.722714,1.087142,2.058815,
        -2.722714,1.098383,-0.696430,
        -2.722714,1.099138,-1.084033,
        -2.722714,1.103208,1.477773,
        -2.722714,1.103783,0.408254,
        -2.722714,1.106275,-2.420800,
        -2.722714,1.108079,2.612943,
        -2.722714,1.109345,-2.172718,
        -2.722714,1.115957,-1.787597,
        -2.722714,1.118930,0.045983,
        -2.722714,1.119777,0.834702,
        -2.722714,1.121943,-2.669042,
        -2.722714,1.124930,2.219917,
        -2.722714,1.125217,1.336294,
        -2.722714,1.130958,2.972931,
        -2.722714,1.131400,-0.306629,
        -2.722714,1.137135,-0.936514,
        -2.722714,1.139125,-1.924528,
        -2.722714,1.139877,0.276382,
        -2.722714,1.142003,-3.058022,
        -2.722714,1.147363,-0.558448,
        -2.722714,1.150348,-0.176418,
        -2.722714,1.151198,1.197016,
        -2.722714,1.153930,-1.597104,
        -2.722714,1.155389,2.837148,
        -2.722714,1.155458,-1.460623,
        -2.722714,1.157331,-2.906362,
        -2.722714,1.158594,2.487618,
        -2.722714,1.162103,-1.322553,
        -2.722714,1.165407,3.092861,
        -2.722714,1.168025,0.717075,
        -2.722714,1.168406,-2.301519,
        -2.722714,1.174782,-2.538755,
        -2.722714,1.177998,1.061256,
        -2.722714,1.178898,1.827305,
        -2.722714,1.181331,-2.060330,
        -2.722714,1.182540,-0.798952,
        -2.722714,1.184569,0.147517,
        -2.722714,1.184734,1.968919,
        -2.722714,1.184804,0.593386,
        -2.722714,1.188278,1.690808,
        -2.722714,1.194360,2.110440,
        -2.722714,1.194776,2.710455,
        -2.722714,1.197745,-1.173705,
        -2.722714,1.199898,-2.777257,
        -2.722714,1.201164,2.355931,
        -2.722714,1.203760,1.556404,
        -2.722714,1.205845,-0.415993,
        -2.722714,1.206585,0.933648,
        -2.722714,1.206950,-0.037248,
        -2.722714,1.209341,0.471162,
        -2.722714,1.213483,-1.039417,
        -2.722714,1.218146,-0.672643,
        -2.722714,1.221608,1.421669,
        -2.722714,1.226043,-1.693533,
        -2.722714,1.232870,-1.822773,
        -2.722714,1.234375,-2.421194,
        -2.722714,1.234531,-2.190148,
        -2.722714,1.239414,0.348959,
        -2.722714,1.240183,2.587255,
        -2.722714,1.242312,2.941234,
        -2.722714,1.244075,1.286644,
        -2.722714,1.245961,-2.997461,
        -2.722714,1.246495,-2.646872,
        -2.722714,1.247736,2.242810,
        -2.722714,1.253604,0.812222,
        -2.722714,1.257816,-0.894317,
        -2.722714,1.258247,-1.953634,
        -2.722714,1.258993,-0.284099,
        -2.722714,1.262380,-0.157609,
        -2.722714,1.262791,-1.549229,
        -2.722714,1.263772,-0.533073,
        -2.722714,1.266731,-1.412461,
        -2.722714,1.268591,-3.122146,
        -2.722714,1.269614,-1.275246,
        -2.722714,1.270745,1.152533,
        -2.722714,1.273606,0.224545,
        -2.722714,1.279716,2.812101,
        -2.722714,1.283786,2.462124,
        -2.722714,1.284798,0.078209,
        -2.722714,1.285402,-2.882049,
        -2.722714,1.286936,1.886053,
        -2.722714,1.289408,0.683122,
        -2.722714,1.294736,1.756184,
        -2.722714,1.296978,-2.311511,
        -2.722714,1.298546,2.020601,
        -2.722714,1.300924,1.022478,
        -2.722714,1.303305,3.043152,
        -2.722714,1.304210,-2.084225,
        -2.722714,1.305241,-2.530249,
        -2.722714,1.307513,1.628003,
        -2.722714,1.310722,0.558267,
        -2.722714,1.314142,2.139329,
        -2.722714,1.316087,-0.777971,
        -2.722714,1.317907,-1.121023,
        -2.722714,1.321019,-2.749776,
        -2.722714,1.321609,2.687717,
        -2.722714,1.322568,1.499415,
        -2.722714,1.322972,-0.395699,
        -2.722714,1.325286,-0.994192,
        -2.722714,1.329146,-0.648513,
        -2.722714,1.329795,-0.042570,
        -2.722714,1.331651,2.344497,
        -2.722714,1.336522,0.435098,
        -2.722714,1.339410,0.900087,
        -2.722714,1.339550,1.370353,
        -2.722714,1.340441,-1.731580,
        -2.722714,1.351855,-1.855234,
        -2.722714,1.360161,-3.018404,
        -2.722714,1.361031,-1.613971,
        -2.722714,1.361196,2.914780,
        -2.722714,1.361998,-2.207254,
        -2.722714,1.362302,1.239384,
        -2.722714,1.365228,2.563808,
        -2.722714,1.365452,0.310522,
        -2.722714,1.367028,-2.420041,
        -2.722714,1.373422,-1.484281,
        -2.722714,1.374245,-1.226417,
        -2.722714,1.375020,-0.170393,
        -2.722714,1.376560,-1.355481,
        -2.722714,1.377070,-1.980984,
        -2.722714,1.377578,-2.636339,
        -2.722714,1.378096,0.774626,
        -2.722714,1.379086,-0.508192,
        -2.722714,1.382562,0.177141,
        -2.722714,1.388849,2.233847,
        -2.722714,1.389268,-0.886349,
        -2.722714,1.391324,1.109396,
        -2.722714,1.391699,-0.288725,
        -2.722714,1.392869,3.136903,
        -2.722714,1.393612,1.952016,
        -2.722714,1.395545,-2.850807,
        -2.722714,1.401926,1.822930,
        -2.722714,1.402149,2.788898,
        -2.722714,1.409552,0.647980,
        -2.722714,1.410840,0.056219,
        -2.722714,1.411944,2.444377,
        -2.722714,1.413639,1.697982,
        -2.722714,1.426479,1.573427,
        -2.722714,1.426725,2.102242,
        -2.722714,1.427868,0.984637,
        -2.722714,1.429580,-2.107232,
        -2.722714,1.431375,-2.313269,
        -2.722714,1.431907,3.015410,
        -2.722714,1.432325,-1.029054,
        -2.722714,1.433341,0.524268,
        -2.722714,1.435181,-0.739771,
        -2.722714,1.437208,-2.529085,
        -2.722714,1.439041,-0.617966,
        -2.722714,1.441473,1.448544,
        -2.722714,1.444527,2.664744,
        -2.722714,1.450513,-0.400625,
        -2.722714,1.451907,-0.066823,
        -2.722714,1.452754,-2.738631,
        -2.722714,1.454061,-2.950416,
        -2.722714,1.456736,1.323154,
        -2.722714,1.458370,0.400433,
        -2.722714,1.459314,-1.765003,
        -2.722714,1.459902,-1.145835,
        -2.722714,1.464952,2.329245,
        -2.722714,1.466050,0.862104,
        -2.722714,1.471330,-1.886160,
        -2.722714,1.475119,-1.648255,
        -2.722714,1.480878,1.194576,
        -2.722714,1.480942,0.273235,
        -2.722714,1.481086,2.890269,
        -2.722714,1.484118,-1.529156,
        -2.722714,1.484741,-1.279219,
        -2.722714,1.487681,-1.406069,
        -2.722714,1.487886,-3.066037,
        -2.722714,1.490555,2.543284,
        -2.722714,1.492424,-0.190932,
        -2.722714,1.495002,-0.839810,
        -2.722714,1.495297,2.007188,
        -2.722714,1.496414,-2.009267,
        -2.722714,1.498261,-2.425270,
        -2.722714,1.498907,-2.207327,
        -2.722714,1.500204,0.737638,
        -2.722714,1.502469,0.151156,
        -2.722714,1.507448,-0.514782,
        -2.722714,1.509287,2.204647,
        -2.722714,1.510518,1.884920,
        -2.722714,1.511948,-2.632360,
        -2.722714,1.515235,1.068565,
        -2.722714,1.522237,3.102018,
        -2.722714,1.522373,1.764991,
        -2.722714,1.522416,-0.954956,
        -2.722714,1.523102,2.765920,
        -2.722714,1.524811,-2.844718,
        -2.722714,1.527261,-0.304916,
        -2.722714,1.529788,0.613208,
        -2.722714,1.533648,1.644808,
        -2.722714,1.535018,0.028869,
        -2.722714,1.541082,2.424608,
        -2.722714,1.546010,1.523462,
        -2.722714,1.551135,-0.642994,
        -2.722714,1.553895,0.489607,
        -2.722714,1.555041,0.947813,
        -2.722714,1.559671,-1.074934,
        -2.722714,1.560345,2.985899,
        -2.722714,1.560487,-2.323400,
        -2.722714,1.561127,1.401111,
        -2.722714,1.567193,2.643172,
        -2.722714,1.568892,2.098362,
        -2.722714,1.571702,-2.098435,
        -2.722714,1.572701,-2.528730,
        -2.722714,1.573572,1.278593,
        -2.722714,1.574849,-0.091505,
        -2.722714,1.576334,0.365236,
        -2.722714,1.577029,-2.962888,
        -2.722714,1.577688,-1.814941,
        -2.722714,1.579542,-0.755917,
        -2.722714,1.579624,-1.199759,
        -2.722714,1.581515,-0.419414,
        -2.722714,1.584581,-2.738702,
        -2.722714,1.586948,-1.696296,
        -2.722714,1.588901,2.301496,
        -2.722714,1.589936,0.825020,
        -2.722714,1.590018,-1.941652,
        -2.722714,1.596968,-1.328358,
        -2.722714,1.597178,-1.575607,
        -2.722714,1.600470,-1.453196,
        -2.722714,1.601408,0.242409,
        -2.722714,1.602259,1.152656,
        -2.722714,1.603064,2.865774,
        -2.722714,1.612085,1.968802,
        -2.722714,1.612720,-3.080498,
        -2.722714,1.616061,2.521531,
        -2.722714,1.616900,-0.207457,
        -2.722714,1.620988,-0.875920,
        -2.722714,1.622114,0.701509,
        -2.722714,1.622749,0.121426,
        -2.722714,1.622993,-2.221629,
        -2.722714,1.629346,-0.542669,
        -2.722714,1.630007,1.844552,
        -2.722714,1.634747,-2.426198,
        -2.722714,1.642346,1.722617,
        -2.722714,1.643925,1.034629,
        -2.722714,1.644848,2.743774,
        -2.722714,1.645258,2.188109,
        -2.722714,1.645417,-2.634725,
        -2.722714,1.647395,3.085437,
        -2.722714,1.648538,-2.855044,
        -2.722714,1.649932,0.576808,
        -2.722714,1.652224,-0.996105,
        -2.722714,1.652871,1.600681,
        -2.722714,1.657997,-0.325127,
        -2.722714,1.661936,0.001037,
        -2.722714,1.665187,1.477402,
        -2.722714,1.665496,2.397872,
        -2.722714,1.673267,-0.665453,
        -2.722714,1.673557,0.452526,
        -2.722714,1.679277,-1.122421,
        -2.722714,1.679696,0.912482,
        -2.722714,1.682420,1.353922,
        -2.722714,1.685031,-2.009038,
        -2.722714,1.685196,2.967046,
        -2.722714,1.690107,2.620910,
        -2.722714,1.690453,-2.123076,
        -2.722714,1.691693,-1.871572,
        -2.722714,1.691817,2.061688,
        -2.722714,1.692644,1.237670,
        -2.722714,1.696386,0.328123,
        -2.722714,1.697162,-1.249026,
        -2.722714,1.697488,-2.322785,
        -2.722714,1.698726,-2.973576,
        -2.722714,1.699172,-1.749717,
        -2.722714,1.704478,-0.106295,
        -2.722714,1.707244,-0.444770,
        -2.722714,1.707500,-2.531497,
        -2.722714,1.707714,-0.787321,
        -2.722714,1.708430,-1.627552,
        -2.722714,1.710583,-1.376890,
        -2.722714,1.712415,-2.747921,
        -2.722714,1.713125,-1.503073,
        -2.722714,1.713452,0.789419,
        -2.722714,1.720612,2.279888,
        -2.722714,1.724852,0.202831,
        -2.722714,1.724868,2.846246,
        -2.722714,1.728940,1.931535,
        -2.722714,1.732838,1.127584,
        -2.722714,1.735719,-3.092485,
        -2.722714,1.739676,2.496017,
        -2.722714,1.739971,-0.228876,
        -2.722714,1.743328,0.663347,
        -2.722714,1.744927,-0.911892,
        -2.722714,1.745952,0.084733,
        -2.722714,1.748077,1.804354,
        -2.722714,1.752959,-0.567548,
        -2.722714,1.760169,1.678132,
        -2.722714,1.763887,-2.218450,
        -2.722714,1.766737,2.722752,
        -2.722714,1.767999,1.004737,
        -2.722714,1.769149,1.550298,
        -2.722714,1.770119,-2.427175,
        -2.722714,1.770460,2.154208,
        -2.722714,1.770514,3.070982,
        -2.722714,1.770735,0.537245,
        -2.722714,1.772673,-2.864760,
        -2.722714,1.773527,-1.039582,
        -2.722714,1.776419,-2.642421,
        -2.722714,1.780769,1.416330,
        -2.722714,1.786634,-0.346247,
        -2.722714,1.788829,-1.947553,
        -2.722714,1.793737,0.410270,
        -2.722714,1.794553,2.374618,
        -2.722714,1.794699,-0.694061,
        -2.722714,1.797469,-1.169105,
        -2.722714,1.803705,0.881076,
        -2.722714,1.804124,1.264249,
        -2.722714,1.807825,-2.090455,
        -2.722714,1.808822,-1.817375,
        -2.722714,1.809020,2.949741,
        -2.722714,1.812466,2.597595,
        -2.722714,1.813159,2.025395,
        -2.722714,1.813816,-1.298435,
        -2.722714,1.815016,0.278951,
        -2.722714,1.816510,-0.008631,
        -2.722714,1.817772,-1.689625,
        -2.722714,1.821360,-2.984352,
        -2.722714,1.822312,-0.127211,
        -2.722714,1.823386,-1.429807,
        -2.722714,1.823834,-1.560483,
        -2.722714,1.832060,-0.821627,
        -2.722714,1.832690,-0.468440,
        -2.722714,1.833945,-2.320278,
        -2.722714,1.834787,0.751968,
        -2.722714,1.840579,-2.536445,
        -2.722714,1.841033,-2.756932,
        -2.722714,1.846350,1.893983,
        -2.722714,1.846644,1.113766,
        -2.722714,1.846833,2.826543,
        -2.722714,1.847325,2.249051,
        -2.722714,1.854119,0.134161,
        -2.722714,1.858863,-3.106022,
        -2.722714,1.865046,1.763384,
        -2.722714,1.865619,-0.952036,
        -2.722714,1.865912,0.624453,
        -2.722714,1.866694,2.473686,
        -2.722714,1.867524,-0.240083,
        -2.722714,1.875561,1.632831,
        -2.722714,1.876049,-0.594378,
        -2.722714,1.880509,1.500204,
        -2.722714,1.883504,1.360457,
        -2.722714,1.886971,-2.196665,
        -2.722714,1.888979,0.982922,
        -2.722714,1.889815,2.701570,
        -2.722714,1.892144,-2.005340,
        -2.722714,1.892348,0.496522,
        -2.722714,1.893281,-1.083976,
        -2.722714,1.893828,3.053453,
        -2.722714,1.895117,2.121107,
        -2.722714,1.898741,-2.875735,
        -2.722714,1.904429,-2.428029,
        -2.722714,1.908408,-2.648903,
        -2.722714,1.911828,-0.364489,
        -2.722714,1.912224,0.366565,
        -2.722714,1.914854,1.214468,
        -2.722714,1.915467,-1.883143,
        -2.722714,1.915763,-0.724679,
        -2.722714,1.915790,-1.216366,
        -2.722714,1.922256,2.347806,
        -2.722714,1.922616,0.844633,
        -2.722714,1.925185,0.231170,
        -2.722714,1.927474,-1.754616,
        -2.722714,1.928342,-1.351174,
        -2.722714,1.933328,-1.623328,
        -2.722714,1.933878,0.034150,
        -2.722714,1.934585,-1.487979,
        -2.722714,1.934802,2.927634,
        -2.722714,1.934862,1.989776,
        -2.722714,1.936327,-0.099350,
        -2.722714,1.938088,2.578507,
        -2.722714,1.945677,-2.999805,
        -2.722714,1.954572,-0.859208,
        -2.722714,1.957868,-0.490784,
        -2.722714,1.958989,0.712772,
        -2.722714,1.960212,-2.307561,
        -2.722714,1.964774,1.855507,
        -2.722714,1.964971,-2.100533,
        -2.722714,1.969112,1.088273,
        -2.722714,1.971438,-2.765415,
        -2.722714,1.972312,2.805134,
        -2.722714,1.974950,2.220596,
        -2.722714,1.975135,-2.539127,
        -2.722714,1.982107,1.720829,
        -2.722714,1.983195,-3.127984,
        -2.722714,1.983318,-0.247182,
        -2.722714,1.986720,-0.993220,
        -2.722714,1.990381,1.586005,
        -2.722714,1.990470,0.580283,
        -2.722714,1.990523,1.450326,
        -2.722714,1.993909,1.312897,
        -2.722714,1.994694,2.452813,
        -2.722714,1.999226,-0.620717,
        -2.722714,2.005145,0.941012,
        -2.722714,2.007301,0.142337,
        -2.722714,2.014149,0.444138,
        -2.722714,2.014338,-1.127392,
        -2.722714,2.018131,3.021778,
        -2.722714,2.019676,-1.980944,
        -2.722714,2.019912,2.686612,
        -2.722714,2.020545,2.088672,
        -2.722714,2.026074,0.300945,
        -2.722714,2.026196,-2.889057,
        -2.722714,2.031959,-1.268178,
        -2.722714,2.032917,-2.199917,
        -2.722714,2.033572,-1.842884,
        -2.722714,2.034712,-2.419505,
        -2.722714,2.035818,-0.379479,
        -2.722714,2.036004,-0.757062,
        -2.722714,2.040501,-1.411488,
        -2.722714,2.042874,-1.701367,
        -2.722714,2.043060,-2.653746,
        -2.722714,2.043729,-1.557938,
        -2.722714,2.048538,0.806623,
        -2.722714,2.049006,1.199930,
        -2.722714,2.049502,-0.137989,
        -2.722714,2.052181,2.326959,
        -2.722714,2.052794,-0.008526,
        -2.722714,2.057608,1.951726,
        -2.722714,2.065568,2.567730,
        -2.722714,2.070371,2.879808,
        -2.722714,2.071256,-3.018961,
        -2.722714,2.079746,-0.897699,
        -2.722714,2.082784,1.042403,
        -2.722714,2.083752,-0.511086,
        -2.722714,2.084401,1.810502,
        -2.722714,2.085731,0.671081,
        -2.722714,2.097496,1.373172,
        -2.722714,2.099185,1.667006,
        -2.722714,2.100482,-2.301084,
        -2.722714,2.103118,2.194745,
        -2.722714,2.103223,1.521685,
        -2.722714,2.103695,-2.777707,
        -2.722714,2.107523,3.129581,
        -2.722714,2.108277,-2.095907,
        -2.722714,2.108891,-2.532166,
        -2.722714,2.111003,-0.268632,
        -2.722714,2.112276,-1.027246,
        -2.722714,2.114380,0.531457,
        -2.722714,2.119166,0.100552,
        -2.722714,2.119564,2.754818,
        -2.722714,2.122474,-0.648314,
        -2.722714,2.125183,2.445598,
        -2.722714,2.129501,0.238290,
        -2.722714,2.131178,0.386663,
        -2.722714,2.131223,-1.174311,
        -2.722714,2.134269,-1.948042,
        -2.722714,2.134684,0.907065,
        -2.722714,2.137615,2.988806,
        -2.722714,2.145635,-1.327708,
        -2.722714,2.146467,2.055081,
        -2.722714,2.146582,-1.793788,
        -2.722714,2.149210,-1.482321,
        -2.722714,2.152901,-1.636788,
        -2.722714,2.156337,1.151558,
        -2.722714,2.156474,-2.912065,
        -2.722714,2.157194,-0.792563,
        -2.722714,2.166724,-0.401638,
        -2.722714,2.166980,-0.134879,
        -2.722714,2.173644,-2.403429,
        -2.722714,2.175711,-2.655523,
        -2.722714,2.177132,2.600419,
        -2.722714,2.178879,0.767722,
        -2.722714,2.180752,1.908485,
        -2.722714,2.181338,2.313721,
        -2.722714,2.185580,1.287673,
        -2.722714,2.195400,-2.217546,
        -2.722714,2.200418,-3.056442,
        -2.722714,2.201829,-0.001735,
        -2.722714,2.203804,1.756176,
        -2.722714,2.211602,1.443930,
        -2.722714,2.212686,-0.544809,
        -2.722714,2.213937,0.621892,
        -2.722714,2.214075,1.600691,
        -2.722714,2.216256,1.020029,
        -2.722714,2.219060,2.886631,
        -2.722714,2.223375,-0.906185,
        -2.722714,2.226002,-1.060248,
        -2.722714,2.226413,-2.064034,
        -2.722714,2.230554,2.171334,
        -2.722714,2.232413,2.735777,
        -2.722714,2.233709,3.070830,
        -2.722714,2.234383,-2.792065,
        -2.722714,2.237183,0.468799,
        -2.722714,2.237695,-0.271388,
        -2.722714,2.239419,0.151693,
        -2.722714,2.243347,-2.521049,
        -2.722714,2.244035,-1.229456,
        -2.722714,2.244355,0.309899,
        -2.722714,2.245175,2.460236,
        -2.722714,2.245628,-1.899216,
        -2.722714,2.245973,-0.702234,
        -2.722714,2.255601,-1.728131,
        -2.722714,2.256702,-1.401367,
        -2.722714,2.261466,-1.557500,
        -2.722714,2.267390,0.878499,
        -2.722714,2.272099,2.018594,
        -2.722714,2.278618,1.170386,
        -2.722714,2.283304,-2.345584,
        -2.722714,2.285175,-2.943090,
        -2.722714,2.293000,-0.124451,
        -2.722714,2.294058,-0.421070,
        -2.722714,2.303283,1.855938,
        -2.722714,2.304968,2.608978,
        -2.722714,2.305347,2.311346,
        -2.722714,2.307188,-2.656134,
        -2.722714,2.307301,1.336200,
        -2.722714,2.309655,0.726020,
        -2.722714,2.317634,-2.185225,
        -2.722714,2.321510,1.685664,
        -2.722714,2.324376,-3.111006,
        -2.722714,2.324731,1.511186,
        -2.722714,2.329587,0.035498,
        -2.722714,2.330158,2.977264,
        -2.722714,2.335096,-0.588468,
        -2.722714,2.336046,-1.115799,
        -2.722714,2.338060,2.803613,
        -2.722714,2.338083,-0.943400,
        -2.722714,2.340210,0.562145,
        -2.722714,2.341931,1.023926,
        -2.722714,2.343265,-0.779061,
        -2.722714,2.343434,-2.013692,
        -2.722714,2.353029,-1.310529,
        -2.722714,2.354868,0.211543,
        -2.722714,2.355120,0.388593,
        -2.722714,2.355165,-1.833173,
        -2.722714,2.355462,2.151441,
        -2.722714,2.362948,-2.487477,
        -2.722714,2.363101,-2.810456,
        -2.722714,2.363394,-0.272918,
        -2.722714,2.364509,-1.644533,
        -2.722714,2.370465,2.466824,
        -2.722714,2.386110,-1.481351,
        -2.722714,2.394960,0.862133,
        -2.722714,2.395716,1.977184,
        -2.722714,2.396339,1.200900,
        -2.722714,2.407941,-2.316753,
        -2.722714,2.409450,-2.988432,
        -2.722714,2.417202,-0.103363,
        -2.722714,2.417386,-0.443097,
        -2.722714,2.417950,2.666662,
        -2.722714,2.423190,1.789783,
        -2.722714,2.423692,1.392696,
        -2.722714,2.431122,3.086379,
        -2.722714,2.431406,-2.647820,
        -2.722714,2.431584,2.307631,
        -2.722714,2.434090,1.592818,
        -2.722714,2.435993,0.683844,
        -2.722714,2.440898,-2.133300,
        -2.722714,2.442222,-1.195837,
        -2.722714,2.443112,2.883761,
        -2.722714,2.449191,-0.633651,
        -2.722714,2.450821,0.084741,
        -2.722714,2.453499,-1.008861,
        -2.722714,2.460010,-1.938011,
        -2.722714,2.460199,-0.822008,
        -2.722714,2.460902,0.490079,
        -2.722714,2.461774,-1.734913,
        -2.722714,2.463986,1.034407,
        -2.722714,2.467647,0.287667,
        -2.722714,2.479849,2.126925,
        -2.722714,2.485907,-2.835471,
        -2.722714,2.485945,-1.374853,
        -2.722714,2.487747,2.505912,
        -2.722714,2.487957,-0.274279,
        -2.722714,2.489741,-2.468428,
        -2.722714,2.512538,1.239900,
        -2.722714,2.517528,1.923289,
        -2.722714,2.518454,-1.565852,
        -2.722714,2.519348,0.843503,
        -2.722714,2.522766,-3.059231,
        -2.722714,2.528686,2.737228,
        -2.722714,2.534798,-2.266671,
        -2.722714,2.536533,1.466274,
        -2.722714,2.538504,1.699812,
        -2.722714,2.539516,-0.476942,
        -2.722714,2.539532,-0.074548,
        -2.722714,2.543197,2.990252,
        -2.722714,2.554022,2.320619,
        -2.722714,2.556919,-2.655005,
        -2.722714,2.558146,0.627191,
        -2.722714,2.558223,-1.164189,
        -2.722714,2.563452,-2.044028,
        -2.722714,2.568462,0.150655,
        -2.722714,2.569130,-0.700837,
        -2.722714,2.570050,-1.804492,
        -2.722714,2.572710,-0.934665,
        -2.722714,2.575994,0.391718,
        -2.722714,2.584774,1.046703,
        -2.722714,2.601432,2.102535,
        -2.722714,2.603559,2.555114,
        -2.722714,2.604150,-2.885208,
        -2.722714,2.608630,-1.378011,
        -2.722714,2.611256,-0.279110,
        -2.722714,2.617344,-2.439480,
        -2.722714,2.627402,1.293552,
        -2.722714,2.635396,1.848850,
        -2.722714,2.635686,2.834950,
        -2.722714,2.637087,3.132730,
        -2.722714,2.642029,0.813274,
        -2.722714,2.645260,1.570203,
        -2.722714,2.646182,-1.614036,
        -2.722714,2.660113,-2.186513,
        -2.722714,2.660483,-0.034608,
        -2.722714,2.661713,-0.529511,
        -2.722714,2.672720,-1.110422,
        -2.722714,2.676626,2.335353,
        -2.722714,2.676638,0.540364,
        -2.722714,2.676780,-1.901158,
        -2.722714,2.681411,-2.667965,
        -2.722714,2.682152,0.245751,
        -2.722714,2.682859,-0.817071,
        -2.722714,2.705107,1.062142,
        -2.722714,2.717705,2.625995,
        -2.722714,2.721001,2.060097,
        -2.722714,2.722319,-2.954069,
        -2.722714,2.729840,-1.368237,
        -2.722714,2.734349,-0.289298,
        -2.722714,2.738409,2.975671,
        -2.722714,2.740366,1.374341,
        -2.722714,2.744633,-2.391311,
        -2.722714,2.747398,1.727166,
        -2.722714,2.763799,0.760372,
        -2.722714,2.768238,-1.679243,
        -2.722714,2.779245,-2.046268,
        -2.722714,2.780208,-0.626453,
        -2.722714,2.780471,0.027732,
        -2.722714,2.784696,-1.013128,
        -2.722714,2.789817,0.395431,
        -2.722714,2.801154,2.353143,
        -2.722714,2.807091,-2.687477,
        -2.722714,2.825762,1.085035,
        -2.722714,2.830209,2.734943,
        -2.722714,2.833517,-3.085508,
        -2.722714,2.838301,1.980061,
        -2.722714,2.849871,1.512012,
        -2.722714,2.850089,-1.343899,
        -2.722714,2.856905,-0.318441,
        -2.722714,2.870445,-2.291001,
        -2.722714,2.885017,0.655562,
        -2.722714,2.886428,-1.793217,
        -2.722714,2.890337,-0.825633,
        -2.722714,2.898932,0.139973,
        -2.722714,2.931927,-2.728768,
        -2.722714,2.932096,2.377790,
        -2.722714,2.937409,2.947245,
        -2.722714,2.948048,1.133127,
        -2.722714,2.950896,1.797889,
        -2.722714,2.970019,-1.282382,
        -2.722714,2.978333,-0.412090,
        -2.722714,2.996437,-2.089121,
        -2.722714,3.011274,0.434110,
        -2.722714,3.050410,3.138492,
        -2.722714,3.064492,1.647259,
        -2.722714,3.088868,-0.989501,
        -2.617994,0.053884,-2.408394,
        -2.617994,0.077645,1.453433,
        -2.617994,0.088408,-0.311160,
        -2.617994,0.142314,2.590387,
        -2.617994,0.147206,-1.385164,
        -2.617994,0.163506,0.573203,
        -2.617994,0.165485,-2.875396,
        -2.617994,0.182564,-2.125821,
        -2.617994,0.190569,1.929141,
        -2.617994,0.201102,1.249082,
        -2.617994,0.201397,-0.716646,
        -2.617994,0.210568,-0.069559,
        -2.617994,0.251672,2.848900,
        -2.617994,0.266262,-1.678068,
        -2.617994,0.269108,-1.215563,
        -2.617994,0.273777,-2.506756,
        -2.617994,0.279854,0.814581,
        -2.617994,0.281664,-2.969147,
        -2.617994,0.283625,2.405603,
        -2.617994,0.284058,0.357350,
        -2.617994,0.294107,1.593880,
        -2.617994,0.302491,-0.400410,
        -2.617994,0.308059,2.021491,
        -2.617994,0.312219,-2.072057,
        -2.617994,0.321905,-0.829121,
        -2.617994,0.329984,1.201621,
        -2.617994,0.339714,-0.013712,
        -2.617994,0.368784,2.974974,
        -2.617994,0.376428,-1.495532,
        -2.617994,0.378103,2.646436,
        -2.617994,0.385088,0.590355,
        -2.617994,0.386248,-2.690943,
        -2.617994,0.389851,-2.340031,
        -2.617994,0.391502,-1.171374,
        -2.617994,0.398249,-1.812694,
        -2.617994,0.400542,-2.995729,
        -2.617994,0.404273,0.918163,
        -2.617994,0.405617,1.773533,
        -2.617994,0.409009,-0.570243,
        -2.617994,0.410700,0.270433,
        -2.617994,0.415893,1.464890,
        -2.617994,0.420680,2.088164,
        -2.617994,0.424766,-0.258729,
        -2.617994,0.426507,2.367872,
        -2.617994,0.441292,-0.890727,
        -2.617994,0.451839,-2.078193,
        -2.617994,0.459270,1.191562,
        -2.617994,0.470622,0.014002,
        -2.617994,0.481292,-2.529098,
        -2.617994,0.482460,2.849843,
        -2.617994,0.492205,-1.397963,
        -2.617994,0.493314,3.117264,
        -2.617994,0.497870,0.727828,
        -2.617994,0.499783,-1.654379,
        -2.617994,0.502651,0.463333,
        -2.617994,0.505463,2.594557,
        -2.617994,0.506104,-2.874910,
        -2.617994,0.513579,-1.147354,
        -2.617994,0.518569,1.644237,
        -2.617994,0.519001,1.900488,
        -2.617994,0.519277,-0.673809,
        -2.617994,0.524769,-0.421955,
        -2.617994,0.525564,-2.288838,
        -2.617994,0.532541,0.978737,
        -2.617994,0.534610,2.154808,
        -2.617994,0.535020,-1.891846,
        -2.617994,0.540208,1.400497,
        -2.617994,0.542006,0.228337,
        -2.617994,0.550762,-0.180771,
        -2.617994,0.560531,2.378474,
        -2.617994,0.561629,-0.924576,
        -2.617994,0.567465,-2.679392,
        -2.617994,0.588527,1.181755,
        -2.617994,0.591643,-2.098364,
        -2.617994,0.594963,2.997575,
        -2.617994,0.599643,-3.045235,
        -2.617994,0.599962,2.773556,
        -2.617994,0.602620,0.029270,
        -2.617994,0.605711,0.598454,
        -2.617994,0.607708,-1.541026,
        -2.617994,0.609360,-2.464379,
        -2.617994,0.610738,-1.326786,
        -2.617994,0.616968,0.820489,
        -2.617994,0.626379,1.773610,
        -2.617994,0.628626,-1.751180,
        -2.617994,0.630671,0.391949,
        -2.617994,0.632622,-0.737642,
        -2.617994,0.632708,2.566722,
        -2.617994,0.632797,1.989858,
        -2.617994,0.634455,-0.534790,
        -2.617994,0.636205,1.561170,
        -2.617994,0.637681,-1.123423,
        -2.617994,0.641441,-2.842109,
        -2.617994,0.644807,-0.327139,
        -2.617994,0.652440,2.197129,
        -2.617994,0.660897,-2.278593,
        -2.617994,0.664008,1.360091,
        -2.617994,0.670850,-1.945257,
        -2.617994,0.671396,1.004859,
        -2.617994,0.673985,0.203855,
        -2.617994,0.680326,-0.133370,
        -2.617994,0.685083,-0.942363,
        -2.617994,0.688204,-2.632204,
        -2.617994,0.689777,2.385011,
        -2.617994,0.700065,3.115430,
        -2.617994,0.704080,2.911636,
        -2.617994,0.712883,1.174216,
        -2.617994,0.715701,0.699906,
        -2.617994,0.718851,-1.442215,
        -2.617994,0.721019,2.721536,
        -2.617994,0.725300,-2.975134,
        -2.617994,0.727136,-1.633288,
        -2.617994,0.727399,-2.120567,
        -2.617994,0.730751,0.522488,
        -2.617994,0.732892,-1.254330,
        -2.617994,0.734333,-2.446184,
        -2.617994,0.735125,0.037270,
        -2.617994,0.737893,1.866463,
        -2.617994,0.740650,1.681075,
        -2.617994,0.745971,-0.781286,
        -2.617994,0.749432,-0.442559,
        -2.617994,0.750102,2.050636,
        -2.617994,0.751046,0.863821,
        -2.617994,0.757018,-0.608833,
        -2.617994,0.757046,-1.816019,
        -2.617994,0.757389,1.501184,
        -2.617994,0.759800,2.546181,
        -2.617994,0.759859,0.351923,
        -2.617994,0.761141,-2.781102,
        -2.617994,0.769035,-1.078574,
        -2.617994,0.770166,-0.266604,
        -2.617994,0.772218,2.227380,
        -2.617994,0.788382,1.331051,
        -2.617994,0.790269,-2.282409,
        -2.617994,0.803410,-1.984567,
        -2.617994,0.803813,0.191520,
        -2.617994,0.806832,3.022155,
        -2.617994,0.807476,-2.597675,
        -2.617994,0.812512,-0.107335,
        -2.617994,0.814161,1.010949,
        -2.617994,0.817732,-3.084486,
        -2.617994,0.817846,2.390688,
        -2.617994,0.819323,2.845960,
        -2.617994,0.819430,-0.915017,
        -2.617994,0.829198,-1.532491,
        -2.617994,0.829531,0.665695,
        -2.617994,0.829816,1.167147,
        -2.617994,0.833613,-1.357492,
        -2.617994,0.842945,-2.907318,
        -2.617994,0.844883,2.679708,
        -2.617994,0.848189,-1.705003,
        -2.617994,0.849888,1.767901,
        -2.617994,0.851126,0.493370,
        -2.617994,0.852911,1.933394,
        -2.617994,0.857516,-2.432863,
        -2.617994,0.858261,-1.191729,
        -2.617994,0.858279,-2.138895,
        -2.617994,0.859723,-0.734928,
        -2.617994,0.859754,1.603292,
        -2.617994,0.861655,-0.394821,
        -2.617994,0.864300,0.043549,
        -2.617994,0.866940,0.851229,
        -2.617994,0.869748,2.095969,
        -2.617994,0.877352,-0.574668,
        -2.617994,0.879254,-2.733338,
        -2.617994,0.882554,1.443757,
        -2.617994,0.884060,0.334199,
        -2.617994,0.885626,-1.866061,
        -2.617994,0.889834,2.529957,
        -2.617994,0.893140,2.251698,
        -2.617994,0.897463,-0.240822,
        -2.617994,0.898758,-1.031948,
        -2.617994,0.913918,1.287618,
        -2.617994,0.916555,-2.287173,
        -2.617994,0.918305,3.107845,
        -2.617994,0.919522,2.948277,
        -2.617994,0.927138,-2.569686,
        -2.617994,0.929031,0.185383,
        -2.617994,0.933766,-3.015917,
        -2.617994,0.934873,-2.010822,
        -2.617994,0.935636,-1.445855,
        -2.617994,0.936702,-0.866461,
        -2.617994,0.937395,0.729108,
        -2.617994,0.937910,2.787575,
        -2.617994,0.939144,-0.092962,
        -2.617994,0.941718,1.125085,
        -2.617994,0.942873,-1.607680,
        -2.617994,0.943701,0.980328,
        -2.617994,0.947774,2.395256,
        -2.617994,0.952433,0.584347,
        -2.617994,0.954456,-1.287293,
        -2.617994,0.958731,-2.852552,
        -2.617994,0.961838,1.839284,
        -2.617994,0.961898,-0.469425,
        -2.617994,0.965426,1.686615,
        -2.617994,0.970108,1.990771,
        -2.617994,0.972820,-1.762780,
        -2.617994,0.973500,-0.711881,
        -2.617994,0.975253,2.641082,
        -2.617994,0.977824,0.441928,
        -2.617994,0.981485,-2.424766,
        -2.617994,0.981877,1.537100,
        -2.617994,0.982609,-1.136920,
        -2.617994,0.984265,-2.156073,
        -2.617994,0.989150,0.045493,
        -2.617994,0.993052,2.137478,
        -2.617994,0.997766,-2.693871,
        -2.617994,0.999847,-0.334738,
        -2.617994,1.001920,0.848054,
        -2.617994,1.006146,1.389734,
        -2.617994,1.012180,0.304066,
        -2.617994,1.016938,2.270674,
        -2.617994,1.017371,-0.979790,
        -2.617994,1.019495,-1.894487,
        -2.617994,1.023796,-0.583691,
        -2.617994,1.026525,2.506945,
        -2.617994,1.027004,-0.203653,
        -2.617994,1.028027,3.036573,
        -2.617994,1.031435,1.241021,
        -2.617994,1.032316,2.884221,
        -2.617994,1.035863,-3.110475,
        -2.617994,1.042085,-2.292950,
        -2.617994,1.043272,-1.522986,
        -2.617994,1.046265,-2.958847,
        -2.617994,1.049272,-1.376344,
        -2.617994,1.050039,-2.552086,
        -2.617994,1.053682,0.689863,
        -2.617994,1.055458,0.170394,
        -2.617994,1.056622,-0.831866,
        -2.617994,1.057515,1.097998,
        -2.617994,1.058706,-2.036438,
        -2.617994,1.061389,-1.672248,
        -2.617994,1.063758,2.743093,
        -2.617994,1.070157,-0.073422,
        -2.617994,1.072154,1.761506,
        -2.617994,1.074029,1.909193,
        -2.617994,1.074204,0.545276,
        -2.617994,1.074714,-2.805147,
        -2.617994,1.076262,0.966104,
        -2.617994,1.077642,-1.227073,
        -2.617994,1.084151,1.618525,
        -2.617994,1.085085,2.377673,
        -2.617994,1.086822,-0.441112,
        -2.617994,1.087142,2.058815,
        -2.617994,1.098383,-0.696430,
        -2.617994,1.099138,-1.084033,
        -2.617994,1.103208,1.477773,
        -2.617994,1.103783,0.408254,
        -2.617994,1.106275,-2.420800,
        -2.617994,1.108079,2.612943,
        -2.617994,1.109345,-2.172718,
        -2.617994,1.115957,-1.787597,
        -2.617994,1.118930,0.045983,
        -2.617994,1.119777,0.834702,
        -2.617994,1.121943,-2.669042,
        -2.617994,1.124930,2.219917,
        -2.617994,1.125217,1.336294,
        -2.617994,1.130958,2.972931,
        -2.617994,1.131400,-0.306629,
        -2.617994,1.137135,-0.936514,
        -2.617994,1.139125,-1.924528,
        -2.617994,1.139877,0.276382,
        -2.617994,1.142003,-3.058022,
        -2.617994,1.147363,-0.558448,
        -2.617994,1.150348,-0.176418,
        -2.617994,1.151198,1.197016,
        -2.617994,1.153930,-1.597104,
        -2.617994,1.155389,2.837148,
        -2.617994,1.155458,-1.460623,
        -2.617994,1.157331,-2.906362,
        -2.617994,1.158594,2.487618,
        -2.617994,1.162103,-1.322553,
        -2.617994,1.165407,3.092861,
        -2.617994,1.168025,0.717075,
        -2.617994,1.168406,-2.301519,
        -2.617994,1.174782,-2.538755,
        -2.617994,1.177998,1.061256,
        -2.617994,1.178898,1.827305,
        -2.617994,1.181331,-2.060330,
        -2.617994,1.182540,-0.798952,
        -2.617994,1.184569,0.147517,
        -2.617994,1.184734,1.968919,
        -2.617994,1.184804,0.593386,
        -2.617994,1.188278,1.690808,
        -2.617994,1.194360,2.110440,
        -2.617994,1.194776,2.710455,
        -2.617994,1.197745,-1.173705,
        -2.617994,1.199898,-2.777257,
        -2.617994,1.201164,2.355931,
        -2.617994,1.203760,1.556404,
        -2.617994,1.205845,-0.415993,
        -2.617994,1.206585,0.933648,
        -2.617994,1.206950,-0.037248,
        -2.617994,1.209341,0.471162,
        -2.617994,1.213483,-1.039417,
        -2.617994,1.218146,-0.672643,
        -2.617994,1.221608,1.421669,
        -2.617994,1.226043,-1.693533,
        -2.617994,1.232870,-1.822773,
        -2.617994,1.234375,-2.421194,
        -2.617994,1.234531,-2.190148,
        -2.617994,1.239414,0.348959,
        -2.617994,1.240183,2.587255,
        -2.617994,1.242312,2.941234,
        -2.617994,1.244075,1.286644,
        -2.617994,1.245961,-2.997461,
        -2.617994,1.246495,-2.646872,
        -2.617994,1.247736,2.242810,
        -2.617994,1.253604,0.812222,
        -2.617994,1.257816,-0.894317,
        -2.617994,1.258247,-1.953634,
        -2.617994,1.258993,-0.284099,
        -2.617994,1.262380,-0.157609,
        -2.617994,1.262791,-1.549229,
        -2.617994,1.263772,-0.533073,
        -2.617994,1.266731,-1.412461,
        -2.617994,1.268591,-3.122146,
        -2.617994,1.269614,-1.275246,
        -2.617994,1.270745,1.152533,
        -2.617994,1.273606,0.224545,
        -2.617994,1.279716,2.812101,
        -2.617994,1.283786,2.462124,
        -2.617994,1.284798,0.078209,
        -2.617994,1.285402,-2.882049,
        -2.617994,1.286936,1.886053,
        -2.617994,1.289408,0.683122,
        -2.617994,1.294736,1.756184,
        -2.617994,1.296978,-2.311511,
        -2.617994,1.298546,2.020601,
        -2.617994,1.300924,1.022478,
        -2.617994,1.303305,3.043152,
        -2.617994,1.304210,-2.084225,
        -2.617994,1.305241,-2.530249,
        -2.617994,1.307513,1.628003,
        -2.617994,1.310722,0.558267,
        -2.617994,1.314142,2.139329,
        -2.617994,1.316087,-0.777971,
        -2.617994,1.317907,-1.121023,
        -2.617994,1.321019,-2.749776,
        -2.617994,1.321609,2.687717,
        -2.617994,1.322568,1.499415,
        -2.617994,1.322972,-0.395699,
        -2.617994,1.325286,-0.994192,
        -2.617994,1.329146,-0.648513,
        -2.617994,1.329795,-0.042570,
        -2.617994,1.331651,2.344497,
        -2.617994,1.336522,0.435098,
        -2.617994,1.339410,0.900087,
        -2.617994,1.339550,1.370353,
        -2.617994,1.340441,-1.731580,
        -2.617994,1.351855,-1.855234,
        -2.617994,1.360161,-3.018404,
        -2.617994,1.361031,-1.613971,
        -2.617994,1.361196,2.914780,
        -2.617994,1.361998,-2.207254,
        -2.617994,1.362302,1.239384,
        -2.617994,1.365228,2.563808,
        -2.617994,1.365452,0.310522,
        -2.617994,1.367028,-2.420041,
        -2.617994,1.373422,-1.484281,
        -2.617994,1.374245,-1.226417,
        -2.617994,1.375020,-0.170393,
        -2.617994,1.376560,-1.355481,
        -2.617994,1.377070,-1.980984,
        -2.617994,1.377578,-2.636339,
        -2.617994,1.378096,0.774626,
        -2.617994,1.379086,-0.508192,
        -2.617994,1.382562,0.177141,
        -2.617994,1.388849,2.233847,
        -2.617994,1.389268,-0.886349,
        -2.617994,1.391324,1.109396,
        -2.617994,1.391699,-0.288725,
        -2.617994,1.392869,3.136903,
        -2.617994,1.393612,1.952016,
        -2.617994,1.395545,-2.850807,
        -2.617994,1.401926,1.822930,
        -2.617994,1.402149,2.788898,
        -2.617994,1.409552,0.647980,
        -2.617994,1.410840,0.056219,
        -2.617994,1.411944,2.444377,
        -2.617994,1.413639,1.697982,
        -2.617994,1.426479,1.573427,
        -2.617994,1.426725,2.102242,
        -2.617994,1.427868,0.984637,
        -2.617994,1.429580,-2.107232,
        -2.617994,1.431375,-2.313269,
        -2.617994,1.431907,3.015410,
        -2.617994,1.432325,-1.029054,
        -2.617994,1.433341,0.524268,
        -2.617994,1.435181,-0.739771,
        -2.617994,1.437208,-2.529085,
        -2.617994,1.439041,-0.617966,
        -2.617994,1.441473,1.448544,
        -2.617994,1.444527,2.664744,
        -2.617994,1.450513,-0.400625,
        -2.617994,1.451907,-0.066823,
        -2.617994,1.452754,-2.738631,
        -2.617994,1.454061,-2.950416,
        -2.617994,1.456736,1.323154,
        -2.617994,1.458370,0.400433,
        -2.617994,1.459314,-1.765003,
        -2.617994,1.459902,-1.145835,
        -2.617994,1.464952,2.329245,
        -2.617994,1.466050,0.862104,
        -2.617994,1.471330,-1.886160,
        -2.617994,1.475119,-1.648255,
        -2.617994,1.480878,1.194576,
        -2.617994,1.480942,0.273235,
        -2.617994,1.481086,2.890269,
        -2.617994,1.484118,-1.529156,
        -2.617994,1.484741,-1.279219,
        -2.617994,1.487681,-1.406069,
        -2.617994,1.487886,-3.066037,
        -2.617994,1.490555,2.543284,
        -2.617994,1.492424,-0.190932,
        -2.617994,1.495002,-0.839810,
        -2.617994,1.495297,2.007188,
        -2.617994,1.496414,-2.009267,
        -2.617994,1.498261,-2.425270,
        -2.617994,1.498907,-2.207327,
        -2.617994,1.500204,0.737638,
        -2.617994,1.502469,0.151156,
        -2.617994,1.507448,-0.514782,
        -2.617994,1.509287,2.204647,
        -2.617994,1.510518,1.884920,
        -2.617994,1.511948,-2.632360,
        -2.617994,1.515235,1.068565,
        -2.617994,1.522237,3.102018,
        -2.617994,1.522373,1.764991,
        -2.617994,1.522416,-0.954956,
        -2.617994,1.523102,2.765920,
        -2.617994,1.524811,-2.844718,
        -2.617994,1.527261,-0.304916,
        -2.617994,1.529788,0.613208,
        -2.617994,1.533648,1.644808,
        -2.617994,1.535018,0.028869,
        -2.617994,1.541082,2.424608,
        -2.617994,1.546010,1.523462,
        -2.617994,1.551135,-0.642994,
        -2.617994,1.553895,0.489607,
        -2.617994,1.555041,0.947813,
        -2.617994,1.559671,-1.074934,
        -2.617994,1.560345,2.985899,
        -2.617994,1.560487,-2.323400,
        -2.617994,1.561127,1.401111,
        -2.617994,1.567193,2.643172,
        -2.617994,1.568892,2.098362,
        -2.617994,1.571702,-2.098435,
        -2.617994,1.572701,-2.528730,
        -2.617994,1.573572,1.278593,
        -2.617994,1.574849,-0.091505,
        -2.617994,1.576334,0.365236,
        -2.617994,1.577029,-2.962888,
        -2.617994,1.577688,-1.814941,
        -2.617994,1.579542,-0.755917,
        -2.617994,1.579624,-1.199759,
        -2.617994,1.581515,-0.419414,
        -2.617994,1.584581,-2.738702,
        -2.617994,1.586948,-1.696296,
        -2.617994,1.588901,2.301496,
        -2.617994,1.589936,0.825020,
        -2.617994,1.590018,-1.941652,
        -2.617994,1.596968,-1.328358,
        -2.617994,1.597178,-1.575607,
        -2.617994,1.600470,-1.453196,
        -2.617994,1.601408,0.242409,
        -2.617994,1.602259,1.152656,
        -2.617994,1.603064,2.865774,
        -2.617994,1.612085,1.968802,
        -2.617994,1.612720,-3.080498,
        -2.617994,1.616061,2.521531,
        -2.617994,1.616900,-0.207457,
        -2.617994,1.620988,-0.875920,
        -2.617994,1.622114,0.701509,
        -2.617994,1.622749,0.121426,
        -2.617994,1.622993,-2.221629,
        -2.617994,1.629346,-0.542669,
        -2.617994,1.630007,1.844552,
        -2.617994,1.634747,-2.426198,
        -2.617994,1.642346,1.722617,
        -2.617994,1.643925,1.034629,
        -2.617994,1.644848,2.743774,
        -2.617994,1.645258,2.188109,
        -2.617994,1.645417,-2.634725,
        -2.617994,1.647395,3.085437,
        -2.617994,1.648538,-2.855044,
        -2.617994,1.649932,0.576808,
        -2.617994,1.652224,-0.996105,
        -2.617994,1.652871,1.600681,
        -2.617994,1.657997,-0.325127,
        -2.617994,1.661936,0.001037,
        -2.617994,1.665187,1.477402,
        -2.617994,1.665496,2.397872,
        -2.617994,1.673267,-0.665453,
        -2.617994,1.673557,0.452526,
        -2.617994,1.679277,-1.122421,
        -2.617994,1.679696,0.912482,
        -2.617994,1.682420,1.353922,
        -2.617994,1.685031,-2.009038,
        -2.617994,1.685196,2.967046,
        -2.617994,1.690107,2.620910,
        -2.617994,1.690453,-2.123076,
        -2.617994,1.691693,-1.871572,
        -2.617994,1.691817,2.061688,
        -2.617994,1.692644,1.237670,
        -2.617994,1.696386,0.328123,
        -2.617994,1.697162,-1.249026,
        -2.617994,1.697488,-2.322785,
        -2.617994,1.698726,-2.973576,
        -2.617994,1.699172,-1.749717,
        -2.617994,1.704478,-0.106295,
        -2.617994,1.707244,-0.444770,
        -2.617994,1.707500,-2.531497,
        -2.617994,1.707714,-0.787321,
        -2.617994,1.708430,-1.627552,
        -2.617994,1.710583,-1.376890,
        -2.617994,1.712415,-2.747921,
        -2.617994,1.713125,-1.503073,
        -2.617994,1.713452,0.789419,
        -2.617994,1.720612,2.279888,
        -2.617994,1.724852,0.202831,
        -2.617994,1.724868,2.846246,
        -2.617994,1.728940,1.931535,
        -2.617994,1.732838,1.127584,
        -2.617994,1.735719,-3.092485,
        -2.617994,1.739676,2.496017,
        -2.617994,1.739971,-0.228876,
        -2.617994,1.743328,0.663347,
        -2.617994,1.744927,-0.911892,
        -2.617994,1.745952,0.084733,
        -2.617994,1.748077,1.804354,
        -2.617994,1.752959,-0.567548,
        -2.617994,1.760169,1.678132,
        -2.617994,1.763887,-2.218450,
        -2.617994,1.766737,2.722752,
        -2.617994,1.767999,1.004737,
        -2.617994,1.769149,1.550298,
        -2.617994,1.770119,-2.427175,
        -2.617994,1.770460,2.154208,
        -2.617994,1.770514,3.070982,
        -2.617994,1.770735,0.537245,
        -2.617994,1.772673,-2.864760,
        -2.617994,1.773527,-1.039582,
        -2.617994,1.776419,-2.642421,
        -2.617994,1.780769,1.416330,
        -2.617994,1.786634,-0.346247,
        -2.617994,1.788829,-1.947553,
        -2.617994,1.793737,0.410270,
        -2.617994,1.794553,2.374618,
        -2.617994,1.794699,-0.694061,
        -2.617994,1.797469,-1.169105,
        -2.617994,1.803705,0.881076,
        -2.617994,1.804124,1.264249,
        -2.617994,1.807825,-2.090455,
        -2.617994,1.808822,-1.817375,
        -2.617994,1.809020,2.949741,
        -2.617994,1.812466,2.597595,
        -2.617994,1.813159,2.025395,
        -2.617994,1.813816,-1.298435,
        -2.617994,1.815016,0.278951,
        -2.617994,1.816510,-0.008631,
        -2.617994,1.817772,-1.689625,
        -2.617994,1.821360,-2.984352,
        -2.617994,1.822312,-0.127211,
        -2.617994,1.823386,-1.429807,
        -2.617994,1.823834,-1.560483,
        -2.617994,1.832060,-0.821627,
        -2.617994,1.832690,-0.468440,
        -2.617994,1.833945,-2.320278,
        -2.617994,1.834787,0.751968,
        -2.617994,1.840579,-2.536445,
        -2.617994,1.841033,-2.756932,
        -2.617994,1.846350,1.893983,
        -2.617994,1.846644,1.113766,
        -2.617994,1.846833,2.826543,
        -2.617994,1.847325,2.249051,
        -2.617994,1.854119,0.134161,
        -2.617994,1.858863,-3.106022,
        -2.617994,1.865046,1.763384,
        -2.617994,1.865619,-0.952036,
        -2.617994,1.865912,0.624453,
        -2.617994,1.866694,2.473686,
        -2.617994,1.867524,-0.240083,
        -2.617994,1.875561,1.632831,
        -2.617994,1.876049,-0.594378,
        -2.617994,1.880509,1.500204,
        -2.617994,1.883504,1.360457,
        -2.617994,1.886971,-2.196665,
        -2.617994,1.888979,0.982922,
        -2.617994,1.889815,2.701570,
        -2.617994,1.892144,-2.005340,
        -2.617994,1.892348,0.496522,
        -2.617994,1.893281,-1.083976,
        -2.617994,1.893828,3.053453,
        -2.617994,1.895117,2.121107,
        -2.617994,1.898741,-2.875735,
        -2.617994,1.904429,-2.428029,
        -2.617994,1.908408,-2.648903,
        -2.617994,1.911828,-0.364489,
        -2.617994,1.912224,0.366565,
        -2.617994,1.914854,1.214468,
        -2.617994,1.915467,-1.883143,
        -2.617994,1.915763,-0.724679,
        -2.617994,1.915790,-1.216366,
        -2.617994,1.922256,2.347806,
        -2.617994,1.922616,0.844633,
        -2.617994,1.925185,0.231170,
        -2.617994,1.927474,-1.754616,
        -2.617994,1.928342,-1.351174,
        -2.617994,1.933328,-1.623328,
        -2.617994,1.933878,0.034150,
        -2.617994,1.934585,-1.487979,
        -2.617994,1.934802,2.927634,
        -2.617994,1.934862,1.989776,
        -2.617994,1.936327,-0.099350,
        -2.617994,1.938088,2.578507,
        -2.617994,1.945677,-2.999805,
        -2.617994,1.954572,-0.859208,
        -2.617994,1.957868,-0.490784,
        -2.617994,1.958989,0.712772,
        -2.617994,1.960212,-2.307561,
        -2.617994,1.964774,1.855507,
        -2.617994,1.964971,-2.100533,
        -2.617994,1.969112,1.088273,
        -2.617994,1.971438,-2.765415,
        -2.617994,1.972312,2.805134,
        -2.617994,1.974950,2.220596,
        -2.617994,1.975135,-2.539127,
        -2.617994,1.982107,1.720829,
        -2.617994,1.983195,-3.127984,
        -2.617994,1.983318,-0.247182,
        -2.617994,1.986720,-0.993220,
        -2.617994,1.990381,1.586005,
        -2.617994,1.990470,0.580283,
        -2.617994,1.990523,1.450326,
        -2.617994,1.993909,1.312897,
        -2.617994,1.994694,2.452813,
        -2.617994,1.999226,-0.620717,
        -2.617994,2.005145,0.941012,
        -2.617994,2.007301,0.142337,
        -2.617994,2.014149,0.444138,
        -2.617994,2.014338,-1.127392,
        -2.617994,2.018131,3.021778,
        -2.617994,2.019676,-1.980944,
        -2.617994,2.019912,2.686612,
        -2.617994,2.020545,2.088672,
        -2.617994,2.026074,0.300945,
        -2.617994,2.026196,-2.889057,
        -2.617994,2.031959,-1.268178,
        -2.617994,2.032917,-2.199917,
        -2.617994,2.033572,-1.842884,
        -2.617994,2.034712,-2.419505,
        -2.617994,2.035818,-0.379479,
        -2.617994,2.036004,-0.757062,
        -2.617994,2.040501,-1.411488,
        -2.617994,2.042874,-1.701367,
        -2.617994,2.043060,-2.653746,
        -2.617994,2.043729,-1.557938,
        -2.617994,2.048538,0.806623,
        -2.617994,2.049006,1.199930,
        -2.617994,2.049502,-0.137989,
        -2.617994,2.052181,2.326959,
        -2.617994,2.052794,-0.008526,
        -2.617994,2.057608,1.951726,
        -2.617994,2.065568,2.567730,
        -2.617994,2.070371,2.879808,
        -2.617994,2.071256,-3.018961,
        -2.617994,2.079746,-0.897699,
        -2.617994,2.082784,1.042403,
        -2.617994,2.083752,-0.511086,
        -2.617994,2.084401,1.810502,
        -2.617994,2.085731,0.671081,
        -2.617994,2.097496,1.373172,
        -2.617994,2.099185,1.667006,
        -2.617994,2.100482,-2.301084,
        -2.617994,2.103118,2.194745,
        -2.617994,2.103223,1.521685,
        -2.617994,2.103695,-2.777707,
        -2.617994,2.107523,3.129581,
        -2.617994,2.108277,-2.095907,
        -2.617994,2.108891,-2.532166,
        -2.617994,2.111003,-0.268632,
        -2.617994,2.112276,-1.027246,
        -2.617994,2.114380,0.531457,
        -2.617994,2.119166,0.100552,
        -2.617994,2.119564,2.754818,
        -2.617994,2.122474,-0.648314,
        -2.617994,2.125183,2.445598,
        -2.617994,2.129501,0.238290,
        -2.617994,2.131178,0.386663,
        -2.617994,2.131223,-1.174311,
        -2.617994,2.134269,-1.948042,
        -2.617994,2.134684,0.907065,
        -2.617994,2.137615,2.988806,
        -2.617994,2.145635,-1.327708,
        -2.617994,2.146467,2.055081,
        -2.617994,2.146582,-1.793788,
        -2.617994,2.149210,-1.482321,
        -2.617994,2.152901,-1.636788,
        -2.617994,2.156337,1.151558,
        -2.617994,2.156474,-2.912065,
        -2.617994,2.157194,-0.792563,
        -2.617994,2.166724,-0.401638,
        -2.617994,2.166980,-0.134879,
        -2.617994,2.173644,-2.403429,
        -2.617994,2.175711,-2.655523,
        -2.617994,2.177132,2.600419,
        -2.617994,2.178879,0.767722,
        -2.617994,2.180752,1.908485,
        -2.617994,2.181338,2.313721,
        -2.617994,2.185580,1.287673,
        -2.617994,2.195400,-2.217546,
        -2.617994,2.200418,-3.056442,
        -2.617994,2.201829,-0.001735,
        -2.617994,2.203804,1.756176,
        -2.617994,2.211602,1.443930,
        -2.617994,2.212686,-0.544809,
        -2.617994,2.213937,0.621892,
        -2.617994,2.214075,1.600691,
        -2.617994,2.216256,1.020029,
        -2.617994,2.219060,2.886631,
        -2.617994,2.223375,-0.906185,
        -2.617994,2.226002,-1.060248,
        -2.617994,2.226413,-2.064034,
        -2.617994,2.230554,2.171334,
        -2.617994,2.232413,2.735777,
        -2.617994,2.233709,3.070830,
        -2.617994,2.234383,-2.792065,
        -2.617994,2.237183,0.468799,
        -2.617994,2.237695,-0.271388,
        -2.617994,2.239419,0.151693,
        -2.617994,2.243347,-2.521049,
        -2.617994,2.244035,-1.229456,
        -2.617994,2.244355,0.309899,
        -2.617994,2.245175,2.460236,
        -2.617994,2.245628,-1.899216,
        -2.617994,2.245973,-0.702234,
        -2.617994,2.255601,-1.728131,
        -2.617994,2.256702,-1.401367,
        -2.617994,2.261466,-1.557500,
        -2.617994,2.267390,0.878499,
        -2.617994,2.272099,2.018594,
        -2.617994,2.278618,1.170386,
        -2.617994,2.283304,-2.345584,
        -2.617994,2.285175,-2.943090,
        -2.617994,2.293000,-0.124451,
        -2.617994,2.294058,-0.421070,
        -2.617994,2.303283,1.855938,
        -2.617994,2.304968,2.608978,
        -2.617994,2.305347,2.311346,
        -2.617994,2.307188,-2.656134,
        -2.617994,2.307301,1.336200,
        -2.617994,2.309655,0.726020,
        -2.617994,2.317634,-2.185225,
        -2.617994,2.321510,1.685664,
        -2.617994,2.324376,-3.111006,
        -2.617994,2.324731,1.511186,
        -2.617994,2.329587,0.035498,
        -2.617994,2.330158,2.977264,
        -2.617994,2.335096,-0.588468,
        -2.617994,2.336046,-1.115799,
        -2.617994,2.338060,2.803613,
        -2.617994,2.338083,-0.943400,
        -2.617994,2.340210,0.562145,
        -2.617994,2.341931,1.023926,
        -2.617994,2.343265,-0.779061,
        -2.617994,2.343434,-2.013692,
        -2.617994,2.353029,-1.310529,
        -2.617994,2.354868,0.211543,
        -2.617994,2.355120,0.388593,
        -2.617994,2.355165,-1.833173,
        -2.617994,2.355462,2.151441,
        -2.617994,2.362948,-2.487477,
        -2.617994,2.363101,-2.810456,
        -2.617994,2.363394,-0.272918,
        -2.617994,2.364509,-1.644533,
        -2.617994,2.370465,2.466824,
        -2.617994,2.386110,-1.481351,
        -2.617994,2.394960,0.862133,
        -2.617994,2.395716,1.977184,
        -2.617994,2.396339,1.200900,
        -2.617994,2.407941,-2.316753,
        -2.617994,2.409450,-2.988432,
        -2.617994,2.417202,-0.103363,
        -2.617994,2.417386,-0.443097,
        -2.617994,2.417950,2.666662,
        -2.617994,2.423190,1.789783,
        -2.617994,2.423692,1.392696,
        -2.617994,2.431122,3.086379,
        -2.617994,2.431406,-2.647820,
        -2.617994,2.431584,2.307631,
        -2.617994,2.434090,1.592818,
        -2.617994,2.435993,0.683844,
        -2.617994,2.440898,-2.133300,
        -2.617994,2.442222,-1.195837,
        -2.617994,2.443112,2.883761,
        -2.617994,2.449191,-0.633651,
        -2.617994,2.450821,0.084741,
        -2.617994,2.453499,-1.008861,
        -2.617994,2.460010,-1.938011,
        -2.617994,2.460199,-0.822008,
        -2.617994,2.460902,0.490079,
        -2.617994,2.461774,-1.734913,
        -2.617994,2.463986,1.034407,
        -2.617994,2.467647,0.287667,
        -2.617994,2.479849,2.126925,
        -2.617994,2.485907,-2.835471,
        -2.617994,2.485945,-1.374853,
        -2.617994,2.487747,2.505912,
        -2.617994,2.487957,-0.274279,
        -2.617994,2.489741,-2.468428,
        -2.617994,2.512538,1.239900,
        -2.617994,2.517528,1.923289,
        -2.617994,2.518454,-1.565852,
        -2.617994,2.519348,0.843503,
        -2.617994,2.522766,-3.059231,
        -2.617994,2.528686,2.737228,
        -2.617994,2.534798,-2.266671,
        -2.617994,2.536533,1.466274,
        -2.617994,2.538504,1.699812,
        -2.617994,2.539516,-0.476942,
        -2.617994,2.539532,-0.074548,
        -2.617994,2.543197,2.990252,
        -2.617994,2.554022,2.320619,
        -2.617994,2.556919,-2.655005,
        -2.617994,2.558146,0.627191,
        -2.617994,2.558223,-1.164189,
        -2.617994,2.563452,-2.044028,
        -2.617994,2.568462,0.150655,
        -2.617994,2.569130,-0.700837,
        -2.617994,2.570050,-1.804492,
        -2.617994,2.572710,-0.934665,
        -2.617994,2.575994,0.391718,
        -2.617994,2.584774,1.046703,
        -2.617994,2.601432,2.102535,
        -2.617994,2.603559,2.555114,
        -2.617994,2.604150,-2.885208,
        -2.617994,2.608630,-1.378011,
        -2.617994,2.611256,-0.279110,
        -2.617994,2.617344,-2.439480,
        -2.617994,2.627402,1.293552,
        -2.617994,2.635396,1.848850,
        -2.617994,2.635686,2.834950,
        -2.617994,2.637087,3.132730,
        -2.617994,2.642029,0.813274,
        -2.617994,2.645260,1.570203,
        -2.617994,2.646182,-1.614036,
        -2.617994,2.660113,-2.186513,
        -2.617994,2.660483,-0.034608,
        -2.617994,2.661713,-0.529511,
        -2.617994,2.672720,-1.110422,
        -2.617994,2.676626,2.335353,
        -2.617994,2.676638,0.540364,
        -2.617994,2.676780,-1.901158,
        -2.617994,2.681411,-2.667965,
        -2.617994,2.682152,0.245751,
        -2.617994,2.682859,-0.817071,
        -2.617994,2.705107,1.062142,
        -2.617994,2.717705,2.625995,
        -2.617994,2.721001,2.060097,
        -2.617994,2.722319,-2.954069,
        -2.617994,2.729840,-1.368237,
        -2.617994,2.734349,-0.289298,
        -2.617994,2.738409,2.975671,
        -2.617994,2.740366,1.374341,
        -2.617994,2.744633,-2.391311,
        -2.617994,2.747398,1.727166,
        -2.617994,2.763799,0.760372,
        -2.617994,2.768238,-1.679243,
        -2.617994,2.779245,-2.046268,
        -2.617994,2.780208,-0.626453,
        -2.617994,2.780471,0.027732,
        -2.617994,2.784696,-1.013128,
        -2.617994,2.789817,0.395431,
        -2.617994,2.801154,2.353143,
        -2.617994,2.807091,-2.687477,
        -2.617994,2.825762,1.085035,
        -2.617994,2.830209,2.734943,
        -2.617994,2.833517,-3.085508,
        -2.617994,2.838301,1.980061,
        -2.617994,2.849871,1.512012,
        -2.617994,2.850089,-1.343899,
        -2.617994,2.856905,-0.318441,
        -2.617994,2.870445,-2.291001,
        -2.617994,2.885017,0.655562,
        -2.617994,2.886428,-1.793217,
        -2.617994,2.890337,-0.825633,
        -2.617994,2.898932,0.139973,
        -2.617994,2.931927,-2.728768,
        -2.617994,2.932096,2.377790,
        -2.617994,2.937409,2.947245,
        -2.617994,2.948048,1.133127,
        -2.617994,2.950896,1.797889,
        -2.617994,2.970019,-1.282382,
        -2.617994,2.978333,-0.412090,
        -2.617994,2.996437,-2.089121,
        -2.617994,3.011274,0.434110,
        -2.617994,3.050410,3.138492,
        -2.617994,3.064492,1.647259,
        -2.617994,3.088868,-0.989501,
        -2.513274,0.053884,-2.408394,
        -2.513274,0.077645,1.453433,
        -2.513274,0.088408,-0.311160,
        -2.513274,0.142314,2.590387,
        -2.513274,0.147206,-1.385164,
        -2.513274,0.163506,0.573203,
        -2.513274,0.165485,-2.875396,
        -2.513274,0.182564,-2.125821,
        -2.513274,0.190569,1.929141,
        -2.513274,0.201102,1.249082,
        -2.513274,0.201397,-0.716646,
        -2.513274,0.210568,-0.069559,
        -2.513274,0.251672,2.848900,
        -2.513274,0.266262,-1.678068,
        -2.513274,0.269108,-1.215563,
        -2.513274,0.273777,-2.506756,
        -2.513274,0.279854,0.814581,
        -2.513274,0.281664,-2.969147,
        -2.513274,0.283625,2.405603,
        -2.513274,0.284058,0.357350,
        -2.513274,0.294107,1.593880,
        -2.513274,0.302491,-0.400410,
        -2.513274,0.308059,2.021491,
        -2.513274,0.312219,-2.072057,
        -2.513274,0.321905,-0.829121,
        -2.513274,0.329984,1.201621,
        -2.513274,0.339714,-0.013712,
        -2.513274,0.368784,2.974974,
        -2.513274,0.376428,-1.495532,
        -2.513274,0.378103,2.646436,
        -2.513274,0.385088,0.590355,
        -2.513274,0.386248,-2.690943,
        -2.513274,0.389851,-2.340031,
        -2.513274,0.391502,-1.171374,
        -2.513274,0.398249,-1.812694,
        -2.513274,0.400542,-2.995729,
        -2.513274,0.404273,0.918163,
        -2.513274,0.405617,1.773533,
        -2.513274,0.409009,-0.570243,
        -2.513274,0.410700,0.270433,
        -2.513274,0.415893,1.464890,
        -2.513274,0.420680,2.088164,
        -2.513274,0.424766,-0.258729,
        -2.513274,0.426507,2.367872,
        -2.513274,0.441292,-0.890727,
        -2.513274,0.451839,-2.078193,
        -2.513274,0.459270,1.191562,
        -2.513274,0.470622,0.014002,
        -2.513274,0.481292,-2.529098,
        -2.513274,0.482460,2.849843,
        -2.513274,0.492205,-1.397963,
        -2.513274,0.493314,3.117264,
        -2.513274,0.497870,0.727828,
        -2.513274,0.499783,-1.654379,
        -2.513274,0.502651,0.463333,
        -2.513274,0.505463,2.594557,
        -2.513274,0.506104,-2.874910,
        -2.513274,0.513579,-1.147354,
        -2.513274,0.518569,1.644237,
        -2.513274,0.519001,1.900488,
        -2.513274,0.519277,-0.673809,
        -2.513274,0.524769,-0.421955,
        -2.513274,0.525564,-2.288838,
        -2.513274,0.532541,0.978737,
        -2.513274,0.534610,2.154808,
        -2.513274,0.535020,-1.891846,
        -2.513274,0.540208,1.400497,
        -2.513274,0.542006,0.228337,
        -2.513274,0.550762,-0.180771,
        -2.513274,0.560531,2.378474,
        -2.513274,0.561629,-0.924576,
        -2.513274,0.567465,-2.679392,
        -2.513274,0.588527,1.181755,
        -2.513274,0.591643,-2.098364,
        -2.513274,0.594963,2.997575,
        -2.513274,0.599643,-3.045235,
        -2.513274,0.599962,2.773556,
        -2.513274,0.602620,0.029270,
        -2.513274,0.605711,0.598454,
        -2.513274,0.607708,-1.541026,
        -2.513274,0.609360,-2.464379,
        -2.513274,0.610738,-1.326786,
        -2.513274,0.616968,0.820489,
        -2.513274,0.626379,1.773610,
        -2.513274,0.628626,-1.751180,
        -2.513274,0.630671,0.391949,
        -2.513274,0.632622,-0.737642,
        -2.513274,0.632708,2.566722,
        -2.513274,0.632797,1.989858,
        -2.513274,0.634455,-0.534790,
        -2.513274,0.636205,1.561170,
        -2.513274,0.637681,-1.123423,
        -2.513274,0.641441,-2.842109,
        -2.513274,0.644807,-0.327139,
        -2.513274,0.652440,2.197129,
        -2.513274,0.660897,-2.278593,
        -2.513274,0.664008,1.360091,
        -2.513274,0.670850,-1.945257,
        -2.513274,0.671396,1.004859,
        -2.513274,0.673985,0.203855,
        -2.513274,0.680326,-0.133370,
        -2.513274,0.685083,-0.942363,
        -2.513274,0.688204,-2.632204,
        -2.513274,0.689777,2.385011,
        -2.513274,0.700065,3.115430,
        -2.513274,0.704080,2.911636,
        -2.513274,0.712883,1.174216,
        -2.513274,0.715701,0.699906,
        -2.513274,0.718851,-1.442215,
        -2.513274,0.721019,2.721536,
        -2.513274,0.725300,-2.975134,
        -2.513274,0.727136,-1.633288,
        -2.513274,0.727399,-2.120567,
        -2.513274,0.730751,0.522488,
        -2.513274,0.732892,-1.254330,
        -2.513274,0.734333,-2.446184,
        -2.513274,0.735125,0.037270,
        -2.513274,0.737893,1.866463,
        -2.513274,0.740650,1.681075,
        -2.513274,0.745971,-0.781286,
        -2.513274,0.749432,-0.442559,
        -2.513274,0.750102,2.050636,
        -2.513274,0.751046,0.863821,
        -2.513274,0.757018,-0.608833,
        -2.513274,0.757046,-1.816019,
        -2.513274,0.757389,1.501184,
        -2.513274,0.759800,2.546181,
        -2.513274,0.759859,0.351923,
        -2.513274,0.761141,-2.781102,
        -2.513274,0.769035,-1.078574,
        -2.513274,0.770166,-0.266604,
        -2.513274,0.772218,2.227380,
        -2.513274,0.788382,1.331051,
        -2.513274,0.790269,-2.282409,
        -2.513274,0.803410,-1.984567,
        -2.513274,0.803813,0.191520,
        -2.513274,0.806832,3.022155,
        -2.513274,0.807476,-2.597675,
        -2.513274,0.812512,-0.107335,
        -2.513274,0.814161,1.010949,
        -2.513274,0.817732,-3.084486,
        -2.513274,0.817846,2.390688,
        -2.513274,0.819323,2.845960,
        -2.513274,0.819430,-0.915017,
        -2.513274,0.829198,-1.532491,
        -2.513274,0.829531,0.665695,
        -2.513274,0.829816,1.167147,
        -2.513274,0.833613,-1.357492,
        -2.513274,0.842945,-2.907318,
        -2.513274,0.844883,2.679708,
        -2.513274,0.848189,-1.705003,
        -2.513274,0.849888,1.767901,
        -2.513274,0.851126,0.493370,
        -2.513274,0.852911,1.933394,
        -2.513274,0.857516,-2.432863,
        -2.513274,0.858261,-1.191729,
        -2.513274,0.858279,-2.138895,
        -2.513274,0.859723,-0.734928,
        -2.513274,0.859754,1.603292,
        -2.513274,0.861655,-0.394821,
        -2.513274,0.864300,0.043549,
        -2.513274,0.866940,0.851229,
        -2.513274,0.869748,2.095969,
        -2.513274,0.877352,-0.574668,
        -2.513274,0.879254,-2.733338,
        -2.513274,0.882554,1.443757,
        -2.513274,0.884060,0.334199,
        -2.513274,0.885626,-1.866061,
        -2.513274,0.889834,2.529957,
        -2.513274,0.893140,2.251698,
        -2.513274,0.897463,-0.240822,
        -2.513274,0.898758,-1.031948,
        -2.513274,0.913918,1.287618,
        -2.513274,0.916555,-2.287173,
        -2.513274,0.918305,3.107845,
        -2.513274,0.919522,2.948277,
        -2.513274,0.927138,-2.569686,
        -2.513274,0.929031,0.185383,
        -2.513274,0.933766,-3.015917,
        -2.513274,0.934873,-2.010822,
        -2.513274,0.935636,-1.445855,
        -2.513274,0.936702,-0.866461,
        -2.513274,0.937395,0.729108,
        -2.513274,0.937910,2.787575,
        -2.513274,0.939144,-0.092962,
        -2.513274,0.941718,1.125085,
        -2.513274,0.942873,-1.607680,
        -2.513274,0.943701,0.980328,
        -2.513274,0.947774,2.395256,
        -2.513274,0.952433,0.584347,
        -2.513274,0.954456,-1.287293,
        -2.513274,0.958731,-2.852552,
        -2.513274,0.961838,1.839284,
        -2.513274,0.961898,-0.469425,
        -2.513274,0.965426,1.686615,
        -2.513274,0.970108,1.990771,
        -2.513274,0.972820,-1.762780,
        -2.513274,0.973500,-0.711881,
        -2.513274,0.975253,2.641082,
        -2.513274,0.977824,0.441928,
        -2.513274,0.981485,-2.424766,
        -2.513274,0.981877,1.537100,
        -2.513274,0.982609,-1.136920,
        -2.513274,0.984265,-2.156073,
        -2.513274,0.989150,0.045493,
        -2.513274,0.993052,2.137478,
        -2.513274,0.997766,-2.693871,
        -2.513274,0.999847,-0.334738,
        -2.513274,1.001920,0.848054,
        -2.513274,1.006146,1.389734,
        -2.513274,1.012180,0.304066,
        -2.513274,1.016938,2.270674,
        -2.513274,1.017371,-0.979790,
        -2.513274,1.019495,-1.894487,
        -2.513274,1.023796,-0.583691,
        -2.513274,1.026525,2.506945,
        -2.513274,1.027004,-0.203653,
        -2.513274,1.028027,3.036573,
        -2.513274,1.031435,1.241021,
        -2.513274,1.032316,2.884221,
        -2.513274,1.035863,-3.110475,
        -2.513274,1.042085,-2.292950,
        -2.513274,1.043272,-1.522986,
        -2.513274,1.046265,-2.958847,
        -2.513274,1.049272,-1.376344,
        -2.513274,1.050039,-2.552086,
        -2.513274,1.053682,0.689863,
        -2.513274,1.055458,0.170394,
        -2.513274,1.056622,-0.831866,
        -2.513274,1.057515,1.097998,
        -2.513274,1.058706,-2.036438,
        -2.513274,1.061389,-1.672248,
        -2.513274,1.063758,2.743093,
        -2.513274,1.070157,-0.073422,
        -2.513274,1.072154,1.761506,
        -2.513274,1.074029,1.909193,
        -2.513274,1.074204,0.545276,
        -2.513274,1.074714,-2.805147,
        -2.513274,1.076262,0.966104,
        -2.513274,1.077642,-1.227073,
        -2.513274,1.084151,1.618525,
        -2.513274,1.085085,2.377673,
        -2.513274,1.086822,-0.441112,
        -2.513274,1.087142,2.058815,
        -2.513274,1.098383,-0.696430,
        -2.513274,1.099138,-1.084033,
        -2.513274,1.103208,1.477773,
        -2.513274,1.103783,0.408254,
        -2.513274,1.106275,-2.420800,
        -2.513274,1.108079,2.612943,
        -2.513274,1.109345,-2.172718,
        -2.513274,1.115957,-1.787597,
        -2.513274,1.118930,0.045983,
        -2.513274,1.119777,0.834702,
        -2.513274,1.121943,-2.669042,
        -2.513274,1.124930,2.219917,
        -2.513274,1.125217,1.336294,
        -2.513274,1.130958,2.972931,
        -2.513274,1.131400,-0.306629,
        -2.513274,1.137135,-0.936514,
        -2.513274,1.139125,-1.924528,
        -2.513274,1.139877,0.276382,
        -2.513274,1.142003,-3.058022,
        -2.513274,1.147363,-0.558448,
        -2.513274,1.150348,-0.176418,
        -2.513274,1.151198,1.197016,
        -2.513274,1.153930,-1.597104,
        -2.513274,1.155389,2.837148,
        -2.513274,1.155458,-1.460623,
        -2.513274,1.157331,-2.906362,
        -2.513274,1.158594,2.487618,
        -2.513274,1.162103,-1.322553,
        -2.513274,1.165407,3.092861,
        -2.513274,1.168025,0.717075,
        -2.513274,1.168406,-2.301519,
        -2.513274,1.174782,-2.538755,
        -2.513274,1.177998,1.061256,
        -2.513274,1.178898,1.827305,
        -2.513274,1.181331,-2.060330,
        -2.513274,1.182540,-0.798952,
        -2.513274,1.184569,0.147517,
        -2.513274,1.184734,1.968919,
        -2.513274,1.184804,0.593386,
        -2.513274,1.188278,1.690808,
        -2.513274,1.194360,2.110440,
        -2.513274,1.194776,2.710455,
        -2.513274,1.197745,-1.173705,
        -2.513274,1.199898,-2.777257,
        -2.513274,1.201164,2.355931,
        -2.513274,1.203760,1.556404,
        -2.513274,1.205845,-0.415993,
        -2.513274,1.206585,0.933648,
        -2.513274,1.206950,-0.037248,
        -2.513274,1.209341,0.471162,
        -2.513274,1.213483,-1.039417,
        -2.513274,1.218146,-0.672643,
        -2.513274,1.221608,1.421669,
        -2.513274,1.226043,-1.693533,
        -2.513274,1.232870,-1.822773,
        -2.513274,1.234375,-2.421194,
        -2.513274,1.234531,-2.190148,
        -2.513274,1.239414,0.348959,
        -2.513274,1.240183,2.587255,
        -2.513274,1.242312,2.941234,
        -2.513274,1.244075,1.286644,
        -2.513274,1.245961,-2.997461,
        -2.513274,1.246495,-2.646872,
        -2.513274,1.247736,2.242810,
        -2.513274,1.253604,0.812222,
        -2.513274,1.257816,-0.894317,
        -2.513274,1.258247,-1.953634,
        -2.513274,1.258993,-0.284099,
        -2.513274,1.262380,-0.157609,
        -2.513274,1.262791,-1.549229,
        -2.513274,1.263772,-0.533073,
        -2.513274,1.266731,-1.412461,
        -2.513274,1.268591,-3.122146,
        -2.513274,1.269614,-1.275246,
        -2.513274,1.270745,1.152533,
        -2.513274,1.273606,0.224545,
        -2.513274,1.279716,2.812101,
        -2.513274,1.283786,2.462124,
        -2.513274,1.284798,0.078209,
        -2.513274,1.285402,-2.882049,
        -2.513274,1.286936,1.886053,
        -2.513274,1.289408,0.683122,
        -2.513274,1.294736,1.756184,
        -2.513274,1.296978,-2.311511,
        -2.513274,1.298546,2.020601,
        -2.513274,1.300924,1.022478,
        -2.513274,1.303305,3.043152,
        -2.513274,1.304210,-2.084225,
        -2.513274,1.305241,-2.530249,
        -2.513274,1.307513,1.628003,
        -2.513274,1.310722,0.558267,
        -2.513274,1.314142,2.139329,
        -2.513274,1.316087,-0.777971,
        -2.513274,1.317907,-1.121023,
        -2.513274,1.321019,-2.749776,
        -2.513274,1.321609,2.687717,
        -2.513274,1.322568,1.499415,
        -2.513274,1.322972,-0.395699,
        -2.513274,1.325286,-0.994192,
        -2.513274,1.329146,-0.648513,
        -2.513274,1.329795,-0.042570,
        -2.513274,1.331651,2.344497,
        -2.513274,1.336522,0.435098,
        -2.513274,1.339410,0.900087,
        -2.513274,1.339550,1.370353,
        -2.513274,1.340441,-1.731580,
        -2.513274,1.351855,-1.855234,
        -2.513274,1.360161,-3.018404,
        -2.513274,1.361031,-1.613971,
        -2.513274,1.361196,2.914780,
        -2.513274,1.361998,-2.207254,
        -2.513274,1.362302,1.239384,
        -2.513274,1.365228,2.563808,
        -2.513274,1.365452,0.310522,
        -2.513274,1.367028,-2.420041,
        -2.513274,1.373422,-1.484281,
        -2.513274,1.374245,-1.226417,
        -2.513274,1.375020,-0.170393,
        -2.513274,1.376560,-1.355481,
        -2.513274,1.377070,-1.980984,
        -2.513274,1.377578,-2.636339,
        -2.513274,1.378096,0.774626,
        -2.513274,1.379086,-0.508192,
        -2.513274,1.382562,0.177141,
        -2.513274,1.388849,2.233847,
        -2.513274,1.389268,-0.886349,
        -2.513274,1.391324,1.109396,
        -2.513274,1.391699,-0.288725,
        -2.513274,1.392869,3.136903,
        -2.513274,1.393612,1.952016,
        -2.513274,1.395545,-2.850807,
        -2.513274,1.401926,1.822930,
        -2.513274,1.402149,2.788898,
        -2.513274,1.409552,0.647980,
        -2.513274,1.410840,0.056219,
        -2.513274,1.411944,2.444377,
        -2.513274,1.413639,1.697982,
        -2.513274,1.426479,1.573427,
        -2.513274,1.426725,2.102242,
        -2.513274,1.427868,0.984637,
        -2.513274,1.429580,-2.107232,
        -2.513274,1.431375,-2.313269,
        -2.513274,1.431907,3.015410,
        -2.513274,1.432325,-1.029054,
        -2.513274,1.433341,0.524268,
        -2.513274,1.435181,-0.739771,
        -2.513274,1.437208,-2.529085,
        -2.513274,1.439041,-0.617966,
        -2.513274,1.441473,1.448544,
        -2.513274,1.444527,2.664744,
        -2.513274,1.450513,-0.400625,
        -2.513274,1.451907,-0.066823,
        -2.513274,1.452754,-2.738631,
        -2.513274,1.454061,-2.950416,
        -2.513274,1.456736,1.323154,
        -2.513274,1.458370,0.400433,
        -2.513274,1.459314,-1.765003,
        -2.513274,1.459902,-1.145835,
        -2.513274,1.464952,2.329245,
        -2.513274,1.466050,0.862104,
        -2.513274,1.471330,-1.886160,
        -2.513274,1.475119,-1.648255,
        -2.513274,1.480878,1.194576,
        -2.513274,1.480942,0.273235,
        -2.513274,1.481086,2.890269,
        -2.513274,1.484118,-1.529156,
        -2.513274,1.484741,-1.279219,
        -2.513274,1.487681,-1.406069,
        -2.513274,1.487886,-3.066037,
        -2.513274,1.490555,2.543284,
        -2.513274,1.492424,-0.190932,
        -2.513274,1.495002,-0.839810,
        -2.513274,1.495297,2.007188,
        -2.513274,1.496414,-2.009267,
        -2.513274,1.498261,-2.425270,
        -2.513274,1.498907,-2.207327,
        -2.513274,1.500204,0.737638,
        -2.513274,1.502469,0.151156,
        -2.513274,1.507448,-0.514782,
        -2.513274,1.509287,2.204647,
        -2.513274,1.510518,1.884920,
        -2.513274,1.511948,-2.632360,
        -2.513274,1.515235,1.068565,
        -2.513274,1.522237,3.102018,
        -2.513274,1.522373,1.764991,
        -2.513274,1.522416,-0.954956,
        -2.513274,1.523102,2.765920,
        -2.513274,1.524811,-2.844718,
        -2.513274,1.527261,-0.304916,
        -2.513274,1.529788,0.613208,
        -2.513274,1.533648,1.644808,
        -2.513274,1.535018,0.028869,
        -2.513274,1.541082,2.424608,
        -2.513274,1.546010,1.523462,
        -2.513274,1.551135,-0.642994,
        -2.513274,1.553895,0.489607,
        -2.513274,1.555041,0.947813,
        -2.513274,1.559671,-1.074934,
        -2.513274,1.560345,2.985899,
        -2.513274,1.560487,-2.323400,
        -2.513274,1.561127,1.401111,
        -2.513274,1.567193,2.643172,
        -2.513274,1.568892,2.098362,
        -2.513274,1.571702,-2.098435,
        -2.513274,1.572701,-2.528730,
        -2.513274,1.573572,1.278593,
        -2.513274,1.574849,-0.091505,
        -2.513274,1.576334,0.365236,
        -2.513274,1.577029,-2.962888,
        -2.513274,1.577688,-1.814941,
        -2.513274,1.579542,-0.755917,
        -2.513274,1.579624,-1.199759,
        -2.513274,1.581515,-0.419414,
        -2.513274,1.584581,-2.738702,
        -2.513274,1.586948,-1.696296,
        -2.513274,1.588901,2.301496,
        -2.513274,1.589936,0.825020,
        -2.513274,1.590018,-1.941652,
        -2.513274,1.596968,-1.328358,
        -2.513274,1.597178,-1.575607,
        -2.513274,1.600470,-1.453196,
        -2.513274,1.601408,0.242409,
        -2.513274,1.602259,1.152656,
        -2.513274,1.603064,2.865774,
        -2.513274,1.612085,1.968802,
        -2.513274,1.612720,-3.080498,
        -2.513274,1.616061,2.521531,
        -2.513274,1.616900,-0.207457,
        -2.513274,1.620988,-0.875920,
        -2.513274,1.622114,0.701509,
        -2.513274,1.622749,0.121426,
        -2.513274,1.622993,-2.221629,
        -2.513274,1.629346,-0.542669,
        -2.513274,1.630007,1.844552,
        -2.513274,1.634747,-2.426198,
        -2.513274,1.642346,1.722617,
        -2.513274,1.643925,1.034629,
        -2.513274,1.644848,2.743774,
        -2.513274,1.645258,2.188109,
        -2.513274,1.645417,-2.634725,
        -2.513274,1.647395,3.085437,
        -2.513274,1.648538,-2.855044,
        -2.513274,1.649932,0.576808,
        -2.513274,1.652224,-0.996105,
        -2.513274,1.652871,1.600681,
        -2.513274,1.657997,-0.325127,
        -2.513274,1.661936,0.001037,
        -2.513274,1.665187,1.477402,
        -2.513274,1.665496,2.397872,
        -2.513274,1.673267,-0.665453,
        -2.513274,1.673557,0.452526,
        -2.513274,1.679277,-1.122421,
        -2.513274,1.679696,0.912482,
        -2.513274,1.682420,1.353922,
        -2.513274,1.685031,-2.009038,
        -2.513274,1.685196,2.967046,
        -2.513274,1.690107,2.620910,
        -2.513274,1.690453,-2.123076,
        -2.513274,1.691693,-1.871572,
        -2.513274,1.691817,2.061688,
        -2.513274,1.692644,1.237670,
        -2.513274,1.696386,0.328123,
        -2.513274,1.697162,-1.249026,
        -2.513274,1.697488,-2.322785,
        -2.513274,1.698726,-2.973576,
        -2.513274,1.699172,-1.749717,
        -2.513274,1.704478,-0.106295,
        -2.513274,1.707244,-0.444770,
        -2.513274,1.707500,-2.531497,
        -2.513274,1.707714,-0.787321,
        -2.513274,1.708430,-1.627552,
        -2.513274,1.710583,-1.376890,
        -2.513274,1.712415,-2.747921,
        -2.513274,1.713125,-1.503073,
        -2.513274,1.713452,0.789419,
        -2.513274,1.720612,2.279888,
        -2.513274,1.724852,0.202831,
        -2.513274,1.724868,2.846246,
        -2.513274,1.728940,1.931535,
        -2.513274,1.732838,1.127584,
        -2.513274,1.735719,-3.092485,
        -2.513274,1.739676,2.496017,
        -2.513274,1.739971,-0.228876,
        -2.513274,1.743328,0.663347,
        -2.513274,1.744927,-0.911892,
        -2.513274,1.745952,0.084733,
        -2.513274,1.748077,1.804354,
        -2.513274,1.752959,-0.567548,
        -2.513274,1.760169,1.678132,
        -2.513274,1.763887,-2.218450,
        -2.513274,1.766737,2.722752,
        -2.513274,1.767999,1.004737,
        -2.513274,1.769149,1.550298,
        -2.513274,1.770119,-2.427175,
        -2.513274,1.770460,2.154208,
        -2.513274,1.770514,3.070982,
        -2.513274,1.770735,0.537245,
        -2.513274,1.772673,-2.864760,
        -2.513274,1.773527,-1.039582,
        -2.513274,1.776419,-2.642421,
        -2.513274,1.780769,1.416330,
        -2.513274,1.786634,-0.346247,
        -2.513274,1.788829,-1.947553,
        -2.513274,1.793737,0.410270,
        -2.513274,1.794553,2.374618,
        -2.513274,1.794699,-0.694061,
        -2.513274,1.797469,-1.169105,
        -2.513274,1.803705,0.881076,
        -2.513274,1.804124,1.264249,
        -2.513274,1.807825,-2.090455,
        -2.513274,1.808822,-1.817375,
        -2.513274,1.809020,2.949741,
        -2.513274,1.812466,2.597595,
        -2.513274,1.813159,2.025395,
        -2.513274,1.813816,-1.298435,
        -2.513274,1.815016,0.278951,
        -2.513274,1.816510,-0.008631,
        -2.513274,1.817772,-1.689625,
        -2.513274,1.821360,-2.984352,
        -2.513274,1.822312,-0.127211,
        -2.513274,1.823386,-1.429807,
        -2.513274,1.823834,-1.560483,
        -2.513274,1.832060,-0.821627,
        -2.513274,1.832690,-0.468440,
        -2.513274,1.833945,-2.320278,
        -2.513274,1.834787,0.751968,
        -2.513274,1.840579,-2.536445,
        -2.513274,1.841033,-2.756932,
        -2.513274,1.846350,1.893983,
        -2.513274,1.846644,1.113766,
        -2.513274,1.846833,2.826543,
        -2.513274,1.847325,2.249051,
        -2.513274,1.854119,0.134161,
        -2.513274,1.858863,-3.106022,
        -2.513274,1.865046,1.763384,
        -2.513274,1.865619,-0.952036,
        -2.513274,1.865912,0.624453,
        -2.513274,1.866694,2.473686,
        -2.513274,1.867524,-0.240083,
        -2.513274,1.875561,1.632831,
        -2.513274,1.876049,-0.594378,
        -2.513274,1.880509,1.500204,
        -2.513274,1.883504,1.360457,
        -2.513274,1.886971,-2.196665,
        -2.513274,1.888979,0.982922,
        -2.513274,1.889815,2.701570,
        -2.513274,1.892144,-2.005340,
        -2.513274,1.892348,0.496522,
        -2.513274,1.893281,-1.083976,
        -2.513274,1.893828,3.053453,
        -2.513274,1.895117,2.121107,
        -2.513274,1.898741,-2.875735,
        -2.513274,1.904429,-2.428029,
        -2.513274,1.908408,-2.648903,
        -2.513274,1.911828,-0.364489,
        -2.513274,1.912224,0.366565,
        -2.513274,1.914854,1.214468,
        -2.513274,1.915467,-1.883143,
        -2.513274,1.915763,-0.724679,
        -2.513274,1.915790,-1.216366,
        -2.513274,1.922256,2.347806,
        -2.513274,1.922616,0.844633,
        -2.513274,1.925185,0.231170,
        -2.513274,1.927474,-1.754616,
        -2.513274,1.928342,-1.351174,
        -2.513274,1.933328,-1.623328,
        -2.513274,1.933878,0.034150,
        -2.513274,1.934585,-1.487979,
        -2.513274,1.934802,2.927634,
        -2.513274,1.934862,1.989776,
        -2.513274,1.936327,-0.099350,
        -2.513274,1.938088,2.578507,
        -2.513274,1.945677,-2.999805,
        -2.513274,1.954572,-0.859208,
        -2.513274,1.957868,-0.490784,
        -2.513274,1.958989,0.712772,
        -2.513274,1.960212,-2.307561,
        -2.513274,1.964774,1.855507,
        -2.513274,1.964971,-2.100533,
        -2.513274,1.969112,1.088273,
        -2.513274,1.971438,-2.765415,
        -2.513274,1.972312,2.805134,
        -2.513274,1.974950,2.220596,
        -2.513274,1.975135,-2.539127,
        -2.513274,1.982107,1.720829,
        -2.513274,1.983195,-3.127984,
        -2.513274,1.983318,-0.247182,
        -2.513274,1.986720,-0.993220,
        -2.513274,1.990381,1.586005,
        -2.513274,1.990470,0.580283,
        -2.513274,1.990523,1.450326,
        -2.513274,1.993909,1.312897,
        -2.513274,1.994694,2.452813,
        -2.513274,1.999226,-0.620717,
        -2.513274,2.005145,0.941012,
        -2.513274,2.007301,0.142337,
        -2.513274,2.014149,0.444138,
        -2.513274,2.014338,-1.127392,
        -2.513274,2.018131,3.021778,
        -2.513274,2.019676,-1.980944,
        -2.513274,2.019912,2.686612,
        -2.513274,2.020545,2.088672,
        -2.513274,2.026074,0.300945,
        -2.513274,2.026196,-2.889057,
        -2.513274,2.031959,-1.268178,
        -2.513274,2.032917,-2.199917,
        -2.513274,2.033572,-1.842884,
        -2.513274,2.034712,-2.419505,
        -2.513274,2.035818,-0.379479,
        -2.513274,2.036004,-0.757062,
        -2.513274,2.040501,-1.411488,
        -2.513274,2.042874,-1.701367,
        -2.513274,2.043060,-2.653746,
        -2.513274,2.043729,-1.557938,
        -2.513274,2.048538,0.806623,
        -2.513274,2.049006,1.199930,
        -2.513274,2.049502,-0.137989,
        -2.513274,2.052181,2.326959,
        -2.513274,2.052794,-0.008526,
        -2.513274,2.057608,1.951726,
        -2.513274,2.065568,2.567730,
        -2.513274,2.070371,2.879808,
        -2.513274,2.071256,-3.018961,
        -2.513274,2.079746,-0.897699,
        -2.513274,2.082784,1.042403,
        -2.513274,2.083752,-0.511086,
        -2.513274,2.084401,1.810502,
        -2.513274,2.085731,0.671081,
        -2.513274,2.097496,1.373172,
        -2.513274,2.099185,1.667006,
        -2.513274,2.100482,-2.301084,
        -2.513274,2.103118,2.194745,
        -2.513274,2.103223,1.521685,
        -2.513274,2.103695,-2.777707,
        -2.513274,2.107523,3.129581,
        -2.513274,2.108277,-2.095907,
        -2.513274,2.108891,-2.532166,
        -2.513274,2.111003,-0.268632,
        -2.513274,2.112276,-1.027246,
        -2.513274,2.114380,0.531457,
        -2.513274,2.119166,0.100552,
        -2.513274,2.119564,2.754818,
        -2.513274,2.122474,-0.648314,
        -2.513274,2.125183,2.445598,
        -2.513274,2.129501,0.238290,
        -2.513274,2.131178,0.386663,
        -2.513274,2.131223,-1.174311,
        -2.513274,2.134269,-1.948042,
        -2.513274,2.134684,0.907065,
        -2.513274,2.137615,2.988806,
        -2.513274,2.145635,-1.327708,
        -2.513274,2.146467,2.055081,
        -2.513274,2.146582,-1.793788,
        -2.513274,2.149210,-1.482321,
        -2.513274,2.152901,-1.636788,
        -2.513274,2.156337,1.151558,
        -2.513274,2.156474,-2.912065,
        -2.513274,2.157194,-0.792563,
        -2.513274,2.166724,-0.401638,
        -2.513274,2.166980,-0.134879,
        -2.513274,2.173644,-2.403429,
        -2.513274,2.175711,-2.655523,
        -2.513274,2.177132,2.600419,
        -2.513274,2.178879,0.767722,
        -2.513274,2.180752,1.908485,
        -2.513274,2.181338,2.313721,
        -2.513274,2.185580,1.287673,
        -2.513274,2.195400,-2.217546,
        -2.513274,2.200418,-3.056442,
        -2.513274,2.201829,-0.001735,
        -2.513274,2.203804,1.756176,
        -2.513274,2.211602,1.443930,
        -2.513274,2.212686,-0.544809,
        -2.513274,2.213937,0.621892,
        -2.513274,2.214075,1.600691,
        -2.513274,2.216256,1.020029,
        -2.513274,2.219060,2.886631,
        -2.513274,2.223375,-0.906185,
        -2.513274,2.226002,-1.060248,
        -2.513274,2.226413,-2.064034,
        -2.513274,2.230554,2.171334,
        -2.513274,2.232413,2.735777,
        -2.513274,2.233709,3.070830,
        -2.513274,2.234383,-2.792065,
        -2.513274,2.237183,0.468799,
        -2.513274,2.237695,-0.271388,
        -2.513274,2.239419,0.151693,
        -2.513274,2.243347,-2.521049,
        -2.513274,2.244035,-1.229456,
        -2.513274,2.244355,0.309899,
        -2.513274,2.245175,2.460236,
        -2.513274,2.245628,-1.899216,
        -2.513274,2.245973,-0.702234,
        -2.513274,2.255601,-1.728131,
        -2.513274,2.256702,-1.401367,
        -2.513274,2.261466,-1.557500,
        -2.513274,2.267390,0.878499,
        -2.513274,2.272099,2.018594,
        -2.513274,2.278618,1.170386,
        -2.513274,2.283304,-2.345584,
        -2.513274,2.285175,-2.943090,
        -2.513274,2.293000,-0.124451,
        -2.513274,2.294058,-0.421070,
        -2.513274,2.303283,1.855938,
        -2.513274,2.304968,2.608978,
        -2.513274,2.305347,2.311346,
        -2.513274,2.307188,-2.656134,
        -2.513274,2.307301,1.336200,
        -2.513274,2.309655,0.726020,
        -2.513274,2.317634,-2.185225,
        -2.513274,2.321510,1.685664,
        -2.513274,2.324376,-3.111006,
        -2.513274,2.324731,1.511186,
        -2.513274,2.329587,0.035498,
        -2.513274,2.330158,2.977264,
        -2.513274,2.335096,-0.588468,
        -2.513274,2.336046,-1.115799,
        -2.513274,2.338060,2.803613,
        -2.513274,2.338083,-0.943400,
        -2.513274,2.340210,0.562145,
        -2.513274,2.341931,1.023926,
        -2.513274,2.343265,-0.779061,
        -2.513274,2.343434,-2.013692,
        -2.513274,2.353029,-1.310529,
        -2.513274,2.354868,0.211543,
        -2.513274,2.355120,0.388593,
        -2.513274,2.355165,-1.833173,
        -2.513274,2.355462,2.151441,
        -2.513274,2.362948,-2.487477,
        -2.513274,2.363101,-2.810456,
        -2.513274,2.363394,-0.272918,
        -2.513274,2.364509,-1.644533,
        -2.513274,2.370465,2.466824,
        -2.513274,2.386110,-1.481351,
        -2.513274,2.394960,0.862133,
        -2.513274,2.395716,1.977184,
        -2.513274,2.396339,1.200900,
        -2.513274,2.407941,-2.316753,
        -2.513274,2.409450,-2.988432,
        -2.513274,2.417202,-0.103363,
        -2.513274,2.417386,-0.443097,
        -2.513274,2.417950,2.666662,
        -2.513274,2.423190,1.789783,
        -2.513274,2.423692,1.392696,
        -2.513274,2.431122,3.086379,
        -2.513274,2.431406,-2.647820,
        -2.513274,2.431584,2.307631,
        -2.513274,2.434090,1.592818,
        -2.513274,2.435993,0.683844,
        -2.513274,2.440898,-2.133300,
        -2.513274,2.442222,-1.195837,
        -2.513274,2.443112,2.883761,
        -2.513274,2.449191,-0.633651,
        -2.513274,2.450821,0.084741,
        -2.513274,2.453499,-1.008861,
        -2.513274,2.460010,-1.938011,
        -2.513274,2.460199,-0.822008,
        -2.513274,2.460902,0.490079,
        -2.513274,2.461774,-1.734913,
        -2.513274,2.463986,1.034407,
        -2.513274,2.467647,0.287667,
        -2.513274,2.479849,2.126925,
        -2.513274,2.485907,-2.835471,
        -2.513274,2.485945,-1.374853,
        -2.513274,2.487747,2.505912,
        -2.513274,2.487957,-0.274279,
        -2.513274,2.489741,-2.468428,
        -2.513274,2.512538,1.239900,
        -2.513274,2.517528,1.923289,
        -2.513274,2.518454,-1.565852,
        -2.513274,2.519348,0.843503,
        -2.513274,2.522766,-3.059231,
        -2.513274,2.528686,2.737228,
        -2.513274,2.534798,-2.266671,
        -2.513274,2.536533,1.466274,
        -2.513274,2.538504,1.699812,
        -2.513274,2.539516,-0.476942,
        -2.513274,2.539532,-0.074548,
        -2.513274,2.543197,2.990252,
        -2.513274,2.554022,2.320619,
        -2.513274,2.556919,-2.655005,
        -2.513274,2.558146,0.627191,
        -2.513274,2.558223,-1.164189,
        -2.513274,2.563452,-2.044028,
        -2.513274,2.568462,0.150655,
        -2.513274,2.569130,-0.700837,
        -2.513274,2.570050,-1.804492,
        -2.513274,2.572710,-0.934665,
        -2.513274,2.575994,0.391718,
        -2.513274,2.584774,1.046703,
        -2.513274,2.601432,2.102535,
        -2.513274,2.603559,2.555114,
        -2.513274,2.604150,-2.885208,
        -2.513274,2.608630,-1.378011,
        -2.513274,2.611256,-0.279110,
        -2.513274,2.617344,-2.439480,
        -2.513274,2.627402,1.293552,
        -2.513274,2.635396,1.848850,
        -2.513274,2.635686,2.834950,
        -2.513274,2.637087,3.132730,
        -2.513274,2.642029,0.813274,
        -2.513274,2.645260,1.570203,
        -2.513274,2.646182,-1.614036,
        -2.513274,2.660113,-2.186513,
        -2.513274,2.660483,-0.034608,
        -2.513274,2.661713,-0.529511,
        -2.513274,2.672720,-1.110422,
        -2.513274,2.676626,2.335353,
        -2.513274,2.676638,0.540364,
        -2.513274,2.676780,-1.901158,
        -2.513274,2.681411,-2.667965,
        -2.513274,2.682152,0.245751,
        -2.513274,2.682859,-0.817071,
        -2.513274,2.705107,1.062142,
        -2.513274,2.717705,2.625995,
        -2.513274,2.721001,2.060097,
        -2.513274,2.722319,-2.954069,
        -2.513274,2.729840,-1.368237,
        -2.513274,2.734349,-0.289298,
        -2.513274,2.738409,2.975671,
        -2.513274,2.740366,1.374341,
        -2.513274,2.744633,-2.391311,
        -2.513274,2.747398,1.727166,
        -2.513274,2.763799,0.760372,
        -2.513274,2.768238,-1.679243,
        -2.513274,2.779245,-2.046268,
        -2.513274,2.780208,-0.626453,
        -2.513274,2.780471,0.027732,
        -2.513274,2.784696,-1.013128,
        -2.513274,2.789817,0.395431,
        -2.513274,2.801154,2.353143,
        -2.513274,2.807091,-2.687477,
        -2.513274,2.825762,1.085035,
        -2.513274,2.830209,2.734943,
        -2.513274,2.833517,-3.085508,
        -2.513274,2.838301,1.980061,
        -2.513274,2.849871,1.512012,
        -2.513274,2.850089,-1.343899,
        -2.513274,2.856905,-0.318441,
        -2.513274,2.870445,-2.291001,
        -2.513274,2.885017,0.655562,
        -2.513274,2.886428,-1.793217,
        -2.513274,2.890337,-0.825633,
        -2.513274,2.898932,0.139973,
        -2.513274,2.931927,-2.728768,
        -2.513274,2.932096,2.377790,
        -2.513274,2.937409,2.947245,
        -2.513274,2.948048,1.133127,
        -2.513274,2.950896,1.797889,
        -2.513274,2.970019,-1.282382,
        -2.513274,2.978333,-0.412090,
        -2.513274,2.996437,-2.089121,
        -2.513274,3.011274,0.434110,
        -2.513274,3.050410,3.138492,
        -2.513274,3.064492,1.647259,
        -2.513274,3.088868,-0.989501,
        -2.408554,0.053884,-2.408394,
        -2.408554,0.077645,1.453433,
        -2.408554,0.088408,-0.311160,
        -2.408554,0.142314,2.590387,
        -2.408554,0.147206,-1.385164,
        -2.408554,0.163506,0.573203,
        -2.408554,0.165485,-2.875396,
        -2.408554,0.182564,-2.125821,
        -2.408554,0.190569,1.929141,
        -2.408554,0.201102,1.249082,
        -2.408554,0.201397,-0.716646,
        -2.408554,0.210568,-0.069559,
        -2.408554,0.251672,2.848900,
        -2.408554,0.266262,-1.678068,
        -2.408554,0.269108,-1.215563,
        -2.408554,0.273777,-2.506756,
        -2.408554,0.279854,0.814581,
        -2.408554,0.281664,-2.969147,
        -2.408554,0.283625,2.405603,
        -2.408554,0.284058,0.357350,
        -2.408554,0.294107,1.593880,
        -2.408554,0.302491,-0.400410,
        -2.408554,0.308059,2.021491,
        -2.408554,0.312219,-2.072057,
        -2.408554,0.321905,-0.829121,
        -2.408554,0.329984,1.201621,
        -2.408554,0.339714,-0.013712,
        -2.408554,0.368784,2.974974,
        -2.408554,0.376428,-1.495532,
        -2.408554,0.378103,2.646436,
        -2.408554,0.385088,0.590355,
        -2.408554,0.386248,-2.690943,
        -2.408554,0.389851,-2.340031,
        -2.408554,0.391502,-1.171374,
        -2.408554,0.398249,-1.812694,
        -2.408554,0.400542,-2.995729,
        -2.408554,0.404273,0.918163,
        -2.408554,0.405617,1.773533,
        -2.408554,0.409009,-0.570243,
        -2.408554,0.410700,0.270433,
        -2.408554,0.415893,1.464890,
        -2.408554,0.420680,2.088164,
        -2.408554,0.424766,-0.258729,
        -2.408554,0.426507,2.367872,
        -2.408554,0.441292,-0.890727,
        -2.408554,0.451839,-2.078193,
        -2.408554,0.459270,1.191562,
        -2.408554,0.470622,0.014002,
        -2.408554,0.481292,-2.529098,
        -2.408554,0.482460,2.849843,
        -2.408554,0.492205,-1.397963,
        -2.408554,0.493314,3.117264,
        -2.408554,0.497870,0.727828,
        -2.408554,0.499783,-1.654379,
        -2.408554,0.502651,0.463333,
        -2.408554,0.505463,2.594557,
        -2.408554,0.506104,-2.874910,
        -2.408554,0.513579,-1.147354,
        -2.408554,0.518569,1.644237,
        -2.408554,0.519001,1.900488,
        -2.408554,0.519277,-0.673809,
        -2.408554,0.524769,-0.421955,
        -2.408554,0.525564,-2.288838,
        -2.408554,0.532541,0.978737,
        -2.408554,0.534610,2.154808,
        -2.408554,0.535020,-1.891846,
        -2.408554,0.540208,1.400497,
        -2.408554,0.542006,0.228337,
        -2.408554,0.550762,-0.180771,
        -2.408554,0.560531,2.378474,
        -2.408554,0.561629,-0.924576,
        -2.408554,0.567465,-2.679392,
        -2.408554,0.588527,1.181755,
        -2.408554,0.591643,-2.098364,
        -2.408554,0.594963,2.997575,
        -2.408554,0.599643,-3.045235,
        -2.408554,0.599962,2.773556,
        -2.408554,0.602620,0.029270,
        -2.408554,0.605711,0.598454,
        -2.408554,0.607708,-1.541026,
        -2.408554,0.609360,-2.464379,
        -2.408554,0.610738,-1.326786,
        -2.408554,0.616968,0.820489,
        -2.408554,0.626379,1.773610,
        -2.408554,0.628626,-1.751180,
        -2.408554,0.630671,0.391949,
        -2.408554,0.632622,-0.737642,
        -2.408554,0.632708,2.566722,
        -2.408554,0.632797,1.989858,
        -2.408554,0.634455,-0.534790,
        -2.408554,0.636205,1.561170,
        -2.408554,0.637681,-1.123423,
        -2.408554,0.641441,-2.842109,
        -2.408554,0.644807,-0.327139,
        -2.408554,0.652440,2.197129,
        -2.408554,0.660897,-2.278593,
        -2.408554,0.664008,1.360091,
        -2.408554,0.670850,-1.945257,
        -2.408554,0.671396,1.004859,
        -2.408554,0.673985,0.203855,
        -2.408554,0.680326,-0.133370,
        -2.408554,0.685083,-0.942363,
        -2.408554,0.688204,-2.632204,
        -2.408554,0.689777,2.385011,
        -2.408554,0.700065,3.115430,
        -2.408554,0.704080,2.911636,
        -2.408554,0.712883,1.174216,
        -2.408554,0.715701,0.699906,
        -2.408554,0.718851,-1.442215,
        -2.408554,0.721019,2.721536,
        -2.408554,0.725300,-2.975134,
        -2.408554,0.727136,-1.633288,
        -2.408554,0.727399,-2.120567,
        -2.408554,0.730751,0.522488,
        -2.408554,0.732892,-1.254330,
        -2.408554,0.734333,-2.446184,
        -2.408554,0.735125,0.037270,
        -2.408554,0.737893,1.866463,
        -2.408554,0.740650,1.681075,
        -2.408554,0.745971,-0.781286,
        -2.408554,0.749432,-0.442559,
        -2.408554,0.750102,2.050636,
        -2.408554,0.751046,0.863821,
        -2.408554,0.757018,-0.608833,
        -2.408554,0.757046,-1.816019,
        -2.408554,0.757389,1.501184,
        -2.408554,0.759800,2.546181,
        -2.408554,0.759859,0.351923,
        -2.408554,0.761141,-2.781102,
        -2.408554,0.769035,-1.078574,
        -2.408554,0.770166,-0.266604,
        -2.408554,0.772218,2.227380,
        -2.408554,0.788382,1.331051,
        -2.408554,0.790269,-2.282409,
        -2.408554,0.803410,-1.984567,
        -2.408554,0.803813,0.191520,
        -2.408554,0.806832,3.022155,
        -2.408554,0.807476,-2.597675,
        -2.408554,0.812512,-0.107335,
        -2.408554,0.814161,1.010949,
        -2.408554,0.817732,-3.084486,
        -2.408554,0.817846,2.390688,
        -2.408554,0.819323,2.845960,
        -2.408554,0.819430,-0.915017,
        -2.408554,0.829198,-1.532491,
        -2.408554,0.829531,0.665695,
        -2.408554,0.829816,1.167147,
        -2.408554,0.833613,-1.357492,
        -2.408554,0.842945,-2.907318,
        -2.408554,0.844883,2.679708,
        -2.408554,0.848189,-1.705003,
        -2.408554,0.849888,1.767901,
        -2.408554,0.851126,0.493370,
        -2.408554,0.852911,1.933394,
        -2.408554,0.857516,-2.432863,
        -2.408554,0.858261,-1.191729,
        -2.408554,0.858279,-2.138895,
        -2.408554,0.859723,-0.734928,
        -2.408554,0.859754,1.603292,
        -2.408554,0.861655,-0.394821,
        -2.408554,0.864300,0.043549,
        -2.408554,0.866940,0.851229,
        -2.408554,0.869748,2.095969,
        -2.408554,0.877352,-0.574668,
        -2.408554,0.879254,-2.733338,
        -2.408554,0.882554,1.443757,
        -2.408554,0.884060,0.334199,
        -2.408554,0.885626,-1.866061,
        -2.408554,0.889834,2.529957,
        -2.408554,0.893140,2.251698,
        -2.408554,0.897463,-0.240822,
        -2.408554,0.898758,-1.031948,
        -2.408554,0.913918,1.287618,
        -2.408554,0.916555,-2.287173,
        -2.408554,0.918305,3.107845,
        -2.408554,0.919522,2.948277,
        -2.408554,0.927138,-2.569686,
        -2.408554,0.929031,0.185383,
        -2.408554,0.933766,-3.015917,
        -2.408554,0.934873,-2.010822,
        -2.408554,0.935636,-1.445855,
        -2.408554,0.936702,-0.866461,
        -2.408554,0.937395,0.729108,
        -2.408554,0.937910,2.787575,
        -2.408554,0.939144,-0.092962,
        -2.408554,0.941718,1.125085,
        -2.408554,0.942873,-1.607680,
        -2.408554,0.943701,0.980328,
        -2.408554,0.947774,2.395256,
        -2.408554,0.952433,0.584347,
        -2.408554,0.954456,-1.287293,
        -2.408554,0.958731,-2.852552,
        -2.408554,0.961838,1.839284,
        -2.408554,0.961898,-0.469425,
        -2.408554,0.965426,1.686615,
        -2.408554,0.970108,1.990771,
        -2.408554,0.972820,-1.762780,
        -2.408554,0.973500,-0.711881,
        -2.408554,0.975253,2.641082,
        -2.408554,0.977824,0.441928,
        -2.408554,0.981485,-2.424766,
        -2.408554,0.981877,1.537100,
        -2.408554,0.982609,-1.136920,
        -2.408554,0.984265,-2.156073,
        -2.408554,0.989150,0.045493,
        -2.408554,0.993052,2.137478,
        -2.408554,0.997766,-2.693871,
        -2.408554,0.999847,-0.334738,
        -2.408554,1.001920,0.848054,
        -2.408554,1.006146,1.389734,
        -2.408554,1.012180,0.304066,
        -2.408554,1.016938,2.270674,
        -2.408554,1.017371,-0.979790,
        -2.408554,1.019495,-1.894487,
        -2.408554,1.023796,-0.583691,
        -2.408554,1.026525,2.506945,
        -2.408554,1.027004,-0.203653,
        -2.408554,1.028027,3.036573,
        -2.408554,1.031435,1.241021,
        -2.408554,1.032316,2.884221,
        -2.408554,1.035863,-3.110475,
        -2.408554,1.042085,-2.292950,
        -2.408554,1.043272,-1.522986,
        -2.408554,1.046265,-2.958847,
        -2.408554,1.049272,-1.376344,
        -2.408554,1.050039,-2.552086,
        -2.408554,1.053682,0.689863,
        -2.408554,1.055458,0.170394,
        -2.408554,1.056622,-0.831866,
        -2.408554,1.057515,1.097998,
        -2.408554,1.058706,-2.036438,
        -2.408554,1.061389,-1.672248,
        -2.408554,1.063758,2.743093,
        -2.408554,1.070157,-0.073422,
        -2.408554,1.072154,1.761506,
        -2.408554,1.074029,1.909193,
        -2.408554,1.074204,0.545276,
        -2.408554,1.074714,-2.805147,
        -2.408554,1.076262,0.966104,
        -2.408554,1.077642,-1.227073,
        -2.408554,1.084151,1.618525,
        -2.408554,1.085085,2.377673,
        -2.408554,1.086822,-0.441112,
        -2.408554,1.087142,2.058815,
        -2.408554,1.098383,-0.696430,
        -2.408554,1.099138,-1.084033,
        -2.408554,1.103208,1.477773,
        -2.408554,1.103783,0.408254,
        -2.408554,1.106275,-2.420800,
        -2.408554,1.108079,2.612943,
        -2.408554,1.109345,-2.172718,
        -2.408554,1.115957,-1.787597,
        -2.408554,1.118930,0.045983,
        -2.408554,1.119777,0.834702,
        -2.408554,1.121943,-2.669042,
        -2.408554,1.124930,2.219917,
        -2.408554,1.125217,1.336294,
        -2.408554,1.130958,2.972931,
        -2.408554,1.131400,-0.306629,
        -2.408554,1.137135,-0.936514,
        -2.408554,1.139125,-1.924528,
        -2.408554,1.139877,0.276382,
        -2.408554,1.142003,-3.058022,
        -2.408554,1.147363,-0.558448,
        -2.408554,1.150348,-0.176418,
        -2.408554,1.151198,1.197016,
        -2.408554,1.153930,-1.597104,
        -2.408554,1.155389,2.837148,
        -2.408554,1.155458,-1.460623,
        -2.408554,1.157331,-2.906362,
        -2.408554,1.158594,2.487618,
        -2.408554,1.162103,-1.322553,
        -2.408554,1.165407,3.092861,
        -2.408554,1.168025,0.717075,
        -2.408554,1.168406,-2.301519,
        -2.408554,1.174782,-2.538755,
        -2.408554,1.177998,1.061256,
        -2.408554,1.178898,1.827305,
        -2.408554,1.181331,-2.060330,
        -2.408554,1.182540,-0.798952,
        -2.408554,1.184569,0.147517,
        -2.408554,1.184734,1.968919,
        -2.408554,1.184804,0.593386,
        -2.408554,1.188278,1.690808,
        -2.408554,1.194360,2.110440,
        -2.408554,1.194776,2.710455,
        -2.408554,1.197745,-1.173705,
        -2.408554,1.199898,-2.777257,
        -2.408554,1.201164,2.355931,
        -2.408554,1.203760,1.556404,
        -2.408554,1.205845,-0.415993,
        -2.408554,1.206585,0.933648,
        -2.408554,1.206950,-0.037248,
        -2.408554,1.209341,0.471162,
        -2.408554,1.213483,-1.039417,
        -2.408554,1.218146,-0.672643,
        -2.408554,1.221608,1.421669,
        -2.408554,1.226043,-1.693533,
        -2.408554,1.232870,-1.822773,
        -2.408554,1.234375,-2.421194,
        -2.408554,1.234531,-2.190148,
        -2.408554,1.239414,0.348959,
        -2.408554,1.240183,2.587255,
        -2.408554,1.242312,2.941234,
        -2.408554,1.244075,1.286644,
        -2.408554,1.245961,-2.997461,
        -2.408554,1.246495,-2.646872,
        -2.408554,1.247736,2.242810,
        -2.408554,1.253604,0.812222,
        -2.408554,1.257816,-0.894317,
        -2.408554,1.258247,-1.953634,
        -2.408554,1.258993,-0.284099,
        -2.408554,1.262380,-0.157609,
        -2.408554,1.262791,-1.549229,
        -2.408554,1.263772,-0.533073,
        -2.408554,1.266731,-1.412461,
        -2.408554,1.268591,-3.122146,
        -2.408554,1.269614,-1.275246,
        -2.408554,1.270745,1.152533,
        -2.408554,1.273606,0.224545,
        -2.408554,1.279716,2.812101,
        -2.408554,1.283786,2.462124,
        -2.408554,1.284798,0.078209,
        -2.408554,1.285402,-2.882049,
        -2.408554,1.286936,1.886053,
        -2.408554,1.289408,0.683122,
        -2.408554,1.294736,1.756184,
        -2.408554,1.296978,-2.311511,
        -2.408554,1.298546,2.020601,
        -2.408554,1.300924,1.022478,
        -2.408554,1.303305,3.043152,
        -2.408554,1.304210,-2.084225,
        -2.408554,1.305241,-2.530249,
        -2.408554,1.307513,1.628003,
        -2.408554,1.310722,0.558267,
        -2.408554,1.314142,2.139329,
        -2.408554,1.316087,-0.777971,
        -2.408554,1.317907,-1.121023,
        -2.408554,1.321019,-2.749776,
        -2.408554,1.321609,2.687717,
        -2.408554,1.322568,1.499415,
        -2.408554,1.322972,-0.395699,
        -2.408554,1.325286,-0.994192,
        -2.408554,1.329146,-0.648513,
        -2.408554,1.329795,-0.042570,
        -2.408554,1.331651,2.344497,
        -2.408554,1.336522,0.435098,
        -2.408554,1.339410,0.900087,
        -2.408554,1.339550,1.370353,
        -2.408554,1.340441,-1.731580,
        -2.408554,1.351855,-1.855234,
        -2.408554,1.360161,-3.018404,
        -2.408554,1.361031,-1.613971,
        -2.408554,1.361196,2.914780,
        -2.408554,1.361998,-2.207254,
        -2.408554,1.362302,1.239384,
        -2.408554,1.365228,2.563808,
        -2.408554,1.365452,0.310522,
        -2.408554,1.367028,-2.420041,
        -2.408554,1.373422,-1.484281,
        -2.408554,1.374245,-1.226417,
        -2.408554,1.375020,-0.170393,
        -2.408554,1.376560,-1.355481,
        -2.408554,1.377070,-1.980984,
        -2.408554,1.377578,-2.636339,
        -2.408554,1.378096,0.774626,
        -2.408554,1.379086,-0.508192,
        -2.408554,1.382562,0.177141,
        -2.408554,1.388849,2.233847,
        -2.408554,1.389268,-0.886349,
        -2.408554,1.391324,1.109396,
        -2.408554,1.391699,-0.288725,
        -2.408554,1.392869,3.136903,
        -2.408554,1.393612,1.952016,
        -2.408554,1.395545,-2.850807,
        -2.408554,1.401926,1.822930,
        -2.408554,1.402149,2.788898,
        -2.408554,1.409552,0.647980,
        -2.408554,1.410840,0.056219,
        -2.408554,1.411944,2.444377,
        -2.408554,1.413639,1.697982,
        -2.408554,1.426479,1.573427,
        -2.408554,1.426725,2.102242,
        -2.408554,1.427868,0.984637,
        -2.408554,1.429580,-2.107232,
        -2.408554,1.431375,-2.313269,
        -2.408554,1.431907,3.015410,
        -2.408554,1.432325,-1.029054,
        -2.408554,1.433341,0.524268,
        -2.408554,1.435181,-0.739771,
        -2.408554,1.437208,-2.529085,
        -2.408554,1.439041,-0.617966,
        -2.408554,1.441473,1.448544,
        -2.408554,1.444527,2.664744,
        -2.408554,1.450513,-0.400625,
        -2.408554,1.451907,-0.066823,
        -2.408554,1.452754,-2.738631,
        -2.408554,1.454061,-2.950416,
        -2.408554,1.456736,1.323154,
        -2.408554,1.458370,0.400433,
        -2.408554,1.459314,-1.765003,
        -2.408554,1.459902,-1.145835,
        -2.408554,1.464952,2.329245,
        -2.408554,1.466050,0.862104,
        -2.408554,1.471330,-1.886160,
        -2.408554,1.475119,-1.648255,
        -2.408554,1.480878,1.194576,
        -2.408554,1.480942,0.273235,
        -2.408554,1.481086,2.890269,
        -2.408554,1.484118,-1.529156,
        -2.408554,1.484741,-1.279219,
        -2.408554,1.487681,-1.406069,
        -2.408554,1.487886,-3.066037,
        -2.408554,1.490555,2.543284,
        -2.408554,1.492424,-0.190932,
        -2.408554,1.495002,-0.839810,
        -2.408554,1.495297,2.007188,
        -2.408554,1.496414,-2.009267,
        -2.408554,1.498261,-2.425270,
        -2.408554,1.498907,-2.207327,
        -2.408554,1.500204,0.737638,
        -2.408554,1.502469,0.151156,
        -2.408554,1.507448,-0.514782,
        -2.408554,1.509287,2.204647,
        -2.408554,1.510518,1.884920,
        -2.408554,1.511948,-2.632360,
        -2.408554,1.515235,1.068565,
        -2.408554,1.522237,3.102018,
        -2.408554,1.522373,1.764991,
        -2.408554,1.522416,-0.954956,
        -2.408554,1.523102,2.765920,
        -2.408554,1.524811,-2.844718,
        -2.408554,1.527261,-0.304916,
        -2.408554,1.529788,0.613208,
        -2.408554,1.533648,1.644808,
        -2.408554,1.535018,0.028869,
        -2.408554,1.541082,2.424608,
        -2.408554,1.546010,1.523462,
        -2.408554,1.551135,-0.642994,
        -2.408554,1.553895,0.489607,
        -2.408554,1.555041,0.947813,
        -2.408554,1.559671,-1.074934,
        -2.408554,1.560345,2.985899,
        -2.408554,1.560487,-2.323400,
        -2.408554,1.561127,1.401111,
        -2.408554,1.567193,2.643172,
        -2.408554,1.568892,2.098362,
        -2.408554,1.571702,-2.098435,
        -2.408554,1.572701,-2.528730,
        -2.408554,1.573572,1.278593,
        -2.408554,1.574849,-0.091505,
        -2.408554,1.576334,0.365236,
        -2.408554,1.577029,-2.962888,
        -2.408554,1.577688,-1.814941,
        -2.408554,1.579542,-0.755917,
        -2.408554,1.579624,-1.199759,
        -2.408554,1.581515,-0.419414,
        -2.408554,1.584581,-2.738702,
        -2.408554,1.586948,-1.696296,
        -2.408554,1.588901,2.301496,
        -2.408554,1.589936,0.825020,
        -2.408554,1.590018,-1.941652,
        -2.408554,1.596968,-1.328358,
        -2.408554,1.597178,-1.575607,
        -2.408554,1.600470,-1.453196,
        -2.408554,1.601408,0.242409,
        -2.408554,1.602259,1.152656,
        -2.408554,1.603064,2.865774,
        -2.408554,1.612085,1.968802,
        -2.408554,1.612720,-3.080498,
        -2.408554,1.616061,2.521531,
        -2.408554,1.616900,-0.207457,
        -2.408554,1.620988,-0.875920,
        -2.408554,1.622114,0.701509,
        -2.408554,1.622749,0.121426,
        -2.408554,1.622993,-2.221629,
        -2.408554,1.629346,-0.542669,
        -2.408554,1.630007,1.844552,
        -2.408554,1.634747,-2.426198,
        -2.408554,1.642346,1.722617,
        -2.408554,1.643925,1.034629,
        -2.408554,1.644848,2.743774,
        -2.408554,1.645258,2.188109,
        -2.408554,1.645417,-2.634725,
        -2.408554,1.647395,3.085437,
        -2.408554,1.648538,-2.855044,
        -2.408554,1.649932,0.576808,
        -2.408554,1.652224,-0.996105,
        -2.408554,1.652871,1.600681,
        -2.408554,1.657997,-0.325127,
        -2.408554,1.661936,0.001037,
        -2.408554,1.665187,1.477402,
        -2.408554,1.665496,2.397872,
        -2.408554,1.673267,-0.665453,
        -2.408554,1.673557,0.452526,
        -2.408554,1.679277,-1.122421,
        -2.408554,1.679696,0.912482,
        -2.408554,1.682420,1.353922,
        -2.408554,1.685031,-2.009038,
        -2.408554,1.685196,2.967046,
        -2.408554,1.690107,2.620910,
        -2.408554,1.690453,-2.123076,
        -2.408554,1.691693,-1.871572,
        -2.408554,1.691817,2.061688,
        -2.408554,1.692644,1.237670,
        -2.408554,1.696386,0.328123,
        -2.408554,1.697162,-1.249026,
        -2.408554,1.697488,-2.322785,
        -2.408554,1.698726,-2.973576,
        -2.408554,1.699172,-1.749717,
        -2.408554,1.704478,-0.106295,
        -2.408554,1.707244,-0.444770,
        -2.408554,1.707500,-2.531497,
        -2.408554,1.707714,-0.787321,
        -2.408554,1.708430,-1.627552,
        -2.408554,1.710583,-1.376890,
        -2.408554,1.712415,-2.747921,
        -2.408554,1.713125,-1.503073,
        -2.408554,1.713452,0.789419,
        -2.408554,1.720612,2.279888,
        -2.408554,1.724852,0.202831,
        -2.408554,1.724868,2.846246,
        -2.408554,1.728940,1.931535,
        -2.408554,1.732838,1.127584,
        -2.408554,1.735719,-3.092485,
        -2.408554,1.739676,2.496017,
        -2.408554,1.739971,-0.228876,
        -2.408554,1.743328,0.663347,
        -2.408554,1.744927,-0.911892,
        -2.408554,1.745952,0.084733,
        -2.408554,1.748077,1.804354,
        -2.408554,1.752959,-0.567548,
        -2.408554,1.760169,1.678132,
        -2.408554,1.763887,-2.218450,
        -2.408554,1.766737,2.722752,
        -2.408554,1.767999,1.004737,
        -2.408554,1.769149,1.550298,
        -2.408554,1.770119,-2.427175,
        -2.408554,1.770460,2.154208,
        -2.408554,1.770514,3.070982,
        -2.408554,1.770735,0.537245,
        -2.408554,1.772673,-2.864760,
        -2.408554,1.773527,-1.039582,
        -2.408554,1.776419,-2.642421,
        -2.408554,1.780769,1.416330,
        -2.408554,1.786634,-0.346247,
        -2.408554,1.788829,-1.947553,
        -2.408554,1.793737,0.410270,
        -2.408554,1.794553,2.374618,
        -2.408554,1.794699,-0.694061,
        -2.408554,1.797469,-1.169105,
        -2.408554,1.803705,0.881076,
        -2.408554,1.804124,1.264249,
        -2.408554,1.807825,-2.090455,
        -2.408554,1.808822,-1.817375,
        -2.408554,1.809020,2.949741,
        -2.408554,1.812466,2.597595,
        -2.408554,1.813159,2.025395,
        -2.408554,1.813816,-1.298435,
        -2.408554,1.815016,0.278951,
        -2.408554,1.816510,-0.008631,
        -2.408554,1.817772,-1.689625,
        -2.408554,1.821360,-2.984352,
        -2.408554,1.822312,-0.127211,
        -2.408554,1.823386,-1.429807,
        -2.408554,1.823834,-1.560483,
        -2.408554,1.832060,-0.821627,
        -2.408554,1.832690,-0.468440,
        -2.408554,1.833945,-2.320278,
        -2.408554,1.834787,0.751968,
        -2.408554,1.840579,-2.536445,
        -2.408554,1.841033,-2.756932,
        -2.408554,1.846350,1.893983,
        -2.408554,1.846644,1.113766,
        -2.408554,1.846833,2.826543,
        -2.408554,1.847325,2.249051,
        -2.408554,1.854119,0.134161,
        -2.408554,1.858863,-3.106022,
        -2.408554,1.865046,1.763384,
        -2.408554,1.865619,-0.952036,
        -2.408554,1.865912,0.624453,
        -2.408554,1.866694,2.473686,
        -2.408554,1.867524,-0.240083,
        -2.408554,1.875561,1.632831,
        -2.408554,1.876049,-0.594378,
        -2.408554,1.880509,1.500204,
        -2.408554,1.883504,1.360457,
        -2.408554,1.886971,-2.196665,
        -2.408554,1.888979,0.982922,
        -2.408554,1.889815,2.701570,
        -2.408554,1.892144,-2.005340,
        -2.408554,1.892348,0.496522,
        -2.408554,1.893281,-1.083976,
        -2.408554,1.893828,3.053453,
        -2.408554,1.895117,2.121107,
        -2.408554,1.898741,-2.875735,
        -2.408554,1.904429,-2.428029,
        -2.408554,1.908408,-2.648903,
        -2.408554,1.911828,-0.364489,
        -2.408554,1.912224,0.366565,
        -2.408554,1.914854,1.214468,
        -2.408554,1.915467,-1.883143,
        -2.408554,1.915763,-0.724679,
        -2.408554,1.915790,-1.216366,
        -2.408554,1.922256,2.347806,
        -2.408554,1.922616,0.844633,
        -2.408554,1.925185,0.231170,
        -2.408554,1.927474,-1.754616,
        -2.408554,1.928342,-1.351174,
        -2.408554,1.933328,-1.623328,
        -2.408554,1.933878,0.034150,
        -2.408554,1.934585,-1.487979,
        -2.408554,1.934802,2.927634,
        -2.408554,1.934862,1.989776,
        -2.408554,1.936327,-0.099350,
        -2.408554,1.938088,2.578507,
        -2.408554,1.945677,-2.999805,
        -2.408554,1.954572,-0.859208,
        -2.408554,1.957868,-0.490784,
        -2.408554,1.958989,0.712772,
        -2.408554,1.960212,-2.307561,
        -2.408554,1.964774,1.855507,
        -2.408554,1.964971,-2.100533,
        -2.408554,1.969112,1.088273,
        -2.408554,1.971438,-2.765415,
        -2.408554,1.972312,2.805134,
        -2.408554,1.974950,2.220596,
        -2.408554,1.975135,-2.539127,
        -2.408554,1.982107,1.720829,
        -2.408554,1.983195,-3.127984,
        -2.408554,1.983318,-0.247182,
        -2.408554,1.986720,-0.993220,
        -2.408554,1.990381,1.586005,
        -2.408554,1.990470,0.580283,
        -2.408554,1.990523,1.450326,
        -2.408554,1.993909,1.312897,
        -2.408554,1.994694,2.452813,
        -2.408554,1.999226,-0.620717,
        -2.408554,2.005145,0.941012,
        -2.408554,2.007301,0.142337,
        -2.408554,2.014149,0.444138,
        -2.408554,2.014338,-1.127392,
        -2.408554,2.018131,3.021778,
        -2.408554,2.019676,-1.980944,
        -2.408554,2.019912,2.686612,
        -2.408554,2.020545,2.088672,
        -2.408554,2.026074,0.300945,
        -2.408554,2.026196,-2.889057,
        -2.408554,2.031959,-1.268178,
        -2.408554,2.032917,-2.199917,
        -2.408554,2.033572,-1.842884,
        -2.408554,2.034712,-2.419505,
        -2.408554,2.035818,-0.379479,
        -2.408554,2.036004,-0.757062,
        -2.408554,2.040501,-1.411488,
        -2.408554,2.042874,-1.701367,
        -2.408554,2.043060,-2.653746,
        -2.408554,2.043729,-1.557938,
        -2.408554,2.048538,0.806623,
        -2.408554,2.049006,1.199930,
        -2.408554,2.049502,-0.137989,
        -2.408554,2.052181,2.326959,
        -2.408554,2.052794,-0.008526,
        -2.408554,2.057608,1.951726,
        -2.408554,2.065568,2.567730,
        -2.408554,2.070371,2.879808,
        -2.408554,2.071256,-3.018961,
        -2.408554,2.079746,-0.897699,
        -2.408554,2.082784,1.042403,
        -2.408554,2.083752,-0.511086,
        -2.408554,2.084401,1.810502,
        -2.408554,2.085731,0.671081,
        -2.408554,2.097496,1.373172,
        -2.408554,2.099185,1.667006,
        -2.408554,2.100482,-2.301084,
        -2.408554,2.103118,2.194745,
        -2.408554,2.103223,1.521685,
        -2.408554,2.103695,-2.777707,
        -2.408554,2.107523,3.129581,
        -2.408554,2.108277,-2.095907,
        -2.408554,2.108891,-2.532166,
        -2.408554,2.111003,-0.268632,
        -2.408554,2.112276,-1.027246,
        -2.408554,2.114380,0.531457,
        -2.408554,2.119166,0.100552,
        -2.408554,2.119564,2.754818,
        -2.408554,2.122474,-0.648314,
        -2.408554,2.125183,2.445598,
        -2.408554,2.129501,0.238290,
        -2.408554,2.131178,0.386663,
        -2.408554,2.131223,-1.174311,
        -2.408554,2.134269,-1.948042,
        -2.408554,2.134684,0.907065,
        -2.408554,2.137615,2.988806,
        -2.408554,2.145635,-1.327708,
        -2.408554,2.146467,2.055081,
        -2.408554,2.146582,-1.793788,
        -2.408554,2.149210,-1.482321,
        -2.408554,2.152901,-1.636788,
        -2.408554,2.156337,1.151558,
        -2.408554,2.156474,-2.912065,
        -2.408554,2.157194,-0.792563,
        -2.408554,2.166724,-0.401638,
        -2.408554,2.166980,-0.134879,
        -2.408554,2.173644,-2.403429,
        -2.408554,2.175711,-2.655523,
        -2.408554,2.177132,2.600419,
        -2.408554,2.178879,0.767722,
        -2.408554,2.180752,1.908485,
        -2.408554,2.181338,2.313721,
        -2.408554,2.185580,1.287673,
        -2.408554,2.195400,-2.217546,
        -2.408554,2.200418,-3.056442,
        -2.408554,2.201829,-0.001735,
        -2.408554,2.203804,1.756176,
        -2.408554,2.211602,1.443930,
        -2.408554,2.212686,-0.544809,
        -2.408554,2.213937,0.621892,
        -2.408554,2.214075,1.600691,
        -2.408554,2.216256,1.020029,
        -2.408554,2.219060,2.886631,
        -2.408554,2.223375,-0.906185,
        -2.408554,2.226002,-1.060248,
        -2.408554,2.226413,-2.064034,
        -2.408554,2.230554,2.171334,
        -2.408554,2.232413,2.735777,
        -2.408554,2.233709,3.070830,
        -2.408554,2.234383,-2.792065,
        -2.408554,2.237183,0.468799,
        -2.408554,2.237695,-0.271388,
        -2.408554,2.239419,0.151693,
        -2.408554,2.243347,-2.521049,
        -2.408554,2.244035,-1.229456,
        -2.408554,2.244355,0.309899,
        -2.408554,2.245175,2.460236,
        -2.408554,2.245628,-1.899216,
        -2.408554,2.245973,-0.702234,
        -2.408554,2.255601,-1.728131,
        -2.408554,2.256702,-1.401367,
        -2.408554,2.261466,-1.557500,
        -2.408554,2.267390,0.878499,
        -2.408554,2.272099,2.018594,
        -2.408554,2.278618,1.170386,
        -2.408554,2.283304,-2.345584,
        -2.408554,2.285175,-2.943090,
        -2.408554,2.293000,-0.124451,
        -2.408554,2.294058,-0.421070,
        -2.408554,2.303283,1.855938,
        -2.408554,2.304968,2.608978,
        -2.408554,2.305347,2.311346,
        -2.408554,2.307188,-2.656134,
        -2.408554,2.307301,1.336200,
        -2.408554,2.309655,0.726020,
        -2.408554,2.317634,-2.185225,
        -2.408554,2.321510,1.685664,
        -2.408554,2.324376,-3.111006,
        -2.408554,2.324731,1.511186,
        -2.408554,2.329587,0.035498,
        -2.408554,2.330158,2.977264,
        -2.408554,2.335096,-0.588468,
        -2.408554,2.336046,-1.115799,
        -2.408554,2.338060,2.803613,
        -2.408554,2.338083,-0.943400,
        -2.408554,2.340210,0.562145,
        -2.408554,2.341931,1.023926,
        -2.408554,2.343265,-0.779061,
        -2.408554,2.343434,-2.013692,
        -2.408554,2.353029,-1.310529,
        -2.408554,2.354868,0.211543,
        -2.408554,2.355120,0.388593,
        -2.408554,2.355165,-1.833173,
        -2.408554,2.355462,2.151441,
        -2.408554,2.362948,-2.487477,
        -2.408554,2.363101,-2.810456,
        -2.408554,2.363394,-0.272918,
        -2.408554,2.364509,-1.644533,
        -2.408554,2.370465,2.466824,
        -2.408554,2.386110,-1.481351,
        -2.408554,2.394960,0.862133,
        -2.408554,2.395716,1.977184,
        -2.408554,2.396339,1.200900,
        -2.408554,2.407941,-2.316753,
        -2.408554,2.409450,-2.988432,
        -2.408554,2.417202,-0.103363,
        -2.408554,2.417386,-0.443097,
        -2.408554,2.417950,2.666662,
        -2.408554,2.423190,1.789783,
        -2.408554,2.423692,1.392696,
        -2.408554,2.431122,3.086379,
        -2.408554,2.431406,-2.647820,
        -2.408554,2.431584,2.307631,
        -2.408554,2.434090,1.592818,
        -2.408554,2.435993,0.683844,
        -2.408554,2.440898,-2.133300,
        -2.408554,2.442222,-1.195837,
        -2.408554,2.443112,2.883761,
        -2.408554,2.449191,-0.633651,
        -2.408554,2.450821,0.084741,
        -2.408554,2.453499,-1.008861,
        -2.408554,2.460010,-1.938011,
        -2.408554,2.460199,-0.822008,
        -2.408554,2.460902,0.490079,
        -2.408554,2.461774,-1.734913,
        -2.408554,2.463986,1.034407,
        -2.408554,2.467647,0.287667,
        -2.408554,2.479849,2.126925,
        -2.408554,2.485907,-2.835471,
        -2.408554,2.485945,-1.374853,
        -2.408554,2.487747,2.505912,
        -2.408554,2.487957,-0.274279,
        -2.408554,2.489741,-2.468428,
        -2.408554,2.512538,1.239900,
        -2.408554,2.517528,1.923289,
        -2.408554,2.518454,-1.565852,
        -2.408554,2.519348,0.843503,
        -2.408554,2.522766,-3.059231,
        -2.408554,2.528686,2.737228,
        -2.408554,2.534798,-2.266671,
        -2.408554,2.536533,1.466274,
        -2.408554,2.538504,1.699812,
        -2.408554,2.539516,-0.476942,
        -2.408554,2.539532,-0.074548,
        -2.408554,2.543197,2.990252,
        -2.408554,2.554022,2.320619,
        -2.408554,2.556919,-2.655005,
        -2.408554,2.558146,0.627191,
        -2.408554,2.558223,-1.164189,
        -2.408554,2.563452,-2.044028,
        -2.408554,2.568462,0.150655,
        -2.408554,2.569130,-0.700837,
        -2.408554,2.570050,-1.804492,
        -2.408554,2.572710,-0.934665,
        -2.408554,2.575994,0.391718,
        -2.408554,2.584774,1.046703,
        -2.408554,2.601432,2.102535,
        -2.408554,2.603559,2.555114,
        -2.408554,2.604150,-2.885208,
        -2.408554,2.608630,-1.378011,
        -2.408554,2.611256,-0.279110,
        -2.408554,2.617344,-2.439480,
        -2.408554,2.627402,1.293552,
        -2.408554,2.635396,1.848850,
        -2.408554,2.635686,2.834950,
        -2.408554,2.637087,3.132730,
        -2.408554,2.642029,0.813274,
        -2.408554,2.645260,1.570203,
        -2.408554,2.646182,-1.614036,
        -2.408554,2.660113,-2.186513,
        -2.408554,2.660483,-0.034608,
        -2.408554,2.661713,-0.529511,
        -2.408554,2.672720,-1.110422,
        -2.408554,2.676626,2.335353,
        -2.408554,2.676638,0.540364,
        -2.408554,2.676780,-1.901158,
        -2.408554,2.681411,-2.667965,
        -2.408554,2.682152,0.245751,
        -2.408554,2.682859,-0.817071,
        -2.408554,2.705107,1.062142,
        -2.408554,2.717705,2.625995,
        -2.408554,2.721001,2.060097,
        -2.408554,2.722319,-2.954069,
        -2.408554,2.729840,-1.368237,
        -2.408554,2.734349,-0.289298,
        -2.408554,2.738409,2.975671,
        -2.408554,2.740366,1.374341,
        -2.408554,2.744633,-2.391311,
        -2.408554,2.747398,1.727166,
        -2.408554,2.763799,0.760372,
        -2.408554,2.768238,-1.679243,
        -2.408554,2.779245,-2.046268,
        -2.408554,2.780208,-0.626453,
        -2.408554,2.780471,0.027732,
        -2.408554,2.784696,-1.013128,
        -2.408554,2.789817,0.395431,
        -2.408554,2.801154,2.353143,
        -2.408554,2.807091,-2.687477,
        -2.408554,2.825762,1.085035,
        -2.408554,2.830209,2.734943,
        -2.408554,2.833517,-3.085508,
        -2.408554,2.838301,1.980061,
        -2.408554,2.849871,1.512012,
        -2.408554,2.850089,-1.343899,
        -2.408554,2.856905,-0.318441,
        -2.408554,2.870445,-2.291001,
        -2.408554,2.885017,0.655562,
        -2.408554,2.886428,-1.793217,
        -2.408554,2.890337,-0.825633,
        -2.408554,2.898932,0.139973,
        -2.408554,2.931927,-2.728768,
        -2.408554,2.932096,2.377790,
        -2.408554,2.937409,2.947245,
        -2.408554,2.948048,1.133127,
        -2.408554,2.950896,1.797889,
        -2.408554,2.970019,-1.282382,
        -2.408554,2.978333,-0.412090,
        -2.408554,2.996437,-2.089121,
        -2.408554,3.011274,0.434110,
        -2.408554,3.050410,3.138492,
        -2.408554,3.064492,1.647259,
        -2.408554,3.088868,-0.989501,
        -2.303835,0.053884,-2.408394,
        -2.303835,0.077645,1.453433,
        -2.303835,0.088408,-0.311160,
        -2.303835,0.142314,2.590387,
        -2.303835,0.147206,-1.385164,
        -2.303835,0.163506,0.573203,
        -2.303835,0.165485,-2.875396,
        -2.303835,0.182564,-2.125821,
        -2.303835,0.190569,1.929141,
        -2.303835,0.201102,1.249082,
        -2.303835,0.201397,-0.716646,
        -2.303835,0.210568,-0.069559,
        -2.303835,0.251672,2.848900,
        -2.303835,0.266262,-1.678068,
        -2.303835,0.269108,-1.215563,
        -2.303835,0.273777,-2.506756,
        -2.303835,0.279854,0.814581,
        -2.303835,0.281664,-2.969147,
        -2.303835,0.283625,2.405603,
        -2.303835,0.284058,0.357350,
        -2.303835,0.294107,1.593880,
        -2.303835,0.302491,-0.400410,
        -2.303835,0.308059,2.021491,
        -2.303835,0.312219,-2.072057,
        -2.303835,0.321905,-0.829121,
        -2.303835,0.329984,1.201621,
        -2.303835,0.339714,-0.013712,
        -2.303835,0.368784,2.974974,
        -2.303835,0.376428,-1.495532,
        -2.303835,0.378103,2.646436,
        -2.303835,0.385088,0.590355,
        -2.303835,0.386248,-2.690943,
        -2.303835,0.389851,-2.340031,
        -2.303835,0.391502,-1.171374,
        -2.303835,0.398249,-1.812694,
        -2.303835,0.400542,-2.995729,
        -2.303835,0.404273,0.918163,
        -2.303835,0.405617,1.773533,
        -2.303835,0.409009,-0.570243,
        -2.303835,0.410700,0.270433,
        -2.303835,0.415893,1.464890,
        -2.303835,0.420680,2.088164,
        -2.303835,0.424766,-0.258729,
        -2.303835,0.426507,2.367872,
        -2.303835,0.441292,-0.890727,
        -2.303835,0.451839,-2.078193,
        -2.303835,0.459270,1.191562,
        -2.303835,0.470622,0.014002,
        -2.303835,0.481292,-2.529098,
        -2.303835,0.482460,2.849843,
        -2.303835,0.492205,-1.397963,
        -2.303835,0.493314,3.117264,
        -2.303835,0.497870,0.727828,
        -2.303835,0.499783,-1.654379,
        -2.303835,0.502651,0.463333,
        -2.303835,0.505463,2.594557,
        -2.303835,0.506104,-2.874910,
        -2.303835,0.513579,-1.147354,
        -2.303835,0.518569,1.644237,
        -2.303835,0.519001,1.900488,
        -2.303835,0.519277,-0.673809,
        -2.303835,0.524769,-0.421955,
        -2.303835,0.525564,-2.288838,
        -2.303835,0.532541,0.978737,
        -2.303835,0.534610,2.154808,
        -2.303835,0.535020,-1.891846,
        -2.303835,0.540208,1.400497,
        -2.303835,0.542006,0.228337,
        -2.303835,0.550762,-0.180771,
        -2.303835,0.560531,2.378474,
        -2.303835,0.561629,-0.924576,
        -2.303835,0.567465,-2.679392,
        -2.303835,0.588527,1.181755,
        -2.303835,0.591643,-2.098364,
        -2.303835,0.594963,2.997575,
        -2.303835,0.599643,-3.045235,
        -2.303835,0.599962,2.773556,
        -2.303835,0.602620,0.029270,
        -2.303835,0.605711,0.598454,
        -2.303835,0.607708,-1.541026,
        -2.303835,0.609360,-2.464379,
        -2.303835,0.610738,-1.326786,
        -2.303835,0.616968,0.820489,
        -2.303835,0.626379,1.773610,
        -2.303835,0.628626,-1.751180,
        -2.303835,0.630671,0.391949,
        -2.303835,0.632622,-0.737642,
        -2.303835,0.632708,2.566722,
        -2.303835,0.632797,1.989858,
        -2.303835,0.634455,-0.534790,
        -2.303835,0.636205,1.561170,
        -2.303835,0.637681,-1.123423,
        -2.303835,0.641441,-2.842109,
        -2.303835,0.644807,-0.327139,
        -2.303835,0.652440,2.197129,
        -2.303835,0.660897,-2.278593,
        -2.303835,0.664008,1.360091,
        -2.303835,0.670850,-1.945257,
        -2.303835,0.671396,1.004859,
        -2.303835,0.673985,0.203855,
        -2.303835,0.680326,-0.133370,
        -2.303835,0.685083,-0.942363,
        -2.303835,0.688204,-2.632204,
        -2.303835,0.689777,2.385011,
        -2.303835,0.700065,3.115430,
        -2.303835,0.704080,2.911636,
        -2.303835,0.712883,1.174216,
        -2.303835,0.715701,0.699906,
        -2.303835,0.718851,-1.442215,
        -2.303835,0.721019,2.721536,
        -2.303835,0.725300,-2.975134,
        -2.303835,0.727136,-1.633288,
        -2.303835,0.727399,-2.120567,
        -2.303835,0.730751,0.522488,
        -2.303835,0.732892,-1.254330,
        -2.303835,0.734333,-2.446184,
        -2.303835,0.735125,0.037270,
        -2.303835,0.737893,1.866463,
        -2.303835,0.740650,1.681075,
        -2.303835,0.745971,-0.781286,
        -2.303835,0.749432,-0.442559,
        -2.303835,0.750102,2.050636,
        -2.303835,0.751046,0.863821,
        -2.303835,0.757018,-0.608833,
        -2.303835,0.757046,-1.816019,
        -2.303835,0.757389,1.501184,
        -2.303835,0.759800,2.546181,
        -2.303835,0.759859,0.351923,
        -2.303835,0.761141,-2.781102,
        -2.303835,0.769035,-1.078574,
        -2.303835,0.770166,-0.266604,
        -2.303835,0.772218,2.227380,
        -2.303835,0.788382,1.331051,
        -2.303835,0.790269,-2.282409,
        -2.303835,0.803410,-1.984567,
        -2.303835,0.803813,0.191520,
        -2.303835,0.806832,3.022155,
        -2.303835,0.807476,-2.597675,
        -2.303835,0.812512,-0.107335,
        -2.303835,0.814161,1.010949,
        -2.303835,0.817732,-3.084486,
        -2.303835,0.817846,2.390688,
        -2.303835,0.819323,2.845960,
        -2.303835,0.819430,-0.915017,
        -2.303835,0.829198,-1.532491,
        -2.303835,0.829531,0.665695,
        -2.303835,0.829816,1.167147,
        -2.303835,0.833613,-1.357492,
        -2.303835,0.842945,-2.907318,
        -2.303835,0.844883,2.679708,
        -2.303835,0.848189,-1.705003,
        -2.303835,0.849888,1.767901,
        -2.303835,0.851126,0.493370,
        -2.303835,0.852911,1.933394,
        -2.303835,0.857516,-2.432863,
        -2.303835,0.858261,-1.191729,
        -2.303835,0.858279,-2.138895,
        -2.303835,0.859723,-0.734928,
        -2.303835,0.859754,1.603292,
        -2.303835,0.861655,-0.394821,
        -2.303835,0.864300,0.043549,
        -2.303835,0.866940,0.851229,
        -2.303835,0.869748,2.095969,
        -2.303835,0.877352,-0.574668,
        -2.303835,0.879254,-2.733338,
        -2.303835,0.882554,1.443757,
        -2.303835,0.884060,0.334199,
        -2.303835,0.885626,-1.866061,
        -2.303835,0.889834,2.529957,
        -2.303835,0.893140,2.251698,
        -2.303835,0.897463,-0.240822,
        -2.303835,0.898758,-1.031948,
        -2.303835,0.913918,1.287618,
        -2.303835,0.916555,-2.287173,
        -2.303835,0.918305,3.107845,
        -2.303835,0.919522,2.948277,
        -2.303835,0.927138,-2.569686,
        -2.303835,0.929031,0.185383,
        -2.303835,0.933766,-3.015917,
        -2.303835,0.934873,-2.010822,
        -2.303835,0.935636,-1.445855,
        -2.303835,0.936702,-0.866461,
        -2.303835,0.937395,0.729108,
        -2.303835,0.937910,2.787575,
        -2.303835,0.939144,-0.092962,
        -2.303835,0.941718,1.125085,
        -2.303835,0.942873,-1.607680,
        -2.303835,0.943701,0.980328,
        -2.303835,0.947774,2.395256,
        -2.303835,0.952433,0.584347,
        -2.303835,0.954456,-1.287293,
        -2.303835,0.958731,-2.852552,
        -2.303835,0.961838,1.839284,
        -2.303835,0.961898,-0.469425,
        -2.303835,0.965426,1.686615,
        -2.303835,0.970108,1.990771,
        -2.303835,0.972820,-1.762780,
        -2.303835,0.973500,-0.711881,
        -2.303835,0.975253,2.641082,
        -2.303835,0.977824,0.441928,
        -2.303835,0.981485,-2.424766,
        -2.303835,0.981877,1.537100,
        -2.303835,0.982609,-1.136920,
        -2.303835,0.984265,-2.156073,
        -2.303835,0.989150,0.045493,
        -2.303835,0.993052,2.137478,
        -2.303835,0.997766,-2.693871,
        -2.303835,0.999847,-0.334738,
        -2.303835,1.001920,0.848054,
        -2.303835,1.006146,1.389734,
        -2.303835,1.012180,0.304066,
        -2.303835,1.016938,2.270674,
        -2.303835,1.017371,-0.979790,
        -2.303835,1.019495,-1.894487,
        -2.303835,1.023796,-0.583691,
        -2.303835,1.026525,2.506945,
        -2.303835,1.027004,-0.203653,
        -2.303835,1.028027,3.036573,
        -2.303835,1.031435,1.241021,
        -2.303835,1.032316,2.884221,
        -2.303835,1.035863,-3.110475,
        -2.303835,1.042085,-2.292950,
        -2.303835,1.043272,-1.522986,
        -2.303835,1.046265,-2.958847,
        -2.303835,1.049272,-1.376344,
        -2.303835,1.050039,-2.552086,
        -2.303835,1.053682,0.689863,
        -2.303835,1.055458,0.170394,
        -2.303835,1.056622,-0.831866,
        -2.303835,1.057515,1.097998,
        -2.303835,1.058706,-2.036438,
        -2.303835,1.061389,-1.672248,
        -2.303835,1.063758,2.743093,
        -2.303835,1.070157,-0.073422,
        -2.303835,1.072154,1.761506,
        -2.303835,1.074029,1.909193,
        -2.303835,1.074204,0.545276,
        -2.303835,1.074714,-2.805147,
        -2.303835,1.076262,0.966104,
        -2.303835,1.077642,-1.227073,
        -2.303835,1.084151,1.618525,
        -2.303835,1.085085,2.377673,
        -2.303835,1.086822,-0.441112,
        -2.303835,1.087142,2.058815,
        -2.303835,1.098383,-0.696430,
        -2.303835,1.099138,-1.084033,
        -2.303835,1.103208,1.477773,
        -2.303835,1.103783,0.408254,
        -2.303835,1.106275,-2.420800,
        -2.303835,1.108079,2.612943,
        -2.303835,1.109345,-2.172718,
        -2.303835,1.115957,-1.787597,
        -2.303835,1.118930,0.045983,
        -2.303835,1.119777,0.834702,
        -2.303835,1.121943,-2.669042,
        -2.303835,1.124930,2.219917,
        -2.303835,1.125217,1.336294,
        -2.303835,1.130958,2.972931,
        -2.303835,1.131400,-0.306629,
        -2.303835,1.137135,-0.936514,
        -2.303835,1.139125,-1.924528,
        -2.303835,1.139877,0.276382,
        -2.303835,1.142003,-3.058022,
        -2.303835,1.147363,-0.558448,
        -2.303835,1.150348,-0.176418,
        -2.303835,1.151198,1.197016,
        -2.303835,1.153930,-1.597104,
        -2.303835,1.155389,2.837148,
        -2.303835,1.155458,-1.460623,
        -2.303835,1.157331,-2.906362,
        -2.303835,1.158594,2.487618,
        -2.303835,1.162103,-1.322553,
        -2.303835,1.165407,3.092861,
        -2.303835,1.168025,0.717075,
        -2.303835,1.168406,-2.301519,
        -2.303835,1.174782,-2.538755,
        -2.303835,1.177998,1.061256,
        -2.303835,1.178898,1.827305,
        -2.303835,1.181331,-2.060330,
        -2.303835,1.182540,-0.798952,
        -2.303835,1.184569,0.147517,
        -2.303835,1.184734,1.968919,
        -2.303835,1.184804,0.593386,
        -2.303835,1.188278,1.690808,
        -2.303835,1.194360,2.110440,
        -2.303835,1.194776,2.710455,
        -2.303835,1.197745,-1.173705,
        -2.303835,1.199898,-2.777257,
        -2.303835,1.201164,2.355931,
        -2.303835,1.203760,1.556404,
        -2.303835,1.205845,-0.415993,
        -2.303835,1.206585,0.933648,
        -2.303835,1.206950,-0.037248,
        -2.303835,1.209341,0.471162,
        -2.303835,1.213483,-1.039417,
        -2.303835,1.218146,-0.672643,
        -2.303835,1.221608,1.421669,
        -2.303835,1.226043,-1.693533,
        -2.303835,1.232870,-1.822773,
        -2.303835,1.234375,-2.421194,
        -2.303835,1.234531,-2.190148,
        -2.303835,1.239414,0.348959,
        -2.303835,1.240183,2.587255,
        -2.303835,1.242312,2.941234,
        -2.303835,1.244075,1.286644,
        -2.303835,1.245961,-2.997461,
        -2.303835,1.246495,-2.646872,
        -2.303835,1.247736,2.242810,
        -2.303835,1.253604,0.812222,
        -2.303835,1.257816,-0.894317,
        -2.303835,1.258247,-1.953634,
        -2.303835,1.258993,-0.284099,
        -2.303835,1.262380,-0.157609,
        -2.303835,1.262791,-1.549229,
        -2.303835,1.263772,-0.533073,
        -2.303835,1.266731,-1.412461,
        -2.303835,1.268591,-3.122146,
        -2.303835,1.269614,-1.275246,
        -2.303835,1.270745,1.152533,
        -2.303835,1.273606,0.224545,
        -2.303835,1.279716,2.812101,
        -2.303835,1.283786,2.462124,
        -2.303835,1.284798,0.078209,
        -2.303835,1.285402,-2.882049,
        -2.303835,1.286936,1.886053,
        -2.303835,1.289408,0.683122,
        -2.303835,1.294736,1.756184,
        -2.303835,1.296978,-2.311511,
        -2.303835,1.298546,2.020601,
        -2.303835,1.300924,1.022478,
        -2.303835,1.303305,3.043152,
        -2.303835,1.304210,-2.084225,
        -2.303835,1.305241,-2.530249,
        -2.303835,1.307513,1.628003,
        -2.303835,1.310722,0.558267,
        -2.303835,1.314142,2.139329,
        -2.303835,1.316087,-0.777971,
        -2.303835,1.317907,-1.121023,
        -2.303835,1.321019,-2.749776,
        -2.303835,1.321609,2.687717,
        -2.303835,1.322568,1.499415,
        -2.303835,1.322972,-0.395699,
        -2.303835,1.325286,-0.994192,
        -2.303835,1.329146,-0.648513,
        -2.303835,1.329795,-0.042570,
        -2.303835,1.331651,2.344497,
        -2.303835,1.336522,0.435098,
        -2.303835,1.339410,0.900087,
        -2.303835,1.339550,1.370353,
        -2.303835,1.340441,-1.731580,
        -2.303835,1.351855,-1.855234,
        -2.303835,1.360161,-3.018404,
        -2.303835,1.361031,-1.613971,
        -2.303835,1.361196,2.914780,
        -2.303835,1.361998,-2.207254,
        -2.303835,1.362302,1.239384,
        -2.303835,1.365228,2.563808,
        -2.303835,1.365452,0.310522,
        -2.303835,1.367028,-2.420041,
        -2.303835,1.373422,-1.484281,
        -2.303835,1.374245,-1.226417,
        -2.303835,1.375020,-0.170393,
        -2.303835,1.376560,-1.355481,
        -2.303835,1.377070,-1.980984,
        -2.303835,1.377578,-2.636339,
        -2.303835,1.378096,0.774626,
        -2.303835,1.379086,-0.508192,
        -2.303835,1.382562,0.177141,
        -2.303835,1.388849,2.233847,
        -2.303835,1.389268,-0.886349,
        -2.303835,1.391324,1.109396,
        -2.303835,1.391699,-0.288725,
        -2.303835,1.392869,3.136903,
        -2.303835,1.393612,1.952016,
        -2.303835,1.395545,-2.850807,
        -2.303835,1.401926,1.822930,
        -2.303835,1.402149,2.788898,
        -2.303835,1.409552,0.647980,
        -2.303835,1.410840,0.056219,
        -2.303835,1.411944,2.444377,
        -2.303835,1.413639,1.697982,
        -2.303835,1.426479,1.573427,
        -2.303835,1.426725,2.102242,
        -2.303835,1.427868,0.984637,
        -2.303835,1.429580,-2.107232,
        -2.303835,1.431375,-2.313269,
        -2.303835,1.431907,3.015410,
        -2.303835,1.432325,-1.029054,
        -2.303835,1.433341,0.524268,
        -2.303835,1.435181,-0.739771,
        -2.303835,1.437208,-2.529085,
        -2.303835,1.439041,-0.617966,
        -2.303835,1.441473,1.448544,
        -2.303835,1.444527,2.664744,
        -2.303835,1.450513,-0.400625,
        -2.303835,1.451907,-0.066823,
        -2.303835,1.452754,-2.738631,
        -2.303835,1.454061,-2.950416,
        -2.303835,1.456736,1.323154,
        -2.303835,1.458370,0.400433,
        -2.303835,1.459314,-1.765003,
        -2.303835,1.459902,-1.145835,
        -2.303835,1.464952,2.329245,
        -2.303835,1.466050,0.862104,
        -2.303835,1.471330,-1.886160,
        -2.303835,1.475119,-1.648255,
        -2.303835,1.480878,1.194576,
        -2.303835,1.480942,0.273235,
        -2.303835,1.481086,2.890269,
        -2.303835,1.484118,-1.529156,
        -2.303835,1.484741,-1.279219,
        -2.303835,1.487681,-1.406069,
        -2.303835,1.487886,-3.066037,
        -2.303835,1.490555,2.543284,
        -2.303835,1.492424,-0.190932,
        -2.303835,1.495002,-0.839810,
        -2.303835,1.495297,2.007188,
        -2.303835,1.496414,-2.009267,
        -2.303835,1.498261,-2.425270,
        -2.303835,1.498907,-2.207327,
        -2.303835,1.500204,0.737638,
        -2.303835,1.502469,0.151156,
        -2.303835,1.507448,-0.514782,
        -2.303835,1.509287,2.204647,
        -2.303835,1.510518,1.884920,
        -2.303835,1.511948,-2.632360,
        -2.303835,1.515235,1.068565,
        -2.303835,1.522237,3.102018,
        -2.303835,1.522373,1.764991,
        -2.303835,1.522416,-0.954956,
        -2.303835,1.523102,2.765920,
        -2.303835,1.524811,-2.844718,
        -2.303835,1.527261,-0.304916,
        -2.303835,1.529788,0.613208,
        -2.303835,1.533648,1.644808,
        -2.303835,1.535018,0.028869,
        -2.303835,1.541082,2.424608,
        -2.303835,1.546010,1.523462,
        -2.303835,1.551135,-0.642994,
        -2.303835,1.553895,0.489607,
        -2.303835,1.555041,0.947813,
        -2.303835,1.559671,-1.074934,
        -2.303835,1.560345,2.985899,
        -2.303835,1.560487,-2.323400,
        -2.303835,1.561127,1.401111,
        -2.303835,1.567193,2.643172,
        -2.303835,1.568892,2.098362,
        -2.303835,1.571702,-2.098435,
        -2.303835,1.572701,-2.528730,
        -2.303835,1.573572,1.278593,
        -2.303835,1.574849,-0.091505,
        -2.303835,1.576334,0.365236,
        -2.303835,1.577029,-2.962888,
        -2.303835,1.577688,-1.814941,
        -2.303835,1.579542,-0.755917,
        -2.303835,1.579624,-1.199759,
        -2.303835,1.581515,-0.419414,
        -2.303835,1.584581,-2.738702,
        -2.303835,1.586948,-1.696296,
        -2.303835,1.588901,2.301496,
        -2.303835,1.589936,0.825020,
        -2.303835,1.590018,-1.941652,
        -2.303835,1.596968,-1.328358,
        -2.303835,1.597178,-1.575607,
        -2.303835,1.600470,-1.453196,
        -2.303835,1.601408,0.242409,
        -2.303835,1.602259,1.152656,
        -2.303835,1.603064,2.865774,
        -2.303835,1.612085,1.968802,
        -2.303835,1.612720,-3.080498,
        -2.303835,1.616061,2.521531,
        -2.303835,1.616900,-0.207457,
        -2.303835,1.620988,-0.875920,
        -2.303835,1.622114,0.701509,
        -2.303835,1.622749,0.121426,
        -2.303835,1.622993,-2.221629,
        -2.303835,1.629346,-0.542669,
        -2.303835,1.630007,1.844552,
        -2.303835,1.634747,-2.426198,
        -2.303835,1.642346,1.722617,
        -2.303835,1.643925,1.034629,
        -2.303835,1.644848,2.743774,
        -2.303835,1.645258,2.188109,
        -2.303835,1.645417,-2.634725,
        -2.303835,1.647395,3.085437,
        -2.303835,1.648538,-2.855044,
        -2.303835,1.649932,0.576808,
        -2.303835,1.652224,-0.996105,
        -2.303835,1.652871,1.600681,
        -2.303835,1.657997,-0.325127,
        -2.303835,1.661936,0.001037,
        -2.303835,1.665187,1.477402,
        -2.303835,1.665496,2.397872,
        -2.303835,1.673267,-0.665453,
        -2.303835,1.673557,0.452526,
        -2.303835,1.679277,-1.122421,
        -2.303835,1.679696,0.912482,
        -2.303835,1.682420,1.353922,
        -2.303835,1.685031,-2.009038,
        -2.303835,1.685196,2.967046,
        -2.303835,1.690107,2.620910,
        -2.303835,1.690453,-2.123076,
        -2.303835,1.691693,-1.871572,
        -2.303835,1.691817,2.061688,
        -2.303835,1.692644,1.237670,
        -2.303835,1.696386,0.328123,
        -2.303835,1.697162,-1.249026,
        -2.303835,1.697488,-2.322785,
        -2.303835,1.698726,-2.973576,
        -2.303835,1.699172,-1.749717,
        -2.303835,1.704478,-0.106295,
        -2.303835,1.707244,-0.444770,
        -2.303835,1.707500,-2.531497,
        -2.303835,1.707714,-0.787321,
        -2.303835,1.708430,-1.627552,
        -2.303835,1.710583,-1.376890,
        -2.303835,1.712415,-2.747921,
        -2.303835,1.713125,-1.503073,
        -2.303835,1.713452,0.789419,
        -2.303835,1.720612,2.279888,
        -2.303835,1.724852,0.202831,
        -2.303835,1.724868,2.846246,
        -2.303835,1.728940,1.931535,
        -2.303835,1.732838,1.127584,
        -2.303835,1.735719,-3.092485,
        -2.303835,1.739676,2.496017,
        -2.303835,1.739971,-0.228876,
        -2.303835,1.743328,0.663347,
        -2.303835,1.744927,-0.911892,
        -2.303835,1.745952,0.084733,
        -2.303835,1.748077,1.804354,
        -2.303835,1.752959,-0.567548,
        -2.303835,1.760169,1.678132,
        -2.303835,1.763887,-2.218450,
        -2.303835,1.766737,2.722752,
        -2.303835,1.767999,1.004737,
        -2.303835,1.769149,1.550298,
        -2.303835,1.770119,-2.427175,
        -2.303835,1.770460,2.154208,
        -2.303835,1.770514,3.070982,
        -2.303835,1.770735,0.537245,
        -2.303835,1.772673,-2.864760,
        -2.303835,1.773527,-1.039582,
        -2.303835,1.776419,-2.642421,
        -2.303835,1.780769,1.416330,
        -2.303835,1.786634,-0.346247,
        -2.303835,1.788829,-1.947553,
        -2.303835,1.793737,0.410270,
        -2.303835,1.794553,2.374618,
        -2.303835,1.794699,-0.694061,
        -2.303835,1.797469,-1.169105,
        -2.303835,1.803705,0.881076,
        -2.303835,1.804124,1.264249,
        -2.303835,1.807825,-2.090455,
        -2.303835,1.808822,-1.817375,
        -2.303835,1.809020,2.949741,
        -2.303835,1.812466,2.597595,
        -2.303835,1.813159,2.025395,
        -2.303835,1.813816,-1.298435,
        -2.303835,1.815016,0.278951,
        -2.303835,1.816510,-0.008631,
        -2.303835,1.817772,-1.689625,
        -2.303835,1.821360,-2.984352,
        -2.303835,1.822312,-0.127211,
        -2.303835,1.823386,-1.429807,
        -2.303835,1.823834,-1.560483,
        -2.303835,1.832060,-0.821627,
        -2.303835,1.832690,-0.468440,
        -2.303835,1.833945,-2.320278,
        -2.303835,1.834787,0.751968,
        -2.303835,1.840579,-2.536445,
        -2.303835,1.841033,-2.756932,
        -2.303835,1.846350,1.893983,
        -2.303835,1.846644,1.113766,
        -2.303835,1.846833,2.826543,
        -2.303835,1.847325,2.249051,
        -2.303835,1.854119,0.134161,
        -2.303835,1.858863,-3.106022,
        -2.303835,1.865046,1.763384,
        -2.303835,1.865619,-0.952036,
        -2.303835,1.865912,0.624453,
        -2.303835,1.866694,2.473686,
        -2.303835,1.867524,-0.240083,
        -2.303835,1.875561,1.632831,
        -2.303835,1.876049,-0.594378,
        -2.303835,1.880509,1.500204,
        -2.303835,1.883504,1.360457,
        -2.303835,1.886971,-2.196665,
        -2.303835,1.888979,0.982922,
        -2.303835,1.889815,2.701570,
        -2.303835,1.892144,-2.005340,
        -2.303835,1.892348,0.496522,
        -2.303835,1.893281,-1.083976,
        -2.303835,1.893828,3.053453,
        -2.303835,1.895117,2.121107,
        -2.303835,1.898741,-2.875735,
        -2.303835,1.904429,-2.428029,
        -2.303835,1.908408,-2.648903,
        -2.303835,1.911828,-0.364489,
        -2.303835,1.912224,0.366565,
        -2.303835,1.914854,1.214468,
        -2.303835,1.915467,-1.883143,
        -2.303835,1.915763,-0.724679,
        -2.303835,1.915790,-1.216366,
        -2.303835,1.922256,2.347806,
        -2.303835,1.922616,0.844633,
        -2.303835,1.925185,0.231170,
        -2.303835,1.927474,-1.754616,
        -2.303835,1.928342,-1.351174,
        -2.303835,1.933328,-1.623328,
        -2.303835,1.933878,0.034150,
        -2.303835,1.934585,-1.487979,
        -2.303835,1.934802,2.927634,
        -2.303835,1.934862,1.989776,
        -2.303835,1.936327,-0.099350,
        -2.303835,1.938088,2.578507,
        -2.303835,1.945677,-2.999805,
        -2.303835,1.954572,-0.859208,
        -2.303835,1.957868,-0.490784,
        -2.303835,1.958989,0.712772,
        -2.303835,1.960212,-2.307561,
        -2.303835,1.964774,1.855507,
        -2.303835,1.964971,-2.100533,
        -2.303835,1.969112,1.088273,
        -2.303835,1.971438,-2.765415,
        -2.303835,1.972312,2.805134,
        -2.303835,1.974950,2.220596,
        -2.303835,1.975135,-2.539127,
        -2.303835,1.982107,1.720829,
        -2.303835,1.983195,-3.127984,
        -2.303835,1.983318,-0.247182,
        -2.303835,1.986720,-0.993220,
        -2.303835,1.990381,1.586005,
        -2.303835,1.990470,0.580283,
        -2.303835,1.990523,1.450326,
        -2.303835,1.993909,1.312897,
        -2.303835,1.994694,2.452813,
        -2.303835,1.999226,-0.620717,
        -2.303835,2.005145,0.941012,
        -2.303835,2.007301,0.142337,
        -2.303835,2.014149,0.444138,
        -2.303835,2.014338,-1.127392,
        -2.303835,2.018131,3.021778,
        -2.303835,2.019676,-1.980944,
        -2.303835,2.019912,2.686612,
        -2.303835,2.020545,2.088672,
        -2.303835,2.026074,0.300945,
        -2.303835,2.026196,-2.889057,
        -2.303835,2.031959,-1.268178,
        -2.303835,2.032917,-2.199917,
        -2.303835,2.033572,-1.842884,
        -2.303835,2.034712,-2.419505,
        -2.303835,2.035818,-0.379479,
        -2.303835,2.036004,-0.757062,
        -2.303835,2.040501,-1.411488,
        -2.303835,2.042874,-1.701367,
        -2.303835,2.043060,-2.653746,
        -2.303835,2.043729,-1.557938,
        -2.303835,2.048538,0.806623,
        -2.303835,2.049006,1.199930,
        -2.303835,2.049502,-0.137989,
        -2.303835,2.052181,2.326959,
        -2.303835,2.052794,-0.008526,
        -2.303835,2.057608,1.951726,
        -2.303835,2.065568,2.567730,
        -2.303835,2.070371,2.879808,
        -2.303835,2.071256,-3.018961,
        -2.303835,2.079746,-0.897699,
        -2.303835,2.082784,1.042403,
        -2.303835,2.083752,-0.511086,
        -2.303835,2.084401,1.810502,
        -2.303835,2.085731,0.671081,
        -2.303835,2.097496,1.373172,
        -2.303835,2.099185,1.667006,
        -2.303835,2.100482,-2.301084,
        -2.303835,2.103118,2.194745,
        -2.303835,2.103223,1.521685,
        -2.303835,2.103695,-2.777707,
        -2.303835,2.107523,3.129581,
        -2.303835,2.108277,-2.095907,
        -2.303835,2.108891,-2.532166,
        -2.303835,2.111003,-0.268632,
        -2.303835,2.112276,-1.027246,
        -2.303835,2.114380,0.531457,
        -2.303835,2.119166,0.100552,
        -2.303835,2.119564,2.754818,
        -2.303835,2.122474,-0.648314,
        -2.303835,2.125183,2.445598,
        -2.303835,2.129501,0.238290,
        -2.303835,2.131178,0.386663,
        -2.303835,2.131223,-1.174311,
        -2.303835,2.134269,-1.948042,
        -2.303835,2.134684,0.907065,
        -2.303835,2.137615,2.988806,
        -2.303835,2.145635,-1.327708,
        -2.303835,2.146467,2.055081,
        -2.303835,2.146582,-1.793788,
        -2.303835,2.149210,-1.482321,
        -2.303835,2.152901,-1.636788,
        -2.303835,2.156337,1.151558,
        -2.303835,2.156474,-2.912065,
        -2.303835,2.157194,-0.792563,
        -2.303835,2.166724,-0.401638,
        -2.303835,2.166980,-0.134879,
        -2.303835,2.173644,-2.403429,
        -2.303835,2.175711,-2.655523,
        -2.303835,2.177132,2.600419,
        -2.303835,2.178879,0.767722,
        -2.303835,2.180752,1.908485,
        -2.303835,2.181338,2.313721,
        -2.303835,2.185580,1.287673,
        -2.303835,2.195400,-2.217546,
        -2.303835,2.200418,-3.056442,
        -2.303835,2.201829,-0.001735,
        -2.303835,2.203804,1.756176,
        -2.303835,2.211602,1.443930,
        -2.303835,2.212686,-0.544809,
        -2.303835,2.213937,0.621892,
        -2.303835,2.214075,1.600691,
        -2.303835,2.216256,1.020029,
        -2.303835,2.219060,2.886631,
        -2.303835,2.223375,-0.906185,
        -2.303835,2.226002,-1.060248,
        -2.303835,2.226413,-2.064034,
        -2.303835,2.230554,2.171334,
        -2.303835,2.232413,2.735777,
        -2.303835,2.233709,3.070830,
        -2.303835,2.234383,-2.792065,
        -2.303835,2.237183,0.468799,
        -2.303835,2.237695,-0.271388,
        -2.303835,2.239419,0.151693,
        -2.303835,2.243347,-2.521049,
        -2.303835,2.244035,-1.229456,
        -2.303835,2.244355,0.309899,
        -2.303835,2.245175,2.460236,
        -2.303835,2.245628,-1.899216,
        -2.303835,2.245973,-0.702234,
        -2.303835,2.255601,-1.728131,
        -2.303835,2.256702,-1.401367,
        -2.303835,2.261466,-1.557500,
        -2.303835,2.267390,0.878499,
        -2.303835,2.272099,2.018594,
        -2.303835,2.278618,1.170386,
        -2.303835,2.283304,-2.345584,
        -2.303835,2.285175,-2.943090,
        -2.303835,2.293000,-0.124451,
        -2.303835,2.294058,-0.421070,
        -2.303835,2.303283,1.855938,
        -2.303835,2.304968,2.608978,
        -2.303835,2.305347,2.311346,
        -2.303835,2.307188,-2.656134,
        -2.303835,2.307301,1.336200,
        -2.303835,2.309655,0.726020,
        -2.303835,2.317634,-2.185225,
        -2.303835,2.321510,1.685664,
        -2.303835,2.324376,-3.111006,
        -2.303835,2.324731,1.511186,
        -2.303835,2.329587,0.035498,
        -2.303835,2.330158,2.977264,
        -2.303835,2.335096,-0.588468,
        -2.303835,2.336046,-1.115799,
        -2.303835,2.338060,2.803613,
        -2.303835,2.338083,-0.943400,
        -2.303835,2.340210,0.562145,
        -2.303835,2.341931,1.023926,
        -2.303835,2.343265,-0.779061,
        -2.303835,2.343434,-2.013692,
        -2.303835,2.353029,-1.310529,
        -2.303835,2.354868,0.211543,
        -2.303835,2.355120,0.388593,
        -2.303835,2.355165,-1.833173,
        -2.303835,2.355462,2.151441,
        -2.303835,2.362948,-2.487477,
        -2.303835,2.363101,-2.810456,
        -2.303835,2.363394,-0.272918,
        -2.303835,2.364509,-1.644533,
        -2.303835,2.370465,2.466824,
        -2.303835,2.386110,-1.481351,
        -2.303835,2.394960,0.862133,
        -2.303835,2.395716,1.977184,
        -2.303835,2.396339,1.200900,
        -2.303835,2.407941,-2.316753,
        -2.303835,2.409450,-2.988432,
        -2.303835,2.417202,-0.103363,
        -2.303835,2.417386,-0.443097,
        -2.303835,2.417950,2.666662,
        -2.303835,2.423190,1.789783,
        -2.303835,2.423692,1.392696,
        -2.303835,2.431122,3.086379,
        -2.303835,2.431406,-2.647820,
        -2.303835,2.431584,2.307631,
        -2.303835,2.434090,1.592818,
        -2.303835,2.435993,0.683844,
        -2.303835,2.440898,-2.133300,
        -2.303835,2.442222,-1.195837,
        -2.303835,2.443112,2.883761,
        -2.303835,2.449191,-0.633651,
        -2.303835,2.450821,0.084741,
        -2.303835,2.453499,-1.008861,
        -2.303835,2.460010,-1.938011,
        -2.303835,2.460199,-0.822008,
        -2.303835,2.460902,0.490079,
        -2.303835,2.461774,-1.734913,
        -2.303835,2.463986,1.034407,
        -2.303835,2.467647,0.287667,
        -2.303835,2.479849,2.126925,
        -2.303835,2.485907,-2.835471,
        -2.303835,2.485945,-1.374853,
        -2.303835,2.487747,2.505912,
        -2.303835,2.487957,-0.274279,
        -2.303835,2.489741,-2.468428,
        -2.303835,2.512538,1.239900,
        -2.303835,2.517528,1.923289,
        -2.303835,2.518454,-1.565852,
        -2.303835,2.519348,0.843503,
        -2.303835,2.522766,-3.059231,
        -2.303835,2.528686,2.737228,
        -2.303835,2.534798,-2.266671,
        -2.303835,2.536533,1.466274,
        -2.303835,2.538504,1.699812,
        -2.303835,2.539516,-0.476942,
        -2.303835,2.539532,-0.074548,
        -2.303835,2.543197,2.990252,
        -2.303835,2.554022,2.320619,
        -2.303835,2.556919,-2.655005,
        -2.303835,2.558146,0.627191,
        -2.303835,2.558223,-1.164189,
        -2.303835,2.563452,-2.044028,
        -2.303835,2.568462,0.150655,
        -2.303835,2.569130,-0.700837,
        -2.303835,2.570050,-1.804492,
        -2.303835,2.572710,-0.934665,
        -2.303835,2.575994,0.391718,
        -2.303835,2.584774,1.046703,
        -2.303835,2.601432,2.102535,
        -2.303835,2.603559,2.555114,
        -2.303835,2.604150,-2.885208,
        -2.303835,2.608630,-1.378011,
        -2.303835,2.611256,-0.279110,
        -2.303835,2.617344,-2.439480,
        -2.303835,2.627402,1.293552,
        -2.303835,2.635396,1.848850,
        -2.303835,2.635686,2.834950,
        -2.303835,2.637087,3.132730,
        -2.303835,2.642029,0.813274,
        -2.303835,2.645260,1.570203,
        -2.303835,2.646182,-1.614036,
        -2.303835,2.660113,-2.186513,
        -2.303835,2.660483,-0.034608,
        -2.303835,2.661713,-0.529511,
        -2.303835,2.672720,-1.110422,
        -2.303835,2.676626,2.335353,
        -2.303835,2.676638,0.540364,
        -2.303835,2.676780,-1.901158,
        -2.303835,2.681411,-2.667965,
        -2.303835,2.682152,0.245751,
        -2.303835,2.682859,-0.817071,
        -2.303835,2.705107,1.062142,
        -2.303835,2.717705,2.625995,
        -2.303835,2.721001,2.060097,
        -2.303835,2.722319,-2.954069,
        -2.303835,2.729840,-1.368237,
        -2.303835,2.734349,-0.289298,
        -2.303835,2.738409,2.975671,
        -2.303835,2.740366,1.374341,
        -2.303835,2.744633,-2.391311,
        -2.303835,2.747398,1.727166,
        -2.303835,2.763799,0.760372,
        -2.303835,2.768238,-1.679243,
        -2.303835,2.779245,-2.046268,
        -2.303835,2.780208,-0.626453,
        -2.303835,2.780471,0.027732,
        -2.303835,2.784696,-1.013128,
        -2.303835,2.789817,0.395431,
        -2.303835,2.801154,2.353143,
        -2.303835,2.807091,-2.687477,
        -2.303835,2.825762,1.085035,
        -2.303835,2.830209,2.734943,
        -2.303835,2.833517,-3.085508,
        -2.303835,2.838301,1.980061,
        -2.303835,2.849871,1.512012,
        -2.303835,2.850089,-1.343899,
        -2.303835,2.856905,-0.318441,
        -2.303835,2.870445,-2.291001,
        -2.303835,2.885017,0.655562,
        -2.303835,2.886428,-1.793217,
        -2.303835,2.890337,-0.825633,
        -2.303835,2.898932,0.139973,
        -2.303835,2.931927,-2.728768,
        -2.303835,2.932096,2.377790,
        -2.303835,2.937409,2.947245,
        -2.303835,2.948048,1.133127,
        -2.303835,2.950896,1.797889,
        -2.303835,2.970019,-1.282382,
        -2.303835,2.978333,-0.412090,
        -2.303835,2.996437,-2.089121,
        -2.303835,3.011274,0.434110,
        -2.303835,3.050410,3.138492,
        -2.303835,3.064492,1.647259,
        -2.303835,3.088868,-0.989501,
        -2.199115,0.053884,-2.408394,
        -2.199115,0.077645,1.453433,
        -2.199115,0.088408,-0.311160,
        -2.199115,0.142314,2.590387,
        -2.199115,0.147206,-1.385164,
        -2.199115,0.163506,0.573203,
        -2.199115,0.165485,-2.875396,
        -2.199115,0.182564,-2.125821,
        -2.199115,0.190569,1.929141,
        -2.199115,0.201102,1.249082,
        -2.199115,0.201397,-0.716646,
        -2.199115,0.210568,-0.069559,
        -2.199115,0.251672,2.848900,
        -2.199115,0.266262,-1.678068,
        -2.199115,0.269108,-1.215563,
        -2.199115,0.273777,-2.506756,
        -2.199115,0.279854,0.814581,
        -2.199115,0.281664,-2.969147,
        -2.199115,0.283625,2.405603,
        -2.199115,0.284058,0.357350,
        -2.199115,0.294107,1.593880,
        -2.199115,0.302491,-0.400410,
        -2.199115,0.308059,2.021491,
        -2.199115,0.312219,-2.072057,
        -2.199115,0.321905,-0.829121,
        -2.199115,0.329984,1.201621,
        -2.199115,0.339714,-0.013712,
        -2.199115,0.368784,2.974974,
        -2.199115,0.376428,-1.495532,
        -2.199115,0.378103,2.646436,
        -2.199115,0.385088,0.590355,
        -2.199115,0.386248,-2.690943,
        -2.199115,0.389851,-2.340031,
        -2.199115,0.391502,-1.171374,
        -2.199115,0.398249,-1.812694,
        -2.199115,0.400542,-2.995729,
        -2.199115,0.404273,0.918163,
        -2.199115,0.405617,1.773533,
        -2.199115,0.409009,-0.570243,
        -2.199115,0.410700,0.270433,
        -2.199115,0.415893,1.464890,
        -2.199115,0.420680,2.088164,
        -2.199115,0.424766,-0.258729,
        -2.199115,0.426507,2.367872,
        -2.199115,0.441292,-0.890727,
        -2.199115,0.451839,-2.078193,
        -2.199115,0.459270,1.191562,
        -2.199115,0.470622,0.014002,
        -2.199115,0.481292,-2.529098,
        -2.199115,0.482460,2.849843,
        -2.199115,0.492205,-1.397963,
        -2.199115,0.493314,3.117264,
        -2.199115,0.497870,0.727828,
        -2.199115,0.499783,-1.654379,
        -2.199115,0.502651,0.463333,
        -2.199115,0.505463,2.594557,
        -2.199115,0.506104,-2.874910,
        -2.199115,0.513579,-1.147354,
        -2.199115,0.518569,1.644237,
        -2.199115,0.519001,1.900488,
        -2.199115,0.519277,-0.673809,
        -2.199115,0.524769,-0.421955,
        -2.199115,0.525564,-2.288838,
        -2.199115,0.532541,0.978737,
        -2.199115,0.534610,2.154808,
        -2.199115,0.535020,-1.891846,
        -2.199115,0.540208,1.400497,
        -2.199115,0.542006,0.228337,
        -2.199115,0.550762,-0.180771,
        -2.199115,0.560531,2.378474,
        -2.199115,0.561629,-0.924576,
        -2.199115,0.567465,-2.679392,
        -2.199115,0.588527,1.181755,
        -2.199115,0.591643,-2.098364,
        -2.199115,0.594963,2.997575,
        -2.199115,0.599643,-3.045235,
        -2.199115,0.599962,2.773556,
        -2.199115,0.602620,0.029270,
        -2.199115,0.605711,0.598454,
        -2.199115,0.607708,-1.541026,
        -2.199115,0.609360,-2.464379,
        -2.199115,0.610738,-1.326786,
        -2.199115,0.616968,0.820489,
        -2.199115,0.626379,1.773610,
        -2.199115,0.628626,-1.751180,
        -2.199115,0.630671,0.391949,
        -2.199115,0.632622,-0.737642,
        -2.199115,0.632708,2.566722,
        -2.199115,0.632797,1.989858,
        -2.199115,0.634455,-0.534790,
        -2.199115,0.636205,1.561170,
        -2.199115,0.637681,-1.123423,
        -2.199115,0.641441,-2.842109,
        -2.199115,0.644807,-0.327139,
        -2.199115,0.652440,2.197129,
        -2.199115,0.660897,-2.278593,
        -2.199115,0.664008,1.360091,
        -2.199115,0.670850,-1.945257,
        -2.199115,0.671396,1.004859,
        -2.199115,0.673985,0.203855,
        -2.199115,0.680326,-0.133370,
        -2.199115,0.685083,-0.942363,
        -2.199115,0.688204,-2.632204,
        -2.199115,0.689777,2.385011,
        -2.199115,0.700065,3.115430,
        -2.199115,0.704080,2.911636,
        -2.199115,0.712883,1.174216,
        -2.199115,0.715701,0.699906,
        -2.199115,0.718851,-1.442215,
        -2.199115,0.721019,2.721536,
        -2.199115,0.725300,-2.975134,
        -2.199115,0.727136,-1.633288,
        -2.199115,0.727399,-2.120567,
        -2.199115,0.730751,0.522488,
        -2.199115,0.732892,-1.254330,
        -2.199115,0.734333,-2.446184,
        -2.199115,0.735125,0.037270,
        -2.199115,0.737893,1.866463,
        -2.199115,0.740650,1.681075,
        -2.199115,0.745971,-0.781286,
        -2.199115,0.749432,-0.442559,
        -2.199115,0.750102,2.050636,
        -2.199115,0.751046,0.863821,
        -2.199115,0.757018,-0.608833,
        -2.199115,0.757046,-1.816019,
        -2.199115,0.757389,1.501184,
        -2.199115,0.759800,2.546181,
        -2.199115,0.759859,0.351923,
        -2.199115,0.761141,-2.781102,
        -2.199115,0.769035,-1.078574,
        -2.199115,0.770166,-0.266604,
        -2.199115,0.772218,2.227380,
        -2.199115,0.788382,1.331051,
        -2.199115,0.790269,-2.282409,
        -2.199115,0.803410,-1.984567,
        -2.199115,0.803813,0.191520,
        -2.199115,0.806832,3.022155,
        -2.199115,0.807476,-2.597675,
        -2.199115,0.812512,-0.107335,
        -2.199115,0.814161,1.010949,
        -2.199115,0.817732,-3.084486,
        -2.199115,0.817846,2.390688,
        -2.199115,0.819323,2.845960,
        -2.199115,0.819430,-0.915017,
        -2.199115,0.829198,-1.532491,
        -2.199115,0.829531,0.665695,
        -2.199115,0.829816,1.167147,
        -2.199115,0.833613,-1.357492,
        -2.199115,0.842945,-2.907318,
        -2.199115,0.844883,2.679708,
        -2.199115,0.848189,-1.705003,
        -2.199115,0.849888,1.767901,
        -2.199115,0.851126,0.493370,
        -2.199115,0.852911,1.933394,
        -2.199115,0.857516,-2.432863,
        -2.199115,0.858261,-1.191729,
        -2.199115,0.858279,-2.138895,
        -2.199115,0.859723,-0.734928,
        -2.199115,0.859754,1.603292,
        -2.199115,0.861655,-0.394821,
        -2.199115,0.864300,0.043549,
        -2.199115,0.866940,0.851229,
        -2.199115,0.869748,2.095969,
        -2.199115,0.877352,-0.574668,
        -2.199115,0.879254,-2.733338,
        -2.199115,0.882554,1.443757,
        -2.199115,0.884060,0.334199,
        -2.199115,0.885626,-1.866061,
        -2.199115,0.889834,2.529957,
        -2.199115,0.893140,2.251698,
        -2.199115,0.897463,-0.240822,
        -2.199115,0.898758,-1.031948,
        -2.199115,0.913918,1.287618,
        -2.199115,0.916555,-2.287173,
        -2.199115,0.918305,3.107845,
        -2.199115,0.919522,2.948277,
        -2.199115,0.927138,-2.569686,
        -2.199115,0.929031,0.185383,
        -2.199115,0.933766,-3.015917,
        -2.199115,0.934873,-2.010822,
        -2.199115,0.935636,-1.445855,
        -2.199115,0.936702,-0.866461,
        -2.199115,0.937395,0.729108,
        -2.199115,0.937910,2.787575,
        -2.199115,0.939144,-0.092962,
        -2.199115,0.941718,1.125085,
        -2.199115,0.942873,-1.607680,
        -2.199115,0.943701,0.980328,
        -2.199115,0.947774,2.395256,
        -2.199115,0.952433,0.584347,
        -2.199115,0.954456,-1.287293,
        -2.199115,0.958731,-2.852552,
        -2.199115,0.961838,1.839284,
        -2.199115,0.961898,-0.469425,
        -2.199115,0.965426,1.686615,
        -2.199115,0.970108,1.990771,
        -2.199115,0.972820,-1.762780,
        -2.199115,0.973500,-0.711881,
        -2.199115,0.975253,2.641082,
        -2.199115,0.977824,0.441928,
        -2.199115,0.981485,-2.424766,
        -2.199115,0.981877,1.537100,
        -2.199115,0.982609,-1.136920,
        -2.199115,0.984265,-2.156073,
        -2.199115,0.989150,0.045493,
        -2.199115,0.993052,2.137478,
        -2.199115,0.997766,-2.693871,
        -2.199115,0.999847,-0.334738,
        -2.199115,1.001920,0.848054,
        -2.199115,1.006146,1.389734,
        -2.199115,1.012180,0.304066,
        -2.199115,1.016938,2.270674,
        -2.199115,1.017371,-0.979790,
        -2.199115,1.019495,-1.894487,
        -2.199115,1.023796,-0.583691,
        -2.199115,1.026525,2.506945,
        -2.199115,1.027004,-0.203653,
        -2.199115,1.028027,3.036573,
        -2.199115,1.031435,1.241021,
        -2.199115,1.032316,2.884221,
        -2.199115,1.035863,-3.110475,
        -2.199115,1.042085,-2.292950,
        -2.199115,1.043272,-1.522986,
        -2.199115,1.046265,-2.958847,
        -2.199115,1.049272,-1.376344,
        -2.199115,1.050039,-2.552086,
        -2.199115,1.053682,0.689863,
        -2.199115,1.055458,0.170394,
        -2.199115,1.056622,-0.831866,
        -2.199115,1.057515,1.097998,
        -2.199115,1.058706,-2.036438,
        -2.199115,1.061389,-1.672248,
        -2.199115,1.063758,2.743093,
        -2.199115,1.070157,-0.073422,
        -2.199115,1.072154,1.761506,
        -2.199115,1.074029,1.909193,
        -2.199115,1.074204,0.545276,
        -2.199115,1.074714,-2.805147,
        -2.199115,1.076262,0.966104,
        -2.199115,1.077642,-1.227073,
        -2.199115,1.084151,1.618525,
        -2.199115,1.085085,2.377673,
        -2.199115,1.086822,-0.441112,
        -2.199115,1.087142,2.058815,
        -2.199115,1.098383,-0.696430,
        -2.199115,1.099138,-1.084033,
        -2.199115,1.103208,1.477773,
        -2.199115,1.103783,0.408254,
        -2.199115,1.106275,-2.420800,
        -2.199115,1.108079,2.612943,
        -2.199115,1.109345,-2.172718,
        -2.199115,1.115957,-1.787597,
        -2.199115,1.118930,0.045983,
        -2.199115,1.119777,0.834702,
        -2.199115,1.121943,-2.669042,
        -2.199115,1.124930,2.219917,
        -2.199115,1.125217,1.336294,
        -2.199115,1.130958,2.972931,
        -2.199115,1.131400,-0.306629,
        -2.199115,1.137135,-0.936514,
        -2.199115,1.139125,-1.924528,
        -2.199115,1.139877,0.276382,
        -2.199115,1.142003,-3.058022,
        -2.199115,1.147363,-0.558448,
        -2.199115,1.150348,-0.176418,
        -2.199115,1.151198,1.197016,
        -2.199115,1.153930,-1.597104,
        -2.199115,1.155389,2.837148,
        -2.199115,1.155458,-1.460623,
        -2.199115,1.157331,-2.906362,
        -2.199115,1.158594,2.487618,
        -2.199115,1.162103,-1.322553,
        -2.199115,1.165407,3.092861,
        -2.199115,1.168025,0.717075,
        -2.199115,1.168406,-2.301519,
        -2.199115,1.174782,-2.538755,
        -2.199115,1.177998,1.061256,
        -2.199115,1.178898,1.827305,
        -2.199115,1.181331,-2.060330,
        -2.199115,1.182540,-0.798952,
        -2.199115,1.184569,0.147517,
        -2.199115,1.184734,1.968919,
        -2.199115,1.184804,0.593386,
        -2.199115,1.188278,1.690808,
        -2.199115,1.194360,2.110440,
        -2.199115,1.194776,2.710455,
        -2.199115,1.197745,-1.173705,
        -2.199115,1.199898,-2.777257,
        -2.199115,1.201164,2.355931,
        -2.199115,1.203760,1.556404,
        -2.199115,1.205845,-0.415993,
        -2.199115,1.206585,0.933648,
        -2.199115,1.206950,-0.037248,
        -2.199115,1.209341,0.471162,
        -2.199115,1.213483,-1.039417,
        -2.199115,1.218146,-0.672643,
        -2.199115,1.221608,1.421669,
        -2.199115,1.226043,-1.693533,
        -2.199115,1.232870,-1.822773,
        -2.199115,1.234375,-2.421194,
        -2.199115,1.234531,-2.190148,
        -2.199115,1.239414,0.348959,
        -2.199115,1.240183,2.587255,
        -2.199115,1.242312,2.941234,
        -2.199115,1.244075,1.286644,
        -2.199115,1.245961,-2.997461,
        -2.199115,1.246495,-2.646872,
        -2.199115,1.247736,2.242810,
        -2.199115,1.253604,0.812222,
        -2.199115,1.257816,-0.894317,
        -2.199115,1.258247,-1.953634,
        -2.199115,1.258993,-0.284099,
        -2.199115,1.262380,-0.157609,
        -2.199115,1.262791,-1.549229,
        -2.199115,1.263772,-0.533073,
        -2.199115,1.266731,-1.412461,
        -2.199115,1.268591,-3.122146,
        -2.199115,1.269614,-1.275246,
        -2.199115,1.270745,1.152533,
        -2.199115,1.273606,0.224545,
        -2.199115,1.279716,2.812101,
        -2.199115,1.283786,2.462124,
        -2.199115,1.284798,0.078209,
        -2.199115,1.285402,-2.882049,
        -2.199115,1.286936,1.886053,
        -2.199115,1.289408,0.683122,
        -2.199115,1.294736,1.756184,
        -2.199115,1.296978,-2.311511,
        -2.199115,1.298546,2.020601,
        -2.199115,1.300924,1.022478,
        -2.199115,1.303305,3.043152,
        -2.199115,1.304210,-2.084225,
        -2.199115,1.305241,-2.530249,
        -2.199115,1.307513,1.628003,
        -2.199115,1.310722,0.558267,
        -2.199115,1.314142,2.139329,
        -2.199115,1.316087,-0.777971,
        -2.199115,1.317907,-1.121023,
        -2.199115,1.321019,-2.749776,
        -2.199115,1.321609,2.687717,
        -2.199115,1.322568,1.499415,
        -2.199115,1.322972,-0.395699,
        -2.199115,1.325286,-0.994192,
        -2.199115,1.329146,-0.648513,
        -2.199115,1.329795,-0.042570,
        -2.199115,1.331651,2.344497,
        -2.199115,1.336522,0.435098,
        -2.199115,1.339410,0.900087,
        -2.199115,1.339550,1.370353,
        -2.199115,1.340441,-1.731580,
        -2.199115,1.351855,-1.855234,
        -2.199115,1.360161,-3.018404,
        -2.199115,1.361031,-1.613971,
        -2.199115,1.361196,2.914780,
        -2.199115,1.361998,-2.207254,
        -2.199115,1.362302,1.239384,
        -2.199115,1.365228,2.563808,
        -2.199115,1.365452,0.310522,
        -2.199115,1.367028,-2.420041,
        -2.199115,1.373422,-1.484281,
        -2.199115,1.374245,-1.226417,
        -2.199115,1.375020,-0.170393,
        -2.199115,1.376560,-1.355481,
        -2.199115,1.377070,-1.980984,
        -2.199115,1.377578,-2.636339,
        -2.199115,1.378096,0.774626,
        -2.199115,1.379086,-0.508192,
        -2.199115,1.382562,0.177141,
        -2.199115,1.388849,2.233847,
        -2.199115,1.389268,-0.886349,
        -2.199115,1.391324,1.109396,
        -2.199115,1.391699,-0.288725,
        -2.199115,1.392869,3.136903,
        -2.199115,1.393612,1.952016,
        -2.199115,1.395545,-2.850807,
        -2.199115,1.401926,1.822930,
        -2.199115,1.402149,2.788898,
        -2.199115,1.409552,0.647980,
        -2.199115,1.410840,0.056219,
        -2.199115,1.411944,2.444377,
        -2.199115,1.413639,1.697982,
        -2.199115,1.426479,1.573427,
        -2.199115,1.426725,2.102242,
        -2.199115,1.427868,0.984637,
        -2.199115,1.429580,-2.107232,
        -2.199115,1.431375,-2.313269,
        -2.199115,1.431907,3.015410,
        -2.199115,1.432325,-1.029054,
        -2.199115,1.433341,0.524268,
        -2.199115,1.435181,-0.739771,
        -2.199115,1.437208,-2.529085,
        -2.199115,1.439041,-0.617966,
        -2.199115,1.441473,1.448544,
        -2.199115,1.444527,2.664744,
        -2.199115,1.450513,-0.400625,
        -2.199115,1.451907,-0.066823,
        -2.199115,1.452754,-2.738631,
        -2.199115,1.454061,-2.950416,
        -2.199115,1.456736,1.323154,
        -2.199115,1.458370,0.400433,
        -2.199115,1.459314,-1.765003,
        -2.199115,1.459902,-1.145835,
        -2.199115,1.464952,2.329245,
        -2.199115,1.466050,0.862104,
        -2.199115,1.471330,-1.886160,
        -2.199115,1.475119,-1.648255,
        -2.199115,1.480878,1.194576,
        -2.199115,1.480942,0.273235,
        -2.199115,1.481086,2.890269,
        -2.199115,1.484118,-1.529156,
        -2.199115,1.484741,-1.279219,
        -2.199115,1.487681,-1.406069,
        -2.199115,1.487886,-3.066037,
        -2.199115,1.490555,2.543284,
        -2.199115,1.492424,-0.190932,
        -2.199115,1.495002,-0.839810,
        -2.199115,1.495297,2.007188,
        -2.199115,1.496414,-2.009267,
        -2.199115,1.498261,-2.425270,
        -2.199115,1.498907,-2.207327,
        -2.199115,1.500204,0.737638,
        -2.199115,1.502469,0.151156,
        -2.199115,1.507448,-0.514782,
        -2.199115,1.509287,2.204647,
        -2.199115,1.510518,1.884920,
        -2.199115,1.511948,-2.632360,
        -2.199115,1.515235,1.068565,
        -2.199115,1.522237,3.102018,
        -2.199115,1.522373,1.764991,
        -2.199115,1.522416,-0.954956,
        -2.199115,1.523102,2.765920,
        -2.199115,1.524811,-2.844718,
        -2.199115,1.527261,-0.304916,
        -2.199115,1.529788,0.613208,
        -2.199115,1.533648,1.644808,
        -2.199115,1.535018,0.028869,
        -2.199115,1.541082,2.424608,
        -2.199115,1.546010,1.523462,
        -2.199115,1.551135,-0.642994,
        -2.199115,1.553895,0.489607,
        -2.199115,1.555041,0.947813,
        -2.199115,1.559671,-1.074934,
        -2.199115,1.560345,2.985899,
        -2.199115,1.560487,-2.323400,
        -2.199115,1.561127,1.401111,
        -2.199115,1.567193,2.643172,
        -2.199115,1.568892,2.098362,
        -2.199115,1.571702,-2.098435,
        -2.199115,1.572701,-2.528730,
        -2.199115,1.573572,1.278593,
        -2.199115,1.574849,-0.091505,
        -2.199115,1.576334,0.365236,
        -2.199115,1.577029,-2.962888,
        -2.199115,1.577688,-1.814941,
        -2.199115,1.579542,-0.755917,
        -2.199115,1.579624,-1.199759,
        -2.199115,1.581515,-0.419414,
        -2.199115,1.584581,-2.738702,
        -2.199115,1.586948,-1.696296,
        -2.199115,1.588901,2.301496,
        -2.199115,1.589936,0.825020,
        -2.199115,1.590018,-1.941652,
        -2.199115,1.596968,-1.328358,
        -2.199115,1.597178,-1.575607,
        -2.199115,1.600470,-1.453196,
        -2.199115,1.601408,0.242409,
        -2.199115,1.602259,1.152656,
        -2.199115,1.603064,2.865774,
        -2.199115,1.612085,1.968802,
        -2.199115,1.612720,-3.080498,
        -2.199115,1.616061,2.521531,
        -2.199115,1.616900,-0.207457,
        -2.199115,1.620988,-0.875920,
        -2.199115,1.622114,0.701509,
        -2.199115,1.622749,0.121426,
        -2.199115,1.622993,-2.221629,
        -2.199115,1.629346,-0.542669,
        -2.199115,1.630007,1.844552,
        -2.199115,1.634747,-2.426198,
        -2.199115,1.642346,1.722617,
        -2.199115,1.643925,1.034629,
        -2.199115,1.644848,2.743774,
        -2.199115,1.645258,2.188109,
        -2.199115,1.645417,-2.634725,
        -2.199115,1.647395,3.085437,
        -2.199115,1.648538,-2.855044,
        -2.199115,1.649932,0.576808,
        -2.199115,1.652224,-0.996105,
        -2.199115,1.652871,1.600681,
        -2.199115,1.657997,-0.325127,
        -2.199115,1.661936,0.001037,
        -2.199115,1.665187,1.477402,
        -2.199115,1.665496,2.397872,
        -2.199115,1.673267,-0.665453,
        -2.199115,1.673557,0.452526,
        -2.199115,1.679277,-1.122421,
        -2.199115,1.679696,0.912482,
        -2.199115,1.682420,1.353922,
        -2.199115,1.685031,-2.009038,
        -2.199115,1.685196,2.967046,
        -2.199115,1.690107,2.620910,
        -2.199115,1.690453,-2.123076,
        -2.199115,1.691693,-1.871572,
        -2.199115,1.691817,2.061688,
        -2.199115,1.692644,1.237670,
        -2.199115,1.696386,0.328123,
        -2.199115,1.697162,-1.249026,
        -2.199115,1.697488,-2.322785,
        -2.199115,1.698726,-2.973576,
        -2.199115,1.699172,-1.749717,
        -2.199115,1.704478,-0.106295,
        -2.199115,1.707244,-0.444770,
        -2.199115,1.707500,-2.531497,
        -2.199115,1.707714,-0.787321,
        -2.199115,1.708430,-1.627552,
        -2.199115,1.710583,-1.376890,
        -2.199115,1.712415,-2.747921,
        -2.199115,1.713125,-1.503073,
        -2.199115,1.713452,0.789419,
        -2.199115,1.720612,2.279888,
        -2.199115,1.724852,0.202831,
        -2.199115,1.724868,2.846246,
        -2.199115,1.728940,1.931535,
        -2.199115,1.732838,1.127584,
        -2.199115,1.735719,-3.092485,
        -2.199115,1.739676,2.496017,
        -2.199115,1.739971,-0.228876,
        -2.199115,1.743328,0.663347,
        -2.199115,1.744927,-0.911892,
        -2.199115,1.745952,0.084733,
        -2.199115,1.748077,1.804354,
        -2.199115,1.752959,-0.567548,
        -2.199115,1.760169,1.678132,
        -2.199115,1.763887,-2.218450,
        -2.199115,1.766737,2.722752,
        -2.199115,1.767999,1.004737,
        -2.199115,1.769149,1.550298,
        -2.199115,1.770119,-2.427175,
        -2.199115,1.770460,2.154208,
        -2.199115,1.770514,3.070982,
        -2.199115,1.770735,0.537245,
        -2.199115,1.772673,-2.864760,
        -2.199115,1.773527,-1.039582,
        -2.199115,1.776419,-2.642421,
        -2.199115,1.780769,1.416330,
        -2.199115,1.786634,-0.346247,
        -2.199115,1.788829,-1.947553,
        -2.199115,1.793737,0.410270,
        -2.199115,1.794553,2.374618,
        -2.199115,1.794699,-0.694061,
        -2.199115,1.797469,-1.169105,
        -2.199115,1.803705,0.881076,
        -2.199115,1.804124,1.264249,
        -2.199115,1.807825,-2.090455,
        -2.199115,1.808822,-1.817375,
        -2.199115,1.809020,2.949741,
        -2.199115,1.812466,2.597595,
        -2.199115,1.813159,2.025395,
        -2.199115,1.813816,-1.298435,
        -2.199115,1.815016,0.278951,
        -2.199115,1.816510,-0.008631,
        -2.199115,1.817772,-1.689625,
        -2.199115,1.821360,-2.984352,
        -2.199115,1.822312,-0.127211,
        -2.199115,1.823386,-1.429807,
        -2.199115,1.823834,-1.560483,
        -2.199115,1.832060,-0.821627,
        -2.199115,1.832690,-0.468440,
        -2.199115,1.833945,-2.320278,
        -2.199115,1.834787,0.751968,
        -2.199115,1.840579,-2.536445,
        -2.199115,1.841033,-2.756932,
        -2.199115,1.846350,1.893983,
        -2.199115,1.846644,1.113766,
        -2.199115,1.846833,2.826543,
        -2.199115,1.847325,2.249051,
        -2.199115,1.854119,0.134161,
        -2.199115,1.858863,-3.106022,
        -2.199115,1.865046,1.763384,
        -2.199115,1.865619,-0.952036,
        -2.199115,1.865912,0.624453,
        -2.199115,1.866694,2.473686,
        -2.199115,1.867524,-0.240083,
        -2.199115,1.875561,1.632831,
        -2.199115,1.876049,-0.594378,
        -2.199115,1.880509,1.500204,
        -2.199115,1.883504,1.360457,
        -2.199115,1.886971,-2.196665,
        -2.199115,1.888979,0.982922,
        -2.199115,1.889815,2.701570,
        -2.199115,1.892144,-2.005340,
        -2.199115,1.892348,0.496522,
        -2.199115,1.893281,-1.083976,
        -2.199115,1.893828,3.053453,
        -2.199115,1.895117,2.121107,
        -2.199115,1.898741,-2.875735,
        -2.199115,1.904429,-2.428029,
        -2.199115,1.908408,-2.648903,
        -2.199115,1.911828,-0.364489,
        -2.199115,1.912224,0.366565,
        -2.199115,1.914854,1.214468,
        -2.199115,1.915467,-1.883143,
        -2.199115,1.915763,-0.724679,
        -2.199115,1.915790,-1.216366,
        -2.199115,1.922256,2.347806,
        -2.199115,1.922616,0.844633,
        -2.199115,1.925185,0.231170,
        -2.199115,1.927474,-1.754616,
        -2.199115,1.928342,-1.351174,
        -2.199115,1.933328,-1.623328,
        -2.199115,1.933878,0.034150,
        -2.199115,1.934585,-1.487979,
        -2.199115,1.934802,2.927634,
        -2.199115,1.934862,1.989776,
        -2.199115,1.936327,-0.099350,
        -2.199115,1.938088,2.578507,
        -2.199115,1.945677,-2.999805,
        -2.199115,1.954572,-0.859208,
        -2.199115,1.957868,-0.490784,
        -2.199115,1.958989,0.712772,
        -2.199115,1.960212,-2.307561,
        -2.199115,1.964774,1.855507,
        -2.199115,1.964971,-2.100533,
        -2.199115,1.969112,1.088273,
        -2.199115,1.971438,-2.765415,
        -2.199115,1.972312,2.805134,
        -2.199115,1.974950,2.220596,
        -2.199115,1.975135,-2.539127,
        -2.199115,1.982107,1.720829,
        -2.199115,1.983195,-3.127984,
        -2.199115,1.983318,-0.247182,
        -2.199115,1.986720,-0.993220,
        -2.199115,1.990381,1.586005,
        -2.199115,1.990470,0.580283,
        -2.199115,1.990523,1.450326,
        -2.199115,1.993909,1.312897,
        -2.199115,1.994694,2.452813,
        -2.199115,1.999226,-0.620717,
        -2.199115,2.005145,0.941012,
        -2.199115,2.007301,0.142337,
        -2.199115,2.014149,0.444138,
        -2.199115,2.014338,-1.127392,
        -2.199115,2.018131,3.021778,
        -2.199115,2.019676,-1.980944,
        -2.199115,2.019912,2.686612,
        -2.199115,2.020545,2.088672,
        -2.199115,2.026074,0.300945,
        -2.199115,2.026196,-2.889057,
        -2.199115,2.031959,-1.268178,
        -2.199115,2.032917,-2.199917,
        -2.199115,2.033572,-1.842884,
        -2.199115,2.034712,-2.419505,
        -2.199115,2.035818,-0.379479,
        -2.199115,2.036004,-0.757062,
        -2.199115,2.040501,-1.411488,
        -2.199115,2.042874,-1.701367,
        -2.199115,2.043060,-2.653746,
        -2.199115,2.043729,-1.557938,
        -2.199115,2.048538,0.806623,
        -2.199115,2.049006,1.199930,
        -2.199115,2.049502,-0.137989,
        -2.199115,2.052181,2.326959,
        -2.199115,2.052794,-0.008526,
        -2.199115,2.057608,1.951726,
        -2.199115,2.065568,2.567730,
        -2.199115,2.070371,2.879808,
        -2.199115,2.071256,-3.018961,
        -2.199115,2.079746,-0.897699,
        -2.199115,2.082784,1.042403,
        -2.199115,2.083752,-0.511086,
        -2.199115,2.084401,1.810502,
        -2.199115,2.085731,0.671081,
        -2.199115,2.097496,1.373172,
        -2.199115,2.099185,1.667006,
        -2.199115,2.100482,-2.301084,
        -2.199115,2.103118,2.194745,
        -2.199115,2.103223,1.521685,
        -2.199115,2.103695,-2.777707,
        -2.199115,2.107523,3.129581,
        -2.199115,2.108277,-2.095907,
        -2.199115,2.108891,-2.532166,
        -2.199115,2.111003,-0.268632,
        -2.199115,2.112276,-1.027246,
        -2.199115,2.114380,0.531457,
        -2.199115,2.119166,0.100552,
        -2.199115,2.119564,2.754818,
        -2.199115,2.122474,-0.648314,
        -2.199115,2.125183,2.445598,
        -2.199115,2.129501,0.238290,
        -2.199115,2.131178,0.386663,
        -2.199115,2.131223,-1.174311,
        -2.199115,2.134269,-1.948042,
        -2.199115,2.134684,0.907065,
        -2.199115,2.137615,2.988806,
        -2.199115,2.145635,-1.327708,
        -2.199115,2.146467,2.055081,
        -2.199115,2.146582,-1.793788,
        -2.199115,2.149210,-1.482321,
        -2.199115,2.152901,-1.636788,
        -2.199115,2.156337,1.151558,
        -2.199115,2.156474,-2.912065,
        -2.199115,2.157194,-0.792563,
        -2.199115,2.166724,-0.401638,
        -2.199115,2.166980,-0.134879,
        -2.199115,2.173644,-2.403429,
        -2.199115,2.175711,-2.655523,
        -2.199115,2.177132,2.600419,
        -2.199115,2.178879,0.767722,
        -2.199115,2.180752,1.908485,
        -2.199115,2.181338,2.313721,
        -2.199115,2.185580,1.287673,
        -2.199115,2.195400,-2.217546,
        -2.199115,2.200418,-3.056442,
        -2.199115,2.201829,-0.001735,
        -2.199115,2.203804,1.756176,
        -2.199115,2.211602,1.443930,
        -2.199115,2.212686,-0.544809,
        -2.199115,2.213937,0.621892,
        -2.199115,2.214075,1.600691,
        -2.199115,2.216256,1.020029,
        -2.199115,2.219060,2.886631,
        -2.199115,2.223375,-0.906185,
        -2.199115,2.226002,-1.060248,
        -2.199115,2.226413,-2.064034,
        -2.199115,2.230554,2.171334,
        -2.199115,2.232413,2.735777,
        -2.199115,2.233709,3.070830,
        -2.199115,2.234383,-2.792065,
        -2.199115,2.237183,0.468799,
        -2.199115,2.237695,-0.271388,
        -2.199115,2.239419,0.151693,
        -2.199115,2.243347,-2.521049,
        -2.199115,2.244035,-1.229456,
        -2.199115,2.244355,0.309899,
        -2.199115,2.245175,2.460236,
        -2.199115,2.245628,-1.899216,
        -2.199115,2.245973,-0.702234,
        -2.199115,2.255601,-1.728131,
        -2.199115,2.256702,-1.401367,
        -2.199115,2.261466,-1.557500,
        -2.199115,2.267390,0.878499,
        -2.199115,2.272099,2.018594,
        -2.199115,2.278618,1.170386,
        -2.199115,2.283304,-2.345584,
        -2.199115,2.285175,-2.943090,
        -2.199115,2.293000,-0.124451,
        -2.199115,2.294058,-0.421070,
        -2.199115,2.303283,1.855938,
        -2.199115,2.304968,2.608978,
        -2.199115,2.305347,2.311346,
        -2.199115,2.307188,-2.656134,
        -2.199115,2.307301,1.336200,
        -2.199115,2.309655,0.726020,
        -2.199115,2.317634,-2.185225,
        -2.199115,2.321510,1.685664,
        -2.199115,2.324376,-3.111006,
        -2.199115,2.324731,1.511186,
        -2.199115,2.329587,0.035498,
        -2.199115,2.330158,2.977264,
        -2.199115,2.335096,-0.588468,
        -2.199115,2.336046,-1.115799,
        -2.199115,2.338060,2.803613,
        -2.199115,2.338083,-0.943400,
        -2.199115,2.340210,0.562145,
        -2.199115,2.341931,1.023926,
        -2.199115,2.343265,-0.779061,
        -2.199115,2.343434,-2.013692,
        -2.199115,2.353029,-1.310529,
        -2.199115,2.354868,0.211543,
        -2.199115,2.355120,0.388593,
        -2.199115,2.355165,-1.833173,
        -2.199115,2.355462,2.151441,
        -2.199115,2.362948,-2.487477,
        -2.199115,2.363101,-2.810456,
        -2.199115,2.363394,-0.272918,
        -2.199115,2.364509,-1.644533,
        -2.199115,2.370465,2.466824,
        -2.199115,2.386110,-1.481351,
        -2.199115,2.394960,0.862133,
        -2.199115,2.395716,1.977184,
        -2.199115,2.396339,1.200900,
        -2.199115,2.407941,-2.316753,
        -2.199115,2.409450,-2.988432,
        -2.199115,2.417202,-0.103363,
        -2.199115,2.417386,-0.443097,
        -2.199115,2.417950,2.666662,
        -2.199115,2.423190,1.789783,
        -2.199115,2.423692,1.392696,
        -2.199115,2.431122,3.086379,
        -2.199115,2.431406,-2.647820,
        -2.199115,2.431584,2.307631,
        -2.199115,2.434090,1.592818,
        -2.199115,2.435993,0.683844,
        -2.199115,2.440898,-2.133300,
        -2.199115,2.442222,-1.195837,
        -2.199115,2.443112,2.883761,
        -2.199115,2.449191,-0.633651,
        -2.199115,2.450821,0.084741,
        -2.199115,2.453499,-1.008861,
        -2.199115,2.460010,-1.938011,
        -2.199115,2.460199,-0.822008,
        -2.199115,2.460902,0.490079,
        -2.199115,2.461774,-1.734913,
        -2.199115,2.463986,1.034407,
        -2.199115,2.467647,0.287667,
        -2.199115,2.479849,2.126925,
        -2.199115,2.485907,-2.835471,
        -2.199115,2.485945,-1.374853,
        -2.199115,2.487747,2.505912,
        -2.199115,2.487957,-0.274279,
        -2.199115,2.489741,-2.468428,
        -2.199115,2.512538,1.239900,
        -2.199115,2.517528,1.923289,
        -2.199115,2.518454,-1.565852,
        -2.199115,2.519348,0.843503,
        -2.199115,2.522766,-3.059231,
        -2.199115,2.528686,2.737228,
        -2.199115,2.534798,-2.266671,
        -2.199115,2.536533,1.466274,
        -2.199115,2.538504,1.699812,
        -2.199115,2.539516,-0.476942,
        -2.199115,2.539532,-0.074548,
        -2.199115,2.543197,2.990252,
        -2.199115,2.554022,2.320619,
        -2.199115,2.556919,-2.655005,
        -2.199115,2.558146,0.627191,
        -2.199115,2.558223,-1.164189,
        -2.199115,2.563452,-2.044028,
        -2.199115,2.568462,0.150655,
        -2.199115,2.569130,-0.700837,
        -2.199115,2.570050,-1.804492,
        -2.199115,2.572710,-0.934665,
        -2.199115,2.575994,0.391718,
        -2.199115,2.584774,1.046703,
        -2.199115,2.601432,2.102535,
        -2.199115,2.603559,2.555114,
        -2.199115,2.604150,-2.885208,
        -2.199115,2.608630,-1.378011,
        -2.199115,2.611256,-0.279110,
        -2.199115,2.617344,-2.439480,
        -2.199115,2.627402,1.293552,
        -2.199115,2.635396,1.848850,
        -2.199115,2.635686,2.834950,
        -2.199115,2.637087,3.132730,
        -2.199115,2.642029,0.813274,
        -2.199115,2.645260,1.570203,
        -2.199115,2.646182,-1.614036,
        -2.199115,2.660113,-2.186513,
        -2.199115,2.660483,-0.034608,
        -2.199115,2.661713,-0.529511,
        -2.199115,2.672720,-1.110422,
        -2.199115,2.676626,2.335353,
        -2.199115,2.676638,0.540364,
        -2.199115,2.676780,-1.901158,
        -2.199115,2.681411,-2.667965,
        -2.199115,2.682152,0.245751,
        -2.199115,2.682859,-0.817071,
        -2.199115,2.705107,1.062142,
        -2.199115,2.717705,2.625995,
        -2.199115,2.721001,2.060097,
        -2.199115,2.722319,-2.954069,
        -2.199115,2.729840,-1.368237,
        -2.199115,2.734349,-0.289298,
        -2.199115,2.738409,2.975671,
        -2.199115,2.740366,1.374341,
        -2.199115,2.744633,-2.391311,
        -2.199115,2.747398,1.727166,
        -2.199115,2.763799,0.760372,
        -2.199115,2.768238,-1.679243,
        -2.199115,2.779245,-2.046268,
        -2.199115,2.780208,-0.626453,
        -2.199115,2.780471,0.027732,
        -2.199115,2.784696,-1.013128,
        -2.199115,2.789817,0.395431,
        -2.199115,2.801154,2.353143,
        -2.199115,2.807091,-2.687477,
        -2.199115,2.825762,1.085035,
        -2.199115,2.830209,2.734943,
        -2.199115,2.833517,-3.085508,
        -2.199115,2.838301,1.980061,
        -2.199115,2.849871,1.512012,
        -2.199115,2.850089,-1.343899,
        -2.199115,2.856905,-0.318441,
        -2.199115,2.870445,-2.291001,
        -2.199115,2.885017,0.655562,
        -2.199115,2.886428,-1.793217,
        -2.199115,2.890337,-0.825633,
        -2.199115,2.898932,0.139973,
        -2.199115,2.931927,-2.728768,
        -2.199115,2.932096,2.377790,
        -2.199115,2.937409,2.947245,
        -2.199115,2.948048,1.133127,
        -2.199115,2.950896,1.797889,
        -2.199115,2.970019,-1.282382,
        -2.199115,2.978333,-0.412090,
        -2.199115,2.996437,-2.089121,
        -2.199115,3.011274,0.434110,
        -2.199115,3.050410,3.138492,
        -2.199115,3.064492,1.647259,
        -2.199115,3.088868,-0.989501,
        -2.094395,0.053884,-2.408394,
        -2.094395,0.077645,1.453433,
        -2.094395,0.088408,-0.311160,
        -2.094395,0.142314,2.590387,
        -2.094395,0.147206,-1.385164,
        -2.094395,0.163506,0.573203,
        -2.094395,0.165485,-2.875396,
        -2.094395,0.182564,-2.125821,
        -2.094395,0.190569,1.929141,
        -2.094395,0.201102,1.249082,
        -2.094395,0.201397,-0.716646,
        -2.094395,0.210568,-0.069559,
        -2.094395,0.251672,2.848900,
        -2.094395,0.266262,-1.678068,
        -2.094395,0.269108,-1.215563,
        -2.094395,0.273777,-2.506756,
        -2.094395,0.279854,0.814581,
        -2.094395,0.281664,-2.969147,
        -2.094395,0.283625,2.405603,
        -2.094395,0.284058,0.357350,
        -2.094395,0.294107,1.593880,
        -2.094395,0.302491,-0.400410,
        -2.094395,0.308059,2.021491,
        -2.094395,0.312219,-2.072057,
        -2.094395,0.321905,-0.829121,
        -2.094395,0.329984,1.201621,
        -2.094395,0.339714,-0.013712,
        -2.094395,0.368784,2.974974,
        -2.094395,0.376428,-1.495532,
        -2.094395,0.378103,2.646436,
        -2.094395,0.385088,0.590355,
        -2.094395,0.386248,-2.690943,
        -2.094395,0.389851,-2.340031,
        -2.094395,0.391502,-1.171374,
        -2.094395,0.398249,-1.812694,
        -2.094395,0.400542,-2.995729,
        -2.094395,0.404273,0.918163,
        -2.094395,0.405617,1.773533,
        -2.094395,0.409009,-0.570243,
        -2.094395,0.410700,0.270433,
        -2.094395,0.415893,1.464890,
        -2.094395,0.420680,2.088164,
        -2.094395,0.424766,-0.258729,
        -2.094395,0.426507,2.367872,
        -2.094395,0.441292,-0.890727,
        -2.094395,0.451839,-2.078193,
        -2.094395,0.459270,1.191562,
        -2.094395,0.470622,0.014002,
        -2.094395,0.481292,-2.529098,
        -2.094395,0.482460,2.849843,
        -2.094395,0.492205,-1.397963,
        -2.094395,0.493314,3.117264,
        -2.094395,0.497870,0.727828,
        -2.094395,0.499783,-1.654379,
        -2.094395,0.502651,0.463333,
        -2.094395,0.505463,2.594557,
        -2.094395,0.506104,-2.874910,
        -2.094395,0.513579,-1.147354,
        -2.094395,0.518569,1.644237,
        -2.094395,0.519001,1.900488,
        -2.094395,0.519277,-0.673809,
        -2.094395,0.524769,-0.421955,
        -2.094395,0.525564,-2.288838,
        -2.094395,0.532541,0.978737,
        -2.094395,0.534610,2.154808,
        -2.094395,0.535020,-1.891846,
        -2.094395,0.540208,1.400497,
        -2.094395,0.542006,0.228337,
        -2.094395,0.550762,-0.180771,
        -2.094395,0.560531,2.378474,
        -2.094395,0.561629,-0.924576,
        -2.094395,0.567465,-2.679392,
        -2.094395,0.588527,1.181755,
        -2.094395,0.591643,-2.098364,
        -2.094395,0.594963,2.997575,
        -2.094395,0.599643,-3.045235,
        -2.094395,0.599962,2.773556,
        -2.094395,0.602620,0.029270,
        -2.094395,0.605711,0.598454,
        -2.094395,0.607708,-1.541026,
        -2.094395,0.609360,-2.464379,
        -2.094395,0.610738,-1.326786,
        -2.094395,0.616968,0.820489,
        -2.094395,0.626379,1.773610,
        -2.094395,0.628626,-1.751180,
        -2.094395,0.630671,0.391949,
        -2.094395,0.632622,-0.737642,
        -2.094395,0.632708,2.566722,
        -2.094395,0.632797,1.989858,
        -2.094395,0.634455,-0.534790,
        -2.094395,0.636205,1.561170,
        -2.094395,0.637681,-1.123423,
        -2.094395,0.641441,-2.842109,
        -2.094395,0.644807,-0.327139,
        -2.094395,0.652440,2.197129,
        -2.094395,0.660897,-2.278593,
        -2.094395,0.664008,1.360091,
        -2.094395,0.670850,-1.945257,
        -2.094395,0.671396,1.004859,
        -2.094395,0.673985,0.203855,
        -2.094395,0.680326,-0.133370,
        -2.094395,0.685083,-0.942363,
        -2.094395,0.688204,-2.632204,
        -2.094395,0.689777,2.385011,
        -2.094395,0.700065,3.115430,
        -2.094395,0.704080,2.911636,
        -2.094395,0.712883,1.174216,
        -2.094395,0.715701,0.699906,
        -2.094395,0.718851,-1.442215,
        -2.094395,0.721019,2.721536,
        -2.094395,0.725300,-2.975134,
        -2.094395,0.727136,-1.633288,
        -2.094395,0.727399,-2.120567,
        -2.094395,0.730751,0.522488,
        -2.094395,0.732892,-1.254330,
        -2.094395,0.734333,-2.446184,
        -2.094395,0.735125,0.037270,
        -2.094395,0.737893,1.866463,
        -2.094395,0.740650,1.681075,
        -2.094395,0.745971,-0.781286,
        -2.094395,0.749432,-0.442559,
        -2.094395,0.750102,2.050636,
        -2.094395,0.751046,0.863821,
        -2.094395,0.757018,-0.608833,
        -2.094395,0.757046,-1.816019,
        -2.094395,0.757389,1.501184,
        -2.094395,0.759800,2.546181,
        -2.094395,0.759859,0.351923,
        -2.094395,0.761141,-2.781102,
        -2.094395,0.769035,-1.078574,
        -2.094395,0.770166,-0.266604,
        -2.094395,0.772218,2.227380,
        -2.094395,0.788382,1.331051,
        -2.094395,0.790269,-2.282409,
        -2.094395,0.803410,-1.984567,
        -2.094395,0.803813,0.191520,
        -2.094395,0.806832,3.022155,
        -2.094395,0.807476,-2.597675,
        -2.094395,0.812512,-0.107335,
        -2.094395,0.814161,1.010949,
        -2.094395,0.817732,-3.084486,
        -2.094395,0.817846,2.390688,
        -2.094395,0.819323,2.845960,
        -2.094395,0.819430,-0.915017,
        -2.094395,0.829198,-1.532491,
        -2.094395,0.829531,0.665695,
        -2.094395,0.829816,1.167147,
        -2.094395,0.833613,-1.357492,
        -2.094395,0.842945,-2.907318,
        -2.094395,0.844883,2.679708,
        -2.094395,0.848189,-1.705003,
        -2.094395,0.849888,1.767901,
        -2.094395,0.851126,0.493370,
        -2.094395,0.852911,1.933394,
        -2.094395,0.857516,-2.432863,
        -2.094395,0.858261,-1.191729,
        -2.094395,0.858279,-2.138895,
        -2.094395,0.859723,-0.734928,
        -2.094395,0.859754,1.603292,
        -2.094395,0.861655,-0.394821,
        -2.094395,0.864300,0.043549,
        -2.094395,0.866940,0.851229,
        -2.094395,0.869748,2.095969,
        -2.094395,0.877352,-0.574668,
        -2.094395,0.879254,-2.733338,
        -2.094395,0.882554,1.443757,
        -2.094395,0.884060,0.334199,
        -2.094395,0.885626,-1.866061,
        -2.094395,0.889834,2.529957,
        -2.094395,0.893140,2.251698,
        -2.094395,0.897463,-0.240822,
        -2.094395,0.898758,-1.031948,
        -2.094395,0.913918,1.287618,
        -2.094395,0.916555,-2.287173,
        -2.094395,0.918305,3.107845,
        -2.094395,0.919522,2.948277,
        -2.094395,0.927138,-2.569686,
        -2.094395,0.929031,0.185383,
        -2.094395,0.933766,-3.015917,
        -2.094395,0.934873,-2.010822,
        -2.094395,0.935636,-1.445855,
        -2.094395,0.936702,-0.866461,
        -2.094395,0.937395,0.729108,
        -2.094395,0.937910,2.787575,
        -2.094395,0.939144,-0.092962,
        -2.094395,0.941718,1.125085,
        -2.094395,0.942873,-1.607680,
        -2.094395,0.943701,0.980328,
        -2.094395,0.947774,2.395256,
        -2.094395,0.952433,0.584347,
        -2.094395,0.954456,-1.287293,
        -2.094395,0.958731,-2.852552,
        -2.094395,0.961838,1.839284,
        -2.094395,0.961898,-0.469425,
        -2.094395,0.965426,1.686615,
        -2.094395,0.970108,1.990771,
        -2.094395,0.972820,-1.762780,
        -2.094395,0.973500,-0.711881,
        -2.094395,0.975253,2.641082,
        -2.094395,0.977824,0.441928,
        -2.094395,0.981485,-2.424766,
        -2.094395,0.981877,1.537100,
        -2.094395,0.982609,-1.136920,
        -2.094395,0.984265,-2.156073,
        -2.094395,0.989150,0.045493,
        -2.094395,0.993052,2.137478,
        -2.094395,0.997766,-2.693871,
        -2.094395,0.999847,-0.334738,
        -2.094395,1.001920,0.848054,
        -2.094395,1.006146,1.389734,
        -2.094395,1.012180,0.304066,
        -2.094395,1.016938,2.270674,
        -2.094395,1.017371,-0.979790,
        -2.094395,1.019495,-1.894487,
        -2.094395,1.023796,-0.583691,
        -2.094395,1.026525,2.506945,
        -2.094395,1.027004,-0.203653,
        -2.094395,1.028027,3.036573,
        -2.094395,1.031435,1.241021,
        -2.094395,1.032316,2.884221,
        -2.094395,1.035863,-3.110475,
        -2.094395,1.042085,-2.292950,
        -2.094395,1.043272,-1.522986,
        -2.094395,1.046265,-2.958847,
        -2.094395,1.049272,-1.376344,
        -2.094395,1.050039,-2.552086,
        -2.094395,1.053682,0.689863,
        -2.094395,1.055458,0.170394,
        -2.094395,1.056622,-0.831866,
        -2.094395,1.057515,1.097998,
        -2.094395,1.058706,-2.036438,
        -2.094395,1.061389,-1.672248,
        -2.094395,1.063758,2.743093,
        -2.094395,1.070157,-0.073422,
        -2.094395,1.072154,1.761506,
        -2.094395,1.074029,1.909193,
        -2.094395,1.074204,0.545276,
        -2.094395,1.074714,-2.805147,
        -2.094395,1.076262,0.966104,
        -2.094395,1.077642,-1.227073,
        -2.094395,1.084151,1.618525,
        -2.094395,1.085085,2.377673,
        -2.094395,1.086822,-0.441112,
        -2.094395,1.087142,2.058815,
        -2.094395,1.098383,-0.696430,
        -2.094395,1.099138,-1.084033,
        -2.094395,1.103208,1.477773,
        -2.094395,1.103783,0.408254,
        -2.094395,1.106275,-2.420800,
        -2.094395,1.108079,2.612943,
        -2.094395,1.109345,-2.172718,
        -2.094395,1.115957,-1.787597,
        -2.094395,1.118930,0.045983,
        -2.094395,1.119777,0.834702,
        -2.094395,1.121943,-2.669042,
        -2.094395,1.124930,2.219917,
        -2.094395,1.125217,1.336294,
        -2.094395,1.130958,2.972931,
        -2.094395,1.131400,-0.306629,
        -2.094395,1.137135,-0.936514,
        -2.094395,1.139125,-1.924528,
        -2.094395,1.139877,0.276382,
        -2.094395,1.142003,-3.058022,
        -2.094395,1.147363,-0.558448,
        -2.094395,1.150348,-0.176418,
        -2.094395,1.151198,1.197016,
        -2.094395,1.153930,-1.597104,
        -2.094395,1.155389,2.837148,
        -2.094395,1.155458,-1.460623,
        -2.094395,1.157331,-2.906362,
        -2.094395,1.158594,2.487618,
        -2.094395,1.162103,-1.322553,
        -2.094395,1.165407,3.092861,
        -2.094395,1.168025,0.717075,
        -2.094395,1.168406,-2.301519,
        -2.094395,1.174782,-2.538755,
        -2.094395,1.177998,1.061256,
        -2.094395,1.178898,1.827305,
        -2.094395,1.181331,-2.060330,
        -2.094395,1.182540,-0.798952,
        -2.094395,1.184569,0.147517,
        -2.094395,1.184734,1.968919,
        -2.094395,1.184804,0.593386,
        -2.094395,1.188278,1.690808,
        -2.094395,1.194360,2.110440,
        -2.094395,1.194776,2.710455,
        -2.094395,1.197745,-1.173705,
        -2.094395,1.199898,-2.777257,
        -2.094395,1.201164,2.355931,
        -2.094395,1.203760,1.556404,
        -2.094395,1.205845,-0.415993,
        -2.094395,1.206585,0.933648,
        -2.094395,1.206950,-0.037248,
        -2.094395,1.209341,0.471162,
        -2.094395,1.213483,-1.039417,
        -2.094395,1.218146,-0.672643,
        -2.094395,1.221608,1.421669,
        -2.094395,1.226043,-1.693533,
        -2.094395,1.232870,-1.822773,
        -2.094395,1.234375,-2.421194,
        -2.094395,1.234531,-2.190148,
        -2.094395,1.239414,0.348959,
        -2.094395,1.240183,2.587255,
        -2.094395,1.242312,2.941234,
        -2.094395,1.244075,1.286644,
        -2.094395,1.245961,-2.997461,
        -2.094395,1.246495,-2.646872,
        -2.094395,1.247736,2.242810,
        -2.094395,1.253604,0.812222,
        -2.094395,1.257816,-0.894317,
        -2.094395,1.258247,-1.953634,
        -2.094395,1.258993,-0.284099,
        -2.094395,1.262380,-0.157609,
        -2.094395,1.262791,-1.549229,
        -2.094395,1.263772,-0.533073,
        -2.094395,1.266731,-1.412461,
        -2.094395,1.268591,-3.122146,
        -2.094395,1.269614,-1.275246,
        -2.094395,1.270745,1.152533,
        -2.094395,1.273606,0.224545,
        -2.094395,1.279716,2.812101,
        -2.094395,1.283786,2.462124,
        -2.094395,1.284798,0.078209,
        -2.094395,1.285402,-2.882049,
        -2.094395,1.286936,1.886053,
        -2.094395,1.289408,0.683122,
        -2.094395,1.294736,1.756184,
        -2.094395,1.296978,-2.311511,
        -2.094395,1.298546,2.020601,
        -2.094395,1.300924,1.022478,
        -2.094395,1.303305,3.043152,
        -2.094395,1.304210,-2.084225,
        -2.094395,1.305241,-2.530249,
        -2.094395,1.307513,1.628003,
        -2.094395,1.310722,0.558267,
        -2.094395,1.314142,2.139329,
        -2.094395,1.316087,-0.777971,
        -2.094395,1.317907,-1.121023,
        -2.094395,1.321019,-2.749776,
        -2.094395,1.321609,2.687717,
        -2.094395,1.322568,1.499415,
        -2.094395,1.322972,-0.395699,
        -2.094395,1.325286,-0.994192,
        -2.094395,1.329146,-0.648513,
        -2.094395,1.329795,-0.042570,
        -2.094395,1.331651,2.344497,
        -2.094395,1.336522,0.435098,
        -2.094395,1.339410,0.900087,
        -2.094395,1.339550,1.370353,
        -2.094395,1.340441,-1.731580,
        -2.094395,1.351855,-1.855234,
        -2.094395,1.360161,-3.018404,
        -2.094395,1.361031,-1.613971,
        -2.094395,1.361196,2.914780,
        -2.094395,1.361998,-2.207254,
        -2.094395,1.362302,1.239384,
        -2.094395,1.365228,2.563808,
        -2.094395,1.365452,0.310522,
        -2.094395,1.367028,-2.420041,
        -2.094395,1.373422,-1.484281,
        -2.094395,1.374245,-1.226417,
        -2.094395,1.375020,-0.170393,
        -2.094395,1.376560,-1.355481,
        -2.094395,1.377070,-1.980984,
        -2.094395,1.377578,-2.636339,
        -2.094395,1.378096,0.774626,
        -2.094395,1.379086,-0.508192,
        -2.094395,1.382562,0.177141,
        -2.094395,1.388849,2.233847,
        -2.094395,1.389268,-0.886349,
        -2.094395,1.391324,1.109396,
        -2.094395,1.391699,-0.288725,
        -2.094395,1.392869,3.136903,
        -2.094395,1.393612,1.952016,
        -2.094395,1.395545,-2.850807,
        -2.094395,1.401926,1.822930,
        -2.094395,1.402149,2.788898,
        -2.094395,1.409552,0.647980,
        -2.094395,1.410840,0.056219,
        -2.094395,1.411944,2.444377,
        -2.094395,1.413639,1.697982,
        -2.094395,1.426479,1.573427,
        -2.094395,1.426725,2.102242,
        -2.094395,1.427868,0.984637,
        -2.094395,1.429580,-2.107232,
        -2.094395,1.431375,-2.313269,
        -2.094395,1.431907,3.015410,
        -2.094395,1.432325,-1.029054,
        -2.094395,1.433341,0.524268,
        -2.094395,1.435181,-0.739771,
        -2.094395,1.437208,-2.529085,
        -2.094395,1.439041,-0.617966,
        -2.094395,1.441473,1.448544,
        -2.094395,1.444527,2.664744,
        -2.094395,1.450513,-0.400625,
        -2.094395,1.451907,-0.066823,
        -2.094395,1.452754,-2.738631,
        -2.094395,1.454061,-2.950416,
        -2.094395,1.456736,1.323154,
        -2.094395,1.458370,0.400433,
        -2.094395,1.459314,-1.765003,
        -2.094395,1.459902,-1.145835,
        -2.094395,1.464952,2.329245,
        -2.094395,1.466050,0.862104,
        -2.094395,1.471330,-1.886160,
        -2.094395,1.475119,-1.648255,
        -2.094395,1.480878,1.194576,
        -2.094395,1.480942,0.273235,
        -2.094395,1.481086,2.890269,
        -2.094395,1.484118,-1.529156,
        -2.094395,1.484741,-1.279219,
        -2.094395,1.487681,-1.406069,
        -2.094395,1.487886,-3.066037,
        -2.094395,1.490555,2.543284,
        -2.094395,1.492424,-0.190932,
        -2.094395,1.495002,-0.839810,
        -2.094395,1.495297,2.007188,
        -2.094395,1.496414,-2.009267,
        -2.094395,1.498261,-2.425270,
        -2.094395,1.498907,-2.207327,
        -2.094395,1.500204,0.737638,
        -2.094395,1.502469,0.151156,
        -2.094395,1.507448,-0.514782,
        -2.094395,1.509287,2.204647,
        -2.094395,1.510518,1.884920,
        -2.094395,1.511948,-2.632360,
        -2.094395,1.515235,1.068565,
        -2.094395,1.522237,3.102018,
        -2.094395,1.522373,1.764991,
        -2.094395,1.522416,-0.954956,
        -2.094395,1.523102,2.765920,
        -2.094395,1.524811,-2.844718,
        -2.094395,1.527261,-0.304916,
        -2.094395,1.529788,0.613208,
        -2.094395,1.533648,1.644808,
        -2.094395,1.535018,0.028869,
        -2.094395,1.541082,2.424608,
        -2.094395,1.546010,1.523462,
        -2.094395,1.551135,-0.642994,
        -2.094395,1.553895,0.489607,
        -2.094395,1.555041,0.947813,
        -2.094395,1.559671,-1.074934,
        -2.094395,1.560345,2.985899,
        -2.094395,1.560487,-2.323400,
        -2.094395,1.561127,1.401111,
        -2.094395,1.567193,2.643172,
        -2.094395,1.568892,2.098362,
        -2.094395,1.571702,-2.098435,
        -2.094395,1.572701,-2.528730,
        -2.094395,1.573572,1.278593,
        -2.094395,1.574849,-0.091505,
        -2.094395,1.576334,0.365236,
        -2.094395,1.577029,-2.962888,
        -2.094395,1.577688,-1.814941,
        -2.094395,1.579542,-0.755917,
        -2.094395,1.579624,-1.199759,
        -2.094395,1.581515,-0.419414,
        -2.094395,1.584581,-2.738702,
        -2.094395,1.586948,-1.696296,
        -2.094395,1.588901,2.301496,
        -2.094395,1.589936,0.825020,
        -2.094395,1.590018,-1.941652,
        -2.094395,1.596968,-1.328358,
        -2.094395,1.597178,-1.575607,
        -2.094395,1.600470,-1.453196,
        -2.094395,1.601408,0.242409,
        -2.094395,1.602259,1.152656,
        -2.094395,1.603064,2.865774,
        -2.094395,1.612085,1.968802,
        -2.094395,1.612720,-3.080498,
        -2.094395,1.616061,2.521531,
        -2.094395,1.616900,-0.207457,
        -2.094395,1.620988,-0.875920,
        -2.094395,1.622114,0.701509,
        -2.094395,1.622749,0.121426,
        -2.094395,1.622993,-2.221629,
        -2.094395,1.629346,-0.542669,
        -2.094395,1.630007,1.844552,
        -2.094395,1.634747,-2.426198,
        -2.094395,1.642346,1.722617,
        -2.094395,1.643925,1.034629,
        -2.094395,1.644848,2.743774,
        -2.094395,1.645258,2.188109,
        -2.094395,1.645417,-2.634725,
        -2.094395,1.647395,3.085437,
        -2.094395,1.648538,-2.855044,
        -2.094395,1.649932,0.576808,
        -2.094395,1.652224,-0.996105,
        -2.094395,1.652871,1.600681,
        -2.094395,1.657997,-0.325127,
        -2.094395,1.661936,0.001037,
        -2.094395,1.665187,1.477402,
        -2.094395,1.665496,2.397872,
        -2.094395,1.673267,-0.665453,
        -2.094395,1.673557,0.452526,
        -2.094395,1.679277,-1.122421,
        -2.094395,1.679696,0.912482,
        -2.094395,1.682420,1.353922,
        -2.094395,1.685031,-2.009038,
        -2.094395,1.685196,2.967046,
        -2.094395,1.690107,2.620910,
        -2.094395,1.690453,-2.123076,
        -2.094395,1.691693,-1.871572,
        -2.094395,1.691817,2.061688,
        -2.094395,1.692644,1.237670,
        -2.094395,1.696386,0.328123,
        -2.094395,1.697162,-1.249026,
        -2.094395,1.697488,-2.322785,
        -2.094395,1.698726,-2.973576,
        -2.094395,1.699172,-1.749717,
        -2.094395,1.704478,-0.106295,
        -2.094395,1.707244,-0.444770,
        -2.094395,1.707500,-2.531497,
        -2.094395,1.707714,-0.787321,
        -2.094395,1.708430,-1.627552,
        -2.094395,1.710583,-1.376890,
        -2.094395,1.712415,-2.747921,
        -2.094395,1.713125,-1.503073,
        -2.094395,1.713452,0.789419,
        -2.094395,1.720612,2.279888,
        -2.094395,1.724852,0.202831,
        -2.094395,1.724868,2.846246,
        -2.094395,1.728940,1.931535,
        -2.094395,1.732838,1.127584,
        -2.094395,1.735719,-3.092485,
        -2.094395,1.739676,2.496017,
        -2.094395,1.739971,-0.228876,
        -2.094395,1.743328,0.663347,
        -2.094395,1.744927,-0.911892,
        -2.094395,1.745952,0.084733,
        -2.094395,1.748077,1.804354,
        -2.094395,1.752959,-0.567548,
        -2.094395,1.760169,1.678132,
        -2.094395,1.763887,-2.218450,
        -2.094395,1.766737,2.722752,
        -2.094395,1.767999,1.004737,
        -2.094395,1.769149,1.550298,
        -2.094395,1.770119,-2.427175,
        -2.094395,1.770460,2.154208,
        -2.094395,1.770514,3.070982,
        -2.094395,1.770735,0.537245,
        -2.094395,1.772673,-2.864760,
        -2.094395,1.773527,-1.039582,
        -2.094395,1.776419,-2.642421,
        -2.094395,1.780769,1.416330,
        -2.094395,1.786634,-0.346247,
        -2.094395,1.788829,-1.947553,
        -2.094395,1.793737,0.410270,
        -2.094395,1.794553,2.374618,
        -2.094395,1.794699,-0.694061,
        -2.094395,1.797469,-1.169105,
        -2.094395,1.803705,0.881076,
        -2.094395,1.804124,1.264249,
        -2.094395,1.807825,-2.090455,
        -2.094395,1.808822,-1.817375,
        -2.094395,1.809020,2.949741,
        -2.094395,1.812466,2.597595,
        -2.094395,1.813159,2.025395,
        -2.094395,1.813816,-1.298435,
        -2.094395,1.815016,0.278951,
        -2.094395,1.816510,-0.008631,
        -2.094395,1.817772,-1.689625,
        -2.094395,1.821360,-2.984352,
        -2.094395,1.822312,-0.127211,
        -2.094395,1.823386,-1.429807,
        -2.094395,1.823834,-1.560483,
        -2.094395,1.832060,-0.821627,
        -2.094395,1.832690,-0.468440,
        -2.094395,1.833945,-2.320278,
        -2.094395,1.834787,0.751968,
        -2.094395,1.840579,-2.536445,
        -2.094395,1.841033,-2.756932,
        -2.094395,1.846350,1.893983,
        -2.094395,1.846644,1.113766,
        -2.094395,1.846833,2.826543,
        -2.094395,1.847325,2.249051,
        -2.094395,1.854119,0.134161,
        -2.094395,1.858863,-3.106022,
        -2.094395,1.865046,1.763384,
        -2.094395,1.865619,-0.952036,
        -2.094395,1.865912,0.624453,
        -2.094395,1.866694,2.473686,
        -2.094395,1.867524,-0.240083,
        -2.094395,1.875561,1.632831,
        -2.094395,1.876049,-0.594378,
        -2.094395,1.880509,1.500204,
        -2.094395,1.883504,1.360457,
        -2.094395,1.886971,-2.196665,
        -2.094395,1.888979,0.982922,
        -2.094395,1.889815,2.701570,
        -2.094395,1.892144,-2.005340,
        -2.094395,1.892348,0.496522,
        -2.094395,1.893281,-1.083976,
        -2.094395,1.893828,3.053453,
        -2.094395,1.895117,2.121107,
        -2.094395,1.898741,-2.875735,
        -2.094395,1.904429,-2.428029,
        -2.094395,1.908408,-2.648903,
        -2.094395,1.911828,-0.364489,
        -2.094395,1.912224,0.366565,
        -2.094395,1.914854,1.214468,
        -2.094395,1.915467,-1.883143,
        -2.094395,1.915763,-0.724679,
        -2.094395,1.915790,-1.216366,
        -2.094395,1.922256,2.347806,
        -2.094395,1.922616,0.844633,
        -2.094395,1.925185,0.231170,
        -2.094395,1.927474,-1.754616,
        -2.094395,1.928342,-1.351174,
        -2.094395,1.933328,-1.623328,
        -2.094395,1.933878,0.034150,
        -2.094395,1.934585,-1.487979,
        -2.094395,1.934802,2.927634,
        -2.094395,1.934862,1.989776,
        -2.094395,1.936327,-0.099350,
        -2.094395,1.938088,2.578507,
        -2.094395,1.945677,-2.999805,
        -2.094395,1.954572,-0.859208,
        -2.094395,1.957868,-0.490784,
        -2.094395,1.958989,0.712772,
        -2.094395,1.960212,-2.307561,
        -2.094395,1.964774,1.855507,
        -2.094395,1.964971,-2.100533,
        -2.094395,1.969112,1.088273,
        -2.094395,1.971438,-2.765415,
        -2.094395,1.972312,2.805134,
        -2.094395,1.974950,2.220596,
        -2.094395,1.975135,-2.539127,
        -2.094395,1.982107,1.720829,
        -2.094395,1.983195,-3.127984,
        -2.094395,1.983318,-0.247182,
        -2.094395,1.986720,-0.993220,
        -2.094395,1.990381,1.586005,
        -2.094395,1.990470,0.580283,
        -2.094395,1.990523,1.450326,
        -2.094395,1.993909,1.312897,
        -2.094395,1.994694,2.452813,
        -2.094395,1.999226,-0.620717,
        -2.094395,2.005145,0.941012,
        -2.094395,2.007301,0.142337,
        -2.094395,2.014149,0.444138,
        -2.094395,2.014338,-1.127392,
        -2.094395,2.018131,3.021778,
        -2.094395,2.019676,-1.980944,
        -2.094395,2.019912,2.686612,
        -2.094395,2.020545,2.088672,
        -2.094395,2.026074,0.300945,
        -2.094395,2.026196,-2.889057,
        -2.094395,2.031959,-1.268178,
        -2.094395,2.032917,-2.199917,
        -2.094395,2.033572,-1.842884,
        -2.094395,2.034712,-2.419505,
        -2.094395,2.035818,-0.379479,
        -2.094395,2.036004,-0.757062,
        -2.094395,2.040501,-1.411488,
        -2.094395,2.042874,-1.701367,
        -2.094395,2.043060,-2.653746,
        -2.094395,2.043729,-1.557938,
        -2.094395,2.048538,0.806623,
        -2.094395,2.049006,1.199930,
        -2.094395,2.049502,-0.137989,
        -2.094395,2.052181,2.326959,
        -2.094395,2.052794,-0.008526,
        -2.094395,2.057608,1.951726,
        -2.094395,2.065568,2.567730,
        -2.094395,2.070371,2.879808,
        -2.094395,2.071256,-3.018961,
        -2.094395,2.079746,-0.897699,
        -2.094395,2.082784,1.042403,
        -2.094395,2.083752,-0.511086,
        -2.094395,2.084401,1.810502,
        -2.094395,2.085731,0.671081,
        -2.094395,2.097496,1.373172,
        -2.094395,2.099185,1.667006,
        -2.094395,2.100482,-2.301084,
        -2.094395,2.103118,2.194745,
        -2.094395,2.103223,1.521685,
        -2.094395,2.103695,-2.777707,
        -2.094395,2.107523,3.129581,
        -2.094395,2.108277,-2.095907,
        -2.094395,2.108891,-2.532166,
        -2.094395,2.111003,-0.268632,
        -2.094395,2.112276,-1.027246,
        -2.094395,2.114380,0.531457,
        -2.094395,2.119166,0.100552,
        -2.094395,2.119564,2.754818,
        -2.094395,2.122474,-0.648314,
        -2.094395,2.125183,2.445598,
        -2.094395,2.129501,0.238290,
        -2.094395,2.131178,0.386663,
        -2.094395,2.131223,-1.174311,
        -2.094395,2.134269,-1.948042,
        -2.094395,2.134684,0.907065,
        -2.094395,2.137615,2.988806,
        -2.094395,2.145635,-1.327708,
        -2.094395,2.146467,2.055081,
        -2.094395,2.146582,-1.793788,
        -2.094395,2.149210,-1.482321,
        -2.094395,2.152901,-1.636788,
        -2.094395,2.156337,1.151558,
        -2.094395,2.156474,-2.912065,
        -2.094395,2.157194,-0.792563,
        -2.094395,2.166724,-0.401638,
        -2.094395,2.166980,-0.134879,
        -2.094395,2.173644,-2.403429,
        -2.094395,2.175711,-2.655523,
        -2.094395,2.177132,2.600419,
        -2.094395,2.178879,0.767722,
        -2.094395,2.180752,1.908485,
        -2.094395,2.181338,2.313721,
        -2.094395,2.185580,1.287673,
        -2.094395,2.195400,-2.217546,
        -2.094395,2.200418,-3.056442,
        -2.094395,2.201829,-0.001735,
        -2.094395,2.203804,1.756176,
        -2.094395,2.211602,1.443930,
        -2.094395,2.212686,-0.544809,
        -2.094395,2.213937,0.621892,
        -2.094395,2.214075,1.600691,
        -2.094395,2.216256,1.020029,
        -2.094395,2.219060,2.886631,
        -2.094395,2.223375,-0.906185,
        -2.094395,2.226002,-1.060248,
        -2.094395,2.226413,-2.064034,
        -2.094395,2.230554,2.171334,
        -2.094395,2.232413,2.735777,
        -2.094395,2.233709,3.070830,
        -2.094395,2.234383,-2.792065,
        -2.094395,2.237183,0.468799,
        -2.094395,2.237695,-0.271388,
        -2.094395,2.239419,0.151693,
        -2.094395,2.243347,-2.521049,
        -2.094395,2.244035,-1.229456,
        -2.094395,2.244355,0.309899,
        -2.094395,2.245175,2.460236,
        -2.094395,2.245628,-1.899216,
        -2.094395,2.245973,-0.702234,
        -2.094395,2.255601,-1.728131,
        -2.094395,2.256702,-1.401367,
        -2.094395,2.261466,-1.557500,
        -2.094395,2.267390,0.878499,
        -2.094395,2.272099,2.018594,
        -2.094395,2.278618,1.170386,
        -2.094395,2.283304,-2.345584,
        -2.094395,2.285175,-2.943090,
        -2.094395,2.293000,-0.124451,
        -2.094395,2.294058,-0.421070,
        -2.094395,2.303283,1.855938,
        -2.094395,2.304968,2.608978,
        -2.094395,2.305347,2.311346,
        -2.094395,2.307188,-2.656134,
        -2.094395,2.307301,1.336200,
        -2.094395,2.309655,0.726020,
        -2.094395,2.317634,-2.185225,
        -2.094395,2.321510,1.685664,
        -2.094395,2.324376,-3.111006,
        -2.094395,2.324731,1.511186,
        -2.094395,2.329587,0.035498,
        -2.094395,2.330158,2.977264,
        -2.094395,2.335096,-0.588468,
        -2.094395,2.336046,-1.115799,
        -2.094395,2.338060,2.803613,
        -2.094395,2.338083,-0.943400,
        -2.094395,2.340210,0.562145,
        -2.094395,2.341931,1.023926,
        -2.094395,2.343265,-0.779061,
        -2.094395,2.343434,-2.013692,
        -2.094395,2.353029,-1.310529,
        -2.094395,2.354868,0.211543,
        -2.094395,2.355120,0.388593,
        -2.094395,2.355165,-1.833173,
        -2.094395,2.355462,2.151441,
        -2.094395,2.362948,-2.487477,
        -2.094395,2.363101,-2.810456,
        -2.094395,2.363394,-0.272918,
        -2.094395,2.364509,-1.644533,
        -2.094395,2.370465,2.466824,
        -2.094395,2.386110,-1.481351,
        -2.094395,2.394960,0.862133,
        -2.094395,2.395716,1.977184,
        -2.094395,2.396339,1.200900,
        -2.094395,2.407941,-2.316753,
        -2.094395,2.409450,-2.988432,
        -2.094395,2.417202,-0.103363,
        -2.094395,2.417386,-0.443097,
        -2.094395,2.417950,2.666662,
        -2.094395,2.423190,1.789783,
        -2.094395,2.423692,1.392696,
        -2.094395,2.431122,3.086379,
        -2.094395,2.431406,-2.647820,
        -2.094395,2.431584,2.307631,
        -2.094395,2.434090,1.592818,
        -2.094395,2.435993,0.683844,
        -2.094395,2.440898,-2.133300,
        -2.094395,2.442222,-1.195837,
        -2.094395,2.443112,2.883761,
        -2.094395,2.449191,-0.633651,
        -2.094395,2.450821,0.084741,
        -2.094395,2.453499,-1.008861,
        -2.094395,2.460010,-1.938011,
        -2.094395,2.460199,-0.822008,
        -2.094395,2.460902,0.490079,
        -2.094395,2.461774,-1.734913,
        -2.094395,2.463986,1.034407,
        -2.094395,2.467647,0.287667,
        -2.094395,2.479849,2.126925,
        -2.094395,2.485907,-2.835471,
        -2.094395,2.485945,-1.374853,
        -2.094395,2.487747,2.505912,
        -2.094395,2.487957,-0.274279,
        -2.094395,2.489741,-2.468428,
        -2.094395,2.512538,1.239900,
        -2.094395,2.517528,1.923289,
        -2.094395,2.518454,-1.565852,
        -2.094395,2.519348,0.843503,
        -2.094395,2.522766,-3.059231,
        -2.094395,2.528686,2.737228,
        -2.094395,2.534798,-2.266671,
        -2.094395,2.536533,1.466274,
        -2.094395,2.538504,1.699812,
        -2.094395,2.539516,-0.476942,
        -2.094395,2.539532,-0.074548,
        -2.094395,2.543197,2.990252,
        -2.094395,2.554022,2.320619,
        -2.094395,2.556919,-2.655005,
        -2.094395,2.558146,0.627191,
        -2.094395,2.558223,-1.164189,
        -2.094395,2.563452,-2.044028,
        -2.094395,2.568462,0.150655,
        -2.094395,2.569130,-0.700837,
        -2.094395,2.570050,-1.804492,
        -2.094395,2.572710,-0.934665,
        -2.094395,2.575994,0.391718,
        -2.094395,2.584774,1.046703,
        -2.094395,2.601432,2.102535,
        -2.094395,2.603559,2.555114,
        -2.094395,2.604150,-2.885208,
        -2.094395,2.608630,-1.378011,
        -2.094395,2.611256,-0.279110,
        -2.094395,2.617344,-2.439480,
        -2.094395,2.627402,1.293552,
        -2.094395,2.635396,1.848850,
        -2.094395,2.635686,2.834950,
        -2.094395,2.637087,3.132730,
        -2.094395,2.642029,0.813274,
        -2.094395,2.645260,1.570203,
        -2.094395,2.646182,-1.614036,
        -2.094395,2.660113,-2.186513,
        -2.094395,2.660483,-0.034608,
        -2.094395,2.661713,-0.529511,
        -2.094395,2.672720,-1.110422,
        -2.094395,2.676626,2.335353,
        -2.094395,2.676638,0.540364,
        -2.094395,2.676780,-1.901158,
        -2.094395,2.681411,-2.667965,
        -2.094395,2.682152,0.245751,
        -2.094395,2.682859,-0.817071,
        -2.094395,2.705107,1.062142,
        -2.094395,2.717705,2.625995,
        -2.094395,2.721001,2.060097,
        -2.094395,2.722319,-2.954069,
        -2.094395,2.729840,-1.368237,
        -2.094395,2.734349,-0.289298,
        -2.094395,2.738409,2.975671,
        -2.094395,2.740366,1.374341,
        -2.094395,2.744633,-2.391311,
        -2.094395,2.747398,1.727166,
        -2.094395,2.763799,0.760372,
        -2.094395,2.768238,-1.679243,
        -2.094395,2.779245,-2.046268,
        -2.094395,2.780208,-0.626453,
        -2.094395,2.780471,0.027732,
        -2.094395,2.784696,-1.013128,
        -2.094395,2.789817,0.395431,
        -2.094395,2.801154,2.353143,
        -2.094395,2.807091,-2.687477,
        -2.094395,2.825762,1.085035,
        -2.094395,2.830209,2.734943,
        -2.094395,2.833517,-3.085508,
        -2.094395,2.838301,1.980061,
        -2.094395,2.849871,1.512012,
        -2.094395,2.850089,-1.343899,
        -2.094395,2.856905,-0.318441,
        -2.094395,2.870445,-2.291001,
        -2.094395,2.885017,0.655562,
        -2.094395,2.886428,-1.793217,
        -2.094395,2.890337,-0.825633,
        -2.094395,2.898932,0.139973,
        -2.094395,2.931927,-2.728768,
        -2.094395,2.932096,2.377790,
        -2.094395,2.937409,2.947245,
        -2.094395,2.948048,1.133127,
        -2.094395,2.950896,1.797889,
        -2.094395,2.970019,-1.282382,
        -2.094395,2.978333,-0.412090,
        -2.094395,2.996437,-2.089121,
        -2.094395,3.011274,0.434110,
        -2.094395,3.050410,3.138492,
        -2.094395,3.064492,1.647259,
        -2.094395,3.088868,-0.989501,
        -1.989675,0.053884,-2.408394,
        -1.989675,0.077645,1.453433,
        -1.989675,0.088408,-0.311160,
        -1.989675,0.142314,2.590387,
        -1.989675,0.147206,-1.385164,
        -1.989675,0.163506,0.573203,
        -1.989675,0.165485,-2.875396,
        -1.989675,0.182564,-2.125821,
        -1.989675,0.190569,1.929141,
        -1.989675,0.201102,1.249082,
        -1.989675,0.201397,-0.716646,
        -1.989675,0.210568,-0.069559,
        -1.989675,0.251672,2.848900,
        -1.989675,0.266262,-1.678068,
        -1.989675,0.269108,-1.215563,
        -1.989675,0.273777,-2.506756,
        -1.989675,0.279854,0.814581,
        -1.989675,0.281664,-2.969147,
        -1.989675,0.283625,2.405603,
        -1.989675,0.284058,0.357350,
        -1.989675,0.294107,1.593880,
        -1.989675,0.302491,-0.400410,
        -1.989675,0.308059,2.021491,
        -1.989675,0.312219,-2.072057,
        -1.989675,0.321905,-0.829121,
        -1.989675,0.329984,1.201621,
        -1.989675,0.339714,-0.013712,
        -1.989675,0.368784,2.974974,
        -1.989675,0.376428,-1.495532,
        -1.989675,0.378103,2.646436,
        -1.989675,0.385088,0.590355,
        -1.989675,0.386248,-2.690943,
        -1.989675,0.389851,-2.340031,
        -1.989675,0.391502,-1.171374,
        -1.989675,0.398249,-1.812694,
        -1.989675,0.400542,-2.995729,
        -1.989675,0.404273,0.918163,
        -1.989675,0.405617,1.773533,
        -1.989675,0.409009,-0.570243,
        -1.989675,0.410700,0.270433,
        -1.989675,0.415893,1.464890,
        -1.989675,0.420680,2.088164,
        -1.989675,0.424766,-0.258729,
        -1.989675,0.426507,2.367872,
        -1.989675,0.441292,-0.890727,
        -1.989675,0.451839,-2.078193,
        -1.989675,0.459270,1.191562,
        -1.989675,0.470622,0.014002,
        -1.989675,0.481292,-2.529098,
        -1.989675,0.482460,2.849843,
        -1.989675,0.492205,-1.397963,
        -1.989675,0.493314,3.117264,
        -1.989675,0.497870,0.727828,
        -1.989675,0.499783,-1.654379,
        -1.989675,0.502651,0.463333,
        -1.989675,0.505463,2.594557,
        -1.989675,0.506104,-2.874910,
        -1.989675,0.513579,-1.147354,
        -1.989675,0.518569,1.644237,
        -1.989675,0.519001,1.900488,
        -1.989675,0.519277,-0.673809,
        -1.989675,0.524769,-0.421955,
        -1.989675,0.525564,-2.288838,
        -1.989675,0.532541,0.978737,
        -1.989675,0.534610,2.154808,
        -1.989675,0.535020,-1.891846,
        -1.989675,0.540208,1.400497,
        -1.989675,0.542006,0.228337,
        -1.989675,0.550762,-0.180771,
        -1.989675,0.560531,2.378474,
        -1.989675,0.561629,-0.924576,
        -1.989675,0.567465,-2.679392,
        -1.989675,0.588527,1.181755,
        -1.989675,0.591643,-2.098364,
        -1.989675,0.594963,2.997575,
        -1.989675,0.599643,-3.045235,
        -1.989675,0.599962,2.773556,
        -1.989675,0.602620,0.029270,
        -1.989675,0.605711,0.598454,
        -1.989675,0.607708,-1.541026,
        -1.989675,0.609360,-2.464379,
        -1.989675,0.610738,-1.326786,
        -1.989675,0.616968,0.820489,
        -1.989675,0.626379,1.773610,
        -1.989675,0.628626,-1.751180,
        -1.989675,0.630671,0.391949,
        -1.989675,0.632622,-0.737642,
        -1.989675,0.632708,2.566722,
        -1.989675,0.632797,1.989858,
        -1.989675,0.634455,-0.534790,
        -1.989675,0.636205,1.561170,
        -1.989675,0.637681,-1.123423,
        -1.989675,0.641441,-2.842109,
        -1.989675,0.644807,-0.327139,
        -1.989675,0.652440,2.197129,
        -1.989675,0.660897,-2.278593,
        -1.989675,0.664008,1.360091,
        -1.989675,0.670850,-1.945257,
        -1.989675,0.671396,1.004859,
        -1.989675,0.673985,0.203855,
        -1.989675,0.680326,-0.133370,
        -1.989675,0.685083,-0.942363,
        -1.989675,0.688204,-2.632204,
        -1.989675,0.689777,2.385011,
        -1.989675,0.700065,3.115430,
        -1.989675,0.704080,2.911636,
        -1.989675,0.712883,1.174216,
        -1.989675,0.715701,0.699906,
        -1.989675,0.718851,-1.442215,
        -1.989675,0.721019,2.721536,
        -1.989675,0.725300,-2.975134,
        -1.989675,0.727136,-1.633288,
        -1.989675,0.727399,-2.120567,
        -1.989675,0.730751,0.522488,
        -1.989675,0.732892,-1.254330,
        -1.989675,0.734333,-2.446184,
        -1.989675,0.735125,0.037270,
        -1.989675,0.737893,1.866463,
        -1.989675,0.740650,1.681075,
        -1.989675,0.745971,-0.781286,
        -1.989675,0.749432,-0.442559,
        -1.989675,0.750102,2.050636,
        -1.989675,0.751046,0.863821,
        -1.989675,0.757018,-0.608833,
        -1.989675,0.757046,-1.816019,
        -1.989675,0.757389,1.501184,
        -1.989675,0.759800,2.546181,
        -1.989675,0.759859,0.351923,
        -1.989675,0.761141,-2.781102,
        -1.989675,0.769035,-1.078574,
        -1.989675,0.770166,-0.266604,
        -1.989675,0.772218,2.227380,
        -1.989675,0.788382,1.331051,
        -1.989675,0.790269,-2.282409,
        -1.989675,0.803410,-1.984567,
        -1.989675,0.803813,0.191520,
        -1.989675,0.806832,3.022155,
        -1.989675,0.807476,-2.597675,
        -1.989675,0.812512,-0.107335,
        -1.989675,0.814161,1.010949,
        -1.989675,0.817732,-3.084486,
        -1.989675,0.817846,2.390688,
        -1.989675,0.819323,2.845960,
        -1.989675,0.819430,-0.915017,
        -1.989675,0.829198,-1.532491,
        -1.989675,0.829531,0.665695,
        -1.989675,0.829816,1.167147,
        -1.989675,0.833613,-1.357492,
        -1.989675,0.842945,-2.907318,
        -1.989675,0.844883,2.679708,
        -1.989675,0.848189,-1.705003,
        -1.989675,0.849888,1.767901,
        -1.989675,0.851126,0.493370,
        -1.989675,0.852911,1.933394,
        -1.989675,0.857516,-2.432863,
        -1.989675,0.858261,-1.191729,
        -1.989675,0.858279,-2.138895,
        -1.989675,0.859723,-0.734928,
        -1.989675,0.859754,1.603292,
        -1.989675,0.861655,-0.394821,
        -1.989675,0.864300,0.043549,
        -1.989675,0.866940,0.851229,
        -1.989675,0.869748,2.095969,
        -1.989675,0.877352,-0.574668,
        -1.989675,0.879254,-2.733338,
        -1.989675,0.882554,1.443757,
        -1.989675,0.884060,0.334199,
        -1.989675,0.885626,-1.866061,
        -1.989675,0.889834,2.529957,
        -1.989675,0.893140,2.251698,
        -1.989675,0.897463,-0.240822,
        -1.989675,0.898758,-1.031948,
        -1.989675,0.913918,1.287618,
        -1.989675,0.916555,-2.287173,
        -1.989675,0.918305,3.107845,
        -1.989675,0.919522,2.948277,
        -1.989675,0.927138,-2.569686,
        -1.989675,0.929031,0.185383,
        -1.989675,0.933766,-3.015917,
        -1.989675,0.934873,-2.010822,
        -1.989675,0.935636,-1.445855,
        -1.989675,0.936702,-0.866461,
        -1.989675,0.937395,0.729108,
        -1.989675,0.937910,2.787575,
        -1.989675,0.939144,-0.092962,
        -1.989675,0.941718,1.125085,
        -1.989675,0.942873,-1.607680,
        -1.989675,0.943701,0.980328,
        -1.989675,0.947774,2.395256,
        -1.989675,0.952433,0.584347,
        -1.989675,0.954456,-1.287293,
        -1.989675,0.958731,-2.852552,
        -1.989675,0.961838,1.839284,
        -1.989675,0.961898,-0.469425,
        -1.989675,0.965426,1.686615,
        -1.989675,0.970108,1.990771,
        -1.989675,0.972820,-1.762780,
        -1.989675,0.973500,-0.711881,
        -1.989675,0.975253,2.641082,
        -1.989675,0.977824,0.441928,
        -1.989675,0.981485,-2.424766,
        -1.989675,0.981877,1.537100,
        -1.989675,0.982609,-1.136920,
        -1.989675,0.984265,-2.156073,
        -1.989675,0.989150,0.045493,
        -1.989675,0.993052,2.137478,
        -1.989675,0.997766,-2.693871,
        -1.989675,0.999847,-0.334738,
        -1.989675,1.001920,0.848054,
        -1.989675,1.006146,1.389734,
        -1.989675,1.012180,0.304066,
        -1.989675,1.016938,2.270674,
        -1.989675,1.017371,-0.979790,
        -1.989675,1.019495,-1.894487,
        -1.989675,1.023796,-0.583691,
        -1.989675,1.026525,2.506945,
        -1.989675,1.027004,-0.203653,
        -1.989675,1.028027,3.036573,
        -1.989675,1.031435,1.241021,
        -1.989675,1.032316,2.884221,
        -1.989675,1.035863,-3.110475,
        -1.989675,1.042085,-2.292950,
        -1.989675,1.043272,-1.522986,
        -1.989675,1.046265,-2.958847,
        -1.989675,1.049272,-1.376344,
        -1.989675,1.050039,-2.552086,
        -1.989675,1.053682,0.689863,
        -1.989675,1.055458,0.170394,
        -1.989675,1.056622,-0.831866,
        -1.989675,1.057515,1.097998,
        -1.989675,1.058706,-2.036438,
        -1.989675,1.061389,-1.672248,
        -1.989675,1.063758,2.743093,
        -1.989675,1.070157,-0.073422,
        -1.989675,1.072154,1.761506,
        -1.989675,1.074029,1.909193,
        -1.989675,1.074204,0.545276,
        -1.989675,1.074714,-2.805147,
        -1.989675,1.076262,0.966104,
        -1.989675,1.077642,-1.227073,
        -1.989675,1.084151,1.618525,
        -1.989675,1.085085,2.377673,
        -1.989675,1.086822,-0.441112,
        -1.989675,1.087142,2.058815,
        -1.989675,1.098383,-0.696430,
        -1.989675,1.099138,-1.084033,
        -1.989675,1.103208,1.477773,
        -1.989675,1.103783,0.408254,
        -1.989675,1.106275,-2.420800,
        -1.989675,1.108079,2.612943,
        -1.989675,1.109345,-2.172718,
        -1.989675,1.115957,-1.787597,
        -1.989675,1.118930,0.045983,
        -1.989675,1.119777,0.834702,
        -1.989675,1.121943,-2.669042,
        -1.989675,1.124930,2.219917,
        -1.989675,1.125217,1.336294,
        -1.989675,1.130958,2.972931,
        -1.989675,1.131400,-0.306629,
        -1.989675,1.137135,-0.936514,
        -1.989675,1.139125,-1.924528,
        -1.989675,1.139877,0.276382,
        -1.989675,1.142003,-3.058022,
        -1.989675,1.147363,-0.558448,
        -1.989675,1.150348,-0.176418,
        -1.989675,1.151198,1.197016,
        -1.989675,1.153930,-1.597104,
        -1.989675,1.155389,2.837148,
        -1.989675,1.155458,-1.460623,
        -1.989675,1.157331,-2.906362,
        -1.989675,1.158594,2.487618,
        -1.989675,1.162103,-1.322553,
        -1.989675,1.165407,3.092861,
        -1.989675,1.168025,0.717075,
        -1.989675,1.168406,-2.301519,
        -1.989675,1.174782,-2.538755,
        -1.989675,1.177998,1.061256,
        -1.989675,1.178898,1.827305,
        -1.989675,1.181331,-2.060330,
        -1.989675,1.182540,-0.798952,
        -1.989675,1.184569,0.147517,
        -1.989675,1.184734,1.968919,
        -1.989675,1.184804,0.593386,
        -1.989675,1.188278,1.690808,
        -1.989675,1.194360,2.110440,
        -1.989675,1.194776,2.710455,
        -1.989675,1.197745,-1.173705,
        -1.989675,1.199898,-2.777257,
        -1.989675,1.201164,2.355931,
        -1.989675,1.203760,1.556404,
        -1.989675,1.205845,-0.415993,
        -1.989675,1.206585,0.933648,
        -1.989675,1.206950,-0.037248,
        -1.989675,1.209341,0.471162,
        -1.989675,1.213483,-1.039417,
        -1.989675,1.218146,-0.672643,
        -1.989675,1.221608,1.421669,
        -1.989675,1.226043,-1.693533,
        -1.989675,1.232870,-1.822773,
        -1.989675,1.234375,-2.421194,
        -1.989675,1.234531,-2.190148,
        -1.989675,1.239414,0.348959,
        -1.989675,1.240183,2.587255,
        -1.989675,1.242312,2.941234,
        -1.989675,1.244075,1.286644,
        -1.989675,1.245961,-2.997461,
        -1.989675,1.246495,-2.646872,
        -1.989675,1.247736,2.242810,
        -1.989675,1.253604,0.812222,
        -1.989675,1.257816,-0.894317,
        -1.989675,1.258247,-1.953634,
        -1.989675,1.258993,-0.284099,
        -1.989675,1.262380,-0.157609,
        -1.989675,1.262791,-1.549229,
        -1.989675,1.263772,-0.533073,
        -1.989675,1.266731,-1.412461,
        -1.989675,1.268591,-3.122146,
        -1.989675,1.269614,-1.275246,
        -1.989675,1.270745,1.152533,
        -1.989675,1.273606,0.224545,
        -1.989675,1.279716,2.812101,
        -1.989675,1.283786,2.462124,
        -1.989675,1.284798,0.078209,
        -1.989675,1.285402,-2.882049,
        -1.989675,1.286936,1.886053,
        -1.989675,1.289408,0.683122,
        -1.989675,1.294736,1.756184,
        -1.989675,1.296978,-2.311511,
        -1.989675,1.298546,2.020601,
        -1.989675,1.300924,1.022478,
        -1.989675,1.303305,3.043152,
        -1.989675,1.304210,-2.084225,
        -1.989675,1.305241,-2.530249,
        -1.989675,1.307513,1.628003,
        -1.989675,1.310722,0.558267,
        -1.989675,1.314142,2.139329,
        -1.989675,1.316087,-0.777971,
        -1.989675,1.317907,-1.121023,
        -1.989675,1.321019,-2.749776,
        -1.989675,1.321609,2.687717,
        -1.989675,1.322568,1.499415,
        -1.989675,1.322972,-0.395699,
        -1.989675,1.325286,-0.994192,
        -1.989675,1.329146,-0.648513,
        -1.989675,1.329795,-0.042570,
        -1.989675,1.331651,2.344497,
        -1.989675,1.336522,0.435098,
        -1.989675,1.339410,0.900087,
        -1.989675,1.339550,1.370353,
        -1.989675,1.340441,-1.731580,
        -1.989675,1.351855,-1.855234,
        -1.989675,1.360161,-3.018404,
        -1.989675,1.361031,-1.613971,
        -1.989675,1.361196,2.914780,
        -1.989675,1.361998,-2.207254,
        -1.989675,1.362302,1.239384,
        -1.989675,1.365228,2.563808,
        -1.989675,1.365452,0.310522,
        -1.989675,1.367028,-2.420041,
        -1.989675,1.373422,-1.484281,
        -1.989675,1.374245,-1.226417,
        -1.989675,1.375020,-0.170393,
        -1.989675,1.376560,-1.355481,
        -1.989675,1.377070,-1.980984,
        -1.989675,1.377578,-2.636339,
        -1.989675,1.378096,0.774626,
        -1.989675,1.379086,-0.508192,
        -1.989675,1.382562,0.177141,
        -1.989675,1.388849,2.233847,
        -1.989675,1.389268,-0.886349,
        -1.989675,1.391324,1.109396,
        -1.989675,1.391699,-0.288725,
        -1.989675,1.392869,3.136903,
        -1.989675,1.393612,1.952016,
        -1.989675,1.395545,-2.850807,
        -1.989675,1.401926,1.822930,
        -1.989675,1.402149,2.788898,
        -1.989675,1.409552,0.647980,
        -1.989675,1.410840,0.056219,
        -1.989675,1.411944,2.444377,
        -1.989675,1.413639,1.697982,
        -1.989675,1.426479,1.573427,
        -1.989675,1.426725,2.102242,
        -1.989675,1.427868,0.984637,
        -1.989675,1.429580,-2.107232,
        -1.989675,1.431375,-2.313269,
        -1.989675,1.431907,3.015410,
        -1.989675,1.432325,-1.029054,
        -1.989675,1.433341,0.524268,
        -1.989675,1.435181,-0.739771,
        -1.989675,1.437208,-2.529085,
        -1.989675,1.439041,-0.617966,
        -1.989675,1.441473,1.448544,
        -1.989675,1.444527,2.664744,
        -1.989675,1.450513,-0.400625,
        -1.989675,1.451907,-0.066823,
        -1.989675,1.452754,-2.738631,
        -1.989675,1.454061,-2.950416,
        -1.989675,1.456736,1.323154,
        -1.989675,1.458370,0.400433,
        -1.989675,1.459314,-1.765003,
        -1.989675,1.459902,-1.145835,
        -1.989675,1.464952,2.329245,
        -1.989675,1.466050,0.862104,
        -1.989675,1.471330,-1.886160,
        -1.989675,1.475119,-1.648255,
        -1.989675,1.480878,1.194576,
        -1.989675,1.480942,0.273235,
        -1.989675,1.481086,2.890269,
        -1.989675,1.484118,-1.529156,
        -1.989675,1.484741,-1.279219,
        -1.989675,1.487681,-1.406069,
        -1.989675,1.487886,-3.066037,
        -1.989675,1.490555,2.543284,
        -1.989675,1.492424,-0.190932,
        -1.989675,1.495002,-0.839810,
        -1.989675,1.495297,2.007188,
        -1.989675,1.496414,-2.009267,
        -1.989675,1.498261,-2.425270,
        -1.989675,1.498907,-2.207327,
        -1.989675,1.500204,0.737638,
        -1.989675,1.502469,0.151156,
        -1.989675,1.507448,-0.514782,
        -1.989675,1.509287,2.204647,
        -1.989675,1.510518,1.884920,
        -1.989675,1.511948,-2.632360,
        -1.989675,1.515235,1.068565,
        -1.989675,1.522237,3.102018,
        -1.989675,1.522373,1.764991,
        -1.989675,1.522416,-0.954956,
        -1.989675,1.523102,2.765920,
        -1.989675,1.524811,-2.844718,
        -1.989675,1.527261,-0.304916,
        -1.989675,1.529788,0.613208,
        -1.989675,1.533648,1.644808,
        -1.989675,1.535018,0.028869,
        -1.989675,1.541082,2.424608,
        -1.989675,1.546010,1.523462,
        -1.989675,1.551135,-0.642994,
        -1.989675,1.553895,0.489607,
        -1.989675,1.555041,0.947813,
        -1.989675,1.559671,-1.074934,
        -1.989675,1.560345,2.985899,
        -1.989675,1.560487,-2.323400,
        -1.989675,1.561127,1.401111,
        -1.989675,1.567193,2.643172,
        -1.989675,1.568892,2.098362,
        -1.989675,1.571702,-2.098435,
        -1.989675,1.572701,-2.528730,
        -1.989675,1.573572,1.278593,
        -1.989675,1.574849,-0.091505,
        -1.989675,1.576334,0.365236,
        -1.989675,1.577029,-2.962888,
        -1.989675,1.577688,-1.814941,
        -1.989675,1.579542,-0.755917,
        -1.989675,1.579624,-1.199759,
        -1.989675,1.581515,-0.419414,
        -1.989675,1.584581,-2.738702,
        -1.989675,1.586948,-1.696296,
        -1.989675,1.588901,2.301496,
        -1.989675,1.589936,0.825020,
        -1.989675,1.590018,-1.941652,
        -1.989675,1.596968,-1.328358,
        -1.989675,1.597178,-1.575607,
        -1.989675,1.600470,-1.453196,
        -1.989675,1.601408,0.242409,
        -1.989675,1.602259,1.152656,
        -1.989675,1.603064,2.865774,
        -1.989675,1.612085,1.968802,
        -1.989675,1.612720,-3.080498,
        -1.989675,1.616061,2.521531,
        -1.989675,1.616900,-0.207457,
        -1.989675,1.620988,-0.875920,
        -1.989675,1.622114,0.701509,
        -1.989675,1.622749,0.121426,
        -1.989675,1.622993,-2.221629,
        -1.989675,1.629346,-0.542669,
        -1.989675,1.630007,1.844552,
        -1.989675,1.634747,-2.426198,
        -1.989675,1.642346,1.722617,
        -1.989675,1.643925,1.034629,
        -1.989675,1.644848,2.743774,
        -1.989675,1.645258,2.188109,
        -1.989675,1.645417,-2.634725,
        -1.989675,1.647395,3.085437,
        -1.989675,1.648538,-2.855044,
        -1.989675,1.649932,0.576808,
        -1.989675,1.652224,-0.996105,
        -1.989675,1.652871,1.600681,
        -1.989675,1.657997,-0.325127,
        -1.989675,1.661936,0.001037,
        -1.989675,1.665187,1.477402,
        -1.989675,1.665496,2.397872,
        -1.989675,1.673267,-0.665453,
        -1.989675,1.673557,0.452526,
        -1.989675,1.679277,-1.122421,
        -1.989675,1.679696,0.912482,
        -1.989675,1.682420,1.353922,
        -1.989675,1.685031,-2.009038,
        -1.989675,1.685196,2.967046,
        -1.989675,1.690107,2.620910,
        -1.989675,1.690453,-2.123076,
        -1.989675,1.691693,-1.871572,
        -1.989675,1.691817,2.061688,
        -1.989675,1.692644,1.237670,
        -1.989675,1.696386,0.328123,
        -1.989675,1.697162,-1.249026,
        -1.989675,1.697488,-2.322785,
        -1.989675,1.698726,-2.973576,
        -1.989675,1.699172,-1.749717,
        -1.989675,1.704478,-0.106295,
        -1.989675,1.707244,-0.444770,
        -1.989675,1.707500,-2.531497,
        -1.989675,1.707714,-0.787321,
        -1.989675,1.708430,-1.627552,
        -1.989675,1.710583,-1.376890,
        -1.989675,1.712415,-2.747921,
        -1.989675,1.713125,-1.503073,
        -1.989675,1.713452,0.789419,
        -1.989675,1.720612,2.279888,
        -1.989675,1.724852,0.202831,
        -1.989675,1.724868,2.846246,
        -1.989675,1.728940,1.931535,
        -1.989675,1.732838,1.127584,
        -1.989675,1.735719,-3.092485,
        -1.989675,1.739676,2.496017,
        -1.989675,1.739971,-0.228876,
        -1.989675,1.743328,0.663347,
        -1.989675,1.744927,-0.911892,
        -1.989675,1.745952,0.084733,
        -1.989675,1.748077,1.804354,
        -1.989675,1.752959,-0.567548,
        -1.989675,1.760169,1.678132,
        -1.989675,1.763887,-2.218450,
        -1.989675,1.766737,2.722752,
        -1.989675,1.767999,1.004737,
        -1.989675,1.769149,1.550298,
        -1.989675,1.770119,-2.427175,
        -1.989675,1.770460,2.154208,
        -1.989675,1.770514,3.070982,
        -1.989675,1.770735,0.537245,
        -1.989675,1.772673,-2.864760,
        -1.989675,1.773527,-1.039582,
        -1.989675,1.776419,-2.642421,
        -1.989675,1.780769,1.416330,
        -1.989675,1.786634,-0.346247,
        -1.989675,1.788829,-1.947553,
        -1.989675,1.793737,0.410270,
        -1.989675,1.794553,2.374618,
        -1.989675,1.794699,-0.694061,
        -1.989675,1.797469,-1.169105,
        -1.989675,1.803705,0.881076,
        -1.989675,1.804124,1.264249,
        -1.989675,1.807825,-2.090455,
        -1.989675,1.808822,-1.817375,
        -1.989675,1.809020,2.949741,
        -1.989675,1.812466,2.597595,
        -1.989675,1.813159,2.025395,
        -1.989675,1.813816,-1.298435,
        -1.989675,1.815016,0.278951,
        -1.989675,1.816510,-0.008631,
        -1.989675,1.817772,-1.689625,
        -1.989675,1.821360,-2.984352,
        -1.989675,1.822312,-0.127211,
        -1.989675,1.823386,-1.429807,
        -1.989675,1.823834,-1.560483,
        -1.989675,1.832060,-0.821627,
        -1.989675,1.832690,-0.468440,
        -1.989675,1.833945,-2.320278,
        -1.989675,1.834787,0.751968,
        -1.989675,1.840579,-2.536445,
        -1.989675,1.841033,-2.756932,
        -1.989675,1.846350,1.893983,
        -1.989675,1.846644,1.113766,
        -1.989675,1.846833,2.826543,
        -1.989675,1.847325,2.249051,
        -1.989675,1.854119,0.134161,
        -1.989675,1.858863,-3.106022,
        -1.989675,1.865046,1.763384,
        -1.989675,1.865619,-0.952036,
        -1.989675,1.865912,0.624453,
        -1.989675,1.866694,2.473686,
        -1.989675,1.867524,-0.240083,
        -1.989675,1.875561,1.632831,
        -1.989675,1.876049,-0.594378,
        -1.989675,1.880509,1.500204,
        -1.989675,1.883504,1.360457,
        -1.989675,1.886971,-2.196665,
        -1.989675,1.888979,0.982922,
        -1.989675,1.889815,2.701570,
        -1.989675,1.892144,-2.005340,
        -1.989675,1.892348,0.496522,
        -1.989675,1.893281,-1.083976,
        -1.989675,1.893828,3.053453,
        -1.989675,1.895117,2.121107,
        -1.989675,1.898741,-2.875735,
        -1.989675,1.904429,-2.428029,
        -1.989675,1.908408,-2.648903,
        -1.989675,1.911828,-0.364489,
        -1.989675,1.912224,0.366565,
        -1.989675,1.914854,1.214468,
        -1.989675,1.915467,-1.883143,
        -1.989675,1.915763,-0.724679,
        -1.989675,1.915790,-1.216366,
        -1.989675,1.922256,2.347806,
        -1.989675,1.922616,0.844633,
        -1.989675,1.925185,0.231170,
        -1.989675,1.927474,-1.754616,
        -1.989675,1.928342,-1.351174,
        -1.989675,1.933328,-1.623328,
        -1.989675,1.933878,0.034150,
        -1.989675,1.934585,-1.487979,
        -1.989675,1.934802,2.927634,
        -1.989675,1.934862,1.989776,
        -1.989675,1.936327,-0.099350,
        -1.989675,1.938088,2.578507,
        -1.989675,1.945677,-2.999805,
        -1.989675,1.954572,-0.859208,
        -1.989675,1.957868,-0.490784,
        -1.989675,1.958989,0.712772,
        -1.989675,1.960212,-2.307561,
        -1.989675,1.964774,1.855507,
        -1.989675,1.964971,-2.100533,
        -1.989675,1.969112,1.088273,
        -1.989675,1.971438,-2.765415,
        -1.989675,1.972312,2.805134,
        -1.989675,1.974950,2.220596,
        -1.989675,1.975135,-2.539127,
        -1.989675,1.982107,1.720829,
        -1.989675,1.983195,-3.127984,
        -1.989675,1.983318,-0.247182,
        -1.989675,1.986720,-0.993220,
        -1.989675,1.990381,1.586005,
        -1.989675,1.990470,0.580283,
        -1.989675,1.990523,1.450326,
        -1.989675,1.993909,1.312897,
        -1.989675,1.994694,2.452813,
        -1.989675,1.999226,-0.620717,
        -1.989675,2.005145,0.941012,
        -1.989675,2.007301,0.142337,
        -1.989675,2.014149,0.444138,
        -1.989675,2.014338,-1.127392,
        -1.989675,2.018131,3.021778,
        -1.989675,2.019676,-1.980944,
        -1.989675,2.019912,2.686612,
        -1.989675,2.020545,2.088672,
        -1.989675,2.026074,0.300945,
        -1.989675,2.026196,-2.889057,
        -1.989675,2.031959,-1.268178,
        -1.989675,2.032917,-2.199917,
        -1.989675,2.033572,-1.842884,
        -1.989675,2.034712,-2.419505,
        -1.989675,2.035818,-0.379479,
        -1.989675,2.036004,-0.757062,
        -1.989675,2.040501,-1.411488,
        -1.989675,2.042874,-1.701367,
        -1.989675,2.043060,-2.653746,
        -1.989675,2.043729,-1.557938,
        -1.989675,2.048538,0.806623,
        -1.989675,2.049006,1.199930,
        -1.989675,2.049502,-0.137989,
        -1.989675,2.052181,2.326959,
        -1.989675,2.052794,-0.008526,
        -1.989675,2.057608,1.951726,
        -1.989675,2.065568,2.567730,
        -1.989675,2.070371,2.879808,
        -1.989675,2.071256,-3.018961,
        -1.989675,2.079746,-0.897699,
        -1.989675,2.082784,1.042403,
        -1.989675,2.083752,-0.511086,
        -1.989675,2.084401,1.810502,
        -1.989675,2.085731,0.671081,
        -1.989675,2.097496,1.373172,
        -1.989675,2.099185,1.667006,
        -1.989675,2.100482,-2.301084,
        -1.989675,2.103118,2.194745,
        -1.989675,2.103223,1.521685,
        -1.989675,2.103695,-2.777707,
        -1.989675,2.107523,3.129581,
        -1.989675,2.108277,-2.095907,
        -1.989675,2.108891,-2.532166,
        -1.989675,2.111003,-0.268632,
        -1.989675,2.112276,-1.027246,
        -1.989675,2.114380,0.531457,
        -1.989675,2.119166,0.100552,
        -1.989675,2.119564,2.754818,
        -1.989675,2.122474,-0.648314,
        -1.989675,2.125183,2.445598,
        -1.989675,2.129501,0.238290,
        -1.989675,2.131178,0.386663,
        -1.989675,2.131223,-1.174311,
        -1.989675,2.134269,-1.948042,
        -1.989675,2.134684,0.907065,
        -1.989675,2.137615,2.988806,
        -1.989675,2.145635,-1.327708,
        -1.989675,2.146467,2.055081,
        -1.989675,2.146582,-1.793788,
        -1.989675,2.149210,-1.482321,
        -1.989675,2.152901,-1.636788,
        -1.989675,2.156337,1.151558,
        -1.989675,2.156474,-2.912065,
        -1.989675,2.157194,-0.792563,
        -1.989675,2.166724,-0.401638,
        -1.989675,2.166980,-0.134879,
        -1.989675,2.173644,-2.403429,
        -1.989675,2.175711,-2.655523,
        -1.989675,2.177132,2.600419,
        -1.989675,2.178879,0.767722,
        -1.989675,2.180752,1.908485,
        -1.989675,2.181338,2.313721,
        -1.989675,2.185580,1.287673,
        -1.989675,2.195400,-2.217546,
        -1.989675,2.200418,-3.056442,
        -1.989675,2.201829,-0.001735,
        -1.989675,2.203804,1.756176,
        -1.989675,2.211602,1.443930,
        -1.989675,2.212686,-0.544809,
        -1.989675,2.213937,0.621892,
        -1.989675,2.214075,1.600691,
        -1.989675,2.216256,1.020029,
        -1.989675,2.219060,2.886631,
        -1.989675,2.223375,-0.906185,
        -1.989675,2.226002,-1.060248,
        -1.989675,2.226413,-2.064034,
        -1.989675,2.230554,2.171334,
        -1.989675,2.232413,2.735777,
        -1.989675,2.233709,3.070830,
        -1.989675,2.234383,-2.792065,
        -1.989675,2.237183,0.468799,
        -1.989675,2.237695,-0.271388,
        -1.989675,2.239419,0.151693,
        -1.989675,2.243347,-2.521049,
        -1.989675,2.244035,-1.229456,
        -1.989675,2.244355,0.309899,
        -1.989675,2.245175,2.460236,
        -1.989675,2.245628,-1.899216,
        -1.989675,2.245973,-0.702234,
        -1.989675,2.255601,-1.728131,
        -1.989675,2.256702,-1.401367,
        -1.989675,2.261466,-1.557500,
        -1.989675,2.267390,0.878499,
        -1.989675,2.272099,2.018594,
        -1.989675,2.278618,1.170386,
        -1.989675,2.283304,-2.345584,
        -1.989675,2.285175,-2.943090,
        -1.989675,2.293000,-0.124451,
        -1.989675,2.294058,-0.421070,
        -1.989675,2.303283,1.855938,
        -1.989675,2.304968,2.608978,
        -1.989675,2.305347,2.311346,
        -1.989675,2.307188,-2.656134,
        -1.989675,2.307301,1.336200,
        -1.989675,2.309655,0.726020,
        -1.989675,2.317634,-2.185225,
        -1.989675,2.321510,1.685664,
        -1.989675,2.324376,-3.111006,
        -1.989675,2.324731,1.511186,
        -1.989675,2.329587,0.035498,
        -1.989675,2.330158,2.977264,
        -1.989675,2.335096,-0.588468,
        -1.989675,2.336046,-1.115799,
        -1.989675,2.338060,2.803613,
        -1.989675,2.338083,-0.943400,
        -1.989675,2.340210,0.562145,
        -1.989675,2.341931,1.023926,
        -1.989675,2.343265,-0.779061,
        -1.989675,2.343434,-2.013692,
        -1.989675,2.353029,-1.310529,
        -1.989675,2.354868,0.211543,
        -1.989675,2.355120,0.388593,
        -1.989675,2.355165,-1.833173,
        -1.989675,2.355462,2.151441,
        -1.989675,2.362948,-2.487477,
        -1.989675,2.363101,-2.810456,
        -1.989675,2.363394,-0.272918,
        -1.989675,2.364509,-1.644533,
        -1.989675,2.370465,2.466824,
        -1.989675,2.386110,-1.481351,
        -1.989675,2.394960,0.862133,
        -1.989675,2.395716,1.977184,
        -1.989675,2.396339,1.200900,
        -1.989675,2.407941,-2.316753,
        -1.989675,2.409450,-2.988432,
        -1.989675,2.417202,-0.103363,
        -1.989675,2.417386,-0.443097,
        -1.989675,2.417950,2.666662,
        -1.989675,2.423190,1.789783,
        -1.989675,2.423692,1.392696,
        -1.989675,2.431122,3.086379,
        -1.989675,2.431406,-2.647820,
        -1.989675,2.431584,2.307631,
        -1.989675,2.434090,1.592818,
        -1.989675,2.435993,0.683844,
        -1.989675,2.440898,-2.133300,
        -1.989675,2.442222,-1.195837,
        -1.989675,2.443112,2.883761,
        -1.989675,2.449191,-0.633651,
        -1.989675,2.450821,0.084741,
        -1.989675,2.453499,-1.008861,
        -1.989675,2.460010,-1.938011,
        -1.989675,2.460199,-0.822008,
        -1.989675,2.460902,0.490079,
        -1.989675,2.461774,-1.734913,
        -1.989675,2.463986,1.034407,
        -1.989675,2.467647,0.287667,
        -1.989675,2.479849,2.126925,
        -1.989675,2.485907,-2.835471,
        -1.989675,2.485945,-1.374853,
        -1.989675,2.487747,2.505912,
        -1.989675,2.487957,-0.274279,
        -1.989675,2.489741,-2.468428,
        -1.989675,2.512538,1.239900,
        -1.989675,2.517528,1.923289,
        -1.989675,2.518454,-1.565852,
        -1.989675,2.519348,0.843503,
        -1.989675,2.522766,-3.059231,
        -1.989675,2.528686,2.737228,
        -1.989675,2.534798,-2.266671,
        -1.989675,2.536533,1.466274,
        -1.989675,2.538504,1.699812,
        -1.989675,2.539516,-0.476942,
        -1.989675,2.539532,-0.074548,
        -1.989675,2.543197,2.990252,
        -1.989675,2.554022,2.320619,
        -1.989675,2.556919,-2.655005,
        -1.989675,2.558146,0.627191,
        -1.989675,2.558223,-1.164189,
        -1.989675,2.563452,-2.044028,
        -1.989675,2.568462,0.150655,
        -1.989675,2.569130,-0.700837,
        -1.989675,2.570050,-1.804492,
        -1.989675,2.572710,-0.934665,
        -1.989675,2.575994,0.391718,
        -1.989675,2.584774,1.046703,
        -1.989675,2.601432,2.102535,
        -1.989675,2.603559,2.555114,
        -1.989675,2.604150,-2.885208,
        -1.989675,2.608630,-1.378011,
        -1.989675,2.611256,-0.279110,
        -1.989675,2.617344,-2.439480,
        -1.989675,2.627402,1.293552,
        -1.989675,2.635396,1.848850,
        -1.989675,2.635686,2.834950,
        -1.989675,2.637087,3.132730,
        -1.989675,2.642029,0.813274,
        -1.989675,2.645260,1.570203,
        -1.989675,2.646182,-1.614036,
        -1.989675,2.660113,-2.186513,
        -1.989675,2.660483,-0.034608,
        -1.989675,2.661713,-0.529511,
        -1.989675,2.672720,-1.110422,
        -1.989675,2.676626,2.335353,
        -1.989675,2.676638,0.540364,
        -1.989675,2.676780,-1.901158,
        -1.989675,2.681411,-2.667965,
        -1.989675,2.682152,0.245751,
        -1.989675,2.682859,-0.817071,
        -1.989675,2.705107,1.062142,
        -1.989675,2.717705,2.625995,
        -1.989675,2.721001,2.060097,
        -1.989675,2.722319,-2.954069,
        -1.989675,2.729840,-1.368237,
        -1.989675,2.734349,-0.289298,
        -1.989675,2.738409,2.975671,
        -1.989675,2.740366,1.374341,
        -1.989675,2.744633,-2.391311,
        -1.989675,2.747398,1.727166,
        -1.989675,2.763799,0.760372,
        -1.989675,2.768238,-1.679243,
        -1.989675,2.779245,-2.046268,
        -1.989675,2.780208,-0.626453,
        -1.989675,2.780471,0.027732,
        -1.989675,2.784696,-1.013128,
        -1.989675,2.789817,0.395431,
        -1.989675,2.801154,2.353143,
        -1.989675,2.807091,-2.687477,
        -1.989675,2.825762,1.085035,
        -1.989675,2.830209,2.734943,
        -1.989675,2.833517,-3.085508,
        -1.989675,2.838301,1.980061,
        -1.989675,2.849871,1.512012,
        -1.989675,2.850089,-1.343899,
        -1.989675,2.856905,-0.318441,
        -1.989675,2.870445,-2.291001,
        -1.989675,2.885017,0.655562,
        -1.989675,2.886428,-1.793217,
        -1.989675,2.890337,-0.825633,
        -1.989675,2.898932,0.139973,
        -1.989675,2.931927,-2.728768,
        -1.989675,2.932096,2.377790,
        -1.989675,2.937409,2.947245,
        -1.989675,2.948048,1.133127,
        -1.989675,2.950896,1.797889,
        -1.989675,2.970019,-1.282382,
        -1.989675,2.978333,-0.412090,
        -1.989675,2.996437,-2.089121,
        -1.989675,3.011274,0.434110,
        -1.989675,3.050410,3.138492,
        -1.989675,3.064492,1.647259,
        -1.989675,3.088868,-0.989501,
        -1.884956,0.053884,-2.408394,
        -1.884956,0.077645,1.453433,
        -1.884956,0.088408,-0.311160,
        -1.884956,0.142314,2.590387,
        -1.884956,0.147206,-1.385164,
        -1.884956,0.163506,0.573203,
        -1.884956,0.165485,-2.875396,
        -1.884956,0.182564,-2.125821,
        -1.884956,0.190569,1.929141,
        -1.884956,0.201102,1.249082,
        -1.884956,0.201397,-0.716646,
        -1.884956,0.210568,-0.069559,
        -1.884956,0.251672,2.848900,
        -1.884956,0.266262,-1.678068,
        -1.884956,0.269108,-1.215563,
        -1.884956,0.273777,-2.506756,
        -1.884956,0.279854,0.814581,
        -1.884956,0.281664,-2.969147,
        -1.884956,0.283625,2.405603,
        -1.884956,0.284058,0.357350,
        -1.884956,0.294107,1.593880,
        -1.884956,0.302491,-0.400410,
        -1.884956,0.308059,2.021491,
        -1.884956,0.312219,-2.072057,
        -1.884956,0.321905,-0.829121,
        -1.884956,0.329984,1.201621,
        -1.884956,0.339714,-0.013712,
        -1.884956,0.368784,2.974974,
        -1.884956,0.376428,-1.495532,
        -1.884956,0.378103,2.646436,
        -1.884956,0.385088,0.590355,
        -1.884956,0.386248,-2.690943,
        -1.884956,0.389851,-2.340031,
        -1.884956,0.391502,-1.171374,
        -1.884956,0.398249,-1.812694,
        -1.884956,0.400542,-2.995729,
        -1.884956,0.404273,0.918163,
        -1.884956,0.405617,1.773533,
        -1.884956,0.409009,-0.570243,
        -1.884956,0.410700,0.270433,
        -1.884956,0.415893,1.464890,
        -1.884956,0.420680,2.088164,
        -1.884956,0.424766,-0.258729,
        -1.884956,0.426507,2.367872,
        -1.884956,0.441292,-0.890727,
        -1.884956,0.451839,-2.078193,
        -1.884956,0.459270,1.191562,
        -1.884956,0.470622,0.014002,
        -1.884956,0.481292,-2.529098,
        -1.884956,0.482460,2.849843,
        -1.884956,0.492205,-1.397963,
        -1.884956,0.493314,3.117264,
        -1.884956,0.497870,0.727828,
        -1.884956,0.499783,-1.654379,
        -1.884956,0.502651,0.463333,
        -1.884956,0.505463,2.594557,
        -1.884956,0.506104,-2.874910,
        -1.884956,0.513579,-1.147354,
        -1.884956,0.518569,1.644237,
        -1.884956,0.519001,1.900488,
        -1.884956,0.519277,-0.673809,
        -1.884956,0.524769,-0.421955,
        -1.884956,0.525564,-2.288838,
        -1.884956,0.532541,0.978737,
        -1.884956,0.534610,2.154808,
        -1.884956,0.535020,-1.891846,
        -1.884956,0.540208,1.400497,
        -1.884956,0.542006,0.228337,
        -1.884956,0.550762,-0.180771,
        -1.884956,0.560531,2.378474,
        -1.884956,0.561629,-0.924576,
        -1.884956,0.567465,-2.679392,
        -1.884956,0.588527,1.181755,
        -1.884956,0.591643,-2.098364,
        -1.884956,0.594963,2.997575,
        -1.884956,0.599643,-3.045235,
        -1.884956,0.599962,2.773556,
        -1.884956,0.602620,0.029270,
        -1.884956,0.605711,0.598454,
        -1.884956,0.607708,-1.541026,
        -1.884956,0.609360,-2.464379,
        -1.884956,0.610738,-1.326786,
        -1.884956,0.616968,0.820489,
        -1.884956,0.626379,1.773610,
        -1.884956,0.628626,-1.751180,
        -1.884956,0.630671,0.391949,
        -1.884956,0.632622,-0.737642,
        -1.884956,0.632708,2.566722,
        -1.884956,0.632797,1.989858,
        -1.884956,0.634455,-0.534790,
        -1.884956,0.636205,1.561170,
        -1.884956,0.637681,-1.123423,
        -1.884956,0.641441,-2.842109,
        -1.884956,0.644807,-0.327139,
        -1.884956,0.652440,2.197129,
        -1.884956,0.660897,-2.278593,
        -1.884956,0.664008,1.360091,
        -1.884956,0.670850,-1.945257,
        -1.884956,0.671396,1.004859,
        -1.884956,0.673985,0.203855,
        -1.884956,0.680326,-0.133370,
        -1.884956,0.685083,-0.942363,
        -1.884956,0.688204,-2.632204,
        -1.884956,0.689777,2.385011,
        -1.884956,0.700065,3.115430,
        -1.884956,0.704080,2.911636,
        -1.884956,0.712883,1.174216,
        -1.884956,0.715701,0.699906,
        -1.884956,0.718851,-1.442215,
        -1.884956,0.721019,2.721536,
        -1.884956,0.725300,-2.975134,
        -1.884956,0.727136,-1.633288,
        -1.884956,0.727399,-2.120567,
        -1.884956,0.730751,0.522488,
        -1.884956,0.732892,-1.254330,
        -1.884956,0.734333,-2.446184,
        -1.884956,0.735125,0.037270,
        -1.884956,0.737893,1.866463,
        -1.884956,0.740650,1.681075,
        -1.884956,0.745971,-0.781286,
        -1.884956,0.749432,-0.442559,
        -1.884956,0.750102,2.050636,
        -1.884956,0.751046,0.863821,
        -1.884956,0.757018,-0.608833,
        -1.884956,0.757046,-1.816019,
        -1.884956,0.757389,1.501184,
        -1.884956,0.759800,2.546181,
        -1.884956,0.759859,0.351923,
        -1.884956,0.761141,-2.781102,
        -1.884956,0.769035,-1.078574,
        -1.884956,0.770166,-0.266604,
        -1.884956,0.772218,2.227380,
        -1.884956,0.788382,1.331051,
        -1.884956,0.790269,-2.282409,
        -1.884956,0.803410,-1.984567,
        -1.884956,0.803813,0.191520,
        -1.884956,0.806832,3.022155,
        -1.884956,0.807476,-2.597675,
        -1.884956,0.812512,-0.107335,
        -1.884956,0.814161,1.010949,
        -1.884956,0.817732,-3.084486,
        -1.884956,0.817846,2.390688,
        -1.884956,0.819323,2.845960,
        -1.884956,0.819430,-0.915017,
        -1.884956,0.829198,-1.532491,
        -1.884956,0.829531,0.665695,
        -1.884956,0.829816,1.167147,
        -1.884956,0.833613,-1.357492,
        -1.884956,0.842945,-2.907318,
        -1.884956,0.844883,2.679708,
        -1.884956,0.848189,-1.705003,
        -1.884956,0.849888,1.767901,
        -1.884956,0.851126,0.493370,
        -1.884956,0.852911,1.933394,
        -1.884956,0.857516,-2.432863,
        -1.884956,0.858261,-1.191729,
        -1.884956,0.858279,-2.138895,
        -1.884956,0.859723,-0.734928,
        -1.884956,0.859754,1.603292,
        -1.884956,0.861655,-0.394821,
        -1.884956,0.864300,0.043549,
        -1.884956,0.866940,0.851229,
        -1.884956,0.869748,2.095969,
        -1.884956,0.877352,-0.574668,
        -1.884956,0.879254,-2.733338,
        -1.884956,0.882554,1.443757,
        -1.884956,0.884060,0.334199,
        -1.884956,0.885626,-1.866061,
        -1.884956,0.889834,2.529957,
        -1.884956,0.893140,2.251698,
        -1.884956,0.897463,-0.240822,
        -1.884956,0.898758,-1.031948,
        -1.884956,0.913918,1.287618,
        -1.884956,0.916555,-2.287173,
        -1.884956,0.918305,3.107845,
        -1.884956,0.919522,2.948277,
        -1.884956,0.927138,-2.569686,
        -1.884956,0.929031,0.185383,
        -1.884956,0.933766,-3.015917,
        -1.884956,0.934873,-2.010822,
        -1.884956,0.935636,-1.445855,
        -1.884956,0.936702,-0.866461,
        -1.884956,0.937395,0.729108,
        -1.884956,0.937910,2.787575,
        -1.884956,0.939144,-0.092962,
        -1.884956,0.941718,1.125085,
        -1.884956,0.942873,-1.607680,
        -1.884956,0.943701,0.980328,
        -1.884956,0.947774,2.395256,
        -1.884956,0.952433,0.584347,
        -1.884956,0.954456,-1.287293,
        -1.884956,0.958731,-2.852552,
        -1.884956,0.961838,1.839284,
        -1.884956,0.961898,-0.469425,
        -1.884956,0.965426,1.686615,
        -1.884956,0.970108,1.990771,
        -1.884956,0.972820,-1.762780,
        -1.884956,0.973500,-0.711881,
        -1.884956,0.975253,2.641082,
        -1.884956,0.977824,0.441928,
        -1.884956,0.981485,-2.424766,
        -1.884956,0.981877,1.537100,
        -1.884956,0.982609,-1.136920,
        -1.884956,0.984265,-2.156073,
        -1.884956,0.989150,0.045493,
        -1.884956,0.993052,2.137478,
        -1.884956,0.997766,-2.693871,
        -1.884956,0.999847,-0.334738,
        -1.884956,1.001920,0.848054,
        -1.884956,1.006146,1.389734,
        -1.884956,1.012180,0.304066,
        -1.884956,1.016938,2.270674,
        -1.884956,1.017371,-0.979790,
        -1.884956,1.019495,-1.894487,
        -1.884956,1.023796,-0.583691,
        -1.884956,1.026525,2.506945,
        -1.884956,1.027004,-0.203653,
        -1.884956,1.028027,3.036573,
        -1.884956,1.031435,1.241021,
        -1.884956,1.032316,2.884221,
        -1.884956,1.035863,-3.110475,
        -1.884956,1.042085,-2.292950,
        -1.884956,1.043272,-1.522986,
        -1.884956,1.046265,-2.958847,
        -1.884956,1.049272,-1.376344,
        -1.884956,1.050039,-2.552086,
        -1.884956,1.053682,0.689863,
        -1.884956,1.055458,0.170394,
        -1.884956,1.056622,-0.831866,
        -1.884956,1.057515,1.097998,
        -1.884956,1.058706,-2.036438,
        -1.884956,1.061389,-1.672248,
        -1.884956,1.063758,2.743093,
        -1.884956,1.070157,-0.073422,
        -1.884956,1.072154,1.761506,
        -1.884956,1.074029,1.909193,
        -1.884956,1.074204,0.545276,
        -1.884956,1.074714,-2.805147,
        -1.884956,1.076262,0.966104,
        -1.884956,1.077642,-1.227073,
        -1.884956,1.084151,1.618525,
        -1.884956,1.085085,2.377673,
        -1.884956,1.086822,-0.441112,
        -1.884956,1.087142,2.058815,
        -1.884956,1.098383,-0.696430,
        -1.884956,1.099138,-1.084033,
        -1.884956,1.103208,1.477773,
        -1.884956,1.103783,0.408254,
        -1.884956,1.106275,-2.420800,
        -1.884956,1.108079,2.612943,
        -1.884956,1.109345,-2.172718,
        -1.884956,1.115957,-1.787597,
        -1.884956,1.118930,0.045983,
        -1.884956,1.119777,0.834702,
        -1.884956,1.121943,-2.669042,
        -1.884956,1.124930,2.219917,
        -1.884956,1.125217,1.336294,
        -1.884956,1.130958,2.972931,
        -1.884956,1.131400,-0.306629,
        -1.884956,1.137135,-0.936514,
        -1.884956,1.139125,-1.924528,
        -1.884956,1.139877,0.276382,
        -1.884956,1.142003,-3.058022,
        -1.884956,1.147363,-0.558448,
        -1.884956,1.150348,-0.176418,
        -1.884956,1.151198,1.197016,
        -1.884956,1.153930,-1.597104,
        -1.884956,1.155389,2.837148,
        -1.884956,1.155458,-1.460623,
        -1.884956,1.157331,-2.906362,
        -1.884956,1.158594,2.487618,
        -1.884956,1.162103,-1.322553,
        -1.884956,1.165407,3.092861,
        -1.884956,1.168025,0.717075,
        -1.884956,1.168406,-2.301519,
        -1.884956,1.174782,-2.538755,
        -1.884956,1.177998,1.061256,
        -1.884956,1.178898,1.827305,
        -1.884956,1.181331,-2.060330,
        -1.884956,1.182540,-0.798952,
        -1.884956,1.184569,0.147517,
        -1.884956,1.184734,1.968919,
        -1.884956,1.184804,0.593386,
        -1.884956,1.188278,1.690808,
        -1.884956,1.194360,2.110440,
        -1.884956,1.194776,2.710455,
        -1.884956,1.197745,-1.173705,
        -1.884956,1.199898,-2.777257,
        -1.884956,1.201164,2.355931,
        -1.884956,1.203760,1.556404,
        -1.884956,1.205845,-0.415993,
        -1.884956,1.206585,0.933648,
        -1.884956,1.206950,-0.037248,
        -1.884956,1.209341,0.471162,
        -1.884956,1.213483,-1.039417,
        -1.884956,1.218146,-0.672643,
        -1.884956,1.221608,1.421669,
        -1.884956,1.226043,-1.693533,
        -1.884956,1.232870,-1.822773,
        -1.884956,1.234375,-2.421194,
        -1.884956,1.234531,-2.190148,
        -1.884956,1.239414,0.348959,
        -1.884956,1.240183,2.587255,
        -1.884956,1.242312,2.941234,
        -1.884956,1.244075,1.286644,
        -1.884956,1.245961,-2.997461,
        -1.884956,1.246495,-2.646872,
        -1.884956,1.247736,2.242810,
        -1.884956,1.253604,0.812222,
        -1.884956,1.257816,-0.894317,
        -1.884956,1.258247,-1.953634,
        -1.884956,1.258993,-0.284099,
        -1.884956,1.262380,-0.157609,
        -1.884956,1.262791,-1.549229,
        -1.884956,1.263772,-0.533073,
        -1.884956,1.266731,-1.412461,
        -1.884956,1.268591,-3.122146,
        -1.884956,1.269614,-1.275246,
        -1.884956,1.270745,1.152533,
        -1.884956,1.273606,0.224545,
        -1.884956,1.279716,2.812101,
        -1.884956,1.283786,2.462124,
        -1.884956,1.284798,0.078209,
        -1.884956,1.285402,-2.882049,
        -1.884956,1.286936,1.886053,
        -1.884956,1.289408,0.683122,
        -1.884956,1.294736,1.756184,
        -1.884956,1.296978,-2.311511,
        -1.884956,1.298546,2.020601,
        -1.884956,1.300924,1.022478,
        -1.884956,1.303305,3.043152,
        -1.884956,1.304210,-2.084225,
        -1.884956,1.305241,-2.530249,
        -1.884956,1.307513,1.628003,
        -1.884956,1.310722,0.558267,
        -1.884956,1.314142,2.139329,
        -1.884956,1.316087,-0.777971,
        -1.884956,1.317907,-1.121023,
        -1.884956,1.321019,-2.749776,
        -1.884956,1.321609,2.687717,
        -1.884956,1.322568,1.499415,
        -1.884956,1.322972,-0.395699,
        -1.884956,1.325286,-0.994192,
        -1.884956,1.329146,-0.648513,
        -1.884956,1.329795,-0.042570,
        -1.884956,1.331651,2.344497,
        -1.884956,1.336522,0.435098,
        -1.884956,1.339410,0.900087,
        -1.884956,1.339550,1.370353,
        -1.884956,1.340441,-1.731580,
        -1.884956,1.351855,-1.855234,
        -1.884956,1.360161,-3.018404,
        -1.884956,1.361031,-1.613971,
        -1.884956,1.361196,2.914780,
        -1.884956,1.361998,-2.207254,
        -1.884956,1.362302,1.239384,
        -1.884956,1.365228,2.563808,
        -1.884956,1.365452,0.310522,
        -1.884956,1.367028,-2.420041,
        -1.884956,1.373422,-1.484281,
        -1.884956,1.374245,-1.226417,
        -1.884956,1.375020,-0.170393,
        -1.884956,1.376560,-1.355481,
        -1.884956,1.377070,-1.980984,
        -1.884956,1.377578,-2.636339,
        -1.884956,1.378096,0.774626,
        -1.884956,1.379086,-0.508192,
        -1.884956,1.382562,0.177141,
        -1.884956,1.388849,2.233847,
        -1.884956,1.389268,-0.886349,
        -1.884956,1.391324,1.109396,
        -1.884956,1.391699,-0.288725,
        -1.884956,1.392869,3.136903,
        -1.884956,1.393612,1.952016,
        -1.884956,1.395545,-2.850807,
        -1.884956,1.401926,1.822930,
        -1.884956,1.402149,2.788898,
        -1.884956,1.409552,0.647980,
        -1.884956,1.410840,0.056219,
        -1.884956,1.411944,2.444377,
        -1.884956,1.413639,1.697982,
        -1.884956,1.426479,1.573427,
        -1.884956,1.426725,2.102242,
        -1.884956,1.427868,0.984637,
        -1.884956,1.429580,-2.107232,
        -1.884956,1.431375,-2.313269,
        -1.884956,1.431907,3.015410,
        -1.884956,1.432325,-1.029054,
        -1.884956,1.433341,0.524268,
        -1.884956,1.435181,-0.739771,
        -1.884956,1.437208,-2.529085,
        -1.884956,1.439041,-0.617966,
        -1.884956,1.441473,1.448544,
        -1.884956,1.444527,2.664744,
        -1.884956,1.450513,-0.400625,
        -1.884956,1.451907,-0.066823,
        -1.884956,1.452754,-2.738631,
        -1.884956,1.454061,-2.950416,
        -1.884956,1.456736,1.323154,
        -1.884956,1.458370,0.400433,
        -1.884956,1.459314,-1.765003,
        -1.884956,1.459902,-1.145835,
        -1.884956,1.464952,2.329245,
        -1.884956,1.466050,0.862104,
        -1.884956,1.471330,-1.886160,
        -1.884956,1.475119,-1.648255,
        -1.884956,1.480878,1.194576,
        -1.884956,1.480942,0.273235,
        -1.884956,1.481086,2.890269,
        -1.884956,1.484118,-1.529156,
        -1.884956,1.484741,-1.279219,
        -1.884956,1.487681,-1.406069,
        -1.884956,1.487886,-3.066037,
        -1.884956,1.490555,2.543284,
        -1.884956,1.492424,-0.190932,
        -1.884956,1.495002,-0.839810,
        -1.884956,1.495297,2.007188,
        -1.884956,1.496414,-2.009267,
        -1.884956,1.498261,-2.425270,
        -1.884956,1.498907,-2.207327,
        -1.884956,1.500204,0.737638,
        -1.884956,1.502469,0.151156,
        -1.884956,1.507448,-0.514782,
        -1.884956,1.509287,2.204647,
        -1.884956,1.510518,1.884920,
        -1.884956,1.511948,-2.632360,
        -1.884956,1.515235,1.068565,
        -1.884956,1.522237,3.102018,
        -1.884956,1.522373,1.764991,
        -1.884956,1.522416,-0.954956,
        -1.884956,1.523102,2.765920,
        -1.884956,1.524811,-2.844718,
        -1.884956,1.527261,-0.304916,
        -1.884956,1.529788,0.613208,
        -1.884956,1.533648,1.644808,
        -1.884956,1.535018,0.028869,
        -1.884956,1.541082,2.424608,
        -1.884956,1.546010,1.523462,
        -1.884956,1.551135,-0.642994,
        -1.884956,1.553895,0.489607,
        -1.884956,1.555041,0.947813,
        -1.884956,1.559671,-1.074934,
        -1.884956,1.560345,2.985899,
        -1.884956,1.560487,-2.323400,
        -1.884956,1.561127,1.401111,
        -1.884956,1.567193,2.643172,
        -1.884956,1.568892,2.098362,
        -1.884956,1.571702,-2.098435,
        -1.884956,1.572701,-2.528730,
        -1.884956,1.573572,1.278593,
        -1.884956,1.574849,-0.091505,
        -1.884956,1.576334,0.365236,
        -1.884956,1.577029,-2.962888,
        -1.884956,1.577688,-1.814941,
        -1.884956,1.579542,-0.755917,
        -1.884956,1.579624,-1.199759,
        -1.884956,1.581515,-0.419414,
        -1.884956,1.584581,-2.738702,
        -1.884956,1.586948,-1.696296,
        -1.884956,1.588901,2.301496,
        -1.884956,1.589936,0.825020,
        -1.884956,1.590018,-1.941652,
        -1.884956,1.596968,-1.328358,
        -1.884956,1.597178,-1.575607,
        -1.884956,1.600470,-1.453196,
        -1.884956,1.601408,0.242409,
        -1.884956,1.602259,1.152656,
        -1.884956,1.603064,2.865774,
        -1.884956,1.612085,1.968802,
        -1.884956,1.612720,-3.080498,
        -1.884956,1.616061,2.521531,
        -1.884956,1.616900,-0.207457,
        -1.884956,1.620988,-0.875920,
        -1.884956,1.622114,0.701509,
        -1.884956,1.622749,0.121426,
        -1.884956,1.622993,-2.221629,
        -1.884956,1.629346,-0.542669,
        -1.884956,1.630007,1.844552,
        -1.884956,1.634747,-2.426198,
        -1.884956,1.642346,1.722617,
        -1.884956,1.643925,1.034629,
        -1.884956,1.644848,2.743774,
        -1.884956,1.645258,2.188109,
        -1.884956,1.645417,-2.634725,
        -1.884956,1.647395,3.085437,
        -1.884956,1.648538,-2.855044,
        -1.884956,1.649932,0.576808,
        -1.884956,1.652224,-0.996105,
        -1.884956,1.652871,1.600681,
        -1.884956,1.657997,-0.325127,
        -1.884956,1.661936,0.001037,
        -1.884956,1.665187,1.477402,
        -1.884956,1.665496,2.397872,
        -1.884956,1.673267,-0.665453,
        -1.884956,1.673557,0.452526,
        -1.884956,1.679277,-1.122421,
        -1.884956,1.679696,0.912482,
        -1.884956,1.682420,1.353922,
        -1.884956,1.685031,-2.009038,
        -1.884956,1.685196,2.967046,
        -1.884956,1.690107,2.620910,
        -1.884956,1.690453,-2.123076,
        -1.884956,1.691693,-1.871572,
        -1.884956,1.691817,2.061688,
        -1.884956,1.692644,1.237670,
        -1.884956,1.696386,0.328123,
        -1.884956,1.697162,-1.249026,
        -1.884956,1.697488,-2.322785,
        -1.884956,1.698726,-2.973576,
        -1.884956,1.699172,-1.749717,
        -1.884956,1.704478,-0.106295,
        -1.884956,1.707244,-0.444770,
        -1.884956,1.707500,-2.531497,
        -1.884956,1.707714,-0.787321,
        -1.884956,1.708430,-1.627552,
        -1.884956,1.710583,-1.376890,
        -1.884956,1.712415,-2.747921,
        -1.884956,1.713125,-1.503073,
        -1.884956,1.713452,0.789419,
        -1.884956,1.720612,2.279888,
        -1.884956,1.724852,0.202831,
        -1.884956,1.724868,2.846246,
        -1.884956,1.728940,1.931535,
        -1.884956,1.732838,1.127584,
        -1.884956,1.735719,-3.092485,
        -1.884956,1.739676,2.496017,
        -1.884956,1.739971,-0.228876,
        -1.884956,1.743328,0.663347,
        -1.884956,1.744927,-0.911892,
        -1.884956,1.745952,0.084733,
        -1.884956,1.748077,1.804354,
        -1.884956,1.752959,-0.567548,
        -1.884956,1.760169,1.678132,
        -1.884956,1.763887,-2.218450,
        -1.884956,1.766737,2.722752,
        -1.884956,1.767999,1.004737,
        -1.884956,1.769149,1.550298,
        -1.884956,1.770119,-2.427175,
        -1.884956,1.770460,2.154208,
        -1.884956,1.770514,3.070982,
        -1.884956,1.770735,0.537245,
        -1.884956,1.772673,-2.864760,
        -1.884956,1.773527,-1.039582,
        -1.884956,1.776419,-2.642421,
        -1.884956,1.780769,1.416330,
        -1.884956,1.786634,-0.346247,
        -1.884956,1.788829,-1.947553,
        -1.884956,1.793737,0.410270,
        -1.884956,1.794553,2.374618,
        -1.884956,1.794699,-0.694061,
        -1.884956,1.797469,-1.169105,
        -1.884956,1.803705,0.881076,
        -1.884956,1.804124,1.264249,
        -1.884956,1.807825,-2.090455,
        -1.884956,1.808822,-1.817375,
        -1.884956,1.809020,2.949741,
        -1.884956,1.812466,2.597595,
        -1.884956,1.813159,2.025395,
        -1.884956,1.813816,-1.298435,
        -1.884956,1.815016,0.278951,
        -1.884956,1.816510,-0.008631,
        -1.884956,1.817772,-1.689625,
        -1.884956,1.821360,-2.984352,
        -1.884956,1.822312,-0.127211,
        -1.884956,1.823386,-1.429807,
        -1.884956,1.823834,-1.560483,
        -1.884956,1.832060,-0.821627,
        -1.884956,1.832690,-0.468440,
        -1.884956,1.833945,-2.320278,
        -1.884956,1.834787,0.751968,
        -1.884956,1.840579,-2.536445,
        -1.884956,1.841033,-2.756932,
        -1.884956,1.846350,1.893983,
        -1.884956,1.846644,1.113766,
        -1.884956,1.846833,2.826543,
        -1.884956,1.847325,2.249051,
        -1.884956,1.854119,0.134161,
        -1.884956,1.858863,-3.106022,
        -1.884956,1.865046,1.763384,
        -1.884956,1.865619,-0.952036,
        -1.884956,1.865912,0.624453,
        -1.884956,1.866694,2.473686,
        -1.884956,1.867524,-0.240083,
        -1.884956,1.875561,1.632831,
        -1.884956,1.876049,-0.594378,
        -1.884956,1.880509,1.500204,
        -1.884956,1.883504,1.360457,
        -1.884956,1.886971,-2.196665,
        -1.884956,1.888979,0.982922,
        -1.884956,1.889815,2.701570,
        -1.884956,1.892144,-2.005340,
        -1.884956,1.892348,0.496522,
        -1.884956,1.893281,-1.083976,
        -1.884956,1.893828,3.053453,
        -1.884956,1.895117,2.121107,
        -1.884956,1.898741,-2.875735,
        -1.884956,1.904429,-2.428029,
        -1.884956,1.908408,-2.648903,
        -1.884956,1.911828,-0.364489,
        -1.884956,1.912224,0.366565,
        -1.884956,1.914854,1.214468,
        -1.884956,1.915467,-1.883143,
        -1.884956,1.915763,-0.724679,
        -1.884956,1.915790,-1.216366,
        -1.884956,1.922256,2.347806,
        -1.884956,1.922616,0.844633,
        -1.884956,1.925185,0.231170,
        -1.884956,1.927474,-1.754616,
        -1.884956,1.928342,-1.351174,
        -1.884956,1.933328,-1.623328,
        -1.884956,1.933878,0.034150,
        -1.884956,1.934585,-1.487979,
        -1.884956,1.934802,2.927634,
        -1.884956,1.934862,1.989776,
        -1.884956,1.936327,-0.099350,
        -1.884956,1.938088,2.578507,
        -1.884956,1.945677,-2.999805,
        -1.884956,1.954572,-0.859208,
        -1.884956,1.957868,-0.490784,
        -1.884956,1.958989,0.712772,
        -1.884956,1.960212,-2.307561,
        -1.884956,1.964774,1.855507,
        -1.884956,1.964971,-2.100533,
        -1.884956,1.969112,1.088273,
        -1.884956,1.971438,-2.765415,
        -1.884956,1.972312,2.805134,
        -1.884956,1.974950,2.220596,
        -1.884956,1.975135,-2.539127,
        -1.884956,1.982107,1.720829,
        -1.884956,1.983195,-3.127984,
        -1.884956,1.983318,-0.247182,
        -1.884956,1.986720,-0.993220,
        -1.884956,1.990381,1.586005,
        -1.884956,1.990470,0.580283,
        -1.884956,1.990523,1.450326,
        -1.884956,1.993909,1.312897,
        -1.884956,1.994694,2.452813,
        -1.884956,1.999226,-0.620717,
        -1.884956,2.005145,0.941012,
        -1.884956,2.007301,0.142337,
        -1.884956,2.014149,0.444138,
        -1.884956,2.014338,-1.127392,
        -1.884956,2.018131,3.021778,
        -1.884956,2.019676,-1.980944,
        -1.884956,2.019912,2.686612,
        -1.884956,2.020545,2.088672,
        -1.884956,2.026074,0.300945,
        -1.884956,2.026196,-2.889057,
        -1.884956,2.031959,-1.268178,
        -1.884956,2.032917,-2.199917,
        -1.884956,2.033572,-1.842884,
        -1.884956,2.034712,-2.419505,
        -1.884956,2.035818,-0.379479,
        -1.884956,2.036004,-0.757062,
        -1.884956,2.040501,-1.411488,
        -1.884956,2.042874,-1.701367,
        -1.884956,2.043060,-2.653746,
        -1.884956,2.043729,-1.557938,
        -1.884956,2.048538,0.806623,
        -1.884956,2.049006,1.199930,
        -1.884956,2.049502,-0.137989,
        -1.884956,2.052181,2.326959,
        -1.884956,2.052794,-0.008526,
        -1.884956,2.057608,1.951726,
        -1.884956,2.065568,2.567730,
        -1.884956,2.070371,2.879808,
        -1.884956,2.071256,-3.018961,
        -1.884956,2.079746,-0.897699,
        -1.884956,2.082784,1.042403,
        -1.884956,2.083752,-0.511086,
        -1.884956,2.084401,1.810502,
        -1.884956,2.085731,0.671081,
        -1.884956,2.097496,1.373172,
        -1.884956,2.099185,1.667006,
        -1.884956,2.100482,-2.301084,
        -1.884956,2.103118,2.194745,
        -1.884956,2.103223,1.521685,
        -1.884956,2.103695,-2.777707,
        -1.884956,2.107523,3.129581,
        -1.884956,2.108277,-2.095907,
        -1.884956,2.108891,-2.532166,
        -1.884956,2.111003,-0.268632,
        -1.884956,2.112276,-1.027246,
        -1.884956,2.114380,0.531457,
        -1.884956,2.119166,0.100552,
        -1.884956,2.119564,2.754818,
        -1.884956,2.122474,-0.648314,
        -1.884956,2.125183,2.445598,
        -1.884956,2.129501,0.238290,
        -1.884956,2.131178,0.386663,
        -1.884956,2.131223,-1.174311,
        -1.884956,2.134269,-1.948042,
        -1.884956,2.134684,0.907065,
        -1.884956,2.137615,2.988806,
        -1.884956,2.145635,-1.327708,
        -1.884956,2.146467,2.055081,
        -1.884956,2.146582,-1.793788,
        -1.884956,2.149210,-1.482321,
        -1.884956,2.152901,-1.636788,
        -1.884956,2.156337,1.151558,
        -1.884956,2.156474,-2.912065,
        -1.884956,2.157194,-0.792563,
        -1.884956,2.166724,-0.401638,
        -1.884956,2.166980,-0.134879,
        -1.884956,2.173644,-2.403429,
        -1.884956,2.175711,-2.655523,
        -1.884956,2.177132,2.600419,
        -1.884956,2.178879,0.767722,
        -1.884956,2.180752,1.908485,
        -1.884956,2.181338,2.313721,
        -1.884956,2.185580,1.287673,
        -1.884956,2.195400,-2.217546,
        -1.884956,2.200418,-3.056442,
        -1.884956,2.201829,-0.001735,
        -1.884956,2.203804,1.756176,
        -1.884956,2.211602,1.443930,
        -1.884956,2.212686,-0.544809,
        -1.884956,2.213937,0.621892,
        -1.884956,2.214075,1.600691,
        -1.884956,2.216256,1.020029,
        -1.884956,2.219060,2.886631,
        -1.884956,2.223375,-0.906185,
        -1.884956,2.226002,-1.060248,
        -1.884956,2.226413,-2.064034,
        -1.884956,2.230554,2.171334,
        -1.884956,2.232413,2.735777,
        -1.884956,2.233709,3.070830,
        -1.884956,2.234383,-2.792065,
        -1.884956,2.237183,0.468799,
        -1.884956,2.237695,-0.271388,
        -1.884956,2.239419,0.151693,
        -1.884956,2.243347,-2.521049,
        -1.884956,2.244035,-1.229456,
        -1.884956,2.244355,0.309899,
        -1.884956,2.245175,2.460236,
        -1.884956,2.245628,-1.899216,
        -1.884956,2.245973,-0.702234,
        -1.884956,2.255601,-1.728131,
        -1.884956,2.256702,-1.401367,
        -1.884956,2.261466,-1.557500,
        -1.884956,2.267390,0.878499,
        -1.884956,2.272099,2.018594,
        -1.884956,2.278618,1.170386,
        -1.884956,2.283304,-2.345584,
        -1.884956,2.285175,-2.943090,
        -1.884956,2.293000,-0.124451,
        -1.884956,2.294058,-0.421070,
        -1.884956,2.303283,1.855938,
        -1.884956,2.304968,2.608978,
        -1.884956,2.305347,2.311346,
        -1.884956,2.307188,-2.656134,
        -1.884956,2.307301,1.336200,
        -1.884956,2.309655,0.726020,
        -1.884956,2.317634,-2.185225,
        -1.884956,2.321510,1.685664,
        -1.884956,2.324376,-3.111006,
        -1.884956,2.324731,1.511186,
        -1.884956,2.329587,0.035498,
        -1.884956,2.330158,2.977264,
        -1.884956,2.335096,-0.588468,
        -1.884956,2.336046,-1.115799,
        -1.884956,2.338060,2.803613,
        -1.884956,2.338083,-0.943400,
        -1.884956,2.340210,0.562145,
        -1.884956,2.341931,1.023926,
        -1.884956,2.343265,-0.779061,
        -1.884956,2.343434,-2.013692,
        -1.884956,2.353029,-1.310529,
        -1.884956,2.354868,0.211543,
        -1.884956,2.355120,0.388593,
        -1.884956,2.355165,-1.833173,
        -1.884956,2.355462,2.151441,
        -1.884956,2.362948,-2.487477,
        -1.884956,2.363101,-2.810456,
        -1.884956,2.363394,-0.272918,
        -1.884956,2.364509,-1.644533,
        -1.884956,2.370465,2.466824,
        -1.884956,2.386110,-1.481351,
        -1.884956,2.394960,0.862133,
        -1.884956,2.395716,1.977184,
        -1.884956,2.396339,1.200900,
        -1.884956,2.407941,-2.316753,
        -1.884956,2.409450,-2.988432,
        -1.884956,2.417202,-0.103363,
        -1.884956,2.417386,-0.443097,
        -1.884956,2.417950,2.666662,
        -1.884956,2.423190,1.789783,
        -1.884956,2.423692,1.392696,
        -1.884956,2.431122,3.086379,
        -1.884956,2.431406,-2.647820,
        -1.884956,2.431584,2.307631,
        -1.884956,2.434090,1.592818,
        -1.884956,2.435993,0.683844,
        -1.884956,2.440898,-2.133300,
        -1.884956,2.442222,-1.195837,
        -1.884956,2.443112,2.883761,
        -1.884956,2.449191,-0.633651,
        -1.884956,2.450821,0.084741,
        -1.884956,2.453499,-1.008861,
        -1.884956,2.460010,-1.938011,
        -1.884956,2.460199,-0.822008,
        -1.884956,2.460902,0.490079,
        -1.884956,2.461774,-1.734913,
        -1.884956,2.463986,1.034407,
        -1.884956,2.467647,0.287667,
        -1.884956,2.479849,2.126925,
        -1.884956,2.485907,-2.835471,
        -1.884956,2.485945,-1.374853,
        -1.884956,2.487747,2.505912,
        -1.884956,2.487957,-0.274279,
        -1.884956,2.489741,-2.468428,
        -1.884956,2.512538,1.239900,
        -1.884956,2.517528,1.923289,
        -1.884956,2.518454,-1.565852,
        -1.884956,2.519348,0.843503,
        -1.884956,2.522766,-3.059231,
        -1.884956,2.528686,2.737228,
        -1.884956,2.534798,-2.266671,
        -1.884956,2.536533,1.466274,
        -1.884956,2.538504,1.699812,
        -1.884956,2.539516,-0.476942,
        -1.884956,2.539532,-0.074548,
        -1.884956,2.543197,2.990252,
        -1.884956,2.554022,2.320619,
        -1.884956,2.556919,-2.655005,
        -1.884956,2.558146,0.627191,
        -1.884956,2.558223,-1.164189,
        -1.884956,2.563452,-2.044028,
        -1.884956,2.568462,0.150655,
        -1.884956,2.569130,-0.700837,
        -1.884956,2.570050,-1.804492,
        -1.884956,2.572710,-0.934665,
        -1.884956,2.575994,0.391718,
        -1.884956,2.584774,1.046703,
        -1.884956,2.601432,2.102535,
        -1.884956,2.603559,2.555114,
        -1.884956,2.604150,-2.885208,
        -1.884956,2.608630,-1.378011,
        -1.884956,2.611256,-0.279110,
        -1.884956,2.617344,-2.439480,
        -1.884956,2.627402,1.293552,
        -1.884956,2.635396,1.848850,
        -1.884956,2.635686,2.834950,
        -1.884956,2.637087,3.132730,
        -1.884956,2.642029,0.813274,
        -1.884956,2.645260,1.570203,
        -1.884956,2.646182,-1.614036,
        -1.884956,2.660113,-2.186513,
        -1.884956,2.660483,-0.034608,
        -1.884956,2.661713,-0.529511,
        -1.884956,2.672720,-1.110422,
        -1.884956,2.676626,2.335353,
        -1.884956,2.676638,0.540364,
        -1.884956,2.676780,-1.901158,
        -1.884956,2.681411,-2.667965,
        -1.884956,2.682152,0.245751,
        -1.884956,2.682859,-0.817071,
        -1.884956,2.705107,1.062142,
        -1.884956,2.717705,2.625995,
        -1.884956,2.721001,2.060097,
        -1.884956,2.722319,-2.954069,
        -1.884956,2.729840,-1.368237,
        -1.884956,2.734349,-0.289298,
        -1.884956,2.738409,2.975671,
        -1.884956,2.740366,1.374341,
        -1.884956,2.744633,-2.391311,
        -1.884956,2.747398,1.727166,
        -1.884956,2.763799,0.760372,
        -1.884956,2.768238,-1.679243,
        -1.884956,2.779245,-2.046268,
        -1.884956,2.780208,-0.626453,
        -1.884956,2.780471,0.027732,
        -1.884956,2.784696,-1.013128,
        -1.884956,2.789817,0.395431,
        -1.884956,2.801154,2.353143,
        -1.884956,2.807091,-2.687477,
        -1.884956,2.825762,1.085035,
        -1.884956,2.830209,2.734943,
        -1.884956,2.833517,-3.085508,
        -1.884956,2.838301,1.980061,
        -1.884956,2.849871,1.512012,
        -1.884956,2.850089,-1.343899,
        -1.884956,2.856905,-0.318441,
        -1.884956,2.870445,-2.291001,
        -1.884956,2.885017,0.655562,
        -1.884956,2.886428,-1.793217,
        -1.884956,2.890337,-0.825633,
        -1.884956,2.898932,0.139973,
        -1.884956,2.931927,-2.728768,
        -1.884956,2.932096,2.377790,
        -1.884956,2.937409,2.947245,
        -1.884956,2.948048,1.133127,
        -1.884956,2.950896,1.797889,
        -1.884956,2.970019,-1.282382,
        -1.884956,2.978333,-0.412090,
        -1.884956,2.996437,-2.089121,
        -1.884956,3.011274,0.434110,
        -1.884956,3.050410,3.138492,
        -1.884956,3.064492,1.647259,
        -1.884956,3.088868,-0.989501,
        -1.780236,0.053884,-2.408394,
        -1.780236,0.077645,1.453433,
        -1.780236,0.088408,-0.311160,
        -1.780236,0.142314,2.590387,
        -1.780236,0.147206,-1.385164,
        -1.780236,0.163506,0.573203,
        -1.780236,0.165485,-2.875396,
        -1.780236,0.182564,-2.125821,
        -1.780236,0.190569,1.929141,
        -1.780236,0.201102,1.249082,
        -1.780236,0.201397,-0.716646,
        -1.780236,0.210568,-0.069559,
        -1.780236,0.251672,2.848900,
        -1.780236,0.266262,-1.678068,
        -1.780236,0.269108,-1.215563,
        -1.780236,0.273777,-2.506756,
        -1.780236,0.279854,0.814581,
        -1.780236,0.281664,-2.969147,
        -1.780236,0.283625,2.405603,
        -1.780236,0.284058,0.357350,
        -1.780236,0.294107,1.593880,
        -1.780236,0.302491,-0.400410,
        -1.780236,0.308059,2.021491,
        -1.780236,0.312219,-2.072057,
        -1.780236,0.321905,-0.829121,
        -1.780236,0.329984,1.201621,
        -1.780236,0.339714,-0.013712,
        -1.780236,0.368784,2.974974,
        -1.780236,0.376428,-1.495532,
        -1.780236,0.378103,2.646436,
        -1.780236,0.385088,0.590355,
        -1.780236,0.386248,-2.690943,
        -1.780236,0.389851,-2.340031,
        -1.780236,0.391502,-1.171374,
        -1.780236,0.398249,-1.812694,
        -1.780236,0.400542,-2.995729,
        -1.780236,0.404273,0.918163,
        -1.780236,0.405617,1.773533,
        -1.780236,0.409009,-0.570243,
        -1.780236,0.410700,0.270433,
        -1.780236,0.415893,1.464890,
        -1.780236,0.420680,2.088164,
        -1.780236,0.424766,-0.258729,
        -1.780236,0.426507,2.367872,
        -1.780236,0.441292,-0.890727,
        -1.780236,0.451839,-2.078193,
        -1.780236,0.459270,1.191562,
        -1.780236,0.470622,0.014002,
        -1.780236,0.481292,-2.529098,
        -1.780236,0.482460,2.849843,
        -1.780236,0.492205,-1.397963,
        -1.780236,0.493314,3.117264,
        -1.780236,0.497870,0.727828,
        -1.780236,0.499783,-1.654379,
        -1.780236,0.502651,0.463333,
        -1.780236,0.505463,2.594557,
        -1.780236,0.506104,-2.874910,
        -1.780236,0.513579,-1.147354,
        -1.780236,0.518569,1.644237,
        -1.780236,0.519001,1.900488,
        -1.780236,0.519277,-0.673809,
        -1.780236,0.524769,-0.421955,
        -1.780236,0.525564,-2.288838,
        -1.780236,0.532541,0.978737,
        -1.780236,0.534610,2.154808,
        -1.780236,0.535020,-1.891846,
        -1.780236,0.540208,1.400497,
        -1.780236,0.542006,0.228337,
        -1.780236,0.550762,-0.180771,
        -1.780236,0.560531,2.378474,
        -1.780236,0.561629,-0.924576,
        -1.780236,0.567465,-2.679392,
        -1.780236,0.588527,1.181755,
        -1.780236,0.591643,-2.098364,
        -1.780236,0.594963,2.997575,
        -1.780236,0.599643,-3.045235,
        -1.780236,0.599962,2.773556,
        -1.780236,0.602620,0.029270,
        -1.780236,0.605711,0.598454,
        -1.780236,0.607708,-1.541026,
        -1.780236,0.609360,-2.464379,
        -1.780236,0.610738,-1.326786,
        -1.780236,0.616968,0.820489,
        -1.780236,0.626379,1.773610,
        -1.780236,0.628626,-1.751180,
        -1.780236,0.630671,0.391949,
        -1.780236,0.632622,-0.737642,
        -1.780236,0.632708,2.566722,
        -1.780236,0.632797,1.989858,
        -1.780236,0.634455,-0.534790,
        -1.780236,0.636205,1.561170,
        -1.780236,0.637681,-1.123423,
        -1.780236,0.641441,-2.842109,
        -1.780236,0.644807,-0.327139,
        -1.780236,0.652440,2.197129,
        -1.780236,0.660897,-2.278593,
        -1.780236,0.664008,1.360091,
        -1.780236,0.670850,-1.945257,
        -1.780236,0.671396,1.004859,
        -1.780236,0.673985,0.203855,
        -1.780236,0.680326,-0.133370,
        -1.780236,0.685083,-0.942363,
        -1.780236,0.688204,-2.632204,
        -1.780236,0.689777,2.385011,
        -1.780236,0.700065,3.115430,
        -1.780236,0.704080,2.911636,
        -1.780236,0.712883,1.174216,
        -1.780236,0.715701,0.699906,
        -1.780236,0.718851,-1.442215,
        -1.780236,0.721019,2.721536,
        -1.780236,0.725300,-2.975134,
        -1.780236,0.727136,-1.633288,
        -1.780236,0.727399,-2.120567,
        -1.780236,0.730751,0.522488,
        -1.780236,0.732892,-1.254330,
        -1.780236,0.734333,-2.446184,
        -1.780236,0.735125,0.037270,
        -1.780236,0.737893,1.866463,
        -1.780236,0.740650,1.681075,
        -1.780236,0.745971,-0.781286,
        -1.780236,0.749432,-0.442559,
        -1.780236,0.750102,2.050636,
        -1.780236,0.751046,0.863821,
        -1.780236,0.757018,-0.608833,
        -1.780236,0.757046,-1.816019,
        -1.780236,0.757389,1.501184,
        -1.780236,0.759800,2.546181,
        -1.780236,0.759859,0.351923,
        -1.780236,0.761141,-2.781102,
        -1.780236,0.769035,-1.078574,
        -1.780236,0.770166,-0.266604,
        -1.780236,0.772218,2.227380,
        -1.780236,0.788382,1.331051,
        -1.780236,0.790269,-2.282409,
        -1.780236,0.803410,-1.984567,
        -1.780236,0.803813,0.191520,
        -1.780236,0.806832,3.022155,
        -1.780236,0.807476,-2.597675,
        -1.780236,0.812512,-0.107335,
        -1.780236,0.814161,1.010949,
        -1.780236,0.817732,-3.084486,
        -1.780236,0.817846,2.390688,
        -1.780236,0.819323,2.845960,
        -1.780236,0.819430,-0.915017,
        -1.780236,0.829198,-1.532491,
        -1.780236,0.829531,0.665695,
        -1.780236,0.829816,1.167147,
        -1.780236,0.833613,-1.357492,
        -1.780236,0.842945,-2.907318,
        -1.780236,0.844883,2.679708,
        -1.780236,0.848189,-1.705003,
        -1.780236,0.849888,1.767901,
        -1.780236,0.851126,0.493370,
        -1.780236,0.852911,1.933394,
        -1.780236,0.857516,-2.432863,
        -1.780236,0.858261,-1.191729,
        -1.780236,0.858279,-2.138895,
        -1.780236,0.859723,-0.734928,
        -1.780236,0.859754,1.603292,
        -1.780236,0.861655,-0.394821,
        -1.780236,0.864300,0.043549,
        -1.780236,0.866940,0.851229,
        -1.780236,0.869748,2.095969,
        -1.780236,0.877352,-0.574668,
        -1.780236,0.879254,-2.733338,
        -1.780236,0.882554,1.443757,
        -1.780236,0.884060,0.334199,
        -1.780236,0.885626,-1.866061,
        -1.780236,0.889834,2.529957,
        -1.780236,0.893140,2.251698,
        -1.780236,0.897463,-0.240822,
        -1.780236,0.898758,-1.031948,
        -1.780236,0.913918,1.287618,
        -1.780236,0.916555,-2.287173,
        -1.780236,0.918305,3.107845,
        -1.780236,0.919522,2.948277,
        -1.780236,0.927138,-2.569686,
        -1.780236,0.929031,0.185383,
        -1.780236,0.933766,-3.015917,
        -1.780236,0.934873,-2.010822,
        -1.780236,0.935636,-1.445855,
        -1.780236,0.936702,-0.866461,
        -1.780236,0.937395,0.729108,
        -1.780236,0.937910,2.787575,
        -1.780236,0.939144,-0.092962,
        -1.780236,0.941718,1.125085,
        -1.780236,0.942873,-1.607680,
        -1.780236,0.943701,0.980328,
        -1.780236,0.947774,2.395256,
        -1.780236,0.952433,0.584347,
        -1.780236,0.954456,-1.287293,
        -1.780236,0.958731,-2.852552,
        -1.780236,0.961838,1.839284,
        -1.780236,0.961898,-0.469425,
        -1.780236,0.965426,1.686615,
        -1.780236,0.970108,1.990771,
        -1.780236,0.972820,-1.762780,
        -1.780236,0.973500,-0.711881,
        -1.780236,0.975253,2.641082,
        -1.780236,0.977824,0.441928,
        -1.780236,0.981485,-2.424766,
        -1.780236,0.981877,1.537100,
        -1.780236,0.982609,-1.136920,
        -1.780236,0.984265,-2.156073,
        -1.780236,0.989150,0.045493,
        -1.780236,0.993052,2.137478,
        -1.780236,0.997766,-2.693871,
        -1.780236,0.999847,-0.334738,
        -1.780236,1.001920,0.848054,
        -1.780236,1.006146,1.389734,
        -1.780236,1.012180,0.304066,
        -1.780236,1.016938,2.270674,
        -1.780236,1.017371,-0.979790,
        -1.780236,1.019495,-1.894487,
        -1.780236,1.023796,-0.583691,
        -1.780236,1.026525,2.506945,
        -1.780236,1.027004,-0.203653,
        -1.780236,1.028027,3.036573,
        -1.780236,1.031435,1.241021,
        -1.780236,1.032316,2.884221,
        -1.780236,1.035863,-3.110475,
        -1.780236,1.042085,-2.292950,
        -1.780236,1.043272,-1.522986,
        -1.780236,1.046265,-2.958847,
        -1.780236,1.049272,-1.376344,
        -1.780236,1.050039,-2.552086,
        -1.780236,1.053682,0.689863,
        -1.780236,1.055458,0.170394,
        -1.780236,1.056622,-0.831866,
        -1.780236,1.057515,1.097998,
        -1.780236,1.058706,-2.036438,
        -1.780236,1.061389,-1.672248,
        -1.780236,1.063758,2.743093,
        -1.780236,1.070157,-0.073422,
        -1.780236,1.072154,1.761506,
        -1.780236,1.074029,1.909193,
        -1.780236,1.074204,0.545276,
        -1.780236,1.074714,-2.805147,
        -1.780236,1.076262,0.966104,
        -1.780236,1.077642,-1.227073,
        -1.780236,1.084151,1.618525,
        -1.780236,1.085085,2.377673,
        -1.780236,1.086822,-0.441112,
        -1.780236,1.087142,2.058815,
        -1.780236,1.098383,-0.696430,
        -1.780236,1.099138,-1.084033,
        -1.780236,1.103208,1.477773,
        -1.780236,1.103783,0.408254,
        -1.780236,1.106275,-2.420800,
        -1.780236,1.108079,2.612943,
        -1.780236,1.109345,-2.172718,
        -1.780236,1.115957,-1.787597,
        -1.780236,1.118930,0.045983,
        -1.780236,1.119777,0.834702,
        -1.780236,1.121943,-2.669042,
        -1.780236,1.124930,2.219917,
        -1.780236,1.125217,1.336294,
        -1.780236,1.130958,2.972931,
        -1.780236,1.131400,-0.306629,
        -1.780236,1.137135,-0.936514,
        -1.780236,1.139125,-1.924528,
        -1.780236,1.139877,0.276382,
        -1.780236,1.142003,-3.058022,
        -1.780236,1.147363,-0.558448,
        -1.780236,1.150348,-0.176418,
        -1.780236,1.151198,1.197016,
        -1.780236,1.153930,-1.597104,
        -1.780236,1.155389,2.837148,
        -1.780236,1.155458,-1.460623,
        -1.780236,1.157331,-2.906362,
        -1.780236,1.158594,2.487618,
        -1.780236,1.162103,-1.322553,
        -1.780236,1.165407,3.092861,
        -1.780236,1.168025,0.717075,
        -1.780236,1.168406,-2.301519,
        -1.780236,1.174782,-2.538755,
        -1.780236,1.177998,1.061256,
        -1.780236,1.178898,1.827305,
        -1.780236,1.181331,-2.060330,
        -1.780236,1.182540,-0.798952,
        -1.780236,1.184569,0.147517,
        -1.780236,1.184734,1.968919,
        -1.780236,1.184804,0.593386,
        -1.780236,1.188278,1.690808,
        -1.780236,1.194360,2.110440,
        -1.780236,1.194776,2.710455,
        -1.780236,1.197745,-1.173705,
        -1.780236,1.199898,-2.777257,
        -1.780236,1.201164,2.355931,
        -1.780236,1.203760,1.556404,
        -1.780236,1.205845,-0.415993,
        -1.780236,1.206585,0.933648,
        -1.780236,1.206950,-0.037248,
        -1.780236,1.209341,0.471162,
        -1.780236,1.213483,-1.039417,
        -1.780236,1.218146,-0.672643,
        -1.780236,1.221608,1.421669,
        -1.780236,1.226043,-1.693533,
        -1.780236,1.232870,-1.822773,
        -1.780236,1.234375,-2.421194,
        -1.780236,1.234531,-2.190148,
        -1.780236,1.239414,0.348959,
        -1.780236,1.240183,2.587255,
        -1.780236,1.242312,2.941234,
        -1.780236,1.244075,1.286644,
        -1.780236,1.245961,-2.997461,
        -1.780236,1.246495,-2.646872,
        -1.780236,1.247736,2.242810,
        -1.780236,1.253604,0.812222,
        -1.780236,1.257816,-0.894317,
        -1.780236,1.258247,-1.953634,
        -1.780236,1.258993,-0.284099,
        -1.780236,1.262380,-0.157609,
        -1.780236,1.262791,-1.549229,
        -1.780236,1.263772,-0.533073,
        -1.780236,1.266731,-1.412461,
        -1.780236,1.268591,-3.122146,
        -1.780236,1.269614,-1.275246,
        -1.780236,1.270745,1.152533,
        -1.780236,1.273606,0.224545,
        -1.780236,1.279716,2.812101,
        -1.780236,1.283786,2.462124,
        -1.780236,1.284798,0.078209,
        -1.780236,1.285402,-2.882049,
        -1.780236,1.286936,1.886053,
        -1.780236,1.289408,0.683122,
        -1.780236,1.294736,1.756184,
        -1.780236,1.296978,-2.311511,
        -1.780236,1.298546,2.020601,
        -1.780236,1.300924,1.022478,
        -1.780236,1.303305,3.043152,
        -1.780236,1.304210,-2.084225,
        -1.780236,1.305241,-2.530249,
        -1.780236,1.307513,1.628003,
        -1.780236,1.310722,0.558267,
        -1.780236,1.314142,2.139329,
        -1.780236,1.316087,-0.777971,
        -1.780236,1.317907,-1.121023,
        -1.780236,1.321019,-2.749776,
        -1.780236,1.321609,2.687717,
        -1.780236,1.322568,1.499415,
        -1.780236,1.322972,-0.395699,
        -1.780236,1.325286,-0.994192,
        -1.780236,1.329146,-0.648513,
        -1.780236,1.329795,-0.042570,
        -1.780236,1.331651,2.344497,
        -1.780236,1.336522,0.435098,
        -1.780236,1.339410,0.900087,
        -1.780236,1.339550,1.370353,
        -1.780236,1.340441,-1.731580,
        -1.780236,1.351855,-1.855234,
        -1.780236,1.360161,-3.018404,
        -1.780236,1.361031,-1.613971,
        -1.780236,1.361196,2.914780,
        -1.780236,1.361998,-2.207254,
        -1.780236,1.362302,1.239384,
        -1.780236,1.365228,2.563808,
        -1.780236,1.365452,0.310522,
        -1.780236,1.367028,-2.420041,
        -1.780236,1.373422,-1.484281,
        -1.780236,1.374245,-1.226417,
        -1.780236,1.375020,-0.170393,
        -1.780236,1.376560,-1.355481,
        -1.780236,1.377070,-1.980984,
        -1.780236,1.377578,-2.636339,
        -1.780236,1.378096,0.774626,
        -1.780236,1.379086,-0.508192,
        -1.780236,1.382562,0.177141,
        -1.780236,1.388849,2.233847,
        -1.780236,1.389268,-0.886349,
        -1.780236,1.391324,1.109396,
        -1.780236,1.391699,-0.288725,
        -1.780236,1.392869,3.136903,
        -1.780236,1.393612,1.952016,
        -1.780236,1.395545,-2.850807,
        -1.780236,1.401926,1.822930,
        -1.780236,1.402149,2.788898,
        -1.780236,1.409552,0.647980,
        -1.780236,1.410840,0.056219,
        -1.780236,1.411944,2.444377,
        -1.780236,1.413639,1.697982,
        -1.780236,1.426479,1.573427,
        -1.780236,1.426725,2.102242,
        -1.780236,1.427868,0.984637,
        -1.780236,1.429580,-2.107232,
        -1.780236,1.431375,-2.313269,
        -1.780236,1.431907,3.015410,
        -1.780236,1.432325,-1.029054,
        -1.780236,1.433341,0.524268,
        -1.780236,1.435181,-0.739771,
        -1.780236,1.437208,-2.529085,
        -1.780236,1.439041,-0.617966,
        -1.780236,1.441473,1.448544,
        -1.780236,1.444527,2.664744,
        -1.780236,1.450513,-0.400625,
        -1.780236,1.451907,-0.066823,
        -1.780236,1.452754,-2.738631,
        -1.780236,1.454061,-2.950416,
        -1.780236,1.456736,1.323154,
        -1.780236,1.458370,0.400433,
        -1.780236,1.459314,-1.765003,
        -1.780236,1.459902,-1.145835,
        -1.780236,1.464952,2.329245,
        -1.780236,1.466050,0.862104,
        -1.780236,1.471330,-1.886160,
        -1.780236,1.475119,-1.648255,
        -1.780236,1.480878,1.194576,
        -1.780236,1.480942,0.273235,
        -1.780236,1.481086,2.890269,
        -1.780236,1.484118,-1.529156,
        -1.780236,1.484741,-1.279219,
        -1.780236,1.487681,-1.406069,
        -1.780236,1.487886,-3.066037,
        -1.780236,1.490555,2.543284,
        -1.780236,1.492424,-0.190932,
        -1.780236,1.495002,-0.839810,
        -1.780236,1.495297,2.007188,
        -1.780236,1.496414,-2.009267,
        -1.780236,1.498261,-2.425270,
        -1.780236,1.498907,-2.207327,
        -1.780236,1.500204,0.737638,
        -1.780236,1.502469,0.151156,
        -1.780236,1.507448,-0.514782,
        -1.780236,1.509287,2.204647,
        -1.780236,1.510518,1.884920,
        -1.780236,1.511948,-2.632360,
        -1.780236,1.515235,1.068565,
        -1.780236,1.522237,3.102018,
        -1.780236,1.522373,1.764991,
        -1.780236,1.522416,-0.954956,
        -1.780236,1.523102,2.765920,
        -1.780236,1.524811,-2.844718,
        -1.780236,1.527261,-0.304916,
        -1.780236,1.529788,0.613208,
        -1.780236,1.533648,1.644808,
        -1.780236,1.535018,0.028869,
        -1.780236,1.541082,2.424608,
        -1.780236,1.546010,1.523462,
        -1.780236,1.551135,-0.642994,
        -1.780236,1.553895,0.489607,
        -1.780236,1.555041,0.947813,
        -1.780236,1.559671,-1.074934,
        -1.780236,1.560345,2.985899,
        -1.780236,1.560487,-2.323400,
        -1.780236,1.561127,1.401111,
        -1.780236,1.567193,2.643172,
        -1.780236,1.568892,2.098362,
        -1.780236,1.571702,-2.098435,
        -1.780236,1.572701,-2.528730,
        -1.780236,1.573572,1.278593,
        -1.780236,1.574849,-0.091505,
        -1.780236,1.576334,0.365236,
        -1.780236,1.577029,-2.962888,
        -1.780236,1.577688,-1.814941,
        -1.780236,1.579542,-0.755917,
        -1.780236,1.579624,-1.199759,
        -1.780236,1.581515,-0.419414,
        -1.780236,1.584581,-2.738702,
        -1.780236,1.586948,-1.696296,
        -1.780236,1.588901,2.301496,
        -1.780236,1.589936,0.825020,
        -1.780236,1.590018,-1.941652,
        -1.780236,1.596968,-1.328358,
        -1.780236,1.597178,-1.575607,
        -1.780236,1.600470,-1.453196,
        -1.780236,1.601408,0.242409,
        -1.780236,1.602259,1.152656,
        -1.780236,1.603064,2.865774,
        -1.780236,1.612085,1.968802,
        -1.780236,1.612720,-3.080498,
        -1.780236,1.616061,2.521531,
        -1.780236,1.616900,-0.207457,
        -1.780236,1.620988,-0.875920,
        -1.780236,1.622114,0.701509,
        -1.780236,1.622749,0.121426,
        -1.780236,1.622993,-2.221629,
        -1.780236,1.629346,-0.542669,
        -1.780236,1.630007,1.844552,
        -1.780236,1.634747,-2.426198,
        -1.780236,1.642346,1.722617,
        -1.780236,1.643925,1.034629,
        -1.780236,1.644848,2.743774,
        -1.780236,1.645258,2.188109,
        -1.780236,1.645417,-2.634725,
        -1.780236,1.647395,3.085437,
        -1.780236,1.648538,-2.855044,
        -1.780236,1.649932,0.576808,
        -1.780236,1.652224,-0.996105,
        -1.780236,1.652871,1.600681,
        -1.780236,1.657997,-0.325127,
        -1.780236,1.661936,0.001037,
        -1.780236,1.665187,1.477402,
        -1.780236,1.665496,2.397872,
        -1.780236,1.673267,-0.665453,
        -1.780236,1.673557,0.452526,
        -1.780236,1.679277,-1.122421,
        -1.780236,1.679696,0.912482,
        -1.780236,1.682420,1.353922,
        -1.780236,1.685031,-2.009038,
        -1.780236,1.685196,2.967046,
        -1.780236,1.690107,2.620910,
        -1.780236,1.690453,-2.123076,
        -1.780236,1.691693,-1.871572,
        -1.780236,1.691817,2.061688,
        -1.780236,1.692644,1.237670,
        -1.780236,1.696386,0.328123,
        -1.780236,1.697162,-1.249026,
        -1.780236,1.697488,-2.322785,
        -1.780236,1.698726,-2.973576,
        -1.780236,1.699172,-1.749717,
        -1.780236,1.704478,-0.106295,
        -1.780236,1.707244,-0.444770,
        -1.780236,1.707500,-2.531497,
        -1.780236,1.707714,-0.787321,
        -1.780236,1.708430,-1.627552,
        -1.780236,1.710583,-1.376890,
        -1.780236,1.712415,-2.747921,
        -1.780236,1.713125,-1.503073,
        -1.780236,1.713452,0.789419,
        -1.780236,1.720612,2.279888,
        -1.780236,1.724852,0.202831,
        -1.780236,1.724868,2.846246,
        -1.780236,1.728940,1.931535,
        -1.780236,1.732838,1.127584,
        -1.780236,1.735719,-3.092485,
        -1.780236,1.739676,2.496017,
        -1.780236,1.739971,-0.228876,
        -1.780236,1.743328,0.663347,
        -1.780236,1.744927,-0.911892,
        -1.780236,1.745952,0.084733,
        -1.780236,1.748077,1.804354,
        -1.780236,1.752959,-0.567548,
        -1.780236,1.760169,1.678132,
        -1.780236,1.763887,-2.218450,
        -1.780236,1.766737,2.722752,
        -1.780236,1.767999,1.004737,
        -1.780236,1.769149,1.550298,
        -1.780236,1.770119,-2.427175,
        -1.780236,1.770460,2.154208,
        -1.780236,1.770514,3.070982,
        -1.780236,1.770735,0.537245,
        -1.780236,1.772673,-2.864760,
        -1.780236,1.773527,-1.039582,
        -1.780236,1.776419,-2.642421,
        -1.780236,1.780769,1.416330,
        -1.780236,1.786634,-0.346247,
        -1.780236,1.788829,-1.947553,
        -1.780236,1.793737,0.410270,
        -1.780236,1.794553,2.374618,
        -1.780236,1.794699,-0.694061,
        -1.780236,1.797469,-1.169105,
        -1.780236,1.803705,0.881076,
        -1.780236,1.804124,1.264249,
        -1.780236,1.807825,-2.090455,
        -1.780236,1.808822,-1.817375,
        -1.780236,1.809020,2.949741,
        -1.780236,1.812466,2.597595,
        -1.780236,1.813159,2.025395,
        -1.780236,1.813816,-1.298435,
        -1.780236,1.815016,0.278951,
        -1.780236,1.816510,-0.008631,
        -1.780236,1.817772,-1.689625,
        -1.780236,1.821360,-2.984352,
        -1.780236,1.822312,-0.127211,
        -1.780236,1.823386,-1.429807,
        -1.780236,1.823834,-1.560483,
        -1.780236,1.832060,-0.821627,
        -1.780236,1.832690,-0.468440,
        -1.780236,1.833945,-2.320278,
        -1.780236,1.834787,0.751968,
        -1.780236,1.840579,-2.536445,
        -1.780236,1.841033,-2.756932,
        -1.780236,1.846350,1.893983,
        -1.780236,1.846644,1.113766,
        -1.780236,1.846833,2.826543,
        -1.780236,1.847325,2.249051,
        -1.780236,1.854119,0.134161,
        -1.780236,1.858863,-3.106022,
        -1.780236,1.865046,1.763384,
        -1.780236,1.865619,-0.952036,
        -1.780236,1.865912,0.624453,
        -1.780236,1.866694,2.473686,
        -1.780236,1.867524,-0.240083,
        -1.780236,1.875561,1.632831,
        -1.780236,1.876049,-0.594378,
        -1.780236,1.880509,1.500204,
        -1.780236,1.883504,1.360457,
        -1.780236,1.886971,-2.196665,
        -1.780236,1.888979,0.982922,
        -1.780236,1.889815,2.701570,
        -1.780236,1.892144,-2.005340,
        -1.780236,1.892348,0.496522,
        -1.780236,1.893281,-1.083976,
        -1.780236,1.893828,3.053453,
        -1.780236,1.895117,2.121107,
        -1.780236,1.898741,-2.875735,
        -1.780236,1.904429,-2.428029,
        -1.780236,1.908408,-2.648903,
        -1.780236,1.911828,-0.364489,
        -1.780236,1.912224,0.366565,
        -1.780236,1.914854,1.214468,
        -1.780236,1.915467,-1.883143,
        -1.780236,1.915763,-0.724679,
        -1.780236,1.915790,-1.216366,
        -1.780236,1.922256,2.347806,
        -1.780236,1.922616,0.844633,
        -1.780236,1.925185,0.231170,
        -1.780236,1.927474,-1.754616,
        -1.780236,1.928342,-1.351174,
        -1.780236,1.933328,-1.623328,
        -1.780236,1.933878,0.034150,
        -1.780236,1.934585,-1.487979,
        -1.780236,1.934802,2.927634,
        -1.780236,1.934862,1.989776,
        -1.780236,1.936327,-0.099350,
        -1.780236,1.938088,2.578507,
        -1.780236,1.945677,-2.999805,
        -1.780236,1.954572,-0.859208,
        -1.780236,1.957868,-0.490784,
        -1.780236,1.958989,0.712772,
        -1.780236,1.960212,-2.307561,
        -1.780236,1.964774,1.855507,
        -1.780236,1.964971,-2.100533,
        -1.780236,1.969112,1.088273,
        -1.780236,1.971438,-2.765415,
        -1.780236,1.972312,2.805134,
        -1.780236,1.974950,2.220596,
        -1.780236,1.975135,-2.539127,
        -1.780236,1.982107,1.720829,
        -1.780236,1.983195,-3.127984,
        -1.780236,1.983318,-0.247182,
        -1.780236,1.986720,-0.993220,
        -1.780236,1.990381,1.586005,
        -1.780236,1.990470,0.580283,
        -1.780236,1.990523,1.450326,
        -1.780236,1.993909,1.312897,
        -1.780236,1.994694,2.452813,
        -1.780236,1.999226,-0.620717,
        -1.780236,2.005145,0.941012,
        -1.780236,2.007301,0.142337,
        -1.780236,2.014149,0.444138,
        -1.780236,2.014338,-1.127392,
        -1.780236,2.018131,3.021778,
        -1.780236,2.019676,-1.980944,
        -1.780236,2.019912,2.686612,
        -1.780236,2.020545,2.088672,
        -1.780236,2.026074,0.300945,
        -1.780236,2.026196,-2.889057,
        -1.780236,2.031959,-1.268178,
        -1.780236,2.032917,-2.199917,
        -1.780236,2.033572,-1.842884,
        -1.780236,2.034712,-2.419505,
        -1.780236,2.035818,-0.379479,
        -1.780236,2.036004,-0.757062,
        -1.780236,2.040501,-1.411488,
        -1.780236,2.042874,-1.701367,
        -1.780236,2.043060,-2.653746,
        -1.780236,2.043729,-1.557938,
        -1.780236,2.048538,0.806623,
        -1.780236,2.049006,1.199930,
        -1.780236,2.049502,-0.137989,
        -1.780236,2.052181,2.326959,
        -1.780236,2.052794,-0.008526,
        -1.780236,2.057608,1.951726,
        -1.780236,2.065568,2.567730,
        -1.780236,2.070371,2.879808,
        -1.780236,2.071256,-3.018961,
        -1.780236,2.079746,-0.897699,
        -1.780236,2.082784,1.042403,
        -1.780236,2.083752,-0.511086,
        -1.780236,2.084401,1.810502,
        -1.780236,2.085731,0.671081,
        -1.780236,2.097496,1.373172,
        -1.780236,2.099185,1.667006,
        -1.780236,2.100482,-2.301084,
        -1.780236,2.103118,2.194745,
        -1.780236,2.103223,1.521685,
        -1.780236,2.103695,-2.777707,
        -1.780236,2.107523,3.129581,
        -1.780236,2.108277,-2.095907,
        -1.780236,2.108891,-2.532166,
        -1.780236,2.111003,-0.268632,
        -1.780236,2.112276,-1.027246,
        -1.780236,2.114380,0.531457,
        -1.780236,2.119166,0.100552,
        -1.780236,2.119564,2.754818,
        -1.780236,2.122474,-0.648314,
        -1.780236,2.125183,2.445598,
        -1.780236,2.129501,0.238290,
        -1.780236,2.131178,0.386663,
        -1.780236,2.131223,-1.174311,
        -1.780236,2.134269,-1.948042,
        -1.780236,2.134684,0.907065,
        -1.780236,2.137615,2.988806,
        -1.780236,2.145635,-1.327708,
        -1.780236,2.146467,2.055081,
        -1.780236,2.146582,-1.793788,
        -1.780236,2.149210,-1.482321,
        -1.780236,2.152901,-1.636788,
        -1.780236,2.156337,1.151558,
        -1.780236,2.156474,-2.912065,
        -1.780236,2.157194,-0.792563,
        -1.780236,2.166724,-0.401638,
        -1.780236,2.166980,-0.134879,
        -1.780236,2.173644,-2.403429,
        -1.780236,2.175711,-2.655523,
        -1.780236,2.177132,2.600419,
        -1.780236,2.178879,0.767722,
        -1.780236,2.180752,1.908485,
        -1.780236,2.181338,2.313721,
        -1.780236,2.185580,1.287673,
        -1.780236,2.195400,-2.217546,
        -1.780236,2.200418,-3.056442,
        -1.780236,2.201829,-0.001735,
        -1.780236,2.203804,1.756176,
        -1.780236,2.211602,1.443930,
        -1.780236,2.212686,-0.544809,
        -1.780236,2.213937,0.621892,
        -1.780236,2.214075,1.600691,
        -1.780236,2.216256,1.020029,
        -1.780236,2.219060,2.886631,
        -1.780236,2.223375,-0.906185,
        -1.780236,2.226002,-1.060248,
        -1.780236,2.226413,-2.064034,
        -1.780236,2.230554,2.171334,
        -1.780236,2.232413,2.735777,
        -1.780236,2.233709,3.070830,
        -1.780236,2.234383,-2.792065,
        -1.780236,2.237183,0.468799,
        -1.780236,2.237695,-0.271388,
        -1.780236,2.239419,0.151693,
        -1.780236,2.243347,-2.521049,
        -1.780236,2.244035,-1.229456,
        -1.780236,2.244355,0.309899,
        -1.780236,2.245175,2.460236,
        -1.780236,2.245628,-1.899216,
        -1.780236,2.245973,-0.702234,
        -1.780236,2.255601,-1.728131,
        -1.780236,2.256702,-1.401367,
        -1.780236,2.261466,-1.557500,
        -1.780236,2.267390,0.878499,
        -1.780236,2.272099,2.018594,
        -1.780236,2.278618,1.170386,
        -1.780236,2.283304,-2.345584,
        -1.780236,2.285175,-2.943090,
        -1.780236,2.293000,-0.124451,
        -1.780236,2.294058,-0.421070,
        -1.780236,2.303283,1.855938,
        -1.780236,2.304968,2.608978,
        -1.780236,2.305347,2.311346,
        -1.780236,2.307188,-2.656134,
        -1.780236,2.307301,1.336200,
        -1.780236,2.309655,0.726020,
        -1.780236,2.317634,-2.185225,
        -1.780236,2.321510,1.685664,
        -1.780236,2.324376,-3.111006,
        -1.780236,2.324731,1.511186,
        -1.780236,2.329587,0.035498,
        -1.780236,2.330158,2.977264,
        -1.780236,2.335096,-0.588468,
        -1.780236,2.336046,-1.115799,
        -1.780236,2.338060,2.803613,
        -1.780236,2.338083,-0.943400,
        -1.780236,2.340210,0.562145,
        -1.780236,2.341931,1.023926,
        -1.780236,2.343265,-0.779061,
        -1.780236,2.343434,-2.013692,
        -1.780236,2.353029,-1.310529,
        -1.780236,2.354868,0.211543,
        -1.780236,2.355120,0.388593,
        -1.780236,2.355165,-1.833173,
        -1.780236,2.355462,2.151441,
        -1.780236,2.362948,-2.487477,
        -1.780236,2.363101,-2.810456,
        -1.780236,2.363394,-0.272918,
        -1.780236,2.364509,-1.644533,
        -1.780236,2.370465,2.466824,
        -1.780236,2.386110,-1.481351,
        -1.780236,2.394960,0.862133,
        -1.780236,2.395716,1.977184,
        -1.780236,2.396339,1.200900,
        -1.780236,2.407941,-2.316753,
        -1.780236,2.409450,-2.988432,
        -1.780236,2.417202,-0.103363,
        -1.780236,2.417386,-0.443097,
        -1.780236,2.417950,2.666662,
        -1.780236,2.423190,1.789783,
        -1.780236,2.423692,1.392696,
        -1.780236,2.431122,3.086379,
        -1.780236,2.431406,-2.647820,
        -1.780236,2.431584,2.307631,
        -1.780236,2.434090,1.592818,
        -1.780236,2.435993,0.683844,
        -1.780236,2.440898,-2.133300,
        -1.780236,2.442222,-1.195837,
        -1.780236,2.443112,2.883761,
        -1.780236,2.449191,-0.633651,
        -1.780236,2.450821,0.084741,
        -1.780236,2.453499,-1.008861,
        -1.780236,2.460010,-1.938011,
        -1.780236,2.460199,-0.822008,
        -1.780236,2.460902,0.490079,
        -1.780236,2.461774,-1.734913,
        -1.780236,2.463986,1.034407,
        -1.780236,2.467647,0.287667,
        -1.780236,2.479849,2.126925,
        -1.780236,2.485907,-2.835471,
        -1.780236,2.485945,-1.374853,
        -1.780236,2.487747,2.505912,
        -1.780236,2.487957,-0.274279,
        -1.780236,2.489741,-2.468428,
        -1.780236,2.512538,1.239900,
        -1.780236,2.517528,1.923289,
        -1.780236,2.518454,-1.565852,
        -1.780236,2.519348,0.843503,
        -1.780236,2.522766,-3.059231,
        -1.780236,2.528686,2.737228,
        -1.780236,2.534798,-2.266671,
        -1.780236,2.536533,1.466274,
        -1.780236,2.538504,1.699812,
        -1.780236,2.539516,-0.476942,
        -1.780236,2.539532,-0.074548,
        -1.780236,2.543197,2.990252,
        -1.780236,2.554022,2.320619,
        -1.780236,2.556919,-2.655005,
        -1.780236,2.558146,0.627191,
        -1.780236,2.558223,-1.164189,
        -1.780236,2.563452,-2.044028,
        -1.780236,2.568462,0.150655,
        -1.780236,2.569130,-0.700837,
        -1.780236,2.570050,-1.804492,
        -1.780236,2.572710,-0.934665,
        -1.780236,2.575994,0.391718,
        -1.780236,2.584774,1.046703,
        -1.780236,2.601432,2.102535,
        -1.780236,2.603559,2.555114,
        -1.780236,2.604150,-2.885208,
        -1.780236,2.608630,-1.378011,
        -1.780236,2.611256,-0.279110,
        -1.780236,2.617344,-2.439480,
        -1.780236,2.627402,1.293552,
        -1.780236,2.635396,1.848850,
        -1.780236,2.635686,2.834950,
        -1.780236,2.637087,3.132730,
        -1.780236,2.642029,0.813274,
        -1.780236,2.645260,1.570203,
        -1.780236,2.646182,-1.614036,
        -1.780236,2.660113,-2.186513,
        -1.780236,2.660483,-0.034608,
        -1.780236,2.661713,-0.529511,
        -1.780236,2.672720,-1.110422,
        -1.780236,2.676626,2.335353,
        -1.780236,2.676638,0.540364,
        -1.780236,2.676780,-1.901158,
        -1.780236,2.681411,-2.667965,
        -1.780236,2.682152,0.245751,
        -1.780236,2.682859,-0.817071,
        -1.780236,2.705107,1.062142,
        -1.780236,2.717705,2.625995,
        -1.780236,2.721001,2.060097,
        -1.780236,2.722319,-2.954069,
        -1.780236,2.729840,-1.368237,
        -1.780236,2.734349,-0.289298,
        -1.780236,2.738409,2.975671,
        -1.780236,2.740366,1.374341,
        -1.780236,2.744633,-2.391311,
        -1.780236,2.747398,1.727166,
        -1.780236,2.763799,0.760372,
        -1.780236,2.768238,-1.679243,
        -1.780236,2.779245,-2.046268,
        -1.780236,2.780208,-0.626453,
        -1.780236,2.780471,0.027732,
        -1.780236,2.784696,-1.013128,
        -1.780236,2.789817,0.395431,
        -1.780236,2.801154,2.353143,
        -1.780236,2.807091,-2.687477,
        -1.780236,2.825762,1.085035,
        -1.780236,2.830209,2.734943,
        -1.780236,2.833517,-3.085508,
        -1.780236,2.838301,1.980061,
        -1.780236,2.849871,1.512012,
        -1.780236,2.850089,-1.343899,
        -1.780236,2.856905,-0.318441,
        -1.780236,2.870445,-2.291001,
        -1.780236,2.885017,0.655562,
        -1.780236,2.886428,-1.793217,
        -1.780236,2.890337,-0.825633,
        -1.780236,2.898932,0.139973,
        -1.780236,2.931927,-2.728768,
        -1.780236,2.932096,2.377790,
        -1.780236,2.937409,2.947245,
        -1.780236,2.948048,1.133127,
        -1.780236,2.950896,1.797889,
        -1.780236,2.970019,-1.282382,
        -1.780236,2.978333,-0.412090,
        -1.780236,2.996437,-2.089121,
        -1.780236,3.011274,0.434110,
        -1.780236,3.050410,3.138492,
        -1.780236,3.064492,1.647259,
        -1.780236,3.088868,-0.989501,
        -1.675516,0.053884,-2.408394,
        -1.675516,0.077645,1.453433,
        -1.675516,0.088408,-0.311160,
        -1.675516,0.142314,2.590387,
        -1.675516,0.147206,-1.385164,
        -1.675516,0.163506,0.573203,
        -1.675516,0.165485,-2.875396,
        -1.675516,0.182564,-2.125821,
        -1.675516,0.190569,1.929141,
        -1.675516,0.201102,1.249082,
        -1.675516,0.201397,-0.716646,
        -1.675516,0.210568,-0.069559,
        -1.675516,0.251672,2.848900,
        -1.675516,0.266262,-1.678068,
        -1.675516,0.269108,-1.215563,
        -1.675516,0.273777,-2.506756,
        -1.675516,0.279854,0.814581,
        -1.675516,0.281664,-2.969147,
        -1.675516,0.283625,2.405603,
        -1.675516,0.284058,0.357350,
        -1.675516,0.294107,1.593880,
        -1.675516,0.302491,-0.400410,
        -1.675516,0.308059,2.021491,
        -1.675516,0.312219,-2.072057,
        -1.675516,0.321905,-0.829121,
        -1.675516,0.329984,1.201621,
        -1.675516,0.339714,-0.013712,
        -1.675516,0.368784,2.974974,
        -1.675516,0.376428,-1.495532,
        -1.675516,0.378103,2.646436,
        -1.675516,0.385088,0.590355,
        -1.675516,0.386248,-2.690943,
        -1.675516,0.389851,-2.340031,
        -1.675516,0.391502,-1.171374,
        -1.675516,0.398249,-1.812694,
        -1.675516,0.400542,-2.995729,
        -1.675516,0.404273,0.918163,
        -1.675516,0.405617,1.773533,
        -1.675516,0.409009,-0.570243,
        -1.675516,0.410700,0.270433,
        -1.675516,0.415893,1.464890,
        -1.675516,0.420680,2.088164,
        -1.675516,0.424766,-0.258729,
        -1.675516,0.426507,2.367872,
        -1.675516,0.441292,-0.890727,
        -1.675516,0.451839,-2.078193,
        -1.675516,0.459270,1.191562,
        -1.675516,0.470622,0.014002,
        -1.675516,0.481292,-2.529098,
        -1.675516,0.482460,2.849843,
        -1.675516,0.492205,-1.397963,
        -1.675516,0.493314,3.117264,
        -1.675516,0.497870,0.727828,
        -1.675516,0.499783,-1.654379,
        -1.675516,0.502651,0.463333,
        -1.675516,0.505463,2.594557,
        -1.675516,0.506104,-2.874910,
        -1.675516,0.513579,-1.147354,
        -1.675516,0.518569,1.644237,
        -1.675516,0.519001,1.900488,
        -1.675516,0.519277,-0.673809,
        -1.675516,0.524769,-0.421955,
        -1.675516,0.525564,-2.288838,
        -1.675516,0.532541,0.978737,
        -1.675516,0.534610,2.154808,
        -1.675516,0.535020,-1.891846,
        -1.675516,0.540208,1.400497,
        -1.675516,0.542006,0.228337,
        -1.675516,0.550762,-0.180771,
        -1.675516,0.560531,2.378474,
        -1.675516,0.561629,-0.924576,
        -1.675516,0.567465,-2.679392,
        -1.675516,0.588527,1.181755,
        -1.675516,0.591643,-2.098364,
        -1.675516,0.594963,2.997575,
        -1.675516,0.599643,-3.045235,
        -1.675516,0.599962,2.773556,
        -1.675516,0.602620,0.029270,
        -1.675516,0.605711,0.598454,
        -1.675516,0.607708,-1.541026,
        -1.675516,0.609360,-2.464379,
        -1.675516,0.610738,-1.326786,
        -1.675516,0.616968,0.820489,
        -1.675516,0.626379,1.773610,
        -1.675516,0.628626,-1.751180,
        -1.675516,0.630671,0.391949,
        -1.675516,0.632622,-0.737642,
        -1.675516,0.632708,2.566722,
        -1.675516,0.632797,1.989858,
        -1.675516,0.634455,-0.534790,
        -1.675516,0.636205,1.561170,
        -1.675516,0.637681,-1.123423,
        -1.675516,0.641441,-2.842109,
        -1.675516,0.644807,-0.327139,
        -1.675516,0.652440,2.197129,
        -1.675516,0.660897,-2.278593,
        -1.675516,0.664008,1.360091,
        -1.675516,0.670850,-1.945257,
        -1.675516,0.671396,1.004859,
        -1.675516,0.673985,0.203855,
        -1.675516,0.680326,-0.133370,
        -1.675516,0.685083,-0.942363,
        -1.675516,0.688204,-2.632204,
        -1.675516,0.689777,2.385011,
        -1.675516,0.700065,3.115430,
        -1.675516,0.704080,2.911636,
        -1.675516,0.712883,1.174216,
        -1.675516,0.715701,0.699906,
        -1.675516,0.718851,-1.442215,
        -1.675516,0.721019,2.721536,
        -1.675516,0.725300,-2.975134,
        -1.675516,0.727136,-1.633288,
        -1.675516,0.727399,-2.120567,
        -1.675516,0.730751,0.522488,
        -1.675516,0.732892,-1.254330,
        -1.675516,0.734333,-2.446184,
        -1.675516,0.735125,0.037270,
        -1.675516,0.737893,1.866463,
        -1.675516,0.740650,1.681075,
        -1.675516,0.745971,-0.781286,
        -1.675516,0.749432,-0.442559,
        -1.675516,0.750102,2.050636,
        -1.675516,0.751046,0.863821,
        -1.675516,0.757018,-0.608833,
        -1.675516,0.757046,-1.816019,
        -1.675516,0.757389,1.501184,
        -1.675516,0.759800,2.546181,
        -1.675516,0.759859,0.351923,
        -1.675516,0.761141,-2.781102,
        -1.675516,0.769035,-1.078574,
        -1.675516,0.770166,-0.266604,
        -1.675516,0.772218,2.227380,
        -1.675516,0.788382,1.331051,
        -1.675516,0.790269,-2.282409,
        -1.675516,0.803410,-1.984567,
        -1.675516,0.803813,0.191520,
        -1.675516,0.806832,3.022155,
        -1.675516,0.807476,-2.597675,
        -1.675516,0.812512,-0.107335,
        -1.675516,0.814161,1.010949,
        -1.675516,0.817732,-3.084486,
        -1.675516,0.817846,2.390688,
        -1.675516,0.819323,2.845960,
        -1.675516,0.819430,-0.915017,
        -1.675516,0.829198,-1.532491,
        -1.675516,0.829531,0.665695,
        -1.675516,0.829816,1.167147,
        -1.675516,0.833613,-1.357492,
        -1.675516,0.842945,-2.907318,
        -1.675516,0.844883,2.679708,
        -1.675516,0.848189,-1.705003,
        -1.675516,0.849888,1.767901,
        -1.675516,0.851126,0.493370,
        -1.675516,0.852911,1.933394,
        -1.675516,0.857516,-2.432863,
        -1.675516,0.858261,-1.191729,
        -1.675516,0.858279,-2.138895,
        -1.675516,0.859723,-0.734928,
        -1.675516,0.859754,1.603292,
        -1.675516,0.861655,-0.394821,
        -1.675516,0.864300,0.043549,
        -1.675516,0.866940,0.851229,
        -1.675516,0.869748,2.095969,
        -1.675516,0.877352,-0.574668,
        -1.675516,0.879254,-2.733338,
        -1.675516,0.882554,1.443757,
        -1.675516,0.884060,0.334199,
        -1.675516,0.885626,-1.866061,
        -1.675516,0.889834,2.529957,
        -1.675516,0.893140,2.251698,
        -1.675516,0.897463,-0.240822,
        -1.675516,0.898758,-1.031948,
        -1.675516,0.913918,1.287618,
        -1.675516,0.916555,-2.287173,
        -1.675516,0.918305,3.107845,
        -1.675516,0.919522,2.948277,
        -1.675516,0.927138,-2.569686,
        -1.675516,0.929031,0.185383,
        -1.675516,0.933766,-3.015917,
        -1.675516,0.934873,-2.010822,
        -1.675516,0.935636,-1.445855,
        -1.675516,0.936702,-0.866461,
        -1.675516,0.937395,0.729108,
        -1.675516,0.937910,2.787575,
        -1.675516,0.939144,-0.092962,
        -1.675516,0.941718,1.125085,
        -1.675516,0.942873,-1.607680,
        -1.675516,0.943701,0.980328,
        -1.675516,0.947774,2.395256,
        -1.675516,0.952433,0.584347,
        -1.675516,0.954456,-1.287293,
        -1.675516,0.958731,-2.852552,
        -1.675516,0.961838,1.839284,
        -1.675516,0.961898,-0.469425,
        -1.675516,0.965426,1.686615,
        -1.675516,0.970108,1.990771,
        -1.675516,0.972820,-1.762780,
        -1.675516,0.973500,-0.711881,
        -1.675516,0.975253,2.641082,
        -1.675516,0.977824,0.441928,
        -1.675516,0.981485,-2.424766,
        -1.675516,0.981877,1.537100,
        -1.675516,0.982609,-1.136920,
        -1.675516,0.984265,-2.156073,
        -1.675516,0.989150,0.045493,
        -1.675516,0.993052,2.137478,
        -1.675516,0.997766,-2.693871,
        -1.675516,0.999847,-0.334738,
        -1.675516,1.001920,0.848054,
        -1.675516,1.006146,1.389734,
        -1.675516,1.012180,0.304066,
        -1.675516,1.016938,2.270674,
        -1.675516,1.017371,-0.979790,
        -1.675516,1.019495,-1.894487,
        -1.675516,1.023796,-0.583691,
        -1.675516,1.026525,2.506945,
        -1.675516,1.027004,-0.203653,
        -1.675516,1.028027,3.036573,
        -1.675516,1.031435,1.241021,
        -1.675516,1.032316,2.884221,
        -1.675516,1.035863,-3.110475,
        -1.675516,1.042085,-2.292950,
        -1.675516,1.043272,-1.522986,
        -1.675516,1.046265,-2.958847,
        -1.675516,1.049272,-1.376344,
        -1.675516,1.050039,-2.552086,
        -1.675516,1.053682,0.689863,
        -1.675516,1.055458,0.170394,
        -1.675516,1.056622,-0.831866,
        -1.675516,1.057515,1.097998,
        -1.675516,1.058706,-2.036438,
        -1.675516,1.061389,-1.672248,
        -1.675516,1.063758,2.743093,
        -1.675516,1.070157,-0.073422,
        -1.675516,1.072154,1.761506,
        -1.675516,1.074029,1.909193,
        -1.675516,1.074204,0.545276,
        -1.675516,1.074714,-2.805147,
        -1.675516,1.076262,0.966104,
        -1.675516,1.077642,-1.227073,
        -1.675516,1.084151,1.618525,
        -1.675516,1.085085,2.377673,
        -1.675516,1.086822,-0.441112,
        -1.675516,1.087142,2.058815,
        -1.675516,1.098383,-0.696430,
        -1.675516,1.099138,-1.084033,
        -1.675516,1.103208,1.477773,
        -1.675516,1.103783,0.408254,
        -1.675516,1.106275,-2.420800,
        -1.675516,1.108079,2.612943,
        -1.675516,1.109345,-2.172718,
        -1.675516,1.115957,-1.787597,
        -1.675516,1.118930,0.045983,
        -1.675516,1.119777,0.834702,
        -1.675516,1.121943,-2.669042,
        -1.675516,1.124930,2.219917,
        -1.675516,1.125217,1.336294,
        -1.675516,1.130958,2.972931,
        -1.675516,1.131400,-0.306629,
        -1.675516,1.137135,-0.936514,
        -1.675516,1.139125,-1.924528,
        -1.675516,1.139877,0.276382,
        -1.675516,1.142003,-3.058022,
        -1.675516,1.147363,-0.558448,
        -1.675516,1.150348,-0.176418,
        -1.675516,1.151198,1.197016,
        -1.675516,1.153930,-1.597104,
        -1.675516,1.155389,2.837148,
        -1.675516,1.155458,-1.460623,
        -1.675516,1.157331,-2.906362,
        -1.675516,1.158594,2.487618,
        -1.675516,1.162103,-1.322553,
        -1.675516,1.165407,3.092861,
        -1.675516,1.168025,0.717075,
        -1.675516,1.168406,-2.301519,
        -1.675516,1.174782,-2.538755,
        -1.675516,1.177998,1.061256,
        -1.675516,1.178898,1.827305,
        -1.675516,1.181331,-2.060330,
        -1.675516,1.182540,-0.798952,
        -1.675516,1.184569,0.147517,
        -1.675516,1.184734,1.968919,
        -1.675516,1.184804,0.593386,
        -1.675516,1.188278,1.690808,
        -1.675516,1.194360,2.110440,
        -1.675516,1.194776,2.710455,
        -1.675516,1.197745,-1.173705,
        -1.675516,1.199898,-2.777257,
        -1.675516,1.201164,2.355931,
        -1.675516,1.203760,1.556404,
        -1.675516,1.205845,-0.415993,
        -1.675516,1.206585,0.933648,
        -1.675516,1.206950,-0.037248,
        -1.675516,1.209341,0.471162,
        -1.675516,1.213483,-1.039417,
        -1.675516,1.218146,-0.672643,
        -1.675516,1.221608,1.421669,
        -1.675516,1.226043,-1.693533,
        -1.675516,1.232870,-1.822773,
        -1.675516,1.234375,-2.421194,
        -1.675516,1.234531,-2.190148,
        -1.675516,1.239414,0.348959,
        -1.675516,1.240183,2.587255,
        -1.675516,1.242312,2.941234,
        -1.675516,1.244075,1.286644,
        -1.675516,1.245961,-2.997461,
        -1.675516,1.246495,-2.646872,
        -1.675516,1.247736,2.242810,
        -1.675516,1.253604,0.812222,
        -1.675516,1.257816,-0.894317,
        -1.675516,1.258247,-1.953634,
        -1.675516,1.258993,-0.284099,
        -1.675516,1.262380,-0.157609,
        -1.675516,1.262791,-1.549229,
        -1.675516,1.263772,-0.533073,
        -1.675516,1.266731,-1.412461,
        -1.675516,1.268591,-3.122146,
        -1.675516,1.269614,-1.275246,
        -1.675516,1.270745,1.152533,
        -1.675516,1.273606,0.224545,
        -1.675516,1.279716,2.812101,
        -1.675516,1.283786,2.462124,
        -1.675516,1.284798,0.078209,
        -1.675516,1.285402,-2.882049,
        -1.675516,1.286936,1.886053,
        -1.675516,1.289408,0.683122,
        -1.675516,1.294736,1.756184,
        -1.675516,1.296978,-2.311511,
        -1.675516,1.298546,2.020601,
        -1.675516,1.300924,1.022478,
        -1.675516,1.303305,3.043152,
        -1.675516,1.304210,-2.084225,
        -1.675516,1.305241,-2.530249,
        -1.675516,1.307513,1.628003,
        -1.675516,1.310722,0.558267,
        -1.675516,1.314142,2.139329,
        -1.675516,1.316087,-0.777971,
        -1.675516,1.317907,-1.121023,
        -1.675516,1.321019,-2.749776,
        -1.675516,1.321609,2.687717,
        -1.675516,1.322568,1.499415,
        -1.675516,1.322972,-0.395699,
        -1.675516,1.325286,-0.994192,
        -1.675516,1.329146,-0.648513,
        -1.675516,1.329795,-0.042570,
        -1.675516,1.331651,2.344497,
        -1.675516,1.336522,0.435098,
        -1.675516,1.339410,0.900087,
        -1.675516,1.339550,1.370353,
        -1.675516,1.340441,-1.731580,
        -1.675516,1.351855,-1.855234,
        -1.675516,1.360161,-3.018404,
        -1.675516,1.361031,-1.613971,
        -1.675516,1.361196,2.914780,
        -1.675516,1.361998,-2.207254,
        -1.675516,1.362302,1.239384,
        -1.675516,1.365228,2.563808,
        -1.675516,1.365452,0.310522,
        -1.675516,1.367028,-2.420041,
        -1.675516,1.373422,-1.484281,
        -1.675516,1.374245,-1.226417,
        -1.675516,1.375020,-0.170393,
        -1.675516,1.376560,-1.355481,
        -1.675516,1.377070,-1.980984,
        -1.675516,1.377578,-2.636339,
        -1.675516,1.378096,0.774626,
        -1.675516,1.379086,-0.508192,
        -1.675516,1.382562,0.177141,
        -1.675516,1.388849,2.233847,
        -1.675516,1.389268,-0.886349,
        -1.675516,1.391324,1.109396,
        -1.675516,1.391699,-0.288725,
        -1.675516,1.392869,3.136903,
        -1.675516,1.393612,1.952016,
        -1.675516,1.395545,-2.850807,
        -1.675516,1.401926,1.822930,
        -1.675516,1.402149,2.788898,
        -1.675516,1.409552,0.647980,
        -1.675516,1.410840,0.056219,
        -1.675516,1.411944,2.444377,
        -1.675516,1.413639,1.697982,
        -1.675516,1.426479,1.573427,
        -1.675516,1.426725,2.102242,
        -1.675516,1.427868,0.984637,
        -1.675516,1.429580,-2.107232,
        -1.675516,1.431375,-2.313269,
        -1.675516,1.431907,3.015410,
        -1.675516,1.432325,-1.029054,
        -1.675516,1.433341,0.524268,
        -1.675516,1.435181,-0.739771,
        -1.675516,1.437208,-2.529085,
        -1.675516,1.439041,-0.617966,
        -1.675516,1.441473,1.448544,
        -1.675516,1.444527,2.664744,
        -1.675516,1.450513,-0.400625,
        -1.675516,1.451907,-0.066823,
        -1.675516,1.452754,-2.738631,
        -1.675516,1.454061,-2.950416,
        -1.675516,1.456736,1.323154,
        -1.675516,1.458370,0.400433,
        -1.675516,1.459314,-1.765003,
        -1.675516,1.459902,-1.145835,
        -1.675516,1.464952,2.329245,
        -1.675516,1.466050,0.862104,
        -1.675516,1.471330,-1.886160,
        -1.675516,1.475119,-1.648255,
        -1.675516,1.480878,1.194576,
        -1.675516,1.480942,0.273235,
        -1.675516,1.481086,2.890269,
        -1.675516,1.484118,-1.529156,
        -1.675516,1.484741,-1.279219,
        -1.675516,1.487681,-1.406069,
        -1.675516,1.487886,-3.066037,
        -1.675516,1.490555,2.543284,
        -1.675516,1.492424,-0.190932,
        -1.675516,1.495002,-0.839810,
        -1.675516,1.495297,2.007188,
        -1.675516,1.496414,-2.009267,
        -1.675516,1.498261,-2.425270,
        -1.675516,1.498907,-2.207327,
        -1.675516,1.500204,0.737638,
        -1.675516,1.502469,0.151156,
        -1.675516,1.507448,-0.514782,
        -1.675516,1.509287,2.204647,
        -1.675516,1.510518,1.884920,
        -1.675516,1.511948,-2.632360,
        -1.675516,1.515235,1.068565,
        -1.675516,1.522237,3.102018,
        -1.675516,1.522373,1.764991,
        -1.675516,1.522416,-0.954956,
        -1.675516,1.523102,2.765920,
        -1.675516,1.524811,-2.844718,
        -1.675516,1.527261,-0.304916,
        -1.675516,1.529788,0.613208,
        -1.675516,1.533648,1.644808,
        -1.675516,1.535018,0.028869,
        -1.675516,1.541082,2.424608,
        -1.675516,1.546010,1.523462,
        -1.675516,1.551135,-0.642994,
        -1.675516,1.553895,0.489607,
        -1.675516,1.555041,0.947813,
        -1.675516,1.559671,-1.074934,
        -1.675516,1.560345,2.985899,
        -1.675516,1.560487,-2.323400,
        -1.675516,1.561127,1.401111,
        -1.675516,1.567193,2.643172,
        -1.675516,1.568892,2.098362,
        -1.675516,1.571702,-2.098435,
        -1.675516,1.572701,-2.528730,
        -1.675516,1.573572,1.278593,
        -1.675516,1.574849,-0.091505,
        -1.675516,1.576334,0.365236,
        -1.675516,1.577029,-2.962888,
        -1.675516,1.577688,-1.814941,
        -1.675516,1.579542,-0.755917,
        -1.675516,1.579624,-1.199759,
        -1.675516,1.581515,-0.419414,
        -1.675516,1.584581,-2.738702,
        -1.675516,1.586948,-1.696296,
        -1.675516,1.588901,2.301496,
        -1.675516,1.589936,0.825020,
        -1.675516,1.590018,-1.941652,
        -1.675516,1.596968,-1.328358,
        -1.675516,1.597178,-1.575607,
        -1.675516,1.600470,-1.453196,
        -1.675516,1.601408,0.242409,
        -1.675516,1.602259,1.152656,
        -1.675516,1.603064,2.865774,
        -1.675516,1.612085,1.968802,
        -1.675516,1.612720,-3.080498,
        -1.675516,1.616061,2.521531,
        -1.675516,1.616900,-0.207457,
        -1.675516,1.620988,-0.875920,
        -1.675516,1.622114,0.701509,
        -1.675516,1.622749,0.121426,
        -1.675516,1.622993,-2.221629,
        -1.675516,1.629346,-0.542669,
        -1.675516,1.630007,1.844552,
        -1.675516,1.634747,-2.426198,
        -1.675516,1.642346,1.722617,
        -1.675516,1.643925,1.034629,
        -1.675516,1.644848,2.743774,
        -1.675516,1.645258,2.188109,
        -1.675516,1.645417,-2.634725,
        -1.675516,1.647395,3.085437,
        -1.675516,1.648538,-2.855044,
        -1.675516,1.649932,0.576808,
        -1.675516,1.652224,-0.996105,
        -1.675516,1.652871,1.600681,
        -1.675516,1.657997,-0.325127,
        -1.675516,1.661936,0.001037,
        -1.675516,1.665187,1.477402,
        -1.675516,1.665496,2.397872,
        -1.675516,1.673267,-0.665453,
        -1.675516,1.673557,0.452526,
        -1.675516,1.679277,-1.122421,
        -1.675516,1.679696,0.912482,
        -1.675516,1.682420,1.353922,
        -1.675516,1.685031,-2.009038,
        -1.675516,1.685196,2.967046,
        -1.675516,1.690107,2.620910,
        -1.675516,1.690453,-2.123076,
        -1.675516,1.691693,-1.871572,
        -1.675516,1.691817,2.061688,
        -1.675516,1.692644,1.237670,
        -1.675516,1.696386,0.328123,
        -1.675516,1.697162,-1.249026,
        -1.675516,1.697488,-2.322785,
        -1.675516,1.698726,-2.973576,
        -1.675516,1.699172,-1.749717,
        -1.675516,1.704478,-0.106295,
        -1.675516,1.707244,-0.444770,
        -1.675516,1.707500,-2.531497,
        -1.675516,1.707714,-0.787321,
        -1.675516,1.708430,-1.627552,
        -1.675516,1.710583,-1.376890,
        -1.675516,1.712415,-2.747921,
        -1.675516,1.713125,-1.503073,
        -1.675516,1.713452,0.789419,
        -1.675516,1.720612,2.279888,
        -1.675516,1.724852,0.202831,
        -1.675516,1.724868,2.846246,
        -1.675516,1.728940,1.931535,
        -1.675516,1.732838,1.127584,
        -1.675516,1.735719,-3.092485,
        -1.675516,1.739676,2.496017,
        -1.675516,1.739971,-0.228876,
        -1.675516,1.743328,0.663347,
        -1.675516,1.744927,-0.911892,
        -1.675516,1.745952,0.084733,
        -1.675516,1.748077,1.804354,
        -1.675516,1.752959,-0.567548,
        -1.675516,1.760169,1.678132,
        -1.675516,1.763887,-2.218450,
        -1.675516,1.766737,2.722752,
        -1.675516,1.767999,1.004737,
        -1.675516,1.769149,1.550298,
        -1.675516,1.770119,-2.427175,
        -1.675516,1.770460,2.154208,
        -1.675516,1.770514,3.070982,
        -1.675516,1.770735,0.537245,
        -1.675516,1.772673,-2.864760,
        -1.675516,1.773527,-1.039582,
        -1.675516,1.776419,-2.642421,
        -1.675516,1.780769,1.416330,
        -1.675516,1.786634,-0.346247,
        -1.675516,1.788829,-1.947553,
        -1.675516,1.793737,0.410270,
        -1.675516,1.794553,2.374618,
        -1.675516,1.794699,-0.694061,
        -1.675516,1.797469,-1.169105,
        -1.675516,1.803705,0.881076,
        -1.675516,1.804124,1.264249,
        -1.675516,1.807825,-2.090455,
        -1.675516,1.808822,-1.817375,
        -1.675516,1.809020,2.949741,
        -1.675516,1.812466,2.597595,
        -1.675516,1.813159,2.025395,
        -1.675516,1.813816,-1.298435,
        -1.675516,1.815016,0.278951,
        -1.675516,1.816510,-0.008631,
        -1.675516,1.817772,-1.689625,
        -1.675516,1.821360,-2.984352,
        -1.675516,1.822312,-0.127211,
        -1.675516,1.823386,-1.429807,
        -1.675516,1.823834,-1.560483,
        -1.675516,1.832060,-0.821627,
        -1.675516,1.832690,-0.468440,
        -1.675516,1.833945,-2.320278,
        -1.675516,1.834787,0.751968,
        -1.675516,1.840579,-2.536445,
        -1.675516,1.841033,-2.756932,
        -1.675516,1.846350,1.893983,
        -1.675516,1.846644,1.113766,
        -1.675516,1.846833,2.826543,
        -1.675516,1.847325,2.249051,
        -1.675516,1.854119,0.134161,
        -1.675516,1.858863,-3.106022,
        -1.675516,1.865046,1.763384,
        -1.675516,1.865619,-0.952036,
        -1.675516,1.865912,0.624453,
        -1.675516,1.866694,2.473686,
        -1.675516,1.867524,-0.240083,
        -1.675516,1.875561,1.632831,
        -1.675516,1.876049,-0.594378,
        -1.675516,1.880509,1.500204,
        -1.675516,1.883504,1.360457,
        -1.675516,1.886971,-2.196665,
        -1.675516,1.888979,0.982922,
        -1.675516,1.889815,2.701570,
        -1.675516,1.892144,-2.005340,
        -1.675516,1.892348,0.496522,
        -1.675516,1.893281,-1.083976,
        -1.675516,1.893828,3.053453,
        -1.675516,1.895117,2.121107,
        -1.675516,1.898741,-2.875735,
        -1.675516,1.904429,-2.428029,
        -1.675516,1.908408,-2.648903,
        -1.675516,1.911828,-0.364489,
        -1.675516,1.912224,0.366565,
        -1.675516,1.914854,1.214468,
        -1.675516,1.915467,-1.883143,
        -1.675516,1.915763,-0.724679,
        -1.675516,1.915790,-1.216366,
        -1.675516,1.922256,2.347806,
        -1.675516,1.922616,0.844633,
        -1.675516,1.925185,0.231170,
        -1.675516,1.927474,-1.754616,
        -1.675516,1.928342,-1.351174,
        -1.675516,1.933328,-1.623328,
        -1.675516,1.933878,0.034150,
        -1.675516,1.934585,-1.487979,
        -1.675516,1.934802,2.927634,
        -1.675516,1.934862,1.989776,
        -1.675516,1.936327,-0.099350,
        -1.675516,1.938088,2.578507,
        -1.675516,1.945677,-2.999805,
        -1.675516,1.954572,-0.859208,
        -1.675516,1.957868,-0.490784,
        -1.675516,1.958989,0.712772,
        -1.675516,1.960212,-2.307561,
        -1.675516,1.964774,1.855507,
        -1.675516,1.964971,-2.100533,
        -1.675516,1.969112,1.088273,
        -1.675516,1.971438,-2.765415,
        -1.675516,1.972312,2.805134,
        -1.675516,1.974950,2.220596,
        -1.675516,1.975135,-2.539127,
        -1.675516,1.982107,1.720829,
        -1.675516,1.983195,-3.127984,
        -1.675516,1.983318,-0.247182,
        -1.675516,1.986720,-0.993220,
        -1.675516,1.990381,1.586005,
        -1.675516,1.990470,0.580283,
        -1.675516,1.990523,1.450326,
        -1.675516,1.993909,1.312897,
        -1.675516,1.994694,2.452813,
        -1.675516,1.999226,-0.620717,
        -1.675516,2.005145,0.941012,
        -1.675516,2.007301,0.142337,
        -1.675516,2.014149,0.444138,
        -1.675516,2.014338,-1.127392,
        -1.675516,2.018131,3.021778,
        -1.675516,2.019676,-1.980944,
        -1.675516,2.019912,2.686612,
        -1.675516,2.020545,2.088672,
        -1.675516,2.026074,0.300945,
        -1.675516,2.026196,-2.889057,
        -1.675516,2.031959,-1.268178,
        -1.675516,2.032917,-2.199917,
        -1.675516,2.033572,-1.842884,
        -1.675516,2.034712,-2.419505,
        -1.675516,2.035818,-0.379479,
        -1.675516,2.036004,-0.757062,
        -1.675516,2.040501,-1.411488,
        -1.675516,2.042874,-1.701367,
        -1.675516,2.043060,-2.653746,
        -1.675516,2.043729,-1.557938,
        -1.675516,2.048538,0.806623,
        -1.675516,2.049006,1.199930,
        -1.675516,2.049502,-0.137989,
        -1.675516,2.052181,2.326959,
        -1.675516,2.052794,-0.008526,
        -1.675516,2.057608,1.951726,
        -1.675516,2.065568,2.567730,
        -1.675516,2.070371,2.879808,
        -1.675516,2.071256,-3.018961,
        -1.675516,2.079746,-0.897699,
        -1.675516,2.082784,1.042403,
        -1.675516,2.083752,-0.511086,
        -1.675516,2.084401,1.810502,
        -1.675516,2.085731,0.671081,
        -1.675516,2.097496,1.373172,
        -1.675516,2.099185,1.667006,
        -1.675516,2.100482,-2.301084,
        -1.675516,2.103118,2.194745,
        -1.675516,2.103223,1.521685,
        -1.675516,2.103695,-2.777707,
        -1.675516,2.107523,3.129581,
        -1.675516,2.108277,-2.095907,
        -1.675516,2.108891,-2.532166,
        -1.675516,2.111003,-0.268632,
        -1.675516,2.112276,-1.027246,
        -1.675516,2.114380,0.531457,
        -1.675516,2.119166,0.100552,
        -1.675516,2.119564,2.754818,
        -1.675516,2.122474,-0.648314,
        -1.675516,2.125183,2.445598,
        -1.675516,2.129501,0.238290,
        -1.675516,2.131178,0.386663,
        -1.675516,2.131223,-1.174311,
        -1.675516,2.134269,-1.948042,
        -1.675516,2.134684,0.907065,
        -1.675516,2.137615,2.988806,
        -1.675516,2.145635,-1.327708,
        -1.675516,2.146467,2.055081,
        -1.675516,2.146582,-1.793788,
        -1.675516,2.149210,-1.482321,
        -1.675516,2.152901,-1.636788,
        -1.675516,2.156337,1.151558,
        -1.675516,2.156474,-2.912065,
        -1.675516,2.157194,-0.792563,
        -1.675516,2.166724,-0.401638,
        -1.675516,2.166980,-0.134879,
        -1.675516,2.173644,-2.403429,
        -1.675516,2.175711,-2.655523,
        -1.675516,2.177132,2.600419,
        -1.675516,2.178879,0.767722,
        -1.675516,2.180752,1.908485,
        -1.675516,2.181338,2.313721,
        -1.675516,2.185580,1.287673,
        -1.675516,2.195400,-2.217546,
        -1.675516,2.200418,-3.056442,
        -1.675516,2.201829,-0.001735,
        -1.675516,2.203804,1.756176,
        -1.675516,2.211602,1.443930,
        -1.675516,2.212686,-0.544809,
        -1.675516,2.213937,0.621892,
        -1.675516,2.214075,1.600691,
        -1.675516,2.216256,1.020029,
        -1.675516,2.219060,2.886631,
        -1.675516,2.223375,-0.906185,
        -1.675516,2.226002,-1.060248,
        -1.675516,2.226413,-2.064034,
        -1.675516,2.230554,2.171334,
        -1.675516,2.232413,2.735777,
        -1.675516,2.233709,3.070830,
        -1.675516,2.234383,-2.792065,
        -1.675516,2.237183,0.468799,
        -1.675516,2.237695,-0.271388,
        -1.675516,2.239419,0.151693,
        -1.675516,2.243347,-2.521049,
        -1.675516,2.244035,-1.229456,
        -1.675516,2.244355,0.309899,
        -1.675516,2.245175,2.460236,
        -1.675516,2.245628,-1.899216,
        -1.675516,2.245973,-0.702234,
        -1.675516,2.255601,-1.728131,
        -1.675516,2.256702,-1.401367,
        -1.675516,2.261466,-1.557500,
        -1.675516,2.267390,0.878499,
        -1.675516,2.272099,2.018594,
        -1.675516,2.278618,1.170386,
        -1.675516,2.283304,-2.345584,
        -1.675516,2.285175,-2.943090,
        -1.675516,2.293000,-0.124451,
        -1.675516,2.294058,-0.421070,
        -1.675516,2.303283,1.855938,
        -1.675516,2.304968,2.608978,
        -1.675516,2.305347,2.311346,
        -1.675516,2.307188,-2.656134,
        -1.675516,2.307301,1.336200,
        -1.675516,2.309655,0.726020,
        -1.675516,2.317634,-2.185225,
        -1.675516,2.321510,1.685664,
        -1.675516,2.324376,-3.111006,
        -1.675516,2.324731,1.511186,
        -1.675516,2.329587,0.035498,
        -1.675516,2.330158,2.977264,
        -1.675516,2.335096,-0.588468,
        -1.675516,2.336046,-1.115799,
        -1.675516,2.338060,2.803613,
        -1.675516,2.338083,-0.943400,
        -1.675516,2.340210,0.562145,
        -1.675516,2.341931,1.023926,
        -1.675516,2.343265,-0.779061,
        -1.675516,2.343434,-2.013692,
        -1.675516,2.353029,-1.310529,
        -1.675516,2.354868,0.211543,
        -1.675516,2.355120,0.388593,
        -1.675516,2.355165,-1.833173,
        -1.675516,2.355462,2.151441,
        -1.675516,2.362948,-2.487477,
        -1.675516,2.363101,-2.810456,
        -1.675516,2.363394,-0.272918,
        -1.675516,2.364509,-1.644533,
        -1.675516,2.370465,2.466824,
        -1.675516,2.386110,-1.481351,
        -1.675516,2.394960,0.862133,
        -1.675516,2.395716,1.977184,
        -1.675516,2.396339,1.200900,
        -1.675516,2.407941,-2.316753,
        -1.675516,2.409450,-2.988432,
        -1.675516,2.417202,-0.103363,
        -1.675516,2.417386,-0.443097,
        -1.675516,2.417950,2.666662,
        -1.675516,2.423190,1.789783,
        -1.675516,2.423692,1.392696,
        -1.675516,2.431122,3.086379,
        -1.675516,2.431406,-2.647820,
        -1.675516,2.431584,2.307631,
        -1.675516,2.434090,1.592818,
        -1.675516,2.435993,0.683844,
        -1.675516,2.440898,-2.133300,
        -1.675516,2.442222,-1.195837,
        -1.675516,2.443112,2.883761,
        -1.675516,2.449191,-0.633651,
        -1.675516,2.450821,0.084741,
        -1.675516,2.453499,-1.008861,
        -1.675516,2.460010,-1.938011,
        -1.675516,2.460199,-0.822008,
        -1.675516,2.460902,0.490079,
        -1.675516,2.461774,-1.734913,
        -1.675516,2.463986,1.034407,
        -1.675516,2.467647,0.287667,
        -1.675516,2.479849,2.126925,
        -1.675516,2.485907,-2.835471,
        -1.675516,2.485945,-1.374853,
        -1.675516,2.487747,2.505912,
        -1.675516,2.487957,-0.274279,
        -1.675516,2.489741,-2.468428,
        -1.675516,2.512538,1.239900,
        -1.675516,2.517528,1.923289,
        -1.675516,2.518454,-1.565852,
        -1.675516,2.519348,0.843503,
        -1.675516,2.522766,-3.059231,
        -1.675516,2.528686,2.737228,
        -1.675516,2.534798,-2.266671,
        -1.675516,2.536533,1.466274,
        -1.675516,2.538504,1.699812,
        -1.675516,2.539516,-0.476942,
        -1.675516,2.539532,-0.074548,
        -1.675516,2.543197,2.990252,
        -1.675516,2.554022,2.320619,
        -1.675516,2.556919,-2.655005,
        -1.675516,2.558146,0.627191,
        -1.675516,2.558223,-1.164189,
        -1.675516,2.563452,-2.044028,
        -1.675516,2.568462,0.150655,
        -1.675516,2.569130,-0.700837,
        -1.675516,2.570050,-1.804492,
        -1.675516,2.572710,-0.934665,
        -1.675516,2.575994,0.391718,
        -1.675516,2.584774,1.046703,
        -1.675516,2.601432,2.102535,
        -1.675516,2.603559,2.555114,
        -1.675516,2.604150,-2.885208,
        -1.675516,2.608630,-1.378011,
        -1.675516,2.611256,-0.279110,
        -1.675516,2.617344,-2.439480,
        -1.675516,2.627402,1.293552,
        -1.675516,2.635396,1.848850,
        -1.675516,2.635686,2.834950,
        -1.675516,2.637087,3.132730,
        -1.675516,2.642029,0.813274,
        -1.675516,2.645260,1.570203,
        -1.675516,2.646182,-1.614036,
        -1.675516,2.660113,-2.186513,
        -1.675516,2.660483,-0.034608,
        -1.675516,2.661713,-0.529511,
        -1.675516,2.672720,-1.110422,
        -1.675516,2.676626,2.335353,
        -1.675516,2.676638,0.540364,
        -1.675516,2.676780,-1.901158,
        -1.675516,2.681411,-2.667965,
        -1.675516,2.682152,0.245751,
        -1.675516,2.682859,-0.817071,
        -1.675516,2.705107,1.062142,
        -1.675516,2.717705,2.625995,
        -1.675516,2.721001,2.060097,
        -1.675516,2.722319,-2.954069,
        -1.675516,2.729840,-1.368237,
        -1.675516,2.734349,-0.289298,
        -1.675516,2.738409,2.975671,
        -1.675516,2.740366,1.374341,
        -1.675516,2.744633,-2.391311,
        -1.675516,2.747398,1.727166,
        -1.675516,2.763799,0.760372,
        -1.675516,2.768238,-1.679243,
        -1.675516,2.779245,-2.046268,
        -1.675516,2.780208,-0.626453,
        -1.675516,2.780471,0.027732,
        -1.675516,2.784696,-1.013128,
        -1.675516,2.789817,0.395431,
        -1.675516,2.801154,2.353143,
        -1.675516,2.807091,-2.687477,
        -1.675516,2.825762,1.085035,
        -1.675516,2.830209,2.734943,
        -1.675516,2.833517,-3.085508,
        -1.675516,2.838301,1.980061,
        -1.675516,2.849871,1.512012,
        -1.675516,2.850089,-1.343899,
        -1.675516,2.856905,-0.318441,
        -1.675516,2.870445,-2.291001,
        -1.675516,2.885017,0.655562,
        -1.675516,2.886428,-1.793217,
        -1.675516,2.890337,-0.825633,
        -1.675516,2.898932,0.139973,
        -1.675516,2.931927,-2.728768,
        -1.675516,2.932096,2.377790,
        -1.675516,2.937409,2.947245,
        -1.675516,2.948048,1.133127,
        -1.675516,2.950896,1.797889,
        -1.675516,2.970019,-1.282382,
        -1.675516,2.978333,-0.412090,
        -1.675516,2.996437,-2.089121,
        -1.675516,3.011274,0.434110,
        -1.675516,3.050410,3.138492,
        -1.675516,3.064492,1.647259,
        -1.675516,3.088868,-0.989501,
        -1.570796,0.053884,-2.408394,
        -1.570796,0.077645,1.453433,
        -1.570796,0.088408,-0.311160,
        -1.570796,0.142314,2.590387,
        -1.570796,0.147206,-1.385164,
        -1.570796,0.163506,0.573203,
        -1.570796,0.165485,-2.875396,
        -1.570796,0.182564,-2.125821,
        -1.570796,0.190569,1.929141,
        -1.570796,0.201102,1.249082,
        -1.570796,0.201397,-0.716646,
        -1.570796,0.210568,-0.069559,
        -1.570796,0.251672,2.848900,
        -1.570796,0.266262,-1.678068,
        -1.570796,0.269108,-1.215563,
        -1.570796,0.273777,-2.506756,
        -1.570796,0.279854,0.814581,
        -1.570796,0.281664,-2.969147,
        -1.570796,0.283625,2.405603,
        -1.570796,0.284058,0.357350,
        -1.570796,0.294107,1.593880,
        -1.570796,0.302491,-0.400410,
        -1.570796,0.308059,2.021491,
        -1.570796,0.312219,-2.072057,
        -1.570796,0.321905,-0.829121,
        -1.570796,0.329984,1.201621,
        -1.570796,0.339714,-0.013712,
        -1.570796,0.368784,2.974974,
        -1.570796,0.376428,-1.495532,
        -1.570796,0.378103,2.646436,
        -1.570796,0.385088,0.590355,
        -1.570796,0.386248,-2.690943,
        -1.570796,0.389851,-2.340031,
        -1.570796,0.391502,-1.171374,
        -1.570796,0.398249,-1.812694,
        -1.570796,0.400542,-2.995729,
        -1.570796,0.404273,0.918163,
        -1.570796,0.405617,1.773533,
        -1.570796,0.409009,-0.570243,
        -1.570796,0.410700,0.270433,
        -1.570796,0.415893,1.464890,
        -1.570796,0.420680,2.088164,
        -1.570796,0.424766,-0.258729,
        -1.570796,0.426507,2.367872,
        -1.570796,0.441292,-0.890727,
        -1.570796,0.451839,-2.078193,
        -1.570796,0.459270,1.191562,
        -1.570796,0.470622,0.014002,
        -1.570796,0.481292,-2.529098,
        -1.570796,0.482460,2.849843,
        -1.570796,0.492205,-1.397963,
        -1.570796,0.493314,3.117264,
        -1.570796,0.497870,0.727828,
        -1.570796,0.499783,-1.654379,
        -1.570796,0.502651,0.463333,
        -1.570796,0.505463,2.594557,
        -1.570796,0.506104,-2.874910,
        -1.570796,0.513579,-1.147354,
        -1.570796,0.518569,1.644237,
        -1.570796,0.519001,1.900488,
        -1.570796,0.519277,-0.673809,
        -1.570796,0.524769,-0.421955,
        -1.570796,0.525564,-2.288838,
        -1.570796,0.532541,0.978737,
        -1.570796,0.534610,2.154808,
        -1.570796,0.535020,-1.891846,
        -1.570796,0.540208,1.400497,
        -1.570796,0.542006,0.228337,
        -1.570796,0.550762,-0.180771,
        -1.570796,0.560531,2.378474,
        -1.570796,0.561629,-0.924576,
        -1.570796,0.567465,-2.679392,
        -1.570796,0.588527,1.181755,
        -1.570796,0.591643,-2.098364,
        -1.570796,0.594963,2.997575,
        -1.570796,0.599643,-3.045235,
        -1.570796,0.599962,2.773556,
        -1.570796,0.602620,0.029270,
        -1.570796,0.605711,0.598454,
        -1.570796,0.607708,-1.541026,
        -1.570796,0.609360,-2.464379,
        -1.570796,0.610738,-1.326786,
        -1.570796,0.616968,0.820489,
        -1.570796,0.626379,1.773610,
        -1.570796,0.628626,-1.751180,
        -1.570796,0.630671,0.391949,
        -1.570796,0.632622,-0.737642,
        -1.570796,0.632708,2.566722,
        -1.570796,0.632797,1.989858,
        -1.570796,0.634455,-0.534790,
        -1.570796,0.636205,1.561170,
        -1.570796,0.637681,-1.123423,
        -1.570796,0.641441,-2.842109,
        -1.570796,0.644807,-0.327139,
        -1.570796,0.652440,2.197129,
        -1.570796,0.660897,-2.278593,
        -1.570796,0.664008,1.360091,
        -1.570796,0.670850,-1.945257,
        -1.570796,0.671396,1.004859,
        -1.570796,0.673985,0.203855,
        -1.570796,0.680326,-0.133370,
        -1.570796,0.685083,-0.942363,
        -1.570796,0.688204,-2.632204,
        -1.570796,0.689777,2.385011,
        -1.570796,0.700065,3.115430,
        -1.570796,0.704080,2.911636,
        -1.570796,0.712883,1.174216,
        -1.570796,0.715701,0.699906,
        -1.570796,0.718851,-1.442215,
        -1.570796,0.721019,2.721536,
        -1.570796,0.725300,-2.975134,
        -1.570796,0.727136,-1.633288,
        -1.570796,0.727399,-2.120567,
        -1.570796,0.730751,0.522488,
        -1.570796,0.732892,-1.254330,
        -1.570796,0.734333,-2.446184,
        -1.570796,0.735125,0.037270,
        -1.570796,0.737893,1.866463,
        -1.570796,0.740650,1.681075,
        -1.570796,0.745971,-0.781286,
        -1.570796,0.749432,-0.442559,
        -1.570796,0.750102,2.050636,
        -1.570796,0.751046,0.863821,
        -1.570796,0.757018,-0.608833,
        -1.570796,0.757046,-1.816019,
        -1.570796,0.757389,1.501184,
        -1.570796,0.759800,2.546181,
        -1.570796,0.759859,0.351923,
        -1.570796,0.761141,-2.781102,
        -1.570796,0.769035,-1.078574,
        -1.570796,0.770166,-0.266604,
        -1.570796,0.772218,2.227380,
        -1.570796,0.788382,1.331051,
        -1.570796,0.790269,-2.282409,
        -1.570796,0.803410,-1.984567,
        -1.570796,0.803813,0.191520,
        -1.570796,0.806832,3.022155,
        -1.570796,0.807476,-2.597675,
        -1.570796,0.812512,-0.107335,
        -1.570796,0.814161,1.010949,
        -1.570796,0.817732,-3.084486,
        -1.570796,0.817846,2.390688,
        -1.570796,0.819323,2.845960,
        -1.570796,0.819430,-0.915017,
        -1.570796,0.829198,-1.532491,
        -1.570796,0.829531,0.665695,
        -1.570796,0.829816,1.167147,
        -1.570796,0.833613,-1.357492,
        -1.570796,0.842945,-2.907318,
        -1.570796,0.844883,2.679708,
        -1.570796,0.848189,-1.705003,
        -1.570796,0.849888,1.767901,
        -1.570796,0.851126,0.493370,
        -1.570796,0.852911,1.933394,
        -1.570796,0.857516,-2.432863,
        -1.570796,0.858261,-1.191729,
        -1.570796,0.858279,-2.138895,
        -1.570796,0.859723,-0.734928,
        -1.570796,0.859754,1.603292,
        -1.570796,0.861655,-0.394821,
        -1.570796,0.864300,0.043549,
        -1.570796,0.866940,0.851229,
        -1.570796,0.869748,2.095969,
        -1.570796,0.877352,-0.574668,
        -1.570796,0.879254,-2.733338,
        -1.570796,0.882554,1.443757,
        -1.570796,0.884060,0.334199,
        -1.570796,0.885626,-1.866061,
        -1.570796,0.889834,2.529957,
        -1.570796,0.893140,2.251698,
        -1.570796,0.897463,-0.240822,
        -1.570796,0.898758,-1.031948,
        -1.570796,0.913918,1.287618,
        -1.570796,0.916555,-2.287173,
        -1.570796,0.918305,3.107845,
        -1.570796,0.919522,2.948277,
        -1.570796,0.927138,-2.569686,
        -1.570796,0.929031,0.185383,
        -1.570796,0.933766,-3.015917,
        -1.570796,0.934873,-2.010822,
        -1.570796,0.935636,-1.445855,
        -1.570796,0.936702,-0.866461,
        -1.570796,0.937395,0.729108,
        -1.570796,0.937910,2.787575,
        -1.570796,0.939144,-0.092962,
        -1.570796,0.941718,1.125085,
        -1.570796,0.942873,-1.607680,
        -1.570796,0.943701,0.980328,
        -1.570796,0.947774,2.395256,
        -1.570796,0.952433,0.584347,
        -1.570796,0.954456,-1.287293,
        -1.570796,0.958731,-2.852552,
        -1.570796,0.961838,1.839284,
        -1.570796,0.961898,-0.469425,
        -1.570796,0.965426,1.686615,
        -1.570796,0.970108,1.990771,
        -1.570796,0.972820,-1.762780,
        -1.570796,0.973500,-0.711881,
        -1.570796,0.975253,2.641082,
        -1.570796,0.977824,0.441928,
        -1.570796,0.981485,-2.424766,
        -1.570796,0.981877,1.537100,
        -1.570796,0.982609,-1.136920,
        -1.570796,0.984265,-2.156073,
        -1.570796,0.989150,0.045493,
        -1.570796,0.993052,2.137478,
        -1.570796,0.997766,-2.693871,
        -1.570796,0.999847,-0.334738,
        -1.570796,1.001920,0.848054,
        -1.570796,1.006146,1.389734,
        -1.570796,1.012180,0.304066,
        -1.570796,1.016938,2.270674,
        -1.570796,1.017371,-0.979790,
        -1.570796,1.019495,-1.894487,
        -1.570796,1.023796,-0.583691,
        -1.570796,1.026525,2.506945,
        -1.570796,1.027004,-0.203653,
        -1.570796,1.028027,3.036573,
        -1.570796,1.031435,1.241021,
        -1.570796,1.032316,2.884221,
        -1.570796,1.035863,-3.110475,
        -1.570796,1.042085,-2.292950,
        -1.570796,1.043272,-1.522986,
        -1.570796,1.046265,-2.958847,
        -1.570796,1.049272,-1.376344,
        -1.570796,1.050039,-2.552086,
        -1.570796,1.053682,0.689863,
        -1.570796,1.055458,0.170394,
        -1.570796,1.056622,-0.831866,
        -1.570796,1.057515,1.097998,
        -1.570796,1.058706,-2.036438,
        -1.570796,1.061389,-1.672248,
        -1.570796,1.063758,2.743093,
        -1.570796,1.070157,-0.073422,
        -1.570796,1.072154,1.761506,
        -1.570796,1.074029,1.909193,
        -1.570796,1.074204,0.545276,
        -1.570796,1.074714,-2.805147,
        -1.570796,1.076262,0.966104,
        -1.570796,1.077642,-1.227073,
        -1.570796,1.084151,1.618525,
        -1.570796,1.085085,2.377673,
        -1.570796,1.086822,-0.441112,
        -1.570796,1.087142,2.058815,
        -1.570796,1.098383,-0.696430,
        -1.570796,1.099138,-1.084033,
        -1.570796,1.103208,1.477773,
        -1.570796,1.103783,0.408254,
        -1.570796,1.106275,-2.420800,
        -1.570796,1.108079,2.612943,
        -1.570796,1.109345,-2.172718,
        -1.570796,1.115957,-1.787597,
        -1.570796,1.118930,0.045983,
        -1.570796,1.119777,0.834702,
        -1.570796,1.121943,-2.669042,
        -1.570796,1.124930,2.219917,
        -1.570796,1.125217,1.336294,
        -1.570796,1.130958,2.972931,
        -1.570796,1.131400,-0.306629,
        -1.570796,1.137135,-0.936514,
        -1.570796,1.139125,-1.924528,
        -1.570796,1.139877,0.276382,
        -1.570796,1.142003,-3.058022,
        -1.570796,1.147363,-0.558448,
        -1.570796,1.150348,-0.176418,
        -1.570796,1.151198,1.197016,
        -1.570796,1.153930,-1.597104,
        -1.570796,1.155389,2.837148,
        -1.570796,1.155458,-1.460623,
        -1.570796,1.157331,-2.906362,
        -1.570796,1.158594,2.487618,
        -1.570796,1.162103,-1.322553,
        -1.570796,1.165407,3.092861,
        -1.570796,1.168025,0.717075,
        -1.570796,1.168406,-2.301519,
        -1.570796,1.174782,-2.538755,
        -1.570796,1.177998,1.061256,
        -1.570796,1.178898,1.827305,
        -1.570796,1.181331,-2.060330,
        -1.570796,1.182540,-0.798952,
        -1.570796,1.184569,0.147517,
        -1.570796,1.184734,1.968919,
        -1.570796,1.184804,0.593386,
        -1.570796,1.188278,1.690808,
        -1.570796,1.194360,2.110440,
        -1.570796,1.194776,2.710455,
        -1.570796,1.197745,-1.173705,
        -1.570796,1.199898,-2.777257,
        -1.570796,1.201164,2.355931,
        -1.570796,1.203760,1.556404,
        -1.570796,1.205845,-0.415993,
        -1.570796,1.206585,0.933648,
        -1.570796,1.206950,-0.037248,
        -1.570796,1.209341,0.471162,
        -1.570796,1.213483,-1.039417,
        -1.570796,1.218146,-0.672643,
        -1.570796,1.221608,1.421669,
        -1.570796,1.226043,-1.693533,
        -1.570796,1.232870,-1.822773,
        -1.570796,1.234375,-2.421194,
        -1.570796,1.234531,-2.190148,
        -1.570796,1.239414,0.348959,
        -1.570796,1.240183,2.587255,
        -1.570796,1.242312,2.941234,
        -1.570796,1.244075,1.286644,
        -1.570796,1.245961,-2.997461,
        -1.570796,1.246495,-2.646872,
        -1.570796,1.247736,2.242810,
        -1.570796,1.253604,0.812222,
        -1.570796,1.257816,-0.894317,
        -1.570796,1.258247,-1.953634,
        -1.570796,1.258993,-0.284099,
        -1.570796,1.262380,-0.157609,
        -1.570796,1.262791,-1.549229,
        -1.570796,1.263772,-0.533073,
        -1.570796,1.266731,-1.412461,
        -1.570796,1.268591,-3.122146,
        -1.570796,1.269614,-1.275246,
        -1.570796,1.270745,1.152533,
        -1.570796,1.273606,0.224545,
        -1.570796,1.279716,2.812101,
        -1.570796,1.283786,2.462124,
        -1.570796,1.284798,0.078209,
        -1.570796,1.285402,-2.882049,
        -1.570796,1.286936,1.886053,
        -1.570796,1.289408,0.683122,
        -1.570796,1.294736,1.756184,
        -1.570796,1.296978,-2.311511,
        -1.570796,1.298546,2.020601,
        -1.570796,1.300924,1.022478,
        -1.570796,1.303305,3.043152,
        -1.570796,1.304210,-2.084225,
        -1.570796,1.305241,-2.530249,
        -1.570796,1.307513,1.628003,
        -1.570796,1.310722,0.558267,
        -1.570796,1.314142,2.139329,
        -1.570796,1.316087,-0.777971,
        -1.570796,1.317907,-1.121023,
        -1.570796,1.321019,-2.749776,
        -1.570796,1.321609,2.687717,
        -1.570796,1.322568,1.499415,
        -1.570796,1.322972,-0.395699,
        -1.570796,1.325286,-0.994192,
        -1.570796,1.329146,-0.648513,
        -1.570796,1.329795,-0.042570,
        -1.570796,1.331651,2.344497,
        -1.570796,1.336522,0.435098,
        -1.570796,1.339410,0.900087,
        -1.570796,1.339550,1.370353,
        -1.570796,1.340441,-1.731580,
        -1.570796,1.351855,-1.855234,
        -1.570796,1.360161,-3.018404,
        -1.570796,1.361031,-1.613971,
        -1.570796,1.361196,2.914780,
        -1.570796,1.361998,-2.207254,
        -1.570796,1.362302,1.239384,
        -1.570796,1.365228,2.563808,
        -1.570796,1.365452,0.310522,
        -1.570796,1.367028,-2.420041,
        -1.570796,1.373422,-1.484281,
        -1.570796,1.374245,-1.226417,
        -1.570796,1.375020,-0.170393,
        -1.570796,1.376560,-1.355481,
        -1.570796,1.377070,-1.980984,
        -1.570796,1.377578,-2.636339,
        -1.570796,1.378096,0.774626,
        -1.570796,1.379086,-0.508192,
        -1.570796,1.382562,0.177141,
        -1.570796,1.388849,2.233847,
        -1.570796,1.389268,-0.886349,
        -1.570796,1.391324,1.109396,
        -1.570796,1.391699,-0.288725,
        -1.570796,1.392869,3.136903,
        -1.570796,1.393612,1.952016,
        -1.570796,1.395545,-2.850807,
        -1.570796,1.401926,1.822930,
        -1.570796,1.402149,2.788898,
        -1.570796,1.409552,0.647980,
        -1.570796,1.410840,0.056219,
        -1.570796,1.411944,2.444377,
        -1.570796,1.413639,1.697982,
        -1.570796,1.426479,1.573427,
        -1.570796,1.426725,2.102242,
        -1.570796,1.427868,0.984637,
        -1.570796,1.429580,-2.107232,
        -1.570796,1.431375,-2.313269,
        -1.570796,1.431907,3.015410,
        -1.570796,1.432325,-1.029054,
        -1.570796,1.433341,0.524268,
        -1.570796,1.435181,-0.739771,
        -1.570796,1.437208,-2.529085,
        -1.570796,1.439041,-0.617966,
        -1.570796,1.441473,1.448544,
        -1.570796,1.444527,2.664744,
        -1.570796,1.450513,-0.400625,
        -1.570796,1.451907,-0.066823,
        -1.570796,1.452754,-2.738631,
        -1.570796,1.454061,-2.950416,
        -1.570796,1.456736,1.323154,
        -1.570796,1.458370,0.400433,
        -1.570796,1.459314,-1.765003,
        -1.570796,1.459902,-1.145835,
        -1.570796,1.464952,2.329245,
        -1.570796,1.466050,0.862104,
        -1.570796,1.471330,-1.886160,
        -1.570796,1.475119,-1.648255,
        -1.570796,1.480878,1.194576,
        -1.570796,1.480942,0.273235,
        -1.570796,1.481086,2.890269,
        -1.570796,1.484118,-1.529156,
        -1.570796,1.484741,-1.279219,
        -1.570796,1.487681,-1.406069,
        -1.570796,1.487886,-3.066037,
        -1.570796,1.490555,2.543284,
        -1.570796,1.492424,-0.190932,
        -1.570796,1.495002,-0.839810,
        -1.570796,1.495297,2.007188,
        -1.570796,1.496414,-2.009267,
        -1.570796,1.498261,-2.425270,
        -1.570796,1.498907,-2.207327,
        -1.570796,1.500204,0.737638,
        -1.570796,1.502469,0.151156,
        -1.570796,1.507448,-0.514782,
        -1.570796,1.509287,2.204647,
        -1.570796,1.510518,1.884920,
        -1.570796,1.511948,-2.632360,
        -1.570796,1.515235,1.068565,
        -1.570796,1.522237,3.102018,
        -1.570796,1.522373,1.764991,
        -1.570796,1.522416,-0.954956,
        -1.570796,1.523102,2.765920,
        -1.570796,1.524811,-2.844718,
        -1.570796,1.527261,-0.304916,
        -1.570796,1.529788,0.613208,
        -1.570796,1.533648,1.644808,
        -1.570796,1.535018,0.028869,
        -1.570796,1.541082,2.424608,
        -1.570796,1.546010,1.523462,
        -1.570796,1.551135,-0.642994,
        -1.570796,1.553895,0.489607,
        -1.570796,1.555041,0.947813,
        -1.570796,1.559671,-1.074934,
        -1.570796,1.560345,2.985899,
        -1.570796,1.560487,-2.323400,
        -1.570796,1.561127,1.401111,
        -1.570796,1.567193,2.643172,
        -1.570796,1.568892,2.098362,
        -1.570796,1.571702,-2.098435,
        -1.570796,1.572701,-2.528730,
        -1.570796,1.573572,1.278593,
        -1.570796,1.574849,-0.091505,
        -1.570796,1.576334,0.365236,
        -1.570796,1.577029,-2.962888,
        -1.570796,1.577688,-1.814941,
        -1.570796,1.579542,-0.755917,
        -1.570796,1.579624,-1.199759,
        -1.570796,1.581515,-0.419414,
        -1.570796,1.584581,-2.738702,
        -1.570796,1.586948,-1.696296,
        -1.570796,1.588901,2.301496,
        -1.570796,1.589936,0.825020,
        -1.570796,1.590018,-1.941652,
        -1.570796,1.596968,-1.328358,
        -1.570796,1.597178,-1.575607,
        -1.570796,1.600470,-1.453196,
        -1.570796,1.601408,0.242409,
        -1.570796,1.602259,1.152656,
        -1.570796,1.603064,2.865774,
        -1.570796,1.612085,1.968802,
        -1.570796,1.612720,-3.080498,
        -1.570796,1.616061,2.521531,
        -1.570796,1.616900,-0.207457,
        -1.570796,1.620988,-0.875920,
        -1.570796,1.622114,0.701509,
        -1.570796,1.622749,0.121426,
        -1.570796,1.622993,-2.221629,
        -1.570796,1.629346,-0.542669,
        -1.570796,1.630007,1.844552,
        -1.570796,1.634747,-2.426198,
        -1.570796,1.642346,1.722617,
        -1.570796,1.643925,1.034629,
        -1.570796,1.644848,2.743774,
        -1.570796,1.645258,2.188109,
        -1.570796,1.645417,-2.634725,
        -1.570796,1.647395,3.085437,
        -1.570796,1.648538,-2.855044,
        -1.570796,1.649932,0.576808,
        -1.570796,1.652224,-0.996105,
        -1.570796,1.652871,1.600681,
        -1.570796,1.657997,-0.325127,
        -1.570796,1.661936,0.001037,
        -1.570796,1.665187,1.477402,
        -1.570796,1.665496,2.397872,
        -1.570796,1.673267,-0.665453,
        -1.570796,1.673557,0.452526,
        -1.570796,1.679277,-1.122421,
        -1.570796,1.679696,0.912482,
        -1.570796,1.682420,1.353922,
        -1.570796,1.685031,-2.009038,
        -1.570796,1.685196,2.967046,
        -1.570796,1.690107,2.620910,
        -1.570796,1.690453,-2.123076,
        -1.570796,1.691693,-1.871572,
        -1.570796,1.691817,2.061688,
        -1.570796,1.692644,1.237670,
        -1.570796,1.696386,0.328123,
        -1.570796,1.697162,-1.249026,
        -1.570796,1.697488,-2.322785,
        -1.570796,1.698726,-2.973576,
        -1.570796,1.699172,-1.749717,
        -1.570796,1.704478,-0.106295,
        -1.570796,1.707244,-0.444770,
        -1.570796,1.707500,-2.531497,
        -1.570796,1.707714,-0.787321,
        -1.570796,1.708430,-1.627552,
        -1.570796,1.710583,-1.376890,
        -1.570796,1.712415,-2.747921,
        -1.570796,1.713125,-1.503073,
        -1.570796,1.713452,0.789419,
        -1.570796,1.720612,2.279888,
        -1.570796,1.724852,0.202831,
        -1.570796,1.724868,2.846246,
        -1.570796,1.728940,1.931535,
        -1.570796,1.732838,1.127584,
        -1.570796,1.735719,-3.092485,
        -1.570796,1.739676,2.496017,
        -1.570796,1.739971,-0.228876,
        -1.570796,1.743328,0.663347,
        -1.570796,1.744927,-0.911892,
        -1.570796,1.745952,0.084733,
        -1.570796,1.748077,1.804354,
        -1.570796,1.752959,-0.567548,
        -1.570796,1.760169,1.678132,
        -1.570796,1.763887,-2.218450,
        -1.570796,1.766737,2.722752,
        -1.570796,1.767999,1.004737,
        -1.570796,1.769149,1.550298,
        -1.570796,1.770119,-2.427175,
        -1.570796,1.770460,2.154208,
        -1.570796,1.770514,3.070982,
        -1.570796,1.770735,0.537245,
        -1.570796,1.772673,-2.864760,
        -1.570796,1.773527,-1.039582,
        -1.570796,1.776419,-2.642421,
        -1.570796,1.780769,1.416330,
        -1.570796,1.786634,-0.346247,
        -1.570796,1.788829,-1.947553,
        -1.570796,1.793737,0.410270,
        -1.570796,1.794553,2.374618,
        -1.570796,1.794699,-0.694061,
        -1.570796,1.797469,-1.169105,
        -1.570796,1.803705,0.881076,
        -1.570796,1.804124,1.264249,
        -1.570796,1.807825,-2.090455,
        -1.570796,1.808822,-1.817375,
        -1.570796,1.809020,2.949741,
        -1.570796,1.812466,2.597595,
        -1.570796,1.813159,2.025395,
        -1.570796,1.813816,-1.298435,
        -1.570796,1.815016,0.278951,
        -1.570796,1.816510,-0.008631,
        -1.570796,1.817772,-1.689625,
        -1.570796,1.821360,-2.984352,
        -1.570796,1.822312,-0.127211,
        -1.570796,1.823386,-1.429807,
        -1.570796,1.823834,-1.560483,
        -1.570796,1.832060,-0.821627,
        -1.570796,1.832690,-0.468440,
        -1.570796,1.833945,-2.320278,
        -1.570796,1.834787,0.751968,
        -1.570796,1.840579,-2.536445,
        -1.570796,1.841033,-2.756932,
        -1.570796,1.846350,1.893983,
        -1.570796,1.846644,1.113766,
        -1.570796,1.846833,2.826543,
        -1.570796,1.847325,2.249051,
        -1.570796,1.854119,0.134161,
        -1.570796,1.858863,-3.106022,
        -1.570796,1.865046,1.763384,
        -1.570796,1.865619,-0.952036,
        -1.570796,1.865912,0.624453,
        -1.570796,1.866694,2.473686,
        -1.570796,1.867524,-0.240083,
        -1.570796,1.875561,1.632831,
        -1.570796,1.876049,-0.594378,
        -1.570796,1.880509,1.500204,
        -1.570796,1.883504,1.360457,
        -1.570796,1.886971,-2.196665,
        -1.570796,1.888979,0.982922,
        -1.570796,1.889815,2.701570,
        -1.570796,1.892144,-2.005340,
        -1.570796,1.892348,0.496522,
        -1.570796,1.893281,-1.083976,
        -1.570796,1.893828,3.053453,
        -1.570796,1.895117,2.121107,
        -1.570796,1.898741,-2.875735,
        -1.570796,1.904429,-2.428029,
        -1.570796,1.908408,-2.648903,
        -1.570796,1.911828,-0.364489,
        -1.570796,1.912224,0.366565,
        -1.570796,1.914854,1.214468,
        -1.570796,1.915467,-1.883143,
        -1.570796,1.915763,-0.724679,
        -1.570796,1.915790,-1.216366,
        -1.570796,1.922256,2.347806,
        -1.570796,1.922616,0.844633,
        -1.570796,1.925185,0.231170,
        -1.570796,1.927474,-1.754616,
        -1.570796,1.928342,-1.351174,
        -1.570796,1.933328,-1.623328,
        -1.570796,1.933878,0.034150,
        -1.570796,1.934585,-1.487979,
        -1.570796,1.934802,2.927634,
        -1.570796,1.934862,1.989776,
        -1.570796,1.936327,-0.099350,
        -1.570796,1.938088,2.578507,
        -1.570796,1.945677,-2.999805,
        -1.570796,1.954572,-0.859208,
        -1.570796,1.957868,-0.490784,
        -1.570796,1.958989,0.712772,
        -1.570796,1.960212,-2.307561,
        -1.570796,1.964774,1.855507,
        -1.570796,1.964971,-2.100533,
        -1.570796,1.969112,1.088273,
        -1.570796,1.971438,-2.765415,
        -1.570796,1.972312,2.805134,
        -1.570796,1.974950,2.220596,
        -1.570796,1.975135,-2.539127,
        -1.570796,1.982107,1.720829,
        -1.570796,1.983195,-3.127984,
        -1.570796,1.983318,-0.247182,
        -1.570796,1.986720,-0.993220,
        -1.570796,1.990381,1.586005,
        -1.570796,1.990470,0.580283,
        -1.570796,1.990523,1.450326,
        -1.570796,1.993909,1.312897,
        -1.570796,1.994694,2.452813,
        -1.570796,1.999226,-0.620717,
        -1.570796,2.005145,0.941012,
        -1.570796,2.007301,0.142337,
        -1.570796,2.014149,0.444138,
        -1.570796,2.014338,-1.127392,
        -1.570796,2.018131,3.021778,
        -1.570796,2.019676,-1.980944,
        -1.570796,2.019912,2.686612,
        -1.570796,2.020545,2.088672,
        -1.570796,2.026074,0.300945,
        -1.570796,2.026196,-2.889057,
        -1.570796,2.031959,-1.268178,
        -1.570796,2.032917,-2.199917,
        -1.570796,2.033572,-1.842884,
        -1.570796,2.034712,-2.419505,
        -1.570796,2.035818,-0.379479,
        -1.570796,2.036004,-0.757062,
        -1.570796,2.040501,-1.411488,
        -1.570796,2.042874,-1.701367,
        -1.570796,2.043060,-2.653746,
        -1.570796,2.043729,-1.557938,
        -1.570796,2.048538,0.806623,
        -1.570796,2.049006,1.199930,
        -1.570796,2.049502,-0.137989,
        -1.570796,2.052181,2.326959,
        -1.570796,2.052794,-0.008526,
        -1.570796,2.057608,1.951726,
        -1.570796,2.065568,2.567730,
        -1.570796,2.070371,2.879808,
        -1.570796,2.071256,-3.018961,
        -1.570796,2.079746,-0.897699,
        -1.570796,2.082784,1.042403,
        -1.570796,2.083752,-0.511086,
        -1.570796,2.084401,1.810502,
        -1.570796,2.085731,0.671081,
        -1.570796,2.097496,1.373172,
        -1.570796,2.099185,1.667006,
        -1.570796,2.100482,-2.301084,
        -1.570796,2.103118,2.194745,
        -1.570796,2.103223,1.521685,
        -1.570796,2.103695,-2.777707,
        -1.570796,2.107523,3.129581,
        -1.570796,2.108277,-2.095907,
        -1.570796,2.108891,-2.532166,
        -1.570796,2.111003,-0.268632,
        -1.570796,2.112276,-1.027246,
        -1.570796,2.114380,0.531457,
        -1.570796,2.119166,0.100552,
        -1.570796,2.119564,2.754818,
        -1.570796,2.122474,-0.648314,
        -1.570796,2.125183,2.445598,
        -1.570796,2.129501,0.238290,
        -1.570796,2.131178,0.386663,
        -1.570796,2.131223,-1.174311,
        -1.570796,2.134269,-1.948042,
        -1.570796,2.134684,0.907065,
        -1.570796,2.137615,2.988806,
        -1.570796,2.145635,-1.327708,
        -1.570796,2.146467,2.055081,
        -1.570796,2.146582,-1.793788,
        -1.570796,2.149210,-1.482321,
        -1.570796,2.152901,-1.636788,
        -1.570796,2.156337,1.151558,
        -1.570796,2.156474,-2.912065,
        -1.570796,2.157194,-0.792563,
        -1.570796,2.166724,-0.401638,
        -1.570796,2.166980,-0.134879,
        -1.570796,2.173644,-2.403429,
        -1.570796,2.175711,-2.655523,
        -1.570796,2.177132,2.600419,
        -1.570796,2.178879,0.767722,
        -1.570796,2.180752,1.908485,
        -1.570796,2.181338,2.313721,
        -1.570796,2.185580,1.287673,
        -1.570796,2.195400,-2.217546,
        -1.570796,2.200418,-3.056442,
        -1.570796,2.201829,-0.001735,
        -1.570796,2.203804,1.756176,
        -1.570796,2.211602,1.443930,
        -1.570796,2.212686,-0.544809,
        -1.570796,2.213937,0.621892,
        -1.570796,2.214075,1.600691,
        -1.570796,2.216256,1.020029,
        -1.570796,2.219060,2.886631,
        -1.570796,2.223375,-0.906185,
        -1.570796,2.226002,-1.060248,
        -1.570796,2.226413,-2.064034,
        -1.570796,2.230554,2.171334,
        -1.570796,2.232413,2.735777,
        -1.570796,2.233709,3.070830,
        -1.570796,2.234383,-2.792065,
        -1.570796,2.237183,0.468799,
        -1.570796,2.237695,-0.271388,
        -1.570796,2.239419,0.151693,
        -1.570796,2.243347,-2.521049,
        -1.570796,2.244035,-1.229456,
        -1.570796,2.244355,0.309899,
        -1.570796,2.245175,2.460236,
        -1.570796,2.245628,-1.899216,
        -1.570796,2.245973,-0.702234,
        -1.570796,2.255601,-1.728131,
        -1.570796,2.256702,-1.401367,
        -1.570796,2.261466,-1.557500,
        -1.570796,2.267390,0.878499,
        -1.570796,2.272099,2.018594,
        -1.570796,2.278618,1.170386,
        -1.570796,2.283304,-2.345584,
        -1.570796,2.285175,-2.943090,
        -1.570796,2.293000,-0.124451,
        -1.570796,2.294058,-0.421070,
        -1.570796,2.303283,1.855938,
        -1.570796,2.304968,2.608978,
        -1.570796,2.305347,2.311346,
        -1.570796,2.307188,-2.656134,
        -1.570796,2.307301,1.336200,
        -1.570796,2.309655,0.726020,
        -1.570796,2.317634,-2.185225,
        -1.570796,2.321510,1.685664,
        -1.570796,2.324376,-3.111006,
        -1.570796,2.324731,1.511186,
        -1.570796,2.329587,0.035498,
        -1.570796,2.330158,2.977264,
        -1.570796,2.335096,-0.588468,
        -1.570796,2.336046,-1.115799,
        -1.570796,2.338060,2.803613,
        -1.570796,2.338083,-0.943400,
        -1.570796,2.340210,0.562145,
        -1.570796,2.341931,1.023926,
        -1.570796,2.343265,-0.779061,
        -1.570796,2.343434,-2.013692,
        -1.570796,2.353029,-1.310529,
        -1.570796,2.354868,0.211543,
        -1.570796,2.355120,0.388593,
        -1.570796,2.355165,-1.833173,
        -1.570796,2.355462,2.151441,
        -1.570796,2.362948,-2.487477,
        -1.570796,2.363101,-2.810456,
        -1.570796,2.363394,-0.272918,
        -1.570796,2.364509,-1.644533,
        -1.570796,2.370465,2.466824,
        -1.570796,2.386110,-1.481351,
        -1.570796,2.394960,0.862133,
        -1.570796,2.395716,1.977184,
        -1.570796,2.396339,1.200900,
        -1.570796,2.407941,-2.316753,
        -1.570796,2.409450,-2.988432,
        -1.570796,2.417202,-0.103363,
        -1.570796,2.417386,-0.443097,
        -1.570796,2.417950,2.666662,
        -1.570796,2.423190,1.789783,
        -1.570796,2.423692,1.392696,
        -1.570796,2.431122,3.086379,
        -1.570796,2.431406,-2.647820,
        -1.570796,2.431584,2.307631,
        -1.570796,2.434090,1.592818,
        -1.570796,2.435993,0.683844,
        -1.570796,2.440898,-2.133300,
        -1.570796,2.442222,-1.195837,
        -1.570796,2.443112,2.883761,
        -1.570796,2.449191,-0.633651,
        -1.570796,2.450821,0.084741,
        -1.570796,2.453499,-1.008861,
        -1.570796,2.460010,-1.938011,
        -1.570796,2.460199,-0.822008,
        -1.570796,2.460902,0.490079,
        -1.570796,2.461774,-1.734913,
        -1.570796,2.463986,1.034407,
        -1.570796,2.467647,0.287667,
        -1.570796,2.479849,2.126925,
        -1.570796,2.485907,-2.835471,
        -1.570796,2.485945,-1.374853,
        -1.570796,2.487747,2.505912,
        -1.570796,2.487957,-0.274279,
        -1.570796,2.489741,-2.468428,
        -1.570796,2.512538,1.239900,
        -1.570796,2.517528,1.923289,
        -1.570796,2.518454,-1.565852,
        -1.570796,2.519348,0.843503,
        -1.570796,2.522766,-3.059231,
        -1.570796,2.528686,2.737228,
        -1.570796,2.534798,-2.266671,
        -1.570796,2.536533,1.466274,
        -1.570796,2.538504,1.699812,
        -1.570796,2.539516,-0.476942,
        -1.570796,2.539532,-0.074548,
        -1.570796,2.543197,2.990252,
        -1.570796,2.554022,2.320619,
        -1.570796,2.556919,-2.655005,
        -1.570796,2.558146,0.627191,
        -1.570796,2.558223,-1.164189,
        -1.570796,2.563452,-2.044028,
        -1.570796,2.568462,0.150655,
        -1.570796,2.569130,-0.700837,
        -1.570796,2.570050,-1.804492,
        -1.570796,2.572710,-0.934665,
        -1.570796,2.575994,0.391718,
        -1.570796,2.584774,1.046703,
        -1.570796,2.601432,2.102535,
        -1.570796,2.603559,2.555114,
        -1.570796,2.604150,-2.885208,
        -1.570796,2.608630,-1.378011,
        -1.570796,2.611256,-0.279110,
        -1.570796,2.617344,-2.439480,
        -1.570796,2.627402,1.293552,
        -1.570796,2.635396,1.848850,
        -1.570796,2.635686,2.834950,
        -1.570796,2.637087,3.132730,
        -1.570796,2.642029,0.813274,
        -1.570796,2.645260,1.570203,
        -1.570796,2.646182,-1.614036,
        -1.570796,2.660113,-2.186513,
        -1.570796,2.660483,-0.034608,
        -1.570796,2.661713,-0.529511,
        -1.570796,2.672720,-1.110422,
        -1.570796,2.676626,2.335353,
        -1.570796,2.676638,0.540364,
        -1.570796,2.676780,-1.901158,
        -1.570796,2.681411,-2.667965,
        -1.570796,2.682152,0.245751,
        -1.570796,2.682859,-0.817071,
        -1.570796,2.705107,1.062142,
        -1.570796,2.717705,2.625995,
        -1.570796,2.721001,2.060097,
        -1.570796,2.722319,-2.954069,
        -1.570796,2.729840,-1.368237,
        -1.570796,2.734349,-0.289298,
        -1.570796,2.738409,2.975671,
        -1.570796,2.740366,1.374341,
        -1.570796,2.744633,-2.391311,
        -1.570796,2.747398,1.727166,
        -1.570796,2.763799,0.760372,
        -1.570796,2.768238,-1.679243,
        -1.570796,2.779245,-2.046268,
        -1.570796,2.780208,-0.626453,
        -1.570796,2.780471,0.027732,
        -1.570796,2.784696,-1.013128,
        -1.570796,2.789817,0.395431,
        -1.570796,2.801154,2.353143,
        -1.570796,2.807091,-2.687477,
        -1.570796,2.825762,1.085035,
        -1.570796,2.830209,2.734943,
        -1.570796,2.833517,-3.085508,
        -1.570796,2.838301,1.980061,
        -1.570796,2.849871,1.512012,
        -1.570796,2.850089,-1.343899,
        -1.570796,2.856905,-0.318441,
        -1.570796,2.870445,-2.291001,
        -1.570796,2.885017,0.655562,
        -1.570796,2.886428,-1.793217,
        -1.570796,2.890337,-0.825633,
        -1.570796,2.898932,0.139973,
        -1.570796,2.931927,-2.728768,
        -1.570796,2.932096,2.377790,
        -1.570796,2.937409,2.947245,
        -1.570796,2.948048,1.133127,
        -1.570796,2.950896,1.797889,
        -1.570796,2.970019,-1.282382,
        -1.570796,2.978333,-0.412090,
        -1.570796,2.996437,-2.089121,
        -1.570796,3.011274,0.434110,
        -1.570796,3.050410,3.138492,
        -1.570796,3.064492,1.647259,
        -1.570796,3.088868,-0.989501,
        -1.466077,0.053884,-2.408394,
        -1.466077,0.077645,1.453433,
        -1.466077,0.088408,-0.311160,
        -1.466077,0.142314,2.590387,
        -1.466077,0.147206,-1.385164,
        -1.466077,0.163506,0.573203,
        -1.466077,0.165485,-2.875396,
        -1.466077,0.182564,-2.125821,
        -1.466077,0.190569,1.929141,
        -1.466077,0.201102,1.249082,
        -1.466077,0.201397,-0.716646,
        -1.466077,0.210568,-0.069559,
        -1.466077,0.251672,2.848900,
        -1.466077,0.266262,-1.678068,
        -1.466077,0.269108,-1.215563,
        -1.466077,0.273777,-2.506756,
        -1.466077,0.279854,0.814581,
        -1.466077,0.281664,-2.969147,
        -1.466077,0.283625,2.405603,
        -1.466077,0.284058,0.357350,
        -1.466077,0.294107,1.593880,
        -1.466077,0.302491,-0.400410,
        -1.466077,0.308059,2.021491,
        -1.466077,0.312219,-2.072057,
        -1.466077,0.321905,-0.829121,
        -1.466077,0.329984,1.201621,
        -1.466077,0.339714,-0.013712,
        -1.466077,0.368784,2.974974,
        -1.466077,0.376428,-1.495532,
        -1.466077,0.378103,2.646436,
        -1.466077,0.385088,0.590355,
        -1.466077,0.386248,-2.690943,
        -1.466077,0.389851,-2.340031,
        -1.466077,0.391502,-1.171374,
        -1.466077,0.398249,-1.812694,
        -1.466077,0.400542,-2.995729,
        -1.466077,0.404273,0.918163,
        -1.466077,0.405617,1.773533,
        -1.466077,0.409009,-0.570243,
        -1.466077,0.410700,0.270433,
        -1.466077,0.415893,1.464890,
        -1.466077,0.420680,2.088164,
        -1.466077,0.424766,-0.258729,
        -1.466077,0.426507,2.367872,
        -1.466077,0.441292,-0.890727,
        -1.466077,0.451839,-2.078193,
        -1.466077,0.459270,1.191562,
        -1.466077,0.470622,0.014002,
        -1.466077,0.481292,-2.529098,
        -1.466077,0.482460,2.849843,
        -1.466077,0.492205,-1.397963,
        -1.466077,0.493314,3.117264,
        -1.466077,0.497870,0.727828,
        -1.466077,0.499783,-1.654379,
        -1.466077,0.502651,0.463333,
        -1.466077,0.505463,2.594557,
        -1.466077,0.506104,-2.874910,
        -1.466077,0.513579,-1.147354,
        -1.466077,0.518569,1.644237,
        -1.466077,0.519001,1.900488,
        -1.466077,0.519277,-0.673809,
        -1.466077,0.524769,-0.421955,
        -1.466077,0.525564,-2.288838,
        -1.466077,0.532541,0.978737,
        -1.466077,0.534610,2.154808,
        -1.466077,0.535020,-1.891846,
        -1.466077,0.540208,1.400497,
        -1.466077,0.542006,0.228337,
        -1.466077,0.550762,-0.180771,
        -1.466077,0.560531,2.378474,
        -1.466077,0.561629,-0.924576,
        -1.466077,0.567465,-2.679392,
        -1.466077,0.588527,1.181755,
        -1.466077,0.591643,-2.098364,
        -1.466077,0.594963,2.997575,
        -1.466077,0.599643,-3.045235,
        -1.466077,0.599962,2.773556,
        -1.466077,0.602620,0.029270,
        -1.466077,0.605711,0.598454,
        -1.466077,0.607708,-1.541026,
        -1.466077,0.609360,-2.464379,
        -1.466077,0.610738,-1.326786,
        -1.466077,0.616968,0.820489,
        -1.466077,0.626379,1.773610,
        -1.466077,0.628626,-1.751180,
        -1.466077,0.630671,0.391949,
        -1.466077,0.632622,-0.737642,
        -1.466077,0.632708,2.566722,
        -1.466077,0.632797,1.989858,
        -1.466077,0.634455,-0.534790,
        -1.466077,0.636205,1.561170,
        -1.466077,0.637681,-1.123423,
        -1.466077,0.641441,-2.842109,
        -1.466077,0.644807,-0.327139,
        -1.466077,0.652440,2.197129,
        -1.466077,0.660897,-2.278593,
        -1.466077,0.664008,1.360091,
        -1.466077,0.670850,-1.945257,
        -1.466077,0.671396,1.004859,
        -1.466077,0.673985,0.203855,
        -1.466077,0.680326,-0.133370,
        -1.466077,0.685083,-0.942363,
        -1.466077,0.688204,-2.632204,
        -1.466077,0.689777,2.385011,
        -1.466077,0.700065,3.115430,
        -1.466077,0.704080,2.911636,
        -1.466077,0.712883,1.174216,
        -1.466077,0.715701,0.699906,
        -1.466077,0.718851,-1.442215,
        -1.466077,0.721019,2.721536,
        -1.466077,0.725300,-2.975134,
        -1.466077,0.727136,-1.633288,
        -1.466077,0.727399,-2.120567,
        -1.466077,0.730751,0.522488,
        -1.466077,0.732892,-1.254330,
        -1.466077,0.734333,-2.446184,
        -1.466077,0.735125,0.037270,
        -1.466077,0.737893,1.866463,
        -1.466077,0.740650,1.681075,
        -1.466077,0.745971,-0.781286,
        -1.466077,0.749432,-0.442559,
        -1.466077,0.750102,2.050636,
        -1.466077,0.751046,0.863821,
        -1.466077,0.757018,-0.608833,
        -1.466077,0.757046,-1.816019,
        -1.466077,0.757389,1.501184,
        -1.466077,0.759800,2.546181,
        -1.466077,0.759859,0.351923,
        -1.466077,0.761141,-2.781102,
        -1.466077,0.769035,-1.078574,
        -1.466077,0.770166,-0.266604,
        -1.466077,0.772218,2.227380,
        -1.466077,0.788382,1.331051,
        -1.466077,0.790269,-2.282409,
        -1.466077,0.803410,-1.984567,
        -1.466077,0.803813,0.191520,
        -1.466077,0.806832,3.022155,
        -1.466077,0.807476,-2.597675,
        -1.466077,0.812512,-0.107335,
        -1.466077,0.814161,1.010949,
        -1.466077,0.817732,-3.084486,
        -1.466077,0.817846,2.390688,
        -1.466077,0.819323,2.845960,
        -1.466077,0.819430,-0.915017,
        -1.466077,0.829198,-1.532491,
        -1.466077,0.829531,0.665695,
        -1.466077,0.829816,1.167147,
        -1.466077,0.833613,-1.357492,
        -1.466077,0.842945,-2.907318,
        -1.466077,0.844883,2.679708,
        -1.466077,0.848189,-1.705003,
        -1.466077,0.849888,1.767901,
        -1.466077,0.851126,0.493370,
        -1.466077,0.852911,1.933394,
        -1.466077,0.857516,-2.432863,
        -1.466077,0.858261,-1.191729,
        -1.466077,0.858279,-2.138895,
        -1.466077,0.859723,-0.734928,
        -1.466077,0.859754,1.603292,
        -1.466077,0.861655,-0.394821,
        -1.466077,0.864300,0.043549,
        -1.466077,0.866940,0.851229,
        -1.466077,0.869748,2.095969,
        -1.466077,0.877352,-0.574668,
        -1.466077,0.879254,-2.733338,
        -1.466077,0.882554,1.443757,
        -1.466077,0.884060,0.334199,
        -1.466077,0.885626,-1.866061,
        -1.466077,0.889834,2.529957,
        -1.466077,0.893140,2.251698,
        -1.466077,0.897463,-0.240822,
        -1.466077,0.898758,-1.031948,
        -1.466077,0.913918,1.287618,
        -1.466077,0.916555,-2.287173,
        -1.466077,0.918305,3.107845,
        -1.466077,0.919522,2.948277,
        -1.466077,0.927138,-2.569686,
        -1.466077,0.929031,0.185383,
        -1.466077,0.933766,-3.015917,
        -1.466077,0.934873,-2.010822,
        -1.466077,0.935636,-1.445855,
        -1.466077,0.936702,-0.866461,
        -1.466077,0.937395,0.729108,
        -1.466077,0.937910,2.787575,
        -1.466077,0.939144,-0.092962,
        -1.466077,0.941718,1.125085,
        -1.466077,0.942873,-1.607680,
        -1.466077,0.943701,0.980328,
        -1.466077,0.947774,2.395256,
        -1.466077,0.952433,0.584347,
        -1.466077,0.954456,-1.287293,
        -1.466077,0.958731,-2.852552,
        -1.466077,0.961838,1.839284,
        -1.466077,0.961898,-0.469425,
        -1.466077,0.965426,1.686615,
        -1.466077,0.970108,1.990771,
        -1.466077,0.972820,-1.762780,
        -1.466077,0.973500,-0.711881,
        -1.466077,0.975253,2.641082,
        -1.466077,0.977824,0.441928,
        -1.466077,0.981485,-2.424766,
        -1.466077,0.981877,1.537100,
        -1.466077,0.982609,-1.136920,
        -1.466077,0.984265,-2.156073,
        -1.466077,0.989150,0.045493,
        -1.466077,0.993052,2.137478,
        -1.466077,0.997766,-2.693871,
        -1.466077,0.999847,-0.334738,
        -1.466077,1.001920,0.848054,
        -1.466077,1.006146,1.389734,
        -1.466077,1.012180,0.304066,
        -1.466077,1.016938,2.270674,
        -1.466077,1.017371,-0.979790,
        -1.466077,1.019495,-1.894487,
        -1.466077,1.023796,-0.583691,
        -1.466077,1.026525,2.506945,
        -1.466077,1.027004,-0.203653,
        -1.466077,1.028027,3.036573,
        -1.466077,1.031435,1.241021,
        -1.466077,1.032316,2.884221,
        -1.466077,1.035863,-3.110475,
        -1.466077,1.042085,-2.292950,
        -1.466077,1.043272,-1.522986,
        -1.466077,1.046265,-2.958847,
        -1.466077,1.049272,-1.376344,
        -1.466077,1.050039,-2.552086,
        -1.466077,1.053682,0.689863,
        -1.466077,1.055458,0.170394,
        -1.466077,1.056622,-0.831866,
        -1.466077,1.057515,1.097998,
        -1.466077,1.058706,-2.036438,
        -1.466077,1.061389,-1.672248,
        -1.466077,1.063758,2.743093,
        -1.466077,1.070157,-0.073422,
        -1.466077,1.072154,1.761506,
        -1.466077,1.074029,1.909193,
        -1.466077,1.074204,0.545276,
        -1.466077,1.074714,-2.805147,
        -1.466077,1.076262,0.966104,
        -1.466077,1.077642,-1.227073,
        -1.466077,1.084151,1.618525,
        -1.466077,1.085085,2.377673,
        -1.466077,1.086822,-0.441112,
        -1.466077,1.087142,2.058815,
        -1.466077,1.098383,-0.696430,
        -1.466077,1.099138,-1.084033,
        -1.466077,1.103208,1.477773,
        -1.466077,1.103783,0.408254,
        -1.466077,1.106275,-2.420800,
        -1.466077,1.108079,2.612943,
        -1.466077,1.109345,-2.172718,
        -1.466077,1.115957,-1.787597,
        -1.466077,1.118930,0.045983,
        -1.466077,1.119777,0.834702,
        -1.466077,1.121943,-2.669042,
        -1.466077,1.124930,2.219917,
        -1.466077,1.125217,1.336294,
        -1.466077,1.130958,2.972931,
        -1.466077,1.131400,-0.306629,
        -1.466077,1.137135,-0.936514,
        -1.466077,1.139125,-1.924528,
        -1.466077,1.139877,0.276382,
        -1.466077,1.142003,-3.058022,
        -1.466077,1.147363,-0.558448,
        -1.466077,1.150348,-0.176418,
        -1.466077,1.151198,1.197016,
        -1.466077,1.153930,-1.597104,
        -1.466077,1.155389,2.837148,
        -1.466077,1.155458,-1.460623,
        -1.466077,1.157331,-2.906362,
        -1.466077,1.158594,2.487618,
        -1.466077,1.162103,-1.322553,
        -1.466077,1.165407,3.092861,
        -1.466077,1.168025,0.717075,
        -1.466077,1.168406,-2.301519,
        -1.466077,1.174782,-2.538755,
        -1.466077,1.177998,1.061256,
        -1.466077,1.178898,1.827305,
        -1.466077,1.181331,-2.060330,
        -1.466077,1.182540,-0.798952,
        -1.466077,1.184569,0.147517,
        -1.466077,1.184734,1.968919,
        -1.466077,1.184804,0.593386,
        -1.466077,1.188278,1.690808,
        -1.466077,1.194360,2.110440,
        -1.466077,1.194776,2.710455,
        -1.466077,1.197745,-1.173705,
        -1.466077,1.199898,-2.777257,
        -1.466077,1.201164,2.355931,
        -1.466077,1.203760,1.556404,
        -1.466077,1.205845,-0.415993,
        -1.466077,1.206585,0.933648,
        -1.466077,1.206950,-0.037248,
        -1.466077,1.209341,0.471162,
        -1.466077,1.213483,-1.039417,
        -1.466077,1.218146,-0.672643,
        -1.466077,1.221608,1.421669,
        -1.466077,1.226043,-1.693533,
        -1.466077,1.232870,-1.822773,
        -1.466077,1.234375,-2.421194,
        -1.466077,1.234531,-2.190148,
        -1.466077,1.239414,0.348959,
        -1.466077,1.240183,2.587255,
        -1.466077,1.242312,2.941234,
        -1.466077,1.244075,1.286644,
        -1.466077,1.245961,-2.997461,
        -1.466077,1.246495,-2.646872,
        -1.466077,1.247736,2.242810,
        -1.466077,1.253604,0.812222,
        -1.466077,1.257816,-0.894317,
        -1.466077,1.258247,-1.953634,
        -1.466077,1.258993,-0.284099,
        -1.466077,1.262380,-0.157609,
        -1.466077,1.262791,-1.549229,
        -1.466077,1.263772,-0.533073,
        -1.466077,1.266731,-1.412461,
        -1.466077,1.268591,-3.122146,
        -1.466077,1.269614,-1.275246,
        -1.466077,1.270745,1.152533,
        -1.466077,1.273606,0.224545,
        -1.466077,1.279716,2.812101,
        -1.466077,1.283786,2.462124,
        -1.466077,1.284798,0.078209,
        -1.466077,1.285402,-2.882049,
        -1.466077,1.286936,1.886053,
        -1.466077,1.289408,0.683122,
        -1.466077,1.294736,1.756184,
        -1.466077,1.296978,-2.311511,
        -1.466077,1.298546,2.020601,
        -1.466077,1.300924,1.022478,
        -1.466077,1.303305,3.043152,
        -1.466077,1.304210,-2.084225,
        -1.466077,1.305241,-2.530249,
        -1.466077,1.307513,1.628003,
        -1.466077,1.310722,0.558267,
        -1.466077,1.314142,2.139329,
        -1.466077,1.316087,-0.777971,
        -1.466077,1.317907,-1.121023,
        -1.466077,1.321019,-2.749776,
        -1.466077,1.321609,2.687717,
        -1.466077,1.322568,1.499415,
        -1.466077,1.322972,-0.395699,
        -1.466077,1.325286,-0.994192,
        -1.466077,1.329146,-0.648513,
        -1.466077,1.329795,-0.042570,
        -1.466077,1.331651,2.344497,
        -1.466077,1.336522,0.435098,
        -1.466077,1.339410,0.900087,
        -1.466077,1.339550,1.370353,
        -1.466077,1.340441,-1.731580,
        -1.466077,1.351855,-1.855234,
        -1.466077,1.360161,-3.018404,
        -1.466077,1.361031,-1.613971,
        -1.466077,1.361196,2.914780,
        -1.466077,1.361998,-2.207254,
        -1.466077,1.362302,1.239384,
        -1.466077,1.365228,2.563808,
        -1.466077,1.365452,0.310522,
        -1.466077,1.367028,-2.420041,
        -1.466077,1.373422,-1.484281,
        -1.466077,1.374245,-1.226417,
        -1.466077,1.375020,-0.170393,
        -1.466077,1.376560,-1.355481,
        -1.466077,1.377070,-1.980984,
        -1.466077,1.377578,-2.636339,
        -1.466077,1.378096,0.774626,
        -1.466077,1.379086,-0.508192,
        -1.466077,1.382562,0.177141,
        -1.466077,1.388849,2.233847,
        -1.466077,1.389268,-0.886349,
        -1.466077,1.391324,1.109396,
        -1.466077,1.391699,-0.288725,
        -1.466077,1.392869,3.136903,
        -1.466077,1.393612,1.952016,
        -1.466077,1.395545,-2.850807,
        -1.466077,1.401926,1.822930,
        -1.466077,1.402149,2.788898,
        -1.466077,1.409552,0.647980,
        -1.466077,1.410840,0.056219,
        -1.466077,1.411944,2.444377,
        -1.466077,1.413639,1.697982,
        -1.466077,1.426479,1.573427,
        -1.466077,1.426725,2.102242,
        -1.466077,1.427868,0.984637,
        -1.466077,1.429580,-2.107232,
        -1.466077,1.431375,-2.313269,
        -1.466077,1.431907,3.015410,
        -1.466077,1.432325,-1.029054,
        -1.466077,1.433341,0.524268,
        -1.466077,1.435181,-0.739771,
        -1.466077,1.437208,-2.529085,
        -1.466077,1.439041,-0.617966,
        -1.466077,1.441473,1.448544,
        -1.466077,1.444527,2.664744,
        -1.466077,1.450513,-0.400625,
        -1.466077,1.451907,-0.066823,
        -1.466077,1.452754,-2.738631,
        -1.466077,1.454061,-2.950416,
        -1.466077,1.456736,1.323154,
        -1.466077,1.458370,0.400433,
        -1.466077,1.459314,-1.765003,
        -1.466077,1.459902,-1.145835,
        -1.466077,1.464952,2.329245,
        -1.466077,1.466050,0.862104,
        -1.466077,1.471330,-1.886160,
        -1.466077,1.475119,-1.648255,
        -1.466077,1.480878,1.194576,
        -1.466077,1.480942,0.273235,
        -1.466077,1.481086,2.890269,
        -1.466077,1.484118,-1.529156,
        -1.466077,1.484741,-1.279219,
        -1.466077,1.487681,-1.406069,
        -1.466077,1.487886,-3.066037,
        -1.466077,1.490555,2.543284,
        -1.466077,1.492424,-0.190932,
        -1.466077,1.495002,-0.839810,
        -1.466077,1.495297,2.007188,
        -1.466077,1.496414,-2.009267,
        -1.466077,1.498261,-2.425270,
        -1.466077,1.498907,-2.207327,
        -1.466077,1.500204,0.737638,
        -1.466077,1.502469,0.151156,
        -1.466077,1.507448,-0.514782,
        -1.466077,1.509287,2.204647,
        -1.466077,1.510518,1.884920,
        -1.466077,1.511948,-2.632360,
        -1.466077,1.515235,1.068565,
        -1.466077,1.522237,3.102018,
        -1.466077,1.522373,1.764991,
        -1.466077,1.522416,-0.954956,
        -1.466077,1.523102,2.765920,
        -1.466077,1.524811,-2.844718,
        -1.466077,1.527261,-0.304916,
        -1.466077,1.529788,0.613208,
        -1.466077,1.533648,1.644808,
        -1.466077,1.535018,0.028869,
        -1.466077,1.541082,2.424608,
        -1.466077,1.546010,1.523462,
        -1.466077,1.551135,-0.642994,
        -1.466077,1.553895,0.489607,
        -1.466077,1.555041,0.947813,
        -1.466077,1.559671,-1.074934,
        -1.466077,1.560345,2.985899,
        -1.466077,1.560487,-2.323400,
        -1.466077,1.561127,1.401111,
        -1.466077,1.567193,2.643172,
        -1.466077,1.568892,2.098362,
        -1.466077,1.571702,-2.098435,
        -1.466077,1.572701,-2.528730,
        -1.466077,1.573572,1.278593,
        -1.466077,1.574849,-0.091505,
        -1.466077,1.576334,0.365236,
        -1.466077,1.577029,-2.962888,
        -1.466077,1.577688,-1.814941,
        -1.466077,1.579542,-0.755917,
        -1.466077,1.579624,-1.199759,
        -1.466077,1.581515,-0.419414,
        -1.466077,1.584581,-2.738702,
        -1.466077,1.586948,-1.696296,
        -1.466077,1.588901,2.301496,
        -1.466077,1.589936,0.825020,
        -1.466077,1.590018,-1.941652,
        -1.466077,1.596968,-1.328358,
        -1.466077,1.597178,-1.575607,
        -1.466077,1.600470,-1.453196,
        -1.466077,1.601408,0.242409,
        -1.466077,1.602259,1.152656,
        -1.466077,1.603064,2.865774,
        -1.466077,1.612085,1.968802,
        -1.466077,1.612720,-3.080498,
        -1.466077,1.616061,2.521531,
        -1.466077,1.616900,-0.207457,
        -1.466077,1.620988,-0.875920,
        -1.466077,1.622114,0.701509,
        -1.466077,1.622749,0.121426,
        -1.466077,1.622993,-2.221629,
        -1.466077,1.629346,-0.542669,
        -1.466077,1.630007,1.844552,
        -1.466077,1.634747,-2.426198,
        -1.466077,1.642346,1.722617,
        -1.466077,1.643925,1.034629,
        -1.466077,1.644848,2.743774,
        -1.466077,1.645258,2.188109,
        -1.466077,1.645417,-2.634725,
        -1.466077,1.647395,3.085437,
        -1.466077,1.648538,-2.855044,
        -1.466077,1.649932,0.576808,
        -1.466077,1.652224,-0.996105,
        -1.466077,1.652871,1.600681,
        -1.466077,1.657997,-0.325127,
        -1.466077,1.661936,0.001037,
        -1.466077,1.665187,1.477402,
        -1.466077,1.665496,2.397872,
        -1.466077,1.673267,-0.665453,
        -1.466077,1.673557,0.452526,
        -1.466077,1.679277,-1.122421,
        -1.466077,1.679696,0.912482,
        -1.466077,1.682420,1.353922,
        -1.466077,1.685031,-2.009038,
        -1.466077,1.685196,2.967046,
        -1.466077,1.690107,2.620910,
        -1.466077,1.690453,-2.123076,
        -1.466077,1.691693,-1.871572,
        -1.466077,1.691817,2.061688,
        -1.466077,1.692644,1.237670,
        -1.466077,1.696386,0.328123,
        -1.466077,1.697162,-1.249026,
        -1.466077,1.697488,-2.322785,
        -1.466077,1.698726,-2.973576,
        -1.466077,1.699172,-1.749717,
        -1.466077,1.704478,-0.106295,
        -1.466077,1.707244,-0.444770,
        -1.466077,1.707500,-2.531497,
        -1.466077,1.707714,-0.787321,
        -1.466077,1.708430,-1.627552,
        -1.466077,1.710583,-1.376890,
        -1.466077,1.712415,-2.747921,
        -1.466077,1.713125,-1.503073,
        -1.466077,1.713452,0.789419,
        -1.466077,1.720612,2.279888,
        -1.466077,1.724852,0.202831,
        -1.466077,1.724868,2.846246,
        -1.466077,1.728940,1.931535,
        -1.466077,1.732838,1.127584,
        -1.466077,1.735719,-3.092485,
        -1.466077,1.739676,2.496017,
        -1.466077,1.739971,-0.228876,
        -1.466077,1.743328,0.663347,
        -1.466077,1.744927,-0.911892,
        -1.466077,1.745952,0.084733,
        -1.466077,1.748077,1.804354,
        -1.466077,1.752959,-0.567548,
        -1.466077,1.760169,1.678132,
        -1.466077,1.763887,-2.218450,
        -1.466077,1.766737,2.722752,
        -1.466077,1.767999,1.004737,
        -1.466077,1.769149,1.550298,
        -1.466077,1.770119,-2.427175,
        -1.466077,1.770460,2.154208,
        -1.466077,1.770514,3.070982,
        -1.466077,1.770735,0.537245,
        -1.466077,1.772673,-2.864760,
        -1.466077,1.773527,-1.039582,
        -1.466077,1.776419,-2.642421,
        -1.466077,1.780769,1.416330,
        -1.466077,1.786634,-0.346247,
        -1.466077,1.788829,-1.947553,
        -1.466077,1.793737,0.410270,
        -1.466077,1.794553,2.374618,
        -1.466077,1.794699,-0.694061,
        -1.466077,1.797469,-1.169105,
        -1.466077,1.803705,0.881076,
        -1.466077,1.804124,1.264249,
        -1.466077,1.807825,-2.090455,
        -1.466077,1.808822,-1.817375,
        -1.466077,1.809020,2.949741,
        -1.466077,1.812466,2.597595,
        -1.466077,1.813159,2.025395,
        -1.466077,1.813816,-1.298435,
        -1.466077,1.815016,0.278951,
        -1.466077,1.816510,-0.008631,
        -1.466077,1.817772,-1.689625,
        -1.466077,1.821360,-2.984352,
        -1.466077,1.822312,-0.127211,
        -1.466077,1.823386,-1.429807,
        -1.466077,1.823834,-1.560483,
        -1.466077,1.832060,-0.821627,
        -1.466077,1.832690,-0.468440,
        -1.466077,1.833945,-2.320278,
        -1.466077,1.834787,0.751968,
        -1.466077,1.840579,-2.536445,
        -1.466077,1.841033,-2.756932,
        -1.466077,1.846350,1.893983,
        -1.466077,1.846644,1.113766,
        -1.466077,1.846833,2.826543,
        -1.466077,1.847325,2.249051,
        -1.466077,1.854119,0.134161,
        -1.466077,1.858863,-3.106022,
        -1.466077,1.865046,1.763384,
        -1.466077,1.865619,-0.952036,
        -1.466077,1.865912,0.624453,
        -1.466077,1.866694,2.473686,
        -1.466077,1.867524,-0.240083,
        -1.466077,1.875561,1.632831,
        -1.466077,1.876049,-0.594378,
        -1.466077,1.880509,1.500204,
        -1.466077,1.883504,1.360457,
        -1.466077,1.886971,-2.196665,
        -1.466077,1.888979,0.982922,
        -1.466077,1.889815,2.701570,
        -1.466077,1.892144,-2.005340,
        -1.466077,1.892348,0.496522,
        -1.466077,1.893281,-1.083976,
        -1.466077,1.893828,3.053453,
        -1.466077,1.895117,2.121107,
        -1.466077,1.898741,-2.875735,
        -1.466077,1.904429,-2.428029,
        -1.466077,1.908408,-2.648903,
        -1.466077,1.911828,-0.364489,
        -1.466077,1.912224,0.366565,
        -1.466077,1.914854,1.214468,
        -1.466077,1.915467,-1.883143,
        -1.466077,1.915763,-0.724679,
        -1.466077,1.915790,-1.216366,
        -1.466077,1.922256,2.347806,
        -1.466077,1.922616,0.844633,
        -1.466077,1.925185,0.231170,
        -1.466077,1.927474,-1.754616,
        -1.466077,1.928342,-1.351174,
        -1.466077,1.933328,-1.623328,
        -1.466077,1.933878,0.034150,
        -1.466077,1.934585,-1.487979,
        -1.466077,1.934802,2.927634,
        -1.466077,1.934862,1.989776,
        -1.466077,1.936327,-0.099350,
        -1.466077,1.938088,2.578507,
        -1.466077,1.945677,-2.999805,
        -1.466077,1.954572,-0.859208,
        -1.466077,1.957868,-0.490784,
        -1.466077,1.958989,0.712772,
        -1.466077,1.960212,-2.307561,
        -1.466077,1.964774,1.855507,
        -1.466077,1.964971,-2.100533,
        -1.466077,1.969112,1.088273,
        -1.466077,1.971438,-2.765415,
        -1.466077,1.972312,2.805134,
        -1.466077,1.974950,2.220596,
        -1.466077,1.975135,-2.539127,
        -1.466077,1.982107,1.720829,
        -1.466077,1.983195,-3.127984,
        -1.466077,1.983318,-0.247182,
        -1.466077,1.986720,-0.993220,
        -1.466077,1.990381,1.586005,
        -1.466077,1.990470,0.580283,
        -1.466077,1.990523,1.450326,
        -1.466077,1.993909,1.312897,
        -1.466077,1.994694,2.452813,
        -1.466077,1.999226,-0.620717,
        -1.466077,2.005145,0.941012,
        -1.466077,2.007301,0.142337,
        -1.466077,2.014149,0.444138,
        -1.466077,2.014338,-1.127392,
        -1.466077,2.018131,3.021778,
        -1.466077,2.019676,-1.980944,
        -1.466077,2.019912,2.686612,
        -1.466077,2.020545,2.088672,
        -1.466077,2.026074,0.300945,
        -1.466077,2.026196,-2.889057,
        -1.466077,2.031959,-1.268178,
        -1.466077,2.032917,-2.199917,
        -1.466077,2.033572,-1.842884,
        -1.466077,2.034712,-2.419505,
        -1.466077,2.035818,-0.379479,
        -1.466077,2.036004,-0.757062,
        -1.466077,2.040501,-1.411488,
        -1.466077,2.042874,-1.701367,
        -1.466077,2.043060,-2.653746,
        -1.466077,2.043729,-1.557938,
        -1.466077,2.048538,0.806623,
        -1.466077,2.049006,1.199930,
        -1.466077,2.049502,-0.137989,
        -1.466077,2.052181,2.326959,
        -1.466077,2.052794,-0.008526,
        -1.466077,2.057608,1.951726,
        -1.466077,2.065568,2.567730,
        -1.466077,2.070371,2.879808,
        -1.466077,2.071256,-3.018961,
        -1.466077,2.079746,-0.897699,
        -1.466077,2.082784,1.042403,
        -1.466077,2.083752,-0.511086,
        -1.466077,2.084401,1.810502,
        -1.466077,2.085731,0.671081,
        -1.466077,2.097496,1.373172,
        -1.466077,2.099185,1.667006,
        -1.466077,2.100482,-2.301084,
        -1.466077,2.103118,2.194745,
        -1.466077,2.103223,1.521685,
        -1.466077,2.103695,-2.777707,
        -1.466077,2.107523,3.129581,
        -1.466077,2.108277,-2.095907,
        -1.466077,2.108891,-2.532166,
        -1.466077,2.111003,-0.268632,
        -1.466077,2.112276,-1.027246,
        -1.466077,2.114380,0.531457,
        -1.466077,2.119166,0.100552,
        -1.466077,2.119564,2.754818,
        -1.466077,2.122474,-0.648314,
        -1.466077,2.125183,2.445598,
        -1.466077,2.129501,0.238290,
        -1.466077,2.131178,0.386663,
        -1.466077,2.131223,-1.174311,
        -1.466077,2.134269,-1.948042,
        -1.466077,2.134684,0.907065,
        -1.466077,2.137615,2.988806,
        -1.466077,2.145635,-1.327708,
        -1.466077,2.146467,2.055081,
        -1.466077,2.146582,-1.793788,
        -1.466077,2.149210,-1.482321,
        -1.466077,2.152901,-1.636788,
        -1.466077,2.156337,1.151558,
        -1.466077,2.156474,-2.912065,
        -1.466077,2.157194,-0.792563,
        -1.466077,2.166724,-0.401638,
        -1.466077,2.166980,-0.134879,
        -1.466077,2.173644,-2.403429,
        -1.466077,2.175711,-2.655523,
        -1.466077,2.177132,2.600419,
        -1.466077,2.178879,0.767722,
        -1.466077,2.180752,1.908485,
        -1.466077,2.181338,2.313721,
        -1.466077,2.185580,1.287673,
        -1.466077,2.195400,-2.217546,
        -1.466077,2.200418,-3.056442,
        -1.466077,2.201829,-0.001735,
        -1.466077,2.203804,1.756176,
        -1.466077,2.211602,1.443930,
        -1.466077,2.212686,-0.544809,
        -1.466077,2.213937,0.621892,
        -1.466077,2.214075,1.600691,
        -1.466077,2.216256,1.020029,
        -1.466077,2.219060,2.886631,
        -1.466077,2.223375,-0.906185,
        -1.466077,2.226002,-1.060248,
        -1.466077,2.226413,-2.064034,
        -1.466077,2.230554,2.171334,
        -1.466077,2.232413,2.735777,
        -1.466077,2.233709,3.070830,
        -1.466077,2.234383,-2.792065,
        -1.466077,2.237183,0.468799,
        -1.466077,2.237695,-0.271388,
        -1.466077,2.239419,0.151693,
        -1.466077,2.243347,-2.521049,
        -1.466077,2.244035,-1.229456,
        -1.466077,2.244355,0.309899,
        -1.466077,2.245175,2.460236,
        -1.466077,2.245628,-1.899216,
        -1.466077,2.245973,-0.702234,
        -1.466077,2.255601,-1.728131,
        -1.466077,2.256702,-1.401367,
        -1.466077,2.261466,-1.557500,
        -1.466077,2.267390,0.878499,
        -1.466077,2.272099,2.018594,
        -1.466077,2.278618,1.170386,
        -1.466077,2.283304,-2.345584,
        -1.466077,2.285175,-2.943090,
        -1.466077,2.293000,-0.124451,
        -1.466077,2.294058,-0.421070,
        -1.466077,2.303283,1.855938,
        -1.466077,2.304968,2.608978,
        -1.466077,2.305347,2.311346,
        -1.466077,2.307188,-2.656134,
        -1.466077,2.307301,1.336200,
        -1.466077,2.309655,0.726020,
        -1.466077,2.317634,-2.185225,
        -1.466077,2.321510,1.685664,
        -1.466077,2.324376,-3.111006,
        -1.466077,2.324731,1.511186,
        -1.466077,2.329587,0.035498,
        -1.466077,2.330158,2.977264,
        -1.466077,2.335096,-0.588468,
        -1.466077,2.336046,-1.115799,
        -1.466077,2.338060,2.803613,
        -1.466077,2.338083,-0.943400,
        -1.466077,2.340210,0.562145,
        -1.466077,2.341931,1.023926,
        -1.466077,2.343265,-0.779061,
        -1.466077,2.343434,-2.013692,
        -1.466077,2.353029,-1.310529,
        -1.466077,2.354868,0.211543,
        -1.466077,2.355120,0.388593,
        -1.466077,2.355165,-1.833173,
        -1.466077,2.355462,2.151441,
        -1.466077,2.362948,-2.487477,
        -1.466077,2.363101,-2.810456,
        -1.466077,2.363394,-0.272918,
        -1.466077,2.364509,-1.644533,
        -1.466077,2.370465,2.466824,
        -1.466077,2.386110,-1.481351,
        -1.466077,2.394960,0.862133,
        -1.466077,2.395716,1.977184,
        -1.466077,2.396339,1.200900,
        -1.466077,2.407941,-2.316753,
        -1.466077,2.409450,-2.988432,
        -1.466077,2.417202,-0.103363,
        -1.466077,2.417386,-0.443097,
        -1.466077,2.417950,2.666662,
        -1.466077,2.423190,1.789783,
        -1.466077,2.423692,1.392696,
        -1.466077,2.431122,3.086379,
        -1.466077,2.431406,-2.647820,
        -1.466077,2.431584,2.307631,
        -1.466077,2.434090,1.592818,
        -1.466077,2.435993,0.683844,
        -1.466077,2.440898,-2.133300,
        -1.466077,2.442222,-1.195837,
        -1.466077,2.443112,2.883761,
        -1.466077,2.449191,-0.633651,
        -1.466077,2.450821,0.084741,
        -1.466077,2.453499,-1.008861,
        -1.466077,2.460010,-1.938011,
        -1.466077,2.460199,-0.822008,
        -1.466077,2.460902,0.490079,
        -1.466077,2.461774,-1.734913,
        -1.466077,2.463986,1.034407,
        -1.466077,2.467647,0.287667,
        -1.466077,2.479849,2.126925,
        -1.466077,2.485907,-2.835471,
        -1.466077,2.485945,-1.374853,
        -1.466077,2.487747,2.505912,
        -1.466077,2.487957,-0.274279,
        -1.466077,2.489741,-2.468428,
        -1.466077,2.512538,1.239900,
        -1.466077,2.517528,1.923289,
        -1.466077,2.518454,-1.565852,
        -1.466077,2.519348,0.843503,
        -1.466077,2.522766,-3.059231,
        -1.466077,2.528686,2.737228,
        -1.466077,2.534798,-2.266671,
        -1.466077,2.536533,1.466274,
        -1.466077,2.538504,1.699812,
        -1.466077,2.539516,-0.476942,
        -1.466077,2.539532,-0.074548,
        -1.466077,2.543197,2.990252,
        -1.466077,2.554022,2.320619,
        -1.466077,2.556919,-2.655005,
        -1.466077,2.558146,0.627191,
        -1.466077,2.558223,-1.164189,
        -1.466077,2.563452,-2.044028,
        -1.466077,2.568462,0.150655,
        -1.466077,2.569130,-0.700837,
        -1.466077,2.570050,-1.804492,
        -1.466077,2.572710,-0.934665,
        -1.466077,2.575994,0.391718,
        -1.466077,2.584774,1.046703,
        -1.466077,2.601432,2.102535,
        -1.466077,2.603559,2.555114,
        -1.466077,2.604150,-2.885208,
        -1.466077,2.608630,-1.378011,
        -1.466077,2.611256,-0.279110,
        -1.466077,2.617344,-2.439480,
        -1.466077,2.627402,1.293552,
        -1.466077,2.635396,1.848850,
        -1.466077,2.635686,2.834950,
        -1.466077,2.637087,3.132730,
        -1.466077,2.642029,0.813274,
        -1.466077,2.645260,1.570203,
        -1.466077,2.646182,-1.614036,
        -1.466077,2.660113,-2.186513,
        -1.466077,2.660483,-0.034608,
        -1.466077,2.661713,-0.529511,
        -1.466077,2.672720,-1.110422,
        -1.466077,2.676626,2.335353,
        -1.466077,2.676638,0.540364,
        -1.466077,2.676780,-1.901158,
        -1.466077,2.681411,-2.667965,
        -1.466077,2.682152,0.245751,
        -1.466077,2.682859,-0.817071,
        -1.466077,2.705107,1.062142,
        -1.466077,2.717705,2.625995,
        -1.466077,2.721001,2.060097,
        -1.466077,2.722319,-2.954069,
        -1.466077,2.729840,-1.368237,
        -1.466077,2.734349,-0.289298,
        -1.466077,2.738409,2.975671,
        -1.466077,2.740366,1.374341,
        -1.466077,2.744633,-2.391311,
        -1.466077,2.747398,1.727166,
        -1.466077,2.763799,0.760372,
        -1.466077,2.768238,-1.679243,
        -1.466077,2.779245,-2.046268,
        -1.466077,2.780208,-0.626453,
        -1.466077,2.780471,0.027732,
        -1.466077,2.784696,-1.013128,
        -1.466077,2.789817,0.395431,
        -1.466077,2.801154,2.353143,
        -1.466077,2.807091,-2.687477,
        -1.466077,2.825762,1.085035,
        -1.466077,2.830209,2.734943,
        -1.466077,2.833517,-3.085508,
        -1.466077,2.838301,1.980061,
        -1.466077,2.849871,1.512012,
        -1.466077,2.850089,-1.343899,
        -1.466077,2.856905,-0.318441,
        -1.466077,2.870445,-2.291001,
        -1.466077,2.885017,0.655562,
        -1.466077,2.886428,-1.793217,
        -1.466077,2.890337,-0.825633,
        -1.466077,2.898932,0.139973,
        -1.466077,2.931927,-2.728768,
        -1.466077,2.932096,2.377790,
        -1.466077,2.937409,2.947245,
        -1.466077,2.948048,1.133127,
        -1.466077,2.950896,1.797889,
        -1.466077,2.970019,-1.282382,
        -1.466077,2.978333,-0.412090,
        -1.466077,2.996437,-2.089121,
        -1.466077,3.011274,0.434110,
        -1.466077,3.050410,3.138492,
        -1.466077,3.064492,1.647259,
        -1.466077,3.088868,-0.989501,
        -1.361357,0.053884,-2.408394,
        -1.361357,0.077645,1.453433,
        -1.361357,0.088408,-0.311160,
        -1.361357,0.142314,2.590387,
        -1.361357,0.147206,-1.385164,
        -1.361357,0.163506,0.573203,
        -1.361357,0.165485,-2.875396,
        -1.361357,0.182564,-2.125821,
        -1.361357,0.190569,1.929141,
        -1.361357,0.201102,1.249082,
        -1.361357,0.201397,-0.716646,
        -1.361357,0.210568,-0.069559,
        -1.361357,0.251672,2.848900,
        -1.361357,0.266262,-1.678068,
        -1.361357,0.269108,-1.215563,
        -1.361357,0.273777,-2.506756,
        -1.361357,0.279854,0.814581,
        -1.361357,0.281664,-2.969147,
        -1.361357,0.283625,2.405603,
        -1.361357,0.284058,0.357350,
        -1.361357,0.294107,1.593880,
        -1.361357,0.302491,-0.400410,
        -1.361357,0.308059,2.021491,
        -1.361357,0.312219,-2.072057,
        -1.361357,0.321905,-0.829121,
        -1.361357,0.329984,1.201621,
        -1.361357,0.339714,-0.013712,
        -1.361357,0.368784,2.974974,
        -1.361357,0.376428,-1.495532,
        -1.361357,0.378103,2.646436,
        -1.361357,0.385088,0.590355,
        -1.361357,0.386248,-2.690943,
        -1.361357,0.389851,-2.340031,
        -1.361357,0.391502,-1.171374,
        -1.361357,0.398249,-1.812694,
        -1.361357,0.400542,-2.995729,
        -1.361357,0.404273,0.918163,
        -1.361357,0.405617,1.773533,
        -1.361357,0.409009,-0.570243,
        -1.361357,0.410700,0.270433,
        -1.361357,0.415893,1.464890,
        -1.361357,0.420680,2.088164,
        -1.361357,0.424766,-0.258729,
        -1.361357,0.426507,2.367872,
        -1.361357,0.441292,-0.890727,
        -1.361357,0.451839,-2.078193,
        -1.361357,0.459270,1.191562,
        -1.361357,0.470622,0.014002,
        -1.361357,0.481292,-2.529098,
        -1.361357,0.482460,2.849843,
        -1.361357,0.492205,-1.397963,
        -1.361357,0.493314,3.117264,
        -1.361357,0.497870,0.727828,
        -1.361357,0.499783,-1.654379,
        -1.361357,0.502651,0.463333,
        -1.361357,0.505463,2.594557,
        -1.361357,0.506104,-2.874910,
        -1.361357,0.513579,-1.147354,
        -1.361357,0.518569,1.644237,
        -1.361357,0.519001,1.900488,
        -1.361357,0.519277,-0.673809,
        -1.361357,0.524769,-0.421955,
        -1.361357,0.525564,-2.288838,
        -1.361357,0.532541,0.978737,
        -1.361357,0.534610,2.154808,
        -1.361357,0.535020,-1.891846,
        -1.361357,0.540208,1.400497,
        -1.361357,0.542006,0.228337,
        -1.361357,0.550762,-0.180771,
        -1.361357,0.560531,2.378474,
        -1.361357,0.561629,-0.924576,
        -1.361357,0.567465,-2.679392,
        -1.361357,0.588527,1.181755,
        -1.361357,0.591643,-2.098364,
        -1.361357,0.594963,2.997575,
        -1.361357,0.599643,-3.045235,
        -1.361357,0.599962,2.773556,
        -1.361357,0.602620,0.029270,
        -1.361357,0.605711,0.598454,
        -1.361357,0.607708,-1.541026,
        -1.361357,0.609360,-2.464379,
        -1.361357,0.610738,-1.326786,
        -1.361357,0.616968,0.820489,
        -1.361357,0.626379,1.773610,
        -1.361357,0.628626,-1.751180,
        -1.361357,0.630671,0.391949,
        -1.361357,0.632622,-0.737642,
        -1.361357,0.632708,2.566722,
        -1.361357,0.632797,1.989858,
        -1.361357,0.634455,-0.534790,
        -1.361357,0.636205,1.561170,
        -1.361357,0.637681,-1.123423,
        -1.361357,0.641441,-2.842109,
        -1.361357,0.644807,-0.327139,
        -1.361357,0.652440,2.197129,
        -1.361357,0.660897,-2.278593,
        -1.361357,0.664008,1.360091,
        -1.361357,0.670850,-1.945257,
        -1.361357,0.671396,1.004859,
        -1.361357,0.673985,0.203855,
        -1.361357,0.680326,-0.133370,
        -1.361357,0.685083,-0.942363,
        -1.361357,0.688204,-2.632204,
        -1.361357,0.689777,2.385011,
        -1.361357,0.700065,3.115430,
        -1.361357,0.704080,2.911636,
        -1.361357,0.712883,1.174216,
        -1.361357,0.715701,0.699906,
        -1.361357,0.718851,-1.442215,
        -1.361357,0.721019,2.721536,
        -1.361357,0.725300,-2.975134,
        -1.361357,0.727136,-1.633288,
        -1.361357,0.727399,-2.120567,
        -1.361357,0.730751,0.522488,
        -1.361357,0.732892,-1.254330,
        -1.361357,0.734333,-2.446184,
        -1.361357,0.735125,0.037270,
        -1.361357,0.737893,1.866463,
        -1.361357,0.740650,1.681075,
        -1.361357,0.745971,-0.781286,
        -1.361357,0.749432,-0.442559,
        -1.361357,0.750102,2.050636,
        -1.361357,0.751046,0.863821,
        -1.361357,0.757018,-0.608833,
        -1.361357,0.757046,-1.816019,
        -1.361357,0.757389,1.501184,
        -1.361357,0.759800,2.546181,
        -1.361357,0.759859,0.351923,
        -1.361357,0.761141,-2.781102,
        -1.361357,0.769035,-1.078574,
        -1.361357,0.770166,-0.266604,
        -1.361357,0.772218,2.227380,
        -1.361357,0.788382,1.331051,
        -1.361357,0.790269,-2.282409,
        -1.361357,0.803410,-1.984567,
        -1.361357,0.803813,0.191520,
        -1.361357,0.806832,3.022155,
        -1.361357,0.807476,-2.597675,
        -1.361357,0.812512,-0.107335,
        -1.361357,0.814161,1.010949,
        -1.361357,0.817732,-3.084486,
        -1.361357,0.817846,2.390688,
        -1.361357,0.819323,2.845960,
        -1.361357,0.819430,-0.915017,
        -1.361357,0.829198,-1.532491,
        -1.361357,0.829531,0.665695,
        -1.361357,0.829816,1.167147,
        -1.361357,0.833613,-1.357492,
        -1.361357,0.842945,-2.907318,
        -1.361357,0.844883,2.679708,
        -1.361357,0.848189,-1.705003,
        -1.361357,0.849888,1.767901,
        -1.361357,0.851126,0.493370,
        -1.361357,0.852911,1.933394,
        -1.361357,0.857516,-2.432863,
        -1.361357,0.858261,-1.191729,
        -1.361357,0.858279,-2.138895,
        -1.361357,0.859723,-0.734928,
        -1.361357,0.859754,1.603292,
        -1.361357,0.861655,-0.394821,
        -1.361357,0.864300,0.043549,
        -1.361357,0.866940,0.851229,
        -1.361357,0.869748,2.095969,
        -1.361357,0.877352,-0.574668,
        -1.361357,0.879254,-2.733338,
        -1.361357,0.882554,1.443757,
        -1.361357,0.884060,0.334199,
        -1.361357,0.885626,-1.866061,
        -1.361357,0.889834,2.529957,
        -1.361357,0.893140,2.251698,
        -1.361357,0.897463,-0.240822,
        -1.361357,0.898758,-1.031948,
        -1.361357,0.913918,1.287618,
        -1.361357,0.916555,-2.287173,
        -1.361357,0.918305,3.107845,
        -1.361357,0.919522,2.948277,
        -1.361357,0.927138,-2.569686,
        -1.361357,0.929031,0.185383,
        -1.361357,0.933766,-3.015917,
        -1.361357,0.934873,-2.010822,
        -1.361357,0.935636,-1.445855,
        -1.361357,0.936702,-0.866461,
        -1.361357,0.937395,0.729108,
        -1.361357,0.937910,2.787575,
        -1.361357,0.939144,-0.092962,
        -1.361357,0.941718,1.125085,
        -1.361357,0.942873,-1.607680,
        -1.361357,0.943701,0.980328,
        -1.361357,0.947774,2.395256,
        -1.361357,0.952433,0.584347,
        -1.361357,0.954456,-1.287293,
        -1.361357,0.958731,-2.852552,
        -1.361357,0.961838,1.839284,
        -1.361357,0.961898,-0.469425,
        -1.361357,0.965426,1.686615,
        -1.361357,0.970108,1.990771,
        -1.361357,0.972820,-1.762780,
        -1.361357,0.973500,-0.711881,
        -1.361357,0.975253,2.641082,
        -1.361357,0.977824,0.441928,
        -1.361357,0.981485,-2.424766,
        -1.361357,0.981877,1.537100,
        -1.361357,0.982609,-1.136920,
        -1.361357,0.984265,-2.156073,
        -1.361357,0.989150,0.045493,
        -1.361357,0.993052,2.137478,
        -1.361357,0.997766,-2.693871,
        -1.361357,0.999847,-0.334738,
        -1.361357,1.001920,0.848054,
        -1.361357,1.006146,1.389734,
        -1.361357,1.012180,0.304066,
        -1.361357,1.016938,2.270674,
        -1.361357,1.017371,-0.979790,
        -1.361357,1.019495,-1.894487,
        -1.361357,1.023796,-0.583691,
        -1.361357,1.026525,2.506945,
        -1.361357,1.027004,-0.203653,
        -1.361357,1.028027,3.036573,
        -1.361357,1.031435,1.241021,
        -1.361357,1.032316,2.884221,
        -1.361357,1.035863,-3.110475,
        -1.361357,1.042085,-2.292950,
        -1.361357,1.043272,-1.522986,
        -1.361357,1.046265,-2.958847,
        -1.361357,1.049272,-1.376344,
        -1.361357,1.050039,-2.552086,
        -1.361357,1.053682,0.689863,
        -1.361357,1.055458,0.170394,
        -1.361357,1.056622,-0.831866,
        -1.361357,1.057515,1.097998,
        -1.361357,1.058706,-2.036438,
        -1.361357,1.061389,-1.672248,
        -1.361357,1.063758,2.743093,
        -1.361357,1.070157,-0.073422,
        -1.361357,1.072154,1.761506,
        -1.361357,1.074029,1.909193,
        -1.361357,1.074204,0.545276,
        -1.361357,1.074714,-2.805147,
        -1.361357,1.076262,0.966104,
        -1.361357,1.077642,-1.227073,
        -1.361357,1.084151,1.618525,
        -1.361357,1.085085,2.377673,
        -1.361357,1.086822,-0.441112,
        -1.361357,1.087142,2.058815,
        -1.361357,1.098383,-0.696430,
        -1.361357,1.099138,-1.084033,
        -1.361357,1.103208,1.477773,
        -1.361357,1.103783,0.408254,
        -1.361357,1.106275,-2.420800,
        -1.361357,1.108079,2.612943,
        -1.361357,1.109345,-2.172718,
        -1.361357,1.115957,-1.787597,
        -1.361357,1.118930,0.045983,
        -1.361357,1.119777,0.834702,
        -1.361357,1.121943,-2.669042,
        -1.361357,1.124930,2.219917,
        -1.361357,1.125217,1.336294,
        -1.361357,1.130958,2.972931,
        -1.361357,1.131400,-0.306629,
        -1.361357,1.137135,-0.936514,
        -1.361357,1.139125,-1.924528,
        -1.361357,1.139877,0.276382,
        -1.361357,1.142003,-3.058022,
        -1.361357,1.147363,-0.558448,
        -1.361357,1.150348,-0.176418,
        -1.361357,1.151198,1.197016,
        -1.361357,1.153930,-1.597104,
        -1.361357,1.155389,2.837148,
        -1.361357,1.155458,-1.460623,
        -1.361357,1.157331,-2.906362,
        -1.361357,1.158594,2.487618,
        -1.361357,1.162103,-1.322553,
        -1.361357,1.165407,3.092861,
        -1.361357,1.168025,0.717075,
        -1.361357,1.168406,-2.301519,
        -1.361357,1.174782,-2.538755,
        -1.361357,1.177998,1.061256,
        -1.361357,1.178898,1.827305,
        -1.361357,1.181331,-2.060330,
        -1.361357,1.182540,-0.798952,
        -1.361357,1.184569,0.147517,
        -1.361357,1.184734,1.968919,
        -1.361357,1.184804,0.593386,
        -1.361357,1.188278,1.690808,
        -1.361357,1.194360,2.110440,
        -1.361357,1.194776,2.710455,
        -1.361357,1.197745,-1.173705,
        -1.361357,1.199898,-2.777257,
        -1.361357,1.201164,2.355931,
        -1.361357,1.203760,1.556404,
        -1.361357,1.205845,-0.415993,
        -1.361357,1.206585,0.933648,
        -1.361357,1.206950,-0.037248,
        -1.361357,1.209341,0.471162,
        -1.361357,1.213483,-1.039417,
        -1.361357,1.218146,-0.672643,
        -1.361357,1.221608,1.421669,
        -1.361357,1.226043,-1.693533,
        -1.361357,1.232870,-1.822773,
        -1.361357,1.234375,-2.421194,
        -1.361357,1.234531,-2.190148,
        -1.361357,1.239414,0.348959,
        -1.361357,1.240183,2.587255,
        -1.361357,1.242312,2.941234,
        -1.361357,1.244075,1.286644,
        -1.361357,1.245961,-2.997461,
        -1.361357,1.246495,-2.646872,
        -1.361357,1.247736,2.242810,
        -1.361357,1.253604,0.812222,
        -1.361357,1.257816,-0.894317,
        -1.361357,1.258247,-1.953634,
        -1.361357,1.258993,-0.284099,
        -1.361357,1.262380,-0.157609,
        -1.361357,1.262791,-1.549229,
        -1.361357,1.263772,-0.533073,
        -1.361357,1.266731,-1.412461,
        -1.361357,1.268591,-3.122146,
        -1.361357,1.269614,-1.275246,
        -1.361357,1.270745,1.152533,
        -1.361357,1.273606,0.224545,
        -1.361357,1.279716,2.812101,
        -1.361357,1.283786,2.462124,
        -1.361357,1.284798,0.078209,
        -1.361357,1.285402,-2.882049,
        -1.361357,1.286936,1.886053,
        -1.361357,1.289408,0.683122,
        -1.361357,1.294736,1.756184,
        -1.361357,1.296978,-2.311511,
        -1.361357,1.298546,2.020601,
        -1.361357,1.300924,1.022478,
        -1.361357,1.303305,3.043152,
        -1.361357,1.304210,-2.084225,
        -1.361357,1.305241,-2.530249,
        -1.361357,1.307513,1.628003,
        -1.361357,1.310722,0.558267,
        -1.361357,1.314142,2.139329,
        -1.361357,1.316087,-0.777971,
        -1.361357,1.317907,-1.121023,
        -1.361357,1.321019,-2.749776,
        -1.361357,1.321609,2.687717,
        -1.361357,1.322568,1.499415,
        -1.361357,1.322972,-0.395699,
        -1.361357,1.325286,-0.994192,
        -1.361357,1.329146,-0.648513,
        -1.361357,1.329795,-0.042570,
        -1.361357,1.331651,2.344497,
        -1.361357,1.336522,0.435098,
        -1.361357,1.339410,0.900087,
        -1.361357,1.339550,1.370353,
        -1.361357,1.340441,-1.731580,
        -1.361357,1.351855,-1.855234,
        -1.361357,1.360161,-3.018404,
        -1.361357,1.361031,-1.613971,
        -1.361357,1.361196,2.914780,
        -1.361357,1.361998,-2.207254,
        -1.361357,1.362302,1.239384,
        -1.361357,1.365228,2.563808,
        -1.361357,1.365452,0.310522,
        -1.361357,1.367028,-2.420041,
        -1.361357,1.373422,-1.484281,
        -1.361357,1.374245,-1.226417,
        -1.361357,1.375020,-0.170393,
        -1.361357,1.376560,-1.355481,
        -1.361357,1.377070,-1.980984,
        -1.361357,1.377578,-2.636339,
        -1.361357,1.378096,0.774626,
        -1.361357,1.379086,-0.508192,
        -1.361357,1.382562,0.177141,
        -1.361357,1.388849,2.233847,
        -1.361357,1.389268,-0.886349,
        -1.361357,1.391324,1.109396,
        -1.361357,1.391699,-0.288725,
        -1.361357,1.392869,3.136903,
        -1.361357,1.393612,1.952016,
        -1.361357,1.395545,-2.850807,
        -1.361357,1.401926,1.822930,
        -1.361357,1.402149,2.788898,
        -1.361357,1.409552,0.647980,
        -1.361357,1.410840,0.056219,
        -1.361357,1.411944,2.444377,
        -1.361357,1.413639,1.697982,
        -1.361357,1.426479,1.573427,
        -1.361357,1.426725,2.102242,
        -1.361357,1.427868,0.984637,
        -1.361357,1.429580,-2.107232,
        -1.361357,1.431375,-2.313269,
        -1.361357,1.431907,3.015410,
        -1.361357,1.432325,-1.029054,
        -1.361357,1.433341,0.524268,
        -1.361357,1.435181,-0.739771,
        -1.361357,1.437208,-2.529085,
        -1.361357,1.439041,-0.617966,
        -1.361357,1.441473,1.448544,
        -1.361357,1.444527,2.664744,
        -1.361357,1.450513,-0.400625,
        -1.361357,1.451907,-0.066823,
        -1.361357,1.452754,-2.738631,
        -1.361357,1.454061,-2.950416,
        -1.361357,1.456736,1.323154,
        -1.361357,1.458370,0.400433,
        -1.361357,1.459314,-1.765003,
        -1.361357,1.459902,-1.145835,
        -1.361357,1.464952,2.329245,
        -1.361357,1.466050,0.862104,
        -1.361357,1.471330,-1.886160,
        -1.361357,1.475119,-1.648255,
        -1.361357,1.480878,1.194576,
        -1.361357,1.480942,0.273235,
        -1.361357,1.481086,2.890269,
        -1.361357,1.484118,-1.529156,
        -1.361357,1.484741,-1.279219,
        -1.361357,1.487681,-1.406069,
        -1.361357,1.487886,-3.066037,
        -1.361357,1.490555,2.543284,
        -1.361357,1.492424,-0.190932,
        -1.361357,1.495002,-0.839810,
        -1.361357,1.495297,2.007188,
        -1.361357,1.496414,-2.009267,
        -1.361357,1.498261,-2.425270,
        -1.361357,1.498907,-2.207327,
        -1.361357,1.500204,0.737638,
        -1.361357,1.502469,0.151156,
        -1.361357,1.507448,-0.514782,
        -1.361357,1.509287,2.204647,
        -1.361357,1.510518,1.884920,
        -1.361357,1.511948,-2.632360,
        -1.361357,1.515235,1.068565,
        -1.361357,1.522237,3.102018,
        -1.361357,1.522373,1.764991,
        -1.361357,1.522416,-0.954956,
        -1.361357,1.523102,2.765920,
        -1.361357,1.524811,-2.844718,
        -1.361357,1.527261,-0.304916,
        -1.361357,1.529788,0.613208,
        -1.361357,1.533648,1.644808,
        -1.361357,1.535018,0.028869,
        -1.361357,1.541082,2.424608,
        -1.361357,1.546010,1.523462,
        -1.361357,1.551135,-0.642994,
        -1.361357,1.553895,0.489607,
        -1.361357,1.555041,0.947813,
        -1.361357,1.559671,-1.074934,
        -1.361357,1.560345,2.985899,
        -1.361357,1.560487,-2.323400,
        -1.361357,1.561127,1.401111,
        -1.361357,1.567193,2.643172,
        -1.361357,1.568892,2.098362,
        -1.361357,1.571702,-2.098435,
        -1.361357,1.572701,-2.528730,
        -1.361357,1.573572,1.278593,
        -1.361357,1.574849,-0.091505,
        -1.361357,1.576334,0.365236,
        -1.361357,1.577029,-2.962888,
        -1.361357,1.577688,-1.814941,
        -1.361357,1.579542,-0.755917,
        -1.361357,1.579624,-1.199759,
        -1.361357,1.581515,-0.419414,
        -1.361357,1.584581,-2.738702,
        -1.361357,1.586948,-1.696296,
        -1.361357,1.588901,2.301496,
        -1.361357,1.589936,0.825020,
        -1.361357,1.590018,-1.941652,
        -1.361357,1.596968,-1.328358,
        -1.361357,1.597178,-1.575607,
        -1.361357,1.600470,-1.453196,
        -1.361357,1.601408,0.242409,
        -1.361357,1.602259,1.152656,
        -1.361357,1.603064,2.865774,
        -1.361357,1.612085,1.968802,
        -1.361357,1.612720,-3.080498,
        -1.361357,1.616061,2.521531,
        -1.361357,1.616900,-0.207457,
        -1.361357,1.620988,-0.875920,
        -1.361357,1.622114,0.701509,
        -1.361357,1.622749,0.121426,
        -1.361357,1.622993,-2.221629,
        -1.361357,1.629346,-0.542669,
        -1.361357,1.630007,1.844552,
        -1.361357,1.634747,-2.426198,
        -1.361357,1.642346,1.722617,
        -1.361357,1.643925,1.034629,
        -1.361357,1.644848,2.743774,
        -1.361357,1.645258,2.188109,
        -1.361357,1.645417,-2.634725,
        -1.361357,1.647395,3.085437,
        -1.361357,1.648538,-2.855044,
        -1.361357,1.649932,0.576808,
        -1.361357,1.652224,-0.996105,
        -1.361357,1.652871,1.600681,
        -1.361357,1.657997,-0.325127,
        -1.361357,1.661936,0.001037,
        -1.361357,1.665187,1.477402,
        -1.361357,1.665496,2.397872,
        -1.361357,1.673267,-0.665453,
        -1.361357,1.673557,0.452526,
        -1.361357,1.679277,-1.122421,
        -1.361357,1.679696,0.912482,
        -1.361357,1.682420,1.353922,
        -1.361357,1.685031,-2.009038,
        -1.361357,1.685196,2.967046,
        -1.361357,1.690107,2.620910,
        -1.361357,1.690453,-2.123076,
        -1.361357,1.691693,-1.871572,
        -1.361357,1.691817,2.061688,
        -1.361357,1.692644,1.237670,
        -1.361357,1.696386,0.328123,
        -1.361357,1.697162,-1.249026,
        -1.361357,1.697488,-2.322785,
        -1.361357,1.698726,-2.973576,
        -1.361357,1.699172,-1.749717,
        -1.361357,1.704478,-0.106295,
        -1.361357,1.707244,-0.444770,
        -1.361357,1.707500,-2.531497,
        -1.361357,1.707714,-0.787321,
        -1.361357,1.708430,-1.627552,
        -1.361357,1.710583,-1.376890,
        -1.361357,1.712415,-2.747921,
        -1.361357,1.713125,-1.503073,
        -1.361357,1.713452,0.789419,
        -1.361357,1.720612,2.279888,
        -1.361357,1.724852,0.202831,
        -1.361357,1.724868,2.846246,
        -1.361357,1.728940,1.931535,
        -1.361357,1.732838,1.127584,
        -1.361357,1.735719,-3.092485,
        -1.361357,1.739676,2.496017,
        -1.361357,1.739971,-0.228876,
        -1.361357,1.743328,0.663347,
        -1.361357,1.744927,-0.911892,
        -1.361357,1.745952,0.084733,
        -1.361357,1.748077,1.804354,
        -1.361357,1.752959,-0.567548,
        -1.361357,1.760169,1.678132,
        -1.361357,1.763887,-2.218450,
        -1.361357,1.766737,2.722752,
        -1.361357,1.767999,1.004737,
        -1.361357,1.769149,1.550298,
        -1.361357,1.770119,-2.427175,
        -1.361357,1.770460,2.154208,
        -1.361357,1.770514,3.070982,
        -1.361357,1.770735,0.537245,
        -1.361357,1.772673,-2.864760,
        -1.361357,1.773527,-1.039582,
        -1.361357,1.776419,-2.642421,
        -1.361357,1.780769,1.416330,
        -1.361357,1.786634,-0.346247,
        -1.361357,1.788829,-1.947553,
        -1.361357,1.793737,0.410270,
        -1.361357,1.794553,2.374618,
        -1.361357,1.794699,-0.694061,
        -1.361357,1.797469,-1.169105,
        -1.361357,1.803705,0.881076,
        -1.361357,1.804124,1.264249,
        -1.361357,1.807825,-2.090455,
        -1.361357,1.808822,-1.817375,
        -1.361357,1.809020,2.949741,
        -1.361357,1.812466,2.597595,
        -1.361357,1.813159,2.025395,
        -1.361357,1.813816,-1.298435,
        -1.361357,1.815016,0.278951,
        -1.361357,1.816510,-0.008631,
        -1.361357,1.817772,-1.689625,
        -1.361357,1.821360,-2.984352,
        -1.361357,1.822312,-0.127211,
        -1.361357,1.823386,-1.429807,
        -1.361357,1.823834,-1.560483,
        -1.361357,1.832060,-0.821627,
        -1.361357,1.832690,-0.468440,
        -1.361357,1.833945,-2.320278,
        -1.361357,1.834787,0.751968,
        -1.361357,1.840579,-2.536445,
        -1.361357,1.841033,-2.756932,
        -1.361357,1.846350,1.893983,
        -1.361357,1.846644,1.113766,
        -1.361357,1.846833,2.826543,
        -1.361357,1.847325,2.249051,
        -1.361357,1.854119,0.134161,
        -1.361357,1.858863,-3.106022,
        -1.361357,1.865046,1.763384,
        -1.361357,1.865619,-0.952036,
        -1.361357,1.865912,0.624453,
        -1.361357,1.866694,2.473686,
        -1.361357,1.867524,-0.240083,
        -1.361357,1.875561,1.632831,
        -1.361357,1.876049,-0.594378,
        -1.361357,1.880509,1.500204,
        -1.361357,1.883504,1.360457,
        -1.361357,1.886971,-2.196665,
        -1.361357,1.888979,0.982922,
        -1.361357,1.889815,2.701570,
        -1.361357,1.892144,-2.005340,
        -1.361357,1.892348,0.496522,
        -1.361357,1.893281,-1.083976,
        -1.361357,1.893828,3.053453,
        -1.361357,1.895117,2.121107,
        -1.361357,1.898741,-2.875735,
        -1.361357,1.904429,-2.428029,
        -1.361357,1.908408,-2.648903,
        -1.361357,1.911828,-0.364489,
        -1.361357,1.912224,0.366565,
        -1.361357,1.914854,1.214468,
        -1.361357,1.915467,-1.883143,
        -1.361357,1.915763,-0.724679,
        -1.361357,1.915790,-1.216366,
        -1.361357,1.922256,2.347806,
        -1.361357,1.922616,0.844633,
        -1.361357,1.925185,0.231170,
        -1.361357,1.927474,-1.754616,
        -1.361357,1.928342,-1.351174,
        -1.361357,1.933328,-1.623328,
        -1.361357,1.933878,0.034150,
        -1.361357,1.934585,-1.487979,
        -1.361357,1.934802,2.927634,
        -1.361357,1.934862,1.989776,
        -1.361357,1.936327,-0.099350,
        -1.361357,1.938088,2.578507,
        -1.361357,1.945677,-2.999805,
        -1.361357,1.954572,-0.859208,
        -1.361357,1.957868,-0.490784,
        -1.361357,1.958989,0.712772,
        -1.361357,1.960212,-2.307561,
        -1.361357,1.964774,1.855507,
        -1.361357,1.964971,-2.100533,
        -1.361357,1.969112,1.088273,
        -1.361357,1.971438,-2.765415,
        -1.361357,1.972312,2.805134,
        -1.361357,1.974950,2.220596,
        -1.361357,1.975135,-2.539127,
        -1.361357,1.982107,1.720829,
        -1.361357,1.983195,-3.127984,
        -1.361357,1.983318,-0.247182,
        -1.361357,1.986720,-0.993220,
        -1.361357,1.990381,1.586005,
        -1.361357,1.990470,0.580283,
        -1.361357,1.990523,1.450326,
        -1.361357,1.993909,1.312897,
        -1.361357,1.994694,2.452813,
        -1.361357,1.999226,-0.620717,
        -1.361357,2.005145,0.941012,
        -1.361357,2.007301,0.142337,
        -1.361357,2.014149,0.444138,
        -1.361357,2.014338,-1.127392,
        -1.361357,2.018131,3.021778,
        -1.361357,2.019676,-1.980944,
        -1.361357,2.019912,2.686612,
        -1.361357,2.020545,2.088672,
        -1.361357,2.026074,0.300945,
        -1.361357,2.026196,-2.889057,
        -1.361357,2.031959,-1.268178,
        -1.361357,2.032917,-2.199917,
        -1.361357,2.033572,-1.842884,
        -1.361357,2.034712,-2.419505,
        -1.361357,2.035818,-0.379479,
        -1.361357,2.036004,-0.757062,
        -1.361357,2.040501,-1.411488,
        -1.361357,2.042874,-1.701367,
        -1.361357,2.043060,-2.653746,
        -1.361357,2.043729,-1.557938,
        -1.361357,2.048538,0.806623,
        -1.361357,2.049006,1.199930,
        -1.361357,2.049502,-0.137989,
        -1.361357,2.052181,2.326959,
        -1.361357,2.052794,-0.008526,
        -1.361357,2.057608,1.951726,
        -1.361357,2.065568,2.567730,
        -1.361357,2.070371,2.879808,
        -1.361357,2.071256,-3.018961,
        -1.361357,2.079746,-0.897699,
        -1.361357,2.082784,1.042403,
        -1.361357,2.083752,-0.511086,
        -1.361357,2.084401,1.810502,
        -1.361357,2.085731,0.671081,
        -1.361357,2.097496,1.373172,
        -1.361357,2.099185,1.667006,
        -1.361357,2.100482,-2.301084,
        -1.361357,2.103118,2.194745,
        -1.361357,2.103223,1.521685,
        -1.361357,2.103695,-2.777707,
        -1.361357,2.107523,3.129581,
        -1.361357,2.108277,-2.095907,
        -1.361357,2.108891,-2.532166,
        -1.361357,2.111003,-0.268632,
        -1.361357,2.112276,-1.027246,
        -1.361357,2.114380,0.531457,
        -1.361357,2.119166,0.100552,
        -1.361357,2.119564,2.754818,
        -1.361357,2.122474,-0.648314,
        -1.361357,2.125183,2.445598,
        -1.361357,2.129501,0.238290,
        -1.361357,2.131178,0.386663,
        -1.361357,2.131223,-1.174311,
        -1.361357,2.134269,-1.948042,
        -1.361357,2.134684,0.907065,
        -1.361357,2.137615,2.988806,
        -1.361357,2.145635,-1.327708,
        -1.361357,2.146467,2.055081,
        -1.361357,2.146582,-1.793788,
        -1.361357,2.149210,-1.482321,
        -1.361357,2.152901,-1.636788,
        -1.361357,2.156337,1.151558,
        -1.361357,2.156474,-2.912065,
        -1.361357,2.157194,-0.792563,
        -1.361357,2.166724,-0.401638,
        -1.361357,2.166980,-0.134879,
        -1.361357,2.173644,-2.403429,
        -1.361357,2.175711,-2.655523,
        -1.361357,2.177132,2.600419,
        -1.361357,2.178879,0.767722,
        -1.361357,2.180752,1.908485,
        -1.361357,2.181338,2.313721,
        -1.361357,2.185580,1.287673,
        -1.361357,2.195400,-2.217546,
        -1.361357,2.200418,-3.056442,
        -1.361357,2.201829,-0.001735,
        -1.361357,2.203804,1.756176,
        -1.361357,2.211602,1.443930,
        -1.361357,2.212686,-0.544809,
        -1.361357,2.213937,0.621892,
        -1.361357,2.214075,1.600691,
        -1.361357,2.216256,1.020029,
        -1.361357,2.219060,2.886631,
        -1.361357,2.223375,-0.906185,
        -1.361357,2.226002,-1.060248,
        -1.361357,2.226413,-2.064034,
        -1.361357,2.230554,2.171334,
        -1.361357,2.232413,2.735777,
        -1.361357,2.233709,3.070830,
        -1.361357,2.234383,-2.792065,
        -1.361357,2.237183,0.468799,
        -1.361357,2.237695,-0.271388,
        -1.361357,2.239419,0.151693,
        -1.361357,2.243347,-2.521049,
        -1.361357,2.244035,-1.229456,
        -1.361357,2.244355,0.309899,
        -1.361357,2.245175,2.460236,
        -1.361357,2.245628,-1.899216,
        -1.361357,2.245973,-0.702234,
        -1.361357,2.255601,-1.728131,
        -1.361357,2.256702,-1.401367,
        -1.361357,2.261466,-1.557500,
        -1.361357,2.267390,0.878499,
        -1.361357,2.272099,2.018594,
        -1.361357,2.278618,1.170386,
        -1.361357,2.283304,-2.345584,
        -1.361357,2.285175,-2.943090,
        -1.361357,2.293000,-0.124451,
        -1.361357,2.294058,-0.421070,
        -1.361357,2.303283,1.855938,
        -1.361357,2.304968,2.608978,
        -1.361357,2.305347,2.311346,
        -1.361357,2.307188,-2.656134,
        -1.361357,2.307301,1.336200,
        -1.361357,2.309655,0.726020,
        -1.361357,2.317634,-2.185225,
        -1.361357,2.321510,1.685664,
        -1.361357,2.324376,-3.111006,
        -1.361357,2.324731,1.511186,
        -1.361357,2.329587,0.035498,
        -1.361357,2.330158,2.977264,
        -1.361357,2.335096,-0.588468,
        -1.361357,2.336046,-1.115799,
        -1.361357,2.338060,2.803613,
        -1.361357,2.338083,-0.943400,
        -1.361357,2.340210,0.562145,
        -1.361357,2.341931,1.023926,
        -1.361357,2.343265,-0.779061,
        -1.361357,2.343434,-2.013692,
        -1.361357,2.353029,-1.310529,
        -1.361357,2.354868,0.211543,
        -1.361357,2.355120,0.388593,
        -1.361357,2.355165,-1.833173,
        -1.361357,2.355462,2.151441,
        -1.361357,2.362948,-2.487477,
        -1.361357,2.363101,-2.810456,
        -1.361357,2.363394,-0.272918,
        -1.361357,2.364509,-1.644533,
        -1.361357,2.370465,2.466824,
        -1.361357,2.386110,-1.481351,
        -1.361357,2.394960,0.862133,
        -1.361357,2.395716,1.977184,
        -1.361357,2.396339,1.200900,
        -1.361357,2.407941,-2.316753,
        -1.361357,2.409450,-2.988432,
        -1.361357,2.417202,-0.103363,
        -1.361357,2.417386,-0.443097,
        -1.361357,2.417950,2.666662,
        -1.361357,2.423190,1.789783,
        -1.361357,2.423692,1.392696,
        -1.361357,2.431122,3.086379,
        -1.361357,2.431406,-2.647820,
        -1.361357,2.431584,2.307631,
        -1.361357,2.434090,1.592818,
        -1.361357,2.435993,0.683844,
        -1.361357,2.440898,-2.133300,
        -1.361357,2.442222,-1.195837,
        -1.361357,2.443112,2.883761,
        -1.361357,2.449191,-0.633651,
        -1.361357,2.450821,0.084741,
        -1.361357,2.453499,-1.008861,
        -1.361357,2.460010,-1.938011,
        -1.361357,2.460199,-0.822008,
        -1.361357,2.460902,0.490079,
        -1.361357,2.461774,-1.734913,
        -1.361357,2.463986,1.034407,
        -1.361357,2.467647,0.287667,
        -1.361357,2.479849,2.126925,
        -1.361357,2.485907,-2.835471,
        -1.361357,2.485945,-1.374853,
        -1.361357,2.487747,2.505912,
        -1.361357,2.487957,-0.274279,
        -1.361357,2.489741,-2.468428,
        -1.361357,2.512538,1.239900,
        -1.361357,2.517528,1.923289,
        -1.361357,2.518454,-1.565852,
        -1.361357,2.519348,0.843503,
        -1.361357,2.522766,-3.059231,
        -1.361357,2.528686,2.737228,
        -1.361357,2.534798,-2.266671,
        -1.361357,2.536533,1.466274,
        -1.361357,2.538504,1.699812,
        -1.361357,2.539516,-0.476942,
        -1.361357,2.539532,-0.074548,
        -1.361357,2.543197,2.990252,
        -1.361357,2.554022,2.320619,
        -1.361357,2.556919,-2.655005,
        -1.361357,2.558146,0.627191,
        -1.361357,2.558223,-1.164189,
        -1.361357,2.563452,-2.044028,
        -1.361357,2.568462,0.150655,
        -1.361357,2.569130,-0.700837,
        -1.361357,2.570050,-1.804492,
        -1.361357,2.572710,-0.934665,
        -1.361357,2.575994,0.391718,
        -1.361357,2.584774,1.046703,
        -1.361357,2.601432,2.102535,
        -1.361357,2.603559,2.555114,
        -1.361357,2.604150,-2.885208,
        -1.361357,2.608630,-1.378011,
        -1.361357,2.611256,-0.279110,
        -1.361357,2.617344,-2.439480,
        -1.361357,2.627402,1.293552,
        -1.361357,2.635396,1.848850,
        -1.361357,2.635686,2.834950,
        -1.361357,2.637087,3.132730,
        -1.361357,2.642029,0.813274,
        -1.361357,2.645260,1.570203,
        -1.361357,2.646182,-1.614036,
        -1.361357,2.660113,-2.186513,
        -1.361357,2.660483,-0.034608,
        -1.361357,2.661713,-0.529511,
        -1.361357,2.672720,-1.110422,
        -1.361357,2.676626,2.335353,
        -1.361357,2.676638,0.540364,
        -1.361357,2.676780,-1.901158,
        -1.361357,2.681411,-2.667965,
        -1.361357,2.682152,0.245751,
        -1.361357,2.682859,-0.817071,
        -1.361357,2.705107,1.062142,
        -1.361357,2.717705,2.625995,
        -1.361357,2.721001,2.060097,
        -1.361357,2.722319,-2.954069,
        -1.361357,2.729840,-1.368237,
        -1.361357,2.734349,-0.289298,
        -1.361357,2.738409,2.975671,
        -1.361357,2.740366,1.374341,
        -1.361357,2.744633,-2.391311,
        -1.361357,2.747398,1.727166,
        -1.361357,2.763799,0.760372,
        -1.361357,2.768238,-1.679243,
        -1.361357,2.779245,-2.046268,
        -1.361357,2.780208,-0.626453,
        -1.361357,2.780471,0.027732,
        -1.361357,2.784696,-1.013128,
        -1.361357,2.789817,0.395431,
        -1.361357,2.801154,2.353143,
        -1.361357,2.807091,-2.687477,
        -1.361357,2.825762,1.085035,
        -1.361357,2.830209,2.734943,
        -1.361357,2.833517,-3.085508,
        -1.361357,2.838301,1.980061,
        -1.361357,2.849871,1.512012,
        -1.361357,2.850089,-1.343899,
        -1.361357,2.856905,-0.318441,
        -1.361357,2.870445,-2.291001,
        -1.361357,2.885017,0.655562,
        -1.361357,2.886428,-1.793217,
        -1.361357,2.890337,-0.825633,
        -1.361357,2.898932,0.139973,
        -1.361357,2.931927,-2.728768,
        -1.361357,2.932096,2.377790,
        -1.361357,2.937409,2.947245,
        -1.361357,2.948048,1.133127,
        -1.361357,2.950896,1.797889,
        -1.361357,2.970019,-1.282382,
        -1.361357,2.978333,-0.412090,
        -1.361357,2.996437,-2.089121,
        -1.361357,3.011274,0.434110,
        -1.361357,3.050410,3.138492,
        -1.361357,3.064492,1.647259,
        -1.361357,3.088868,-0.989501,
        -1.256637,0.053884,-2.408394,
        -1.256637,0.077645,1.453433,
        -1.256637,0.088408,-0.311160,
        -1.256637,0.142314,2.590387,
        -1.256637,0.147206,-1.385164,
        -1.256637,0.163506,0.573203,
        -1.256637,0.165485,-2.875396,
        -1.256637,0.182564,-2.125821,
        -1.256637,0.190569,1.929141,
        -1.256637,0.201102,1.249082,
        -1.256637,0.201397,-0.716646,
        -1.256637,0.210568,-0.069559,
        -1.256637,0.251672,2.848900,
        -1.256637,0.266262,-1.678068,
        -1.256637,0.269108,-1.215563,
        -1.256637,0.273777,-2.506756,
        -1.256637,0.279854,0.814581,
        -1.256637,0.281664,-2.969147,
        -1.256637,0.283625,2.405603,
        -1.256637,0.284058,0.357350,
        -1.256637,0.294107,1.593880,
        -1.256637,0.302491,-0.400410,
        -1.256637,0.308059,2.021491,
        -1.256637,0.312219,-2.072057,
        -1.256637,0.321905,-0.829121,
        -1.256637,0.329984,1.201621,
        -1.256637,0.339714,-0.013712,
        -1.256637,0.368784,2.974974,
        -1.256637,0.376428,-1.495532,
        -1.256637,0.378103,2.646436,
        -1.256637,0.385088,0.590355,
        -1.256637,0.386248,-2.690943,
        -1.256637,0.389851,-2.340031,
        -1.256637,0.391502,-1.171374,
        -1.256637,0.398249,-1.812694,
        -1.256637,0.400542,-2.995729,
        -1.256637,0.404273,0.918163,
        -1.256637,0.405617,1.773533,
        -1.256637,0.409009,-0.570243,
        -1.256637,0.410700,0.270433,
        -1.256637,0.415893,1.464890,
        -1.256637,0.420680,2.088164,
        -1.256637,0.424766,-0.258729,
        -1.256637,0.426507,2.367872,
        -1.256637,0.441292,-0.890727,
        -1.256637,0.451839,-2.078193,
        -1.256637,0.459270,1.191562,
        -1.256637,0.470622,0.014002,
        -1.256637,0.481292,-2.529098,
        -1.256637,0.482460,2.849843,
        -1.256637,0.492205,-1.397963,
        -1.256637,0.493314,3.117264,
        -1.256637,0.497870,0.727828,
        -1.256637,0.499783,-1.654379,
        -1.256637,0.502651,0.463333,
        -1.256637,0.505463,2.594557,
        -1.256637,0.506104,-2.874910,
        -1.256637,0.513579,-1.147354,
        -1.256637,0.518569,1.644237,
        -1.256637,0.519001,1.900488,
        -1.256637,0.519277,-0.673809,
        -1.256637,0.524769,-0.421955,
        -1.256637,0.525564,-2.288838,
        -1.256637,0.532541,0.978737,
        -1.256637,0.534610,2.154808,
        -1.256637,0.535020,-1.891846,
        -1.256637,0.540208,1.400497,
        -1.256637,0.542006,0.228337,
        -1.256637,0.550762,-0.180771,
        -1.256637,0.560531,2.378474,
        -1.256637,0.561629,-0.924576,
        -1.256637,0.567465,-2.679392,
        -1.256637,0.588527,1.181755,
        -1.256637,0.591643,-2.098364,
        -1.256637,0.594963,2.997575,
        -1.256637,0.599643,-3.045235,
        -1.256637,0.599962,2.773556,
        -1.256637,0.602620,0.029270,
        -1.256637,0.605711,0.598454,
        -1.256637,0.607708,-1.541026,
        -1.256637,0.609360,-2.464379,
        -1.256637,0.610738,-1.326786,
        -1.256637,0.616968,0.820489,
        -1.256637,0.626379,1.773610,
        -1.256637,0.628626,-1.751180,
        -1.256637,0.630671,0.391949,
        -1.256637,0.632622,-0.737642,
        -1.256637,0.632708,2.566722,
        -1.256637,0.632797,1.989858,
        -1.256637,0.634455,-0.534790,
        -1.256637,0.636205,1.561170,
        -1.256637,0.637681,-1.123423,
        -1.256637,0.641441,-2.842109,
        -1.256637,0.644807,-0.327139,
        -1.256637,0.652440,2.197129,
        -1.256637,0.660897,-2.278593,
        -1.256637,0.664008,1.360091,
        -1.256637,0.670850,-1.945257,
        -1.256637,0.671396,1.004859,
        -1.256637,0.673985,0.203855,
        -1.256637,0.680326,-0.133370,
        -1.256637,0.685083,-0.942363,
        -1.256637,0.688204,-2.632204,
        -1.256637,0.689777,2.385011,
        -1.256637,0.700065,3.115430,
        -1.256637,0.704080,2.911636,
        -1.256637,0.712883,1.174216,
        -1.256637,0.715701,0.699906,
        -1.256637,0.718851,-1.442215,
        -1.256637,0.721019,2.721536,
        -1.256637,0.725300,-2.975134,
        -1.256637,0.727136,-1.633288,
        -1.256637,0.727399,-2.120567,
        -1.256637,0.730751,0.522488,
        -1.256637,0.732892,-1.254330,
        -1.256637,0.734333,-2.446184,
        -1.256637,0.735125,0.037270,
        -1.256637,0.737893,1.866463,
        -1.256637,0.740650,1.681075,
        -1.256637,0.745971,-0.781286,
        -1.256637,0.749432,-0.442559,
        -1.256637,0.750102,2.050636,
        -1.256637,0.751046,0.863821,
        -1.256637,0.757018,-0.608833,
        -1.256637,0.757046,-1.816019,
        -1.256637,0.757389,1.501184,
        -1.256637,0.759800,2.546181,
        -1.256637,0.759859,0.351923,
        -1.256637,0.761141,-2.781102,
        -1.256637,0.769035,-1.078574,
        -1.256637,0.770166,-0.266604,
        -1.256637,0.772218,2.227380,
        -1.256637,0.788382,1.331051,
        -1.256637,0.790269,-2.282409,
        -1.256637,0.803410,-1.984567,
        -1.256637,0.803813,0.191520,
        -1.256637,0.806832,3.022155,
        -1.256637,0.807476,-2.597675,
        -1.256637,0.812512,-0.107335,
        -1.256637,0.814161,1.010949,
        -1.256637,0.817732,-3.084486,
        -1.256637,0.817846,2.390688,
        -1.256637,0.819323,2.845960,
        -1.256637,0.819430,-0.915017,
        -1.256637,0.829198,-1.532491,
        -1.256637,0.829531,0.665695,
        -1.256637,0.829816,1.167147,
        -1.256637,0.833613,-1.357492,
        -1.256637,0.842945,-2.907318,
        -1.256637,0.844883,2.679708,
        -1.256637,0.848189,-1.705003,
        -1.256637,0.849888,1.767901,
        -1.256637,0.851126,0.493370,
        -1.256637,0.852911,1.933394,
        -1.256637,0.857516,-2.432863,
        -1.256637,0.858261,-1.191729,
        -1.256637,0.858279,-2.138895,
        -1.256637,0.859723,-0.734928,
        -1.256637,0.859754,1.603292,
        -1.256637,0.861655,-0.394821,
        -1.256637,0.864300,0.043549,
        -1.256637,0.866940,0.851229,
        -1.256637,0.869748,2.095969,
        -1.256637,0.877352,-0.574668,
        -1.256637,0.879254,-2.733338,
        -1.256637,0.882554,1.443757,
        -1.256637,0.884060,0.334199,
        -1.256637,0.885626,-1.866061,
        -1.256637,0.889834,2.529957,
        -1.256637,0.893140,2.251698,
        -1.256637,0.897463,-0.240822,
        -1.256637,0.898758,-1.031948,
        -1.256637,0.913918,1.287618,
        -1.256637,0.916555,-2.287173,
        -1.256637,0.918305,3.107845,
        -1.256637,0.919522,2.948277,
        -1.256637,0.927138,-2.569686,
        -1.256637,0.929031,0.185383,
        -1.256637,0.933766,-3.015917,
        -1.256637,0.934873,-2.010822,
        -1.256637,0.935636,-1.445855,
        -1.256637,0.936702,-0.866461,
        -1.256637,0.937395,0.729108,
        -1.256637,0.937910,2.787575,
        -1.256637,0.939144,-0.092962,
        -1.256637,0.941718,1.125085,
        -1.256637,0.942873,-1.607680,
        -1.256637,0.943701,0.980328,
        -1.256637,0.947774,2.395256,
        -1.256637,0.952433,0.584347,
        -1.256637,0.954456,-1.287293,
        -1.256637,0.958731,-2.852552,
        -1.256637,0.961838,1.839284,
        -1.256637,0.961898,-0.469425,
        -1.256637,0.965426,1.686615,
        -1.256637,0.970108,1.990771,
        -1.256637,0.972820,-1.762780,
        -1.256637,0.973500,-0.711881,
        -1.256637,0.975253,2.641082,
        -1.256637,0.977824,0.441928,
        -1.256637,0.981485,-2.424766,
        -1.256637,0.981877,1.537100,
        -1.256637,0.982609,-1.136920,
        -1.256637,0.984265,-2.156073,
        -1.256637,0.989150,0.045493,
        -1.256637,0.993052,2.137478,
        -1.256637,0.997766,-2.693871,
        -1.256637,0.999847,-0.334738,
        -1.256637,1.001920,0.848054,
        -1.256637,1.006146,1.389734,
        -1.256637,1.012180,0.304066,
        -1.256637,1.016938,2.270674,
        -1.256637,1.017371,-0.979790,
        -1.256637,1.019495,-1.894487,
        -1.256637,1.023796,-0.583691,
        -1.256637,1.026525,2.506945,
        -1.256637,1.027004,-0.203653,
        -1.256637,1.028027,3.036573,
        -1.256637,1.031435,1.241021,
        -1.256637,1.032316,2.884221,
        -1.256637,1.035863,-3.110475,
        -1.256637,1.042085,-2.292950,
        -1.256637,1.043272,-1.522986,
        -1.256637,1.046265,-2.958847,
        -1.256637,1.049272,-1.376344,
        -1.256637,1.050039,-2.552086,
        -1.256637,1.053682,0.689863,
        -1.256637,1.055458,0.170394,
        -1.256637,1.056622,-0.831866,
        -1.256637,1.057515,1.097998,
        -1.256637,1.058706,-2.036438,
        -1.256637,1.061389,-1.672248,
        -1.256637,1.063758,2.743093,
        -1.256637,1.070157,-0.073422,
        -1.256637,1.072154,1.761506,
        -1.256637,1.074029,1.909193,
        -1.256637,1.074204,0.545276,
        -1.256637,1.074714,-2.805147,
        -1.256637,1.076262,0.966104,
        -1.256637,1.077642,-1.227073,
        -1.256637,1.084151,1.618525,
        -1.256637,1.085085,2.377673,
        -1.256637,1.086822,-0.441112,
        -1.256637,1.087142,2.058815,
        -1.256637,1.098383,-0.696430,
        -1.256637,1.099138,-1.084033,
        -1.256637,1.103208,1.477773,
        -1.256637,1.103783,0.408254,
        -1.256637,1.106275,-2.420800,
        -1.256637,1.108079,2.612943,
        -1.256637,1.109345,-2.172718,
        -1.256637,1.115957,-1.787597,
        -1.256637,1.118930,0.045983,
        -1.256637,1.119777,0.834702,
        -1.256637,1.121943,-2.669042,
        -1.256637,1.124930,2.219917,
        -1.256637,1.125217,1.336294,
        -1.256637,1.130958,2.972931,
        -1.256637,1.131400,-0.306629,
        -1.256637,1.137135,-0.936514,
        -1.256637,1.139125,-1.924528,
        -1.256637,1.139877,0.276382,
        -1.256637,1.142003,-3.058022,
        -1.256637,1.147363,-0.558448,
        -1.256637,1.150348,-0.176418,
        -1.256637,1.151198,1.197016,
        -1.256637,1.153930,-1.597104,
        -1.256637,1.155389,2.837148,
        -1.256637,1.155458,-1.460623,
        -1.256637,1.157331,-2.906362,
        -1.256637,1.158594,2.487618,
        -1.256637,1.162103,-1.322553,
        -1.256637,1.165407,3.092861,
        -1.256637,1.168025,0.717075,
        -1.256637,1.168406,-2.301519,
        -1.256637,1.174782,-2.538755,
        -1.256637,1.177998,1.061256,
        -1.256637,1.178898,1.827305,
        -1.256637,1.181331,-2.060330,
        -1.256637,1.182540,-0.798952,
        -1.256637,1.184569,0.147517,
        -1.256637,1.184734,1.968919,
        -1.256637,1.184804,0.593386,
        -1.256637,1.188278,1.690808,
        -1.256637,1.194360,2.110440,
        -1.256637,1.194776,2.710455,
        -1.256637,1.197745,-1.173705,
        -1.256637,1.199898,-2.777257,
        -1.256637,1.201164,2.355931,
        -1.256637,1.203760,1.556404,
        -1.256637,1.205845,-0.415993,
        -1.256637,1.206585,0.933648,
        -1.256637,1.206950,-0.037248,
        -1.256637,1.209341,0.471162,
        -1.256637,1.213483,-1.039417,
        -1.256637,1.218146,-0.672643,
        -1.256637,1.221608,1.421669,
        -1.256637,1.226043,-1.693533,
        -1.256637,1.232870,-1.822773,
        -1.256637,1.234375,-2.421194,
        -1.256637,1.234531,-2.190148,
        -1.256637,1.239414,0.348959,
        -1.256637,1.240183,2.587255,
        -1.256637,1.242312,2.941234,
        -1.256637,1.244075,1.286644,
        -1.256637,1.245961,-2.997461,
        -1.256637,1.246495,-2.646872,
        -1.256637,1.247736,2.242810,
        -1.256637,1.253604,0.812222,
        -1.256637,1.257816,-0.894317,
        -1.256637,1.258247,-1.953634,
        -1.256637,1.258993,-0.284099,
        -1.256637,1.262380,-0.157609,
        -1.256637,1.262791,-1.549229,
        -1.256637,1.263772,-0.533073,
        -1.256637,1.266731,-1.412461,
        -1.256637,1.268591,-3.122146,
        -1.256637,1.269614,-1.275246,
        -1.256637,1.270745,1.152533,
        -1.256637,1.273606,0.224545,
        -1.256637,1.279716,2.812101,
        -1.256637,1.283786,2.462124,
        -1.256637,1.284798,0.078209,
        -1.256637,1.285402,-2.882049,
        -1.256637,1.286936,1.886053,
        -1.256637,1.289408,0.683122,
        -1.256637,1.294736,1.756184,
        -1.256637,1.296978,-2.311511,
        -1.256637,1.298546,2.020601,
        -1.256637,1.300924,1.022478,
        -1.256637,1.303305,3.043152,
        -1.256637,1.304210,-2.084225,
        -1.256637,1.305241,-2.530249,
        -1.256637,1.307513,1.628003,
        -1.256637,1.310722,0.558267,
        -1.256637,1.314142,2.139329,
        -1.256637,1.316087,-0.777971,
        -1.256637,1.317907,-1.121023,
        -1.256637,1.321019,-2.749776,
        -1.256637,1.321609,2.687717,
        -1.256637,1.322568,1.499415,
        -1.256637,1.322972,-0.395699,
        -1.256637,1.325286,-0.994192,
        -1.256637,1.329146,-0.648513,
        -1.256637,1.329795,-0.042570,
        -1.256637,1.331651,2.344497,
        -1.256637,1.336522,0.435098,
        -1.256637,1.339410,0.900087,
        -1.256637,1.339550,1.370353,
        -1.256637,1.340441,-1.731580,
        -1.256637,1.351855,-1.855234,
        -1.256637,1.360161,-3.018404,
        -1.256637,1.361031,-1.613971,
        -1.256637,1.361196,2.914780,
        -1.256637,1.361998,-2.207254,
        -1.256637,1.362302,1.239384,
        -1.256637,1.365228,2.563808,
        -1.256637,1.365452,0.310522,
        -1.256637,1.367028,-2.420041,
        -1.256637,1.373422,-1.484281,
        -1.256637,1.374245,-1.226417,
        -1.256637,1.375020,-0.170393,
        -1.256637,1.376560,-1.355481,
        -1.256637,1.377070,-1.980984,
        -1.256637,1.377578,-2.636339,
        -1.256637,1.378096,0.774626,
        -1.256637,1.379086,-0.508192,
        -1.256637,1.382562,0.177141,
        -1.256637,1.388849,2.233847,
        -1.256637,1.389268,-0.886349,
        -1.256637,1.391324,1.109396,
        -1.256637,1.391699,-0.288725,
        -1.256637,1.392869,3.136903,
        -1.256637,1.393612,1.952016,
        -1.256637,1.395545,-2.850807,
        -1.256637,1.401926,1.822930,
        -1.256637,1.402149,2.788898,
        -1.256637,1.409552,0.647980,
        -1.256637,1.410840,0.056219,
        -1.256637,1.411944,2.444377,
        -1.256637,1.413639,1.697982,
        -1.256637,1.426479,1.573427,
        -1.256637,1.426725,2.102242,
        -1.256637,1.427868,0.984637,
        -1.256637,1.429580,-2.107232,
        -1.256637,1.431375,-2.313269,
        -1.256637,1.431907,3.015410,
        -1.256637,1.432325,-1.029054,
        -1.256637,1.433341,0.524268,
        -1.256637,1.435181,-0.739771,
        -1.256637,1.437208,-2.529085,
        -1.256637,1.439041,-0.617966,
        -1.256637,1.441473,1.448544,
        -1.256637,1.444527,2.664744,
        -1.256637,1.450513,-0.400625,
        -1.256637,1.451907,-0.066823,
        -1.256637,1.452754,-2.738631,
        -1.256637,1.454061,-2.950416,
        -1.256637,1.456736,1.323154,
        -1.256637,1.458370,0.400433,
        -1.256637,1.459314,-1.765003,
        -1.256637,1.459902,-1.145835,
        -1.256637,1.464952,2.329245,
        -1.256637,1.466050,0.862104,
        -1.256637,1.471330,-1.886160,
        -1.256637,1.475119,-1.648255,
        -1.256637,1.480878,1.194576,
        -1.256637,1.480942,0.273235,
        -1.256637,1.481086,2.890269,
        -1.256637,1.484118,-1.529156,
        -1.256637,1.484741,-1.279219,
        -1.256637,1.487681,-1.406069,
        -1.256637,1.487886,-3.066037,
        -1.256637,1.490555,2.543284,
        -1.256637,1.492424,-0.190932,
        -1.256637,1.495002,-0.839810,
        -1.256637,1.495297,2.007188,
        -1.256637,1.496414,-2.009267,
        -1.256637,1.498261,-2.425270,
        -1.256637,1.498907,-2.207327,
        -1.256637,1.500204,0.737638,
        -1.256637,1.502469,0.151156,
        -1.256637,1.507448,-0.514782,
        -1.256637,1.509287,2.204647,
        -1.256637,1.510518,1.884920,
        -1.256637,1.511948,-2.632360,
        -1.256637,1.515235,1.068565,
        -1.256637,1.522237,3.102018,
        -1.256637,1.522373,1.764991,
        -1.256637,1.522416,-0.954956,
        -1.256637,1.523102,2.765920,
        -1.256637,1.524811,-2.844718,
        -1.256637,1.527261,-0.304916,
        -1.256637,1.529788,0.613208,
        -1.256637,1.533648,1.644808,
        -1.256637,1.535018,0.028869,
        -1.256637,1.541082,2.424608,
        -1.256637,1.546010,1.523462,
        -1.256637,1.551135,-0.642994,
        -1.256637,1.553895,0.489607,
        -1.256637,1.555041,0.947813,
        -1.256637,1.559671,-1.074934,
        -1.256637,1.560345,2.985899,
        -1.256637,1.560487,-2.323400,
        -1.256637,1.561127,1.401111,
        -1.256637,1.567193,2.643172,
        -1.256637,1.568892,2.098362,
        -1.256637,1.571702,-2.098435,
        -1.256637,1.572701,-2.528730,
        -1.256637,1.573572,1.278593,
        -1.256637,1.574849,-0.091505,
        -1.256637,1.576334,0.365236,
        -1.256637,1.577029,-2.962888,
        -1.256637,1.577688,-1.814941,
        -1.256637,1.579542,-0.755917,
        -1.256637,1.579624,-1.199759,
        -1.256637,1.581515,-0.419414,
        -1.256637,1.584581,-2.738702,
        -1.256637,1.586948,-1.696296,
        -1.256637,1.588901,2.301496,
        -1.256637,1.589936,0.825020,
        -1.256637,1.590018,-1.941652,
        -1.256637,1.596968,-1.328358,
        -1.256637,1.597178,-1.575607,
        -1.256637,1.600470,-1.453196,
        -1.256637,1.601408,0.242409,
        -1.256637,1.602259,1.152656,
        -1.256637,1.603064,2.865774,
        -1.256637,1.612085,1.968802,
        -1.256637,1.612720,-3.080498,
        -1.256637,1.616061,2.521531,
        -1.256637,1.616900,-0.207457,
        -1.256637,1.620988,-0.875920,
        -1.256637,1.622114,0.701509,
        -1.256637,1.622749,0.121426,
        -1.256637,1.622993,-2.221629,
        -1.256637,1.629346,-0.542669,
        -1.256637,1.630007,1.844552,
        -1.256637,1.634747,-2.426198,
        -1.256637,1.642346,1.722617,
        -1.256637,1.643925,1.034629,
        -1.256637,1.644848,2.743774,
        -1.256637,1.645258,2.188109,
        -1.256637,1.645417,-2.634725,
        -1.256637,1.647395,3.085437,
        -1.256637,1.648538,-2.855044,
        -1.256637,1.649932,0.576808,
        -1.256637,1.652224,-0.996105,
        -1.256637,1.652871,1.600681,
        -1.256637,1.657997,-0.325127,
        -1.256637,1.661936,0.001037,
        -1.256637,1.665187,1.477402,
        -1.256637,1.665496,2.397872,
        -1.256637,1.673267,-0.665453,
        -1.256637,1.673557,0.452526,
        -1.256637,1.679277,-1.122421,
        -1.256637,1.679696,0.912482,
        -1.256637,1.682420,1.353922,
        -1.256637,1.685031,-2.009038,
        -1.256637,1.685196,2.967046,
        -1.256637,1.690107,2.620910,
        -1.256637,1.690453,-2.123076,
        -1.256637,1.691693,-1.871572,
        -1.256637,1.691817,2.061688,
        -1.256637,1.692644,1.237670,
        -1.256637,1.696386,0.328123,
        -1.256637,1.697162,-1.249026,
        -1.256637,1.697488,-2.322785,
        -1.256637,1.698726,-2.973576,
        -1.256637,1.699172,-1.749717,
        -1.256637,1.704478,-0.106295,
        -1.256637,1.707244,-0.444770,
        -1.256637,1.707500,-2.531497,
        -1.256637,1.707714,-0.787321,
        -1.256637,1.708430,-1.627552,
        -1.256637,1.710583,-1.376890,
        -1.256637,1.712415,-2.747921,
        -1.256637,1.713125,-1.503073,
        -1.256637,1.713452,0.789419,
        -1.256637,1.720612,2.279888,
        -1.256637,1.724852,0.202831,
        -1.256637,1.724868,2.846246,
        -1.256637,1.728940,1.931535,
        -1.256637,1.732838,1.127584,
        -1.256637,1.735719,-3.092485,
        -1.256637,1.739676,2.496017,
        -1.256637,1.739971,-0.228876,
        -1.256637,1.743328,0.663347,
        -1.256637,1.744927,-0.911892,
        -1.256637,1.745952,0.084733,
        -1.256637,1.748077,1.804354,
        -1.256637,1.752959,-0.567548,
        -1.256637,1.760169,1.678132,
        -1.256637,1.763887,-2.218450,
        -1.256637,1.766737,2.722752,
        -1.256637,1.767999,1.004737,
        -1.256637,1.769149,1.550298,
        -1.256637,1.770119,-2.427175,
        -1.256637,1.770460,2.154208,
        -1.256637,1.770514,3.070982,
        -1.256637,1.770735,0.537245,
        -1.256637,1.772673,-2.864760,
        -1.256637,1.773527,-1.039582,
        -1.256637,1.776419,-2.642421,
        -1.256637,1.780769,1.416330,
        -1.256637,1.786634,-0.346247,
        -1.256637,1.788829,-1.947553,
        -1.256637,1.793737,0.410270,
        -1.256637,1.794553,2.374618,
        -1.256637,1.794699,-0.694061,
        -1.256637,1.797469,-1.169105,
        -1.256637,1.803705,0.881076,
        -1.256637,1.804124,1.264249,
        -1.256637,1.807825,-2.090455,
        -1.256637,1.808822,-1.817375,
        -1.256637,1.809020,2.949741,
        -1.256637,1.812466,2.597595,
        -1.256637,1.813159,2.025395,
        -1.256637,1.813816,-1.298435,
        -1.256637,1.815016,0.278951,
        -1.256637,1.816510,-0.008631,
        -1.256637,1.817772,-1.689625,
        -1.256637,1.821360,-2.984352,
        -1.256637,1.822312,-0.127211,
        -1.256637,1.823386,-1.429807,
        -1.256637,1.823834,-1.560483,
        -1.256637,1.832060,-0.821627,
        -1.256637,1.832690,-0.468440,
        -1.256637,1.833945,-2.320278,
        -1.256637,1.834787,0.751968,
        -1.256637,1.840579,-2.536445,
        -1.256637,1.841033,-2.756932,
        -1.256637,1.846350,1.893983,
        -1.256637,1.846644,1.113766,
        -1.256637,1.846833,2.826543,
        -1.256637,1.847325,2.249051,
        -1.256637,1.854119,0.134161,
        -1.256637,1.858863,-3.106022,
        -1.256637,1.865046,1.763384,
        -1.256637,1.865619,-0.952036,
        -1.256637,1.865912,0.624453,
        -1.256637,1.866694,2.473686,
        -1.256637,1.867524,-0.240083,
        -1.256637,1.875561,1.632831,
        -1.256637,1.876049,-0.594378,
        -1.256637,1.880509,1.500204,
        -1.256637,1.883504,1.360457,
        -1.256637,1.886971,-2.196665,
        -1.256637,1.888979,0.982922,
        -1.256637,1.889815,2.701570,
        -1.256637,1.892144,-2.005340,
        -1.256637,1.892348,0.496522,
        -1.256637,1.893281,-1.083976,
        -1.256637,1.893828,3.053453,
        -1.256637,1.895117,2.121107,
        -1.256637,1.898741,-2.875735,
        -1.256637,1.904429,-2.428029,
        -1.256637,1.908408,-2.648903,
        -1.256637,1.911828,-0.364489,
        -1.256637,1.912224,0.366565,
        -1.256637,1.914854,1.214468,
        -1.256637,1.915467,-1.883143,
        -1.256637,1.915763,-0.724679,
        -1.256637,1.915790,-1.216366,
        -1.256637,1.922256,2.347806,
        -1.256637,1.922616,0.844633,
        -1.256637,1.925185,0.231170,
        -1.256637,1.927474,-1.754616,
        -1.256637,1.928342,-1.351174,
        -1.256637,1.933328,-1.623328,
        -1.256637,1.933878,0.034150,
        -1.256637,1.934585,-1.487979,
        -1.256637,1.934802,2.927634,
        -1.256637,1.934862,1.989776,
        -1.256637,1.936327,-0.099350,
        -1.256637,1.938088,2.578507,
        -1.256637,1.945677,-2.999805,
        -1.256637,1.954572,-0.859208,
        -1.256637,1.957868,-0.490784,
        -1.256637,1.958989,0.712772,
        -1.256637,1.960212,-2.307561,
        -1.256637,1.964774,1.855507,
        -1.256637,1.964971,-2.100533,
        -1.256637,1.969112,1.088273,
        -1.256637,1.971438,-2.765415,
        -1.256637,1.972312,2.805134,
        -1.256637,1.974950,2.220596,
        -1.256637,1.975135,-2.539127,
        -1.256637,1.982107,1.720829,
        -1.256637,1.983195,-3.127984,
        -1.256637,1.983318,-0.247182,
        -1.256637,1.986720,-0.993220,
        -1.256637,1.990381,1.586005,
        -1.256637,1.990470,0.580283,
        -1.256637,1.990523,1.450326,
        -1.256637,1.993909,1.312897,
        -1.256637,1.994694,2.452813,
        -1.256637,1.999226,-0.620717,
        -1.256637,2.005145,0.941012,
        -1.256637,2.007301,0.142337,
        -1.256637,2.014149,0.444138,
        -1.256637,2.014338,-1.127392,
        -1.256637,2.018131,3.021778,
        -1.256637,2.019676,-1.980944,
        -1.256637,2.019912,2.686612,
        -1.256637,2.020545,2.088672,
        -1.256637,2.026074,0.300945,
        -1.256637,2.026196,-2.889057,
        -1.256637,2.031959,-1.268178,
        -1.256637,2.032917,-2.199917,
        -1.256637,2.033572,-1.842884,
        -1.256637,2.034712,-2.419505,
        -1.256637,2.035818,-0.379479,
        -1.256637,2.036004,-0.757062,
        -1.256637,2.040501,-1.411488,
        -1.256637,2.042874,-1.701367,
        -1.256637,2.043060,-2.653746,
        -1.256637,2.043729,-1.557938,
        -1.256637,2.048538,0.806623,
        -1.256637,2.049006,1.199930,
        -1.256637,2.049502,-0.137989,
        -1.256637,2.052181,2.326959,
        -1.256637,2.052794,-0.008526,
        -1.256637,2.057608,1.951726,
        -1.256637,2.065568,2.567730,
        -1.256637,2.070371,2.879808,
        -1.256637,2.071256,-3.018961,
        -1.256637,2.079746,-0.897699,
        -1.256637,2.082784,1.042403,
        -1.256637,2.083752,-0.511086,
        -1.256637,2.084401,1.810502,
        -1.256637,2.085731,0.671081,
        -1.256637,2.097496,1.373172,
        -1.256637,2.099185,1.667006,
        -1.256637,2.100482,-2.301084,
        -1.256637,2.103118,2.194745,
        -1.256637,2.103223,1.521685,
        -1.256637,2.103695,-2.777707,
        -1.256637,2.107523,3.129581,
        -1.256637,2.108277,-2.095907,
        -1.256637,2.108891,-2.532166,
        -1.256637,2.111003,-0.268632,
        -1.256637,2.112276,-1.027246,
        -1.256637,2.114380,0.531457,
        -1.256637,2.119166,0.100552,
        -1.256637,2.119564,2.754818,
        -1.256637,2.122474,-0.648314,
        -1.256637,2.125183,2.445598,
        -1.256637,2.129501,0.238290,
        -1.256637,2.131178,0.386663,
        -1.256637,2.131223,-1.174311,
        -1.256637,2.134269,-1.948042,
        -1.256637,2.134684,0.907065,
        -1.256637,2.137615,2.988806,
        -1.256637,2.145635,-1.327708,
        -1.256637,2.146467,2.055081,
        -1.256637,2.146582,-1.793788,
        -1.256637,2.149210,-1.482321,
        -1.256637,2.152901,-1.636788,
        -1.256637,2.156337,1.151558,
        -1.256637,2.156474,-2.912065,
        -1.256637,2.157194,-0.792563,
        -1.256637,2.166724,-0.401638,
        -1.256637,2.166980,-0.134879,
        -1.256637,2.173644,-2.403429,
        -1.256637,2.175711,-2.655523,
        -1.256637,2.177132,2.600419,
        -1.256637,2.178879,0.767722,
        -1.256637,2.180752,1.908485,
        -1.256637,2.181338,2.313721,
        -1.256637,2.185580,1.287673,
        -1.256637,2.195400,-2.217546,
        -1.256637,2.200418,-3.056442,
        -1.256637,2.201829,-0.001735,
        -1.256637,2.203804,1.756176,
        -1.256637,2.211602,1.443930,
        -1.256637,2.212686,-0.544809,
        -1.256637,2.213937,0.621892,
        -1.256637,2.214075,1.600691,
        -1.256637,2.216256,1.020029,
        -1.256637,2.219060,2.886631,
        -1.256637,2.223375,-0.906185,
        -1.256637,2.226002,-1.060248,
        -1.256637,2.226413,-2.064034,
        -1.256637,2.230554,2.171334,
        -1.256637,2.232413,2.735777,
        -1.256637,2.233709,3.070830,
        -1.256637,2.234383,-2.792065,
        -1.256637,2.237183,0.468799,
        -1.256637,2.237695,-0.271388,
        -1.256637,2.239419,0.151693,
        -1.256637,2.243347,-2.521049,
        -1.256637,2.244035,-1.229456,
        -1.256637,2.244355,0.309899,
        -1.256637,2.245175,2.460236,
        -1.256637,2.245628,-1.899216,
        -1.256637,2.245973,-0.702234,
        -1.256637,2.255601,-1.728131,
        -1.256637,2.256702,-1.401367,
        -1.256637,2.261466,-1.557500,
        -1.256637,2.267390,0.878499,
        -1.256637,2.272099,2.018594,
        -1.256637,2.278618,1.170386,
        -1.256637,2.283304,-2.345584,
        -1.256637,2.285175,-2.943090,
        -1.256637,2.293000,-0.124451,
        -1.256637,2.294058,-0.421070,
        -1.256637,2.303283,1.855938,
        -1.256637,2.304968,2.608978,
        -1.256637,2.305347,2.311346,
        -1.256637,2.307188,-2.656134,
        -1.256637,2.307301,1.336200,
        -1.256637,2.309655,0.726020,
        -1.256637,2.317634,-2.185225,
        -1.256637,2.321510,1.685664,
        -1.256637,2.324376,-3.111006,
        -1.256637,2.324731,1.511186,
        -1.256637,2.329587,0.035498,
        -1.256637,2.330158,2.977264,
        -1.256637,2.335096,-0.588468,
        -1.256637,2.336046,-1.115799,
        -1.256637,2.338060,2.803613,
        -1.256637,2.338083,-0.943400,
        -1.256637,2.340210,0.562145,
        -1.256637,2.341931,1.023926,
        -1.256637,2.343265,-0.779061,
        -1.256637,2.343434,-2.013692,
        -1.256637,2.353029,-1.310529,
        -1.256637,2.354868,0.211543,
        -1.256637,2.355120,0.388593,
        -1.256637,2.355165,-1.833173,
        -1.256637,2.355462,2.151441,
        -1.256637,2.362948,-2.487477,
        -1.256637,2.363101,-2.810456,
        -1.256637,2.363394,-0.272918,
        -1.256637,2.364509,-1.644533,
        -1.256637,2.370465,2.466824,
        -1.256637,2.386110,-1.481351,
        -1.256637,2.394960,0.862133,
        -1.256637,2.395716,1.977184,
        -1.256637,2.396339,1.200900,
        -1.256637,2.407941,-2.316753,
        -1.256637,2.409450,-2.988432,
        -1.256637,2.417202,-0.103363,
        -1.256637,2.417386,-0.443097,
        -1.256637,2.417950,2.666662,
        -1.256637,2.423190,1.789783,
        -1.256637,2.423692,1.392696,
        -1.256637,2.431122,3.086379,
        -1.256637,2.431406,-2.647820,
        -1.256637,2.431584,2.307631,
        -1.256637,2.434090,1.592818,
        -1.256637,2.435993,0.683844,
        -1.256637,2.440898,-2.133300,
        -1.256637,2.442222,-1.195837,
        -1.256637,2.443112,2.883761,
        -1.256637,2.449191,-0.633651,
        -1.256637,2.450821,0.084741,
        -1.256637,2.453499,-1.008861,
        -1.256637,2.460010,-1.938011,
        -1.256637,2.460199,-0.822008,
        -1.256637,2.460902,0.490079,
        -1.256637,2.461774,-1.734913,
        -1.256637,2.463986,1.034407,
        -1.256637,2.467647,0.287667,
        -1.256637,2.479849,2.126925,
        -1.256637,2.485907,-2.835471,
        -1.256637,2.485945,-1.374853,
        -1.256637,2.487747,2.505912,
        -1.256637,2.487957,-0.274279,
        -1.256637,2.489741,-2.468428,
        -1.256637,2.512538,1.239900,
        -1.256637,2.517528,1.923289,
        -1.256637,2.518454,-1.565852,
        -1.256637,2.519348,0.843503,
        -1.256637,2.522766,-3.059231,
        -1.256637,2.528686,2.737228,
        -1.256637,2.534798,-2.266671,
        -1.256637,2.536533,1.466274,
        -1.256637,2.538504,1.699812,
        -1.256637,2.539516,-0.476942,
        -1.256637,2.539532,-0.074548,
        -1.256637,2.543197,2.990252,
        -1.256637,2.554022,2.320619,
        -1.256637,2.556919,-2.655005,
        -1.256637,2.558146,0.627191,
        -1.256637,2.558223,-1.164189,
        -1.256637,2.563452,-2.044028,
        -1.256637,2.568462,0.150655,
        -1.256637,2.569130,-0.700837,
        -1.256637,2.570050,-1.804492,
        -1.256637,2.572710,-0.934665,
        -1.256637,2.575994,0.391718,
        -1.256637,2.584774,1.046703,
        -1.256637,2.601432,2.102535,
        -1.256637,2.603559,2.555114,
        -1.256637,2.604150,-2.885208,
        -1.256637,2.608630,-1.378011,
        -1.256637,2.611256,-0.279110,
        -1.256637,2.617344,-2.439480,
        -1.256637,2.627402,1.293552,
        -1.256637,2.635396,1.848850,
        -1.256637,2.635686,2.834950,
        -1.256637,2.637087,3.132730,
        -1.256637,2.642029,0.813274,
        -1.256637,2.645260,1.570203,
        -1.256637,2.646182,-1.614036,
        -1.256637,2.660113,-2.186513,
        -1.256637,2.660483,-0.034608,
        -1.256637,2.661713,-0.529511,
        -1.256637,2.672720,-1.110422,
        -1.256637,2.676626,2.335353,
        -1.256637,2.676638,0.540364,
        -1.256637,2.676780,-1.901158,
        -1.256637,2.681411,-2.667965,
        -1.256637,2.682152,0.245751,
        -1.256637,2.682859,-0.817071,
        -1.256637,2.705107,1.062142,
        -1.256637,2.717705,2.625995,
        -1.256637,2.721001,2.060097,
        -1.256637,2.722319,-2.954069,
        -1.256637,2.729840,-1.368237,
        -1.256637,2.734349,-0.289298,
        -1.256637,2.738409,2.975671,
        -1.256637,2.740366,1.374341,
        -1.256637,2.744633,-2.391311,
        -1.256637,2.747398,1.727166,
        -1.256637,2.763799,0.760372,
        -1.256637,2.768238,-1.679243,
        -1.256637,2.779245,-2.046268,
        -1.256637,2.780208,-0.626453,
        -1.256637,2.780471,0.027732,
        -1.256637,2.784696,-1.013128,
        -1.256637,2.789817,0.395431,
        -1.256637,2.801154,2.353143,
        -1.256637,2.807091,-2.687477,
        -1.256637,2.825762,1.085035,
        -1.256637,2.830209,2.734943,
        -1.256637,2.833517,-3.085508,
        -1.256637,2.838301,1.980061,
        -1.256637,2.849871,1.512012,
        -1.256637,2.850089,-1.343899,
        -1.256637,2.856905,-0.318441,
        -1.256637,2.870445,-2.291001,
        -1.256637,2.885017,0.655562,
        -1.256637,2.886428,-1.793217,
        -1.256637,2.890337,-0.825633,
        -1.256637,2.898932,0.139973,
        -1.256637,2.931927,-2.728768,
        -1.256637,2.932096,2.377790,
        -1.256637,2.937409,2.947245,
        -1.256637,2.948048,1.133127,
        -1.256637,2.950896,1.797889,
        -1.256637,2.970019,-1.282382,
        -1.256637,2.978333,-0.412090,
        -1.256637,2.996437,-2.089121,
        -1.256637,3.011274,0.434110,
        -1.256637,3.050410,3.138492,
        -1.256637,3.064492,1.647259,
        -1.256637,3.088868,-0.989501,
        -1.151917,0.053884,-2.408394,
        -1.151917,0.077645,1.453433,
        -1.151917,0.088408,-0.311160,
        -1.151917,0.142314,2.590387,
        -1.151917,0.147206,-1.385164,
        -1.151917,0.163506,0.573203,
        -1.151917,0.165485,-2.875396,
        -1.151917,0.182564,-2.125821,
        -1.151917,0.190569,1.929141,
        -1.151917,0.201102,1.249082,
        -1.151917,0.201397,-0.716646,
        -1.151917,0.210568,-0.069559,
        -1.151917,0.251672,2.848900,
        -1.151917,0.266262,-1.678068,
        -1.151917,0.269108,-1.215563,
        -1.151917,0.273777,-2.506756,
        -1.151917,0.279854,0.814581,
        -1.151917,0.281664,-2.969147,
        -1.151917,0.283625,2.405603,
        -1.151917,0.284058,0.357350,
        -1.151917,0.294107,1.593880,
        -1.151917,0.302491,-0.400410,
        -1.151917,0.308059,2.021491,
        -1.151917,0.312219,-2.072057,
        -1.151917,0.321905,-0.829121,
        -1.151917,0.329984,1.201621,
        -1.151917,0.339714,-0.013712,
        -1.151917,0.368784,2.974974,
        -1.151917,0.376428,-1.495532,
        -1.151917,0.378103,2.646436,
        -1.151917,0.385088,0.590355,
        -1.151917,0.386248,-2.690943,
        -1.151917,0.389851,-2.340031,
        -1.151917,0.391502,-1.171374,
        -1.151917,0.398249,-1.812694,
        -1.151917,0.400542,-2.995729,
        -1.151917,0.404273,0.918163,
        -1.151917,0.405617,1.773533,
        -1.151917,0.409009,-0.570243,
        -1.151917,0.410700,0.270433,
        -1.151917,0.415893,1.464890,
        -1.151917,0.420680,2.088164,
        -1.151917,0.424766,-0.258729,
        -1.151917,0.426507,2.367872,
        -1.151917,0.441292,-0.890727,
        -1.151917,0.451839,-2.078193,
        -1.151917,0.459270,1.191562,
        -1.151917,0.470622,0.014002,
        -1.151917,0.481292,-2.529098,
        -1.151917,0.482460,2.849843,
        -1.151917,0.492205,-1.397963,
        -1.151917,0.493314,3.117264,
        -1.151917,0.497870,0.727828,
        -1.151917,0.499783,-1.654379,
        -1.151917,0.502651,0.463333,
        -1.151917,0.505463,2.594557,
        -1.151917,0.506104,-2.874910,
        -1.151917,0.513579,-1.147354,
        -1.151917,0.518569,1.644237,
        -1.151917,0.519001,1.900488,
        -1.151917,0.519277,-0.673809,
        -1.151917,0.524769,-0.421955,
        -1.151917,0.525564,-2.288838,
        -1.151917,0.532541,0.978737,
        -1.151917,0.534610,2.154808,
        -1.151917,0.535020,-1.891846,
        -1.151917,0.540208,1.400497,
        -1.151917,0.542006,0.228337,
        -1.151917,0.550762,-0.180771,
        -1.151917,0.560531,2.378474,
        -1.151917,0.561629,-0.924576,
        -1.151917,0.567465,-2.679392,
        -1.151917,0.588527,1.181755,
        -1.151917,0.591643,-2.098364,
        -1.151917,0.594963,2.997575,
        -1.151917,0.599643,-3.045235,
        -1.151917,0.599962,2.773556,
        -1.151917,0.602620,0.029270,
        -1.151917,0.605711,0.598454,
        -1.151917,0.607708,-1.541026,
        -1.151917,0.609360,-2.464379,
        -1.151917,0.610738,-1.326786,
        -1.151917,0.616968,0.820489,
        -1.151917,0.626379,1.773610,
        -1.151917,0.628626,-1.751180,
        -1.151917,0.630671,0.391949,
        -1.151917,0.632622,-0.737642,
        -1.151917,0.632708,2.566722,
        -1.151917,0.632797,1.989858,
        -1.151917,0.634455,-0.534790,
        -1.151917,0.636205,1.561170,
        -1.151917,0.637681,-1.123423,
        -1.151917,0.641441,-2.842109,
        -1.151917,0.644807,-0.327139,
        -1.151917,0.652440,2.197129,
        -1.151917,0.660897,-2.278593,
        -1.151917,0.664008,1.360091,
        -1.151917,0.670850,-1.945257,
        -1.151917,0.671396,1.004859,
        -1.151917,0.673985,0.203855,
        -1.151917,0.680326,-0.133370,
        -1.151917,0.685083,-0.942363,
        -1.151917,0.688204,-2.632204,
        -1.151917,0.689777,2.385011,
        -1.151917,0.700065,3.115430,
        -1.151917,0.704080,2.911636,
        -1.151917,0.712883,1.174216,
        -1.151917,0.715701,0.699906,
        -1.151917,0.718851,-1.442215,
        -1.151917,0.721019,2.721536,
        -1.151917,0.725300,-2.975134,
        -1.151917,0.727136,-1.633288,
        -1.151917,0.727399,-2.120567,
        -1.151917,0.730751,0.522488,
        -1.151917,0.732892,-1.254330,
        -1.151917,0.734333,-2.446184,
        -1.151917,0.735125,0.037270,
        -1.151917,0.737893,1.866463,
        -1.151917,0.740650,1.681075,
        -1.151917,0.745971,-0.781286,
        -1.151917,0.749432,-0.442559,
        -1.151917,0.750102,2.050636,
        -1.151917,0.751046,0.863821,
        -1.151917,0.757018,-0.608833,
        -1.151917,0.757046,-1.816019,
        -1.151917,0.757389,1.501184,
        -1.151917,0.759800,2.546181,
        -1.151917,0.759859,0.351923,
        -1.151917,0.761141,-2.781102,
        -1.151917,0.769035,-1.078574,
        -1.151917,0.770166,-0.266604,
        -1.151917,0.772218,2.227380,
        -1.151917,0.788382,1.331051,
        -1.151917,0.790269,-2.282409,
        -1.151917,0.803410,-1.984567,
        -1.151917,0.803813,0.191520,
        -1.151917,0.806832,3.022155,
        -1.151917,0.807476,-2.597675,
        -1.151917,0.812512,-0.107335,
        -1.151917,0.814161,1.010949,
        -1.151917,0.817732,-3.084486,
        -1.151917,0.817846,2.390688,
        -1.151917,0.819323,2.845960,
        -1.151917,0.819430,-0.915017,
        -1.151917,0.829198,-1.532491,
        -1.151917,0.829531,0.665695,
        -1.151917,0.829816,1.167147,
        -1.151917,0.833613,-1.357492,
        -1.151917,0.842945,-2.907318,
        -1.151917,0.844883,2.679708,
        -1.151917,0.848189,-1.705003,
        -1.151917,0.849888,1.767901,
        -1.151917,0.851126,0.493370,
        -1.151917,0.852911,1.933394,
        -1.151917,0.857516,-2.432863,
        -1.151917,0.858261,-1.191729,
        -1.151917,0.858279,-2.138895,
        -1.151917,0.859723,-0.734928,
        -1.151917,0.859754,1.603292,
        -1.151917,0.861655,-0.394821,
        -1.151917,0.864300,0.043549,
        -1.151917,0.866940,0.851229,
        -1.151917,0.869748,2.095969,
        -1.151917,0.877352,-0.574668,
        -1.151917,0.879254,-2.733338,
        -1.151917,0.882554,1.443757,
        -1.151917,0.884060,0.334199,
        -1.151917,0.885626,-1.866061,
        -1.151917,0.889834,2.529957,
        -1.151917,0.893140,2.251698,
        -1.151917,0.897463,-0.240822,
        -1.151917,0.898758,-1.031948,
        -1.151917,0.913918,1.287618,
        -1.151917,0.916555,-2.287173,
        -1.151917,0.918305,3.107845,
        -1.151917,0.919522,2.948277,
        -1.151917,0.927138,-2.569686,
        -1.151917,0.929031,0.185383,
        -1.151917,0.933766,-3.015917,
        -1.151917,0.934873,-2.010822,
        -1.151917,0.935636,-1.445855,
        -1.151917,0.936702,-0.866461,
        -1.151917,0.937395,0.729108,
        -1.151917,0.937910,2.787575,
        -1.151917,0.939144,-0.092962,
        -1.151917,0.941718,1.125085,
        -1.151917,0.942873,-1.607680,
        -1.151917,0.943701,0.980328,
        -1.151917,0.947774,2.395256,
        -1.151917,0.952433,0.584347,
        -1.151917,0.954456,-1.287293,
        -1.151917,0.958731,-2.852552,
        -1.151917,0.961838,1.839284,
        -1.151917,0.961898,-0.469425,
        -1.151917,0.965426,1.686615,
        -1.151917,0.970108,1.990771,
        -1.151917,0.972820,-1.762780,
        -1.151917,0.973500,-0.711881,
        -1.151917,0.975253,2.641082,
        -1.151917,0.977824,0.441928,
        -1.151917,0.981485,-2.424766,
        -1.151917,0.981877,1.537100,
        -1.151917,0.982609,-1.136920,
        -1.151917,0.984265,-2.156073,
        -1.151917,0.989150,0.045493,
        -1.151917,0.993052,2.137478,
        -1.151917,0.997766,-2.693871,
        -1.151917,0.999847,-0.334738,
        -1.151917,1.001920,0.848054,
        -1.151917,1.006146,1.389734,
        -1.151917,1.012180,0.304066,
        -1.151917,1.016938,2.270674,
        -1.151917,1.017371,-0.979790,
        -1.151917,1.019495,-1.894487,
        -1.151917,1.023796,-0.583691,
        -1.151917,1.026525,2.506945,
        -1.151917,1.027004,-0.203653,
        -1.151917,1.028027,3.036573,
        -1.151917,1.031435,1.241021,
        -1.151917,1.032316,2.884221,
        -1.151917,1.035863,-3.110475,
        -1.151917,1.042085,-2.292950,
        -1.151917,1.043272,-1.522986,
        -1.151917,1.046265,-2.958847,
        -1.151917,1.049272,-1.376344,
        -1.151917,1.050039,-2.552086,
        -1.151917,1.053682,0.689863,
        -1.151917,1.055458,0.170394,
        -1.151917,1.056622,-0.831866,
        -1.151917,1.057515,1.097998,
        -1.151917,1.058706,-2.036438,
        -1.151917,1.061389,-1.672248,
        -1.151917,1.063758,2.743093,
        -1.151917,1.070157,-0.073422,
        -1.151917,1.072154,1.761506,
        -1.151917,1.074029,1.909193,
        -1.151917,1.074204,0.545276,
        -1.151917,1.074714,-2.805147,
        -1.151917,1.076262,0.966104,
        -1.151917,1.077642,-1.227073,
        -1.151917,1.084151,1.618525,
        -1.151917,1.085085,2.377673,
        -1.151917,1.086822,-0.441112,
        -1.151917,1.087142,2.058815,
        -1.151917,1.098383,-0.696430,
        -1.151917,1.099138,-1.084033,
        -1.151917,1.103208,1.477773,
        -1.151917,1.103783,0.408254,
        -1.151917,1.106275,-2.420800,
        -1.151917,1.108079,2.612943,
        -1.151917,1.109345,-2.172718,
        -1.151917,1.115957,-1.787597,
        -1.151917,1.118930,0.045983,
        -1.151917,1.119777,0.834702,
        -1.151917,1.121943,-2.669042,
        -1.151917,1.124930,2.219917,
        -1.151917,1.125217,1.336294,
        -1.151917,1.130958,2.972931,
        -1.151917,1.131400,-0.306629,
        -1.151917,1.137135,-0.936514,
        -1.151917,1.139125,-1.924528,
        -1.151917,1.139877,0.276382,
        -1.151917,1.142003,-3.058022,
        -1.151917,1.147363,-0.558448,
        -1.151917,1.150348,-0.176418,
        -1.151917,1.151198,1.197016,
        -1.151917,1.153930,-1.597104,
        -1.151917,1.155389,2.837148,
        -1.151917,1.155458,-1.460623,
        -1.151917,1.157331,-2.906362,
        -1.151917,1.158594,2.487618,
        -1.151917,1.162103,-1.322553,
        -1.151917,1.165407,3.092861,
        -1.151917,1.168025,0.717075,
        -1.151917,1.168406,-2.301519,
        -1.151917,1.174782,-2.538755,
        -1.151917,1.177998,1.061256,
        -1.151917,1.178898,1.827305,
        -1.151917,1.181331,-2.060330,
        -1.151917,1.182540,-0.798952,
        -1.151917,1.184569,0.147517,
        -1.151917,1.184734,1.968919,
        -1.151917,1.184804,0.593386,
        -1.151917,1.188278,1.690808,
        -1.151917,1.194360,2.110440,
        -1.151917,1.194776,2.710455,
        -1.151917,1.197745,-1.173705,
        -1.151917,1.199898,-2.777257,
        -1.151917,1.201164,2.355931,
        -1.151917,1.203760,1.556404,
        -1.151917,1.205845,-0.415993,
        -1.151917,1.206585,0.933648,
        -1.151917,1.206950,-0.037248,
        -1.151917,1.209341,0.471162,
        -1.151917,1.213483,-1.039417,
        -1.151917,1.218146,-0.672643,
        -1.151917,1.221608,1.421669,
        -1.151917,1.226043,-1.693533,
        -1.151917,1.232870,-1.822773,
        -1.151917,1.234375,-2.421194,
        -1.151917,1.234531,-2.190148,
        -1.151917,1.239414,0.348959,
        -1.151917,1.240183,2.587255,
        -1.151917,1.242312,2.941234,
        -1.151917,1.244075,1.286644,
        -1.151917,1.245961,-2.997461,
        -1.151917,1.246495,-2.646872,
        -1.151917,1.247736,2.242810,
        -1.151917,1.253604,0.812222,
        -1.151917,1.257816,-0.894317,
        -1.151917,1.258247,-1.953634,
        -1.151917,1.258993,-0.284099,
        -1.151917,1.262380,-0.157609,
        -1.151917,1.262791,-1.549229,
        -1.151917,1.263772,-0.533073,
        -1.151917,1.266731,-1.412461,
        -1.151917,1.268591,-3.122146,
        -1.151917,1.269614,-1.275246,
        -1.151917,1.270745,1.152533,
        -1.151917,1.273606,0.224545,
        -1.151917,1.279716,2.812101,
        -1.151917,1.283786,2.462124,
        -1.151917,1.284798,0.078209,
        -1.151917,1.285402,-2.882049,
        -1.151917,1.286936,1.886053,
        -1.151917,1.289408,0.683122,
        -1.151917,1.294736,1.756184,
        -1.151917,1.296978,-2.311511,
        -1.151917,1.298546,2.020601,
        -1.151917,1.300924,1.022478,
        -1.151917,1.303305,3.043152,
        -1.151917,1.304210,-2.084225,
        -1.151917,1.305241,-2.530249,
        -1.151917,1.307513,1.628003,
        -1.151917,1.310722,0.558267,
        -1.151917,1.314142,2.139329,
        -1.151917,1.316087,-0.777971,
        -1.151917,1.317907,-1.121023,
        -1.151917,1.321019,-2.749776,
        -1.151917,1.321609,2.687717,
        -1.151917,1.322568,1.499415,
        -1.151917,1.322972,-0.395699,
        -1.151917,1.325286,-0.994192,
        -1.151917,1.329146,-0.648513,
        -1.151917,1.329795,-0.042570,
        -1.151917,1.331651,2.344497,
        -1.151917,1.336522,0.435098,
        -1.151917,1.339410,0.900087,
        -1.151917,1.339550,1.370353,
        -1.151917,1.340441,-1.731580,
        -1.151917,1.351855,-1.855234,
        -1.151917,1.360161,-3.018404,
        -1.151917,1.361031,-1.613971,
        -1.151917,1.361196,2.914780,
        -1.151917,1.361998,-2.207254,
        -1.151917,1.362302,1.239384,
        -1.151917,1.365228,2.563808,
        -1.151917,1.365452,0.310522,
        -1.151917,1.367028,-2.420041,
        -1.151917,1.373422,-1.484281,
        -1.151917,1.374245,-1.226417,
        -1.151917,1.375020,-0.170393,
        -1.151917,1.376560,-1.355481,
        -1.151917,1.377070,-1.980984,
        -1.151917,1.377578,-2.636339,
        -1.151917,1.378096,0.774626,
        -1.151917,1.379086,-0.508192,
        -1.151917,1.382562,0.177141,
        -1.151917,1.388849,2.233847,
        -1.151917,1.389268,-0.886349,
        -1.151917,1.391324,1.109396,
        -1.151917,1.391699,-0.288725,
        -1.151917,1.392869,3.136903,
        -1.151917,1.393612,1.952016,
        -1.151917,1.395545,-2.850807,
        -1.151917,1.401926,1.822930,
        -1.151917,1.402149,2.788898,
        -1.151917,1.409552,0.647980,
        -1.151917,1.410840,0.056219,
        -1.151917,1.411944,2.444377,
        -1.151917,1.413639,1.697982,
        -1.151917,1.426479,1.573427,
        -1.151917,1.426725,2.102242,
        -1.151917,1.427868,0.984637,
        -1.151917,1.429580,-2.107232,
        -1.151917,1.431375,-2.313269,
        -1.151917,1.431907,3.015410,
        -1.151917,1.432325,-1.029054,
        -1.151917,1.433341,0.524268,
        -1.151917,1.435181,-0.739771,
        -1.151917,1.437208,-2.529085,
        -1.151917,1.439041,-0.617966,
        -1.151917,1.441473,1.448544,
        -1.151917,1.444527,2.664744,
        -1.151917,1.450513,-0.400625,
        -1.151917,1.451907,-0.066823,
        -1.151917,1.452754,-2.738631,
        -1.151917,1.454061,-2.950416,
        -1.151917,1.456736,1.323154,
        -1.151917,1.458370,0.400433,
        -1.151917,1.459314,-1.765003,
        -1.151917,1.459902,-1.145835,
        -1.151917,1.464952,2.329245,
        -1.151917,1.466050,0.862104,
        -1.151917,1.471330,-1.886160,
        -1.151917,1.475119,-1.648255,
        -1.151917,1.480878,1.194576,
        -1.151917,1.480942,0.273235,
        -1.151917,1.481086,2.890269,
        -1.151917,1.484118,-1.529156,
        -1.151917,1.484741,-1.279219,
        -1.151917,1.487681,-1.406069,
        -1.151917,1.487886,-3.066037,
        -1.151917,1.490555,2.543284,
        -1.151917,1.492424,-0.190932,
        -1.151917,1.495002,-0.839810,
        -1.151917,1.495297,2.007188,
        -1.151917,1.496414,-2.009267,
        -1.151917,1.498261,-2.425270,
        -1.151917,1.498907,-2.207327,
        -1.151917,1.500204,0.737638,
        -1.151917,1.502469,0.151156,
        -1.151917,1.507448,-0.514782,
        -1.151917,1.509287,2.204647,
        -1.151917,1.510518,1.884920,
        -1.151917,1.511948,-2.632360,
        -1.151917,1.515235,1.068565,
        -1.151917,1.522237,3.102018,
        -1.151917,1.522373,1.764991,
        -1.151917,1.522416,-0.954956,
        -1.151917,1.523102,2.765920,
        -1.151917,1.524811,-2.844718,
        -1.151917,1.527261,-0.304916,
        -1.151917,1.529788,0.613208,
        -1.151917,1.533648,1.644808,
        -1.151917,1.535018,0.028869,
        -1.151917,1.541082,2.424608,
        -1.151917,1.546010,1.523462,
        -1.151917,1.551135,-0.642994,
        -1.151917,1.553895,0.489607,
        -1.151917,1.555041,0.947813,
        -1.151917,1.559671,-1.074934,
        -1.151917,1.560345,2.985899,
        -1.151917,1.560487,-2.323400,
        -1.151917,1.561127,1.401111,
        -1.151917,1.567193,2.643172,
        -1.151917,1.568892,2.098362,
        -1.151917,1.571702,-2.098435,
        -1.151917,1.572701,-2.528730,
        -1.151917,1.573572,1.278593,
        -1.151917,1.574849,-0.091505,
        -1.151917,1.576334,0.365236,
        -1.151917,1.577029,-2.962888,
        -1.151917,1.577688,-1.814941,
        -1.151917,1.579542,-0.755917,
        -1.151917,1.579624,-1.199759,
        -1.151917,1.581515,-0.419414,
        -1.151917,1.584581,-2.738702,
        -1.151917,1.586948,-1.696296,
        -1.151917,1.588901,2.301496,
        -1.151917,1.589936,0.825020,
        -1.151917,1.590018,-1.941652,
        -1.151917,1.596968,-1.328358,
        -1.151917,1.597178,-1.575607,
        -1.151917,1.600470,-1.453196,
        -1.151917,1.601408,0.242409,
        -1.151917,1.602259,1.152656,
        -1.151917,1.603064,2.865774,
        -1.151917,1.612085,1.968802,
        -1.151917,1.612720,-3.080498,
        -1.151917,1.616061,2.521531,
        -1.151917,1.616900,-0.207457,
        -1.151917,1.620988,-0.875920,
        -1.151917,1.622114,0.701509,
        -1.151917,1.622749,0.121426,
        -1.151917,1.622993,-2.221629,
        -1.151917,1.629346,-0.542669,
        -1.151917,1.630007,1.844552,
        -1.151917,1.634747,-2.426198,
        -1.151917,1.642346,1.722617,
        -1.151917,1.643925,1.034629,
        -1.151917,1.644848,2.743774,
        -1.151917,1.645258,2.188109,
        -1.151917,1.645417,-2.634725,
        -1.151917,1.647395,3.085437,
        -1.151917,1.648538,-2.855044,
        -1.151917,1.649932,0.576808,
        -1.151917,1.652224,-0.996105,
        -1.151917,1.652871,1.600681,
        -1.151917,1.657997,-0.325127,
        -1.151917,1.661936,0.001037,
        -1.151917,1.665187,1.477402,
        -1.151917,1.665496,2.397872,
        -1.151917,1.673267,-0.665453,
        -1.151917,1.673557,0.452526,
        -1.151917,1.679277,-1.122421,
        -1.151917,1.679696,0.912482,
        -1.151917,1.682420,1.353922,
        -1.151917,1.685031,-2.009038,
        -1.151917,1.685196,2.967046,
        -1.151917,1.690107,2.620910,
        -1.151917,1.690453,-2.123076,
        -1.151917,1.691693,-1.871572,
        -1.151917,1.691817,2.061688,
        -1.151917,1.692644,1.237670,
        -1.151917,1.696386,0.328123,
        -1.151917,1.697162,-1.249026,
        -1.151917,1.697488,-2.322785,
        -1.151917,1.698726,-2.973576,
        -1.151917,1.699172,-1.749717,
        -1.151917,1.704478,-0.106295,
        -1.151917,1.707244,-0.444770,
        -1.151917,1.707500,-2.531497,
        -1.151917,1.707714,-0.787321,
        -1.151917,1.708430,-1.627552,
        -1.151917,1.710583,-1.376890,
        -1.151917,1.712415,-2.747921,
        -1.151917,1.713125,-1.503073,
        -1.151917,1.713452,0.789419,
        -1.151917,1.720612,2.279888,
        -1.151917,1.724852,0.202831,
        -1.151917,1.724868,2.846246,
        -1.151917,1.728940,1.931535,
        -1.151917,1.732838,1.127584,
        -1.151917,1.735719,-3.092485,
        -1.151917,1.739676,2.496017,
        -1.151917,1.739971,-0.228876,
        -1.151917,1.743328,0.663347,
        -1.151917,1.744927,-0.911892,
        -1.151917,1.745952,0.084733,
        -1.151917,1.748077,1.804354,
        -1.151917,1.752959,-0.567548,
        -1.151917,1.760169,1.678132,
        -1.151917,1.763887,-2.218450,
        -1.151917,1.766737,2.722752,
        -1.151917,1.767999,1.004737,
        -1.151917,1.769149,1.550298,
        -1.151917,1.770119,-2.427175,
        -1.151917,1.770460,2.154208,
        -1.151917,1.770514,3.070982,
        -1.151917,1.770735,0.537245,
        -1.151917,1.772673,-2.864760,
        -1.151917,1.773527,-1.039582,
        -1.151917,1.776419,-2.642421,
        -1.151917,1.780769,1.416330,
        -1.151917,1.786634,-0.346247,
        -1.151917,1.788829,-1.947553,
        -1.151917,1.793737,0.410270,
        -1.151917,1.794553,2.374618,
        -1.151917,1.794699,-0.694061,
        -1.151917,1.797469,-1.169105,
        -1.151917,1.803705,0.881076,
        -1.151917,1.804124,1.264249,
        -1.151917,1.807825,-2.090455,
        -1.151917,1.808822,-1.817375,
        -1.151917,1.809020,2.949741,
        -1.151917,1.812466,2.597595,
        -1.151917,1.813159,2.025395,
        -1.151917,1.813816,-1.298435,
        -1.151917,1.815016,0.278951,
        -1.151917,1.816510,-0.008631,
        -1.151917,1.817772,-1.689625,
        -1.151917,1.821360,-2.984352,
        -1.151917,1.822312,-0.127211,
        -1.151917,1.823386,-1.429807,
        -1.151917,1.823834,-1.560483,
        -1.151917,1.832060,-0.821627,
        -1.151917,1.832690,-0.468440,
        -1.151917,1.833945,-2.320278,
        -1.151917,1.834787,0.751968,
        -1.151917,1.840579,-2.536445,
        -1.151917,1.841033,-2.756932,
        -1.151917,1.846350,1.893983,
        -1.151917,1.846644,1.113766,
        -1.151917,1.846833,2.826543,
        -1.151917,1.847325,2.249051,
        -1.151917,1.854119,0.134161,
        -1.151917,1.858863,-3.106022,
        -1.151917,1.865046,1.763384,
        -1.151917,1.865619,-0.952036,
        -1.151917,1.865912,0.624453,
        -1.151917,1.866694,2.473686,
        -1.151917,1.867524,-0.240083,
        -1.151917,1.875561,1.632831,
        -1.151917,1.876049,-0.594378,
        -1.151917,1.880509,1.500204,
        -1.151917,1.883504,1.360457,
        -1.151917,1.886971,-2.196665,
        -1.151917,1.888979,0.982922,
        -1.151917,1.889815,2.701570,
        -1.151917,1.892144,-2.005340,
        -1.151917,1.892348,0.496522,
        -1.151917,1.893281,-1.083976,
        -1.151917,1.893828,3.053453,
        -1.151917,1.895117,2.121107,
        -1.151917,1.898741,-2.875735,
        -1.151917,1.904429,-2.428029,
        -1.151917,1.908408,-2.648903,
        -1.151917,1.911828,-0.364489,
        -1.151917,1.912224,0.366565,
        -1.151917,1.914854,1.214468,
        -1.151917,1.915467,-1.883143,
        -1.151917,1.915763,-0.724679,
        -1.151917,1.915790,-1.216366,
        -1.151917,1.922256,2.347806,
        -1.151917,1.922616,0.844633,
        -1.151917,1.925185,0.231170,
        -1.151917,1.927474,-1.754616,
        -1.151917,1.928342,-1.351174,
        -1.151917,1.933328,-1.623328,
        -1.151917,1.933878,0.034150,
        -1.151917,1.934585,-1.487979,
        -1.151917,1.934802,2.927634,
        -1.151917,1.934862,1.989776,
        -1.151917,1.936327,-0.099350,
        -1.151917,1.938088,2.578507,
        -1.151917,1.945677,-2.999805,
        -1.151917,1.954572,-0.859208,
        -1.151917,1.957868,-0.490784,
        -1.151917,1.958989,0.712772,
        -1.151917,1.960212,-2.307561,
        -1.151917,1.964774,1.855507,
        -1.151917,1.964971,-2.100533,
        -1.151917,1.969112,1.088273,
        -1.151917,1.971438,-2.765415,
        -1.151917,1.972312,2.805134,
        -1.151917,1.974950,2.220596,
        -1.151917,1.975135,-2.539127,
        -1.151917,1.982107,1.720829,
        -1.151917,1.983195,-3.127984,
        -1.151917,1.983318,-0.247182,
        -1.151917,1.986720,-0.993220,
        -1.151917,1.990381,1.586005,
        -1.151917,1.990470,0.580283,
        -1.151917,1.990523,1.450326,
        -1.151917,1.993909,1.312897,
        -1.151917,1.994694,2.452813,
        -1.151917,1.999226,-0.620717,
        -1.151917,2.005145,0.941012,
        -1.151917,2.007301,0.142337,
        -1.151917,2.014149,0.444138,
        -1.151917,2.014338,-1.127392,
        -1.151917,2.018131,3.021778,
        -1.151917,2.019676,-1.980944,
        -1.151917,2.019912,2.686612,
        -1.151917,2.020545,2.088672,
        -1.151917,2.026074,0.300945,
        -1.151917,2.026196,-2.889057,
        -1.151917,2.031959,-1.268178,
        -1.151917,2.032917,-2.199917,
        -1.151917,2.033572,-1.842884,
        -1.151917,2.034712,-2.419505,
        -1.151917,2.035818,-0.379479,
        -1.151917,2.036004,-0.757062,
        -1.151917,2.040501,-1.411488,
        -1.151917,2.042874,-1.701367,
        -1.151917,2.043060,-2.653746,
        -1.151917,2.043729,-1.557938,
        -1.151917,2.048538,0.806623,
        -1.151917,2.049006,1.199930,
        -1.151917,2.049502,-0.137989,
        -1.151917,2.052181,2.326959,
        -1.151917,2.052794,-0.008526,
        -1.151917,2.057608,1.951726,
        -1.151917,2.065568,2.567730,
        -1.151917,2.070371,2.879808,
        -1.151917,2.071256,-3.018961,
        -1.151917,2.079746,-0.897699,
        -1.151917,2.082784,1.042403,
        -1.151917,2.083752,-0.511086,
        -1.151917,2.084401,1.810502,
        -1.151917,2.085731,0.671081,
        -1.151917,2.097496,1.373172,
        -1.151917,2.099185,1.667006,
        -1.151917,2.100482,-2.301084,
        -1.151917,2.103118,2.194745,
        -1.151917,2.103223,1.521685,
        -1.151917,2.103695,-2.777707,
        -1.151917,2.107523,3.129581,
        -1.151917,2.108277,-2.095907,
        -1.151917,2.108891,-2.532166,
        -1.151917,2.111003,-0.268632,
        -1.151917,2.112276,-1.027246,
        -1.151917,2.114380,0.531457,
        -1.151917,2.119166,0.100552,
        -1.151917,2.119564,2.754818,
        -1.151917,2.122474,-0.648314,
        -1.151917,2.125183,2.445598,
        -1.151917,2.129501,0.238290,
        -1.151917,2.131178,0.386663,
        -1.151917,2.131223,-1.174311,
        -1.151917,2.134269,-1.948042,
        -1.151917,2.134684,0.907065,
        -1.151917,2.137615,2.988806,
        -1.151917,2.145635,-1.327708,
        -1.151917,2.146467,2.055081,
        -1.151917,2.146582,-1.793788,
        -1.151917,2.149210,-1.482321,
        -1.151917,2.152901,-1.636788,
        -1.151917,2.156337,1.151558,
        -1.151917,2.156474,-2.912065,
        -1.151917,2.157194,-0.792563,
        -1.151917,2.166724,-0.401638,
        -1.151917,2.166980,-0.134879,
        -1.151917,2.173644,-2.403429,
        -1.151917,2.175711,-2.655523,
        -1.151917,2.177132,2.600419,
        -1.151917,2.178879,0.767722,
        -1.151917,2.180752,1.908485,
        -1.151917,2.181338,2.313721,
        -1.151917,2.185580,1.287673,
        -1.151917,2.195400,-2.217546,
        -1.151917,2.200418,-3.056442,
        -1.151917,2.201829,-0.001735,
        -1.151917,2.203804,1.756176,
        -1.151917,2.211602,1.443930,
        -1.151917,2.212686,-0.544809,
        -1.151917,2.213937,0.621892,
        -1.151917,2.214075,1.600691,
        -1.151917,2.216256,1.020029,
        -1.151917,2.219060,2.886631,
        -1.151917,2.223375,-0.906185,
        -1.151917,2.226002,-1.060248,
        -1.151917,2.226413,-2.064034,
        -1.151917,2.230554,2.171334,
        -1.151917,2.232413,2.735777,
        -1.151917,2.233709,3.070830,
        -1.151917,2.234383,-2.792065,
        -1.151917,2.237183,0.468799,
        -1.151917,2.237695,-0.271388,
        -1.151917,2.239419,0.151693,
        -1.151917,2.243347,-2.521049,
        -1.151917,2.244035,-1.229456,
        -1.151917,2.244355,0.309899,
        -1.151917,2.245175,2.460236,
        -1.151917,2.245628,-1.899216,
        -1.151917,2.245973,-0.702234,
        -1.151917,2.255601,-1.728131,
        -1.151917,2.256702,-1.401367,
        -1.151917,2.261466,-1.557500,
        -1.151917,2.267390,0.878499,
        -1.151917,2.272099,2.018594,
        -1.151917,2.278618,1.170386,
        -1.151917,2.283304,-2.345584,
        -1.151917,2.285175,-2.943090,
        -1.151917,2.293000,-0.124451,
        -1.151917,2.294058,-0.421070,
        -1.151917,2.303283,1.855938,
        -1.151917,2.304968,2.608978,
        -1.151917,2.305347,2.311346,
        -1.151917,2.307188,-2.656134,
        -1.151917,2.307301,1.336200,
        -1.151917,2.309655,0.726020,
        -1.151917,2.317634,-2.185225,
        -1.151917,2.321510,1.685664,
        -1.151917,2.324376,-3.111006,
        -1.151917,2.324731,1.511186,
        -1.151917,2.329587,0.035498,
        -1.151917,2.330158,2.977264,
        -1.151917,2.335096,-0.588468,
        -1.151917,2.336046,-1.115799,
        -1.151917,2.338060,2.803613,
        -1.151917,2.338083,-0.943400,
        -1.151917,2.340210,0.562145,
        -1.151917,2.341931,1.023926,
        -1.151917,2.343265,-0.779061,
        -1.151917,2.343434,-2.013692,
        -1.151917,2.353029,-1.310529,
        -1.151917,2.354868,0.211543,
        -1.151917,2.355120,0.388593,
        -1.151917,2.355165,-1.833173,
        -1.151917,2.355462,2.151441,
        -1.151917,2.362948,-2.487477,
        -1.151917,2.363101,-2.810456,
        -1.151917,2.363394,-0.272918,
        -1.151917,2.364509,-1.644533,
        -1.151917,2.370465,2.466824,
        -1.151917,2.386110,-1.481351,
        -1.151917,2.394960,0.862133,
        -1.151917,2.395716,1.977184,
        -1.151917,2.396339,1.200900,
        -1.151917,2.407941,-2.316753,
        -1.151917,2.409450,-2.988432,
        -1.151917,2.417202,-0.103363,
        -1.151917,2.417386,-0.443097,
        -1.151917,2.417950,2.666662,
        -1.151917,2.423190,1.789783,
        -1.151917,2.423692,1.392696,
        -1.151917,2.431122,3.086379,
        -1.151917,2.431406,-2.647820,
        -1.151917,2.431584,2.307631,
        -1.151917,2.434090,1.592818,
        -1.151917,2.435993,0.683844,
        -1.151917,2.440898,-2.133300,
        -1.151917,2.442222,-1.195837,
        -1.151917,2.443112,2.883761,
        -1.151917,2.449191,-0.633651,
        -1.151917,2.450821,0.084741,
        -1.151917,2.453499,-1.008861,
        -1.151917,2.460010,-1.938011,
        -1.151917,2.460199,-0.822008,
        -1.151917,2.460902,0.490079,
        -1.151917,2.461774,-1.734913,
        -1.151917,2.463986,1.034407,
        -1.151917,2.467647,0.287667,
        -1.151917,2.479849,2.126925,
        -1.151917,2.485907,-2.835471,
        -1.151917,2.485945,-1.374853,
        -1.151917,2.487747,2.505912,
        -1.151917,2.487957,-0.274279,
        -1.151917,2.489741,-2.468428,
        -1.151917,2.512538,1.239900,
        -1.151917,2.517528,1.923289,
        -1.151917,2.518454,-1.565852,
        -1.151917,2.519348,0.843503,
        -1.151917,2.522766,-3.059231,
        -1.151917,2.528686,2.737228,
        -1.151917,2.534798,-2.266671,
        -1.151917,2.536533,1.466274,
        -1.151917,2.538504,1.699812,
        -1.151917,2.539516,-0.476942,
        -1.151917,2.539532,-0.074548,
        -1.151917,2.543197,2.990252,
        -1.151917,2.554022,2.320619,
        -1.151917,2.556919,-2.655005,
        -1.151917,2.558146,0.627191,
        -1.151917,2.558223,-1.164189,
        -1.151917,2.563452,-2.044028,
        -1.151917,2.568462,0.150655,
        -1.151917,2.569130,-0.700837,
        -1.151917,2.570050,-1.804492,
        -1.151917,2.572710,-0.934665,
        -1.151917,2.575994,0.391718,
        -1.151917,2.584774,1.046703,
        -1.151917,2.601432,2.102535,
        -1.151917,2.603559,2.555114,
        -1.151917,2.604150,-2.885208,
        -1.151917,2.608630,-1.378011,
        -1.151917,2.611256,-0.279110,
        -1.151917,2.617344,-2.439480,
        -1.151917,2.627402,1.293552,
        -1.151917,2.635396,1.848850,
        -1.151917,2.635686,2.834950,
        -1.151917,2.637087,3.132730,
        -1.151917,2.642029,0.813274,
        -1.151917,2.645260,1.570203,
        -1.151917,2.646182,-1.614036,
        -1.151917,2.660113,-2.186513,
        -1.151917,2.660483,-0.034608,
        -1.151917,2.661713,-0.529511,
        -1.151917,2.672720,-1.110422,
        -1.151917,2.676626,2.335353,
        -1.151917,2.676638,0.540364,
        -1.151917,2.676780,-1.901158,
        -1.151917,2.681411,-2.667965,
        -1.151917,2.682152,0.245751,
        -1.151917,2.682859,-0.817071,
        -1.151917,2.705107,1.062142,
        -1.151917,2.717705,2.625995,
        -1.151917,2.721001,2.060097,
        -1.151917,2.722319,-2.954069,
        -1.151917,2.729840,-1.368237,
        -1.151917,2.734349,-0.289298,
        -1.151917,2.738409,2.975671,
        -1.151917,2.740366,1.374341,
        -1.151917,2.744633,-2.391311,
        -1.151917,2.747398,1.727166,
        -1.151917,2.763799,0.760372,
        -1.151917,2.768238,-1.679243,
        -1.151917,2.779245,-2.046268,
        -1.151917,2.780208,-0.626453,
        -1.151917,2.780471,0.027732,
        -1.151917,2.784696,-1.013128,
        -1.151917,2.789817,0.395431,
        -1.151917,2.801154,2.353143,
        -1.151917,2.807091,-2.687477,
        -1.151917,2.825762,1.085035,
        -1.151917,2.830209,2.734943,
        -1.151917,2.833517,-3.085508,
        -1.151917,2.838301,1.980061,
        -1.151917,2.849871,1.512012,
        -1.151917,2.850089,-1.343899,
        -1.151917,2.856905,-0.318441,
        -1.151917,2.870445,-2.291001,
        -1.151917,2.885017,0.655562,
        -1.151917,2.886428,-1.793217,
        -1.151917,2.890337,-0.825633,
        -1.151917,2.898932,0.139973,
        -1.151917,2.931927,-2.728768,
        -1.151917,2.932096,2.377790,
        -1.151917,2.937409,2.947245,
        -1.151917,2.948048,1.133127,
        -1.151917,2.950896,1.797889,
        -1.151917,2.970019,-1.282382,
        -1.151917,2.978333,-0.412090,
        -1.151917,2.996437,-2.089121,
        -1.151917,3.011274,0.434110,
        -1.151917,3.050410,3.138492,
        -1.151917,3.064492,1.647259,
        -1.151917,3.088868,-0.989501,
        -1.047198,0.053884,-2.408394,
        -1.047198,0.077645,1.453433,
        -1.047198,0.088408,-0.311160,
        -1.047198,0.142314,2.590387,
        -1.047198,0.147206,-1.385164,
        -1.047198,0.163506,0.573203,
        -1.047198,0.165485,-2.875396,
        -1.047198,0.182564,-2.125821,
        -1.047198,0.190569,1.929141,
        -1.047198,0.201102,1.249082,
        -1.047198,0.201397,-0.716646,
        -1.047198,0.210568,-0.069559,
        -1.047198,0.251672,2.848900,
        -1.047198,0.266262,-1.678068,
        -1.047198,0.269108,-1.215563,
        -1.047198,0.273777,-2.506756,
        -1.047198,0.279854,0.814581,
        -1.047198,0.281664,-2.969147,
        -1.047198,0.283625,2.405603,
        -1.047198,0.284058,0.357350,
        -1.047198,0.294107,1.593880,
        -1.047198,0.302491,-0.400410,
        -1.047198,0.308059,2.021491,
        -1.047198,0.312219,-2.072057,
        -1.047198,0.321905,-0.829121,
        -1.047198,0.329984,1.201621,
        -1.047198,0.339714,-0.013712,
        -1.047198,0.368784,2.974974,
        -1.047198,0.376428,-1.495532,
        -1.047198,0.378103,2.646436,
        -1.047198,0.385088,0.590355,
        -1.047198,0.386248,-2.690943,
        -1.047198,0.389851,-2.340031,
        -1.047198,0.391502,-1.171374,
        -1.047198,0.398249,-1.812694,
        -1.047198,0.400542,-2.995729,
        -1.047198,0.404273,0.918163,
        -1.047198,0.405617,1.773533,
        -1.047198,0.409009,-0.570243,
        -1.047198,0.410700,0.270433,
        -1.047198,0.415893,1.464890,
        -1.047198,0.420680,2.088164,
        -1.047198,0.424766,-0.258729,
        -1.047198,0.426507,2.367872,
        -1.047198,0.441292,-0.890727,
        -1.047198,0.451839,-2.078193,
        -1.047198,0.459270,1.191562,
        -1.047198,0.470622,0.014002,
        -1.047198,0.481292,-2.529098,
        -1.047198,0.482460,2.849843,
        -1.047198,0.492205,-1.397963,
        -1.047198,0.493314,3.117264,
        -1.047198,0.497870,0.727828,
        -1.047198,0.499783,-1.654379,
        -1.047198,0.502651,0.463333,
        -1.047198,0.505463,2.594557,
        -1.047198,0.506104,-2.874910,
        -1.047198,0.513579,-1.147354,
        -1.047198,0.518569,1.644237,
        -1.047198,0.519001,1.900488,
        -1.047198,0.519277,-0.673809,
        -1.047198,0.524769,-0.421955,
        -1.047198,0.525564,-2.288838,
        -1.047198,0.532541,0.978737,
        -1.047198,0.534610,2.154808,
        -1.047198,0.535020,-1.891846,
        -1.047198,0.540208,1.400497,
        -1.047198,0.542006,0.228337,
        -1.047198,0.550762,-0.180771,
        -1.047198,0.560531,2.378474,
        -1.047198,0.561629,-0.924576,
        -1.047198,0.567465,-2.679392,
        -1.047198,0.588527,1.181755,
        -1.047198,0.591643,-2.098364,
        -1.047198,0.594963,2.997575,
        -1.047198,0.599643,-3.045235,
        -1.047198,0.599962,2.773556,
        -1.047198,0.602620,0.029270,
        -1.047198,0.605711,0.598454,
        -1.047198,0.607708,-1.541026,
        -1.047198,0.609360,-2.464379,
        -1.047198,0.610738,-1.326786,
        -1.047198,0.616968,0.820489,
        -1.047198,0.626379,1.773610,
        -1.047198,0.628626,-1.751180,
        -1.047198,0.630671,0.391949,
        -1.047198,0.632622,-0.737642,
        -1.047198,0.632708,2.566722,
        -1.047198,0.632797,1.989858,
        -1.047198,0.634455,-0.534790,
        -1.047198,0.636205,1.561170,
        -1.047198,0.637681,-1.123423,
        -1.047198,0.641441,-2.842109,
        -1.047198,0.644807,-0.327139,
        -1.047198,0.652440,2.197129,
        -1.047198,0.660897,-2.278593,
        -1.047198,0.664008,1.360091,
        -1.047198,0.670850,-1.945257,
        -1.047198,0.671396,1.004859,
        -1.047198,0.673985,0.203855,
        -1.047198,0.680326,-0.133370,
        -1.047198,0.685083,-0.942363,
        -1.047198,0.688204,-2.632204,
        -1.047198,0.689777,2.385011,
        -1.047198,0.700065,3.115430,
        -1.047198,0.704080,2.911636,
        -1.047198,0.712883,1.174216,
        -1.047198,0.715701,0.699906,
        -1.047198,0.718851,-1.442215,
        -1.047198,0.721019,2.721536,
        -1.047198,0.725300,-2.975134,
        -1.047198,0.727136,-1.633288,
        -1.047198,0.727399,-2.120567,
        -1.047198,0.730751,0.522488,
        -1.047198,0.732892,-1.254330,
        -1.047198,0.734333,-2.446184,
        -1.047198,0.735125,0.037270,
        -1.047198,0.737893,1.866463,
        -1.047198,0.740650,1.681075,
        -1.047198,0.745971,-0.781286,
        -1.047198,0.749432,-0.442559,
        -1.047198,0.750102,2.050636,
        -1.047198,0.751046,0.863821,
        -1.047198,0.757018,-0.608833,
        -1.047198,0.757046,-1.816019,
        -1.047198,0.757389,1.501184,
        -1.047198,0.759800,2.546181,
        -1.047198,0.759859,0.351923,
        -1.047198,0.761141,-2.781102,
        -1.047198,0.769035,-1.078574,
        -1.047198,0.770166,-0.266604,
        -1.047198,0.772218,2.227380,
        -1.047198,0.788382,1.331051,
        -1.047198,0.790269,-2.282409,
        -1.047198,0.803410,-1.984567,
        -1.047198,0.803813,0.191520,
        -1.047198,0.806832,3.022155,
        -1.047198,0.807476,-2.597675,
        -1.047198,0.812512,-0.107335,
        -1.047198,0.814161,1.010949,
        -1.047198,0.817732,-3.084486,
        -1.047198,0.817846,2.390688,
        -1.047198,0.819323,2.845960,
        -1.047198,0.819430,-0.915017,
        -1.047198,0.829198,-1.532491,
        -1.047198,0.829531,0.665695,
        -1.047198,0.829816,1.167147,
        -1.047198,0.833613,-1.357492,
        -1.047198,0.842945,-2.907318,
        -1.047198,0.844883,2.679708,
        -1.047198,0.848189,-1.705003,
        -1.047198,0.849888,1.767901,
        -1.047198,0.851126,0.493370,
        -1.047198,0.852911,1.933394,
        -1.047198,0.857516,-2.432863,
        -1.047198,0.858261,-1.191729,
        -1.047198,0.858279,-2.138895,
        -1.047198,0.859723,-0.734928,
        -1.047198,0.859754,1.603292,
        -1.047198,0.861655,-0.394821,
        -1.047198,0.864300,0.043549,
        -1.047198,0.866940,0.851229,
        -1.047198,0.869748,2.095969,
        -1.047198,0.877352,-0.574668,
        -1.047198,0.879254,-2.733338,
        -1.047198,0.882554,1.443757,
        -1.047198,0.884060,0.334199,
        -1.047198,0.885626,-1.866061,
        -1.047198,0.889834,2.529957,
        -1.047198,0.893140,2.251698,
        -1.047198,0.897463,-0.240822,
        -1.047198,0.898758,-1.031948,
        -1.047198,0.913918,1.287618,
        -1.047198,0.916555,-2.287173,
        -1.047198,0.918305,3.107845,
        -1.047198,0.919522,2.948277,
        -1.047198,0.927138,-2.569686,
        -1.047198,0.929031,0.185383,
        -1.047198,0.933766,-3.015917,
        -1.047198,0.934873,-2.010822,
        -1.047198,0.935636,-1.445855,
        -1.047198,0.936702,-0.866461,
        -1.047198,0.937395,0.729108,
        -1.047198,0.937910,2.787575,
        -1.047198,0.939144,-0.092962,
        -1.047198,0.941718,1.125085,
        -1.047198,0.942873,-1.607680,
        -1.047198,0.943701,0.980328,
        -1.047198,0.947774,2.395256,
        -1.047198,0.952433,0.584347,
        -1.047198,0.954456,-1.287293,
        -1.047198,0.958731,-2.852552,
        -1.047198,0.961838,1.839284,
        -1.047198,0.961898,-0.469425,
        -1.047198,0.965426,1.686615,
        -1.047198,0.970108,1.990771,
        -1.047198,0.972820,-1.762780,
        -1.047198,0.973500,-0.711881,
        -1.047198,0.975253,2.641082,
        -1.047198,0.977824,0.441928,
        -1.047198,0.981485,-2.424766,
        -1.047198,0.981877,1.537100,
        -1.047198,0.982609,-1.136920,
        -1.047198,0.984265,-2.156073,
        -1.047198,0.989150,0.045493,
        -1.047198,0.993052,2.137478,
        -1.047198,0.997766,-2.693871,
        -1.047198,0.999847,-0.334738,
        -1.047198,1.001920,0.848054,
        -1.047198,1.006146,1.389734,
        -1.047198,1.012180,0.304066,
        -1.047198,1.016938,2.270674,
        -1.047198,1.017371,-0.979790,
        -1.047198,1.019495,-1.894487,
        -1.047198,1.023796,-0.583691,
        -1.047198,1.026525,2.506945,
        -1.047198,1.027004,-0.203653,
        -1.047198,1.028027,3.036573,
        -1.047198,1.031435,1.241021,
        -1.047198,1.032316,2.884221,
        -1.047198,1.035863,-3.110475,
        -1.047198,1.042085,-2.292950,
        -1.047198,1.043272,-1.522986,
        -1.047198,1.046265,-2.958847,
        -1.047198,1.049272,-1.376344,
        -1.047198,1.050039,-2.552086,
        -1.047198,1.053682,0.689863,
        -1.047198,1.055458,0.170394,
        -1.047198,1.056622,-0.831866,
        -1.047198,1.057515,1.097998,
        -1.047198,1.058706,-2.036438,
        -1.047198,1.061389,-1.672248,
        -1.047198,1.063758,2.743093,
        -1.047198,1.070157,-0.073422,
        -1.047198,1.072154,1.761506,
        -1.047198,1.074029,1.909193,
        -1.047198,1.074204,0.545276,
        -1.047198,1.074714,-2.805147,
        -1.047198,1.076262,0.966104,
        -1.047198,1.077642,-1.227073,
        -1.047198,1.084151,1.618525,
        -1.047198,1.085085,2.377673,
        -1.047198,1.086822,-0.441112,
        -1.047198,1.087142,2.058815,
        -1.047198,1.098383,-0.696430,
        -1.047198,1.099138,-1.084033,
        -1.047198,1.103208,1.477773,
        -1.047198,1.103783,0.408254,
        -1.047198,1.106275,-2.420800,
        -1.047198,1.108079,2.612943,
        -1.047198,1.109345,-2.172718,
        -1.047198,1.115957,-1.787597,
        -1.047198,1.118930,0.045983,
        -1.047198,1.119777,0.834702,
        -1.047198,1.121943,-2.669042,
        -1.047198,1.124930,2.219917,
        -1.047198,1.125217,1.336294,
        -1.047198,1.130958,2.972931,
        -1.047198,1.131400,-0.306629,
        -1.047198,1.137135,-0.936514,
        -1.047198,1.139125,-1.924528,
        -1.047198,1.139877,0.276382,
        -1.047198,1.142003,-3.058022,
        -1.047198,1.147363,-0.558448,
        -1.047198,1.150348,-0.176418,
        -1.047198,1.151198,1.197016,
        -1.047198,1.153930,-1.597104,
        -1.047198,1.155389,2.837148,
        -1.047198,1.155458,-1.460623,
        -1.047198,1.157331,-2.906362,
        -1.047198,1.158594,2.487618,
        -1.047198,1.162103,-1.322553,
        -1.047198,1.165407,3.092861,
        -1.047198,1.168025,0.717075,
        -1.047198,1.168406,-2.301519,
        -1.047198,1.174782,-2.538755,
        -1.047198,1.177998,1.061256,
        -1.047198,1.178898,1.827305,
        -1.047198,1.181331,-2.060330,
        -1.047198,1.182540,-0.798952,
        -1.047198,1.184569,0.147517,
        -1.047198,1.184734,1.968919,
        -1.047198,1.184804,0.593386,
        -1.047198,1.188278,1.690808,
        -1.047198,1.194360,2.110440,
        -1.047198,1.194776,2.710455,
        -1.047198,1.197745,-1.173705,
        -1.047198,1.199898,-2.777257,
        -1.047198,1.201164,2.355931,
        -1.047198,1.203760,1.556404,
        -1.047198,1.205845,-0.415993,
        -1.047198,1.206585,0.933648,
        -1.047198,1.206950,-0.037248,
        -1.047198,1.209341,0.471162,
        -1.047198,1.213483,-1.039417,
        -1.047198,1.218146,-0.672643,
        -1.047198,1.221608,1.421669,
        -1.047198,1.226043,-1.693533,
        -1.047198,1.232870,-1.822773,
        -1.047198,1.234375,-2.421194,
        -1.047198,1.234531,-2.190148,
        -1.047198,1.239414,0.348959,
        -1.047198,1.240183,2.587255,
        -1.047198,1.242312,2.941234,
        -1.047198,1.244075,1.286644,
        -1.047198,1.245961,-2.997461,
        -1.047198,1.246495,-2.646872,
        -1.047198,1.247736,2.242810,
        -1.047198,1.253604,0.812222,
        -1.047198,1.257816,-0.894317,
        -1.047198,1.258247,-1.953634,
        -1.047198,1.258993,-0.284099,
        -1.047198,1.262380,-0.157609,
        -1.047198,1.262791,-1.549229,
        -1.047198,1.263772,-0.533073,
        -1.047198,1.266731,-1.412461,
        -1.047198,1.268591,-3.122146,
        -1.047198,1.269614,-1.275246,
        -1.047198,1.270745,1.152533,
        -1.047198,1.273606,0.224545,
        -1.047198,1.279716,2.812101,
        -1.047198,1.283786,2.462124,
        -1.047198,1.284798,0.078209,
        -1.047198,1.285402,-2.882049,
        -1.047198,1.286936,1.886053,
        -1.047198,1.289408,0.683122,
        -1.047198,1.294736,1.756184,
        -1.047198,1.296978,-2.311511,
        -1.047198,1.298546,2.020601,
        -1.047198,1.300924,1.022478,
        -1.047198,1.303305,3.043152,
        -1.047198,1.304210,-2.084225,
        -1.047198,1.305241,-2.530249,
        -1.047198,1.307513,1.628003,
        -1.047198,1.310722,0.558267,
        -1.047198,1.314142,2.139329,
        -1.047198,1.316087,-0.777971,
        -1.047198,1.317907,-1.121023,
        -1.047198,1.321019,-2.749776,
        -1.047198,1.321609,2.687717,
        -1.047198,1.322568,1.499415,
        -1.047198,1.322972,-0.395699,
        -1.047198,1.325286,-0.994192,
        -1.047198,1.329146,-0.648513,
        -1.047198,1.329795,-0.042570,
        -1.047198,1.331651,2.344497,
        -1.047198,1.336522,0.435098,
        -1.047198,1.339410,0.900087,
        -1.047198,1.339550,1.370353,
        -1.047198,1.340441,-1.731580,
        -1.047198,1.351855,-1.855234,
        -1.047198,1.360161,-3.018404,
        -1.047198,1.361031,-1.613971,
        -1.047198,1.361196,2.914780,
        -1.047198,1.361998,-2.207254,
        -1.047198,1.362302,1.239384,
        -1.047198,1.365228,2.563808,
        -1.047198,1.365452,0.310522,
        -1.047198,1.367028,-2.420041,
        -1.047198,1.373422,-1.484281,
        -1.047198,1.374245,-1.226417,
        -1.047198,1.375020,-0.170393,
        -1.047198,1.376560,-1.355481,
        -1.047198,1.377070,-1.980984,
        -1.047198,1.377578,-2.636339,
        -1.047198,1.378096,0.774626,
        -1.047198,1.379086,-0.508192,
        -1.047198,1.382562,0.177141,
        -1.047198,1.388849,2.233847,
        -1.047198,1.389268,-0.886349,
        -1.047198,1.391324,1.109396,
        -1.047198,1.391699,-0.288725,
        -1.047198,1.392869,3.136903,
        -1.047198,1.393612,1.952016,
        -1.047198,1.395545,-2.850807,
        -1.047198,1.401926,1.822930,
        -1.047198,1.402149,2.788898,
        -1.047198,1.409552,0.647980,
        -1.047198,1.410840,0.056219,
        -1.047198,1.411944,2.444377,
        -1.047198,1.413639,1.697982,
        -1.047198,1.426479,1.573427,
        -1.047198,1.426725,2.102242,
        -1.047198,1.427868,0.984637,
        -1.047198,1.429580,-2.107232,
        -1.047198,1.431375,-2.313269,
        -1.047198,1.431907,3.015410,
        -1.047198,1.432325,-1.029054,
        -1.047198,1.433341,0.524268,
        -1.047198,1.435181,-0.739771,
        -1.047198,1.437208,-2.529085,
        -1.047198,1.439041,-0.617966,
        -1.047198,1.441473,1.448544,
        -1.047198,1.444527,2.664744,
        -1.047198,1.450513,-0.400625,
        -1.047198,1.451907,-0.066823,
        -1.047198,1.452754,-2.738631,
        -1.047198,1.454061,-2.950416,
        -1.047198,1.456736,1.323154,
        -1.047198,1.458370,0.400433,
        -1.047198,1.459314,-1.765003,
        -1.047198,1.459902,-1.145835,
        -1.047198,1.464952,2.329245,
        -1.047198,1.466050,0.862104,
        -1.047198,1.471330,-1.886160,
        -1.047198,1.475119,-1.648255,
        -1.047198,1.480878,1.194576,
        -1.047198,1.480942,0.273235,
        -1.047198,1.481086,2.890269,
        -1.047198,1.484118,-1.529156,
        -1.047198,1.484741,-1.279219,
        -1.047198,1.487681,-1.406069,
        -1.047198,1.487886,-3.066037,
        -1.047198,1.490555,2.543284,
        -1.047198,1.492424,-0.190932,
        -1.047198,1.495002,-0.839810,
        -1.047198,1.495297,2.007188,
        -1.047198,1.496414,-2.009267,
        -1.047198,1.498261,-2.425270,
        -1.047198,1.498907,-2.207327,
        -1.047198,1.500204,0.737638,
        -1.047198,1.502469,0.151156,
        -1.047198,1.507448,-0.514782,
        -1.047198,1.509287,2.204647,
        -1.047198,1.510518,1.884920,
        -1.047198,1.511948,-2.632360,
        -1.047198,1.515235,1.068565,
        -1.047198,1.522237,3.102018,
        -1.047198,1.522373,1.764991,
        -1.047198,1.522416,-0.954956,
        -1.047198,1.523102,2.765920,
        -1.047198,1.524811,-2.844718,
        -1.047198,1.527261,-0.304916,
        -1.047198,1.529788,0.613208,
        -1.047198,1.533648,1.644808,
        -1.047198,1.535018,0.028869,
        -1.047198,1.541082,2.424608,
        -1.047198,1.546010,1.523462,
        -1.047198,1.551135,-0.642994,
        -1.047198,1.553895,0.489607,
        -1.047198,1.555041,0.947813,
        -1.047198,1.559671,-1.074934,
        -1.047198,1.560345,2.985899,
        -1.047198,1.560487,-2.323400,
        -1.047198,1.561127,1.401111,
        -1.047198,1.567193,2.643172,
        -1.047198,1.568892,2.098362,
        -1.047198,1.571702,-2.098435,
        -1.047198,1.572701,-2.528730,
        -1.047198,1.573572,1.278593,
        -1.047198,1.574849,-0.091505,
        -1.047198,1.576334,0.365236,
        -1.047198,1.577029,-2.962888,
        -1.047198,1.577688,-1.814941,
        -1.047198,1.579542,-0.755917,
        -1.047198,1.579624,-1.199759,
        -1.047198,1.581515,-0.419414,
        -1.047198,1.584581,-2.738702,
        -1.047198,1.586948,-1.696296,
        -1.047198,1.588901,2.301496,
        -1.047198,1.589936,0.825020,
        -1.047198,1.590018,-1.941652,
        -1.047198,1.596968,-1.328358,
        -1.047198,1.597178,-1.575607,
        -1.047198,1.600470,-1.453196,
        -1.047198,1.601408,0.242409,
        -1.047198,1.602259,1.152656,
        -1.047198,1.603064,2.865774,
        -1.047198,1.612085,1.968802,
        -1.047198,1.612720,-3.080498,
        -1.047198,1.616061,2.521531,
        -1.047198,1.616900,-0.207457,
        -1.047198,1.620988,-0.875920,
        -1.047198,1.622114,0.701509,
        -1.047198,1.622749,0.121426,
        -1.047198,1.622993,-2.221629,
        -1.047198,1.629346,-0.542669,
        -1.047198,1.630007,1.844552,
        -1.047198,1.634747,-2.426198,
        -1.047198,1.642346,1.722617,
        -1.047198,1.643925,1.034629,
        -1.047198,1.644848,2.743774,
        -1.047198,1.645258,2.188109,
        -1.047198,1.645417,-2.634725,
        -1.047198,1.647395,3.085437,
        -1.047198,1.648538,-2.855044,
        -1.047198,1.649932,0.576808,
        -1.047198,1.652224,-0.996105,
        -1.047198,1.652871,1.600681,
        -1.047198,1.657997,-0.325127,
        -1.047198,1.661936,0.001037,
        -1.047198,1.665187,1.477402,
        -1.047198,1.665496,2.397872,
        -1.047198,1.673267,-0.665453,
        -1.047198,1.673557,0.452526,
        -1.047198,1.679277,-1.122421,
        -1.047198,1.679696,0.912482,
        -1.047198,1.682420,1.353922,
        -1.047198,1.685031,-2.009038,
        -1.047198,1.685196,2.967046,
        -1.047198,1.690107,2.620910,
        -1.047198,1.690453,-2.123076,
        -1.047198,1.691693,-1.871572,
        -1.047198,1.691817,2.061688,
        -1.047198,1.692644,1.237670,
        -1.047198,1.696386,0.328123,
        -1.047198,1.697162,-1.249026,
        -1.047198,1.697488,-2.322785,
        -1.047198,1.698726,-2.973576,
        -1.047198,1.699172,-1.749717,
        -1.047198,1.704478,-0.106295,
        -1.047198,1.707244,-0.444770,
        -1.047198,1.707500,-2.531497,
        -1.047198,1.707714,-0.787321,
        -1.047198,1.708430,-1.627552,
        -1.047198,1.710583,-1.376890,
        -1.047198,1.712415,-2.747921,
        -1.047198,1.713125,-1.503073,
        -1.047198,1.713452,0.789419,
        -1.047198,1.720612,2.279888,
        -1.047198,1.724852,0.202831,
        -1.047198,1.724868,2.846246,
        -1.047198,1.728940,1.931535,
        -1.047198,1.732838,1.127584,
        -1.047198,1.735719,-3.092485,
        -1.047198,1.739676,2.496017,
        -1.047198,1.739971,-0.228876,
        -1.047198,1.743328,0.663347,
        -1.047198,1.744927,-0.911892,
        -1.047198,1.745952,0.084733,
        -1.047198,1.748077,1.804354,
        -1.047198,1.752959,-0.567548,
        -1.047198,1.760169,1.678132,
        -1.047198,1.763887,-2.218450,
        -1.047198,1.766737,2.722752,
        -1.047198,1.767999,1.004737,
        -1.047198,1.769149,1.550298,
        -1.047198,1.770119,-2.427175,
        -1.047198,1.770460,2.154208,
        -1.047198,1.770514,3.070982,
        -1.047198,1.770735,0.537245,
        -1.047198,1.772673,-2.864760,
        -1.047198,1.773527,-1.039582,
        -1.047198,1.776419,-2.642421,
        -1.047198,1.780769,1.416330,
        -1.047198,1.786634,-0.346247,
        -1.047198,1.788829,-1.947553,
        -1.047198,1.793737,0.410270,
        -1.047198,1.794553,2.374618,
        -1.047198,1.794699,-0.694061,
        -1.047198,1.797469,-1.169105,
        -1.047198,1.803705,0.881076,
        -1.047198,1.804124,1.264249,
        -1.047198,1.807825,-2.090455,
        -1.047198,1.808822,-1.817375,
        -1.047198,1.809020,2.949741,
        -1.047198,1.812466,2.597595,
        -1.047198,1.813159,2.025395,
        -1.047198,1.813816,-1.298435,
        -1.047198,1.815016,0.278951,
        -1.047198,1.816510,-0.008631,
        -1.047198,1.817772,-1.689625,
        -1.047198,1.821360,-2.984352,
        -1.047198,1.822312,-0.127211,
        -1.047198,1.823386,-1.429807,
        -1.047198,1.823834,-1.560483,
        -1.047198,1.832060,-0.821627,
        -1.047198,1.832690,-0.468440,
        -1.047198,1.833945,-2.320278,
        -1.047198,1.834787,0.751968,
        -1.047198,1.840579,-2.536445,
        -1.047198,1.841033,-2.756932,
        -1.047198,1.846350,1.893983,
        -1.047198,1.846644,1.113766,
        -1.047198,1.846833,2.826543,
        -1.047198,1.847325,2.249051,
        -1.047198,1.854119,0.134161,
        -1.047198,1.858863,-3.106022,
        -1.047198,1.865046,1.763384,
        -1.047198,1.865619,-0.952036,
        -1.047198,1.865912,0.624453,
        -1.047198,1.866694,2.473686,
        -1.047198,1.867524,-0.240083,
        -1.047198,1.875561,1.632831,
        -1.047198,1.876049,-0.594378,
        -1.047198,1.880509,1.500204,
        -1.047198,1.883504,1.360457,
        -1.047198,1.886971,-2.196665,
        -1.047198,1.888979,0.982922,
        -1.047198,1.889815,2.701570,
        -1.047198,1.892144,-2.005340,
        -1.047198,1.892348,0.496522,
        -1.047198,1.893281,-1.083976,
        -1.047198,1.893828,3.053453,
        -1.047198,1.895117,2.121107,
        -1.047198,1.898741,-2.875735,
        -1.047198,1.904429,-2.428029,
        -1.047198,1.908408,-2.648903,
        -1.047198,1.911828,-0.364489,
        -1.047198,1.912224,0.366565,
        -1.047198,1.914854,1.214468,
        -1.047198,1.915467,-1.883143,
        -1.047198,1.915763,-0.724679,
        -1.047198,1.915790,-1.216366,
        -1.047198,1.922256,2.347806,
        -1.047198,1.922616,0.844633,
        -1.047198,1.925185,0.231170,
        -1.047198,1.927474,-1.754616,
        -1.047198,1.928342,-1.351174,
        -1.047198,1.933328,-1.623328,
        -1.047198,1.933878,0.034150,
        -1.047198,1.934585,-1.487979,
        -1.047198,1.934802,2.927634,
        -1.047198,1.934862,1.989776,
        -1.047198,1.936327,-0.099350,
        -1.047198,1.938088,2.578507,
        -1.047198,1.945677,-2.999805,
        -1.047198,1.954572,-0.859208,
        -1.047198,1.957868,-0.490784,
        -1.047198,1.958989,0.712772,
        -1.047198,1.960212,-2.307561,
        -1.047198,1.964774,1.855507,
        -1.047198,1.964971,-2.100533,
        -1.047198,1.969112,1.088273,
        -1.047198,1.971438,-2.765415,
        -1.047198,1.972312,2.805134,
        -1.047198,1.974950,2.220596,
        -1.047198,1.975135,-2.539127,
        -1.047198,1.982107,1.720829,
        -1.047198,1.983195,-3.127984,
        -1.047198,1.983318,-0.247182,
        -1.047198,1.986720,-0.993220,
        -1.047198,1.990381,1.586005,
        -1.047198,1.990470,0.580283,
        -1.047198,1.990523,1.450326,
        -1.047198,1.993909,1.312897,
        -1.047198,1.994694,2.452813,
        -1.047198,1.999226,-0.620717,
        -1.047198,2.005145,0.941012,
        -1.047198,2.007301,0.142337,
        -1.047198,2.014149,0.444138,
        -1.047198,2.014338,-1.127392,
        -1.047198,2.018131,3.021778,
        -1.047198,2.019676,-1.980944,
        -1.047198,2.019912,2.686612,
        -1.047198,2.020545,2.088672,
        -1.047198,2.026074,0.300945,
        -1.047198,2.026196,-2.889057,
        -1.047198,2.031959,-1.268178,
        -1.047198,2.032917,-2.199917,
        -1.047198,2.033572,-1.842884,
        -1.047198,2.034712,-2.419505,
        -1.047198,2.035818,-0.379479,
        -1.047198,2.036004,-0.757062,
        -1.047198,2.040501,-1.411488,
        -1.047198,2.042874,-1.701367,
        -1.047198,2.043060,-2.653746,
        -1.047198,2.043729,-1.557938,
        -1.047198,2.048538,0.806623,
        -1.047198,2.049006,1.199930,
        -1.047198,2.049502,-0.137989,
        -1.047198,2.052181,2.326959,
        -1.047198,2.052794,-0.008526,
        -1.047198,2.057608,1.951726,
        -1.047198,2.065568,2.567730,
        -1.047198,2.070371,2.879808,
        -1.047198,2.071256,-3.018961,
        -1.047198,2.079746,-0.897699,
        -1.047198,2.082784,1.042403,
        -1.047198,2.083752,-0.511086,
        -1.047198,2.084401,1.810502,
        -1.047198,2.085731,0.671081,
        -1.047198,2.097496,1.373172,
        -1.047198,2.099185,1.667006,
        -1.047198,2.100482,-2.301084,
        -1.047198,2.103118,2.194745,
        -1.047198,2.103223,1.521685,
        -1.047198,2.103695,-2.777707,
        -1.047198,2.107523,3.129581,
        -1.047198,2.108277,-2.095907,
        -1.047198,2.108891,-2.532166,
        -1.047198,2.111003,-0.268632,
        -1.047198,2.112276,-1.027246,
        -1.047198,2.114380,0.531457,
        -1.047198,2.119166,0.100552,
        -1.047198,2.119564,2.754818,
        -1.047198,2.122474,-0.648314,
        -1.047198,2.125183,2.445598,
        -1.047198,2.129501,0.238290,
        -1.047198,2.131178,0.386663,
        -1.047198,2.131223,-1.174311,
        -1.047198,2.134269,-1.948042,
        -1.047198,2.134684,0.907065,
        -1.047198,2.137615,2.988806,
        -1.047198,2.145635,-1.327708,
        -1.047198,2.146467,2.055081,
        -1.047198,2.146582,-1.793788,
        -1.047198,2.149210,-1.482321,
        -1.047198,2.152901,-1.636788,
        -1.047198,2.156337,1.151558,
        -1.047198,2.156474,-2.912065,
        -1.047198,2.157194,-0.792563,
        -1.047198,2.166724,-0.401638,
        -1.047198,2.166980,-0.134879,
        -1.047198,2.173644,-2.403429,
        -1.047198,2.175711,-2.655523,
        -1.047198,2.177132,2.600419,
        -1.047198,2.178879,0.767722,
        -1.047198,2.180752,1.908485,
        -1.047198,2.181338,2.313721,
        -1.047198,2.185580,1.287673,
        -1.047198,2.195400,-2.217546,
        -1.047198,2.200418,-3.056442,
        -1.047198,2.201829,-0.001735,
        -1.047198,2.203804,1.756176,
        -1.047198,2.211602,1.443930,
        -1.047198,2.212686,-0.544809,
        -1.047198,2.213937,0.621892,
        -1.047198,2.214075,1.600691,
        -1.047198,2.216256,1.020029,
        -1.047198,2.219060,2.886631,
        -1.047198,2.223375,-0.906185,
        -1.047198,2.226002,-1.060248,
        -1.047198,2.226413,-2.064034,
        -1.047198,2.230554,2.171334,
        -1.047198,2.232413,2.735777,
        -1.047198,2.233709,3.070830,
        -1.047198,2.234383,-2.792065,
        -1.047198,2.237183,0.468799,
        -1.047198,2.237695,-0.271388,
        -1.047198,2.239419,0.151693,
        -1.047198,2.243347,-2.521049,
        -1.047198,2.244035,-1.229456,
        -1.047198,2.244355,0.309899,
        -1.047198,2.245175,2.460236,
        -1.047198,2.245628,-1.899216,
        -1.047198,2.245973,-0.702234,
        -1.047198,2.255601,-1.728131,
        -1.047198,2.256702,-1.401367,
        -1.047198,2.261466,-1.557500,
        -1.047198,2.267390,0.878499,
        -1.047198,2.272099,2.018594,
        -1.047198,2.278618,1.170386,
        -1.047198,2.283304,-2.345584,
        -1.047198,2.285175,-2.943090,
        -1.047198,2.293000,-0.124451,
        -1.047198,2.294058,-0.421070,
        -1.047198,2.303283,1.855938,
        -1.047198,2.304968,2.608978,
        -1.047198,2.305347,2.311346,
        -1.047198,2.307188,-2.656134,
        -1.047198,2.307301,1.336200,
        -1.047198,2.309655,0.726020,
        -1.047198,2.317634,-2.185225,
        -1.047198,2.321510,1.685664,
        -1.047198,2.324376,-3.111006,
        -1.047198,2.324731,1.511186,
        -1.047198,2.329587,0.035498,
        -1.047198,2.330158,2.977264,
        -1.047198,2.335096,-0.588468,
        -1.047198,2.336046,-1.115799,
        -1.047198,2.338060,2.803613,
        -1.047198,2.338083,-0.943400,
        -1.047198,2.340210,0.562145,
        -1.047198,2.341931,1.023926,
        -1.047198,2.343265,-0.779061,
        -1.047198,2.343434,-2.013692,
        -1.047198,2.353029,-1.310529,
        -1.047198,2.354868,0.211543,
        -1.047198,2.355120,0.388593,
        -1.047198,2.355165,-1.833173,
        -1.047198,2.355462,2.151441,
        -1.047198,2.362948,-2.487477,
        -1.047198,2.363101,-2.810456,
        -1.047198,2.363394,-0.272918,
        -1.047198,2.364509,-1.644533,
        -1.047198,2.370465,2.466824,
        -1.047198,2.386110,-1.481351,
        -1.047198,2.394960,0.862133,
        -1.047198,2.395716,1.977184,
        -1.047198,2.396339,1.200900,
        -1.047198,2.407941,-2.316753,
        -1.047198,2.409450,-2.988432,
        -1.047198,2.417202,-0.103363,
        -1.047198,2.417386,-0.443097,
        -1.047198,2.417950,2.666662,
        -1.047198,2.423190,1.789783,
        -1.047198,2.423692,1.392696,
        -1.047198,2.431122,3.086379,
        -1.047198,2.431406,-2.647820,
        -1.047198,2.431584,2.307631,
        -1.047198,2.434090,1.592818,
        -1.047198,2.435993,0.683844,
        -1.047198,2.440898,-2.133300,
        -1.047198,2.442222,-1.195837,
        -1.047198,2.443112,2.883761,
        -1.047198,2.449191,-0.633651,
        -1.047198,2.450821,0.084741,
        -1.047198,2.453499,-1.008861,
        -1.047198,2.460010,-1.938011,
        -1.047198,2.460199,-0.822008,
        -1.047198,2.460902,0.490079,
        -1.047198,2.461774,-1.734913,
        -1.047198,2.463986,1.034407,
        -1.047198,2.467647,0.287667,
        -1.047198,2.479849,2.126925,
        -1.047198,2.485907,-2.835471,
        -1.047198,2.485945,-1.374853,
        -1.047198,2.487747,2.505912,
        -1.047198,2.487957,-0.274279,
        -1.047198,2.489741,-2.468428,
        -1.047198,2.512538,1.239900,
        -1.047198,2.517528,1.923289,
        -1.047198,2.518454,-1.565852,
        -1.047198,2.519348,0.843503,
        -1.047198,2.522766,-3.059231,
        -1.047198,2.528686,2.737228,
        -1.047198,2.534798,-2.266671,
        -1.047198,2.536533,1.466274,
        -1.047198,2.538504,1.699812,
        -1.047198,2.539516,-0.476942,
        -1.047198,2.539532,-0.074548,
        -1.047198,2.543197,2.990252,
        -1.047198,2.554022,2.320619,
        -1.047198,2.556919,-2.655005,
        -1.047198,2.558146,0.627191,
        -1.047198,2.558223,-1.164189,
        -1.047198,2.563452,-2.044028,
        -1.047198,2.568462,0.150655,
        -1.047198,2.569130,-0.700837,
        -1.047198,2.570050,-1.804492,
        -1.047198,2.572710,-0.934665,
        -1.047198,2.575994,0.391718,
        -1.047198,2.584774,1.046703,
        -1.047198,2.601432,2.102535,
        -1.047198,2.603559,2.555114,
        -1.047198,2.604150,-2.885208,
        -1.047198,2.608630,-1.378011,
        -1.047198,2.611256,-0.279110,
        -1.047198,2.617344,-2.439480,
        -1.047198,2.627402,1.293552,
        -1.047198,2.635396,1.848850,
        -1.047198,2.635686,2.834950,
        -1.047198,2.637087,3.132730,
        -1.047198,2.642029,0.813274,
        -1.047198,2.645260,1.570203,
        -1.047198,2.646182,-1.614036,
        -1.047198,2.660113,-2.186513,
        -1.047198,2.660483,-0.034608,
        -1.047198,2.661713,-0.529511,
        -1.047198,2.672720,-1.110422,
        -1.047198,2.676626,2.335353,
        -1.047198,2.676638,0.540364,
        -1.047198,2.676780,-1.901158,
        -1.047198,2.681411,-2.667965,
        -1.047198,2.682152,0.245751,
        -1.047198,2.682859,-0.817071,
        -1.047198,2.705107,1.062142,
        -1.047198,2.717705,2.625995,
        -1.047198,2.721001,2.060097,
        -1.047198,2.722319,-2.954069,
        -1.047198,2.729840,-1.368237,
        -1.047198,2.734349,-0.289298,
        -1.047198,2.738409,2.975671,
        -1.047198,2.740366,1.374341,
        -1.047198,2.744633,-2.391311,
        -1.047198,2.747398,1.727166,
        -1.047198,2.763799,0.760372,
        -1.047198,2.768238,-1.679243,
        -1.047198,2.779245,-2.046268,
        -1.047198,2.780208,-0.626453,
        -1.047198,2.780471,0.027732,
        -1.047198,2.784696,-1.013128,
        -1.047198,2.789817,0.395431,
        -1.047198,2.801154,2.353143,
        -1.047198,2.807091,-2.687477,
        -1.047198,2.825762,1.085035,
        -1.047198,2.830209,2.734943,
        -1.047198,2.833517,-3.085508,
        -1.047198,2.838301,1.980061,
        -1.047198,2.849871,1.512012,
        -1.047198,2.850089,-1.343899,
        -1.047198,2.856905,-0.318441,
        -1.047198,2.870445,-2.291001,
        -1.047198,2.885017,0.655562,
        -1.047198,2.886428,-1.793217,
        -1.047198,2.890337,-0.825633,
        -1.047198,2.898932,0.139973,
        -1.047198,2.931927,-2.728768,
        -1.047198,2.932096,2.377790,
        -1.047198,2.937409,2.947245,
        -1.047198,2.948048,1.133127,
        -1.047198,2.950896,1.797889,
        -1.047198,2.970019,-1.282382,
        -1.047198,2.978333,-0.412090,
        -1.047198,2.996437,-2.089121,
        -1.047198,3.011274,0.434110,
        -1.047198,3.050410,3.138492,
        -1.047198,3.064492,1.647259,
        -1.047198,3.088868,-0.989501,
        -0.942478,0.053884,-2.408394,
        -0.942478,0.077645,1.453433,
        -0.942478,0.088408,-0.311160,
        -0.942478,0.142314,2.590387,
        -0.942478,0.147206,-1.385164,
        -0.942478,0.163506,0.573203,
        -0.942478,0.165485,-2.875396,
        -0.942478,0.182564,-2.125821,
        -0.942478,0.190569,1.929141,
        -0.942478,0.201102,1.249082,
        -0.942478,0.201397,-0.716646,
        -0.942478,0.210568,-0.069559,
        -0.942478,0.251672,2.848900,
        -0.942478,0.266262,-1.678068,
        -0.942478,0.269108,-1.215563,
        -0.942478,0.273777,-2.506756,
        -0.942478,0.279854,0.814581,
        -0.942478,0.281664,-2.969147,
        -0.942478,0.283625,2.405603,
        -0.942478,0.284058,0.357350,
        -0.942478,0.294107,1.593880,
        -0.942478,0.302491,-0.400410,
        -0.942478,0.308059,2.021491,
        -0.942478,0.312219,-2.072057,
        -0.942478,0.321905,-0.829121,
        -0.942478,0.329984,1.201621,
        -0.942478,0.339714,-0.013712,
        -0.942478,0.368784,2.974974,
        -0.942478,0.376428,-1.495532,
        -0.942478,0.378103,2.646436,
        -0.942478,0.385088,0.590355,
        -0.942478,0.386248,-2.690943,
        -0.942478,0.389851,-2.340031,
        -0.942478,0.391502,-1.171374,
        -0.942478,0.398249,-1.812694,
        -0.942478,0.400542,-2.995729,
        -0.942478,0.404273,0.918163,
        -0.942478,0.405617,1.773533,
        -0.942478,0.409009,-0.570243,
        -0.942478,0.410700,0.270433,
        -0.942478,0.415893,1.464890,
        -0.942478,0.420680,2.088164,
        -0.942478,0.424766,-0.258729,
        -0.942478,0.426507,2.367872,
        -0.942478,0.441292,-0.890727,
        -0.942478,0.451839,-2.078193,
        -0.942478,0.459270,1.191562,
        -0.942478,0.470622,0.014002,
        -0.942478,0.481292,-2.529098,
        -0.942478,0.482460,2.849843,
        -0.942478,0.492205,-1.397963,
        -0.942478,0.493314,3.117264,
        -0.942478,0.497870,0.727828,
        -0.942478,0.499783,-1.654379,
        -0.942478,0.502651,0.463333,
        -0.942478,0.505463,2.594557,
        -0.942478,0.506104,-2.874910,
        -0.942478,0.513579,-1.147354,
        -0.942478,0.518569,1.644237,
        -0.942478,0.519001,1.900488,
        -0.942478,0.519277,-0.673809,
        -0.942478,0.524769,-0.421955,
        -0.942478,0.525564,-2.288838,
        -0.942478,0.532541,0.978737,
        -0.942478,0.534610,2.154808,
        -0.942478,0.535020,-1.891846,
        -0.942478,0.540208,1.400497,
        -0.942478,0.542006,0.228337,
        -0.942478,0.550762,-0.180771,
        -0.942478,0.560531,2.378474,
        -0.942478,0.561629,-0.924576,
        -0.942478,0.567465,-2.679392,
        -0.942478,0.588527,1.181755,
        -0.942478,0.591643,-2.098364,
        -0.942478,0.594963,2.997575,
        -0.942478,0.599643,-3.045235,
        -0.942478,0.599962,2.773556,
        -0.942478,0.602620,0.029270,
        -0.942478,0.605711,0.598454,
        -0.942478,0.607708,-1.541026,
        -0.942478,0.609360,-2.464379,
        -0.942478,0.610738,-1.326786,
        -0.942478,0.616968,0.820489,
        -0.942478,0.626379,1.773610,
        -0.942478,0.628626,-1.751180,
        -0.942478,0.630671,0.391949,
        -0.942478,0.632622,-0.737642,
        -0.942478,0.632708,2.566722,
        -0.942478,0.632797,1.989858,
        -0.942478,0.634455,-0.534790,
        -0.942478,0.636205,1.561170,
        -0.942478,0.637681,-1.123423,
        -0.942478,0.641441,-2.842109,
        -0.942478,0.644807,-0.327139,
        -0.942478,0.652440,2.197129,
        -0.942478,0.660897,-2.278593,
        -0.942478,0.664008,1.360091,
        -0.942478,0.670850,-1.945257,
        -0.942478,0.671396,1.004859,
        -0.942478,0.673985,0.203855,
        -0.942478,0.680326,-0.133370,
        -0.942478,0.685083,-0.942363,
        -0.942478,0.688204,-2.632204,
        -0.942478,0.689777,2.385011,
        -0.942478,0.700065,3.115430,
        -0.942478,0.704080,2.911636,
        -0.942478,0.712883,1.174216,
        -0.942478,0.715701,0.699906,
        -0.942478,0.718851,-1.442215,
        -0.942478,0.721019,2.721536,
        -0.942478,0.725300,-2.975134,
        -0.942478,0.727136,-1.633288,
        -0.942478,0.727399,-2.120567,
        -0.942478,0.730751,0.522488,
        -0.942478,0.732892,-1.254330,
        -0.942478,0.734333,-2.446184,
        -0.942478,0.735125,0.037270,
        -0.942478,0.737893,1.866463,
        -0.942478,0.740650,1.681075,
        -0.942478,0.745971,-0.781286,
        -0.942478,0.749432,-0.442559,
        -0.942478,0.750102,2.050636,
        -0.942478,0.751046,0.863821,
        -0.942478,0.757018,-0.608833,
        -0.942478,0.757046,-1.816019,
        -0.942478,0.757389,1.501184,
        -0.942478,0.759800,2.546181,
        -0.942478,0.759859,0.351923,
        -0.942478,0.761141,-2.781102,
        -0.942478,0.769035,-1.078574,
        -0.942478,0.770166,-0.266604,
        -0.942478,0.772218,2.227380,
        -0.942478,0.788382,1.331051,
        -0.942478,0.790269,-2.282409,
        -0.942478,0.803410,-1.984567,
        -0.942478,0.803813,0.191520,
        -0.942478,0.806832,3.022155,
        -0.942478,0.807476,-2.597675,
        -0.942478,0.812512,-0.107335,
        -0.942478,0.814161,1.010949,
        -0.942478,0.817732,-3.084486,
        -0.942478,0.817846,2.390688,
        -0.942478,0.819323,2.845960,
        -0.942478,0.819430,-0.915017,
        -0.942478,0.829198,-1.532491,
        -0.942478,0.829531,0.665695,
        -0.942478,0.829816,1.167147,
        -0.942478,0.833613,-1.357492,
        -0.942478,0.842945,-2.907318,
        -0.942478,0.844883,2.679708,
        -0.942478,0.848189,-1.705003,
        -0.942478,0.849888,1.767901,
        -0.942478,0.851126,0.493370,
        -0.942478,0.852911,1.933394,
        -0.942478,0.857516,-2.432863,
        -0.942478,0.858261,-1.191729,
        -0.942478,0.858279,-2.138895,
        -0.942478,0.859723,-0.734928,
        -0.942478,0.859754,1.603292,
        -0.942478,0.861655,-0.394821,
        -0.942478,0.864300,0.043549,
        -0.942478,0.866940,0.851229,
        -0.942478,0.869748,2.095969,
        -0.942478,0.877352,-0.574668,
        -0.942478,0.879254,-2.733338,
        -0.942478,0.882554,1.443757,
        -0.942478,0.884060,0.334199,
        -0.942478,0.885626,-1.866061,
        -0.942478,0.889834,2.529957,
        -0.942478,0.893140,2.251698,
        -0.942478,0.897463,-0.240822,
        -0.942478,0.898758,-1.031948,
        -0.942478,0.913918,1.287618,
        -0.942478,0.916555,-2.287173,
        -0.942478,0.918305,3.107845,
        -0.942478,0.919522,2.948277,
        -0.942478,0.927138,-2.569686,
        -0.942478,0.929031,0.185383,
        -0.942478,0.933766,-3.015917,
        -0.942478,0.934873,-2.010822,
        -0.942478,0.935636,-1.445855,
        -0.942478,0.936702,-0.866461,
        -0.942478,0.937395,0.729108,
        -0.942478,0.937910,2.787575,
        -0.942478,0.939144,-0.092962,
        -0.942478,0.941718,1.125085,
        -0.942478,0.942873,-1.607680,
        -0.942478,0.943701,0.980328,
        -0.942478,0.947774,2.395256,
        -0.942478,0.952433,0.584347,
        -0.942478,0.954456,-1.287293,
        -0.942478,0.958731,-2.852552,
        -0.942478,0.961838,1.839284,
        -0.942478,0.961898,-0.469425,
        -0.942478,0.965426,1.686615,
        -0.942478,0.970108,1.990771,
        -0.942478,0.972820,-1.762780,
        -0.942478,0.973500,-0.711881,
        -0.942478,0.975253,2.641082,
        -0.942478,0.977824,0.441928,
        -0.942478,0.981485,-2.424766,
        -0.942478,0.981877,1.537100,
        -0.942478,0.982609,-1.136920,
        -0.942478,0.984265,-2.156073,
        -0.942478,0.989150,0.045493,
        -0.942478,0.993052,2.137478,
        -0.942478,0.997766,-2.693871,
        -0.942478,0.999847,-0.334738,
        -0.942478,1.001920,0.848054,
        -0.942478,1.006146,1.389734,
        -0.942478,1.012180,0.304066,
        -0.942478,1.016938,2.270674,
        -0.942478,1.017371,-0.979790,
        -0.942478,1.019495,-1.894487,
        -0.942478,1.023796,-0.583691,
        -0.942478,1.026525,2.506945,
        -0.942478,1.027004,-0.203653,
        -0.942478,1.028027,3.036573,
        -0.942478,1.031435,1.241021,
        -0.942478,1.032316,2.884221,
        -0.942478,1.035863,-3.110475,
        -0.942478,1.042085,-2.292950,
        -0.942478,1.043272,-1.522986,
        -0.942478,1.046265,-2.958847,
        -0.942478,1.049272,-1.376344,
        -0.942478,1.050039,-2.552086,
        -0.942478,1.053682,0.689863,
        -0.942478,1.055458,0.170394,
        -0.942478,1.056622,-0.831866,
        -0.942478,1.057515,1.097998,
        -0.942478,1.058706,-2.036438,
        -0.942478,1.061389,-1.672248,
        -0.942478,1.063758,2.743093,
        -0.942478,1.070157,-0.073422,
        -0.942478,1.072154,1.761506,
        -0.942478,1.074029,1.909193,
        -0.942478,1.074204,0.545276,
        -0.942478,1.074714,-2.805147,
        -0.942478,1.076262,0.966104,
        -0.942478,1.077642,-1.227073,
        -0.942478,1.084151,1.618525,
        -0.942478,1.085085,2.377673,
        -0.942478,1.086822,-0.441112,
        -0.942478,1.087142,2.058815,
        -0.942478,1.098383,-0.696430,
        -0.942478,1.099138,-1.084033,
        -0.942478,1.103208,1.477773,
        -0.942478,1.103783,0.408254,
        -0.942478,1.106275,-2.420800,
        -0.942478,1.108079,2.612943,
        -0.942478,1.109345,-2.172718,
        -0.942478,1.115957,-1.787597,
        -0.942478,1.118930,0.045983,
        -0.942478,1.119777,0.834702,
        -0.942478,1.121943,-2.669042,
        -0.942478,1.124930,2.219917,
        -0.942478,1.125217,1.336294,
        -0.942478,1.130958,2.972931,
        -0.942478,1.131400,-0.306629,
        -0.942478,1.137135,-0.936514,
        -0.942478,1.139125,-1.924528,
        -0.942478,1.139877,0.276382,
        -0.942478,1.142003,-3.058022,
        -0.942478,1.147363,-0.558448,
        -0.942478,1.150348,-0.176418,
        -0.942478,1.151198,1.197016,
        -0.942478,1.153930,-1.597104,
        -0.942478,1.155389,2.837148,
        -0.942478,1.155458,-1.460623,
        -0.942478,1.157331,-2.906362,
        -0.942478,1.158594,2.487618,
        -0.942478,1.162103,-1.322553,
        -0.942478,1.165407,3.092861,
        -0.942478,1.168025,0.717075,
        -0.942478,1.168406,-2.301519,
        -0.942478,1.174782,-2.538755,
        -0.942478,1.177998,1.061256,
        -0.942478,1.178898,1.827305,
        -0.942478,1.181331,-2.060330,
        -0.942478,1.182540,-0.798952,
        -0.942478,1.184569,0.147517,
        -0.942478,1.184734,1.968919,
        -0.942478,1.184804,0.593386,
        -0.942478,1.188278,1.690808,
        -0.942478,1.194360,2.110440,
        -0.942478,1.194776,2.710455,
        -0.942478,1.197745,-1.173705,
        -0.942478,1.199898,-2.777257,
        -0.942478,1.201164,2.355931,
        -0.942478,1.203760,1.556404,
        -0.942478,1.205845,-0.415993,
        -0.942478,1.206585,0.933648,
        -0.942478,1.206950,-0.037248,
        -0.942478,1.209341,0.471162,
        -0.942478,1.213483,-1.039417,
        -0.942478,1.218146,-0.672643,
        -0.942478,1.221608,1.421669,
        -0.942478,1.226043,-1.693533,
        -0.942478,1.232870,-1.822773,
        -0.942478,1.234375,-2.421194,
        -0.942478,1.234531,-2.190148,
        -0.942478,1.239414,0.348959,
        -0.942478,1.240183,2.587255,
        -0.942478,1.242312,2.941234,
        -0.942478,1.244075,1.286644,
        -0.942478,1.245961,-2.997461,
        -0.942478,1.246495,-2.646872,
        -0.942478,1.247736,2.242810,
        -0.942478,1.253604,0.812222,
        -0.942478,1.257816,-0.894317,
        -0.942478,1.258247,-1.953634,
        -0.942478,1.258993,-0.284099,
        -0.942478,1.262380,-0.157609,
        -0.942478,1.262791,-1.549229,
        -0.942478,1.263772,-0.533073,
        -0.942478,1.266731,-1.412461,
        -0.942478,1.268591,-3.122146,
        -0.942478,1.269614,-1.275246,
        -0.942478,1.270745,1.152533,
        -0.942478,1.273606,0.224545,
        -0.942478,1.279716,2.812101,
        -0.942478,1.283786,2.462124,
        -0.942478,1.284798,0.078209,
        -0.942478,1.285402,-2.882049,
        -0.942478,1.286936,1.886053,
        -0.942478,1.289408,0.683122,
        -0.942478,1.294736,1.756184,
        -0.942478,1.296978,-2.311511,
        -0.942478,1.298546,2.020601,
        -0.942478,1.300924,1.022478,
        -0.942478,1.303305,3.043152,
        -0.942478,1.304210,-2.084225,
        -0.942478,1.305241,-2.530249,
        -0.942478,1.307513,1.628003,
        -0.942478,1.310722,0.558267,
        -0.942478,1.314142,2.139329,
        -0.942478,1.316087,-0.777971,
        -0.942478,1.317907,-1.121023,
        -0.942478,1.321019,-2.749776,
        -0.942478,1.321609,2.687717,
        -0.942478,1.322568,1.499415,
        -0.942478,1.322972,-0.395699,
        -0.942478,1.325286,-0.994192,
        -0.942478,1.329146,-0.648513,
        -0.942478,1.329795,-0.042570,
        -0.942478,1.331651,2.344497,
        -0.942478,1.336522,0.435098,
        -0.942478,1.339410,0.900087,
        -0.942478,1.339550,1.370353,
        -0.942478,1.340441,-1.731580,
        -0.942478,1.351855,-1.855234,
        -0.942478,1.360161,-3.018404,
        -0.942478,1.361031,-1.613971,
        -0.942478,1.361196,2.914780,
        -0.942478,1.361998,-2.207254,
        -0.942478,1.362302,1.239384,
        -0.942478,1.365228,2.563808,
        -0.942478,1.365452,0.310522,
        -0.942478,1.367028,-2.420041,
        -0.942478,1.373422,-1.484281,
        -0.942478,1.374245,-1.226417,
        -0.942478,1.375020,-0.170393,
        -0.942478,1.376560,-1.355481,
        -0.942478,1.377070,-1.980984,
        -0.942478,1.377578,-2.636339,
        -0.942478,1.378096,0.774626,
        -0.942478,1.379086,-0.508192,
        -0.942478,1.382562,0.177141,
        -0.942478,1.388849,2.233847,
        -0.942478,1.389268,-0.886349,
        -0.942478,1.391324,1.109396,
        -0.942478,1.391699,-0.288725,
        -0.942478,1.392869,3.136903,
        -0.942478,1.393612,1.952016,
        -0.942478,1.395545,-2.850807,
        -0.942478,1.401926,1.822930,
        -0.942478,1.402149,2.788898,
        -0.942478,1.409552,0.647980,
        -0.942478,1.410840,0.056219,
        -0.942478,1.411944,2.444377,
        -0.942478,1.413639,1.697982,
        -0.942478,1.426479,1.573427,
        -0.942478,1.426725,2.102242,
        -0.942478,1.427868,0.984637,
        -0.942478,1.429580,-2.107232,
        -0.942478,1.431375,-2.313269,
        -0.942478,1.431907,3.015410,
        -0.942478,1.432325,-1.029054,
        -0.942478,1.433341,0.524268,
        -0.942478,1.435181,-0.739771,
        -0.942478,1.437208,-2.529085,
        -0.942478,1.439041,-0.617966,
        -0.942478,1.441473,1.448544,
        -0.942478,1.444527,2.664744,
        -0.942478,1.450513,-0.400625,
        -0.942478,1.451907,-0.066823,
        -0.942478,1.452754,-2.738631,
        -0.942478,1.454061,-2.950416,
        -0.942478,1.456736,1.323154,
        -0.942478,1.458370,0.400433,
        -0.942478,1.459314,-1.765003,
        -0.942478,1.459902,-1.145835,
        -0.942478,1.464952,2.329245,
        -0.942478,1.466050,0.862104,
        -0.942478,1.471330,-1.886160,
        -0.942478,1.475119,-1.648255,
        -0.942478,1.480878,1.194576,
        -0.942478,1.480942,0.273235,
        -0.942478,1.481086,2.890269,
        -0.942478,1.484118,-1.529156,
        -0.942478,1.484741,-1.279219,
        -0.942478,1.487681,-1.406069,
        -0.942478,1.487886,-3.066037,
        -0.942478,1.490555,2.543284,
        -0.942478,1.492424,-0.190932,
        -0.942478,1.495002,-0.839810,
        -0.942478,1.495297,2.007188,
        -0.942478,1.496414,-2.009267,
        -0.942478,1.498261,-2.425270,
        -0.942478,1.498907,-2.207327,
        -0.942478,1.500204,0.737638,
        -0.942478,1.502469,0.151156,
        -0.942478,1.507448,-0.514782,
        -0.942478,1.509287,2.204647,
        -0.942478,1.510518,1.884920,
        -0.942478,1.511948,-2.632360,
        -0.942478,1.515235,1.068565,
        -0.942478,1.522237,3.102018,
        -0.942478,1.522373,1.764991,
        -0.942478,1.522416,-0.954956,
        -0.942478,1.523102,2.765920,
        -0.942478,1.524811,-2.844718,
        -0.942478,1.527261,-0.304916,
        -0.942478,1.529788,0.613208,
        -0.942478,1.533648,1.644808,
        -0.942478,1.535018,0.028869,
        -0.942478,1.541082,2.424608,
        -0.942478,1.546010,1.523462,
        -0.942478,1.551135,-0.642994,
        -0.942478,1.553895,0.489607,
        -0.942478,1.555041,0.947813,
        -0.942478,1.559671,-1.074934,
        -0.942478,1.560345,2.985899,
        -0.942478,1.560487,-2.323400,
        -0.942478,1.561127,1.401111,
        -0.942478,1.567193,2.643172,
        -0.942478,1.568892,2.098362,
        -0.942478,1.571702,-2.098435,
        -0.942478,1.572701,-2.528730,
        -0.942478,1.573572,1.278593,
        -0.942478,1.574849,-0.091505,
        -0.942478,1.576334,0.365236,
        -0.942478,1.577029,-2.962888,
        -0.942478,1.577688,-1.814941,
        -0.942478,1.579542,-0.755917,
        -0.942478,1.579624,-1.199759,
        -0.942478,1.581515,-0.419414,
        -0.942478,1.584581,-2.738702,
        -0.942478,1.586948,-1.696296,
        -0.942478,1.588901,2.301496,
        -0.942478,1.589936,0.825020,
        -0.942478,1.590018,-1.941652,
        -0.942478,1.596968,-1.328358,
        -0.942478,1.597178,-1.575607,
        -0.942478,1.600470,-1.453196,
        -0.942478,1.601408,0.242409,
        -0.942478,1.602259,1.152656,
        -0.942478,1.603064,2.865774,
        -0.942478,1.612085,1.968802,
        -0.942478,1.612720,-3.080498,
        -0.942478,1.616061,2.521531,
        -0.942478,1.616900,-0.207457,
        -0.942478,1.620988,-0.875920,
        -0.942478,1.622114,0.701509,
        -0.942478,1.622749,0.121426,
        -0.942478,1.622993,-2.221629,
        -0.942478,1.629346,-0.542669,
        -0.942478,1.630007,1.844552,
        -0.942478,1.634747,-2.426198,
        -0.942478,1.642346,1.722617,
        -0.942478,1.643925,1.034629,
        -0.942478,1.644848,2.743774,
        -0.942478,1.645258,2.188109,
        -0.942478,1.645417,-2.634725,
        -0.942478,1.647395,3.085437,
        -0.942478,1.648538,-2.855044,
        -0.942478,1.649932,0.576808,
        -0.942478,1.652224,-0.996105,
        -0.942478,1.652871,1.600681,
        -0.942478,1.657997,-0.325127,
        -0.942478,1.661936,0.001037,
        -0.942478,1.665187,1.477402,
        -0.942478,1.665496,2.397872,
        -0.942478,1.673267,-0.665453,
        -0.942478,1.673557,0.452526,
        -0.942478,1.679277,-1.122421,
        -0.942478,1.679696,0.912482,
        -0.942478,1.682420,1.353922,
        -0.942478,1.685031,-2.009038,
        -0.942478,1.685196,2.967046,
        -0.942478,1.690107,2.620910,
        -0.942478,1.690453,-2.123076,
        -0.942478,1.691693,-1.871572,
        -0.942478,1.691817,2.061688,
        -0.942478,1.692644,1.237670,
        -0.942478,1.696386,0.328123,
        -0.942478,1.697162,-1.249026,
        -0.942478,1.697488,-2.322785,
        -0.942478,1.698726,-2.973576,
        -0.942478,1.699172,-1.749717,
        -0.942478,1.704478,-0.106295,
        -0.942478,1.707244,-0.444770,
        -0.942478,1.707500,-2.531497,
        -0.942478,1.707714,-0.787321,
        -0.942478,1.708430,-1.627552,
        -0.942478,1.710583,-1.376890,
        -0.942478,1.712415,-2.747921,
        -0.942478,1.713125,-1.503073,
        -0.942478,1.713452,0.789419,
        -0.942478,1.720612,2.279888,
        -0.942478,1.724852,0.202831,
        -0.942478,1.724868,2.846246,
        -0.942478,1.728940,1.931535,
        -0.942478,1.732838,1.127584,
        -0.942478,1.735719,-3.092485,
        -0.942478,1.739676,2.496017,
        -0.942478,1.739971,-0.228876,
        -0.942478,1.743328,0.663347,
        -0.942478,1.744927,-0.911892,
        -0.942478,1.745952,0.084733,
        -0.942478,1.748077,1.804354,
        -0.942478,1.752959,-0.567548,
        -0.942478,1.760169,1.678132,
        -0.942478,1.763887,-2.218450,
        -0.942478,1.766737,2.722752,
        -0.942478,1.767999,1.004737,
        -0.942478,1.769149,1.550298,
        -0.942478,1.770119,-2.427175,
        -0.942478,1.770460,2.154208,
        -0.942478,1.770514,3.070982,
        -0.942478,1.770735,0.537245,
        -0.942478,1.772673,-2.864760,
        -0.942478,1.773527,-1.039582,
        -0.942478,1.776419,-2.642421,
        -0.942478,1.780769,1.416330,
        -0.942478,1.786634,-0.346247,
        -0.942478,1.788829,-1.947553,
        -0.942478,1.793737,0.410270,
        -0.942478,1.794553,2.374618,
        -0.942478,1.794699,-0.694061,
        -0.942478,1.797469,-1.169105,
        -0.942478,1.803705,0.881076,
        -0.942478,1.804124,1.264249,
        -0.942478,1.807825,-2.090455,
        -0.942478,1.808822,-1.817375,
        -0.942478,1.809020,2.949741,
        -0.942478,1.812466,2.597595,
        -0.942478,1.813159,2.025395,
        -0.942478,1.813816,-1.298435,
        -0.942478,1.815016,0.278951,
        -0.942478,1.816510,-0.008631,
        -0.942478,1.817772,-1.689625,
        -0.942478,1.821360,-2.984352,
        -0.942478,1.822312,-0.127211,
        -0.942478,1.823386,-1.429807,
        -0.942478,1.823834,-1.560483,
        -0.942478,1.832060,-0.821627,
        -0.942478,1.832690,-0.468440,
        -0.942478,1.833945,-2.320278,
        -0.942478,1.834787,0.751968,
        -0.942478,1.840579,-2.536445,
        -0.942478,1.841033,-2.756932,
        -0.942478,1.846350,1.893983,
        -0.942478,1.846644,1.113766,
        -0.942478,1.846833,2.826543,
        -0.942478,1.847325,2.249051,
        -0.942478,1.854119,0.134161,
        -0.942478,1.858863,-3.106022,
        -0.942478,1.865046,1.763384,
        -0.942478,1.865619,-0.952036,
        -0.942478,1.865912,0.624453,
        -0.942478,1.866694,2.473686,
        -0.942478,1.867524,-0.240083,
        -0.942478,1.875561,1.632831,
        -0.942478,1.876049,-0.594378,
        -0.942478,1.880509,1.500204,
        -0.942478,1.883504,1.360457,
        -0.942478,1.886971,-2.196665,
        -0.942478,1.888979,0.982922,
        -0.942478,1.889815,2.701570,
        -0.942478,1.892144,-2.005340,
        -0.942478,1.892348,0.496522,
        -0.942478,1.893281,-1.083976,
        -0.942478,1.893828,3.053453,
        -0.942478,1.895117,2.121107,
        -0.942478,1.898741,-2.875735,
        -0.942478,1.904429,-2.428029,
        -0.942478,1.908408,-2.648903,
        -0.942478,1.911828,-0.364489,
        -0.942478,1.912224,0.366565,
        -0.942478,1.914854,1.214468,
        -0.942478,1.915467,-1.883143,
        -0.942478,1.915763,-0.724679,
        -0.942478,1.915790,-1.216366,
        -0.942478,1.922256,2.347806,
        -0.942478,1.922616,0.844633,
        -0.942478,1.925185,0.231170,
        -0.942478,1.927474,-1.754616,
        -0.942478,1.928342,-1.351174,
        -0.942478,1.933328,-1.623328,
        -0.942478,1.933878,0.034150,
        -0.942478,1.934585,-1.487979,
        -0.942478,1.934802,2.927634,
        -0.942478,1.934862,1.989776,
        -0.942478,1.936327,-0.099350,
        -0.942478,1.938088,2.578507,
        -0.942478,1.945677,-2.999805,
        -0.942478,1.954572,-0.859208,
        -0.942478,1.957868,-0.490784,
        -0.942478,1.958989,0.712772,
        -0.942478,1.960212,-2.307561,
        -0.942478,1.964774,1.855507,
        -0.942478,1.964971,-2.100533,
        -0.942478,1.969112,1.088273,
        -0.942478,1.971438,-2.765415,
        -0.942478,1.972312,2.805134,
        -0.942478,1.974950,2.220596,
        -0.942478,1.975135,-2.539127,
        -0.942478,1.982107,1.720829,
        -0.942478,1.983195,-3.127984,
        -0.942478,1.983318,-0.247182,
        -0.942478,1.986720,-0.993220,
        -0.942478,1.990381,1.586005,
        -0.942478,1.990470,0.580283,
        -0.942478,1.990523,1.450326,
        -0.942478,1.993909,1.312897,
        -0.942478,1.994694,2.452813,
        -0.942478,1.999226,-0.620717,
        -0.942478,2.005145,0.941012,
        -0.942478,2.007301,0.142337,
        -0.942478,2.014149,0.444138,
        -0.942478,2.014338,-1.127392,
        -0.942478,2.018131,3.021778,
        -0.942478,2.019676,-1.980944,
        -0.942478,2.019912,2.686612,
        -0.942478,2.020545,2.088672,
        -0.942478,2.026074,0.300945,
        -0.942478,2.026196,-2.889057,
        -0.942478,2.031959,-1.268178,
        -0.942478,2.032917,-2.199917,
        -0.942478,2.033572,-1.842884,
        -0.942478,2.034712,-2.419505,
        -0.942478,2.035818,-0.379479,
        -0.942478,2.036004,-0.757062,
        -0.942478,2.040501,-1.411488,
        -0.942478,2.042874,-1.701367,
        -0.942478,2.043060,-2.653746,
        -0.942478,2.043729,-1.557938,
        -0.942478,2.048538,0.806623,
        -0.942478,2.049006,1.199930,
        -0.942478,2.049502,-0.137989,
        -0.942478,2.052181,2.326959,
        -0.942478,2.052794,-0.008526,
        -0.942478,2.057608,1.951726,
        -0.942478,2.065568,2.567730,
        -0.942478,2.070371,2.879808,
        -0.942478,2.071256,-3.018961,
        -0.942478,2.079746,-0.897699,
        -0.942478,2.082784,1.042403,
        -0.942478,2.083752,-0.511086,
        -0.942478,2.084401,1.810502,
        -0.942478,2.085731,0.671081,
        -0.942478,2.097496,1.373172,
        -0.942478,2.099185,1.667006,
        -0.942478,2.100482,-2.301084,
        -0.942478,2.103118,2.194745,
        -0.942478,2.103223,1.521685,
        -0.942478,2.103695,-2.777707,
        -0.942478,2.107523,3.129581,
        -0.942478,2.108277,-2.095907,
        -0.942478,2.108891,-2.532166,
        -0.942478,2.111003,-0.268632,
        -0.942478,2.112276,-1.027246,
        -0.942478,2.114380,0.531457,
        -0.942478,2.119166,0.100552,
        -0.942478,2.119564,2.754818,
        -0.942478,2.122474,-0.648314,
        -0.942478,2.125183,2.445598,
        -0.942478,2.129501,0.238290,
        -0.942478,2.131178,0.386663,
        -0.942478,2.131223,-1.174311,
        -0.942478,2.134269,-1.948042,
        -0.942478,2.134684,0.907065,
        -0.942478,2.137615,2.988806,
        -0.942478,2.145635,-1.327708,
        -0.942478,2.146467,2.055081,
        -0.942478,2.146582,-1.793788,
        -0.942478,2.149210,-1.482321,
        -0.942478,2.152901,-1.636788,
        -0.942478,2.156337,1.151558,
        -0.942478,2.156474,-2.912065,
        -0.942478,2.157194,-0.792563,
        -0.942478,2.166724,-0.401638,
        -0.942478,2.166980,-0.134879,
        -0.942478,2.173644,-2.403429,
        -0.942478,2.175711,-2.655523,
        -0.942478,2.177132,2.600419,
        -0.942478,2.178879,0.767722,
        -0.942478,2.180752,1.908485,
        -0.942478,2.181338,2.313721,
        -0.942478,2.185580,1.287673,
        -0.942478,2.195400,-2.217546,
        -0.942478,2.200418,-3.056442,
        -0.942478,2.201829,-0.001735,
        -0.942478,2.203804,1.756176,
        -0.942478,2.211602,1.443930,
        -0.942478,2.212686,-0.544809,
        -0.942478,2.213937,0.621892,
        -0.942478,2.214075,1.600691,
        -0.942478,2.216256,1.020029,
        -0.942478,2.219060,2.886631,
        -0.942478,2.223375,-0.906185,
        -0.942478,2.226002,-1.060248,
        -0.942478,2.226413,-2.064034,
        -0.942478,2.230554,2.171334,
        -0.942478,2.232413,2.735777,
        -0.942478,2.233709,3.070830,
        -0.942478,2.234383,-2.792065,
        -0.942478,2.237183,0.468799,
        -0.942478,2.237695,-0.271388,
        -0.942478,2.239419,0.151693,
        -0.942478,2.243347,-2.521049,
        -0.942478,2.244035,-1.229456,
        -0.942478,2.244355,0.309899,
        -0.942478,2.245175,2.460236,
        -0.942478,2.245628,-1.899216,
        -0.942478,2.245973,-0.702234,
        -0.942478,2.255601,-1.728131,
        -0.942478,2.256702,-1.401367,
        -0.942478,2.261466,-1.557500,
        -0.942478,2.267390,0.878499,
        -0.942478,2.272099,2.018594,
        -0.942478,2.278618,1.170386,
        -0.942478,2.283304,-2.345584,
        -0.942478,2.285175,-2.943090,
        -0.942478,2.293000,-0.124451,
        -0.942478,2.294058,-0.421070,
        -0.942478,2.303283,1.855938,
        -0.942478,2.304968,2.608978,
        -0.942478,2.305347,2.311346,
        -0.942478,2.307188,-2.656134,
        -0.942478,2.307301,1.336200,
        -0.942478,2.309655,0.726020,
        -0.942478,2.317634,-2.185225,
        -0.942478,2.321510,1.685664,
        -0.942478,2.324376,-3.111006,
        -0.942478,2.324731,1.511186,
        -0.942478,2.329587,0.035498,
        -0.942478,2.330158,2.977264,
        -0.942478,2.335096,-0.588468,
        -0.942478,2.336046,-1.115799,
        -0.942478,2.338060,2.803613,
        -0.942478,2.338083,-0.943400,
        -0.942478,2.340210,0.562145,
        -0.942478,2.341931,1.023926,
        -0.942478,2.343265,-0.779061,
        -0.942478,2.343434,-2.013692,
        -0.942478,2.353029,-1.310529,
        -0.942478,2.354868,0.211543,
        -0.942478,2.355120,0.388593,
        -0.942478,2.355165,-1.833173,
        -0.942478,2.355462,2.151441,
        -0.942478,2.362948,-2.487477,
        -0.942478,2.363101,-2.810456,
        -0.942478,2.363394,-0.272918,
        -0.942478,2.364509,-1.644533,
        -0.942478,2.370465,2.466824,
        -0.942478,2.386110,-1.481351,
        -0.942478,2.394960,0.862133,
        -0.942478,2.395716,1.977184,
        -0.942478,2.396339,1.200900,
        -0.942478,2.407941,-2.316753,
        -0.942478,2.409450,-2.988432,
        -0.942478,2.417202,-0.103363,
        -0.942478,2.417386,-0.443097,
        -0.942478,2.417950,2.666662,
        -0.942478,2.423190,1.789783,
        -0.942478,2.423692,1.392696,
        -0.942478,2.431122,3.086379,
        -0.942478,2.431406,-2.647820,
        -0.942478,2.431584,2.307631,
        -0.942478,2.434090,1.592818,
        -0.942478,2.435993,0.683844,
        -0.942478,2.440898,-2.133300,
        -0.942478,2.442222,-1.195837,
        -0.942478,2.443112,2.883761,
        -0.942478,2.449191,-0.633651,
        -0.942478,2.450821,0.084741,
        -0.942478,2.453499,-1.008861,
        -0.942478,2.460010,-1.938011,
        -0.942478,2.460199,-0.822008,
        -0.942478,2.460902,0.490079,
        -0.942478,2.461774,-1.734913,
        -0.942478,2.463986,1.034407,
        -0.942478,2.467647,0.287667,
        -0.942478,2.479849,2.126925,
        -0.942478,2.485907,-2.835471,
        -0.942478,2.485945,-1.374853,
        -0.942478,2.487747,2.505912,
        -0.942478,2.487957,-0.274279,
        -0.942478,2.489741,-2.468428,
        -0.942478,2.512538,1.239900,
        -0.942478,2.517528,1.923289,
        -0.942478,2.518454,-1.565852,
        -0.942478,2.519348,0.843503,
        -0.942478,2.522766,-3.059231,
        -0.942478,2.528686,2.737228,
        -0.942478,2.534798,-2.266671,
        -0.942478,2.536533,1.466274,
        -0.942478,2.538504,1.699812,
        -0.942478,2.539516,-0.476942,
        -0.942478,2.539532,-0.074548,
        -0.942478,2.543197,2.990252,
        -0.942478,2.554022,2.320619,
        -0.942478,2.556919,-2.655005,
        -0.942478,2.558146,0.627191,
        -0.942478,2.558223,-1.164189,
        -0.942478,2.563452,-2.044028,
        -0.942478,2.568462,0.150655,
        -0.942478,2.569130,-0.700837,
        -0.942478,2.570050,-1.804492,
        -0.942478,2.572710,-0.934665,
        -0.942478,2.575994,0.391718,
        -0.942478,2.584774,1.046703,
        -0.942478,2.601432,2.102535,
        -0.942478,2.603559,2.555114,
        -0.942478,2.604150,-2.885208,
        -0.942478,2.608630,-1.378011,
        -0.942478,2.611256,-0.279110,
        -0.942478,2.617344,-2.439480,
        -0.942478,2.627402,1.293552,
        -0.942478,2.635396,1.848850,
        -0.942478,2.635686,2.834950,
        -0.942478,2.637087,3.132730,
        -0.942478,2.642029,0.813274,
        -0.942478,2.645260,1.570203,
        -0.942478,2.646182,-1.614036,
        -0.942478,2.660113,-2.186513,
        -0.942478,2.660483,-0.034608,
        -0.942478,2.661713,-0.529511,
        -0.942478,2.672720,-1.110422,
        -0.942478,2.676626,2.335353,
        -0.942478,2.676638,0.540364,
        -0.942478,2.676780,-1.901158,
        -0.942478,2.681411,-2.667965,
        -0.942478,2.682152,0.245751,
        -0.942478,2.682859,-0.817071,
        -0.942478,2.705107,1.062142,
        -0.942478,2.717705,2.625995,
        -0.942478,2.721001,2.060097,
        -0.942478,2.722319,-2.954069,
        -0.942478,2.729840,-1.368237,
        -0.942478,2.734349,-0.289298,
        -0.942478,2.738409,2.975671,
        -0.942478,2.740366,1.374341,
        -0.942478,2.744633,-2.391311,
        -0.942478,2.747398,1.727166,
        -0.942478,2.763799,0.760372,
        -0.942478,2.768238,-1.679243,
        -0.942478,2.779245,-2.046268,
        -0.942478,2.780208,-0.626453,
        -0.942478,2.780471,0.027732,
        -0.942478,2.784696,-1.013128,
        -0.942478,2.789817,0.395431,
        -0.942478,2.801154,2.353143,
        -0.942478,2.807091,-2.687477,
        -0.942478,2.825762,1.085035,
        -0.942478,2.830209,2.734943,
        -0.942478,2.833517,-3.085508,
        -0.942478,2.838301,1.980061,
        -0.942478,2.849871,1.512012,
        -0.942478,2.850089,-1.343899,
        -0.942478,2.856905,-0.318441,
        -0.942478,2.870445,-2.291001,
        -0.942478,2.885017,0.655562,
        -0.942478,2.886428,-1.793217,
        -0.942478,2.890337,-0.825633,
        -0.942478,2.898932,0.139973,
        -0.942478,2.931927,-2.728768,
        -0.942478,2.932096,2.377790,
        -0.942478,2.937409,2.947245,
        -0.942478,2.948048,1.133127,
        -0.942478,2.950896,1.797889,
        -0.942478,2.970019,-1.282382,
        -0.942478,2.978333,-0.412090,
        -0.942478,2.996437,-2.089121,
        -0.942478,3.011274,0.434110,
        -0.942478,3.050410,3.138492,
        -0.942478,3.064492,1.647259,
        -0.942478,3.088868,-0.989501,
        -0.837758,0.053884,-2.408394,
        -0.837758,0.077645,1.453433,
        -0.837758,0.088408,-0.311160,
        -0.837758,0.142314,2.590387,
        -0.837758,0.147206,-1.385164,
        -0.837758,0.163506,0.573203,
        -0.837758,0.165485,-2.875396,
        -0.837758,0.182564,-2.125821,
        -0.837758,0.190569,1.929141,
        -0.837758,0.201102,1.249082,
        -0.837758,0.201397,-0.716646,
        -0.837758,0.210568,-0.069559,
        -0.837758,0.251672,2.848900,
        -0.837758,0.266262,-1.678068,
        -0.837758,0.269108,-1.215563,
        -0.837758,0.273777,-2.506756,
        -0.837758,0.279854,0.814581,
        -0.837758,0.281664,-2.969147,
        -0.837758,0.283625,2.405603,
        -0.837758,0.284058,0.357350,
        -0.837758,0.294107,1.593880,
        -0.837758,0.302491,-0.400410,
        -0.837758,0.308059,2.021491,
        -0.837758,0.312219,-2.072057,
        -0.837758,0.321905,-0.829121,
        -0.837758,0.329984,1.201621,
        -0.837758,0.339714,-0.013712,
        -0.837758,0.368784,2.974974,
        -0.837758,0.376428,-1.495532,
        -0.837758,0.378103,2.646436,
        -0.837758,0.385088,0.590355,
        -0.837758,0.386248,-2.690943,
        -0.837758,0.389851,-2.340031,
        -0.837758,0.391502,-1.171374,
        -0.837758,0.398249,-1.812694,
        -0.837758,0.400542,-2.995729,
        -0.837758,0.404273,0.918163,
        -0.837758,0.405617,1.773533,
        -0.837758,0.409009,-0.570243,
        -0.837758,0.410700,0.270433,
        -0.837758,0.415893,1.464890,
        -0.837758,0.420680,2.088164,
        -0.837758,0.424766,-0.258729,
        -0.837758,0.426507,2.367872,
        -0.837758,0.441292,-0.890727,
        -0.837758,0.451839,-2.078193,
        -0.837758,0.459270,1.191562,
        -0.837758,0.470622,0.014002,
        -0.837758,0.481292,-2.529098,
        -0.837758,0.482460,2.849843,
        -0.837758,0.492205,-1.397963,
        -0.837758,0.493314,3.117264,
        -0.837758,0.497870,0.727828,
        -0.837758,0.499783,-1.654379,
        -0.837758,0.502651,0.463333,
        -0.837758,0.505463,2.594557,
        -0.837758,0.506104,-2.874910,
        -0.837758,0.513579,-1.147354,
        -0.837758,0.518569,1.644237,
        -0.837758,0.519001,1.900488,
        -0.837758,0.519277,-0.673809,
        -0.837758,0.524769,-0.421955,
        -0.837758,0.525564,-2.288838,
        -0.837758,0.532541,0.978737,
        -0.837758,0.534610,2.154808,
        -0.837758,0.535020,-1.891846,
        -0.837758,0.540208,1.400497,
        -0.837758,0.542006,0.228337,
        -0.837758,0.550762,-0.180771,
        -0.837758,0.560531,2.378474,
        -0.837758,0.561629,-0.924576,
        -0.837758,0.567465,-2.679392,
        -0.837758,0.588527,1.181755,
        -0.837758,0.591643,-2.098364,
        -0.837758,0.594963,2.997575,
        -0.837758,0.599643,-3.045235,
        -0.837758,0.599962,2.773556,
        -0.837758,0.602620,0.029270,
        -0.837758,0.605711,0.598454,
        -0.837758,0.607708,-1.541026,
        -0.837758,0.609360,-2.464379,
        -0.837758,0.610738,-1.326786,
        -0.837758,0.616968,0.820489,
        -0.837758,0.626379,1.773610,
        -0.837758,0.628626,-1.751180,
        -0.837758,0.630671,0.391949,
        -0.837758,0.632622,-0.737642,
        -0.837758,0.632708,2.566722,
        -0.837758,0.632797,1.989858,
        -0.837758,0.634455,-0.534790,
        -0.837758,0.636205,1.561170,
        -0.837758,0.637681,-1.123423,
        -0.837758,0.641441,-2.842109,
        -0.837758,0.644807,-0.327139,
        -0.837758,0.652440,2.197129,
        -0.837758,0.660897,-2.278593,
        -0.837758,0.664008,1.360091,
        -0.837758,0.670850,-1.945257,
        -0.837758,0.671396,1.004859,
        -0.837758,0.673985,0.203855,
        -0.837758,0.680326,-0.133370,
        -0.837758,0.685083,-0.942363,
        -0.837758,0.688204,-2.632204,
        -0.837758,0.689777,2.385011,
        -0.837758,0.700065,3.115430,
        -0.837758,0.704080,2.911636,
        -0.837758,0.712883,1.174216,
        -0.837758,0.715701,0.699906,
        -0.837758,0.718851,-1.442215,
        -0.837758,0.721019,2.721536,
        -0.837758,0.725300,-2.975134,
        -0.837758,0.727136,-1.633288,
        -0.837758,0.727399,-2.120567,
        -0.837758,0.730751,0.522488,
        -0.837758,0.732892,-1.254330,
        -0.837758,0.734333,-2.446184,
        -0.837758,0.735125,0.037270,
        -0.837758,0.737893,1.866463,
        -0.837758,0.740650,1.681075,
        -0.837758,0.745971,-0.781286,
        -0.837758,0.749432,-0.442559,
        -0.837758,0.750102,2.050636,
        -0.837758,0.751046,0.863821,
        -0.837758,0.757018,-0.608833,
        -0.837758,0.757046,-1.816019,
        -0.837758,0.757389,1.501184,
        -0.837758,0.759800,2.546181,
        -0.837758,0.759859,0.351923,
        -0.837758,0.761141,-2.781102,
        -0.837758,0.769035,-1.078574,
        -0.837758,0.770166,-0.266604,
        -0.837758,0.772218,2.227380,
        -0.837758,0.788382,1.331051,
        -0.837758,0.790269,-2.282409,
        -0.837758,0.803410,-1.984567,
        -0.837758,0.803813,0.191520,
        -0.837758,0.806832,3.022155,
        -0.837758,0.807476,-2.597675,
        -0.837758,0.812512,-0.107335,
        -0.837758,0.814161,1.010949,
        -0.837758,0.817732,-3.084486,
        -0.837758,0.817846,2.390688,
        -0.837758,0.819323,2.845960,
        -0.837758,0.819430,-0.915017,
        -0.837758,0.829198,-1.532491,
        -0.837758,0.829531,0.665695,
        -0.837758,0.829816,1.167147,
        -0.837758,0.833613,-1.357492,
        -0.837758,0.842945,-2.907318,
        -0.837758,0.844883,2.679708,
        -0.837758,0.848189,-1.705003,
        -0.837758,0.849888,1.767901,
        -0.837758,0.851126,0.493370,
        -0.837758,0.852911,1.933394,
        -0.837758,0.857516,-2.432863,
        -0.837758,0.858261,-1.191729,
        -0.837758,0.858279,-2.138895,
        -0.837758,0.859723,-0.734928,
        -0.837758,0.859754,1.603292,
        -0.837758,0.861655,-0.394821,
        -0.837758,0.864300,0.043549,
        -0.837758,0.866940,0.851229,
        -0.837758,0.869748,2.095969,
        -0.837758,0.877352,-0.574668,
        -0.837758,0.879254,-2.733338,
        -0.837758,0.882554,1.443757,
        -0.837758,0.884060,0.334199,
        -0.837758,0.885626,-1.866061,
        -0.837758,0.889834,2.529957,
        -0.837758,0.893140,2.251698,
        -0.837758,0.897463,-0.240822,
        -0.837758,0.898758,-1.031948,
        -0.837758,0.913918,1.287618,
        -0.837758,0.916555,-2.287173,
        -0.837758,0.918305,3.107845,
        -0.837758,0.919522,2.948277,
        -0.837758,0.927138,-2.569686,
        -0.837758,0.929031,0.185383,
        -0.837758,0.933766,-3.015917,
        -0.837758,0.934873,-2.010822,
        -0.837758,0.935636,-1.445855,
        -0.837758,0.936702,-0.866461,
        -0.837758,0.937395,0.729108,
        -0.837758,0.937910,2.787575,
        -0.837758,0.939144,-0.092962,
        -0.837758,0.941718,1.125085,
        -0.837758,0.942873,-1.607680,
        -0.837758,0.943701,0.980328,
        -0.837758,0.947774,2.395256,
        -0.837758,0.952433,0.584347,
        -0.837758,0.954456,-1.287293,
        -0.837758,0.958731,-2.852552,
        -0.837758,0.961838,1.839284,
        -0.837758,0.961898,-0.469425,
        -0.837758,0.965426,1.686615,
        -0.837758,0.970108,1.990771,
        -0.837758,0.972820,-1.762780,
        -0.837758,0.973500,-0.711881,
        -0.837758,0.975253,2.641082,
        -0.837758,0.977824,0.441928,
        -0.837758,0.981485,-2.424766,
        -0.837758,0.981877,1.537100,
        -0.837758,0.982609,-1.136920,
        -0.837758,0.984265,-2.156073,
        -0.837758,0.989150,0.045493,
        -0.837758,0.993052,2.137478,
        -0.837758,0.997766,-2.693871,
        -0.837758,0.999847,-0.334738,
        -0.837758,1.001920,0.848054,
        -0.837758,1.006146,1.389734,
        -0.837758,1.012180,0.304066,
        -0.837758,1.016938,2.270674,
        -0.837758,1.017371,-0.979790,
        -0.837758,1.019495,-1.894487,
        -0.837758,1.023796,-0.583691,
        -0.837758,1.026525,2.506945,
        -0.837758,1.027004,-0.203653,
        -0.837758,1.028027,3.036573,
        -0.837758,1.031435,1.241021,
        -0.837758,1.032316,2.884221,
        -0.837758,1.035863,-3.110475,
        -0.837758,1.042085,-2.292950,
        -0.837758,1.043272,-1.522986,
        -0.837758,1.046265,-2.958847,
        -0.837758,1.049272,-1.376344,
        -0.837758,1.050039,-2.552086,
        -0.837758,1.053682,0.689863,
        -0.837758,1.055458,0.170394,
        -0.837758,1.056622,-0.831866,
        -0.837758,1.057515,1.097998,
        -0.837758,1.058706,-2.036438,
        -0.837758,1.061389,-1.672248,
        -0.837758,1.063758,2.743093,
        -0.837758,1.070157,-0.073422,
        -0.837758,1.072154,1.761506,
        -0.837758,1.074029,1.909193,
        -0.837758,1.074204,0.545276,
        -0.837758,1.074714,-2.805147,
        -0.837758,1.076262,0.966104,
        -0.837758,1.077642,-1.227073,
        -0.837758,1.084151,1.618525,
        -0.837758,1.085085,2.377673,
        -0.837758,1.086822,-0.441112,
        -0.837758,1.087142,2.058815,
        -0.837758,1.098383,-0.696430,
        -0.837758,1.099138,-1.084033,
        -0.837758,1.103208,1.477773,
        -0.837758,1.103783,0.408254,
        -0.837758,1.106275,-2.420800,
        -0.837758,1.108079,2.612943,
        -0.837758,1.109345,-2.172718,
        -0.837758,1.115957,-1.787597,
        -0.837758,1.118930,0.045983,
        -0.837758,1.119777,0.834702,
        -0.837758,1.121943,-2.669042,
        -0.837758,1.124930,2.219917,
        -0.837758,1.125217,1.336294,
        -0.837758,1.130958,2.972931,
        -0.837758,1.131400,-0.306629,
        -0.837758,1.137135,-0.936514,
        -0.837758,1.139125,-1.924528,
        -0.837758,1.139877,0.276382,
        -0.837758,1.142003,-3.058022,
        -0.837758,1.147363,-0.558448,
        -0.837758,1.150348,-0.176418,
        -0.837758,1.151198,1.197016,
        -0.837758,1.153930,-1.597104,
        -0.837758,1.155389,2.837148,
        -0.837758,1.155458,-1.460623,
        -0.837758,1.157331,-2.906362,
        -0.837758,1.158594,2.487618,
        -0.837758,1.162103,-1.322553,
        -0.837758,1.165407,3.092861,
        -0.837758,1.168025,0.717075,
        -0.837758,1.168406,-2.301519,
        -0.837758,1.174782,-2.538755,
        -0.837758,1.177998,1.061256,
        -0.837758,1.178898,1.827305,
        -0.837758,1.181331,-2.060330,
        -0.837758,1.182540,-0.798952,
        -0.837758,1.184569,0.147517,
        -0.837758,1.184734,1.968919,
        -0.837758,1.184804,0.593386,
        -0.837758,1.188278,1.690808,
        -0.837758,1.194360,2.110440,
        -0.837758,1.194776,2.710455,
        -0.837758,1.197745,-1.173705,
        -0.837758,1.199898,-2.777257,
        -0.837758,1.201164,2.355931,
        -0.837758,1.203760,1.556404,
        -0.837758,1.205845,-0.415993,
        -0.837758,1.206585,0.933648,
        -0.837758,1.206950,-0.037248,
        -0.837758,1.209341,0.471162,
        -0.837758,1.213483,-1.039417,
        -0.837758,1.218146,-0.672643,
        -0.837758,1.221608,1.421669,
        -0.837758,1.226043,-1.693533,
        -0.837758,1.232870,-1.822773,
        -0.837758,1.234375,-2.421194,
        -0.837758,1.234531,-2.190148,
        -0.837758,1.239414,0.348959,
        -0.837758,1.240183,2.587255,
        -0.837758,1.242312,2.941234,
        -0.837758,1.244075,1.286644,
        -0.837758,1.245961,-2.997461,
        -0.837758,1.246495,-2.646872,
        -0.837758,1.247736,2.242810,
        -0.837758,1.253604,0.812222,
        -0.837758,1.257816,-0.894317,
        -0.837758,1.258247,-1.953634,
        -0.837758,1.258993,-0.284099,
        -0.837758,1.262380,-0.157609,
        -0.837758,1.262791,-1.549229,
        -0.837758,1.263772,-0.533073,
        -0.837758,1.266731,-1.412461,
        -0.837758,1.268591,-3.122146,
        -0.837758,1.269614,-1.275246,
        -0.837758,1.270745,1.152533,
        -0.837758,1.273606,0.224545,
        -0.837758,1.279716,2.812101,
        -0.837758,1.283786,2.462124,
        -0.837758,1.284798,0.078209,
        -0.837758,1.285402,-2.882049,
        -0.837758,1.286936,1.886053,
        -0.837758,1.289408,0.683122,
        -0.837758,1.294736,1.756184,
        -0.837758,1.296978,-2.311511,
        -0.837758,1.298546,2.020601,
        -0.837758,1.300924,1.022478,
        -0.837758,1.303305,3.043152,
        -0.837758,1.304210,-2.084225,
        -0.837758,1.305241,-2.530249,
        -0.837758,1.307513,1.628003,
        -0.837758,1.310722,0.558267,
        -0.837758,1.314142,2.139329,
        -0.837758,1.316087,-0.777971,
        -0.837758,1.317907,-1.121023,
        -0.837758,1.321019,-2.749776,
        -0.837758,1.321609,2.687717,
        -0.837758,1.322568,1.499415,
        -0.837758,1.322972,-0.395699,
        -0.837758,1.325286,-0.994192,
        -0.837758,1.329146,-0.648513,
        -0.837758,1.329795,-0.042570,
        -0.837758,1.331651,2.344497,
        -0.837758,1.336522,0.435098,
        -0.837758,1.339410,0.900087,
        -0.837758,1.339550,1.370353,
        -0.837758,1.340441,-1.731580,
        -0.837758,1.351855,-1.855234,
        -0.837758,1.360161,-3.018404,
        -0.837758,1.361031,-1.613971,
        -0.837758,1.361196,2.914780,
        -0.837758,1.361998,-2.207254,
        -0.837758,1.362302,1.239384,
        -0.837758,1.365228,2.563808,
        -0.837758,1.365452,0.310522,
        -0.837758,1.367028,-2.420041,
        -0.837758,1.373422,-1.484281,
        -0.837758,1.374245,-1.226417,
        -0.837758,1.375020,-0.170393,
        -0.837758,1.376560,-1.355481,
        -0.837758,1.377070,-1.980984,
        -0.837758,1.377578,-2.636339,
        -0.837758,1.378096,0.774626,
        -0.837758,1.379086,-0.508192,
        -0.837758,1.382562,0.177141,
        -0.837758,1.388849,2.233847,
        -0.837758,1.389268,-0.886349,
        -0.837758,1.391324,1.109396,
        -0.837758,1.391699,-0.288725,
        -0.837758,1.392869,3.136903,
        -0.837758,1.393612,1.952016,
        -0.837758,1.395545,-2.850807,
        -0.837758,1.401926,1.822930,
        -0.837758,1.402149,2.788898,
        -0.837758,1.409552,0.647980,
        -0.837758,1.410840,0.056219,
        -0.837758,1.411944,2.444377,
        -0.837758,1.413639,1.697982,
        -0.837758,1.426479,1.573427,
        -0.837758,1.426725,2.102242,
        -0.837758,1.427868,0.984637,
        -0.837758,1.429580,-2.107232,
        -0.837758,1.431375,-2.313269,
        -0.837758,1.431907,3.015410,
        -0.837758,1.432325,-1.029054,
        -0.837758,1.433341,0.524268,
        -0.837758,1.435181,-0.739771,
        -0.837758,1.437208,-2.529085,
        -0.837758,1.439041,-0.617966,
        -0.837758,1.441473,1.448544,
        -0.837758,1.444527,2.664744,
        -0.837758,1.450513,-0.400625,
        -0.837758,1.451907,-0.066823,
        -0.837758,1.452754,-2.738631,
        -0.837758,1.454061,-2.950416,
        -0.837758,1.456736,1.323154,
        -0.837758,1.458370,0.400433,
        -0.837758,1.459314,-1.765003,
        -0.837758,1.459902,-1.145835,
        -0.837758,1.464952,2.329245,
        -0.837758,1.466050,0.862104,
        -0.837758,1.471330,-1.886160,
        -0.837758,1.475119,-1.648255,
        -0.837758,1.480878,1.194576,
        -0.837758,1.480942,0.273235,
        -0.837758,1.481086,2.890269,
        -0.837758,1.484118,-1.529156,
        -0.837758,1.484741,-1.279219,
        -0.837758,1.487681,-1.406069,
        -0.837758,1.487886,-3.066037,
        -0.837758,1.490555,2.543284,
        -0.837758,1.492424,-0.190932,
        -0.837758,1.495002,-0.839810,
        -0.837758,1.495297,2.007188,
        -0.837758,1.496414,-2.009267,
        -0.837758,1.498261,-2.425270,
        -0.837758,1.498907,-2.207327,
        -0.837758,1.500204,0.737638,
        -0.837758,1.502469,0.151156,
        -0.837758,1.507448,-0.514782,
        -0.837758,1.509287,2.204647,
        -0.837758,1.510518,1.884920,
        -0.837758,1.511948,-2.632360,
        -0.837758,1.515235,1.068565,
        -0.837758,1.522237,3.102018,
        -0.837758,1.522373,1.764991,
        -0.837758,1.522416,-0.954956,
        -0.837758,1.523102,2.765920,
        -0.837758,1.524811,-2.844718,
        -0.837758,1.527261,-0.304916,
        -0.837758,1.529788,0.613208,
        -0.837758,1.533648,1.644808,
        -0.837758,1.535018,0.028869,
        -0.837758,1.541082,2.424608,
        -0.837758,1.546010,1.523462,
        -0.837758,1.551135,-0.642994,
        -0.837758,1.553895,0.489607,
        -0.837758,1.555041,0.947813,
        -0.837758,1.559671,-1.074934,
        -0.837758,1.560345,2.985899,
        -0.837758,1.560487,-2.323400,
        -0.837758,1.561127,1.401111,
        -0.837758,1.567193,2.643172,
        -0.837758,1.568892,2.098362,
        -0.837758,1.571702,-2.098435,
        -0.837758,1.572701,-2.528730,
        -0.837758,1.573572,1.278593,
        -0.837758,1.574849,-0.091505,
        -0.837758,1.576334,0.365236,
        -0.837758,1.577029,-2.962888,
        -0.837758,1.577688,-1.814941,
        -0.837758,1.579542,-0.755917,
        -0.837758,1.579624,-1.199759,
        -0.837758,1.581515,-0.419414,
        -0.837758,1.584581,-2.738702,
        -0.837758,1.586948,-1.696296,
        -0.837758,1.588901,2.301496,
        -0.837758,1.589936,0.825020,
        -0.837758,1.590018,-1.941652,
        -0.837758,1.596968,-1.328358,
        -0.837758,1.597178,-1.575607,
        -0.837758,1.600470,-1.453196,
        -0.837758,1.601408,0.242409,
        -0.837758,1.602259,1.152656,
        -0.837758,1.603064,2.865774,
        -0.837758,1.612085,1.968802,
        -0.837758,1.612720,-3.080498,
        -0.837758,1.616061,2.521531,
        -0.837758,1.616900,-0.207457,
        -0.837758,1.620988,-0.875920,
        -0.837758,1.622114,0.701509,
        -0.837758,1.622749,0.121426,
        -0.837758,1.622993,-2.221629,
        -0.837758,1.629346,-0.542669,
        -0.837758,1.630007,1.844552,
        -0.837758,1.634747,-2.426198,
        -0.837758,1.642346,1.722617,
        -0.837758,1.643925,1.034629,
        -0.837758,1.644848,2.743774,
        -0.837758,1.645258,2.188109,
        -0.837758,1.645417,-2.634725,
        -0.837758,1.647395,3.085437,
        -0.837758,1.648538,-2.855044,
        -0.837758,1.649932,0.576808,
        -0.837758,1.652224,-0.996105,
        -0.837758,1.652871,1.600681,
        -0.837758,1.657997,-0.325127,
        -0.837758,1.661936,0.001037,
        -0.837758,1.665187,1.477402,
        -0.837758,1.665496,2.397872,
        -0.837758,1.673267,-0.665453,
        -0.837758,1.673557,0.452526,
        -0.837758,1.679277,-1.122421,
        -0.837758,1.679696,0.912482,
        -0.837758,1.682420,1.353922,
        -0.837758,1.685031,-2.009038,
        -0.837758,1.685196,2.967046,
        -0.837758,1.690107,2.620910,
        -0.837758,1.690453,-2.123076,
        -0.837758,1.691693,-1.871572,
        -0.837758,1.691817,2.061688,
        -0.837758,1.692644,1.237670,
        -0.837758,1.696386,0.328123,
        -0.837758,1.697162,-1.249026,
        -0.837758,1.697488,-2.322785,
        -0.837758,1.698726,-2.973576,
        -0.837758,1.699172,-1.749717,
        -0.837758,1.704478,-0.106295,
        -0.837758,1.707244,-0.444770,
        -0.837758,1.707500,-2.531497,
        -0.837758,1.707714,-0.787321,
        -0.837758,1.708430,-1.627552,
        -0.837758,1.710583,-1.376890,
        -0.837758,1.712415,-2.747921,
        -0.837758,1.713125,-1.503073,
        -0.837758,1.713452,0.789419,
        -0.837758,1.720612,2.279888,
        -0.837758,1.724852,0.202831,
        -0.837758,1.724868,2.846246,
        -0.837758,1.728940,1.931535,
        -0.837758,1.732838,1.127584,
        -0.837758,1.735719,-3.092485,
        -0.837758,1.739676,2.496017,
        -0.837758,1.739971,-0.228876,
        -0.837758,1.743328,0.663347,
        -0.837758,1.744927,-0.911892,
        -0.837758,1.745952,0.084733,
        -0.837758,1.748077,1.804354,
        -0.837758,1.752959,-0.567548,
        -0.837758,1.760169,1.678132,
        -0.837758,1.763887,-2.218450,
        -0.837758,1.766737,2.722752,
        -0.837758,1.767999,1.004737,
        -0.837758,1.769149,1.550298,
        -0.837758,1.770119,-2.427175,
        -0.837758,1.770460,2.154208,
        -0.837758,1.770514,3.070982,
        -0.837758,1.770735,0.537245,
        -0.837758,1.772673,-2.864760,
        -0.837758,1.773527,-1.039582,
        -0.837758,1.776419,-2.642421,
        -0.837758,1.780769,1.416330,
        -0.837758,1.786634,-0.346247,
        -0.837758,1.788829,-1.947553,
        -0.837758,1.793737,0.410270,
        -0.837758,1.794553,2.374618,
        -0.837758,1.794699,-0.694061,
        -0.837758,1.797469,-1.169105,
        -0.837758,1.803705,0.881076,
        -0.837758,1.804124,1.264249,
        -0.837758,1.807825,-2.090455,
        -0.837758,1.808822,-1.817375,
        -0.837758,1.809020,2.949741,
        -0.837758,1.812466,2.597595,
        -0.837758,1.813159,2.025395,
        -0.837758,1.813816,-1.298435,
        -0.837758,1.815016,0.278951,
        -0.837758,1.816510,-0.008631,
        -0.837758,1.817772,-1.689625,
        -0.837758,1.821360,-2.984352,
        -0.837758,1.822312,-0.127211,
        -0.837758,1.823386,-1.429807,
        -0.837758,1.823834,-1.560483,
        -0.837758,1.832060,-0.821627,
        -0.837758,1.832690,-0.468440,
        -0.837758,1.833945,-2.320278,
        -0.837758,1.834787,0.751968,
        -0.837758,1.840579,-2.536445,
        -0.837758,1.841033,-2.756932,
        -0.837758,1.846350,1.893983,
        -0.837758,1.846644,1.113766,
        -0.837758,1.846833,2.826543,
        -0.837758,1.847325,2.249051,
        -0.837758,1.854119,0.134161,
        -0.837758,1.858863,-3.106022,
        -0.837758,1.865046,1.763384,
        -0.837758,1.865619,-0.952036,
        -0.837758,1.865912,0.624453,
        -0.837758,1.866694,2.473686,
        -0.837758,1.867524,-0.240083,
        -0.837758,1.875561,1.632831,
        -0.837758,1.876049,-0.594378,
        -0.837758,1.880509,1.500204,
        -0.837758,1.883504,1.360457,
        -0.837758,1.886971,-2.196665,
        -0.837758,1.888979,0.982922,
        -0.837758,1.889815,2.701570,
        -0.837758,1.892144,-2.005340,
        -0.837758,1.892348,0.496522,
        -0.837758,1.893281,-1.083976,
        -0.837758,1.893828,3.053453,
        -0.837758,1.895117,2.121107,
        -0.837758,1.898741,-2.875735,
        -0.837758,1.904429,-2.428029,
        -0.837758,1.908408,-2.648903,
        -0.837758,1.911828,-0.364489,
        -0.837758,1.912224,0.366565,
        -0.837758,1.914854,1.214468,
        -0.837758,1.915467,-1.883143,
        -0.837758,1.915763,-0.724679,
        -0.837758,1.915790,-1.216366,
        -0.837758,1.922256,2.347806,
        -0.837758,1.922616,0.844633,
        -0.837758,1.925185,0.231170,
        -0.837758,1.927474,-1.754616,
        -0.837758,1.928342,-1.351174,
        -0.837758,1.933328,-1.623328,
        -0.837758,1.933878,0.034150,
        -0.837758,1.934585,-1.487979,
        -0.837758,1.934802,2.927634,
        -0.837758,1.934862,1.989776,
        -0.837758,1.936327,-0.099350,
        -0.837758,1.938088,2.578507,
        -0.837758,1.945677,-2.999805,
        -0.837758,1.954572,-0.859208,
        -0.837758,1.957868,-0.490784,
        -0.837758,1.958989,0.712772,
        -0.837758,1.960212,-2.307561,
        -0.837758,1.964774,1.855507,
        -0.837758,1.964971,-2.100533,
        -0.837758,1.969112,1.088273,
        -0.837758,1.971438,-2.765415,
        -0.837758,1.972312,2.805134,
        -0.837758,1.974950,2.220596,
        -0.837758,1.975135,-2.539127,
        -0.837758,1.982107,1.720829,
        -0.837758,1.983195,-3.127984,
        -0.837758,1.983318,-0.247182,
        -0.837758,1.986720,-0.993220,
        -0.837758,1.990381,1.586005,
        -0.837758,1.990470,0.580283,
        -0.837758,1.990523,1.450326,
        -0.837758,1.993909,1.312897,
        -0.837758,1.994694,2.452813,
        -0.837758,1.999226,-0.620717,
        -0.837758,2.005145,0.941012,
        -0.837758,2.007301,0.142337,
        -0.837758,2.014149,0.444138,
        -0.837758,2.014338,-1.127392,
        -0.837758,2.018131,3.021778,
        -0.837758,2.019676,-1.980944,
        -0.837758,2.019912,2.686612,
        -0.837758,2.020545,2.088672,
        -0.837758,2.026074,0.300945,
        -0.837758,2.026196,-2.889057,
        -0.837758,2.031959,-1.268178,
        -0.837758,2.032917,-2.199917,
        -0.837758,2.033572,-1.842884,
        -0.837758,2.034712,-2.419505,
        -0.837758,2.035818,-0.379479,
        -0.837758,2.036004,-0.757062,
        -0.837758,2.040501,-1.411488,
        -0.837758,2.042874,-1.701367,
        -0.837758,2.043060,-2.653746,
        -0.837758,2.043729,-1.557938,
        -0.837758,2.048538,0.806623,
        -0.837758,2.049006,1.199930,
        -0.837758,2.049502,-0.137989,
        -0.837758,2.052181,2.326959,
        -0.837758,2.052794,-0.008526,
        -0.837758,2.057608,1.951726,
        -0.837758,2.065568,2.567730,
        -0.837758,2.070371,2.879808,
        -0.837758,2.071256,-3.018961,
        -0.837758,2.079746,-0.897699,
        -0.837758,2.082784,1.042403,
        -0.837758,2.083752,-0.511086,
        -0.837758,2.084401,1.810502,
        -0.837758,2.085731,0.671081,
        -0.837758,2.097496,1.373172,
        -0.837758,2.099185,1.667006,
        -0.837758,2.100482,-2.301084,
        -0.837758,2.103118,2.194745,
        -0.837758,2.103223,1.521685,
        -0.837758,2.103695,-2.777707,
        -0.837758,2.107523,3.129581,
        -0.837758,2.108277,-2.095907,
        -0.837758,2.108891,-2.532166,
        -0.837758,2.111003,-0.268632,
        -0.837758,2.112276,-1.027246,
        -0.837758,2.114380,0.531457,
        -0.837758,2.119166,0.100552,
        -0.837758,2.119564,2.754818,
        -0.837758,2.122474,-0.648314,
        -0.837758,2.125183,2.445598,
        -0.837758,2.129501,0.238290,
        -0.837758,2.131178,0.386663,
        -0.837758,2.131223,-1.174311,
        -0.837758,2.134269,-1.948042,
        -0.837758,2.134684,0.907065,
        -0.837758,2.137615,2.988806,
        -0.837758,2.145635,-1.327708,
        -0.837758,2.146467,2.055081,
        -0.837758,2.146582,-1.793788,
        -0.837758,2.149210,-1.482321,
        -0.837758,2.152901,-1.636788,
        -0.837758,2.156337,1.151558,
        -0.837758,2.156474,-2.912065,
        -0.837758,2.157194,-0.792563,
        -0.837758,2.166724,-0.401638,
        -0.837758,2.166980,-0.134879,
        -0.837758,2.173644,-2.403429,
        -0.837758,2.175711,-2.655523,
        -0.837758,2.177132,2.600419,
        -0.837758,2.178879,0.767722,
        -0.837758,2.180752,1.908485,
        -0.837758,2.181338,2.313721,
        -0.837758,2.185580,1.287673,
        -0.837758,2.195400,-2.217546,
        -0.837758,2.200418,-3.056442,
        -0.837758,2.201829,-0.001735,
        -0.837758,2.203804,1.756176,
        -0.837758,2.211602,1.443930,
        -0.837758,2.212686,-0.544809,
        -0.837758,2.213937,0.621892,
        -0.837758,2.214075,1.600691,
        -0.837758,2.216256,1.020029,
        -0.837758,2.219060,2.886631,
        -0.837758,2.223375,-0.906185,
        -0.837758,2.226002,-1.060248,
        -0.837758,2.226413,-2.064034,
        -0.837758,2.230554,2.171334,
        -0.837758,2.232413,2.735777,
        -0.837758,2.233709,3.070830,
        -0.837758,2.234383,-2.792065,
        -0.837758,2.237183,0.468799,
        -0.837758,2.237695,-0.271388,
        -0.837758,2.239419,0.151693,
        -0.837758,2.243347,-2.521049,
        -0.837758,2.244035,-1.229456,
        -0.837758,2.244355,0.309899,
        -0.837758,2.245175,2.460236,
        -0.837758,2.245628,-1.899216,
        -0.837758,2.245973,-0.702234,
        -0.837758,2.255601,-1.728131,
        -0.837758,2.256702,-1.401367,
        -0.837758,2.261466,-1.557500,
        -0.837758,2.267390,0.878499,
        -0.837758,2.272099,2.018594,
        -0.837758,2.278618,1.170386,
        -0.837758,2.283304,-2.345584,
        -0.837758,2.285175,-2.943090,
        -0.837758,2.293000,-0.124451,
        -0.837758,2.294058,-0.421070,
        -0.837758,2.303283,1.855938,
        -0.837758,2.304968,2.608978,
        -0.837758,2.305347,2.311346,
        -0.837758,2.307188,-2.656134,
        -0.837758,2.307301,1.336200,
        -0.837758,2.309655,0.726020,
        -0.837758,2.317634,-2.185225,
        -0.837758,2.321510,1.685664,
        -0.837758,2.324376,-3.111006,
        -0.837758,2.324731,1.511186,
        -0.837758,2.329587,0.035498,
        -0.837758,2.330158,2.977264,
        -0.837758,2.335096,-0.588468,
        -0.837758,2.336046,-1.115799,
        -0.837758,2.338060,2.803613,
        -0.837758,2.338083,-0.943400,
        -0.837758,2.340210,0.562145,
        -0.837758,2.341931,1.023926,
        -0.837758,2.343265,-0.779061,
        -0.837758,2.343434,-2.013692,
        -0.837758,2.353029,-1.310529,
        -0.837758,2.354868,0.211543,
        -0.837758,2.355120,0.388593,
        -0.837758,2.355165,-1.833173,
        -0.837758,2.355462,2.151441,
        -0.837758,2.362948,-2.487477,
        -0.837758,2.363101,-2.810456,
        -0.837758,2.363394,-0.272918,
        -0.837758,2.364509,-1.644533,
        -0.837758,2.370465,2.466824,
        -0.837758,2.386110,-1.481351,
        -0.837758,2.394960,0.862133,
        -0.837758,2.395716,1.977184,
        -0.837758,2.396339,1.200900,
        -0.837758,2.407941,-2.316753,
        -0.837758,2.409450,-2.988432,
        -0.837758,2.417202,-0.103363,
        -0.837758,2.417386,-0.443097,
        -0.837758,2.417950,2.666662,
        -0.837758,2.423190,1.789783,
        -0.837758,2.423692,1.392696,
        -0.837758,2.431122,3.086379,
        -0.837758,2.431406,-2.647820,
        -0.837758,2.431584,2.307631,
        -0.837758,2.434090,1.592818,
        -0.837758,2.435993,0.683844,
        -0.837758,2.440898,-2.133300,
        -0.837758,2.442222,-1.195837,
        -0.837758,2.443112,2.883761,
        -0.837758,2.449191,-0.633651,
        -0.837758,2.450821,0.084741,
        -0.837758,2.453499,-1.008861,
        -0.837758,2.460010,-1.938011,
        -0.837758,2.460199,-0.822008,
        -0.837758,2.460902,0.490079,
        -0.837758,2.461774,-1.734913,
        -0.837758,2.463986,1.034407,
        -0.837758,2.467647,0.287667,
        -0.837758,2.479849,2.126925,
        -0.837758,2.485907,-2.835471,
        -0.837758,2.485945,-1.374853,
        -0.837758,2.487747,2.505912,
        -0.837758,2.487957,-0.274279,
        -0.837758,2.489741,-2.468428,
        -0.837758,2.512538,1.239900,
        -0.837758,2.517528,1.923289,
        -0.837758,2.518454,-1.565852,
        -0.837758,2.519348,0.843503,
        -0.837758,2.522766,-3.059231,
        -0.837758,2.528686,2.737228,
        -0.837758,2.534798,-2.266671,
        -0.837758,2.536533,1.466274,
        -0.837758,2.538504,1.699812,
        -0.837758,2.539516,-0.476942,
        -0.837758,2.539532,-0.074548,
        -0.837758,2.543197,2.990252,
        -0.837758,2.554022,2.320619,
        -0.837758,2.556919,-2.655005,
        -0.837758,2.558146,0.627191,
        -0.837758,2.558223,-1.164189,
        -0.837758,2.563452,-2.044028,
        -0.837758,2.568462,0.150655,
        -0.837758,2.569130,-0.700837,
        -0.837758,2.570050,-1.804492,
        -0.837758,2.572710,-0.934665,
        -0.837758,2.575994,0.391718,
        -0.837758,2.584774,1.046703,
        -0.837758,2.601432,2.102535,
        -0.837758,2.603559,2.555114,
        -0.837758,2.604150,-2.885208,
        -0.837758,2.608630,-1.378011,
        -0.837758,2.611256,-0.279110,
        -0.837758,2.617344,-2.439480,
        -0.837758,2.627402,1.293552,
        -0.837758,2.635396,1.848850,
        -0.837758,2.635686,2.834950,
        -0.837758,2.637087,3.132730,
        -0.837758,2.642029,0.813274,
        -0.837758,2.645260,1.570203,
        -0.837758,2.646182,-1.614036,
        -0.837758,2.660113,-2.186513,
        -0.837758,2.660483,-0.034608,
        -0.837758,2.661713,-0.529511,
        -0.837758,2.672720,-1.110422,
        -0.837758,2.676626,2.335353,
        -0.837758,2.676638,0.540364,
        -0.837758,2.676780,-1.901158,
        -0.837758,2.681411,-2.667965,
        -0.837758,2.682152,0.245751,
        -0.837758,2.682859,-0.817071,
        -0.837758,2.705107,1.062142,
        -0.837758,2.717705,2.625995,
        -0.837758,2.721001,2.060097,
        -0.837758,2.722319,-2.954069,
        -0.837758,2.729840,-1.368237,
        -0.837758,2.734349,-0.289298,
        -0.837758,2.738409,2.975671,
        -0.837758,2.740366,1.374341,
        -0.837758,2.744633,-2.391311,
        -0.837758,2.747398,1.727166,
        -0.837758,2.763799,0.760372,
        -0.837758,2.768238,-1.679243,
        -0.837758,2.779245,-2.046268,
        -0.837758,2.780208,-0.626453,
        -0.837758,2.780471,0.027732,
        -0.837758,2.784696,-1.013128,
        -0.837758,2.789817,0.395431,
        -0.837758,2.801154,2.353143,
        -0.837758,2.807091,-2.687477,
        -0.837758,2.825762,1.085035,
        -0.837758,2.830209,2.734943,
        -0.837758,2.833517,-3.085508,
        -0.837758,2.838301,1.980061,
        -0.837758,2.849871,1.512012,
        -0.837758,2.850089,-1.343899,
        -0.837758,2.856905,-0.318441,
        -0.837758,2.870445,-2.291001,
        -0.837758,2.885017,0.655562,
        -0.837758,2.886428,-1.793217,
        -0.837758,2.890337,-0.825633,
        -0.837758,2.898932,0.139973,
        -0.837758,2.931927,-2.728768,
        -0.837758,2.932096,2.377790,
        -0.837758,2.937409,2.947245,
        -0.837758,2.948048,1.133127,
        -0.837758,2.950896,1.797889,
        -0.837758,2.970019,-1.282382,
        -0.837758,2.978333,-0.412090,
        -0.837758,2.996437,-2.089121,
        -0.837758,3.011274,0.434110,
        -0.837758,3.050410,3.138492,
        -0.837758,3.064492,1.647259,
        -0.837758,3.088868,-0.989501,
        -0.733038,0.053884,-2.408394,
        -0.733038,0.077645,1.453433,
        -0.733038,0.088408,-0.311160,
        -0.733038,0.142314,2.590387,
        -0.733038,0.147206,-1.385164,
        -0.733038,0.163506,0.573203,
        -0.733038,0.165485,-2.875396,
        -0.733038,0.182564,-2.125821,
        -0.733038,0.190569,1.929141,
        -0.733038,0.201102,1.249082,
        -0.733038,0.201397,-0.716646,
        -0.733038,0.210568,-0.069559,
        -0.733038,0.251672,2.848900,
        -0.733038,0.266262,-1.678068,
        -0.733038,0.269108,-1.215563,
        -0.733038,0.273777,-2.506756,
        -0.733038,0.279854,0.814581,
        -0.733038,0.281664,-2.969147,
        -0.733038,0.283625,2.405603,
        -0.733038,0.284058,0.357350,
        -0.733038,0.294107,1.593880,
        -0.733038,0.302491,-0.400410,
        -0.733038,0.308059,2.021491,
        -0.733038,0.312219,-2.072057,
        -0.733038,0.321905,-0.829121,
        -0.733038,0.329984,1.201621,
        -0.733038,0.339714,-0.013712,
        -0.733038,0.368784,2.974974,
        -0.733038,0.376428,-1.495532,
        -0.733038,0.378103,2.646436,
        -0.733038,0.385088,0.590355,
        -0.733038,0.386248,-2.690943,
        -0.733038,0.389851,-2.340031,
        -0.733038,0.391502,-1.171374,
        -0.733038,0.398249,-1.812694,
        -0.733038,0.400542,-2.995729,
        -0.733038,0.404273,0.918163,
        -0.733038,0.405617,1.773533,
        -0.733038,0.409009,-0.570243,
        -0.733038,0.410700,0.270433,
        -0.733038,0.415893,1.464890,
        -0.733038,0.420680,2.088164,
        -0.733038,0.424766,-0.258729,
        -0.733038,0.426507,2.367872,
        -0.733038,0.441292,-0.890727,
        -0.733038,0.451839,-2.078193,
        -0.733038,0.459270,1.191562,
        -0.733038,0.470622,0.014002,
        -0.733038,0.481292,-2.529098,
        -0.733038,0.482460,2.849843,
        -0.733038,0.492205,-1.397963,
        -0.733038,0.493314,3.117264,
        -0.733038,0.497870,0.727828,
        -0.733038,0.499783,-1.654379,
        -0.733038,0.502651,0.463333,
        -0.733038,0.505463,2.594557,
        -0.733038,0.506104,-2.874910,
        -0.733038,0.513579,-1.147354,
        -0.733038,0.518569,1.644237,
        -0.733038,0.519001,1.900488,
        -0.733038,0.519277,-0.673809,
        -0.733038,0.524769,-0.421955,
        -0.733038,0.525564,-2.288838,
        -0.733038,0.532541,0.978737,
        -0.733038,0.534610,2.154808,
        -0.733038,0.535020,-1.891846,
        -0.733038,0.540208,1.400497,
        -0.733038,0.542006,0.228337,
        -0.733038,0.550762,-0.180771,
        -0.733038,0.560531,2.378474,
        -0.733038,0.561629,-0.924576,
        -0.733038,0.567465,-2.679392,
        -0.733038,0.588527,1.181755,
        -0.733038,0.591643,-2.098364,
        -0.733038,0.594963,2.997575,
        -0.733038,0.599643,-3.045235,
        -0.733038,0.599962,2.773556,
        -0.733038,0.602620,0.029270,
        -0.733038,0.605711,0.598454,
        -0.733038,0.607708,-1.541026,
        -0.733038,0.609360,-2.464379,
        -0.733038,0.610738,-1.326786,
        -0.733038,0.616968,0.820489,
        -0.733038,0.626379,1.773610,
        -0.733038,0.628626,-1.751180,
        -0.733038,0.630671,0.391949,
        -0.733038,0.632622,-0.737642,
        -0.733038,0.632708,2.566722,
        -0.733038,0.632797,1.989858,
        -0.733038,0.634455,-0.534790,
        -0.733038,0.636205,1.561170,
        -0.733038,0.637681,-1.123423,
        -0.733038,0.641441,-2.842109,
        -0.733038,0.644807,-0.327139,
        -0.733038,0.652440,2.197129,
        -0.733038,0.660897,-2.278593,
        -0.733038,0.664008,1.360091,
        -0.733038,0.670850,-1.945257,
        -0.733038,0.671396,1.004859,
        -0.733038,0.673985,0.203855,
        -0.733038,0.680326,-0.133370,
        -0.733038,0.685083,-0.942363,
        -0.733038,0.688204,-2.632204,
        -0.733038,0.689777,2.385011,
        -0.733038,0.700065,3.115430,
        -0.733038,0.704080,2.911636,
        -0.733038,0.712883,1.174216,
        -0.733038,0.715701,0.699906,
        -0.733038,0.718851,-1.442215,
        -0.733038,0.721019,2.721536,
        -0.733038,0.725300,-2.975134,
        -0.733038,0.727136,-1.633288,
        -0.733038,0.727399,-2.120567,
        -0.733038,0.730751,0.522488,
        -0.733038,0.732892,-1.254330,
        -0.733038,0.734333,-2.446184,
        -0.733038,0.735125,0.037270,
        -0.733038,0.737893,1.866463,
        -0.733038,0.740650,1.681075,
        -0.733038,0.745971,-0.781286,
        -0.733038,0.749432,-0.442559,
        -0.733038,0.750102,2.050636,
        -0.733038,0.751046,0.863821,
        -0.733038,0.757018,-0.608833,
        -0.733038,0.757046,-1.816019,
        -0.733038,0.757389,1.501184,
        -0.733038,0.759800,2.546181,
        -0.733038,0.759859,0.351923,
        -0.733038,0.761141,-2.781102,
        -0.733038,0.769035,-1.078574,
        -0.733038,0.770166,-0.266604,
        -0.733038,0.772218,2.227380,
        -0.733038,0.788382,1.331051,
        -0.733038,0.790269,-2.282409,
        -0.733038,0.803410,-1.984567,
        -0.733038,0.803813,0.191520,
        -0.733038,0.806832,3.022155,
        -0.733038,0.807476,-2.597675,
        -0.733038,0.812512,-0.107335,
        -0.733038,0.814161,1.010949,
        -0.733038,0.817732,-3.084486,
        -0.733038,0.817846,2.390688,
        -0.733038,0.819323,2.845960,
        -0.733038,0.819430,-0.915017,
        -0.733038,0.829198,-1.532491,
        -0.733038,0.829531,0.665695,
        -0.733038,0.829816,1.167147,
        -0.733038,0.833613,-1.357492,
        -0.733038,0.842945,-2.907318,
        -0.733038,0.844883,2.679708,
        -0.733038,0.848189,-1.705003,
        -0.733038,0.849888,1.767901,
        -0.733038,0.851126,0.493370,
        -0.733038,0.852911,1.933394,
        -0.733038,0.857516,-2.432863,
        -0.733038,0.858261,-1.191729,
        -0.733038,0.858279,-2.138895,
        -0.733038,0.859723,-0.734928,
        -0.733038,0.859754,1.603292,
        -0.733038,0.861655,-0.394821,
        -0.733038,0.864300,0.043549,
        -0.733038,0.866940,0.851229,
        -0.733038,0.869748,2.095969,
        -0.733038,0.877352,-0.574668,
        -0.733038,0.879254,-2.733338,
        -0.733038,0.882554,1.443757,
        -0.733038,0.884060,0.334199,
        -0.733038,0.885626,-1.866061,
        -0.733038,0.889834,2.529957,
        -0.733038,0.893140,2.251698,
        -0.733038,0.897463,-0.240822,
        -0.733038,0.898758,-1.031948,
        -0.733038,0.913918,1.287618,
        -0.733038,0.916555,-2.287173,
        -0.733038,0.918305,3.107845,
        -0.733038,0.919522,2.948277,
        -0.733038,0.927138,-2.569686,
        -0.733038,0.929031,0.185383,
        -0.733038,0.933766,-3.015917,
        -0.733038,0.934873,-2.010822,
        -0.733038,0.935636,-1.445855,
        -0.733038,0.936702,-0.866461,
        -0.733038,0.937395,0.729108,
        -0.733038,0.937910,2.787575,
        -0.733038,0.939144,-0.092962,
        -0.733038,0.941718,1.125085,
        -0.733038,0.942873,-1.607680,
        -0.733038,0.943701,0.980328,
        -0.733038,0.947774,2.395256,
        -0.733038,0.952433,0.584347,
        -0.733038,0.954456,-1.287293,
        -0.733038,0.958731,-2.852552,
        -0.733038,0.961838,1.839284,
        -0.733038,0.961898,-0.469425,
        -0.733038,0.965426,1.686615,
        -0.733038,0.970108,1.990771,
        -0.733038,0.972820,-1.762780,
        -0.733038,0.973500,-0.711881,
        -0.733038,0.975253,2.641082,
        -0.733038,0.977824,0.441928,
        -0.733038,0.981485,-2.424766,
        -0.733038,0.981877,1.537100,
        -0.733038,0.982609,-1.136920,
        -0.733038,0.984265,-2.156073,
        -0.733038,0.989150,0.045493,
        -0.733038,0.993052,2.137478,
        -0.733038,0.997766,-2.693871,
        -0.733038,0.999847,-0.334738,
        -0.733038,1.001920,0.848054,
        -0.733038,1.006146,1.389734,
        -0.733038,1.012180,0.304066,
        -0.733038,1.016938,2.270674,
        -0.733038,1.017371,-0.979790,
        -0.733038,1.019495,-1.894487,
        -0.733038,1.023796,-0.583691,
        -0.733038,1.026525,2.506945,
        -0.733038,1.027004,-0.203653,
        -0.733038,1.028027,3.036573,
        -0.733038,1.031435,1.241021,
        -0.733038,1.032316,2.884221,
        -0.733038,1.035863,-3.110475,
        -0.733038,1.042085,-2.292950,
        -0.733038,1.043272,-1.522986,
        -0.733038,1.046265,-2.958847,
        -0.733038,1.049272,-1.376344,
        -0.733038,1.050039,-2.552086,
        -0.733038,1.053682,0.689863,
        -0.733038,1.055458,0.170394,
        -0.733038,1.056622,-0.831866,
        -0.733038,1.057515,1.097998,
        -0.733038,1.058706,-2.036438,
        -0.733038,1.061389,-1.672248,
        -0.733038,1.063758,2.743093,
        -0.733038,1.070157,-0.073422,
        -0.733038,1.072154,1.761506,
        -0.733038,1.074029,1.909193,
        -0.733038,1.074204,0.545276,
        -0.733038,1.074714,-2.805147,
        -0.733038,1.076262,0.966104,
        -0.733038,1.077642,-1.227073,
        -0.733038,1.084151,1.618525,
        -0.733038,1.085085,2.377673,
        -0.733038,1.086822,-0.441112,
        -0.733038,1.087142,2.058815,
        -0.733038,1.098383,-0.696430,
        -0.733038,1.099138,-1.084033,
        -0.733038,1.103208,1.477773,
        -0.733038,1.103783,0.408254,
        -0.733038,1.106275,-2.420800,
        -0.733038,1.108079,2.612943,
        -0.733038,1.109345,-2.172718,
        -0.733038,1.115957,-1.787597,
        -0.733038,1.118930,0.045983,
        -0.733038,1.119777,0.834702,
        -0.733038,1.121943,-2.669042,
        -0.733038,1.124930,2.219917,
        -0.733038,1.125217,1.336294,
        -0.733038,1.130958,2.972931,
        -0.733038,1.131400,-0.306629,
        -0.733038,1.137135,-0.936514,
        -0.733038,1.139125,-1.924528,
        -0.733038,1.139877,0.276382,
        -0.733038,1.142003,-3.058022,
        -0.733038,1.147363,-0.558448,
        -0.733038,1.150348,-0.176418,
        -0.733038,1.151198,1.197016,
        -0.733038,1.153930,-1.597104,
        -0.733038,1.155389,2.837148,
        -0.733038,1.155458,-1.460623,
        -0.733038,1.157331,-2.906362,
        -0.733038,1.158594,2.487618,
        -0.733038,1.162103,-1.322553,
        -0.733038,1.165407,3.092861,
        -0.733038,1.168025,0.717075,
        -0.733038,1.168406,-2.301519,
        -0.733038,1.174782,-2.538755,
        -0.733038,1.177998,1.061256,
        -0.733038,1.178898,1.827305,
        -0.733038,1.181331,-2.060330,
        -0.733038,1.182540,-0.798952,
        -0.733038,1.184569,0.147517,
        -0.733038,1.184734,1.968919,
        -0.733038,1.184804,0.593386,
        -0.733038,1.188278,1.690808,
        -0.733038,1.194360,2.110440,
        -0.733038,1.194776,2.710455,
        -0.733038,1.197745,-1.173705,
        -0.733038,1.199898,-2.777257,
        -0.733038,1.201164,2.355931,
        -0.733038,1.203760,1.556404,
        -0.733038,1.205845,-0.415993,
        -0.733038,1.206585,0.933648,
        -0.733038,1.206950,-0.037248,
        -0.733038,1.209341,0.471162,
        -0.733038,1.213483,-1.039417,
        -0.733038,1.218146,-0.672643,
        -0.733038,1.221608,1.421669,
        -0.733038,1.226043,-1.693533,
        -0.733038,1.232870,-1.822773,
        -0.733038,1.234375,-2.421194,
        -0.733038,1.234531,-2.190148,
        -0.733038,1.239414,0.348959,
        -0.733038,1.240183,2.587255,
        -0.733038,1.242312,2.941234,
        -0.733038,1.244075,1.286644,
        -0.733038,1.245961,-2.997461,
        -0.733038,1.246495,-2.646872,
        -0.733038,1.247736,2.242810,
        -0.733038,1.253604,0.812222,
        -0.733038,1.257816,-0.894317,
        -0.733038,1.258247,-1.953634,
        -0.733038,1.258993,-0.284099,
        -0.733038,1.262380,-0.157609,
        -0.733038,1.262791,-1.549229,
        -0.733038,1.263772,-0.533073,
        -0.733038,1.266731,-1.412461,
        -0.733038,1.268591,-3.122146,
        -0.733038,1.269614,-1.275246,
        -0.733038,1.270745,1.152533,
        -0.733038,1.273606,0.224545,
        -0.733038,1.279716,2.812101,
        -0.733038,1.283786,2.462124,
        -0.733038,1.284798,0.078209,
        -0.733038,1.285402,-2.882049,
        -0.733038,1.286936,1.886053,
        -0.733038,1.289408,0.683122,
        -0.733038,1.294736,1.756184,
        -0.733038,1.296978,-2.311511,
        -0.733038,1.298546,2.020601,
        -0.733038,1.300924,1.022478,
        -0.733038,1.303305,3.043152,
        -0.733038,1.304210,-2.084225,
        -0.733038,1.305241,-2.530249,
        -0.733038,1.307513,1.628003,
        -0.733038,1.310722,0.558267,
        -0.733038,1.314142,2.139329,
        -0.733038,1.316087,-0.777971,
        -0.733038,1.317907,-1.121023,
        -0.733038,1.321019,-2.749776,
        -0.733038,1.321609,2.687717,
        -0.733038,1.322568,1.499415,
        -0.733038,1.322972,-0.395699,
        -0.733038,1.325286,-0.994192,
        -0.733038,1.329146,-0.648513,
        -0.733038,1.329795,-0.042570,
        -0.733038,1.331651,2.344497,
        -0.733038,1.336522,0.435098,
        -0.733038,1.339410,0.900087,
        -0.733038,1.339550,1.370353,
        -0.733038,1.340441,-1.731580,
        -0.733038,1.351855,-1.855234,
        -0.733038,1.360161,-3.018404,
        -0.733038,1.361031,-1.613971,
        -0.733038,1.361196,2.914780,
        -0.733038,1.361998,-2.207254,
        -0.733038,1.362302,1.239384,
        -0.733038,1.365228,2.563808,
        -0.733038,1.365452,0.310522,
        -0.733038,1.367028,-2.420041,
        -0.733038,1.373422,-1.484281,
        -0.733038,1.374245,-1.226417,
        -0.733038,1.375020,-0.170393,
        -0.733038,1.376560,-1.355481,
        -0.733038,1.377070,-1.980984,
        -0.733038,1.377578,-2.636339,
        -0.733038,1.378096,0.774626,
        -0.733038,1.379086,-0.508192,
        -0.733038,1.382562,0.177141,
        -0.733038,1.388849,2.233847,
        -0.733038,1.389268,-0.886349,
        -0.733038,1.391324,1.109396,
        -0.733038,1.391699,-0.288725,
        -0.733038,1.392869,3.136903,
        -0.733038,1.393612,1.952016,
        -0.733038,1.395545,-2.850807,
        -0.733038,1.401926,1.822930,
        -0.733038,1.402149,2.788898,
        -0.733038,1.409552,0.647980,
        -0.733038,1.410840,0.056219,
        -0.733038,1.411944,2.444377,
        -0.733038,1.413639,1.697982,
        -0.733038,1.426479,1.573427,
        -0.733038,1.426725,2.102242,
        -0.733038,1.427868,0.984637,
        -0.733038,1.429580,-2.107232,
        -0.733038,1.431375,-2.313269,
        -0.733038,1.431907,3.015410,
        -0.733038,1.432325,-1.029054,
        -0.733038,1.433341,0.524268,
        -0.733038,1.435181,-0.739771,
        -0.733038,1.437208,-2.529085,
        -0.733038,1.439041,-0.617966,
        -0.733038,1.441473,1.448544,
        -0.733038,1.444527,2.664744,
        -0.733038,1.450513,-0.400625,
        -0.733038,1.451907,-0.066823,
        -0.733038,1.452754,-2.738631,
        -0.733038,1.454061,-2.950416,
        -0.733038,1.456736,1.323154,
        -0.733038,1.458370,0.400433,
        -0.733038,1.459314,-1.765003,
        -0.733038,1.459902,-1.145835,
        -0.733038,1.464952,2.329245,
        -0.733038,1.466050,0.862104,
        -0.733038,1.471330,-1.886160,
        -0.733038,1.475119,-1.648255,
        -0.733038,1.480878,1.194576,
        -0.733038,1.480942,0.273235,
        -0.733038,1.481086,2.890269,
        -0.733038,1.484118,-1.529156,
        -0.733038,1.484741,-1.279219,
        -0.733038,1.487681,-1.406069,
        -0.733038,1.487886,-3.066037,
        -0.733038,1.490555,2.543284,
        -0.733038,1.492424,-0.190932,
        -0.733038,1.495002,-0.839810,
        -0.733038,1.495297,2.007188,
        -0.733038,1.496414,-2.009267,
        -0.733038,1.498261,-2.425270,
        -0.733038,1.498907,-2.207327,
        -0.733038,1.500204,0.737638,
        -0.733038,1.502469,0.151156,
        -0.733038,1.507448,-0.514782,
        -0.733038,1.509287,2.204647,
        -0.733038,1.510518,1.884920,
        -0.733038,1.511948,-2.632360,
        -0.733038,1.515235,1.068565,
        -0.733038,1.522237,3.102018,
        -0.733038,1.522373,1.764991,
        -0.733038,1.522416,-0.954956,
        -0.733038,1.523102,2.765920,
        -0.733038,1.524811,-2.844718,
        -0.733038,1.527261,-0.304916,
        -0.733038,1.529788,0.613208,
        -0.733038,1.533648,1.644808,
        -0.733038,1.535018,0.028869,
        -0.733038,1.541082,2.424608,
        -0.733038,1.546010,1.523462,
        -0.733038,1.551135,-0.642994,
        -0.733038,1.553895,0.489607,
        -0.733038,1.555041,0.947813,
        -0.733038,1.559671,-1.074934,
        -0.733038,1.560345,2.985899,
        -0.733038,1.560487,-2.323400,
        -0.733038,1.561127,1.401111,
        -0.733038,1.567193,2.643172,
        -0.733038,1.568892,2.098362,
        -0.733038,1.571702,-2.098435,
        -0.733038,1.572701,-2.528730,
        -0.733038,1.573572,1.278593,
        -0.733038,1.574849,-0.091505,
        -0.733038,1.576334,0.365236,
        -0.733038,1.577029,-2.962888,
        -0.733038,1.577688,-1.814941,
        -0.733038,1.579542,-0.755917,
        -0.733038,1.579624,-1.199759,
        -0.733038,1.581515,-0.419414,
        -0.733038,1.584581,-2.738702,
        -0.733038,1.586948,-1.696296,
        -0.733038,1.588901,2.301496,
        -0.733038,1.589936,0.825020,
        -0.733038,1.590018,-1.941652,
        -0.733038,1.596968,-1.328358,
        -0.733038,1.597178,-1.575607,
        -0.733038,1.600470,-1.453196,
        -0.733038,1.601408,0.242409,
        -0.733038,1.602259,1.152656,
        -0.733038,1.603064,2.865774,
        -0.733038,1.612085,1.968802,
        -0.733038,1.612720,-3.080498,
        -0.733038,1.616061,2.521531,
        -0.733038,1.616900,-0.207457,
        -0.733038,1.620988,-0.875920,
        -0.733038,1.622114,0.701509,
        -0.733038,1.622749,0.121426,
        -0.733038,1.622993,-2.221629,
        -0.733038,1.629346,-0.542669,
        -0.733038,1.630007,1.844552,
        -0.733038,1.634747,-2.426198,
        -0.733038,1.642346,1.722617,
        -0.733038,1.643925,1.034629,
        -0.733038,1.644848,2.743774,
        -0.733038,1.645258,2.188109,
        -0.733038,1.645417,-2.634725,
        -0.733038,1.647395,3.085437,
        -0.733038,1.648538,-2.855044,
        -0.733038,1.649932,0.576808,
        -0.733038,1.652224,-0.996105,
        -0.733038,1.652871,1.600681,
        -0.733038,1.657997,-0.325127,
        -0.733038,1.661936,0.001037,
        -0.733038,1.665187,1.477402,
        -0.733038,1.665496,2.397872,
        -0.733038,1.673267,-0.665453,
        -0.733038,1.673557,0.452526,
        -0.733038,1.679277,-1.122421,
        -0.733038,1.679696,0.912482,
        -0.733038,1.682420,1.353922,
        -0.733038,1.685031,-2.009038,
        -0.733038,1.685196,2.967046,
        -0.733038,1.690107,2.620910,
        -0.733038,1.690453,-2.123076,
        -0.733038,1.691693,-1.871572,
        -0.733038,1.691817,2.061688,
        -0.733038,1.692644,1.237670,
        -0.733038,1.696386,0.328123,
        -0.733038,1.697162,-1.249026,
        -0.733038,1.697488,-2.322785,
        -0.733038,1.698726,-2.973576,
        -0.733038,1.699172,-1.749717,
        -0.733038,1.704478,-0.106295,
        -0.733038,1.707244,-0.444770,
        -0.733038,1.707500,-2.531497,
        -0.733038,1.707714,-0.787321,
        -0.733038,1.708430,-1.627552,
        -0.733038,1.710583,-1.376890,
        -0.733038,1.712415,-2.747921,
        -0.733038,1.713125,-1.503073,
        -0.733038,1.713452,0.789419,
        -0.733038,1.720612,2.279888,
        -0.733038,1.724852,0.202831,
        -0.733038,1.724868,2.846246,
        -0.733038,1.728940,1.931535,
        -0.733038,1.732838,1.127584,
        -0.733038,1.735719,-3.092485,
        -0.733038,1.739676,2.496017,
        -0.733038,1.739971,-0.228876,
        -0.733038,1.743328,0.663347,
        -0.733038,1.744927,-0.911892,
        -0.733038,1.745952,0.084733,
        -0.733038,1.748077,1.804354,
        -0.733038,1.752959,-0.567548,
        -0.733038,1.760169,1.678132,
        -0.733038,1.763887,-2.218450,
        -0.733038,1.766737,2.722752,
        -0.733038,1.767999,1.004737,
        -0.733038,1.769149,1.550298,
        -0.733038,1.770119,-2.427175,
        -0.733038,1.770460,2.154208,
        -0.733038,1.770514,3.070982,
        -0.733038,1.770735,0.537245,
        -0.733038,1.772673,-2.864760,
        -0.733038,1.773527,-1.039582,
        -0.733038,1.776419,-2.642421,
        -0.733038,1.780769,1.416330,
        -0.733038,1.786634,-0.346247,
        -0.733038,1.788829,-1.947553,
        -0.733038,1.793737,0.410270,
        -0.733038,1.794553,2.374618,
        -0.733038,1.794699,-0.694061,
        -0.733038,1.797469,-1.169105,
        -0.733038,1.803705,0.881076,
        -0.733038,1.804124,1.264249,
        -0.733038,1.807825,-2.090455,
        -0.733038,1.808822,-1.817375,
        -0.733038,1.809020,2.949741,
        -0.733038,1.812466,2.597595,
        -0.733038,1.813159,2.025395,
        -0.733038,1.813816,-1.298435,
        -0.733038,1.815016,0.278951,
        -0.733038,1.816510,-0.008631,
        -0.733038,1.817772,-1.689625,
        -0.733038,1.821360,-2.984352,
        -0.733038,1.822312,-0.127211,
        -0.733038,1.823386,-1.429807,
        -0.733038,1.823834,-1.560483,
        -0.733038,1.832060,-0.821627,
        -0.733038,1.832690,-0.468440,
        -0.733038,1.833945,-2.320278,
        -0.733038,1.834787,0.751968,
        -0.733038,1.840579,-2.536445,
        -0.733038,1.841033,-2.756932,
        -0.733038,1.846350,1.893983,
        -0.733038,1.846644,1.113766,
        -0.733038,1.846833,2.826543,
        -0.733038,1.847325,2.249051,
        -0.733038,1.854119,0.134161,
        -0.733038,1.858863,-3.106022,
        -0.733038,1.865046,1.763384,
        -0.733038,1.865619,-0.952036,
        -0.733038,1.865912,0.624453,
        -0.733038,1.866694,2.473686,
        -0.733038,1.867524,-0.240083,
        -0.733038,1.875561,1.632831,
        -0.733038,1.876049,-0.594378,
        -0.733038,1.880509,1.500204,
        -0.733038,1.883504,1.360457,
        -0.733038,1.886971,-2.196665,
        -0.733038,1.888979,0.982922,
        -0.733038,1.889815,2.701570,
        -0.733038,1.892144,-2.005340,
        -0.733038,1.892348,0.496522,
        -0.733038,1.893281,-1.083976,
        -0.733038,1.893828,3.053453,
        -0.733038,1.895117,2.121107,
        -0.733038,1.898741,-2.875735,
        -0.733038,1.904429,-2.428029,
        -0.733038,1.908408,-2.648903,
        -0.733038,1.911828,-0.364489,
        -0.733038,1.912224,0.366565,
        -0.733038,1.914854,1.214468,
        -0.733038,1.915467,-1.883143,
        -0.733038,1.915763,-0.724679,
        -0.733038,1.915790,-1.216366,
        -0.733038,1.922256,2.347806,
        -0.733038,1.922616,0.844633,
        -0.733038,1.925185,0.231170,
        -0.733038,1.927474,-1.754616,
        -0.733038,1.928342,-1.351174,
        -0.733038,1.933328,-1.623328,
        -0.733038,1.933878,0.034150,
        -0.733038,1.934585,-1.487979,
        -0.733038,1.934802,2.927634,
        -0.733038,1.934862,1.989776,
        -0.733038,1.936327,-0.099350,
        -0.733038,1.938088,2.578507,
        -0.733038,1.945677,-2.999805,
        -0.733038,1.954572,-0.859208,
        -0.733038,1.957868,-0.490784,
        -0.733038,1.958989,0.712772,
        -0.733038,1.960212,-2.307561,
        -0.733038,1.964774,1.855507,
        -0.733038,1.964971,-2.100533,
        -0.733038,1.969112,1.088273,
        -0.733038,1.971438,-2.765415,
        -0.733038,1.972312,2.805134,
        -0.733038,1.974950,2.220596,
        -0.733038,1.975135,-2.539127,
        -0.733038,1.982107,1.720829,
        -0.733038,1.983195,-3.127984,
        -0.733038,1.983318,-0.247182,
        -0.733038,1.986720,-0.993220,
        -0.733038,1.990381,1.586005,
        -0.733038,1.990470,0.580283,
        -0.733038,1.990523,1.450326,
        -0.733038,1.993909,1.312897,
        -0.733038,1.994694,2.452813,
        -0.733038,1.999226,-0.620717,
        -0.733038,2.005145,0.941012,
        -0.733038,2.007301,0.142337,
        -0.733038,2.014149,0.444138,
        -0.733038,2.014338,-1.127392,
        -0.733038,2.018131,3.021778,
        -0.733038,2.019676,-1.980944,
        -0.733038,2.019912,2.686612,
        -0.733038,2.020545,2.088672,
        -0.733038,2.026074,0.300945,
        -0.733038,2.026196,-2.889057,
        -0.733038,2.031959,-1.268178,
        -0.733038,2.032917,-2.199917,
        -0.733038,2.033572,-1.842884,
        -0.733038,2.034712,-2.419505,
        -0.733038,2.035818,-0.379479,
        -0.733038,2.036004,-0.757062,
        -0.733038,2.040501,-1.411488,
        -0.733038,2.042874,-1.701367,
        -0.733038,2.043060,-2.653746,
        -0.733038,2.043729,-1.557938,
        -0.733038,2.048538,0.806623,
        -0.733038,2.049006,1.199930,
        -0.733038,2.049502,-0.137989,
        -0.733038,2.052181,2.326959,
        -0.733038,2.052794,-0.008526,
        -0.733038,2.057608,1.951726,
        -0.733038,2.065568,2.567730,
        -0.733038,2.070371,2.879808,
        -0.733038,2.071256,-3.018961,
        -0.733038,2.079746,-0.897699,
        -0.733038,2.082784,1.042403,
        -0.733038,2.083752,-0.511086,
        -0.733038,2.084401,1.810502,
        -0.733038,2.085731,0.671081,
        -0.733038,2.097496,1.373172,
        -0.733038,2.099185,1.667006,
        -0.733038,2.100482,-2.301084,
        -0.733038,2.103118,2.194745,
        -0.733038,2.103223,1.521685,
        -0.733038,2.103695,-2.777707,
        -0.733038,2.107523,3.129581,
        -0.733038,2.108277,-2.095907,
        -0.733038,2.108891,-2.532166,
        -0.733038,2.111003,-0.268632,
        -0.733038,2.112276,-1.027246,
        -0.733038,2.114380,0.531457,
        -0.733038,2.119166,0.100552,
        -0.733038,2.119564,2.754818,
        -0.733038,2.122474,-0.648314,
        -0.733038,2.125183,2.445598,
        -0.733038,2.129501,0.238290,
        -0.733038,2.131178,0.386663,
        -0.733038,2.131223,-1.174311,
        -0.733038,2.134269,-1.948042,
        -0.733038,2.134684,0.907065,
        -0.733038,2.137615,2.988806,
        -0.733038,2.145635,-1.327708,
        -0.733038,2.146467,2.055081,
        -0.733038,2.146582,-1.793788,
        -0.733038,2.149210,-1.482321,
        -0.733038,2.152901,-1.636788,
        -0.733038,2.156337,1.151558,
        -0.733038,2.156474,-2.912065,
        -0.733038,2.157194,-0.792563,
        -0.733038,2.166724,-0.401638,
        -0.733038,2.166980,-0.134879,
        -0.733038,2.173644,-2.403429,
        -0.733038,2.175711,-2.655523,
        -0.733038,2.177132,2.600419,
        -0.733038,2.178879,0.767722,
        -0.733038,2.180752,1.908485,
        -0.733038,2.181338,2.313721,
        -0.733038,2.185580,1.287673,
        -0.733038,2.195400,-2.217546,
        -0.733038,2.200418,-3.056442,
        -0.733038,2.201829,-0.001735,
        -0.733038,2.203804,1.756176,
        -0.733038,2.211602,1.443930,
        -0.733038,2.212686,-0.544809,
        -0.733038,2.213937,0.621892,
        -0.733038,2.214075,1.600691,
        -0.733038,2.216256,1.020029,
        -0.733038,2.219060,2.886631,
        -0.733038,2.223375,-0.906185,
        -0.733038,2.226002,-1.060248,
        -0.733038,2.226413,-2.064034,
        -0.733038,2.230554,2.171334,
        -0.733038,2.232413,2.735777,
        -0.733038,2.233709,3.070830,
        -0.733038,2.234383,-2.792065,
        -0.733038,2.237183,0.468799,
        -0.733038,2.237695,-0.271388,
        -0.733038,2.239419,0.151693,
        -0.733038,2.243347,-2.521049,
        -0.733038,2.244035,-1.229456,
        -0.733038,2.244355,0.309899,
        -0.733038,2.245175,2.460236,
        -0.733038,2.245628,-1.899216,
        -0.733038,2.245973,-0.702234,
        -0.733038,2.255601,-1.728131,
        -0.733038,2.256702,-1.401367,
        -0.733038,2.261466,-1.557500,
        -0.733038,2.267390,0.878499,
        -0.733038,2.272099,2.018594,
        -0.733038,2.278618,1.170386,
        -0.733038,2.283304,-2.345584,
        -0.733038,2.285175,-2.943090,
        -0.733038,2.293000,-0.124451,
        -0.733038,2.294058,-0.421070,
        -0.733038,2.303283,1.855938,
        -0.733038,2.304968,2.608978,
        -0.733038,2.305347,2.311346,
        -0.733038,2.307188,-2.656134,
        -0.733038,2.307301,1.336200,
        -0.733038,2.309655,0.726020,
        -0.733038,2.317634,-2.185225,
        -0.733038,2.321510,1.685664,
        -0.733038,2.324376,-3.111006,
        -0.733038,2.324731,1.511186,
        -0.733038,2.329587,0.035498,
        -0.733038,2.330158,2.977264,
        -0.733038,2.335096,-0.588468,
        -0.733038,2.336046,-1.115799,
        -0.733038,2.338060,2.803613,
        -0.733038,2.338083,-0.943400,
        -0.733038,2.340210,0.562145,
        -0.733038,2.341931,1.023926,
        -0.733038,2.343265,-0.779061,
        -0.733038,2.343434,-2.013692,
        -0.733038,2.353029,-1.310529,
        -0.733038,2.354868,0.211543,
        -0.733038,2.355120,0.388593,
        -0.733038,2.355165,-1.833173,
        -0.733038,2.355462,2.151441,
        -0.733038,2.362948,-2.487477,
        -0.733038,2.363101,-2.810456,
        -0.733038,2.363394,-0.272918,
        -0.733038,2.364509,-1.644533,
        -0.733038,2.370465,2.466824,
        -0.733038,2.386110,-1.481351,
        -0.733038,2.394960,0.862133,
        -0.733038,2.395716,1.977184,
        -0.733038,2.396339,1.200900,
        -0.733038,2.407941,-2.316753,
        -0.733038,2.409450,-2.988432,
        -0.733038,2.417202,-0.103363,
        -0.733038,2.417386,-0.443097,
        -0.733038,2.417950,2.666662,
        -0.733038,2.423190,1.789783,
        -0.733038,2.423692,1.392696,
        -0.733038,2.431122,3.086379,
        -0.733038,2.431406,-2.647820,
        -0.733038,2.431584,2.307631,
        -0.733038,2.434090,1.592818,
        -0.733038,2.435993,0.683844,
        -0.733038,2.440898,-2.133300,
        -0.733038,2.442222,-1.195837,
        -0.733038,2.443112,2.883761,
        -0.733038,2.449191,-0.633651,
        -0.733038,2.450821,0.084741,
        -0.733038,2.453499,-1.008861,
        -0.733038,2.460010,-1.938011,
        -0.733038,2.460199,-0.822008,
        -0.733038,2.460902,0.490079,
        -0.733038,2.461774,-1.734913,
        -0.733038,2.463986,1.034407,
        -0.733038,2.467647,0.287667,
        -0.733038,2.479849,2.126925,
        -0.733038,2.485907,-2.835471,
        -0.733038,2.485945,-1.374853,
        -0.733038,2.487747,2.505912,
        -0.733038,2.487957,-0.274279,
        -0.733038,2.489741,-2.468428,
        -0.733038,2.512538,1.239900,
        -0.733038,2.517528,1.923289,
        -0.733038,2.518454,-1.565852,
        -0.733038,2.519348,0.843503,
        -0.733038,2.522766,-3.059231,
        -0.733038,2.528686,2.737228,
        -0.733038,2.534798,-2.266671,
        -0.733038,2.536533,1.466274,
        -0.733038,2.538504,1.699812,
        -0.733038,2.539516,-0.476942,
        -0.733038,2.539532,-0.074548,
        -0.733038,2.543197,2.990252,
        -0.733038,2.554022,2.320619,
        -0.733038,2.556919,-2.655005,
        -0.733038,2.558146,0.627191,
        -0.733038,2.558223,-1.164189,
        -0.733038,2.563452,-2.044028,
        -0.733038,2.568462,0.150655,
        -0.733038,2.569130,-0.700837,
        -0.733038,2.570050,-1.804492,
        -0.733038,2.572710,-0.934665,
        -0.733038,2.575994,0.391718,
        -0.733038,2.584774,1.046703,
        -0.733038,2.601432,2.102535,
        -0.733038,2.603559,2.555114,
        -0.733038,2.604150,-2.885208,
        -0.733038,2.608630,-1.378011,
        -0.733038,2.611256,-0.279110,
        -0.733038,2.617344,-2.439480,
        -0.733038,2.627402,1.293552,
        -0.733038,2.635396,1.848850,
        -0.733038,2.635686,2.834950,
        -0.733038,2.637087,3.132730,
        -0.733038,2.642029,0.813274,
        -0.733038,2.645260,1.570203,
        -0.733038,2.646182,-1.614036,
        -0.733038,2.660113,-2.186513,
        -0.733038,2.660483,-0.034608,
        -0.733038,2.661713,-0.529511,
        -0.733038,2.672720,-1.110422,
        -0.733038,2.676626,2.335353,
        -0.733038,2.676638,0.540364,
        -0.733038,2.676780,-1.901158,
        -0.733038,2.681411,-2.667965,
        -0.733038,2.682152,0.245751,
        -0.733038,2.682859,-0.817071,
        -0.733038,2.705107,1.062142,
        -0.733038,2.717705,2.625995,
        -0.733038,2.721001,2.060097,
        -0.733038,2.722319,-2.954069,
        -0.733038,2.729840,-1.368237,
        -0.733038,2.734349,-0.289298,
        -0.733038,2.738409,2.975671,
        -0.733038,2.740366,1.374341,
        -0.733038,2.744633,-2.391311,
        -0.733038,2.747398,1.727166,
        -0.733038,2.763799,0.760372,
        -0.733038,2.768238,-1.679243,
        -0.733038,2.779245,-2.046268,
        -0.733038,2.780208,-0.626453,
        -0.733038,2.780471,0.027732,
        -0.733038,2.784696,-1.013128,
        -0.733038,2.789817,0.395431,
        -0.733038,2.801154,2.353143,
        -0.733038,2.807091,-2.687477,
        -0.733038,2.825762,1.085035,
        -0.733038,2.830209,2.734943,
        -0.733038,2.833517,-3.085508,
        -0.733038,2.838301,1.980061,
        -0.733038,2.849871,1.512012,
        -0.733038,2.850089,-1.343899,
        -0.733038,2.856905,-0.318441,
        -0.733038,2.870445,-2.291001,
        -0.733038,2.885017,0.655562,
        -0.733038,2.886428,-1.793217,
        -0.733038,2.890337,-0.825633,
        -0.733038,2.898932,0.139973,
        -0.733038,2.931927,-2.728768,
        -0.733038,2.932096,2.377790,
        -0.733038,2.937409,2.947245,
        -0.733038,2.948048,1.133127,
        -0.733038,2.950896,1.797889,
        -0.733038,2.970019,-1.282382,
        -0.733038,2.978333,-0.412090,
        -0.733038,2.996437,-2.089121,
        -0.733038,3.011274,0.434110,
        -0.733038,3.050410,3.138492,
        -0.733038,3.064492,1.647259,
        -0.733038,3.088868,-0.989501,
        -0.628319,0.053884,-2.408394,
        -0.628319,0.077645,1.453433,
        -0.628319,0.088408,-0.311160,
        -0.628319,0.142314,2.590387,
        -0.628319,0.147206,-1.385164,
        -0.628319,0.163506,0.573203,
        -0.628319,0.165485,-2.875396,
        -0.628319,0.182564,-2.125821,
        -0.628319,0.190569,1.929141,
        -0.628319,0.201102,1.249082,
        -0.628319,0.201397,-0.716646,
        -0.628319,0.210568,-0.069559,
        -0.628319,0.251672,2.848900,
        -0.628319,0.266262,-1.678068,
        -0.628319,0.269108,-1.215563,
        -0.628319,0.273777,-2.506756,
        -0.628319,0.279854,0.814581,
        -0.628319,0.281664,-2.969147,
        -0.628319,0.283625,2.405603,
        -0.628319,0.284058,0.357350,
        -0.628319,0.294107,1.593880,
        -0.628319,0.302491,-0.400410,
        -0.628319,0.308059,2.021491,
        -0.628319,0.312219,-2.072057,
        -0.628319,0.321905,-0.829121,
        -0.628319,0.329984,1.201621,
        -0.628319,0.339714,-0.013712,
        -0.628319,0.368784,2.974974,
        -0.628319,0.376428,-1.495532,
        -0.628319,0.378103,2.646436,
        -0.628319,0.385088,0.590355,
        -0.628319,0.386248,-2.690943,
        -0.628319,0.389851,-2.340031,
        -0.628319,0.391502,-1.171374,
        -0.628319,0.398249,-1.812694,
        -0.628319,0.400542,-2.995729,
        -0.628319,0.404273,0.918163,
        -0.628319,0.405617,1.773533,
        -0.628319,0.409009,-0.570243,
        -0.628319,0.410700,0.270433,
        -0.628319,0.415893,1.464890,
        -0.628319,0.420680,2.088164,
        -0.628319,0.424766,-0.258729,
        -0.628319,0.426507,2.367872,
        -0.628319,0.441292,-0.890727,
        -0.628319,0.451839,-2.078193,
        -0.628319,0.459270,1.191562,
        -0.628319,0.470622,0.014002,
        -0.628319,0.481292,-2.529098,
        -0.628319,0.482460,2.849843,
        -0.628319,0.492205,-1.397963,
        -0.628319,0.493314,3.117264,
        -0.628319,0.497870,0.727828,
        -0.628319,0.499783,-1.654379,
        -0.628319,0.502651,0.463333,
        -0.628319,0.505463,2.594557,
        -0.628319,0.506104,-2.874910,
        -0.628319,0.513579,-1.147354,
        -0.628319,0.518569,1.644237,
        -0.628319,0.519001,1.900488,
        -0.628319,0.519277,-0.673809,
        -0.628319,0.524769,-0.421955,
        -0.628319,0.525564,-2.288838,
        -0.628319,0.532541,0.978737,
        -0.628319,0.534610,2.154808,
        -0.628319,0.535020,-1.891846,
        -0.628319,0.540208,1.400497,
        -0.628319,0.542006,0.228337,
        -0.628319,0.550762,-0.180771,
        -0.628319,0.560531,2.378474,
        -0.628319,0.561629,-0.924576,
        -0.628319,0.567465,-2.679392,
        -0.628319,0.588527,1.181755,
        -0.628319,0.591643,-2.098364,
        -0.628319,0.594963,2.997575,
        -0.628319,0.599643,-3.045235,
        -0.628319,0.599962,2.773556,
        -0.628319,0.602620,0.029270,
        -0.628319,0.605711,0.598454,
        -0.628319,0.607708,-1.541026,
        -0.628319,0.609360,-2.464379,
        -0.628319,0.610738,-1.326786,
        -0.628319,0.616968,0.820489,
        -0.628319,0.626379,1.773610,
        -0.628319,0.628626,-1.751180,
        -0.628319,0.630671,0.391949,
        -0.628319,0.632622,-0.737642,
        -0.628319,0.632708,2.566722,
        -0.628319,0.632797,1.989858,
        -0.628319,0.634455,-0.534790,
        -0.628319,0.636205,1.561170,
        -0.628319,0.637681,-1.123423,
        -0.628319,0.641441,-2.842109,
        -0.628319,0.644807,-0.327139,
        -0.628319,0.652440,2.197129,
        -0.628319,0.660897,-2.278593,
        -0.628319,0.664008,1.360091,
        -0.628319,0.670850,-1.945257,
        -0.628319,0.671396,1.004859,
        -0.628319,0.673985,0.203855,
        -0.628319,0.680326,-0.133370,
        -0.628319,0.685083,-0.942363,
        -0.628319,0.688204,-2.632204,
        -0.628319,0.689777,2.385011,
        -0.628319,0.700065,3.115430,
        -0.628319,0.704080,2.911636,
        -0.628319,0.712883,1.174216,
        -0.628319,0.715701,0.699906,
        -0.628319,0.718851,-1.442215,
        -0.628319,0.721019,2.721536,
        -0.628319,0.725300,-2.975134,
        -0.628319,0.727136,-1.633288,
        -0.628319,0.727399,-2.120567,
        -0.628319,0.730751,0.522488,
        -0.628319,0.732892,-1.254330,
        -0.628319,0.734333,-2.446184,
        -0.628319,0.735125,0.037270,
        -0.628319,0.737893,1.866463,
        -0.628319,0.740650,1.681075,
        -0.628319,0.745971,-0.781286,
        -0.628319,0.749432,-0.442559,
        -0.628319,0.750102,2.050636,
        -0.628319,0.751046,0.863821,
        -0.628319,0.757018,-0.608833,
        -0.628319,0.757046,-1.816019,
        -0.628319,0.757389,1.501184,
        -0.628319,0.759800,2.546181,
        -0.628319,0.759859,0.351923,
        -0.628319,0.761141,-2.781102,
        -0.628319,0.769035,-1.078574,
        -0.628319,0.770166,-0.266604,
        -0.628319,0.772218,2.227380,
        -0.628319,0.788382,1.331051,
        -0.628319,0.790269,-2.282409,
        -0.628319,0.803410,-1.984567,
        -0.628319,0.803813,0.191520,
        -0.628319,0.806832,3.022155,
        -0.628319,0.807476,-2.597675,
        -0.628319,0.812512,-0.107335,
        -0.628319,0.814161,1.010949,
        -0.628319,0.817732,-3.084486,
        -0.628319,0.817846,2.390688,
        -0.628319,0.819323,2.845960,
        -0.628319,0.819430,-0.915017,
        -0.628319,0.829198,-1.532491,
        -0.628319,0.829531,0.665695,
        -0.628319,0.829816,1.167147,
        -0.628319,0.833613,-1.357492,
        -0.628319,0.842945,-2.907318,
        -0.628319,0.844883,2.679708,
        -0.628319,0.848189,-1.705003,
        -0.628319,0.849888,1.767901,
        -0.628319,0.851126,0.493370,
        -0.628319,0.852911,1.933394,
        -0.628319,0.857516,-2.432863,
        -0.628319,0.858261,-1.191729,
        -0.628319,0.858279,-2.138895,
        -0.628319,0.859723,-0.734928,
        -0.628319,0.859754,1.603292,
        -0.628319,0.861655,-0.394821,
        -0.628319,0.864300,0.043549,
        -0.628319,0.866940,0.851229,
        -0.628319,0.869748,2.095969,
        -0.628319,0.877352,-0.574668,
        -0.628319,0.879254,-2.733338,
        -0.628319,0.882554,1.443757,
        -0.628319,0.884060,0.334199,
        -0.628319,0.885626,-1.866061,
        -0.628319,0.889834,2.529957,
        -0.628319,0.893140,2.251698,
        -0.628319,0.897463,-0.240822,
        -0.628319,0.898758,-1.031948,
        -0.628319,0.913918,1.287618,
        -0.628319,0.916555,-2.287173,
        -0.628319,0.918305,3.107845,
        -0.628319,0.919522,2.948277,
        -0.628319,0.927138,-2.569686,
        -0.628319,0.929031,0.185383,
        -0.628319,0.933766,-3.015917,
        -0.628319,0.934873,-2.010822,
        -0.628319,0.935636,-1.445855,
        -0.628319,0.936702,-0.866461,
        -0.628319,0.937395,0.729108,
        -0.628319,0.937910,2.787575,
        -0.628319,0.939144,-0.092962,
        -0.628319,0.941718,1.125085,
        -0.628319,0.942873,-1.607680,
        -0.628319,0.943701,0.980328,
        -0.628319,0.947774,2.395256,
        -0.628319,0.952433,0.584347,
        -0.628319,0.954456,-1.287293,
        -0.628319,0.958731,-2.852552,
        -0.628319,0.961838,1.839284,
        -0.628319,0.961898,-0.469425,
        -0.628319,0.965426,1.686615,
        -0.628319,0.970108,1.990771,
        -0.628319,0.972820,-1.762780,
        -0.628319,0.973500,-0.711881,
        -0.628319,0.975253,2.641082,
        -0.628319,0.977824,0.441928,
        -0.628319,0.981485,-2.424766,
        -0.628319,0.981877,1.537100,
        -0.628319,0.982609,-1.136920,
        -0.628319,0.984265,-2.156073,
        -0.628319,0.989150,0.045493,
        -0.628319,0.993052,2.137478,
        -0.628319,0.997766,-2.693871,
        -0.628319,0.999847,-0.334738,
        -0.628319,1.001920,0.848054,
        -0.628319,1.006146,1.389734,
        -0.628319,1.012180,0.304066,
        -0.628319,1.016938,2.270674,
        -0.628319,1.017371,-0.979790,
        -0.628319,1.019495,-1.894487,
        -0.628319,1.023796,-0.583691,
        -0.628319,1.026525,2.506945,
        -0.628319,1.027004,-0.203653,
        -0.628319,1.028027,3.036573,
        -0.628319,1.031435,1.241021,
        -0.628319,1.032316,2.884221,
        -0.628319,1.035863,-3.110475,
        -0.628319,1.042085,-2.292950,
        -0.628319,1.043272,-1.522986,
        -0.628319,1.046265,-2.958847,
        -0.628319,1.049272,-1.376344,
        -0.628319,1.050039,-2.552086,
        -0.628319,1.053682,0.689863,
        -0.628319,1.055458,0.170394,
        -0.628319,1.056622,-0.831866,
        -0.628319,1.057515,1.097998,
        -0.628319,1.058706,-2.036438,
        -0.628319,1.061389,-1.672248,
        -0.628319,1.063758,2.743093,
        -0.628319,1.070157,-0.073422,
        -0.628319,1.072154,1.761506,
        -0.628319,1.074029,1.909193,
        -0.628319,1.074204,0.545276,
        -0.628319,1.074714,-2.805147,
        -0.628319,1.076262,0.966104,
        -0.628319,1.077642,-1.227073,
        -0.628319,1.084151,1.618525,
        -0.628319,1.085085,2.377673,
        -0.628319,1.086822,-0.441112,
        -0.628319,1.087142,2.058815,
        -0.628319,1.098383,-0.696430,
        -0.628319,1.099138,-1.084033,
        -0.628319,1.103208,1.477773,
        -0.628319,1.103783,0.408254,
        -0.628319,1.106275,-2.420800,
        -0.628319,1.108079,2.612943,
        -0.628319,1.109345,-2.172718,
        -0.628319,1.115957,-1.787597,
        -0.628319,1.118930,0.045983,
        -0.628319,1.119777,0.834702,
        -0.628319,1.121943,-2.669042,
        -0.628319,1.124930,2.219917,
        -0.628319,1.125217,1.336294,
        -0.628319,1.130958,2.972931,
        -0.628319,1.131400,-0.306629,
        -0.628319,1.137135,-0.936514,
        -0.628319,1.139125,-1.924528,
        -0.628319,1.139877,0.276382,
        -0.628319,1.142003,-3.058022,
        -0.628319,1.147363,-0.558448,
        -0.628319,1.150348,-0.176418,
        -0.628319,1.151198,1.197016,
        -0.628319,1.153930,-1.597104,
        -0.628319,1.155389,2.837148,
        -0.628319,1.155458,-1.460623,
        -0.628319,1.157331,-2.906362,
        -0.628319,1.158594,2.487618,
        -0.628319,1.162103,-1.322553,
        -0.628319,1.165407,3.092861,
        -0.628319,1.168025,0.717075,
        -0.628319,1.168406,-2.301519,
        -0.628319,1.174782,-2.538755,
        -0.628319,1.177998,1.061256,
        -0.628319,1.178898,1.827305,
        -0.628319,1.181331,-2.060330,
        -0.628319,1.182540,-0.798952,
        -0.628319,1.184569,0.147517,
        -0.628319,1.184734,1.968919,
        -0.628319,1.184804,0.593386,
        -0.628319,1.188278,1.690808,
        -0.628319,1.194360,2.110440,
        -0.628319,1.194776,2.710455,
        -0.628319,1.197745,-1.173705,
        -0.628319,1.199898,-2.777257,
        -0.628319,1.201164,2.355931,
        -0.628319,1.203760,1.556404,
        -0.628319,1.205845,-0.415993,
        -0.628319,1.206585,0.933648,
        -0.628319,1.206950,-0.037248,
        -0.628319,1.209341,0.471162,
        -0.628319,1.213483,-1.039417,
        -0.628319,1.218146,-0.672643,
        -0.628319,1.221608,1.421669,
        -0.628319,1.226043,-1.693533,
        -0.628319,1.232870,-1.822773,
        -0.628319,1.234375,-2.421194,
        -0.628319,1.234531,-2.190148,
        -0.628319,1.239414,0.348959,
        -0.628319,1.240183,2.587255,
        -0.628319,1.242312,2.941234,
        -0.628319,1.244075,1.286644,
        -0.628319,1.245961,-2.997461,
        -0.628319,1.246495,-2.646872,
        -0.628319,1.247736,2.242810,
        -0.628319,1.253604,0.812222,
        -0.628319,1.257816,-0.894317,
        -0.628319,1.258247,-1.953634,
        -0.628319,1.258993,-0.284099,
        -0.628319,1.262380,-0.157609,
        -0.628319,1.262791,-1.549229,
        -0.628319,1.263772,-0.533073,
        -0.628319,1.266731,-1.412461,
        -0.628319,1.268591,-3.122146,
        -0.628319,1.269614,-1.275246,
        -0.628319,1.270745,1.152533,
        -0.628319,1.273606,0.224545,
        -0.628319,1.279716,2.812101,
        -0.628319,1.283786,2.462124,
        -0.628319,1.284798,0.078209,
        -0.628319,1.285402,-2.882049,
        -0.628319,1.286936,1.886053,
        -0.628319,1.289408,0.683122,
        -0.628319,1.294736,1.756184,
        -0.628319,1.296978,-2.311511,
        -0.628319,1.298546,2.020601,
        -0.628319,1.300924,1.022478,
        -0.628319,1.303305,3.043152,
        -0.628319,1.304210,-2.084225,
        -0.628319,1.305241,-2.530249,
        -0.628319,1.307513,1.628003,
        -0.628319,1.310722,0.558267,
        -0.628319,1.314142,2.139329,
        -0.628319,1.316087,-0.777971,
        -0.628319,1.317907,-1.121023,
        -0.628319,1.321019,-2.749776,
        -0.628319,1.321609,2.687717,
        -0.628319,1.322568,1.499415,
        -0.628319,1.322972,-0.395699,
        -0.628319,1.325286,-0.994192,
        -0.628319,1.329146,-0.648513,
        -0.628319,1.329795,-0.042570,
        -0.628319,1.331651,2.344497,
        -0.628319,1.336522,0.435098,
        -0.628319,1.339410,0.900087,
        -0.628319,1.339550,1.370353,
        -0.628319,1.340441,-1.731580,
        -0.628319,1.351855,-1.855234,
        -0.628319,1.360161,-3.018404,
        -0.628319,1.361031,-1.613971,
        -0.628319,1.361196,2.914780,
        -0.628319,1.361998,-2.207254,
        -0.628319,1.362302,1.239384,
        -0.628319,1.365228,2.563808,
        -0.628319,1.365452,0.310522,
        -0.628319,1.367028,-2.420041,
        -0.628319,1.373422,-1.484281,
        -0.628319,1.374245,-1.226417,
        -0.628319,1.375020,-0.170393,
        -0.628319,1.376560,-1.355481,
        -0.628319,1.377070,-1.980984,
        -0.628319,1.377578,-2.636339,
        -0.628319,1.378096,0.774626,
        -0.628319,1.379086,-0.508192,
        -0.628319,1.382562,0.177141,
        -0.628319,1.388849,2.233847,
        -0.628319,1.389268,-0.886349,
        -0.628319,1.391324,1.109396,
        -0.628319,1.391699,-0.288725,
        -0.628319,1.392869,3.136903,
        -0.628319,1.393612,1.952016,
        -0.628319,1.395545,-2.850807,
        -0.628319,1.401926,1.822930,
        -0.628319,1.402149,2.788898,
        -0.628319,1.409552,0.647980,
        -0.628319,1.410840,0.056219,
        -0.628319,1.411944,2.444377,
        -0.628319,1.413639,1.697982,
        -0.628319,1.426479,1.573427,
        -0.628319,1.426725,2.102242,
        -0.628319,1.427868,0.984637,
        -0.628319,1.429580,-2.107232,
        -0.628319,1.431375,-2.313269,
        -0.628319,1.431907,3.015410,
        -0.628319,1.432325,-1.029054,
        -0.628319,1.433341,0.524268,
        -0.628319,1.435181,-0.739771,
        -0.628319,1.437208,-2.529085,
        -0.628319,1.439041,-0.617966,
        -0.628319,1.441473,1.448544,
        -0.628319,1.444527,2.664744,
        -0.628319,1.450513,-0.400625,
        -0.628319,1.451907,-0.066823,
        -0.628319,1.452754,-2.738631,
        -0.628319,1.454061,-2.950416,
        -0.628319,1.456736,1.323154,
        -0.628319,1.458370,0.400433,
        -0.628319,1.459314,-1.765003,
        -0.628319,1.459902,-1.145835,
        -0.628319,1.464952,2.329245,
        -0.628319,1.466050,0.862104,
        -0.628319,1.471330,-1.886160,
        -0.628319,1.475119,-1.648255,
        -0.628319,1.480878,1.194576,
        -0.628319,1.480942,0.273235,
        -0.628319,1.481086,2.890269,
        -0.628319,1.484118,-1.529156,
        -0.628319,1.484741,-1.279219,
        -0.628319,1.487681,-1.406069,
        -0.628319,1.487886,-3.066037,
        -0.628319,1.490555,2.543284,
        -0.628319,1.492424,-0.190932,
        -0.628319,1.495002,-0.839810,
        -0.628319,1.495297,2.007188,
        -0.628319,1.496414,-2.009267,
        -0.628319,1.498261,-2.425270,
        -0.628319,1.498907,-2.207327,
        -0.628319,1.500204,0.737638,
        -0.628319,1.502469,0.151156,
        -0.628319,1.507448,-0.514782,
        -0.628319,1.509287,2.204647,
        -0.628319,1.510518,1.884920,
        -0.628319,1.511948,-2.632360,
        -0.628319,1.515235,1.068565,
        -0.628319,1.522237,3.102018,
        -0.628319,1.522373,1.764991,
        -0.628319,1.522416,-0.954956,
        -0.628319,1.523102,2.765920,
        -0.628319,1.524811,-2.844718,
        -0.628319,1.527261,-0.304916,
        -0.628319,1.529788,0.613208,
        -0.628319,1.533648,1.644808,
        -0.628319,1.535018,0.028869,
        -0.628319,1.541082,2.424608,
        -0.628319,1.546010,1.523462,
        -0.628319,1.551135,-0.642994,
        -0.628319,1.553895,0.489607,
        -0.628319,1.555041,0.947813,
        -0.628319,1.559671,-1.074934,
        -0.628319,1.560345,2.985899,
        -0.628319,1.560487,-2.323400,
        -0.628319,1.561127,1.401111,
        -0.628319,1.567193,2.643172,
        -0.628319,1.568892,2.098362,
        -0.628319,1.571702,-2.098435,
        -0.628319,1.572701,-2.528730,
        -0.628319,1.573572,1.278593,
        -0.628319,1.574849,-0.091505,
        -0.628319,1.576334,0.365236,
        -0.628319,1.577029,-2.962888,
        -0.628319,1.577688,-1.814941,
        -0.628319,1.579542,-0.755917,
        -0.628319,1.579624,-1.199759,
        -0.628319,1.581515,-0.419414,
        -0.628319,1.584581,-2.738702,
        -0.628319,1.586948,-1.696296,
        -0.628319,1.588901,2.301496,
        -0.628319,1.589936,0.825020,
        -0.628319,1.590018,-1.941652,
        -0.628319,1.596968,-1.328358,
        -0.628319,1.597178,-1.575607,
        -0.628319,1.600470,-1.453196,
        -0.628319,1.601408,0.242409,
        -0.628319,1.602259,1.152656,
        -0.628319,1.603064,2.865774,
        -0.628319,1.612085,1.968802,
        -0.628319,1.612720,-3.080498,
        -0.628319,1.616061,2.521531,
        -0.628319,1.616900,-0.207457,
        -0.628319,1.620988,-0.875920,
        -0.628319,1.622114,0.701509,
        -0.628319,1.622749,0.121426,
        -0.628319,1.622993,-2.221629,
        -0.628319,1.629346,-0.542669,
        -0.628319,1.630007,1.844552,
        -0.628319,1.634747,-2.426198,
        -0.628319,1.642346,1.722617,
        -0.628319,1.643925,1.034629,
        -0.628319,1.644848,2.743774,
        -0.628319,1.645258,2.188109,
        -0.628319,1.645417,-2.634725,
        -0.628319,1.647395,3.085437,
        -0.628319,1.648538,-2.855044,
        -0.628319,1.649932,0.576808,
        -0.628319,1.652224,-0.996105,
        -0.628319,1.652871,1.600681,
        -0.628319,1.657997,-0.325127,
        -0.628319,1.661936,0.001037,
        -0.628319,1.665187,1.477402,
        -0.628319,1.665496,2.397872,
        -0.628319,1.673267,-0.665453,
        -0.628319,1.673557,0.452526,
        -0.628319,1.679277,-1.122421,
        -0.628319,1.679696,0.912482,
        -0.628319,1.682420,1.353922,
        -0.628319,1.685031,-2.009038,
        -0.628319,1.685196,2.967046,
        -0.628319,1.690107,2.620910,
        -0.628319,1.690453,-2.123076,
        -0.628319,1.691693,-1.871572,
        -0.628319,1.691817,2.061688,
        -0.628319,1.692644,1.237670,
        -0.628319,1.696386,0.328123,
        -0.628319,1.697162,-1.249026,
        -0.628319,1.697488,-2.322785,
        -0.628319,1.698726,-2.973576,
        -0.628319,1.699172,-1.749717,
        -0.628319,1.704478,-0.106295,
        -0.628319,1.707244,-0.444770,
        -0.628319,1.707500,-2.531497,
        -0.628319,1.707714,-0.787321,
        -0.628319,1.708430,-1.627552,
        -0.628319,1.710583,-1.376890,
        -0.628319,1.712415,-2.747921,
        -0.628319,1.713125,-1.503073,
        -0.628319,1.713452,0.789419,
        -0.628319,1.720612,2.279888,
        -0.628319,1.724852,0.202831,
        -0.628319,1.724868,2.846246,
        -0.628319,1.728940,1.931535,
        -0.628319,1.732838,1.127584,
        -0.628319,1.735719,-3.092485,
        -0.628319,1.739676,2.496017,
        -0.628319,1.739971,-0.228876,
        -0.628319,1.743328,0.663347,
        -0.628319,1.744927,-0.911892,
        -0.628319,1.745952,0.084733,
        -0.628319,1.748077,1.804354,
        -0.628319,1.752959,-0.567548,
        -0.628319,1.760169,1.678132,
        -0.628319,1.763887,-2.218450,
        -0.628319,1.766737,2.722752,
        -0.628319,1.767999,1.004737,
        -0.628319,1.769149,1.550298,
        -0.628319,1.770119,-2.427175,
        -0.628319,1.770460,2.154208,
        -0.628319,1.770514,3.070982,
        -0.628319,1.770735,0.537245,
        -0.628319,1.772673,-2.864760,
        -0.628319,1.773527,-1.039582,
        -0.628319,1.776419,-2.642421,
        -0.628319,1.780769,1.416330,
        -0.628319,1.786634,-0.346247,
        -0.628319,1.788829,-1.947553,
        -0.628319,1.793737,0.410270,
        -0.628319,1.794553,2.374618,
        -0.628319,1.794699,-0.694061,
        -0.628319,1.797469,-1.169105,
        -0.628319,1.803705,0.881076,
        -0.628319,1.804124,1.264249,
        -0.628319,1.807825,-2.090455,
        -0.628319,1.808822,-1.817375,
        -0.628319,1.809020,2.949741,
        -0.628319,1.812466,2.597595,
        -0.628319,1.813159,2.025395,
        -0.628319,1.813816,-1.298435,
        -0.628319,1.815016,0.278951,
        -0.628319,1.816510,-0.008631,
        -0.628319,1.817772,-1.689625,
        -0.628319,1.821360,-2.984352,
        -0.628319,1.822312,-0.127211,
        -0.628319,1.823386,-1.429807,
        -0.628319,1.823834,-1.560483,
        -0.628319,1.832060,-0.821627,
        -0.628319,1.832690,-0.468440,
        -0.628319,1.833945,-2.320278,
        -0.628319,1.834787,0.751968,
        -0.628319,1.840579,-2.536445,
        -0.628319,1.841033,-2.756932,
        -0.628319,1.846350,1.893983,
        -0.628319,1.846644,1.113766,
        -0.628319,1.846833,2.826543,
        -0.628319,1.847325,2.249051,
        -0.628319,1.854119,0.134161,
        -0.628319,1.858863,-3.106022,
        -0.628319,1.865046,1.763384,
        -0.628319,1.865619,-0.952036,
        -0.628319,1.865912,0.624453,
        -0.628319,1.866694,2.473686,
        -0.628319,1.867524,-0.240083,
        -0.628319,1.875561,1.632831,
        -0.628319,1.876049,-0.594378,
        -0.628319,1.880509,1.500204,
        -0.628319,1.883504,1.360457,
        -0.628319,1.886971,-2.196665,
        -0.628319,1.888979,0.982922,
        -0.628319,1.889815,2.701570,
        -0.628319,1.892144,-2.005340,
        -0.628319,1.892348,0.496522,
        -0.628319,1.893281,-1.083976,
        -0.628319,1.893828,3.053453,
        -0.628319,1.895117,2.121107,
        -0.628319,1.898741,-2.875735,
        -0.628319,1.904429,-2.428029,
        -0.628319,1.908408,-2.648903,
        -0.628319,1.911828,-0.364489,
        -0.628319,1.912224,0.366565,
        -0.628319,1.914854,1.214468,
        -0.628319,1.915467,-1.883143,
        -0.628319,1.915763,-0.724679,
        -0.628319,1.915790,-1.216366,
        -0.628319,1.922256,2.347806,
        -0.628319,1.922616,0.844633,
        -0.628319,1.925185,0.231170,
        -0.628319,1.927474,-1.754616,
        -0.628319,1.928342,-1.351174,
        -0.628319,1.933328,-1.623328,
        -0.628319,1.933878,0.034150,
        -0.628319,1.934585,-1.487979,
        -0.628319,1.934802,2.927634,
        -0.628319,1.934862,1.989776,
        -0.628319,1.936327,-0.099350,
        -0.628319,1.938088,2.578507,
        -0.628319,1.945677,-2.999805,
        -0.628319,1.954572,-0.859208,
        -0.628319,1.957868,-0.490784,
        -0.628319,1.958989,0.712772,
        -0.628319,1.960212,-2.307561,
        -0.628319,1.964774,1.855507,
        -0.628319,1.964971,-2.100533,
        -0.628319,1.969112,1.088273,
        -0.628319,1.971438,-2.765415,
        -0.628319,1.972312,2.805134,
        -0.628319,1.974950,2.220596,
        -0.628319,1.975135,-2.539127,
        -0.628319,1.982107,1.720829,
        -0.628319,1.983195,-3.127984,
        -0.628319,1.983318,-0.247182,
        -0.628319,1.986720,-0.993220,
        -0.628319,1.990381,1.586005,
        -0.628319,1.990470,0.580283,
        -0.628319,1.990523,1.450326,
        -0.628319,1.993909,1.312897,
        -0.628319,1.994694,2.452813,
        -0.628319,1.999226,-0.620717,
        -0.628319,2.005145,0.941012,
        -0.628319,2.007301,0.142337,
        -0.628319,2.014149,0.444138,
        -0.628319,2.014338,-1.127392,
        -0.628319,2.018131,3.021778,
        -0.628319,2.019676,-1.980944,
        -0.628319,2.019912,2.686612,
        -0.628319,2.020545,2.088672,
        -0.628319,2.026074,0.300945,
        -0.628319,2.026196,-2.889057,
        -0.628319,2.031959,-1.268178,
        -0.628319,2.032917,-2.199917,
        -0.628319,2.033572,-1.842884,
        -0.628319,2.034712,-2.419505,
        -0.628319,2.035818,-0.379479,
        -0.628319,2.036004,-0.757062,
        -0.628319,2.040501,-1.411488,
        -0.628319,2.042874,-1.701367,
        -0.628319,2.043060,-2.653746,
        -0.628319,2.043729,-1.557938,
        -0.628319,2.048538,0.806623,
        -0.628319,2.049006,1.199930,
        -0.628319,2.049502,-0.137989,
        -0.628319,2.052181,2.326959,
        -0.628319,2.052794,-0.008526,
        -0.628319,2.057608,1.951726,
        -0.628319,2.065568,2.567730,
        -0.628319,2.070371,2.879808,
        -0.628319,2.071256,-3.018961,
        -0.628319,2.079746,-0.897699,
        -0.628319,2.082784,1.042403,
        -0.628319,2.083752,-0.511086,
        -0.628319,2.084401,1.810502,
        -0.628319,2.085731,0.671081,
        -0.628319,2.097496,1.373172,
        -0.628319,2.099185,1.667006,
        -0.628319,2.100482,-2.301084,
        -0.628319,2.103118,2.194745,
        -0.628319,2.103223,1.521685,
        -0.628319,2.103695,-2.777707,
        -0.628319,2.107523,3.129581,
        -0.628319,2.108277,-2.095907,
        -0.628319,2.108891,-2.532166,
        -0.628319,2.111003,-0.268632,
        -0.628319,2.112276,-1.027246,
        -0.628319,2.114380,0.531457,
        -0.628319,2.119166,0.100552,
        -0.628319,2.119564,2.754818,
        -0.628319,2.122474,-0.648314,
        -0.628319,2.125183,2.445598,
        -0.628319,2.129501,0.238290,
        -0.628319,2.131178,0.386663,
        -0.628319,2.131223,-1.174311,
        -0.628319,2.134269,-1.948042,
        -0.628319,2.134684,0.907065,
        -0.628319,2.137615,2.988806,
        -0.628319,2.145635,-1.327708,
        -0.628319,2.146467,2.055081,
        -0.628319,2.146582,-1.793788,
        -0.628319,2.149210,-1.482321,
        -0.628319,2.152901,-1.636788,
        -0.628319,2.156337,1.151558,
        -0.628319,2.156474,-2.912065,
        -0.628319,2.157194,-0.792563,
        -0.628319,2.166724,-0.401638,
        -0.628319,2.166980,-0.134879,
        -0.628319,2.173644,-2.403429,
        -0.628319,2.175711,-2.655523,
        -0.628319,2.177132,2.600419,
        -0.628319,2.178879,0.767722,
        -0.628319,2.180752,1.908485,
        -0.628319,2.181338,2.313721,
        -0.628319,2.185580,1.287673,
        -0.628319,2.195400,-2.217546,
        -0.628319,2.200418,-3.056442,
        -0.628319,2.201829,-0.001735,
        -0.628319,2.203804,1.756176,
        -0.628319,2.211602,1.443930,
        -0.628319,2.212686,-0.544809,
        -0.628319,2.213937,0.621892,
        -0.628319,2.214075,1.600691,
        -0.628319,2.216256,1.020029,
        -0.628319,2.219060,2.886631,
        -0.628319,2.223375,-0.906185,
        -0.628319,2.226002,-1.060248,
        -0.628319,2.226413,-2.064034,
        -0.628319,2.230554,2.171334,
        -0.628319,2.232413,2.735777,
        -0.628319,2.233709,3.070830,
        -0.628319,2.234383,-2.792065,
        -0.628319,2.237183,0.468799,
        -0.628319,2.237695,-0.271388,
        -0.628319,2.239419,0.151693,
        -0.628319,2.243347,-2.521049,
        -0.628319,2.244035,-1.229456,
        -0.628319,2.244355,0.309899,
        -0.628319,2.245175,2.460236,
        -0.628319,2.245628,-1.899216,
        -0.628319,2.245973,-0.702234,
        -0.628319,2.255601,-1.728131,
        -0.628319,2.256702,-1.401367,
        -0.628319,2.261466,-1.557500,
        -0.628319,2.267390,0.878499,
        -0.628319,2.272099,2.018594,
        -0.628319,2.278618,1.170386,
        -0.628319,2.283304,-2.345584,
        -0.628319,2.285175,-2.943090,
        -0.628319,2.293000,-0.124451,
        -0.628319,2.294058,-0.421070,
        -0.628319,2.303283,1.855938,
        -0.628319,2.304968,2.608978,
        -0.628319,2.305347,2.311346,
        -0.628319,2.307188,-2.656134,
        -0.628319,2.307301,1.336200,
        -0.628319,2.309655,0.726020,
        -0.628319,2.317634,-2.185225,
        -0.628319,2.321510,1.685664,
        -0.628319,2.324376,-3.111006,
        -0.628319,2.324731,1.511186,
        -0.628319,2.329587,0.035498,
        -0.628319,2.330158,2.977264,
        -0.628319,2.335096,-0.588468,
        -0.628319,2.336046,-1.115799,
        -0.628319,2.338060,2.803613,
        -0.628319,2.338083,-0.943400,
        -0.628319,2.340210,0.562145,
        -0.628319,2.341931,1.023926,
        -0.628319,2.343265,-0.779061,
        -0.628319,2.343434,-2.013692,
        -0.628319,2.353029,-1.310529,
        -0.628319,2.354868,0.211543,
        -0.628319,2.355120,0.388593,
        -0.628319,2.355165,-1.833173,
        -0.628319,2.355462,2.151441,
        -0.628319,2.362948,-2.487477,
        -0.628319,2.363101,-2.810456,
        -0.628319,2.363394,-0.272918,
        -0.628319,2.364509,-1.644533,
        -0.628319,2.370465,2.466824,
        -0.628319,2.386110,-1.481351,
        -0.628319,2.394960,0.862133,
        -0.628319,2.395716,1.977184,
        -0.628319,2.396339,1.200900,
        -0.628319,2.407941,-2.316753,
        -0.628319,2.409450,-2.988432,
        -0.628319,2.417202,-0.103363,
        -0.628319,2.417386,-0.443097,
        -0.628319,2.417950,2.666662,
        -0.628319,2.423190,1.789783,
        -0.628319,2.423692,1.392696,
        -0.628319,2.431122,3.086379,
        -0.628319,2.431406,-2.647820,
        -0.628319,2.431584,2.307631,
        -0.628319,2.434090,1.592818,
        -0.628319,2.435993,0.683844,
        -0.628319,2.440898,-2.133300,
        -0.628319,2.442222,-1.195837,
        -0.628319,2.443112,2.883761,
        -0.628319,2.449191,-0.633651,
        -0.628319,2.450821,0.084741,
        -0.628319,2.453499,-1.008861,
        -0.628319,2.460010,-1.938011,
        -0.628319,2.460199,-0.822008,
        -0.628319,2.460902,0.490079,
        -0.628319,2.461774,-1.734913,
        -0.628319,2.463986,1.034407,
        -0.628319,2.467647,0.287667,
        -0.628319,2.479849,2.126925,
        -0.628319,2.485907,-2.835471,
        -0.628319,2.485945,-1.374853,
        -0.628319,2.487747,2.505912,
        -0.628319,2.487957,-0.274279,
        -0.628319,2.489741,-2.468428,
        -0.628319,2.512538,1.239900,
        -0.628319,2.517528,1.923289,
        -0.628319,2.518454,-1.565852,
        -0.628319,2.519348,0.843503,
        -0.628319,2.522766,-3.059231,
        -0.628319,2.528686,2.737228,
        -0.628319,2.534798,-2.266671,
        -0.628319,2.536533,1.466274,
        -0.628319,2.538504,1.699812,
        -0.628319,2.539516,-0.476942,
        -0.628319,2.539532,-0.074548,
        -0.628319,2.543197,2.990252,
        -0.628319,2.554022,2.320619,
        -0.628319,2.556919,-2.655005,
        -0.628319,2.558146,0.627191,
        -0.628319,2.558223,-1.164189,
        -0.628319,2.563452,-2.044028,
        -0.628319,2.568462,0.150655,
        -0.628319,2.569130,-0.700837,
        -0.628319,2.570050,-1.804492,
        -0.628319,2.572710,-0.934665,
        -0.628319,2.575994,0.391718,
        -0.628319,2.584774,1.046703,
        -0.628319,2.601432,2.102535,
        -0.628319,2.603559,2.555114,
        -0.628319,2.604150,-2.885208,
        -0.628319,2.608630,-1.378011,
        -0.628319,2.611256,-0.279110,
        -0.628319,2.617344,-2.439480,
        -0.628319,2.627402,1.293552,
        -0.628319,2.635396,1.848850,
        -0.628319,2.635686,2.834950,
        -0.628319,2.637087,3.132730,
        -0.628319,2.642029,0.813274,
        -0.628319,2.645260,1.570203,
        -0.628319,2.646182,-1.614036,
        -0.628319,2.660113,-2.186513,
        -0.628319,2.660483,-0.034608,
        -0.628319,2.661713,-0.529511,
        -0.628319,2.672720,-1.110422,
        -0.628319,2.676626,2.335353,
        -0.628319,2.676638,0.540364,
        -0.628319,2.676780,-1.901158,
        -0.628319,2.681411,-2.667965,
        -0.628319,2.682152,0.245751,
        -0.628319,2.682859,-0.817071,
        -0.628319,2.705107,1.062142,
        -0.628319,2.717705,2.625995,
        -0.628319,2.721001,2.060097,
        -0.628319,2.722319,-2.954069,
        -0.628319,2.729840,-1.368237,
        -0.628319,2.734349,-0.289298,
        -0.628319,2.738409,2.975671,
        -0.628319,2.740366,1.374341,
        -0.628319,2.744633,-2.391311,
        -0.628319,2.747398,1.727166,
        -0.628319,2.763799,0.760372,
        -0.628319,2.768238,-1.679243,
        -0.628319,2.779245,-2.046268,
        -0.628319,2.780208,-0.626453,
        -0.628319,2.780471,0.027732,
        -0.628319,2.784696,-1.013128,
        -0.628319,2.789817,0.395431,
        -0.628319,2.801154,2.353143,
        -0.628319,2.807091,-2.687477,
        -0.628319,2.825762,1.085035,
        -0.628319,2.830209,2.734943,
        -0.628319,2.833517,-3.085508,
        -0.628319,2.838301,1.980061,
        -0.628319,2.849871,1.512012,
        -0.628319,2.850089,-1.343899,
        -0.628319,2.856905,-0.318441,
        -0.628319,2.870445,-2.291001,
        -0.628319,2.885017,0.655562,
        -0.628319,2.886428,-1.793217,
        -0.628319,2.890337,-0.825633,
        -0.628319,2.898932,0.139973,
        -0.628319,2.931927,-2.728768,
        -0.628319,2.932096,2.377790,
        -0.628319,2.937409,2.947245,
        -0.628319,2.948048,1.133127,
        -0.628319,2.950896,1.797889,
        -0.628319,2.970019,-1.282382,
        -0.628319,2.978333,-0.412090,
        -0.628319,2.996437,-2.089121,
        -0.628319,3.011274,0.434110,
        -0.628319,3.050410,3.138492,
        -0.628319,3.064492,1.647259,
        -0.628319,3.088868,-0.989501,
        -0.523599,0.053884,-2.408394,
        -0.523599,0.077645,1.453433,
        -0.523599,0.088408,-0.311160,
        -0.523599,0.142314,2.590387,
        -0.523599,0.147206,-1.385164,
        -0.523599,0.163506,0.573203,
        -0.523599,0.165485,-2.875396,
        -0.523599,0.182564,-2.125821,
        -0.523599,0.190569,1.929141,
        -0.523599,0.201102,1.249082,
        -0.523599,0.201397,-0.716646,
        -0.523599,0.210568,-0.069559,
        -0.523599,0.251672,2.848900,
        -0.523599,0.266262,-1.678068,
        -0.523599,0.269108,-1.215563,
        -0.523599,0.273777,-2.506756,
        -0.523599,0.279854,0.814581,
        -0.523599,0.281664,-2.969147,
        -0.523599,0.283625,2.405603,
        -0.523599,0.284058,0.357350,
        -0.523599,0.294107,1.593880,
        -0.523599,0.302491,-0.400410,
        -0.523599,0.308059,2.021491,
        -0.523599,0.312219,-2.072057,
        -0.523599,0.321905,-0.829121,
        -0.523599,0.329984,1.201621,
        -0.523599,0.339714,-0.013712,
        -0.523599,0.368784,2.974974,
        -0.523599,0.376428,-1.495532,
        -0.523599,0.378103,2.646436,
        -0.523599,0.385088,0.590355,
        -0.523599,0.386248,-2.690943,
        -0.523599,0.389851,-2.340031,
        -0.523599,0.391502,-1.171374,
        -0.523599,0.398249,-1.812694,
        -0.523599,0.400542,-2.995729,
        -0.523599,0.404273,0.918163,
        -0.523599,0.405617,1.773533,
        -0.523599,0.409009,-0.570243,
        -0.523599,0.410700,0.270433,
        -0.523599,0.415893,1.464890,
        -0.523599,0.420680,2.088164,
        -0.523599,0.424766,-0.258729,
        -0.523599,0.426507,2.367872,
        -0.523599,0.441292,-0.890727,
        -0.523599,0.451839,-2.078193,
        -0.523599,0.459270,1.191562,
        -0.523599,0.470622,0.014002,
        -0.523599,0.481292,-2.529098,
        -0.523599,0.482460,2.849843,
        -0.523599,0.492205,-1.397963,
        -0.523599,0.493314,3.117264,
        -0.523599,0.497870,0.727828,
        -0.523599,0.499783,-1.654379,
        -0.523599,0.502651,0.463333,
        -0.523599,0.505463,2.594557,
        -0.523599,0.506104,-2.874910,
        -0.523599,0.513579,-1.147354,
        -0.523599,0.518569,1.644237,
        -0.523599,0.519001,1.900488,
        -0.523599,0.519277,-0.673809,
        -0.523599,0.524769,-0.421955,
        -0.523599,0.525564,-2.288838,
        -0.523599,0.532541,0.978737,
        -0.523599,0.534610,2.154808,
        -0.523599,0.535020,-1.891846,
        -0.523599,0.540208,1.400497,
        -0.523599,0.542006,0.228337,
        -0.523599,0.550762,-0.180771,
        -0.523599,0.560531,2.378474,
        -0.523599,0.561629,-0.924576,
        -0.523599,0.567465,-2.679392,
        -0.523599,0.588527,1.181755,
        -0.523599,0.591643,-2.098364,
        -0.523599,0.594963,2.997575,
        -0.523599,0.599643,-3.045235,
        -0.523599,0.599962,2.773556,
        -0.523599,0.602620,0.029270,
        -0.523599,0.605711,0.598454,
        -0.523599,0.607708,-1.541026,
        -0.523599,0.609360,-2.464379,
        -0.523599,0.610738,-1.326786,
        -0.523599,0.616968,0.820489,
        -0.523599,0.626379,1.773610,
        -0.523599,0.628626,-1.751180,
        -0.523599,0.630671,0.391949,
        -0.523599,0.632622,-0.737642,
        -0.523599,0.632708,2.566722,
        -0.523599,0.632797,1.989858,
        -0.523599,0.634455,-0.534790,
        -0.523599,0.636205,1.561170,
        -0.523599,0.637681,-1.123423,
        -0.523599,0.641441,-2.842109,
        -0.523599,0.644807,-0.327139,
        -0.523599,0.652440,2.197129,
        -0.523599,0.660897,-2.278593,
        -0.523599,0.664008,1.360091,
        -0.523599,0.670850,-1.945257,
        -0.523599,0.671396,1.004859,
        -0.523599,0.673985,0.203855,
        -0.523599,0.680326,-0.133370,
        -0.523599,0.685083,-0.942363,
        -0.523599,0.688204,-2.632204,
        -0.523599,0.689777,2.385011,
        -0.523599,0.700065,3.115430,
        -0.523599,0.704080,2.911636,
        -0.523599,0.712883,1.174216,
        -0.523599,0.715701,0.699906,
        -0.523599,0.718851,-1.442215,
        -0.523599,0.721019,2.721536,
        -0.523599,0.725300,-2.975134,
        -0.523599,0.727136,-1.633288,
        -0.523599,0.727399,-2.120567,
        -0.523599,0.730751,0.522488,
        -0.523599,0.732892,-1.254330,
        -0.523599,0.734333,-2.446184,
        -0.523599,0.735125,0.037270,
        -0.523599,0.737893,1.866463,
        -0.523599,0.740650,1.681075,
        -0.523599,0.745971,-0.781286,
        -0.523599,0.749432,-0.442559,
        -0.523599,0.750102,2.050636,
        -0.523599,0.751046,0.863821,
        -0.523599,0.757018,-0.608833,
        -0.523599,0.757046,-1.816019,
        -0.523599,0.757389,1.501184,
        -0.523599,0.759800,2.546181,
        -0.523599,0.759859,0.351923,
        -0.523599,0.761141,-2.781102,
        -0.523599,0.769035,-1.078574,
        -0.523599,0.770166,-0.266604,
        -0.523599,0.772218,2.227380,
        -0.523599,0.788382,1.331051,
        -0.523599,0.790269,-2.282409,
        -0.523599,0.803410,-1.984567,
        -0.523599,0.803813,0.191520,
        -0.523599,0.806832,3.022155,
        -0.523599,0.807476,-2.597675,
        -0.523599,0.812512,-0.107335,
        -0.523599,0.814161,1.010949,
        -0.523599,0.817732,-3.084486,
        -0.523599,0.817846,2.390688,
        -0.523599,0.819323,2.845960,
        -0.523599,0.819430,-0.915017,
        -0.523599,0.829198,-1.532491,
        -0.523599,0.829531,0.665695,
        -0.523599,0.829816,1.167147,
        -0.523599,0.833613,-1.357492,
        -0.523599,0.842945,-2.907318,
        -0.523599,0.844883,2.679708,
        -0.523599,0.848189,-1.705003,
        -0.523599,0.849888,1.767901,
        -0.523599,0.851126,0.493370,
        -0.523599,0.852911,1.933394,
        -0.523599,0.857516,-2.432863,
        -0.523599,0.858261,-1.191729,
        -0.523599,0.858279,-2.138895,
        -0.523599,0.859723,-0.734928,
        -0.523599,0.859754,1.603292,
        -0.523599,0.861655,-0.394821,
        -0.523599,0.864300,0.043549,
        -0.523599,0.866940,0.851229,
        -0.523599,0.869748,2.095969,
        -0.523599,0.877352,-0.574668,
        -0.523599,0.879254,-2.733338,
        -0.523599,0.882554,1.443757,
        -0.523599,0.884060,0.334199,
        -0.523599,0.885626,-1.866061,
        -0.523599,0.889834,2.529957,
        -0.523599,0.893140,2.251698,
        -0.523599,0.897463,-0.240822,
        -0.523599,0.898758,-1.031948,
        -0.523599,0.913918,1.287618,
        -0.523599,0.916555,-2.287173,
        -0.523599,0.918305,3.107845,
        -0.523599,0.919522,2.948277,
        -0.523599,0.927138,-2.569686,
        -0.523599,0.929031,0.185383,
        -0.523599,0.933766,-3.015917,
        -0.523599,0.934873,-2.010822,
        -0.523599,0.935636,-1.445855,
        -0.523599,0.936702,-0.866461,
        -0.523599,0.937395,0.729108,
        -0.523599,0.937910,2.787575,
        -0.523599,0.939144,-0.092962,
        -0.523599,0.941718,1.125085,
        -0.523599,0.942873,-1.607680,
        -0.523599,0.943701,0.980328,
        -0.523599,0.947774,2.395256,
        -0.523599,0.952433,0.584347,
        -0.523599,0.954456,-1.287293,
        -0.523599,0.958731,-2.852552,
        -0.523599,0.961838,1.839284,
        -0.523599,0.961898,-0.469425,
        -0.523599,0.965426,1.686615,
        -0.523599,0.970108,1.990771,
        -0.523599,0.972820,-1.762780,
        -0.523599,0.973500,-0.711881,
        -0.523599,0.975253,2.641082,
        -0.523599,0.977824,0.441928,
        -0.523599,0.981485,-2.424766,
        -0.523599,0.981877,1.537100,
        -0.523599,0.982609,-1.136920,
        -0.523599,0.984265,-2.156073,
        -0.523599,0.989150,0.045493,
        -0.523599,0.993052,2.137478,
        -0.523599,0.997766,-2.693871,
        -0.523599,0.999847,-0.334738,
        -0.523599,1.001920,0.848054,
        -0.523599,1.006146,1.389734,
        -0.523599,1.012180,0.304066,
        -0.523599,1.016938,2.270674,
        -0.523599,1.017371,-0.979790,
        -0.523599,1.019495,-1.894487,
        -0.523599,1.023796,-0.583691,
        -0.523599,1.026525,2.506945,
        -0.523599,1.027004,-0.203653,
        -0.523599,1.028027,3.036573,
        -0.523599,1.031435,1.241021,
        -0.523599,1.032316,2.884221,
        -0.523599,1.035863,-3.110475,
        -0.523599,1.042085,-2.292950,
        -0.523599,1.043272,-1.522986,
        -0.523599,1.046265,-2.958847,
        -0.523599,1.049272,-1.376344,
        -0.523599,1.050039,-2.552086,
        -0.523599,1.053682,0.689863,
        -0.523599,1.055458,0.170394,
        -0.523599,1.056622,-0.831866,
        -0.523599,1.057515,1.097998,
        -0.523599,1.058706,-2.036438,
        -0.523599,1.061389,-1.672248,
        -0.523599,1.063758,2.743093,
        -0.523599,1.070157,-0.073422,
        -0.523599,1.072154,1.761506,
        -0.523599,1.074029,1.909193,
        -0.523599,1.074204,0.545276,
        -0.523599,1.074714,-2.805147,
        -0.523599,1.076262,0.966104,
        -0.523599,1.077642,-1.227073,
        -0.523599,1.084151,1.618525,
        -0.523599,1.085085,2.377673,
        -0.523599,1.086822,-0.441112,
        -0.523599,1.087142,2.058815,
        -0.523599,1.098383,-0.696430,
        -0.523599,1.099138,-1.084033,
        -0.523599,1.103208,1.477773,
        -0.523599,1.103783,0.408254,
        -0.523599,1.106275,-2.420800,
        -0.523599,1.108079,2.612943,
        -0.523599,1.109345,-2.172718,
        -0.523599,1.115957,-1.787597,
        -0.523599,1.118930,0.045983,
        -0.523599,1.119777,0.834702,
        -0.523599,1.121943,-2.669042,
        -0.523599,1.124930,2.219917,
        -0.523599,1.125217,1.336294,
        -0.523599,1.130958,2.972931,
        -0.523599,1.131400,-0.306629,
        -0.523599,1.137135,-0.936514,
        -0.523599,1.139125,-1.924528,
        -0.523599,1.139877,0.276382,
        -0.523599,1.142003,-3.058022,
        -0.523599,1.147363,-0.558448,
        -0.523599,1.150348,-0.176418,
        -0.523599,1.151198,1.197016,
        -0.523599,1.153930,-1.597104,
        -0.523599,1.155389,2.837148,
        -0.523599,1.155458,-1.460623,
        -0.523599,1.157331,-2.906362,
        -0.523599,1.158594,2.487618,
        -0.523599,1.162103,-1.322553,
        -0.523599,1.165407,3.092861,
        -0.523599,1.168025,0.717075,
        -0.523599,1.168406,-2.301519,
        -0.523599,1.174782,-2.538755,
        -0.523599,1.177998,1.061256,
        -0.523599,1.178898,1.827305,
        -0.523599,1.181331,-2.060330,
        -0.523599,1.182540,-0.798952,
        -0.523599,1.184569,0.147517,
        -0.523599,1.184734,1.968919,
        -0.523599,1.184804,0.593386,
        -0.523599,1.188278,1.690808,
        -0.523599,1.194360,2.110440,
        -0.523599,1.194776,2.710455,
        -0.523599,1.197745,-1.173705,
        -0.523599,1.199898,-2.777257,
        -0.523599,1.201164,2.355931,
        -0.523599,1.203760,1.556404,
        -0.523599,1.205845,-0.415993,
        -0.523599,1.206585,0.933648,
        -0.523599,1.206950,-0.037248,
        -0.523599,1.209341,0.471162,
        -0.523599,1.213483,-1.039417,
        -0.523599,1.218146,-0.672643,
        -0.523599,1.221608,1.421669,
        -0.523599,1.226043,-1.693533,
        -0.523599,1.232870,-1.822773,
        -0.523599,1.234375,-2.421194,
        -0.523599,1.234531,-2.190148,
        -0.523599,1.239414,0.348959,
        -0.523599,1.240183,2.587255,
        -0.523599,1.242312,2.941234,
        -0.523599,1.244075,1.286644,
        -0.523599,1.245961,-2.997461,
        -0.523599,1.246495,-2.646872,
        -0.523599,1.247736,2.242810,
        -0.523599,1.253604,0.812222,
        -0.523599,1.257816,-0.894317,
        -0.523599,1.258247,-1.953634,
        -0.523599,1.258993,-0.284099,
        -0.523599,1.262380,-0.157609,
        -0.523599,1.262791,-1.549229,
        -0.523599,1.263772,-0.533073,
        -0.523599,1.266731,-1.412461,
        -0.523599,1.268591,-3.122146,
        -0.523599,1.269614,-1.275246,
        -0.523599,1.270745,1.152533,
        -0.523599,1.273606,0.224545,
        -0.523599,1.279716,2.812101,
        -0.523599,1.283786,2.462124,
        -0.523599,1.284798,0.078209,
        -0.523599,1.285402,-2.882049,
        -0.523599,1.286936,1.886053,
        -0.523599,1.289408,0.683122,
        -0.523599,1.294736,1.756184,
        -0.523599,1.296978,-2.311511,
        -0.523599,1.298546,2.020601,
        -0.523599,1.300924,1.022478,
        -0.523599,1.303305,3.043152,
        -0.523599,1.304210,-2.084225,
        -0.523599,1.305241,-2.530249,
        -0.523599,1.307513,1.628003,
        -0.523599,1.310722,0.558267,
        -0.523599,1.314142,2.139329,
        -0.523599,1.316087,-0.777971,
        -0.523599,1.317907,-1.121023,
        -0.523599,1.321019,-2.749776,
        -0.523599,1.321609,2.687717,
        -0.523599,1.322568,1.499415,
        -0.523599,1.322972,-0.395699,
        -0.523599,1.325286,-0.994192,
        -0.523599,1.329146,-0.648513,
        -0.523599,1.329795,-0.042570,
        -0.523599,1.331651,2.344497,
        -0.523599,1.336522,0.435098,
        -0.523599,1.339410,0.900087,
        -0.523599,1.339550,1.370353,
        -0.523599,1.340441,-1.731580,
        -0.523599,1.351855,-1.855234,
        -0.523599,1.360161,-3.018404,
        -0.523599,1.361031,-1.613971,
        -0.523599,1.361196,2.914780,
        -0.523599,1.361998,-2.207254,
        -0.523599,1.362302,1.239384,
        -0.523599,1.365228,2.563808,
        -0.523599,1.365452,0.310522,
        -0.523599,1.367028,-2.420041,
        -0.523599,1.373422,-1.484281,
        -0.523599,1.374245,-1.226417,
        -0.523599,1.375020,-0.170393,
        -0.523599,1.376560,-1.355481,
        -0.523599,1.377070,-1.980984,
        -0.523599,1.377578,-2.636339,
        -0.523599,1.378096,0.774626,
        -0.523599,1.379086,-0.508192,
        -0.523599,1.382562,0.177141,
        -0.523599,1.388849,2.233847,
        -0.523599,1.389268,-0.886349,
        -0.523599,1.391324,1.109396,
        -0.523599,1.391699,-0.288725,
        -0.523599,1.392869,3.136903,
        -0.523599,1.393612,1.952016,
        -0.523599,1.395545,-2.850807,
        -0.523599,1.401926,1.822930,
        -0.523599,1.402149,2.788898,
        -0.523599,1.409552,0.647980,
        -0.523599,1.410840,0.056219,
        -0.523599,1.411944,2.444377,
        -0.523599,1.413639,1.697982,
        -0.523599,1.426479,1.573427,
        -0.523599,1.426725,2.102242,
        -0.523599,1.427868,0.984637,
        -0.523599,1.429580,-2.107232,
        -0.523599,1.431375,-2.313269,
        -0.523599,1.431907,3.015410,
        -0.523599,1.432325,-1.029054,
        -0.523599,1.433341,0.524268,
        -0.523599,1.435181,-0.739771,
        -0.523599,1.437208,-2.529085,
        -0.523599,1.439041,-0.617966,
        -0.523599,1.441473,1.448544,
        -0.523599,1.444527,2.664744,
        -0.523599,1.450513,-0.400625,
        -0.523599,1.451907,-0.066823,
        -0.523599,1.452754,-2.738631,
        -0.523599,1.454061,-2.950416,
        -0.523599,1.456736,1.323154,
        -0.523599,1.458370,0.400433,
        -0.523599,1.459314,-1.765003,
        -0.523599,1.459902,-1.145835,
        -0.523599,1.464952,2.329245,
        -0.523599,1.466050,0.862104,
        -0.523599,1.471330,-1.886160,
        -0.523599,1.475119,-1.648255,
        -0.523599,1.480878,1.194576,
        -0.523599,1.480942,0.273235,
        -0.523599,1.481086,2.890269,
        -0.523599,1.484118,-1.529156,
        -0.523599,1.484741,-1.279219,
        -0.523599,1.487681,-1.406069,
        -0.523599,1.487886,-3.066037,
        -0.523599,1.490555,2.543284,
        -0.523599,1.492424,-0.190932,
        -0.523599,1.495002,-0.839810,
        -0.523599,1.495297,2.007188,
        -0.523599,1.496414,-2.009267,
        -0.523599,1.498261,-2.425270,
        -0.523599,1.498907,-2.207327,
        -0.523599,1.500204,0.737638,
        -0.523599,1.502469,0.151156,
        -0.523599,1.507448,-0.514782,
        -0.523599,1.509287,2.204647,
        -0.523599,1.510518,1.884920,
        -0.523599,1.511948,-2.632360,
        -0.523599,1.515235,1.068565,
        -0.523599,1.522237,3.102018,
        -0.523599,1.522373,1.764991,
        -0.523599,1.522416,-0.954956,
        -0.523599,1.523102,2.765920,
        -0.523599,1.524811,-2.844718,
        -0.523599,1.527261,-0.304916,
        -0.523599,1.529788,0.613208,
        -0.523599,1.533648,1.644808,
        -0.523599,1.535018,0.028869,
        -0.523599,1.541082,2.424608,
        -0.523599,1.546010,1.523462,
        -0.523599,1.551135,-0.642994,
        -0.523599,1.553895,0.489607,
        -0.523599,1.555041,0.947813,
        -0.523599,1.559671,-1.074934,
        -0.523599,1.560345,2.985899,
        -0.523599,1.560487,-2.323400,
        -0.523599,1.561127,1.401111,
        -0.523599,1.567193,2.643172,
        -0.523599,1.568892,2.098362,
        -0.523599,1.571702,-2.098435,
        -0.523599,1.572701,-2.528730,
        -0.523599,1.573572,1.278593,
        -0.523599,1.574849,-0.091505,
        -0.523599,1.576334,0.365236,
        -0.523599,1.577029,-2.962888,
        -0.523599,1.577688,-1.814941,
        -0.523599,1.579542,-0.755917,
        -0.523599,1.579624,-1.199759,
        -0.523599,1.581515,-0.419414,
        -0.523599,1.584581,-2.738702,
        -0.523599,1.586948,-1.696296,
        -0.523599,1.588901,2.301496,
        -0.523599,1.589936,0.825020,
        -0.523599,1.590018,-1.941652,
        -0.523599,1.596968,-1.328358,
        -0.523599,1.597178,-1.575607,
        -0.523599,1.600470,-1.453196,
        -0.523599,1.601408,0.242409,
        -0.523599,1.602259,1.152656,
        -0.523599,1.603064,2.865774,
        -0.523599,1.612085,1.968802,
        -0.523599,1.612720,-3.080498,
        -0.523599,1.616061,2.521531,
        -0.523599,1.616900,-0.207457,
        -0.523599,1.620988,-0.875920,
        -0.523599,1.622114,0.701509,
        -0.523599,1.622749,0.121426,
        -0.523599,1.622993,-2.221629,
        -0.523599,1.629346,-0.542669,
        -0.523599,1.630007,1.844552,
        -0.523599,1.634747,-2.426198,
        -0.523599,1.642346,1.722617,
        -0.523599,1.643925,1.034629,
        -0.523599,1.644848,2.743774,
        -0.523599,1.645258,2.188109,
        -0.523599,1.645417,-2.634725,
        -0.523599,1.647395,3.085437,
        -0.523599,1.648538,-2.855044,
        -0.523599,1.649932,0.576808,
        -0.523599,1.652224,-0.996105,
        -0.523599,1.652871,1.600681,
        -0.523599,1.657997,-0.325127,
        -0.523599,1.661936,0.001037,
        -0.523599,1.665187,1.477402,
        -0.523599,1.665496,2.397872,
        -0.523599,1.673267,-0.665453,
        -0.523599,1.673557,0.452526,
        -0.523599,1.679277,-1.122421,
        -0.523599,1.679696,0.912482,
        -0.523599,1.682420,1.353922,
        -0.523599,1.685031,-2.009038,
        -0.523599,1.685196,2.967046,
        -0.523599,1.690107,2.620910,
        -0.523599,1.690453,-2.123076,
        -0.523599,1.691693,-1.871572,
        -0.523599,1.691817,2.061688,
        -0.523599,1.692644,1.237670,
        -0.523599,1.696386,0.328123,
        -0.523599,1.697162,-1.249026,
        -0.523599,1.697488,-2.322785,
        -0.523599,1.698726,-2.973576,
        -0.523599,1.699172,-1.749717,
        -0.523599,1.704478,-0.106295,
        -0.523599,1.707244,-0.444770,
        -0.523599,1.707500,-2.531497,
        -0.523599,1.707714,-0.787321,
        -0.523599,1.708430,-1.627552,
        -0.523599,1.710583,-1.376890,
        -0.523599,1.712415,-2.747921,
        -0.523599,1.713125,-1.503073,
        -0.523599,1.713452,0.789419,
        -0.523599,1.720612,2.279888,
        -0.523599,1.724852,0.202831,
        -0.523599,1.724868,2.846246,
        -0.523599,1.728940,1.931535,
        -0.523599,1.732838,1.127584,
        -0.523599,1.735719,-3.092485,
        -0.523599,1.739676,2.496017,
        -0.523599,1.739971,-0.228876,
        -0.523599,1.743328,0.663347,
        -0.523599,1.744927,-0.911892,
        -0.523599,1.745952,0.084733,
        -0.523599,1.748077,1.804354,
        -0.523599,1.752959,-0.567548,
        -0.523599,1.760169,1.678132,
        -0.523599,1.763887,-2.218450,
        -0.523599,1.766737,2.722752,
        -0.523599,1.767999,1.004737,
        -0.523599,1.769149,1.550298,
        -0.523599,1.770119,-2.427175,
        -0.523599,1.770460,2.154208,
        -0.523599,1.770514,3.070982,
        -0.523599,1.770735,0.537245,
        -0.523599,1.772673,-2.864760,
        -0.523599,1.773527,-1.039582,
        -0.523599,1.776419,-2.642421,
        -0.523599,1.780769,1.416330,
        -0.523599,1.786634,-0.346247,
        -0.523599,1.788829,-1.947553,
        -0.523599,1.793737,0.410270,
        -0.523599,1.794553,2.374618,
        -0.523599,1.794699,-0.694061,
        -0.523599,1.797469,-1.169105,
        -0.523599,1.803705,0.881076,
        -0.523599,1.804124,1.264249,
        -0.523599,1.807825,-2.090455,
        -0.523599,1.808822,-1.817375,
        -0.523599,1.809020,2.949741,
        -0.523599,1.812466,2.597595,
        -0.523599,1.813159,2.025395,
        -0.523599,1.813816,-1.298435,
        -0.523599,1.815016,0.278951,
        -0.523599,1.816510,-0.008631,
        -0.523599,1.817772,-1.689625,
        -0.523599,1.821360,-2.984352,
        -0.523599,1.822312,-0.127211,
        -0.523599,1.823386,-1.429807,
        -0.523599,1.823834,-1.560483,
        -0.523599,1.832060,-0.821627,
        -0.523599,1.832690,-0.468440,
        -0.523599,1.833945,-2.320278,
        -0.523599,1.834787,0.751968,
        -0.523599,1.840579,-2.536445,
        -0.523599,1.841033,-2.756932,
        -0.523599,1.846350,1.893983,
        -0.523599,1.846644,1.113766,
        -0.523599,1.846833,2.826543,
        -0.523599,1.847325,2.249051,
        -0.523599,1.854119,0.134161,
        -0.523599,1.858863,-3.106022,
        -0.523599,1.865046,1.763384,
        -0.523599,1.865619,-0.952036,
        -0.523599,1.865912,0.624453,
        -0.523599,1.866694,2.473686,
        -0.523599,1.867524,-0.240083,
        -0.523599,1.875561,1.632831,
        -0.523599,1.876049,-0.594378,
        -0.523599,1.880509,1.500204,
        -0.523599,1.883504,1.360457,
        -0.523599,1.886971,-2.196665,
        -0.523599,1.888979,0.982922,
        -0.523599,1.889815,2.701570,
        -0.523599,1.892144,-2.005340,
        -0.523599,1.892348,0.496522,
        -0.523599,1.893281,-1.083976,
        -0.523599,1.893828,3.053453,
        -0.523599,1.895117,2.121107,
        -0.523599,1.898741,-2.875735,
        -0.523599,1.904429,-2.428029,
        -0.523599,1.908408,-2.648903,
        -0.523599,1.911828,-0.364489,
        -0.523599,1.912224,0.366565,
        -0.523599,1.914854,1.214468,
        -0.523599,1.915467,-1.883143,
        -0.523599,1.915763,-0.724679,
        -0.523599,1.915790,-1.216366,
        -0.523599,1.922256,2.347806,
        -0.523599,1.922616,0.844633,
        -0.523599,1.925185,0.231170,
        -0.523599,1.927474,-1.754616,
        -0.523599,1.928342,-1.351174,
        -0.523599,1.933328,-1.623328,
        -0.523599,1.933878,0.034150,
        -0.523599,1.934585,-1.487979,
        -0.523599,1.934802,2.927634,
        -0.523599,1.934862,1.989776,
        -0.523599,1.936327,-0.099350,
        -0.523599,1.938088,2.578507,
        -0.523599,1.945677,-2.999805,
        -0.523599,1.954572,-0.859208,
        -0.523599,1.957868,-0.490784,
        -0.523599,1.958989,0.712772,
        -0.523599,1.960212,-2.307561,
        -0.523599,1.964774,1.855507,
        -0.523599,1.964971,-2.100533,
        -0.523599,1.969112,1.088273,
        -0.523599,1.971438,-2.765415,
        -0.523599,1.972312,2.805134,
        -0.523599,1.974950,2.220596,
        -0.523599,1.975135,-2.539127,
        -0.523599,1.982107,1.720829,
        -0.523599,1.983195,-3.127984,
        -0.523599,1.983318,-0.247182,
        -0.523599,1.986720,-0.993220,
        -0.523599,1.990381,1.586005,
        -0.523599,1.990470,0.580283,
        -0.523599,1.990523,1.450326,
        -0.523599,1.993909,1.312897,
        -0.523599,1.994694,2.452813,
        -0.523599,1.999226,-0.620717,
        -0.523599,2.005145,0.941012,
        -0.523599,2.007301,0.142337,
        -0.523599,2.014149,0.444138,
        -0.523599,2.014338,-1.127392,
        -0.523599,2.018131,3.021778,
        -0.523599,2.019676,-1.980944,
        -0.523599,2.019912,2.686612,
        -0.523599,2.020545,2.088672,
        -0.523599,2.026074,0.300945,
        -0.523599,2.026196,-2.889057,
        -0.523599,2.031959,-1.268178,
        -0.523599,2.032917,-2.199917,
        -0.523599,2.033572,-1.842884,
        -0.523599,2.034712,-2.419505,
        -0.523599,2.035818,-0.379479,
        -0.523599,2.036004,-0.757062,
        -0.523599,2.040501,-1.411488,
        -0.523599,2.042874,-1.701367,
        -0.523599,2.043060,-2.653746,
        -0.523599,2.043729,-1.557938,
        -0.523599,2.048538,0.806623,
        -0.523599,2.049006,1.199930,
        -0.523599,2.049502,-0.137989,
        -0.523599,2.052181,2.326959,
        -0.523599,2.052794,-0.008526,
        -0.523599,2.057608,1.951726,
        -0.523599,2.065568,2.567730,
        -0.523599,2.070371,2.879808,
        -0.523599,2.071256,-3.018961,
        -0.523599,2.079746,-0.897699,
        -0.523599,2.082784,1.042403,
        -0.523599,2.083752,-0.511086,
        -0.523599,2.084401,1.810502,
        -0.523599,2.085731,0.671081,
        -0.523599,2.097496,1.373172,
        -0.523599,2.099185,1.667006,
        -0.523599,2.100482,-2.301084,
        -0.523599,2.103118,2.194745,
        -0.523599,2.103223,1.521685,
        -0.523599,2.103695,-2.777707,
        -0.523599,2.107523,3.129581,
        -0.523599,2.108277,-2.095907,
        -0.523599,2.108891,-2.532166,
        -0.523599,2.111003,-0.268632,
        -0.523599,2.112276,-1.027246,
        -0.523599,2.114380,0.531457,
        -0.523599,2.119166,0.100552,
        -0.523599,2.119564,2.754818,
        -0.523599,2.122474,-0.648314,
        -0.523599,2.125183,2.445598,
        -0.523599,2.129501,0.238290,
        -0.523599,2.131178,0.386663,
        -0.523599,2.131223,-1.174311,
        -0.523599,2.134269,-1.948042,
        -0.523599,2.134684,0.907065,
        -0.523599,2.137615,2.988806,
        -0.523599,2.145635,-1.327708,
        -0.523599,2.146467,2.055081,
        -0.523599,2.146582,-1.793788,
        -0.523599,2.149210,-1.482321,
        -0.523599,2.152901,-1.636788,
        -0.523599,2.156337,1.151558,
        -0.523599,2.156474,-2.912065,
        -0.523599,2.157194,-0.792563,
        -0.523599,2.166724,-0.401638,
        -0.523599,2.166980,-0.134879,
        -0.523599,2.173644,-2.403429,
        -0.523599,2.175711,-2.655523,
        -0.523599,2.177132,2.600419,
        -0.523599,2.178879,0.767722,
        -0.523599,2.180752,1.908485,
        -0.523599,2.181338,2.313721,
        -0.523599,2.185580,1.287673,
        -0.523599,2.195400,-2.217546,
        -0.523599,2.200418,-3.056442,
        -0.523599,2.201829,-0.001735,
        -0.523599,2.203804,1.756176,
        -0.523599,2.211602,1.443930,
        -0.523599,2.212686,-0.544809,
        -0.523599,2.213937,0.621892,
        -0.523599,2.214075,1.600691,
        -0.523599,2.216256,1.020029,
        -0.523599,2.219060,2.886631,
        -0.523599,2.223375,-0.906185,
        -0.523599,2.226002,-1.060248,
        -0.523599,2.226413,-2.064034,
        -0.523599,2.230554,2.171334,
        -0.523599,2.232413,2.735777,
        -0.523599,2.233709,3.070830,
        -0.523599,2.234383,-2.792065,
        -0.523599,2.237183,0.468799,
        -0.523599,2.237695,-0.271388,
        -0.523599,2.239419,0.151693,
        -0.523599,2.243347,-2.521049,
        -0.523599,2.244035,-1.229456,
        -0.523599,2.244355,0.309899,
        -0.523599,2.245175,2.460236,
        -0.523599,2.245628,-1.899216,
        -0.523599,2.245973,-0.702234,
        -0.523599,2.255601,-1.728131,
        -0.523599,2.256702,-1.401367,
        -0.523599,2.261466,-1.557500,
        -0.523599,2.267390,0.878499,
        -0.523599,2.272099,2.018594,
        -0.523599,2.278618,1.170386,
        -0.523599,2.283304,-2.345584,
        -0.523599,2.285175,-2.943090,
        -0.523599,2.293000,-0.124451,
        -0.523599,2.294058,-0.421070,
        -0.523599,2.303283,1.855938,
        -0.523599,2.304968,2.608978,
        -0.523599,2.305347,2.311346,
        -0.523599,2.307188,-2.656134,
        -0.523599,2.307301,1.336200,
        -0.523599,2.309655,0.726020,
        -0.523599,2.317634,-2.185225,
        -0.523599,2.321510,1.685664,
        -0.523599,2.324376,-3.111006,
        -0.523599,2.324731,1.511186,
        -0.523599,2.329587,0.035498,
        -0.523599,2.330158,2.977264,
        -0.523599,2.335096,-0.588468,
        -0.523599,2.336046,-1.115799,
        -0.523599,2.338060,2.803613,
        -0.523599,2.338083,-0.943400,
        -0.523599,2.340210,0.562145,
        -0.523599,2.341931,1.023926,
        -0.523599,2.343265,-0.779061,
        -0.523599,2.343434,-2.013692,
        -0.523599,2.353029,-1.310529,
        -0.523599,2.354868,0.211543,
        -0.523599,2.355120,0.388593,
        -0.523599,2.355165,-1.833173,
        -0.523599,2.355462,2.151441,
        -0.523599,2.362948,-2.487477,
        -0.523599,2.363101,-2.810456,
        -0.523599,2.363394,-0.272918,
        -0.523599,2.364509,-1.644533,
        -0.523599,2.370465,2.466824,
        -0.523599,2.386110,-1.481351,
        -0.523599,2.394960,0.862133,
        -0.523599,2.395716,1.977184,
        -0.523599,2.396339,1.200900,
        -0.523599,2.407941,-2.316753,
        -0.523599,2.409450,-2.988432,
        -0.523599,2.417202,-0.103363,
        -0.523599,2.417386,-0.443097,
        -0.523599,2.417950,2.666662,
        -0.523599,2.423190,1.789783,
        -0.523599,2.423692,1.392696,
        -0.523599,2.431122,3.086379,
        -0.523599,2.431406,-2.647820,
        -0.523599,2.431584,2.307631,
        -0.523599,2.434090,1.592818,
        -0.523599,2.435993,0.683844,
        -0.523599,2.440898,-2.133300,
        -0.523599,2.442222,-1.195837,
        -0.523599,2.443112,2.883761,
        -0.523599,2.449191,-0.633651,
        -0.523599,2.450821,0.084741,
        -0.523599,2.453499,-1.008861,
        -0.523599,2.460010,-1.938011,
        -0.523599,2.460199,-0.822008,
        -0.523599,2.460902,0.490079,
        -0.523599,2.461774,-1.734913,
        -0.523599,2.463986,1.034407,
        -0.523599,2.467647,0.287667,
        -0.523599,2.479849,2.126925,
        -0.523599,2.485907,-2.835471,
        -0.523599,2.485945,-1.374853,
        -0.523599,2.487747,2.505912,
        -0.523599,2.487957,-0.274279,
        -0.523599,2.489741,-2.468428,
        -0.523599,2.512538,1.239900,
        -0.523599,2.517528,1.923289,
        -0.523599,2.518454,-1.565852,
        -0.523599,2.519348,0.843503,
        -0.523599,2.522766,-3.059231,
        -0.523599,2.528686,2.737228,
        -0.523599,2.534798,-2.266671,
        -0.523599,2.536533,1.466274,
        -0.523599,2.538504,1.699812,
        -0.523599,2.539516,-0.476942,
        -0.523599,2.539532,-0.074548,
        -0.523599,2.543197,2.990252,
        -0.523599,2.554022,2.320619,
        -0.523599,2.556919,-2.655005,
        -0.523599,2.558146,0.627191,
        -0.523599,2.558223,-1.164189,
        -0.523599,2.563452,-2.044028,
        -0.523599,2.568462,0.150655,
        -0.523599,2.569130,-0.700837,
        -0.523599,2.570050,-1.804492,
        -0.523599,2.572710,-0.934665,
        -0.523599,2.575994,0.391718,
        -0.523599,2.584774,1.046703,
        -0.523599,2.601432,2.102535,
        -0.523599,2.603559,2.555114,
        -0.523599,2.604150,-2.885208,
        -0.523599,2.608630,-1.378011,
        -0.523599,2.611256,-0.279110,
        -0.523599,2.617344,-2.439480,
        -0.523599,2.627402,1.293552,
        -0.523599,2.635396,1.848850,
        -0.523599,2.635686,2.834950,
        -0.523599,2.637087,3.132730,
        -0.523599,2.642029,0.813274,
        -0.523599,2.645260,1.570203,
        -0.523599,2.646182,-1.614036,
        -0.523599,2.660113,-2.186513,
        -0.523599,2.660483,-0.034608,
        -0.523599,2.661713,-0.529511,
        -0.523599,2.672720,-1.110422,
        -0.523599,2.676626,2.335353,
        -0.523599,2.676638,0.540364,
        -0.523599,2.676780,-1.901158,
        -0.523599,2.681411,-2.667965,
        -0.523599,2.682152,0.245751,
        -0.523599,2.682859,-0.817071,
        -0.523599,2.705107,1.062142,
        -0.523599,2.717705,2.625995,
        -0.523599,2.721001,2.060097,
        -0.523599,2.722319,-2.954069,
        -0.523599,2.729840,-1.368237,
        -0.523599,2.734349,-0.289298,
        -0.523599,2.738409,2.975671,
        -0.523599,2.740366,1.374341,
        -0.523599,2.744633,-2.391311,
        -0.523599,2.747398,1.727166,
        -0.523599,2.763799,0.760372,
        -0.523599,2.768238,-1.679243,
        -0.523599,2.779245,-2.046268,
        -0.523599,2.780208,-0.626453,
        -0.523599,2.780471,0.027732,
        -0.523599,2.784696,-1.013128,
        -0.523599,2.789817,0.395431,
        -0.523599,2.801154,2.353143,
        -0.523599,2.807091,-2.687477,
        -0.523599,2.825762,1.085035,
        -0.523599,2.830209,2.734943,
        -0.523599,2.833517,-3.085508,
        -0.523599,2.838301,1.980061,
        -0.523599,2.849871,1.512012,
        -0.523599,2.850089,-1.343899,
        -0.523599,2.856905,-0.318441,
        -0.523599,2.870445,-2.291001,
        -0.523599,2.885017,0.655562,
        -0.523599,2.886428,-1.793217,
        -0.523599,2.890337,-0.825633,
        -0.523599,2.898932,0.139973,
        -0.523599,2.931927,-2.728768,
        -0.523599,2.932096,2.377790,
        -0.523599,2.937409,2.947245,
        -0.523599,2.948048,1.133127,
        -0.523599,2.950896,1.797889,
        -0.523599,2.970019,-1.282382,
        -0.523599,2.978333,-0.412090,
        -0.523599,2.996437,-2.089121,
        -0.523599,3.011274,0.434110,
        -0.523599,3.050410,3.138492,
        -0.523599,3.064492,1.647259,
        -0.523599,3.088868,-0.989501,
        -0.418879,0.053884,-2.408394,
        -0.418879,0.077645,1.453433,
        -0.418879,0.088408,-0.311160,
        -0.418879,0.142314,2.590387,
        -0.418879,0.147206,-1.385164,
        -0.418879,0.163506,0.573203,
        -0.418879,0.165485,-2.875396,
        -0.418879,0.182564,-2.125821,
        -0.418879,0.190569,1.929141,
        -0.418879,0.201102,1.249082,
        -0.418879,0.201397,-0.716646,
        -0.418879,0.210568,-0.069559,
        -0.418879,0.251672,2.848900,
        -0.418879,0.266262,-1.678068,
        -0.418879,0.269108,-1.215563,
        -0.418879,0.273777,-2.506756,
        -0.418879,0.279854,0.814581,
        -0.418879,0.281664,-2.969147,
        -0.418879,0.283625,2.405603,
        -0.418879,0.284058,0.357350,
        -0.418879,0.294107,1.593880,
        -0.418879,0.302491,-0.400410,
        -0.418879,0.308059,2.021491,
        -0.418879,0.312219,-2.072057,
        -0.418879,0.321905,-0.829121,
        -0.418879,0.329984,1.201621,
        -0.418879,0.339714,-0.013712,
        -0.418879,0.368784,2.974974,
        -0.418879,0.376428,-1.495532,
        -0.418879,0.378103,2.646436,
        -0.418879,0.385088,0.590355,
        -0.418879,0.386248,-2.690943,
        -0.418879,0.389851,-2.340031,
        -0.418879,0.391502,-1.171374,
        -0.418879,0.398249,-1.812694,
        -0.418879,0.400542,-2.995729,
        -0.418879,0.404273,0.918163,
        -0.418879,0.405617,1.773533,
        -0.418879,0.409009,-0.570243,
        -0.418879,0.410700,0.270433,
        -0.418879,0.415893,1.464890,
        -0.418879,0.420680,2.088164,
        -0.418879,0.424766,-0.258729,
        -0.418879,0.426507,2.367872,
        -0.418879,0.441292,-0.890727,
        -0.418879,0.451839,-2.078193,
        -0.418879,0.459270,1.191562,
        -0.418879,0.470622,0.014002,
        -0.418879,0.481292,-2.529098,
        -0.418879,0.482460,2.849843,
        -0.418879,0.492205,-1.397963,
        -0.418879,0.493314,3.117264,
        -0.418879,0.497870,0.727828,
        -0.418879,0.499783,-1.654379,
        -0.418879,0.502651,0.463333,
        -0.418879,0.505463,2.594557,
        -0.418879,0.506104,-2.874910,
        -0.418879,0.513579,-1.147354,
        -0.418879,0.518569,1.644237,
        -0.418879,0.519001,1.900488,
        -0.418879,0.519277,-0.673809,
        -0.418879,0.524769,-0.421955,
        -0.418879,0.525564,-2.288838,
        -0.418879,0.532541,0.978737,
        -0.418879,0.534610,2.154808,
        -0.418879,0.535020,-1.891846,
        -0.418879,0.540208,1.400497,
        -0.418879,0.542006,0.228337,
        -0.418879,0.550762,-0.180771,
        -0.418879,0.560531,2.378474,
        -0.418879,0.561629,-0.924576,
        -0.418879,0.567465,-2.679392,
        -0.418879,0.588527,1.181755,
        -0.418879,0.591643,-2.098364,
        -0.418879,0.594963,2.997575,
        -0.418879,0.599643,-3.045235,
        -0.418879,0.599962,2.773556,
        -0.418879,0.602620,0.029270,
        -0.418879,0.605711,0.598454,
        -0.418879,0.607708,-1.541026,
        -0.418879,0.609360,-2.464379,
        -0.418879,0.610738,-1.326786,
        -0.418879,0.616968,0.820489,
        -0.418879,0.626379,1.773610,
        -0.418879,0.628626,-1.751180,
        -0.418879,0.630671,0.391949,
        -0.418879,0.632622,-0.737642,
        -0.418879,0.632708,2.566722,
        -0.418879,0.632797,1.989858,
        -0.418879,0.634455,-0.534790,
        -0.418879,0.636205,1.561170,
        -0.418879,0.637681,-1.123423,
        -0.418879,0.641441,-2.842109,
        -0.418879,0.644807,-0.327139,
        -0.418879,0.652440,2.197129,
        -0.418879,0.660897,-2.278593,
        -0.418879,0.664008,1.360091,
        -0.418879,0.670850,-1.945257,
        -0.418879,0.671396,1.004859,
        -0.418879,0.673985,0.203855,
        -0.418879,0.680326,-0.133370,
        -0.418879,0.685083,-0.942363,
        -0.418879,0.688204,-2.632204,
        -0.418879,0.689777,2.385011,
        -0.418879,0.700065,3.115430,
        -0.418879,0.704080,2.911636,
        -0.418879,0.712883,1.174216,
        -0.418879,0.715701,0.699906,
        -0.418879,0.718851,-1.442215,
        -0.418879,0.721019,2.721536,
        -0.418879,0.725300,-2.975134,
        -0.418879,0.727136,-1.633288,
        -0.418879,0.727399,-2.120567,
        -0.418879,0.730751,0.522488,
        -0.418879,0.732892,-1.254330,
        -0.418879,0.734333,-2.446184,
        -0.418879,0.735125,0.037270,
        -0.418879,0.737893,1.866463,
        -0.418879,0.740650,1.681075,
        -0.418879,0.745971,-0.781286,
        -0.418879,0.749432,-0.442559,
        -0.418879,0.750102,2.050636,
        -0.418879,0.751046,0.863821,
        -0.418879,0.757018,-0.608833,
        -0.418879,0.757046,-1.816019,
        -0.418879,0.757389,1.501184,
        -0.418879,0.759800,2.546181,
        -0.418879,0.759859,0.351923,
        -0.418879,0.761141,-2.781102,
        -0.418879,0.769035,-1.078574,
        -0.418879,0.770166,-0.266604,
        -0.418879,0.772218,2.227380,
        -0.418879,0.788382,1.331051,
        -0.418879,0.790269,-2.282409,
        -0.418879,0.803410,-1.984567,
        -0.418879,0.803813,0.191520,
        -0.418879,0.806832,3.022155,
        -0.418879,0.807476,-2.597675,
        -0.418879,0.812512,-0.107335,
        -0.418879,0.814161,1.010949,
        -0.418879,0.817732,-3.084486,
        -0.418879,0.817846,2.390688,
        -0.418879,0.819323,2.845960,
        -0.418879,0.819430,-0.915017,
        -0.418879,0.829198,-1.532491,
        -0.418879,0.829531,0.665695,
        -0.418879,0.829816,1.167147,
        -0.418879,0.833613,-1.357492,
        -0.418879,0.842945,-2.907318,
        -0.418879,0.844883,2.679708,
        -0.418879,0.848189,-1.705003,
        -0.418879,0.849888,1.767901,
        -0.418879,0.851126,0.493370,
        -0.418879,0.852911,1.933394,
        -0.418879,0.857516,-2.432863,
        -0.418879,0.858261,-1.191729,
        -0.418879,0.858279,-2.138895,
        -0.418879,0.859723,-0.734928,
        -0.418879,0.859754,1.603292,
        -0.418879,0.861655,-0.394821,
        -0.418879,0.864300,0.043549,
        -0.418879,0.866940,0.851229,
        -0.418879,0.869748,2.095969,
        -0.418879,0.877352,-0.574668,
        -0.418879,0.879254,-2.733338,
        -0.418879,0.882554,1.443757,
        -0.418879,0.884060,0.334199,
        -0.418879,0.885626,-1.866061,
        -0.418879,0.889834,2.529957,
        -0.418879,0.893140,2.251698,
        -0.418879,0.897463,-0.240822,
        -0.418879,0.898758,-1.031948,
        -0.418879,0.913918,1.287618,
        -0.418879,0.916555,-2.287173,
        -0.418879,0.918305,3.107845,
        -0.418879,0.919522,2.948277,
        -0.418879,0.927138,-2.569686,
        -0.418879,0.929031,0.185383,
        -0.418879,0.933766,-3.015917,
        -0.418879,0.934873,-2.010822,
        -0.418879,0.935636,-1.445855,
        -0.418879,0.936702,-0.866461,
        -0.418879,0.937395,0.729108,
        -0.418879,0.937910,2.787575,
        -0.418879,0.939144,-0.092962,
        -0.418879,0.941718,1.125085,
        -0.418879,0.942873,-1.607680,
        -0.418879,0.943701,0.980328,
        -0.418879,0.947774,2.395256,
        -0.418879,0.952433,0.584347,
        -0.418879,0.954456,-1.287293,
        -0.418879,0.958731,-2.852552,
        -0.418879,0.961838,1.839284,
        -0.418879,0.961898,-0.469425,
        -0.418879,0.965426,1.686615,
        -0.418879,0.970108,1.990771,
        -0.418879,0.972820,-1.762780,
        -0.418879,0.973500,-0.711881,
        -0.418879,0.975253,2.641082,
        -0.418879,0.977824,0.441928,
        -0.418879,0.981485,-2.424766,
        -0.418879,0.981877,1.537100,
        -0.418879,0.982609,-1.136920,
        -0.418879,0.984265,-2.156073,
        -0.418879,0.989150,0.045493,
        -0.418879,0.993052,2.137478,
        -0.418879,0.997766,-2.693871,
        -0.418879,0.999847,-0.334738,
        -0.418879,1.001920,0.848054,
        -0.418879,1.006146,1.389734,
        -0.418879,1.012180,0.304066,
        -0.418879,1.016938,2.270674,
        -0.418879,1.017371,-0.979790,
        -0.418879,1.019495,-1.894487,
        -0.418879,1.023796,-0.583691,
        -0.418879,1.026525,2.506945,
        -0.418879,1.027004,-0.203653,
        -0.418879,1.028027,3.036573,
        -0.418879,1.031435,1.241021,
        -0.418879,1.032316,2.884221,
        -0.418879,1.035863,-3.110475,
        -0.418879,1.042085,-2.292950,
        -0.418879,1.043272,-1.522986,
        -0.418879,1.046265,-2.958847,
        -0.418879,1.049272,-1.376344,
        -0.418879,1.050039,-2.552086,
        -0.418879,1.053682,0.689863,
        -0.418879,1.055458,0.170394,
        -0.418879,1.056622,-0.831866,
        -0.418879,1.057515,1.097998,
        -0.418879,1.058706,-2.036438,
        -0.418879,1.061389,-1.672248,
        -0.418879,1.063758,2.743093,
        -0.418879,1.070157,-0.073422,
        -0.418879,1.072154,1.761506,
        -0.418879,1.074029,1.909193,
        -0.418879,1.074204,0.545276,
        -0.418879,1.074714,-2.805147,
        -0.418879,1.076262,0.966104,
        -0.418879,1.077642,-1.227073,
        -0.418879,1.084151,1.618525,
        -0.418879,1.085085,2.377673,
        -0.418879,1.086822,-0.441112,
        -0.418879,1.087142,2.058815,
        -0.418879,1.098383,-0.696430,
        -0.418879,1.099138,-1.084033,
        -0.418879,1.103208,1.477773,
        -0.418879,1.103783,0.408254,
        -0.418879,1.106275,-2.420800,
        -0.418879,1.108079,2.612943,
        -0.418879,1.109345,-2.172718,
        -0.418879,1.115957,-1.787597,
        -0.418879,1.118930,0.045983,
        -0.418879,1.119777,0.834702,
        -0.418879,1.121943,-2.669042,
        -0.418879,1.124930,2.219917,
        -0.418879,1.125217,1.336294,
        -0.418879,1.130958,2.972931,
        -0.418879,1.131400,-0.306629,
        -0.418879,1.137135,-0.936514,
        -0.418879,1.139125,-1.924528,
        -0.418879,1.139877,0.276382,
        -0.418879,1.142003,-3.058022,
        -0.418879,1.147363,-0.558448,
        -0.418879,1.150348,-0.176418,
        -0.418879,1.151198,1.197016,
        -0.418879,1.153930,-1.597104,
        -0.418879,1.155389,2.837148,
        -0.418879,1.155458,-1.460623,
        -0.418879,1.157331,-2.906362,
        -0.418879,1.158594,2.487618,
        -0.418879,1.162103,-1.322553,
        -0.418879,1.165407,3.092861,
        -0.418879,1.168025,0.717075,
        -0.418879,1.168406,-2.301519,
        -0.418879,1.174782,-2.538755,
        -0.418879,1.177998,1.061256,
        -0.418879,1.178898,1.827305,
        -0.418879,1.181331,-2.060330,
        -0.418879,1.182540,-0.798952,
        -0.418879,1.184569,0.147517,
        -0.418879,1.184734,1.968919,
        -0.418879,1.184804,0.593386,
        -0.418879,1.188278,1.690808,
        -0.418879,1.194360,2.110440,
        -0.418879,1.194776,2.710455,
        -0.418879,1.197745,-1.173705,
        -0.418879,1.199898,-2.777257,
        -0.418879,1.201164,2.355931,
        -0.418879,1.203760,1.556404,
        -0.418879,1.205845,-0.415993,
        -0.418879,1.206585,0.933648,
        -0.418879,1.206950,-0.037248,
        -0.418879,1.209341,0.471162,
        -0.418879,1.213483,-1.039417,
        -0.418879,1.218146,-0.672643,
        -0.418879,1.221608,1.421669,
        -0.418879,1.226043,-1.693533,
        -0.418879,1.232870,-1.822773,
        -0.418879,1.234375,-2.421194,
        -0.418879,1.234531,-2.190148,
        -0.418879,1.239414,0.348959,
        -0.418879,1.240183,2.587255,
        -0.418879,1.242312,2.941234,
        -0.418879,1.244075,1.286644,
        -0.418879,1.245961,-2.997461,
        -0.418879,1.246495,-2.646872,
        -0.418879,1.247736,2.242810,
        -0.418879,1.253604,0.812222,
        -0.418879,1.257816,-0.894317,
        -0.418879,1.258247,-1.953634,
        -0.418879,1.258993,-0.284099,
        -0.418879,1.262380,-0.157609,
        -0.418879,1.262791,-1.549229,
        -0.418879,1.263772,-0.533073,
        -0.418879,1.266731,-1.412461,
        -0.418879,1.268591,-3.122146,
        -0.418879,1.269614,-1.275246,
        -0.418879,1.270745,1.152533,
        -0.418879,1.273606,0.224545,
        -0.418879,1.279716,2.812101,
        -0.418879,1.283786,2.462124,
        -0.418879,1.284798,0.078209,
        -0.418879,1.285402,-2.882049,
        -0.418879,1.286936,1.886053,
        -0.418879,1.289408,0.683122,
        -0.418879,1.294736,1.756184,
        -0.418879,1.296978,-2.311511,
        -0.418879,1.298546,2.020601,
        -0.418879,1.300924,1.022478,
        -0.418879,1.303305,3.043152,
        -0.418879,1.304210,-2.084225,
        -0.418879,1.305241,-2.530249,
        -0.418879,1.307513,1.628003,
        -0.418879,1.310722,0.558267,
        -0.418879,1.314142,2.139329,
        -0.418879,1.316087,-0.777971,
        -0.418879,1.317907,-1.121023,
        -0.418879,1.321019,-2.749776,
        -0.418879,1.321609,2.687717,
        -0.418879,1.322568,1.499415,
        -0.418879,1.322972,-0.395699,
        -0.418879,1.325286,-0.994192,
        -0.418879,1.329146,-0.648513,
        -0.418879,1.329795,-0.042570,
        -0.418879,1.331651,2.344497,
        -0.418879,1.336522,0.435098,
        -0.418879,1.339410,0.900087,
        -0.418879,1.339550,1.370353,
        -0.418879,1.340441,-1.731580,
        -0.418879,1.351855,-1.855234,
        -0.418879,1.360161,-3.018404,
        -0.418879,1.361031,-1.613971,
        -0.418879,1.361196,2.914780,
        -0.418879,1.361998,-2.207254,
        -0.418879,1.362302,1.239384,
        -0.418879,1.365228,2.563808,
        -0.418879,1.365452,0.310522,
        -0.418879,1.367028,-2.420041,
        -0.418879,1.373422,-1.484281,
        -0.418879,1.374245,-1.226417,
        -0.418879,1.375020,-0.170393,
        -0.418879,1.376560,-1.355481,
        -0.418879,1.377070,-1.980984,
        -0.418879,1.377578,-2.636339,
        -0.418879,1.378096,0.774626,
        -0.418879,1.379086,-0.508192,
        -0.418879,1.382562,0.177141,
        -0.418879,1.388849,2.233847,
        -0.418879,1.389268,-0.886349,
        -0.418879,1.391324,1.109396,
        -0.418879,1.391699,-0.288725,
        -0.418879,1.392869,3.136903,
        -0.418879,1.393612,1.952016,
        -0.418879,1.395545,-2.850807,
        -0.418879,1.401926,1.822930,
        -0.418879,1.402149,2.788898,
        -0.418879,1.409552,0.647980,
        -0.418879,1.410840,0.056219,
        -0.418879,1.411944,2.444377,
        -0.418879,1.413639,1.697982,
        -0.418879,1.426479,1.573427,
        -0.418879,1.426725,2.102242,
        -0.418879,1.427868,0.984637,
        -0.418879,1.429580,-2.107232,
        -0.418879,1.431375,-2.313269,
        -0.418879,1.431907,3.015410,
        -0.418879,1.432325,-1.029054,
        -0.418879,1.433341,0.524268,
        -0.418879,1.435181,-0.739771,
        -0.418879,1.437208,-2.529085,
        -0.418879,1.439041,-0.617966,
        -0.418879,1.441473,1.448544,
        -0.418879,1.444527,2.664744,
        -0.418879,1.450513,-0.400625,
        -0.418879,1.451907,-0.066823,
        -0.418879,1.452754,-2.738631,
        -0.418879,1.454061,-2.950416,
        -0.418879,1.456736,1.323154,
        -0.418879,1.458370,0.400433,
        -0.418879,1.459314,-1.765003,
        -0.418879,1.459902,-1.145835,
        -0.418879,1.464952,2.329245,
        -0.418879,1.466050,0.862104,
        -0.418879,1.471330,-1.886160,
        -0.418879,1.475119,-1.648255,
        -0.418879,1.480878,1.194576,
        -0.418879,1.480942,0.273235,
        -0.418879,1.481086,2.890269,
        -0.418879,1.484118,-1.529156,
        -0.418879,1.484741,-1.279219,
        -0.418879,1.487681,-1.406069,
        -0.418879,1.487886,-3.066037,
        -0.418879,1.490555,2.543284,
        -0.418879,1.492424,-0.190932,
        -0.418879,1.495002,-0.839810,
        -0.418879,1.495297,2.007188,
        -0.418879,1.496414,-2.009267,
        -0.418879,1.498261,-2.425270,
        -0.418879,1.498907,-2.207327,
        -0.418879,1.500204,0.737638,
        -0.418879,1.502469,0.151156,
        -0.418879,1.507448,-0.514782,
        -0.418879,1.509287,2.204647,
        -0.418879,1.510518,1.884920,
        -0.418879,1.511948,-2.632360,
        -0.418879,1.515235,1.068565,
        -0.418879,1.522237,3.102018,
        -0.418879,1.522373,1.764991,
        -0.418879,1.522416,-0.954956,
        -0.418879,1.523102,2.765920,
        -0.418879,1.524811,-2.844718,
        -0.418879,1.527261,-0.304916,
        -0.418879,1.529788,0.613208,
        -0.418879,1.533648,1.644808,
        -0.418879,1.535018,0.028869,
        -0.418879,1.541082,2.424608,
        -0.418879,1.546010,1.523462,
        -0.418879,1.551135,-0.642994,
        -0.418879,1.553895,0.489607,
        -0.418879,1.555041,0.947813,
        -0.418879,1.559671,-1.074934,
        -0.418879,1.560345,2.985899,
        -0.418879,1.560487,-2.323400,
        -0.418879,1.561127,1.401111,
        -0.418879,1.567193,2.643172,
        -0.418879,1.568892,2.098362,
        -0.418879,1.571702,-2.098435,
        -0.418879,1.572701,-2.528730,
        -0.418879,1.573572,1.278593,
        -0.418879,1.574849,-0.091505,
        -0.418879,1.576334,0.365236,
        -0.418879,1.577029,-2.962888,
        -0.418879,1.577688,-1.814941,
        -0.418879,1.579542,-0.755917,
        -0.418879,1.579624,-1.199759,
        -0.418879,1.581515,-0.419414,
        -0.418879,1.584581,-2.738702,
        -0.418879,1.586948,-1.696296,
        -0.418879,1.588901,2.301496,
        -0.418879,1.589936,0.825020,
        -0.418879,1.590018,-1.941652,
        -0.418879,1.596968,-1.328358,
        -0.418879,1.597178,-1.575607,
        -0.418879,1.600470,-1.453196,
        -0.418879,1.601408,0.242409,
        -0.418879,1.602259,1.152656,
        -0.418879,1.603064,2.865774,
        -0.418879,1.612085,1.968802,
        -0.418879,1.612720,-3.080498,
        -0.418879,1.616061,2.521531,
        -0.418879,1.616900,-0.207457,
        -0.418879,1.620988,-0.875920,
        -0.418879,1.622114,0.701509,
        -0.418879,1.622749,0.121426,
        -0.418879,1.622993,-2.221629,
        -0.418879,1.629346,-0.542669,
        -0.418879,1.630007,1.844552,
        -0.418879,1.634747,-2.426198,
        -0.418879,1.642346,1.722617,
        -0.418879,1.643925,1.034629,
        -0.418879,1.644848,2.743774,
        -0.418879,1.645258,2.188109,
        -0.418879,1.645417,-2.634725,
        -0.418879,1.647395,3.085437,
        -0.418879,1.648538,-2.855044,
        -0.418879,1.649932,0.576808,
        -0.418879,1.652224,-0.996105,
        -0.418879,1.652871,1.600681,
        -0.418879,1.657997,-0.325127,
        -0.418879,1.661936,0.001037,
        -0.418879,1.665187,1.477402,
        -0.418879,1.665496,2.397872,
        -0.418879,1.673267,-0.665453,
        -0.418879,1.673557,0.452526,
        -0.418879,1.679277,-1.122421,
        -0.418879,1.679696,0.912482,
        -0.418879,1.682420,1.353922,
        -0.418879,1.685031,-2.009038,
        -0.418879,1.685196,2.967046,
        -0.418879,1.690107,2.620910,
        -0.418879,1.690453,-2.123076,
        -0.418879,1.691693,-1.871572,
        -0.418879,1.691817,2.061688,
        -0.418879,1.692644,1.237670,
        -0.418879,1.696386,0.328123,
        -0.418879,1.697162,-1.249026,
        -0.418879,1.697488,-2.322785,
        -0.418879,1.698726,-2.973576,
        -0.418879,1.699172,-1.749717,
        -0.418879,1.704478,-0.106295,
        -0.418879,1.707244,-0.444770,
        -0.418879,1.707500,-2.531497,
        -0.418879,1.707714,-0.787321,
        -0.418879,1.708430,-1.627552,
        -0.418879,1.710583,-1.376890,
        -0.418879,1.712415,-2.747921,
        -0.418879,1.713125,-1.503073,
        -0.418879,1.713452,0.789419,
        -0.418879,1.720612,2.279888,
        -0.418879,1.724852,0.202831,
        -0.418879,1.724868,2.846246,
        -0.418879,1.728940,1.931535,
        -0.418879,1.732838,1.127584,
        -0.418879,1.735719,-3.092485,
        -0.418879,1.739676,2.496017,
        -0.418879,1.739971,-0.228876,
        -0.418879,1.743328,0.663347,
        -0.418879,1.744927,-0.911892,
        -0.418879,1.745952,0.084733,
        -0.418879,1.748077,1.804354,
        -0.418879,1.752959,-0.567548,
        -0.418879,1.760169,1.678132,
        -0.418879,1.763887,-2.218450,
        -0.418879,1.766737,2.722752,
        -0.418879,1.767999,1.004737,
        -0.418879,1.769149,1.550298,
        -0.418879,1.770119,-2.427175,
        -0.418879,1.770460,2.154208,
        -0.418879,1.770514,3.070982,
        -0.418879,1.770735,0.537245,
        -0.418879,1.772673,-2.864760,
        -0.418879,1.773527,-1.039582,
        -0.418879,1.776419,-2.642421,
        -0.418879,1.780769,1.416330,
        -0.418879,1.786634,-0.346247,
        -0.418879,1.788829,-1.947553,
        -0.418879,1.793737,0.410270,
        -0.418879,1.794553,2.374618,
        -0.418879,1.794699,-0.694061,
        -0.418879,1.797469,-1.169105,
        -0.418879,1.803705,0.881076,
        -0.418879,1.804124,1.264249,
        -0.418879,1.807825,-2.090455,
        -0.418879,1.808822,-1.817375,
        -0.418879,1.809020,2.949741,
        -0.418879,1.812466,2.597595,
        -0.418879,1.813159,2.025395,
        -0.418879,1.813816,-1.298435,
        -0.418879,1.815016,0.278951,
        -0.418879,1.816510,-0.008631,
        -0.418879,1.817772,-1.689625,
        -0.418879,1.821360,-2.984352,
        -0.418879,1.822312,-0.127211,
        -0.418879,1.823386,-1.429807,
        -0.418879,1.823834,-1.560483,
        -0.418879,1.832060,-0.821627,
        -0.418879,1.832690,-0.468440,
        -0.418879,1.833945,-2.320278,
        -0.418879,1.834787,0.751968,
        -0.418879,1.840579,-2.536445,
        -0.418879,1.841033,-2.756932,
        -0.418879,1.846350,1.893983,
        -0.418879,1.846644,1.113766,
        -0.418879,1.846833,2.826543,
        -0.418879,1.847325,2.249051,
        -0.418879,1.854119,0.134161,
        -0.418879,1.858863,-3.106022,
        -0.418879,1.865046,1.763384,
        -0.418879,1.865619,-0.952036,
        -0.418879,1.865912,0.624453,
        -0.418879,1.866694,2.473686,
        -0.418879,1.867524,-0.240083,
        -0.418879,1.875561,1.632831,
        -0.418879,1.876049,-0.594378,
        -0.418879,1.880509,1.500204,
        -0.418879,1.883504,1.360457,
        -0.418879,1.886971,-2.196665,
        -0.418879,1.888979,0.982922,
        -0.418879,1.889815,2.701570,
        -0.418879,1.892144,-2.005340,
        -0.418879,1.892348,0.496522,
        -0.418879,1.893281,-1.083976,
        -0.418879,1.893828,3.053453,
        -0.418879,1.895117,2.121107,
        -0.418879,1.898741,-2.875735,
        -0.418879,1.904429,-2.428029,
        -0.418879,1.908408,-2.648903,
        -0.418879,1.911828,-0.364489,
        -0.418879,1.912224,0.366565,
        -0.418879,1.914854,1.214468,
        -0.418879,1.915467,-1.883143,
        -0.418879,1.915763,-0.724679,
        -0.418879,1.915790,-1.216366,
        -0.418879,1.922256,2.347806,
        -0.418879,1.922616,0.844633,
        -0.418879,1.925185,0.231170,
        -0.418879,1.927474,-1.754616,
        -0.418879,1.928342,-1.351174,
        -0.418879,1.933328,-1.623328,
        -0.418879,1.933878,0.034150,
        -0.418879,1.934585,-1.487979,
        -0.418879,1.934802,2.927634,
        -0.418879,1.934862,1.989776,
        -0.418879,1.936327,-0.099350,
        -0.418879,1.938088,2.578507,
        -0.418879,1.945677,-2.999805,
        -0.418879,1.954572,-0.859208,
        -0.418879,1.957868,-0.490784,
        -0.418879,1.958989,0.712772,
        -0.418879,1.960212,-2.307561,
        -0.418879,1.964774,1.855507,
        -0.418879,1.964971,-2.100533,
        -0.418879,1.969112,1.088273,
        -0.418879,1.971438,-2.765415,
        -0.418879,1.972312,2.805134,
        -0.418879,1.974950,2.220596,
        -0.418879,1.975135,-2.539127,
        -0.418879,1.982107,1.720829,
        -0.418879,1.983195,-3.127984,
        -0.418879,1.983318,-0.247182,
        -0.418879,1.986720,-0.993220,
        -0.418879,1.990381,1.586005,
        -0.418879,1.990470,0.580283,
        -0.418879,1.990523,1.450326,
        -0.418879,1.993909,1.312897,
        -0.418879,1.994694,2.452813,
        -0.418879,1.999226,-0.620717,
        -0.418879,2.005145,0.941012,
        -0.418879,2.007301,0.142337,
        -0.418879,2.014149,0.444138,
        -0.418879,2.014338,-1.127392,
        -0.418879,2.018131,3.021778,
        -0.418879,2.019676,-1.980944,
        -0.418879,2.019912,2.686612,
        -0.418879,2.020545,2.088672,
        -0.418879,2.026074,0.300945,
        -0.418879,2.026196,-2.889057,
        -0.418879,2.031959,-1.268178,
        -0.418879,2.032917,-2.199917,
        -0.418879,2.033572,-1.842884,
        -0.418879,2.034712,-2.419505,
        -0.418879,2.035818,-0.379479,
        -0.418879,2.036004,-0.757062,
        -0.418879,2.040501,-1.411488,
        -0.418879,2.042874,-1.701367,
        -0.418879,2.043060,-2.653746,
        -0.418879,2.043729,-1.557938,
        -0.418879,2.048538,0.806623,
        -0.418879,2.049006,1.199930,
        -0.418879,2.049502,-0.137989,
        -0.418879,2.052181,2.326959,
        -0.418879,2.052794,-0.008526,
        -0.418879,2.057608,1.951726,
        -0.418879,2.065568,2.567730,
        -0.418879,2.070371,2.879808,
        -0.418879,2.071256,-3.018961,
        -0.418879,2.079746,-0.897699,
        -0.418879,2.082784,1.042403,
        -0.418879,2.083752,-0.511086,
        -0.418879,2.084401,1.810502,
        -0.418879,2.085731,0.671081,
        -0.418879,2.097496,1.373172,
        -0.418879,2.099185,1.667006,
        -0.418879,2.100482,-2.301084,
        -0.418879,2.103118,2.194745,
        -0.418879,2.103223,1.521685,
        -0.418879,2.103695,-2.777707,
        -0.418879,2.107523,3.129581,
        -0.418879,2.108277,-2.095907,
        -0.418879,2.108891,-2.532166,
        -0.418879,2.111003,-0.268632,
        -0.418879,2.112276,-1.027246,
        -0.418879,2.114380,0.531457,
        -0.418879,2.119166,0.100552,
        -0.418879,2.119564,2.754818,
        -0.418879,2.122474,-0.648314,
        -0.418879,2.125183,2.445598,
        -0.418879,2.129501,0.238290,
        -0.418879,2.131178,0.386663,
        -0.418879,2.131223,-1.174311,
        -0.418879,2.134269,-1.948042,
        -0.418879,2.134684,0.907065,
        -0.418879,2.137615,2.988806,
        -0.418879,2.145635,-1.327708,
        -0.418879,2.146467,2.055081,
        -0.418879,2.146582,-1.793788,
        -0.418879,2.149210,-1.482321,
        -0.418879,2.152901,-1.636788,
        -0.418879,2.156337,1.151558,
        -0.418879,2.156474,-2.912065,
        -0.418879,2.157194,-0.792563,
        -0.418879,2.166724,-0.401638,
        -0.418879,2.166980,-0.134879,
        -0.418879,2.173644,-2.403429,
        -0.418879,2.175711,-2.655523,
        -0.418879,2.177132,2.600419,
        -0.418879,2.178879,0.767722,
        -0.418879,2.180752,1.908485,
        -0.418879,2.181338,2.313721,
        -0.418879,2.185580,1.287673,
        -0.418879,2.195400,-2.217546,
        -0.418879,2.200418,-3.056442,
        -0.418879,2.201829,-0.001735,
        -0.418879,2.203804,1.756176,
        -0.418879,2.211602,1.443930,
        -0.418879,2.212686,-0.544809,
        -0.418879,2.213937,0.621892,
        -0.418879,2.214075,1.600691,
        -0.418879,2.216256,1.020029,
        -0.418879,2.219060,2.886631,
        -0.418879,2.223375,-0.906185,
        -0.418879,2.226002,-1.060248,
        -0.418879,2.226413,-2.064034,
        -0.418879,2.230554,2.171334,
        -0.418879,2.232413,2.735777,
        -0.418879,2.233709,3.070830,
        -0.418879,2.234383,-2.792065,
        -0.418879,2.237183,0.468799,
        -0.418879,2.237695,-0.271388,
        -0.418879,2.239419,0.151693,
        -0.418879,2.243347,-2.521049,
        -0.418879,2.244035,-1.229456,
        -0.418879,2.244355,0.309899,
        -0.418879,2.245175,2.460236,
        -0.418879,2.245628,-1.899216,
        -0.418879,2.245973,-0.702234,
        -0.418879,2.255601,-1.728131,
        -0.418879,2.256702,-1.401367,
        -0.418879,2.261466,-1.557500,
        -0.418879,2.267390,0.878499,
        -0.418879,2.272099,2.018594,
        -0.418879,2.278618,1.170386,
        -0.418879,2.283304,-2.345584,
        -0.418879,2.285175,-2.943090,
        -0.418879,2.293000,-0.124451,
        -0.418879,2.294058,-0.421070,
        -0.418879,2.303283,1.855938,
        -0.418879,2.304968,2.608978,
        -0.418879,2.305347,2.311346,
        -0.418879,2.307188,-2.656134,
        -0.418879,2.307301,1.336200,
        -0.418879,2.309655,0.726020,
        -0.418879,2.317634,-2.185225,
        -0.418879,2.321510,1.685664,
        -0.418879,2.324376,-3.111006,
        -0.418879,2.324731,1.511186,
        -0.418879,2.329587,0.035498,
        -0.418879,2.330158,2.977264,
        -0.418879,2.335096,-0.588468,
        -0.418879,2.336046,-1.115799,
        -0.418879,2.338060,2.803613,
        -0.418879,2.338083,-0.943400,
        -0.418879,2.340210,0.562145,
        -0.418879,2.341931,1.023926,
        -0.418879,2.343265,-0.779061,
        -0.418879,2.343434,-2.013692,
        -0.418879,2.353029,-1.310529,
        -0.418879,2.354868,0.211543,
        -0.418879,2.355120,0.388593,
        -0.418879,2.355165,-1.833173,
        -0.418879,2.355462,2.151441,
        -0.418879,2.362948,-2.487477,
        -0.418879,2.363101,-2.810456,
        -0.418879,2.363394,-0.272918,
        -0.418879,2.364509,-1.644533,
        -0.418879,2.370465,2.466824,
        -0.418879,2.386110,-1.481351,
        -0.418879,2.394960,0.862133,
        -0.418879,2.395716,1.977184,
        -0.418879,2.396339,1.200900,
        -0.418879,2.407941,-2.316753,
        -0.418879,2.409450,-2.988432,
        -0.418879,2.417202,-0.103363,
        -0.418879,2.417386,-0.443097,
        -0.418879,2.417950,2.666662,
        -0.418879,2.423190,1.789783,
        -0.418879,2.423692,1.392696,
        -0.418879,2.431122,3.086379,
        -0.418879,2.431406,-2.647820,
        -0.418879,2.431584,2.307631,
        -0.418879,2.434090,1.592818,
        -0.418879,2.435993,0.683844,
        -0.418879,2.440898,-2.133300,
        -0.418879,2.442222,-1.195837,
        -0.418879,2.443112,2.883761,
        -0.418879,2.449191,-0.633651,
        -0.418879,2.450821,0.084741,
        -0.418879,2.453499,-1.008861,
        -0.418879,2.460010,-1.938011,
        -0.418879,2.460199,-0.822008,
        -0.418879,2.460902,0.490079,
        -0.418879,2.461774,-1.734913,
        -0.418879,2.463986,1.034407,
        -0.418879,2.467647,0.287667,
        -0.418879,2.479849,2.126925,
        -0.418879,2.485907,-2.835471,
        -0.418879,2.485945,-1.374853,
        -0.418879,2.487747,2.505912,
        -0.418879,2.487957,-0.274279,
        -0.418879,2.489741,-2.468428,
        -0.418879,2.512538,1.239900,
        -0.418879,2.517528,1.923289,
        -0.418879,2.518454,-1.565852,
        -0.418879,2.519348,0.843503,
        -0.418879,2.522766,-3.059231,
        -0.418879,2.528686,2.737228,
        -0.418879,2.534798,-2.266671,
        -0.418879,2.536533,1.466274,
        -0.418879,2.538504,1.699812,
        -0.418879,2.539516,-0.476942,
        -0.418879,2.539532,-0.074548,
        -0.418879,2.543197,2.990252,
        -0.418879,2.554022,2.320619,
        -0.418879,2.556919,-2.655005,
        -0.418879,2.558146,0.627191,
        -0.418879,2.558223,-1.164189,
        -0.418879,2.563452,-2.044028,
        -0.418879,2.568462,0.150655,
        -0.418879,2.569130,-0.700837,
        -0.418879,2.570050,-1.804492,
        -0.418879,2.572710,-0.934665,
        -0.418879,2.575994,0.391718,
        -0.418879,2.584774,1.046703,
        -0.418879,2.601432,2.102535,
        -0.418879,2.603559,2.555114,
        -0.418879,2.604150,-2.885208,
        -0.418879,2.608630,-1.378011,
        -0.418879,2.611256,-0.279110,
        -0.418879,2.617344,-2.439480,
        -0.418879,2.627402,1.293552,
        -0.418879,2.635396,1.848850,
        -0.418879,2.635686,2.834950,
        -0.418879,2.637087,3.132730,
        -0.418879,2.642029,0.813274,
        -0.418879,2.645260,1.570203,
        -0.418879,2.646182,-1.614036,
        -0.418879,2.660113,-2.186513,
        -0.418879,2.660483,-0.034608,
        -0.418879,2.661713,-0.529511,
        -0.418879,2.672720,-1.110422,
        -0.418879,2.676626,2.335353,
        -0.418879,2.676638,0.540364,
        -0.418879,2.676780,-1.901158,
        -0.418879,2.681411,-2.667965,
        -0.418879,2.682152,0.245751,
        -0.418879,2.682859,-0.817071,
        -0.418879,2.705107,1.062142,
        -0.418879,2.717705,2.625995,
        -0.418879,2.721001,2.060097,
        -0.418879,2.722319,-2.954069,
        -0.418879,2.729840,-1.368237,
        -0.418879,2.734349,-0.289298,
        -0.418879,2.738409,2.975671,
        -0.418879,2.740366,1.374341,
        -0.418879,2.744633,-2.391311,
        -0.418879,2.747398,1.727166,
        -0.418879,2.763799,0.760372,
        -0.418879,2.768238,-1.679243,
        -0.418879,2.779245,-2.046268,
        -0.418879,2.780208,-0.626453,
        -0.418879,2.780471,0.027732,
        -0.418879,2.784696,-1.013128,
        -0.418879,2.789817,0.395431,
        -0.418879,2.801154,2.353143,
        -0.418879,2.807091,-2.687477,
        -0.418879,2.825762,1.085035,
        -0.418879,2.830209,2.734943,
        -0.418879,2.833517,-3.085508,
        -0.418879,2.838301,1.980061,
        -0.418879,2.849871,1.512012,
        -0.418879,2.850089,-1.343899,
        -0.418879,2.856905,-0.318441,
        -0.418879,2.870445,-2.291001,
        -0.418879,2.885017,0.655562,
        -0.418879,2.886428,-1.793217,
        -0.418879,2.890337,-0.825633,
        -0.418879,2.898932,0.139973,
        -0.418879,2.931927,-2.728768,
        -0.418879,2.932096,2.377790,
        -0.418879,2.937409,2.947245,
        -0.418879,2.948048,1.133127,
        -0.418879,2.950896,1.797889,
        -0.418879,2.970019,-1.282382,
        -0.418879,2.978333,-0.412090,
        -0.418879,2.996437,-2.089121,
        -0.418879,3.011274,0.434110,
        -0.418879,3.050410,3.138492,
        -0.418879,3.064492,1.647259,
        -0.418879,3.088868,-0.989501,
        -0.314159,0.053884,-2.408394,
        -0.314159,0.077645,1.453433,
        -0.314159,0.088408,-0.311160,
        -0.314159,0.142314,2.590387,
        -0.314159,0.147206,-1.385164,
        -0.314159,0.163506,0.573203,
        -0.314159,0.165485,-2.875396,
        -0.314159,0.182564,-2.125821,
        -0.314159,0.190569,1.929141,
        -0.314159,0.201102,1.249082,
        -0.314159,0.201397,-0.716646,
        -0.314159,0.210568,-0.069559,
        -0.314159,0.251672,2.848900,
        -0.314159,0.266262,-1.678068,
        -0.314159,0.269108,-1.215563,
        -0.314159,0.273777,-2.506756,
        -0.314159,0.279854,0.814581,
        -0.314159,0.281664,-2.969147,
        -0.314159,0.283625,2.405603,
        -0.314159,0.284058,0.357350,
        -0.314159,0.294107,1.593880,
        -0.314159,0.302491,-0.400410,
        -0.314159,0.308059,2.021491,
        -0.314159,0.312219,-2.072057,
        -0.314159,0.321905,-0.829121,
        -0.314159,0.329984,1.201621,
        -0.314159,0.339714,-0.013712,
        -0.314159,0.368784,2.974974,
        -0.314159,0.376428,-1.495532,
        -0.314159,0.378103,2.646436,
        -0.314159,0.385088,0.590355,
        -0.314159,0.386248,-2.690943,
        -0.314159,0.389851,-2.340031,
        -0.314159,0.391502,-1.171374,
        -0.314159,0.398249,-1.812694,
        -0.314159,0.400542,-2.995729,
        -0.314159,0.404273,0.918163,
        -0.314159,0.405617,1.773533,
        -0.314159,0.409009,-0.570243,
        -0.314159,0.410700,0.270433,
        -0.314159,0.415893,1.464890,
        -0.314159,0.420680,2.088164,
        -0.314159,0.424766,-0.258729,
        -0.314159,0.426507,2.367872,
        -0.314159,0.441292,-0.890727,
        -0.314159,0.451839,-2.078193,
        -0.314159,0.459270,1.191562,
        -0.314159,0.470622,0.014002,
        -0.314159,0.481292,-2.529098,
        -0.314159,0.482460,2.849843,
        -0.314159,0.492205,-1.397963,
        -0.314159,0.493314,3.117264,
        -0.314159,0.497870,0.727828,
        -0.314159,0.499783,-1.654379,
        -0.314159,0.502651,0.463333,
        -0.314159,0.505463,2.594557,
        -0.314159,0.506104,-2.874910,
        -0.314159,0.513579,-1.147354,
        -0.314159,0.518569,1.644237,
        -0.314159,0.519001,1.900488,
        -0.314159,0.519277,-0.673809,
        -0.314159,0.524769,-0.421955,
        -0.314159,0.525564,-2.288838,
        -0.314159,0.532541,0.978737,
        -0.314159,0.534610,2.154808,
        -0.314159,0.535020,-1.891846,
        -0.314159,0.540208,1.400497,
        -0.314159,0.542006,0.228337,
        -0.314159,0.550762,-0.180771,
        -0.314159,0.560531,2.378474,
        -0.314159,0.561629,-0.924576,
        -0.314159,0.567465,-2.679392,
        -0.314159,0.588527,1.181755,
        -0.314159,0.591643,-2.098364,
        -0.314159,0.594963,2.997575,
        -0.314159,0.599643,-3.045235,
        -0.314159,0.599962,2.773556,
        -0.314159,0.602620,0.029270,
        -0.314159,0.605711,0.598454,
        -0.314159,0.607708,-1.541026,
        -0.314159,0.609360,-2.464379,
        -0.314159,0.610738,-1.326786,
        -0.314159,0.616968,0.820489,
        -0.314159,0.626379,1.773610,
        -0.314159,0.628626,-1.751180,
        -0.314159,0.630671,0.391949,
        -0.314159,0.632622,-0.737642,
        -0.314159,0.632708,2.566722,
        -0.314159,0.632797,1.989858,
        -0.314159,0.634455,-0.534790,
        -0.314159,0.636205,1.561170,
        -0.314159,0.637681,-1.123423,
        -0.314159,0.641441,-2.842109,
        -0.314159,0.644807,-0.327139,
        -0.314159,0.652440,2.197129,
        -0.314159,0.660897,-2.278593,
        -0.314159,0.664008,1.360091,
        -0.314159,0.670850,-1.945257,
        -0.314159,0.671396,1.004859,
        -0.314159,0.673985,0.203855,
        -0.314159,0.680326,-0.133370,
        -0.314159,0.685083,-0.942363,
        -0.314159,0.688204,-2.632204,
        -0.314159,0.689777,2.385011,
        -0.314159,0.700065,3.115430,
        -0.314159,0.704080,2.911636,
        -0.314159,0.712883,1.174216,
        -0.314159,0.715701,0.699906,
        -0.314159,0.718851,-1.442215,
        -0.314159,0.721019,2.721536,
        -0.314159,0.725300,-2.975134,
        -0.314159,0.727136,-1.633288,
        -0.314159,0.727399,-2.120567,
        -0.314159,0.730751,0.522488,
        -0.314159,0.732892,-1.254330,
        -0.314159,0.734333,-2.446184,
        -0.314159,0.735125,0.037270,
        -0.314159,0.737893,1.866463,
        -0.314159,0.740650,1.681075,
        -0.314159,0.745971,-0.781286,
        -0.314159,0.749432,-0.442559,
        -0.314159,0.750102,2.050636,
        -0.314159,0.751046,0.863821,
        -0.314159,0.757018,-0.608833,
        -0.314159,0.757046,-1.816019,
        -0.314159,0.757389,1.501184,
        -0.314159,0.759800,2.546181,
        -0.314159,0.759859,0.351923,
        -0.314159,0.761141,-2.781102,
        -0.314159,0.769035,-1.078574,
        -0.314159,0.770166,-0.266604,
        -0.314159,0.772218,2.227380,
        -0.314159,0.788382,1.331051,
        -0.314159,0.790269,-2.282409,
        -0.314159,0.803410,-1.984567,
        -0.314159,0.803813,0.191520,
        -0.314159,0.806832,3.022155,
        -0.314159,0.807476,-2.597675,
        -0.314159,0.812512,-0.107335,
        -0.314159,0.814161,1.010949,
        -0.314159,0.817732,-3.084486,
        -0.314159,0.817846,2.390688,
        -0.314159,0.819323,2.845960,
        -0.314159,0.819430,-0.915017,
        -0.314159,0.829198,-1.532491,
        -0.314159,0.829531,0.665695,
        -0.314159,0.829816,1.167147,
        -0.314159,0.833613,-1.357492,
        -0.314159,0.842945,-2.907318,
        -0.314159,0.844883,2.679708,
        -0.314159,0.848189,-1.705003,
        -0.314159,0.849888,1.767901,
        -0.314159,0.851126,0.493370,
        -0.314159,0.852911,1.933394,
        -0.314159,0.857516,-2.432863,
        -0.314159,0.858261,-1.191729,
        -0.314159,0.858279,-2.138895,
        -0.314159,0.859723,-0.734928,
        -0.314159,0.859754,1.603292,
        -0.314159,0.861655,-0.394821,
        -0.314159,0.864300,0.043549,
        -0.314159,0.866940,0.851229,
        -0.314159,0.869748,2.095969,
        -0.314159,0.877352,-0.574668,
        -0.314159,0.879254,-2.733338,
        -0.314159,0.882554,1.443757,
        -0.314159,0.884060,0.334199,
        -0.314159,0.885626,-1.866061,
        -0.314159,0.889834,2.529957,
        -0.314159,0.893140,2.251698,
        -0.314159,0.897463,-0.240822,
        -0.314159,0.898758,-1.031948,
        -0.314159,0.913918,1.287618,
        -0.314159,0.916555,-2.287173,
        -0.314159,0.918305,3.107845,
        -0.314159,0.919522,2.948277,
        -0.314159,0.927138,-2.569686,
        -0.314159,0.929031,0.185383,
        -0.314159,0.933766,-3.015917,
        -0.314159,0.934873,-2.010822,
        -0.314159,0.935636,-1.445855,
        -0.314159,0.936702,-0.866461,
        -0.314159,0.937395,0.729108,
        -0.314159,0.937910,2.787575,
        -0.314159,0.939144,-0.092962,
        -0.314159,0.941718,1.125085,
        -0.314159,0.942873,-1.607680,
        -0.314159,0.943701,0.980328,
        -0.314159,0.947774,2.395256,
        -0.314159,0.952433,0.584347,
        -0.314159,0.954456,-1.287293,
        -0.314159,0.958731,-2.852552,
        -0.314159,0.961838,1.839284,
        -0.314159,0.961898,-0.469425,
        -0.314159,0.965426,1.686615,
        -0.314159,0.970108,1.990771,
        -0.314159,0.972820,-1.762780,
        -0.314159,0.973500,-0.711881,
        -0.314159,0.975253,2.641082,
        -0.314159,0.977824,0.441928,
        -0.314159,0.981485,-2.424766,
        -0.314159,0.981877,1.537100,
        -0.314159,0.982609,-1.136920,
        -0.314159,0.984265,-2.156073,
        -0.314159,0.989150,0.045493,
        -0.314159,0.993052,2.137478,
        -0.314159,0.997766,-2.693871,
        -0.314159,0.999847,-0.334738,
        -0.314159,1.001920,0.848054,
        -0.314159,1.006146,1.389734,
        -0.314159,1.012180,0.304066,
        -0.314159,1.016938,2.270674,
        -0.314159,1.017371,-0.979790,
        -0.314159,1.019495,-1.894487,
        -0.314159,1.023796,-0.583691,
        -0.314159,1.026525,2.506945,
        -0.314159,1.027004,-0.203653,
        -0.314159,1.028027,3.036573,
        -0.314159,1.031435,1.241021,
        -0.314159,1.032316,2.884221,
        -0.314159,1.035863,-3.110475,
        -0.314159,1.042085,-2.292950,
        -0.314159,1.043272,-1.522986,
        -0.314159,1.046265,-2.958847,
        -0.314159,1.049272,-1.376344,
        -0.314159,1.050039,-2.552086,
        -0.314159,1.053682,0.689863,
        -0.314159,1.055458,0.170394,
        -0.314159,1.056622,-0.831866,
        -0.314159,1.057515,1.097998,
        -0.314159,1.058706,-2.036438,
        -0.314159,1.061389,-1.672248,
        -0.314159,1.063758,2.743093,
        -0.314159,1.070157,-0.073422,
        -0.314159,1.072154,1.761506,
        -0.314159,1.074029,1.909193,
        -0.314159,1.074204,0.545276,
        -0.314159,1.074714,-2.805147,
        -0.314159,1.076262,0.966104,
        -0.314159,1.077642,-1.227073,
        -0.314159,1.084151,1.618525,
        -0.314159,1.085085,2.377673,
        -0.314159,1.086822,-0.441112,
        -0.314159,1.087142,2.058815,
        -0.314159,1.098383,-0.696430,
        -0.314159,1.099138,-1.084033,
        -0.314159,1.103208,1.477773,
        -0.314159,1.103783,0.408254,
        -0.314159,1.106275,-2.420800,
        -0.314159,1.108079,2.612943,
        -0.314159,1.109345,-2.172718,
        -0.314159,1.115957,-1.787597,
        -0.314159,1.118930,0.045983,
        -0.314159,1.119777,0.834702,
        -0.314159,1.121943,-2.669042,
        -0.314159,1.124930,2.219917,
        -0.314159,1.125217,1.336294,
        -0.314159,1.130958,2.972931,
        -0.314159,1.131400,-0.306629,
        -0.314159,1.137135,-0.936514,
        -0.314159,1.139125,-1.924528,
        -0.314159,1.139877,0.276382,
        -0.314159,1.142003,-3.058022,
        -0.314159,1.147363,-0.558448,
        -0.314159,1.150348,-0.176418,
        -0.314159,1.151198,1.197016,
        -0.314159,1.153930,-1.597104,
        -0.314159,1.155389,2.837148,
        -0.314159,1.155458,-1.460623,
        -0.314159,1.157331,-2.906362,
        -0.314159,1.158594,2.487618,
        -0.314159,1.162103,-1.322553,
        -0.314159,1.165407,3.092861,
        -0.314159,1.168025,0.717075,
        -0.314159,1.168406,-2.301519,
        -0.314159,1.174782,-2.538755,
        -0.314159,1.177998,1.061256,
        -0.314159,1.178898,1.827305,
        -0.314159,1.181331,-2.060330,
        -0.314159,1.182540,-0.798952,
        -0.314159,1.184569,0.147517,
        -0.314159,1.184734,1.968919,
        -0.314159,1.184804,0.593386,
        -0.314159,1.188278,1.690808,
        -0.314159,1.194360,2.110440,
        -0.314159,1.194776,2.710455,
        -0.314159,1.197745,-1.173705,
        -0.314159,1.199898,-2.777257,
        -0.314159,1.201164,2.355931,
        -0.314159,1.203760,1.556404,
        -0.314159,1.205845,-0.415993,
        -0.314159,1.206585,0.933648,
        -0.314159,1.206950,-0.037248,
        -0.314159,1.209341,0.471162,
        -0.314159,1.213483,-1.039417,
        -0.314159,1.218146,-0.672643,
        -0.314159,1.221608,1.421669,
        -0.314159,1.226043,-1.693533,
        -0.314159,1.232870,-1.822773,
        -0.314159,1.234375,-2.421194,
        -0.314159,1.234531,-2.190148,
        -0.314159,1.239414,0.348959,
        -0.314159,1.240183,2.587255,
        -0.314159,1.242312,2.941234,
        -0.314159,1.244075,1.286644,
        -0.314159,1.245961,-2.997461,
        -0.314159,1.246495,-2.646872,
        -0.314159,1.247736,2.242810,
        -0.314159,1.253604,0.812222,
        -0.314159,1.257816,-0.894317,
        -0.314159,1.258247,-1.953634,
        -0.314159,1.258993,-0.284099,
        -0.314159,1.262380,-0.157609,
        -0.314159,1.262791,-1.549229,
        -0.314159,1.263772,-0.533073,
        -0.314159,1.266731,-1.412461,
        -0.314159,1.268591,-3.122146,
        -0.314159,1.269614,-1.275246,
        -0.314159,1.270745,1.152533,
        -0.314159,1.273606,0.224545,
        -0.314159,1.279716,2.812101,
        -0.314159,1.283786,2.462124,
        -0.314159,1.284798,0.078209,
        -0.314159,1.285402,-2.882049,
        -0.314159,1.286936,1.886053,
        -0.314159,1.289408,0.683122,
        -0.314159,1.294736,1.756184,
        -0.314159,1.296978,-2.311511,
        -0.314159,1.298546,2.020601,
        -0.314159,1.300924,1.022478,
        -0.314159,1.303305,3.043152,
        -0.314159,1.304210,-2.084225,
        -0.314159,1.305241,-2.530249,
        -0.314159,1.307513,1.628003,
        -0.314159,1.310722,0.558267,
        -0.314159,1.314142,2.139329,
        -0.314159,1.316087,-0.777971,
        -0.314159,1.317907,-1.121023,
        -0.314159,1.321019,-2.749776,
        -0.314159,1.321609,2.687717,
        -0.314159,1.322568,1.499415,
        -0.314159,1.322972,-0.395699,
        -0.314159,1.325286,-0.994192,
        -0.314159,1.329146,-0.648513,
        -0.314159,1.329795,-0.042570,
        -0.314159,1.331651,2.344497,
        -0.314159,1.336522,0.435098,
        -0.314159,1.339410,0.900087,
        -0.314159,1.339550,1.370353,
        -0.314159,1.340441,-1.731580,
        -0.314159,1.351855,-1.855234,
        -0.314159,1.360161,-3.018404,
        -0.314159,1.361031,-1.613971,
        -0.314159,1.361196,2.914780,
        -0.314159,1.361998,-2.207254,
        -0.314159,1.362302,1.239384,
        -0.314159,1.365228,2.563808,
        -0.314159,1.365452,0.310522,
        -0.314159,1.367028,-2.420041,
        -0.314159,1.373422,-1.484281,
        -0.314159,1.374245,-1.226417,
        -0.314159,1.375020,-0.170393,
        -0.314159,1.376560,-1.355481,
        -0.314159,1.377070,-1.980984,
        -0.314159,1.377578,-2.636339,
        -0.314159,1.378096,0.774626,
        -0.314159,1.379086,-0.508192,
        -0.314159,1.382562,0.177141,
        -0.314159,1.388849,2.233847,
        -0.314159,1.389268,-0.886349,
        -0.314159,1.391324,1.109396,
        -0.314159,1.391699,-0.288725,
        -0.314159,1.392869,3.136903,
        -0.314159,1.393612,1.952016,
        -0.314159,1.395545,-2.850807,
        -0.314159,1.401926,1.822930,
        -0.314159,1.402149,2.788898,
        -0.314159,1.409552,0.647980,
        -0.314159,1.410840,0.056219,
        -0.314159,1.411944,2.444377,
        -0.314159,1.413639,1.697982,
        -0.314159,1.426479,1.573427,
        -0.314159,1.426725,2.102242,
        -0.314159,1.427868,0.984637,
        -0.314159,1.429580,-2.107232,
        -0.314159,1.431375,-2.313269,
        -0.314159,1.431907,3.015410,
        -0.314159,1.432325,-1.029054,
        -0.314159,1.433341,0.524268,
        -0.314159,1.435181,-0.739771,
        -0.314159,1.437208,-2.529085,
        -0.314159,1.439041,-0.617966,
        -0.314159,1.441473,1.448544,
        -0.314159,1.444527,2.664744,
        -0.314159,1.450513,-0.400625,
        -0.314159,1.451907,-0.066823,
        -0.314159,1.452754,-2.738631,
        -0.314159,1.454061,-2.950416,
        -0.314159,1.456736,1.323154,
        -0.314159,1.458370,0.400433,
        -0.314159,1.459314,-1.765003,
        -0.314159,1.459902,-1.145835,
        -0.314159,1.464952,2.329245,
        -0.314159,1.466050,0.862104,
        -0.314159,1.471330,-1.886160,
        -0.314159,1.475119,-1.648255,
        -0.314159,1.480878,1.194576,
        -0.314159,1.480942,0.273235,
        -0.314159,1.481086,2.890269,
        -0.314159,1.484118,-1.529156,
        -0.314159,1.484741,-1.279219,
        -0.314159,1.487681,-1.406069,
        -0.314159,1.487886,-3.066037,
        -0.314159,1.490555,2.543284,
        -0.314159,1.492424,-0.190932,
        -0.314159,1.495002,-0.839810,
        -0.314159,1.495297,2.007188,
        -0.314159,1.496414,-2.009267,
        -0.314159,1.498261,-2.425270,
        -0.314159,1.498907,-2.207327,
        -0.314159,1.500204,0.737638,
        -0.314159,1.502469,0.151156,
        -0.314159,1.507448,-0.514782,
        -0.314159,1.509287,2.204647,
        -0.314159,1.510518,1.884920,
        -0.314159,1.511948,-2.632360,
        -0.314159,1.515235,1.068565,
        -0.314159,1.522237,3.102018,
        -0.314159,1.522373,1.764991,
        -0.314159,1.522416,-0.954956,
        -0.314159,1.523102,2.765920,
        -0.314159,1.524811,-2.844718,
        -0.314159,1.527261,-0.304916,
        -0.314159,1.529788,0.613208,
        -0.314159,1.533648,1.644808,
        -0.314159,1.535018,0.028869,
        -0.314159,1.541082,2.424608,
        -0.314159,1.546010,1.523462,
        -0.314159,1.551135,-0.642994,
        -0.314159,1.553895,0.489607,
        -0.314159,1.555041,0.947813,
        -0.314159,1.559671,-1.074934,
        -0.314159,1.560345,2.985899,
        -0.314159,1.560487,-2.323400,
        -0.314159,1.561127,1.401111,
        -0.314159,1.567193,2.643172,
        -0.314159,1.568892,2.098362,
        -0.314159,1.571702,-2.098435,
        -0.314159,1.572701,-2.528730,
        -0.314159,1.573572,1.278593,
        -0.314159,1.574849,-0.091505,
        -0.314159,1.576334,0.365236,
        -0.314159,1.577029,-2.962888,
        -0.314159,1.577688,-1.814941,
        -0.314159,1.579542,-0.755917,
        -0.314159,1.579624,-1.199759,
        -0.314159,1.581515,-0.419414,
        -0.314159,1.584581,-2.738702,
        -0.314159,1.586948,-1.696296,
        -0.314159,1.588901,2.301496,
        -0.314159,1.589936,0.825020,
        -0.314159,1.590018,-1.941652,
        -0.314159,1.596968,-1.328358,
        -0.314159,1.597178,-1.575607,
        -0.314159,1.600470,-1.453196,
        -0.314159,1.601408,0.242409,
        -0.314159,1.602259,1.152656,
        -0.314159,1.603064,2.865774,
        -0.314159,1.612085,1.968802,
        -0.314159,1.612720,-3.080498,
        -0.314159,1.616061,2.521531,
        -0.314159,1.616900,-0.207457,
        -0.314159,1.620988,-0.875920,
        -0.314159,1.622114,0.701509,
        -0.314159,1.622749,0.121426,
        -0.314159,1.622993,-2.221629,
        -0.314159,1.629346,-0.542669,
        -0.314159,1.630007,1.844552,
        -0.314159,1.634747,-2.426198,
        -0.314159,1.642346,1.722617,
        -0.314159,1.643925,1.034629,
        -0.314159,1.644848,2.743774,
        -0.314159,1.645258,2.188109,
        -0.314159,1.645417,-2.634725,
        -0.314159,1.647395,3.085437,
        -0.314159,1.648538,-2.855044,
        -0.314159,1.649932,0.576808,
        -0.314159,1.652224,-0.996105,
        -0.314159,1.652871,1.600681,
        -0.314159,1.657997,-0.325127,
        -0.314159,1.661936,0.001037,
        -0.314159,1.665187,1.477402,
        -0.314159,1.665496,2.397872,
        -0.314159,1.673267,-0.665453,
        -0.314159,1.673557,0.452526,
        -0.314159,1.679277,-1.122421,
        -0.314159,1.679696,0.912482,
        -0.314159,1.682420,1.353922,
        -0.314159,1.685031,-2.009038,
        -0.314159,1.685196,2.967046,
        -0.314159,1.690107,2.620910,
        -0.314159,1.690453,-2.123076,
        -0.314159,1.691693,-1.871572,
        -0.314159,1.691817,2.061688,
        -0.314159,1.692644,1.237670,
        -0.314159,1.696386,0.328123,
        -0.314159,1.697162,-1.249026,
        -0.314159,1.697488,-2.322785,
        -0.314159,1.698726,-2.973576,
        -0.314159,1.699172,-1.749717,
        -0.314159,1.704478,-0.106295,
        -0.314159,1.707244,-0.444770,
        -0.314159,1.707500,-2.531497,
        -0.314159,1.707714,-0.787321,
        -0.314159,1.708430,-1.627552,
        -0.314159,1.710583,-1.376890,
        -0.314159,1.712415,-2.747921,
        -0.314159,1.713125,-1.503073,
        -0.314159,1.713452,0.789419,
        -0.314159,1.720612,2.279888,
        -0.314159,1.724852,0.202831,
        -0.314159,1.724868,2.846246,
        -0.314159,1.728940,1.931535,
        -0.314159,1.732838,1.127584,
        -0.314159,1.735719,-3.092485,
        -0.314159,1.739676,2.496017,
        -0.314159,1.739971,-0.228876,
        -0.314159,1.743328,0.663347,
        -0.314159,1.744927,-0.911892,
        -0.314159,1.745952,0.084733,
        -0.314159,1.748077,1.804354,
        -0.314159,1.752959,-0.567548,
        -0.314159,1.760169,1.678132,
        -0.314159,1.763887,-2.218450,
        -0.314159,1.766737,2.722752,
        -0.314159,1.767999,1.004737,
        -0.314159,1.769149,1.550298,
        -0.314159,1.770119,-2.427175,
        -0.314159,1.770460,2.154208,
        -0.314159,1.770514,3.070982,
        -0.314159,1.770735,0.537245,
        -0.314159,1.772673,-2.864760,
        -0.314159,1.773527,-1.039582,
        -0.314159,1.776419,-2.642421,
        -0.314159,1.780769,1.416330,
        -0.314159,1.786634,-0.346247,
        -0.314159,1.788829,-1.947553,
        -0.314159,1.793737,0.410270,
        -0.314159,1.794553,2.374618,
        -0.314159,1.794699,-0.694061,
        -0.314159,1.797469,-1.169105,
        -0.314159,1.803705,0.881076,
        -0.314159,1.804124,1.264249,
        -0.314159,1.807825,-2.090455,
        -0.314159,1.808822,-1.817375,
        -0.314159,1.809020,2.949741,
        -0.314159,1.812466,2.597595,
        -0.314159,1.813159,2.025395,
        -0.314159,1.813816,-1.298435,
        -0.314159,1.815016,0.278951,
        -0.314159,1.816510,-0.008631,
        -0.314159,1.817772,-1.689625,
        -0.314159,1.821360,-2.984352,
        -0.314159,1.822312,-0.127211,
        -0.314159,1.823386,-1.429807,
        -0.314159,1.823834,-1.560483,
        -0.314159,1.832060,-0.821627,
        -0.314159,1.832690,-0.468440,
        -0.314159,1.833945,-2.320278,
        -0.314159,1.834787,0.751968,
        -0.314159,1.840579,-2.536445,
        -0.314159,1.841033,-2.756932,
        -0.314159,1.846350,1.893983,
        -0.314159,1.846644,1.113766,
        -0.314159,1.846833,2.826543,
        -0.314159,1.847325,2.249051,
        -0.314159,1.854119,0.134161,
        -0.314159,1.858863,-3.106022,
        -0.314159,1.865046,1.763384,
        -0.314159,1.865619,-0.952036,
        -0.314159,1.865912,0.624453,
        -0.314159,1.866694,2.473686,
        -0.314159,1.867524,-0.240083,
        -0.314159,1.875561,1.632831,
        -0.314159,1.876049,-0.594378,
        -0.314159,1.880509,1.500204,
        -0.314159,1.883504,1.360457,
        -0.314159,1.886971,-2.196665,
        -0.314159,1.888979,0.982922,
        -0.314159,1.889815,2.701570,
        -0.314159,1.892144,-2.005340,
        -0.314159,1.892348,0.496522,
        -0.314159,1.893281,-1.083976,
        -0.314159,1.893828,3.053453,
        -0.314159,1.895117,2.121107,
        -0.314159,1.898741,-2.875735,
        -0.314159,1.904429,-2.428029,
        -0.314159,1.908408,-2.648903,
        -0.314159,1.911828,-0.364489,
        -0.314159,1.912224,0.366565,
        -0.314159,1.914854,1.214468,
        -0.314159,1.915467,-1.883143,
        -0.314159,1.915763,-0.724679,
        -0.314159,1.915790,-1.216366,
        -0.314159,1.922256,2.347806,
        -0.314159,1.922616,0.844633,
        -0.314159,1.925185,0.231170,
        -0.314159,1.927474,-1.754616,
        -0.314159,1.928342,-1.351174,
        -0.314159,1.933328,-1.623328,
        -0.314159,1.933878,0.034150,
        -0.314159,1.934585,-1.487979,
        -0.314159,1.934802,2.927634,
        -0.314159,1.934862,1.989776,
        -0.314159,1.936327,-0.099350,
        -0.314159,1.938088,2.578507,
        -0.314159,1.945677,-2.999805,
        -0.314159,1.954572,-0.859208,
        -0.314159,1.957868,-0.490784,
        -0.314159,1.958989,0.712772,
        -0.314159,1.960212,-2.307561,
        -0.314159,1.964774,1.855507,
        -0.314159,1.964971,-2.100533,
        -0.314159,1.969112,1.088273,
        -0.314159,1.971438,-2.765415,
        -0.314159,1.972312,2.805134,
        -0.314159,1.974950,2.220596,
        -0.314159,1.975135,-2.539127,
        -0.314159,1.982107,1.720829,
        -0.314159,1.983195,-3.127984,
        -0.314159,1.983318,-0.247182,
        -0.314159,1.986720,-0.993220,
        -0.314159,1.990381,1.586005,
        -0.314159,1.990470,0.580283,
        -0.314159,1.990523,1.450326,
        -0.314159,1.993909,1.312897,
        -0.314159,1.994694,2.452813,
        -0.314159,1.999226,-0.620717,
        -0.314159,2.005145,0.941012,
        -0.314159,2.007301,0.142337,
        -0.314159,2.014149,0.444138,
        -0.314159,2.014338,-1.127392,
        -0.314159,2.018131,3.021778,
        -0.314159,2.019676,-1.980944,
        -0.314159,2.019912,2.686612,
        -0.314159,2.020545,2.088672,
        -0.314159,2.026074,0.300945,
        -0.314159,2.026196,-2.889057,
        -0.314159,2.031959,-1.268178,
        -0.314159,2.032917,-2.199917,
        -0.314159,2.033572,-1.842884,
        -0.314159,2.034712,-2.419505,
        -0.314159,2.035818,-0.379479,
        -0.314159,2.036004,-0.757062,
        -0.314159,2.040501,-1.411488,
        -0.314159,2.042874,-1.701367,
        -0.314159,2.043060,-2.653746,
        -0.314159,2.043729,-1.557938,
        -0.314159,2.048538,0.806623,
        -0.314159,2.049006,1.199930,
        -0.314159,2.049502,-0.137989,
        -0.314159,2.052181,2.326959,
        -0.314159,2.052794,-0.008526,
        -0.314159,2.057608,1.951726,
        -0.314159,2.065568,2.567730,
        -0.314159,2.070371,2.879808,
        -0.314159,2.071256,-3.018961,
        -0.314159,2.079746,-0.897699,
        -0.314159,2.082784,1.042403,
        -0.314159,2.083752,-0.511086,
        -0.314159,2.084401,1.810502,
        -0.314159,2.085731,0.671081,
        -0.314159,2.097496,1.373172,
        -0.314159,2.099185,1.667006,
        -0.314159,2.100482,-2.301084,
        -0.314159,2.103118,2.194745,
        -0.314159,2.103223,1.521685,
        -0.314159,2.103695,-2.777707,
        -0.314159,2.107523,3.129581,
        -0.314159,2.108277,-2.095907,
        -0.314159,2.108891,-2.532166,
        -0.314159,2.111003,-0.268632,
        -0.314159,2.112276,-1.027246,
        -0.314159,2.114380,0.531457,
        -0.314159,2.119166,0.100552,
        -0.314159,2.119564,2.754818,
        -0.314159,2.122474,-0.648314,
        -0.314159,2.125183,2.445598,
        -0.314159,2.129501,0.238290,
        -0.314159,2.131178,0.386663,
        -0.314159,2.131223,-1.174311,
        -0.314159,2.134269,-1.948042,
        -0.314159,2.134684,0.907065,
        -0.314159,2.137615,2.988806,
        -0.314159,2.145635,-1.327708,
        -0.314159,2.146467,2.055081,
        -0.314159,2.146582,-1.793788,
        -0.314159,2.149210,-1.482321,
        -0.314159,2.152901,-1.636788,
        -0.314159,2.156337,1.151558,
        -0.314159,2.156474,-2.912065,
        -0.314159,2.157194,-0.792563,
        -0.314159,2.166724,-0.401638,
        -0.314159,2.166980,-0.134879,
        -0.314159,2.173644,-2.403429,
        -0.314159,2.175711,-2.655523,
        -0.314159,2.177132,2.600419,
        -0.314159,2.178879,0.767722,
        -0.314159,2.180752,1.908485,
        -0.314159,2.181338,2.313721,
        -0.314159,2.185580,1.287673,
        -0.314159,2.195400,-2.217546,
        -0.314159,2.200418,-3.056442,
        -0.314159,2.201829,-0.001735,
        -0.314159,2.203804,1.756176,
        -0.314159,2.211602,1.443930,
        -0.314159,2.212686,-0.544809,
        -0.314159,2.213937,0.621892,
        -0.314159,2.214075,1.600691,
        -0.314159,2.216256,1.020029,
        -0.314159,2.219060,2.886631,
        -0.314159,2.223375,-0.906185,
        -0.314159,2.226002,-1.060248,
        -0.314159,2.226413,-2.064034,
        -0.314159,2.230554,2.171334,
        -0.314159,2.232413,2.735777,
        -0.314159,2.233709,3.070830,
        -0.314159,2.234383,-2.792065,
        -0.314159,2.237183,0.468799,
        -0.314159,2.237695,-0.271388,
        -0.314159,2.239419,0.151693,
        -0.314159,2.243347,-2.521049,
        -0.314159,2.244035,-1.229456,
        -0.314159,2.244355,0.309899,
        -0.314159,2.245175,2.460236,
        -0.314159,2.245628,-1.899216,
        -0.314159,2.245973,-0.702234,
        -0.314159,2.255601,-1.728131,
        -0.314159,2.256702,-1.401367,
        -0.314159,2.261466,-1.557500,
        -0.314159,2.267390,0.878499,
        -0.314159,2.272099,2.018594,
        -0.314159,2.278618,1.170386,
        -0.314159,2.283304,-2.345584,
        -0.314159,2.285175,-2.943090,
        -0.314159,2.293000,-0.124451,
        -0.314159,2.294058,-0.421070,
        -0.314159,2.303283,1.855938,
        -0.314159,2.304968,2.608978,
        -0.314159,2.305347,2.311346,
        -0.314159,2.307188,-2.656134,
        -0.314159,2.307301,1.336200,
        -0.314159,2.309655,0.726020,
        -0.314159,2.317634,-2.185225,
        -0.314159,2.321510,1.685664,
        -0.314159,2.324376,-3.111006,
        -0.314159,2.324731,1.511186,
        -0.314159,2.329587,0.035498,
        -0.314159,2.330158,2.977264,
        -0.314159,2.335096,-0.588468,
        -0.314159,2.336046,-1.115799,
        -0.314159,2.338060,2.803613,
        -0.314159,2.338083,-0.943400,
        -0.314159,2.340210,0.562145,
        -0.314159,2.341931,1.023926,
        -0.314159,2.343265,-0.779061,
        -0.314159,2.343434,-2.013692,
        -0.314159,2.353029,-1.310529,
        -0.314159,2.354868,0.211543,
        -0.314159,2.355120,0.388593,
        -0.314159,2.355165,-1.833173,
        -0.314159,2.355462,2.151441,
        -0.314159,2.362948,-2.487477,
        -0.314159,2.363101,-2.810456,
        -0.314159,2.363394,-0.272918,
        -0.314159,2.364509,-1.644533,
        -0.314159,2.370465,2.466824,
        -0.314159,2.386110,-1.481351,
        -0.314159,2.394960,0.862133,
        -0.314159,2.395716,1.977184,
        -0.314159,2.396339,1.200900,
        -0.314159,2.407941,-2.316753,
        -0.314159,2.409450,-2.988432,
        -0.314159,2.417202,-0.103363,
        -0.314159,2.417386,-0.443097,
        -0.314159,2.417950,2.666662,
        -0.314159,2.423190,1.789783,
        -0.314159,2.423692,1.392696,
        -0.314159,2.431122,3.086379,
        -0.314159,2.431406,-2.647820,
        -0.314159,2.431584,2.307631,
        -0.314159,2.434090,1.592818,
        -0.314159,2.435993,0.683844,
        -0.314159,2.440898,-2.133300,
        -0.314159,2.442222,-1.195837,
        -0.314159,2.443112,2.883761,
        -0.314159,2.449191,-0.633651,
        -0.314159,2.450821,0.084741,
        -0.314159,2.453499,-1.008861,
        -0.314159,2.460010,-1.938011,
        -0.314159,2.460199,-0.822008,
        -0.314159,2.460902,0.490079,
        -0.314159,2.461774,-1.734913,
        -0.314159,2.463986,1.034407,
        -0.314159,2.467647,0.287667,
        -0.314159,2.479849,2.126925,
        -0.314159,2.485907,-2.835471,
        -0.314159,2.485945,-1.374853,
        -0.314159,2.487747,2.505912,
        -0.314159,2.487957,-0.274279,
        -0.314159,2.489741,-2.468428,
        -0.314159,2.512538,1.239900,
        -0.314159,2.517528,1.923289,
        -0.314159,2.518454,-1.565852,
        -0.314159,2.519348,0.843503,
        -0.314159,2.522766,-3.059231,
        -0.314159,2.528686,2.737228,
        -0.314159,2.534798,-2.266671,
        -0.314159,2.536533,1.466274,
        -0.314159,2.538504,1.699812,
        -0.314159,2.539516,-0.476942,
        -0.314159,2.539532,-0.074548,
        -0.314159,2.543197,2.990252,
        -0.314159,2.554022,2.320619,
        -0.314159,2.556919,-2.655005,
        -0.314159,2.558146,0.627191,
        -0.314159,2.558223,-1.164189,
        -0.314159,2.563452,-2.044028,
        -0.314159,2.568462,0.150655,
        -0.314159,2.569130,-0.700837,
        -0.314159,2.570050,-1.804492,
        -0.314159,2.572710,-0.934665,
        -0.314159,2.575994,0.391718,
        -0.314159,2.584774,1.046703,
        -0.314159,2.601432,2.102535,
        -0.314159,2.603559,2.555114,
        -0.314159,2.604150,-2.885208,
        -0.314159,2.608630,-1.378011,
        -0.314159,2.611256,-0.279110,
        -0.314159,2.617344,-2.439480,
        -0.314159,2.627402,1.293552,
        -0.314159,2.635396,1.848850,
        -0.314159,2.635686,2.834950,
        -0.314159,2.637087,3.132730,
        -0.314159,2.642029,0.813274,
        -0.314159,2.645260,1.570203,
        -0.314159,2.646182,-1.614036,
        -0.314159,2.660113,-2.186513,
        -0.314159,2.660483,-0.034608,
        -0.314159,2.661713,-0.529511,
        -0.314159,2.672720,-1.110422,
        -0.314159,2.676626,2.335353,
        -0.314159,2.676638,0.540364,
        -0.314159,2.676780,-1.901158,
        -0.314159,2.681411,-2.667965,
        -0.314159,2.682152,0.245751,
        -0.314159,2.682859,-0.817071,
        -0.314159,2.705107,1.062142,
        -0.314159,2.717705,2.625995,
        -0.314159,2.721001,2.060097,
        -0.314159,2.722319,-2.954069,
        -0.314159,2.729840,-1.368237,
        -0.314159,2.734349,-0.289298,
        -0.314159,2.738409,2.975671,
        -0.314159,2.740366,1.374341,
        -0.314159,2.744633,-2.391311,
        -0.314159,2.747398,1.727166,
        -0.314159,2.763799,0.760372,
        -0.314159,2.768238,-1.679243,
        -0.314159,2.779245,-2.046268,
        -0.314159,2.780208,-0.626453,
        -0.314159,2.780471,0.027732,
        -0.314159,2.784696,-1.013128,
        -0.314159,2.789817,0.395431,
        -0.314159,2.801154,2.353143,
        -0.314159,2.807091,-2.687477,
        -0.314159,2.825762,1.085035,
        -0.314159,2.830209,2.734943,
        -0.314159,2.833517,-3.085508,
        -0.314159,2.838301,1.980061,
        -0.314159,2.849871,1.512012,
        -0.314159,2.850089,-1.343899,
        -0.314159,2.856905,-0.318441,
        -0.314159,2.870445,-2.291001,
        -0.314159,2.885017,0.655562,
        -0.314159,2.886428,-1.793217,
        -0.314159,2.890337,-0.825633,
        -0.314159,2.898932,0.139973,
        -0.314159,2.931927,-2.728768,
        -0.314159,2.932096,2.377790,
        -0.314159,2.937409,2.947245,
        -0.314159,2.948048,1.133127,
        -0.314159,2.950896,1.797889,
        -0.314159,2.970019,-1.282382,
        -0.314159,2.978333,-0.412090,
        -0.314159,2.996437,-2.089121,
        -0.314159,3.011274,0.434110,
        -0.314159,3.050410,3.138492,
        -0.314159,3.064492,1.647259,
        -0.314159,3.088868,-0.989501,
        -0.209440,0.053884,-2.408394,
        -0.209440,0.077645,1.453433,
        -0.209440,0.088408,-0.311160,
        -0.209440,0.142314,2.590387,
        -0.209440,0.147206,-1.385164,
        -0.209440,0.163506,0.573203,
        -0.209440,0.165485,-2.875396,
        -0.209440,0.182564,-2.125821,
        -0.209440,0.190569,1.929141,
        -0.209440,0.201102,1.249082,
        -0.209440,0.201397,-0.716646,
        -0.209440,0.210568,-0.069559,
        -0.209440,0.251672,2.848900,
        -0.209440,0.266262,-1.678068,
        -0.209440,0.269108,-1.215563,
        -0.209440,0.273777,-2.506756,
        -0.209440,0.279854,0.814581,
        -0.209440,0.281664,-2.969147,
        -0.209440,0.283625,2.405603,
        -0.209440,0.284058,0.357350,
        -0.209440,0.294107,1.593880,
        -0.209440,0.302491,-0.400410,
        -0.209440,0.308059,2.021491,
        -0.209440,0.312219,-2.072057,
        -0.209440,0.321905,-0.829121,
        -0.209440,0.329984,1.201621,
        -0.209440,0.339714,-0.013712,
        -0.209440,0.368784,2.974974,
        -0.209440,0.376428,-1.495532,
        -0.209440,0.378103,2.646436,
        -0.209440,0.385088,0.590355,
        -0.209440,0.386248,-2.690943,
        -0.209440,0.389851,-2.340031,
        -0.209440,0.391502,-1.171374,
        -0.209440,0.398249,-1.812694,
        -0.209440,0.400542,-2.995729,
        -0.209440,0.404273,0.918163,
        -0.209440,0.405617,1.773533,
        -0.209440,0.409009,-0.570243,
        -0.209440,0.410700,0.270433,
        -0.209440,0.415893,1.464890,
        -0.209440,0.420680,2.088164,
        -0.209440,0.424766,-0.258729,
        -0.209440,0.426507,2.367872,
        -0.209440,0.441292,-0.890727,
        -0.209440,0.451839,-2.078193,
        -0.209440,0.459270,1.191562,
        -0.209440,0.470622,0.014002,
        -0.209440,0.481292,-2.529098,
        -0.209440,0.482460,2.849843,
        -0.209440,0.492205,-1.397963,
        -0.209440,0.493314,3.117264,
        -0.209440,0.497870,0.727828,
        -0.209440,0.499783,-1.654379,
        -0.209440,0.502651,0.463333,
        -0.209440,0.505463,2.594557,
        -0.209440,0.506104,-2.874910,
        -0.209440,0.513579,-1.147354,
        -0.209440,0.518569,1.644237,
        -0.209440,0.519001,1.900488,
        -0.209440,0.519277,-0.673809,
        -0.209440,0.524769,-0.421955,
        -0.209440,0.525564,-2.288838,
        -0.209440,0.532541,0.978737,
        -0.209440,0.534610,2.154808,
        -0.209440,0.535020,-1.891846,
        -0.209440,0.540208,1.400497,
        -0.209440,0.542006,0.228337,
        -0.209440,0.550762,-0.180771,
        -0.209440,0.560531,2.378474,
        -0.209440,0.561629,-0.924576,
        -0.209440,0.567465,-2.679392,
        -0.209440,0.588527,1.181755,
        -0.209440,0.591643,-2.098364,
        -0.209440,0.594963,2.997575,
        -0.209440,0.599643,-3.045235,
        -0.209440,0.599962,2.773556,
        -0.209440,0.602620,0.029270,
        -0.209440,0.605711,0.598454,
        -0.209440,0.607708,-1.541026,
        -0.209440,0.609360,-2.464379,
        -0.209440,0.610738,-1.326786,
        -0.209440,0.616968,0.820489,
        -0.209440,0.626379,1.773610,
        -0.209440,0.628626,-1.751180,
        -0.209440,0.630671,0.391949,
        -0.209440,0.632622,-0.737642,
        -0.209440,0.632708,2.566722,
        -0.209440,0.632797,1.989858,
        -0.209440,0.634455,-0.534790,
        -0.209440,0.636205,1.561170,
        -0.209440,0.637681,-1.123423,
        -0.209440,0.641441,-2.842109,
        -0.209440,0.644807,-0.327139,
        -0.209440,0.652440,2.197129,
        -0.209440,0.660897,-2.278593,
        -0.209440,0.664008,1.360091,
        -0.209440,0.670850,-1.945257,
        -0.209440,0.671396,1.004859,
        -0.209440,0.673985,0.203855,
        -0.209440,0.680326,-0.133370,
        -0.209440,0.685083,-0.942363,
        -0.209440,0.688204,-2.632204,
        -0.209440,0.689777,2.385011,
        -0.209440,0.700065,3.115430,
        -0.209440,0.704080,2.911636,
        -0.209440,0.712883,1.174216,
        -0.209440,0.715701,0.699906,
        -0.209440,0.718851,-1.442215,
        -0.209440,0.721019,2.721536,
        -0.209440,0.725300,-2.975134,
        -0.209440,0.727136,-1.633288,
        -0.209440,0.727399,-2.120567,
        -0.209440,0.730751,0.522488,
        -0.209440,0.732892,-1.254330,
        -0.209440,0.734333,-2.446184,
        -0.209440,0.735125,0.037270,
        -0.209440,0.737893,1.866463,
        -0.209440,0.740650,1.681075,
        -0.209440,0.745971,-0.781286,
        -0.209440,0.749432,-0.442559,
        -0.209440,0.750102,2.050636,
        -0.209440,0.751046,0.863821,
        -0.209440,0.757018,-0.608833,
        -0.209440,0.757046,-1.816019,
        -0.209440,0.757389,1.501184,
        -0.209440,0.759800,2.546181,
        -0.209440,0.759859,0.351923,
        -0.209440,0.761141,-2.781102,
        -0.209440,0.769035,-1.078574,
        -0.209440,0.770166,-0.266604,
        -0.209440,0.772218,2.227380,
        -0.209440,0.788382,1.331051,
        -0.209440,0.790269,-2.282409,
        -0.209440,0.803410,-1.984567,
        -0.209440,0.803813,0.191520,
        -0.209440,0.806832,3.022155,
        -0.209440,0.807476,-2.597675,
        -0.209440,0.812512,-0.107335,
        -0.209440,0.814161,1.010949,
        -0.209440,0.817732,-3.084486,
        -0.209440,0.817846,2.390688,
        -0.209440,0.819323,2.845960,
        -0.209440,0.819430,-0.915017,
        -0.209440,0.829198,-1.532491,
        -0.209440,0.829531,0.665695,
        -0.209440,0.829816,1.167147,
        -0.209440,0.833613,-1.357492,
        -0.209440,0.842945,-2.907318,
        -0.209440,0.844883,2.679708,
        -0.209440,0.848189,-1.705003,
        -0.209440,0.849888,1.767901,
        -0.209440,0.851126,0.493370,
        -0.209440,0.852911,1.933394,
        -0.209440,0.857516,-2.432863,
        -0.209440,0.858261,-1.191729,
        -0.209440,0.858279,-2.138895,
        -0.209440,0.859723,-0.734928,
        -0.209440,0.859754,1.603292,
        -0.209440,0.861655,-0.394821,
        -0.209440,0.864300,0.043549,
        -0.209440,0.866940,0.851229,
        -0.209440,0.869748,2.095969,
        -0.209440,0.877352,-0.574668,
        -0.209440,0.879254,-2.733338,
        -0.209440,0.882554,1.443757,
        -0.209440,0.884060,0.334199,
        -0.209440,0.885626,-1.866061,
        -0.209440,0.889834,2.529957,
        -0.209440,0.893140,2.251698,
        -0.209440,0.897463,-0.240822,
        -0.209440,0.898758,-1.031948,
        -0.209440,0.913918,1.287618,
        -0.209440,0.916555,-2.287173,
        -0.209440,0.918305,3.107845,
        -0.209440,0.919522,2.948277,
        -0.209440,0.927138,-2.569686,
        -0.209440,0.929031,0.185383,
        -0.209440,0.933766,-3.015917,
        -0.209440,0.934873,-2.010822,
        -0.209440,0.935636,-1.445855,
        -0.209440,0.936702,-0.866461,
        -0.209440,0.937395,0.729108,
        -0.209440,0.937910,2.787575,
        -0.209440,0.939144,-0.092962,
        -0.209440,0.941718,1.125085,
        -0.209440,0.942873,-1.607680,
        -0.209440,0.943701,0.980328,
        -0.209440,0.947774,2.395256,
        -0.209440,0.952433,0.584347,
        -0.209440,0.954456,-1.287293,
        -0.209440,0.958731,-2.852552,
        -0.209440,0.961838,1.839284,
        -0.209440,0.961898,-0.469425,
        -0.209440,0.965426,1.686615,
        -0.209440,0.970108,1.990771,
        -0.209440,0.972820,-1.762780,
        -0.209440,0.973500,-0.711881,
        -0.209440,0.975253,2.641082,
        -0.209440,0.977824,0.441928,
        -0.209440,0.981485,-2.424766,
        -0.209440,0.981877,1.537100,
        -0.209440,0.982609,-1.136920,
        -0.209440,0.984265,-2.156073,
        -0.209440,0.989150,0.045493,
        -0.209440,0.993052,2.137478,
        -0.209440,0.997766,-2.693871,
        -0.209440,0.999847,-0.334738,
        -0.209440,1.001920,0.848054,
        -0.209440,1.006146,1.389734,
        -0.209440,1.012180,0.304066,
        -0.209440,1.016938,2.270674,
        -0.209440,1.017371,-0.979790,
        -0.209440,1.019495,-1.894487,
        -0.209440,1.023796,-0.583691,
        -0.209440,1.026525,2.506945,
        -0.209440,1.027004,-0.203653,
        -0.209440,1.028027,3.036573,
        -0.209440,1.031435,1.241021,
        -0.209440,1.032316,2.884221,
        -0.209440,1.035863,-3.110475,
        -0.209440,1.042085,-2.292950,
        -0.209440,1.043272,-1.522986,
        -0.209440,1.046265,-2.958847,
        -0.209440,1.049272,-1.376344,
        -0.209440,1.050039,-2.552086,
        -0.209440,1.053682,0.689863,
        -0.209440,1.055458,0.170394,
        -0.209440,1.056622,-0.831866,
        -0.209440,1.057515,1.097998,
        -0.209440,1.058706,-2.036438,
        -0.209440,1.061389,-1.672248,
        -0.209440,1.063758,2.743093,
        -0.209440,1.070157,-0.073422,
        -0.209440,1.072154,1.761506,
        -0.209440,1.074029,1.909193,
        -0.209440,1.074204,0.545276,
        -0.209440,1.074714,-2.805147,
        -0.209440,1.076262,0.966104,
        -0.209440,1.077642,-1.227073,
        -0.209440,1.084151,1.618525,
        -0.209440,1.085085,2.377673,
        -0.209440,1.086822,-0.441112,
        -0.209440,1.087142,2.058815,
        -0.209440,1.098383,-0.696430,
        -0.209440,1.099138,-1.084033,
        -0.209440,1.103208,1.477773,
        -0.209440,1.103783,0.408254,
        -0.209440,1.106275,-2.420800,
        -0.209440,1.108079,2.612943,
        -0.209440,1.109345,-2.172718,
        -0.209440,1.115957,-1.787597,
        -0.209440,1.118930,0.045983,
        -0.209440,1.119777,0.834702,
        -0.209440,1.121943,-2.669042,
        -0.209440,1.124930,2.219917,
        -0.209440,1.125217,1.336294,
        -0.209440,1.130958,2.972931,
        -0.209440,1.131400,-0.306629,
        -0.209440,1.137135,-0.936514,
        -0.209440,1.139125,-1.924528,
        -0.209440,1.139877,0.276382,
        -0.209440,1.142003,-3.058022,
        -0.209440,1.147363,-0.558448,
        -0.209440,1.150348,-0.176418,
        -0.209440,1.151198,1.197016,
        -0.209440,1.153930,-1.597104,
        -0.209440,1.155389,2.837148,
        -0.209440,1.155458,-1.460623,
        -0.209440,1.157331,-2.906362,
        -0.209440,1.158594,2.487618,
        -0.209440,1.162103,-1.322553,
        -0.209440,1.165407,3.092861,
        -0.209440,1.168025,0.717075,
        -0.209440,1.168406,-2.301519,
        -0.209440,1.174782,-2.538755,
        -0.209440,1.177998,1.061256,
        -0.209440,1.178898,1.827305,
        -0.209440,1.181331,-2.060330,
        -0.209440,1.182540,-0.798952,
        -0.209440,1.184569,0.147517,
        -0.209440,1.184734,1.968919,
        -0.209440,1.184804,0.593386,
        -0.209440,1.188278,1.690808,
        -0.209440,1.194360,2.110440,
        -0.209440,1.194776,2.710455,
        -0.209440,1.197745,-1.173705,
        -0.209440,1.199898,-2.777257,
        -0.209440,1.201164,2.355931,
        -0.209440,1.203760,1.556404,
        -0.209440,1.205845,-0.415993,
        -0.209440,1.206585,0.933648,
        -0.209440,1.206950,-0.037248,
        -0.209440,1.209341,0.471162,
        -0.209440,1.213483,-1.039417,
        -0.209440,1.218146,-0.672643,
        -0.209440,1.221608,1.421669,
        -0.209440,1.226043,-1.693533,
        -0.209440,1.232870,-1.822773,
        -0.209440,1.234375,-2.421194,
        -0.209440,1.234531,-2.190148,
        -0.209440,1.239414,0.348959,
        -0.209440,1.240183,2.587255,
        -0.209440,1.242312,2.941234,
        -0.209440,1.244075,1.286644,
        -0.209440,1.245961,-2.997461,
        -0.209440,1.246495,-2.646872,
        -0.209440,1.247736,2.242810,
        -0.209440,1.253604,0.812222,
        -0.209440,1.257816,-0.894317,
        -0.209440,1.258247,-1.953634,
        -0.209440,1.258993,-0.284099,
        -0.209440,1.262380,-0.157609,
        -0.209440,1.262791,-1.549229,
        -0.209440,1.263772,-0.533073,
        -0.209440,1.266731,-1.412461,
        -0.209440,1.268591,-3.122146,
        -0.209440,1.269614,-1.275246,
        -0.209440,1.270745,1.152533,
        -0.209440,1.273606,0.224545,
        -0.209440,1.279716,2.812101,
        -0.209440,1.283786,2.462124,
        -0.209440,1.284798,0.078209,
        -0.209440,1.285402,-2.882049,
        -0.209440,1.286936,1.886053,
        -0.209440,1.289408,0.683122,
        -0.209440,1.294736,1.756184,
        -0.209440,1.296978,-2.311511,
        -0.209440,1.298546,2.020601,
        -0.209440,1.300924,1.022478,
        -0.209440,1.303305,3.043152,
        -0.209440,1.304210,-2.084225,
        -0.209440,1.305241,-2.530249,
        -0.209440,1.307513,1.628003,
        -0.209440,1.310722,0.558267,
        -0.209440,1.314142,2.139329,
        -0.209440,1.316087,-0.777971,
        -0.209440,1.317907,-1.121023,
        -0.209440,1.321019,-2.749776,
        -0.209440,1.321609,2.687717,
        -0.209440,1.322568,1.499415,
        -0.209440,1.322972,-0.395699,
        -0.209440,1.325286,-0.994192,
        -0.209440,1.329146,-0.648513,
        -0.209440,1.329795,-0.042570,
        -0.209440,1.331651,2.344497,
        -0.209440,1.336522,0.435098,
        -0.209440,1.339410,0.900087,
        -0.209440,1.339550,1.370353,
        -0.209440,1.340441,-1.731580,
        -0.209440,1.351855,-1.855234,
        -0.209440,1.360161,-3.018404,
        -0.209440,1.361031,-1.613971,
        -0.209440,1.361196,2.914780,
        -0.209440,1.361998,-2.207254,
        -0.209440,1.362302,1.239384,
        -0.209440,1.365228,2.563808,
        -0.209440,1.365452,0.310522,
        -0.209440,1.367028,-2.420041,
        -0.209440,1.373422,-1.484281,
        -0.209440,1.374245,-1.226417,
        -0.209440,1.375020,-0.170393,
        -0.209440,1.376560,-1.355481,
        -0.209440,1.377070,-1.980984,
        -0.209440,1.377578,-2.636339,
        -0.209440,1.378096,0.774626,
        -0.209440,1.379086,-0.508192,
        -0.209440,1.382562,0.177141,
        -0.209440,1.388849,2.233847,
        -0.209440,1.389268,-0.886349,
        -0.209440,1.391324,1.109396,
        -0.209440,1.391699,-0.288725,
        -0.209440,1.392869,3.136903,
        -0.209440,1.393612,1.952016,
        -0.209440,1.395545,-2.850807,
        -0.209440,1.401926,1.822930,
        -0.209440,1.402149,2.788898,
        -0.209440,1.409552,0.647980,
        -0.209440,1.410840,0.056219,
        -0.209440,1.411944,2.444377,
        -0.209440,1.413639,1.697982,
        -0.209440,1.426479,1.573427,
        -0.209440,1.426725,2.102242,
        -0.209440,1.427868,0.984637,
        -0.209440,1.429580,-2.107232,
        -0.209440,1.431375,-2.313269,
        -0.209440,1.431907,3.015410,
        -0.209440,1.432325,-1.029054,
        -0.209440,1.433341,0.524268,
        -0.209440,1.435181,-0.739771,
        -0.209440,1.437208,-2.529085,
        -0.209440,1.439041,-0.617966,
        -0.209440,1.441473,1.448544,
        -0.209440,1.444527,2.664744,
        -0.209440,1.450513,-0.400625,
        -0.209440,1.451907,-0.066823,
        -0.209440,1.452754,-2.738631,
        -0.209440,1.454061,-2.950416,
        -0.209440,1.456736,1.323154,
        -0.209440,1.458370,0.400433,
        -0.209440,1.459314,-1.765003,
        -0.209440,1.459902,-1.145835,
        -0.209440,1.464952,2.329245,
        -0.209440,1.466050,0.862104,
        -0.209440,1.471330,-1.886160,
        -0.209440,1.475119,-1.648255,
        -0.209440,1.480878,1.194576,
        -0.209440,1.480942,0.273235,
        -0.209440,1.481086,2.890269,
        -0.209440,1.484118,-1.529156,
        -0.209440,1.484741,-1.279219,
        -0.209440,1.487681,-1.406069,
        -0.209440,1.487886,-3.066037,
        -0.209440,1.490555,2.543284,
        -0.209440,1.492424,-0.190932,
        -0.209440,1.495002,-0.839810,
        -0.209440,1.495297,2.007188,
        -0.209440,1.496414,-2.009267,
        -0.209440,1.498261,-2.425270,
        -0.209440,1.498907,-2.207327,
        -0.209440,1.500204,0.737638,
        -0.209440,1.502469,0.151156,
        -0.209440,1.507448,-0.514782,
        -0.209440,1.509287,2.204647,
        -0.209440,1.510518,1.884920,
        -0.209440,1.511948,-2.632360,
        -0.209440,1.515235,1.068565,
        -0.209440,1.522237,3.102018,
        -0.209440,1.522373,1.764991,
        -0.209440,1.522416,-0.954956,
        -0.209440,1.523102,2.765920,
        -0.209440,1.524811,-2.844718,
        -0.209440,1.527261,-0.304916,
        -0.209440,1.529788,0.613208,
        -0.209440,1.533648,1.644808,
        -0.209440,1.535018,0.028869,
        -0.209440,1.541082,2.424608,
        -0.209440,1.546010,1.523462,
        -0.209440,1.551135,-0.642994,
        -0.209440,1.553895,0.489607,
        -0.209440,1.555041,0.947813,
        -0.209440,1.559671,-1.074934,
        -0.209440,1.560345,2.985899,
        -0.209440,1.560487,-2.323400,
        -0.209440,1.561127,1.401111,
        -0.209440,1.567193,2.643172,
        -0.209440,1.568892,2.098362,
        -0.209440,1.571702,-2.098435,
        -0.209440,1.572701,-2.528730,
        -0.209440,1.573572,1.278593,
        -0.209440,1.574849,-0.091505,
        -0.209440,1.576334,0.365236,
        -0.209440,1.577029,-2.962888,
        -0.209440,1.577688,-1.814941,
        -0.209440,1.579542,-0.755917,
        -0.209440,1.579624,-1.199759,
        -0.209440,1.581515,-0.419414,
        -0.209440,1.584581,-2.738702,
        -0.209440,1.586948,-1.696296,
        -0.209440,1.588901,2.301496,
        -0.209440,1.589936,0.825020,
        -0.209440,1.590018,-1.941652,
        -0.209440,1.596968,-1.328358,
        -0.209440,1.597178,-1.575607,
        -0.209440,1.600470,-1.453196,
        -0.209440,1.601408,0.242409,
        -0.209440,1.602259,1.152656,
        -0.209440,1.603064,2.865774,
        -0.209440,1.612085,1.968802,
        -0.209440,1.612720,-3.080498,
        -0.209440,1.616061,2.521531,
        -0.209440,1.616900,-0.207457,
        -0.209440,1.620988,-0.875920,
        -0.209440,1.622114,0.701509,
        -0.209440,1.622749,0.121426,
        -0.209440,1.622993,-2.221629,
        -0.209440,1.629346,-0.542669,
        -0.209440,1.630007,1.844552,
        -0.209440,1.634747,-2.426198,
        -0.209440,1.642346,1.722617,
        -0.209440,1.643925,1.034629,
        -0.209440,1.644848,2.743774,
        -0.209440,1.645258,2.188109,
        -0.209440,1.645417,-2.634725,
        -0.209440,1.647395,3.085437,
        -0.209440,1.648538,-2.855044,
        -0.209440,1.649932,0.576808,
        -0.209440,1.652224,-0.996105,
        -0.209440,1.652871,1.600681,
        -0.209440,1.657997,-0.325127,
        -0.209440,1.661936,0.001037,
        -0.209440,1.665187,1.477402,
        -0.209440,1.665496,2.397872,
        -0.209440,1.673267,-0.665453,
        -0.209440,1.673557,0.452526,
        -0.209440,1.679277,-1.122421,
        -0.209440,1.679696,0.912482,
        -0.209440,1.682420,1.353922,
        -0.209440,1.685031,-2.009038,
        -0.209440,1.685196,2.967046,
        -0.209440,1.690107,2.620910,
        -0.209440,1.690453,-2.123076,
        -0.209440,1.691693,-1.871572,
        -0.209440,1.691817,2.061688,
        -0.209440,1.692644,1.237670,
        -0.209440,1.696386,0.328123,
        -0.209440,1.697162,-1.249026,
        -0.209440,1.697488,-2.322785,
        -0.209440,1.698726,-2.973576,
        -0.209440,1.699172,-1.749717,
        -0.209440,1.704478,-0.106295,
        -0.209440,1.707244,-0.444770,
        -0.209440,1.707500,-2.531497,
        -0.209440,1.707714,-0.787321,
        -0.209440,1.708430,-1.627552,
        -0.209440,1.710583,-1.376890,
        -0.209440,1.712415,-2.747921,
        -0.209440,1.713125,-1.503073,
        -0.209440,1.713452,0.789419,
        -0.209440,1.720612,2.279888,
        -0.209440,1.724852,0.202831,
        -0.209440,1.724868,2.846246,
        -0.209440,1.728940,1.931535,
        -0.209440,1.732838,1.127584,
        -0.209440,1.735719,-3.092485,
        -0.209440,1.739676,2.496017,
        -0.209440,1.739971,-0.228876,
        -0.209440,1.743328,0.663347,
        -0.209440,1.744927,-0.911892,
        -0.209440,1.745952,0.084733,
        -0.209440,1.748077,1.804354,
        -0.209440,1.752959,-0.567548,
        -0.209440,1.760169,1.678132,
        -0.209440,1.763887,-2.218450,
        -0.209440,1.766737,2.722752,
        -0.209440,1.767999,1.004737,
        -0.209440,1.769149,1.550298,
        -0.209440,1.770119,-2.427175,
        -0.209440,1.770460,2.154208,
        -0.209440,1.770514,3.070982,
        -0.209440,1.770735,0.537245,
        -0.209440,1.772673,-2.864760,
        -0.209440,1.773527,-1.039582,
        -0.209440,1.776419,-2.642421,
        -0.209440,1.780769,1.416330,
        -0.209440,1.786634,-0.346247,
        -0.209440,1.788829,-1.947553,
        -0.209440,1.793737,0.410270,
        -0.209440,1.794553,2.374618,
        -0.209440,1.794699,-0.694061,
        -0.209440,1.797469,-1.169105,
        -0.209440,1.803705,0.881076,
        -0.209440,1.804124,1.264249,
        -0.209440,1.807825,-2.090455,
        -0.209440,1.808822,-1.817375,
        -0.209440,1.809020,2.949741,
        -0.209440,1.812466,2.597595,
        -0.209440,1.813159,2.025395,
        -0.209440,1.813816,-1.298435,
        -0.209440,1.815016,0.278951,
        -0.209440,1.816510,-0.008631,
        -0.209440,1.817772,-1.689625,
        -0.209440,1.821360,-2.984352,
        -0.209440,1.822312,-0.127211,
        -0.209440,1.823386,-1.429807,
        -0.209440,1.823834,-1.560483,
        -0.209440,1.832060,-0.821627,
        -0.209440,1.832690,-0.468440,
        -0.209440,1.833945,-2.320278,
        -0.209440,1.834787,0.751968,
        -0.209440,1.840579,-2.536445,
        -0.209440,1.841033,-2.756932,
        -0.209440,1.846350,1.893983,
        -0.209440,1.846644,1.113766,
        -0.209440,1.846833,2.826543,
        -0.209440,1.847325,2.249051,
        -0.209440,1.854119,0.134161,
        -0.209440,1.858863,-3.106022,
        -0.209440,1.865046,1.763384,
        -0.209440,1.865619,-0.952036,
        -0.209440,1.865912,0.624453,
        -0.209440,1.866694,2.473686,
        -0.209440,1.867524,-0.240083,
        -0.209440,1.875561,1.632831,
        -0.209440,1.876049,-0.594378,
        -0.209440,1.880509,1.500204,
        -0.209440,1.883504,1.360457,
        -0.209440,1.886971,-2.196665,
        -0.209440,1.888979,0.982922,
        -0.209440,1.889815,2.701570,
        -0.209440,1.892144,-2.005340,
        -0.209440,1.892348,0.496522,
        -0.209440,1.893281,-1.083976,
        -0.209440,1.893828,3.053453,
        -0.209440,1.895117,2.121107,
        -0.209440,1.898741,-2.875735,
        -0.209440,1.904429,-2.428029,
        -0.209440,1.908408,-2.648903,
        -0.209440,1.911828,-0.364489,
        -0.209440,1.912224,0.366565,
        -0.209440,1.914854,1.214468,
        -0.209440,1.915467,-1.883143,
        -0.209440,1.915763,-0.724679,
        -0.209440,1.915790,-1.216366,
        -0.209440,1.922256,2.347806,
        -0.209440,1.922616,0.844633,
        -0.209440,1.925185,0.231170,
        -0.209440,1.927474,-1.754616,
        -0.209440,1.928342,-1.351174,
        -0.209440,1.933328,-1.623328,
        -0.209440,1.933878,0.034150,
        -0.209440,1.934585,-1.487979,
        -0.209440,1.934802,2.927634,
        -0.209440,1.934862,1.989776,
        -0.209440,1.936327,-0.099350,
        -0.209440,1.938088,2.578507,
        -0.209440,1.945677,-2.999805,
        -0.209440,1.954572,-0.859208,
        -0.209440,1.957868,-0.490784,
        -0.209440,1.958989,0.712772,
        -0.209440,1.960212,-2.307561,
        -0.209440,1.964774,1.855507,
        -0.209440,1.964971,-2.100533,
        -0.209440,1.969112,1.088273,
        -0.209440,1.971438,-2.765415,
        -0.209440,1.972312,2.805134,
        -0.209440,1.974950,2.220596,
        -0.209440,1.975135,-2.539127,
        -0.209440,1.982107,1.720829,
        -0.209440,1.983195,-3.127984,
        -0.209440,1.983318,-0.247182,
        -0.209440,1.986720,-0.993220,
        -0.209440,1.990381,1.586005,
        -0.209440,1.990470,0.580283,
        -0.209440,1.990523,1.450326,
        -0.209440,1.993909,1.312897,
        -0.209440,1.994694,2.452813,
        -0.209440,1.999226,-0.620717,
        -0.209440,2.005145,0.941012,
        -0.209440,2.007301,0.142337,
        -0.209440,2.014149,0.444138,
        -0.209440,2.014338,-1.127392,
        -0.209440,2.018131,3.021778,
        -0.209440,2.019676,-1.980944,
        -0.209440,2.019912,2.686612,
        -0.209440,2.020545,2.088672,
        -0.209440,2.026074,0.300945,
        -0.209440,2.026196,-2.889057,
        -0.209440,2.031959,-1.268178,
        -0.209440,2.032917,-2.199917,
        -0.209440,2.033572,-1.842884,
        -0.209440,2.034712,-2.419505,
        -0.209440,2.035818,-0.379479,
        -0.209440,2.036004,-0.757062,
        -0.209440,2.040501,-1.411488,
        -0.209440,2.042874,-1.701367,
        -0.209440,2.043060,-2.653746,
        -0.209440,2.043729,-1.557938,
        -0.209440,2.048538,0.806623,
        -0.209440,2.049006,1.199930,
        -0.209440,2.049502,-0.137989,
        -0.209440,2.052181,2.326959,
        -0.209440,2.052794,-0.008526,
        -0.209440,2.057608,1.951726,
        -0.209440,2.065568,2.567730,
        -0.209440,2.070371,2.879808,
        -0.209440,2.071256,-3.018961,
        -0.209440,2.079746,-0.897699,
        -0.209440,2.082784,1.042403,
        -0.209440,2.083752,-0.511086,
        -0.209440,2.084401,1.810502,
        -0.209440,2.085731,0.671081,
        -0.209440,2.097496,1.373172,
        -0.209440,2.099185,1.667006,
        -0.209440,2.100482,-2.301084,
        -0.209440,2.103118,2.194745,
        -0.209440,2.103223,1.521685,
        -0.209440,2.103695,-2.777707,
        -0.209440,2.107523,3.129581,
        -0.209440,2.108277,-2.095907,
        -0.209440,2.108891,-2.532166,
        -0.209440,2.111003,-0.268632,
        -0.209440,2.112276,-1.027246,
        -0.209440,2.114380,0.531457,
        -0.209440,2.119166,0.100552,
        -0.209440,2.119564,2.754818,
        -0.209440,2.122474,-0.648314,
        -0.209440,2.125183,2.445598,
        -0.209440,2.129501,0.238290,
        -0.209440,2.131178,0.386663,
        -0.209440,2.131223,-1.174311,
        -0.209440,2.134269,-1.948042,
        -0.209440,2.134684,0.907065,
        -0.209440,2.137615,2.988806,
        -0.209440,2.145635,-1.327708,
        -0.209440,2.146467,2.055081,
        -0.209440,2.146582,-1.793788,
        -0.209440,2.149210,-1.482321,
        -0.209440,2.152901,-1.636788,
        -0.209440,2.156337,1.151558,
        -0.209440,2.156474,-2.912065,
        -0.209440,2.157194,-0.792563,
        -0.209440,2.166724,-0.401638,
        -0.209440,2.166980,-0.134879,
        -0.209440,2.173644,-2.403429,
        -0.209440,2.175711,-2.655523,
        -0.209440,2.177132,2.600419,
        -0.209440,2.178879,0.767722,
        -0.209440,2.180752,1.908485,
        -0.209440,2.181338,2.313721,
        -0.209440,2.185580,1.287673,
        -0.209440,2.195400,-2.217546,
        -0.209440,2.200418,-3.056442,
        -0.209440,2.201829,-0.001735,
        -0.209440,2.203804,1.756176,
        -0.209440,2.211602,1.443930,
        -0.209440,2.212686,-0.544809,
        -0.209440,2.213937,0.621892,
        -0.209440,2.214075,1.600691,
        -0.209440,2.216256,1.020029,
        -0.209440,2.219060,2.886631,
        -0.209440,2.223375,-0.906185,
        -0.209440,2.226002,-1.060248,
        -0.209440,2.226413,-2.064034,
        -0.209440,2.230554,2.171334,
        -0.209440,2.232413,2.735777,
        -0.209440,2.233709,3.070830,
        -0.209440,2.234383,-2.792065,
        -0.209440,2.237183,0.468799,
        -0.209440,2.237695,-0.271388,
        -0.209440,2.239419,0.151693,
        -0.209440,2.243347,-2.521049,
        -0.209440,2.244035,-1.229456,
        -0.209440,2.244355,0.309899,
        -0.209440,2.245175,2.460236,
        -0.209440,2.245628,-1.899216,
        -0.209440,2.245973,-0.702234,
        -0.209440,2.255601,-1.728131,
        -0.209440,2.256702,-1.401367,
        -0.209440,2.261466,-1.557500,
        -0.209440,2.267390,0.878499,
        -0.209440,2.272099,2.018594,
        -0.209440,2.278618,1.170386,
        -0.209440,2.283304,-2.345584,
        -0.209440,2.285175,-2.943090,
        -0.209440,2.293000,-0.124451,
        -0.209440,2.294058,-0.421070,
        -0.209440,2.303283,1.855938,
        -0.209440,2.304968,2.608978,
        -0.209440,2.305347,2.311346,
        -0.209440,2.307188,-2.656134,
        -0.209440,2.307301,1.336200,
        -0.209440,2.309655,0.726020,
        -0.209440,2.317634,-2.185225,
        -0.209440,2.321510,1.685664,
        -0.209440,2.324376,-3.111006,
        -0.209440,2.324731,1.511186,
        -0.209440,2.329587,0.035498,
        -0.209440,2.330158,2.977264,
        -0.209440,2.335096,-0.588468,
        -0.209440,2.336046,-1.115799,
        -0.209440,2.338060,2.803613,
        -0.209440,2.338083,-0.943400,
        -0.209440,2.340210,0.562145,
        -0.209440,2.341931,1.023926,
        -0.209440,2.343265,-0.779061,
        -0.209440,2.343434,-2.013692,
        -0.209440,2.353029,-1.310529,
        -0.209440,2.354868,0.211543,
        -0.209440,2.355120,0.388593,
        -0.209440,2.355165,-1.833173,
        -0.209440,2.355462,2.151441,
        -0.209440,2.362948,-2.487477,
        -0.209440,2.363101,-2.810456,
        -0.209440,2.363394,-0.272918,
        -0.209440,2.364509,-1.644533,
        -0.209440,2.370465,2.466824,
        -0.209440,2.386110,-1.481351,
        -0.209440,2.394960,0.862133,
        -0.209440,2.395716,1.977184,
        -0.209440,2.396339,1.200900,
        -0.209440,2.407941,-2.316753,
        -0.209440,2.409450,-2.988432,
        -0.209440,2.417202,-0.103363,
        -0.209440,2.417386,-0.443097,
        -0.209440,2.417950,2.666662,
        -0.209440,2.423190,1.789783,
        -0.209440,2.423692,1.392696,
        -0.209440,2.431122,3.086379,
        -0.209440,2.431406,-2.647820,
        -0.209440,2.431584,2.307631,
        -0.209440,2.434090,1.592818,
        -0.209440,2.435993,0.683844,
        -0.209440,2.440898,-2.133300,
        -0.209440,2.442222,-1.195837,
        -0.209440,2.443112,2.883761,
        -0.209440,2.449191,-0.633651,
        -0.209440,2.450821,0.084741,
        -0.209440,2.453499,-1.008861,
        -0.209440,2.460010,-1.938011,
        -0.209440,2.460199,-0.822008,
        -0.209440,2.460902,0.490079,
        -0.209440,2.461774,-1.734913,
        -0.209440,2.463986,1.034407,
        -0.209440,2.467647,0.287667,
        -0.209440,2.479849,2.126925,
        -0.209440,2.485907,-2.835471,
        -0.209440,2.485945,-1.374853,
        -0.209440,2.487747,2.505912,
        -0.209440,2.487957,-0.274279,
        -0.209440,2.489741,-2.468428,
        -0.209440,2.512538,1.239900,
        -0.209440,2.517528,1.923289,
        -0.209440,2.518454,-1.565852,
        -0.209440,2.519348,0.843503,
        -0.209440,2.522766,-3.059231,
        -0.209440,2.528686,2.737228,
        -0.209440,2.534798,-2.266671,
        -0.209440,2.536533,1.466274,
        -0.209440,2.538504,1.699812,
        -0.209440,2.539516,-0.476942,
        -0.209440,2.539532,-0.074548,
        -0.209440,2.543197,2.990252,
        -0.209440,2.554022,2.320619,
        -0.209440,2.556919,-2.655005,
        -0.209440,2.558146,0.627191,
        -0.209440,2.558223,-1.164189,
        -0.209440,2.563452,-2.044028,
        -0.209440,2.568462,0.150655,
        -0.209440,2.569130,-0.700837,
        -0.209440,2.570050,-1.804492,
        -0.209440,2.572710,-0.934665,
        -0.209440,2.575994,0.391718,
        -0.209440,2.584774,1.046703,
        -0.209440,2.601432,2.102535,
        -0.209440,2.603559,2.555114,
        -0.209440,2.604150,-2.885208,
        -0.209440,2.608630,-1.378011,
        -0.209440,2.611256,-0.279110,
        -0.209440,2.617344,-2.439480,
        -0.209440,2.627402,1.293552,
        -0.209440,2.635396,1.848850,
        -0.209440,2.635686,2.834950,
        -0.209440,2.637087,3.132730,
        -0.209440,2.642029,0.813274,
        -0.209440,2.645260,1.570203,
        -0.209440,2.646182,-1.614036,
        -0.209440,2.660113,-2.186513,
        -0.209440,2.660483,-0.034608,
        -0.209440,2.661713,-0.529511,
        -0.209440,2.672720,-1.110422,
        -0.209440,2.676626,2.335353,
        -0.209440,2.676638,0.540364,
        -0.209440,2.676780,-1.901158,
        -0.209440,2.681411,-2.667965,
        -0.209440,2.682152,0.245751,
        -0.209440,2.682859,-0.817071,
        -0.209440,2.705107,1.062142,
        -0.209440,2.717705,2.625995,
        -0.209440,2.721001,2.060097,
        -0.209440,2.722319,-2.954069,
        -0.209440,2.729840,-1.368237,
        -0.209440,2.734349,-0.289298,
        -0.209440,2.738409,2.975671,
        -0.209440,2.740366,1.374341,
        -0.209440,2.744633,-2.391311,
        -0.209440,2.747398,1.727166,
        -0.209440,2.763799,0.760372,
        -0.209440,2.768238,-1.679243,
        -0.209440,2.779245,-2.046268,
        -0.209440,2.780208,-0.626453,
        -0.209440,2.780471,0.027732,
        -0.209440,2.784696,-1.013128,
        -0.209440,2.789817,0.395431,
        -0.209440,2.801154,2.353143,
        -0.209440,2.807091,-2.687477,
        -0.209440,2.825762,1.085035,
        -0.209440,2.830209,2.734943,
        -0.209440,2.833517,-3.085508,
        -0.209440,2.838301,1.980061,
        -0.209440,2.849871,1.512012,
        -0.209440,2.850089,-1.343899,
        -0.209440,2.856905,-0.318441,
        -0.209440,2.870445,-2.291001,
        -0.209440,2.885017,0.655562,
        -0.209440,2.886428,-1.793217,
        -0.209440,2.890337,-0.825633,
        -0.209440,2.898932,0.139973,
        -0.209440,2.931927,-2.728768,
        -0.209440,2.932096,2.377790,
        -0.209440,2.937409,2.947245,
        -0.209440,2.948048,1.133127,
        -0.209440,2.950896,1.797889,
        -0.209440,2.970019,-1.282382,
        -0.209440,2.978333,-0.412090,
        -0.209440,2.996437,-2.089121,
        -0.209440,3.011274,0.434110,
        -0.209440,3.050410,3.138492,
        -0.209440,3.064492,1.647259,
        -0.209440,3.088868,-0.989501,
        -0.104720,0.053884,-2.408394,
        -0.104720,0.077645,1.453433,
        -0.104720,0.088408,-0.311160,
        -0.104720,0.142314,2.590387,
        -0.104720,0.147206,-1.385164,
        -0.104720,0.163506,0.573203,
        -0.104720,0.165485,-2.875396,
        -0.104720,0.182564,-2.125821,
        -0.104720,0.190569,1.929141,
        -0.104720,0.201102,1.249082,
        -0.104720,0.201397,-0.716646,
        -0.104720,0.210568,-0.069559,
        -0.104720,0.251672,2.848900,
        -0.104720,0.266262,-1.678068,
        -0.104720,0.269108,-1.215563,
        -0.104720,0.273777,-2.506756,
        -0.104720,0.279854,0.814581,
        -0.104720,0.281664,-2.969147,
        -0.104720,0.283625,2.405603,
        -0.104720,0.284058,0.357350,
        -0.104720,0.294107,1.593880,
        -0.104720,0.302491,-0.400410,
        -0.104720,0.308059,2.021491,
        -0.104720,0.312219,-2.072057,
        -0.104720,0.321905,-0.829121,
        -0.104720,0.329984,1.201621,
        -0.104720,0.339714,-0.013712,
        -0.104720,0.368784,2.974974,
        -0.104720,0.376428,-1.495532,
        -0.104720,0.378103,2.646436,
        -0.104720,0.385088,0.590355,
        -0.104720,0.386248,-2.690943,
        -0.104720,0.389851,-2.340031,
        -0.104720,0.391502,-1.171374,
        -0.104720,0.398249,-1.812694,
        -0.104720,0.400542,-2.995729,
        -0.104720,0.404273,0.918163,
        -0.104720,0.405617,1.773533,
        -0.104720,0.409009,-0.570243,
        -0.104720,0.410700,0.270433,
        -0.104720,0.415893,1.464890,
        -0.104720,0.420680,2.088164,
        -0.104720,0.424766,-0.258729,
        -0.104720,0.426507,2.367872,
        -0.104720,0.441292,-0.890727,
        -0.104720,0.451839,-2.078193,
        -0.104720,0.459270,1.191562,
        -0.104720,0.470622,0.014002,
        -0.104720,0.481292,-2.529098,
        -0.104720,0.482460,2.849843,
        -0.104720,0.492205,-1.397963,
        -0.104720,0.493314,3.117264,
        -0.104720,0.497870,0.727828,
        -0.104720,0.499783,-1.654379,
        -0.104720,0.502651,0.463333,
        -0.104720,0.505463,2.594557,
        -0.104720,0.506104,-2.874910,
        -0.104720,0.513579,-1.147354,
        -0.104720,0.518569,1.644237,
        -0.104720,0.519001,1.900488,
        -0.104720,0.519277,-0.673809,
        -0.104720,0.524769,-0.421955,
        -0.104720,0.525564,-2.288838,
        -0.104720,0.532541,0.978737,
        -0.104720,0.534610,2.154808,
        -0.104720,0.535020,-1.891846,
        -0.104720,0.540208,1.400497,
        -0.104720,0.542006,0.228337,
        -0.104720,0.550762,-0.180771,
        -0.104720,0.560531,2.378474,
        -0.104720,0.561629,-0.924576,
        -0.104720,0.567465,-2.679392,
        -0.104720,0.588527,1.181755,
        -0.104720,0.591643,-2.098364,
        -0.104720,0.594963,2.997575,
        -0.104720,0.599643,-3.045235,
        -0.104720,0.599962,2.773556,
        -0.104720,0.602620,0.029270,
        -0.104720,0.605711,0.598454,
        -0.104720,0.607708,-1.541026,
        -0.104720,0.609360,-2.464379,
        -0.104720,0.610738,-1.326786,
        -0.104720,0.616968,0.820489,
        -0.104720,0.626379,1.773610,
        -0.104720,0.628626,-1.751180,
        -0.104720,0.630671,0.391949,
        -0.104720,0.632622,-0.737642,
        -0.104720,0.632708,2.566722,
        -0.104720,0.632797,1.989858,
        -0.104720,0.634455,-0.534790,
        -0.104720,0.636205,1.561170,
        -0.104720,0.637681,-1.123423,
        -0.104720,0.641441,-2.842109,
        -0.104720,0.644807,-0.327139,
        -0.104720,0.652440,2.197129,
        -0.104720,0.660897,-2.278593,
        -0.104720,0.664008,1.360091,
        -0.104720,0.670850,-1.945257,
        -0.104720,0.671396,1.004859,
        -0.104720,0.673985,0.203855,
        -0.104720,0.680326,-0.133370,
        -0.104720,0.685083,-0.942363,
        -0.104720,0.688204,-2.632204,
        -0.104720,0.689777,2.385011,
        -0.104720,0.700065,3.115430,
        -0.104720,0.704080,2.911636,
        -0.104720,0.712883,1.174216,
        -0.104720,0.715701,0.699906,
        -0.104720,0.718851,-1.442215,
        -0.104720,0.721019,2.721536,
        -0.104720,0.725300,-2.975134,
        -0.104720,0.727136,-1.633288,
        -0.104720,0.727399,-2.120567,
        -0.104720,0.730751,0.522488,
        -0.104720,0.732892,-1.254330,
        -0.104720,0.734333,-2.446184,
        -0.104720,0.735125,0.037270,
        -0.104720,0.737893,1.866463,
        -0.104720,0.740650,1.681075,
        -0.104720,0.745971,-0.781286,
        -0.104720,0.749432,-0.442559,
        -0.104720,0.750102,2.050636,
        -0.104720,0.751046,0.863821,
        -0.104720,0.757018,-0.608833,
        -0.104720,0.757046,-1.816019,
        -0.104720,0.757389,1.501184,
        -0.104720,0.759800,2.546181,
        -0.104720,0.759859,0.351923,
        -0.104720,0.761141,-2.781102,
        -0.104720,0.769035,-1.078574,
        -0.104720,0.770166,-0.266604,
        -0.104720,0.772218,2.227380,
        -0.104720,0.788382,1.331051,
        -0.104720,0.790269,-2.282409,
        -0.104720,0.803410,-1.984567,
        -0.104720,0.803813,0.191520,
        -0.104720,0.806832,3.022155,
        -0.104720,0.807476,-2.597675,
        -0.104720,0.812512,-0.107335,
        -0.104720,0.814161,1.010949,
        -0.104720,0.817732,-3.084486,
        -0.104720,0.817846,2.390688,
        -0.104720,0.819323,2.845960,
        -0.104720,0.819430,-0.915017,
        -0.104720,0.829198,-1.532491,
        -0.104720,0.829531,0.665695,
        -0.104720,0.829816,1.167147,
        -0.104720,0.833613,-1.357492,
        -0.104720,0.842945,-2.907318,
        -0.104720,0.844883,2.679708,
        -0.104720,0.848189,-1.705003,
        -0.104720,0.849888,1.767901,
        -0.104720,0.851126,0.493370,
        -0.104720,0.852911,1.933394,
        -0.104720,0.857516,-2.432863,
        -0.104720,0.858261,-1.191729,
        -0.104720,0.858279,-2.138895,
        -0.104720,0.859723,-0.734928,
        -0.104720,0.859754,1.603292,
        -0.104720,0.861655,-0.394821,
        -0.104720,0.864300,0.043549,
        -0.104720,0.866940,0.851229,
        -0.104720,0.869748,2.095969,
        -0.104720,0.877352,-0.574668,
        -0.104720,0.879254,-2.733338,
        -0.104720,0.882554,1.443757,
        -0.104720,0.884060,0.334199,
        -0.104720,0.885626,-1.866061,
        -0.104720,0.889834,2.529957,
        -0.104720,0.893140,2.251698,
        -0.104720,0.897463,-0.240822,
        -0.104720,0.898758,-1.031948,
        -0.104720,0.913918,1.287618,
        -0.104720,0.916555,-2.287173,
        -0.104720,0.918305,3.107845,
        -0.104720,0.919522,2.948277,
        -0.104720,0.927138,-2.569686,
        -0.104720,0.929031,0.185383,
        -0.104720,0.933766,-3.015917,
        -0.104720,0.934873,-2.010822,
        -0.104720,0.935636,-1.445855,
        -0.104720,0.936702,-0.866461,
        -0.104720,0.937395,0.729108,
        -0.104720,0.937910,2.787575,
        -0.104720,0.939144,-0.092962,
        -0.104720,0.941718,1.125085,
        -0.104720,0.942873,-1.607680,
        -0.104720,0.943701,0.980328,
        -0.104720,0.947774,2.395256,
        -0.104720,0.952433,0.584347,
        -0.104720,0.954456,-1.287293,
        -0.104720,0.958731,-2.852552,
        -0.104720,0.961838,1.839284,
        -0.104720,0.961898,-0.469425,
        -0.104720,0.965426,1.686615,
        -0.104720,0.970108,1.990771,
        -0.104720,0.972820,-1.762780,
        -0.104720,0.973500,-0.711881,
        -0.104720,0.975253,2.641082,
        -0.104720,0.977824,0.441928,
        -0.104720,0.981485,-2.424766,
        -0.104720,0.981877,1.537100,
        -0.104720,0.982609,-1.136920,
        -0.104720,0.984265,-2.156073,
        -0.104720,0.989150,0.045493,
        -0.104720,0.993052,2.137478,
        -0.104720,0.997766,-2.693871,
        -0.104720,0.999847,-0.334738,
        -0.104720,1.001920,0.848054,
        -0.104720,1.006146,1.389734,
        -0.104720,1.012180,0.304066,
        -0.104720,1.016938,2.270674,
        -0.104720,1.017371,-0.979790,
        -0.104720,1.019495,-1.894487,
        -0.104720,1.023796,-0.583691,
        -0.104720,1.026525,2.506945,
        -0.104720,1.027004,-0.203653,
        -0.104720,1.028027,3.036573,
        -0.104720,1.031435,1.241021,
        -0.104720,1.032316,2.884221,
        -0.104720,1.035863,-3.110475,
        -0.104720,1.042085,-2.292950,
        -0.104720,1.043272,-1.522986,
        -0.104720,1.046265,-2.958847,
        -0.104720,1.049272,-1.376344,
        -0.104720,1.050039,-2.552086,
        -0.104720,1.053682,0.689863,
        -0.104720,1.055458,0.170394,
        -0.104720,1.056622,-0.831866,
        -0.104720,1.057515,1.097998,
        -0.104720,1.058706,-2.036438,
        -0.104720,1.061389,-1.672248,
        -0.104720,1.063758,2.743093,
        -0.104720,1.070157,-0.073422,
        -0.104720,1.072154,1.761506,
        -0.104720,1.074029,1.909193,
        -0.104720,1.074204,0.545276,
        -0.104720,1.074714,-2.805147,
        -0.104720,1.076262,0.966104,
        -0.104720,1.077642,-1.227073,
        -0.104720,1.084151,1.618525,
        -0.104720,1.085085,2.377673,
        -0.104720,1.086822,-0.441112,
        -0.104720,1.087142,2.058815,
        -0.104720,1.098383,-0.696430,
        -0.104720,1.099138,-1.084033,
        -0.104720,1.103208,1.477773,
        -0.104720,1.103783,0.408254,
        -0.104720,1.106275,-2.420800,
        -0.104720,1.108079,2.612943,
        -0.104720,1.109345,-2.172718,
        -0.104720,1.115957,-1.787597,
        -0.104720,1.118930,0.045983,
        -0.104720,1.119777,0.834702,
        -0.104720,1.121943,-2.669042,
        -0.104720,1.124930,2.219917,
        -0.104720,1.125217,1.336294,
        -0.104720,1.130958,2.972931,
        -0.104720,1.131400,-0.306629,
        -0.104720,1.137135,-0.936514,
        -0.104720,1.139125,-1.924528,
        -0.104720,1.139877,0.276382,
        -0.104720,1.142003,-3.058022,
        -0.104720,1.147363,-0.558448,
        -0.104720,1.150348,-0.176418,
        -0.104720,1.151198,1.197016,
        -0.104720,1.153930,-1.597104,
        -0.104720,1.155389,2.837148,
        -0.104720,1.155458,-1.460623,
        -0.104720,1.157331,-2.906362,
        -0.104720,1.158594,2.487618,
        -0.104720,1.162103,-1.322553,
        -0.104720,1.165407,3.092861,
        -0.104720,1.168025,0.717075,
        -0.104720,1.168406,-2.301519,
        -0.104720,1.174782,-2.538755,
        -0.104720,1.177998,1.061256,
        -0.104720,1.178898,1.827305,
        -0.104720,1.181331,-2.060330,
        -0.104720,1.182540,-0.798952,
        -0.104720,1.184569,0.147517,
        -0.104720,1.184734,1.968919,
        -0.104720,1.184804,0.593386,
        -0.104720,1.188278,1.690808,
        -0.104720,1.194360,2.110440,
        -0.104720,1.194776,2.710455,
        -0.104720,1.197745,-1.173705,
        -0.104720,1.199898,-2.777257,
        -0.104720,1.201164,2.355931,
        -0.104720,1.203760,1.556404,
        -0.104720,1.205845,-0.415993,
        -0.104720,1.206585,0.933648,
        -0.104720,1.206950,-0.037248,
        -0.104720,1.209341,0.471162,
        -0.104720,1.213483,-1.039417,
        -0.104720,1.218146,-0.672643,
        -0.104720,1.221608,1.421669,
        -0.104720,1.226043,-1.693533,
        -0.104720,1.232870,-1.822773,
        -0.104720,1.234375,-2.421194,
        -0.104720,1.234531,-2.190148,
        -0.104720,1.239414,0.348959,
        -0.104720,1.240183,2.587255,
        -0.104720,1.242312,2.941234,
        -0.104720,1.244075,1.286644,
        -0.104720,1.245961,-2.997461,
        -0.104720,1.246495,-2.646872,
        -0.104720,1.247736,2.242810,
        -0.104720,1.253604,0.812222,
        -0.104720,1.257816,-0.894317,
        -0.104720,1.258247,-1.953634,
        -0.104720,1.258993,-0.284099,
        -0.104720,1.262380,-0.157609,
        -0.104720,1.262791,-1.549229,
        -0.104720,1.263772,-0.533073,
        -0.104720,1.266731,-1.412461,
        -0.104720,1.268591,-3.122146,
        -0.104720,1.269614,-1.275246,
        -0.104720,1.270745,1.152533,
        -0.104720,1.273606,0.224545,
        -0.104720,1.279716,2.812101,
        -0.104720,1.283786,2.462124,
        -0.104720,1.284798,0.078209,
        -0.104720,1.285402,-2.882049,
        -0.104720,1.286936,1.886053,
        -0.104720,1.289408,0.683122,
        -0.104720,1.294736,1.756184,
        -0.104720,1.296978,-2.311511,
        -0.104720,1.298546,2.020601,
        -0.104720,1.300924,1.022478,
        -0.104720,1.303305,3.043152,
        -0.104720,1.304210,-2.084225,
        -0.104720,1.305241,-2.530249,
        -0.104720,1.307513,1.628003,
        -0.104720,1.310722,0.558267,
        -0.104720,1.314142,2.139329,
        -0.104720,1.316087,-0.777971,
        -0.104720,1.317907,-1.121023,
        -0.104720,1.321019,-2.749776,
        -0.104720,1.321609,2.687717,
        -0.104720,1.322568,1.499415,
        -0.104720,1.322972,-0.395699,
        -0.104720,1.325286,-0.994192,
        -0.104720,1.329146,-0.648513,
        -0.104720,1.329795,-0.042570,
        -0.104720,1.331651,2.344497,
        -0.104720,1.336522,0.435098,
        -0.104720,1.339410,0.900087,
        -0.104720,1.339550,1.370353,
        -0.104720,1.340441,-1.731580,
        -0.104720,1.351855,-1.855234,
        -0.104720,1.360161,-3.018404,
        -0.104720,1.361031,-1.613971,
        -0.104720,1.361196,2.914780,
        -0.104720,1.361998,-2.207254,
        -0.104720,1.362302,1.239384,
        -0.104720,1.365228,2.563808,
        -0.104720,1.365452,0.310522,
        -0.104720,1.367028,-2.420041,
        -0.104720,1.373422,-1.484281,
        -0.104720,1.374245,-1.226417,
        -0.104720,1.375020,-0.170393,
        -0.104720,1.376560,-1.355481,
        -0.104720,1.377070,-1.980984,
        -0.104720,1.377578,-2.636339,
        -0.104720,1.378096,0.774626,
        -0.104720,1.379086,-0.508192,
        -0.104720,1.382562,0.177141,
        -0.104720,1.388849,2.233847,
        -0.104720,1.389268,-0.886349,
        -0.104720,1.391324,1.109396,
        -0.104720,1.391699,-0.288725,
        -0.104720,1.392869,3.136903,
        -0.104720,1.393612,1.952016,
        -0.104720,1.395545,-2.850807,
        -0.104720,1.401926,1.822930,
        -0.104720,1.402149,2.788898,
        -0.104720,1.409552,0.647980,
        -0.104720,1.410840,0.056219,
        -0.104720,1.411944,2.444377,
        -0.104720,1.413639,1.697982,
        -0.104720,1.426479,1.573427,
        -0.104720,1.426725,2.102242,
        -0.104720,1.427868,0.984637,
        -0.104720,1.429580,-2.107232,
        -0.104720,1.431375,-2.313269,
        -0.104720,1.431907,3.015410,
        -0.104720,1.432325,-1.029054,
        -0.104720,1.433341,0.524268,
        -0.104720,1.435181,-0.739771,
        -0.104720,1.437208,-2.529085,
        -0.104720,1.439041,-0.617966,
        -0.104720,1.441473,1.448544,
        -0.104720,1.444527,2.664744,
        -0.104720,1.450513,-0.400625,
        -0.104720,1.451907,-0.066823,
        -0.104720,1.452754,-2.738631,
        -0.104720,1.454061,-2.950416,
        -0.104720,1.456736,1.323154,
        -0.104720,1.458370,0.400433,
        -0.104720,1.459314,-1.765003,
        -0.104720,1.459902,-1.145835,
        -0.104720,1.464952,2.329245,
        -0.104720,1.466050,0.862104,
        -0.104720,1.471330,-1.886160,
        -0.104720,1.475119,-1.648255,
        -0.104720,1.480878,1.194576,
        -0.104720,1.480942,0.273235,
        -0.104720,1.481086,2.890269,
        -0.104720,1.484118,-1.529156,
        -0.104720,1.484741,-1.279219,
        -0.104720,1.487681,-1.406069,
        -0.104720,1.487886,-3.066037,
        -0.104720,1.490555,2.543284,
        -0.104720,1.492424,-0.190932,
        -0.104720,1.495002,-0.839810,
        -0.104720,1.495297,2.007188,
        -0.104720,1.496414,-2.009267,
        -0.104720,1.498261,-2.425270,
        -0.104720,1.498907,-2.207327,
        -0.104720,1.500204,0.737638,
        -0.104720,1.502469,0.151156,
        -0.104720,1.507448,-0.514782,
        -0.104720,1.509287,2.204647,
        -0.104720,1.510518,1.884920,
        -0.104720,1.511948,-2.632360,
        -0.104720,1.515235,1.068565,
        -0.104720,1.522237,3.102018,
        -0.104720,1.522373,1.764991,
        -0.104720,1.522416,-0.954956,
        -0.104720,1.523102,2.765920,
        -0.104720,1.524811,-2.844718,
        -0.104720,1.527261,-0.304916,
        -0.104720,1.529788,0.613208,
        -0.104720,1.533648,1.644808,
        -0.104720,1.535018,0.028869,
        -0.104720,1.541082,2.424608,
        -0.104720,1.546010,1.523462,
        -0.104720,1.551135,-0.642994,
        -0.104720,1.553895,0.489607,
        -0.104720,1.555041,0.947813,
        -0.104720,1.559671,-1.074934,
        -0.104720,1.560345,2.985899,
        -0.104720,1.560487,-2.323400,
        -0.104720,1.561127,1.401111,
        -0.104720,1.567193,2.643172,
        -0.104720,1.568892,2.098362,
        -0.104720,1.571702,-2.098435,
        -0.104720,1.572701,-2.528730,
        -0.104720,1.573572,1.278593,
        -0.104720,1.574849,-0.091505,
        -0.104720,1.576334,0.365236,
        -0.104720,1.577029,-2.962888,
        -0.104720,1.577688,-1.814941,
        -0.104720,1.579542,-0.755917,
        -0.104720,1.579624,-1.199759,
        -0.104720,1.581515,-0.419414,
        -0.104720,1.584581,-2.738702,
        -0.104720,1.586948,-1.696296,
        -0.104720,1.588901,2.301496,
        -0.104720,1.589936,0.825020,
        -0.104720,1.590018,-1.941652,
        -0.104720,1.596968,-1.328358,
        -0.104720,1.597178,-1.575607,
        -0.104720,1.600470,-1.453196,
        -0.104720,1.601408,0.242409,
        -0.104720,1.602259,1.152656,
        -0.104720,1.603064,2.865774,
        -0.104720,1.612085,1.968802,
        -0.104720,1.612720,-3.080498,
        -0.104720,1.616061,2.521531,
        -0.104720,1.616900,-0.207457,
        -0.104720,1.620988,-0.875920,
        -0.104720,1.622114,0.701509,
        -0.104720,1.622749,0.121426,
        -0.104720,1.622993,-2.221629,
        -0.104720,1.629346,-0.542669,
        -0.104720,1.630007,1.844552,
        -0.104720,1.634747,-2.426198,
        -0.104720,1.642346,1.722617,
        -0.104720,1.643925,1.034629,
        -0.104720,1.644848,2.743774,
        -0.104720,1.645258,2.188109,
        -0.104720,1.645417,-2.634725,
        -0.104720,1.647395,3.085437,
        -0.104720,1.648538,-2.855044,
        -0.104720,1.649932,0.576808,
        -0.104720,1.652224,-0.996105,
        -0.104720,1.652871,1.600681,
        -0.104720,1.657997,-0.325127,
        -0.104720,1.661936,0.001037,
        -0.104720,1.665187,1.477402,
        -0.104720,1.665496,2.397872,
        -0.104720,1.673267,-0.665453,
        -0.104720,1.673557,0.452526,
        -0.104720,1.679277,-1.122421,
        -0.104720,1.679696,0.912482,
        -0.104720,1.682420,1.353922,
        -0.104720,1.685031,-2.009038,
        -0.104720,1.685196,2.967046,
        -0.104720,1.690107,2.620910,
        -0.104720,1.690453,-2.123076,
        -0.104720,1.691693,-1.871572,
        -0.104720,1.691817,2.061688,
        -0.104720,1.692644,1.237670,
        -0.104720,1.696386,0.328123,
        -0.104720,1.697162,-1.249026,
        -0.104720,1.697488,-2.322785,
        -0.104720,1.698726,-2.973576,
        -0.104720,1.699172,-1.749717,
        -0.104720,1.704478,-0.106295,
        -0.104720,1.707244,-0.444770,
        -0.104720,1.707500,-2.531497,
        -0.104720,1.707714,-0.787321,
        -0.104720,1.708430,-1.627552,
        -0.104720,1.710583,-1.376890,
        -0.104720,1.712415,-2.747921,
        -0.104720,1.713125,-1.503073,
        -0.104720,1.713452,0.789419,
        -0.104720,1.720612,2.279888,
        -0.104720,1.724852,0.202831,
        -0.104720,1.724868,2.846246,
        -0.104720,1.728940,1.931535,
        -0.104720,1.732838,1.127584,
        -0.104720,1.735719,-3.092485,
        -0.104720,1.739676,2.496017,
        -0.104720,1.739971,-0.228876,
        -0.104720,1.743328,0.663347,
        -0.104720,1.744927,-0.911892,
        -0.104720,1.745952,0.084733,
        -0.104720,1.748077,1.804354,
        -0.104720,1.752959,-0.567548,
        -0.104720,1.760169,1.678132,
        -0.104720,1.763887,-2.218450,
        -0.104720,1.766737,2.722752,
        -0.104720,1.767999,1.004737,
        -0.104720,1.769149,1.550298,
        -0.104720,1.770119,-2.427175,
        -0.104720,1.770460,2.154208,
        -0.104720,1.770514,3.070982,
        -0.104720,1.770735,0.537245,
        -0.104720,1.772673,-2.864760,
        -0.104720,1.773527,-1.039582,
        -0.104720,1.776419,-2.642421,
        -0.104720,1.780769,1.416330,
        -0.104720,1.786634,-0.346247,
        -0.104720,1.788829,-1.947553,
        -0.104720,1.793737,0.410270,
        -0.104720,1.794553,2.374618,
        -0.104720,1.794699,-0.694061,
        -0.104720,1.797469,-1.169105,
        -0.104720,1.803705,0.881076,
        -0.104720,1.804124,1.264249,
        -0.104720,1.807825,-2.090455,
        -0.104720,1.808822,-1.817375,
        -0.104720,1.809020,2.949741,
        -0.104720,1.812466,2.597595,
        -0.104720,1.813159,2.025395,
        -0.104720,1.813816,-1.298435,
        -0.104720,1.815016,0.278951,
        -0.104720,1.816510,-0.008631,
        -0.104720,1.817772,-1.689625,
        -0.104720,1.821360,-2.984352,
        -0.104720,1.822312,-0.127211,
        -0.104720,1.823386,-1.429807,
        -0.104720,1.823834,-1.560483,
        -0.104720,1.832060,-0.821627,
        -0.104720,1.832690,-0.468440,
        -0.104720,1.833945,-2.320278,
        -0.104720,1.834787,0.751968,
        -0.104720,1.840579,-2.536445,
        -0.104720,1.841033,-2.756932,
        -0.104720,1.846350,1.893983,
        -0.104720,1.846644,1.113766,
        -0.104720,1.846833,2.826543,
        -0.104720,1.847325,2.249051,
        -0.104720,1.854119,0.134161,
        -0.104720,1.858863,-3.106022,
        -0.104720,1.865046,1.763384,
        -0.104720,1.865619,-0.952036,
        -0.104720,1.865912,0.624453,
        -0.104720,1.866694,2.473686,
        -0.104720,1.867524,-0.240083,
        -0.104720,1.875561,1.632831,
        -0.104720,1.876049,-0.594378,
        -0.104720,1.880509,1.500204,
        -0.104720,1.883504,1.360457,
        -0.104720,1.886971,-2.196665,
        -0.104720,1.888979,0.982922,
        -0.104720,1.889815,2.701570,
        -0.104720,1.892144,-2.005340,
        -0.104720,1.892348,0.496522,
        -0.104720,1.893281,-1.083976,
        -0.104720,1.893828,3.053453,
        -0.104720,1.895117,2.121107,
        -0.104720,1.898741,-2.875735,
        -0.104720,1.904429,-2.428029,
        -0.104720,1.908408,-2.648903,
        -0.104720,1.911828,-0.364489,
        -0.104720,1.912224,0.366565,
        -0.104720,1.914854,1.214468,
        -0.104720,1.915467,-1.883143,
        -0.104720,1.915763,-0.724679,
        -0.104720,1.915790,-1.216366,
        -0.104720,1.922256,2.347806,
        -0.104720,1.922616,0.844633,
        -0.104720,1.925185,0.231170,
        -0.104720,1.927474,-1.754616,
        -0.104720,1.928342,-1.351174,
        -0.104720,1.933328,-1.623328,
        -0.104720,1.933878,0.034150,
        -0.104720,1.934585,-1.487979,
        -0.104720,1.934802,2.927634,
        -0.104720,1.934862,1.989776,
        -0.104720,1.936327,-0.099350,
        -0.104720,1.938088,2.578507,
        -0.104720,1.945677,-2.999805,
        -0.104720,1.954572,-0.859208,
        -0.104720,1.957868,-0.490784,
        -0.104720,1.958989,0.712772,
        -0.104720,1.960212,-2.307561,
        -0.104720,1.964774,1.855507,
        -0.104720,1.964971,-2.100533,
        -0.104720,1.969112,1.088273,
        -0.104720,1.971438,-2.765415,
        -0.104720,1.972312,2.805134,
        -0.104720,1.974950,2.220596,
        -0.104720,1.975135,-2.539127,
        -0.104720,1.982107,1.720829,
        -0.104720,1.983195,-3.127984,
        -0.104720,1.983318,-0.247182,
        -0.104720,1.986720,-0.993220,
        -0.104720,1.990381,1.586005,
        -0.104720,1.990470,0.580283,
        -0.104720,1.990523,1.450326,
        -0.104720,1.993909,1.312897,
        -0.104720,1.994694,2.452813,
        -0.104720,1.999226,-0.620717,
        -0.104720,2.005145,0.941012,
        -0.104720,2.007301,0.142337,
        -0.104720,2.014149,0.444138,
        -0.104720,2.014338,-1.127392,
        -0.104720,2.018131,3.021778,
        -0.104720,2.019676,-1.980944,
        -0.104720,2.019912,2.686612,
        -0.104720,2.020545,2.088672,
        -0.104720,2.026074,0.300945,
        -0.104720,2.026196,-2.889057,
        -0.104720,2.031959,-1.268178,
        -0.104720,2.032917,-2.199917,
        -0.104720,2.033572,-1.842884,
        -0.104720,2.034712,-2.419505,
        -0.104720,2.035818,-0.379479,
        -0.104720,2.036004,-0.757062,
        -0.104720,2.040501,-1.411488,
        -0.104720,2.042874,-1.701367,
        -0.104720,2.043060,-2.653746,
        -0.104720,2.043729,-1.557938,
        -0.104720,2.048538,0.806623,
        -0.104720,2.049006,1.199930,
        -0.104720,2.049502,-0.137989,
        -0.104720,2.052181,2.326959,
        -0.104720,2.052794,-0.008526,
        -0.104720,2.057608,1.951726,
        -0.104720,2.065568,2.567730,
        -0.104720,2.070371,2.879808,
        -0.104720,2.071256,-3.018961,
        -0.104720,2.079746,-0.897699,
        -0.104720,2.082784,1.042403,
        -0.104720,2.083752,-0.511086,
        -0.104720,2.084401,1.810502,
        -0.104720,2.085731,0.671081,
        -0.104720,2.097496,1.373172,
        -0.104720,2.099185,1.667006,
        -0.104720,2.100482,-2.301084,
        -0.104720,2.103118,2.194745,
        -0.104720,2.103223,1.521685,
        -0.104720,2.103695,-2.777707,
        -0.104720,2.107523,3.129581,
        -0.104720,2.108277,-2.095907,
        -0.104720,2.108891,-2.532166,
        -0.104720,2.111003,-0.268632,
        -0.104720,2.112276,-1.027246,
        -0.104720,2.114380,0.531457,
        -0.104720,2.119166,0.100552,
        -0.104720,2.119564,2.754818,
        -0.104720,2.122474,-0.648314,
        -0.104720,2.125183,2.445598,
        -0.104720,2.129501,0.238290,
        -0.104720,2.131178,0.386663,
        -0.104720,2.131223,-1.174311,
        -0.104720,2.134269,-1.948042,
        -0.104720,2.134684,0.907065,
        -0.104720,2.137615,2.988806,
        -0.104720,2.145635,-1.327708,
        -0.104720,2.146467,2.055081,
        -0.104720,2.146582,-1.793788,
        -0.104720,2.149210,-1.482321,
        -0.104720,2.152901,-1.636788,
        -0.104720,2.156337,1.151558,
        -0.104720,2.156474,-2.912065,
        -0.104720,2.157194,-0.792563,
        -0.104720,2.166724,-0.401638,
        -0.104720,2.166980,-0.134879,
        -0.104720,2.173644,-2.403429,
        -0.104720,2.175711,-2.655523,
        -0.104720,2.177132,2.600419,
        -0.104720,2.178879,0.767722,
        -0.104720,2.180752,1.908485,
        -0.104720,2.181338,2.313721,
        -0.104720,2.185580,1.287673,
        -0.104720,2.195400,-2.217546,
        -0.104720,2.200418,-3.056442,
        -0.104720,2.201829,-0.001735,
        -0.104720,2.203804,1.756176,
        -0.104720,2.211602,1.443930,
        -0.104720,2.212686,-0.544809,
        -0.104720,2.213937,0.621892,
        -0.104720,2.214075,1.600691,
        -0.104720,2.216256,1.020029,
        -0.104720,2.219060,2.886631,
        -0.104720,2.223375,-0.906185,
        -0.104720,2.226002,-1.060248,
        -0.104720,2.226413,-2.064034,
        -0.104720,2.230554,2.171334,
        -0.104720,2.232413,2.735777,
        -0.104720,2.233709,3.070830,
        -0.104720,2.234383,-2.792065,
        -0.104720,2.237183,0.468799,
        -0.104720,2.237695,-0.271388,
        -0.104720,2.239419,0.151693,
        -0.104720,2.243347,-2.521049,
        -0.104720,2.244035,-1.229456,
        -0.104720,2.244355,0.309899,
        -0.104720,2.245175,2.460236,
        -0.104720,2.245628,-1.899216,
        -0.104720,2.245973,-0.702234,
        -0.104720,2.255601,-1.728131,
        -0.104720,2.256702,-1.401367,
        -0.104720,2.261466,-1.557500,
        -0.104720,2.267390,0.878499,
        -0.104720,2.272099,2.018594,
        -0.104720,2.278618,1.170386,
        -0.104720,2.283304,-2.345584,
        -0.104720,2.285175,-2.943090,
        -0.104720,2.293000,-0.124451,
        -0.104720,2.294058,-0.421070,
        -0.104720,2.303283,1.855938,
        -0.104720,2.304968,2.608978,
        -0.104720,2.305347,2.311346,
        -0.104720,2.307188,-2.656134,
        -0.104720,2.307301,1.336200,
        -0.104720,2.309655,0.726020,
        -0.104720,2.317634,-2.185225,
        -0.104720,2.321510,1.685664,
        -0.104720,2.324376,-3.111006,
        -0.104720,2.324731,1.511186,
        -0.104720,2.329587,0.035498,
        -0.104720,2.330158,2.977264,
        -0.104720,2.335096,-0.588468,
        -0.104720,2.336046,-1.115799,
        -0.104720,2.338060,2.803613,
        -0.104720,2.338083,-0.943400,
        -0.104720,2.340210,0.562145,
        -0.104720,2.341931,1.023926,
        -0.104720,2.343265,-0.779061,
        -0.104720,2.343434,-2.013692,
        -0.104720,2.353029,-1.310529,
        -0.104720,2.354868,0.211543,
        -0.104720,2.355120,0.388593,
        -0.104720,2.355165,-1.833173,
        -0.104720,2.355462,2.151441,
        -0.104720,2.362948,-2.487477,
        -0.104720,2.363101,-2.810456,
        -0.104720,2.363394,-0.272918,
        -0.104720,2.364509,-1.644533,
        -0.104720,2.370465,2.466824,
        -0.104720,2.386110,-1.481351,
        -0.104720,2.394960,0.862133,
        -0.104720,2.395716,1.977184,
        -0.104720,2.396339,1.200900,
        -0.104720,2.407941,-2.316753,
        -0.104720,2.409450,-2.988432,
        -0.104720,2.417202,-0.103363,
        -0.104720,2.417386,-0.443097,
        -0.104720,2.417950,2.666662,
        -0.104720,2.423190,1.789783,
        -0.104720,2.423692,1.392696,
        -0.104720,2.431122,3.086379,
        -0.104720,2.431406,-2.647820,
        -0.104720,2.431584,2.307631,
        -0.104720,2.434090,1.592818,
        -0.104720,2.435993,0.683844,
        -0.104720,2.440898,-2.133300,
        -0.104720,2.442222,-1.195837,
        -0.104720,2.443112,2.883761,
        -0.104720,2.449191,-0.633651,
        -0.104720,2.450821,0.084741,
        -0.104720,2.453499,-1.008861,
        -0.104720,2.460010,-1.938011,
        -0.104720,2.460199,-0.822008,
        -0.104720,2.460902,0.490079,
        -0.104720,2.461774,-1.734913,
        -0.104720,2.463986,1.034407,
        -0.104720,2.467647,0.287667,
        -0.104720,2.479849,2.126925,
        -0.104720,2.485907,-2.835471,
        -0.104720,2.485945,-1.374853,
        -0.104720,2.487747,2.505912,
        -0.104720,2.487957,-0.274279,
        -0.104720,2.489741,-2.468428,
        -0.104720,2.512538,1.239900,
        -0.104720,2.517528,1.923289,
        -0.104720,2.518454,-1.565852,
        -0.104720,2.519348,0.843503,
        -0.104720,2.522766,-3.059231,
        -0.104720,2.528686,2.737228,
        -0.104720,2.534798,-2.266671,
        -0.104720,2.536533,1.466274,
        -0.104720,2.538504,1.699812,
        -0.104720,2.539516,-0.476942,
        -0.104720,2.539532,-0.074548,
        -0.104720,2.543197,2.990252,
        -0.104720,2.554022,2.320619,
        -0.104720,2.556919,-2.655005,
        -0.104720,2.558146,0.627191,
        -0.104720,2.558223,-1.164189,
        -0.104720,2.563452,-2.044028,
        -0.104720,2.568462,0.150655,
        -0.104720,2.569130,-0.700837,
        -0.104720,2.570050,-1.804492,
        -0.104720,2.572710,-0.934665,
        -0.104720,2.575994,0.391718,
        -0.104720,2.584774,1.046703,
        -0.104720,2.601432,2.102535,
        -0.104720,2.603559,2.555114,
        -0.104720,2.604150,-2.885208,
        -0.104720,2.608630,-1.378011,
        -0.104720,2.611256,-0.279110,
        -0.104720,2.617344,-2.439480,
        -0.104720,2.627402,1.293552,
        -0.104720,2.635396,1.848850,
        -0.104720,2.635686,2.834950,
        -0.104720,2.637087,3.132730,
        -0.104720,2.642029,0.813274,
        -0.104720,2.645260,1.570203,
        -0.104720,2.646182,-1.614036,
        -0.104720,2.660113,-2.186513,
        -0.104720,2.660483,-0.034608,
        -0.104720,2.661713,-0.529511,
        -0.104720,2.672720,-1.110422,
        -0.104720,2.676626,2.335353,
        -0.104720,2.676638,0.540364,
        -0.104720,2.676780,-1.901158,
        -0.104720,2.681411,-2.667965,
        -0.104720,2.682152,0.245751,
        -0.104720,2.682859,-0.817071,
        -0.104720,2.705107,1.062142,
        -0.104720,2.717705,2.625995,
        -0.104720,2.721001,2.060097,
        -0.104720,2.722319,-2.954069,
        -0.104720,2.729840,-1.368237,
        -0.104720,2.734349,-0.289298,
        -0.104720,2.738409,2.975671,
        -0.104720,2.740366,1.374341,
        -0.104720,2.744633,-2.391311,
        -0.104720,2.747398,1.727166,
        -0.104720,2.763799,0.760372,
        -0.104720,2.768238,-1.679243,
        -0.104720,2.779245,-2.046268,
        -0.104720,2.780208,-0.626453,
        -0.104720,2.780471,0.027732,
        -0.104720,2.784696,-1.013128,
        -0.104720,2.789817,0.395431,
        -0.104720,2.801154,2.353143,
        -0.104720,2.807091,-2.687477,
        -0.104720,2.825762,1.085035,
        -0.104720,2.830209,2.734943,
        -0.104720,2.833517,-3.085508,
        -0.104720,2.838301,1.980061,
        -0.104720,2.849871,1.512012,
        -0.104720,2.850089,-1.343899,
        -0.104720,2.856905,-0.318441,
        -0.104720,2.870445,-2.291001,
        -0.104720,2.885017,0.655562,
        -0.104720,2.886428,-1.793217,
        -0.104720,2.890337,-0.825633,
        -0.104720,2.898932,0.139973,
        -0.104720,2.931927,-2.728768,
        -0.104720,2.932096,2.377790,
        -0.104720,2.937409,2.947245,
        -0.104720,2.948048,1.133127,
        -0.104720,2.950896,1.797889,
        -0.104720,2.970019,-1.282382,
        -0.104720,2.978333,-0.412090,
        -0.104720,2.996437,-2.089121,
        -0.104720,3.011274,0.434110,
        -0.104720,3.050410,3.138492,
        -0.104720,3.064492,1.647259,
        -0.104720,3.088868,-0.989501,
        0.000000,0.053884,-2.408394,
        0.000000,0.077645,1.453433,
        0.000000,0.088408,-0.311160,
        0.000000,0.142314,2.590387,
        0.000000,0.147206,-1.385164,
        0.000000,0.163506,0.573203,
        0.000000,0.165485,-2.875396,
        0.000000,0.182564,-2.125821,
        0.000000,0.190569,1.929141,
        0.000000,0.201102,1.249082,
        0.000000,0.201397,-0.716646,
        0.000000,0.210568,-0.069559,
        0.000000,0.251672,2.848900,
        0.000000,0.266262,-1.678068,
        0.000000,0.269108,-1.215563,
        0.000000,0.273777,-2.506756,
        0.000000,0.279854,0.814581,
        0.000000,0.281664,-2.969147,
        0.000000,0.283625,2.405603,
        0.000000,0.284058,0.357350,
        0.000000,0.294107,1.593880,
        0.000000,0.302491,-0.400410,
        0.000000,0.308059,2.021491,
        0.000000,0.312219,-2.072057,
        0.000000,0.321905,-0.829121,
        0.000000,0.329984,1.201621,
        0.000000,0.339714,-0.013712,
        0.000000,0.368784,2.974974,
        0.000000,0.376428,-1.495532,
        0.000000,0.378103,2.646436,
        0.000000,0.385088,0.590355,
        0.000000,0.386248,-2.690943,
        0.000000,0.389851,-2.340031,
        0.000000,0.391502,-1.171374,
        0.000000,0.398249,-1.812694,
        0.000000,0.400542,-2.995729,
        0.000000,0.404273,0.918163,
        0.000000,0.405617,1.773533,
        0.000000,0.409009,-0.570243,
        0.000000,0.410700,0.270433,
        0.000000,0.415893,1.464890,
        0.000000,0.420680,2.088164,
        0.000000,0.424766,-0.258729,
        0.000000,0.426507,2.367872,
        0.000000,0.441292,-0.890727,
        0.000000,0.451839,-2.078193,
        0.000000,0.459270,1.191562,
        0.000000,0.470622,0.014002,
        0.000000,0.481292,-2.529098,
        0.000000,0.482460,2.849843,
        0.000000,0.492205,-1.397963,
        0.000000,0.493314,3.117264,
        0.000000,0.497870,0.727828,
        0.000000,0.499783,-1.654379,
        0.000000,0.502651,0.463333,
        0.000000,0.505463,2.594557,
        0.000000,0.506104,-2.874910,
        0.000000,0.513579,-1.147354,
        0.000000,0.518569,1.644237,
        0.000000,0.519001,1.900488,
        0.000000,0.519277,-0.673809,
        0.000000,0.524769,-0.421955,
        0.000000,0.525564,-2.288838,
        0.000000,0.532541,0.978737,
        0.000000,0.534610,2.154808,
        0.000000,0.535020,-1.891846,
        0.000000,0.540208,1.400497,
        0.000000,0.542006,0.228337,
        0.000000,0.550762,-0.180771,
        0.000000,0.560531,2.378474,
        0.000000,0.561629,-0.924576,
        0.000000,0.567465,-2.679392,
        0.000000,0.588527,1.181755,
        0.000000,0.591643,-2.098364,
        0.000000,0.594963,2.997575,
        0.000000,0.599643,-3.045235,
        0.000000,0.599962,2.773556,
        0.000000,0.602620,0.029270,
        0.000000,0.605711,0.598454,
        0.000000,0.607708,-1.541026,
        0.000000,0.609360,-2.464379,
        0.000000,0.610738,-1.326786,
        0.000000,0.616968,0.820489,
        0.000000,0.626379,1.773610,
        0.000000,0.628626,-1.751180,
        0.000000,0.630671,0.391949,
        0.000000,0.632622,-0.737642,
        0.000000,0.632708,2.566722,
        0.000000,0.632797,1.989858,
        0.000000,0.634455,-0.534790,
        0.000000,0.636205,1.561170,
        0.000000,0.637681,-1.123423,
        0.000000,0.641441,-2.842109,
        0.000000,0.644807,-0.327139,
        0.000000,0.652440,2.197129,
        0.000000,0.660897,-2.278593,
        0.000000,0.664008,1.360091,
        0.000000,0.670850,-1.945257,
        0.000000,0.671396,1.004859,
        0.000000,0.673985,0.203855,
        0.000000,0.680326,-0.133370,
        0.000000,0.685083,-0.942363,
        0.000000,0.688204,-2.632204,
        0.000000,0.689777,2.385011,
        0.000000,0.700065,3.115430,
        0.000000,0.704080,2.911636,
        0.000000,0.712883,1.174216,
        0.000000,0.715701,0.699906,
        0.000000,0.718851,-1.442215,
        0.000000,0.721019,2.721536,
        0.000000,0.725300,-2.975134,
        0.000000,0.727136,-1.633288,
        0.000000,0.727399,-2.120567,
        0.000000,0.730751,0.522488,
        0.000000,0.732892,-1.254330,
        0.000000,0.734333,-2.446184,
        0.000000,0.735125,0.037270,
        0.000000,0.737893,1.866463,
        0.000000,0.740650,1.681075,
        0.000000,0.745971,-0.781286,
        0.000000,0.749432,-0.442559,
        0.000000,0.750102,2.050636,
        0.000000,0.751046,0.863821,
        0.000000,0.757018,-0.608833,
        0.000000,0.757046,-1.816019,
        0.000000,0.757389,1.501184,
        0.000000,0.759800,2.546181,
        0.000000,0.759859,0.351923,
        0.000000,0.761141,-2.781102,
        0.000000,0.769035,-1.078574,
        0.000000,0.770166,-0.266604,
        0.000000,0.772218,2.227380,
        0.000000,0.788382,1.331051,
        0.000000,0.790269,-2.282409,
        0.000000,0.803410,-1.984567,
        0.000000,0.803813,0.191520,
        0.000000,0.806832,3.022155,
        0.000000,0.807476,-2.597675,
        0.000000,0.812512,-0.107335,
        0.000000,0.814161,1.010949,
        0.000000,0.817732,-3.084486,
        0.000000,0.817846,2.390688,
        0.000000,0.819323,2.845960,
        0.000000,0.819430,-0.915017,
        0.000000,0.829198,-1.532491,
        0.000000,0.829531,0.665695,
        0.000000,0.829816,1.167147,
        0.000000,0.833613,-1.357492,
        0.000000,0.842945,-2.907318,
        0.000000,0.844883,2.679708,
        0.000000,0.848189,-1.705003,
        0.000000,0.849888,1.767901,
        0.000000,0.851126,0.493370,
        0.000000,0.852911,1.933394,
        0.000000,0.857516,-2.432863,
        0.000000,0.858261,-1.191729,
        0.000000,0.858279,-2.138895,
        0.000000,0.859723,-0.734928,
        0.000000,0.859754,1.603292,
        0.000000,0.861655,-0.394821,
        0.000000,0.864300,0.043549,
        0.000000,0.866940,0.851229,
        0.000000,0.869748,2.095969,
        0.000000,0.877352,-0.574668,
        0.000000,0.879254,-2.733338,
        0.000000,0.882554,1.443757,
        0.000000,0.884060,0.334199,
        0.000000,0.885626,-1.866061,
        0.000000,0.889834,2.529957,
        0.000000,0.893140,2.251698,
        0.000000,0.897463,-0.240822,
        0.000000,0.898758,-1.031948,
        0.000000,0.913918,1.287618,
        0.000000,0.916555,-2.287173,
        0.000000,0.918305,3.107845,
        0.000000,0.919522,2.948277,
        0.000000,0.927138,-2.569686,
        0.000000,0.929031,0.185383,
        0.000000,0.933766,-3.015917,
        0.000000,0.934873,-2.010822,
        0.000000,0.935636,-1.445855,
        0.000000,0.936702,-0.866461,
        0.000000,0.937395,0.729108,
        0.000000,0.937910,2.787575,
        0.000000,0.939144,-0.092962,
        0.000000,0.941718,1.125085,
        0.000000,0.942873,-1.607680,
        0.000000,0.943701,0.980328,
        0.000000,0.947774,2.395256,
        0.000000,0.952433,0.584347,
        0.000000,0.954456,-1.287293,
        0.000000,0.958731,-2.852552,
        0.000000,0.961838,1.839284,
        0.000000,0.961898,-0.469425,
        0.000000,0.965426,1.686615,
        0.000000,0.970108,1.990771,
        0.000000,0.972820,-1.762780,
        0.000000,0.973500,-0.711881,
        0.000000,0.975253,2.641082,
        0.000000,0.977824,0.441928,
        0.000000,0.981485,-2.424766,
        0.000000,0.981877,1.537100,
        0.000000,0.982609,-1.136920,
        0.000000,0.984265,-2.156073,
        0.000000,0.989150,0.045493,
        0.000000,0.993052,2.137478,
        0.000000,0.997766,-2.693871,
        0.000000,0.999847,-0.334738,
        0.000000,1.001920,0.848054,
        0.000000,1.006146,1.389734,
        0.000000,1.012180,0.304066,
        0.000000,1.016938,2.270674,
        0.000000,1.017371,-0.979790,
        0.000000,1.019495,-1.894487,
        0.000000,1.023796,-0.583691,
        0.000000,1.026525,2.506945,
        0.000000,1.027004,-0.203653,
        0.000000,1.028027,3.036573,
        0.000000,1.031435,1.241021,
        0.000000,1.032316,2.884221,
        0.000000,1.035863,-3.110475,
        0.000000,1.042085,-2.292950,
        0.000000,1.043272,-1.522986,
        0.000000,1.046265,-2.958847,
        0.000000,1.049272,-1.376344,
        0.000000,1.050039,-2.552086,
        0.000000,1.053682,0.689863,
        0.000000,1.055458,0.170394,
        0.000000,1.056622,-0.831866,
        0.000000,1.057515,1.097998,
        0.000000,1.058706,-2.036438,
        0.000000,1.061389,-1.672248,
        0.000000,1.063758,2.743093,
        0.000000,1.070157,-0.073422,
        0.000000,1.072154,1.761506,
        0.000000,1.074029,1.909193,
        0.000000,1.074204,0.545276,
        0.000000,1.074714,-2.805147,
        0.000000,1.076262,0.966104,
        0.000000,1.077642,-1.227073,
        0.000000,1.084151,1.618525,
        0.000000,1.085085,2.377673,
        0.000000,1.086822,-0.441112,
        0.000000,1.087142,2.058815,
        0.000000,1.098383,-0.696430,
        0.000000,1.099138,-1.084033,
        0.000000,1.103208,1.477773,
        0.000000,1.103783,0.408254,
        0.000000,1.106275,-2.420800,
        0.000000,1.108079,2.612943,
        0.000000,1.109345,-2.172718,
        0.000000,1.115957,-1.787597,
        0.000000,1.118930,0.045983,
        0.000000,1.119777,0.834702,
        0.000000,1.121943,-2.669042,
        0.000000,1.124930,2.219917,
        0.000000,1.125217,1.336294,
        0.000000,1.130958,2.972931,
        0.000000,1.131400,-0.306629,
        0.000000,1.137135,-0.936514,
        0.000000,1.139125,-1.924528,
        0.000000,1.139877,0.276382,
        0.000000,1.142003,-3.058022,
        0.000000,1.147363,-0.558448,
        0.000000,1.150348,-0.176418,
        0.000000,1.151198,1.197016,
        0.000000,1.153930,-1.597104,
        0.000000,1.155389,2.837148,
        0.000000,1.155458,-1.460623,
        0.000000,1.157331,-2.906362,
        0.000000,1.158594,2.487618,
        0.000000,1.162103,-1.322553,
        0.000000,1.165407,3.092861,
        0.000000,1.168025,0.717075,
        0.000000,1.168406,-2.301519,
        0.000000,1.174782,-2.538755,
        0.000000,1.177998,1.061256,
        0.000000,1.178898,1.827305,
        0.000000,1.181331,-2.060330,
        0.000000,1.182540,-0.798952,
        0.000000,1.184569,0.147517,
        0.000000,1.184734,1.968919,
        0.000000,1.184804,0.593386,
        0.000000,1.188278,1.690808,
        0.000000,1.194360,2.110440,
        0.000000,1.194776,2.710455,
        0.000000,1.197745,-1.173705,
        0.000000,1.199898,-2.777257,
        0.000000,1.201164,2.355931,
        0.000000,1.203760,1.556404,
        0.000000,1.205845,-0.415993,
        0.000000,1.206585,0.933648,
        0.000000,1.206950,-0.037248,
        0.000000,1.209341,0.471162,
        0.000000,1.213483,-1.039417,
        0.000000,1.218146,-0.672643,
        0.000000,1.221608,1.421669,
        0.000000,1.226043,-1.693533,
        0.000000,1.232870,-1.822773,
        0.000000,1.234375,-2.421194,
        0.000000,1.234531,-2.190148,
        0.000000,1.239414,0.348959,
        0.000000,1.240183,2.587255,
        0.000000,1.242312,2.941234,
        0.000000,1.244075,1.286644,
        0.000000,1.245961,-2.997461,
        0.000000,1.246495,-2.646872,
        0.000000,1.247736,2.242810,
        0.000000,1.253604,0.812222,
        0.000000,1.257816,-0.894317,
        0.000000,1.258247,-1.953634,
        0.000000,1.258993,-0.284099,
        0.000000,1.262380,-0.157609,
        0.000000,1.262791,-1.549229,
        0.000000,1.263772,-0.533073,
        0.000000,1.266731,-1.412461,
        0.000000,1.268591,-3.122146,
        0.000000,1.269614,-1.275246,
        0.000000,1.270745,1.152533,
        0.000000,1.273606,0.224545,
        0.000000,1.279716,2.812101,
        0.000000,1.283786,2.462124,
        0.000000,1.284798,0.078209,
        0.000000,1.285402,-2.882049,
        0.000000,1.286936,1.886053,
        0.000000,1.289408,0.683122,
        0.000000,1.294736,1.756184,
        0.000000,1.296978,-2.311511,
        0.000000,1.298546,2.020601,
        0.000000,1.300924,1.022478,
        0.000000,1.303305,3.043152,
        0.000000,1.304210,-2.084225,
        0.000000,1.305241,-2.530249,
        0.000000,1.307513,1.628003,
        0.000000,1.310722,0.558267,
        0.000000,1.314142,2.139329,
        0.000000,1.316087,-0.777971,
        0.000000,1.317907,-1.121023,
        0.000000,1.321019,-2.749776,
        0.000000,1.321609,2.687717,
        0.000000,1.322568,1.499415,
        0.000000,1.322972,-0.395699,
        0.000000,1.325286,-0.994192,
        0.000000,1.329146,-0.648513,
        0.000000,1.329795,-0.042570,
        0.000000,1.331651,2.344497,
        0.000000,1.336522,0.435098,
        0.000000,1.339410,0.900087,
        0.000000,1.339550,1.370353,
        0.000000,1.340441,-1.731580,
        0.000000,1.351855,-1.855234,
        0.000000,1.360161,-3.018404,
        0.000000,1.361031,-1.613971,
        0.000000,1.361196,2.914780,
        0.000000,1.361998,-2.207254,
        0.000000,1.362302,1.239384,
        0.000000,1.365228,2.563808,
        0.000000,1.365452,0.310522,
        0.000000,1.367028,-2.420041,
        0.000000,1.373422,-1.484281,
        0.000000,1.374245,-1.226417,
        0.000000,1.375020,-0.170393,
        0.000000,1.376560,-1.355481,
        0.000000,1.377070,-1.980984,
        0.000000,1.377578,-2.636339,
        0.000000,1.378096,0.774626,
        0.000000,1.379086,-0.508192,
        0.000000,1.382562,0.177141,
        0.000000,1.388849,2.233847,
        0.000000,1.389268,-0.886349,
        0.000000,1.391324,1.109396,
        0.000000,1.391699,-0.288725,
        0.000000,1.392869,3.136903,
        0.000000,1.393612,1.952016,
        0.000000,1.395545,-2.850807,
        0.000000,1.401926,1.822930,
        0.000000,1.402149,2.788898,
        0.000000,1.409552,0.647980,
        0.000000,1.410840,0.056219,
        0.000000,1.411944,2.444377,
        0.000000,1.413639,1.697982,
        0.000000,1.426479,1.573427,
        0.000000,1.426725,2.102242,
        0.000000,1.427868,0.984637,
        0.000000,1.429580,-2.107232,
        0.000000,1.431375,-2.313269,
        0.000000,1.431907,3.015410,
        0.000000,1.432325,-1.029054,
        0.000000,1.433341,0.524268,
        0.000000,1.435181,-0.739771,
        0.000000,1.437208,-2.529085,
        0.000000,1.439041,-0.617966,
        0.000000,1.441473,1.448544,
        0.000000,1.444527,2.664744,
        0.000000,1.450513,-0.400625,
        0.000000,1.451907,-0.066823,
        0.000000,1.452754,-2.738631,
        0.000000,1.454061,-2.950416,
        0.000000,1.456736,1.323154,
        0.000000,1.458370,0.400433,
        0.000000,1.459314,-1.765003,
        0.000000,1.459902,-1.145835,
        0.000000,1.464952,2.329245,
        0.000000,1.466050,0.862104,
        0.000000,1.471330,-1.886160,
        0.000000,1.475119,-1.648255,
        0.000000,1.480878,1.194576,
        0.000000,1.480942,0.273235,
        0.000000,1.481086,2.890269,
        0.000000,1.484118,-1.529156,
        0.000000,1.484741,-1.279219,
        0.000000,1.487681,-1.406069,
        0.000000,1.487886,-3.066037,
        0.000000,1.490555,2.543284,
        0.000000,1.492424,-0.190932,
        0.000000,1.495002,-0.839810,
        0.000000,1.495297,2.007188,
        0.000000,1.496414,-2.009267,
        0.000000,1.498261,-2.425270,
        0.000000,1.498907,-2.207327,
        0.000000,1.500204,0.737638,
        0.000000,1.502469,0.151156,
        0.000000,1.507448,-0.514782,
        0.000000,1.509287,2.204647,
        0.000000,1.510518,1.884920,
        0.000000,1.511948,-2.632360,
        0.000000,1.515235,1.068565,
        0.000000,1.522237,3.102018,
        0.000000,1.522373,1.764991,
        0.000000,1.522416,-0.954956,
        0.000000,1.523102,2.765920,
        0.000000,1.524811,-2.844718,
        0.000000,1.527261,-0.304916,
        0.000000,1.529788,0.613208,
        0.000000,1.533648,1.644808,
        0.000000,1.535018,0.028869,
        0.000000,1.541082,2.424608,
        0.000000,1.546010,1.523462,
        0.000000,1.551135,-0.642994,
        0.000000,1.553895,0.489607,
        0.000000,1.555041,0.947813,
        0.000000,1.559671,-1.074934,
        0.000000,1.560345,2.985899,
        0.000000,1.560487,-2.323400,
        0.000000,1.561127,1.401111,
        0.000000,1.567193,2.643172,
        0.000000,1.568892,2.098362,
        0.000000,1.571702,-2.098435,
        0.000000,1.572701,-2.528730,
        0.000000,1.573572,1.278593,
        0.000000,1.574849,-0.091505,
        0.000000,1.576334,0.365236,
        0.000000,1.577029,-2.962888,
        0.000000,1.577688,-1.814941,
        0.000000,1.579542,-0.755917,
        0.000000,1.579624,-1.199759,
        0.000000,1.581515,-0.419414,
        0.000000,1.584581,-2.738702,
        0.000000,1.586948,-1.696296,
        0.000000,1.588901,2.301496,
        0.000000,1.589936,0.825020,
        0.000000,1.590018,-1.941652,
        0.000000,1.596968,-1.328358,
        0.000000,1.597178,-1.575607,
        0.000000,1.600470,-1.453196,
        0.000000,1.601408,0.242409,
        0.000000,1.602259,1.152656,
        0.000000,1.603064,2.865774,
        0.000000,1.612085,1.968802,
        0.000000,1.612720,-3.080498,
        0.000000,1.616061,2.521531,
        0.000000,1.616900,-0.207457,
        0.000000,1.620988,-0.875920,
        0.000000,1.622114,0.701509,
        0.000000,1.622749,0.121426,
        0.000000,1.622993,-2.221629,
        0.000000,1.629346,-0.542669,
        0.000000,1.630007,1.844552,
        0.000000,1.634747,-2.426198,
        0.000000,1.642346,1.722617,
        0.000000,1.643925,1.034629,
        0.000000,1.644848,2.743774,
        0.000000,1.645258,2.188109,
        0.000000,1.645417,-2.634725,
        0.000000,1.647395,3.085437,
        0.000000,1.648538,-2.855044,
        0.000000,1.649932,0.576808,
        0.000000,1.652224,-0.996105,
        0.000000,1.652871,1.600681,
        0.000000,1.657997,-0.325127,
        0.000000,1.661936,0.001037,
        0.000000,1.665187,1.477402,
        0.000000,1.665496,2.397872,
        0.000000,1.673267,-0.665453,
        0.000000,1.673557,0.452526,
        0.000000,1.679277,-1.122421,
        0.000000,1.679696,0.912482,
        0.000000,1.682420,1.353922,
        0.000000,1.685031,-2.009038,
        0.000000,1.685196,2.967046,
        0.000000,1.690107,2.620910,
        0.000000,1.690453,-2.123076,
        0.000000,1.691693,-1.871572,
        0.000000,1.691817,2.061688,
        0.000000,1.692644,1.237670,
        0.000000,1.696386,0.328123,
        0.000000,1.697162,-1.249026,
        0.000000,1.697488,-2.322785,
        0.000000,1.698726,-2.973576,
        0.000000,1.699172,-1.749717,
        0.000000,1.704478,-0.106295,
        0.000000,1.707244,-0.444770,
        0.000000,1.707500,-2.531497,
        0.000000,1.707714,-0.787321,
        0.000000,1.708430,-1.627552,
        0.000000,1.710583,-1.376890,
        0.000000,1.712415,-2.747921,
        0.000000,1.713125,-1.503073,
        0.000000,1.713452,0.789419,
        0.000000,1.720612,2.279888,
        0.000000,1.724852,0.202831,
        0.000000,1.724868,2.846246,
        0.000000,1.728940,1.931535,
        0.000000,1.732838,1.127584,
        0.000000,1.735719,-3.092485,
        0.000000,1.739676,2.496017,
        0.000000,1.739971,-0.228876,
        0.000000,1.743328,0.663347,
        0.000000,1.744927,-0.911892,
        0.000000,1.745952,0.084733,
        0.000000,1.748077,1.804354,
        0.000000,1.752959,-0.567548,
        0.000000,1.760169,1.678132,
        0.000000,1.763887,-2.218450,
        0.000000,1.766737,2.722752,
        0.000000,1.767999,1.004737,
        0.000000,1.769149,1.550298,
        0.000000,1.770119,-2.427175,
        0.000000,1.770460,2.154208,
        0.000000,1.770514,3.070982,
        0.000000,1.770735,0.537245,
        0.000000,1.772673,-2.864760,
        0.000000,1.773527,-1.039582,
        0.000000,1.776419,-2.642421,
        0.000000,1.780769,1.416330,
        0.000000,1.786634,-0.346247,
        0.000000,1.788829,-1.947553,
        0.000000,1.793737,0.410270,
        0.000000,1.794553,2.374618,
        0.000000,1.794699,-0.694061,
        0.000000,1.797469,-1.169105,
        0.000000,1.803705,0.881076,
        0.000000,1.804124,1.264249,
        0.000000,1.807825,-2.090455,
        0.000000,1.808822,-1.817375,
        0.000000,1.809020,2.949741,
        0.000000,1.812466,2.597595,
        0.000000,1.813159,2.025395,
        0.000000,1.813816,-1.298435,
        0.000000,1.815016,0.278951,
        0.000000,1.816510,-0.008631,
        0.000000,1.817772,-1.689625,
        0.000000,1.821360,-2.984352,
        0.000000,1.822312,-0.127211,
        0.000000,1.823386,-1.429807,
        0.000000,1.823834,-1.560483,
        0.000000,1.832060,-0.821627,
        0.000000,1.832690,-0.468440,
        0.000000,1.833945,-2.320278,
        0.000000,1.834787,0.751968,
        0.000000,1.840579,-2.536445,
        0.000000,1.841033,-2.756932,
        0.000000,1.846350,1.893983,
        0.000000,1.846644,1.113766,
        0.000000,1.846833,2.826543,
        0.000000,1.847325,2.249051,
        0.000000,1.854119,0.134161,
        0.000000,1.858863,-3.106022,
        0.000000,1.865046,1.763384,
        0.000000,1.865619,-0.952036,
        0.000000,1.865912,0.624453,
        0.000000,1.866694,2.473686,
        0.000000,1.867524,-0.240083,
        0.000000,1.875561,1.632831,
        0.000000,1.876049,-0.594378,
        0.000000,1.880509,1.500204,
        0.000000,1.883504,1.360457,
        0.000000,1.886971,-2.196665,
        0.000000,1.888979,0.982922,
        0.000000,1.889815,2.701570,
        0.000000,1.892144,-2.005340,
        0.000000,1.892348,0.496522,
        0.000000,1.893281,-1.083976,
        0.000000,1.893828,3.053453,
        0.000000,1.895117,2.121107,
        0.000000,1.898741,-2.875735,
        0.000000,1.904429,-2.428029,
        0.000000,1.908408,-2.648903,
        0.000000,1.911828,-0.364489,
        0.000000,1.912224,0.366565,
        0.000000,1.914854,1.214468,
        0.000000,1.915467,-1.883143,
        0.000000,1.915763,-0.724679,
        0.000000,1.915790,-1.216366,
        0.000000,1.922256,2.347806,
        0.000000,1.922616,0.844633,
        0.000000,1.925185,0.231170,
        0.000000,1.927474,-1.754616,
        0.000000,1.928342,-1.351174,
        0.000000,1.933328,-1.623328,
        0.000000,1.933878,0.034150,
        0.000000,1.934585,-1.487979,
        0.000000,1.934802,2.927634,
        0.000000,1.934862,1.989776,
        0.000000,1.936327,-0.099350,
        0.000000,1.938088,2.578507,
        0.000000,1.945677,-2.999805,
        0.000000,1.954572,-0.859208,
        0.000000,1.957868,-0.490784,
        0.000000,1.958989,0.712772,
        0.000000,1.960212,-2.307561,
        0.000000,1.964774,1.855507,
        0.000000,1.964971,-2.100533,
        0.000000,1.969112,1.088273,
        0.000000,1.971438,-2.765415,
        0.000000,1.972312,2.805134,
        0.000000,1.974950,2.220596,
        0.000000,1.975135,-2.539127,
        0.000000,1.982107,1.720829,
        0.000000,1.983195,-3.127984,
        0.000000,1.983318,-0.247182,
        0.000000,1.986720,-0.993220,
        0.000000,1.990381,1.586005,
        0.000000,1.990470,0.580283,
        0.000000,1.990523,1.450326,
        0.000000,1.993909,1.312897,
        0.000000,1.994694,2.452813,
        0.000000,1.999226,-0.620717,
        0.000000,2.005145,0.941012,
        0.000000,2.007301,0.142337,
        0.000000,2.014149,0.444138,
        0.000000,2.014338,-1.127392,
        0.000000,2.018131,3.021778,
        0.000000,2.019676,-1.980944,
        0.000000,2.019912,2.686612,
        0.000000,2.020545,2.088672,
        0.000000,2.026074,0.300945,
        0.000000,2.026196,-2.889057,
        0.000000,2.031959,-1.268178,
        0.000000,2.032917,-2.199917,
        0.000000,2.033572,-1.842884,
        0.000000,2.034712,-2.419505,
        0.000000,2.035818,-0.379479,
        0.000000,2.036004,-0.757062,
        0.000000,2.040501,-1.411488,
        0.000000,2.042874,-1.701367,
        0.000000,2.043060,-2.653746,
        0.000000,2.043729,-1.557938,
        0.000000,2.048538,0.806623,
        0.000000,2.049006,1.199930,
        0.000000,2.049502,-0.137989,
        0.000000,2.052181,2.326959,
        0.000000,2.052794,-0.008526,
        0.000000,2.057608,1.951726,
        0.000000,2.065568,2.567730,
        0.000000,2.070371,2.879808,
        0.000000,2.071256,-3.018961,
        0.000000,2.079746,-0.897699,
        0.000000,2.082784,1.042403,
        0.000000,2.083752,-0.511086,
        0.000000,2.084401,1.810502,
        0.000000,2.085731,0.671081,
        0.000000,2.097496,1.373172,
        0.000000,2.099185,1.667006,
        0.000000,2.100482,-2.301084,
        0.000000,2.103118,2.194745,
        0.000000,2.103223,1.521685,
        0.000000,2.103695,-2.777707,
        0.000000,2.107523,3.129581,
        0.000000,2.108277,-2.095907,
        0.000000,2.108891,-2.532166,
        0.000000,2.111003,-0.268632,
        0.000000,2.112276,-1.027246,
        0.000000,2.114380,0.531457,
        0.000000,2.119166,0.100552,
        0.000000,2.119564,2.754818,
        0.000000,2.122474,-0.648314,
        0.000000,2.125183,2.445598,
        0.000000,2.129501,0.238290,
        0.000000,2.131178,0.386663,
        0.000000,2.131223,-1.174311,
        0.000000,2.134269,-1.948042,
        0.000000,2.134684,0.907065,
        0.000000,2.137615,2.988806,
        0.000000,2.145635,-1.327708,
        0.000000,2.146467,2.055081,
        0.000000,2.146582,-1.793788,
        0.000000,2.149210,-1.482321,
        0.000000,2.152901,-1.636788,
        0.000000,2.156337,1.151558,
        0.000000,2.156474,-2.912065,
        0.000000,2.157194,-0.792563,
        0.000000,2.166724,-0.401638,
        0.000000,2.166980,-0.134879,
        0.000000,2.173644,-2.403429,
        0.000000,2.175711,-2.655523,
        0.000000,2.177132,2.600419,
        0.000000,2.178879,0.767722,
        0.000000,2.180752,1.908485,
        0.000000,2.181338,2.313721,
        0.000000,2.185580,1.287673,
        0.000000,2.195400,-2.217546,
        0.000000,2.200418,-3.056442,
        0.000000,2.201829,-0.001735,
        0.000000,2.203804,1.756176,
        0.000000,2.211602,1.443930,
        0.000000,2.212686,-0.544809,
        0.000000,2.213937,0.621892,
        0.000000,2.214075,1.600691,
        0.000000,2.216256,1.020029,
        0.000000,2.219060,2.886631,
        0.000000,2.223375,-0.906185,
        0.000000,2.226002,-1.060248,
        0.000000,2.226413,-2.064034,
        0.000000,2.230554,2.171334,
        0.000000,2.232413,2.735777,
        0.000000,2.233709,3.070830,
        0.000000,2.234383,-2.792065,
        0.000000,2.237183,0.468799,
        0.000000,2.237695,-0.271388,
        0.000000,2.239419,0.151693,
        0.000000,2.243347,-2.521049,
        0.000000,2.244035,-1.229456,
        0.000000,2.244355,0.309899,
        0.000000,2.245175,2.460236,
        0.000000,2.245628,-1.899216,
        0.000000,2.245973,-0.702234,
        0.000000,2.255601,-1.728131,
        0.000000,2.256702,-1.401367,
        0.000000,2.261466,-1.557500,
        0.000000,2.267390,0.878499,
        0.000000,2.272099,2.018594,
        0.000000,2.278618,1.170386,
        0.000000,2.283304,-2.345584,
        0.000000,2.285175,-2.943090,
        0.000000,2.293000,-0.124451,
        0.000000,2.294058,-0.421070,
        0.000000,2.303283,1.855938,
        0.000000,2.304968,2.608978,
        0.000000,2.305347,2.311346,
        0.000000,2.307188,-2.656134,
        0.000000,2.307301,1.336200,
        0.000000,2.309655,0.726020,
        0.000000,2.317634,-2.185225,
        0.000000,2.321510,1.685664,
        0.000000,2.324376,-3.111006,
        0.000000,2.324731,1.511186,
        0.000000,2.329587,0.035498,
        0.000000,2.330158,2.977264,
        0.000000,2.335096,-0.588468,
        0.000000,2.336046,-1.115799,
        0.000000,2.338060,2.803613,
        0.000000,2.338083,-0.943400,
        0.000000,2.340210,0.562145,
        0.000000,2.341931,1.023926,
        0.000000,2.343265,-0.779061,
        0.000000,2.343434,-2.013692,
        0.000000,2.353029,-1.310529,
        0.000000,2.354868,0.211543,
        0.000000,2.355120,0.388593,
        0.000000,2.355165,-1.833173,
        0.000000,2.355462,2.151441,
        0.000000,2.362948,-2.487477,
        0.000000,2.363101,-2.810456,
        0.000000,2.363394,-0.272918,
        0.000000,2.364509,-1.644533,
        0.000000,2.370465,2.466824,
        0.000000,2.386110,-1.481351,
        0.000000,2.394960,0.862133,
        0.000000,2.395716,1.977184,
        0.000000,2.396339,1.200900,
        0.000000,2.407941,-2.316753,
        0.000000,2.409450,-2.988432,
        0.000000,2.417202,-0.103363,
        0.000000,2.417386,-0.443097,
        0.000000,2.417950,2.666662,
        0.000000,2.423190,1.789783,
        0.000000,2.423692,1.392696,
        0.000000,2.431122,3.086379,
        0.000000,2.431406,-2.647820,
        0.000000,2.431584,2.307631,
        0.000000,2.434090,1.592818,
        0.000000,2.435993,0.683844,
        0.000000,2.440898,-2.133300,
        0.000000,2.442222,-1.195837,
        0.000000,2.443112,2.883761,
        0.000000,2.449191,-0.633651,
        0.000000,2.450821,0.084741,
        0.000000,2.453499,-1.008861,
        0.000000,2.460010,-1.938011,
        0.000000,2.460199,-0.822008,
        0.000000,2.460902,0.490079,
        0.000000,2.461774,-1.734913,
        0.000000,2.463986,1.034407,
        0.000000,2.467647,0.287667,
        0.000000,2.479849,2.126925,
        0.000000,2.485907,-2.835471,
        0.000000,2.485945,-1.374853,
        0.000000,2.487747,2.505912,
        0.000000,2.487957,-0.274279,
        0.000000,2.489741,-2.468428,
        0.000000,2.512538,1.239900,
        0.000000,2.517528,1.923289,
        0.000000,2.518454,-1.565852,
        0.000000,2.519348,0.843503,
        0.000000,2.522766,-3.059231,
        0.000000,2.528686,2.737228,
        0.000000,2.534798,-2.266671,
        0.000000,2.536533,1.466274,
        0.000000,2.538504,1.699812,
        0.000000,2.539516,-0.476942,
        0.000000,2.539532,-0.074548,
        0.000000,2.543197,2.990252,
        0.000000,2.554022,2.320619,
        0.000000,2.556919,-2.655005,
        0.000000,2.558146,0.627191,
        0.000000,2.558223,-1.164189,
        0.000000,2.563452,-2.044028,
        0.000000,2.568462,0.150655,
        0.000000,2.569130,-0.700837,
        0.000000,2.570050,-1.804492,
        0.000000,2.572710,-0.934665,
        0.000000,2.575994,0.391718,
        0.000000,2.584774,1.046703,
        0.000000,2.601432,2.102535,
        0.000000,2.603559,2.555114,
        0.000000,2.604150,-2.885208,
        0.000000,2.608630,-1.378011,
        0.000000,2.611256,-0.279110,
        0.000000,2.617344,-2.439480,
        0.000000,2.627402,1.293552,
        0.000000,2.635396,1.848850,
        0.000000,2.635686,2.834950,
        0.000000,2.637087,3.132730,
        0.000000,2.642029,0.813274,
        0.000000,2.645260,1.570203,
        0.000000,2.646182,-1.614036,
        0.000000,2.660113,-2.186513,
        0.000000,2.660483,-0.034608,
        0.000000,2.661713,-0.529511,
        0.000000,2.672720,-1.110422,
        0.000000,2.676626,2.335353,
        0.000000,2.676638,0.540364,
        0.000000,2.676780,-1.901158,
        0.000000,2.681411,-2.667965,
        0.000000,2.682152,0.245751,
        0.000000,2.682859,-0.817071,
        0.000000,2.705107,1.062142,
        0.000000,2.717705,2.625995,
        0.000000,2.721001,2.060097,
        0.000000,2.722319,-2.954069,
        0.000000,2.729840,-1.368237,
        0.000000,2.734349,-0.289298,
        0.000000,2.738409,2.975671,
        0.000000,2.740366,1.374341,
        0.000000,2.744633,-2.391311,
        0.000000,2.747398,1.727166,
        0.000000,2.763799,0.760372,
        0.000000,2.768238,-1.679243,
        0.000000,2.779245,-2.046268,
        0.000000,2.780208,-0.626453,
        0.000000,2.780471,0.027732,
        0.000000,2.784696,-1.013128,
        0.000000,2.789817,0.395431,
        0.000000,2.801154,2.353143,
        0.000000,2.807091,-2.687477,
        0.000000,2.825762,1.085035,
        0.000000,2.830209,2.734943,
        0.000000,2.833517,-3.085508,
        0.000000,2.838301,1.980061,
        0.000000,2.849871,1.512012,
        0.000000,2.850089,-1.343899,
        0.000000,2.856905,-0.318441,
        0.000000,2.870445,-2.291001,
        0.000000,2.885017,0.655562,
        0.000000,2.886428,-1.793217,
        0.000000,2.890337,-0.825633,
        0.000000,2.898932,0.139973,
        0.000000,2.931927,-2.728768,
        0.000000,2.932096,2.377790,
        0.000000,2.937409,2.947245,
        0.000000,2.948048,1.133127,
        0.000000,2.950896,1.797889,
        0.000000,2.970019,-1.282382,
        0.000000,2.978333,-0.412090,
        0.000000,2.996437,-2.089121,
        0.000000,3.011274,0.434110,
        0.000000,3.050410,3.138492,
        0.000000,3.064492,1.647259,
        0.000000,3.088868,-0.989501,
        0.104720,0.053884,-2.408394,
        0.104720,0.077645,1.453433,
        0.104720,0.088408,-0.311160,
        0.104720,0.142314,2.590387,
        0.104720,0.147206,-1.385164,
        0.104720,0.163506,0.573203,
        0.104720,0.165485,-2.875396,
        0.104720,0.182564,-2.125821,
        0.104720,0.190569,1.929141,
        0.104720,0.201102,1.249082,
        0.104720,0.201397,-0.716646,
        0.104720,0.210568,-0.069559,
        0.104720,0.251672,2.848900,
        0.104720,0.266262,-1.678068,
        0.104720,0.269108,-1.215563,
        0.104720,0.273777,-2.506756,
        0.104720,0.279854,0.814581,
        0.104720,0.281664,-2.969147,
        0.104720,0.283625,2.405603,
        0.104720,0.284058,0.357350,
        0.104720,0.294107,1.593880,
        0.104720,0.302491,-0.400410,
        0.104720,0.308059,2.021491,
        0.104720,0.312219,-2.072057,
        0.104720,0.321905,-0.829121,
        0.104720,0.329984,1.201621,
        0.104720,0.339714,-0.013712,
        0.104720,0.368784,2.974974,
        0.104720,0.376428,-1.495532,
        0.104720,0.378103,2.646436,
        0.104720,0.385088,0.590355,
        0.104720,0.386248,-2.690943,
        0.104720,0.389851,-2.340031,
        0.104720,0.391502,-1.171374,
        0.104720,0.398249,-1.812694,
        0.104720,0.400542,-2.995729,
        0.104720,0.404273,0.918163,
        0.104720,0.405617,1.773533,
        0.104720,0.409009,-0.570243,
        0.104720,0.410700,0.270433,
        0.104720,0.415893,1.464890,
        0.104720,0.420680,2.088164,
        0.104720,0.424766,-0.258729,
        0.104720,0.426507,2.367872,
        0.104720,0.441292,-0.890727,
        0.104720,0.451839,-2.078193,
        0.104720,0.459270,1.191562,
        0.104720,0.470622,0.014002,
        0.104720,0.481292,-2.529098,
        0.104720,0.482460,2.849843,
        0.104720,0.492205,-1.397963,
        0.104720,0.493314,3.117264,
        0.104720,0.497870,0.727828,
        0.104720,0.499783,-1.654379,
        0.104720,0.502651,0.463333,
        0.104720,0.505463,2.594557,
        0.104720,0.506104,-2.874910,
        0.104720,0.513579,-1.147354,
        0.104720,0.518569,1.644237,
        0.104720,0.519001,1.900488,
        0.104720,0.519277,-0.673809,
        0.104720,0.524769,-0.421955,
        0.104720,0.525564,-2.288838,
        0.104720,0.532541,0.978737,
        0.104720,0.534610,2.154808,
        0.104720,0.535020,-1.891846,
        0.104720,0.540208,1.400497,
        0.104720,0.542006,0.228337,
        0.104720,0.550762,-0.180771,
        0.104720,0.560531,2.378474,
        0.104720,0.561629,-0.924576,
        0.104720,0.567465,-2.679392,
        0.104720,0.588527,1.181755,
        0.104720,0.591643,-2.098364,
        0.104720,0.594963,2.997575,
        0.104720,0.599643,-3.045235,
        0.104720,0.599962,2.773556,
        0.104720,0.602620,0.029270,
        0.104720,0.605711,0.598454,
        0.104720,0.607708,-1.541026,
        0.104720,0.609360,-2.464379,
        0.104720,0.610738,-1.326786,
        0.104720,0.616968,0.820489,
        0.104720,0.626379,1.773610,
        0.104720,0.628626,-1.751180,
        0.104720,0.630671,0.391949,
        0.104720,0.632622,-0.737642,
        0.104720,0.632708,2.566722,
        0.104720,0.632797,1.989858,
        0.104720,0.634455,-0.534790,
        0.104720,0.636205,1.561170,
        0.104720,0.637681,-1.123423,
        0.104720,0.641441,-2.842109,
        0.104720,0.644807,-0.327139,
        0.104720,0.652440,2.197129,
        0.104720,0.660897,-2.278593,
        0.104720,0.664008,1.360091,
        0.104720,0.670850,-1.945257,
        0.104720,0.671396,1.004859,
        0.104720,0.673985,0.203855,
        0.104720,0.680326,-0.133370,
        0.104720,0.685083,-0.942363,
        0.104720,0.688204,-2.632204,
        0.104720,0.689777,2.385011,
        0.104720,0.700065,3.115430,
        0.104720,0.704080,2.911636,
        0.104720,0.712883,1.174216,
        0.104720,0.715701,0.699906,
        0.104720,0.718851,-1.442215,
        0.104720,0.721019,2.721536,
        0.104720,0.725300,-2.975134,
        0.104720,0.727136,-1.633288,
        0.104720,0.727399,-2.120567,
        0.104720,0.730751,0.522488,
        0.104720,0.732892,-1.254330,
        0.104720,0.734333,-2.446184,
        0.104720,0.735125,0.037270,
        0.104720,0.737893,1.866463,
        0.104720,0.740650,1.681075,
        0.104720,0.745971,-0.781286,
        0.104720,0.749432,-0.442559,
        0.104720,0.750102,2.050636,
        0.104720,0.751046,0.863821,
        0.104720,0.757018,-0.608833,
        0.104720,0.757046,-1.816019,
        0.104720,0.757389,1.501184,
        0.104720,0.759800,2.546181,
        0.104720,0.759859,0.351923,
        0.104720,0.761141,-2.781102,
        0.104720,0.769035,-1.078574,
        0.104720,0.770166,-0.266604,
        0.104720,0.772218,2.227380,
        0.104720,0.788382,1.331051,
        0.104720,0.790269,-2.282409,
        0.104720,0.803410,-1.984567,
        0.104720,0.803813,0.191520,
        0.104720,0.806832,3.022155,
        0.104720,0.807476,-2.597675,
        0.104720,0.812512,-0.107335,
        0.104720,0.814161,1.010949,
        0.104720,0.817732,-3.084486,
        0.104720,0.817846,2.390688,
        0.104720,0.819323,2.845960,
        0.104720,0.819430,-0.915017,
        0.104720,0.829198,-1.532491,
        0.104720,0.829531,0.665695,
        0.104720,0.829816,1.167147,
        0.104720,0.833613,-1.357492,
        0.104720,0.842945,-2.907318,
        0.104720,0.844883,2.679708,
        0.104720,0.848189,-1.705003,
        0.104720,0.849888,1.767901,
        0.104720,0.851126,0.493370,
        0.104720,0.852911,1.933394,
        0.104720,0.857516,-2.432863,
        0.104720,0.858261,-1.191729,
        0.104720,0.858279,-2.138895,
        0.104720,0.859723,-0.734928,
        0.104720,0.859754,1.603292,
        0.104720,0.861655,-0.394821,
        0.104720,0.864300,0.043549,
        0.104720,0.866940,0.851229,
        0.104720,0.869748,2.095969,
        0.104720,0.877352,-0.574668,
        0.104720,0.879254,-2.733338,
        0.104720,0.882554,1.443757,
        0.104720,0.884060,0.334199,
        0.104720,0.885626,-1.866061,
        0.104720,0.889834,2.529957,
        0.104720,0.893140,2.251698,
        0.104720,0.897463,-0.240822,
        0.104720,0.898758,-1.031948,
        0.104720,0.913918,1.287618,
        0.104720,0.916555,-2.287173,
        0.104720,0.918305,3.107845,
        0.104720,0.919522,2.948277,
        0.104720,0.927138,-2.569686,
        0.104720,0.929031,0.185383,
        0.104720,0.933766,-3.015917,
        0.104720,0.934873,-2.010822,
        0.104720,0.935636,-1.445855,
        0.104720,0.936702,-0.866461,
        0.104720,0.937395,0.729108,
        0.104720,0.937910,2.787575,
        0.104720,0.939144,-0.092962,
        0.104720,0.941718,1.125085,
        0.104720,0.942873,-1.607680,
        0.104720,0.943701,0.980328,
        0.104720,0.947774,2.395256,
        0.104720,0.952433,0.584347,
        0.104720,0.954456,-1.287293,
        0.104720,0.958731,-2.852552,
        0.104720,0.961838,1.839284,
        0.104720,0.961898,-0.469425,
        0.104720,0.965426,1.686615,
        0.104720,0.970108,1.990771,
        0.104720,0.972820,-1.762780,
        0.104720,0.973500,-0.711881,
        0.104720,0.975253,2.641082,
        0.104720,0.977824,0.441928,
        0.104720,0.981485,-2.424766,
        0.104720,0.981877,1.537100,
        0.104720,0.982609,-1.136920,
        0.104720,0.984265,-2.156073,
        0.104720,0.989150,0.045493,
        0.104720,0.993052,2.137478,
        0.104720,0.997766,-2.693871,
        0.104720,0.999847,-0.334738,
        0.104720,1.001920,0.848054,
        0.104720,1.006146,1.389734,
        0.104720,1.012180,0.304066,
        0.104720,1.016938,2.270674,
        0.104720,1.017371,-0.979790,
        0.104720,1.019495,-1.894487,
        0.104720,1.023796,-0.583691,
        0.104720,1.026525,2.506945,
        0.104720,1.027004,-0.203653,
        0.104720,1.028027,3.036573,
        0.104720,1.031435,1.241021,
        0.104720,1.032316,2.884221,
        0.104720,1.035863,-3.110475,
        0.104720,1.042085,-2.292950,
        0.104720,1.043272,-1.522986,
        0.104720,1.046265,-2.958847,
        0.104720,1.049272,-1.376344,
        0.104720,1.050039,-2.552086,
        0.104720,1.053682,0.689863,
        0.104720,1.055458,0.170394,
        0.104720,1.056622,-0.831866,
        0.104720,1.057515,1.097998,
        0.104720,1.058706,-2.036438,
        0.104720,1.061389,-1.672248,
        0.104720,1.063758,2.743093,
        0.104720,1.070157,-0.073422,
        0.104720,1.072154,1.761506,
        0.104720,1.074029,1.909193,
        0.104720,1.074204,0.545276,
        0.104720,1.074714,-2.805147,
        0.104720,1.076262,0.966104,
        0.104720,1.077642,-1.227073,
        0.104720,1.084151,1.618525,
        0.104720,1.085085,2.377673,
        0.104720,1.086822,-0.441112,
        0.104720,1.087142,2.058815,
        0.104720,1.098383,-0.696430,
        0.104720,1.099138,-1.084033,
        0.104720,1.103208,1.477773,
        0.104720,1.103783,0.408254,
        0.104720,1.106275,-2.420800,
        0.104720,1.108079,2.612943,
        0.104720,1.109345,-2.172718,
        0.104720,1.115957,-1.787597,
        0.104720,1.118930,0.045983,
        0.104720,1.119777,0.834702,
        0.104720,1.121943,-2.669042,
        0.104720,1.124930,2.219917,
        0.104720,1.125217,1.336294,
        0.104720,1.130958,2.972931,
        0.104720,1.131400,-0.306629,
        0.104720,1.137135,-0.936514,
        0.104720,1.139125,-1.924528,
        0.104720,1.139877,0.276382,
        0.104720,1.142003,-3.058022,
        0.104720,1.147363,-0.558448,
        0.104720,1.150348,-0.176418,
        0.104720,1.151198,1.197016,
        0.104720,1.153930,-1.597104,
        0.104720,1.155389,2.837148,
        0.104720,1.155458,-1.460623,
        0.104720,1.157331,-2.906362,
        0.104720,1.158594,2.487618,
        0.104720,1.162103,-1.322553,
        0.104720,1.165407,3.092861,
        0.104720,1.168025,0.717075,
        0.104720,1.168406,-2.301519,
        0.104720,1.174782,-2.538755,
        0.104720,1.177998,1.061256,
        0.104720,1.178898,1.827305,
        0.104720,1.181331,-2.060330,
        0.104720,1.182540,-0.798952,
        0.104720,1.184569,0.147517,
        0.104720,1.184734,1.968919,
        0.104720,1.184804,0.593386,
        0.104720,1.188278,1.690808,
        0.104720,1.194360,2.110440,
        0.104720,1.194776,2.710455,
        0.104720,1.197745,-1.173705,
        0.104720,1.199898,-2.777257,
        0.104720,1.201164,2.355931,
        0.104720,1.203760,1.556404,
        0.104720,1.205845,-0.415993,
        0.104720,1.206585,0.933648,
        0.104720,1.206950,-0.037248,
        0.104720,1.209341,0.471162,
        0.104720,1.213483,-1.039417,
        0.104720,1.218146,-0.672643,
        0.104720,1.221608,1.421669,
        0.104720,1.226043,-1.693533,
        0.104720,1.232870,-1.822773,
        0.104720,1.234375,-2.421194,
        0.104720,1.234531,-2.190148,
        0.104720,1.239414,0.348959,
        0.104720,1.240183,2.587255,
        0.104720,1.242312,2.941234,
        0.104720,1.244075,1.286644,
        0.104720,1.245961,-2.997461,
        0.104720,1.246495,-2.646872,
        0.104720,1.247736,2.242810,
        0.104720,1.253604,0.812222,
        0.104720,1.257816,-0.894317,
        0.104720,1.258247,-1.953634,
        0.104720,1.258993,-0.284099,
        0.104720,1.262380,-0.157609,
        0.104720,1.262791,-1.549229,
        0.104720,1.263772,-0.533073,
        0.104720,1.266731,-1.412461,
        0.104720,1.268591,-3.122146,
        0.104720,1.269614,-1.275246,
        0.104720,1.270745,1.152533,
        0.104720,1.273606,0.224545,
        0.104720,1.279716,2.812101,
        0.104720,1.283786,2.462124,
        0.104720,1.284798,0.078209,
        0.104720,1.285402,-2.882049,
        0.104720,1.286936,1.886053,
        0.104720,1.289408,0.683122,
        0.104720,1.294736,1.756184,
        0.104720,1.296978,-2.311511,
        0.104720,1.298546,2.020601,
        0.104720,1.300924,1.022478,
        0.104720,1.303305,3.043152,
        0.104720,1.304210,-2.084225,
        0.104720,1.305241,-2.530249,
        0.104720,1.307513,1.628003,
        0.104720,1.310722,0.558267,
        0.104720,1.314142,2.139329,
        0.104720,1.316087,-0.777971,
        0.104720,1.317907,-1.121023,
        0.104720,1.321019,-2.749776,
        0.104720,1.321609,2.687717,
        0.104720,1.322568,1.499415,
        0.104720,1.322972,-0.395699,
        0.104720,1.325286,-0.994192,
        0.104720,1.329146,-0.648513,
        0.104720,1.329795,-0.042570,
        0.104720,1.331651,2.344497,
        0.104720,1.336522,0.435098,
        0.104720,1.339410,0.900087,
        0.104720,1.339550,1.370353,
        0.104720,1.340441,-1.731580,
        0.104720,1.351855,-1.855234,
        0.104720,1.360161,-3.018404,
        0.104720,1.361031,-1.613971,
        0.104720,1.361196,2.914780,
        0.104720,1.361998,-2.207254,
        0.104720,1.362302,1.239384,
        0.104720,1.365228,2.563808,
        0.104720,1.365452,0.310522,
        0.104720,1.367028,-2.420041,
        0.104720,1.373422,-1.484281,
        0.104720,1.374245,-1.226417,
        0.104720,1.375020,-0.170393,
        0.104720,1.376560,-1.355481,
        0.104720,1.377070,-1.980984,
        0.104720,1.377578,-2.636339,
        0.104720,1.378096,0.774626,
        0.104720,1.379086,-0.508192,
        0.104720,1.382562,0.177141,
        0.104720,1.388849,2.233847,
        0.104720,1.389268,-0.886349,
        0.104720,1.391324,1.109396,
        0.104720,1.391699,-0.288725,
        0.104720,1.392869,3.136903,
        0.104720,1.393612,1.952016,
        0.104720,1.395545,-2.850807,
        0.104720,1.401926,1.822930,
        0.104720,1.402149,2.788898,
        0.104720,1.409552,0.647980,
        0.104720,1.410840,0.056219,
        0.104720,1.411944,2.444377,
        0.104720,1.413639,1.697982,
        0.104720,1.426479,1.573427,
        0.104720,1.426725,2.102242,
        0.104720,1.427868,0.984637,
        0.104720,1.429580,-2.107232,
        0.104720,1.431375,-2.313269,
        0.104720,1.431907,3.015410,
        0.104720,1.432325,-1.029054,
        0.104720,1.433341,0.524268,
        0.104720,1.435181,-0.739771,
        0.104720,1.437208,-2.529085,
        0.104720,1.439041,-0.617966,
        0.104720,1.441473,1.448544,
        0.104720,1.444527,2.664744,
        0.104720,1.450513,-0.400625,
        0.104720,1.451907,-0.066823,
        0.104720,1.452754,-2.738631,
        0.104720,1.454061,-2.950416,
        0.104720,1.456736,1.323154,
        0.104720,1.458370,0.400433,
        0.104720,1.459314,-1.765003,
        0.104720,1.459902,-1.145835,
        0.104720,1.464952,2.329245,
        0.104720,1.466050,0.862104,
        0.104720,1.471330,-1.886160,
        0.104720,1.475119,-1.648255,
        0.104720,1.480878,1.194576,
        0.104720,1.480942,0.273235,
        0.104720,1.481086,2.890269,
        0.104720,1.484118,-1.529156,
        0.104720,1.484741,-1.279219,
        0.104720,1.487681,-1.406069,
        0.104720,1.487886,-3.066037,
        0.104720,1.490555,2.543284,
        0.104720,1.492424,-0.190932,
        0.104720,1.495002,-0.839810,
        0.104720,1.495297,2.007188,
        0.104720,1.496414,-2.009267,
        0.104720,1.498261,-2.425270,
        0.104720,1.498907,-2.207327,
        0.104720,1.500204,0.737638,
        0.104720,1.502469,0.151156,
        0.104720,1.507448,-0.514782,
        0.104720,1.509287,2.204647,
        0.104720,1.510518,1.884920,
        0.104720,1.511948,-2.632360,
        0.104720,1.515235,1.068565,
        0.104720,1.522237,3.102018,
        0.104720,1.522373,1.764991,
        0.104720,1.522416,-0.954956,
        0.104720,1.523102,2.765920,
        0.104720,1.524811,-2.844718,
        0.104720,1.527261,-0.304916,
        0.104720,1.529788,0.613208,
        0.104720,1.533648,1.644808,
        0.104720,1.535018,0.028869,
        0.104720,1.541082,2.424608,
        0.104720,1.546010,1.523462,
        0.104720,1.551135,-0.642994,
        0.104720,1.553895,0.489607,
        0.104720,1.555041,0.947813,
        0.104720,1.559671,-1.074934,
        0.104720,1.560345,2.985899,
        0.104720,1.560487,-2.323400,
        0.104720,1.561127,1.401111,
        0.104720,1.567193,2.643172,
        0.104720,1.568892,2.098362,
        0.104720,1.571702,-2.098435,
        0.104720,1.572701,-2.528730,
        0.104720,1.573572,1.278593,
        0.104720,1.574849,-0.091505,
        0.104720,1.576334,0.365236,
        0.104720,1.577029,-2.962888,
        0.104720,1.577688,-1.814941,
        0.104720,1.579542,-0.755917,
        0.104720,1.579624,-1.199759,
        0.104720,1.581515,-0.419414,
        0.104720,1.584581,-2.738702,
        0.104720,1.586948,-1.696296,
        0.104720,1.588901,2.301496,
        0.104720,1.589936,0.825020,
        0.104720,1.590018,-1.941652,
        0.104720,1.596968,-1.328358,
        0.104720,1.597178,-1.575607,
        0.104720,1.600470,-1.453196,
        0.104720,1.601408,0.242409,
        0.104720,1.602259,1.152656,
        0.104720,1.603064,2.865774,
        0.104720,1.612085,1.968802,
        0.104720,1.612720,-3.080498,
        0.104720,1.616061,2.521531,
        0.104720,1.616900,-0.207457,
        0.104720,1.620988,-0.875920,
        0.104720,1.622114,0.701509,
        0.104720,1.622749,0.121426,
        0.104720,1.622993,-2.221629,
        0.104720,1.629346,-0.542669,
        0.104720,1.630007,1.844552,
        0.104720,1.634747,-2.426198,
        0.104720,1.642346,1.722617,
        0.104720,1.643925,1.034629,
        0.104720,1.644848,2.743774,
        0.104720,1.645258,2.188109,
        0.104720,1.645417,-2.634725,
        0.104720,1.647395,3.085437,
        0.104720,1.648538,-2.855044,
        0.104720,1.649932,0.576808,
        0.104720,1.652224,-0.996105,
        0.104720,1.652871,1.600681,
        0.104720,1.657997,-0.325127,
        0.104720,1.661936,0.001037,
        0.104720,1.665187,1.477402,
        0.104720,1.665496,2.397872,
        0.104720,1.673267,-0.665453,
        0.104720,1.673557,0.452526,
        0.104720,1.679277,-1.122421,
        0.104720,1.679696,0.912482,
        0.104720,1.682420,1.353922,
        0.104720,1.685031,-2.009038,
        0.104720,1.685196,2.967046,
        0.104720,1.690107,2.620910,
        0.104720,1.690453,-2.123076,
        0.104720,1.691693,-1.871572,
        0.104720,1.691817,2.061688,
        0.104720,1.692644,1.237670,
        0.104720,1.696386,0.328123,
        0.104720,1.697162,-1.249026,
        0.104720,1.697488,-2.322785,
        0.104720,1.698726,-2.973576,
        0.104720,1.699172,-1.749717,
        0.104720,1.704478,-0.106295,
        0.104720,1.707244,-0.444770,
        0.104720,1.707500,-2.531497,
        0.104720,1.707714,-0.787321,
        0.104720,1.708430,-1.627552,
        0.104720,1.710583,-1.376890,
        0.104720,1.712415,-2.747921,
        0.104720,1.713125,-1.503073,
        0.104720,1.713452,0.789419,
        0.104720,1.720612,2.279888,
        0.104720,1.724852,0.202831,
        0.104720,1.724868,2.846246,
        0.104720,1.728940,1.931535,
        0.104720,1.732838,1.127584,
        0.104720,1.735719,-3.092485,
        0.104720,1.739676,2.496017,
        0.104720,1.739971,-0.228876,
        0.104720,1.743328,0.663347,
        0.104720,1.744927,-0.911892,
        0.104720,1.745952,0.084733,
        0.104720,1.748077,1.804354,
        0.104720,1.752959,-0.567548,
        0.104720,1.760169,1.678132,
        0.104720,1.763887,-2.218450,
        0.104720,1.766737,2.722752,
        0.104720,1.767999,1.004737,
        0.104720,1.769149,1.550298,
        0.104720,1.770119,-2.427175,
        0.104720,1.770460,2.154208,
        0.104720,1.770514,3.070982,
        0.104720,1.770735,0.537245,
        0.104720,1.772673,-2.864760,
        0.104720,1.773527,-1.039582,
        0.104720,1.776419,-2.642421,
        0.104720,1.780769,1.416330,
        0.104720,1.786634,-0.346247,
        0.104720,1.788829,-1.947553,
        0.104720,1.793737,0.410270,
        0.104720,1.794553,2.374618,
        0.104720,1.794699,-0.694061,
        0.104720,1.797469,-1.169105,
        0.104720,1.803705,0.881076,
        0.104720,1.804124,1.264249,
        0.104720,1.807825,-2.090455,
        0.104720,1.808822,-1.817375,
        0.104720,1.809020,2.949741,
        0.104720,1.812466,2.597595,
        0.104720,1.813159,2.025395,
        0.104720,1.813816,-1.298435,
        0.104720,1.815016,0.278951,
        0.104720,1.816510,-0.008631,
        0.104720,1.817772,-1.689625,
        0.104720,1.821360,-2.984352,
        0.104720,1.822312,-0.127211,
        0.104720,1.823386,-1.429807,
        0.104720,1.823834,-1.560483,
        0.104720,1.832060,-0.821627,
        0.104720,1.832690,-0.468440,
        0.104720,1.833945,-2.320278,
        0.104720,1.834787,0.751968,
        0.104720,1.840579,-2.536445,
        0.104720,1.841033,-2.756932,
        0.104720,1.846350,1.893983,
        0.104720,1.846644,1.113766,
        0.104720,1.846833,2.826543,
        0.104720,1.847325,2.249051,
        0.104720,1.854119,0.134161,
        0.104720,1.858863,-3.106022,
        0.104720,1.865046,1.763384,
        0.104720,1.865619,-0.952036,
        0.104720,1.865912,0.624453,
        0.104720,1.866694,2.473686,
        0.104720,1.867524,-0.240083,
        0.104720,1.875561,1.632831,
        0.104720,1.876049,-0.594378,
        0.104720,1.880509,1.500204,
        0.104720,1.883504,1.360457,
        0.104720,1.886971,-2.196665,
        0.104720,1.888979,0.982922,
        0.104720,1.889815,2.701570,
        0.104720,1.892144,-2.005340,
        0.104720,1.892348,0.496522,
        0.104720,1.893281,-1.083976,
        0.104720,1.893828,3.053453,
        0.104720,1.895117,2.121107,
        0.104720,1.898741,-2.875735,
        0.104720,1.904429,-2.428029,
        0.104720,1.908408,-2.648903,
        0.104720,1.911828,-0.364489,
        0.104720,1.912224,0.366565,
        0.104720,1.914854,1.214468,
        0.104720,1.915467,-1.883143,
        0.104720,1.915763,-0.724679,
        0.104720,1.915790,-1.216366,
        0.104720,1.922256,2.347806,
        0.104720,1.922616,0.844633,
        0.104720,1.925185,0.231170,
        0.104720,1.927474,-1.754616,
        0.104720,1.928342,-1.351174,
        0.104720,1.933328,-1.623328,
        0.104720,1.933878,0.034150,
        0.104720,1.934585,-1.487979,
        0.104720,1.934802,2.927634,
        0.104720,1.934862,1.989776,
        0.104720,1.936327,-0.099350,
        0.104720,1.938088,2.578507,
        0.104720,1.945677,-2.999805,
        0.104720,1.954572,-0.859208,
        0.104720,1.957868,-0.490784,
        0.104720,1.958989,0.712772,
        0.104720,1.960212,-2.307561,
        0.104720,1.964774,1.855507,
        0.104720,1.964971,-2.100533,
        0.104720,1.969112,1.088273,
        0.104720,1.971438,-2.765415,
        0.104720,1.972312,2.805134,
        0.104720,1.974950,2.220596,
        0.104720,1.975135,-2.539127,
        0.104720,1.982107,1.720829,
        0.104720,1.983195,-3.127984,
        0.104720,1.983318,-0.247182,
        0.104720,1.986720,-0.993220,
        0.104720,1.990381,1.586005,
        0.104720,1.990470,0.580283,
        0.104720,1.990523,1.450326,
        0.104720,1.993909,1.312897,
        0.104720,1.994694,2.452813,
        0.104720,1.999226,-0.620717,
        0.104720,2.005145,0.941012,
        0.104720,2.007301,0.142337,
        0.104720,2.014149,0.444138,
        0.104720,2.014338,-1.127392,
        0.104720,2.018131,3.021778,
        0.104720,2.019676,-1.980944,
        0.104720,2.019912,2.686612,
        0.104720,2.020545,2.088672,
        0.104720,2.026074,0.300945,
        0.104720,2.026196,-2.889057,
        0.104720,2.031959,-1.268178,
        0.104720,2.032917,-2.199917,
        0.104720,2.033572,-1.842884,
        0.104720,2.034712,-2.419505,
        0.104720,2.035818,-0.379479,
        0.104720,2.036004,-0.757062,
        0.104720,2.040501,-1.411488,
        0.104720,2.042874,-1.701367,
        0.104720,2.043060,-2.653746,
        0.104720,2.043729,-1.557938,
        0.104720,2.048538,0.806623,
        0.104720,2.049006,1.199930,
        0.104720,2.049502,-0.137989,
        0.104720,2.052181,2.326959,
        0.104720,2.052794,-0.008526,
        0.104720,2.057608,1.951726,
        0.104720,2.065568,2.567730,
        0.104720,2.070371,2.879808,
        0.104720,2.071256,-3.018961,
        0.104720,2.079746,-0.897699,
        0.104720,2.082784,1.042403,
        0.104720,2.083752,-0.511086,
        0.104720,2.084401,1.810502,
        0.104720,2.085731,0.671081,
        0.104720,2.097496,1.373172,
        0.104720,2.099185,1.667006,
        0.104720,2.100482,-2.301084,
        0.104720,2.103118,2.194745,
        0.104720,2.103223,1.521685,
        0.104720,2.103695,-2.777707,
        0.104720,2.107523,3.129581,
        0.104720,2.108277,-2.095907,
        0.104720,2.108891,-2.532166,
        0.104720,2.111003,-0.268632,
        0.104720,2.112276,-1.027246,
        0.104720,2.114380,0.531457,
        0.104720,2.119166,0.100552,
        0.104720,2.119564,2.754818,
        0.104720,2.122474,-0.648314,
        0.104720,2.125183,2.445598,
        0.104720,2.129501,0.238290,
        0.104720,2.131178,0.386663,
        0.104720,2.131223,-1.174311,
        0.104720,2.134269,-1.948042,
        0.104720,2.134684,0.907065,
        0.104720,2.137615,2.988806,
        0.104720,2.145635,-1.327708,
        0.104720,2.146467,2.055081,
        0.104720,2.146582,-1.793788,
        0.104720,2.149210,-1.482321,
        0.104720,2.152901,-1.636788,
        0.104720,2.156337,1.151558,
        0.104720,2.156474,-2.912065,
        0.104720,2.157194,-0.792563,
        0.104720,2.166724,-0.401638,
        0.104720,2.166980,-0.134879,
        0.104720,2.173644,-2.403429,
        0.104720,2.175711,-2.655523,
        0.104720,2.177132,2.600419,
        0.104720,2.178879,0.767722,
        0.104720,2.180752,1.908485,
        0.104720,2.181338,2.313721,
        0.104720,2.185580,1.287673,
        0.104720,2.195400,-2.217546,
        0.104720,2.200418,-3.056442,
        0.104720,2.201829,-0.001735,
        0.104720,2.203804,1.756176,
        0.104720,2.211602,1.443930,
        0.104720,2.212686,-0.544809,
        0.104720,2.213937,0.621892,
        0.104720,2.214075,1.600691,
        0.104720,2.216256,1.020029,
        0.104720,2.219060,2.886631,
        0.104720,2.223375,-0.906185,
        0.104720,2.226002,-1.060248,
        0.104720,2.226413,-2.064034,
        0.104720,2.230554,2.171334,
        0.104720,2.232413,2.735777,
        0.104720,2.233709,3.070830,
        0.104720,2.234383,-2.792065,
        0.104720,2.237183,0.468799,
        0.104720,2.237695,-0.271388,
        0.104720,2.239419,0.151693,
        0.104720,2.243347,-2.521049,
        0.104720,2.244035,-1.229456,
        0.104720,2.244355,0.309899,
        0.104720,2.245175,2.460236,
        0.104720,2.245628,-1.899216,
        0.104720,2.245973,-0.702234,
        0.104720,2.255601,-1.728131,
        0.104720,2.256702,-1.401367,
        0.104720,2.261466,-1.557500,
        0.104720,2.267390,0.878499,
        0.104720,2.272099,2.018594,
        0.104720,2.278618,1.170386,
        0.104720,2.283304,-2.345584,
        0.104720,2.285175,-2.943090,
        0.104720,2.293000,-0.124451,
        0.104720,2.294058,-0.421070,
        0.104720,2.303283,1.855938,
        0.104720,2.304968,2.608978,
        0.104720,2.305347,2.311346,
        0.104720,2.307188,-2.656134,
        0.104720,2.307301,1.336200,
        0.104720,2.309655,0.726020,
        0.104720,2.317634,-2.185225,
        0.104720,2.321510,1.685664,
        0.104720,2.324376,-3.111006,
        0.104720,2.324731,1.511186,
        0.104720,2.329587,0.035498,
        0.104720,2.330158,2.977264,
        0.104720,2.335096,-0.588468,
        0.104720,2.336046,-1.115799,
        0.104720,2.338060,2.803613,
        0.104720,2.338083,-0.943400,
        0.104720,2.340210,0.562145,
        0.104720,2.341931,1.023926,
        0.104720,2.343265,-0.779061,
        0.104720,2.343434,-2.013692,
        0.104720,2.353029,-1.310529,
        0.104720,2.354868,0.211543,
        0.104720,2.355120,0.388593,
        0.104720,2.355165,-1.833173,
        0.104720,2.355462,2.151441,
        0.104720,2.362948,-2.487477,
        0.104720,2.363101,-2.810456,
        0.104720,2.363394,-0.272918,
        0.104720,2.364509,-1.644533,
        0.104720,2.370465,2.466824,
        0.104720,2.386110,-1.481351,
        0.104720,2.394960,0.862133,
        0.104720,2.395716,1.977184,
        0.104720,2.396339,1.200900,
        0.104720,2.407941,-2.316753,
        0.104720,2.409450,-2.988432,
        0.104720,2.417202,-0.103363,
        0.104720,2.417386,-0.443097,
        0.104720,2.417950,2.666662,
        0.104720,2.423190,1.789783,
        0.104720,2.423692,1.392696,
        0.104720,2.431122,3.086379,
        0.104720,2.431406,-2.647820,
        0.104720,2.431584,2.307631,
        0.104720,2.434090,1.592818,
        0.104720,2.435993,0.683844,
        0.104720,2.440898,-2.133300,
        0.104720,2.442222,-1.195837,
        0.104720,2.443112,2.883761,
        0.104720,2.449191,-0.633651,
        0.104720,2.450821,0.084741,
        0.104720,2.453499,-1.008861,
        0.104720,2.460010,-1.938011,
        0.104720,2.460199,-0.822008,
        0.104720,2.460902,0.490079,
        0.104720,2.461774,-1.734913,
        0.104720,2.463986,1.034407,
        0.104720,2.467647,0.287667,
        0.104720,2.479849,2.126925,
        0.104720,2.485907,-2.835471,
        0.104720,2.485945,-1.374853,
        0.104720,2.487747,2.505912,
        0.104720,2.487957,-0.274279,
        0.104720,2.489741,-2.468428,
        0.104720,2.512538,1.239900,
        0.104720,2.517528,1.923289,
        0.104720,2.518454,-1.565852,
        0.104720,2.519348,0.843503,
        0.104720,2.522766,-3.059231,
        0.104720,2.528686,2.737228,
        0.104720,2.534798,-2.266671,
        0.104720,2.536533,1.466274,
        0.104720,2.538504,1.699812,
        0.104720,2.539516,-0.476942,
        0.104720,2.539532,-0.074548,
        0.104720,2.543197,2.990252,
        0.104720,2.554022,2.320619,
        0.104720,2.556919,-2.655005,
        0.104720,2.558146,0.627191,
        0.104720,2.558223,-1.164189,
        0.104720,2.563452,-2.044028,
        0.104720,2.568462,0.150655,
        0.104720,2.569130,-0.700837,
        0.104720,2.570050,-1.804492,
        0.104720,2.572710,-0.934665,
        0.104720,2.575994,0.391718,
        0.104720,2.584774,1.046703,
        0.104720,2.601432,2.102535,
        0.104720,2.603559,2.555114,
        0.104720,2.604150,-2.885208,
        0.104720,2.608630,-1.378011,
        0.104720,2.611256,-0.279110,
        0.104720,2.617344,-2.439480,
        0.104720,2.627402,1.293552,
        0.104720,2.635396,1.848850,
        0.104720,2.635686,2.834950,
        0.104720,2.637087,3.132730,
        0.104720,2.642029,0.813274,
        0.104720,2.645260,1.570203,
        0.104720,2.646182,-1.614036,
        0.104720,2.660113,-2.186513,
        0.104720,2.660483,-0.034608,
        0.104720,2.661713,-0.529511,
        0.104720,2.672720,-1.110422,
        0.104720,2.676626,2.335353,
        0.104720,2.676638,0.540364,
        0.104720,2.676780,-1.901158,
        0.104720,2.681411,-2.667965,
        0.104720,2.682152,0.245751,
        0.104720,2.682859,-0.817071,
        0.104720,2.705107,1.062142,
        0.104720,2.717705,2.625995,
        0.104720,2.721001,2.060097,
        0.104720,2.722319,-2.954069,
        0.104720,2.729840,-1.368237,
        0.104720,2.734349,-0.289298,
        0.104720,2.738409,2.975671,
        0.104720,2.740366,1.374341,
        0.104720,2.744633,-2.391311,
        0.104720,2.747398,1.727166,
        0.104720,2.763799,0.760372,
        0.104720,2.768238,-1.679243,
        0.104720,2.779245,-2.046268,
        0.104720,2.780208,-0.626453,
        0.104720,2.780471,0.027732,
        0.104720,2.784696,-1.013128,
        0.104720,2.789817,0.395431,
        0.104720,2.801154,2.353143,
        0.104720,2.807091,-2.687477,
        0.104720,2.825762,1.085035,
        0.104720,2.830209,2.734943,
        0.104720,2.833517,-3.085508,
        0.104720,2.838301,1.980061,
        0.104720,2.849871,1.512012,
        0.104720,2.850089,-1.343899,
        0.104720,2.856905,-0.318441,
        0.104720,2.870445,-2.291001,
        0.104720,2.885017,0.655562,
        0.104720,2.886428,-1.793217,
        0.104720,2.890337,-0.825633,
        0.104720,2.898932,0.139973,
        0.104720,2.931927,-2.728768,
        0.104720,2.932096,2.377790,
        0.104720,2.937409,2.947245,
        0.104720,2.948048,1.133127,
        0.104720,2.950896,1.797889,
        0.104720,2.970019,-1.282382,
        0.104720,2.978333,-0.412090,
        0.104720,2.996437,-2.089121,
        0.104720,3.011274,0.434110,
        0.104720,3.050410,3.138492,
        0.104720,3.064492,1.647259,
        0.104720,3.088868,-0.989501,
        0.209440,0.053884,-2.408394,
        0.209440,0.077645,1.453433,
        0.209440,0.088408,-0.311160,
        0.209440,0.142314,2.590387,
        0.209440,0.147206,-1.385164,
        0.209440,0.163506,0.573203,
        0.209440,0.165485,-2.875396,
        0.209440,0.182564,-2.125821,
        0.209440,0.190569,1.929141,
        0.209440,0.201102,1.249082,
        0.209440,0.201397,-0.716646,
        0.209440,0.210568,-0.069559,
        0.209440,0.251672,2.848900,
        0.209440,0.266262,-1.678068,
        0.209440,0.269108,-1.215563,
        0.209440,0.273777,-2.506756,
        0.209440,0.279854,0.814581,
        0.209440,0.281664,-2.969147,
        0.209440,0.283625,2.405603,
        0.209440,0.284058,0.357350,
        0.209440,0.294107,1.593880,
        0.209440,0.302491,-0.400410,
        0.209440,0.308059,2.021491,
        0.209440,0.312219,-2.072057,
        0.209440,0.321905,-0.829121,
        0.209440,0.329984,1.201621,
        0.209440,0.339714,-0.013712,
        0.209440,0.368784,2.974974,
        0.209440,0.376428,-1.495532,
        0.209440,0.378103,2.646436,
        0.209440,0.385088,0.590355,
        0.209440,0.386248,-2.690943,
        0.209440,0.389851,-2.340031,
        0.209440,0.391502,-1.171374,
        0.209440,0.398249,-1.812694,
        0.209440,0.400542,-2.995729,
        0.209440,0.404273,0.918163,
        0.209440,0.405617,1.773533,
        0.209440,0.409009,-0.570243,
        0.209440,0.410700,0.270433,
        0.209440,0.415893,1.464890,
        0.209440,0.420680,2.088164,
        0.209440,0.424766,-0.258729,
        0.209440,0.426507,2.367872,
        0.209440,0.441292,-0.890727,
        0.209440,0.451839,-2.078193,
        0.209440,0.459270,1.191562,
        0.209440,0.470622,0.014002,
        0.209440,0.481292,-2.529098,
        0.209440,0.482460,2.849843,
        0.209440,0.492205,-1.397963,
        0.209440,0.493314,3.117264,
        0.209440,0.497870,0.727828,
        0.209440,0.499783,-1.654379,
        0.209440,0.502651,0.463333,
        0.209440,0.505463,2.594557,
        0.209440,0.506104,-2.874910,
        0.209440,0.513579,-1.147354,
        0.209440,0.518569,1.644237,
        0.209440,0.519001,1.900488,
        0.209440,0.519277,-0.673809,
        0.209440,0.524769,-0.421955,
        0.209440,0.525564,-2.288838,
        0.209440,0.532541,0.978737,
        0.209440,0.534610,2.154808,
        0.209440,0.535020,-1.891846,
        0.209440,0.540208,1.400497,
        0.209440,0.542006,0.228337,
        0.209440,0.550762,-0.180771,
        0.209440,0.560531,2.378474,
        0.209440,0.561629,-0.924576,
        0.209440,0.567465,-2.679392,
        0.209440,0.588527,1.181755,
        0.209440,0.591643,-2.098364,
        0.209440,0.594963,2.997575,
        0.209440,0.599643,-3.045235,
        0.209440,0.599962,2.773556,
        0.209440,0.602620,0.029270,
        0.209440,0.605711,0.598454,
        0.209440,0.607708,-1.541026,
        0.209440,0.609360,-2.464379,
        0.209440,0.610738,-1.326786,
        0.209440,0.616968,0.820489,
        0.209440,0.626379,1.773610,
        0.209440,0.628626,-1.751180,
        0.209440,0.630671,0.391949,
        0.209440,0.632622,-0.737642,
        0.209440,0.632708,2.566722,
        0.209440,0.632797,1.989858,
        0.209440,0.634455,-0.534790,
        0.209440,0.636205,1.561170,
        0.209440,0.637681,-1.123423,
        0.209440,0.641441,-2.842109,
        0.209440,0.644807,-0.327139,
        0.209440,0.652440,2.197129,
        0.209440,0.660897,-2.278593,
        0.209440,0.664008,1.360091,
        0.209440,0.670850,-1.945257,
        0.209440,0.671396,1.004859,
        0.209440,0.673985,0.203855,
        0.209440,0.680326,-0.133370,
        0.209440,0.685083,-0.942363,
        0.209440,0.688204,-2.632204,
        0.209440,0.689777,2.385011,
        0.209440,0.700065,3.115430,
        0.209440,0.704080,2.911636,
        0.209440,0.712883,1.174216,
        0.209440,0.715701,0.699906,
        0.209440,0.718851,-1.442215,
        0.209440,0.721019,2.721536,
        0.209440,0.725300,-2.975134,
        0.209440,0.727136,-1.633288,
        0.209440,0.727399,-2.120567,
        0.209440,0.730751,0.522488,
        0.209440,0.732892,-1.254330,
        0.209440,0.734333,-2.446184,
        0.209440,0.735125,0.037270,
        0.209440,0.737893,1.866463,
        0.209440,0.740650,1.681075,
        0.209440,0.745971,-0.781286,
        0.209440,0.749432,-0.442559,
        0.209440,0.750102,2.050636,
        0.209440,0.751046,0.863821,
        0.209440,0.757018,-0.608833,
        0.209440,0.757046,-1.816019,
        0.209440,0.757389,1.501184,
        0.209440,0.759800,2.546181,
        0.209440,0.759859,0.351923,
        0.209440,0.761141,-2.781102,
        0.209440,0.769035,-1.078574,
        0.209440,0.770166,-0.266604,
        0.209440,0.772218,2.227380,
        0.209440,0.788382,1.331051,
        0.209440,0.790269,-2.282409,
        0.209440,0.803410,-1.984567,
        0.209440,0.803813,0.191520,
        0.209440,0.806832,3.022155,
        0.209440,0.807476,-2.597675,
        0.209440,0.812512,-0.107335,
        0.209440,0.814161,1.010949,
        0.209440,0.817732,-3.084486,
        0.209440,0.817846,2.390688,
        0.209440,0.819323,2.845960,
        0.209440,0.819430,-0.915017,
        0.209440,0.829198,-1.532491,
        0.209440,0.829531,0.665695,
        0.209440,0.829816,1.167147,
        0.209440,0.833613,-1.357492,
        0.209440,0.842945,-2.907318,
        0.209440,0.844883,2.679708,
        0.209440,0.848189,-1.705003,
        0.209440,0.849888,1.767901,
        0.209440,0.851126,0.493370,
        0.209440,0.852911,1.933394,
        0.209440,0.857516,-2.432863,
        0.209440,0.858261,-1.191729,
        0.209440,0.858279,-2.138895,
        0.209440,0.859723,-0.734928,
        0.209440,0.859754,1.603292,
        0.209440,0.861655,-0.394821,
        0.209440,0.864300,0.043549,
        0.209440,0.866940,0.851229,
        0.209440,0.869748,2.095969,
        0.209440,0.877352,-0.574668,
        0.209440,0.879254,-2.733338,
        0.209440,0.882554,1.443757,
        0.209440,0.884060,0.334199,
        0.209440,0.885626,-1.866061,
        0.209440,0.889834,2.529957,
        0.209440,0.893140,2.251698,
        0.209440,0.897463,-0.240822,
        0.209440,0.898758,-1.031948,
        0.209440,0.913918,1.287618,
        0.209440,0.916555,-2.287173,
        0.209440,0.918305,3.107845,
        0.209440,0.919522,2.948277,
        0.209440,0.927138,-2.569686,
        0.209440,0.929031,0.185383,
        0.209440,0.933766,-3.015917,
        0.209440,0.934873,-2.010822,
        0.209440,0.935636,-1.445855,
        0.209440,0.936702,-0.866461,
        0.209440,0.937395,0.729108,
        0.209440,0.937910,2.787575,
        0.209440,0.939144,-0.092962,
        0.209440,0.941718,1.125085,
        0.209440,0.942873,-1.607680,
        0.209440,0.943701,0.980328,
        0.209440,0.947774,2.395256,
        0.209440,0.952433,0.584347,
        0.209440,0.954456,-1.287293,
        0.209440,0.958731,-2.852552,
        0.209440,0.961838,1.839284,
        0.209440,0.961898,-0.469425,
        0.209440,0.965426,1.686615,
        0.209440,0.970108,1.990771,
        0.209440,0.972820,-1.762780,
        0.209440,0.973500,-0.711881,
        0.209440,0.975253,2.641082,
        0.209440,0.977824,0.441928,
        0.209440,0.981485,-2.424766,
        0.209440,0.981877,1.537100,
        0.209440,0.982609,-1.136920,
        0.209440,0.984265,-2.156073,
        0.209440,0.989150,0.045493,
        0.209440,0.993052,2.137478,
        0.209440,0.997766,-2.693871,
        0.209440,0.999847,-0.334738,
        0.209440,1.001920,0.848054,
        0.209440,1.006146,1.389734,
        0.209440,1.012180,0.304066,
        0.209440,1.016938,2.270674,
        0.209440,1.017371,-0.979790,
        0.209440,1.019495,-1.894487,
        0.209440,1.023796,-0.583691,
        0.209440,1.026525,2.506945,
        0.209440,1.027004,-0.203653,
        0.209440,1.028027,3.036573,
        0.209440,1.031435,1.241021,
        0.209440,1.032316,2.884221,
        0.209440,1.035863,-3.110475,
        0.209440,1.042085,-2.292950,
        0.209440,1.043272,-1.522986,
        0.209440,1.046265,-2.958847,
        0.209440,1.049272,-1.376344,
        0.209440,1.050039,-2.552086,
        0.209440,1.053682,0.689863,
        0.209440,1.055458,0.170394,
        0.209440,1.056622,-0.831866,
        0.209440,1.057515,1.097998,
        0.209440,1.058706,-2.036438,
        0.209440,1.061389,-1.672248,
        0.209440,1.063758,2.743093,
        0.209440,1.070157,-0.073422,
        0.209440,1.072154,1.761506,
        0.209440,1.074029,1.909193,
        0.209440,1.074204,0.545276,
        0.209440,1.074714,-2.805147,
        0.209440,1.076262,0.966104,
        0.209440,1.077642,-1.227073,
        0.209440,1.084151,1.618525,
        0.209440,1.085085,2.377673,
        0.209440,1.086822,-0.441112,
        0.209440,1.087142,2.058815,
        0.209440,1.098383,-0.696430,
        0.209440,1.099138,-1.084033,
        0.209440,1.103208,1.477773,
        0.209440,1.103783,0.408254,
        0.209440,1.106275,-2.420800,
        0.209440,1.108079,2.612943,
        0.209440,1.109345,-2.172718,
        0.209440,1.115957,-1.787597,
        0.209440,1.118930,0.045983,
        0.209440,1.119777,0.834702,
        0.209440,1.121943,-2.669042,
        0.209440,1.124930,2.219917,
        0.209440,1.125217,1.336294,
        0.209440,1.130958,2.972931,
        0.209440,1.131400,-0.306629,
        0.209440,1.137135,-0.936514,
        0.209440,1.139125,-1.924528,
        0.209440,1.139877,0.276382,
        0.209440,1.142003,-3.058022,
        0.209440,1.147363,-0.558448,
        0.209440,1.150348,-0.176418,
        0.209440,1.151198,1.197016,
        0.209440,1.153930,-1.597104,
        0.209440,1.155389,2.837148,
        0.209440,1.155458,-1.460623,
        0.209440,1.157331,-2.906362,
        0.209440,1.158594,2.487618,
        0.209440,1.162103,-1.322553,
        0.209440,1.165407,3.092861,
        0.209440,1.168025,0.717075,
        0.209440,1.168406,-2.301519,
        0.209440,1.174782,-2.538755,
        0.209440,1.177998,1.061256,
        0.209440,1.178898,1.827305,
        0.209440,1.181331,-2.060330,
        0.209440,1.182540,-0.798952,
        0.209440,1.184569,0.147517,
        0.209440,1.184734,1.968919,
        0.209440,1.184804,0.593386,
        0.209440,1.188278,1.690808,
        0.209440,1.194360,2.110440,
        0.209440,1.194776,2.710455,
        0.209440,1.197745,-1.173705,
        0.209440,1.199898,-2.777257,
        0.209440,1.201164,2.355931,
        0.209440,1.203760,1.556404,
        0.209440,1.205845,-0.415993,
        0.209440,1.206585,0.933648,
        0.209440,1.206950,-0.037248,
        0.209440,1.209341,0.471162,
        0.209440,1.213483,-1.039417,
        0.209440,1.218146,-0.672643,
        0.209440,1.221608,1.421669,
        0.209440,1.226043,-1.693533,
        0.209440,1.232870,-1.822773,
        0.209440,1.234375,-2.421194,
        0.209440,1.234531,-2.190148,
        0.209440,1.239414,0.348959,
        0.209440,1.240183,2.587255,
        0.209440,1.242312,2.941234,
        0.209440,1.244075,1.286644,
        0.209440,1.245961,-2.997461,
        0.209440,1.246495,-2.646872,
        0.209440,1.247736,2.242810,
        0.209440,1.253604,0.812222,
        0.209440,1.257816,-0.894317,
        0.209440,1.258247,-1.953634,
        0.209440,1.258993,-0.284099,
        0.209440,1.262380,-0.157609,
        0.209440,1.262791,-1.549229,
        0.209440,1.263772,-0.533073,
        0.209440,1.266731,-1.412461,
        0.209440,1.268591,-3.122146,
        0.209440,1.269614,-1.275246,
        0.209440,1.270745,1.152533,
        0.209440,1.273606,0.224545,
        0.209440,1.279716,2.812101,
        0.209440,1.283786,2.462124,
        0.209440,1.284798,0.078209,
        0.209440,1.285402,-2.882049,
        0.209440,1.286936,1.886053,
        0.209440,1.289408,0.683122,
        0.209440,1.294736,1.756184,
        0.209440,1.296978,-2.311511,
        0.209440,1.298546,2.020601,
        0.209440,1.300924,1.022478,
        0.209440,1.303305,3.043152,
        0.209440,1.304210,-2.084225,
        0.209440,1.305241,-2.530249,
        0.209440,1.307513,1.628003,
        0.209440,1.310722,0.558267,
        0.209440,1.314142,2.139329,
        0.209440,1.316087,-0.777971,
        0.209440,1.317907,-1.121023,
        0.209440,1.321019,-2.749776,
        0.209440,1.321609,2.687717,
        0.209440,1.322568,1.499415,
        0.209440,1.322972,-0.395699,
        0.209440,1.325286,-0.994192,
        0.209440,1.329146,-0.648513,
        0.209440,1.329795,-0.042570,
        0.209440,1.331651,2.344497,
        0.209440,1.336522,0.435098,
        0.209440,1.339410,0.900087,
        0.209440,1.339550,1.370353,
        0.209440,1.340441,-1.731580,
        0.209440,1.351855,-1.855234,
        0.209440,1.360161,-3.018404,
        0.209440,1.361031,-1.613971,
        0.209440,1.361196,2.914780,
        0.209440,1.361998,-2.207254,
        0.209440,1.362302,1.239384,
        0.209440,1.365228,2.563808,
        0.209440,1.365452,0.310522,
        0.209440,1.367028,-2.420041,
        0.209440,1.373422,-1.484281,
        0.209440,1.374245,-1.226417,
        0.209440,1.375020,-0.170393,
        0.209440,1.376560,-1.355481,
        0.209440,1.377070,-1.980984,
        0.209440,1.377578,-2.636339,
        0.209440,1.378096,0.774626,
        0.209440,1.379086,-0.508192,
        0.209440,1.382562,0.177141,
        0.209440,1.388849,2.233847,
        0.209440,1.389268,-0.886349,
        0.209440,1.391324,1.109396,
        0.209440,1.391699,-0.288725,
        0.209440,1.392869,3.136903,
        0.209440,1.393612,1.952016,
        0.209440,1.395545,-2.850807,
        0.209440,1.401926,1.822930,
        0.209440,1.402149,2.788898,
        0.209440,1.409552,0.647980,
        0.209440,1.410840,0.056219,
        0.209440,1.411944,2.444377,
        0.209440,1.413639,1.697982,
        0.209440,1.426479,1.573427,
        0.209440,1.426725,2.102242,
        0.209440,1.427868,0.984637,
        0.209440,1.429580,-2.107232,
        0.209440,1.431375,-2.313269,
        0.209440,1.431907,3.015410,
        0.209440,1.432325,-1.029054,
        0.209440,1.433341,0.524268,
        0.209440,1.435181,-0.739771,
        0.209440,1.437208,-2.529085,
        0.209440,1.439041,-0.617966,
        0.209440,1.441473,1.448544,
        0.209440,1.444527,2.664744,
        0.209440,1.450513,-0.400625,
        0.209440,1.451907,-0.066823,
        0.209440,1.452754,-2.738631,
        0.209440,1.454061,-2.950416,
        0.209440,1.456736,1.323154,
        0.209440,1.458370,0.400433,
        0.209440,1.459314,-1.765003,
        0.209440,1.459902,-1.145835,
        0.209440,1.464952,2.329245,
        0.209440,1.466050,0.862104,
        0.209440,1.471330,-1.886160,
        0.209440,1.475119,-1.648255,
        0.209440,1.480878,1.194576,
        0.209440,1.480942,0.273235,
        0.209440,1.481086,2.890269,
        0.209440,1.484118,-1.529156,
        0.209440,1.484741,-1.279219,
        0.209440,1.487681,-1.406069,
        0.209440,1.487886,-3.066037,
        0.209440,1.490555,2.543284,
        0.209440,1.492424,-0.190932,
        0.209440,1.495002,-0.839810,
        0.209440,1.495297,2.007188,
        0.209440,1.496414,-2.009267,
        0.209440,1.498261,-2.425270,
        0.209440,1.498907,-2.207327,
        0.209440,1.500204,0.737638,
        0.209440,1.502469,0.151156,
        0.209440,1.507448,-0.514782,
        0.209440,1.509287,2.204647,
        0.209440,1.510518,1.884920,
        0.209440,1.511948,-2.632360,
        0.209440,1.515235,1.068565,
        0.209440,1.522237,3.102018,
        0.209440,1.522373,1.764991,
        0.209440,1.522416,-0.954956,
        0.209440,1.523102,2.765920,
        0.209440,1.524811,-2.844718,
        0.209440,1.527261,-0.304916,
        0.209440,1.529788,0.613208,
        0.209440,1.533648,1.644808,
        0.209440,1.535018,0.028869,
        0.209440,1.541082,2.424608,
        0.209440,1.546010,1.523462,
        0.209440,1.551135,-0.642994,
        0.209440,1.553895,0.489607,
        0.209440,1.555041,0.947813,
        0.209440,1.559671,-1.074934,
        0.209440,1.560345,2.985899,
        0.209440,1.560487,-2.323400,
        0.209440,1.561127,1.401111,
        0.209440,1.567193,2.643172,
        0.209440,1.568892,2.098362,
        0.209440,1.571702,-2.098435,
        0.209440,1.572701,-2.528730,
        0.209440,1.573572,1.278593,
        0.209440,1.574849,-0.091505,
        0.209440,1.576334,0.365236,
        0.209440,1.577029,-2.962888,
        0.209440,1.577688,-1.814941,
        0.209440,1.579542,-0.755917,
        0.209440,1.579624,-1.199759,
        0.209440,1.581515,-0.419414,
        0.209440,1.584581,-2.738702,
        0.209440,1.586948,-1.696296,
        0.209440,1.588901,2.301496,
        0.209440,1.589936,0.825020,
        0.209440,1.590018,-1.941652,
        0.209440,1.596968,-1.328358,
        0.209440,1.597178,-1.575607,
        0.209440,1.600470,-1.453196,
        0.209440,1.601408,0.242409,
        0.209440,1.602259,1.152656,
        0.209440,1.603064,2.865774,
        0.209440,1.612085,1.968802,
        0.209440,1.612720,-3.080498,
        0.209440,1.616061,2.521531,
        0.209440,1.616900,-0.207457,
        0.209440,1.620988,-0.875920,
        0.209440,1.622114,0.701509,
        0.209440,1.622749,0.121426,
        0.209440,1.622993,-2.221629,
        0.209440,1.629346,-0.542669,
        0.209440,1.630007,1.844552,
        0.209440,1.634747,-2.426198,
        0.209440,1.642346,1.722617,
        0.209440,1.643925,1.034629,
        0.209440,1.644848,2.743774,
        0.209440,1.645258,2.188109,
        0.209440,1.645417,-2.634725,
        0.209440,1.647395,3.085437,
        0.209440,1.648538,-2.855044,
        0.209440,1.649932,0.576808,
        0.209440,1.652224,-0.996105,
        0.209440,1.652871,1.600681,
        0.209440,1.657997,-0.325127,
        0.209440,1.661936,0.001037,
        0.209440,1.665187,1.477402,
        0.209440,1.665496,2.397872,
        0.209440,1.673267,-0.665453,
        0.209440,1.673557,0.452526,
        0.209440,1.679277,-1.122421,
        0.209440,1.679696,0.912482,
        0.209440,1.682420,1.353922,
        0.209440,1.685031,-2.009038,
        0.209440,1.685196,2.967046,
        0.209440,1.690107,2.620910,
        0.209440,1.690453,-2.123076,
        0.209440,1.691693,-1.871572,
        0.209440,1.691817,2.061688,
        0.209440,1.692644,1.237670,
        0.209440,1.696386,0.328123,
        0.209440,1.697162,-1.249026,
        0.209440,1.697488,-2.322785,
        0.209440,1.698726,-2.973576,
        0.209440,1.699172,-1.749717,
        0.209440,1.704478,-0.106295,
        0.209440,1.707244,-0.444770,
        0.209440,1.707500,-2.531497,
        0.209440,1.707714,-0.787321,
        0.209440,1.708430,-1.627552,
        0.209440,1.710583,-1.376890,
        0.209440,1.712415,-2.747921,
        0.209440,1.713125,-1.503073,
        0.209440,1.713452,0.789419,
        0.209440,1.720612,2.279888,
        0.209440,1.724852,0.202831,
        0.209440,1.724868,2.846246,
        0.209440,1.728940,1.931535,
        0.209440,1.732838,1.127584,
        0.209440,1.735719,-3.092485,
        0.209440,1.739676,2.496017,
        0.209440,1.739971,-0.228876,
        0.209440,1.743328,0.663347,
        0.209440,1.744927,-0.911892,
        0.209440,1.745952,0.084733,
        0.209440,1.748077,1.804354,
        0.209440,1.752959,-0.567548,
        0.209440,1.760169,1.678132,
        0.209440,1.763887,-2.218450,
        0.209440,1.766737,2.722752,
        0.209440,1.767999,1.004737,
        0.209440,1.769149,1.550298,
        0.209440,1.770119,-2.427175,
        0.209440,1.770460,2.154208,
        0.209440,1.770514,3.070982,
        0.209440,1.770735,0.537245,
        0.209440,1.772673,-2.864760,
        0.209440,1.773527,-1.039582,
        0.209440,1.776419,-2.642421,
        0.209440,1.780769,1.416330,
        0.209440,1.786634,-0.346247,
        0.209440,1.788829,-1.947553,
        0.209440,1.793737,0.410270,
        0.209440,1.794553,2.374618,
        0.209440,1.794699,-0.694061,
        0.209440,1.797469,-1.169105,
        0.209440,1.803705,0.881076,
        0.209440,1.804124,1.264249,
        0.209440,1.807825,-2.090455,
        0.209440,1.808822,-1.817375,
        0.209440,1.809020,2.949741,
        0.209440,1.812466,2.597595,
        0.209440,1.813159,2.025395,
        0.209440,1.813816,-1.298435,
        0.209440,1.815016,0.278951,
        0.209440,1.816510,-0.008631,
        0.209440,1.817772,-1.689625,
        0.209440,1.821360,-2.984352,
        0.209440,1.822312,-0.127211,
        0.209440,1.823386,-1.429807,
        0.209440,1.823834,-1.560483,
        0.209440,1.832060,-0.821627,
        0.209440,1.832690,-0.468440,
        0.209440,1.833945,-2.320278,
        0.209440,1.834787,0.751968,
        0.209440,1.840579,-2.536445,
        0.209440,1.841033,-2.756932,
        0.209440,1.846350,1.893983,
        0.209440,1.846644,1.113766,
        0.209440,1.846833,2.826543,
        0.209440,1.847325,2.249051,
        0.209440,1.854119,0.134161,
        0.209440,1.858863,-3.106022,
        0.209440,1.865046,1.763384,
        0.209440,1.865619,-0.952036,
        0.209440,1.865912,0.624453,
        0.209440,1.866694,2.473686,
        0.209440,1.867524,-0.240083,
        0.209440,1.875561,1.632831,
        0.209440,1.876049,-0.594378,
        0.209440,1.880509,1.500204,
        0.209440,1.883504,1.360457,
        0.209440,1.886971,-2.196665,
        0.209440,1.888979,0.982922,
        0.209440,1.889815,2.701570,
        0.209440,1.892144,-2.005340,
        0.209440,1.892348,0.496522,
        0.209440,1.893281,-1.083976,
        0.209440,1.893828,3.053453,
        0.209440,1.895117,2.121107,
        0.209440,1.898741,-2.875735,
        0.209440,1.904429,-2.428029,
        0.209440,1.908408,-2.648903,
        0.209440,1.911828,-0.364489,
        0.209440,1.912224,0.366565,
        0.209440,1.914854,1.214468,
        0.209440,1.915467,-1.883143,
        0.209440,1.915763,-0.724679,
        0.209440,1.915790,-1.216366,
        0.209440,1.922256,2.347806,
        0.209440,1.922616,0.844633,
        0.209440,1.925185,0.231170,
        0.209440,1.927474,-1.754616,
        0.209440,1.928342,-1.351174,
        0.209440,1.933328,-1.623328,
        0.209440,1.933878,0.034150,
        0.209440,1.934585,-1.487979,
        0.209440,1.934802,2.927634,
        0.209440,1.934862,1.989776,
        0.209440,1.936327,-0.099350,
        0.209440,1.938088,2.578507,
        0.209440,1.945677,-2.999805,
        0.209440,1.954572,-0.859208,
        0.209440,1.957868,-0.490784,
        0.209440,1.958989,0.712772,
        0.209440,1.960212,-2.307561,
        0.209440,1.964774,1.855507,
        0.209440,1.964971,-2.100533,
        0.209440,1.969112,1.088273,
        0.209440,1.971438,-2.765415,
        0.209440,1.972312,2.805134,
        0.209440,1.974950,2.220596,
        0.209440,1.975135,-2.539127,
        0.209440,1.982107,1.720829,
        0.209440,1.983195,-3.127984,
        0.209440,1.983318,-0.247182,
        0.209440,1.986720,-0.993220,
        0.209440,1.990381,1.586005,
        0.209440,1.990470,0.580283,
        0.209440,1.990523,1.450326,
        0.209440,1.993909,1.312897,
        0.209440,1.994694,2.452813,
        0.209440,1.999226,-0.620717,
        0.209440,2.005145,0.941012,
        0.209440,2.007301,0.142337,
        0.209440,2.014149,0.444138,
        0.209440,2.014338,-1.127392,
        0.209440,2.018131,3.021778,
        0.209440,2.019676,-1.980944,
        0.209440,2.019912,2.686612,
        0.209440,2.020545,2.088672,
        0.209440,2.026074,0.300945,
        0.209440,2.026196,-2.889057,
        0.209440,2.031959,-1.268178,
        0.209440,2.032917,-2.199917,
        0.209440,2.033572,-1.842884,
        0.209440,2.034712,-2.419505,
        0.209440,2.035818,-0.379479,
        0.209440,2.036004,-0.757062,
        0.209440,2.040501,-1.411488,
        0.209440,2.042874,-1.701367,
        0.209440,2.043060,-2.653746,
        0.209440,2.043729,-1.557938,
        0.209440,2.048538,0.806623,
        0.209440,2.049006,1.199930,
        0.209440,2.049502,-0.137989,
        0.209440,2.052181,2.326959,
        0.209440,2.052794,-0.008526,
        0.209440,2.057608,1.951726,
        0.209440,2.065568,2.567730,
        0.209440,2.070371,2.879808,
        0.209440,2.071256,-3.018961,
        0.209440,2.079746,-0.897699,
        0.209440,2.082784,1.042403,
        0.209440,2.083752,-0.511086,
        0.209440,2.084401,1.810502,
        0.209440,2.085731,0.671081,
        0.209440,2.097496,1.373172,
        0.209440,2.099185,1.667006,
        0.209440,2.100482,-2.301084,
        0.209440,2.103118,2.194745,
        0.209440,2.103223,1.521685,
        0.209440,2.103695,-2.777707,
        0.209440,2.107523,3.129581,
        0.209440,2.108277,-2.095907,
        0.209440,2.108891,-2.532166,
        0.209440,2.111003,-0.268632,
        0.209440,2.112276,-1.027246,
        0.209440,2.114380,0.531457,
        0.209440,2.119166,0.100552,
        0.209440,2.119564,2.754818,
        0.209440,2.122474,-0.648314,
        0.209440,2.125183,2.445598,
        0.209440,2.129501,0.238290,
        0.209440,2.131178,0.386663,
        0.209440,2.131223,-1.174311,
        0.209440,2.134269,-1.948042,
        0.209440,2.134684,0.907065,
        0.209440,2.137615,2.988806,
        0.209440,2.145635,-1.327708,
        0.209440,2.146467,2.055081,
        0.209440,2.146582,-1.793788,
        0.209440,2.149210,-1.482321,
        0.209440,2.152901,-1.636788,
        0.209440,2.156337,1.151558,
        0.209440,2.156474,-2.912065,
        0.209440,2.157194,-0.792563,
        0.209440,2.166724,-0.401638,
        0.209440,2.166980,-0.134879,
        0.209440,2.173644,-2.403429,
        0.209440,2.175711,-2.655523,
        0.209440,2.177132,2.600419,
        0.209440,2.178879,0.767722,
        0.209440,2.180752,1.908485,
        0.209440,2.181338,2.313721,
        0.209440,2.185580,1.287673,
        0.209440,2.195400,-2.217546,
        0.209440,2.200418,-3.056442,
        0.209440,2.201829,-0.001735,
        0.209440,2.203804,1.756176,
        0.209440,2.211602,1.443930,
        0.209440,2.212686,-0.544809,
        0.209440,2.213937,0.621892,
        0.209440,2.214075,1.600691,
        0.209440,2.216256,1.020029,
        0.209440,2.219060,2.886631,
        0.209440,2.223375,-0.906185,
        0.209440,2.226002,-1.060248,
        0.209440,2.226413,-2.064034,
        0.209440,2.230554,2.171334,
        0.209440,2.232413,2.735777,
        0.209440,2.233709,3.070830,
        0.209440,2.234383,-2.792065,
        0.209440,2.237183,0.468799,
        0.209440,2.237695,-0.271388,
        0.209440,2.239419,0.151693,
        0.209440,2.243347,-2.521049,
        0.209440,2.244035,-1.229456,
        0.209440,2.244355,0.309899,
        0.209440,2.245175,2.460236,
        0.209440,2.245628,-1.899216,
        0.209440,2.245973,-0.702234,
        0.209440,2.255601,-1.728131,
        0.209440,2.256702,-1.401367,
        0.209440,2.261466,-1.557500,
        0.209440,2.267390,0.878499,
        0.209440,2.272099,2.018594,
        0.209440,2.278618,1.170386,
        0.209440,2.283304,-2.345584,
        0.209440,2.285175,-2.943090,
        0.209440,2.293000,-0.124451,
        0.209440,2.294058,-0.421070,
        0.209440,2.303283,1.855938,
        0.209440,2.304968,2.608978,
        0.209440,2.305347,2.311346,
        0.209440,2.307188,-2.656134,
        0.209440,2.307301,1.336200,
        0.209440,2.309655,0.726020,
        0.209440,2.317634,-2.185225,
        0.209440,2.321510,1.685664,
        0.209440,2.324376,-3.111006,
        0.209440,2.324731,1.511186,
        0.209440,2.329587,0.035498,
        0.209440,2.330158,2.977264,
        0.209440,2.335096,-0.588468,
        0.209440,2.336046,-1.115799,
        0.209440,2.338060,2.803613,
        0.209440,2.338083,-0.943400,
        0.209440,2.340210,0.562145,
        0.209440,2.341931,1.023926,
        0.209440,2.343265,-0.779061,
        0.209440,2.343434,-2.013692,
        0.209440,2.353029,-1.310529,
        0.209440,2.354868,0.211543,
        0.209440,2.355120,0.388593,
        0.209440,2.355165,-1.833173,
        0.209440,2.355462,2.151441,
        0.209440,2.362948,-2.487477,
        0.209440,2.363101,-2.810456,
        0.209440,2.363394,-0.272918,
        0.209440,2.364509,-1.644533,
        0.209440,2.370465,2.466824,
        0.209440,2.386110,-1.481351,
        0.209440,2.394960,0.862133,
        0.209440,2.395716,1.977184,
        0.209440,2.396339,1.200900,
        0.209440,2.407941,-2.316753,
        0.209440,2.409450,-2.988432,
        0.209440,2.417202,-0.103363,
        0.209440,2.417386,-0.443097,
        0.209440,2.417950,2.666662,
        0.209440,2.423190,1.789783,
        0.209440,2.423692,1.392696,
        0.209440,2.431122,3.086379,
        0.209440,2.431406,-2.647820,
        0.209440,2.431584,2.307631,
        0.209440,2.434090,1.592818,
        0.209440,2.435993,0.683844,
        0.209440,2.440898,-2.133300,
        0.209440,2.442222,-1.195837,
        0.209440,2.443112,2.883761,
        0.209440,2.449191,-0.633651,
        0.209440,2.450821,0.084741,
        0.209440,2.453499,-1.008861,
        0.209440,2.460010,-1.938011,
        0.209440,2.460199,-0.822008,
        0.209440,2.460902,0.490079,
        0.209440,2.461774,-1.734913,
        0.209440,2.463986,1.034407,
        0.209440,2.467647,0.287667,
        0.209440,2.479849,2.126925,
        0.209440,2.485907,-2.835471,
        0.209440,2.485945,-1.374853,
        0.209440,2.487747,2.505912,
        0.209440,2.487957,-0.274279,
        0.209440,2.489741,-2.468428,
        0.209440,2.512538,1.239900,
        0.209440,2.517528,1.923289,
        0.209440,2.518454,-1.565852,
        0.209440,2.519348,0.843503,
        0.209440,2.522766,-3.059231,
        0.209440,2.528686,2.737228,
        0.209440,2.534798,-2.266671,
        0.209440,2.536533,1.466274,
        0.209440,2.538504,1.699812,
        0.209440,2.539516,-0.476942,
        0.209440,2.539532,-0.074548,
        0.209440,2.543197,2.990252,
        0.209440,2.554022,2.320619,
        0.209440,2.556919,-2.655005,
        0.209440,2.558146,0.627191,
        0.209440,2.558223,-1.164189,
        0.209440,2.563452,-2.044028,
        0.209440,2.568462,0.150655,
        0.209440,2.569130,-0.700837,
        0.209440,2.570050,-1.804492,
        0.209440,2.572710,-0.934665,
        0.209440,2.575994,0.391718,
        0.209440,2.584774,1.046703,
        0.209440,2.601432,2.102535,
        0.209440,2.603559,2.555114,
        0.209440,2.604150,-2.885208,
        0.209440,2.608630,-1.378011,
        0.209440,2.611256,-0.279110,
        0.209440,2.617344,-2.439480,
        0.209440,2.627402,1.293552,
        0.209440,2.635396,1.848850,
        0.209440,2.635686,2.834950,
        0.209440,2.637087,3.132730,
        0.209440,2.642029,0.813274,
        0.209440,2.645260,1.570203,
        0.209440,2.646182,-1.614036,
        0.209440,2.660113,-2.186513,
        0.209440,2.660483,-0.034608,
        0.209440,2.661713,-0.529511,
        0.209440,2.672720,-1.110422,
        0.209440,2.676626,2.335353,
        0.209440,2.676638,0.540364,
        0.209440,2.676780,-1.901158,
        0.209440,2.681411,-2.667965,
        0.209440,2.682152,0.245751,
        0.209440,2.682859,-0.817071,
        0.209440,2.705107,1.062142,
        0.209440,2.717705,2.625995,
        0.209440,2.721001,2.060097,
        0.209440,2.722319,-2.954069,
        0.209440,2.729840,-1.368237,
        0.209440,2.734349,-0.289298,
        0.209440,2.738409,2.975671,
        0.209440,2.740366,1.374341,
        0.209440,2.744633,-2.391311,
        0.209440,2.747398,1.727166,
        0.209440,2.763799,0.760372,
        0.209440,2.768238,-1.679243,
        0.209440,2.779245,-2.046268,
        0.209440,2.780208,-0.626453,
        0.209440,2.780471,0.027732,
        0.209440,2.784696,-1.013128,
        0.209440,2.789817,0.395431,
        0.209440,2.801154,2.353143,
        0.209440,2.807091,-2.687477,
        0.209440,2.825762,1.085035,
        0.209440,2.830209,2.734943,
        0.209440,2.833517,-3.085508,
        0.209440,2.838301,1.980061,
        0.209440,2.849871,1.512012,
        0.209440,2.850089,-1.343899,
        0.209440,2.856905,-0.318441,
        0.209440,2.870445,-2.291001,
        0.209440,2.885017,0.655562,
        0.209440,2.886428,-1.793217,
        0.209440,2.890337,-0.825633,
        0.209440,2.898932,0.139973,
        0.209440,2.931927,-2.728768,
        0.209440,2.932096,2.377790,
        0.209440,2.937409,2.947245,
        0.209440,2.948048,1.133127,
        0.209440,2.950896,1.797889,
        0.209440,2.970019,-1.282382,
        0.209440,2.978333,-0.412090,
        0.209440,2.996437,-2.089121,
        0.209440,3.011274,0.434110,
        0.209440,3.050410,3.138492,
        0.209440,3.064492,1.647259,
        0.209440,3.088868,-0.989501,
        0.314159,0.053884,-2.408394,
        0.314159,0.077645,1.453433,
        0.314159,0.088408,-0.311160,
        0.314159,0.142314,2.590387,
        0.314159,0.147206,-1.385164,
        0.314159,0.163506,0.573203,
        0.314159,0.165485,-2.875396,
        0.314159,0.182564,-2.125821,
        0.314159,0.190569,1.929141,
        0.314159,0.201102,1.249082,
        0.314159,0.201397,-0.716646,
        0.314159,0.210568,-0.069559,
        0.314159,0.251672,2.848900,
        0.314159,0.266262,-1.678068,
        0.314159,0.269108,-1.215563,
        0.314159,0.273777,-2.506756,
        0.314159,0.279854,0.814581,
        0.314159,0.281664,-2.969147,
        0.314159,0.283625,2.405603,
        0.314159,0.284058,0.357350,
        0.314159,0.294107,1.593880,
        0.314159,0.302491,-0.400410,
        0.314159,0.308059,2.021491,
        0.314159,0.312219,-2.072057,
        0.314159,0.321905,-0.829121,
        0.314159,0.329984,1.201621,
        0.314159,0.339714,-0.013712,
        0.314159,0.368784,2.974974,
        0.314159,0.376428,-1.495532,
        0.314159,0.378103,2.646436,
        0.314159,0.385088,0.590355,
        0.314159,0.386248,-2.690943,
        0.314159,0.389851,-2.340031,
        0.314159,0.391502,-1.171374,
        0.314159,0.398249,-1.812694,
        0.314159,0.400542,-2.995729,
        0.314159,0.404273,0.918163,
        0.314159,0.405617,1.773533,
        0.314159,0.409009,-0.570243,
        0.314159,0.410700,0.270433,
        0.314159,0.415893,1.464890,
        0.314159,0.420680,2.088164,
        0.314159,0.424766,-0.258729,
        0.314159,0.426507,2.367872,
        0.314159,0.441292,-0.890727,
        0.314159,0.451839,-2.078193,
        0.314159,0.459270,1.191562,
        0.314159,0.470622,0.014002,
        0.314159,0.481292,-2.529098,
        0.314159,0.482460,2.849843,
        0.314159,0.492205,-1.397963,
        0.314159,0.493314,3.117264,
        0.314159,0.497870,0.727828,
        0.314159,0.499783,-1.654379,
        0.314159,0.502651,0.463333,
        0.314159,0.505463,2.594557,
        0.314159,0.506104,-2.874910,
        0.314159,0.513579,-1.147354,
        0.314159,0.518569,1.644237,
        0.314159,0.519001,1.900488,
        0.314159,0.519277,-0.673809,
        0.314159,0.524769,-0.421955,
        0.314159,0.525564,-2.288838,
        0.314159,0.532541,0.978737,
        0.314159,0.534610,2.154808,
        0.314159,0.535020,-1.891846,
        0.314159,0.540208,1.400497,
        0.314159,0.542006,0.228337,
        0.314159,0.550762,-0.180771,
        0.314159,0.560531,2.378474,
        0.314159,0.561629,-0.924576,
        0.314159,0.567465,-2.679392,
        0.314159,0.588527,1.181755,
        0.314159,0.591643,-2.098364,
        0.314159,0.594963,2.997575,
        0.314159,0.599643,-3.045235,
        0.314159,0.599962,2.773556,
        0.314159,0.602620,0.029270,
        0.314159,0.605711,0.598454,
        0.314159,0.607708,-1.541026,
        0.314159,0.609360,-2.464379,
        0.314159,0.610738,-1.326786,
        0.314159,0.616968,0.820489,
        0.314159,0.626379,1.773610,
        0.314159,0.628626,-1.751180,
        0.314159,0.630671,0.391949,
        0.314159,0.632622,-0.737642,
        0.314159,0.632708,2.566722,
        0.314159,0.632797,1.989858,
        0.314159,0.634455,-0.534790,
        0.314159,0.636205,1.561170,
        0.314159,0.637681,-1.123423,
        0.314159,0.641441,-2.842109,
        0.314159,0.644807,-0.327139,
        0.314159,0.652440,2.197129,
        0.314159,0.660897,-2.278593,
        0.314159,0.664008,1.360091,
        0.314159,0.670850,-1.945257,
        0.314159,0.671396,1.004859,
        0.314159,0.673985,0.203855,
        0.314159,0.680326,-0.133370,
        0.314159,0.685083,-0.942363,
        0.314159,0.688204,-2.632204,
        0.314159,0.689777,2.385011,
        0.314159,0.700065,3.115430,
        0.314159,0.704080,2.911636,
        0.314159,0.712883,1.174216,
        0.314159,0.715701,0.699906,
        0.314159,0.718851,-1.442215,
        0.314159,0.721019,2.721536,
        0.314159,0.725300,-2.975134,
        0.314159,0.727136,-1.633288,
        0.314159,0.727399,-2.120567,
        0.314159,0.730751,0.522488,
        0.314159,0.732892,-1.254330,
        0.314159,0.734333,-2.446184,
        0.314159,0.735125,0.037270,
        0.314159,0.737893,1.866463,
        0.314159,0.740650,1.681075,
        0.314159,0.745971,-0.781286,
        0.314159,0.749432,-0.442559,
        0.314159,0.750102,2.050636,
        0.314159,0.751046,0.863821,
        0.314159,0.757018,-0.608833,
        0.314159,0.757046,-1.816019,
        0.314159,0.757389,1.501184,
        0.314159,0.759800,2.546181,
        0.314159,0.759859,0.351923,
        0.314159,0.761141,-2.781102,
        0.314159,0.769035,-1.078574,
        0.314159,0.770166,-0.266604,
        0.314159,0.772218,2.227380,
        0.314159,0.788382,1.331051,
        0.314159,0.790269,-2.282409,
        0.314159,0.803410,-1.984567,
        0.314159,0.803813,0.191520,
        0.314159,0.806832,3.022155,
        0.314159,0.807476,-2.597675,
        0.314159,0.812512,-0.107335,
        0.314159,0.814161,1.010949,
        0.314159,0.817732,-3.084486,
        0.314159,0.817846,2.390688,
        0.314159,0.819323,2.845960,
        0.314159,0.819430,-0.915017,
        0.314159,0.829198,-1.532491,
        0.314159,0.829531,0.665695,
        0.314159,0.829816,1.167147,
        0.314159,0.833613,-1.357492,
        0.314159,0.842945,-2.907318,
        0.314159,0.844883,2.679708,
        0.314159,0.848189,-1.705003,
        0.314159,0.849888,1.767901,
        0.314159,0.851126,0.493370,
        0.314159,0.852911,1.933394,
        0.314159,0.857516,-2.432863,
        0.314159,0.858261,-1.191729,
        0.314159,0.858279,-2.138895,
        0.314159,0.859723,-0.734928,
        0.314159,0.859754,1.603292,
        0.314159,0.861655,-0.394821,
        0.314159,0.864300,0.043549,
        0.314159,0.866940,0.851229,
        0.314159,0.869748,2.095969,
        0.314159,0.877352,-0.574668,
        0.314159,0.879254,-2.733338,
        0.314159,0.882554,1.443757,
        0.314159,0.884060,0.334199,
        0.314159,0.885626,-1.866061,
        0.314159,0.889834,2.529957,
        0.314159,0.893140,2.251698,
        0.314159,0.897463,-0.240822,
        0.314159,0.898758,-1.031948,
        0.314159,0.913918,1.287618,
        0.314159,0.916555,-2.287173,
        0.314159,0.918305,3.107845,
        0.314159,0.919522,2.948277,
        0.314159,0.927138,-2.569686,
        0.314159,0.929031,0.185383,
        0.314159,0.933766,-3.015917,
        0.314159,0.934873,-2.010822,
        0.314159,0.935636,-1.445855,
        0.314159,0.936702,-0.866461,
        0.314159,0.937395,0.729108,
        0.314159,0.937910,2.787575,
        0.314159,0.939144,-0.092962,
        0.314159,0.941718,1.125085,
        0.314159,0.942873,-1.607680,
        0.314159,0.943701,0.980328,
        0.314159,0.947774,2.395256,
        0.314159,0.952433,0.584347,
        0.314159,0.954456,-1.287293,
        0.314159,0.958731,-2.852552,
        0.314159,0.961838,1.839284,
        0.314159,0.961898,-0.469425,
        0.314159,0.965426,1.686615,
        0.314159,0.970108,1.990771,
        0.314159,0.972820,-1.762780,
        0.314159,0.973500,-0.711881,
        0.314159,0.975253,2.641082,
        0.314159,0.977824,0.441928,
        0.314159,0.981485,-2.424766,
        0.314159,0.981877,1.537100,
        0.314159,0.982609,-1.136920,
        0.314159,0.984265,-2.156073,
        0.314159,0.989150,0.045493,
        0.314159,0.993052,2.137478,
        0.314159,0.997766,-2.693871,
        0.314159,0.999847,-0.334738,
        0.314159,1.001920,0.848054,
        0.314159,1.006146,1.389734,
        0.314159,1.012180,0.304066,
        0.314159,1.016938,2.270674,
        0.314159,1.017371,-0.979790,
        0.314159,1.019495,-1.894487,
        0.314159,1.023796,-0.583691,
        0.314159,1.026525,2.506945,
        0.314159,1.027004,-0.203653,
        0.314159,1.028027,3.036573,
        0.314159,1.031435,1.241021,
        0.314159,1.032316,2.884221,
        0.314159,1.035863,-3.110475,
        0.314159,1.042085,-2.292950,
        0.314159,1.043272,-1.522986,
        0.314159,1.046265,-2.958847,
        0.314159,1.049272,-1.376344,
        0.314159,1.050039,-2.552086,
        0.314159,1.053682,0.689863,
        0.314159,1.055458,0.170394,
        0.314159,1.056622,-0.831866,
        0.314159,1.057515,1.097998,
        0.314159,1.058706,-2.036438,
        0.314159,1.061389,-1.672248,
        0.314159,1.063758,2.743093,
        0.314159,1.070157,-0.073422,
        0.314159,1.072154,1.761506,
        0.314159,1.074029,1.909193,
        0.314159,1.074204,0.545276,
        0.314159,1.074714,-2.805147,
        0.314159,1.076262,0.966104,
        0.314159,1.077642,-1.227073,
        0.314159,1.084151,1.618525,
        0.314159,1.085085,2.377673,
        0.314159,1.086822,-0.441112,
        0.314159,1.087142,2.058815,
        0.314159,1.098383,-0.696430,
        0.314159,1.099138,-1.084033,
        0.314159,1.103208,1.477773,
        0.314159,1.103783,0.408254,
        0.314159,1.106275,-2.420800,
        0.314159,1.108079,2.612943,
        0.314159,1.109345,-2.172718,
        0.314159,1.115957,-1.787597,
        0.314159,1.118930,0.045983,
        0.314159,1.119777,0.834702,
        0.314159,1.121943,-2.669042,
        0.314159,1.124930,2.219917,
        0.314159,1.125217,1.336294,
        0.314159,1.130958,2.972931,
        0.314159,1.131400,-0.306629,
        0.314159,1.137135,-0.936514,
        0.314159,1.139125,-1.924528,
        0.314159,1.139877,0.276382,
        0.314159,1.142003,-3.058022,
        0.314159,1.147363,-0.558448,
        0.314159,1.150348,-0.176418,
        0.314159,1.151198,1.197016,
        0.314159,1.153930,-1.597104,
        0.314159,1.155389,2.837148,
        0.314159,1.155458,-1.460623,
        0.314159,1.157331,-2.906362,
        0.314159,1.158594,2.487618,
        0.314159,1.162103,-1.322553,
        0.314159,1.165407,3.092861,
        0.314159,1.168025,0.717075,
        0.314159,1.168406,-2.301519,
        0.314159,1.174782,-2.538755,
        0.314159,1.177998,1.061256,
        0.314159,1.178898,1.827305,
        0.314159,1.181331,-2.060330,
        0.314159,1.182540,-0.798952,
        0.314159,1.184569,0.147517,
        0.314159,1.184734,1.968919,
        0.314159,1.184804,0.593386,
        0.314159,1.188278,1.690808,
        0.314159,1.194360,2.110440,
        0.314159,1.194776,2.710455,
        0.314159,1.197745,-1.173705,
        0.314159,1.199898,-2.777257,
        0.314159,1.201164,2.355931,
        0.314159,1.203760,1.556404,
        0.314159,1.205845,-0.415993,
        0.314159,1.206585,0.933648,
        0.314159,1.206950,-0.037248,
        0.314159,1.209341,0.471162,
        0.314159,1.213483,-1.039417,
        0.314159,1.218146,-0.672643,
        0.314159,1.221608,1.421669,
        0.314159,1.226043,-1.693533,
        0.314159,1.232870,-1.822773,
        0.314159,1.234375,-2.421194,
        0.314159,1.234531,-2.190148,
        0.314159,1.239414,0.348959,
        0.314159,1.240183,2.587255,
        0.314159,1.242312,2.941234,
        0.314159,1.244075,1.286644,
        0.314159,1.245961,-2.997461,
        0.314159,1.246495,-2.646872,
        0.314159,1.247736,2.242810,
        0.314159,1.253604,0.812222,
        0.314159,1.257816,-0.894317,
        0.314159,1.258247,-1.953634,
        0.314159,1.258993,-0.284099,
        0.314159,1.262380,-0.157609,
        0.314159,1.262791,-1.549229,
        0.314159,1.263772,-0.533073,
        0.314159,1.266731,-1.412461,
        0.314159,1.268591,-3.122146,
        0.314159,1.269614,-1.275246,
        0.314159,1.270745,1.152533,
        0.314159,1.273606,0.224545,
        0.314159,1.279716,2.812101,
        0.314159,1.283786,2.462124,
        0.314159,1.284798,0.078209,
        0.314159,1.285402,-2.882049,
        0.314159,1.286936,1.886053,
        0.314159,1.289408,0.683122,
        0.314159,1.294736,1.756184,
        0.314159,1.296978,-2.311511,
        0.314159,1.298546,2.020601,
        0.314159,1.300924,1.022478,
        0.314159,1.303305,3.043152,
        0.314159,1.304210,-2.084225,
        0.314159,1.305241,-2.530249,
        0.314159,1.307513,1.628003,
        0.314159,1.310722,0.558267,
        0.314159,1.314142,2.139329,
        0.314159,1.316087,-0.777971,
        0.314159,1.317907,-1.121023,
        0.314159,1.321019,-2.749776,
        0.314159,1.321609,2.687717,
        0.314159,1.322568,1.499415,
        0.314159,1.322972,-0.395699,
        0.314159,1.325286,-0.994192,
        0.314159,1.329146,-0.648513,
        0.314159,1.329795,-0.042570,
        0.314159,1.331651,2.344497,
        0.314159,1.336522,0.435098,
        0.314159,1.339410,0.900087,
        0.314159,1.339550,1.370353,
        0.314159,1.340441,-1.731580,
        0.314159,1.351855,-1.855234,
        0.314159,1.360161,-3.018404,
        0.314159,1.361031,-1.613971,
        0.314159,1.361196,2.914780,
        0.314159,1.361998,-2.207254,
        0.314159,1.362302,1.239384,
        0.314159,1.365228,2.563808,
        0.314159,1.365452,0.310522,
        0.314159,1.367028,-2.420041,
        0.314159,1.373422,-1.484281,
        0.314159,1.374245,-1.226417,
        0.314159,1.375020,-0.170393,
        0.314159,1.376560,-1.355481,
        0.314159,1.377070,-1.980984,
        0.314159,1.377578,-2.636339,
        0.314159,1.378096,0.774626,
        0.314159,1.379086,-0.508192,
        0.314159,1.382562,0.177141,
        0.314159,1.388849,2.233847,
        0.314159,1.389268,-0.886349,
        0.314159,1.391324,1.109396,
        0.314159,1.391699,-0.288725,
        0.314159,1.392869,3.136903,
        0.314159,1.393612,1.952016,
        0.314159,1.395545,-2.850807,
        0.314159,1.401926,1.822930,
        0.314159,1.402149,2.788898,
        0.314159,1.409552,0.647980,
        0.314159,1.410840,0.056219,
        0.314159,1.411944,2.444377,
        0.314159,1.413639,1.697982,
        0.314159,1.426479,1.573427,
        0.314159,1.426725,2.102242,
        0.314159,1.427868,0.984637,
        0.314159,1.429580,-2.107232,
        0.314159,1.431375,-2.313269,
        0.314159,1.431907,3.015410,
        0.314159,1.432325,-1.029054,
        0.314159,1.433341,0.524268,
        0.314159,1.435181,-0.739771,
        0.314159,1.437208,-2.529085,
        0.314159,1.439041,-0.617966,
        0.314159,1.441473,1.448544,
        0.314159,1.444527,2.664744,
        0.314159,1.450513,-0.400625,
        0.314159,1.451907,-0.066823,
        0.314159,1.452754,-2.738631,
        0.314159,1.454061,-2.950416,
        0.314159,1.456736,1.323154,
        0.314159,1.458370,0.400433,
        0.314159,1.459314,-1.765003,
        0.314159,1.459902,-1.145835,
        0.314159,1.464952,2.329245,
        0.314159,1.466050,0.862104,
        0.314159,1.471330,-1.886160,
        0.314159,1.475119,-1.648255,
        0.314159,1.480878,1.194576,
        0.314159,1.480942,0.273235,
        0.314159,1.481086,2.890269,
        0.314159,1.484118,-1.529156,
        0.314159,1.484741,-1.279219,
        0.314159,1.487681,-1.406069,
        0.314159,1.487886,-3.066037,
        0.314159,1.490555,2.543284,
        0.314159,1.492424,-0.190932,
        0.314159,1.495002,-0.839810,
        0.314159,1.495297,2.007188,
        0.314159,1.496414,-2.009267,
        0.314159,1.498261,-2.425270,
        0.314159,1.498907,-2.207327,
        0.314159,1.500204,0.737638,
        0.314159,1.502469,0.151156,
        0.314159,1.507448,-0.514782,
        0.314159,1.509287,2.204647,
        0.314159,1.510518,1.884920,
        0.314159,1.511948,-2.632360,
        0.314159,1.515235,1.068565,
        0.314159,1.522237,3.102018,
        0.314159,1.522373,1.764991,
        0.314159,1.522416,-0.954956,
        0.314159,1.523102,2.765920,
        0.314159,1.524811,-2.844718,
        0.314159,1.527261,-0.304916,
        0.314159,1.529788,0.613208,
        0.314159,1.533648,1.644808,
        0.314159,1.535018,0.028869,
        0.314159,1.541082,2.424608,
        0.314159,1.546010,1.523462,
        0.314159,1.551135,-0.642994,
        0.314159,1.553895,0.489607,
        0.314159,1.555041,0.947813,
        0.314159,1.559671,-1.074934,
        0.314159,1.560345,2.985899,
        0.314159,1.560487,-2.323400,
        0.314159,1.561127,1.401111,
        0.314159,1.567193,2.643172,
        0.314159,1.568892,2.098362,
        0.314159,1.571702,-2.098435,
        0.314159,1.572701,-2.528730,
        0.314159,1.573572,1.278593,
        0.314159,1.574849,-0.091505,
        0.314159,1.576334,0.365236,
        0.314159,1.577029,-2.962888,
        0.314159,1.577688,-1.814941,
        0.314159,1.579542,-0.755917,
        0.314159,1.579624,-1.199759,
        0.314159,1.581515,-0.419414,
        0.314159,1.584581,-2.738702,
        0.314159,1.586948,-1.696296,
        0.314159,1.588901,2.301496,
        0.314159,1.589936,0.825020,
        0.314159,1.590018,-1.941652,
        0.314159,1.596968,-1.328358,
        0.314159,1.597178,-1.575607,
        0.314159,1.600470,-1.453196,
        0.314159,1.601408,0.242409,
        0.314159,1.602259,1.152656,
        0.314159,1.603064,2.865774,
        0.314159,1.612085,1.968802,
        0.314159,1.612720,-3.080498,
        0.314159,1.616061,2.521531,
        0.314159,1.616900,-0.207457,
        0.314159,1.620988,-0.875920,
        0.314159,1.622114,0.701509,
        0.314159,1.622749,0.121426,
        0.314159,1.622993,-2.221629,
        0.314159,1.629346,-0.542669,
        0.314159,1.630007,1.844552,
        0.314159,1.634747,-2.426198,
        0.314159,1.642346,1.722617,
        0.314159,1.643925,1.034629,
        0.314159,1.644848,2.743774,
        0.314159,1.645258,2.188109,
        0.314159,1.645417,-2.634725,
        0.314159,1.647395,3.085437,
        0.314159,1.648538,-2.855044,
        0.314159,1.649932,0.576808,
        0.314159,1.652224,-0.996105,
        0.314159,1.652871,1.600681,
        0.314159,1.657997,-0.325127,
        0.314159,1.661936,0.001037,
        0.314159,1.665187,1.477402,
        0.314159,1.665496,2.397872,
        0.314159,1.673267,-0.665453,
        0.314159,1.673557,0.452526,
        0.314159,1.679277,-1.122421,
        0.314159,1.679696,0.912482,
        0.314159,1.682420,1.353922,
        0.314159,1.685031,-2.009038,
        0.314159,1.685196,2.967046,
        0.314159,1.690107,2.620910,
        0.314159,1.690453,-2.123076,
        0.314159,1.691693,-1.871572,
        0.314159,1.691817,2.061688,
        0.314159,1.692644,1.237670,
        0.314159,1.696386,0.328123,
        0.314159,1.697162,-1.249026,
        0.314159,1.697488,-2.322785,
        0.314159,1.698726,-2.973576,
        0.314159,1.699172,-1.749717,
        0.314159,1.704478,-0.106295,
        0.314159,1.707244,-0.444770,
        0.314159,1.707500,-2.531497,
        0.314159,1.707714,-0.787321,
        0.314159,1.708430,-1.627552,
        0.314159,1.710583,-1.376890,
        0.314159,1.712415,-2.747921,
        0.314159,1.713125,-1.503073,
        0.314159,1.713452,0.789419,
        0.314159,1.720612,2.279888,
        0.314159,1.724852,0.202831,
        0.314159,1.724868,2.846246,
        0.314159,1.728940,1.931535,
        0.314159,1.732838,1.127584,
        0.314159,1.735719,-3.092485,
        0.314159,1.739676,2.496017,
        0.314159,1.739971,-0.228876,
        0.314159,1.743328,0.663347,
        0.314159,1.744927,-0.911892,
        0.314159,1.745952,0.084733,
        0.314159,1.748077,1.804354,
        0.314159,1.752959,-0.567548,
        0.314159,1.760169,1.678132,
        0.314159,1.763887,-2.218450,
        0.314159,1.766737,2.722752,
        0.314159,1.767999,1.004737,
        0.314159,1.769149,1.550298,
        0.314159,1.770119,-2.427175,
        0.314159,1.770460,2.154208,
        0.314159,1.770514,3.070982,
        0.314159,1.770735,0.537245,
        0.314159,1.772673,-2.864760,
        0.314159,1.773527,-1.039582,
        0.314159,1.776419,-2.642421,
        0.314159,1.780769,1.416330,
        0.314159,1.786634,-0.346247,
        0.314159,1.788829,-1.947553,
        0.314159,1.793737,0.410270,
        0.314159,1.794553,2.374618,
        0.314159,1.794699,-0.694061,
        0.314159,1.797469,-1.169105,
        0.314159,1.803705,0.881076,
        0.314159,1.804124,1.264249,
        0.314159,1.807825,-2.090455,
        0.314159,1.808822,-1.817375,
        0.314159,1.809020,2.949741,
        0.314159,1.812466,2.597595,
        0.314159,1.813159,2.025395,
        0.314159,1.813816,-1.298435,
        0.314159,1.815016,0.278951,
        0.314159,1.816510,-0.008631,
        0.314159,1.817772,-1.689625,
        0.314159,1.821360,-2.984352,
        0.314159,1.822312,-0.127211,
        0.314159,1.823386,-1.429807,
        0.314159,1.823834,-1.560483,
        0.314159,1.832060,-0.821627,
        0.314159,1.832690,-0.468440,
        0.314159,1.833945,-2.320278,
        0.314159,1.834787,0.751968,
        0.314159,1.840579,-2.536445,
        0.314159,1.841033,-2.756932,
        0.314159,1.846350,1.893983,
        0.314159,1.846644,1.113766,
        0.314159,1.846833,2.826543,
        0.314159,1.847325,2.249051,
        0.314159,1.854119,0.134161,
        0.314159,1.858863,-3.106022,
        0.314159,1.865046,1.763384,
        0.314159,1.865619,-0.952036,
        0.314159,1.865912,0.624453,
        0.314159,1.866694,2.473686,
        0.314159,1.867524,-0.240083,
        0.314159,1.875561,1.632831,
        0.314159,1.876049,-0.594378,
        0.314159,1.880509,1.500204,
        0.314159,1.883504,1.360457,
        0.314159,1.886971,-2.196665,
        0.314159,1.888979,0.982922,
        0.314159,1.889815,2.701570,
        0.314159,1.892144,-2.005340,
        0.314159,1.892348,0.496522,
        0.314159,1.893281,-1.083976,
        0.314159,1.893828,3.053453,
        0.314159,1.895117,2.121107,
        0.314159,1.898741,-2.875735,
        0.314159,1.904429,-2.428029,
        0.314159,1.908408,-2.648903,
        0.314159,1.911828,-0.364489,
        0.314159,1.912224,0.366565,
        0.314159,1.914854,1.214468,
        0.314159,1.915467,-1.883143,
        0.314159,1.915763,-0.724679,
        0.314159,1.915790,-1.216366,
        0.314159,1.922256,2.347806,
        0.314159,1.922616,0.844633,
        0.314159,1.925185,0.231170,
        0.314159,1.927474,-1.754616,
        0.314159,1.928342,-1.351174,
        0.314159,1.933328,-1.623328,
        0.314159,1.933878,0.034150,
        0.314159,1.934585,-1.487979,
        0.314159,1.934802,2.927634,
        0.314159,1.934862,1.989776,
        0.314159,1.936327,-0.099350,
        0.314159,1.938088,2.578507,
        0.314159,1.945677,-2.999805,
        0.314159,1.954572,-0.859208,
        0.314159,1.957868,-0.490784,
        0.314159,1.958989,0.712772,
        0.314159,1.960212,-2.307561,
        0.314159,1.964774,1.855507,
        0.314159,1.964971,-2.100533,
        0.314159,1.969112,1.088273,
        0.314159,1.971438,-2.765415,
        0.314159,1.972312,2.805134,
        0.314159,1.974950,2.220596,
        0.314159,1.975135,-2.539127,
        0.314159,1.982107,1.720829,
        0.314159,1.983195,-3.127984,
        0.314159,1.983318,-0.247182,
        0.314159,1.986720,-0.993220,
        0.314159,1.990381,1.586005,
        0.314159,1.990470,0.580283,
        0.314159,1.990523,1.450326,
        0.314159,1.993909,1.312897,
        0.314159,1.994694,2.452813,
        0.314159,1.999226,-0.620717,
        0.314159,2.005145,0.941012,
        0.314159,2.007301,0.142337,
        0.314159,2.014149,0.444138,
        0.314159,2.014338,-1.127392,
        0.314159,2.018131,3.021778,
        0.314159,2.019676,-1.980944,
        0.314159,2.019912,2.686612,
        0.314159,2.020545,2.088672,
        0.314159,2.026074,0.300945,
        0.314159,2.026196,-2.889057,
        0.314159,2.031959,-1.268178,
        0.314159,2.032917,-2.199917,
        0.314159,2.033572,-1.842884,
        0.314159,2.034712,-2.419505,
        0.314159,2.035818,-0.379479,
        0.314159,2.036004,-0.757062,
        0.314159,2.040501,-1.411488,
        0.314159,2.042874,-1.701367,
        0.314159,2.043060,-2.653746,
        0.314159,2.043729,-1.557938,
        0.314159,2.048538,0.806623,
        0.314159,2.049006,1.199930,
        0.314159,2.049502,-0.137989,
        0.314159,2.052181,2.326959,
        0.314159,2.052794,-0.008526,
        0.314159,2.057608,1.951726,
        0.314159,2.065568,2.567730,
        0.314159,2.070371,2.879808,
        0.314159,2.071256,-3.018961,
        0.314159,2.079746,-0.897699,
        0.314159,2.082784,1.042403,
        0.314159,2.083752,-0.511086,
        0.314159,2.084401,1.810502,
        0.314159,2.085731,0.671081,
        0.314159,2.097496,1.373172,
        0.314159,2.099185,1.667006,
        0.314159,2.100482,-2.301084,
        0.314159,2.103118,2.194745,
        0.314159,2.103223,1.521685,
        0.314159,2.103695,-2.777707,
        0.314159,2.107523,3.129581,
        0.314159,2.108277,-2.095907,
        0.314159,2.108891,-2.532166,
        0.314159,2.111003,-0.268632,
        0.314159,2.112276,-1.027246,
        0.314159,2.114380,0.531457,
        0.314159,2.119166,0.100552,
        0.314159,2.119564,2.754818,
        0.314159,2.122474,-0.648314,
        0.314159,2.125183,2.445598,
        0.314159,2.129501,0.238290,
        0.314159,2.131178,0.386663,
        0.314159,2.131223,-1.174311,
        0.314159,2.134269,-1.948042,
        0.314159,2.134684,0.907065,
        0.314159,2.137615,2.988806,
        0.314159,2.145635,-1.327708,
        0.314159,2.146467,2.055081,
        0.314159,2.146582,-1.793788,
        0.314159,2.149210,-1.482321,
        0.314159,2.152901,-1.636788,
        0.314159,2.156337,1.151558,
        0.314159,2.156474,-2.912065,
        0.314159,2.157194,-0.792563,
        0.314159,2.166724,-0.401638,
        0.314159,2.166980,-0.134879,
        0.314159,2.173644,-2.403429,
        0.314159,2.175711,-2.655523,
        0.314159,2.177132,2.600419,
        0.314159,2.178879,0.767722,
        0.314159,2.180752,1.908485,
        0.314159,2.181338,2.313721,
        0.314159,2.185580,1.287673,
        0.314159,2.195400,-2.217546,
        0.314159,2.200418,-3.056442,
        0.314159,2.201829,-0.001735,
        0.314159,2.203804,1.756176,
        0.314159,2.211602,1.443930,
        0.314159,2.212686,-0.544809,
        0.314159,2.213937,0.621892,
        0.314159,2.214075,1.600691,
        0.314159,2.216256,1.020029,
        0.314159,2.219060,2.886631,
        0.314159,2.223375,-0.906185,
        0.314159,2.226002,-1.060248,
        0.314159,2.226413,-2.064034,
        0.314159,2.230554,2.171334,
        0.314159,2.232413,2.735777,
        0.314159,2.233709,3.070830,
        0.314159,2.234383,-2.792065,
        0.314159,2.237183,0.468799,
        0.314159,2.237695,-0.271388,
        0.314159,2.239419,0.151693,
        0.314159,2.243347,-2.521049,
        0.314159,2.244035,-1.229456,
        0.314159,2.244355,0.309899,
        0.314159,2.245175,2.460236,
        0.314159,2.245628,-1.899216,
        0.314159,2.245973,-0.702234,
        0.314159,2.255601,-1.728131,
        0.314159,2.256702,-1.401367,
        0.314159,2.261466,-1.557500,
        0.314159,2.267390,0.878499,
        0.314159,2.272099,2.018594,
        0.314159,2.278618,1.170386,
        0.314159,2.283304,-2.345584,
        0.314159,2.285175,-2.943090,
        0.314159,2.293000,-0.124451,
        0.314159,2.294058,-0.421070,
        0.314159,2.303283,1.855938,
        0.314159,2.304968,2.608978,
        0.314159,2.305347,2.311346,
        0.314159,2.307188,-2.656134,
        0.314159,2.307301,1.336200,
        0.314159,2.309655,0.726020,
        0.314159,2.317634,-2.185225,
        0.314159,2.321510,1.685664,
        0.314159,2.324376,-3.111006,
        0.314159,2.324731,1.511186,
        0.314159,2.329587,0.035498,
        0.314159,2.330158,2.977264,
        0.314159,2.335096,-0.588468,
        0.314159,2.336046,-1.115799,
        0.314159,2.338060,2.803613,
        0.314159,2.338083,-0.943400,
        0.314159,2.340210,0.562145,
        0.314159,2.341931,1.023926,
        0.314159,2.343265,-0.779061,
        0.314159,2.343434,-2.013692,
        0.314159,2.353029,-1.310529,
        0.314159,2.354868,0.211543,
        0.314159,2.355120,0.388593,
        0.314159,2.355165,-1.833173,
        0.314159,2.355462,2.151441,
        0.314159,2.362948,-2.487477,
        0.314159,2.363101,-2.810456,
        0.314159,2.363394,-0.272918,
        0.314159,2.364509,-1.644533,
        0.314159,2.370465,2.466824,
        0.314159,2.386110,-1.481351,
        0.314159,2.394960,0.862133,
        0.314159,2.395716,1.977184,
        0.314159,2.396339,1.200900,
        0.314159,2.407941,-2.316753,
        0.314159,2.409450,-2.988432,
        0.314159,2.417202,-0.103363,
        0.314159,2.417386,-0.443097,
        0.314159,2.417950,2.666662,
        0.314159,2.423190,1.789783,
        0.314159,2.423692,1.392696,
        0.314159,2.431122,3.086379,
        0.314159,2.431406,-2.647820,
        0.314159,2.431584,2.307631,
        0.314159,2.434090,1.592818,
        0.314159,2.435993,0.683844,
        0.314159,2.440898,-2.133300,
        0.314159,2.442222,-1.195837,
        0.314159,2.443112,2.883761,
        0.314159,2.449191,-0.633651,
        0.314159,2.450821,0.084741,
        0.314159,2.453499,-1.008861,
        0.314159,2.460010,-1.938011,
        0.314159,2.460199,-0.822008,
        0.314159,2.460902,0.490079,
        0.314159,2.461774,-1.734913,
        0.314159,2.463986,1.034407,
        0.314159,2.467647,0.287667,
        0.314159,2.479849,2.126925,
        0.314159,2.485907,-2.835471,
        0.314159,2.485945,-1.374853,
        0.314159,2.487747,2.505912,
        0.314159,2.487957,-0.274279,
        0.314159,2.489741,-2.468428,
        0.314159,2.512538,1.239900,
        0.314159,2.517528,1.923289,
        0.314159,2.518454,-1.565852,
        0.314159,2.519348,0.843503,
        0.314159,2.522766,-3.059231,
        0.314159,2.528686,2.737228,
        0.314159,2.534798,-2.266671,
        0.314159,2.536533,1.466274,
        0.314159,2.538504,1.699812,
        0.314159,2.539516,-0.476942,
        0.314159,2.539532,-0.074548,
        0.314159,2.543197,2.990252,
        0.314159,2.554022,2.320619,
        0.314159,2.556919,-2.655005,
        0.314159,2.558146,0.627191,
        0.314159,2.558223,-1.164189,
        0.314159,2.563452,-2.044028,
        0.314159,2.568462,0.150655,
        0.314159,2.569130,-0.700837,
        0.314159,2.570050,-1.804492,
        0.314159,2.572710,-0.934665,
        0.314159,2.575994,0.391718,
        0.314159,2.584774,1.046703,
        0.314159,2.601432,2.102535,
        0.314159,2.603559,2.555114,
        0.314159,2.604150,-2.885208,
        0.314159,2.608630,-1.378011,
        0.314159,2.611256,-0.279110,
        0.314159,2.617344,-2.439480,
        0.314159,2.627402,1.293552,
        0.314159,2.635396,1.848850,
        0.314159,2.635686,2.834950,
        0.314159,2.637087,3.132730,
        0.314159,2.642029,0.813274,
        0.314159,2.645260,1.570203,
        0.314159,2.646182,-1.614036,
        0.314159,2.660113,-2.186513,
        0.314159,2.660483,-0.034608,
        0.314159,2.661713,-0.529511,
        0.314159,2.672720,-1.110422,
        0.314159,2.676626,2.335353,
        0.314159,2.676638,0.540364,
        0.314159,2.676780,-1.901158,
        0.314159,2.681411,-2.667965,
        0.314159,2.682152,0.245751,
        0.314159,2.682859,-0.817071,
        0.314159,2.705107,1.062142,
        0.314159,2.717705,2.625995,
        0.314159,2.721001,2.060097,
        0.314159,2.722319,-2.954069,
        0.314159,2.729840,-1.368237,
        0.314159,2.734349,-0.289298,
        0.314159,2.738409,2.975671,
        0.314159,2.740366,1.374341,
        0.314159,2.744633,-2.391311,
        0.314159,2.747398,1.727166,
        0.314159,2.763799,0.760372,
        0.314159,2.768238,-1.679243,
        0.314159,2.779245,-2.046268,
        0.314159,2.780208,-0.626453,
        0.314159,2.780471,0.027732,
        0.314159,2.784696,-1.013128,
        0.314159,2.789817,0.395431,
        0.314159,2.801154,2.353143,
        0.314159,2.807091,-2.687477,
        0.314159,2.825762,1.085035,
        0.314159,2.830209,2.734943,
        0.314159,2.833517,-3.085508,
        0.314159,2.838301,1.980061,
        0.314159,2.849871,1.512012,
        0.314159,2.850089,-1.343899,
        0.314159,2.856905,-0.318441,
        0.314159,2.870445,-2.291001,
        0.314159,2.885017,0.655562,
        0.314159,2.886428,-1.793217,
        0.314159,2.890337,-0.825633,
        0.314159,2.898932,0.139973,
        0.314159,2.931927,-2.728768,
        0.314159,2.932096,2.377790,
        0.314159,2.937409,2.947245,
        0.314159,2.948048,1.133127,
        0.314159,2.950896,1.797889,
        0.314159,2.970019,-1.282382,
        0.314159,2.978333,-0.412090,
        0.314159,2.996437,-2.089121,
        0.314159,3.011274,0.434110,
        0.314159,3.050410,3.138492,
        0.314159,3.064492,1.647259,
        0.314159,3.088868,-0.989501,
        0.418879,0.053884,-2.408394,
        0.418879,0.077645,1.453433,
        0.418879,0.088408,-0.311160,
        0.418879,0.142314,2.590387,
        0.418879,0.147206,-1.385164,
        0.418879,0.163506,0.573203,
        0.418879,0.165485,-2.875396,
        0.418879,0.182564,-2.125821,
        0.418879,0.190569,1.929141,
        0.418879,0.201102,1.249082,
        0.418879,0.201397,-0.716646,
        0.418879,0.210568,-0.069559,
        0.418879,0.251672,2.848900,
        0.418879,0.266262,-1.678068,
        0.418879,0.269108,-1.215563,
        0.418879,0.273777,-2.506756,
        0.418879,0.279854,0.814581,
        0.418879,0.281664,-2.969147,
        0.418879,0.283625,2.405603,
        0.418879,0.284058,0.357350,
        0.418879,0.294107,1.593880,
        0.418879,0.302491,-0.400410,
        0.418879,0.308059,2.021491,
        0.418879,0.312219,-2.072057,
        0.418879,0.321905,-0.829121,
        0.418879,0.329984,1.201621,
        0.418879,0.339714,-0.013712,
        0.418879,0.368784,2.974974,
        0.418879,0.376428,-1.495532,
        0.418879,0.378103,2.646436,
        0.418879,0.385088,0.590355,
        0.418879,0.386248,-2.690943,
        0.418879,0.389851,-2.340031,
        0.418879,0.391502,-1.171374,
        0.418879,0.398249,-1.812694,
        0.418879,0.400542,-2.995729,
        0.418879,0.404273,0.918163,
        0.418879,0.405617,1.773533,
        0.418879,0.409009,-0.570243,
        0.418879,0.410700,0.270433,
        0.418879,0.415893,1.464890,
        0.418879,0.420680,2.088164,
        0.418879,0.424766,-0.258729,
        0.418879,0.426507,2.367872,
        0.418879,0.441292,-0.890727,
        0.418879,0.451839,-2.078193,
        0.418879,0.459270,1.191562,
        0.418879,0.470622,0.014002,
        0.418879,0.481292,-2.529098,
        0.418879,0.482460,2.849843,
        0.418879,0.492205,-1.397963,
        0.418879,0.493314,3.117264,
        0.418879,0.497870,0.727828,
        0.418879,0.499783,-1.654379,
        0.418879,0.502651,0.463333,
        0.418879,0.505463,2.594557,
        0.418879,0.506104,-2.874910,
        0.418879,0.513579,-1.147354,
        0.418879,0.518569,1.644237,
        0.418879,0.519001,1.900488,
        0.418879,0.519277,-0.673809,
        0.418879,0.524769,-0.421955,
        0.418879,0.525564,-2.288838,
        0.418879,0.532541,0.978737,
        0.418879,0.534610,2.154808,
        0.418879,0.535020,-1.891846,
        0.418879,0.540208,1.400497,
        0.418879,0.542006,0.228337,
        0.418879,0.550762,-0.180771,
        0.418879,0.560531,2.378474,
        0.418879,0.561629,-0.924576,
        0.418879,0.567465,-2.679392,
        0.418879,0.588527,1.181755,
        0.418879,0.591643,-2.098364,
        0.418879,0.594963,2.997575,
        0.418879,0.599643,-3.045235,
        0.418879,0.599962,2.773556,
        0.418879,0.602620,0.029270,
        0.418879,0.605711,0.598454,
        0.418879,0.607708,-1.541026,
        0.418879,0.609360,-2.464379,
        0.418879,0.610738,-1.326786,
        0.418879,0.616968,0.820489,
        0.418879,0.626379,1.773610,
        0.418879,0.628626,-1.751180,
        0.418879,0.630671,0.391949,
        0.418879,0.632622,-0.737642,
        0.418879,0.632708,2.566722,
        0.418879,0.632797,1.989858,
        0.418879,0.634455,-0.534790,
        0.418879,0.636205,1.561170,
        0.418879,0.637681,-1.123423,
        0.418879,0.641441,-2.842109,
        0.418879,0.644807,-0.327139,
        0.418879,0.652440,2.197129,
        0.418879,0.660897,-2.278593,
        0.418879,0.664008,1.360091,
        0.418879,0.670850,-1.945257,
        0.418879,0.671396,1.004859,
        0.418879,0.673985,0.203855,
        0.418879,0.680326,-0.133370,
        0.418879,0.685083,-0.942363,
        0.418879,0.688204,-2.632204,
        0.418879,0.689777,2.385011,
        0.418879,0.700065,3.115430,
        0.418879,0.704080,2.911636,
        0.418879,0.712883,1.174216,
        0.418879,0.715701,0.699906,
        0.418879,0.718851,-1.442215,
        0.418879,0.721019,2.721536,
        0.418879,0.725300,-2.975134,
        0.418879,0.727136,-1.633288,
        0.418879,0.727399,-2.120567,
        0.418879,0.730751,0.522488,
        0.418879,0.732892,-1.254330,
        0.418879,0.734333,-2.446184,
        0.418879,0.735125,0.037270,
        0.418879,0.737893,1.866463,
        0.418879,0.740650,1.681075,
        0.418879,0.745971,-0.781286,
        0.418879,0.749432,-0.442559,
        0.418879,0.750102,2.050636,
        0.418879,0.751046,0.863821,
        0.418879,0.757018,-0.608833,
        0.418879,0.757046,-1.816019,
        0.418879,0.757389,1.501184,
        0.418879,0.759800,2.546181,
        0.418879,0.759859,0.351923,
        0.418879,0.761141,-2.781102,
        0.418879,0.769035,-1.078574,
        0.418879,0.770166,-0.266604,
        0.418879,0.772218,2.227380,
        0.418879,0.788382,1.331051,
        0.418879,0.790269,-2.282409,
        0.418879,0.803410,-1.984567,
        0.418879,0.803813,0.191520,
        0.418879,0.806832,3.022155,
        0.418879,0.807476,-2.597675,
        0.418879,0.812512,-0.107335,
        0.418879,0.814161,1.010949,
        0.418879,0.817732,-3.084486,
        0.418879,0.817846,2.390688,
        0.418879,0.819323,2.845960,
        0.418879,0.819430,-0.915017,
        0.418879,0.829198,-1.532491,
        0.418879,0.829531,0.665695,
        0.418879,0.829816,1.167147,
        0.418879,0.833613,-1.357492,
        0.418879,0.842945,-2.907318,
        0.418879,0.844883,2.679708,
        0.418879,0.848189,-1.705003,
        0.418879,0.849888,1.767901,
        0.418879,0.851126,0.493370,
        0.418879,0.852911,1.933394,
        0.418879,0.857516,-2.432863,
        0.418879,0.858261,-1.191729,
        0.418879,0.858279,-2.138895,
        0.418879,0.859723,-0.734928,
        0.418879,0.859754,1.603292,
        0.418879,0.861655,-0.394821,
        0.418879,0.864300,0.043549,
        0.418879,0.866940,0.851229,
        0.418879,0.869748,2.095969,
        0.418879,0.877352,-0.574668,
        0.418879,0.879254,-2.733338,
        0.418879,0.882554,1.443757,
        0.418879,0.884060,0.334199,
        0.418879,0.885626,-1.866061,
        0.418879,0.889834,2.529957,
        0.418879,0.893140,2.251698,
        0.418879,0.897463,-0.240822,
        0.418879,0.898758,-1.031948,
        0.418879,0.913918,1.287618,
        0.418879,0.916555,-2.287173,
        0.418879,0.918305,3.107845,
        0.418879,0.919522,2.948277,
        0.418879,0.927138,-2.569686,
        0.418879,0.929031,0.185383,
        0.418879,0.933766,-3.015917,
        0.418879,0.934873,-2.010822,
        0.418879,0.935636,-1.445855,
        0.418879,0.936702,-0.866461,
        0.418879,0.937395,0.729108,
        0.418879,0.937910,2.787575,
        0.418879,0.939144,-0.092962,
        0.418879,0.941718,1.125085,
        0.418879,0.942873,-1.607680,
        0.418879,0.943701,0.980328,
        0.418879,0.947774,2.395256,
        0.418879,0.952433,0.584347,
        0.418879,0.954456,-1.287293,
        0.418879,0.958731,-2.852552,
        0.418879,0.961838,1.839284,
        0.418879,0.961898,-0.469425,
        0.418879,0.965426,1.686615,
        0.418879,0.970108,1.990771,
        0.418879,0.972820,-1.762780,
        0.418879,0.973500,-0.711881,
        0.418879,0.975253,2.641082,
        0.418879,0.977824,0.441928,
        0.418879,0.981485,-2.424766,
        0.418879,0.981877,1.537100,
        0.418879,0.982609,-1.136920,
        0.418879,0.984265,-2.156073,
        0.418879,0.989150,0.045493,
        0.418879,0.993052,2.137478,
        0.418879,0.997766,-2.693871,
        0.418879,0.999847,-0.334738,
        0.418879,1.001920,0.848054,
        0.418879,1.006146,1.389734,
        0.418879,1.012180,0.304066,
        0.418879,1.016938,2.270674,
        0.418879,1.017371,-0.979790,
        0.418879,1.019495,-1.894487,
        0.418879,1.023796,-0.583691,
        0.418879,1.026525,2.506945,
        0.418879,1.027004,-0.203653,
        0.418879,1.028027,3.036573,
        0.418879,1.031435,1.241021,
        0.418879,1.032316,2.884221,
        0.418879,1.035863,-3.110475,
        0.418879,1.042085,-2.292950,
        0.418879,1.043272,-1.522986,
        0.418879,1.046265,-2.958847,
        0.418879,1.049272,-1.376344,
        0.418879,1.050039,-2.552086,
        0.418879,1.053682,0.689863,
        0.418879,1.055458,0.170394,
        0.418879,1.056622,-0.831866,
        0.418879,1.057515,1.097998,
        0.418879,1.058706,-2.036438,
        0.418879,1.061389,-1.672248,
        0.418879,1.063758,2.743093,
        0.418879,1.070157,-0.073422,
        0.418879,1.072154,1.761506,
        0.418879,1.074029,1.909193,
        0.418879,1.074204,0.545276,
        0.418879,1.074714,-2.805147,
        0.418879,1.076262,0.966104,
        0.418879,1.077642,-1.227073,
        0.418879,1.084151,1.618525,
        0.418879,1.085085,2.377673,
        0.418879,1.086822,-0.441112,
        0.418879,1.087142,2.058815,
        0.418879,1.098383,-0.696430,
        0.418879,1.099138,-1.084033,
        0.418879,1.103208,1.477773,
        0.418879,1.103783,0.408254,
        0.418879,1.106275,-2.420800,
        0.418879,1.108079,2.612943,
        0.418879,1.109345,-2.172718,
        0.418879,1.115957,-1.787597,
        0.418879,1.118930,0.045983,
        0.418879,1.119777,0.834702,
        0.418879,1.121943,-2.669042,
        0.418879,1.124930,2.219917,
        0.418879,1.125217,1.336294,
        0.418879,1.130958,2.972931,
        0.418879,1.131400,-0.306629,
        0.418879,1.137135,-0.936514,
        0.418879,1.139125,-1.924528,
        0.418879,1.139877,0.276382,
        0.418879,1.142003,-3.058022,
        0.418879,1.147363,-0.558448,
        0.418879,1.150348,-0.176418,
        0.418879,1.151198,1.197016,
        0.418879,1.153930,-1.597104,
        0.418879,1.155389,2.837148,
        0.418879,1.155458,-1.460623,
        0.418879,1.157331,-2.906362,
        0.418879,1.158594,2.487618,
        0.418879,1.162103,-1.322553,
        0.418879,1.165407,3.092861,
        0.418879,1.168025,0.717075,
        0.418879,1.168406,-2.301519,
        0.418879,1.174782,-2.538755,
        0.418879,1.177998,1.061256,
        0.418879,1.178898,1.827305,
        0.418879,1.181331,-2.060330,
        0.418879,1.182540,-0.798952,
        0.418879,1.184569,0.147517,
        0.418879,1.184734,1.968919,
        0.418879,1.184804,0.593386,
        0.418879,1.188278,1.690808,
        0.418879,1.194360,2.110440,
        0.418879,1.194776,2.710455,
        0.418879,1.197745,-1.173705,
        0.418879,1.199898,-2.777257,
        0.418879,1.201164,2.355931,
        0.418879,1.203760,1.556404,
        0.418879,1.205845,-0.415993,
        0.418879,1.206585,0.933648,
        0.418879,1.206950,-0.037248,
        0.418879,1.209341,0.471162,
        0.418879,1.213483,-1.039417,
        0.418879,1.218146,-0.672643,
        0.418879,1.221608,1.421669,
        0.418879,1.226043,-1.693533,
        0.418879,1.232870,-1.822773,
        0.418879,1.234375,-2.421194,
        0.418879,1.234531,-2.190148,
        0.418879,1.239414,0.348959,
        0.418879,1.240183,2.587255,
        0.418879,1.242312,2.941234,
        0.418879,1.244075,1.286644,
        0.418879,1.245961,-2.997461,
        0.418879,1.246495,-2.646872,
        0.418879,1.247736,2.242810,
        0.418879,1.253604,0.812222,
        0.418879,1.257816,-0.894317,
        0.418879,1.258247,-1.953634,
        0.418879,1.258993,-0.284099,
        0.418879,1.262380,-0.157609,
        0.418879,1.262791,-1.549229,
        0.418879,1.263772,-0.533073,
        0.418879,1.266731,-1.412461,
        0.418879,1.268591,-3.122146,
        0.418879,1.269614,-1.275246,
        0.418879,1.270745,1.152533,
        0.418879,1.273606,0.224545,
        0.418879,1.279716,2.812101,
        0.418879,1.283786,2.462124,
        0.418879,1.284798,0.078209,
        0.418879,1.285402,-2.882049,
        0.418879,1.286936,1.886053,
        0.418879,1.289408,0.683122,
        0.418879,1.294736,1.756184,
        0.418879,1.296978,-2.311511,
        0.418879,1.298546,2.020601,
        0.418879,1.300924,1.022478,
        0.418879,1.303305,3.043152,
        0.418879,1.304210,-2.084225,
        0.418879,1.305241,-2.530249,
        0.418879,1.307513,1.628003,
        0.418879,1.310722,0.558267,
        0.418879,1.314142,2.139329,
        0.418879,1.316087,-0.777971,
        0.418879,1.317907,-1.121023,
        0.418879,1.321019,-2.749776,
        0.418879,1.321609,2.687717,
        0.418879,1.322568,1.499415,
        0.418879,1.322972,-0.395699,
        0.418879,1.325286,-0.994192,
        0.418879,1.329146,-0.648513,
        0.418879,1.329795,-0.042570,
        0.418879,1.331651,2.344497,
        0.418879,1.336522,0.435098,
        0.418879,1.339410,0.900087,
        0.418879,1.339550,1.370353,
        0.418879,1.340441,-1.731580,
        0.418879,1.351855,-1.855234,
        0.418879,1.360161,-3.018404,
        0.418879,1.361031,-1.613971,
        0.418879,1.361196,2.914780,
        0.418879,1.361998,-2.207254,
        0.418879,1.362302,1.239384,
        0.418879,1.365228,2.563808,
        0.418879,1.365452,0.310522,
        0.418879,1.367028,-2.420041,
        0.418879,1.373422,-1.484281,
        0.418879,1.374245,-1.226417,
        0.418879,1.375020,-0.170393,
        0.418879,1.376560,-1.355481,
        0.418879,1.377070,-1.980984,
        0.418879,1.377578,-2.636339,
        0.418879,1.378096,0.774626,
        0.418879,1.379086,-0.508192,
        0.418879,1.382562,0.177141,
        0.418879,1.388849,2.233847,
        0.418879,1.389268,-0.886349,
        0.418879,1.391324,1.109396,
        0.418879,1.391699,-0.288725,
        0.418879,1.392869,3.136903,
        0.418879,1.393612,1.952016,
        0.418879,1.395545,-2.850807,
        0.418879,1.401926,1.822930,
        0.418879,1.402149,2.788898,
        0.418879,1.409552,0.647980,
        0.418879,1.410840,0.056219,
        0.418879,1.411944,2.444377,
        0.418879,1.413639,1.697982,
        0.418879,1.426479,1.573427,
        0.418879,1.426725,2.102242,
        0.418879,1.427868,0.984637,
        0.418879,1.429580,-2.107232,
        0.418879,1.431375,-2.313269,
        0.418879,1.431907,3.015410,
        0.418879,1.432325,-1.029054,
        0.418879,1.433341,0.524268,
        0.418879,1.435181,-0.739771,
        0.418879,1.437208,-2.529085,
        0.418879,1.439041,-0.617966,
        0.418879,1.441473,1.448544,
        0.418879,1.444527,2.664744,
        0.418879,1.450513,-0.400625,
        0.418879,1.451907,-0.066823,
        0.418879,1.452754,-2.738631,
        0.418879,1.454061,-2.950416,
        0.418879,1.456736,1.323154,
        0.418879,1.458370,0.400433,
        0.418879,1.459314,-1.765003,
        0.418879,1.459902,-1.145835,
        0.418879,1.464952,2.329245,
        0.418879,1.466050,0.862104,
        0.418879,1.471330,-1.886160,
        0.418879,1.475119,-1.648255,
        0.418879,1.480878,1.194576,
        0.418879,1.480942,0.273235,
        0.418879,1.481086,2.890269,
        0.418879,1.484118,-1.529156,
        0.418879,1.484741,-1.279219,
        0.418879,1.487681,-1.406069,
        0.418879,1.487886,-3.066037,
        0.418879,1.490555,2.543284,
        0.418879,1.492424,-0.190932,
        0.418879,1.495002,-0.839810,
        0.418879,1.495297,2.007188,
        0.418879,1.496414,-2.009267,
        0.418879,1.498261,-2.425270,
        0.418879,1.498907,-2.207327,
        0.418879,1.500204,0.737638,
        0.418879,1.502469,0.151156,
        0.418879,1.507448,-0.514782,
        0.418879,1.509287,2.204647,
        0.418879,1.510518,1.884920,
        0.418879,1.511948,-2.632360,
        0.418879,1.515235,1.068565,
        0.418879,1.522237,3.102018,
        0.418879,1.522373,1.764991,
        0.418879,1.522416,-0.954956,
        0.418879,1.523102,2.765920,
        0.418879,1.524811,-2.844718,
        0.418879,1.527261,-0.304916,
        0.418879,1.529788,0.613208,
        0.418879,1.533648,1.644808,
        0.418879,1.535018,0.028869,
        0.418879,1.541082,2.424608,
        0.418879,1.546010,1.523462,
        0.418879,1.551135,-0.642994,
        0.418879,1.553895,0.489607,
        0.418879,1.555041,0.947813,
        0.418879,1.559671,-1.074934,
        0.418879,1.560345,2.985899,
        0.418879,1.560487,-2.323400,
        0.418879,1.561127,1.401111,
        0.418879,1.567193,2.643172,
        0.418879,1.568892,2.098362,
        0.418879,1.571702,-2.098435,
        0.418879,1.572701,-2.528730,
        0.418879,1.573572,1.278593,
        0.418879,1.574849,-0.091505,
        0.418879,1.576334,0.365236,
        0.418879,1.577029,-2.962888,
        0.418879,1.577688,-1.814941,
        0.418879,1.579542,-0.755917,
        0.418879,1.579624,-1.199759,
        0.418879,1.581515,-0.419414,
        0.418879,1.584581,-2.738702,
        0.418879,1.586948,-1.696296,
        0.418879,1.588901,2.301496,
        0.418879,1.589936,0.825020,
        0.418879,1.590018,-1.941652,
        0.418879,1.596968,-1.328358,
        0.418879,1.597178,-1.575607,
        0.418879,1.600470,-1.453196,
        0.418879,1.601408,0.242409,
        0.418879,1.602259,1.152656,
        0.418879,1.603064,2.865774,
        0.418879,1.612085,1.968802,
        0.418879,1.612720,-3.080498,
        0.418879,1.616061,2.521531,
        0.418879,1.616900,-0.207457,
        0.418879,1.620988,-0.875920,
        0.418879,1.622114,0.701509,
        0.418879,1.622749,0.121426,
        0.418879,1.622993,-2.221629,
        0.418879,1.629346,-0.542669,
        0.418879,1.630007,1.844552,
        0.418879,1.634747,-2.426198,
        0.418879,1.642346,1.722617,
        0.418879,1.643925,1.034629,
        0.418879,1.644848,2.743774,
        0.418879,1.645258,2.188109,
        0.418879,1.645417,-2.634725,
        0.418879,1.647395,3.085437,
        0.418879,1.648538,-2.855044,
        0.418879,1.649932,0.576808,
        0.418879,1.652224,-0.996105,
        0.418879,1.652871,1.600681,
        0.418879,1.657997,-0.325127,
        0.418879,1.661936,0.001037,
        0.418879,1.665187,1.477402,
        0.418879,1.665496,2.397872,
        0.418879,1.673267,-0.665453,
        0.418879,1.673557,0.452526,
        0.418879,1.679277,-1.122421,
        0.418879,1.679696,0.912482,
        0.418879,1.682420,1.353922,
        0.418879,1.685031,-2.009038,
        0.418879,1.685196,2.967046,
        0.418879,1.690107,2.620910,
        0.418879,1.690453,-2.123076,
        0.418879,1.691693,-1.871572,
        0.418879,1.691817,2.061688,
        0.418879,1.692644,1.237670,
        0.418879,1.696386,0.328123,
        0.418879,1.697162,-1.249026,
        0.418879,1.697488,-2.322785,
        0.418879,1.698726,-2.973576,
        0.418879,1.699172,-1.749717,
        0.418879,1.704478,-0.106295,
        0.418879,1.707244,-0.444770,
        0.418879,1.707500,-2.531497,
        0.418879,1.707714,-0.787321,
        0.418879,1.708430,-1.627552,
        0.418879,1.710583,-1.376890,
        0.418879,1.712415,-2.747921,
        0.418879,1.713125,-1.503073,
        0.418879,1.713452,0.789419,
        0.418879,1.720612,2.279888,
        0.418879,1.724852,0.202831,
        0.418879,1.724868,2.846246,
        0.418879,1.728940,1.931535,
        0.418879,1.732838,1.127584,
        0.418879,1.735719,-3.092485,
        0.418879,1.739676,2.496017,
        0.418879,1.739971,-0.228876,
        0.418879,1.743328,0.663347,
        0.418879,1.744927,-0.911892,
        0.418879,1.745952,0.084733,
        0.418879,1.748077,1.804354,
        0.418879,1.752959,-0.567548,
        0.418879,1.760169,1.678132,
        0.418879,1.763887,-2.218450,
        0.418879,1.766737,2.722752,
        0.418879,1.767999,1.004737,
        0.418879,1.769149,1.550298,
        0.418879,1.770119,-2.427175,
        0.418879,1.770460,2.154208,
        0.418879,1.770514,3.070982,
        0.418879,1.770735,0.537245,
        0.418879,1.772673,-2.864760,
        0.418879,1.773527,-1.039582,
        0.418879,1.776419,-2.642421,
        0.418879,1.780769,1.416330,
        0.418879,1.786634,-0.346247,
        0.418879,1.788829,-1.947553,
        0.418879,1.793737,0.410270,
        0.418879,1.794553,2.374618,
        0.418879,1.794699,-0.694061,
        0.418879,1.797469,-1.169105,
        0.418879,1.803705,0.881076,
        0.418879,1.804124,1.264249,
        0.418879,1.807825,-2.090455,
        0.418879,1.808822,-1.817375,
        0.418879,1.809020,2.949741,
        0.418879,1.812466,2.597595,
        0.418879,1.813159,2.025395,
        0.418879,1.813816,-1.298435,
        0.418879,1.815016,0.278951,
        0.418879,1.816510,-0.008631,
        0.418879,1.817772,-1.689625,
        0.418879,1.821360,-2.984352,
        0.418879,1.822312,-0.127211,
        0.418879,1.823386,-1.429807,
        0.418879,1.823834,-1.560483,
        0.418879,1.832060,-0.821627,
        0.418879,1.832690,-0.468440,
        0.418879,1.833945,-2.320278,
        0.418879,1.834787,0.751968,
        0.418879,1.840579,-2.536445,
        0.418879,1.841033,-2.756932,
        0.418879,1.846350,1.893983,
        0.418879,1.846644,1.113766,
        0.418879,1.846833,2.826543,
        0.418879,1.847325,2.249051,
        0.418879,1.854119,0.134161,
        0.418879,1.858863,-3.106022,
        0.418879,1.865046,1.763384,
        0.418879,1.865619,-0.952036,
        0.418879,1.865912,0.624453,
        0.418879,1.866694,2.473686,
        0.418879,1.867524,-0.240083,
        0.418879,1.875561,1.632831,
        0.418879,1.876049,-0.594378,
        0.418879,1.880509,1.500204,
        0.418879,1.883504,1.360457,
        0.418879,1.886971,-2.196665,
        0.418879,1.888979,0.982922,
        0.418879,1.889815,2.701570,
        0.418879,1.892144,-2.005340,
        0.418879,1.892348,0.496522,
        0.418879,1.893281,-1.083976,
        0.418879,1.893828,3.053453,
        0.418879,1.895117,2.121107,
        0.418879,1.898741,-2.875735,
        0.418879,1.904429,-2.428029,
        0.418879,1.908408,-2.648903,
        0.418879,1.911828,-0.364489,
        0.418879,1.912224,0.366565,
        0.418879,1.914854,1.214468,
        0.418879,1.915467,-1.883143,
        0.418879,1.915763,-0.724679,
        0.418879,1.915790,-1.216366,
        0.418879,1.922256,2.347806,
        0.418879,1.922616,0.844633,
        0.418879,1.925185,0.231170,
        0.418879,1.927474,-1.754616,
        0.418879,1.928342,-1.351174,
        0.418879,1.933328,-1.623328,
        0.418879,1.933878,0.034150,
        0.418879,1.934585,-1.487979,
        0.418879,1.934802,2.927634,
        0.418879,1.934862,1.989776,
        0.418879,1.936327,-0.099350,
        0.418879,1.938088,2.578507,
        0.418879,1.945677,-2.999805,
        0.418879,1.954572,-0.859208,
        0.418879,1.957868,-0.490784,
        0.418879,1.958989,0.712772,
        0.418879,1.960212,-2.307561,
        0.418879,1.964774,1.855507,
        0.418879,1.964971,-2.100533,
        0.418879,1.969112,1.088273,
        0.418879,1.971438,-2.765415,
        0.418879,1.972312,2.805134,
        0.418879,1.974950,2.220596,
        0.418879,1.975135,-2.539127,
        0.418879,1.982107,1.720829,
        0.418879,1.983195,-3.127984,
        0.418879,1.983318,-0.247182,
        0.418879,1.986720,-0.993220,
        0.418879,1.990381,1.586005,
        0.418879,1.990470,0.580283,
        0.418879,1.990523,1.450326,
        0.418879,1.993909,1.312897,
        0.418879,1.994694,2.452813,
        0.418879,1.999226,-0.620717,
        0.418879,2.005145,0.941012,
        0.418879,2.007301,0.142337,
        0.418879,2.014149,0.444138,
        0.418879,2.014338,-1.127392,
        0.418879,2.018131,3.021778,
        0.418879,2.019676,-1.980944,
        0.418879,2.019912,2.686612,
        0.418879,2.020545,2.088672,
        0.418879,2.026074,0.300945,
        0.418879,2.026196,-2.889057,
        0.418879,2.031959,-1.268178,
        0.418879,2.032917,-2.199917,
        0.418879,2.033572,-1.842884,
        0.418879,2.034712,-2.419505,
        0.418879,2.035818,-0.379479,
        0.418879,2.036004,-0.757062,
        0.418879,2.040501,-1.411488,
        0.418879,2.042874,-1.701367,
        0.418879,2.043060,-2.653746,
        0.418879,2.043729,-1.557938,
        0.418879,2.048538,0.806623,
        0.418879,2.049006,1.199930,
        0.418879,2.049502,-0.137989,
        0.418879,2.052181,2.326959,
        0.418879,2.052794,-0.008526,
        0.418879,2.057608,1.951726,
        0.418879,2.065568,2.567730,
        0.418879,2.070371,2.879808,
        0.418879,2.071256,-3.018961,
        0.418879,2.079746,-0.897699,
        0.418879,2.082784,1.042403,
        0.418879,2.083752,-0.511086,
        0.418879,2.084401,1.810502,
        0.418879,2.085731,0.671081,
        0.418879,2.097496,1.373172,
        0.418879,2.099185,1.667006,
        0.418879,2.100482,-2.301084,
        0.418879,2.103118,2.194745,
        0.418879,2.103223,1.521685,
        0.418879,2.103695,-2.777707,
        0.418879,2.107523,3.129581,
        0.418879,2.108277,-2.095907,
        0.418879,2.108891,-2.532166,
        0.418879,2.111003,-0.268632,
        0.418879,2.112276,-1.027246,
        0.418879,2.114380,0.531457,
        0.418879,2.119166,0.100552,
        0.418879,2.119564,2.754818,
        0.418879,2.122474,-0.648314,
        0.418879,2.125183,2.445598,
        0.418879,2.129501,0.238290,
        0.418879,2.131178,0.386663,
        0.418879,2.131223,-1.174311,
        0.418879,2.134269,-1.948042,
        0.418879,2.134684,0.907065,
        0.418879,2.137615,2.988806,
        0.418879,2.145635,-1.327708,
        0.418879,2.146467,2.055081,
        0.418879,2.146582,-1.793788,
        0.418879,2.149210,-1.482321,
        0.418879,2.152901,-1.636788,
        0.418879,2.156337,1.151558,
        0.418879,2.156474,-2.912065,
        0.418879,2.157194,-0.792563,
        0.418879,2.166724,-0.401638,
        0.418879,2.166980,-0.134879,
        0.418879,2.173644,-2.403429,
        0.418879,2.175711,-2.655523,
        0.418879,2.177132,2.600419,
        0.418879,2.178879,0.767722,
        0.418879,2.180752,1.908485,
        0.418879,2.181338,2.313721,
        0.418879,2.185580,1.287673,
        0.418879,2.195400,-2.217546,
        0.418879,2.200418,-3.056442,
        0.418879,2.201829,-0.001735,
        0.418879,2.203804,1.756176,
        0.418879,2.211602,1.443930,
        0.418879,2.212686,-0.544809,
        0.418879,2.213937,0.621892,
        0.418879,2.214075,1.600691,
        0.418879,2.216256,1.020029,
        0.418879,2.219060,2.886631,
        0.418879,2.223375,-0.906185,
        0.418879,2.226002,-1.060248,
        0.418879,2.226413,-2.064034,
        0.418879,2.230554,2.171334,
        0.418879,2.232413,2.735777,
        0.418879,2.233709,3.070830,
        0.418879,2.234383,-2.792065,
        0.418879,2.237183,0.468799,
        0.418879,2.237695,-0.271388,
        0.418879,2.239419,0.151693,
        0.418879,2.243347,-2.521049,
        0.418879,2.244035,-1.229456,
        0.418879,2.244355,0.309899,
        0.418879,2.245175,2.460236,
        0.418879,2.245628,-1.899216,
        0.418879,2.245973,-0.702234,
        0.418879,2.255601,-1.728131,
        0.418879,2.256702,-1.401367,
        0.418879,2.261466,-1.557500,
        0.418879,2.267390,0.878499,
        0.418879,2.272099,2.018594,
        0.418879,2.278618,1.170386,
        0.418879,2.283304,-2.345584,
        0.418879,2.285175,-2.943090,
        0.418879,2.293000,-0.124451,
        0.418879,2.294058,-0.421070,
        0.418879,2.303283,1.855938,
        0.418879,2.304968,2.608978,
        0.418879,2.305347,2.311346,
        0.418879,2.307188,-2.656134,
        0.418879,2.307301,1.336200,
        0.418879,2.309655,0.726020,
        0.418879,2.317634,-2.185225,
        0.418879,2.321510,1.685664,
        0.418879,2.324376,-3.111006,
        0.418879,2.324731,1.511186,
        0.418879,2.329587,0.035498,
        0.418879,2.330158,2.977264,
        0.418879,2.335096,-0.588468,
        0.418879,2.336046,-1.115799,
        0.418879,2.338060,2.803613,
        0.418879,2.338083,-0.943400,
        0.418879,2.340210,0.562145,
        0.418879,2.341931,1.023926,
        0.418879,2.343265,-0.779061,
        0.418879,2.343434,-2.013692,
        0.418879,2.353029,-1.310529,
        0.418879,2.354868,0.211543,
        0.418879,2.355120,0.388593,
        0.418879,2.355165,-1.833173,
        0.418879,2.355462,2.151441,
        0.418879,2.362948,-2.487477,
        0.418879,2.363101,-2.810456,
        0.418879,2.363394,-0.272918,
        0.418879,2.364509,-1.644533,
        0.418879,2.370465,2.466824,
        0.418879,2.386110,-1.481351,
        0.418879,2.394960,0.862133,
        0.418879,2.395716,1.977184,
        0.418879,2.396339,1.200900,
        0.418879,2.407941,-2.316753,
        0.418879,2.409450,-2.988432,
        0.418879,2.417202,-0.103363,
        0.418879,2.417386,-0.443097,
        0.418879,2.417950,2.666662,
        0.418879,2.423190,1.789783,
        0.418879,2.423692,1.392696,
        0.418879,2.431122,3.086379,
        0.418879,2.431406,-2.647820,
        0.418879,2.431584,2.307631,
        0.418879,2.434090,1.592818,
        0.418879,2.435993,0.683844,
        0.418879,2.440898,-2.133300,
        0.418879,2.442222,-1.195837,
        0.418879,2.443112,2.883761,
        0.418879,2.449191,-0.633651,
        0.418879,2.450821,0.084741,
        0.418879,2.453499,-1.008861,
        0.418879,2.460010,-1.938011,
        0.418879,2.460199,-0.822008,
        0.418879,2.460902,0.490079,
        0.418879,2.461774,-1.734913,
        0.418879,2.463986,1.034407,
        0.418879,2.467647,0.287667,
        0.418879,2.479849,2.126925,
        0.418879,2.485907,-2.835471,
        0.418879,2.485945,-1.374853,
        0.418879,2.487747,2.505912,
        0.418879,2.487957,-0.274279,
        0.418879,2.489741,-2.468428,
        0.418879,2.512538,1.239900,
        0.418879,2.517528,1.923289,
        0.418879,2.518454,-1.565852,
        0.418879,2.519348,0.843503,
        0.418879,2.522766,-3.059231,
        0.418879,2.528686,2.737228,
        0.418879,2.534798,-2.266671,
        0.418879,2.536533,1.466274,
        0.418879,2.538504,1.699812,
        0.418879,2.539516,-0.476942,
        0.418879,2.539532,-0.074548,
        0.418879,2.543197,2.990252,
        0.418879,2.554022,2.320619,
        0.418879,2.556919,-2.655005,
        0.418879,2.558146,0.627191,
        0.418879,2.558223,-1.164189,
        0.418879,2.563452,-2.044028,
        0.418879,2.568462,0.150655,
        0.418879,2.569130,-0.700837,
        0.418879,2.570050,-1.804492,
        0.418879,2.572710,-0.934665,
        0.418879,2.575994,0.391718,
        0.418879,2.584774,1.046703,
        0.418879,2.601432,2.102535,
        0.418879,2.603559,2.555114,
        0.418879,2.604150,-2.885208,
        0.418879,2.608630,-1.378011,
        0.418879,2.611256,-0.279110,
        0.418879,2.617344,-2.439480,
        0.418879,2.627402,1.293552,
        0.418879,2.635396,1.848850,
        0.418879,2.635686,2.834950,
        0.418879,2.637087,3.132730,
        0.418879,2.642029,0.813274,
        0.418879,2.645260,1.570203,
        0.418879,2.646182,-1.614036,
        0.418879,2.660113,-2.186513,
        0.418879,2.660483,-0.034608,
        0.418879,2.661713,-0.529511,
        0.418879,2.672720,-1.110422,
        0.418879,2.676626,2.335353,
        0.418879,2.676638,0.540364,
        0.418879,2.676780,-1.901158,
        0.418879,2.681411,-2.667965,
        0.418879,2.682152,0.245751,
        0.418879,2.682859,-0.817071,
        0.418879,2.705107,1.062142,
        0.418879,2.717705,2.625995,
        0.418879,2.721001,2.060097,
        0.418879,2.722319,-2.954069,
        0.418879,2.729840,-1.368237,
        0.418879,2.734349,-0.289298,
        0.418879,2.738409,2.975671,
        0.418879,2.740366,1.374341,
        0.418879,2.744633,-2.391311,
        0.418879,2.747398,1.727166,
        0.418879,2.763799,0.760372,
        0.418879,2.768238,-1.679243,
        0.418879,2.779245,-2.046268,
        0.418879,2.780208,-0.626453,
        0.418879,2.780471,0.027732,
        0.418879,2.784696,-1.013128,
        0.418879,2.789817,0.395431,
        0.418879,2.801154,2.353143,
        0.418879,2.807091,-2.687477,
        0.418879,2.825762,1.085035,
        0.418879,2.830209,2.734943,
        0.418879,2.833517,-3.085508,
        0.418879,2.838301,1.980061,
        0.418879,2.849871,1.512012,
        0.418879,2.850089,-1.343899,
        0.418879,2.856905,-0.318441,
        0.418879,2.870445,-2.291001,
        0.418879,2.885017,0.655562,
        0.418879,2.886428,-1.793217,
        0.418879,2.890337,-0.825633,
        0.418879,2.898932,0.139973,
        0.418879,2.931927,-2.728768,
        0.418879,2.932096,2.377790,
        0.418879,2.937409,2.947245,
        0.418879,2.948048,1.133127,
        0.418879,2.950896,1.797889,
        0.418879,2.970019,-1.282382,
        0.418879,2.978333,-0.412090,
        0.418879,2.996437,-2.089121,
        0.418879,3.011274,0.434110,
        0.418879,3.050410,3.138492,
        0.418879,3.064492,1.647259,
        0.418879,3.088868,-0.989501,
        0.523599,0.053884,-2.408394,
        0.523599,0.077645,1.453433,
        0.523599,0.088408,-0.311160,
        0.523599,0.142314,2.590387,
        0.523599,0.147206,-1.385164,
        0.523599,0.163506,0.573203,
        0.523599,0.165485,-2.875396,
        0.523599,0.182564,-2.125821,
        0.523599,0.190569,1.929141,
        0.523599,0.201102,1.249082,
        0.523599,0.201397,-0.716646,
        0.523599,0.210568,-0.069559,
        0.523599,0.251672,2.848900,
        0.523599,0.266262,-1.678068,
        0.523599,0.269108,-1.215563,
        0.523599,0.273777,-2.506756,
        0.523599,0.279854,0.814581,
        0.523599,0.281664,-2.969147,
        0.523599,0.283625,2.405603,
        0.523599,0.284058,0.357350,
        0.523599,0.294107,1.593880,
        0.523599,0.302491,-0.400410,
        0.523599,0.308059,2.021491,
        0.523599,0.312219,-2.072057,
        0.523599,0.321905,-0.829121,
        0.523599,0.329984,1.201621,
        0.523599,0.339714,-0.013712,
        0.523599,0.368784,2.974974,
        0.523599,0.376428,-1.495532,
        0.523599,0.378103,2.646436,
        0.523599,0.385088,0.590355,
        0.523599,0.386248,-2.690943,
        0.523599,0.389851,-2.340031,
        0.523599,0.391502,-1.171374,
        0.523599,0.398249,-1.812694,
        0.523599,0.400542,-2.995729,
        0.523599,0.404273,0.918163,
        0.523599,0.405617,1.773533,
        0.523599,0.409009,-0.570243,
        0.523599,0.410700,0.270433,
        0.523599,0.415893,1.464890,
        0.523599,0.420680,2.088164,
        0.523599,0.424766,-0.258729,
        0.523599,0.426507,2.367872,
        0.523599,0.441292,-0.890727,
        0.523599,0.451839,-2.078193,
        0.523599,0.459270,1.191562,
        0.523599,0.470622,0.014002,
        0.523599,0.481292,-2.529098,
        0.523599,0.482460,2.849843,
        0.523599,0.492205,-1.397963,
        0.523599,0.493314,3.117264,
        0.523599,0.497870,0.727828,
        0.523599,0.499783,-1.654379,
        0.523599,0.502651,0.463333,
        0.523599,0.505463,2.594557,
        0.523599,0.506104,-2.874910,
        0.523599,0.513579,-1.147354,
        0.523599,0.518569,1.644237,
        0.523599,0.519001,1.900488,
        0.523599,0.519277,-0.673809,
        0.523599,0.524769,-0.421955,
        0.523599,0.525564,-2.288838,
        0.523599,0.532541,0.978737,
        0.523599,0.534610,2.154808,
        0.523599,0.535020,-1.891846,
        0.523599,0.540208,1.400497,
        0.523599,0.542006,0.228337,
        0.523599,0.550762,-0.180771,
        0.523599,0.560531,2.378474,
        0.523599,0.561629,-0.924576,
        0.523599,0.567465,-2.679392,
        0.523599,0.588527,1.181755,
        0.523599,0.591643,-2.098364,
        0.523599,0.594963,2.997575,
        0.523599,0.599643,-3.045235,
        0.523599,0.599962,2.773556,
        0.523599,0.602620,0.029270,
        0.523599,0.605711,0.598454,
        0.523599,0.607708,-1.541026,
        0.523599,0.609360,-2.464379,
        0.523599,0.610738,-1.326786,
        0.523599,0.616968,0.820489,
        0.523599,0.626379,1.773610,
        0.523599,0.628626,-1.751180,
        0.523599,0.630671,0.391949,
        0.523599,0.632622,-0.737642,
        0.523599,0.632708,2.566722,
        0.523599,0.632797,1.989858,
        0.523599,0.634455,-0.534790,
        0.523599,0.636205,1.561170,
        0.523599,0.637681,-1.123423,
        0.523599,0.641441,-2.842109,
        0.523599,0.644807,-0.327139,
        0.523599,0.652440,2.197129,
        0.523599,0.660897,-2.278593,
        0.523599,0.664008,1.360091,
        0.523599,0.670850,-1.945257,
        0.523599,0.671396,1.004859,
        0.523599,0.673985,0.203855,
        0.523599,0.680326,-0.133370,
        0.523599,0.685083,-0.942363,
        0.523599,0.688204,-2.632204,
        0.523599,0.689777,2.385011,
        0.523599,0.700065,3.115430,
        0.523599,0.704080,2.911636,
        0.523599,0.712883,1.174216,
        0.523599,0.715701,0.699906,
        0.523599,0.718851,-1.442215,
        0.523599,0.721019,2.721536,
        0.523599,0.725300,-2.975134,
        0.523599,0.727136,-1.633288,
        0.523599,0.727399,-2.120567,
        0.523599,0.730751,0.522488,
        0.523599,0.732892,-1.254330,
        0.523599,0.734333,-2.446184,
        0.523599,0.735125,0.037270,
        0.523599,0.737893,1.866463,
        0.523599,0.740650,1.681075,
        0.523599,0.745971,-0.781286,
        0.523599,0.749432,-0.442559,
        0.523599,0.750102,2.050636,
        0.523599,0.751046,0.863821,
        0.523599,0.757018,-0.608833,
        0.523599,0.757046,-1.816019,
        0.523599,0.757389,1.501184,
        0.523599,0.759800,2.546181,
        0.523599,0.759859,0.351923,
        0.523599,0.761141,-2.781102,
        0.523599,0.769035,-1.078574,
        0.523599,0.770166,-0.266604,
        0.523599,0.772218,2.227380,
        0.523599,0.788382,1.331051,
        0.523599,0.790269,-2.282409,
        0.523599,0.803410,-1.984567,
        0.523599,0.803813,0.191520,
        0.523599,0.806832,3.022155,
        0.523599,0.807476,-2.597675,
        0.523599,0.812512,-0.107335,
        0.523599,0.814161,1.010949,
        0.523599,0.817732,-3.084486,
        0.523599,0.817846,2.390688,
        0.523599,0.819323,2.845960,
        0.523599,0.819430,-0.915017,
        0.523599,0.829198,-1.532491,
        0.523599,0.829531,0.665695,
        0.523599,0.829816,1.167147,
        0.523599,0.833613,-1.357492,
        0.523599,0.842945,-2.907318,
        0.523599,0.844883,2.679708,
        0.523599,0.848189,-1.705003,
        0.523599,0.849888,1.767901,
        0.523599,0.851126,0.493370,
        0.523599,0.852911,1.933394,
        0.523599,0.857516,-2.432863,
        0.523599,0.858261,-1.191729,
        0.523599,0.858279,-2.138895,
        0.523599,0.859723,-0.734928,
        0.523599,0.859754,1.603292,
        0.523599,0.861655,-0.394821,
        0.523599,0.864300,0.043549,
        0.523599,0.866940,0.851229,
        0.523599,0.869748,2.095969,
        0.523599,0.877352,-0.574668,
        0.523599,0.879254,-2.733338,
        0.523599,0.882554,1.443757,
        0.523599,0.884060,0.334199,
        0.523599,0.885626,-1.866061,
        0.523599,0.889834,2.529957,
        0.523599,0.893140,2.251698,
        0.523599,0.897463,-0.240822,
        0.523599,0.898758,-1.031948,
        0.523599,0.913918,1.287618,
        0.523599,0.916555,-2.287173,
        0.523599,0.918305,3.107845,
        0.523599,0.919522,2.948277,
        0.523599,0.927138,-2.569686,
        0.523599,0.929031,0.185383,
        0.523599,0.933766,-3.015917,
        0.523599,0.934873,-2.010822,
        0.523599,0.935636,-1.445855,
        0.523599,0.936702,-0.866461,
        0.523599,0.937395,0.729108,
        0.523599,0.937910,2.787575,
        0.523599,0.939144,-0.092962,
        0.523599,0.941718,1.125085,
        0.523599,0.942873,-1.607680,
        0.523599,0.943701,0.980328,
        0.523599,0.947774,2.395256,
        0.523599,0.952433,0.584347,
        0.523599,0.954456,-1.287293,
        0.523599,0.958731,-2.852552,
        0.523599,0.961838,1.839284,
        0.523599,0.961898,-0.469425,
        0.523599,0.965426,1.686615,
        0.523599,0.970108,1.990771,
        0.523599,0.972820,-1.762780,
        0.523599,0.973500,-0.711881,
        0.523599,0.975253,2.641082,
        0.523599,0.977824,0.441928,
        0.523599,0.981485,-2.424766,
        0.523599,0.981877,1.537100,
        0.523599,0.982609,-1.136920,
        0.523599,0.984265,-2.156073,
        0.523599,0.989150,0.045493,
        0.523599,0.993052,2.137478,
        0.523599,0.997766,-2.693871,
        0.523599,0.999847,-0.334738,
        0.523599,1.001920,0.848054,
        0.523599,1.006146,1.389734,
        0.523599,1.012180,0.304066,
        0.523599,1.016938,2.270674,
        0.523599,1.017371,-0.979790,
        0.523599,1.019495,-1.894487,
        0.523599,1.023796,-0.583691,
        0.523599,1.026525,2.506945,
        0.523599,1.027004,-0.203653,
        0.523599,1.028027,3.036573,
        0.523599,1.031435,1.241021,
        0.523599,1.032316,2.884221,
        0.523599,1.035863,-3.110475,
        0.523599,1.042085,-2.292950,
        0.523599,1.043272,-1.522986,
        0.523599,1.046265,-2.958847,
        0.523599,1.049272,-1.376344,
        0.523599,1.050039,-2.552086,
        0.523599,1.053682,0.689863,
        0.523599,1.055458,0.170394,
        0.523599,1.056622,-0.831866,
        0.523599,1.057515,1.097998,
        0.523599,1.058706,-2.036438,
        0.523599,1.061389,-1.672248,
        0.523599,1.063758,2.743093,
        0.523599,1.070157,-0.073422,
        0.523599,1.072154,1.761506,
        0.523599,1.074029,1.909193,
        0.523599,1.074204,0.545276,
        0.523599,1.074714,-2.805147,
        0.523599,1.076262,0.966104,
        0.523599,1.077642,-1.227073,
        0.523599,1.084151,1.618525,
        0.523599,1.085085,2.377673,
        0.523599,1.086822,-0.441112,
        0.523599,1.087142,2.058815,
        0.523599,1.098383,-0.696430,
        0.523599,1.099138,-1.084033,
        0.523599,1.103208,1.477773,
        0.523599,1.103783,0.408254,
        0.523599,1.106275,-2.420800,
        0.523599,1.108079,2.612943,
        0.523599,1.109345,-2.172718,
        0.523599,1.115957,-1.787597,
        0.523599,1.118930,0.045983,
        0.523599,1.119777,0.834702,
        0.523599,1.121943,-2.669042,
        0.523599,1.124930,2.219917,
        0.523599,1.125217,1.336294,
        0.523599,1.130958,2.972931,
        0.523599,1.131400,-0.306629,
        0.523599,1.137135,-0.936514,
        0.523599,1.139125,-1.924528,
        0.523599,1.139877,0.276382,
        0.523599,1.142003,-3.058022,
        0.523599,1.147363,-0.558448,
        0.523599,1.150348,-0.176418,
        0.523599,1.151198,1.197016,
        0.523599,1.153930,-1.597104,
        0.523599,1.155389,2.837148,
        0.523599,1.155458,-1.460623,
        0.523599,1.157331,-2.906362,
        0.523599,1.158594,2.487618,
        0.523599,1.162103,-1.322553,
        0.523599,1.165407,3.092861,
        0.523599,1.168025,0.717075,
        0.523599,1.168406,-2.301519,
        0.523599,1.174782,-2.538755,
        0.523599,1.177998,1.061256,
        0.523599,1.178898,1.827305,
        0.523599,1.181331,-2.060330,
        0.523599,1.182540,-0.798952,
        0.523599,1.184569,0.147517,
        0.523599,1.184734,1.968919,
        0.523599,1.184804,0.593386,
        0.523599,1.188278,1.690808,
        0.523599,1.194360,2.110440,
        0.523599,1.194776,2.710455,
        0.523599,1.197745,-1.173705,
        0.523599,1.199898,-2.777257,
        0.523599,1.201164,2.355931,
        0.523599,1.203760,1.556404,
        0.523599,1.205845,-0.415993,
        0.523599,1.206585,0.933648,
        0.523599,1.206950,-0.037248,
        0.523599,1.209341,0.471162,
        0.523599,1.213483,-1.039417,
        0.523599,1.218146,-0.672643,
        0.523599,1.221608,1.421669,
        0.523599,1.226043,-1.693533,
        0.523599,1.232870,-1.822773,
        0.523599,1.234375,-2.421194,
        0.523599,1.234531,-2.190148,
        0.523599,1.239414,0.348959,
        0.523599,1.240183,2.587255,
        0.523599,1.242312,2.941234,
        0.523599,1.244075,1.286644,
        0.523599,1.245961,-2.997461,
        0.523599,1.246495,-2.646872,
        0.523599,1.247736,2.242810,
        0.523599,1.253604,0.812222,
        0.523599,1.257816,-0.894317,
        0.523599,1.258247,-1.953634,
        0.523599,1.258993,-0.284099,
        0.523599,1.262380,-0.157609,
        0.523599,1.262791,-1.549229,
        0.523599,1.263772,-0.533073,
        0.523599,1.266731,-1.412461,
        0.523599,1.268591,-3.122146,
        0.523599,1.269614,-1.275246,
        0.523599,1.270745,1.152533,
        0.523599,1.273606,0.224545,
        0.523599,1.279716,2.812101,
        0.523599,1.283786,2.462124,
        0.523599,1.284798,0.078209,
        0.523599,1.285402,-2.882049,
        0.523599,1.286936,1.886053,
        0.523599,1.289408,0.683122,
        0.523599,1.294736,1.756184,
        0.523599,1.296978,-2.311511,
        0.523599,1.298546,2.020601,
        0.523599,1.300924,1.022478,
        0.523599,1.303305,3.043152,
        0.523599,1.304210,-2.084225,
        0.523599,1.305241,-2.530249,
        0.523599,1.307513,1.628003,
        0.523599,1.310722,0.558267,
        0.523599,1.314142,2.139329,
        0.523599,1.316087,-0.777971,
        0.523599,1.317907,-1.121023,
        0.523599,1.321019,-2.749776,
        0.523599,1.321609,2.687717,
        0.523599,1.322568,1.499415,
        0.523599,1.322972,-0.395699,
        0.523599,1.325286,-0.994192,
        0.523599,1.329146,-0.648513,
        0.523599,1.329795,-0.042570,
        0.523599,1.331651,2.344497,
        0.523599,1.336522,0.435098,
        0.523599,1.339410,0.900087,
        0.523599,1.339550,1.370353,
        0.523599,1.340441,-1.731580,
        0.523599,1.351855,-1.855234,
        0.523599,1.360161,-3.018404,
        0.523599,1.361031,-1.613971,
        0.523599,1.361196,2.914780,
        0.523599,1.361998,-2.207254,
        0.523599,1.362302,1.239384,
        0.523599,1.365228,2.563808,
        0.523599,1.365452,0.310522,
        0.523599,1.367028,-2.420041,
        0.523599,1.373422,-1.484281,
        0.523599,1.374245,-1.226417,
        0.523599,1.375020,-0.170393,
        0.523599,1.376560,-1.355481,
        0.523599,1.377070,-1.980984,
        0.523599,1.377578,-2.636339,
        0.523599,1.378096,0.774626,
        0.523599,1.379086,-0.508192,
        0.523599,1.382562,0.177141,
        0.523599,1.388849,2.233847,
        0.523599,1.389268,-0.886349,
        0.523599,1.391324,1.109396,
        0.523599,1.391699,-0.288725,
        0.523599,1.392869,3.136903,
        0.523599,1.393612,1.952016,
        0.523599,1.395545,-2.850807,
        0.523599,1.401926,1.822930,
        0.523599,1.402149,2.788898,
        0.523599,1.409552,0.647980,
        0.523599,1.410840,0.056219,
        0.523599,1.411944,2.444377,
        0.523599,1.413639,1.697982,
        0.523599,1.426479,1.573427,
        0.523599,1.426725,2.102242,
        0.523599,1.427868,0.984637,
        0.523599,1.429580,-2.107232,
        0.523599,1.431375,-2.313269,
        0.523599,1.431907,3.015410,
        0.523599,1.432325,-1.029054,
        0.523599,1.433341,0.524268,
        0.523599,1.435181,-0.739771,
        0.523599,1.437208,-2.529085,
        0.523599,1.439041,-0.617966,
        0.523599,1.441473,1.448544,
        0.523599,1.444527,2.664744,
        0.523599,1.450513,-0.400625,
        0.523599,1.451907,-0.066823,
        0.523599,1.452754,-2.738631,
        0.523599,1.454061,-2.950416,
        0.523599,1.456736,1.323154,
        0.523599,1.458370,0.400433,
        0.523599,1.459314,-1.765003,
        0.523599,1.459902,-1.145835,
        0.523599,1.464952,2.329245,
        0.523599,1.466050,0.862104,
        0.523599,1.471330,-1.886160,
        0.523599,1.475119,-1.648255,
        0.523599,1.480878,1.194576,
        0.523599,1.480942,0.273235,
        0.523599,1.481086,2.890269,
        0.523599,1.484118,-1.529156,
        0.523599,1.484741,-1.279219,
        0.523599,1.487681,-1.406069,
        0.523599,1.487886,-3.066037,
        0.523599,1.490555,2.543284,
        0.523599,1.492424,-0.190932,
        0.523599,1.495002,-0.839810,
        0.523599,1.495297,2.007188,
        0.523599,1.496414,-2.009267,
        0.523599,1.498261,-2.425270,
        0.523599,1.498907,-2.207327,
        0.523599,1.500204,0.737638,
        0.523599,1.502469,0.151156,
        0.523599,1.507448,-0.514782,
        0.523599,1.509287,2.204647,
        0.523599,1.510518,1.884920,
        0.523599,1.511948,-2.632360,
        0.523599,1.515235,1.068565,
        0.523599,1.522237,3.102018,
        0.523599,1.522373,1.764991,
        0.523599,1.522416,-0.954956,
        0.523599,1.523102,2.765920,
        0.523599,1.524811,-2.844718,
        0.523599,1.527261,-0.304916,
        0.523599,1.529788,0.613208,
        0.523599,1.533648,1.644808,
        0.523599,1.535018,0.028869,
        0.523599,1.541082,2.424608,
        0.523599,1.546010,1.523462,
        0.523599,1.551135,-0.642994,
        0.523599,1.553895,0.489607,
        0.523599,1.555041,0.947813,
        0.523599,1.559671,-1.074934,
        0.523599,1.560345,2.985899,
        0.523599,1.560487,-2.323400,
        0.523599,1.561127,1.401111,
        0.523599,1.567193,2.643172,
        0.523599,1.568892,2.098362,
        0.523599,1.571702,-2.098435,
        0.523599,1.572701,-2.528730,
        0.523599,1.573572,1.278593,
        0.523599,1.574849,-0.091505,
        0.523599,1.576334,0.365236,
        0.523599,1.577029,-2.962888,
        0.523599,1.577688,-1.814941,
        0.523599,1.579542,-0.755917,
        0.523599,1.579624,-1.199759,
        0.523599,1.581515,-0.419414,
        0.523599,1.584581,-2.738702,
        0.523599,1.586948,-1.696296,
        0.523599,1.588901,2.301496,
        0.523599,1.589936,0.825020,
        0.523599,1.590018,-1.941652,
        0.523599,1.596968,-1.328358,
        0.523599,1.597178,-1.575607,
        0.523599,1.600470,-1.453196,
        0.523599,1.601408,0.242409,
        0.523599,1.602259,1.152656,
        0.523599,1.603064,2.865774,
        0.523599,1.612085,1.968802,
        0.523599,1.612720,-3.080498,
        0.523599,1.616061,2.521531,
        0.523599,1.616900,-0.207457,
        0.523599,1.620988,-0.875920,
        0.523599,1.622114,0.701509,
        0.523599,1.622749,0.121426,
        0.523599,1.622993,-2.221629,
        0.523599,1.629346,-0.542669,
        0.523599,1.630007,1.844552,
        0.523599,1.634747,-2.426198,
        0.523599,1.642346,1.722617,
        0.523599,1.643925,1.034629,
        0.523599,1.644848,2.743774,
        0.523599,1.645258,2.188109,
        0.523599,1.645417,-2.634725,
        0.523599,1.647395,3.085437,
        0.523599,1.648538,-2.855044,
        0.523599,1.649932,0.576808,
        0.523599,1.652224,-0.996105,
        0.523599,1.652871,1.600681,
        0.523599,1.657997,-0.325127,
        0.523599,1.661936,0.001037,
        0.523599,1.665187,1.477402,
        0.523599,1.665496,2.397872,
        0.523599,1.673267,-0.665453,
        0.523599,1.673557,0.452526,
        0.523599,1.679277,-1.122421,
        0.523599,1.679696,0.912482,
        0.523599,1.682420,1.353922,
        0.523599,1.685031,-2.009038,
        0.523599,1.685196,2.967046,
        0.523599,1.690107,2.620910,
        0.523599,1.690453,-2.123076,
        0.523599,1.691693,-1.871572,
        0.523599,1.691817,2.061688,
        0.523599,1.692644,1.237670,
        0.523599,1.696386,0.328123,
        0.523599,1.697162,-1.249026,
        0.523599,1.697488,-2.322785,
        0.523599,1.698726,-2.973576,
        0.523599,1.699172,-1.749717,
        0.523599,1.704478,-0.106295,
        0.523599,1.707244,-0.444770,
        0.523599,1.707500,-2.531497,
        0.523599,1.707714,-0.787321,
        0.523599,1.708430,-1.627552,
        0.523599,1.710583,-1.376890,
        0.523599,1.712415,-2.747921,
        0.523599,1.713125,-1.503073,
        0.523599,1.713452,0.789419,
        0.523599,1.720612,2.279888,
        0.523599,1.724852,0.202831,
        0.523599,1.724868,2.846246,
        0.523599,1.728940,1.931535,
        0.523599,1.732838,1.127584,
        0.523599,1.735719,-3.092485,
        0.523599,1.739676,2.496017,
        0.523599,1.739971,-0.228876,
        0.523599,1.743328,0.663347,
        0.523599,1.744927,-0.911892,
        0.523599,1.745952,0.084733,
        0.523599,1.748077,1.804354,
        0.523599,1.752959,-0.567548,
        0.523599,1.760169,1.678132,
        0.523599,1.763887,-2.218450,
        0.523599,1.766737,2.722752,
        0.523599,1.767999,1.004737,
        0.523599,1.769149,1.550298,
        0.523599,1.770119,-2.427175,
        0.523599,1.770460,2.154208,
        0.523599,1.770514,3.070982,
        0.523599,1.770735,0.537245,
        0.523599,1.772673,-2.864760,
        0.523599,1.773527,-1.039582,
        0.523599,1.776419,-2.642421,
        0.523599,1.780769,1.416330,
        0.523599,1.786634,-0.346247,
        0.523599,1.788829,-1.947553,
        0.523599,1.793737,0.410270,
        0.523599,1.794553,2.374618,
        0.523599,1.794699,-0.694061,
        0.523599,1.797469,-1.169105,
        0.523599,1.803705,0.881076,
        0.523599,1.804124,1.264249,
        0.523599,1.807825,-2.090455,
        0.523599,1.808822,-1.817375,
        0.523599,1.809020,2.949741,
        0.523599,1.812466,2.597595,
        0.523599,1.813159,2.025395,
        0.523599,1.813816,-1.298435,
        0.523599,1.815016,0.278951,
        0.523599,1.816510,-0.008631,
        0.523599,1.817772,-1.689625,
        0.523599,1.821360,-2.984352,
        0.523599,1.822312,-0.127211,
        0.523599,1.823386,-1.429807,
        0.523599,1.823834,-1.560483,
        0.523599,1.832060,-0.821627,
        0.523599,1.832690,-0.468440,
        0.523599,1.833945,-2.320278,
        0.523599,1.834787,0.751968,
        0.523599,1.840579,-2.536445,
        0.523599,1.841033,-2.756932,
        0.523599,1.846350,1.893983,
        0.523599,1.846644,1.113766,
        0.523599,1.846833,2.826543,
        0.523599,1.847325,2.249051,
        0.523599,1.854119,0.134161,
        0.523599,1.858863,-3.106022,
        0.523599,1.865046,1.763384,
        0.523599,1.865619,-0.952036,
        0.523599,1.865912,0.624453,
        0.523599,1.866694,2.473686,
        0.523599,1.867524,-0.240083,
        0.523599,1.875561,1.632831,
        0.523599,1.876049,-0.594378,
        0.523599,1.880509,1.500204,
        0.523599,1.883504,1.360457,
        0.523599,1.886971,-2.196665,
        0.523599,1.888979,0.982922,
        0.523599,1.889815,2.701570,
        0.523599,1.892144,-2.005340,
        0.523599,1.892348,0.496522,
        0.523599,1.893281,-1.083976,
        0.523599,1.893828,3.053453,
        0.523599,1.895117,2.121107,
        0.523599,1.898741,-2.875735,
        0.523599,1.904429,-2.428029,
        0.523599,1.908408,-2.648903,
        0.523599,1.911828,-0.364489,
        0.523599,1.912224,0.366565,
        0.523599,1.914854,1.214468,
        0.523599,1.915467,-1.883143,
        0.523599,1.915763,-0.724679,
        0.523599,1.915790,-1.216366,
        0.523599,1.922256,2.347806,
        0.523599,1.922616,0.844633,
        0.523599,1.925185,0.231170,
        0.523599,1.927474,-1.754616,
        0.523599,1.928342,-1.351174,
        0.523599,1.933328,-1.623328,
        0.523599,1.933878,0.034150,
        0.523599,1.934585,-1.487979,
        0.523599,1.934802,2.927634,
        0.523599,1.934862,1.989776,
        0.523599,1.936327,-0.099350,
        0.523599,1.938088,2.578507,
        0.523599,1.945677,-2.999805,
        0.523599,1.954572,-0.859208,
        0.523599,1.957868,-0.490784,
        0.523599,1.958989,0.712772,
        0.523599,1.960212,-2.307561,
        0.523599,1.964774,1.855507,
        0.523599,1.964971,-2.100533,
        0.523599,1.969112,1.088273,
        0.523599,1.971438,-2.765415,
        0.523599,1.972312,2.805134,
        0.523599,1.974950,2.220596,
        0.523599,1.975135,-2.539127,
        0.523599,1.982107,1.720829,
        0.523599,1.983195,-3.127984,
        0.523599,1.983318,-0.247182,
        0.523599,1.986720,-0.993220,
        0.523599,1.990381,1.586005,
        0.523599,1.990470,0.580283,
        0.523599,1.990523,1.450326,
        0.523599,1.993909,1.312897,
        0.523599,1.994694,2.452813,
        0.523599,1.999226,-0.620717,
        0.523599,2.005145,0.941012,
        0.523599,2.007301,0.142337,
        0.523599,2.014149,0.444138,
        0.523599,2.014338,-1.127392,
        0.523599,2.018131,3.021778,
        0.523599,2.019676,-1.980944,
        0.523599,2.019912,2.686612,
        0.523599,2.020545,2.088672,
        0.523599,2.026074,0.300945,
        0.523599,2.026196,-2.889057,
        0.523599,2.031959,-1.268178,
        0.523599,2.032917,-2.199917,
        0.523599,2.033572,-1.842884,
        0.523599,2.034712,-2.419505,
        0.523599,2.035818,-0.379479,
        0.523599,2.036004,-0.757062,
        0.523599,2.040501,-1.411488,
        0.523599,2.042874,-1.701367,
        0.523599,2.043060,-2.653746,
        0.523599,2.043729,-1.557938,
        0.523599,2.048538,0.806623,
        0.523599,2.049006,1.199930,
        0.523599,2.049502,-0.137989,
        0.523599,2.052181,2.326959,
        0.523599,2.052794,-0.008526,
        0.523599,2.057608,1.951726,
        0.523599,2.065568,2.567730,
        0.523599,2.070371,2.879808,
        0.523599,2.071256,-3.018961,
        0.523599,2.079746,-0.897699,
        0.523599,2.082784,1.042403,
        0.523599,2.083752,-0.511086,
        0.523599,2.084401,1.810502,
        0.523599,2.085731,0.671081,
        0.523599,2.097496,1.373172,
        0.523599,2.099185,1.667006,
        0.523599,2.100482,-2.301084,
        0.523599,2.103118,2.194745,
        0.523599,2.103223,1.521685,
        0.523599,2.103695,-2.777707,
        0.523599,2.107523,3.129581,
        0.523599,2.108277,-2.095907,
        0.523599,2.108891,-2.532166,
        0.523599,2.111003,-0.268632,
        0.523599,2.112276,-1.027246,
        0.523599,2.114380,0.531457,
        0.523599,2.119166,0.100552,
        0.523599,2.119564,2.754818,
        0.523599,2.122474,-0.648314,
        0.523599,2.125183,2.445598,
        0.523599,2.129501,0.238290,
        0.523599,2.131178,0.386663,
        0.523599,2.131223,-1.174311,
        0.523599,2.134269,-1.948042,
        0.523599,2.134684,0.907065,
        0.523599,2.137615,2.988806,
        0.523599,2.145635,-1.327708,
        0.523599,2.146467,2.055081,
        0.523599,2.146582,-1.793788,
        0.523599,2.149210,-1.482321,
        0.523599,2.152901,-1.636788,
        0.523599,2.156337,1.151558,
        0.523599,2.156474,-2.912065,
        0.523599,2.157194,-0.792563,
        0.523599,2.166724,-0.401638,
        0.523599,2.166980,-0.134879,
        0.523599,2.173644,-2.403429,
        0.523599,2.175711,-2.655523,
        0.523599,2.177132,2.600419,
        0.523599,2.178879,0.767722,
        0.523599,2.180752,1.908485,
        0.523599,2.181338,2.313721,
        0.523599,2.185580,1.287673,
        0.523599,2.195400,-2.217546,
        0.523599,2.200418,-3.056442,
        0.523599,2.201829,-0.001735,
        0.523599,2.203804,1.756176,
        0.523599,2.211602,1.443930,
        0.523599,2.212686,-0.544809,
        0.523599,2.213937,0.621892,
        0.523599,2.214075,1.600691,
        0.523599,2.216256,1.020029,
        0.523599,2.219060,2.886631,
        0.523599,2.223375,-0.906185,
        0.523599,2.226002,-1.060248,
        0.523599,2.226413,-2.064034,
        0.523599,2.230554,2.171334,
        0.523599,2.232413,2.735777,
        0.523599,2.233709,3.070830,
        0.523599,2.234383,-2.792065,
        0.523599,2.237183,0.468799,
        0.523599,2.237695,-0.271388,
        0.523599,2.239419,0.151693,
        0.523599,2.243347,-2.521049,
        0.523599,2.244035,-1.229456,
        0.523599,2.244355,0.309899,
        0.523599,2.245175,2.460236,
        0.523599,2.245628,-1.899216,
        0.523599,2.245973,-0.702234,
        0.523599,2.255601,-1.728131,
        0.523599,2.256702,-1.401367,
        0.523599,2.261466,-1.557500,
        0.523599,2.267390,0.878499,
        0.523599,2.272099,2.018594,
        0.523599,2.278618,1.170386,
        0.523599,2.283304,-2.345584,
        0.523599,2.285175,-2.943090,
        0.523599,2.293000,-0.124451,
        0.523599,2.294058,-0.421070,
        0.523599,2.303283,1.855938,
        0.523599,2.304968,2.608978,
        0.523599,2.305347,2.311346,
        0.523599,2.307188,-2.656134,
        0.523599,2.307301,1.336200,
        0.523599,2.309655,0.726020,
        0.523599,2.317634,-2.185225,
        0.523599,2.321510,1.685664,
        0.523599,2.324376,-3.111006,
        0.523599,2.324731,1.511186,
        0.523599,2.329587,0.035498,
        0.523599,2.330158,2.977264,
        0.523599,2.335096,-0.588468,
        0.523599,2.336046,-1.115799,
        0.523599,2.338060,2.803613,
        0.523599,2.338083,-0.943400,
        0.523599,2.340210,0.562145,
        0.523599,2.341931,1.023926,
        0.523599,2.343265,-0.779061,
        0.523599,2.343434,-2.013692,
        0.523599,2.353029,-1.310529,
        0.523599,2.354868,0.211543,
        0.523599,2.355120,0.388593,
        0.523599,2.355165,-1.833173,
        0.523599,2.355462,2.151441,
        0.523599,2.362948,-2.487477,
        0.523599,2.363101,-2.810456,
        0.523599,2.363394,-0.272918,
        0.523599,2.364509,-1.644533,
        0.523599,2.370465,2.466824,
        0.523599,2.386110,-1.481351,
        0.523599,2.394960,0.862133,
        0.523599,2.395716,1.977184,
        0.523599,2.396339,1.200900,
        0.523599,2.407941,-2.316753,
        0.523599,2.409450,-2.988432,
        0.523599,2.417202,-0.103363,
        0.523599,2.417386,-0.443097,
        0.523599,2.417950,2.666662,
        0.523599,2.423190,1.789783,
        0.523599,2.423692,1.392696,
        0.523599,2.431122,3.086379,
        0.523599,2.431406,-2.647820,
        0.523599,2.431584,2.307631,
        0.523599,2.434090,1.592818,
        0.523599,2.435993,0.683844,
        0.523599,2.440898,-2.133300,
        0.523599,2.442222,-1.195837,
        0.523599,2.443112,2.883761,
        0.523599,2.449191,-0.633651,
        0.523599,2.450821,0.084741,
        0.523599,2.453499,-1.008861,
        0.523599,2.460010,-1.938011,
        0.523599,2.460199,-0.822008,
        0.523599,2.460902,0.490079,
        0.523599,2.461774,-1.734913,
        0.523599,2.463986,1.034407,
        0.523599,2.467647,0.287667,
        0.523599,2.479849,2.126925,
        0.523599,2.485907,-2.835471,
        0.523599,2.485945,-1.374853,
        0.523599,2.487747,2.505912,
        0.523599,2.487957,-0.274279,
        0.523599,2.489741,-2.468428,
        0.523599,2.512538,1.239900,
        0.523599,2.517528,1.923289,
        0.523599,2.518454,-1.565852,
        0.523599,2.519348,0.843503,
        0.523599,2.522766,-3.059231,
        0.523599,2.528686,2.737228,
        0.523599,2.534798,-2.266671,
        0.523599,2.536533,1.466274,
        0.523599,2.538504,1.699812,
        0.523599,2.539516,-0.476942,
        0.523599,2.539532,-0.074548,
        0.523599,2.543197,2.990252,
        0.523599,2.554022,2.320619,
        0.523599,2.556919,-2.655005,
        0.523599,2.558146,0.627191,
        0.523599,2.558223,-1.164189,
        0.523599,2.563452,-2.044028,
        0.523599,2.568462,0.150655,
        0.523599,2.569130,-0.700837,
        0.523599,2.570050,-1.804492,
        0.523599,2.572710,-0.934665,
        0.523599,2.575994,0.391718,
        0.523599,2.584774,1.046703,
        0.523599,2.601432,2.102535,
        0.523599,2.603559,2.555114,
        0.523599,2.604150,-2.885208,
        0.523599,2.608630,-1.378011,
        0.523599,2.611256,-0.279110,
        0.523599,2.617344,-2.439480,
        0.523599,2.627402,1.293552,
        0.523599,2.635396,1.848850,
        0.523599,2.635686,2.834950,
        0.523599,2.637087,3.132730,
        0.523599,2.642029,0.813274,
        0.523599,2.645260,1.570203,
        0.523599,2.646182,-1.614036,
        0.523599,2.660113,-2.186513,
        0.523599,2.660483,-0.034608,
        0.523599,2.661713,-0.529511,
        0.523599,2.672720,-1.110422,
        0.523599,2.676626,2.335353,
        0.523599,2.676638,0.540364,
        0.523599,2.676780,-1.901158,
        0.523599,2.681411,-2.667965,
        0.523599,2.682152,0.245751,
        0.523599,2.682859,-0.817071,
        0.523599,2.705107,1.062142,
        0.523599,2.717705,2.625995,
        0.523599,2.721001,2.060097,
        0.523599,2.722319,-2.954069,
        0.523599,2.729840,-1.368237,
        0.523599,2.734349,-0.289298,
        0.523599,2.738409,2.975671,
        0.523599,2.740366,1.374341,
        0.523599,2.744633,-2.391311,
        0.523599,2.747398,1.727166,
        0.523599,2.763799,0.760372,
        0.523599,2.768238,-1.679243,
        0.523599,2.779245,-2.046268,
        0.523599,2.780208,-0.626453,
        0.523599,2.780471,0.027732,
        0.523599,2.784696,-1.013128,
        0.523599,2.789817,0.395431,
        0.523599,2.801154,2.353143,
        0.523599,2.807091,-2.687477,
        0.523599,2.825762,1.085035,
        0.523599,2.830209,2.734943,
        0.523599,2.833517,-3.085508,
        0.523599,2.838301,1.980061,
        0.523599,2.849871,1.512012,
        0.523599,2.850089,-1.343899,
        0.523599,2.856905,-0.318441,
        0.523599,2.870445,-2.291001,
        0.523599,2.885017,0.655562,
        0.523599,2.886428,-1.793217,
        0.523599,2.890337,-0.825633,
        0.523599,2.898932,0.139973,
        0.523599,2.931927,-2.728768,
        0.523599,2.932096,2.377790,
        0.523599,2.937409,2.947245,
        0.523599,2.948048,1.133127,
        0.523599,2.950896,1.797889,
        0.523599,2.970019,-1.282382,
        0.523599,2.978333,-0.412090,
        0.523599,2.996437,-2.089121,
        0.523599,3.011274,0.434110,
        0.523599,3.050410,3.138492,
        0.523599,3.064492,1.647259,
        0.523599,3.088868,-0.989501,
        0.628319,0.053884,-2.408394,
        0.628319,0.077645,1.453433,
        0.628319,0.088408,-0.311160,
        0.628319,0.142314,2.590387,
        0.628319,0.147206,-1.385164,
        0.628319,0.163506,0.573203,
        0.628319,0.165485,-2.875396,
        0.628319,0.182564,-2.125821,
        0.628319,0.190569,1.929141,
        0.628319,0.201102,1.249082,
        0.628319,0.201397,-0.716646,
        0.628319,0.210568,-0.069559,
        0.628319,0.251672,2.848900,
        0.628319,0.266262,-1.678068,
        0.628319,0.269108,-1.215563,
        0.628319,0.273777,-2.506756,
        0.628319,0.279854,0.814581,
        0.628319,0.281664,-2.969147,
        0.628319,0.283625,2.405603,
        0.628319,0.284058,0.357350,
        0.628319,0.294107,1.593880,
        0.628319,0.302491,-0.400410,
        0.628319,0.308059,2.021491,
        0.628319,0.312219,-2.072057,
        0.628319,0.321905,-0.829121,
        0.628319,0.329984,1.201621,
        0.628319,0.339714,-0.013712,
        0.628319,0.368784,2.974974,
        0.628319,0.376428,-1.495532,
        0.628319,0.378103,2.646436,
        0.628319,0.385088,0.590355,
        0.628319,0.386248,-2.690943,
        0.628319,0.389851,-2.340031,
        0.628319,0.391502,-1.171374,
        0.628319,0.398249,-1.812694,
        0.628319,0.400542,-2.995729,
        0.628319,0.404273,0.918163,
        0.628319,0.405617,1.773533,
        0.628319,0.409009,-0.570243,
        0.628319,0.410700,0.270433,
        0.628319,0.415893,1.464890,
        0.628319,0.420680,2.088164,
        0.628319,0.424766,-0.258729,
        0.628319,0.426507,2.367872,
        0.628319,0.441292,-0.890727,
        0.628319,0.451839,-2.078193,
        0.628319,0.459270,1.191562,
        0.628319,0.470622,0.014002,
        0.628319,0.481292,-2.529098,
        0.628319,0.482460,2.849843,
        0.628319,0.492205,-1.397963,
        0.628319,0.493314,3.117264,
        0.628319,0.497870,0.727828,
        0.628319,0.499783,-1.654379,
        0.628319,0.502651,0.463333,
        0.628319,0.505463,2.594557,
        0.628319,0.506104,-2.874910,
        0.628319,0.513579,-1.147354,
        0.628319,0.518569,1.644237,
        0.628319,0.519001,1.900488,
        0.628319,0.519277,-0.673809,
        0.628319,0.524769,-0.421955,
        0.628319,0.525564,-2.288838,
        0.628319,0.532541,0.978737,
        0.628319,0.534610,2.154808,
        0.628319,0.535020,-1.891846,
        0.628319,0.540208,1.400497,
        0.628319,0.542006,0.228337,
        0.628319,0.550762,-0.180771,
        0.628319,0.560531,2.378474,
        0.628319,0.561629,-0.924576,
        0.628319,0.567465,-2.679392,
        0.628319,0.588527,1.181755,
        0.628319,0.591643,-2.098364,
        0.628319,0.594963,2.997575,
        0.628319,0.599643,-3.045235,
        0.628319,0.599962,2.773556,
        0.628319,0.602620,0.029270,
        0.628319,0.605711,0.598454,
        0.628319,0.607708,-1.541026,
        0.628319,0.609360,-2.464379,
        0.628319,0.610738,-1.326786,
        0.628319,0.616968,0.820489,
        0.628319,0.626379,1.773610,
        0.628319,0.628626,-1.751180,
        0.628319,0.630671,0.391949,
        0.628319,0.632622,-0.737642,
        0.628319,0.632708,2.566722,
        0.628319,0.632797,1.989858,
        0.628319,0.634455,-0.534790,
        0.628319,0.636205,1.561170,
        0.628319,0.637681,-1.123423,
        0.628319,0.641441,-2.842109,
        0.628319,0.644807,-0.327139,
        0.628319,0.652440,2.197129,
        0.628319,0.660897,-2.278593,
        0.628319,0.664008,1.360091,
        0.628319,0.670850,-1.945257,
        0.628319,0.671396,1.004859,
        0.628319,0.673985,0.203855,
        0.628319,0.680326,-0.133370,
        0.628319,0.685083,-0.942363,
        0.628319,0.688204,-2.632204,
        0.628319,0.689777,2.385011,
        0.628319,0.700065,3.115430,
        0.628319,0.704080,2.911636,
        0.628319,0.712883,1.174216,
        0.628319,0.715701,0.699906,
        0.628319,0.718851,-1.442215,
        0.628319,0.721019,2.721536,
        0.628319,0.725300,-2.975134,
        0.628319,0.727136,-1.633288,
        0.628319,0.727399,-2.120567,
        0.628319,0.730751,0.522488,
        0.628319,0.732892,-1.254330,
        0.628319,0.734333,-2.446184,
        0.628319,0.735125,0.037270,
        0.628319,0.737893,1.866463,
        0.628319,0.740650,1.681075,
        0.628319,0.745971,-0.781286,
        0.628319,0.749432,-0.442559,
        0.628319,0.750102,2.050636,
        0.628319,0.751046,0.863821,
        0.628319,0.757018,-0.608833,
        0.628319,0.757046,-1.816019,
        0.628319,0.757389,1.501184,
        0.628319,0.759800,2.546181,
        0.628319,0.759859,0.351923,
        0.628319,0.761141,-2.781102,
        0.628319,0.769035,-1.078574,
        0.628319,0.770166,-0.266604,
        0.628319,0.772218,2.227380,
        0.628319,0.788382,1.331051,
        0.628319,0.790269,-2.282409,
        0.628319,0.803410,-1.984567,
        0.628319,0.803813,0.191520,
        0.628319,0.806832,3.022155,
        0.628319,0.807476,-2.597675,
        0.628319,0.812512,-0.107335,
        0.628319,0.814161,1.010949,
        0.628319,0.817732,-3.084486,
        0.628319,0.817846,2.390688,
        0.628319,0.819323,2.845960,
        0.628319,0.819430,-0.915017,
        0.628319,0.829198,-1.532491,
        0.628319,0.829531,0.665695,
        0.628319,0.829816,1.167147,
        0.628319,0.833613,-1.357492,
        0.628319,0.842945,-2.907318,
        0.628319,0.844883,2.679708,
        0.628319,0.848189,-1.705003,
        0.628319,0.849888,1.767901,
        0.628319,0.851126,0.493370,
        0.628319,0.852911,1.933394,
        0.628319,0.857516,-2.432863,
        0.628319,0.858261,-1.191729,
        0.628319,0.858279,-2.138895,
        0.628319,0.859723,-0.734928,
        0.628319,0.859754,1.603292,
        0.628319,0.861655,-0.394821,
        0.628319,0.864300,0.043549,
        0.628319,0.866940,0.851229,
        0.628319,0.869748,2.095969,
        0.628319,0.877352,-0.574668,
        0.628319,0.879254,-2.733338,
        0.628319,0.882554,1.443757,
        0.628319,0.884060,0.334199,
        0.628319,0.885626,-1.866061,
        0.628319,0.889834,2.529957,
        0.628319,0.893140,2.251698,
        0.628319,0.897463,-0.240822,
        0.628319,0.898758,-1.031948,
        0.628319,0.913918,1.287618,
        0.628319,0.916555,-2.287173,
        0.628319,0.918305,3.107845,
        0.628319,0.919522,2.948277,
        0.628319,0.927138,-2.569686,
        0.628319,0.929031,0.185383,
        0.628319,0.933766,-3.015917,
        0.628319,0.934873,-2.010822,
        0.628319,0.935636,-1.445855,
        0.628319,0.936702,-0.866461,
        0.628319,0.937395,0.729108,
        0.628319,0.937910,2.787575,
        0.628319,0.939144,-0.092962,
        0.628319,0.941718,1.125085,
        0.628319,0.942873,-1.607680,
        0.628319,0.943701,0.980328,
        0.628319,0.947774,2.395256,
        0.628319,0.952433,0.584347,
        0.628319,0.954456,-1.287293,
        0.628319,0.958731,-2.852552,
        0.628319,0.961838,1.839284,
        0.628319,0.961898,-0.469425,
        0.628319,0.965426,1.686615,
        0.628319,0.970108,1.990771,
        0.628319,0.972820,-1.762780,
        0.628319,0.973500,-0.711881,
        0.628319,0.975253,2.641082,
        0.628319,0.977824,0.441928,
        0.628319,0.981485,-2.424766,
        0.628319,0.981877,1.537100,
        0.628319,0.982609,-1.136920,
        0.628319,0.984265,-2.156073,
        0.628319,0.989150,0.045493,
        0.628319,0.993052,2.137478,
        0.628319,0.997766,-2.693871,
        0.628319,0.999847,-0.334738,
        0.628319,1.001920,0.848054,
        0.628319,1.006146,1.389734,
        0.628319,1.012180,0.304066,
        0.628319,1.016938,2.270674,
        0.628319,1.017371,-0.979790,
        0.628319,1.019495,-1.894487,
        0.628319,1.023796,-0.583691,
        0.628319,1.026525,2.506945,
        0.628319,1.027004,-0.203653,
        0.628319,1.028027,3.036573,
        0.628319,1.031435,1.241021,
        0.628319,1.032316,2.884221,
        0.628319,1.035863,-3.110475,
        0.628319,1.042085,-2.292950,
        0.628319,1.043272,-1.522986,
        0.628319,1.046265,-2.958847,
        0.628319,1.049272,-1.376344,
        0.628319,1.050039,-2.552086,
        0.628319,1.053682,0.689863,
        0.628319,1.055458,0.170394,
        0.628319,1.056622,-0.831866,
        0.628319,1.057515,1.097998,
        0.628319,1.058706,-2.036438,
        0.628319,1.061389,-1.672248,
        0.628319,1.063758,2.743093,
        0.628319,1.070157,-0.073422,
        0.628319,1.072154,1.761506,
        0.628319,1.074029,1.909193,
        0.628319,1.074204,0.545276,
        0.628319,1.074714,-2.805147,
        0.628319,1.076262,0.966104,
        0.628319,1.077642,-1.227073,
        0.628319,1.084151,1.618525,
        0.628319,1.085085,2.377673,
        0.628319,1.086822,-0.441112,
        0.628319,1.087142,2.058815,
        0.628319,1.098383,-0.696430,
        0.628319,1.099138,-1.084033,
        0.628319,1.103208,1.477773,
        0.628319,1.103783,0.408254,
        0.628319,1.106275,-2.420800,
        0.628319,1.108079,2.612943,
        0.628319,1.109345,-2.172718,
        0.628319,1.115957,-1.787597,
        0.628319,1.118930,0.045983,
        0.628319,1.119777,0.834702,
        0.628319,1.121943,-2.669042,
        0.628319,1.124930,2.219917,
        0.628319,1.125217,1.336294,
        0.628319,1.130958,2.972931,
        0.628319,1.131400,-0.306629,
        0.628319,1.137135,-0.936514,
        0.628319,1.139125,-1.924528,
        0.628319,1.139877,0.276382,
        0.628319,1.142003,-3.058022,
        0.628319,1.147363,-0.558448,
        0.628319,1.150348,-0.176418,
        0.628319,1.151198,1.197016,
        0.628319,1.153930,-1.597104,
        0.628319,1.155389,2.837148,
        0.628319,1.155458,-1.460623,
        0.628319,1.157331,-2.906362,
        0.628319,1.158594,2.487618,
        0.628319,1.162103,-1.322553,
        0.628319,1.165407,3.092861,
        0.628319,1.168025,0.717075,
        0.628319,1.168406,-2.301519,
        0.628319,1.174782,-2.538755,
        0.628319,1.177998,1.061256,
        0.628319,1.178898,1.827305,
        0.628319,1.181331,-2.060330,
        0.628319,1.182540,-0.798952,
        0.628319,1.184569,0.147517,
        0.628319,1.184734,1.968919,
        0.628319,1.184804,0.593386,
        0.628319,1.188278,1.690808,
        0.628319,1.194360,2.110440,
        0.628319,1.194776,2.710455,
        0.628319,1.197745,-1.173705,
        0.628319,1.199898,-2.777257,
        0.628319,1.201164,2.355931,
        0.628319,1.203760,1.556404,
        0.628319,1.205845,-0.415993,
        0.628319,1.206585,0.933648,
        0.628319,1.206950,-0.037248,
        0.628319,1.209341,0.471162,
        0.628319,1.213483,-1.039417,
        0.628319,1.218146,-0.672643,
        0.628319,1.221608,1.421669,
        0.628319,1.226043,-1.693533,
        0.628319,1.232870,-1.822773,
        0.628319,1.234375,-2.421194,
        0.628319,1.234531,-2.190148,
        0.628319,1.239414,0.348959,
        0.628319,1.240183,2.587255,
        0.628319,1.242312,2.941234,
        0.628319,1.244075,1.286644,
        0.628319,1.245961,-2.997461,
        0.628319,1.246495,-2.646872,
        0.628319,1.247736,2.242810,
        0.628319,1.253604,0.812222,
        0.628319,1.257816,-0.894317,
        0.628319,1.258247,-1.953634,
        0.628319,1.258993,-0.284099,
        0.628319,1.262380,-0.157609,
        0.628319,1.262791,-1.549229,
        0.628319,1.263772,-0.533073,
        0.628319,1.266731,-1.412461,
        0.628319,1.268591,-3.122146,
        0.628319,1.269614,-1.275246,
        0.628319,1.270745,1.152533,
        0.628319,1.273606,0.224545,
        0.628319,1.279716,2.812101,
        0.628319,1.283786,2.462124,
        0.628319,1.284798,0.078209,
        0.628319,1.285402,-2.882049,
        0.628319,1.286936,1.886053,
        0.628319,1.289408,0.683122,
        0.628319,1.294736,1.756184,
        0.628319,1.296978,-2.311511,
        0.628319,1.298546,2.020601,
        0.628319,1.300924,1.022478,
        0.628319,1.303305,3.043152,
        0.628319,1.304210,-2.084225,
        0.628319,1.305241,-2.530249,
        0.628319,1.307513,1.628003,
        0.628319,1.310722,0.558267,
        0.628319,1.314142,2.139329,
        0.628319,1.316087,-0.777971,
        0.628319,1.317907,-1.121023,
        0.628319,1.321019,-2.749776,
        0.628319,1.321609,2.687717,
        0.628319,1.322568,1.499415,
        0.628319,1.322972,-0.395699,
        0.628319,1.325286,-0.994192,
        0.628319,1.329146,-0.648513,
        0.628319,1.329795,-0.042570,
        0.628319,1.331651,2.344497,
        0.628319,1.336522,0.435098,
        0.628319,1.339410,0.900087,
        0.628319,1.339550,1.370353,
        0.628319,1.340441,-1.731580,
        0.628319,1.351855,-1.855234,
        0.628319,1.360161,-3.018404,
        0.628319,1.361031,-1.613971,
        0.628319,1.361196,2.914780,
        0.628319,1.361998,-2.207254,
        0.628319,1.362302,1.239384,
        0.628319,1.365228,2.563808,
        0.628319,1.365452,0.310522,
        0.628319,1.367028,-2.420041,
        0.628319,1.373422,-1.484281,
        0.628319,1.374245,-1.226417,
        0.628319,1.375020,-0.170393,
        0.628319,1.376560,-1.355481,
        0.628319,1.377070,-1.980984,
        0.628319,1.377578,-2.636339,
        0.628319,1.378096,0.774626,
        0.628319,1.379086,-0.508192,
        0.628319,1.382562,0.177141,
        0.628319,1.388849,2.233847,
        0.628319,1.389268,-0.886349,
        0.628319,1.391324,1.109396,
        0.628319,1.391699,-0.288725,
        0.628319,1.392869,3.136903,
        0.628319,1.393612,1.952016,
        0.628319,1.395545,-2.850807,
        0.628319,1.401926,1.822930,
        0.628319,1.402149,2.788898,
        0.628319,1.409552,0.647980,
        0.628319,1.410840,0.056219,
        0.628319,1.411944,2.444377,
        0.628319,1.413639,1.697982,
        0.628319,1.426479,1.573427,
        0.628319,1.426725,2.102242,
        0.628319,1.427868,0.984637,
        0.628319,1.429580,-2.107232,
        0.628319,1.431375,-2.313269,
        0.628319,1.431907,3.015410,
        0.628319,1.432325,-1.029054,
        0.628319,1.433341,0.524268,
        0.628319,1.435181,-0.739771,
        0.628319,1.437208,-2.529085,
        0.628319,1.439041,-0.617966,
        0.628319,1.441473,1.448544,
        0.628319,1.444527,2.664744,
        0.628319,1.450513,-0.400625,
        0.628319,1.451907,-0.066823,
        0.628319,1.452754,-2.738631,
        0.628319,1.454061,-2.950416,
        0.628319,1.456736,1.323154,
        0.628319,1.458370,0.400433,
        0.628319,1.459314,-1.765003,
        0.628319,1.459902,-1.145835,
        0.628319,1.464952,2.329245,
        0.628319,1.466050,0.862104,
        0.628319,1.471330,-1.886160,
        0.628319,1.475119,-1.648255,
        0.628319,1.480878,1.194576,
        0.628319,1.480942,0.273235,
        0.628319,1.481086,2.890269,
        0.628319,1.484118,-1.529156,
        0.628319,1.484741,-1.279219,
        0.628319,1.487681,-1.406069,
        0.628319,1.487886,-3.066037,
        0.628319,1.490555,2.543284,
        0.628319,1.492424,-0.190932,
        0.628319,1.495002,-0.839810,
        0.628319,1.495297,2.007188,
        0.628319,1.496414,-2.009267,
        0.628319,1.498261,-2.425270,
        0.628319,1.498907,-2.207327,
        0.628319,1.500204,0.737638,
        0.628319,1.502469,0.151156,
        0.628319,1.507448,-0.514782,
        0.628319,1.509287,2.204647,
        0.628319,1.510518,1.884920,
        0.628319,1.511948,-2.632360,
        0.628319,1.515235,1.068565,
        0.628319,1.522237,3.102018,
        0.628319,1.522373,1.764991,
        0.628319,1.522416,-0.954956,
        0.628319,1.523102,2.765920,
        0.628319,1.524811,-2.844718,
        0.628319,1.527261,-0.304916,
        0.628319,1.529788,0.613208,
        0.628319,1.533648,1.644808,
        0.628319,1.535018,0.028869,
        0.628319,1.541082,2.424608,
        0.628319,1.546010,1.523462,
        0.628319,1.551135,-0.642994,
        0.628319,1.553895,0.489607,
        0.628319,1.555041,0.947813,
        0.628319,1.559671,-1.074934,
        0.628319,1.560345,2.985899,
        0.628319,1.560487,-2.323400,
        0.628319,1.561127,1.401111,
        0.628319,1.567193,2.643172,
        0.628319,1.568892,2.098362,
        0.628319,1.571702,-2.098435,
        0.628319,1.572701,-2.528730,
        0.628319,1.573572,1.278593,
        0.628319,1.574849,-0.091505,
        0.628319,1.576334,0.365236,
        0.628319,1.577029,-2.962888,
        0.628319,1.577688,-1.814941,
        0.628319,1.579542,-0.755917,
        0.628319,1.579624,-1.199759,
        0.628319,1.581515,-0.419414,
        0.628319,1.584581,-2.738702,
        0.628319,1.586948,-1.696296,
        0.628319,1.588901,2.301496,
        0.628319,1.589936,0.825020,
        0.628319,1.590018,-1.941652,
        0.628319,1.596968,-1.328358,
        0.628319,1.597178,-1.575607,
        0.628319,1.600470,-1.453196,
        0.628319,1.601408,0.242409,
        0.628319,1.602259,1.152656,
        0.628319,1.603064,2.865774,
        0.628319,1.612085,1.968802,
        0.628319,1.612720,-3.080498,
        0.628319,1.616061,2.521531,
        0.628319,1.616900,-0.207457,
        0.628319,1.620988,-0.875920,
        0.628319,1.622114,0.701509,
        0.628319,1.622749,0.121426,
        0.628319,1.622993,-2.221629,
        0.628319,1.629346,-0.542669,
        0.628319,1.630007,1.844552,
        0.628319,1.634747,-2.426198,
        0.628319,1.642346,1.722617,
        0.628319,1.643925,1.034629,
        0.628319,1.644848,2.743774,
        0.628319,1.645258,2.188109,
        0.628319,1.645417,-2.634725,
        0.628319,1.647395,3.085437,
        0.628319,1.648538,-2.855044,
        0.628319,1.649932,0.576808,
        0.628319,1.652224,-0.996105,
        0.628319,1.652871,1.600681,
        0.628319,1.657997,-0.325127,
        0.628319,1.661936,0.001037,
        0.628319,1.665187,1.477402,
        0.628319,1.665496,2.397872,
        0.628319,1.673267,-0.665453,
        0.628319,1.673557,0.452526,
        0.628319,1.679277,-1.122421,
        0.628319,1.679696,0.912482,
        0.628319,1.682420,1.353922,
        0.628319,1.685031,-2.009038,
        0.628319,1.685196,2.967046,
        0.628319,1.690107,2.620910,
        0.628319,1.690453,-2.123076,
        0.628319,1.691693,-1.871572,
        0.628319,1.691817,2.061688,
        0.628319,1.692644,1.237670,
        0.628319,1.696386,0.328123,
        0.628319,1.697162,-1.249026,
        0.628319,1.697488,-2.322785,
        0.628319,1.698726,-2.973576,
        0.628319,1.699172,-1.749717,
        0.628319,1.704478,-0.106295,
        0.628319,1.707244,-0.444770,
        0.628319,1.707500,-2.531497,
        0.628319,1.707714,-0.787321,
        0.628319,1.708430,-1.627552,
        0.628319,1.710583,-1.376890,
        0.628319,1.712415,-2.747921,
        0.628319,1.713125,-1.503073,
        0.628319,1.713452,0.789419,
        0.628319,1.720612,2.279888,
        0.628319,1.724852,0.202831,
        0.628319,1.724868,2.846246,
        0.628319,1.728940,1.931535,
        0.628319,1.732838,1.127584,
        0.628319,1.735719,-3.092485,
        0.628319,1.739676,2.496017,
        0.628319,1.739971,-0.228876,
        0.628319,1.743328,0.663347,
        0.628319,1.744927,-0.911892,
        0.628319,1.745952,0.084733,
        0.628319,1.748077,1.804354,
        0.628319,1.752959,-0.567548,
        0.628319,1.760169,1.678132,
        0.628319,1.763887,-2.218450,
        0.628319,1.766737,2.722752,
        0.628319,1.767999,1.004737,
        0.628319,1.769149,1.550298,
        0.628319,1.770119,-2.427175,
        0.628319,1.770460,2.154208,
        0.628319,1.770514,3.070982,
        0.628319,1.770735,0.537245,
        0.628319,1.772673,-2.864760,
        0.628319,1.773527,-1.039582,
        0.628319,1.776419,-2.642421,
        0.628319,1.780769,1.416330,
        0.628319,1.786634,-0.346247,
        0.628319,1.788829,-1.947553,
        0.628319,1.793737,0.410270,
        0.628319,1.794553,2.374618,
        0.628319,1.794699,-0.694061,
        0.628319,1.797469,-1.169105,
        0.628319,1.803705,0.881076,
        0.628319,1.804124,1.264249,
        0.628319,1.807825,-2.090455,
        0.628319,1.808822,-1.817375,
        0.628319,1.809020,2.949741,
        0.628319,1.812466,2.597595,
        0.628319,1.813159,2.025395,
        0.628319,1.813816,-1.298435,
        0.628319,1.815016,0.278951,
        0.628319,1.816510,-0.008631,
        0.628319,1.817772,-1.689625,
        0.628319,1.821360,-2.984352,
        0.628319,1.822312,-0.127211,
        0.628319,1.823386,-1.429807,
        0.628319,1.823834,-1.560483,
        0.628319,1.832060,-0.821627,
        0.628319,1.832690,-0.468440,
        0.628319,1.833945,-2.320278,
        0.628319,1.834787,0.751968,
        0.628319,1.840579,-2.536445,
        0.628319,1.841033,-2.756932,
        0.628319,1.846350,1.893983,
        0.628319,1.846644,1.113766,
        0.628319,1.846833,2.826543,
        0.628319,1.847325,2.249051,
        0.628319,1.854119,0.134161,
        0.628319,1.858863,-3.106022,
        0.628319,1.865046,1.763384,
        0.628319,1.865619,-0.952036,
        0.628319,1.865912,0.624453,
        0.628319,1.866694,2.473686,
        0.628319,1.867524,-0.240083,
        0.628319,1.875561,1.632831,
        0.628319,1.876049,-0.594378,
        0.628319,1.880509,1.500204,
        0.628319,1.883504,1.360457,
        0.628319,1.886971,-2.196665,
        0.628319,1.888979,0.982922,
        0.628319,1.889815,2.701570,
        0.628319,1.892144,-2.005340,
        0.628319,1.892348,0.496522,
        0.628319,1.893281,-1.083976,
        0.628319,1.893828,3.053453,
        0.628319,1.895117,2.121107,
        0.628319,1.898741,-2.875735,
        0.628319,1.904429,-2.428029,
        0.628319,1.908408,-2.648903,
        0.628319,1.911828,-0.364489,
        0.628319,1.912224,0.366565,
        0.628319,1.914854,1.214468,
        0.628319,1.915467,-1.883143,
        0.628319,1.915763,-0.724679,
        0.628319,1.915790,-1.216366,
        0.628319,1.922256,2.347806,
        0.628319,1.922616,0.844633,
        0.628319,1.925185,0.231170,
        0.628319,1.927474,-1.754616,
        0.628319,1.928342,-1.351174,
        0.628319,1.933328,-1.623328,
        0.628319,1.933878,0.034150,
        0.628319,1.934585,-1.487979,
        0.628319,1.934802,2.927634,
        0.628319,1.934862,1.989776,
        0.628319,1.936327,-0.099350,
        0.628319,1.938088,2.578507,
        0.628319,1.945677,-2.999805,
        0.628319,1.954572,-0.859208,
        0.628319,1.957868,-0.490784,
        0.628319,1.958989,0.712772,
        0.628319,1.960212,-2.307561,
        0.628319,1.964774,1.855507,
        0.628319,1.964971,-2.100533,
        0.628319,1.969112,1.088273,
        0.628319,1.971438,-2.765415,
        0.628319,1.972312,2.805134,
        0.628319,1.974950,2.220596,
        0.628319,1.975135,-2.539127,
        0.628319,1.982107,1.720829,
        0.628319,1.983195,-3.127984,
        0.628319,1.983318,-0.247182,
        0.628319,1.986720,-0.993220,
        0.628319,1.990381,1.586005,
        0.628319,1.990470,0.580283,
        0.628319,1.990523,1.450326,
        0.628319,1.993909,1.312897,
        0.628319,1.994694,2.452813,
        0.628319,1.999226,-0.620717,
        0.628319,2.005145,0.941012,
        0.628319,2.007301,0.142337,
        0.628319,2.014149,0.444138,
        0.628319,2.014338,-1.127392,
        0.628319,2.018131,3.021778,
        0.628319,2.019676,-1.980944,
        0.628319,2.019912,2.686612,
        0.628319,2.020545,2.088672,
        0.628319,2.026074,0.300945,
        0.628319,2.026196,-2.889057,
        0.628319,2.031959,-1.268178,
        0.628319,2.032917,-2.199917,
        0.628319,2.033572,-1.842884,
        0.628319,2.034712,-2.419505,
        0.628319,2.035818,-0.379479,
        0.628319,2.036004,-0.757062,
        0.628319,2.040501,-1.411488,
        0.628319,2.042874,-1.701367,
        0.628319,2.043060,-2.653746,
        0.628319,2.043729,-1.557938,
        0.628319,2.048538,0.806623,
        0.628319,2.049006,1.199930,
        0.628319,2.049502,-0.137989,
        0.628319,2.052181,2.326959,
        0.628319,2.052794,-0.008526,
        0.628319,2.057608,1.951726,
        0.628319,2.065568,2.567730,
        0.628319,2.070371,2.879808,
        0.628319,2.071256,-3.018961,
        0.628319,2.079746,-0.897699,
        0.628319,2.082784,1.042403,
        0.628319,2.083752,-0.511086,
        0.628319,2.084401,1.810502,
        0.628319,2.085731,0.671081,
        0.628319,2.097496,1.373172,
        0.628319,2.099185,1.667006,
        0.628319,2.100482,-2.301084,
        0.628319,2.103118,2.194745,
        0.628319,2.103223,1.521685,
        0.628319,2.103695,-2.777707,
        0.628319,2.107523,3.129581,
        0.628319,2.108277,-2.095907,
        0.628319,2.108891,-2.532166,
        0.628319,2.111003,-0.268632,
        0.628319,2.112276,-1.027246,
        0.628319,2.114380,0.531457,
        0.628319,2.119166,0.100552,
        0.628319,2.119564,2.754818,
        0.628319,2.122474,-0.648314,
        0.628319,2.125183,2.445598,
        0.628319,2.129501,0.238290,
        0.628319,2.131178,0.386663,
        0.628319,2.131223,-1.174311,
        0.628319,2.134269,-1.948042,
        0.628319,2.134684,0.907065,
        0.628319,2.137615,2.988806,
        0.628319,2.145635,-1.327708,
        0.628319,2.146467,2.055081,
        0.628319,2.146582,-1.793788,
        0.628319,2.149210,-1.482321,
        0.628319,2.152901,-1.636788,
        0.628319,2.156337,1.151558,
        0.628319,2.156474,-2.912065,
        0.628319,2.157194,-0.792563,
        0.628319,2.166724,-0.401638,
        0.628319,2.166980,-0.134879,
        0.628319,2.173644,-2.403429,
        0.628319,2.175711,-2.655523,
        0.628319,2.177132,2.600419,
        0.628319,2.178879,0.767722,
        0.628319,2.180752,1.908485,
        0.628319,2.181338,2.313721,
        0.628319,2.185580,1.287673,
        0.628319,2.195400,-2.217546,
        0.628319,2.200418,-3.056442,
        0.628319,2.201829,-0.001735,
        0.628319,2.203804,1.756176,
        0.628319,2.211602,1.443930,
        0.628319,2.212686,-0.544809,
        0.628319,2.213937,0.621892,
        0.628319,2.214075,1.600691,
        0.628319,2.216256,1.020029,
        0.628319,2.219060,2.886631,
        0.628319,2.223375,-0.906185,
        0.628319,2.226002,-1.060248,
        0.628319,2.226413,-2.064034,
        0.628319,2.230554,2.171334,
        0.628319,2.232413,2.735777,
        0.628319,2.233709,3.070830,
        0.628319,2.234383,-2.792065,
        0.628319,2.237183,0.468799,
        0.628319,2.237695,-0.271388,
        0.628319,2.239419,0.151693,
        0.628319,2.243347,-2.521049,
        0.628319,2.244035,-1.229456,
        0.628319,2.244355,0.309899,
        0.628319,2.245175,2.460236,
        0.628319,2.245628,-1.899216,
        0.628319,2.245973,-0.702234,
        0.628319,2.255601,-1.728131,
        0.628319,2.256702,-1.401367,
        0.628319,2.261466,-1.557500,
        0.628319,2.267390,0.878499,
        0.628319,2.272099,2.018594,
        0.628319,2.278618,1.170386,
        0.628319,2.283304,-2.345584,
        0.628319,2.285175,-2.943090,
        0.628319,2.293000,-0.124451,
        0.628319,2.294058,-0.421070,
        0.628319,2.303283,1.855938,
        0.628319,2.304968,2.608978,
        0.628319,2.305347,2.311346,
        0.628319,2.307188,-2.656134,
        0.628319,2.307301,1.336200,
        0.628319,2.309655,0.726020,
        0.628319,2.317634,-2.185225,
        0.628319,2.321510,1.685664,
        0.628319,2.324376,-3.111006,
        0.628319,2.324731,1.511186,
        0.628319,2.329587,0.035498,
        0.628319,2.330158,2.977264,
        0.628319,2.335096,-0.588468,
        0.628319,2.336046,-1.115799,
        0.628319,2.338060,2.803613,
        0.628319,2.338083,-0.943400,
        0.628319,2.340210,0.562145,
        0.628319,2.341931,1.023926,
        0.628319,2.343265,-0.779061,
        0.628319,2.343434,-2.013692,
        0.628319,2.353029,-1.310529,
        0.628319,2.354868,0.211543,
        0.628319,2.355120,0.388593,
        0.628319,2.355165,-1.833173,
        0.628319,2.355462,2.151441,
        0.628319,2.362948,-2.487477,
        0.628319,2.363101,-2.810456,
        0.628319,2.363394,-0.272918,
        0.628319,2.364509,-1.644533,
        0.628319,2.370465,2.466824,
        0.628319,2.386110,-1.481351,
        0.628319,2.394960,0.862133,
        0.628319,2.395716,1.977184,
        0.628319,2.396339,1.200900,
        0.628319,2.407941,-2.316753,
        0.628319,2.409450,-2.988432,
        0.628319,2.417202,-0.103363,
        0.628319,2.417386,-0.443097,
        0.628319,2.417950,2.666662,
        0.628319,2.423190,1.789783,
        0.628319,2.423692,1.392696,
        0.628319,2.431122,3.086379,
        0.628319,2.431406,-2.647820,
        0.628319,2.431584,2.307631,
        0.628319,2.434090,1.592818,
        0.628319,2.435993,0.683844,
        0.628319,2.440898,-2.133300,
        0.628319,2.442222,-1.195837,
        0.628319,2.443112,2.883761,
        0.628319,2.449191,-0.633651,
        0.628319,2.450821,0.084741,
        0.628319,2.453499,-1.008861,
        0.628319,2.460010,-1.938011,
        0.628319,2.460199,-0.822008,
        0.628319,2.460902,0.490079,
        0.628319,2.461774,-1.734913,
        0.628319,2.463986,1.034407,
        0.628319,2.467647,0.287667,
        0.628319,2.479849,2.126925,
        0.628319,2.485907,-2.835471,
        0.628319,2.485945,-1.374853,
        0.628319,2.487747,2.505912,
        0.628319,2.487957,-0.274279,
        0.628319,2.489741,-2.468428,
        0.628319,2.512538,1.239900,
        0.628319,2.517528,1.923289,
        0.628319,2.518454,-1.565852,
        0.628319,2.519348,0.843503,
        0.628319,2.522766,-3.059231,
        0.628319,2.528686,2.737228,
        0.628319,2.534798,-2.266671,
        0.628319,2.536533,1.466274,
        0.628319,2.538504,1.699812,
        0.628319,2.539516,-0.476942,
        0.628319,2.539532,-0.074548,
        0.628319,2.543197,2.990252,
        0.628319,2.554022,2.320619,
        0.628319,2.556919,-2.655005,
        0.628319,2.558146,0.627191,
        0.628319,2.558223,-1.164189,
        0.628319,2.563452,-2.044028,
        0.628319,2.568462,0.150655,
        0.628319,2.569130,-0.700837,
        0.628319,2.570050,-1.804492,
        0.628319,2.572710,-0.934665,
        0.628319,2.575994,0.391718,
        0.628319,2.584774,1.046703,
        0.628319,2.601432,2.102535,
        0.628319,2.603559,2.555114,
        0.628319,2.604150,-2.885208,
        0.628319,2.608630,-1.378011,
        0.628319,2.611256,-0.279110,
        0.628319,2.617344,-2.439480,
        0.628319,2.627402,1.293552,
        0.628319,2.635396,1.848850,
        0.628319,2.635686,2.834950,
        0.628319,2.637087,3.132730,
        0.628319,2.642029,0.813274,
        0.628319,2.645260,1.570203,
        0.628319,2.646182,-1.614036,
        0.628319,2.660113,-2.186513,
        0.628319,2.660483,-0.034608,
        0.628319,2.661713,-0.529511,
        0.628319,2.672720,-1.110422,
        0.628319,2.676626,2.335353,
        0.628319,2.676638,0.540364,
        0.628319,2.676780,-1.901158,
        0.628319,2.681411,-2.667965,
        0.628319,2.682152,0.245751,
        0.628319,2.682859,-0.817071,
        0.628319,2.705107,1.062142,
        0.628319,2.717705,2.625995,
        0.628319,2.721001,2.060097,
        0.628319,2.722319,-2.954069,
        0.628319,2.729840,-1.368237,
        0.628319,2.734349,-0.289298,
        0.628319,2.738409,2.975671,
        0.628319,2.740366,1.374341,
        0.628319,2.744633,-2.391311,
        0.628319,2.747398,1.727166,
        0.628319,2.763799,0.760372,
        0.628319,2.768238,-1.679243,
        0.628319,2.779245,-2.046268,
        0.628319,2.780208,-0.626453,
        0.628319,2.780471,0.027732,
        0.628319,2.784696,-1.013128,
        0.628319,2.789817,0.395431,
        0.628319,2.801154,2.353143,
        0.628319,2.807091,-2.687477,
        0.628319,2.825762,1.085035,
        0.628319,2.830209,2.734943,
        0.628319,2.833517,-3.085508,
        0.628319,2.838301,1.980061,
        0.628319,2.849871,1.512012,
        0.628319,2.850089,-1.343899,
        0.628319,2.856905,-0.318441,
        0.628319,2.870445,-2.291001,
        0.628319,2.885017,0.655562,
        0.628319,2.886428,-1.793217,
        0.628319,2.890337,-0.825633,
        0.628319,2.898932,0.139973,
        0.628319,2.931927,-2.728768,
        0.628319,2.932096,2.377790,
        0.628319,2.937409,2.947245,
        0.628319,2.948048,1.133127,
        0.628319,2.950896,1.797889,
        0.628319,2.970019,-1.282382,
        0.628319,2.978333,-0.412090,
        0.628319,2.996437,-2.089121,
        0.628319,3.011274,0.434110,
        0.628319,3.050410,3.138492,
        0.628319,3.064492,1.647259,
        0.628319,3.088868,-0.989501,
        0.733038,0.053884,-2.408394,
        0.733038,0.077645,1.453433,
        0.733038,0.088408,-0.311160,
        0.733038,0.142314,2.590387,
        0.733038,0.147206,-1.385164,
        0.733038,0.163506,0.573203,
        0.733038,0.165485,-2.875396,
        0.733038,0.182564,-2.125821,
        0.733038,0.190569,1.929141,
        0.733038,0.201102,1.249082,
        0.733038,0.201397,-0.716646,
        0.733038,0.210568,-0.069559,
        0.733038,0.251672,2.848900,
        0.733038,0.266262,-1.678068,
        0.733038,0.269108,-1.215563,
        0.733038,0.273777,-2.506756,
        0.733038,0.279854,0.814581,
        0.733038,0.281664,-2.969147,
        0.733038,0.283625,2.405603,
        0.733038,0.284058,0.357350,
        0.733038,0.294107,1.593880,
        0.733038,0.302491,-0.400410,
        0.733038,0.308059,2.021491,
        0.733038,0.312219,-2.072057,
        0.733038,0.321905,-0.829121,
        0.733038,0.329984,1.201621,
        0.733038,0.339714,-0.013712,
        0.733038,0.368784,2.974974,
        0.733038,0.376428,-1.495532,
        0.733038,0.378103,2.646436,
        0.733038,0.385088,0.590355,
        0.733038,0.386248,-2.690943,
        0.733038,0.389851,-2.340031,
        0.733038,0.391502,-1.171374,
        0.733038,0.398249,-1.812694,
        0.733038,0.400542,-2.995729,
        0.733038,0.404273,0.918163,
        0.733038,0.405617,1.773533,
        0.733038,0.409009,-0.570243,
        0.733038,0.410700,0.270433,
        0.733038,0.415893,1.464890,
        0.733038,0.420680,2.088164,
        0.733038,0.424766,-0.258729,
        0.733038,0.426507,2.367872,
        0.733038,0.441292,-0.890727,
        0.733038,0.451839,-2.078193,
        0.733038,0.459270,1.191562,
        0.733038,0.470622,0.014002,
        0.733038,0.481292,-2.529098,
        0.733038,0.482460,2.849843,
        0.733038,0.492205,-1.397963,
        0.733038,0.493314,3.117264,
        0.733038,0.497870,0.727828,
        0.733038,0.499783,-1.654379,
        0.733038,0.502651,0.463333,
        0.733038,0.505463,2.594557,
        0.733038,0.506104,-2.874910,
        0.733038,0.513579,-1.147354,
        0.733038,0.518569,1.644237,
        0.733038,0.519001,1.900488,
        0.733038,0.519277,-0.673809,
        0.733038,0.524769,-0.421955,
        0.733038,0.525564,-2.288838,
        0.733038,0.532541,0.978737,
        0.733038,0.534610,2.154808,
        0.733038,0.535020,-1.891846,
        0.733038,0.540208,1.400497,
        0.733038,0.542006,0.228337,
        0.733038,0.550762,-0.180771,
        0.733038,0.560531,2.378474,
        0.733038,0.561629,-0.924576,
        0.733038,0.567465,-2.679392,
        0.733038,0.588527,1.181755,
        0.733038,0.591643,-2.098364,
        0.733038,0.594963,2.997575,
        0.733038,0.599643,-3.045235,
        0.733038,0.599962,2.773556,
        0.733038,0.602620,0.029270,
        0.733038,0.605711,0.598454,
        0.733038,0.607708,-1.541026,
        0.733038,0.609360,-2.464379,
        0.733038,0.610738,-1.326786,
        0.733038,0.616968,0.820489,
        0.733038,0.626379,1.773610,
        0.733038,0.628626,-1.751180,
        0.733038,0.630671,0.391949,
        0.733038,0.632622,-0.737642,
        0.733038,0.632708,2.566722,
        0.733038,0.632797,1.989858,
        0.733038,0.634455,-0.534790,
        0.733038,0.636205,1.561170,
        0.733038,0.637681,-1.123423,
        0.733038,0.641441,-2.842109,
        0.733038,0.644807,-0.327139,
        0.733038,0.652440,2.197129,
        0.733038,0.660897,-2.278593,
        0.733038,0.664008,1.360091,
        0.733038,0.670850,-1.945257,
        0.733038,0.671396,1.004859,
        0.733038,0.673985,0.203855,
        0.733038,0.680326,-0.133370,
        0.733038,0.685083,-0.942363,
        0.733038,0.688204,-2.632204,
        0.733038,0.689777,2.385011,
        0.733038,0.700065,3.115430,
        0.733038,0.704080,2.911636,
        0.733038,0.712883,1.174216,
        0.733038,0.715701,0.699906,
        0.733038,0.718851,-1.442215,
        0.733038,0.721019,2.721536,
        0.733038,0.725300,-2.975134,
        0.733038,0.727136,-1.633288,
        0.733038,0.727399,-2.120567,
        0.733038,0.730751,0.522488,
        0.733038,0.732892,-1.254330,
        0.733038,0.734333,-2.446184,
        0.733038,0.735125,0.037270,
        0.733038,0.737893,1.866463,
        0.733038,0.740650,1.681075,
        0.733038,0.745971,-0.781286,
        0.733038,0.749432,-0.442559,
        0.733038,0.750102,2.050636,
        0.733038,0.751046,0.863821,
        0.733038,0.757018,-0.608833,
        0.733038,0.757046,-1.816019,
        0.733038,0.757389,1.501184,
        0.733038,0.759800,2.546181,
        0.733038,0.759859,0.351923,
        0.733038,0.761141,-2.781102,
        0.733038,0.769035,-1.078574,
        0.733038,0.770166,-0.266604,
        0.733038,0.772218,2.227380,
        0.733038,0.788382,1.331051,
        0.733038,0.790269,-2.282409,
        0.733038,0.803410,-1.984567,
        0.733038,0.803813,0.191520,
        0.733038,0.806832,3.022155,
        0.733038,0.807476,-2.597675,
        0.733038,0.812512,-0.107335,
        0.733038,0.814161,1.010949,
        0.733038,0.817732,-3.084486,
        0.733038,0.817846,2.390688,
        0.733038,0.819323,2.845960,
        0.733038,0.819430,-0.915017,
        0.733038,0.829198,-1.532491,
        0.733038,0.829531,0.665695,
        0.733038,0.829816,1.167147,
        0.733038,0.833613,-1.357492,
        0.733038,0.842945,-2.907318,
        0.733038,0.844883,2.679708,
        0.733038,0.848189,-1.705003,
        0.733038,0.849888,1.767901,
        0.733038,0.851126,0.493370,
        0.733038,0.852911,1.933394,
        0.733038,0.857516,-2.432863,
        0.733038,0.858261,-1.191729,
        0.733038,0.858279,-2.138895,
        0.733038,0.859723,-0.734928,
        0.733038,0.859754,1.603292,
        0.733038,0.861655,-0.394821,
        0.733038,0.864300,0.043549,
        0.733038,0.866940,0.851229,
        0.733038,0.869748,2.095969,
        0.733038,0.877352,-0.574668,
        0.733038,0.879254,-2.733338,
        0.733038,0.882554,1.443757,
        0.733038,0.884060,0.334199,
        0.733038,0.885626,-1.866061,
        0.733038,0.889834,2.529957,
        0.733038,0.893140,2.251698,
        0.733038,0.897463,-0.240822,
        0.733038,0.898758,-1.031948,
        0.733038,0.913918,1.287618,
        0.733038,0.916555,-2.287173,
        0.733038,0.918305,3.107845,
        0.733038,0.919522,2.948277,
        0.733038,0.927138,-2.569686,
        0.733038,0.929031,0.185383,
        0.733038,0.933766,-3.015917,
        0.733038,0.934873,-2.010822,
        0.733038,0.935636,-1.445855,
        0.733038,0.936702,-0.866461,
        0.733038,0.937395,0.729108,
        0.733038,0.937910,2.787575,
        0.733038,0.939144,-0.092962,
        0.733038,0.941718,1.125085,
        0.733038,0.942873,-1.607680,
        0.733038,0.943701,0.980328,
        0.733038,0.947774,2.395256,
        0.733038,0.952433,0.584347,
        0.733038,0.954456,-1.287293,
        0.733038,0.958731,-2.852552,
        0.733038,0.961838,1.839284,
        0.733038,0.961898,-0.469425,
        0.733038,0.965426,1.686615,
        0.733038,0.970108,1.990771,
        0.733038,0.972820,-1.762780,
        0.733038,0.973500,-0.711881,
        0.733038,0.975253,2.641082,
        0.733038,0.977824,0.441928,
        0.733038,0.981485,-2.424766,
        0.733038,0.981877,1.537100,
        0.733038,0.982609,-1.136920,
        0.733038,0.984265,-2.156073,
        0.733038,0.989150,0.045493,
        0.733038,0.993052,2.137478,
        0.733038,0.997766,-2.693871,
        0.733038,0.999847,-0.334738,
        0.733038,1.001920,0.848054,
        0.733038,1.006146,1.389734,
        0.733038,1.012180,0.304066,
        0.733038,1.016938,2.270674,
        0.733038,1.017371,-0.979790,
        0.733038,1.019495,-1.894487,
        0.733038,1.023796,-0.583691,
        0.733038,1.026525,2.506945,
        0.733038,1.027004,-0.203653,
        0.733038,1.028027,3.036573,
        0.733038,1.031435,1.241021,
        0.733038,1.032316,2.884221,
        0.733038,1.035863,-3.110475,
        0.733038,1.042085,-2.292950,
        0.733038,1.043272,-1.522986,
        0.733038,1.046265,-2.958847,
        0.733038,1.049272,-1.376344,
        0.733038,1.050039,-2.552086,
        0.733038,1.053682,0.689863,
        0.733038,1.055458,0.170394,
        0.733038,1.056622,-0.831866,
        0.733038,1.057515,1.097998,
        0.733038,1.058706,-2.036438,
        0.733038,1.061389,-1.672248,
        0.733038,1.063758,2.743093,
        0.733038,1.070157,-0.073422,
        0.733038,1.072154,1.761506,
        0.733038,1.074029,1.909193,
        0.733038,1.074204,0.545276,
        0.733038,1.074714,-2.805147,
        0.733038,1.076262,0.966104,
        0.733038,1.077642,-1.227073,
        0.733038,1.084151,1.618525,
        0.733038,1.085085,2.377673,
        0.733038,1.086822,-0.441112,
        0.733038,1.087142,2.058815,
        0.733038,1.098383,-0.696430,
        0.733038,1.099138,-1.084033,
        0.733038,1.103208,1.477773,
        0.733038,1.103783,0.408254,
        0.733038,1.106275,-2.420800,
        0.733038,1.108079,2.612943,
        0.733038,1.109345,-2.172718,
        0.733038,1.115957,-1.787597,
        0.733038,1.118930,0.045983,
        0.733038,1.119777,0.834702,
        0.733038,1.121943,-2.669042,
        0.733038,1.124930,2.219917,
        0.733038,1.125217,1.336294,
        0.733038,1.130958,2.972931,
        0.733038,1.131400,-0.306629,
        0.733038,1.137135,-0.936514,
        0.733038,1.139125,-1.924528,
        0.733038,1.139877,0.276382,
        0.733038,1.142003,-3.058022,
        0.733038,1.147363,-0.558448,
        0.733038,1.150348,-0.176418,
        0.733038,1.151198,1.197016,
        0.733038,1.153930,-1.597104,
        0.733038,1.155389,2.837148,
        0.733038,1.155458,-1.460623,
        0.733038,1.157331,-2.906362,
        0.733038,1.158594,2.487618,
        0.733038,1.162103,-1.322553,
        0.733038,1.165407,3.092861,
        0.733038,1.168025,0.717075,
        0.733038,1.168406,-2.301519,
        0.733038,1.174782,-2.538755,
        0.733038,1.177998,1.061256,
        0.733038,1.178898,1.827305,
        0.733038,1.181331,-2.060330,
        0.733038,1.182540,-0.798952,
        0.733038,1.184569,0.147517,
        0.733038,1.184734,1.968919,
        0.733038,1.184804,0.593386,
        0.733038,1.188278,1.690808,
        0.733038,1.194360,2.110440,
        0.733038,1.194776,2.710455,
        0.733038,1.197745,-1.173705,
        0.733038,1.199898,-2.777257,
        0.733038,1.201164,2.355931,
        0.733038,1.203760,1.556404,
        0.733038,1.205845,-0.415993,
        0.733038,1.206585,0.933648,
        0.733038,1.206950,-0.037248,
        0.733038,1.209341,0.471162,
        0.733038,1.213483,-1.039417,
        0.733038,1.218146,-0.672643,
        0.733038,1.221608,1.421669,
        0.733038,1.226043,-1.693533,
        0.733038,1.232870,-1.822773,
        0.733038,1.234375,-2.421194,
        0.733038,1.234531,-2.190148,
        0.733038,1.239414,0.348959,
        0.733038,1.240183,2.587255,
        0.733038,1.242312,2.941234,
        0.733038,1.244075,1.286644,
        0.733038,1.245961,-2.997461,
        0.733038,1.246495,-2.646872,
        0.733038,1.247736,2.242810,
        0.733038,1.253604,0.812222,
        0.733038,1.257816,-0.894317,
        0.733038,1.258247,-1.953634,
        0.733038,1.258993,-0.284099,
        0.733038,1.262380,-0.157609,
        0.733038,1.262791,-1.549229,
        0.733038,1.263772,-0.533073,
        0.733038,1.266731,-1.412461,
        0.733038,1.268591,-3.122146,
        0.733038,1.269614,-1.275246,
        0.733038,1.270745,1.152533,
        0.733038,1.273606,0.224545,
        0.733038,1.279716,2.812101,
        0.733038,1.283786,2.462124,
        0.733038,1.284798,0.078209,
        0.733038,1.285402,-2.882049,
        0.733038,1.286936,1.886053,
        0.733038,1.289408,0.683122,
        0.733038,1.294736,1.756184,
        0.733038,1.296978,-2.311511,
        0.733038,1.298546,2.020601,
        0.733038,1.300924,1.022478,
        0.733038,1.303305,3.043152,
        0.733038,1.304210,-2.084225,
        0.733038,1.305241,-2.530249,
        0.733038,1.307513,1.628003,
        0.733038,1.310722,0.558267,
        0.733038,1.314142,2.139329,
        0.733038,1.316087,-0.777971,
        0.733038,1.317907,-1.121023,
        0.733038,1.321019,-2.749776,
        0.733038,1.321609,2.687717,
        0.733038,1.322568,1.499415,
        0.733038,1.322972,-0.395699,
        0.733038,1.325286,-0.994192,
        0.733038,1.329146,-0.648513,
        0.733038,1.329795,-0.042570,
        0.733038,1.331651,2.344497,
        0.733038,1.336522,0.435098,
        0.733038,1.339410,0.900087,
        0.733038,1.339550,1.370353,
        0.733038,1.340441,-1.731580,
        0.733038,1.351855,-1.855234,
        0.733038,1.360161,-3.018404,
        0.733038,1.361031,-1.613971,
        0.733038,1.361196,2.914780,
        0.733038,1.361998,-2.207254,
        0.733038,1.362302,1.239384,
        0.733038,1.365228,2.563808,
        0.733038,1.365452,0.310522,
        0.733038,1.367028,-2.420041,
        0.733038,1.373422,-1.484281,
        0.733038,1.374245,-1.226417,
        0.733038,1.375020,-0.170393,
        0.733038,1.376560,-1.355481,
        0.733038,1.377070,-1.980984,
        0.733038,1.377578,-2.636339,
        0.733038,1.378096,0.774626,
        0.733038,1.379086,-0.508192,
        0.733038,1.382562,0.177141,
        0.733038,1.388849,2.233847,
        0.733038,1.389268,-0.886349,
        0.733038,1.391324,1.109396,
        0.733038,1.391699,-0.288725,
        0.733038,1.392869,3.136903,
        0.733038,1.393612,1.952016,
        0.733038,1.395545,-2.850807,
        0.733038,1.401926,1.822930,
        0.733038,1.402149,2.788898,
        0.733038,1.409552,0.647980,
        0.733038,1.410840,0.056219,
        0.733038,1.411944,2.444377,
        0.733038,1.413639,1.697982,
        0.733038,1.426479,1.573427,
        0.733038,1.426725,2.102242,
        0.733038,1.427868,0.984637,
        0.733038,1.429580,-2.107232,
        0.733038,1.431375,-2.313269,
        0.733038,1.431907,3.015410,
        0.733038,1.432325,-1.029054,
        0.733038,1.433341,0.524268,
        0.733038,1.435181,-0.739771,
        0.733038,1.437208,-2.529085,
        0.733038,1.439041,-0.617966,
        0.733038,1.441473,1.448544,
        0.733038,1.444527,2.664744,
        0.733038,1.450513,-0.400625,
        0.733038,1.451907,-0.066823,
        0.733038,1.452754,-2.738631,
        0.733038,1.454061,-2.950416,
        0.733038,1.456736,1.323154,
        0.733038,1.458370,0.400433,
        0.733038,1.459314,-1.765003,
        0.733038,1.459902,-1.145835,
        0.733038,1.464952,2.329245,
        0.733038,1.466050,0.862104,
        0.733038,1.471330,-1.886160,
        0.733038,1.475119,-1.648255,
        0.733038,1.480878,1.194576,
        0.733038,1.480942,0.273235,
        0.733038,1.481086,2.890269,
        0.733038,1.484118,-1.529156,
        0.733038,1.484741,-1.279219,
        0.733038,1.487681,-1.406069,
        0.733038,1.487886,-3.066037,
        0.733038,1.490555,2.543284,
        0.733038,1.492424,-0.190932,
        0.733038,1.495002,-0.839810,
        0.733038,1.495297,2.007188,
        0.733038,1.496414,-2.009267,
        0.733038,1.498261,-2.425270,
        0.733038,1.498907,-2.207327,
        0.733038,1.500204,0.737638,
        0.733038,1.502469,0.151156,
        0.733038,1.507448,-0.514782,
        0.733038,1.509287,2.204647,
        0.733038,1.510518,1.884920,
        0.733038,1.511948,-2.632360,
        0.733038,1.515235,1.068565,
        0.733038,1.522237,3.102018,
        0.733038,1.522373,1.764991,
        0.733038,1.522416,-0.954956,
        0.733038,1.523102,2.765920,
        0.733038,1.524811,-2.844718,
        0.733038,1.527261,-0.304916,
        0.733038,1.529788,0.613208,
        0.733038,1.533648,1.644808,
        0.733038,1.535018,0.028869,
        0.733038,1.541082,2.424608,
        0.733038,1.546010,1.523462,
        0.733038,1.551135,-0.642994,
        0.733038,1.553895,0.489607,
        0.733038,1.555041,0.947813,
        0.733038,1.559671,-1.074934,
        0.733038,1.560345,2.985899,
        0.733038,1.560487,-2.323400,
        0.733038,1.561127,1.401111,
        0.733038,1.567193,2.643172,
        0.733038,1.568892,2.098362,
        0.733038,1.571702,-2.098435,
        0.733038,1.572701,-2.528730,
        0.733038,1.573572,1.278593,
        0.733038,1.574849,-0.091505,
        0.733038,1.576334,0.365236,
        0.733038,1.577029,-2.962888,
        0.733038,1.577688,-1.814941,
        0.733038,1.579542,-0.755917,
        0.733038,1.579624,-1.199759,
        0.733038,1.581515,-0.419414,
        0.733038,1.584581,-2.738702,
        0.733038,1.586948,-1.696296,
        0.733038,1.588901,2.301496,
        0.733038,1.589936,0.825020,
        0.733038,1.590018,-1.941652,
        0.733038,1.596968,-1.328358,
        0.733038,1.597178,-1.575607,
        0.733038,1.600470,-1.453196,
        0.733038,1.601408,0.242409,
        0.733038,1.602259,1.152656,
        0.733038,1.603064,2.865774,
        0.733038,1.612085,1.968802,
        0.733038,1.612720,-3.080498,
        0.733038,1.616061,2.521531,
        0.733038,1.616900,-0.207457,
        0.733038,1.620988,-0.875920,
        0.733038,1.622114,0.701509,
        0.733038,1.622749,0.121426,
        0.733038,1.622993,-2.221629,
        0.733038,1.629346,-0.542669,
        0.733038,1.630007,1.844552,
        0.733038,1.634747,-2.426198,
        0.733038,1.642346,1.722617,
        0.733038,1.643925,1.034629,
        0.733038,1.644848,2.743774,
        0.733038,1.645258,2.188109,
        0.733038,1.645417,-2.634725,
        0.733038,1.647395,3.085437,
        0.733038,1.648538,-2.855044,
        0.733038,1.649932,0.576808,
        0.733038,1.652224,-0.996105,
        0.733038,1.652871,1.600681,
        0.733038,1.657997,-0.325127,
        0.733038,1.661936,0.001037,
        0.733038,1.665187,1.477402,
        0.733038,1.665496,2.397872,
        0.733038,1.673267,-0.665453,
        0.733038,1.673557,0.452526,
        0.733038,1.679277,-1.122421,
        0.733038,1.679696,0.912482,
        0.733038,1.682420,1.353922,
        0.733038,1.685031,-2.009038,
        0.733038,1.685196,2.967046,
        0.733038,1.690107,2.620910,
        0.733038,1.690453,-2.123076,
        0.733038,1.691693,-1.871572,
        0.733038,1.691817,2.061688,
        0.733038,1.692644,1.237670,
        0.733038,1.696386,0.328123,
        0.733038,1.697162,-1.249026,
        0.733038,1.697488,-2.322785,
        0.733038,1.698726,-2.973576,
        0.733038,1.699172,-1.749717,
        0.733038,1.704478,-0.106295,
        0.733038,1.707244,-0.444770,
        0.733038,1.707500,-2.531497,
        0.733038,1.707714,-0.787321,
        0.733038,1.708430,-1.627552,
        0.733038,1.710583,-1.376890,
        0.733038,1.712415,-2.747921,
        0.733038,1.713125,-1.503073,
        0.733038,1.713452,0.789419,
        0.733038,1.720612,2.279888,
        0.733038,1.724852,0.202831,
        0.733038,1.724868,2.846246,
        0.733038,1.728940,1.931535,
        0.733038,1.732838,1.127584,
        0.733038,1.735719,-3.092485,
        0.733038,1.739676,2.496017,
        0.733038,1.739971,-0.228876,
        0.733038,1.743328,0.663347,
        0.733038,1.744927,-0.911892,
        0.733038,1.745952,0.084733,
        0.733038,1.748077,1.804354,
        0.733038,1.752959,-0.567548,
        0.733038,1.760169,1.678132,
        0.733038,1.763887,-2.218450,
        0.733038,1.766737,2.722752,
        0.733038,1.767999,1.004737,
        0.733038,1.769149,1.550298,
        0.733038,1.770119,-2.427175,
        0.733038,1.770460,2.154208,
        0.733038,1.770514,3.070982,
        0.733038,1.770735,0.537245,
        0.733038,1.772673,-2.864760,
        0.733038,1.773527,-1.039582,
        0.733038,1.776419,-2.642421,
        0.733038,1.780769,1.416330,
        0.733038,1.786634,-0.346247,
        0.733038,1.788829,-1.947553,
        0.733038,1.793737,0.410270,
        0.733038,1.794553,2.374618,
        0.733038,1.794699,-0.694061,
        0.733038,1.797469,-1.169105,
        0.733038,1.803705,0.881076,
        0.733038,1.804124,1.264249,
        0.733038,1.807825,-2.090455,
        0.733038,1.808822,-1.817375,
        0.733038,1.809020,2.949741,
        0.733038,1.812466,2.597595,
        0.733038,1.813159,2.025395,
        0.733038,1.813816,-1.298435,
        0.733038,1.815016,0.278951,
        0.733038,1.816510,-0.008631,
        0.733038,1.817772,-1.689625,
        0.733038,1.821360,-2.984352,
        0.733038,1.822312,-0.127211,
        0.733038,1.823386,-1.429807,
        0.733038,1.823834,-1.560483,
        0.733038,1.832060,-0.821627,
        0.733038,1.832690,-0.468440,
        0.733038,1.833945,-2.320278,
        0.733038,1.834787,0.751968,
        0.733038,1.840579,-2.536445,
        0.733038,1.841033,-2.756932,
        0.733038,1.846350,1.893983,
        0.733038,1.846644,1.113766,
        0.733038,1.846833,2.826543,
        0.733038,1.847325,2.249051,
        0.733038,1.854119,0.134161,
        0.733038,1.858863,-3.106022,
        0.733038,1.865046,1.763384,
        0.733038,1.865619,-0.952036,
        0.733038,1.865912,0.624453,
        0.733038,1.866694,2.473686,
        0.733038,1.867524,-0.240083,
        0.733038,1.875561,1.632831,
        0.733038,1.876049,-0.594378,
        0.733038,1.880509,1.500204,
        0.733038,1.883504,1.360457,
        0.733038,1.886971,-2.196665,
        0.733038,1.888979,0.982922,
        0.733038,1.889815,2.701570,
        0.733038,1.892144,-2.005340,
        0.733038,1.892348,0.496522,
        0.733038,1.893281,-1.083976,
        0.733038,1.893828,3.053453,
        0.733038,1.895117,2.121107,
        0.733038,1.898741,-2.875735,
        0.733038,1.904429,-2.428029,
        0.733038,1.908408,-2.648903,
        0.733038,1.911828,-0.364489,
        0.733038,1.912224,0.366565,
        0.733038,1.914854,1.214468,
        0.733038,1.915467,-1.883143,
        0.733038,1.915763,-0.724679,
        0.733038,1.915790,-1.216366,
        0.733038,1.922256,2.347806,
        0.733038,1.922616,0.844633,
        0.733038,1.925185,0.231170,
        0.733038,1.927474,-1.754616,
        0.733038,1.928342,-1.351174,
        0.733038,1.933328,-1.623328,
        0.733038,1.933878,0.034150,
        0.733038,1.934585,-1.487979,
        0.733038,1.934802,2.927634,
        0.733038,1.934862,1.989776,
        0.733038,1.936327,-0.099350,
        0.733038,1.938088,2.578507,
        0.733038,1.945677,-2.999805,
        0.733038,1.954572,-0.859208,
        0.733038,1.957868,-0.490784,
        0.733038,1.958989,0.712772,
        0.733038,1.960212,-2.307561,
        0.733038,1.964774,1.855507,
        0.733038,1.964971,-2.100533,
        0.733038,1.969112,1.088273,
        0.733038,1.971438,-2.765415,
        0.733038,1.972312,2.805134,
        0.733038,1.974950,2.220596,
        0.733038,1.975135,-2.539127,
        0.733038,1.982107,1.720829,
        0.733038,1.983195,-3.127984,
        0.733038,1.983318,-0.247182,
        0.733038,1.986720,-0.993220,
        0.733038,1.990381,1.586005,
        0.733038,1.990470,0.580283,
        0.733038,1.990523,1.450326,
        0.733038,1.993909,1.312897,
        0.733038,1.994694,2.452813,
        0.733038,1.999226,-0.620717,
        0.733038,2.005145,0.941012,
        0.733038,2.007301,0.142337,
        0.733038,2.014149,0.444138,
        0.733038,2.014338,-1.127392,
        0.733038,2.018131,3.021778,
        0.733038,2.019676,-1.980944,
        0.733038,2.019912,2.686612,
        0.733038,2.020545,2.088672,
        0.733038,2.026074,0.300945,
        0.733038,2.026196,-2.889057,
        0.733038,2.031959,-1.268178,
        0.733038,2.032917,-2.199917,
        0.733038,2.033572,-1.842884,
        0.733038,2.034712,-2.419505,
        0.733038,2.035818,-0.379479,
        0.733038,2.036004,-0.757062,
        0.733038,2.040501,-1.411488,
        0.733038,2.042874,-1.701367,
        0.733038,2.043060,-2.653746,
        0.733038,2.043729,-1.557938,
        0.733038,2.048538,0.806623,
        0.733038,2.049006,1.199930,
        0.733038,2.049502,-0.137989,
        0.733038,2.052181,2.326959,
        0.733038,2.052794,-0.008526,
        0.733038,2.057608,1.951726,
        0.733038,2.065568,2.567730,
        0.733038,2.070371,2.879808,
        0.733038,2.071256,-3.018961,
        0.733038,2.079746,-0.897699,
        0.733038,2.082784,1.042403,
        0.733038,2.083752,-0.511086,
        0.733038,2.084401,1.810502,
        0.733038,2.085731,0.671081,
        0.733038,2.097496,1.373172,
        0.733038,2.099185,1.667006,
        0.733038,2.100482,-2.301084,
        0.733038,2.103118,2.194745,
        0.733038,2.103223,1.521685,
        0.733038,2.103695,-2.777707,
        0.733038,2.107523,3.129581,
        0.733038,2.108277,-2.095907,
        0.733038,2.108891,-2.532166,
        0.733038,2.111003,-0.268632,
        0.733038,2.112276,-1.027246,
        0.733038,2.114380,0.531457,
        0.733038,2.119166,0.100552,
        0.733038,2.119564,2.754818,
        0.733038,2.122474,-0.648314,
        0.733038,2.125183,2.445598,
        0.733038,2.129501,0.238290,
        0.733038,2.131178,0.386663,
        0.733038,2.131223,-1.174311,
        0.733038,2.134269,-1.948042,
        0.733038,2.134684,0.907065,
        0.733038,2.137615,2.988806,
        0.733038,2.145635,-1.327708,
        0.733038,2.146467,2.055081,
        0.733038,2.146582,-1.793788,
        0.733038,2.149210,-1.482321,
        0.733038,2.152901,-1.636788,
        0.733038,2.156337,1.151558,
        0.733038,2.156474,-2.912065,
        0.733038,2.157194,-0.792563,
        0.733038,2.166724,-0.401638,
        0.733038,2.166980,-0.134879,
        0.733038,2.173644,-2.403429,
        0.733038,2.175711,-2.655523,
        0.733038,2.177132,2.600419,
        0.733038,2.178879,0.767722,
        0.733038,2.180752,1.908485,
        0.733038,2.181338,2.313721,
        0.733038,2.185580,1.287673,
        0.733038,2.195400,-2.217546,
        0.733038,2.200418,-3.056442,
        0.733038,2.201829,-0.001735,
        0.733038,2.203804,1.756176,
        0.733038,2.211602,1.443930,
        0.733038,2.212686,-0.544809,
        0.733038,2.213937,0.621892,
        0.733038,2.214075,1.600691,
        0.733038,2.216256,1.020029,
        0.733038,2.219060,2.886631,
        0.733038,2.223375,-0.906185,
        0.733038,2.226002,-1.060248,
        0.733038,2.226413,-2.064034,
        0.733038,2.230554,2.171334,
        0.733038,2.232413,2.735777,
        0.733038,2.233709,3.070830,
        0.733038,2.234383,-2.792065,
        0.733038,2.237183,0.468799,
        0.733038,2.237695,-0.271388,
        0.733038,2.239419,0.151693,
        0.733038,2.243347,-2.521049,
        0.733038,2.244035,-1.229456,
        0.733038,2.244355,0.309899,
        0.733038,2.245175,2.460236,
        0.733038,2.245628,-1.899216,
        0.733038,2.245973,-0.702234,
        0.733038,2.255601,-1.728131,
        0.733038,2.256702,-1.401367,
        0.733038,2.261466,-1.557500,
        0.733038,2.267390,0.878499,
        0.733038,2.272099,2.018594,
        0.733038,2.278618,1.170386,
        0.733038,2.283304,-2.345584,
        0.733038,2.285175,-2.943090,
        0.733038,2.293000,-0.124451,
        0.733038,2.294058,-0.421070,
        0.733038,2.303283,1.855938,
        0.733038,2.304968,2.608978,
        0.733038,2.305347,2.311346,
        0.733038,2.307188,-2.656134,
        0.733038,2.307301,1.336200,
        0.733038,2.309655,0.726020,
        0.733038,2.317634,-2.185225,
        0.733038,2.321510,1.685664,
        0.733038,2.324376,-3.111006,
        0.733038,2.324731,1.511186,
        0.733038,2.329587,0.035498,
        0.733038,2.330158,2.977264,
        0.733038,2.335096,-0.588468,
        0.733038,2.336046,-1.115799,
        0.733038,2.338060,2.803613,
        0.733038,2.338083,-0.943400,
        0.733038,2.340210,0.562145,
        0.733038,2.341931,1.023926,
        0.733038,2.343265,-0.779061,
        0.733038,2.343434,-2.013692,
        0.733038,2.353029,-1.310529,
        0.733038,2.354868,0.211543,
        0.733038,2.355120,0.388593,
        0.733038,2.355165,-1.833173,
        0.733038,2.355462,2.151441,
        0.733038,2.362948,-2.487477,
        0.733038,2.363101,-2.810456,
        0.733038,2.363394,-0.272918,
        0.733038,2.364509,-1.644533,
        0.733038,2.370465,2.466824,
        0.733038,2.386110,-1.481351,
        0.733038,2.394960,0.862133,
        0.733038,2.395716,1.977184,
        0.733038,2.396339,1.200900,
        0.733038,2.407941,-2.316753,
        0.733038,2.409450,-2.988432,
        0.733038,2.417202,-0.103363,
        0.733038,2.417386,-0.443097,
        0.733038,2.417950,2.666662,
        0.733038,2.423190,1.789783,
        0.733038,2.423692,1.392696,
        0.733038,2.431122,3.086379,
        0.733038,2.431406,-2.647820,
        0.733038,2.431584,2.307631,
        0.733038,2.434090,1.592818,
        0.733038,2.435993,0.683844,
        0.733038,2.440898,-2.133300,
        0.733038,2.442222,-1.195837,
        0.733038,2.443112,2.883761,
        0.733038,2.449191,-0.633651,
        0.733038,2.450821,0.084741,
        0.733038,2.453499,-1.008861,
        0.733038,2.460010,-1.938011,
        0.733038,2.460199,-0.822008,
        0.733038,2.460902,0.490079,
        0.733038,2.461774,-1.734913,
        0.733038,2.463986,1.034407,
        0.733038,2.467647,0.287667,
        0.733038,2.479849,2.126925,
        0.733038,2.485907,-2.835471,
        0.733038,2.485945,-1.374853,
        0.733038,2.487747,2.505912,
        0.733038,2.487957,-0.274279,
        0.733038,2.489741,-2.468428,
        0.733038,2.512538,1.239900,
        0.733038,2.517528,1.923289,
        0.733038,2.518454,-1.565852,
        0.733038,2.519348,0.843503,
        0.733038,2.522766,-3.059231,
        0.733038,2.528686,2.737228,
        0.733038,2.534798,-2.266671,
        0.733038,2.536533,1.466274,
        0.733038,2.538504,1.699812,
        0.733038,2.539516,-0.476942,
        0.733038,2.539532,-0.074548,
        0.733038,2.543197,2.990252,
        0.733038,2.554022,2.320619,
        0.733038,2.556919,-2.655005,
        0.733038,2.558146,0.627191,
        0.733038,2.558223,-1.164189,
        0.733038,2.563452,-2.044028,
        0.733038,2.568462,0.150655,
        0.733038,2.569130,-0.700837,
        0.733038,2.570050,-1.804492,
        0.733038,2.572710,-0.934665,
        0.733038,2.575994,0.391718,
        0.733038,2.584774,1.046703,
        0.733038,2.601432,2.102535,
        0.733038,2.603559,2.555114,
        0.733038,2.604150,-2.885208,
        0.733038,2.608630,-1.378011,
        0.733038,2.611256,-0.279110,
        0.733038,2.617344,-2.439480,
        0.733038,2.627402,1.293552,
        0.733038,2.635396,1.848850,
        0.733038,2.635686,2.834950,
        0.733038,2.637087,3.132730,
        0.733038,2.642029,0.813274,
        0.733038,2.645260,1.570203,
        0.733038,2.646182,-1.614036,
        0.733038,2.660113,-2.186513,
        0.733038,2.660483,-0.034608,
        0.733038,2.661713,-0.529511,
        0.733038,2.672720,-1.110422,
        0.733038,2.676626,2.335353,
        0.733038,2.676638,0.540364,
        0.733038,2.676780,-1.901158,
        0.733038,2.681411,-2.667965,
        0.733038,2.682152,0.245751,
        0.733038,2.682859,-0.817071,
        0.733038,2.705107,1.062142,
        0.733038,2.717705,2.625995,
        0.733038,2.721001,2.060097,
        0.733038,2.722319,-2.954069,
        0.733038,2.729840,-1.368237,
        0.733038,2.734349,-0.289298,
        0.733038,2.738409,2.975671,
        0.733038,2.740366,1.374341,
        0.733038,2.744633,-2.391311,
        0.733038,2.747398,1.727166,
        0.733038,2.763799,0.760372,
        0.733038,2.768238,-1.679243,
        0.733038,2.779245,-2.046268,
        0.733038,2.780208,-0.626453,
        0.733038,2.780471,0.027732,
        0.733038,2.784696,-1.013128,
        0.733038,2.789817,0.395431,
        0.733038,2.801154,2.353143,
        0.733038,2.807091,-2.687477,
        0.733038,2.825762,1.085035,
        0.733038,2.830209,2.734943,
        0.733038,2.833517,-3.085508,
        0.733038,2.838301,1.980061,
        0.733038,2.849871,1.512012,
        0.733038,2.850089,-1.343899,
        0.733038,2.856905,-0.318441,
        0.733038,2.870445,-2.291001,
        0.733038,2.885017,0.655562,
        0.733038,2.886428,-1.793217,
        0.733038,2.890337,-0.825633,
        0.733038,2.898932,0.139973,
        0.733038,2.931927,-2.728768,
        0.733038,2.932096,2.377790,
        0.733038,2.937409,2.947245,
        0.733038,2.948048,1.133127,
        0.733038,2.950896,1.797889,
        0.733038,2.970019,-1.282382,
        0.733038,2.978333,-0.412090,
        0.733038,2.996437,-2.089121,
        0.733038,3.011274,0.434110,
        0.733038,3.050410,3.138492,
        0.733038,3.064492,1.647259,
        0.733038,3.088868,-0.989501,
        0.837758,0.053884,-2.408394,
        0.837758,0.077645,1.453433,
        0.837758,0.088408,-0.311160,
        0.837758,0.142314,2.590387,
        0.837758,0.147206,-1.385164,
        0.837758,0.163506,0.573203,
        0.837758,0.165485,-2.875396,
        0.837758,0.182564,-2.125821,
        0.837758,0.190569,1.929141,
        0.837758,0.201102,1.249082,
        0.837758,0.201397,-0.716646,
        0.837758,0.210568,-0.069559,
        0.837758,0.251672,2.848900,
        0.837758,0.266262,-1.678068,
        0.837758,0.269108,-1.215563,
        0.837758,0.273777,-2.506756,
        0.837758,0.279854,0.814581,
        0.837758,0.281664,-2.969147,
        0.837758,0.283625,2.405603,
        0.837758,0.284058,0.357350,
        0.837758,0.294107,1.593880,
        0.837758,0.302491,-0.400410,
        0.837758,0.308059,2.021491,
        0.837758,0.312219,-2.072057,
        0.837758,0.321905,-0.829121,
        0.837758,0.329984,1.201621,
        0.837758,0.339714,-0.013712,
        0.837758,0.368784,2.974974,
        0.837758,0.376428,-1.495532,
        0.837758,0.378103,2.646436,
        0.837758,0.385088,0.590355,
        0.837758,0.386248,-2.690943,
        0.837758,0.389851,-2.340031,
        0.837758,0.391502,-1.171374,
        0.837758,0.398249,-1.812694,
        0.837758,0.400542,-2.995729,
        0.837758,0.404273,0.918163,
        0.837758,0.405617,1.773533,
        0.837758,0.409009,-0.570243,
        0.837758,0.410700,0.270433,
        0.837758,0.415893,1.464890,
        0.837758,0.420680,2.088164,
        0.837758,0.424766,-0.258729,
        0.837758,0.426507,2.367872,
        0.837758,0.441292,-0.890727,
        0.837758,0.451839,-2.078193,
        0.837758,0.459270,1.191562,
        0.837758,0.470622,0.014002,
        0.837758,0.481292,-2.529098,
        0.837758,0.482460,2.849843,
        0.837758,0.492205,-1.397963,
        0.837758,0.493314,3.117264,
        0.837758,0.497870,0.727828,
        0.837758,0.499783,-1.654379,
        0.837758,0.502651,0.463333,
        0.837758,0.505463,2.594557,
        0.837758,0.506104,-2.874910,
        0.837758,0.513579,-1.147354,
        0.837758,0.518569,1.644237,
        0.837758,0.519001,1.900488,
        0.837758,0.519277,-0.673809,
        0.837758,0.524769,-0.421955,
        0.837758,0.525564,-2.288838,
        0.837758,0.532541,0.978737,
        0.837758,0.534610,2.154808,
        0.837758,0.535020,-1.891846,
        0.837758,0.540208,1.400497,
        0.837758,0.542006,0.228337,
        0.837758,0.550762,-0.180771,
        0.837758,0.560531,2.378474,
        0.837758,0.561629,-0.924576,
        0.837758,0.567465,-2.679392,
        0.837758,0.588527,1.181755,
        0.837758,0.591643,-2.098364,
        0.837758,0.594963,2.997575,
        0.837758,0.599643,-3.045235,
        0.837758,0.599962,2.773556,
        0.837758,0.602620,0.029270,
        0.837758,0.605711,0.598454,
        0.837758,0.607708,-1.541026,
        0.837758,0.609360,-2.464379,
        0.837758,0.610738,-1.326786,
        0.837758,0.616968,0.820489,
        0.837758,0.626379,1.773610,
        0.837758,0.628626,-1.751180,
        0.837758,0.630671,0.391949,
        0.837758,0.632622,-0.737642,
        0.837758,0.632708,2.566722,
        0.837758,0.632797,1.989858,
        0.837758,0.634455,-0.534790,
        0.837758,0.636205,1.561170,
        0.837758,0.637681,-1.123423,
        0.837758,0.641441,-2.842109,
        0.837758,0.644807,-0.327139,
        0.837758,0.652440,2.197129,
        0.837758,0.660897,-2.278593,
        0.837758,0.664008,1.360091,
        0.837758,0.670850,-1.945257,
        0.837758,0.671396,1.004859,
        0.837758,0.673985,0.203855,
        0.837758,0.680326,-0.133370,
        0.837758,0.685083,-0.942363,
        0.837758,0.688204,-2.632204,
        0.837758,0.689777,2.385011,
        0.837758,0.700065,3.115430,
        0.837758,0.704080,2.911636,
        0.837758,0.712883,1.174216,
        0.837758,0.715701,0.699906,
        0.837758,0.718851,-1.442215,
        0.837758,0.721019,2.721536,
        0.837758,0.725300,-2.975134,
        0.837758,0.727136,-1.633288,
        0.837758,0.727399,-2.120567,
        0.837758,0.730751,0.522488,
        0.837758,0.732892,-1.254330,
        0.837758,0.734333,-2.446184,
        0.837758,0.735125,0.037270,
        0.837758,0.737893,1.866463,
        0.837758,0.740650,1.681075,
        0.837758,0.745971,-0.781286,
        0.837758,0.749432,-0.442559,
        0.837758,0.750102,2.050636,
        0.837758,0.751046,0.863821,
        0.837758,0.757018,-0.608833,
        0.837758,0.757046,-1.816019,
        0.837758,0.757389,1.501184,
        0.837758,0.759800,2.546181,
        0.837758,0.759859,0.351923,
        0.837758,0.761141,-2.781102,
        0.837758,0.769035,-1.078574,
        0.837758,0.770166,-0.266604,
        0.837758,0.772218,2.227380,
        0.837758,0.788382,1.331051,
        0.837758,0.790269,-2.282409,
        0.837758,0.803410,-1.984567,
        0.837758,0.803813,0.191520,
        0.837758,0.806832,3.022155,
        0.837758,0.807476,-2.597675,
        0.837758,0.812512,-0.107335,
        0.837758,0.814161,1.010949,
        0.837758,0.817732,-3.084486,
        0.837758,0.817846,2.390688,
        0.837758,0.819323,2.845960,
        0.837758,0.819430,-0.915017,
        0.837758,0.829198,-1.532491,
        0.837758,0.829531,0.665695,
        0.837758,0.829816,1.167147,
        0.837758,0.833613,-1.357492,
        0.837758,0.842945,-2.907318,
        0.837758,0.844883,2.679708,
        0.837758,0.848189,-1.705003,
        0.837758,0.849888,1.767901,
        0.837758,0.851126,0.493370,
        0.837758,0.852911,1.933394,
        0.837758,0.857516,-2.432863,
        0.837758,0.858261,-1.191729,
        0.837758,0.858279,-2.138895,
        0.837758,0.859723,-0.734928,
        0.837758,0.859754,1.603292,
        0.837758,0.861655,-0.394821,
        0.837758,0.864300,0.043549,
        0.837758,0.866940,0.851229,
        0.837758,0.869748,2.095969,
        0.837758,0.877352,-0.574668,
        0.837758,0.879254,-2.733338,
        0.837758,0.882554,1.443757,
        0.837758,0.884060,0.334199,
        0.837758,0.885626,-1.866061,
        0.837758,0.889834,2.529957,
        0.837758,0.893140,2.251698,
        0.837758,0.897463,-0.240822,
        0.837758,0.898758,-1.031948,
        0.837758,0.913918,1.287618,
        0.837758,0.916555,-2.287173,
        0.837758,0.918305,3.107845,
        0.837758,0.919522,2.948277,
        0.837758,0.927138,-2.569686,
        0.837758,0.929031,0.185383,
        0.837758,0.933766,-3.015917,
        0.837758,0.934873,-2.010822,
        0.837758,0.935636,-1.445855,
        0.837758,0.936702,-0.866461,
        0.837758,0.937395,0.729108,
        0.837758,0.937910,2.787575,
        0.837758,0.939144,-0.092962,
        0.837758,0.941718,1.125085,
        0.837758,0.942873,-1.607680,
        0.837758,0.943701,0.980328,
        0.837758,0.947774,2.395256,
        0.837758,0.952433,0.584347,
        0.837758,0.954456,-1.287293,
        0.837758,0.958731,-2.852552,
        0.837758,0.961838,1.839284,
        0.837758,0.961898,-0.469425,
        0.837758,0.965426,1.686615,
        0.837758,0.970108,1.990771,
        0.837758,0.972820,-1.762780,
        0.837758,0.973500,-0.711881,
        0.837758,0.975253,2.641082,
        0.837758,0.977824,0.441928,
        0.837758,0.981485,-2.424766,
        0.837758,0.981877,1.537100,
        0.837758,0.982609,-1.136920,
        0.837758,0.984265,-2.156073,
        0.837758,0.989150,0.045493,
        0.837758,0.993052,2.137478,
        0.837758,0.997766,-2.693871,
        0.837758,0.999847,-0.334738,
        0.837758,1.001920,0.848054,
        0.837758,1.006146,1.389734,
        0.837758,1.012180,0.304066,
        0.837758,1.016938,2.270674,
        0.837758,1.017371,-0.979790,
        0.837758,1.019495,-1.894487,
        0.837758,1.023796,-0.583691,
        0.837758,1.026525,2.506945,
        0.837758,1.027004,-0.203653,
        0.837758,1.028027,3.036573,
        0.837758,1.031435,1.241021,
        0.837758,1.032316,2.884221,
        0.837758,1.035863,-3.110475,
        0.837758,1.042085,-2.292950,
        0.837758,1.043272,-1.522986,
        0.837758,1.046265,-2.958847,
        0.837758,1.049272,-1.376344,
        0.837758,1.050039,-2.552086,
        0.837758,1.053682,0.689863,
        0.837758,1.055458,0.170394,
        0.837758,1.056622,-0.831866,
        0.837758,1.057515,1.097998,
        0.837758,1.058706,-2.036438,
        0.837758,1.061389,-1.672248,
        0.837758,1.063758,2.743093,
        0.837758,1.070157,-0.073422,
        0.837758,1.072154,1.761506,
        0.837758,1.074029,1.909193,
        0.837758,1.074204,0.545276,
        0.837758,1.074714,-2.805147,
        0.837758,1.076262,0.966104,
        0.837758,1.077642,-1.227073,
        0.837758,1.084151,1.618525,
        0.837758,1.085085,2.377673,
        0.837758,1.086822,-0.441112,
        0.837758,1.087142,2.058815,
        0.837758,1.098383,-0.696430,
        0.837758,1.099138,-1.084033,
        0.837758,1.103208,1.477773,
        0.837758,1.103783,0.408254,
        0.837758,1.106275,-2.420800,
        0.837758,1.108079,2.612943,
        0.837758,1.109345,-2.172718,
        0.837758,1.115957,-1.787597,
        0.837758,1.118930,0.045983,
        0.837758,1.119777,0.834702,
        0.837758,1.121943,-2.669042,
        0.837758,1.124930,2.219917,
        0.837758,1.125217,1.336294,
        0.837758,1.130958,2.972931,
        0.837758,1.131400,-0.306629,
        0.837758,1.137135,-0.936514,
        0.837758,1.139125,-1.924528,
        0.837758,1.139877,0.276382,
        0.837758,1.142003,-3.058022,
        0.837758,1.147363,-0.558448,
        0.837758,1.150348,-0.176418,
        0.837758,1.151198,1.197016,
        0.837758,1.153930,-1.597104,
        0.837758,1.155389,2.837148,
        0.837758,1.155458,-1.460623,
        0.837758,1.157331,-2.906362,
        0.837758,1.158594,2.487618,
        0.837758,1.162103,-1.322553,
        0.837758,1.165407,3.092861,
        0.837758,1.168025,0.717075,
        0.837758,1.168406,-2.301519,
        0.837758,1.174782,-2.538755,
        0.837758,1.177998,1.061256,
        0.837758,1.178898,1.827305,
        0.837758,1.181331,-2.060330,
        0.837758,1.182540,-0.798952,
        0.837758,1.184569,0.147517,
        0.837758,1.184734,1.968919,
        0.837758,1.184804,0.593386,
        0.837758,1.188278,1.690808,
        0.837758,1.194360,2.110440,
        0.837758,1.194776,2.710455,
        0.837758,1.197745,-1.173705,
        0.837758,1.199898,-2.777257,
        0.837758,1.201164,2.355931,
        0.837758,1.203760,1.556404,
        0.837758,1.205845,-0.415993,
        0.837758,1.206585,0.933648,
        0.837758,1.206950,-0.037248,
        0.837758,1.209341,0.471162,
        0.837758,1.213483,-1.039417,
        0.837758,1.218146,-0.672643,
        0.837758,1.221608,1.421669,
        0.837758,1.226043,-1.693533,
        0.837758,1.232870,-1.822773,
        0.837758,1.234375,-2.421194,
        0.837758,1.234531,-2.190148,
        0.837758,1.239414,0.348959,
        0.837758,1.240183,2.587255,
        0.837758,1.242312,2.941234,
        0.837758,1.244075,1.286644,
        0.837758,1.245961,-2.997461,
        0.837758,1.246495,-2.646872,
        0.837758,1.247736,2.242810,
        0.837758,1.253604,0.812222,
        0.837758,1.257816,-0.894317,
        0.837758,1.258247,-1.953634,
        0.837758,1.258993,-0.284099,
        0.837758,1.262380,-0.157609,
        0.837758,1.262791,-1.549229,
        0.837758,1.263772,-0.533073,
        0.837758,1.266731,-1.412461,
        0.837758,1.268591,-3.122146,
        0.837758,1.269614,-1.275246,
        0.837758,1.270745,1.152533,
        0.837758,1.273606,0.224545,
        0.837758,1.279716,2.812101,
        0.837758,1.283786,2.462124,
        0.837758,1.284798,0.078209,
        0.837758,1.285402,-2.882049,
        0.837758,1.286936,1.886053,
        0.837758,1.289408,0.683122,
        0.837758,1.294736,1.756184,
        0.837758,1.296978,-2.311511,
        0.837758,1.298546,2.020601,
        0.837758,1.300924,1.022478,
        0.837758,1.303305,3.043152,
        0.837758,1.304210,-2.084225,
        0.837758,1.305241,-2.530249,
        0.837758,1.307513,1.628003,
        0.837758,1.310722,0.558267,
        0.837758,1.314142,2.139329,
        0.837758,1.316087,-0.777971,
        0.837758,1.317907,-1.121023,
        0.837758,1.321019,-2.749776,
        0.837758,1.321609,2.687717,
        0.837758,1.322568,1.499415,
        0.837758,1.322972,-0.395699,
        0.837758,1.325286,-0.994192,
        0.837758,1.329146,-0.648513,
        0.837758,1.329795,-0.042570,
        0.837758,1.331651,2.344497,
        0.837758,1.336522,0.435098,
        0.837758,1.339410,0.900087,
        0.837758,1.339550,1.370353,
        0.837758,1.340441,-1.731580,
        0.837758,1.351855,-1.855234,
        0.837758,1.360161,-3.018404,
        0.837758,1.361031,-1.613971,
        0.837758,1.361196,2.914780,
        0.837758,1.361998,-2.207254,
        0.837758,1.362302,1.239384,
        0.837758,1.365228,2.563808,
        0.837758,1.365452,0.310522,
        0.837758,1.367028,-2.420041,
        0.837758,1.373422,-1.484281,
        0.837758,1.374245,-1.226417,
        0.837758,1.375020,-0.170393,
        0.837758,1.376560,-1.355481,
        0.837758,1.377070,-1.980984,
        0.837758,1.377578,-2.636339,
        0.837758,1.378096,0.774626,
        0.837758,1.379086,-0.508192,
        0.837758,1.382562,0.177141,
        0.837758,1.388849,2.233847,
        0.837758,1.389268,-0.886349,
        0.837758,1.391324,1.109396,
        0.837758,1.391699,-0.288725,
        0.837758,1.392869,3.136903,
        0.837758,1.393612,1.952016,
        0.837758,1.395545,-2.850807,
        0.837758,1.401926,1.822930,
        0.837758,1.402149,2.788898,
        0.837758,1.409552,0.647980,
        0.837758,1.410840,0.056219,
        0.837758,1.411944,2.444377,
        0.837758,1.413639,1.697982,
        0.837758,1.426479,1.573427,
        0.837758,1.426725,2.102242,
        0.837758,1.427868,0.984637,
        0.837758,1.429580,-2.107232,
        0.837758,1.431375,-2.313269,
        0.837758,1.431907,3.015410,
        0.837758,1.432325,-1.029054,
        0.837758,1.433341,0.524268,
        0.837758,1.435181,-0.739771,
        0.837758,1.437208,-2.529085,
        0.837758,1.439041,-0.617966,
        0.837758,1.441473,1.448544,
        0.837758,1.444527,2.664744,
        0.837758,1.450513,-0.400625,
        0.837758,1.451907,-0.066823,
        0.837758,1.452754,-2.738631,
        0.837758,1.454061,-2.950416,
        0.837758,1.456736,1.323154,
        0.837758,1.458370,0.400433,
        0.837758,1.459314,-1.765003,
        0.837758,1.459902,-1.145835,
        0.837758,1.464952,2.329245,
        0.837758,1.466050,0.862104,
        0.837758,1.471330,-1.886160,
        0.837758,1.475119,-1.648255,
        0.837758,1.480878,1.194576,
        0.837758,1.480942,0.273235,
        0.837758,1.481086,2.890269,
        0.837758,1.484118,-1.529156,
        0.837758,1.484741,-1.279219,
        0.837758,1.487681,-1.406069,
        0.837758,1.487886,-3.066037,
        0.837758,1.490555,2.543284,
        0.837758,1.492424,-0.190932,
        0.837758,1.495002,-0.839810,
        0.837758,1.495297,2.007188,
        0.837758,1.496414,-2.009267,
        0.837758,1.498261,-2.425270,
        0.837758,1.498907,-2.207327,
        0.837758,1.500204,0.737638,
        0.837758,1.502469,0.151156,
        0.837758,1.507448,-0.514782,
        0.837758,1.509287,2.204647,
        0.837758,1.510518,1.884920,
        0.837758,1.511948,-2.632360,
        0.837758,1.515235,1.068565,
        0.837758,1.522237,3.102018,
        0.837758,1.522373,1.764991,
        0.837758,1.522416,-0.954956,
        0.837758,1.523102,2.765920,
        0.837758,1.524811,-2.844718,
        0.837758,1.527261,-0.304916,
        0.837758,1.529788,0.613208,
        0.837758,1.533648,1.644808,
        0.837758,1.535018,0.028869,
        0.837758,1.541082,2.424608,
        0.837758,1.546010,1.523462,
        0.837758,1.551135,-0.642994,
        0.837758,1.553895,0.489607,
        0.837758,1.555041,0.947813,
        0.837758,1.559671,-1.074934,
        0.837758,1.560345,2.985899,
        0.837758,1.560487,-2.323400,
        0.837758,1.561127,1.401111,
        0.837758,1.567193,2.643172,
        0.837758,1.568892,2.098362,
        0.837758,1.571702,-2.098435,
        0.837758,1.572701,-2.528730,
        0.837758,1.573572,1.278593,
        0.837758,1.574849,-0.091505,
        0.837758,1.576334,0.365236,
        0.837758,1.577029,-2.962888,
        0.837758,1.577688,-1.814941,
        0.837758,1.579542,-0.755917,
        0.837758,1.579624,-1.199759,
        0.837758,1.581515,-0.419414,
        0.837758,1.584581,-2.738702,
        0.837758,1.586948,-1.696296,
        0.837758,1.588901,2.301496,
        0.837758,1.589936,0.825020,
        0.837758,1.590018,-1.941652,
        0.837758,1.596968,-1.328358,
        0.837758,1.597178,-1.575607,
        0.837758,1.600470,-1.453196,
        0.837758,1.601408,0.242409,
        0.837758,1.602259,1.152656,
        0.837758,1.603064,2.865774,
        0.837758,1.612085,1.968802,
        0.837758,1.612720,-3.080498,
        0.837758,1.616061,2.521531,
        0.837758,1.616900,-0.207457,
        0.837758,1.620988,-0.875920,
        0.837758,1.622114,0.701509,
        0.837758,1.622749,0.121426,
        0.837758,1.622993,-2.221629,
        0.837758,1.629346,-0.542669,
        0.837758,1.630007,1.844552,
        0.837758,1.634747,-2.426198,
        0.837758,1.642346,1.722617,
        0.837758,1.643925,1.034629,
        0.837758,1.644848,2.743774,
        0.837758,1.645258,2.188109,
        0.837758,1.645417,-2.634725,
        0.837758,1.647395,3.085437,
        0.837758,1.648538,-2.855044,
        0.837758,1.649932,0.576808,
        0.837758,1.652224,-0.996105,
        0.837758,1.652871,1.600681,
        0.837758,1.657997,-0.325127,
        0.837758,1.661936,0.001037,
        0.837758,1.665187,1.477402,
        0.837758,1.665496,2.397872,
        0.837758,1.673267,-0.665453,
        0.837758,1.673557,0.452526,
        0.837758,1.679277,-1.122421,
        0.837758,1.679696,0.912482,
        0.837758,1.682420,1.353922,
        0.837758,1.685031,-2.009038,
        0.837758,1.685196,2.967046,
        0.837758,1.690107,2.620910,
        0.837758,1.690453,-2.123076,
        0.837758,1.691693,-1.871572,
        0.837758,1.691817,2.061688,
        0.837758,1.692644,1.237670,
        0.837758,1.696386,0.328123,
        0.837758,1.697162,-1.249026,
        0.837758,1.697488,-2.322785,
        0.837758,1.698726,-2.973576,
        0.837758,1.699172,-1.749717,
        0.837758,1.704478,-0.106295,
        0.837758,1.707244,-0.444770,
        0.837758,1.707500,-2.531497,
        0.837758,1.707714,-0.787321,
        0.837758,1.708430,-1.627552,
        0.837758,1.710583,-1.376890,
        0.837758,1.712415,-2.747921,
        0.837758,1.713125,-1.503073,
        0.837758,1.713452,0.789419,
        0.837758,1.720612,2.279888,
        0.837758,1.724852,0.202831,
        0.837758,1.724868,2.846246,
        0.837758,1.728940,1.931535,
        0.837758,1.732838,1.127584,
        0.837758,1.735719,-3.092485,
        0.837758,1.739676,2.496017,
        0.837758,1.739971,-0.228876,
        0.837758,1.743328,0.663347,
        0.837758,1.744927,-0.911892,
        0.837758,1.745952,0.084733,
        0.837758,1.748077,1.804354,
        0.837758,1.752959,-0.567548,
        0.837758,1.760169,1.678132,
        0.837758,1.763887,-2.218450,
        0.837758,1.766737,2.722752,
        0.837758,1.767999,1.004737,
        0.837758,1.769149,1.550298,
        0.837758,1.770119,-2.427175,
        0.837758,1.770460,2.154208,
        0.837758,1.770514,3.070982,
        0.837758,1.770735,0.537245,
        0.837758,1.772673,-2.864760,
        0.837758,1.773527,-1.039582,
        0.837758,1.776419,-2.642421,
        0.837758,1.780769,1.416330,
        0.837758,1.786634,-0.346247,
        0.837758,1.788829,-1.947553,
        0.837758,1.793737,0.410270,
        0.837758,1.794553,2.374618,
        0.837758,1.794699,-0.694061,
        0.837758,1.797469,-1.169105,
        0.837758,1.803705,0.881076,
        0.837758,1.804124,1.264249,
        0.837758,1.807825,-2.090455,
        0.837758,1.808822,-1.817375,
        0.837758,1.809020,2.949741,
        0.837758,1.812466,2.597595,
        0.837758,1.813159,2.025395,
        0.837758,1.813816,-1.298435,
        0.837758,1.815016,0.278951,
        0.837758,1.816510,-0.008631,
        0.837758,1.817772,-1.689625,
        0.837758,1.821360,-2.984352,
        0.837758,1.822312,-0.127211,
        0.837758,1.823386,-1.429807,
        0.837758,1.823834,-1.560483,
        0.837758,1.832060,-0.821627,
        0.837758,1.832690,-0.468440,
        0.837758,1.833945,-2.320278,
        0.837758,1.834787,0.751968,
        0.837758,1.840579,-2.536445,
        0.837758,1.841033,-2.756932,
        0.837758,1.846350,1.893983,
        0.837758,1.846644,1.113766,
        0.837758,1.846833,2.826543,
        0.837758,1.847325,2.249051,
        0.837758,1.854119,0.134161,
        0.837758,1.858863,-3.106022,
        0.837758,1.865046,1.763384,
        0.837758,1.865619,-0.952036,
        0.837758,1.865912,0.624453,
        0.837758,1.866694,2.473686,
        0.837758,1.867524,-0.240083,
        0.837758,1.875561,1.632831,
        0.837758,1.876049,-0.594378,
        0.837758,1.880509,1.500204,
        0.837758,1.883504,1.360457,
        0.837758,1.886971,-2.196665,
        0.837758,1.888979,0.982922,
        0.837758,1.889815,2.701570,
        0.837758,1.892144,-2.005340,
        0.837758,1.892348,0.496522,
        0.837758,1.893281,-1.083976,
        0.837758,1.893828,3.053453,
        0.837758,1.895117,2.121107,
        0.837758,1.898741,-2.875735,
        0.837758,1.904429,-2.428029,
        0.837758,1.908408,-2.648903,
        0.837758,1.911828,-0.364489,
        0.837758,1.912224,0.366565,
        0.837758,1.914854,1.214468,
        0.837758,1.915467,-1.883143,
        0.837758,1.915763,-0.724679,
        0.837758,1.915790,-1.216366,
        0.837758,1.922256,2.347806,
        0.837758,1.922616,0.844633,
        0.837758,1.925185,0.231170,
        0.837758,1.927474,-1.754616,
        0.837758,1.928342,-1.351174,
        0.837758,1.933328,-1.623328,
        0.837758,1.933878,0.034150,
        0.837758,1.934585,-1.487979,
        0.837758,1.934802,2.927634,
        0.837758,1.934862,1.989776,
        0.837758,1.936327,-0.099350,
        0.837758,1.938088,2.578507,
        0.837758,1.945677,-2.999805,
        0.837758,1.954572,-0.859208,
        0.837758,1.957868,-0.490784,
        0.837758,1.958989,0.712772,
        0.837758,1.960212,-2.307561,
        0.837758,1.964774,1.855507,
        0.837758,1.964971,-2.100533,
        0.837758,1.969112,1.088273,
        0.837758,1.971438,-2.765415,
        0.837758,1.972312,2.805134,
        0.837758,1.974950,2.220596,
        0.837758,1.975135,-2.539127,
        0.837758,1.982107,1.720829,
        0.837758,1.983195,-3.127984,
        0.837758,1.983318,-0.247182,
        0.837758,1.986720,-0.993220,
        0.837758,1.990381,1.586005,
        0.837758,1.990470,0.580283,
        0.837758,1.990523,1.450326,
        0.837758,1.993909,1.312897,
        0.837758,1.994694,2.452813,
        0.837758,1.999226,-0.620717,
        0.837758,2.005145,0.941012,
        0.837758,2.007301,0.142337,
        0.837758,2.014149,0.444138,
        0.837758,2.014338,-1.127392,
        0.837758,2.018131,3.021778,
        0.837758,2.019676,-1.980944,
        0.837758,2.019912,2.686612,
        0.837758,2.020545,2.088672,
        0.837758,2.026074,0.300945,
        0.837758,2.026196,-2.889057,
        0.837758,2.031959,-1.268178,
        0.837758,2.032917,-2.199917,
        0.837758,2.033572,-1.842884,
        0.837758,2.034712,-2.419505,
        0.837758,2.035818,-0.379479,
        0.837758,2.036004,-0.757062,
        0.837758,2.040501,-1.411488,
        0.837758,2.042874,-1.701367,
        0.837758,2.043060,-2.653746,
        0.837758,2.043729,-1.557938,
        0.837758,2.048538,0.806623,
        0.837758,2.049006,1.199930,
        0.837758,2.049502,-0.137989,
        0.837758,2.052181,2.326959,
        0.837758,2.052794,-0.008526,
        0.837758,2.057608,1.951726,
        0.837758,2.065568,2.567730,
        0.837758,2.070371,2.879808,
        0.837758,2.071256,-3.018961,
        0.837758,2.079746,-0.897699,
        0.837758,2.082784,1.042403,
        0.837758,2.083752,-0.511086,
        0.837758,2.084401,1.810502,
        0.837758,2.085731,0.671081,
        0.837758,2.097496,1.373172,
        0.837758,2.099185,1.667006,
        0.837758,2.100482,-2.301084,
        0.837758,2.103118,2.194745,
        0.837758,2.103223,1.521685,
        0.837758,2.103695,-2.777707,
        0.837758,2.107523,3.129581,
        0.837758,2.108277,-2.095907,
        0.837758,2.108891,-2.532166,
        0.837758,2.111003,-0.268632,
        0.837758,2.112276,-1.027246,
        0.837758,2.114380,0.531457,
        0.837758,2.119166,0.100552,
        0.837758,2.119564,2.754818,
        0.837758,2.122474,-0.648314,
        0.837758,2.125183,2.445598,
        0.837758,2.129501,0.238290,
        0.837758,2.131178,0.386663,
        0.837758,2.131223,-1.174311,
        0.837758,2.134269,-1.948042,
        0.837758,2.134684,0.907065,
        0.837758,2.137615,2.988806,
        0.837758,2.145635,-1.327708,
        0.837758,2.146467,2.055081,
        0.837758,2.146582,-1.793788,
        0.837758,2.149210,-1.482321,
        0.837758,2.152901,-1.636788,
        0.837758,2.156337,1.151558,
        0.837758,2.156474,-2.912065,
        0.837758,2.157194,-0.792563,
        0.837758,2.166724,-0.401638,
        0.837758,2.166980,-0.134879,
        0.837758,2.173644,-2.403429,
        0.837758,2.175711,-2.655523,
        0.837758,2.177132,2.600419,
        0.837758,2.178879,0.767722,
        0.837758,2.180752,1.908485,
        0.837758,2.181338,2.313721,
        0.837758,2.185580,1.287673,
        0.837758,2.195400,-2.217546,
        0.837758,2.200418,-3.056442,
        0.837758,2.201829,-0.001735,
        0.837758,2.203804,1.756176,
        0.837758,2.211602,1.443930,
        0.837758,2.212686,-0.544809,
        0.837758,2.213937,0.621892,
        0.837758,2.214075,1.600691,
        0.837758,2.216256,1.020029,
        0.837758,2.219060,2.886631,
        0.837758,2.223375,-0.906185,
        0.837758,2.226002,-1.060248,
        0.837758,2.226413,-2.064034,
        0.837758,2.230554,2.171334,
        0.837758,2.232413,2.735777,
        0.837758,2.233709,3.070830,
        0.837758,2.234383,-2.792065,
        0.837758,2.237183,0.468799,
        0.837758,2.237695,-0.271388,
        0.837758,2.239419,0.151693,
        0.837758,2.243347,-2.521049,
        0.837758,2.244035,-1.229456,
        0.837758,2.244355,0.309899,
        0.837758,2.245175,2.460236,
        0.837758,2.245628,-1.899216,
        0.837758,2.245973,-0.702234,
        0.837758,2.255601,-1.728131,
        0.837758,2.256702,-1.401367,
        0.837758,2.261466,-1.557500,
        0.837758,2.267390,0.878499,
        0.837758,2.272099,2.018594,
        0.837758,2.278618,1.170386,
        0.837758,2.283304,-2.345584,
        0.837758,2.285175,-2.943090,
        0.837758,2.293000,-0.124451,
        0.837758,2.294058,-0.421070,
        0.837758,2.303283,1.855938,
        0.837758,2.304968,2.608978,
        0.837758,2.305347,2.311346,
        0.837758,2.307188,-2.656134,
        0.837758,2.307301,1.336200,
        0.837758,2.309655,0.726020,
        0.837758,2.317634,-2.185225,
        0.837758,2.321510,1.685664,
        0.837758,2.324376,-3.111006,
        0.837758,2.324731,1.511186,
        0.837758,2.329587,0.035498,
        0.837758,2.330158,2.977264,
        0.837758,2.335096,-0.588468,
        0.837758,2.336046,-1.115799,
        0.837758,2.338060,2.803613,
        0.837758,2.338083,-0.943400,
        0.837758,2.340210,0.562145,
        0.837758,2.341931,1.023926,
        0.837758,2.343265,-0.779061,
        0.837758,2.343434,-2.013692,
        0.837758,2.353029,-1.310529,
        0.837758,2.354868,0.211543,
        0.837758,2.355120,0.388593,
        0.837758,2.355165,-1.833173,
        0.837758,2.355462,2.151441,
        0.837758,2.362948,-2.487477,
        0.837758,2.363101,-2.810456,
        0.837758,2.363394,-0.272918,
        0.837758,2.364509,-1.644533,
        0.837758,2.370465,2.466824,
        0.837758,2.386110,-1.481351,
        0.837758,2.394960,0.862133,
        0.837758,2.395716,1.977184,
        0.837758,2.396339,1.200900,
        0.837758,2.407941,-2.316753,
        0.837758,2.409450,-2.988432,
        0.837758,2.417202,-0.103363,
        0.837758,2.417386,-0.443097,
        0.837758,2.417950,2.666662,
        0.837758,2.423190,1.789783,
        0.837758,2.423692,1.392696,
        0.837758,2.431122,3.086379,
        0.837758,2.431406,-2.647820,
        0.837758,2.431584,2.307631,
        0.837758,2.434090,1.592818,
        0.837758,2.435993,0.683844,
        0.837758,2.440898,-2.133300,
        0.837758,2.442222,-1.195837,
        0.837758,2.443112,2.883761,
        0.837758,2.449191,-0.633651,
        0.837758,2.450821,0.084741,
        0.837758,2.453499,-1.008861,
        0.837758,2.460010,-1.938011,
        0.837758,2.460199,-0.822008,
        0.837758,2.460902,0.490079,
        0.837758,2.461774,-1.734913,
        0.837758,2.463986,1.034407,
        0.837758,2.467647,0.287667,
        0.837758,2.479849,2.126925,
        0.837758,2.485907,-2.835471,
        0.837758,2.485945,-1.374853,
        0.837758,2.487747,2.505912,
        0.837758,2.487957,-0.274279,
        0.837758,2.489741,-2.468428,
        0.837758,2.512538,1.239900,
        0.837758,2.517528,1.923289,
        0.837758,2.518454,-1.565852,
        0.837758,2.519348,0.843503,
        0.837758,2.522766,-3.059231,
        0.837758,2.528686,2.737228,
        0.837758,2.534798,-2.266671,
        0.837758,2.536533,1.466274,
        0.837758,2.538504,1.699812,
        0.837758,2.539516,-0.476942,
        0.837758,2.539532,-0.074548,
        0.837758,2.543197,2.990252,
        0.837758,2.554022,2.320619,
        0.837758,2.556919,-2.655005,
        0.837758,2.558146,0.627191,
        0.837758,2.558223,-1.164189,
        0.837758,2.563452,-2.044028,
        0.837758,2.568462,0.150655,
        0.837758,2.569130,-0.700837,
        0.837758,2.570050,-1.804492,
        0.837758,2.572710,-0.934665,
        0.837758,2.575994,0.391718,
        0.837758,2.584774,1.046703,
        0.837758,2.601432,2.102535,
        0.837758,2.603559,2.555114,
        0.837758,2.604150,-2.885208,
        0.837758,2.608630,-1.378011,
        0.837758,2.611256,-0.279110,
        0.837758,2.617344,-2.439480,
        0.837758,2.627402,1.293552,
        0.837758,2.635396,1.848850,
        0.837758,2.635686,2.834950,
        0.837758,2.637087,3.132730,
        0.837758,2.642029,0.813274,
        0.837758,2.645260,1.570203,
        0.837758,2.646182,-1.614036,
        0.837758,2.660113,-2.186513,
        0.837758,2.660483,-0.034608,
        0.837758,2.661713,-0.529511,
        0.837758,2.672720,-1.110422,
        0.837758,2.676626,2.335353,
        0.837758,2.676638,0.540364,
        0.837758,2.676780,-1.901158,
        0.837758,2.681411,-2.667965,
        0.837758,2.682152,0.245751,
        0.837758,2.682859,-0.817071,
        0.837758,2.705107,1.062142,
        0.837758,2.717705,2.625995,
        0.837758,2.721001,2.060097,
        0.837758,2.722319,-2.954069,
        0.837758,2.729840,-1.368237,
        0.837758,2.734349,-0.289298,
        0.837758,2.738409,2.975671,
        0.837758,2.740366,1.374341,
        0.837758,2.744633,-2.391311,
        0.837758,2.747398,1.727166,
        0.837758,2.763799,0.760372,
        0.837758,2.768238,-1.679243,
        0.837758,2.779245,-2.046268,
        0.837758,2.780208,-0.626453,
        0.837758,2.780471,0.027732,
        0.837758,2.784696,-1.013128,
        0.837758,2.789817,0.395431,
        0.837758,2.801154,2.353143,
        0.837758,2.807091,-2.687477,
        0.837758,2.825762,1.085035,
        0.837758,2.830209,2.734943,
        0.837758,2.833517,-3.085508,
        0.837758,2.838301,1.980061,
        0.837758,2.849871,1.512012,
        0.837758,2.850089,-1.343899,
        0.837758,2.856905,-0.318441,
        0.837758,2.870445,-2.291001,
        0.837758,2.885017,0.655562,
        0.837758,2.886428,-1.793217,
        0.837758,2.890337,-0.825633,
        0.837758,2.898932,0.139973,
        0.837758,2.931927,-2.728768,
        0.837758,2.932096,2.377790,
        0.837758,2.937409,2.947245,
        0.837758,2.948048,1.133127,
        0.837758,2.950896,1.797889,
        0.837758,2.970019,-1.282382,
        0.837758,2.978333,-0.412090,
        0.837758,2.996437,-2.089121,
        0.837758,3.011274,0.434110,
        0.837758,3.050410,3.138492,
        0.837758,3.064492,1.647259,
        0.837758,3.088868,-0.989501,
        0.942478,0.053884,-2.408394,
        0.942478,0.077645,1.453433,
        0.942478,0.088408,-0.311160,
        0.942478,0.142314,2.590387,
        0.942478,0.147206,-1.385164,
        0.942478,0.163506,0.573203,
        0.942478,0.165485,-2.875396,
        0.942478,0.182564,-2.125821,
        0.942478,0.190569,1.929141,
        0.942478,0.201102,1.249082,
        0.942478,0.201397,-0.716646,
        0.942478,0.210568,-0.069559,
        0.942478,0.251672,2.848900,
        0.942478,0.266262,-1.678068,
        0.942478,0.269108,-1.215563,
        0.942478,0.273777,-2.506756,
        0.942478,0.279854,0.814581,
        0.942478,0.281664,-2.969147,
        0.942478,0.283625,2.405603,
        0.942478,0.284058,0.357350,
        0.942478,0.294107,1.593880,
        0.942478,0.302491,-0.400410,
        0.942478,0.308059,2.021491,
        0.942478,0.312219,-2.072057,
        0.942478,0.321905,-0.829121,
        0.942478,0.329984,1.201621,
        0.942478,0.339714,-0.013712,
        0.942478,0.368784,2.974974,
        0.942478,0.376428,-1.495532,
        0.942478,0.378103,2.646436,
        0.942478,0.385088,0.590355,
        0.942478,0.386248,-2.690943,
        0.942478,0.389851,-2.340031,
        0.942478,0.391502,-1.171374,
        0.942478,0.398249,-1.812694,
        0.942478,0.400542,-2.995729,
        0.942478,0.404273,0.918163,
        0.942478,0.405617,1.773533,
        0.942478,0.409009,-0.570243,
        0.942478,0.410700,0.270433,
        0.942478,0.415893,1.464890,
        0.942478,0.420680,2.088164,
        0.942478,0.424766,-0.258729,
        0.942478,0.426507,2.367872,
        0.942478,0.441292,-0.890727,
        0.942478,0.451839,-2.078193,
        0.942478,0.459270,1.191562,
        0.942478,0.470622,0.014002,
        0.942478,0.481292,-2.529098,
        0.942478,0.482460,2.849843,
        0.942478,0.492205,-1.397963,
        0.942478,0.493314,3.117264,
        0.942478,0.497870,0.727828,
        0.942478,0.499783,-1.654379,
        0.942478,0.502651,0.463333,
        0.942478,0.505463,2.594557,
        0.942478,0.506104,-2.874910,
        0.942478,0.513579,-1.147354,
        0.942478,0.518569,1.644237,
        0.942478,0.519001,1.900488,
        0.942478,0.519277,-0.673809,
        0.942478,0.524769,-0.421955,
        0.942478,0.525564,-2.288838,
        0.942478,0.532541,0.978737,
        0.942478,0.534610,2.154808,
        0.942478,0.535020,-1.891846,
        0.942478,0.540208,1.400497,
        0.942478,0.542006,0.228337,
        0.942478,0.550762,-0.180771,
        0.942478,0.560531,2.378474,
        0.942478,0.561629,-0.924576,
        0.942478,0.567465,-2.679392,
        0.942478,0.588527,1.181755,
        0.942478,0.591643,-2.098364,
        0.942478,0.594963,2.997575,
        0.942478,0.599643,-3.045235,
        0.942478,0.599962,2.773556,
        0.942478,0.602620,0.029270,
        0.942478,0.605711,0.598454,
        0.942478,0.607708,-1.541026,
        0.942478,0.609360,-2.464379,
        0.942478,0.610738,-1.326786,
        0.942478,0.616968,0.820489,
        0.942478,0.626379,1.773610,
        0.942478,0.628626,-1.751180,
        0.942478,0.630671,0.391949,
        0.942478,0.632622,-0.737642,
        0.942478,0.632708,2.566722,
        0.942478,0.632797,1.989858,
        0.942478,0.634455,-0.534790,
        0.942478,0.636205,1.561170,
        0.942478,0.637681,-1.123423,
        0.942478,0.641441,-2.842109,
        0.942478,0.644807,-0.327139,
        0.942478,0.652440,2.197129,
        0.942478,0.660897,-2.278593,
        0.942478,0.664008,1.360091,
        0.942478,0.670850,-1.945257,
        0.942478,0.671396,1.004859,
        0.942478,0.673985,0.203855,
        0.942478,0.680326,-0.133370,
        0.942478,0.685083,-0.942363,
        0.942478,0.688204,-2.632204,
        0.942478,0.689777,2.385011,
        0.942478,0.700065,3.115430,
        0.942478,0.704080,2.911636,
        0.942478,0.712883,1.174216,
        0.942478,0.715701,0.699906,
        0.942478,0.718851,-1.442215,
        0.942478,0.721019,2.721536,
        0.942478,0.725300,-2.975134,
        0.942478,0.727136,-1.633288,
        0.942478,0.727399,-2.120567,
        0.942478,0.730751,0.522488,
        0.942478,0.732892,-1.254330,
        0.942478,0.734333,-2.446184,
        0.942478,0.735125,0.037270,
        0.942478,0.737893,1.866463,
        0.942478,0.740650,1.681075,
        0.942478,0.745971,-0.781286,
        0.942478,0.749432,-0.442559,
        0.942478,0.750102,2.050636,
        0.942478,0.751046,0.863821,
        0.942478,0.757018,-0.608833,
        0.942478,0.757046,-1.816019,
        0.942478,0.757389,1.501184,
        0.942478,0.759800,2.546181,
        0.942478,0.759859,0.351923,
        0.942478,0.761141,-2.781102,
        0.942478,0.769035,-1.078574,
        0.942478,0.770166,-0.266604,
        0.942478,0.772218,2.227380,
        0.942478,0.788382,1.331051,
        0.942478,0.790269,-2.282409,
        0.942478,0.803410,-1.984567,
        0.942478,0.803813,0.191520,
        0.942478,0.806832,3.022155,
        0.942478,0.807476,-2.597675,
        0.942478,0.812512,-0.107335,
        0.942478,0.814161,1.010949,
        0.942478,0.817732,-3.084486,
        0.942478,0.817846,2.390688,
        0.942478,0.819323,2.845960,
        0.942478,0.819430,-0.915017,
        0.942478,0.829198,-1.532491,
        0.942478,0.829531,0.665695,
        0.942478,0.829816,1.167147,
        0.942478,0.833613,-1.357492,
        0.942478,0.842945,-2.907318,
        0.942478,0.844883,2.679708,
        0.942478,0.848189,-1.705003,
        0.942478,0.849888,1.767901,
        0.942478,0.851126,0.493370,
        0.942478,0.852911,1.933394,
        0.942478,0.857516,-2.432863,
        0.942478,0.858261,-1.191729,
        0.942478,0.858279,-2.138895,
        0.942478,0.859723,-0.734928,
        0.942478,0.859754,1.603292,
        0.942478,0.861655,-0.394821,
        0.942478,0.864300,0.043549,
        0.942478,0.866940,0.851229,
        0.942478,0.869748,2.095969,
        0.942478,0.877352,-0.574668,
        0.942478,0.879254,-2.733338,
        0.942478,0.882554,1.443757,
        0.942478,0.884060,0.334199,
        0.942478,0.885626,-1.866061,
        0.942478,0.889834,2.529957,
        0.942478,0.893140,2.251698,
        0.942478,0.897463,-0.240822,
        0.942478,0.898758,-1.031948,
        0.942478,0.913918,1.287618,
        0.942478,0.916555,-2.287173,
        0.942478,0.918305,3.107845,
        0.942478,0.919522,2.948277,
        0.942478,0.927138,-2.569686,
        0.942478,0.929031,0.185383,
        0.942478,0.933766,-3.015917,
        0.942478,0.934873,-2.010822,
        0.942478,0.935636,-1.445855,
        0.942478,0.936702,-0.866461,
        0.942478,0.937395,0.729108,
        0.942478,0.937910,2.787575,
        0.942478,0.939144,-0.092962,
        0.942478,0.941718,1.125085,
        0.942478,0.942873,-1.607680,
        0.942478,0.943701,0.980328,
        0.942478,0.947774,2.395256,
        0.942478,0.952433,0.584347,
        0.942478,0.954456,-1.287293,
        0.942478,0.958731,-2.852552,
        0.942478,0.961838,1.839284,
        0.942478,0.961898,-0.469425,
        0.942478,0.965426,1.686615,
        0.942478,0.970108,1.990771,
        0.942478,0.972820,-1.762780,
        0.942478,0.973500,-0.711881,
        0.942478,0.975253,2.641082,
        0.942478,0.977824,0.441928,
        0.942478,0.981485,-2.424766,
        0.942478,0.981877,1.537100,
        0.942478,0.982609,-1.136920,
        0.942478,0.984265,-2.156073,
        0.942478,0.989150,0.045493,
        0.942478,0.993052,2.137478,
        0.942478,0.997766,-2.693871,
        0.942478,0.999847,-0.334738,
        0.942478,1.001920,0.848054,
        0.942478,1.006146,1.389734,
        0.942478,1.012180,0.304066,
        0.942478,1.016938,2.270674,
        0.942478,1.017371,-0.979790,
        0.942478,1.019495,-1.894487,
        0.942478,1.023796,-0.583691,
        0.942478,1.026525,2.506945,
        0.942478,1.027004,-0.203653,
        0.942478,1.028027,3.036573,
        0.942478,1.031435,1.241021,
        0.942478,1.032316,2.884221,
        0.942478,1.035863,-3.110475,
        0.942478,1.042085,-2.292950,
        0.942478,1.043272,-1.522986,
        0.942478,1.046265,-2.958847,
        0.942478,1.049272,-1.376344,
        0.942478,1.050039,-2.552086,
        0.942478,1.053682,0.689863,
        0.942478,1.055458,0.170394,
        0.942478,1.056622,-0.831866,
        0.942478,1.057515,1.097998,
        0.942478,1.058706,-2.036438,
        0.942478,1.061389,-1.672248,
        0.942478,1.063758,2.743093,
        0.942478,1.070157,-0.073422,
        0.942478,1.072154,1.761506,
        0.942478,1.074029,1.909193,
        0.942478,1.074204,0.545276,
        0.942478,1.074714,-2.805147,
        0.942478,1.076262,0.966104,
        0.942478,1.077642,-1.227073,
        0.942478,1.084151,1.618525,
        0.942478,1.085085,2.377673,
        0.942478,1.086822,-0.441112,
        0.942478,1.087142,2.058815,
        0.942478,1.098383,-0.696430,
        0.942478,1.099138,-1.084033,
        0.942478,1.103208,1.477773,
        0.942478,1.103783,0.408254,
        0.942478,1.106275,-2.420800,
        0.942478,1.108079,2.612943,
        0.942478,1.109345,-2.172718,
        0.942478,1.115957,-1.787597,
        0.942478,1.118930,0.045983,
        0.942478,1.119777,0.834702,
        0.942478,1.121943,-2.669042,
        0.942478,1.124930,2.219917,
        0.942478,1.125217,1.336294,
        0.942478,1.130958,2.972931,
        0.942478,1.131400,-0.306629,
        0.942478,1.137135,-0.936514,
        0.942478,1.139125,-1.924528,
        0.942478,1.139877,0.276382,
        0.942478,1.142003,-3.058022,
        0.942478,1.147363,-0.558448,
        0.942478,1.150348,-0.176418,
        0.942478,1.151198,1.197016,
        0.942478,1.153930,-1.597104,
        0.942478,1.155389,2.837148,
        0.942478,1.155458,-1.460623,
        0.942478,1.157331,-2.906362,
        0.942478,1.158594,2.487618,
        0.942478,1.162103,-1.322553,
        0.942478,1.165407,3.092861,
        0.942478,1.168025,0.717075,
        0.942478,1.168406,-2.301519,
        0.942478,1.174782,-2.538755,
        0.942478,1.177998,1.061256,
        0.942478,1.178898,1.827305,
        0.942478,1.181331,-2.060330,
        0.942478,1.182540,-0.798952,
        0.942478,1.184569,0.147517,
        0.942478,1.184734,1.968919,
        0.942478,1.184804,0.593386,
        0.942478,1.188278,1.690808,
        0.942478,1.194360,2.110440,
        0.942478,1.194776,2.710455,
        0.942478,1.197745,-1.173705,
        0.942478,1.199898,-2.777257,
        0.942478,1.201164,2.355931,
        0.942478,1.203760,1.556404,
        0.942478,1.205845,-0.415993,
        0.942478,1.206585,0.933648,
        0.942478,1.206950,-0.037248,
        0.942478,1.209341,0.471162,
        0.942478,1.213483,-1.039417,
        0.942478,1.218146,-0.672643,
        0.942478,1.221608,1.421669,
        0.942478,1.226043,-1.693533,
        0.942478,1.232870,-1.822773,
        0.942478,1.234375,-2.421194,
        0.942478,1.234531,-2.190148,
        0.942478,1.239414,0.348959,
        0.942478,1.240183,2.587255,
        0.942478,1.242312,2.941234,
        0.942478,1.244075,1.286644,
        0.942478,1.245961,-2.997461,
        0.942478,1.246495,-2.646872,
        0.942478,1.247736,2.242810,
        0.942478,1.253604,0.812222,
        0.942478,1.257816,-0.894317,
        0.942478,1.258247,-1.953634,
        0.942478,1.258993,-0.284099,
        0.942478,1.262380,-0.157609,
        0.942478,1.262791,-1.549229,
        0.942478,1.263772,-0.533073,
        0.942478,1.266731,-1.412461,
        0.942478,1.268591,-3.122146,
        0.942478,1.269614,-1.275246,
        0.942478,1.270745,1.152533,
        0.942478,1.273606,0.224545,
        0.942478,1.279716,2.812101,
        0.942478,1.283786,2.462124,
        0.942478,1.284798,0.078209,
        0.942478,1.285402,-2.882049,
        0.942478,1.286936,1.886053,
        0.942478,1.289408,0.683122,
        0.942478,1.294736,1.756184,
        0.942478,1.296978,-2.311511,
        0.942478,1.298546,2.020601,
        0.942478,1.300924,1.022478,
        0.942478,1.303305,3.043152,
        0.942478,1.304210,-2.084225,
        0.942478,1.305241,-2.530249,
        0.942478,1.307513,1.628003,
        0.942478,1.310722,0.558267,
        0.942478,1.314142,2.139329,
        0.942478,1.316087,-0.777971,
        0.942478,1.317907,-1.121023,
        0.942478,1.321019,-2.749776,
        0.942478,1.321609,2.687717,
        0.942478,1.322568,1.499415,
        0.942478,1.322972,-0.395699,
        0.942478,1.325286,-0.994192,
        0.942478,1.329146,-0.648513,
        0.942478,1.329795,-0.042570,
        0.942478,1.331651,2.344497,
        0.942478,1.336522,0.435098,
        0.942478,1.339410,0.900087,
        0.942478,1.339550,1.370353,
        0.942478,1.340441,-1.731580,
        0.942478,1.351855,-1.855234,
        0.942478,1.360161,-3.018404,
        0.942478,1.361031,-1.613971,
        0.942478,1.361196,2.914780,
        0.942478,1.361998,-2.207254,
        0.942478,1.362302,1.239384,
        0.942478,1.365228,2.563808,
        0.942478,1.365452,0.310522,
        0.942478,1.367028,-2.420041,
        0.942478,1.373422,-1.484281,
        0.942478,1.374245,-1.226417,
        0.942478,1.375020,-0.170393,
        0.942478,1.376560,-1.355481,
        0.942478,1.377070,-1.980984,
        0.942478,1.377578,-2.636339,
        0.942478,1.378096,0.774626,
        0.942478,1.379086,-0.508192,
        0.942478,1.382562,0.177141,
        0.942478,1.388849,2.233847,
        0.942478,1.389268,-0.886349,
        0.942478,1.391324,1.109396,
        0.942478,1.391699,-0.288725,
        0.942478,1.392869,3.136903,
        0.942478,1.393612,1.952016,
        0.942478,1.395545,-2.850807,
        0.942478,1.401926,1.822930,
        0.942478,1.402149,2.788898,
        0.942478,1.409552,0.647980,
        0.942478,1.410840,0.056219,
        0.942478,1.411944,2.444377,
        0.942478,1.413639,1.697982,
        0.942478,1.426479,1.573427,
        0.942478,1.426725,2.102242,
        0.942478,1.427868,0.984637,
        0.942478,1.429580,-2.107232,
        0.942478,1.431375,-2.313269,
        0.942478,1.431907,3.015410,
        0.942478,1.432325,-1.029054,
        0.942478,1.433341,0.524268,
        0.942478,1.435181,-0.739771,
        0.942478,1.437208,-2.529085,
        0.942478,1.439041,-0.617966,
        0.942478,1.441473,1.448544,
        0.942478,1.444527,2.664744,
        0.942478,1.450513,-0.400625,
        0.942478,1.451907,-0.066823,
        0.942478,1.452754,-2.738631,
        0.942478,1.454061,-2.950416,
        0.942478,1.456736,1.323154,
        0.942478,1.458370,0.400433,
        0.942478,1.459314,-1.765003,
        0.942478,1.459902,-1.145835,
        0.942478,1.464952,2.329245,
        0.942478,1.466050,0.862104,
        0.942478,1.471330,-1.886160,
        0.942478,1.475119,-1.648255,
        0.942478,1.480878,1.194576,
        0.942478,1.480942,0.273235,
        0.942478,1.481086,2.890269,
        0.942478,1.484118,-1.529156,
        0.942478,1.484741,-1.279219,
        0.942478,1.487681,-1.406069,
        0.942478,1.487886,-3.066037,
        0.942478,1.490555,2.543284,
        0.942478,1.492424,-0.190932,
        0.942478,1.495002,-0.839810,
        0.942478,1.495297,2.007188,
        0.942478,1.496414,-2.009267,
        0.942478,1.498261,-2.425270,
        0.942478,1.498907,-2.207327,
        0.942478,1.500204,0.737638,
        0.942478,1.502469,0.151156,
        0.942478,1.507448,-0.514782,
        0.942478,1.509287,2.204647,
        0.942478,1.510518,1.884920,
        0.942478,1.511948,-2.632360,
        0.942478,1.515235,1.068565,
        0.942478,1.522237,3.102018,
        0.942478,1.522373,1.764991,
        0.942478,1.522416,-0.954956,
        0.942478,1.523102,2.765920,
        0.942478,1.524811,-2.844718,
        0.942478,1.527261,-0.304916,
        0.942478,1.529788,0.613208,
        0.942478,1.533648,1.644808,
        0.942478,1.535018,0.028869,
        0.942478,1.541082,2.424608,
        0.942478,1.546010,1.523462,
        0.942478,1.551135,-0.642994,
        0.942478,1.553895,0.489607,
        0.942478,1.555041,0.947813,
        0.942478,1.559671,-1.074934,
        0.942478,1.560345,2.985899,
        0.942478,1.560487,-2.323400,
        0.942478,1.561127,1.401111,
        0.942478,1.567193,2.643172,
        0.942478,1.568892,2.098362,
        0.942478,1.571702,-2.098435,
        0.942478,1.572701,-2.528730,
        0.942478,1.573572,1.278593,
        0.942478,1.574849,-0.091505,
        0.942478,1.576334,0.365236,
        0.942478,1.577029,-2.962888,
        0.942478,1.577688,-1.814941,
        0.942478,1.579542,-0.755917,
        0.942478,1.579624,-1.199759,
        0.942478,1.581515,-0.419414,
        0.942478,1.584581,-2.738702,
        0.942478,1.586948,-1.696296,
        0.942478,1.588901,2.301496,
        0.942478,1.589936,0.825020,
        0.942478,1.590018,-1.941652,
        0.942478,1.596968,-1.328358,
        0.942478,1.597178,-1.575607,
        0.942478,1.600470,-1.453196,
        0.942478,1.601408,0.242409,
        0.942478,1.602259,1.152656,
        0.942478,1.603064,2.865774,
        0.942478,1.612085,1.968802,
        0.942478,1.612720,-3.080498,
        0.942478,1.616061,2.521531,
        0.942478,1.616900,-0.207457,
        0.942478,1.620988,-0.875920,
        0.942478,1.622114,0.701509,
        0.942478,1.622749,0.121426,
        0.942478,1.622993,-2.221629,
        0.942478,1.629346,-0.542669,
        0.942478,1.630007,1.844552,
        0.942478,1.634747,-2.426198,
        0.942478,1.642346,1.722617,
        0.942478,1.643925,1.034629,
        0.942478,1.644848,2.743774,
        0.942478,1.645258,2.188109,
        0.942478,1.645417,-2.634725,
        0.942478,1.647395,3.085437,
        0.942478,1.648538,-2.855044,
        0.942478,1.649932,0.576808,
        0.942478,1.652224,-0.996105,
        0.942478,1.652871,1.600681,
        0.942478,1.657997,-0.325127,
        0.942478,1.661936,0.001037,
        0.942478,1.665187,1.477402,
        0.942478,1.665496,2.397872,
        0.942478,1.673267,-0.665453,
        0.942478,1.673557,0.452526,
        0.942478,1.679277,-1.122421,
        0.942478,1.679696,0.912482,
        0.942478,1.682420,1.353922,
        0.942478,1.685031,-2.009038,
        0.942478,1.685196,2.967046,
        0.942478,1.690107,2.620910,
        0.942478,1.690453,-2.123076,
        0.942478,1.691693,-1.871572,
        0.942478,1.691817,2.061688,
        0.942478,1.692644,1.237670,
        0.942478,1.696386,0.328123,
        0.942478,1.697162,-1.249026,
        0.942478,1.697488,-2.322785,
        0.942478,1.698726,-2.973576,
        0.942478,1.699172,-1.749717,
        0.942478,1.704478,-0.106295,
        0.942478,1.707244,-0.444770,
        0.942478,1.707500,-2.531497,
        0.942478,1.707714,-0.787321,
        0.942478,1.708430,-1.627552,
        0.942478,1.710583,-1.376890,
        0.942478,1.712415,-2.747921,
        0.942478,1.713125,-1.503073,
        0.942478,1.713452,0.789419,
        0.942478,1.720612,2.279888,
        0.942478,1.724852,0.202831,
        0.942478,1.724868,2.846246,
        0.942478,1.728940,1.931535,
        0.942478,1.732838,1.127584,
        0.942478,1.735719,-3.092485,
        0.942478,1.739676,2.496017,
        0.942478,1.739971,-0.228876,
        0.942478,1.743328,0.663347,
        0.942478,1.744927,-0.911892,
        0.942478,1.745952,0.084733,
        0.942478,1.748077,1.804354,
        0.942478,1.752959,-0.567548,
        0.942478,1.760169,1.678132,
        0.942478,1.763887,-2.218450,
        0.942478,1.766737,2.722752,
        0.942478,1.767999,1.004737,
        0.942478,1.769149,1.550298,
        0.942478,1.770119,-2.427175,
        0.942478,1.770460,2.154208,
        0.942478,1.770514,3.070982,
        0.942478,1.770735,0.537245,
        0.942478,1.772673,-2.864760,
        0.942478,1.773527,-1.039582,
        0.942478,1.776419,-2.642421,
        0.942478,1.780769,1.416330,
        0.942478,1.786634,-0.346247,
        0.942478,1.788829,-1.947553,
        0.942478,1.793737,0.410270,
        0.942478,1.794553,2.374618,
        0.942478,1.794699,-0.694061,
        0.942478,1.797469,-1.169105,
        0.942478,1.803705,0.881076,
        0.942478,1.804124,1.264249,
        0.942478,1.807825,-2.090455,
        0.942478,1.808822,-1.817375,
        0.942478,1.809020,2.949741,
        0.942478,1.812466,2.597595,
        0.942478,1.813159,2.025395,
        0.942478,1.813816,-1.298435,
        0.942478,1.815016,0.278951,
        0.942478,1.816510,-0.008631,
        0.942478,1.817772,-1.689625,
        0.942478,1.821360,-2.984352,
        0.942478,1.822312,-0.127211,
        0.942478,1.823386,-1.429807,
        0.942478,1.823834,-1.560483,
        0.942478,1.832060,-0.821627,
        0.942478,1.832690,-0.468440,
        0.942478,1.833945,-2.320278,
        0.942478,1.834787,0.751968,
        0.942478,1.840579,-2.536445,
        0.942478,1.841033,-2.756932,
        0.942478,1.846350,1.893983,
        0.942478,1.846644,1.113766,
        0.942478,1.846833,2.826543,
        0.942478,1.847325,2.249051,
        0.942478,1.854119,0.134161,
        0.942478,1.858863,-3.106022,
        0.942478,1.865046,1.763384,
        0.942478,1.865619,-0.952036,
        0.942478,1.865912,0.624453,
        0.942478,1.866694,2.473686,
        0.942478,1.867524,-0.240083,
        0.942478,1.875561,1.632831,
        0.942478,1.876049,-0.594378,
        0.942478,1.880509,1.500204,
        0.942478,1.883504,1.360457,
        0.942478,1.886971,-2.196665,
        0.942478,1.888979,0.982922,
        0.942478,1.889815,2.701570,
        0.942478,1.892144,-2.005340,
        0.942478,1.892348,0.496522,
        0.942478,1.893281,-1.083976,
        0.942478,1.893828,3.053453,
        0.942478,1.895117,2.121107,
        0.942478,1.898741,-2.875735,
        0.942478,1.904429,-2.428029,
        0.942478,1.908408,-2.648903,
        0.942478,1.911828,-0.364489,
        0.942478,1.912224,0.366565,
        0.942478,1.914854,1.214468,
        0.942478,1.915467,-1.883143,
        0.942478,1.915763,-0.724679,
        0.942478,1.915790,-1.216366,
        0.942478,1.922256,2.347806,
        0.942478,1.922616,0.844633,
        0.942478,1.925185,0.231170,
        0.942478,1.927474,-1.754616,
        0.942478,1.928342,-1.351174,
        0.942478,1.933328,-1.623328,
        0.942478,1.933878,0.034150,
        0.942478,1.934585,-1.487979,
        0.942478,1.934802,2.927634,
        0.942478,1.934862,1.989776,
        0.942478,1.936327,-0.099350,
        0.942478,1.938088,2.578507,
        0.942478,1.945677,-2.999805,
        0.942478,1.954572,-0.859208,
        0.942478,1.957868,-0.490784,
        0.942478,1.958989,0.712772,
        0.942478,1.960212,-2.307561,
        0.942478,1.964774,1.855507,
        0.942478,1.964971,-2.100533,
        0.942478,1.969112,1.088273,
        0.942478,1.971438,-2.765415,
        0.942478,1.972312,2.805134,
        0.942478,1.974950,2.220596,
        0.942478,1.975135,-2.539127,
        0.942478,1.982107,1.720829,
        0.942478,1.983195,-3.127984,
        0.942478,1.983318,-0.247182,
        0.942478,1.986720,-0.993220,
        0.942478,1.990381,1.586005,
        0.942478,1.990470,0.580283,
        0.942478,1.990523,1.450326,
        0.942478,1.993909,1.312897,
        0.942478,1.994694,2.452813,
        0.942478,1.999226,-0.620717,
        0.942478,2.005145,0.941012,
        0.942478,2.007301,0.142337,
        0.942478,2.014149,0.444138,
        0.942478,2.014338,-1.127392,
        0.942478,2.018131,3.021778,
        0.942478,2.019676,-1.980944,
        0.942478,2.019912,2.686612,
        0.942478,2.020545,2.088672,
        0.942478,2.026074,0.300945,
        0.942478,2.026196,-2.889057,
        0.942478,2.031959,-1.268178,
        0.942478,2.032917,-2.199917,
        0.942478,2.033572,-1.842884,
        0.942478,2.034712,-2.419505,
        0.942478,2.035818,-0.379479,
        0.942478,2.036004,-0.757062,
        0.942478,2.040501,-1.411488,
        0.942478,2.042874,-1.701367,
        0.942478,2.043060,-2.653746,
        0.942478,2.043729,-1.557938,
        0.942478,2.048538,0.806623,
        0.942478,2.049006,1.199930,
        0.942478,2.049502,-0.137989,
        0.942478,2.052181,2.326959,
        0.942478,2.052794,-0.008526,
        0.942478,2.057608,1.951726,
        0.942478,2.065568,2.567730,
        0.942478,2.070371,2.879808,
        0.942478,2.071256,-3.018961,
        0.942478,2.079746,-0.897699,
        0.942478,2.082784,1.042403,
        0.942478,2.083752,-0.511086,
        0.942478,2.084401,1.810502,
        0.942478,2.085731,0.671081,
        0.942478,2.097496,1.373172,
        0.942478,2.099185,1.667006,
        0.942478,2.100482,-2.301084,
        0.942478,2.103118,2.194745,
        0.942478,2.103223,1.521685,
        0.942478,2.103695,-2.777707,
        0.942478,2.107523,3.129581,
        0.942478,2.108277,-2.095907,
        0.942478,2.108891,-2.532166,
        0.942478,2.111003,-0.268632,
        0.942478,2.112276,-1.027246,
        0.942478,2.114380,0.531457,
        0.942478,2.119166,0.100552,
        0.942478,2.119564,2.754818,
        0.942478,2.122474,-0.648314,
        0.942478,2.125183,2.445598,
        0.942478,2.129501,0.238290,
        0.942478,2.131178,0.386663,
        0.942478,2.131223,-1.174311,
        0.942478,2.134269,-1.948042,
        0.942478,2.134684,0.907065,
        0.942478,2.137615,2.988806,
        0.942478,2.145635,-1.327708,
        0.942478,2.146467,2.055081,
        0.942478,2.146582,-1.793788,
        0.942478,2.149210,-1.482321,
        0.942478,2.152901,-1.636788,
        0.942478,2.156337,1.151558,
        0.942478,2.156474,-2.912065,
        0.942478,2.157194,-0.792563,
        0.942478,2.166724,-0.401638,
        0.942478,2.166980,-0.134879,
        0.942478,2.173644,-2.403429,
        0.942478,2.175711,-2.655523,
        0.942478,2.177132,2.600419,
        0.942478,2.178879,0.767722,
        0.942478,2.180752,1.908485,
        0.942478,2.181338,2.313721,
        0.942478,2.185580,1.287673,
        0.942478,2.195400,-2.217546,
        0.942478,2.200418,-3.056442,
        0.942478,2.201829,-0.001735,
        0.942478,2.203804,1.756176,
        0.942478,2.211602,1.443930,
        0.942478,2.212686,-0.544809,
        0.942478,2.213937,0.621892,
        0.942478,2.214075,1.600691,
        0.942478,2.216256,1.020029,
        0.942478,2.219060,2.886631,
        0.942478,2.223375,-0.906185,
        0.942478,2.226002,-1.060248,
        0.942478,2.226413,-2.064034,
        0.942478,2.230554,2.171334,
        0.942478,2.232413,2.735777,
        0.942478,2.233709,3.070830,
        0.942478,2.234383,-2.792065,
        0.942478,2.237183,0.468799,
        0.942478,2.237695,-0.271388,
        0.942478,2.239419,0.151693,
        0.942478,2.243347,-2.521049,
        0.942478,2.244035,-1.229456,
        0.942478,2.244355,0.309899,
        0.942478,2.245175,2.460236,
        0.942478,2.245628,-1.899216,
        0.942478,2.245973,-0.702234,
        0.942478,2.255601,-1.728131,
        0.942478,2.256702,-1.401367,
        0.942478,2.261466,-1.557500,
        0.942478,2.267390,0.878499,
        0.942478,2.272099,2.018594,
        0.942478,2.278618,1.170386,
        0.942478,2.283304,-2.345584,
        0.942478,2.285175,-2.943090,
        0.942478,2.293000,-0.124451,
        0.942478,2.294058,-0.421070,
        0.942478,2.303283,1.855938,
        0.942478,2.304968,2.608978,
        0.942478,2.305347,2.311346,
        0.942478,2.307188,-2.656134,
        0.942478,2.307301,1.336200,
        0.942478,2.309655,0.726020,
        0.942478,2.317634,-2.185225,
        0.942478,2.321510,1.685664,
        0.942478,2.324376,-3.111006,
        0.942478,2.324731,1.511186,
        0.942478,2.329587,0.035498,
        0.942478,2.330158,2.977264,
        0.942478,2.335096,-0.588468,
        0.942478,2.336046,-1.115799,
        0.942478,2.338060,2.803613,
        0.942478,2.338083,-0.943400,
        0.942478,2.340210,0.562145,
        0.942478,2.341931,1.023926,
        0.942478,2.343265,-0.779061,
        0.942478,2.343434,-2.013692,
        0.942478,2.353029,-1.310529,
        0.942478,2.354868,0.211543,
        0.942478,2.355120,0.388593,
        0.942478,2.355165,-1.833173,
        0.942478,2.355462,2.151441,
        0.942478,2.362948,-2.487477,
        0.942478,2.363101,-2.810456,
        0.942478,2.363394,-0.272918,
        0.942478,2.364509,-1.644533,
        0.942478,2.370465,2.466824,
        0.942478,2.386110,-1.481351,
        0.942478,2.394960,0.862133,
        0.942478,2.395716,1.977184,
        0.942478,2.396339,1.200900,
        0.942478,2.407941,-2.316753,
        0.942478,2.409450,-2.988432,
        0.942478,2.417202,-0.103363,
        0.942478,2.417386,-0.443097,
        0.942478,2.417950,2.666662,
        0.942478,2.423190,1.789783,
        0.942478,2.423692,1.392696,
        0.942478,2.431122,3.086379,
        0.942478,2.431406,-2.647820,
        0.942478,2.431584,2.307631,
        0.942478,2.434090,1.592818,
        0.942478,2.435993,0.683844,
        0.942478,2.440898,-2.133300,
        0.942478,2.442222,-1.195837,
        0.942478,2.443112,2.883761,
        0.942478,2.449191,-0.633651,
        0.942478,2.450821,0.084741,
        0.942478,2.453499,-1.008861,
        0.942478,2.460010,-1.938011,
        0.942478,2.460199,-0.822008,
        0.942478,2.460902,0.490079,
        0.942478,2.461774,-1.734913,
        0.942478,2.463986,1.034407,
        0.942478,2.467647,0.287667,
        0.942478,2.479849,2.126925,
        0.942478,2.485907,-2.835471,
        0.942478,2.485945,-1.374853,
        0.942478,2.487747,2.505912,
        0.942478,2.487957,-0.274279,
        0.942478,2.489741,-2.468428,
        0.942478,2.512538,1.239900,
        0.942478,2.517528,1.923289,
        0.942478,2.518454,-1.565852,
        0.942478,2.519348,0.843503,
        0.942478,2.522766,-3.059231,
        0.942478,2.528686,2.737228,
        0.942478,2.534798,-2.266671,
        0.942478,2.536533,1.466274,
        0.942478,2.538504,1.699812,
        0.942478,2.539516,-0.476942,
        0.942478,2.539532,-0.074548,
        0.942478,2.543197,2.990252,
        0.942478,2.554022,2.320619,
        0.942478,2.556919,-2.655005,
        0.942478,2.558146,0.627191,
        0.942478,2.558223,-1.164189,
        0.942478,2.563452,-2.044028,
        0.942478,2.568462,0.150655,
        0.942478,2.569130,-0.700837,
        0.942478,2.570050,-1.804492,
        0.942478,2.572710,-0.934665,
        0.942478,2.575994,0.391718,
        0.942478,2.584774,1.046703,
        0.942478,2.601432,2.102535,
        0.942478,2.603559,2.555114,
        0.942478,2.604150,-2.885208,
        0.942478,2.608630,-1.378011,
        0.942478,2.611256,-0.279110,
        0.942478,2.617344,-2.439480,
        0.942478,2.627402,1.293552,
        0.942478,2.635396,1.848850,
        0.942478,2.635686,2.834950,
        0.942478,2.637087,3.132730,
        0.942478,2.642029,0.813274,
        0.942478,2.645260,1.570203,
        0.942478,2.646182,-1.614036,
        0.942478,2.660113,-2.186513,
        0.942478,2.660483,-0.034608,
        0.942478,2.661713,-0.529511,
        0.942478,2.672720,-1.110422,
        0.942478,2.676626,2.335353,
        0.942478,2.676638,0.540364,
        0.942478,2.676780,-1.901158,
        0.942478,2.681411,-2.667965,
        0.942478,2.682152,0.245751,
        0.942478,2.682859,-0.817071,
        0.942478,2.705107,1.062142,
        0.942478,2.717705,2.625995,
        0.942478,2.721001,2.060097,
        0.942478,2.722319,-2.954069,
        0.942478,2.729840,-1.368237,
        0.942478,2.734349,-0.289298,
        0.942478,2.738409,2.975671,
        0.942478,2.740366,1.374341,
        0.942478,2.744633,-2.391311,
        0.942478,2.747398,1.727166,
        0.942478,2.763799,0.760372,
        0.942478,2.768238,-1.679243,
        0.942478,2.779245,-2.046268,
        0.942478,2.780208,-0.626453,
        0.942478,2.780471,0.027732,
        0.942478,2.784696,-1.013128,
        0.942478,2.789817,0.395431,
        0.942478,2.801154,2.353143,
        0.942478,2.807091,-2.687477,
        0.942478,2.825762,1.085035,
        0.942478,2.830209,2.734943,
        0.942478,2.833517,-3.085508,
        0.942478,2.838301,1.980061,
        0.942478,2.849871,1.512012,
        0.942478,2.850089,-1.343899,
        0.942478,2.856905,-0.318441,
        0.942478,2.870445,-2.291001,
        0.942478,2.885017,0.655562,
        0.942478,2.886428,-1.793217,
        0.942478,2.890337,-0.825633,
        0.942478,2.898932,0.139973,
        0.942478,2.931927,-2.728768,
        0.942478,2.932096,2.377790,
        0.942478,2.937409,2.947245,
        0.942478,2.948048,1.133127,
        0.942478,2.950896,1.797889,
        0.942478,2.970019,-1.282382,
        0.942478,2.978333,-0.412090,
        0.942478,2.996437,-2.089121,
        0.942478,3.011274,0.434110,
        0.942478,3.050410,3.138492,
        0.942478,3.064492,1.647259,
        0.942478,3.088868,-0.989501,
        1.047198,0.053884,-2.408394,
        1.047198,0.077645,1.453433,
        1.047198,0.088408,-0.311160,
        1.047198,0.142314,2.590387,
        1.047198,0.147206,-1.385164,
        1.047198,0.163506,0.573203,
        1.047198,0.165485,-2.875396,
        1.047198,0.182564,-2.125821,
        1.047198,0.190569,1.929141,
        1.047198,0.201102,1.249082,
        1.047198,0.201397,-0.716646,
        1.047198,0.210568,-0.069559,
        1.047198,0.251672,2.848900,
        1.047198,0.266262,-1.678068,
        1.047198,0.269108,-1.215563,
        1.047198,0.273777,-2.506756,
        1.047198,0.279854,0.814581,
        1.047198,0.281664,-2.969147,
        1.047198,0.283625,2.405603,
        1.047198,0.284058,0.357350,
        1.047198,0.294107,1.593880,
        1.047198,0.302491,-0.400410,
        1.047198,0.308059,2.021491,
        1.047198,0.312219,-2.072057,
        1.047198,0.321905,-0.829121,
        1.047198,0.329984,1.201621,
        1.047198,0.339714,-0.013712,
        1.047198,0.368784,2.974974,
        1.047198,0.376428,-1.495532,
        1.047198,0.378103,2.646436,
        1.047198,0.385088,0.590355,
        1.047198,0.386248,-2.690943,
        1.047198,0.389851,-2.340031,
        1.047198,0.391502,-1.171374,
        1.047198,0.398249,-1.812694,
        1.047198,0.400542,-2.995729,
        1.047198,0.404273,0.918163,
        1.047198,0.405617,1.773533,
        1.047198,0.409009,-0.570243,
        1.047198,0.410700,0.270433,
        1.047198,0.415893,1.464890,
        1.047198,0.420680,2.088164,
        1.047198,0.424766,-0.258729,
        1.047198,0.426507,2.367872,
        1.047198,0.441292,-0.890727,
        1.047198,0.451839,-2.078193,
        1.047198,0.459270,1.191562,
        1.047198,0.470622,0.014002,
        1.047198,0.481292,-2.529098,
        1.047198,0.482460,2.849843,
        1.047198,0.492205,-1.397963,
        1.047198,0.493314,3.117264,
        1.047198,0.497870,0.727828,
        1.047198,0.499783,-1.654379,
        1.047198,0.502651,0.463333,
        1.047198,0.505463,2.594557,
        1.047198,0.506104,-2.874910,
        1.047198,0.513579,-1.147354,
        1.047198,0.518569,1.644237,
        1.047198,0.519001,1.900488,
        1.047198,0.519277,-0.673809,
        1.047198,0.524769,-0.421955,
        1.047198,0.525564,-2.288838,
        1.047198,0.532541,0.978737,
        1.047198,0.534610,2.154808,
        1.047198,0.535020,-1.891846,
        1.047198,0.540208,1.400497,
        1.047198,0.542006,0.228337,
        1.047198,0.550762,-0.180771,
        1.047198,0.560531,2.378474,
        1.047198,0.561629,-0.924576,
        1.047198,0.567465,-2.679392,
        1.047198,0.588527,1.181755,
        1.047198,0.591643,-2.098364,
        1.047198,0.594963,2.997575,
        1.047198,0.599643,-3.045235,
        1.047198,0.599962,2.773556,
        1.047198,0.602620,0.029270,
        1.047198,0.605711,0.598454,
        1.047198,0.607708,-1.541026,
        1.047198,0.609360,-2.464379,
        1.047198,0.610738,-1.326786,
        1.047198,0.616968,0.820489,
        1.047198,0.626379,1.773610,
        1.047198,0.628626,-1.751180,
        1.047198,0.630671,0.391949,
        1.047198,0.632622,-0.737642,
        1.047198,0.632708,2.566722,
        1.047198,0.632797,1.989858,
        1.047198,0.634455,-0.534790,
        1.047198,0.636205,1.561170,
        1.047198,0.637681,-1.123423,
        1.047198,0.641441,-2.842109,
        1.047198,0.644807,-0.327139,
        1.047198,0.652440,2.197129,
        1.047198,0.660897,-2.278593,
        1.047198,0.664008,1.360091,
        1.047198,0.670850,-1.945257,
        1.047198,0.671396,1.004859,
        1.047198,0.673985,0.203855,
        1.047198,0.680326,-0.133370,
        1.047198,0.685083,-0.942363,
        1.047198,0.688204,-2.632204,
        1.047198,0.689777,2.385011,
        1.047198,0.700065,3.115430,
        1.047198,0.704080,2.911636,
        1.047198,0.712883,1.174216,
        1.047198,0.715701,0.699906,
        1.047198,0.718851,-1.442215,
        1.047198,0.721019,2.721536,
        1.047198,0.725300,-2.975134,
        1.047198,0.727136,-1.633288,
        1.047198,0.727399,-2.120567,
        1.047198,0.730751,0.522488,
        1.047198,0.732892,-1.254330,
        1.047198,0.734333,-2.446184,
        1.047198,0.735125,0.037270,
        1.047198,0.737893,1.866463,
        1.047198,0.740650,1.681075,
        1.047198,0.745971,-0.781286,
        1.047198,0.749432,-0.442559,
        1.047198,0.750102,2.050636,
        1.047198,0.751046,0.863821,
        1.047198,0.757018,-0.608833,
        1.047198,0.757046,-1.816019,
        1.047198,0.757389,1.501184,
        1.047198,0.759800,2.546181,
        1.047198,0.759859,0.351923,
        1.047198,0.761141,-2.781102,
        1.047198,0.769035,-1.078574,
        1.047198,0.770166,-0.266604,
        1.047198,0.772218,2.227380,
        1.047198,0.788382,1.331051,
        1.047198,0.790269,-2.282409,
        1.047198,0.803410,-1.984567,
        1.047198,0.803813,0.191520,
        1.047198,0.806832,3.022155,
        1.047198,0.807476,-2.597675,
        1.047198,0.812512,-0.107335,
        1.047198,0.814161,1.010949,
        1.047198,0.817732,-3.084486,
        1.047198,0.817846,2.390688,
        1.047198,0.819323,2.845960,
        1.047198,0.819430,-0.915017,
        1.047198,0.829198,-1.532491,
        1.047198,0.829531,0.665695,
        1.047198,0.829816,1.167147,
        1.047198,0.833613,-1.357492,
        1.047198,0.842945,-2.907318,
        1.047198,0.844883,2.679708,
        1.047198,0.848189,-1.705003,
        1.047198,0.849888,1.767901,
        1.047198,0.851126,0.493370,
        1.047198,0.852911,1.933394,
        1.047198,0.857516,-2.432863,
        1.047198,0.858261,-1.191729,
        1.047198,0.858279,-2.138895,
        1.047198,0.859723,-0.734928,
        1.047198,0.859754,1.603292,
        1.047198,0.861655,-0.394821,
        1.047198,0.864300,0.043549,
        1.047198,0.866940,0.851229,
        1.047198,0.869748,2.095969,
        1.047198,0.877352,-0.574668,
        1.047198,0.879254,-2.733338,
        1.047198,0.882554,1.443757,
        1.047198,0.884060,0.334199,
        1.047198,0.885626,-1.866061,
        1.047198,0.889834,2.529957,
        1.047198,0.893140,2.251698,
        1.047198,0.897463,-0.240822,
        1.047198,0.898758,-1.031948,
        1.047198,0.913918,1.287618,
        1.047198,0.916555,-2.287173,
        1.047198,0.918305,3.107845,
        1.047198,0.919522,2.948277,
        1.047198,0.927138,-2.569686,
        1.047198,0.929031,0.185383,
        1.047198,0.933766,-3.015917,
        1.047198,0.934873,-2.010822,
        1.047198,0.935636,-1.445855,
        1.047198,0.936702,-0.866461,
        1.047198,0.937395,0.729108,
        1.047198,0.937910,2.787575,
        1.047198,0.939144,-0.092962,
        1.047198,0.941718,1.125085,
        1.047198,0.942873,-1.607680,
        1.047198,0.943701,0.980328,
        1.047198,0.947774,2.395256,
        1.047198,0.952433,0.584347,
        1.047198,0.954456,-1.287293,
        1.047198,0.958731,-2.852552,
        1.047198,0.961838,1.839284,
        1.047198,0.961898,-0.469425,
        1.047198,0.965426,1.686615,
        1.047198,0.970108,1.990771,
        1.047198,0.972820,-1.762780,
        1.047198,0.973500,-0.711881,
        1.047198,0.975253,2.641082,
        1.047198,0.977824,0.441928,
        1.047198,0.981485,-2.424766,
        1.047198,0.981877,1.537100,
        1.047198,0.982609,-1.136920,
        1.047198,0.984265,-2.156073,
        1.047198,0.989150,0.045493,
        1.047198,0.993052,2.137478,
        1.047198,0.997766,-2.693871,
        1.047198,0.999847,-0.334738,
        1.047198,1.001920,0.848054,
        1.047198,1.006146,1.389734,
        1.047198,1.012180,0.304066,
        1.047198,1.016938,2.270674,
        1.047198,1.017371,-0.979790,
        1.047198,1.019495,-1.894487,
        1.047198,1.023796,-0.583691,
        1.047198,1.026525,2.506945,
        1.047198,1.027004,-0.203653,
        1.047198,1.028027,3.036573,
        1.047198,1.031435,1.241021,
        1.047198,1.032316,2.884221,
        1.047198,1.035863,-3.110475,
        1.047198,1.042085,-2.292950,
        1.047198,1.043272,-1.522986,
        1.047198,1.046265,-2.958847,
        1.047198,1.049272,-1.376344,
        1.047198,1.050039,-2.552086,
        1.047198,1.053682,0.689863,
        1.047198,1.055458,0.170394,
        1.047198,1.056622,-0.831866,
        1.047198,1.057515,1.097998,
        1.047198,1.058706,-2.036438,
        1.047198,1.061389,-1.672248,
        1.047198,1.063758,2.743093,
        1.047198,1.070157,-0.073422,
        1.047198,1.072154,1.761506,
        1.047198,1.074029,1.909193,
        1.047198,1.074204,0.545276,
        1.047198,1.074714,-2.805147,
        1.047198,1.076262,0.966104,
        1.047198,1.077642,-1.227073,
        1.047198,1.084151,1.618525,
        1.047198,1.085085,2.377673,
        1.047198,1.086822,-0.441112,
        1.047198,1.087142,2.058815,
        1.047198,1.098383,-0.696430,
        1.047198,1.099138,-1.084033,
        1.047198,1.103208,1.477773,
        1.047198,1.103783,0.408254,
        1.047198,1.106275,-2.420800,
        1.047198,1.108079,2.612943,
        1.047198,1.109345,-2.172718,
        1.047198,1.115957,-1.787597,
        1.047198,1.118930,0.045983,
        1.047198,1.119777,0.834702,
        1.047198,1.121943,-2.669042,
        1.047198,1.124930,2.219917,
        1.047198,1.125217,1.336294,
        1.047198,1.130958,2.972931,
        1.047198,1.131400,-0.306629,
        1.047198,1.137135,-0.936514,
        1.047198,1.139125,-1.924528,
        1.047198,1.139877,0.276382,
        1.047198,1.142003,-3.058022,
        1.047198,1.147363,-0.558448,
        1.047198,1.150348,-0.176418,
        1.047198,1.151198,1.197016,
        1.047198,1.153930,-1.597104,
        1.047198,1.155389,2.837148,
        1.047198,1.155458,-1.460623,
        1.047198,1.157331,-2.906362,
        1.047198,1.158594,2.487618,
        1.047198,1.162103,-1.322553,
        1.047198,1.165407,3.092861,
        1.047198,1.168025,0.717075,
        1.047198,1.168406,-2.301519,
        1.047198,1.174782,-2.538755,
        1.047198,1.177998,1.061256,
        1.047198,1.178898,1.827305,
        1.047198,1.181331,-2.060330,
        1.047198,1.182540,-0.798952,
        1.047198,1.184569,0.147517,
        1.047198,1.184734,1.968919,
        1.047198,1.184804,0.593386,
        1.047198,1.188278,1.690808,
        1.047198,1.194360,2.110440,
        1.047198,1.194776,2.710455,
        1.047198,1.197745,-1.173705,
        1.047198,1.199898,-2.777257,
        1.047198,1.201164,2.355931,
        1.047198,1.203760,1.556404,
        1.047198,1.205845,-0.415993,
        1.047198,1.206585,0.933648,
        1.047198,1.206950,-0.037248,
        1.047198,1.209341,0.471162,
        1.047198,1.213483,-1.039417,
        1.047198,1.218146,-0.672643,
        1.047198,1.221608,1.421669,
        1.047198,1.226043,-1.693533,
        1.047198,1.232870,-1.822773,
        1.047198,1.234375,-2.421194,
        1.047198,1.234531,-2.190148,
        1.047198,1.239414,0.348959,
        1.047198,1.240183,2.587255,
        1.047198,1.242312,2.941234,
        1.047198,1.244075,1.286644,
        1.047198,1.245961,-2.997461,
        1.047198,1.246495,-2.646872,
        1.047198,1.247736,2.242810,
        1.047198,1.253604,0.812222,
        1.047198,1.257816,-0.894317,
        1.047198,1.258247,-1.953634,
        1.047198,1.258993,-0.284099,
        1.047198,1.262380,-0.157609,
        1.047198,1.262791,-1.549229,
        1.047198,1.263772,-0.533073,
        1.047198,1.266731,-1.412461,
        1.047198,1.268591,-3.122146,
        1.047198,1.269614,-1.275246,
        1.047198,1.270745,1.152533,
        1.047198,1.273606,0.224545,
        1.047198,1.279716,2.812101,
        1.047198,1.283786,2.462124,
        1.047198,1.284798,0.078209,
        1.047198,1.285402,-2.882049,
        1.047198,1.286936,1.886053,
        1.047198,1.289408,0.683122,
        1.047198,1.294736,1.756184,
        1.047198,1.296978,-2.311511,
        1.047198,1.298546,2.020601,
        1.047198,1.300924,1.022478,
        1.047198,1.303305,3.043152,
        1.047198,1.304210,-2.084225,
        1.047198,1.305241,-2.530249,
        1.047198,1.307513,1.628003,
        1.047198,1.310722,0.558267,
        1.047198,1.314142,2.139329,
        1.047198,1.316087,-0.777971,
        1.047198,1.317907,-1.121023,
        1.047198,1.321019,-2.749776,
        1.047198,1.321609,2.687717,
        1.047198,1.322568,1.499415,
        1.047198,1.322972,-0.395699,
        1.047198,1.325286,-0.994192,
        1.047198,1.329146,-0.648513,
        1.047198,1.329795,-0.042570,
        1.047198,1.331651,2.344497,
        1.047198,1.336522,0.435098,
        1.047198,1.339410,0.900087,
        1.047198,1.339550,1.370353,
        1.047198,1.340441,-1.731580,
        1.047198,1.351855,-1.855234,
        1.047198,1.360161,-3.018404,
        1.047198,1.361031,-1.613971,
        1.047198,1.361196,2.914780,
        1.047198,1.361998,-2.207254,
        1.047198,1.362302,1.239384,
        1.047198,1.365228,2.563808,
        1.047198,1.365452,0.310522,
        1.047198,1.367028,-2.420041,
        1.047198,1.373422,-1.484281,
        1.047198,1.374245,-1.226417,
        1.047198,1.375020,-0.170393,
        1.047198,1.376560,-1.355481,
        1.047198,1.377070,-1.980984,
        1.047198,1.377578,-2.636339,
        1.047198,1.378096,0.774626,
        1.047198,1.379086,-0.508192,
        1.047198,1.382562,0.177141,
        1.047198,1.388849,2.233847,
        1.047198,1.389268,-0.886349,
        1.047198,1.391324,1.109396,
        1.047198,1.391699,-0.288725,
        1.047198,1.392869,3.136903,
        1.047198,1.393612,1.952016,
        1.047198,1.395545,-2.850807,
        1.047198,1.401926,1.822930,
        1.047198,1.402149,2.788898,
        1.047198,1.409552,0.647980,
        1.047198,1.410840,0.056219,
        1.047198,1.411944,2.444377,
        1.047198,1.413639,1.697982,
        1.047198,1.426479,1.573427,
        1.047198,1.426725,2.102242,
        1.047198,1.427868,0.984637,
        1.047198,1.429580,-2.107232,
        1.047198,1.431375,-2.313269,
        1.047198,1.431907,3.015410,
        1.047198,1.432325,-1.029054,
        1.047198,1.433341,0.524268,
        1.047198,1.435181,-0.739771,
        1.047198,1.437208,-2.529085,
        1.047198,1.439041,-0.617966,
        1.047198,1.441473,1.448544,
        1.047198,1.444527,2.664744,
        1.047198,1.450513,-0.400625,
        1.047198,1.451907,-0.066823,
        1.047198,1.452754,-2.738631,
        1.047198,1.454061,-2.950416,
        1.047198,1.456736,1.323154,
        1.047198,1.458370,0.400433,
        1.047198,1.459314,-1.765003,
        1.047198,1.459902,-1.145835,
        1.047198,1.464952,2.329245,
        1.047198,1.466050,0.862104,
        1.047198,1.471330,-1.886160,
        1.047198,1.475119,-1.648255,
        1.047198,1.480878,1.194576,
        1.047198,1.480942,0.273235,
        1.047198,1.481086,2.890269,
        1.047198,1.484118,-1.529156,
        1.047198,1.484741,-1.279219,
        1.047198,1.487681,-1.406069,
        1.047198,1.487886,-3.066037,
        1.047198,1.490555,2.543284,
        1.047198,1.492424,-0.190932,
        1.047198,1.495002,-0.839810,
        1.047198,1.495297,2.007188,
        1.047198,1.496414,-2.009267,
        1.047198,1.498261,-2.425270,
        1.047198,1.498907,-2.207327,
        1.047198,1.500204,0.737638,
        1.047198,1.502469,0.151156,
        1.047198,1.507448,-0.514782,
        1.047198,1.509287,2.204647,
        1.047198,1.510518,1.884920,
        1.047198,1.511948,-2.632360,
        1.047198,1.515235,1.068565,
        1.047198,1.522237,3.102018,
        1.047198,1.522373,1.764991,
        1.047198,1.522416,-0.954956,
        1.047198,1.523102,2.765920,
        1.047198,1.524811,-2.844718,
        1.047198,1.527261,-0.304916,
        1.047198,1.529788,0.613208,
        1.047198,1.533648,1.644808,
        1.047198,1.535018,0.028869,
        1.047198,1.541082,2.424608,
        1.047198,1.546010,1.523462,
        1.047198,1.551135,-0.642994,
        1.047198,1.553895,0.489607,
        1.047198,1.555041,0.947813,
        1.047198,1.559671,-1.074934,
        1.047198,1.560345,2.985899,
        1.047198,1.560487,-2.323400,
        1.047198,1.561127,1.401111,
        1.047198,1.567193,2.643172,
        1.047198,1.568892,2.098362,
        1.047198,1.571702,-2.098435,
        1.047198,1.572701,-2.528730,
        1.047198,1.573572,1.278593,
        1.047198,1.574849,-0.091505,
        1.047198,1.576334,0.365236,
        1.047198,1.577029,-2.962888,
        1.047198,1.577688,-1.814941,
        1.047198,1.579542,-0.755917,
        1.047198,1.579624,-1.199759,
        1.047198,1.581515,-0.419414,
        1.047198,1.584581,-2.738702,
        1.047198,1.586948,-1.696296,
        1.047198,1.588901,2.301496,
        1.047198,1.589936,0.825020,
        1.047198,1.590018,-1.941652,
        1.047198,1.596968,-1.328358,
        1.047198,1.597178,-1.575607,
        1.047198,1.600470,-1.453196,
        1.047198,1.601408,0.242409,
        1.047198,1.602259,1.152656,
        1.047198,1.603064,2.865774,
        1.047198,1.612085,1.968802,
        1.047198,1.612720,-3.080498,
        1.047198,1.616061,2.521531,
        1.047198,1.616900,-0.207457,
        1.047198,1.620988,-0.875920,
        1.047198,1.622114,0.701509,
        1.047198,1.622749,0.121426,
        1.047198,1.622993,-2.221629,
        1.047198,1.629346,-0.542669,
        1.047198,1.630007,1.844552,
        1.047198,1.634747,-2.426198,
        1.047198,1.642346,1.722617,
        1.047198,1.643925,1.034629,
        1.047198,1.644848,2.743774,
        1.047198,1.645258,2.188109,
        1.047198,1.645417,-2.634725,
        1.047198,1.647395,3.085437,
        1.047198,1.648538,-2.855044,
        1.047198,1.649932,0.576808,
        1.047198,1.652224,-0.996105,
        1.047198,1.652871,1.600681,
        1.047198,1.657997,-0.325127,
        1.047198,1.661936,0.001037,
        1.047198,1.665187,1.477402,
        1.047198,1.665496,2.397872,
        1.047198,1.673267,-0.665453,
        1.047198,1.673557,0.452526,
        1.047198,1.679277,-1.122421,
        1.047198,1.679696,0.912482,
        1.047198,1.682420,1.353922,
        1.047198,1.685031,-2.009038,
        1.047198,1.685196,2.967046,
        1.047198,1.690107,2.620910,
        1.047198,1.690453,-2.123076,
        1.047198,1.691693,-1.871572,
        1.047198,1.691817,2.061688,
        1.047198,1.692644,1.237670,
        1.047198,1.696386,0.328123,
        1.047198,1.697162,-1.249026,
        1.047198,1.697488,-2.322785,
        1.047198,1.698726,-2.973576,
        1.047198,1.699172,-1.749717,
        1.047198,1.704478,-0.106295,
        1.047198,1.707244,-0.444770,
        1.047198,1.707500,-2.531497,
        1.047198,1.707714,-0.787321,
        1.047198,1.708430,-1.627552,
        1.047198,1.710583,-1.376890,
        1.047198,1.712415,-2.747921,
        1.047198,1.713125,-1.503073,
        1.047198,1.713452,0.789419,
        1.047198,1.720612,2.279888,
        1.047198,1.724852,0.202831,
        1.047198,1.724868,2.846246,
        1.047198,1.728940,1.931535,
        1.047198,1.732838,1.127584,
        1.047198,1.735719,-3.092485,
        1.047198,1.739676,2.496017,
        1.047198,1.739971,-0.228876,
        1.047198,1.743328,0.663347,
        1.047198,1.744927,-0.911892,
        1.047198,1.745952,0.084733,
        1.047198,1.748077,1.804354,
        1.047198,1.752959,-0.567548,
        1.047198,1.760169,1.678132,
        1.047198,1.763887,-2.218450,
        1.047198,1.766737,2.722752,
        1.047198,1.767999,1.004737,
        1.047198,1.769149,1.550298,
        1.047198,1.770119,-2.427175,
        1.047198,1.770460,2.154208,
        1.047198,1.770514,3.070982,
        1.047198,1.770735,0.537245,
        1.047198,1.772673,-2.864760,
        1.047198,1.773527,-1.039582,
        1.047198,1.776419,-2.642421,
        1.047198,1.780769,1.416330,
        1.047198,1.786634,-0.346247,
        1.047198,1.788829,-1.947553,
        1.047198,1.793737,0.410270,
        1.047198,1.794553,2.374618,
        1.047198,1.794699,-0.694061,
        1.047198,1.797469,-1.169105,
        1.047198,1.803705,0.881076,
        1.047198,1.804124,1.264249,
        1.047198,1.807825,-2.090455,
        1.047198,1.808822,-1.817375,
        1.047198,1.809020,2.949741,
        1.047198,1.812466,2.597595,
        1.047198,1.813159,2.025395,
        1.047198,1.813816,-1.298435,
        1.047198,1.815016,0.278951,
        1.047198,1.816510,-0.008631,
        1.047198,1.817772,-1.689625,
        1.047198,1.821360,-2.984352,
        1.047198,1.822312,-0.127211,
        1.047198,1.823386,-1.429807,
        1.047198,1.823834,-1.560483,
        1.047198,1.832060,-0.821627,
        1.047198,1.832690,-0.468440,
        1.047198,1.833945,-2.320278,
        1.047198,1.834787,0.751968,
        1.047198,1.840579,-2.536445,
        1.047198,1.841033,-2.756932,
        1.047198,1.846350,1.893983,
        1.047198,1.846644,1.113766,
        1.047198,1.846833,2.826543,
        1.047198,1.847325,2.249051,
        1.047198,1.854119,0.134161,
        1.047198,1.858863,-3.106022,
        1.047198,1.865046,1.763384,
        1.047198,1.865619,-0.952036,
        1.047198,1.865912,0.624453,
        1.047198,1.866694,2.473686,
        1.047198,1.867524,-0.240083,
        1.047198,1.875561,1.632831,
        1.047198,1.876049,-0.594378,
        1.047198,1.880509,1.500204,
        1.047198,1.883504,1.360457,
        1.047198,1.886971,-2.196665,
        1.047198,1.888979,0.982922,
        1.047198,1.889815,2.701570,
        1.047198,1.892144,-2.005340,
        1.047198,1.892348,0.496522,
        1.047198,1.893281,-1.083976,
        1.047198,1.893828,3.053453,
        1.047198,1.895117,2.121107,
        1.047198,1.898741,-2.875735,
        1.047198,1.904429,-2.428029,
        1.047198,1.908408,-2.648903,
        1.047198,1.911828,-0.364489,
        1.047198,1.912224,0.366565,
        1.047198,1.914854,1.214468,
        1.047198,1.915467,-1.883143,
        1.047198,1.915763,-0.724679,
        1.047198,1.915790,-1.216366,
        1.047198,1.922256,2.347806,
        1.047198,1.922616,0.844633,
        1.047198,1.925185,0.231170,
        1.047198,1.927474,-1.754616,
        1.047198,1.928342,-1.351174,
        1.047198,1.933328,-1.623328,
        1.047198,1.933878,0.034150,
        1.047198,1.934585,-1.487979,
        1.047198,1.934802,2.927634,
        1.047198,1.934862,1.989776,
        1.047198,1.936327,-0.099350,
        1.047198,1.938088,2.578507,
        1.047198,1.945677,-2.999805,
        1.047198,1.954572,-0.859208,
        1.047198,1.957868,-0.490784,
        1.047198,1.958989,0.712772,
        1.047198,1.960212,-2.307561,
        1.047198,1.964774,1.855507,
        1.047198,1.964971,-2.100533,
        1.047198,1.969112,1.088273,
        1.047198,1.971438,-2.765415,
        1.047198,1.972312,2.805134,
        1.047198,1.974950,2.220596,
        1.047198,1.975135,-2.539127,
        1.047198,1.982107,1.720829,
        1.047198,1.983195,-3.127984,
        1.047198,1.983318,-0.247182,
        1.047198,1.986720,-0.993220,
        1.047198,1.990381,1.586005,
        1.047198,1.990470,0.580283,
        1.047198,1.990523,1.450326,
        1.047198,1.993909,1.312897,
        1.047198,1.994694,2.452813,
        1.047198,1.999226,-0.620717,
        1.047198,2.005145,0.941012,
        1.047198,2.007301,0.142337,
        1.047198,2.014149,0.444138,
        1.047198,2.014338,-1.127392,
        1.047198,2.018131,3.021778,
        1.047198,2.019676,-1.980944,
        1.047198,2.019912,2.686612,
        1.047198,2.020545,2.088672,
        1.047198,2.026074,0.300945,
        1.047198,2.026196,-2.889057,
        1.047198,2.031959,-1.268178,
        1.047198,2.032917,-2.199917,
        1.047198,2.033572,-1.842884,
        1.047198,2.034712,-2.419505,
        1.047198,2.035818,-0.379479,
        1.047198,2.036004,-0.757062,
        1.047198,2.040501,-1.411488,
        1.047198,2.042874,-1.701367,
        1.047198,2.043060,-2.653746,
        1.047198,2.043729,-1.557938,
        1.047198,2.048538,0.806623,
        1.047198,2.049006,1.199930,
        1.047198,2.049502,-0.137989,
        1.047198,2.052181,2.326959,
        1.047198,2.052794,-0.008526,
        1.047198,2.057608,1.951726,
        1.047198,2.065568,2.567730,
        1.047198,2.070371,2.879808,
        1.047198,2.071256,-3.018961,
        1.047198,2.079746,-0.897699,
        1.047198,2.082784,1.042403,
        1.047198,2.083752,-0.511086,
        1.047198,2.084401,1.810502,
        1.047198,2.085731,0.671081,
        1.047198,2.097496,1.373172,
        1.047198,2.099185,1.667006,
        1.047198,2.100482,-2.301084,
        1.047198,2.103118,2.194745,
        1.047198,2.103223,1.521685,
        1.047198,2.103695,-2.777707,
        1.047198,2.107523,3.129581,
        1.047198,2.108277,-2.095907,
        1.047198,2.108891,-2.532166,
        1.047198,2.111003,-0.268632,
        1.047198,2.112276,-1.027246,
        1.047198,2.114380,0.531457,
        1.047198,2.119166,0.100552,
        1.047198,2.119564,2.754818,
        1.047198,2.122474,-0.648314,
        1.047198,2.125183,2.445598,
        1.047198,2.129501,0.238290,
        1.047198,2.131178,0.386663,
        1.047198,2.131223,-1.174311,
        1.047198,2.134269,-1.948042,
        1.047198,2.134684,0.907065,
        1.047198,2.137615,2.988806,
        1.047198,2.145635,-1.327708,
        1.047198,2.146467,2.055081,
        1.047198,2.146582,-1.793788,
        1.047198,2.149210,-1.482321,
        1.047198,2.152901,-1.636788,
        1.047198,2.156337,1.151558,
        1.047198,2.156474,-2.912065,
        1.047198,2.157194,-0.792563,
        1.047198,2.166724,-0.401638,
        1.047198,2.166980,-0.134879,
        1.047198,2.173644,-2.403429,
        1.047198,2.175711,-2.655523,
        1.047198,2.177132,2.600419,
        1.047198,2.178879,0.767722,
        1.047198,2.180752,1.908485,
        1.047198,2.181338,2.313721,
        1.047198,2.185580,1.287673,
        1.047198,2.195400,-2.217546,
        1.047198,2.200418,-3.056442,
        1.047198,2.201829,-0.001735,
        1.047198,2.203804,1.756176,
        1.047198,2.211602,1.443930,
        1.047198,2.212686,-0.544809,
        1.047198,2.213937,0.621892,
        1.047198,2.214075,1.600691,
        1.047198,2.216256,1.020029,
        1.047198,2.219060,2.886631,
        1.047198,2.223375,-0.906185,
        1.047198,2.226002,-1.060248,
        1.047198,2.226413,-2.064034,
        1.047198,2.230554,2.171334,
        1.047198,2.232413,2.735777,
        1.047198,2.233709,3.070830,
        1.047198,2.234383,-2.792065,
        1.047198,2.237183,0.468799,
        1.047198,2.237695,-0.271388,
        1.047198,2.239419,0.151693,
        1.047198,2.243347,-2.521049,
        1.047198,2.244035,-1.229456,
        1.047198,2.244355,0.309899,
        1.047198,2.245175,2.460236,
        1.047198,2.245628,-1.899216,
        1.047198,2.245973,-0.702234,
        1.047198,2.255601,-1.728131,
        1.047198,2.256702,-1.401367,
        1.047198,2.261466,-1.557500,
        1.047198,2.267390,0.878499,
        1.047198,2.272099,2.018594,
        1.047198,2.278618,1.170386,
        1.047198,2.283304,-2.345584,
        1.047198,2.285175,-2.943090,
        1.047198,2.293000,-0.124451,
        1.047198,2.294058,-0.421070,
        1.047198,2.303283,1.855938,
        1.047198,2.304968,2.608978,
        1.047198,2.305347,2.311346,
        1.047198,2.307188,-2.656134,
        1.047198,2.307301,1.336200,
        1.047198,2.309655,0.726020,
        1.047198,2.317634,-2.185225,
        1.047198,2.321510,1.685664,
        1.047198,2.324376,-3.111006,
        1.047198,2.324731,1.511186,
        1.047198,2.329587,0.035498,
        1.047198,2.330158,2.977264,
        1.047198,2.335096,-0.588468,
        1.047198,2.336046,-1.115799,
        1.047198,2.338060,2.803613,
        1.047198,2.338083,-0.943400,
        1.047198,2.340210,0.562145,
        1.047198,2.341931,1.023926,
        1.047198,2.343265,-0.779061,
        1.047198,2.343434,-2.013692,
        1.047198,2.353029,-1.310529,
        1.047198,2.354868,0.211543,
        1.047198,2.355120,0.388593,
        1.047198,2.355165,-1.833173,
        1.047198,2.355462,2.151441,
        1.047198,2.362948,-2.487477,
        1.047198,2.363101,-2.810456,
        1.047198,2.363394,-0.272918,
        1.047198,2.364509,-1.644533,
        1.047198,2.370465,2.466824,
        1.047198,2.386110,-1.481351,
        1.047198,2.394960,0.862133,
        1.047198,2.395716,1.977184,
        1.047198,2.396339,1.200900,
        1.047198,2.407941,-2.316753,
        1.047198,2.409450,-2.988432,
        1.047198,2.417202,-0.103363,
        1.047198,2.417386,-0.443097,
        1.047198,2.417950,2.666662,
        1.047198,2.423190,1.789783,
        1.047198,2.423692,1.392696,
        1.047198,2.431122,3.086379,
        1.047198,2.431406,-2.647820,
        1.047198,2.431584,2.307631,
        1.047198,2.434090,1.592818,
        1.047198,2.435993,0.683844,
        1.047198,2.440898,-2.133300,
        1.047198,2.442222,-1.195837,
        1.047198,2.443112,2.883761,
        1.047198,2.449191,-0.633651,
        1.047198,2.450821,0.084741,
        1.047198,2.453499,-1.008861,
        1.047198,2.460010,-1.938011,
        1.047198,2.460199,-0.822008,
        1.047198,2.460902,0.490079,
        1.047198,2.461774,-1.734913,
        1.047198,2.463986,1.034407,
        1.047198,2.467647,0.287667,
        1.047198,2.479849,2.126925,
        1.047198,2.485907,-2.835471,
        1.047198,2.485945,-1.374853,
        1.047198,2.487747,2.505912,
        1.047198,2.487957,-0.274279,
        1.047198,2.489741,-2.468428,
        1.047198,2.512538,1.239900,
        1.047198,2.517528,1.923289,
        1.047198,2.518454,-1.565852,
        1.047198,2.519348,0.843503,
        1.047198,2.522766,-3.059231,
        1.047198,2.528686,2.737228,
        1.047198,2.534798,-2.266671,
        1.047198,2.536533,1.466274,
        1.047198,2.538504,1.699812,
        1.047198,2.539516,-0.476942,
        1.047198,2.539532,-0.074548,
        1.047198,2.543197,2.990252,
        1.047198,2.554022,2.320619,
        1.047198,2.556919,-2.655005,
        1.047198,2.558146,0.627191,
        1.047198,2.558223,-1.164189,
        1.047198,2.563452,-2.044028,
        1.047198,2.568462,0.150655,
        1.047198,2.569130,-0.700837,
        1.047198,2.570050,-1.804492,
        1.047198,2.572710,-0.934665,
        1.047198,2.575994,0.391718,
        1.047198,2.584774,1.046703,
        1.047198,2.601432,2.102535,
        1.047198,2.603559,2.555114,
        1.047198,2.604150,-2.885208,
        1.047198,2.608630,-1.378011,
        1.047198,2.611256,-0.279110,
        1.047198,2.617344,-2.439480,
        1.047198,2.627402,1.293552,
        1.047198,2.635396,1.848850,
        1.047198,2.635686,2.834950,
        1.047198,2.637087,3.132730,
        1.047198,2.642029,0.813274,
        1.047198,2.645260,1.570203,
        1.047198,2.646182,-1.614036,
        1.047198,2.660113,-2.186513,
        1.047198,2.660483,-0.034608,
        1.047198,2.661713,-0.529511,
        1.047198,2.672720,-1.110422,
        1.047198,2.676626,2.335353,
        1.047198,2.676638,0.540364,
        1.047198,2.676780,-1.901158,
        1.047198,2.681411,-2.667965,
        1.047198,2.682152,0.245751,
        1.047198,2.682859,-0.817071,
        1.047198,2.705107,1.062142,
        1.047198,2.717705,2.625995,
        1.047198,2.721001,2.060097,
        1.047198,2.722319,-2.954069,
        1.047198,2.729840,-1.368237,
        1.047198,2.734349,-0.289298,
        1.047198,2.738409,2.975671,
        1.047198,2.740366,1.374341,
        1.047198,2.744633,-2.391311,
        1.047198,2.747398,1.727166,
        1.047198,2.763799,0.760372,
        1.047198,2.768238,-1.679243,
        1.047198,2.779245,-2.046268,
        1.047198,2.780208,-0.626453,
        1.047198,2.780471,0.027732,
        1.047198,2.784696,-1.013128,
        1.047198,2.789817,0.395431,
        1.047198,2.801154,2.353143,
        1.047198,2.807091,-2.687477,
        1.047198,2.825762,1.085035,
        1.047198,2.830209,2.734943,
        1.047198,2.833517,-3.085508,
        1.047198,2.838301,1.980061,
        1.047198,2.849871,1.512012,
        1.047198,2.850089,-1.343899,
        1.047198,2.856905,-0.318441,
        1.047198,2.870445,-2.291001,
        1.047198,2.885017,0.655562,
        1.047198,2.886428,-1.793217,
        1.047198,2.890337,-0.825633,
        1.047198,2.898932,0.139973,
        1.047198,2.931927,-2.728768,
        1.047198,2.932096,2.377790,
        1.047198,2.937409,2.947245,
        1.047198,2.948048,1.133127,
        1.047198,2.950896,1.797889,
        1.047198,2.970019,-1.282382,
        1.047198,2.978333,-0.412090,
        1.047198,2.996437,-2.089121,
        1.047198,3.011274,0.434110,
        1.047198,3.050410,3.138492,
        1.047198,3.064492,1.647259,
        1.047198,3.088868,-0.989501,
        1.151917,0.053884,-2.408394,
        1.151917,0.077645,1.453433,
        1.151917,0.088408,-0.311160,
        1.151917,0.142314,2.590387,
        1.151917,0.147206,-1.385164,
        1.151917,0.163506,0.573203,
        1.151917,0.165485,-2.875396,
        1.151917,0.182564,-2.125821,
        1.151917,0.190569,1.929141,
        1.151917,0.201102,1.249082,
        1.151917,0.201397,-0.716646,
        1.151917,0.210568,-0.069559,
        1.151917,0.251672,2.848900,
        1.151917,0.266262,-1.678068,
        1.151917,0.269108,-1.215563,
        1.151917,0.273777,-2.506756,
        1.151917,0.279854,0.814581,
        1.151917,0.281664,-2.969147,
        1.151917,0.283625,2.405603,
        1.151917,0.284058,0.357350,
        1.151917,0.294107,1.593880,
        1.151917,0.302491,-0.400410,
        1.151917,0.308059,2.021491,
        1.151917,0.312219,-2.072057,
        1.151917,0.321905,-0.829121,
        1.151917,0.329984,1.201621,
        1.151917,0.339714,-0.013712,
        1.151917,0.368784,2.974974,
        1.151917,0.376428,-1.495532,
        1.151917,0.378103,2.646436,
        1.151917,0.385088,0.590355,
        1.151917,0.386248,-2.690943,
        1.151917,0.389851,-2.340031,
        1.151917,0.391502,-1.171374,
        1.151917,0.398249,-1.812694,
        1.151917,0.400542,-2.995729,
        1.151917,0.404273,0.918163,
        1.151917,0.405617,1.773533,
        1.151917,0.409009,-0.570243,
        1.151917,0.410700,0.270433,
        1.151917,0.415893,1.464890,
        1.151917,0.420680,2.088164,
        1.151917,0.424766,-0.258729,
        1.151917,0.426507,2.367872,
        1.151917,0.441292,-0.890727,
        1.151917,0.451839,-2.078193,
        1.151917,0.459270,1.191562,
        1.151917,0.470622,0.014002,
        1.151917,0.481292,-2.529098,
        1.151917,0.482460,2.849843,
        1.151917,0.492205,-1.397963,
        1.151917,0.493314,3.117264,
        1.151917,0.497870,0.727828,
        1.151917,0.499783,-1.654379,
        1.151917,0.502651,0.463333,
        1.151917,0.505463,2.594557,
        1.151917,0.506104,-2.874910,
        1.151917,0.513579,-1.147354,
        1.151917,0.518569,1.644237,
        1.151917,0.519001,1.900488,
        1.151917,0.519277,-0.673809,
        1.151917,0.524769,-0.421955,
        1.151917,0.525564,-2.288838,
        1.151917,0.532541,0.978737,
        1.151917,0.534610,2.154808,
        1.151917,0.535020,-1.891846,
        1.151917,0.540208,1.400497,
        1.151917,0.542006,0.228337,
        1.151917,0.550762,-0.180771,
        1.151917,0.560531,2.378474,
        1.151917,0.561629,-0.924576,
        1.151917,0.567465,-2.679392,
        1.151917,0.588527,1.181755,
        1.151917,0.591643,-2.098364,
        1.151917,0.594963,2.997575,
        1.151917,0.599643,-3.045235,
        1.151917,0.599962,2.773556,
        1.151917,0.602620,0.029270,
        1.151917,0.605711,0.598454,
        1.151917,0.607708,-1.541026,
        1.151917,0.609360,-2.464379,
        1.151917,0.610738,-1.326786,
        1.151917,0.616968,0.820489,
        1.151917,0.626379,1.773610,
        1.151917,0.628626,-1.751180,
        1.151917,0.630671,0.391949,
        1.151917,0.632622,-0.737642,
        1.151917,0.632708,2.566722,
        1.151917,0.632797,1.989858,
        1.151917,0.634455,-0.534790,
        1.151917,0.636205,1.561170,
        1.151917,0.637681,-1.123423,
        1.151917,0.641441,-2.842109,
        1.151917,0.644807,-0.327139,
        1.151917,0.652440,2.197129,
        1.151917,0.660897,-2.278593,
        1.151917,0.664008,1.360091,
        1.151917,0.670850,-1.945257,
        1.151917,0.671396,1.004859,
        1.151917,0.673985,0.203855,
        1.151917,0.680326,-0.133370,
        1.151917,0.685083,-0.942363,
        1.151917,0.688204,-2.632204,
        1.151917,0.689777,2.385011,
        1.151917,0.700065,3.115430,
        1.151917,0.704080,2.911636,
        1.151917,0.712883,1.174216,
        1.151917,0.715701,0.699906,
        1.151917,0.718851,-1.442215,
        1.151917,0.721019,2.721536,
        1.151917,0.725300,-2.975134,
        1.151917,0.727136,-1.633288,
        1.151917,0.727399,-2.120567,
        1.151917,0.730751,0.522488,
        1.151917,0.732892,-1.254330,
        1.151917,0.734333,-2.446184,
        1.151917,0.735125,0.037270,
        1.151917,0.737893,1.866463,
        1.151917,0.740650,1.681075,
        1.151917,0.745971,-0.781286,
        1.151917,0.749432,-0.442559,
        1.151917,0.750102,2.050636,
        1.151917,0.751046,0.863821,
        1.151917,0.757018,-0.608833,
        1.151917,0.757046,-1.816019,
        1.151917,0.757389,1.501184,
        1.151917,0.759800,2.546181,
        1.151917,0.759859,0.351923,
        1.151917,0.761141,-2.781102,
        1.151917,0.769035,-1.078574,
        1.151917,0.770166,-0.266604,
        1.151917,0.772218,2.227380,
        1.151917,0.788382,1.331051,
        1.151917,0.790269,-2.282409,
        1.151917,0.803410,-1.984567,
        1.151917,0.803813,0.191520,
        1.151917,0.806832,3.022155,
        1.151917,0.807476,-2.597675,
        1.151917,0.812512,-0.107335,
        1.151917,0.814161,1.010949,
        1.151917,0.817732,-3.084486,
        1.151917,0.817846,2.390688,
        1.151917,0.819323,2.845960,
        1.151917,0.819430,-0.915017,
        1.151917,0.829198,-1.532491,
        1.151917,0.829531,0.665695,
        1.151917,0.829816,1.167147,
        1.151917,0.833613,-1.357492,
        1.151917,0.842945,-2.907318,
        1.151917,0.844883,2.679708,
        1.151917,0.848189,-1.705003,
        1.151917,0.849888,1.767901,
        1.151917,0.851126,0.493370,
        1.151917,0.852911,1.933394,
        1.151917,0.857516,-2.432863,
        1.151917,0.858261,-1.191729,
        1.151917,0.858279,-2.138895,
        1.151917,0.859723,-0.734928,
        1.151917,0.859754,1.603292,
        1.151917,0.861655,-0.394821,
        1.151917,0.864300,0.043549,
        1.151917,0.866940,0.851229,
        1.151917,0.869748,2.095969,
        1.151917,0.877352,-0.574668,
        1.151917,0.879254,-2.733338,
        1.151917,0.882554,1.443757,
        1.151917,0.884060,0.334199,
        1.151917,0.885626,-1.866061,
        1.151917,0.889834,2.529957,
        1.151917,0.893140,2.251698,
        1.151917,0.897463,-0.240822,
        1.151917,0.898758,-1.031948,
        1.151917,0.913918,1.287618,
        1.151917,0.916555,-2.287173,
        1.151917,0.918305,3.107845,
        1.151917,0.919522,2.948277,
        1.151917,0.927138,-2.569686,
        1.151917,0.929031,0.185383,
        1.151917,0.933766,-3.015917,
        1.151917,0.934873,-2.010822,
        1.151917,0.935636,-1.445855,
        1.151917,0.936702,-0.866461,
        1.151917,0.937395,0.729108,
        1.151917,0.937910,2.787575,
        1.151917,0.939144,-0.092962,
        1.151917,0.941718,1.125085,
        1.151917,0.942873,-1.607680,
        1.151917,0.943701,0.980328,
        1.151917,0.947774,2.395256,
        1.151917,0.952433,0.584347,
        1.151917,0.954456,-1.287293,
        1.151917,0.958731,-2.852552,
        1.151917,0.961838,1.839284,
        1.151917,0.961898,-0.469425,
        1.151917,0.965426,1.686615,
        1.151917,0.970108,1.990771,
        1.151917,0.972820,-1.762780,
        1.151917,0.973500,-0.711881,
        1.151917,0.975253,2.641082,
        1.151917,0.977824,0.441928,
        1.151917,0.981485,-2.424766,
        1.151917,0.981877,1.537100,
        1.151917,0.982609,-1.136920,
        1.151917,0.984265,-2.156073,
        1.151917,0.989150,0.045493,
        1.151917,0.993052,2.137478,
        1.151917,0.997766,-2.693871,
        1.151917,0.999847,-0.334738,
        1.151917,1.001920,0.848054,
        1.151917,1.006146,1.389734,
        1.151917,1.012180,0.304066,
        1.151917,1.016938,2.270674,
        1.151917,1.017371,-0.979790,
        1.151917,1.019495,-1.894487,
        1.151917,1.023796,-0.583691,
        1.151917,1.026525,2.506945,
        1.151917,1.027004,-0.203653,
        1.151917,1.028027,3.036573,
        1.151917,1.031435,1.241021,
        1.151917,1.032316,2.884221,
        1.151917,1.035863,-3.110475,
        1.151917,1.042085,-2.292950,
        1.151917,1.043272,-1.522986,
        1.151917,1.046265,-2.958847,
        1.151917,1.049272,-1.376344,
        1.151917,1.050039,-2.552086,
        1.151917,1.053682,0.689863,
        1.151917,1.055458,0.170394,
        1.151917,1.056622,-0.831866,
        1.151917,1.057515,1.097998,
        1.151917,1.058706,-2.036438,
        1.151917,1.061389,-1.672248,
        1.151917,1.063758,2.743093,
        1.151917,1.070157,-0.073422,
        1.151917,1.072154,1.761506,
        1.151917,1.074029,1.909193,
        1.151917,1.074204,0.545276,
        1.151917,1.074714,-2.805147,
        1.151917,1.076262,0.966104,
        1.151917,1.077642,-1.227073,
        1.151917,1.084151,1.618525,
        1.151917,1.085085,2.377673,
        1.151917,1.086822,-0.441112,
        1.151917,1.087142,2.058815,
        1.151917,1.098383,-0.696430,
        1.151917,1.099138,-1.084033,
        1.151917,1.103208,1.477773,
        1.151917,1.103783,0.408254,
        1.151917,1.106275,-2.420800,
        1.151917,1.108079,2.612943,
        1.151917,1.109345,-2.172718,
        1.151917,1.115957,-1.787597,
        1.151917,1.118930,0.045983,
        1.151917,1.119777,0.834702,
        1.151917,1.121943,-2.669042,
        1.151917,1.124930,2.219917,
        1.151917,1.125217,1.336294,
        1.151917,1.130958,2.972931,
        1.151917,1.131400,-0.306629,
        1.151917,1.137135,-0.936514,
        1.151917,1.139125,-1.924528,
        1.151917,1.139877,0.276382,
        1.151917,1.142003,-3.058022,
        1.151917,1.147363,-0.558448,
        1.151917,1.150348,-0.176418,
        1.151917,1.151198,1.197016,
        1.151917,1.153930,-1.597104,
        1.151917,1.155389,2.837148,
        1.151917,1.155458,-1.460623,
        1.151917,1.157331,-2.906362,
        1.151917,1.158594,2.487618,
        1.151917,1.162103,-1.322553,
        1.151917,1.165407,3.092861,
        1.151917,1.168025,0.717075,
        1.151917,1.168406,-2.301519,
        1.151917,1.174782,-2.538755,
        1.151917,1.177998,1.061256,
        1.151917,1.178898,1.827305,
        1.151917,1.181331,-2.060330,
        1.151917,1.182540,-0.798952,
        1.151917,1.184569,0.147517,
        1.151917,1.184734,1.968919,
        1.151917,1.184804,0.593386,
        1.151917,1.188278,1.690808,
        1.151917,1.194360,2.110440,
        1.151917,1.194776,2.710455,
        1.151917,1.197745,-1.173705,
        1.151917,1.199898,-2.777257,
        1.151917,1.201164,2.355931,
        1.151917,1.203760,1.556404,
        1.151917,1.205845,-0.415993,
        1.151917,1.206585,0.933648,
        1.151917,1.206950,-0.037248,
        1.151917,1.209341,0.471162,
        1.151917,1.213483,-1.039417,
        1.151917,1.218146,-0.672643,
        1.151917,1.221608,1.421669,
        1.151917,1.226043,-1.693533,
        1.151917,1.232870,-1.822773,
        1.151917,1.234375,-2.421194,
        1.151917,1.234531,-2.190148,
        1.151917,1.239414,0.348959,
        1.151917,1.240183,2.587255,
        1.151917,1.242312,2.941234,
        1.151917,1.244075,1.286644,
        1.151917,1.245961,-2.997461,
        1.151917,1.246495,-2.646872,
        1.151917,1.247736,2.242810,
        1.151917,1.253604,0.812222,
        1.151917,1.257816,-0.894317,
        1.151917,1.258247,-1.953634,
        1.151917,1.258993,-0.284099,
        1.151917,1.262380,-0.157609,
        1.151917,1.262791,-1.549229,
        1.151917,1.263772,-0.533073,
        1.151917,1.266731,-1.412461,
        1.151917,1.268591,-3.122146,
        1.151917,1.269614,-1.275246,
        1.151917,1.270745,1.152533,
        1.151917,1.273606,0.224545,
        1.151917,1.279716,2.812101,
        1.151917,1.283786,2.462124,
        1.151917,1.284798,0.078209,
        1.151917,1.285402,-2.882049,
        1.151917,1.286936,1.886053,
        1.151917,1.289408,0.683122,
        1.151917,1.294736,1.756184,
        1.151917,1.296978,-2.311511,
        1.151917,1.298546,2.020601,
        1.151917,1.300924,1.022478,
        1.151917,1.303305,3.043152,
        1.151917,1.304210,-2.084225,
        1.151917,1.305241,-2.530249,
        1.151917,1.307513,1.628003,
        1.151917,1.310722,0.558267,
        1.151917,1.314142,2.139329,
        1.151917,1.316087,-0.777971,
        1.151917,1.317907,-1.121023,
        1.151917,1.321019,-2.749776,
        1.151917,1.321609,2.687717,
        1.151917,1.322568,1.499415,
        1.151917,1.322972,-0.395699,
        1.151917,1.325286,-0.994192,
        1.151917,1.329146,-0.648513,
        1.151917,1.329795,-0.042570,
        1.151917,1.331651,2.344497,
        1.151917,1.336522,0.435098,
        1.151917,1.339410,0.900087,
        1.151917,1.339550,1.370353,
        1.151917,1.340441,-1.731580,
        1.151917,1.351855,-1.855234,
        1.151917,1.360161,-3.018404,
        1.151917,1.361031,-1.613971,
        1.151917,1.361196,2.914780,
        1.151917,1.361998,-2.207254,
        1.151917,1.362302,1.239384,
        1.151917,1.365228,2.563808,
        1.151917,1.365452,0.310522,
        1.151917,1.367028,-2.420041,
        1.151917,1.373422,-1.484281,
        1.151917,1.374245,-1.226417,
        1.151917,1.375020,-0.170393,
        1.151917,1.376560,-1.355481,
        1.151917,1.377070,-1.980984,
        1.151917,1.377578,-2.636339,
        1.151917,1.378096,0.774626,
        1.151917,1.379086,-0.508192,
        1.151917,1.382562,0.177141,
        1.151917,1.388849,2.233847,
        1.151917,1.389268,-0.886349,
        1.151917,1.391324,1.109396,
        1.151917,1.391699,-0.288725,
        1.151917,1.392869,3.136903,
        1.151917,1.393612,1.952016,
        1.151917,1.395545,-2.850807,
        1.151917,1.401926,1.822930,
        1.151917,1.402149,2.788898,
        1.151917,1.409552,0.647980,
        1.151917,1.410840,0.056219,
        1.151917,1.411944,2.444377,
        1.151917,1.413639,1.697982,
        1.151917,1.426479,1.573427,
        1.151917,1.426725,2.102242,
        1.151917,1.427868,0.984637,
        1.151917,1.429580,-2.107232,
        1.151917,1.431375,-2.313269,
        1.151917,1.431907,3.015410,
        1.151917,1.432325,-1.029054,
        1.151917,1.433341,0.524268,
        1.151917,1.435181,-0.739771,
        1.151917,1.437208,-2.529085,
        1.151917,1.439041,-0.617966,
        1.151917,1.441473,1.448544,
        1.151917,1.444527,2.664744,
        1.151917,1.450513,-0.400625,
        1.151917,1.451907,-0.066823,
        1.151917,1.452754,-2.738631,
        1.151917,1.454061,-2.950416,
        1.151917,1.456736,1.323154,
        1.151917,1.458370,0.400433,
        1.151917,1.459314,-1.765003,
        1.151917,1.459902,-1.145835,
        1.151917,1.464952,2.329245,
        1.151917,1.466050,0.862104,
        1.151917,1.471330,-1.886160,
        1.151917,1.475119,-1.648255,
        1.151917,1.480878,1.194576,
        1.151917,1.480942,0.273235,
        1.151917,1.481086,2.890269,
        1.151917,1.484118,-1.529156,
        1.151917,1.484741,-1.279219,
        1.151917,1.487681,-1.406069,
        1.151917,1.487886,-3.066037,
        1.151917,1.490555,2.543284,
        1.151917,1.492424,-0.190932,
        1.151917,1.495002,-0.839810,
        1.151917,1.495297,2.007188,
        1.151917,1.496414,-2.009267,
        1.151917,1.498261,-2.425270,
        1.151917,1.498907,-2.207327,
        1.151917,1.500204,0.737638,
        1.151917,1.502469,0.151156,
        1.151917,1.507448,-0.514782,
        1.151917,1.509287,2.204647,
        1.151917,1.510518,1.884920,
        1.151917,1.511948,-2.632360,
        1.151917,1.515235,1.068565,
        1.151917,1.522237,3.102018,
        1.151917,1.522373,1.764991,
        1.151917,1.522416,-0.954956,
        1.151917,1.523102,2.765920,
        1.151917,1.524811,-2.844718,
        1.151917,1.527261,-0.304916,
        1.151917,1.529788,0.613208,
        1.151917,1.533648,1.644808,
        1.151917,1.535018,0.028869,
        1.151917,1.541082,2.424608,
        1.151917,1.546010,1.523462,
        1.151917,1.551135,-0.642994,
        1.151917,1.553895,0.489607,
        1.151917,1.555041,0.947813,
        1.151917,1.559671,-1.074934,
        1.151917,1.560345,2.985899,
        1.151917,1.560487,-2.323400,
        1.151917,1.561127,1.401111,
        1.151917,1.567193,2.643172,
        1.151917,1.568892,2.098362,
        1.151917,1.571702,-2.098435,
        1.151917,1.572701,-2.528730,
        1.151917,1.573572,1.278593,
        1.151917,1.574849,-0.091505,
        1.151917,1.576334,0.365236,
        1.151917,1.577029,-2.962888,
        1.151917,1.577688,-1.814941,
        1.151917,1.579542,-0.755917,
        1.151917,1.579624,-1.199759,
        1.151917,1.581515,-0.419414,
        1.151917,1.584581,-2.738702,
        1.151917,1.586948,-1.696296,
        1.151917,1.588901,2.301496,
        1.151917,1.589936,0.825020,
        1.151917,1.590018,-1.941652,
        1.151917,1.596968,-1.328358,
        1.151917,1.597178,-1.575607,
        1.151917,1.600470,-1.453196,
        1.151917,1.601408,0.242409,
        1.151917,1.602259,1.152656,
        1.151917,1.603064,2.865774,
        1.151917,1.612085,1.968802,
        1.151917,1.612720,-3.080498,
        1.151917,1.616061,2.521531,
        1.151917,1.616900,-0.207457,
        1.151917,1.620988,-0.875920,
        1.151917,1.622114,0.701509,
        1.151917,1.622749,0.121426,
        1.151917,1.622993,-2.221629,
        1.151917,1.629346,-0.542669,
        1.151917,1.630007,1.844552,
        1.151917,1.634747,-2.426198,
        1.151917,1.642346,1.722617,
        1.151917,1.643925,1.034629,
        1.151917,1.644848,2.743774,
        1.151917,1.645258,2.188109,
        1.151917,1.645417,-2.634725,
        1.151917,1.647395,3.085437,
        1.151917,1.648538,-2.855044,
        1.151917,1.649932,0.576808,
        1.151917,1.652224,-0.996105,
        1.151917,1.652871,1.600681,
        1.151917,1.657997,-0.325127,
        1.151917,1.661936,0.001037,
        1.151917,1.665187,1.477402,
        1.151917,1.665496,2.397872,
        1.151917,1.673267,-0.665453,
        1.151917,1.673557,0.452526,
        1.151917,1.679277,-1.122421,
        1.151917,1.679696,0.912482,
        1.151917,1.682420,1.353922,
        1.151917,1.685031,-2.009038,
        1.151917,1.685196,2.967046,
        1.151917,1.690107,2.620910,
        1.151917,1.690453,-2.123076,
        1.151917,1.691693,-1.871572,
        1.151917,1.691817,2.061688,
        1.151917,1.692644,1.237670,
        1.151917,1.696386,0.328123,
        1.151917,1.697162,-1.249026,
        1.151917,1.697488,-2.322785,
        1.151917,1.698726,-2.973576,
        1.151917,1.699172,-1.749717,
        1.151917,1.704478,-0.106295,
        1.151917,1.707244,-0.444770,
        1.151917,1.707500,-2.531497,
        1.151917,1.707714,-0.787321,
        1.151917,1.708430,-1.627552,
        1.151917,1.710583,-1.376890,
        1.151917,1.712415,-2.747921,
        1.151917,1.713125,-1.503073,
        1.151917,1.713452,0.789419,
        1.151917,1.720612,2.279888,
        1.151917,1.724852,0.202831,
        1.151917,1.724868,2.846246,
        1.151917,1.728940,1.931535,
        1.151917,1.732838,1.127584,
        1.151917,1.735719,-3.092485,
        1.151917,1.739676,2.496017,
        1.151917,1.739971,-0.228876,
        1.151917,1.743328,0.663347,
        1.151917,1.744927,-0.911892,
        1.151917,1.745952,0.084733,
        1.151917,1.748077,1.804354,
        1.151917,1.752959,-0.567548,
        1.151917,1.760169,1.678132,
        1.151917,1.763887,-2.218450,
        1.151917,1.766737,2.722752,
        1.151917,1.767999,1.004737,
        1.151917,1.769149,1.550298,
        1.151917,1.770119,-2.427175,
        1.151917,1.770460,2.154208,
        1.151917,1.770514,3.070982,
        1.151917,1.770735,0.537245,
        1.151917,1.772673,-2.864760,
        1.151917,1.773527,-1.039582,
        1.151917,1.776419,-2.642421,
        1.151917,1.780769,1.416330,
        1.151917,1.786634,-0.346247,
        1.151917,1.788829,-1.947553,
        1.151917,1.793737,0.410270,
        1.151917,1.794553,2.374618,
        1.151917,1.794699,-0.694061,
        1.151917,1.797469,-1.169105,
        1.151917,1.803705,0.881076,
        1.151917,1.804124,1.264249,
        1.151917,1.807825,-2.090455,
        1.151917,1.808822,-1.817375,
        1.151917,1.809020,2.949741,
        1.151917,1.812466,2.597595,
        1.151917,1.813159,2.025395,
        1.151917,1.813816,-1.298435,
        1.151917,1.815016,0.278951,
        1.151917,1.816510,-0.008631,
        1.151917,1.817772,-1.689625,
        1.151917,1.821360,-2.984352,
        1.151917,1.822312,-0.127211,
        1.151917,1.823386,-1.429807,
        1.151917,1.823834,-1.560483,
        1.151917,1.832060,-0.821627,
        1.151917,1.832690,-0.468440,
        1.151917,1.833945,-2.320278,
        1.151917,1.834787,0.751968,
        1.151917,1.840579,-2.536445,
        1.151917,1.841033,-2.756932,
        1.151917,1.846350,1.893983,
        1.151917,1.846644,1.113766,
        1.151917,1.846833,2.826543,
        1.151917,1.847325,2.249051,
        1.151917,1.854119,0.134161,
        1.151917,1.858863,-3.106022,
        1.151917,1.865046,1.763384,
        1.151917,1.865619,-0.952036,
        1.151917,1.865912,0.624453,
        1.151917,1.866694,2.473686,
        1.151917,1.867524,-0.240083,
        1.151917,1.875561,1.632831,
        1.151917,1.876049,-0.594378,
        1.151917,1.880509,1.500204,
        1.151917,1.883504,1.360457,
        1.151917,1.886971,-2.196665,
        1.151917,1.888979,0.982922,
        1.151917,1.889815,2.701570,
        1.151917,1.892144,-2.005340,
        1.151917,1.892348,0.496522,
        1.151917,1.893281,-1.083976,
        1.151917,1.893828,3.053453,
        1.151917,1.895117,2.121107,
        1.151917,1.898741,-2.875735,
        1.151917,1.904429,-2.428029,
        1.151917,1.908408,-2.648903,
        1.151917,1.911828,-0.364489,
        1.151917,1.912224,0.366565,
        1.151917,1.914854,1.214468,
        1.151917,1.915467,-1.883143,
        1.151917,1.915763,-0.724679,
        1.151917,1.915790,-1.216366,
        1.151917,1.922256,2.347806,
        1.151917,1.922616,0.844633,
        1.151917,1.925185,0.231170,
        1.151917,1.927474,-1.754616,
        1.151917,1.928342,-1.351174,
        1.151917,1.933328,-1.623328,
        1.151917,1.933878,0.034150,
        1.151917,1.934585,-1.487979,
        1.151917,1.934802,2.927634,
        1.151917,1.934862,1.989776,
        1.151917,1.936327,-0.099350,
        1.151917,1.938088,2.578507,
        1.151917,1.945677,-2.999805,
        1.151917,1.954572,-0.859208,
        1.151917,1.957868,-0.490784,
        1.151917,1.958989,0.712772,
        1.151917,1.960212,-2.307561,
        1.151917,1.964774,1.855507,
        1.151917,1.964971,-2.100533,
        1.151917,1.969112,1.088273,
        1.151917,1.971438,-2.765415,
        1.151917,1.972312,2.805134,
        1.151917,1.974950,2.220596,
        1.151917,1.975135,-2.539127,
        1.151917,1.982107,1.720829,
        1.151917,1.983195,-3.127984,
        1.151917,1.983318,-0.247182,
        1.151917,1.986720,-0.993220,
        1.151917,1.990381,1.586005,
        1.151917,1.990470,0.580283,
        1.151917,1.990523,1.450326,
        1.151917,1.993909,1.312897,
        1.151917,1.994694,2.452813,
        1.151917,1.999226,-0.620717,
        1.151917,2.005145,0.941012,
        1.151917,2.007301,0.142337,
        1.151917,2.014149,0.444138,
        1.151917,2.014338,-1.127392,
        1.151917,2.018131,3.021778,
        1.151917,2.019676,-1.980944,
        1.151917,2.019912,2.686612,
        1.151917,2.020545,2.088672,
        1.151917,2.026074,0.300945,
        1.151917,2.026196,-2.889057,
        1.151917,2.031959,-1.268178,
        1.151917,2.032917,-2.199917,
        1.151917,2.033572,-1.842884,
        1.151917,2.034712,-2.419505,
        1.151917,2.035818,-0.379479,
        1.151917,2.036004,-0.757062,
        1.151917,2.040501,-1.411488,
        1.151917,2.042874,-1.701367,
        1.151917,2.043060,-2.653746,
        1.151917,2.043729,-1.557938,
        1.151917,2.048538,0.806623,
        1.151917,2.049006,1.199930,
        1.151917,2.049502,-0.137989,
        1.151917,2.052181,2.326959,
        1.151917,2.052794,-0.008526,
        1.151917,2.057608,1.951726,
        1.151917,2.065568,2.567730,
        1.151917,2.070371,2.879808,
        1.151917,2.071256,-3.018961,
        1.151917,2.079746,-0.897699,
        1.151917,2.082784,1.042403,
        1.151917,2.083752,-0.511086,
        1.151917,2.084401,1.810502,
        1.151917,2.085731,0.671081,
        1.151917,2.097496,1.373172,
        1.151917,2.099185,1.667006,
        1.151917,2.100482,-2.301084,
        1.151917,2.103118,2.194745,
        1.151917,2.103223,1.521685,
        1.151917,2.103695,-2.777707,
        1.151917,2.107523,3.129581,
        1.151917,2.108277,-2.095907,
        1.151917,2.108891,-2.532166,
        1.151917,2.111003,-0.268632,
        1.151917,2.112276,-1.027246,
        1.151917,2.114380,0.531457,
        1.151917,2.119166,0.100552,
        1.151917,2.119564,2.754818,
        1.151917,2.122474,-0.648314,
        1.151917,2.125183,2.445598,
        1.151917,2.129501,0.238290,
        1.151917,2.131178,0.386663,
        1.151917,2.131223,-1.174311,
        1.151917,2.134269,-1.948042,
        1.151917,2.134684,0.907065,
        1.151917,2.137615,2.988806,
        1.151917,2.145635,-1.327708,
        1.151917,2.146467,2.055081,
        1.151917,2.146582,-1.793788,
        1.151917,2.149210,-1.482321,
        1.151917,2.152901,-1.636788,
        1.151917,2.156337,1.151558,
        1.151917,2.156474,-2.912065,
        1.151917,2.157194,-0.792563,
        1.151917,2.166724,-0.401638,
        1.151917,2.166980,-0.134879,
        1.151917,2.173644,-2.403429,
        1.151917,2.175711,-2.655523,
        1.151917,2.177132,2.600419,
        1.151917,2.178879,0.767722,
        1.151917,2.180752,1.908485,
        1.151917,2.181338,2.313721,
        1.151917,2.185580,1.287673,
        1.151917,2.195400,-2.217546,
        1.151917,2.200418,-3.056442,
        1.151917,2.201829,-0.001735,
        1.151917,2.203804,1.756176,
        1.151917,2.211602,1.443930,
        1.151917,2.212686,-0.544809,
        1.151917,2.213937,0.621892,
        1.151917,2.214075,1.600691,
        1.151917,2.216256,1.020029,
        1.151917,2.219060,2.886631,
        1.151917,2.223375,-0.906185,
        1.151917,2.226002,-1.060248,
        1.151917,2.226413,-2.064034,
        1.151917,2.230554,2.171334,
        1.151917,2.232413,2.735777,
        1.151917,2.233709,3.070830,
        1.151917,2.234383,-2.792065,
        1.151917,2.237183,0.468799,
        1.151917,2.237695,-0.271388,
        1.151917,2.239419,0.151693,
        1.151917,2.243347,-2.521049,
        1.151917,2.244035,-1.229456,
        1.151917,2.244355,0.309899,
        1.151917,2.245175,2.460236,
        1.151917,2.245628,-1.899216,
        1.151917,2.245973,-0.702234,
        1.151917,2.255601,-1.728131,
        1.151917,2.256702,-1.401367,
        1.151917,2.261466,-1.557500,
        1.151917,2.267390,0.878499,
        1.151917,2.272099,2.018594,
        1.151917,2.278618,1.170386,
        1.151917,2.283304,-2.345584,
        1.151917,2.285175,-2.943090,
        1.151917,2.293000,-0.124451,
        1.151917,2.294058,-0.421070,
        1.151917,2.303283,1.855938,
        1.151917,2.304968,2.608978,
        1.151917,2.305347,2.311346,
        1.151917,2.307188,-2.656134,
        1.151917,2.307301,1.336200,
        1.151917,2.309655,0.726020,
        1.151917,2.317634,-2.185225,
        1.151917,2.321510,1.685664,
        1.151917,2.324376,-3.111006,
        1.151917,2.324731,1.511186,
        1.151917,2.329587,0.035498,
        1.151917,2.330158,2.977264,
        1.151917,2.335096,-0.588468,
        1.151917,2.336046,-1.115799,
        1.151917,2.338060,2.803613,
        1.151917,2.338083,-0.943400,
        1.151917,2.340210,0.562145,
        1.151917,2.341931,1.023926,
        1.151917,2.343265,-0.779061,
        1.151917,2.343434,-2.013692,
        1.151917,2.353029,-1.310529,
        1.151917,2.354868,0.211543,
        1.151917,2.355120,0.388593,
        1.151917,2.355165,-1.833173,
        1.151917,2.355462,2.151441,
        1.151917,2.362948,-2.487477,
        1.151917,2.363101,-2.810456,
        1.151917,2.363394,-0.272918,
        1.151917,2.364509,-1.644533,
        1.151917,2.370465,2.466824,
        1.151917,2.386110,-1.481351,
        1.151917,2.394960,0.862133,
        1.151917,2.395716,1.977184,
        1.151917,2.396339,1.200900,
        1.151917,2.407941,-2.316753,
        1.151917,2.409450,-2.988432,
        1.151917,2.417202,-0.103363,
        1.151917,2.417386,-0.443097,
        1.151917,2.417950,2.666662,
        1.151917,2.423190,1.789783,
        1.151917,2.423692,1.392696,
        1.151917,2.431122,3.086379,
        1.151917,2.431406,-2.647820,
        1.151917,2.431584,2.307631,
        1.151917,2.434090,1.592818,
        1.151917,2.435993,0.683844,
        1.151917,2.440898,-2.133300,
        1.151917,2.442222,-1.195837,
        1.151917,2.443112,2.883761,
        1.151917,2.449191,-0.633651,
        1.151917,2.450821,0.084741,
        1.151917,2.453499,-1.008861,
        1.151917,2.460010,-1.938011,
        1.151917,2.460199,-0.822008,
        1.151917,2.460902,0.490079,
        1.151917,2.461774,-1.734913,
        1.151917,2.463986,1.034407,
        1.151917,2.467647,0.287667,
        1.151917,2.479849,2.126925,
        1.151917,2.485907,-2.835471,
        1.151917,2.485945,-1.374853,
        1.151917,2.487747,2.505912,
        1.151917,2.487957,-0.274279,
        1.151917,2.489741,-2.468428,
        1.151917,2.512538,1.239900,
        1.151917,2.517528,1.923289,
        1.151917,2.518454,-1.565852,
        1.151917,2.519348,0.843503,
        1.151917,2.522766,-3.059231,
        1.151917,2.528686,2.737228,
        1.151917,2.534798,-2.266671,
        1.151917,2.536533,1.466274,
        1.151917,2.538504,1.699812,
        1.151917,2.539516,-0.476942,
        1.151917,2.539532,-0.074548,
        1.151917,2.543197,2.990252,
        1.151917,2.554022,2.320619,
        1.151917,2.556919,-2.655005,
        1.151917,2.558146,0.627191,
        1.151917,2.558223,-1.164189,
        1.151917,2.563452,-2.044028,
        1.151917,2.568462,0.150655,
        1.151917,2.569130,-0.700837,
        1.151917,2.570050,-1.804492,
        1.151917,2.572710,-0.934665,
        1.151917,2.575994,0.391718,
        1.151917,2.584774,1.046703,
        1.151917,2.601432,2.102535,
        1.151917,2.603559,2.555114,
        1.151917,2.604150,-2.885208,
        1.151917,2.608630,-1.378011,
        1.151917,2.611256,-0.279110,
        1.151917,2.617344,-2.439480,
        1.151917,2.627402,1.293552,
        1.151917,2.635396,1.848850,
        1.151917,2.635686,2.834950,
        1.151917,2.637087,3.132730,
        1.151917,2.642029,0.813274,
        1.151917,2.645260,1.570203,
        1.151917,2.646182,-1.614036,
        1.151917,2.660113,-2.186513,
        1.151917,2.660483,-0.034608,
        1.151917,2.661713,-0.529511,
        1.151917,2.672720,-1.110422,
        1.151917,2.676626,2.335353,
        1.151917,2.676638,0.540364,
        1.151917,2.676780,-1.901158,
        1.151917,2.681411,-2.667965,
        1.151917,2.682152,0.245751,
        1.151917,2.682859,-0.817071,
        1.151917,2.705107,1.062142,
        1.151917,2.717705,2.625995,
        1.151917,2.721001,2.060097,
        1.151917,2.722319,-2.954069,
        1.151917,2.729840,-1.368237,
        1.151917,2.734349,-0.289298,
        1.151917,2.738409,2.975671,
        1.151917,2.740366,1.374341,
        1.151917,2.744633,-2.391311,
        1.151917,2.747398,1.727166,
        1.151917,2.763799,0.760372,
        1.151917,2.768238,-1.679243,
        1.151917,2.779245,-2.046268,
        1.151917,2.780208,-0.626453,
        1.151917,2.780471,0.027732,
        1.151917,2.784696,-1.013128,
        1.151917,2.789817,0.395431,
        1.151917,2.801154,2.353143,
        1.151917,2.807091,-2.687477,
        1.151917,2.825762,1.085035,
        1.151917,2.830209,2.734943,
        1.151917,2.833517,-3.085508,
        1.151917,2.838301,1.980061,
        1.151917,2.849871,1.512012,
        1.151917,2.850089,-1.343899,
        1.151917,2.856905,-0.318441,
        1.151917,2.870445,-2.291001,
        1.151917,2.885017,0.655562,
        1.151917,2.886428,-1.793217,
        1.151917,2.890337,-0.825633,
        1.151917,2.898932,0.139973,
        1.151917,2.931927,-2.728768,
        1.151917,2.932096,2.377790,
        1.151917,2.937409,2.947245,
        1.151917,2.948048,1.133127,
        1.151917,2.950896,1.797889,
        1.151917,2.970019,-1.282382,
        1.151917,2.978333,-0.412090,
        1.151917,2.996437,-2.089121,
        1.151917,3.011274,0.434110,
        1.151917,3.050410,3.138492,
        1.151917,3.064492,1.647259,
        1.151917,3.088868,-0.989501,
        1.256637,0.053884,-2.408394,
        1.256637,0.077645,1.453433,
        1.256637,0.088408,-0.311160,
        1.256637,0.142314,2.590387,
        1.256637,0.147206,-1.385164,
        1.256637,0.163506,0.573203,
        1.256637,0.165485,-2.875396,
        1.256637,0.182564,-2.125821,
        1.256637,0.190569,1.929141,
        1.256637,0.201102,1.249082,
        1.256637,0.201397,-0.716646,
        1.256637,0.210568,-0.069559,
        1.256637,0.251672,2.848900,
        1.256637,0.266262,-1.678068,
        1.256637,0.269108,-1.215563,
        1.256637,0.273777,-2.506756,
        1.256637,0.279854,0.814581,
        1.256637,0.281664,-2.969147,
        1.256637,0.283625,2.405603,
        1.256637,0.284058,0.357350,
        1.256637,0.294107,1.593880,
        1.256637,0.302491,-0.400410,
        1.256637,0.308059,2.021491,
        1.256637,0.312219,-2.072057,
        1.256637,0.321905,-0.829121,
        1.256637,0.329984,1.201621,
        1.256637,0.339714,-0.013712,
        1.256637,0.368784,2.974974,
        1.256637,0.376428,-1.495532,
        1.256637,0.378103,2.646436,
        1.256637,0.385088,0.590355,
        1.256637,0.386248,-2.690943,
        1.256637,0.389851,-2.340031,
        1.256637,0.391502,-1.171374,
        1.256637,0.398249,-1.812694,
        1.256637,0.400542,-2.995729,
        1.256637,0.404273,0.918163,
        1.256637,0.405617,1.773533,
        1.256637,0.409009,-0.570243,
        1.256637,0.410700,0.270433,
        1.256637,0.415893,1.464890,
        1.256637,0.420680,2.088164,
        1.256637,0.424766,-0.258729,
        1.256637,0.426507,2.367872,
        1.256637,0.441292,-0.890727,
        1.256637,0.451839,-2.078193,
        1.256637,0.459270,1.191562,
        1.256637,0.470622,0.014002,
        1.256637,0.481292,-2.529098,
        1.256637,0.482460,2.849843,
        1.256637,0.492205,-1.397963,
        1.256637,0.493314,3.117264,
        1.256637,0.497870,0.727828,
        1.256637,0.499783,-1.654379,
        1.256637,0.502651,0.463333,
        1.256637,0.505463,2.594557,
        1.256637,0.506104,-2.874910,
        1.256637,0.513579,-1.147354,
        1.256637,0.518569,1.644237,
        1.256637,0.519001,1.900488,
        1.256637,0.519277,-0.673809,
        1.256637,0.524769,-0.421955,
        1.256637,0.525564,-2.288838,
        1.256637,0.532541,0.978737,
        1.256637,0.534610,2.154808,
        1.256637,0.535020,-1.891846,
        1.256637,0.540208,1.400497,
        1.256637,0.542006,0.228337,
        1.256637,0.550762,-0.180771,
        1.256637,0.560531,2.378474,
        1.256637,0.561629,-0.924576,
        1.256637,0.567465,-2.679392,
        1.256637,0.588527,1.181755,
        1.256637,0.591643,-2.098364,
        1.256637,0.594963,2.997575,
        1.256637,0.599643,-3.045235,
        1.256637,0.599962,2.773556,
        1.256637,0.602620,0.029270,
        1.256637,0.605711,0.598454,
        1.256637,0.607708,-1.541026,
        1.256637,0.609360,-2.464379,
        1.256637,0.610738,-1.326786,
        1.256637,0.616968,0.820489,
        1.256637,0.626379,1.773610,
        1.256637,0.628626,-1.751180,
        1.256637,0.630671,0.391949,
        1.256637,0.632622,-0.737642,
        1.256637,0.632708,2.566722,
        1.256637,0.632797,1.989858,
        1.256637,0.634455,-0.534790,
        1.256637,0.636205,1.561170,
        1.256637,0.637681,-1.123423,
        1.256637,0.641441,-2.842109,
        1.256637,0.644807,-0.327139,
        1.256637,0.652440,2.197129,
        1.256637,0.660897,-2.278593,
        1.256637,0.664008,1.360091,
        1.256637,0.670850,-1.945257,
        1.256637,0.671396,1.004859,
        1.256637,0.673985,0.203855,
        1.256637,0.680326,-0.133370,
        1.256637,0.685083,-0.942363,
        1.256637,0.688204,-2.632204,
        1.256637,0.689777,2.385011,
        1.256637,0.700065,3.115430,
        1.256637,0.704080,2.911636,
        1.256637,0.712883,1.174216,
        1.256637,0.715701,0.699906,
        1.256637,0.718851,-1.442215,
        1.256637,0.721019,2.721536,
        1.256637,0.725300,-2.975134,
        1.256637,0.727136,-1.633288,
        1.256637,0.727399,-2.120567,
        1.256637,0.730751,0.522488,
        1.256637,0.732892,-1.254330,
        1.256637,0.734333,-2.446184,
        1.256637,0.735125,0.037270,
        1.256637,0.737893,1.866463,
        1.256637,0.740650,1.681075,
        1.256637,0.745971,-0.781286,
        1.256637,0.749432,-0.442559,
        1.256637,0.750102,2.050636,
        1.256637,0.751046,0.863821,
        1.256637,0.757018,-0.608833,
        1.256637,0.757046,-1.816019,
        1.256637,0.757389,1.501184,
        1.256637,0.759800,2.546181,
        1.256637,0.759859,0.351923,
        1.256637,0.761141,-2.781102,
        1.256637,0.769035,-1.078574,
        1.256637,0.770166,-0.266604,
        1.256637,0.772218,2.227380,
        1.256637,0.788382,1.331051,
        1.256637,0.790269,-2.282409,
        1.256637,0.803410,-1.984567,
        1.256637,0.803813,0.191520,
        1.256637,0.806832,3.022155,
        1.256637,0.807476,-2.597675,
        1.256637,0.812512,-0.107335,
        1.256637,0.814161,1.010949,
        1.256637,0.817732,-3.084486,
        1.256637,0.817846,2.390688,
        1.256637,0.819323,2.845960,
        1.256637,0.819430,-0.915017,
        1.256637,0.829198,-1.532491,
        1.256637,0.829531,0.665695,
        1.256637,0.829816,1.167147,
        1.256637,0.833613,-1.357492,
        1.256637,0.842945,-2.907318,
        1.256637,0.844883,2.679708,
        1.256637,0.848189,-1.705003,
        1.256637,0.849888,1.767901,
        1.256637,0.851126,0.493370,
        1.256637,0.852911,1.933394,
        1.256637,0.857516,-2.432863,
        1.256637,0.858261,-1.191729,
        1.256637,0.858279,-2.138895,
        1.256637,0.859723,-0.734928,
        1.256637,0.859754,1.603292,
        1.256637,0.861655,-0.394821,
        1.256637,0.864300,0.043549,
        1.256637,0.866940,0.851229,
        1.256637,0.869748,2.095969,
        1.256637,0.877352,-0.574668,
        1.256637,0.879254,-2.733338,
        1.256637,0.882554,1.443757,
        1.256637,0.884060,0.334199,
        1.256637,0.885626,-1.866061,
        1.256637,0.889834,2.529957,
        1.256637,0.893140,2.251698,
        1.256637,0.897463,-0.240822,
        1.256637,0.898758,-1.031948,
        1.256637,0.913918,1.287618,
        1.256637,0.916555,-2.287173,
        1.256637,0.918305,3.107845,
        1.256637,0.919522,2.948277,
        1.256637,0.927138,-2.569686,
        1.256637,0.929031,0.185383,
        1.256637,0.933766,-3.015917,
        1.256637,0.934873,-2.010822,
        1.256637,0.935636,-1.445855,
        1.256637,0.936702,-0.866461,
        1.256637,0.937395,0.729108,
        1.256637,0.937910,2.787575,
        1.256637,0.939144,-0.092962,
        1.256637,0.941718,1.125085,
        1.256637,0.942873,-1.607680,
        1.256637,0.943701,0.980328,
        1.256637,0.947774,2.395256,
        1.256637,0.952433,0.584347,
        1.256637,0.954456,-1.287293,
        1.256637,0.958731,-2.852552,
        1.256637,0.961838,1.839284,
        1.256637,0.961898,-0.469425,
        1.256637,0.965426,1.686615,
        1.256637,0.970108,1.990771,
        1.256637,0.972820,-1.762780,
        1.256637,0.973500,-0.711881,
        1.256637,0.975253,2.641082,
        1.256637,0.977824,0.441928,
        1.256637,0.981485,-2.424766,
        1.256637,0.981877,1.537100,
        1.256637,0.982609,-1.136920,
        1.256637,0.984265,-2.156073,
        1.256637,0.989150,0.045493,
        1.256637,0.993052,2.137478,
        1.256637,0.997766,-2.693871,
        1.256637,0.999847,-0.334738,
        1.256637,1.001920,0.848054,
        1.256637,1.006146,1.389734,
        1.256637,1.012180,0.304066,
        1.256637,1.016938,2.270674,
        1.256637,1.017371,-0.979790,
        1.256637,1.019495,-1.894487,
        1.256637,1.023796,-0.583691,
        1.256637,1.026525,2.506945,
        1.256637,1.027004,-0.203653,
        1.256637,1.028027,3.036573,
        1.256637,1.031435,1.241021,
        1.256637,1.032316,2.884221,
        1.256637,1.035863,-3.110475,
        1.256637,1.042085,-2.292950,
        1.256637,1.043272,-1.522986,
        1.256637,1.046265,-2.958847,
        1.256637,1.049272,-1.376344,
        1.256637,1.050039,-2.552086,
        1.256637,1.053682,0.689863,
        1.256637,1.055458,0.170394,
        1.256637,1.056622,-0.831866,
        1.256637,1.057515,1.097998,
        1.256637,1.058706,-2.036438,
        1.256637,1.061389,-1.672248,
        1.256637,1.063758,2.743093,
        1.256637,1.070157,-0.073422,
        1.256637,1.072154,1.761506,
        1.256637,1.074029,1.909193,
        1.256637,1.074204,0.545276,
        1.256637,1.074714,-2.805147,
        1.256637,1.076262,0.966104,
        1.256637,1.077642,-1.227073,
        1.256637,1.084151,1.618525,
        1.256637,1.085085,2.377673,
        1.256637,1.086822,-0.441112,
        1.256637,1.087142,2.058815,
        1.256637,1.098383,-0.696430,
        1.256637,1.099138,-1.084033,
        1.256637,1.103208,1.477773,
        1.256637,1.103783,0.408254,
        1.256637,1.106275,-2.420800,
        1.256637,1.108079,2.612943,
        1.256637,1.109345,-2.172718,
        1.256637,1.115957,-1.787597,
        1.256637,1.118930,0.045983,
        1.256637,1.119777,0.834702,
        1.256637,1.121943,-2.669042,
        1.256637,1.124930,2.219917,
        1.256637,1.125217,1.336294,
        1.256637,1.130958,2.972931,
        1.256637,1.131400,-0.306629,
        1.256637,1.137135,-0.936514,
        1.256637,1.139125,-1.924528,
        1.256637,1.139877,0.276382,
        1.256637,1.142003,-3.058022,
        1.256637,1.147363,-0.558448,
        1.256637,1.150348,-0.176418,
        1.256637,1.151198,1.197016,
        1.256637,1.153930,-1.597104,
        1.256637,1.155389,2.837148,
        1.256637,1.155458,-1.460623,
        1.256637,1.157331,-2.906362,
        1.256637,1.158594,2.487618,
        1.256637,1.162103,-1.322553,
        1.256637,1.165407,3.092861,
        1.256637,1.168025,0.717075,
        1.256637,1.168406,-2.301519,
        1.256637,1.174782,-2.538755,
        1.256637,1.177998,1.061256,
        1.256637,1.178898,1.827305,
        1.256637,1.181331,-2.060330,
        1.256637,1.182540,-0.798952,
        1.256637,1.184569,0.147517,
        1.256637,1.184734,1.968919,
        1.256637,1.184804,0.593386,
        1.256637,1.188278,1.690808,
        1.256637,1.194360,2.110440,
        1.256637,1.194776,2.710455,
        1.256637,1.197745,-1.173705,
        1.256637,1.199898,-2.777257,
        1.256637,1.201164,2.355931,
        1.256637,1.203760,1.556404,
        1.256637,1.205845,-0.415993,
        1.256637,1.206585,0.933648,
        1.256637,1.206950,-0.037248,
        1.256637,1.209341,0.471162,
        1.256637,1.213483,-1.039417,
        1.256637,1.218146,-0.672643,
        1.256637,1.221608,1.421669,
        1.256637,1.226043,-1.693533,
        1.256637,1.232870,-1.822773,
        1.256637,1.234375,-2.421194,
        1.256637,1.234531,-2.190148,
        1.256637,1.239414,0.348959,
        1.256637,1.240183,2.587255,
        1.256637,1.242312,2.941234,
        1.256637,1.244075,1.286644,
        1.256637,1.245961,-2.997461,
        1.256637,1.246495,-2.646872,
        1.256637,1.247736,2.242810,
        1.256637,1.253604,0.812222,
        1.256637,1.257816,-0.894317,
        1.256637,1.258247,-1.953634,
        1.256637,1.258993,-0.284099,
        1.256637,1.262380,-0.157609,
        1.256637,1.262791,-1.549229,
        1.256637,1.263772,-0.533073,
        1.256637,1.266731,-1.412461,
        1.256637,1.268591,-3.122146,
        1.256637,1.269614,-1.275246,
        1.256637,1.270745,1.152533,
        1.256637,1.273606,0.224545,
        1.256637,1.279716,2.812101,
        1.256637,1.283786,2.462124,
        1.256637,1.284798,0.078209,
        1.256637,1.285402,-2.882049,
        1.256637,1.286936,1.886053,
        1.256637,1.289408,0.683122,
        1.256637,1.294736,1.756184,
        1.256637,1.296978,-2.311511,
        1.256637,1.298546,2.020601,
        1.256637,1.300924,1.022478,
        1.256637,1.303305,3.043152,
        1.256637,1.304210,-2.084225,
        1.256637,1.305241,-2.530249,
        1.256637,1.307513,1.628003,
        1.256637,1.310722,0.558267,
        1.256637,1.314142,2.139329,
        1.256637,1.316087,-0.777971,
        1.256637,1.317907,-1.121023,
        1.256637,1.321019,-2.749776,
        1.256637,1.321609,2.687717,
        1.256637,1.322568,1.499415,
        1.256637,1.322972,-0.395699,
        1.256637,1.325286,-0.994192,
        1.256637,1.329146,-0.648513,
        1.256637,1.329795,-0.042570,
        1.256637,1.331651,2.344497,
        1.256637,1.336522,0.435098,
        1.256637,1.339410,0.900087,
        1.256637,1.339550,1.370353,
        1.256637,1.340441,-1.731580,
        1.256637,1.351855,-1.855234,
        1.256637,1.360161,-3.018404,
        1.256637,1.361031,-1.613971,
        1.256637,1.361196,2.914780,
        1.256637,1.361998,-2.207254,
        1.256637,1.362302,1.239384,
        1.256637,1.365228,2.563808,
        1.256637,1.365452,0.310522,
        1.256637,1.367028,-2.420041,
        1.256637,1.373422,-1.484281,
        1.256637,1.374245,-1.226417,
        1.256637,1.375020,-0.170393,
        1.256637,1.376560,-1.355481,
        1.256637,1.377070,-1.980984,
        1.256637,1.377578,-2.636339,
        1.256637,1.378096,0.774626,
        1.256637,1.379086,-0.508192,
        1.256637,1.382562,0.177141,
        1.256637,1.388849,2.233847,
        1.256637,1.389268,-0.886349,
        1.256637,1.391324,1.109396,
        1.256637,1.391699,-0.288725,
        1.256637,1.392869,3.136903,
        1.256637,1.393612,1.952016,
        1.256637,1.395545,-2.850807,
        1.256637,1.401926,1.822930,
        1.256637,1.402149,2.788898,
        1.256637,1.409552,0.647980,
        1.256637,1.410840,0.056219,
        1.256637,1.411944,2.444377,
        1.256637,1.413639,1.697982,
        1.256637,1.426479,1.573427,
        1.256637,1.426725,2.102242,
        1.256637,1.427868,0.984637,
        1.256637,1.429580,-2.107232,
        1.256637,1.431375,-2.313269,
        1.256637,1.431907,3.015410,
        1.256637,1.432325,-1.029054,
        1.256637,1.433341,0.524268,
        1.256637,1.435181,-0.739771,
        1.256637,1.437208,-2.529085,
        1.256637,1.439041,-0.617966,
        1.256637,1.441473,1.448544,
        1.256637,1.444527,2.664744,
        1.256637,1.450513,-0.400625,
        1.256637,1.451907,-0.066823,
        1.256637,1.452754,-2.738631,
        1.256637,1.454061,-2.950416,
        1.256637,1.456736,1.323154,
        1.256637,1.458370,0.400433,
        1.256637,1.459314,-1.765003,
        1.256637,1.459902,-1.145835,
        1.256637,1.464952,2.329245,
        1.256637,1.466050,0.862104,
        1.256637,1.471330,-1.886160,
        1.256637,1.475119,-1.648255,
        1.256637,1.480878,1.194576,
        1.256637,1.480942,0.273235,
        1.256637,1.481086,2.890269,
        1.256637,1.484118,-1.529156,
        1.256637,1.484741,-1.279219,
        1.256637,1.487681,-1.406069,
        1.256637,1.487886,-3.066037,
        1.256637,1.490555,2.543284,
        1.256637,1.492424,-0.190932,
        1.256637,1.495002,-0.839810,
        1.256637,1.495297,2.007188,
        1.256637,1.496414,-2.009267,
        1.256637,1.498261,-2.425270,
        1.256637,1.498907,-2.207327,
        1.256637,1.500204,0.737638,
        1.256637,1.502469,0.151156,
        1.256637,1.507448,-0.514782,
        1.256637,1.509287,2.204647,
        1.256637,1.510518,1.884920,
        1.256637,1.511948,-2.632360,
        1.256637,1.515235,1.068565,
        1.256637,1.522237,3.102018,
        1.256637,1.522373,1.764991,
        1.256637,1.522416,-0.954956,
        1.256637,1.523102,2.765920,
        1.256637,1.524811,-2.844718,
        1.256637,1.527261,-0.304916,
        1.256637,1.529788,0.613208,
        1.256637,1.533648,1.644808,
        1.256637,1.535018,0.028869,
        1.256637,1.541082,2.424608,
        1.256637,1.546010,1.523462,
        1.256637,1.551135,-0.642994,
        1.256637,1.553895,0.489607,
        1.256637,1.555041,0.947813,
        1.256637,1.559671,-1.074934,
        1.256637,1.560345,2.985899,
        1.256637,1.560487,-2.323400,
        1.256637,1.561127,1.401111,
        1.256637,1.567193,2.643172,
        1.256637,1.568892,2.098362,
        1.256637,1.571702,-2.098435,
        1.256637,1.572701,-2.528730,
        1.256637,1.573572,1.278593,
        1.256637,1.574849,-0.091505,
        1.256637,1.576334,0.365236,
        1.256637,1.577029,-2.962888,
        1.256637,1.577688,-1.814941,
        1.256637,1.579542,-0.755917,
        1.256637,1.579624,-1.199759,
        1.256637,1.581515,-0.419414,
        1.256637,1.584581,-2.738702,
        1.256637,1.586948,-1.696296,
        1.256637,1.588901,2.301496,
        1.256637,1.589936,0.825020,
        1.256637,1.590018,-1.941652,
        1.256637,1.596968,-1.328358,
        1.256637,1.597178,-1.575607,
        1.256637,1.600470,-1.453196,
        1.256637,1.601408,0.242409,
        1.256637,1.602259,1.152656,
        1.256637,1.603064,2.865774,
        1.256637,1.612085,1.968802,
        1.256637,1.612720,-3.080498,
        1.256637,1.616061,2.521531,
        1.256637,1.616900,-0.207457,
        1.256637,1.620988,-0.875920,
        1.256637,1.622114,0.701509,
        1.256637,1.622749,0.121426,
        1.256637,1.622993,-2.221629,
        1.256637,1.629346,-0.542669,
        1.256637,1.630007,1.844552,
        1.256637,1.634747,-2.426198,
        1.256637,1.642346,1.722617,
        1.256637,1.643925,1.034629,
        1.256637,1.644848,2.743774,
        1.256637,1.645258,2.188109,
        1.256637,1.645417,-2.634725,
        1.256637,1.647395,3.085437,
        1.256637,1.648538,-2.855044,
        1.256637,1.649932,0.576808,
        1.256637,1.652224,-0.996105,
        1.256637,1.652871,1.600681,
        1.256637,1.657997,-0.325127,
        1.256637,1.661936,0.001037,
        1.256637,1.665187,1.477402,
        1.256637,1.665496,2.397872,
        1.256637,1.673267,-0.665453,
        1.256637,1.673557,0.452526,
        1.256637,1.679277,-1.122421,
        1.256637,1.679696,0.912482,
        1.256637,1.682420,1.353922,
        1.256637,1.685031,-2.009038,
        1.256637,1.685196,2.967046,
        1.256637,1.690107,2.620910,
        1.256637,1.690453,-2.123076,
        1.256637,1.691693,-1.871572,
        1.256637,1.691817,2.061688,
        1.256637,1.692644,1.237670,
        1.256637,1.696386,0.328123,
        1.256637,1.697162,-1.249026,
        1.256637,1.697488,-2.322785,
        1.256637,1.698726,-2.973576,
        1.256637,1.699172,-1.749717,
        1.256637,1.704478,-0.106295,
        1.256637,1.707244,-0.444770,
        1.256637,1.707500,-2.531497,
        1.256637,1.707714,-0.787321,
        1.256637,1.708430,-1.627552,
        1.256637,1.710583,-1.376890,
        1.256637,1.712415,-2.747921,
        1.256637,1.713125,-1.503073,
        1.256637,1.713452,0.789419,
        1.256637,1.720612,2.279888,
        1.256637,1.724852,0.202831,
        1.256637,1.724868,2.846246,
        1.256637,1.728940,1.931535,
        1.256637,1.732838,1.127584,
        1.256637,1.735719,-3.092485,
        1.256637,1.739676,2.496017,
        1.256637,1.739971,-0.228876,
        1.256637,1.743328,0.663347,
        1.256637,1.744927,-0.911892,
        1.256637,1.745952,0.084733,
        1.256637,1.748077,1.804354,
        1.256637,1.752959,-0.567548,
        1.256637,1.760169,1.678132,
        1.256637,1.763887,-2.218450,
        1.256637,1.766737,2.722752,
        1.256637,1.767999,1.004737,
        1.256637,1.769149,1.550298,
        1.256637,1.770119,-2.427175,
        1.256637,1.770460,2.154208,
        1.256637,1.770514,3.070982,
        1.256637,1.770735,0.537245,
        1.256637,1.772673,-2.864760,
        1.256637,1.773527,-1.039582,
        1.256637,1.776419,-2.642421,
        1.256637,1.780769,1.416330,
        1.256637,1.786634,-0.346247,
        1.256637,1.788829,-1.947553,
        1.256637,1.793737,0.410270,
        1.256637,1.794553,2.374618,
        1.256637,1.794699,-0.694061,
        1.256637,1.797469,-1.169105,
        1.256637,1.803705,0.881076,
        1.256637,1.804124,1.264249,
        1.256637,1.807825,-2.090455,
        1.256637,1.808822,-1.817375,
        1.256637,1.809020,2.949741,
        1.256637,1.812466,2.597595,
        1.256637,1.813159,2.025395,
        1.256637,1.813816,-1.298435,
        1.256637,1.815016,0.278951,
        1.256637,1.816510,-0.008631,
        1.256637,1.817772,-1.689625,
        1.256637,1.821360,-2.984352,
        1.256637,1.822312,-0.127211,
        1.256637,1.823386,-1.429807,
        1.256637,1.823834,-1.560483,
        1.256637,1.832060,-0.821627,
        1.256637,1.832690,-0.468440,
        1.256637,1.833945,-2.320278,
        1.256637,1.834787,0.751968,
        1.256637,1.840579,-2.536445,
        1.256637,1.841033,-2.756932,
        1.256637,1.846350,1.893983,
        1.256637,1.846644,1.113766,
        1.256637,1.846833,2.826543,
        1.256637,1.847325,2.249051,
        1.256637,1.854119,0.134161,
        1.256637,1.858863,-3.106022,
        1.256637,1.865046,1.763384,
        1.256637,1.865619,-0.952036,
        1.256637,1.865912,0.624453,
        1.256637,1.866694,2.473686,
        1.256637,1.867524,-0.240083,
        1.256637,1.875561,1.632831,
        1.256637,1.876049,-0.594378,
        1.256637,1.880509,1.500204,
        1.256637,1.883504,1.360457,
        1.256637,1.886971,-2.196665,
        1.256637,1.888979,0.982922,
        1.256637,1.889815,2.701570,
        1.256637,1.892144,-2.005340,
        1.256637,1.892348,0.496522,
        1.256637,1.893281,-1.083976,
        1.256637,1.893828,3.053453,
        1.256637,1.895117,2.121107,
        1.256637,1.898741,-2.875735,
        1.256637,1.904429,-2.428029,
        1.256637,1.908408,-2.648903,
        1.256637,1.911828,-0.364489,
        1.256637,1.912224,0.366565,
        1.256637,1.914854,1.214468,
        1.256637,1.915467,-1.883143,
        1.256637,1.915763,-0.724679,
        1.256637,1.915790,-1.216366,
        1.256637,1.922256,2.347806,
        1.256637,1.922616,0.844633,
        1.256637,1.925185,0.231170,
        1.256637,1.927474,-1.754616,
        1.256637,1.928342,-1.351174,
        1.256637,1.933328,-1.623328,
        1.256637,1.933878,0.034150,
        1.256637,1.934585,-1.487979,
        1.256637,1.934802,2.927634,
        1.256637,1.934862,1.989776,
        1.256637,1.936327,-0.099350,
        1.256637,1.938088,2.578507,
        1.256637,1.945677,-2.999805,
        1.256637,1.954572,-0.859208,
        1.256637,1.957868,-0.490784,
        1.256637,1.958989,0.712772,
        1.256637,1.960212,-2.307561,
        1.256637,1.964774,1.855507,
        1.256637,1.964971,-2.100533,
        1.256637,1.969112,1.088273,
        1.256637,1.971438,-2.765415,
        1.256637,1.972312,2.805134,
        1.256637,1.974950,2.220596,
        1.256637,1.975135,-2.539127,
        1.256637,1.982107,1.720829,
        1.256637,1.983195,-3.127984,
        1.256637,1.983318,-0.247182,
        1.256637,1.986720,-0.993220,
        1.256637,1.990381,1.586005,
        1.256637,1.990470,0.580283,
        1.256637,1.990523,1.450326,
        1.256637,1.993909,1.312897,
        1.256637,1.994694,2.452813,
        1.256637,1.999226,-0.620717,
        1.256637,2.005145,0.941012,
        1.256637,2.007301,0.142337,
        1.256637,2.014149,0.444138,
        1.256637,2.014338,-1.127392,
        1.256637,2.018131,3.021778,
        1.256637,2.019676,-1.980944,
        1.256637,2.019912,2.686612,
        1.256637,2.020545,2.088672,
        1.256637,2.026074,0.300945,
        1.256637,2.026196,-2.889057,
        1.256637,2.031959,-1.268178,
        1.256637,2.032917,-2.199917,
        1.256637,2.033572,-1.842884,
        1.256637,2.034712,-2.419505,
        1.256637,2.035818,-0.379479,
        1.256637,2.036004,-0.757062,
        1.256637,2.040501,-1.411488,
        1.256637,2.042874,-1.701367,
        1.256637,2.043060,-2.653746,
        1.256637,2.043729,-1.557938,
        1.256637,2.048538,0.806623,
        1.256637,2.049006,1.199930,
        1.256637,2.049502,-0.137989,
        1.256637,2.052181,2.326959,
        1.256637,2.052794,-0.008526,
        1.256637,2.057608,1.951726,
        1.256637,2.065568,2.567730,
        1.256637,2.070371,2.879808,
        1.256637,2.071256,-3.018961,
        1.256637,2.079746,-0.897699,
        1.256637,2.082784,1.042403,
        1.256637,2.083752,-0.511086,
        1.256637,2.084401,1.810502,
        1.256637,2.085731,0.671081,
        1.256637,2.097496,1.373172,
        1.256637,2.099185,1.667006,
        1.256637,2.100482,-2.301084,
        1.256637,2.103118,2.194745,
        1.256637,2.103223,1.521685,
        1.256637,2.103695,-2.777707,
        1.256637,2.107523,3.129581,
        1.256637,2.108277,-2.095907,
        1.256637,2.108891,-2.532166,
        1.256637,2.111003,-0.268632,
        1.256637,2.112276,-1.027246,
        1.256637,2.114380,0.531457,
        1.256637,2.119166,0.100552,
        1.256637,2.119564,2.754818,
        1.256637,2.122474,-0.648314,
        1.256637,2.125183,2.445598,
        1.256637,2.129501,0.238290,
        1.256637,2.131178,0.386663,
        1.256637,2.131223,-1.174311,
        1.256637,2.134269,-1.948042,
        1.256637,2.134684,0.907065,
        1.256637,2.137615,2.988806,
        1.256637,2.145635,-1.327708,
        1.256637,2.146467,2.055081,
        1.256637,2.146582,-1.793788,
        1.256637,2.149210,-1.482321,
        1.256637,2.152901,-1.636788,
        1.256637,2.156337,1.151558,
        1.256637,2.156474,-2.912065,
        1.256637,2.157194,-0.792563,
        1.256637,2.166724,-0.401638,
        1.256637,2.166980,-0.134879,
        1.256637,2.173644,-2.403429,
        1.256637,2.175711,-2.655523,
        1.256637,2.177132,2.600419,
        1.256637,2.178879,0.767722,
        1.256637,2.180752,1.908485,
        1.256637,2.181338,2.313721,
        1.256637,2.185580,1.287673,
        1.256637,2.195400,-2.217546,
        1.256637,2.200418,-3.056442,
        1.256637,2.201829,-0.001735,
        1.256637,2.203804,1.756176,
        1.256637,2.211602,1.443930,
        1.256637,2.212686,-0.544809,
        1.256637,2.213937,0.621892,
        1.256637,2.214075,1.600691,
        1.256637,2.216256,1.020029,
        1.256637,2.219060,2.886631,
        1.256637,2.223375,-0.906185,
        1.256637,2.226002,-1.060248,
        1.256637,2.226413,-2.064034,
        1.256637,2.230554,2.171334,
        1.256637,2.232413,2.735777,
        1.256637,2.233709,3.070830,
        1.256637,2.234383,-2.792065,
        1.256637,2.237183,0.468799,
        1.256637,2.237695,-0.271388,
        1.256637,2.239419,0.151693,
        1.256637,2.243347,-2.521049,
        1.256637,2.244035,-1.229456,
        1.256637,2.244355,0.309899,
        1.256637,2.245175,2.460236,
        1.256637,2.245628,-1.899216,
        1.256637,2.245973,-0.702234,
        1.256637,2.255601,-1.728131,
        1.256637,2.256702,-1.401367,
        1.256637,2.261466,-1.557500,
        1.256637,2.267390,0.878499,
        1.256637,2.272099,2.018594,
        1.256637,2.278618,1.170386,
        1.256637,2.283304,-2.345584,
        1.256637,2.285175,-2.943090,
        1.256637,2.293000,-0.124451,
        1.256637,2.294058,-0.421070,
        1.256637,2.303283,1.855938,
        1.256637,2.304968,2.608978,
        1.256637,2.305347,2.311346,
        1.256637,2.307188,-2.656134,
        1.256637,2.307301,1.336200,
        1.256637,2.309655,0.726020,
        1.256637,2.317634,-2.185225,
        1.256637,2.321510,1.685664,
        1.256637,2.324376,-3.111006,
        1.256637,2.324731,1.511186,
        1.256637,2.329587,0.035498,
        1.256637,2.330158,2.977264,
        1.256637,2.335096,-0.588468,
        1.256637,2.336046,-1.115799,
        1.256637,2.338060,2.803613,
        1.256637,2.338083,-0.943400,
        1.256637,2.340210,0.562145,
        1.256637,2.341931,1.023926,
        1.256637,2.343265,-0.779061,
        1.256637,2.343434,-2.013692,
        1.256637,2.353029,-1.310529,
        1.256637,2.354868,0.211543,
        1.256637,2.355120,0.388593,
        1.256637,2.355165,-1.833173,
        1.256637,2.355462,2.151441,
        1.256637,2.362948,-2.487477,
        1.256637,2.363101,-2.810456,
        1.256637,2.363394,-0.272918,
        1.256637,2.364509,-1.644533,
        1.256637,2.370465,2.466824,
        1.256637,2.386110,-1.481351,
        1.256637,2.394960,0.862133,
        1.256637,2.395716,1.977184,
        1.256637,2.396339,1.200900,
        1.256637,2.407941,-2.316753,
        1.256637,2.409450,-2.988432,
        1.256637,2.417202,-0.103363,
        1.256637,2.417386,-0.443097,
        1.256637,2.417950,2.666662,
        1.256637,2.423190,1.789783,
        1.256637,2.423692,1.392696,
        1.256637,2.431122,3.086379,
        1.256637,2.431406,-2.647820,
        1.256637,2.431584,2.307631,
        1.256637,2.434090,1.592818,
        1.256637,2.435993,0.683844,
        1.256637,2.440898,-2.133300,
        1.256637,2.442222,-1.195837,
        1.256637,2.443112,2.883761,
        1.256637,2.449191,-0.633651,
        1.256637,2.450821,0.084741,
        1.256637,2.453499,-1.008861,
        1.256637,2.460010,-1.938011,
        1.256637,2.460199,-0.822008,
        1.256637,2.460902,0.490079,
        1.256637,2.461774,-1.734913,
        1.256637,2.463986,1.034407,
        1.256637,2.467647,0.287667,
        1.256637,2.479849,2.126925,
        1.256637,2.485907,-2.835471,
        1.256637,2.485945,-1.374853,
        1.256637,2.487747,2.505912,
        1.256637,2.487957,-0.274279,
        1.256637,2.489741,-2.468428,
        1.256637,2.512538,1.239900,
        1.256637,2.517528,1.923289,
        1.256637,2.518454,-1.565852,
        1.256637,2.519348,0.843503,
        1.256637,2.522766,-3.059231,
        1.256637,2.528686,2.737228,
        1.256637,2.534798,-2.266671,
        1.256637,2.536533,1.466274,
        1.256637,2.538504,1.699812,
        1.256637,2.539516,-0.476942,
        1.256637,2.539532,-0.074548,
        1.256637,2.543197,2.990252,
        1.256637,2.554022,2.320619,
        1.256637,2.556919,-2.655005,
        1.256637,2.558146,0.627191,
        1.256637,2.558223,-1.164189,
        1.256637,2.563452,-2.044028,
        1.256637,2.568462,0.150655,
        1.256637,2.569130,-0.700837,
        1.256637,2.570050,-1.804492,
        1.256637,2.572710,-0.934665,
        1.256637,2.575994,0.391718,
        1.256637,2.584774,1.046703,
        1.256637,2.601432,2.102535,
        1.256637,2.603559,2.555114,
        1.256637,2.604150,-2.885208,
        1.256637,2.608630,-1.378011,
        1.256637,2.611256,-0.279110,
        1.256637,2.617344,-2.439480,
        1.256637,2.627402,1.293552,
        1.256637,2.635396,1.848850,
        1.256637,2.635686,2.834950,
        1.256637,2.637087,3.132730,
        1.256637,2.642029,0.813274,
        1.256637,2.645260,1.570203,
        1.256637,2.646182,-1.614036,
        1.256637,2.660113,-2.186513,
        1.256637,2.660483,-0.034608,
        1.256637,2.661713,-0.529511,
        1.256637,2.672720,-1.110422,
        1.256637,2.676626,2.335353,
        1.256637,2.676638,0.540364,
        1.256637,2.676780,-1.901158,
        1.256637,2.681411,-2.667965,
        1.256637,2.682152,0.245751,
        1.256637,2.682859,-0.817071,
        1.256637,2.705107,1.062142,
        1.256637,2.717705,2.625995,
        1.256637,2.721001,2.060097,
        1.256637,2.722319,-2.954069,
        1.256637,2.729840,-1.368237,
        1.256637,2.734349,-0.289298,
        1.256637,2.738409,2.975671,
        1.256637,2.740366,1.374341,
        1.256637,2.744633,-2.391311,
        1.256637,2.747398,1.727166,
        1.256637,2.763799,0.760372,
        1.256637,2.768238,-1.679243,
        1.256637,2.779245,-2.046268,
        1.256637,2.780208,-0.626453,
        1.256637,2.780471,0.027732,
        1.256637,2.784696,-1.013128,
        1.256637,2.789817,0.395431,
        1.256637,2.801154,2.353143,
        1.256637,2.807091,-2.687477,
        1.256637,2.825762,1.085035,
        1.256637,2.830209,2.734943,
        1.256637,2.833517,-3.085508,
        1.256637,2.838301,1.980061,
        1.256637,2.849871,1.512012,
        1.256637,2.850089,-1.343899,
        1.256637,2.856905,-0.318441,
        1.256637,2.870445,-2.291001,
        1.256637,2.885017,0.655562,
        1.256637,2.886428,-1.793217,
        1.256637,2.890337,-0.825633,
        1.256637,2.898932,0.139973,
        1.256637,2.931927,-2.728768,
        1.256637,2.932096,2.377790,
        1.256637,2.937409,2.947245,
        1.256637,2.948048,1.133127,
        1.256637,2.950896,1.797889,
        1.256637,2.970019,-1.282382,
        1.256637,2.978333,-0.412090,
        1.256637,2.996437,-2.089121,
        1.256637,3.011274,0.434110,
        1.256637,3.050410,3.138492,
        1.256637,3.064492,1.647259,
        1.256637,3.088868,-0.989501,
        1.361357,0.053884,-2.408394,
        1.361357,0.077645,1.453433,
        1.361357,0.088408,-0.311160,
        1.361357,0.142314,2.590387,
        1.361357,0.147206,-1.385164,
        1.361357,0.163506,0.573203,
        1.361357,0.165485,-2.875396,
        1.361357,0.182564,-2.125821,
        1.361357,0.190569,1.929141,
        1.361357,0.201102,1.249082,
        1.361357,0.201397,-0.716646,
        1.361357,0.210568,-0.069559,
        1.361357,0.251672,2.848900,
        1.361357,0.266262,-1.678068,
        1.361357,0.269108,-1.215563,
        1.361357,0.273777,-2.506756,
        1.361357,0.279854,0.814581,
        1.361357,0.281664,-2.969147,
        1.361357,0.283625,2.405603,
        1.361357,0.284058,0.357350,
        1.361357,0.294107,1.593880,
        1.361357,0.302491,-0.400410,
        1.361357,0.308059,2.021491,
        1.361357,0.312219,-2.072057,
        1.361357,0.321905,-0.829121,
        1.361357,0.329984,1.201621,
        1.361357,0.339714,-0.013712,
        1.361357,0.368784,2.974974,
        1.361357,0.376428,-1.495532,
        1.361357,0.378103,2.646436,
        1.361357,0.385088,0.590355,
        1.361357,0.386248,-2.690943,
        1.361357,0.389851,-2.340031,
        1.361357,0.391502,-1.171374,
        1.361357,0.398249,-1.812694,
        1.361357,0.400542,-2.995729,
        1.361357,0.404273,0.918163,
        1.361357,0.405617,1.773533,
        1.361357,0.409009,-0.570243,
        1.361357,0.410700,0.270433,
        1.361357,0.415893,1.464890,
        1.361357,0.420680,2.088164,
        1.361357,0.424766,-0.258729,
        1.361357,0.426507,2.367872,
        1.361357,0.441292,-0.890727,
        1.361357,0.451839,-2.078193,
        1.361357,0.459270,1.191562,
        1.361357,0.470622,0.014002,
        1.361357,0.481292,-2.529098,
        1.361357,0.482460,2.849843,
        1.361357,0.492205,-1.397963,
        1.361357,0.493314,3.117264,
        1.361357,0.497870,0.727828,
        1.361357,0.499783,-1.654379,
        1.361357,0.502651,0.463333,
        1.361357,0.505463,2.594557,
        1.361357,0.506104,-2.874910,
        1.361357,0.513579,-1.147354,
        1.361357,0.518569,1.644237,
        1.361357,0.519001,1.900488,
        1.361357,0.519277,-0.673809,
        1.361357,0.524769,-0.421955,
        1.361357,0.525564,-2.288838,
        1.361357,0.532541,0.978737,
        1.361357,0.534610,2.154808,
        1.361357,0.535020,-1.891846,
        1.361357,0.540208,1.400497,
        1.361357,0.542006,0.228337,
        1.361357,0.550762,-0.180771,
        1.361357,0.560531,2.378474,
        1.361357,0.561629,-0.924576,
        1.361357,0.567465,-2.679392,
        1.361357,0.588527,1.181755,
        1.361357,0.591643,-2.098364,
        1.361357,0.594963,2.997575,
        1.361357,0.599643,-3.045235,
        1.361357,0.599962,2.773556,
        1.361357,0.602620,0.029270,
        1.361357,0.605711,0.598454,
        1.361357,0.607708,-1.541026,
        1.361357,0.609360,-2.464379,
        1.361357,0.610738,-1.326786,
        1.361357,0.616968,0.820489,
        1.361357,0.626379,1.773610,
        1.361357,0.628626,-1.751180,
        1.361357,0.630671,0.391949,
        1.361357,0.632622,-0.737642,
        1.361357,0.632708,2.566722,
        1.361357,0.632797,1.989858,
        1.361357,0.634455,-0.534790,
        1.361357,0.636205,1.561170,
        1.361357,0.637681,-1.123423,
        1.361357,0.641441,-2.842109,
        1.361357,0.644807,-0.327139,
        1.361357,0.652440,2.197129,
        1.361357,0.660897,-2.278593,
        1.361357,0.664008,1.360091,
        1.361357,0.670850,-1.945257,
        1.361357,0.671396,1.004859,
        1.361357,0.673985,0.203855,
        1.361357,0.680326,-0.133370,
        1.361357,0.685083,-0.942363,
        1.361357,0.688204,-2.632204,
        1.361357,0.689777,2.385011,
        1.361357,0.700065,3.115430,
        1.361357,0.704080,2.911636,
        1.361357,0.712883,1.174216,
        1.361357,0.715701,0.699906,
        1.361357,0.718851,-1.442215,
        1.361357,0.721019,2.721536,
        1.361357,0.725300,-2.975134,
        1.361357,0.727136,-1.633288,
        1.361357,0.727399,-2.120567,
        1.361357,0.730751,0.522488,
        1.361357,0.732892,-1.254330,
        1.361357,0.734333,-2.446184,
        1.361357,0.735125,0.037270,
        1.361357,0.737893,1.866463,
        1.361357,0.740650,1.681075,
        1.361357,0.745971,-0.781286,
        1.361357,0.749432,-0.442559,
        1.361357,0.750102,2.050636,
        1.361357,0.751046,0.863821,
        1.361357,0.757018,-0.608833,
        1.361357,0.757046,-1.816019,
        1.361357,0.757389,1.501184,
        1.361357,0.759800,2.546181,
        1.361357,0.759859,0.351923,
        1.361357,0.761141,-2.781102,
        1.361357,0.769035,-1.078574,
        1.361357,0.770166,-0.266604,
        1.361357,0.772218,2.227380,
        1.361357,0.788382,1.331051,
        1.361357,0.790269,-2.282409,
        1.361357,0.803410,-1.984567,
        1.361357,0.803813,0.191520,
        1.361357,0.806832,3.022155,
        1.361357,0.807476,-2.597675,
        1.361357,0.812512,-0.107335,
        1.361357,0.814161,1.010949,
        1.361357,0.817732,-3.084486,
        1.361357,0.817846,2.390688,
        1.361357,0.819323,2.845960,
        1.361357,0.819430,-0.915017,
        1.361357,0.829198,-1.532491,
        1.361357,0.829531,0.665695,
        1.361357,0.829816,1.167147,
        1.361357,0.833613,-1.357492,
        1.361357,0.842945,-2.907318,
        1.361357,0.844883,2.679708,
        1.361357,0.848189,-1.705003,
        1.361357,0.849888,1.767901,
        1.361357,0.851126,0.493370,
        1.361357,0.852911,1.933394,
        1.361357,0.857516,-2.432863,
        1.361357,0.858261,-1.191729,
        1.361357,0.858279,-2.138895,
        1.361357,0.859723,-0.734928,
        1.361357,0.859754,1.603292,
        1.361357,0.861655,-0.394821,
        1.361357,0.864300,0.043549,
        1.361357,0.866940,0.851229,
        1.361357,0.869748,2.095969,
        1.361357,0.877352,-0.574668,
        1.361357,0.879254,-2.733338,
        1.361357,0.882554,1.443757,
        1.361357,0.884060,0.334199,
        1.361357,0.885626,-1.866061,
        1.361357,0.889834,2.529957,
        1.361357,0.893140,2.251698,
        1.361357,0.897463,-0.240822,
        1.361357,0.898758,-1.031948,
        1.361357,0.913918,1.287618,
        1.361357,0.916555,-2.287173,
        1.361357,0.918305,3.107845,
        1.361357,0.919522,2.948277,
        1.361357,0.927138,-2.569686,
        1.361357,0.929031,0.185383,
        1.361357,0.933766,-3.015917,
        1.361357,0.934873,-2.010822,
        1.361357,0.935636,-1.445855,
        1.361357,0.936702,-0.866461,
        1.361357,0.937395,0.729108,
        1.361357,0.937910,2.787575,
        1.361357,0.939144,-0.092962,
        1.361357,0.941718,1.125085,
        1.361357,0.942873,-1.607680,
        1.361357,0.943701,0.980328,
        1.361357,0.947774,2.395256,
        1.361357,0.952433,0.584347,
        1.361357,0.954456,-1.287293,
        1.361357,0.958731,-2.852552,
        1.361357,0.961838,1.839284,
        1.361357,0.961898,-0.469425,
        1.361357,0.965426,1.686615,
        1.361357,0.970108,1.990771,
        1.361357,0.972820,-1.762780,
        1.361357,0.973500,-0.711881,
        1.361357,0.975253,2.641082,
        1.361357,0.977824,0.441928,
        1.361357,0.981485,-2.424766,
        1.361357,0.981877,1.537100,
        1.361357,0.982609,-1.136920,
        1.361357,0.984265,-2.156073,
        1.361357,0.989150,0.045493,
        1.361357,0.993052,2.137478,
        1.361357,0.997766,-2.693871,
        1.361357,0.999847,-0.334738,
        1.361357,1.001920,0.848054,
        1.361357,1.006146,1.389734,
        1.361357,1.012180,0.304066,
        1.361357,1.016938,2.270674,
        1.361357,1.017371,-0.979790,
        1.361357,1.019495,-1.894487,
        1.361357,1.023796,-0.583691,
        1.361357,1.026525,2.506945,
        1.361357,1.027004,-0.203653,
        1.361357,1.028027,3.036573,
        1.361357,1.031435,1.241021,
        1.361357,1.032316,2.884221,
        1.361357,1.035863,-3.110475,
        1.361357,1.042085,-2.292950,
        1.361357,1.043272,-1.522986,
        1.361357,1.046265,-2.958847,
        1.361357,1.049272,-1.376344,
        1.361357,1.050039,-2.552086,
        1.361357,1.053682,0.689863,
        1.361357,1.055458,0.170394,
        1.361357,1.056622,-0.831866,
        1.361357,1.057515,1.097998,
        1.361357,1.058706,-2.036438,
        1.361357,1.061389,-1.672248,
        1.361357,1.063758,2.743093,
        1.361357,1.070157,-0.073422,
        1.361357,1.072154,1.761506,
        1.361357,1.074029,1.909193,
        1.361357,1.074204,0.545276,
        1.361357,1.074714,-2.805147,
        1.361357,1.076262,0.966104,
        1.361357,1.077642,-1.227073,
        1.361357,1.084151,1.618525,
        1.361357,1.085085,2.377673,
        1.361357,1.086822,-0.441112,
        1.361357,1.087142,2.058815,
        1.361357,1.098383,-0.696430,
        1.361357,1.099138,-1.084033,
        1.361357,1.103208,1.477773,
        1.361357,1.103783,0.408254,
        1.361357,1.106275,-2.420800,
        1.361357,1.108079,2.612943,
        1.361357,1.109345,-2.172718,
        1.361357,1.115957,-1.787597,
        1.361357,1.118930,0.045983,
        1.361357,1.119777,0.834702,
        1.361357,1.121943,-2.669042,
        1.361357,1.124930,2.219917,
        1.361357,1.125217,1.336294,
        1.361357,1.130958,2.972931,
        1.361357,1.131400,-0.306629,
        1.361357,1.137135,-0.936514,
        1.361357,1.139125,-1.924528,
        1.361357,1.139877,0.276382,
        1.361357,1.142003,-3.058022,
        1.361357,1.147363,-0.558448,
        1.361357,1.150348,-0.176418,
        1.361357,1.151198,1.197016,
        1.361357,1.153930,-1.597104,
        1.361357,1.155389,2.837148,
        1.361357,1.155458,-1.460623,
        1.361357,1.157331,-2.906362,
        1.361357,1.158594,2.487618,
        1.361357,1.162103,-1.322553,
        1.361357,1.165407,3.092861,
        1.361357,1.168025,0.717075,
        1.361357,1.168406,-2.301519,
        1.361357,1.174782,-2.538755,
        1.361357,1.177998,1.061256,
        1.361357,1.178898,1.827305,
        1.361357,1.181331,-2.060330,
        1.361357,1.182540,-0.798952,
        1.361357,1.184569,0.147517,
        1.361357,1.184734,1.968919,
        1.361357,1.184804,0.593386,
        1.361357,1.188278,1.690808,
        1.361357,1.194360,2.110440,
        1.361357,1.194776,2.710455,
        1.361357,1.197745,-1.173705,
        1.361357,1.199898,-2.777257,
        1.361357,1.201164,2.355931,
        1.361357,1.203760,1.556404,
        1.361357,1.205845,-0.415993,
        1.361357,1.206585,0.933648,
        1.361357,1.206950,-0.037248,
        1.361357,1.209341,0.471162,
        1.361357,1.213483,-1.039417,
        1.361357,1.218146,-0.672643,
        1.361357,1.221608,1.421669,
        1.361357,1.226043,-1.693533,
        1.361357,1.232870,-1.822773,
        1.361357,1.234375,-2.421194,
        1.361357,1.234531,-2.190148,
        1.361357,1.239414,0.348959,
        1.361357,1.240183,2.587255,
        1.361357,1.242312,2.941234,
        1.361357,1.244075,1.286644,
        1.361357,1.245961,-2.997461,
        1.361357,1.246495,-2.646872,
        1.361357,1.247736,2.242810,
        1.361357,1.253604,0.812222,
        1.361357,1.257816,-0.894317,
        1.361357,1.258247,-1.953634,
        1.361357,1.258993,-0.284099,
        1.361357,1.262380,-0.157609,
        1.361357,1.262791,-1.549229,
        1.361357,1.263772,-0.533073,
        1.361357,1.266731,-1.412461,
        1.361357,1.268591,-3.122146,
        1.361357,1.269614,-1.275246,
        1.361357,1.270745,1.152533,
        1.361357,1.273606,0.224545,
        1.361357,1.279716,2.812101,
        1.361357,1.283786,2.462124,
        1.361357,1.284798,0.078209,
        1.361357,1.285402,-2.882049,
        1.361357,1.286936,1.886053,
        1.361357,1.289408,0.683122,
        1.361357,1.294736,1.756184,
        1.361357,1.296978,-2.311511,
        1.361357,1.298546,2.020601,
        1.361357,1.300924,1.022478,
        1.361357,1.303305,3.043152,
        1.361357,1.304210,-2.084225,
        1.361357,1.305241,-2.530249,
        1.361357,1.307513,1.628003,
        1.361357,1.310722,0.558267,
        1.361357,1.314142,2.139329,
        1.361357,1.316087,-0.777971,
        1.361357,1.317907,-1.121023,
        1.361357,1.321019,-2.749776,
        1.361357,1.321609,2.687717,
        1.361357,1.322568,1.499415,
        1.361357,1.322972,-0.395699,
        1.361357,1.325286,-0.994192,
        1.361357,1.329146,-0.648513,
        1.361357,1.329795,-0.042570,
        1.361357,1.331651,2.344497,
        1.361357,1.336522,0.435098,
        1.361357,1.339410,0.900087,
        1.361357,1.339550,1.370353,
        1.361357,1.340441,-1.731580,
        1.361357,1.351855,-1.855234,
        1.361357,1.360161,-3.018404,
        1.361357,1.361031,-1.613971,
        1.361357,1.361196,2.914780,
        1.361357,1.361998,-2.207254,
        1.361357,1.362302,1.239384,
        1.361357,1.365228,2.563808,
        1.361357,1.365452,0.310522,
        1.361357,1.367028,-2.420041,
        1.361357,1.373422,-1.484281,
        1.361357,1.374245,-1.226417,
        1.361357,1.375020,-0.170393,
        1.361357,1.376560,-1.355481,
        1.361357,1.377070,-1.980984,
        1.361357,1.377578,-2.636339,
        1.361357,1.378096,0.774626,
        1.361357,1.379086,-0.508192,
        1.361357,1.382562,0.177141,
        1.361357,1.388849,2.233847,
        1.361357,1.389268,-0.886349,
        1.361357,1.391324,1.109396,
        1.361357,1.391699,-0.288725,
        1.361357,1.392869,3.136903,
        1.361357,1.393612,1.952016,
        1.361357,1.395545,-2.850807,
        1.361357,1.401926,1.822930,
        1.361357,1.402149,2.788898,
        1.361357,1.409552,0.647980,
        1.361357,1.410840,0.056219,
        1.361357,1.411944,2.444377,
        1.361357,1.413639,1.697982,
        1.361357,1.426479,1.573427,
        1.361357,1.426725,2.102242,
        1.361357,1.427868,0.984637,
        1.361357,1.429580,-2.107232,
        1.361357,1.431375,-2.313269,
        1.361357,1.431907,3.015410,
        1.361357,1.432325,-1.029054,
        1.361357,1.433341,0.524268,
        1.361357,1.435181,-0.739771,
        1.361357,1.437208,-2.529085,
        1.361357,1.439041,-0.617966,
        1.361357,1.441473,1.448544,
        1.361357,1.444527,2.664744,
        1.361357,1.450513,-0.400625,
        1.361357,1.451907,-0.066823,
        1.361357,1.452754,-2.738631,
        1.361357,1.454061,-2.950416,
        1.361357,1.456736,1.323154,
        1.361357,1.458370,0.400433,
        1.361357,1.459314,-1.765003,
        1.361357,1.459902,-1.145835,
        1.361357,1.464952,2.329245,
        1.361357,1.466050,0.862104,
        1.361357,1.471330,-1.886160,
        1.361357,1.475119,-1.648255,
        1.361357,1.480878,1.194576,
        1.361357,1.480942,0.273235,
        1.361357,1.481086,2.890269,
        1.361357,1.484118,-1.529156,
        1.361357,1.484741,-1.279219,
        1.361357,1.487681,-1.406069,
        1.361357,1.487886,-3.066037,
        1.361357,1.490555,2.543284,
        1.361357,1.492424,-0.190932,
        1.361357,1.495002,-0.839810,
        1.361357,1.495297,2.007188,
        1.361357,1.496414,-2.009267,
        1.361357,1.498261,-2.425270,
        1.361357,1.498907,-2.207327,
        1.361357,1.500204,0.737638,
        1.361357,1.502469,0.151156,
        1.361357,1.507448,-0.514782,
        1.361357,1.509287,2.204647,
        1.361357,1.510518,1.884920,
        1.361357,1.511948,-2.632360,
        1.361357,1.515235,1.068565,
        1.361357,1.522237,3.102018,
        1.361357,1.522373,1.764991,
        1.361357,1.522416,-0.954956,
        1.361357,1.523102,2.765920,
        1.361357,1.524811,-2.844718,
        1.361357,1.527261,-0.304916,
        1.361357,1.529788,0.613208,
        1.361357,1.533648,1.644808,
        1.361357,1.535018,0.028869,
        1.361357,1.541082,2.424608,
        1.361357,1.546010,1.523462,
        1.361357,1.551135,-0.642994,
        1.361357,1.553895,0.489607,
        1.361357,1.555041,0.947813,
        1.361357,1.559671,-1.074934,
        1.361357,1.560345,2.985899,
        1.361357,1.560487,-2.323400,
        1.361357,1.561127,1.401111,
        1.361357,1.567193,2.643172,
        1.361357,1.568892,2.098362,
        1.361357,1.571702,-2.098435,
        1.361357,1.572701,-2.528730,
        1.361357,1.573572,1.278593,
        1.361357,1.574849,-0.091505,
        1.361357,1.576334,0.365236,
        1.361357,1.577029,-2.962888,
        1.361357,1.577688,-1.814941,
        1.361357,1.579542,-0.755917,
        1.361357,1.579624,-1.199759,
        1.361357,1.581515,-0.419414,
        1.361357,1.584581,-2.738702,
        1.361357,1.586948,-1.696296,
        1.361357,1.588901,2.301496,
        1.361357,1.589936,0.825020,
        1.361357,1.590018,-1.941652,
        1.361357,1.596968,-1.328358,
        1.361357,1.597178,-1.575607,
        1.361357,1.600470,-1.453196,
        1.361357,1.601408,0.242409,
        1.361357,1.602259,1.152656,
        1.361357,1.603064,2.865774,
        1.361357,1.612085,1.968802,
        1.361357,1.612720,-3.080498,
        1.361357,1.616061,2.521531,
        1.361357,1.616900,-0.207457,
        1.361357,1.620988,-0.875920,
        1.361357,1.622114,0.701509,
        1.361357,1.622749,0.121426,
        1.361357,1.622993,-2.221629,
        1.361357,1.629346,-0.542669,
        1.361357,1.630007,1.844552,
        1.361357,1.634747,-2.426198,
        1.361357,1.642346,1.722617,
        1.361357,1.643925,1.034629,
        1.361357,1.644848,2.743774,
        1.361357,1.645258,2.188109,
        1.361357,1.645417,-2.634725,
        1.361357,1.647395,3.085437,
        1.361357,1.648538,-2.855044,
        1.361357,1.649932,0.576808,
        1.361357,1.652224,-0.996105,
        1.361357,1.652871,1.600681,
        1.361357,1.657997,-0.325127,
        1.361357,1.661936,0.001037,
        1.361357,1.665187,1.477402,
        1.361357,1.665496,2.397872,
        1.361357,1.673267,-0.665453,
        1.361357,1.673557,0.452526,
        1.361357,1.679277,-1.122421,
        1.361357,1.679696,0.912482,
        1.361357,1.682420,1.353922,
        1.361357,1.685031,-2.009038,
        1.361357,1.685196,2.967046,
        1.361357,1.690107,2.620910,
        1.361357,1.690453,-2.123076,
        1.361357,1.691693,-1.871572,
        1.361357,1.691817,2.061688,
        1.361357,1.692644,1.237670,
        1.361357,1.696386,0.328123,
        1.361357,1.697162,-1.249026,
        1.361357,1.697488,-2.322785,
        1.361357,1.698726,-2.973576,
        1.361357,1.699172,-1.749717,
        1.361357,1.704478,-0.106295,
        1.361357,1.707244,-0.444770,
        1.361357,1.707500,-2.531497,
        1.361357,1.707714,-0.787321,
        1.361357,1.708430,-1.627552,
        1.361357,1.710583,-1.376890,
        1.361357,1.712415,-2.747921,
        1.361357,1.713125,-1.503073,
        1.361357,1.713452,0.789419,
        1.361357,1.720612,2.279888,
        1.361357,1.724852,0.202831,
        1.361357,1.724868,2.846246,
        1.361357,1.728940,1.931535,
        1.361357,1.732838,1.127584,
        1.361357,1.735719,-3.092485,
        1.361357,1.739676,2.496017,
        1.361357,1.739971,-0.228876,
        1.361357,1.743328,0.663347,
        1.361357,1.744927,-0.911892,
        1.361357,1.745952,0.084733,
        1.361357,1.748077,1.804354,
        1.361357,1.752959,-0.567548,
        1.361357,1.760169,1.678132,
        1.361357,1.763887,-2.218450,
        1.361357,1.766737,2.722752,
        1.361357,1.767999,1.004737,
        1.361357,1.769149,1.550298,
        1.361357,1.770119,-2.427175,
        1.361357,1.770460,2.154208,
        1.361357,1.770514,3.070982,
        1.361357,1.770735,0.537245,
        1.361357,1.772673,-2.864760,
        1.361357,1.773527,-1.039582,
        1.361357,1.776419,-2.642421,
        1.361357,1.780769,1.416330,
        1.361357,1.786634,-0.346247,
        1.361357,1.788829,-1.947553,
        1.361357,1.793737,0.410270,
        1.361357,1.794553,2.374618,
        1.361357,1.794699,-0.694061,
        1.361357,1.797469,-1.169105,
        1.361357,1.803705,0.881076,
        1.361357,1.804124,1.264249,
        1.361357,1.807825,-2.090455,
        1.361357,1.808822,-1.817375,
        1.361357,1.809020,2.949741,
        1.361357,1.812466,2.597595,
        1.361357,1.813159,2.025395,
        1.361357,1.813816,-1.298435,
        1.361357,1.815016,0.278951,
        1.361357,1.816510,-0.008631,
        1.361357,1.817772,-1.689625,
        1.361357,1.821360,-2.984352,
        1.361357,1.822312,-0.127211,
        1.361357,1.823386,-1.429807,
        1.361357,1.823834,-1.560483,
        1.361357,1.832060,-0.821627,
        1.361357,1.832690,-0.468440,
        1.361357,1.833945,-2.320278,
        1.361357,1.834787,0.751968,
        1.361357,1.840579,-2.536445,
        1.361357,1.841033,-2.756932,
        1.361357,1.846350,1.893983,
        1.361357,1.846644,1.113766,
        1.361357,1.846833,2.826543,
        1.361357,1.847325,2.249051,
        1.361357,1.854119,0.134161,
        1.361357,1.858863,-3.106022,
        1.361357,1.865046,1.763384,
        1.361357,1.865619,-0.952036,
        1.361357,1.865912,0.624453,
        1.361357,1.866694,2.473686,
        1.361357,1.867524,-0.240083,
        1.361357,1.875561,1.632831,
        1.361357,1.876049,-0.594378,
        1.361357,1.880509,1.500204,
        1.361357,1.883504,1.360457,
        1.361357,1.886971,-2.196665,
        1.361357,1.888979,0.982922,
        1.361357,1.889815,2.701570,
        1.361357,1.892144,-2.005340,
        1.361357,1.892348,0.496522,
        1.361357,1.893281,-1.083976,
        1.361357,1.893828,3.053453,
        1.361357,1.895117,2.121107,
        1.361357,1.898741,-2.875735,
        1.361357,1.904429,-2.428029,
        1.361357,1.908408,-2.648903,
        1.361357,1.911828,-0.364489,
        1.361357,1.912224,0.366565,
        1.361357,1.914854,1.214468,
        1.361357,1.915467,-1.883143,
        1.361357,1.915763,-0.724679,
        1.361357,1.915790,-1.216366,
        1.361357,1.922256,2.347806,
        1.361357,1.922616,0.844633,
        1.361357,1.925185,0.231170,
        1.361357,1.927474,-1.754616,
        1.361357,1.928342,-1.351174,
        1.361357,1.933328,-1.623328,
        1.361357,1.933878,0.034150,
        1.361357,1.934585,-1.487979,
        1.361357,1.934802,2.927634,
        1.361357,1.934862,1.989776,
        1.361357,1.936327,-0.099350,
        1.361357,1.938088,2.578507,
        1.361357,1.945677,-2.999805,
        1.361357,1.954572,-0.859208,
        1.361357,1.957868,-0.490784,
        1.361357,1.958989,0.712772,
        1.361357,1.960212,-2.307561,
        1.361357,1.964774,1.855507,
        1.361357,1.964971,-2.100533,
        1.361357,1.969112,1.088273,
        1.361357,1.971438,-2.765415,
        1.361357,1.972312,2.805134,
        1.361357,1.974950,2.220596,
        1.361357,1.975135,-2.539127,
        1.361357,1.982107,1.720829,
        1.361357,1.983195,-3.127984,
        1.361357,1.983318,-0.247182,
        1.361357,1.986720,-0.993220,
        1.361357,1.990381,1.586005,
        1.361357,1.990470,0.580283,
        1.361357,1.990523,1.450326,
        1.361357,1.993909,1.312897,
        1.361357,1.994694,2.452813,
        1.361357,1.999226,-0.620717,
        1.361357,2.005145,0.941012,
        1.361357,2.007301,0.142337,
        1.361357,2.014149,0.444138,
        1.361357,2.014338,-1.127392,
        1.361357,2.018131,3.021778,
        1.361357,2.019676,-1.980944,
        1.361357,2.019912,2.686612,
        1.361357,2.020545,2.088672,
        1.361357,2.026074,0.300945,
        1.361357,2.026196,-2.889057,
        1.361357,2.031959,-1.268178,
        1.361357,2.032917,-2.199917,
        1.361357,2.033572,-1.842884,
        1.361357,2.034712,-2.419505,
        1.361357,2.035818,-0.379479,
        1.361357,2.036004,-0.757062,
        1.361357,2.040501,-1.411488,
        1.361357,2.042874,-1.701367,
        1.361357,2.043060,-2.653746,
        1.361357,2.043729,-1.557938,
        1.361357,2.048538,0.806623,
        1.361357,2.049006,1.199930,
        1.361357,2.049502,-0.137989,
        1.361357,2.052181,2.326959,
        1.361357,2.052794,-0.008526,
        1.361357,2.057608,1.951726,
        1.361357,2.065568,2.567730,
        1.361357,2.070371,2.879808,
        1.361357,2.071256,-3.018961,
        1.361357,2.079746,-0.897699,
        1.361357,2.082784,1.042403,
        1.361357,2.083752,-0.511086,
        1.361357,2.084401,1.810502,
        1.361357,2.085731,0.671081,
        1.361357,2.097496,1.373172,
        1.361357,2.099185,1.667006,
        1.361357,2.100482,-2.301084,
        1.361357,2.103118,2.194745,
        1.361357,2.103223,1.521685,
        1.361357,2.103695,-2.777707,
        1.361357,2.107523,3.129581,
        1.361357,2.108277,-2.095907,
        1.361357,2.108891,-2.532166,
        1.361357,2.111003,-0.268632,
        1.361357,2.112276,-1.027246,
        1.361357,2.114380,0.531457,
        1.361357,2.119166,0.100552,
        1.361357,2.119564,2.754818,
        1.361357,2.122474,-0.648314,
        1.361357,2.125183,2.445598,
        1.361357,2.129501,0.238290,
        1.361357,2.131178,0.386663,
        1.361357,2.131223,-1.174311,
        1.361357,2.134269,-1.948042,
        1.361357,2.134684,0.907065,
        1.361357,2.137615,2.988806,
        1.361357,2.145635,-1.327708,
        1.361357,2.146467,2.055081,
        1.361357,2.146582,-1.793788,
        1.361357,2.149210,-1.482321,
        1.361357,2.152901,-1.636788,
        1.361357,2.156337,1.151558,
        1.361357,2.156474,-2.912065,
        1.361357,2.157194,-0.792563,
        1.361357,2.166724,-0.401638,
        1.361357,2.166980,-0.134879,
        1.361357,2.173644,-2.403429,
        1.361357,2.175711,-2.655523,
        1.361357,2.177132,2.600419,
        1.361357,2.178879,0.767722,
        1.361357,2.180752,1.908485,
        1.361357,2.181338,2.313721,
        1.361357,2.185580,1.287673,
        1.361357,2.195400,-2.217546,
        1.361357,2.200418,-3.056442,
        1.361357,2.201829,-0.001735,
        1.361357,2.203804,1.756176,
        1.361357,2.211602,1.443930,
        1.361357,2.212686,-0.544809,
        1.361357,2.213937,0.621892,
        1.361357,2.214075,1.600691,
        1.361357,2.216256,1.020029,
        1.361357,2.219060,2.886631,
        1.361357,2.223375,-0.906185,
        1.361357,2.226002,-1.060248,
        1.361357,2.226413,-2.064034,
        1.361357,2.230554,2.171334,
        1.361357,2.232413,2.735777,
        1.361357,2.233709,3.070830,
        1.361357,2.234383,-2.792065,
        1.361357,2.237183,0.468799,
        1.361357,2.237695,-0.271388,
        1.361357,2.239419,0.151693,
        1.361357,2.243347,-2.521049,
        1.361357,2.244035,-1.229456,
        1.361357,2.244355,0.309899,
        1.361357,2.245175,2.460236,
        1.361357,2.245628,-1.899216,
        1.361357,2.245973,-0.702234,
        1.361357,2.255601,-1.728131,
        1.361357,2.256702,-1.401367,
        1.361357,2.261466,-1.557500,
        1.361357,2.267390,0.878499,
        1.361357,2.272099,2.018594,
        1.361357,2.278618,1.170386,
        1.361357,2.283304,-2.345584,
        1.361357,2.285175,-2.943090,
        1.361357,2.293000,-0.124451,
        1.361357,2.294058,-0.421070,
        1.361357,2.303283,1.855938,
        1.361357,2.304968,2.608978,
        1.361357,2.305347,2.311346,
        1.361357,2.307188,-2.656134,
        1.361357,2.307301,1.336200,
        1.361357,2.309655,0.726020,
        1.361357,2.317634,-2.185225,
        1.361357,2.321510,1.685664,
        1.361357,2.324376,-3.111006,
        1.361357,2.324731,1.511186,
        1.361357,2.329587,0.035498,
        1.361357,2.330158,2.977264,
        1.361357,2.335096,-0.588468,
        1.361357,2.336046,-1.115799,
        1.361357,2.338060,2.803613,
        1.361357,2.338083,-0.943400,
        1.361357,2.340210,0.562145,
        1.361357,2.341931,1.023926,
        1.361357,2.343265,-0.779061,
        1.361357,2.343434,-2.013692,
        1.361357,2.353029,-1.310529,
        1.361357,2.354868,0.211543,
        1.361357,2.355120,0.388593,
        1.361357,2.355165,-1.833173,
        1.361357,2.355462,2.151441,
        1.361357,2.362948,-2.487477,
        1.361357,2.363101,-2.810456,
        1.361357,2.363394,-0.272918,
        1.361357,2.364509,-1.644533,
        1.361357,2.370465,2.466824,
        1.361357,2.386110,-1.481351,
        1.361357,2.394960,0.862133,
        1.361357,2.395716,1.977184,
        1.361357,2.396339,1.200900,
        1.361357,2.407941,-2.316753,
        1.361357,2.409450,-2.988432,
        1.361357,2.417202,-0.103363,
        1.361357,2.417386,-0.443097,
        1.361357,2.417950,2.666662,
        1.361357,2.423190,1.789783,
        1.361357,2.423692,1.392696,
        1.361357,2.431122,3.086379,
        1.361357,2.431406,-2.647820,
        1.361357,2.431584,2.307631,
        1.361357,2.434090,1.592818,
        1.361357,2.435993,0.683844,
        1.361357,2.440898,-2.133300,
        1.361357,2.442222,-1.195837,
        1.361357,2.443112,2.883761,
        1.361357,2.449191,-0.633651,
        1.361357,2.450821,0.084741,
        1.361357,2.453499,-1.008861,
        1.361357,2.460010,-1.938011,
        1.361357,2.460199,-0.822008,
        1.361357,2.460902,0.490079,
        1.361357,2.461774,-1.734913,
        1.361357,2.463986,1.034407,
        1.361357,2.467647,0.287667,
        1.361357,2.479849,2.126925,
        1.361357,2.485907,-2.835471,
        1.361357,2.485945,-1.374853,
        1.361357,2.487747,2.505912,
        1.361357,2.487957,-0.274279,
        1.361357,2.489741,-2.468428,
        1.361357,2.512538,1.239900,
        1.361357,2.517528,1.923289,
        1.361357,2.518454,-1.565852,
        1.361357,2.519348,0.843503,
        1.361357,2.522766,-3.059231,
        1.361357,2.528686,2.737228,
        1.361357,2.534798,-2.266671,
        1.361357,2.536533,1.466274,
        1.361357,2.538504,1.699812,
        1.361357,2.539516,-0.476942,
        1.361357,2.539532,-0.074548,
        1.361357,2.543197,2.990252,
        1.361357,2.554022,2.320619,
        1.361357,2.556919,-2.655005,
        1.361357,2.558146,0.627191,
        1.361357,2.558223,-1.164189,
        1.361357,2.563452,-2.044028,
        1.361357,2.568462,0.150655,
        1.361357,2.569130,-0.700837,
        1.361357,2.570050,-1.804492,
        1.361357,2.572710,-0.934665,
        1.361357,2.575994,0.391718,
        1.361357,2.584774,1.046703,
        1.361357,2.601432,2.102535,
        1.361357,2.603559,2.555114,
        1.361357,2.604150,-2.885208,
        1.361357,2.608630,-1.378011,
        1.361357,2.611256,-0.279110,
        1.361357,2.617344,-2.439480,
        1.361357,2.627402,1.293552,
        1.361357,2.635396,1.848850,
        1.361357,2.635686,2.834950,
        1.361357,2.637087,3.132730,
        1.361357,2.642029,0.813274,
        1.361357,2.645260,1.570203,
        1.361357,2.646182,-1.614036,
        1.361357,2.660113,-2.186513,
        1.361357,2.660483,-0.034608,
        1.361357,2.661713,-0.529511,
        1.361357,2.672720,-1.110422,
        1.361357,2.676626,2.335353,
        1.361357,2.676638,0.540364,
        1.361357,2.676780,-1.901158,
        1.361357,2.681411,-2.667965,
        1.361357,2.682152,0.245751,
        1.361357,2.682859,-0.817071,
        1.361357,2.705107,1.062142,
        1.361357,2.717705,2.625995,
        1.361357,2.721001,2.060097,
        1.361357,2.722319,-2.954069,
        1.361357,2.729840,-1.368237,
        1.361357,2.734349,-0.289298,
        1.361357,2.738409,2.975671,
        1.361357,2.740366,1.374341,
        1.361357,2.744633,-2.391311,
        1.361357,2.747398,1.727166,
        1.361357,2.763799,0.760372,
        1.361357,2.768238,-1.679243,
        1.361357,2.779245,-2.046268,
        1.361357,2.780208,-0.626453,
        1.361357,2.780471,0.027732,
        1.361357,2.784696,-1.013128,
        1.361357,2.789817,0.395431,
        1.361357,2.801154,2.353143,
        1.361357,2.807091,-2.687477,
        1.361357,2.825762,1.085035,
        1.361357,2.830209,2.734943,
        1.361357,2.833517,-3.085508,
        1.361357,2.838301,1.980061,
        1.361357,2.849871,1.512012,
        1.361357,2.850089,-1.343899,
        1.361357,2.856905,-0.318441,
        1.361357,2.870445,-2.291001,
        1.361357,2.885017,0.655562,
        1.361357,2.886428,-1.793217,
        1.361357,2.890337,-0.825633,
        1.361357,2.898932,0.139973,
        1.361357,2.931927,-2.728768,
        1.361357,2.932096,2.377790,
        1.361357,2.937409,2.947245,
        1.361357,2.948048,1.133127,
        1.361357,2.950896,1.797889,
        1.361357,2.970019,-1.282382,
        1.361357,2.978333,-0.412090,
        1.361357,2.996437,-2.089121,
        1.361357,3.011274,0.434110,
        1.361357,3.050410,3.138492,
        1.361357,3.064492,1.647259,
        1.361357,3.088868,-0.989501,
        1.466077,0.053884,-2.408394,
        1.466077,0.077645,1.453433,
        1.466077,0.088408,-0.311160,
        1.466077,0.142314,2.590387,
        1.466077,0.147206,-1.385164,
        1.466077,0.163506,0.573203,
        1.466077,0.165485,-2.875396,
        1.466077,0.182564,-2.125821,
        1.466077,0.190569,1.929141,
        1.466077,0.201102,1.249082,
        1.466077,0.201397,-0.716646,
        1.466077,0.210568,-0.069559,
        1.466077,0.251672,2.848900,
        1.466077,0.266262,-1.678068,
        1.466077,0.269108,-1.215563,
        1.466077,0.273777,-2.506756,
        1.466077,0.279854,0.814581,
        1.466077,0.281664,-2.969147,
        1.466077,0.283625,2.405603,
        1.466077,0.284058,0.357350,
        1.466077,0.294107,1.593880,
        1.466077,0.302491,-0.400410,
        1.466077,0.308059,2.021491,
        1.466077,0.312219,-2.072057,
        1.466077,0.321905,-0.829121,
        1.466077,0.329984,1.201621,
        1.466077,0.339714,-0.013712,
        1.466077,0.368784,2.974974,
        1.466077,0.376428,-1.495532,
        1.466077,0.378103,2.646436,
        1.466077,0.385088,0.590355,
        1.466077,0.386248,-2.690943,
        1.466077,0.389851,-2.340031,
        1.466077,0.391502,-1.171374,
        1.466077,0.398249,-1.812694,
        1.466077,0.400542,-2.995729,
        1.466077,0.404273,0.918163,
        1.466077,0.405617,1.773533,
        1.466077,0.409009,-0.570243,
        1.466077,0.410700,0.270433,
        1.466077,0.415893,1.464890,
        1.466077,0.420680,2.088164,
        1.466077,0.424766,-0.258729,
        1.466077,0.426507,2.367872,
        1.466077,0.441292,-0.890727,
        1.466077,0.451839,-2.078193,
        1.466077,0.459270,1.191562,
        1.466077,0.470622,0.014002,
        1.466077,0.481292,-2.529098,
        1.466077,0.482460,2.849843,
        1.466077,0.492205,-1.397963,
        1.466077,0.493314,3.117264,
        1.466077,0.497870,0.727828,
        1.466077,0.499783,-1.654379,
        1.466077,0.502651,0.463333,
        1.466077,0.505463,2.594557,
        1.466077,0.506104,-2.874910,
        1.466077,0.513579,-1.147354,
        1.466077,0.518569,1.644237,
        1.466077,0.519001,1.900488,
        1.466077,0.519277,-0.673809,
        1.466077,0.524769,-0.421955,
        1.466077,0.525564,-2.288838,
        1.466077,0.532541,0.978737,
        1.466077,0.534610,2.154808,
        1.466077,0.535020,-1.891846,
        1.466077,0.540208,1.400497,
        1.466077,0.542006,0.228337,
        1.466077,0.550762,-0.180771,
        1.466077,0.560531,2.378474,
        1.466077,0.561629,-0.924576,
        1.466077,0.567465,-2.679392,
        1.466077,0.588527,1.181755,
        1.466077,0.591643,-2.098364,
        1.466077,0.594963,2.997575,
        1.466077,0.599643,-3.045235,
        1.466077,0.599962,2.773556,
        1.466077,0.602620,0.029270,
        1.466077,0.605711,0.598454,
        1.466077,0.607708,-1.541026,
        1.466077,0.609360,-2.464379,
        1.466077,0.610738,-1.326786,
        1.466077,0.616968,0.820489,
        1.466077,0.626379,1.773610,
        1.466077,0.628626,-1.751180,
        1.466077,0.630671,0.391949,
        1.466077,0.632622,-0.737642,
        1.466077,0.632708,2.566722,
        1.466077,0.632797,1.989858,
        1.466077,0.634455,-0.534790,
        1.466077,0.636205,1.561170,
        1.466077,0.637681,-1.123423,
        1.466077,0.641441,-2.842109,
        1.466077,0.644807,-0.327139,
        1.466077,0.652440,2.197129,
        1.466077,0.660897,-2.278593,
        1.466077,0.664008,1.360091,
        1.466077,0.670850,-1.945257,
        1.466077,0.671396,1.004859,
        1.466077,0.673985,0.203855,
        1.466077,0.680326,-0.133370,
        1.466077,0.685083,-0.942363,
        1.466077,0.688204,-2.632204,
        1.466077,0.689777,2.385011,
        1.466077,0.700065,3.115430,
        1.466077,0.704080,2.911636,
        1.466077,0.712883,1.174216,
        1.466077,0.715701,0.699906,
        1.466077,0.718851,-1.442215,
        1.466077,0.721019,2.721536,
        1.466077,0.725300,-2.975134,
        1.466077,0.727136,-1.633288,
        1.466077,0.727399,-2.120567,
        1.466077,0.730751,0.522488,
        1.466077,0.732892,-1.254330,
        1.466077,0.734333,-2.446184,
        1.466077,0.735125,0.037270,
        1.466077,0.737893,1.866463,
        1.466077,0.740650,1.681075,
        1.466077,0.745971,-0.781286,
        1.466077,0.749432,-0.442559,
        1.466077,0.750102,2.050636,
        1.466077,0.751046,0.863821,
        1.466077,0.757018,-0.608833,
        1.466077,0.757046,-1.816019,
        1.466077,0.757389,1.501184,
        1.466077,0.759800,2.546181,
        1.466077,0.759859,0.351923,
        1.466077,0.761141,-2.781102,
        1.466077,0.769035,-1.078574,
        1.466077,0.770166,-0.266604,
        1.466077,0.772218,2.227380,
        1.466077,0.788382,1.331051,
        1.466077,0.790269,-2.282409,
        1.466077,0.803410,-1.984567,
        1.466077,0.803813,0.191520,
        1.466077,0.806832,3.022155,
        1.466077,0.807476,-2.597675,
        1.466077,0.812512,-0.107335,
        1.466077,0.814161,1.010949,
        1.466077,0.817732,-3.084486,
        1.466077,0.817846,2.390688,
        1.466077,0.819323,2.845960,
        1.466077,0.819430,-0.915017,
        1.466077,0.829198,-1.532491,
        1.466077,0.829531,0.665695,
        1.466077,0.829816,1.167147,
        1.466077,0.833613,-1.357492,
        1.466077,0.842945,-2.907318,
        1.466077,0.844883,2.679708,
        1.466077,0.848189,-1.705003,
        1.466077,0.849888,1.767901,
        1.466077,0.851126,0.493370,
        1.466077,0.852911,1.933394,
        1.466077,0.857516,-2.432863,
        1.466077,0.858261,-1.191729,
        1.466077,0.858279,-2.138895,
        1.466077,0.859723,-0.734928,
        1.466077,0.859754,1.603292,
        1.466077,0.861655,-0.394821,
        1.466077,0.864300,0.043549,
        1.466077,0.866940,0.851229,
        1.466077,0.869748,2.095969,
        1.466077,0.877352,-0.574668,
        1.466077,0.879254,-2.733338,
        1.466077,0.882554,1.443757,
        1.466077,0.884060,0.334199,
        1.466077,0.885626,-1.866061,
        1.466077,0.889834,2.529957,
        1.466077,0.893140,2.251698,
        1.466077,0.897463,-0.240822,
        1.466077,0.898758,-1.031948,
        1.466077,0.913918,1.287618,
        1.466077,0.916555,-2.287173,
        1.466077,0.918305,3.107845,
        1.466077,0.919522,2.948277,
        1.466077,0.927138,-2.569686,
        1.466077,0.929031,0.185383,
        1.466077,0.933766,-3.015917,
        1.466077,0.934873,-2.010822,
        1.466077,0.935636,-1.445855,
        1.466077,0.936702,-0.866461,
        1.466077,0.937395,0.729108,
        1.466077,0.937910,2.787575,
        1.466077,0.939144,-0.092962,
        1.466077,0.941718,1.125085,
        1.466077,0.942873,-1.607680,
        1.466077,0.943701,0.980328,
        1.466077,0.947774,2.395256,
        1.466077,0.952433,0.584347,
        1.466077,0.954456,-1.287293,
        1.466077,0.958731,-2.852552,
        1.466077,0.961838,1.839284,
        1.466077,0.961898,-0.469425,
        1.466077,0.965426,1.686615,
        1.466077,0.970108,1.990771,
        1.466077,0.972820,-1.762780,
        1.466077,0.973500,-0.711881,
        1.466077,0.975253,2.641082,
        1.466077,0.977824,0.441928,
        1.466077,0.981485,-2.424766,
        1.466077,0.981877,1.537100,
        1.466077,0.982609,-1.136920,
        1.466077,0.984265,-2.156073,
        1.466077,0.989150,0.045493,
        1.466077,0.993052,2.137478,
        1.466077,0.997766,-2.693871,
        1.466077,0.999847,-0.334738,
        1.466077,1.001920,0.848054,
        1.466077,1.006146,1.389734,
        1.466077,1.012180,0.304066,
        1.466077,1.016938,2.270674,
        1.466077,1.017371,-0.979790,
        1.466077,1.019495,-1.894487,
        1.466077,1.023796,-0.583691,
        1.466077,1.026525,2.506945,
        1.466077,1.027004,-0.203653,
        1.466077,1.028027,3.036573,
        1.466077,1.031435,1.241021,
        1.466077,1.032316,2.884221,
        1.466077,1.035863,-3.110475,
        1.466077,1.042085,-2.292950,
        1.466077,1.043272,-1.522986,
        1.466077,1.046265,-2.958847,
        1.466077,1.049272,-1.376344,
        1.466077,1.050039,-2.552086,
        1.466077,1.053682,0.689863,
        1.466077,1.055458,0.170394,
        1.466077,1.056622,-0.831866,
        1.466077,1.057515,1.097998,
        1.466077,1.058706,-2.036438,
        1.466077,1.061389,-1.672248,
        1.466077,1.063758,2.743093,
        1.466077,1.070157,-0.073422,
        1.466077,1.072154,1.761506,
        1.466077,1.074029,1.909193,
        1.466077,1.074204,0.545276,
        1.466077,1.074714,-2.805147,
        1.466077,1.076262,0.966104,
        1.466077,1.077642,-1.227073,
        1.466077,1.084151,1.618525,
        1.466077,1.085085,2.377673,
        1.466077,1.086822,-0.441112,
        1.466077,1.087142,2.058815,
        1.466077,1.098383,-0.696430,
        1.466077,1.099138,-1.084033,
        1.466077,1.103208,1.477773,
        1.466077,1.103783,0.408254,
        1.466077,1.106275,-2.420800,
        1.466077,1.108079,2.612943,
        1.466077,1.109345,-2.172718,
        1.466077,1.115957,-1.787597,
        1.466077,1.118930,0.045983,
        1.466077,1.119777,0.834702,
        1.466077,1.121943,-2.669042,
        1.466077,1.124930,2.219917,
        1.466077,1.125217,1.336294,
        1.466077,1.130958,2.972931,
        1.466077,1.131400,-0.306629,
        1.466077,1.137135,-0.936514,
        1.466077,1.139125,-1.924528,
        1.466077,1.139877,0.276382,
        1.466077,1.142003,-3.058022,
        1.466077,1.147363,-0.558448,
        1.466077,1.150348,-0.176418,
        1.466077,1.151198,1.197016,
        1.466077,1.153930,-1.597104,
        1.466077,1.155389,2.837148,
        1.466077,1.155458,-1.460623,
        1.466077,1.157331,-2.906362,
        1.466077,1.158594,2.487618,
        1.466077,1.162103,-1.322553,
        1.466077,1.165407,3.092861,
        1.466077,1.168025,0.717075,
        1.466077,1.168406,-2.301519,
        1.466077,1.174782,-2.538755,
        1.466077,1.177998,1.061256,
        1.466077,1.178898,1.827305,
        1.466077,1.181331,-2.060330,
        1.466077,1.182540,-0.798952,
        1.466077,1.184569,0.147517,
        1.466077,1.184734,1.968919,
        1.466077,1.184804,0.593386,
        1.466077,1.188278,1.690808,
        1.466077,1.194360,2.110440,
        1.466077,1.194776,2.710455,
        1.466077,1.197745,-1.173705,
        1.466077,1.199898,-2.777257,
        1.466077,1.201164,2.355931,
        1.466077,1.203760,1.556404,
        1.466077,1.205845,-0.415993,
        1.466077,1.206585,0.933648,
        1.466077,1.206950,-0.037248,
        1.466077,1.209341,0.471162,
        1.466077,1.213483,-1.039417,
        1.466077,1.218146,-0.672643,
        1.466077,1.221608,1.421669,
        1.466077,1.226043,-1.693533,
        1.466077,1.232870,-1.822773,
        1.466077,1.234375,-2.421194,
        1.466077,1.234531,-2.190148,
        1.466077,1.239414,0.348959,
        1.466077,1.240183,2.587255,
        1.466077,1.242312,2.941234,
        1.466077,1.244075,1.286644,
        1.466077,1.245961,-2.997461,
        1.466077,1.246495,-2.646872,
        1.466077,1.247736,2.242810,
        1.466077,1.253604,0.812222,
        1.466077,1.257816,-0.894317,
        1.466077,1.258247,-1.953634,
        1.466077,1.258993,-0.284099,
        1.466077,1.262380,-0.157609,
        1.466077,1.262791,-1.549229,
        1.466077,1.263772,-0.533073,
        1.466077,1.266731,-1.412461,
        1.466077,1.268591,-3.122146,
        1.466077,1.269614,-1.275246,
        1.466077,1.270745,1.152533,
        1.466077,1.273606,0.224545,
        1.466077,1.279716,2.812101,
        1.466077,1.283786,2.462124,
        1.466077,1.284798,0.078209,
        1.466077,1.285402,-2.882049,
        1.466077,1.286936,1.886053,
        1.466077,1.289408,0.683122,
        1.466077,1.294736,1.756184,
        1.466077,1.296978,-2.311511,
        1.466077,1.298546,2.020601,
        1.466077,1.300924,1.022478,
        1.466077,1.303305,3.043152,
        1.466077,1.304210,-2.084225,
        1.466077,1.305241,-2.530249,
        1.466077,1.307513,1.628003,
        1.466077,1.310722,0.558267,
        1.466077,1.314142,2.139329,
        1.466077,1.316087,-0.777971,
        1.466077,1.317907,-1.121023,
        1.466077,1.321019,-2.749776,
        1.466077,1.321609,2.687717,
        1.466077,1.322568,1.499415,
        1.466077,1.322972,-0.395699,
        1.466077,1.325286,-0.994192,
        1.466077,1.329146,-0.648513,
        1.466077,1.329795,-0.042570,
        1.466077,1.331651,2.344497,
        1.466077,1.336522,0.435098,
        1.466077,1.339410,0.900087,
        1.466077,1.339550,1.370353,
        1.466077,1.340441,-1.731580,
        1.466077,1.351855,-1.855234,
        1.466077,1.360161,-3.018404,
        1.466077,1.361031,-1.613971,
        1.466077,1.361196,2.914780,
        1.466077,1.361998,-2.207254,
        1.466077,1.362302,1.239384,
        1.466077,1.365228,2.563808,
        1.466077,1.365452,0.310522,
        1.466077,1.367028,-2.420041,
        1.466077,1.373422,-1.484281,
        1.466077,1.374245,-1.226417,
        1.466077,1.375020,-0.170393,
        1.466077,1.376560,-1.355481,
        1.466077,1.377070,-1.980984,
        1.466077,1.377578,-2.636339,
        1.466077,1.378096,0.774626,
        1.466077,1.379086,-0.508192,
        1.466077,1.382562,0.177141,
        1.466077,1.388849,2.233847,
        1.466077,1.389268,-0.886349,
        1.466077,1.391324,1.109396,
        1.466077,1.391699,-0.288725,
        1.466077,1.392869,3.136903,
        1.466077,1.393612,1.952016,
        1.466077,1.395545,-2.850807,
        1.466077,1.401926,1.822930,
        1.466077,1.402149,2.788898,
        1.466077,1.409552,0.647980,
        1.466077,1.410840,0.056219,
        1.466077,1.411944,2.444377,
        1.466077,1.413639,1.697982,
        1.466077,1.426479,1.573427,
        1.466077,1.426725,2.102242,
        1.466077,1.427868,0.984637,
        1.466077,1.429580,-2.107232,
        1.466077,1.431375,-2.313269,
        1.466077,1.431907,3.015410,
        1.466077,1.432325,-1.029054,
        1.466077,1.433341,0.524268,
        1.466077,1.435181,-0.739771,
        1.466077,1.437208,-2.529085,
        1.466077,1.439041,-0.617966,
        1.466077,1.441473,1.448544,
        1.466077,1.444527,2.664744,
        1.466077,1.450513,-0.400625,
        1.466077,1.451907,-0.066823,
        1.466077,1.452754,-2.738631,
        1.466077,1.454061,-2.950416,
        1.466077,1.456736,1.323154,
        1.466077,1.458370,0.400433,
        1.466077,1.459314,-1.765003,
        1.466077,1.459902,-1.145835,
        1.466077,1.464952,2.329245,
        1.466077,1.466050,0.862104,
        1.466077,1.471330,-1.886160,
        1.466077,1.475119,-1.648255,
        1.466077,1.480878,1.194576,
        1.466077,1.480942,0.273235,
        1.466077,1.481086,2.890269,
        1.466077,1.484118,-1.529156,
        1.466077,1.484741,-1.279219,
        1.466077,1.487681,-1.406069,
        1.466077,1.487886,-3.066037,
        1.466077,1.490555,2.543284,
        1.466077,1.492424,-0.190932,
        1.466077,1.495002,-0.839810,
        1.466077,1.495297,2.007188,
        1.466077,1.496414,-2.009267,
        1.466077,1.498261,-2.425270,
        1.466077,1.498907,-2.207327,
        1.466077,1.500204,0.737638,
        1.466077,1.502469,0.151156,
        1.466077,1.507448,-0.514782,
        1.466077,1.509287,2.204647,
        1.466077,1.510518,1.884920,
        1.466077,1.511948,-2.632360,
        1.466077,1.515235,1.068565,
        1.466077,1.522237,3.102018,
        1.466077,1.522373,1.764991,
        1.466077,1.522416,-0.954956,
        1.466077,1.523102,2.765920,
        1.466077,1.524811,-2.844718,
        1.466077,1.527261,-0.304916,
        1.466077,1.529788,0.613208,
        1.466077,1.533648,1.644808,
        1.466077,1.535018,0.028869,
        1.466077,1.541082,2.424608,
        1.466077,1.546010,1.523462,
        1.466077,1.551135,-0.642994,
        1.466077,1.553895,0.489607,
        1.466077,1.555041,0.947813,
        1.466077,1.559671,-1.074934,
        1.466077,1.560345,2.985899,
        1.466077,1.560487,-2.323400,
        1.466077,1.561127,1.401111,
        1.466077,1.567193,2.643172,
        1.466077,1.568892,2.098362,
        1.466077,1.571702,-2.098435,
        1.466077,1.572701,-2.528730,
        1.466077,1.573572,1.278593,
        1.466077,1.574849,-0.091505,
        1.466077,1.576334,0.365236,
        1.466077,1.577029,-2.962888,
        1.466077,1.577688,-1.814941,
        1.466077,1.579542,-0.755917,
        1.466077,1.579624,-1.199759,
        1.466077,1.581515,-0.419414,
        1.466077,1.584581,-2.738702,
        1.466077,1.586948,-1.696296,
        1.466077,1.588901,2.301496,
        1.466077,1.589936,0.825020,
        1.466077,1.590018,-1.941652,
        1.466077,1.596968,-1.328358,
        1.466077,1.597178,-1.575607,
        1.466077,1.600470,-1.453196,
        1.466077,1.601408,0.242409,
        1.466077,1.602259,1.152656,
        1.466077,1.603064,2.865774,
        1.466077,1.612085,1.968802,
        1.466077,1.612720,-3.080498,
        1.466077,1.616061,2.521531,
        1.466077,1.616900,-0.207457,
        1.466077,1.620988,-0.875920,
        1.466077,1.622114,0.701509,
        1.466077,1.622749,0.121426,
        1.466077,1.622993,-2.221629,
        1.466077,1.629346,-0.542669,
        1.466077,1.630007,1.844552,
        1.466077,1.634747,-2.426198,
        1.466077,1.642346,1.722617,
        1.466077,1.643925,1.034629,
        1.466077,1.644848,2.743774,
        1.466077,1.645258,2.188109,
        1.466077,1.645417,-2.634725,
        1.466077,1.647395,3.085437,
        1.466077,1.648538,-2.855044,
        1.466077,1.649932,0.576808,
        1.466077,1.652224,-0.996105,
        1.466077,1.652871,1.600681,
        1.466077,1.657997,-0.325127,
        1.466077,1.661936,0.001037,
        1.466077,1.665187,1.477402,
        1.466077,1.665496,2.397872,
        1.466077,1.673267,-0.665453,
        1.466077,1.673557,0.452526,
        1.466077,1.679277,-1.122421,
        1.466077,1.679696,0.912482,
        1.466077,1.682420,1.353922,
        1.466077,1.685031,-2.009038,
        1.466077,1.685196,2.967046,
        1.466077,1.690107,2.620910,
        1.466077,1.690453,-2.123076,
        1.466077,1.691693,-1.871572,
        1.466077,1.691817,2.061688,
        1.466077,1.692644,1.237670,
        1.466077,1.696386,0.328123,
        1.466077,1.697162,-1.249026,
        1.466077,1.697488,-2.322785,
        1.466077,1.698726,-2.973576,
        1.466077,1.699172,-1.749717,
        1.466077,1.704478,-0.106295,
        1.466077,1.707244,-0.444770,
        1.466077,1.707500,-2.531497,
        1.466077,1.707714,-0.787321,
        1.466077,1.708430,-1.627552,
        1.466077,1.710583,-1.376890,
        1.466077,1.712415,-2.747921,
        1.466077,1.713125,-1.503073,
        1.466077,1.713452,0.789419,
        1.466077,1.720612,2.279888,
        1.466077,1.724852,0.202831,
        1.466077,1.724868,2.846246,
        1.466077,1.728940,1.931535,
        1.466077,1.732838,1.127584,
        1.466077,1.735719,-3.092485,
        1.466077,1.739676,2.496017,
        1.466077,1.739971,-0.228876,
        1.466077,1.743328,0.663347,
        1.466077,1.744927,-0.911892,
        1.466077,1.745952,0.084733,
        1.466077,1.748077,1.804354,
        1.466077,1.752959,-0.567548,
        1.466077,1.760169,1.678132,
        1.466077,1.763887,-2.218450,
        1.466077,1.766737,2.722752,
        1.466077,1.767999,1.004737,
        1.466077,1.769149,1.550298,
        1.466077,1.770119,-2.427175,
        1.466077,1.770460,2.154208,
        1.466077,1.770514,3.070982,
        1.466077,1.770735,0.537245,
        1.466077,1.772673,-2.864760,
        1.466077,1.773527,-1.039582,
        1.466077,1.776419,-2.642421,
        1.466077,1.780769,1.416330,
        1.466077,1.786634,-0.346247,
        1.466077,1.788829,-1.947553,
        1.466077,1.793737,0.410270,
        1.466077,1.794553,2.374618,
        1.466077,1.794699,-0.694061,
        1.466077,1.797469,-1.169105,
        1.466077,1.803705,0.881076,
        1.466077,1.804124,1.264249,
        1.466077,1.807825,-2.090455,
        1.466077,1.808822,-1.817375,
        1.466077,1.809020,2.949741,
        1.466077,1.812466,2.597595,
        1.466077,1.813159,2.025395,
        1.466077,1.813816,-1.298435,
        1.466077,1.815016,0.278951,
        1.466077,1.816510,-0.008631,
        1.466077,1.817772,-1.689625,
        1.466077,1.821360,-2.984352,
        1.466077,1.822312,-0.127211,
        1.466077,1.823386,-1.429807,
        1.466077,1.823834,-1.560483,
        1.466077,1.832060,-0.821627,
        1.466077,1.832690,-0.468440,
        1.466077,1.833945,-2.320278,
        1.466077,1.834787,0.751968,
        1.466077,1.840579,-2.536445,
        1.466077,1.841033,-2.756932,
        1.466077,1.846350,1.893983,
        1.466077,1.846644,1.113766,
        1.466077,1.846833,2.826543,
        1.466077,1.847325,2.249051,
        1.466077,1.854119,0.134161,
        1.466077,1.858863,-3.106022,
        1.466077,1.865046,1.763384,
        1.466077,1.865619,-0.952036,
        1.466077,1.865912,0.624453,
        1.466077,1.866694,2.473686,
        1.466077,1.867524,-0.240083,
        1.466077,1.875561,1.632831,
        1.466077,1.876049,-0.594378,
        1.466077,1.880509,1.500204,
        1.466077,1.883504,1.360457,
        1.466077,1.886971,-2.196665,
        1.466077,1.888979,0.982922,
        1.466077,1.889815,2.701570,
        1.466077,1.892144,-2.005340,
        1.466077,1.892348,0.496522,
        1.466077,1.893281,-1.083976,
        1.466077,1.893828,3.053453,
        1.466077,1.895117,2.121107,
        1.466077,1.898741,-2.875735,
        1.466077,1.904429,-2.428029,
        1.466077,1.908408,-2.648903,
        1.466077,1.911828,-0.364489,
        1.466077,1.912224,0.366565,
        1.466077,1.914854,1.214468,
        1.466077,1.915467,-1.883143,
        1.466077,1.915763,-0.724679,
        1.466077,1.915790,-1.216366,
        1.466077,1.922256,2.347806,
        1.466077,1.922616,0.844633,
        1.466077,1.925185,0.231170,
        1.466077,1.927474,-1.754616,
        1.466077,1.928342,-1.351174,
        1.466077,1.933328,-1.623328,
        1.466077,1.933878,0.034150,
        1.466077,1.934585,-1.487979,
        1.466077,1.934802,2.927634,
        1.466077,1.934862,1.989776,
        1.466077,1.936327,-0.099350,
        1.466077,1.938088,2.578507,
        1.466077,1.945677,-2.999805,
        1.466077,1.954572,-0.859208,
        1.466077,1.957868,-0.490784,
        1.466077,1.958989,0.712772,
        1.466077,1.960212,-2.307561,
        1.466077,1.964774,1.855507,
        1.466077,1.964971,-2.100533,
        1.466077,1.969112,1.088273,
        1.466077,1.971438,-2.765415,
        1.466077,1.972312,2.805134,
        1.466077,1.974950,2.220596,
        1.466077,1.975135,-2.539127,
        1.466077,1.982107,1.720829,
        1.466077,1.983195,-3.127984,
        1.466077,1.983318,-0.247182,
        1.466077,1.986720,-0.993220,
        1.466077,1.990381,1.586005,
        1.466077,1.990470,0.580283,
        1.466077,1.990523,1.450326,
        1.466077,1.993909,1.312897,
        1.466077,1.994694,2.452813,
        1.466077,1.999226,-0.620717,
        1.466077,2.005145,0.941012,
        1.466077,2.007301,0.142337,
        1.466077,2.014149,0.444138,
        1.466077,2.014338,-1.127392,
        1.466077,2.018131,3.021778,
        1.466077,2.019676,-1.980944,
        1.466077,2.019912,2.686612,
        1.466077,2.020545,2.088672,
        1.466077,2.026074,0.300945,
        1.466077,2.026196,-2.889057,
        1.466077,2.031959,-1.268178,
        1.466077,2.032917,-2.199917,
        1.466077,2.033572,-1.842884,
        1.466077,2.034712,-2.419505,
        1.466077,2.035818,-0.379479,
        1.466077,2.036004,-0.757062,
        1.466077,2.040501,-1.411488,
        1.466077,2.042874,-1.701367,
        1.466077,2.043060,-2.653746,
        1.466077,2.043729,-1.557938,
        1.466077,2.048538,0.806623,
        1.466077,2.049006,1.199930,
        1.466077,2.049502,-0.137989,
        1.466077,2.052181,2.326959,
        1.466077,2.052794,-0.008526,
        1.466077,2.057608,1.951726,
        1.466077,2.065568,2.567730,
        1.466077,2.070371,2.879808,
        1.466077,2.071256,-3.018961,
        1.466077,2.079746,-0.897699,
        1.466077,2.082784,1.042403,
        1.466077,2.083752,-0.511086,
        1.466077,2.084401,1.810502,
        1.466077,2.085731,0.671081,
        1.466077,2.097496,1.373172,
        1.466077,2.099185,1.667006,
        1.466077,2.100482,-2.301084,
        1.466077,2.103118,2.194745,
        1.466077,2.103223,1.521685,
        1.466077,2.103695,-2.777707,
        1.466077,2.107523,3.129581,
        1.466077,2.108277,-2.095907,
        1.466077,2.108891,-2.532166,
        1.466077,2.111003,-0.268632,
        1.466077,2.112276,-1.027246,
        1.466077,2.114380,0.531457,
        1.466077,2.119166,0.100552,
        1.466077,2.119564,2.754818,
        1.466077,2.122474,-0.648314,
        1.466077,2.125183,2.445598,
        1.466077,2.129501,0.238290,
        1.466077,2.131178,0.386663,
        1.466077,2.131223,-1.174311,
        1.466077,2.134269,-1.948042,
        1.466077,2.134684,0.907065,
        1.466077,2.137615,2.988806,
        1.466077,2.145635,-1.327708,
        1.466077,2.146467,2.055081,
        1.466077,2.146582,-1.793788,
        1.466077,2.149210,-1.482321,
        1.466077,2.152901,-1.636788,
        1.466077,2.156337,1.151558,
        1.466077,2.156474,-2.912065,
        1.466077,2.157194,-0.792563,
        1.466077,2.166724,-0.401638,
        1.466077,2.166980,-0.134879,
        1.466077,2.173644,-2.403429,
        1.466077,2.175711,-2.655523,
        1.466077,2.177132,2.600419,
        1.466077,2.178879,0.767722,
        1.466077,2.180752,1.908485,
        1.466077,2.181338,2.313721,
        1.466077,2.185580,1.287673,
        1.466077,2.195400,-2.217546,
        1.466077,2.200418,-3.056442,
        1.466077,2.201829,-0.001735,
        1.466077,2.203804,1.756176,
        1.466077,2.211602,1.443930,
        1.466077,2.212686,-0.544809,
        1.466077,2.213937,0.621892,
        1.466077,2.214075,1.600691,
        1.466077,2.216256,1.020029,
        1.466077,2.219060,2.886631,
        1.466077,2.223375,-0.906185,
        1.466077,2.226002,-1.060248,
        1.466077,2.226413,-2.064034,
        1.466077,2.230554,2.171334,
        1.466077,2.232413,2.735777,
        1.466077,2.233709,3.070830,
        1.466077,2.234383,-2.792065,
        1.466077,2.237183,0.468799,
        1.466077,2.237695,-0.271388,
        1.466077,2.239419,0.151693,
        1.466077,2.243347,-2.521049,
        1.466077,2.244035,-1.229456,
        1.466077,2.244355,0.309899,
        1.466077,2.245175,2.460236,
        1.466077,2.245628,-1.899216,
        1.466077,2.245973,-0.702234,
        1.466077,2.255601,-1.728131,
        1.466077,2.256702,-1.401367,
        1.466077,2.261466,-1.557500,
        1.466077,2.267390,0.878499,
        1.466077,2.272099,2.018594,
        1.466077,2.278618,1.170386,
        1.466077,2.283304,-2.345584,
        1.466077,2.285175,-2.943090,
        1.466077,2.293000,-0.124451,
        1.466077,2.294058,-0.421070,
        1.466077,2.303283,1.855938,
        1.466077,2.304968,2.608978,
        1.466077,2.305347,2.311346,
        1.466077,2.307188,-2.656134,
        1.466077,2.307301,1.336200,
        1.466077,2.309655,0.726020,
        1.466077,2.317634,-2.185225,
        1.466077,2.321510,1.685664,
        1.466077,2.324376,-3.111006,
        1.466077,2.324731,1.511186,
        1.466077,2.329587,0.035498,
        1.466077,2.330158,2.977264,
        1.466077,2.335096,-0.588468,
        1.466077,2.336046,-1.115799,
        1.466077,2.338060,2.803613,
        1.466077,2.338083,-0.943400,
        1.466077,2.340210,0.562145,
        1.466077,2.341931,1.023926,
        1.466077,2.343265,-0.779061,
        1.466077,2.343434,-2.013692,
        1.466077,2.353029,-1.310529,
        1.466077,2.354868,0.211543,
        1.466077,2.355120,0.388593,
        1.466077,2.355165,-1.833173,
        1.466077,2.355462,2.151441,
        1.466077,2.362948,-2.487477,
        1.466077,2.363101,-2.810456,
        1.466077,2.363394,-0.272918,
        1.466077,2.364509,-1.644533,
        1.466077,2.370465,2.466824,
        1.466077,2.386110,-1.481351,
        1.466077,2.394960,0.862133,
        1.466077,2.395716,1.977184,
        1.466077,2.396339,1.200900,
        1.466077,2.407941,-2.316753,
        1.466077,2.409450,-2.988432,
        1.466077,2.417202,-0.103363,
        1.466077,2.417386,-0.443097,
        1.466077,2.417950,2.666662,
        1.466077,2.423190,1.789783,
        1.466077,2.423692,1.392696,
        1.466077,2.431122,3.086379,
        1.466077,2.431406,-2.647820,
        1.466077,2.431584,2.307631,
        1.466077,2.434090,1.592818,
        1.466077,2.435993,0.683844,
        1.466077,2.440898,-2.133300,
        1.466077,2.442222,-1.195837,
        1.466077,2.443112,2.883761,
        1.466077,2.449191,-0.633651,
        1.466077,2.450821,0.084741,
        1.466077,2.453499,-1.008861,
        1.466077,2.460010,-1.938011,
        1.466077,2.460199,-0.822008,
        1.466077,2.460902,0.490079,
        1.466077,2.461774,-1.734913,
        1.466077,2.463986,1.034407,
        1.466077,2.467647,0.287667,
        1.466077,2.479849,2.126925,
        1.466077,2.485907,-2.835471,
        1.466077,2.485945,-1.374853,
        1.466077,2.487747,2.505912,
        1.466077,2.487957,-0.274279,
        1.466077,2.489741,-2.468428,
        1.466077,2.512538,1.239900,
        1.466077,2.517528,1.923289,
        1.466077,2.518454,-1.565852,
        1.466077,2.519348,0.843503,
        1.466077,2.522766,-3.059231,
        1.466077,2.528686,2.737228,
        1.466077,2.534798,-2.266671,
        1.466077,2.536533,1.466274,
        1.466077,2.538504,1.699812,
        1.466077,2.539516,-0.476942,
        1.466077,2.539532,-0.074548,
        1.466077,2.543197,2.990252,
        1.466077,2.554022,2.320619,
        1.466077,2.556919,-2.655005,
        1.466077,2.558146,0.627191,
        1.466077,2.558223,-1.164189,
        1.466077,2.563452,-2.044028,
        1.466077,2.568462,0.150655,
        1.466077,2.569130,-0.700837,
        1.466077,2.570050,-1.804492,
        1.466077,2.572710,-0.934665,
        1.466077,2.575994,0.391718,
        1.466077,2.584774,1.046703,
        1.466077,2.601432,2.102535,
        1.466077,2.603559,2.555114,
        1.466077,2.604150,-2.885208,
        1.466077,2.608630,-1.378011,
        1.466077,2.611256,-0.279110,
        1.466077,2.617344,-2.439480,
        1.466077,2.627402,1.293552,
        1.466077,2.635396,1.848850,
        1.466077,2.635686,2.834950,
        1.466077,2.637087,3.132730,
        1.466077,2.642029,0.813274,
        1.466077,2.645260,1.570203,
        1.466077,2.646182,-1.614036,
        1.466077,2.660113,-2.186513,
        1.466077,2.660483,-0.034608,
        1.466077,2.661713,-0.529511,
        1.466077,2.672720,-1.110422,
        1.466077,2.676626,2.335353,
        1.466077,2.676638,0.540364,
        1.466077,2.676780,-1.901158,
        1.466077,2.681411,-2.667965,
        1.466077,2.682152,0.245751,
        1.466077,2.682859,-0.817071,
        1.466077,2.705107,1.062142,
        1.466077,2.717705,2.625995,
        1.466077,2.721001,2.060097,
        1.466077,2.722319,-2.954069,
        1.466077,2.729840,-1.368237,
        1.466077,2.734349,-0.289298,
        1.466077,2.738409,2.975671,
        1.466077,2.740366,1.374341,
        1.466077,2.744633,-2.391311,
        1.466077,2.747398,1.727166,
        1.466077,2.763799,0.760372,
        1.466077,2.768238,-1.679243,
        1.466077,2.779245,-2.046268,
        1.466077,2.780208,-0.626453,
        1.466077,2.780471,0.027732,
        1.466077,2.784696,-1.013128,
        1.466077,2.789817,0.395431,
        1.466077,2.801154,2.353143,
        1.466077,2.807091,-2.687477,
        1.466077,2.825762,1.085035,
        1.466077,2.830209,2.734943,
        1.466077,2.833517,-3.085508,
        1.466077,2.838301,1.980061,
        1.466077,2.849871,1.512012,
        1.466077,2.850089,-1.343899,
        1.466077,2.856905,-0.318441,
        1.466077,2.870445,-2.291001,
        1.466077,2.885017,0.655562,
        1.466077,2.886428,-1.793217,
        1.466077,2.890337,-0.825633,
        1.466077,2.898932,0.139973,
        1.466077,2.931927,-2.728768,
        1.466077,2.932096,2.377790,
        1.466077,2.937409,2.947245,
        1.466077,2.948048,1.133127,
        1.466077,2.950896,1.797889,
        1.466077,2.970019,-1.282382,
        1.466077,2.978333,-0.412090,
        1.466077,2.996437,-2.089121,
        1.466077,3.011274,0.434110,
        1.466077,3.050410,3.138492,
        1.466077,3.064492,1.647259,
        1.466077,3.088868,-0.989501,
        1.570796,0.053884,-2.408394,
        1.570796,0.077645,1.453433,
        1.570796,0.088408,-0.311160,
        1.570796,0.142314,2.590387,
        1.570796,0.147206,-1.385164,
        1.570796,0.163506,0.573203,
        1.570796,0.165485,-2.875396,
        1.570796,0.182564,-2.125821,
        1.570796,0.190569,1.929141,
        1.570796,0.201102,1.249082,
        1.570796,0.201397,-0.716646,
        1.570796,0.210568,-0.069559,
        1.570796,0.251672,2.848900,
        1.570796,0.266262,-1.678068,
        1.570796,0.269108,-1.215563,
        1.570796,0.273777,-2.506756,
        1.570796,0.279854,0.814581,
        1.570796,0.281664,-2.969147,
        1.570796,0.283625,2.405603,
        1.570796,0.284058,0.357350,
        1.570796,0.294107,1.593880,
        1.570796,0.302491,-0.400410,
        1.570796,0.308059,2.021491,
        1.570796,0.312219,-2.072057,
        1.570796,0.321905,-0.829121,
        1.570796,0.329984,1.201621,
        1.570796,0.339714,-0.013712,
        1.570796,0.368784,2.974974,
        1.570796,0.376428,-1.495532,
        1.570796,0.378103,2.646436,
        1.570796,0.385088,0.590355,
        1.570796,0.386248,-2.690943,
        1.570796,0.389851,-2.340031,
        1.570796,0.391502,-1.171374,
        1.570796,0.398249,-1.812694,
        1.570796,0.400542,-2.995729,
        1.570796,0.404273,0.918163,
        1.570796,0.405617,1.773533,
        1.570796,0.409009,-0.570243,
        1.570796,0.410700,0.270433,
        1.570796,0.415893,1.464890,
        1.570796,0.420680,2.088164,
        1.570796,0.424766,-0.258729,
        1.570796,0.426507,2.367872,
        1.570796,0.441292,-0.890727,
        1.570796,0.451839,-2.078193,
        1.570796,0.459270,1.191562,
        1.570796,0.470622,0.014002,
        1.570796,0.481292,-2.529098,
        1.570796,0.482460,2.849843,
        1.570796,0.492205,-1.397963,
        1.570796,0.493314,3.117264,
        1.570796,0.497870,0.727828,
        1.570796,0.499783,-1.654379,
        1.570796,0.502651,0.463333,
        1.570796,0.505463,2.594557,
        1.570796,0.506104,-2.874910,
        1.570796,0.513579,-1.147354,
        1.570796,0.518569,1.644237,
        1.570796,0.519001,1.900488,
        1.570796,0.519277,-0.673809,
        1.570796,0.524769,-0.421955,
        1.570796,0.525564,-2.288838,
        1.570796,0.532541,0.978737,
        1.570796,0.534610,2.154808,
        1.570796,0.535020,-1.891846,
        1.570796,0.540208,1.400497,
        1.570796,0.542006,0.228337,
        1.570796,0.550762,-0.180771,
        1.570796,0.560531,2.378474,
        1.570796,0.561629,-0.924576,
        1.570796,0.567465,-2.679392,
        1.570796,0.588527,1.181755,
        1.570796,0.591643,-2.098364,
        1.570796,0.594963,2.997575,
        1.570796,0.599643,-3.045235,
        1.570796,0.599962,2.773556,
        1.570796,0.602620,0.029270,
        1.570796,0.605711,0.598454,
        1.570796,0.607708,-1.541026,
        1.570796,0.609360,-2.464379,
        1.570796,0.610738,-1.326786,
        1.570796,0.616968,0.820489,
        1.570796,0.626379,1.773610,
        1.570796,0.628626,-1.751180,
        1.570796,0.630671,0.391949,
        1.570796,0.632622,-0.737642,
        1.570796,0.632708,2.566722,
        1.570796,0.632797,1.989858,
        1.570796,0.634455,-0.534790,
        1.570796,0.636205,1.561170,
        1.570796,0.637681,-1.123423,
        1.570796,0.641441,-2.842109,
        1.570796,0.644807,-0.327139,
        1.570796,0.652440,2.197129,
        1.570796,0.660897,-2.278593,
        1.570796,0.664008,1.360091,
        1.570796,0.670850,-1.945257,
        1.570796,0.671396,1.004859,
        1.570796,0.673985,0.203855,
        1.570796,0.680326,-0.133370,
        1.570796,0.685083,-0.942363,
        1.570796,0.688204,-2.632204,
        1.570796,0.689777,2.385011,
        1.570796,0.700065,3.115430,
        1.570796,0.704080,2.911636,
        1.570796,0.712883,1.174216,
        1.570796,0.715701,0.699906,
        1.570796,0.718851,-1.442215,
        1.570796,0.721019,2.721536,
        1.570796,0.725300,-2.975134,
        1.570796,0.727136,-1.633288,
        1.570796,0.727399,-2.120567,
        1.570796,0.730751,0.522488,
        1.570796,0.732892,-1.254330,
        1.570796,0.734333,-2.446184,
        1.570796,0.735125,0.037270,
        1.570796,0.737893,1.866463,
        1.570796,0.740650,1.681075,
        1.570796,0.745971,-0.781286,
        1.570796,0.749432,-0.442559,
        1.570796,0.750102,2.050636,
        1.570796,0.751046,0.863821,
        1.570796,0.757018,-0.608833,
        1.570796,0.757046,-1.816019,
        1.570796,0.757389,1.501184,
        1.570796,0.759800,2.546181,
        1.570796,0.759859,0.351923,
        1.570796,0.761141,-2.781102,
        1.570796,0.769035,-1.078574,
        1.570796,0.770166,-0.266604,
        1.570796,0.772218,2.227380,
        1.570796,0.788382,1.331051,
        1.570796,0.790269,-2.282409,
        1.570796,0.803410,-1.984567,
        1.570796,0.803813,0.191520,
        1.570796,0.806832,3.022155,
        1.570796,0.807476,-2.597675,
        1.570796,0.812512,-0.107335,
        1.570796,0.814161,1.010949,
        1.570796,0.817732,-3.084486,
        1.570796,0.817846,2.390688,
        1.570796,0.819323,2.845960,
        1.570796,0.819430,-0.915017,
        1.570796,0.829198,-1.532491,
        1.570796,0.829531,0.665695,
        1.570796,0.829816,1.167147,
        1.570796,0.833613,-1.357492,
        1.570796,0.842945,-2.907318,
        1.570796,0.844883,2.679708,
        1.570796,0.848189,-1.705003,
        1.570796,0.849888,1.767901,
        1.570796,0.851126,0.493370,
        1.570796,0.852911,1.933394,
        1.570796,0.857516,-2.432863,
        1.570796,0.858261,-1.191729,
        1.570796,0.858279,-2.138895,
        1.570796,0.859723,-0.734928,
        1.570796,0.859754,1.603292,
        1.570796,0.861655,-0.394821,
        1.570796,0.864300,0.043549,
        1.570796,0.866940,0.851229,
        1.570796,0.869748,2.095969,
        1.570796,0.877352,-0.574668,
        1.570796,0.879254,-2.733338,
        1.570796,0.882554,1.443757,
        1.570796,0.884060,0.334199,
        1.570796,0.885626,-1.866061,
        1.570796,0.889834,2.529957,
        1.570796,0.893140,2.251698,
        1.570796,0.897463,-0.240822,
        1.570796,0.898758,-1.031948,
        1.570796,0.913918,1.287618,
        1.570796,0.916555,-2.287173,
        1.570796,0.918305,3.107845,
        1.570796,0.919522,2.948277,
        1.570796,0.927138,-2.569686,
        1.570796,0.929031,0.185383,
        1.570796,0.933766,-3.015917,
        1.570796,0.934873,-2.010822,
        1.570796,0.935636,-1.445855,
        1.570796,0.936702,-0.866461,
        1.570796,0.937395,0.729108,
        1.570796,0.937910,2.787575,
        1.570796,0.939144,-0.092962,
        1.570796,0.941718,1.125085,
        1.570796,0.942873,-1.607680,
        1.570796,0.943701,0.980328,
        1.570796,0.947774,2.395256,
        1.570796,0.952433,0.584347,
        1.570796,0.954456,-1.287293,
        1.570796,0.958731,-2.852552,
        1.570796,0.961838,1.839284,
        1.570796,0.961898,-0.469425,
        1.570796,0.965426,1.686615,
        1.570796,0.970108,1.990771,
        1.570796,0.972820,-1.762780,
        1.570796,0.973500,-0.711881,
        1.570796,0.975253,2.641082,
        1.570796,0.977824,0.441928,
        1.570796,0.981485,-2.424766,
        1.570796,0.981877,1.537100,
        1.570796,0.982609,-1.136920,
        1.570796,0.984265,-2.156073,
        1.570796,0.989150,0.045493,
        1.570796,0.993052,2.137478,
        1.570796,0.997766,-2.693871,
        1.570796,0.999847,-0.334738,
        1.570796,1.001920,0.848054,
        1.570796,1.006146,1.389734,
        1.570796,1.012180,0.304066,
        1.570796,1.016938,2.270674,
        1.570796,1.017371,-0.979790,
        1.570796,1.019495,-1.894487,
        1.570796,1.023796,-0.583691,
        1.570796,1.026525,2.506945,
        1.570796,1.027004,-0.203653,
        1.570796,1.028027,3.036573,
        1.570796,1.031435,1.241021,
        1.570796,1.032316,2.884221,
        1.570796,1.035863,-3.110475,
        1.570796,1.042085,-2.292950,
        1.570796,1.043272,-1.522986,
        1.570796,1.046265,-2.958847,
        1.570796,1.049272,-1.376344,
        1.570796,1.050039,-2.552086,
        1.570796,1.053682,0.689863,
        1.570796,1.055458,0.170394,
        1.570796,1.056622,-0.831866,
        1.570796,1.057515,1.097998,
        1.570796,1.058706,-2.036438,
        1.570796,1.061389,-1.672248,
        1.570796,1.063758,2.743093,
        1.570796,1.070157,-0.073422,
        1.570796,1.072154,1.761506,
        1.570796,1.074029,1.909193,
        1.570796,1.074204,0.545276,
        1.570796,1.074714,-2.805147,
        1.570796,1.076262,0.966104,
        1.570796,1.077642,-1.227073,
        1.570796,1.084151,1.618525,
        1.570796,1.085085,2.377673,
        1.570796,1.086822,-0.441112,
        1.570796,1.087142,2.058815,
        1.570796,1.098383,-0.696430,
        1.570796,1.099138,-1.084033,
        1.570796,1.103208,1.477773,
        1.570796,1.103783,0.408254,
        1.570796,1.106275,-2.420800,
        1.570796,1.108079,2.612943,
        1.570796,1.109345,-2.172718,
        1.570796,1.115957,-1.787597,
        1.570796,1.118930,0.045983,
        1.570796,1.119777,0.834702,
        1.570796,1.121943,-2.669042,
        1.570796,1.124930,2.219917,
        1.570796,1.125217,1.336294,
        1.570796,1.130958,2.972931,
        1.570796,1.131400,-0.306629,
        1.570796,1.137135,-0.936514,
        1.570796,1.139125,-1.924528,
        1.570796,1.139877,0.276382,
        1.570796,1.142003,-3.058022,
        1.570796,1.147363,-0.558448,
        1.570796,1.150348,-0.176418,
        1.570796,1.151198,1.197016,
        1.570796,1.153930,-1.597104,
        1.570796,1.155389,2.837148,
        1.570796,1.155458,-1.460623,
        1.570796,1.157331,-2.906362,
        1.570796,1.158594,2.487618,
        1.570796,1.162103,-1.322553,
        1.570796,1.165407,3.092861,
        1.570796,1.168025,0.717075,
        1.570796,1.168406,-2.301519,
        1.570796,1.174782,-2.538755,
        1.570796,1.177998,1.061256,
        1.570796,1.178898,1.827305,
        1.570796,1.181331,-2.060330,
        1.570796,1.182540,-0.798952,
        1.570796,1.184569,0.147517,
        1.570796,1.184734,1.968919,
        1.570796,1.184804,0.593386,
        1.570796,1.188278,1.690808,
        1.570796,1.194360,2.110440,
        1.570796,1.194776,2.710455,
        1.570796,1.197745,-1.173705,
        1.570796,1.199898,-2.777257,
        1.570796,1.201164,2.355931,
        1.570796,1.203760,1.556404,
        1.570796,1.205845,-0.415993,
        1.570796,1.206585,0.933648,
        1.570796,1.206950,-0.037248,
        1.570796,1.209341,0.471162,
        1.570796,1.213483,-1.039417,
        1.570796,1.218146,-0.672643,
        1.570796,1.221608,1.421669,
        1.570796,1.226043,-1.693533,
        1.570796,1.232870,-1.822773,
        1.570796,1.234375,-2.421194,
        1.570796,1.234531,-2.190148,
        1.570796,1.239414,0.348959,
        1.570796,1.240183,2.587255,
        1.570796,1.242312,2.941234,
        1.570796,1.244075,1.286644,
        1.570796,1.245961,-2.997461,
        1.570796,1.246495,-2.646872,
        1.570796,1.247736,2.242810,
        1.570796,1.253604,0.812222,
        1.570796,1.257816,-0.894317,
        1.570796,1.258247,-1.953634,
        1.570796,1.258993,-0.284099,
        1.570796,1.262380,-0.157609,
        1.570796,1.262791,-1.549229,
        1.570796,1.263772,-0.533073,
        1.570796,1.266731,-1.412461,
        1.570796,1.268591,-3.122146,
        1.570796,1.269614,-1.275246,
        1.570796,1.270745,1.152533,
        1.570796,1.273606,0.224545,
        1.570796,1.279716,2.812101,
        1.570796,1.283786,2.462124,
        1.570796,1.284798,0.078209,
        1.570796,1.285402,-2.882049,
        1.570796,1.286936,1.886053,
        1.570796,1.289408,0.683122,
        1.570796,1.294736,1.756184,
        1.570796,1.296978,-2.311511,
        1.570796,1.298546,2.020601,
        1.570796,1.300924,1.022478,
        1.570796,1.303305,3.043152,
        1.570796,1.304210,-2.084225,
        1.570796,1.305241,-2.530249,
        1.570796,1.307513,1.628003,
        1.570796,1.310722,0.558267,
        1.570796,1.314142,2.139329,
        1.570796,1.316087,-0.777971,
        1.570796,1.317907,-1.121023,
        1.570796,1.321019,-2.749776,
        1.570796,1.321609,2.687717,
        1.570796,1.322568,1.499415,
        1.570796,1.322972,-0.395699,
        1.570796,1.325286,-0.994192,
        1.570796,1.329146,-0.648513,
        1.570796,1.329795,-0.042570,
        1.570796,1.331651,2.344497,
        1.570796,1.336522,0.435098,
        1.570796,1.339410,0.900087,
        1.570796,1.339550,1.370353,
        1.570796,1.340441,-1.731580,
        1.570796,1.351855,-1.855234,
        1.570796,1.360161,-3.018404,
        1.570796,1.361031,-1.613971,
        1.570796,1.361196,2.914780,
        1.570796,1.361998,-2.207254,
        1.570796,1.362302,1.239384,
        1.570796,1.365228,2.563808,
        1.570796,1.365452,0.310522,
        1.570796,1.367028,-2.420041,
        1.570796,1.373422,-1.484281,
        1.570796,1.374245,-1.226417,
        1.570796,1.375020,-0.170393,
        1.570796,1.376560,-1.355481,
        1.570796,1.377070,-1.980984,
        1.570796,1.377578,-2.636339,
        1.570796,1.378096,0.774626,
        1.570796,1.379086,-0.508192,
        1.570796,1.382562,0.177141,
        1.570796,1.388849,2.233847,
        1.570796,1.389268,-0.886349,
        1.570796,1.391324,1.109396,
        1.570796,1.391699,-0.288725,
        1.570796,1.392869,3.136903,
        1.570796,1.393612,1.952016,
        1.570796,1.395545,-2.850807,
        1.570796,1.401926,1.822930,
        1.570796,1.402149,2.788898,
        1.570796,1.409552,0.647980,
        1.570796,1.410840,0.056219,
        1.570796,1.411944,2.444377,
        1.570796,1.413639,1.697982,
        1.570796,1.426479,1.573427,
        1.570796,1.426725,2.102242,
        1.570796,1.427868,0.984637,
        1.570796,1.429580,-2.107232,
        1.570796,1.431375,-2.313269,
        1.570796,1.431907,3.015410,
        1.570796,1.432325,-1.029054,
        1.570796,1.433341,0.524268,
        1.570796,1.435181,-0.739771,
        1.570796,1.437208,-2.529085,
        1.570796,1.439041,-0.617966,
        1.570796,1.441473,1.448544,
        1.570796,1.444527,2.664744,
        1.570796,1.450513,-0.400625,
        1.570796,1.451907,-0.066823,
        1.570796,1.452754,-2.738631,
        1.570796,1.454061,-2.950416,
        1.570796,1.456736,1.323154,
        1.570796,1.458370,0.400433,
        1.570796,1.459314,-1.765003,
        1.570796,1.459902,-1.145835,
        1.570796,1.464952,2.329245,
        1.570796,1.466050,0.862104,
        1.570796,1.471330,-1.886160,
        1.570796,1.475119,-1.648255,
        1.570796,1.480878,1.194576,
        1.570796,1.480942,0.273235,
        1.570796,1.481086,2.890269,
        1.570796,1.484118,-1.529156,
        1.570796,1.484741,-1.279219,
        1.570796,1.487681,-1.406069,
        1.570796,1.487886,-3.066037,
        1.570796,1.490555,2.543284,
        1.570796,1.492424,-0.190932,
        1.570796,1.495002,-0.839810,
        1.570796,1.495297,2.007188,
        1.570796,1.496414,-2.009267,
        1.570796,1.498261,-2.425270,
        1.570796,1.498907,-2.207327,
        1.570796,1.500204,0.737638,
        1.570796,1.502469,0.151156,
        1.570796,1.507448,-0.514782,
        1.570796,1.509287,2.204647,
        1.570796,1.510518,1.884920,
        1.570796,1.511948,-2.632360,
        1.570796,1.515235,1.068565,
        1.570796,1.522237,3.102018,
        1.570796,1.522373,1.764991,
        1.570796,1.522416,-0.954956,
        1.570796,1.523102,2.765920,
        1.570796,1.524811,-2.844718,
        1.570796,1.527261,-0.304916,
        1.570796,1.529788,0.613208,
        1.570796,1.533648,1.644808,
        1.570796,1.535018,0.028869,
        1.570796,1.541082,2.424608,
        1.570796,1.546010,1.523462,
        1.570796,1.551135,-0.642994,
        1.570796,1.553895,0.489607,
        1.570796,1.555041,0.947813,
        1.570796,1.559671,-1.074934,
        1.570796,1.560345,2.985899,
        1.570796,1.560487,-2.323400,
        1.570796,1.561127,1.401111,
        1.570796,1.567193,2.643172,
        1.570796,1.568892,2.098362,
        1.570796,1.571702,-2.098435,
        1.570796,1.572701,-2.528730,
        1.570796,1.573572,1.278593,
        1.570796,1.574849,-0.091505,
        1.570796,1.576334,0.365236,
        1.570796,1.577029,-2.962888,
        1.570796,1.577688,-1.814941,
        1.570796,1.579542,-0.755917,
        1.570796,1.579624,-1.199759,
        1.570796,1.581515,-0.419414,
        1.570796,1.584581,-2.738702,
        1.570796,1.586948,-1.696296,
        1.570796,1.588901,2.301496,
        1.570796,1.589936,0.825020,
        1.570796,1.590018,-1.941652,
        1.570796,1.596968,-1.328358,
        1.570796,1.597178,-1.575607,
        1.570796,1.600470,-1.453196,
        1.570796,1.601408,0.242409,
        1.570796,1.602259,1.152656,
        1.570796,1.603064,2.865774,
        1.570796,1.612085,1.968802,
        1.570796,1.612720,-3.080498,
        1.570796,1.616061,2.521531,
        1.570796,1.616900,-0.207457,
        1.570796,1.620988,-0.875920,
        1.570796,1.622114,0.701509,
        1.570796,1.622749,0.121426,
        1.570796,1.622993,-2.221629,
        1.570796,1.629346,-0.542669,
        1.570796,1.630007,1.844552,
        1.570796,1.634747,-2.426198,
        1.570796,1.642346,1.722617,
        1.570796,1.643925,1.034629,
        1.570796,1.644848,2.743774,
        1.570796,1.645258,2.188109,
        1.570796,1.645417,-2.634725,
        1.570796,1.647395,3.085437,
        1.570796,1.648538,-2.855044,
        1.570796,1.649932,0.576808,
        1.570796,1.652224,-0.996105,
        1.570796,1.652871,1.600681,
        1.570796,1.657997,-0.325127,
        1.570796,1.661936,0.001037,
        1.570796,1.665187,1.477402,
        1.570796,1.665496,2.397872,
        1.570796,1.673267,-0.665453,
        1.570796,1.673557,0.452526,
        1.570796,1.679277,-1.122421,
        1.570796,1.679696,0.912482,
        1.570796,1.682420,1.353922,
        1.570796,1.685031,-2.009038,
        1.570796,1.685196,2.967046,
        1.570796,1.690107,2.620910,
        1.570796,1.690453,-2.123076,
        1.570796,1.691693,-1.871572,
        1.570796,1.691817,2.061688,
        1.570796,1.692644,1.237670,
        1.570796,1.696386,0.328123,
        1.570796,1.697162,-1.249026,
        1.570796,1.697488,-2.322785,
        1.570796,1.698726,-2.973576,
        1.570796,1.699172,-1.749717,
        1.570796,1.704478,-0.106295,
        1.570796,1.707244,-0.444770,
        1.570796,1.707500,-2.531497,
        1.570796,1.707714,-0.787321,
        1.570796,1.708430,-1.627552,
        1.570796,1.710583,-1.376890,
        1.570796,1.712415,-2.747921,
        1.570796,1.713125,-1.503073,
        1.570796,1.713452,0.789419,
        1.570796,1.720612,2.279888,
        1.570796,1.724852,0.202831,
        1.570796,1.724868,2.846246,
        1.570796,1.728940,1.931535,
        1.570796,1.732838,1.127584,
        1.570796,1.735719,-3.092485,
        1.570796,1.739676,2.496017,
        1.570796,1.739971,-0.228876,
        1.570796,1.743328,0.663347,
        1.570796,1.744927,-0.911892,
        1.570796,1.745952,0.084733,
        1.570796,1.748077,1.804354,
        1.570796,1.752959,-0.567548,
        1.570796,1.760169,1.678132,
        1.570796,1.763887,-2.218450,
        1.570796,1.766737,2.722752,
        1.570796,1.767999,1.004737,
        1.570796,1.769149,1.550298,
        1.570796,1.770119,-2.427175,
        1.570796,1.770460,2.154208,
        1.570796,1.770514,3.070982,
        1.570796,1.770735,0.537245,
        1.570796,1.772673,-2.864760,
        1.570796,1.773527,-1.039582,
        1.570796,1.776419,-2.642421,
        1.570796,1.780769,1.416330,
        1.570796,1.786634,-0.346247,
        1.570796,1.788829,-1.947553,
        1.570796,1.793737,0.410270,
        1.570796,1.794553,2.374618,
        1.570796,1.794699,-0.694061,
        1.570796,1.797469,-1.169105,
        1.570796,1.803705,0.881076,
        1.570796,1.804124,1.264249,
        1.570796,1.807825,-2.090455,
        1.570796,1.808822,-1.817375,
        1.570796,1.809020,2.949741,
        1.570796,1.812466,2.597595,
        1.570796,1.813159,2.025395,
        1.570796,1.813816,-1.298435,
        1.570796,1.815016,0.278951,
        1.570796,1.816510,-0.008631,
        1.570796,1.817772,-1.689625,
        1.570796,1.821360,-2.984352,
        1.570796,1.822312,-0.127211,
        1.570796,1.823386,-1.429807,
        1.570796,1.823834,-1.560483,
        1.570796,1.832060,-0.821627,
        1.570796,1.832690,-0.468440,
        1.570796,1.833945,-2.320278,
        1.570796,1.834787,0.751968,
        1.570796,1.840579,-2.536445,
        1.570796,1.841033,-2.756932,
        1.570796,1.846350,1.893983,
        1.570796,1.846644,1.113766,
        1.570796,1.846833,2.826543,
        1.570796,1.847325,2.249051,
        1.570796,1.854119,0.134161,
        1.570796,1.858863,-3.106022,
        1.570796,1.865046,1.763384,
        1.570796,1.865619,-0.952036,
        1.570796,1.865912,0.624453,
        1.570796,1.866694,2.473686,
        1.570796,1.867524,-0.240083,
        1.570796,1.875561,1.632831,
        1.570796,1.876049,-0.594378,
        1.570796,1.880509,1.500204,
        1.570796,1.883504,1.360457,
        1.570796,1.886971,-2.196665,
        1.570796,1.888979,0.982922,
        1.570796,1.889815,2.701570,
        1.570796,1.892144,-2.005340,
        1.570796,1.892348,0.496522,
        1.570796,1.893281,-1.083976,
        1.570796,1.893828,3.053453,
        1.570796,1.895117,2.121107,
        1.570796,1.898741,-2.875735,
        1.570796,1.904429,-2.428029,
        1.570796,1.908408,-2.648903,
        1.570796,1.911828,-0.364489,
        1.570796,1.912224,0.366565,
        1.570796,1.914854,1.214468,
        1.570796,1.915467,-1.883143,
        1.570796,1.915763,-0.724679,
        1.570796,1.915790,-1.216366,
        1.570796,1.922256,2.347806,
        1.570796,1.922616,0.844633,
        1.570796,1.925185,0.231170,
        1.570796,1.927474,-1.754616,
        1.570796,1.928342,-1.351174,
        1.570796,1.933328,-1.623328,
        1.570796,1.933878,0.034150,
        1.570796,1.934585,-1.487979,
        1.570796,1.934802,2.927634,
        1.570796,1.934862,1.989776,
        1.570796,1.936327,-0.099350,
        1.570796,1.938088,2.578507,
        1.570796,1.945677,-2.999805,
        1.570796,1.954572,-0.859208,
        1.570796,1.957868,-0.490784,
        1.570796,1.958989,0.712772,
        1.570796,1.960212,-2.307561,
        1.570796,1.964774,1.855507,
        1.570796,1.964971,-2.100533,
        1.570796,1.969112,1.088273,
        1.570796,1.971438,-2.765415,
        1.570796,1.972312,2.805134,
        1.570796,1.974950,2.220596,
        1.570796,1.975135,-2.539127,
        1.570796,1.982107,1.720829,
        1.570796,1.983195,-3.127984,
        1.570796,1.983318,-0.247182,
        1.570796,1.986720,-0.993220,
        1.570796,1.990381,1.586005,
        1.570796,1.990470,0.580283,
        1.570796,1.990523,1.450326,
        1.570796,1.993909,1.312897,
        1.570796,1.994694,2.452813,
        1.570796,1.999226,-0.620717,
        1.570796,2.005145,0.941012,
        1.570796,2.007301,0.142337,
        1.570796,2.014149,0.444138,
        1.570796,2.014338,-1.127392,
        1.570796,2.018131,3.021778,
        1.570796,2.019676,-1.980944,
        1.570796,2.019912,2.686612,
        1.570796,2.020545,2.088672,
        1.570796,2.026074,0.300945,
        1.570796,2.026196,-2.889057,
        1.570796,2.031959,-1.268178,
        1.570796,2.032917,-2.199917,
        1.570796,2.033572,-1.842884,
        1.570796,2.034712,-2.419505,
        1.570796,2.035818,-0.379479,
        1.570796,2.036004,-0.757062,
        1.570796,2.040501,-1.411488,
        1.570796,2.042874,-1.701367,
        1.570796,2.043060,-2.653746,
        1.570796,2.043729,-1.557938,
        1.570796,2.048538,0.806623,
        1.570796,2.049006,1.199930,
        1.570796,2.049502,-0.137989,
        1.570796,2.052181,2.326959,
        1.570796,2.052794,-0.008526,
        1.570796,2.057608,1.951726,
        1.570796,2.065568,2.567730,
        1.570796,2.070371,2.879808,
        1.570796,2.071256,-3.018961,
        1.570796,2.079746,-0.897699,
        1.570796,2.082784,1.042403,
        1.570796,2.083752,-0.511086,
        1.570796,2.084401,1.810502,
        1.570796,2.085731,0.671081,
        1.570796,2.097496,1.373172,
        1.570796,2.099185,1.667006,
        1.570796,2.100482,-2.301084,
        1.570796,2.103118,2.194745,
        1.570796,2.103223,1.521685,
        1.570796,2.103695,-2.777707,
        1.570796,2.107523,3.129581,
        1.570796,2.108277,-2.095907,
        1.570796,2.108891,-2.532166,
        1.570796,2.111003,-0.268632,
        1.570796,2.112276,-1.027246,
        1.570796,2.114380,0.531457,
        1.570796,2.119166,0.100552,
        1.570796,2.119564,2.754818,
        1.570796,2.122474,-0.648314,
        1.570796,2.125183,2.445598,
        1.570796,2.129501,0.238290,
        1.570796,2.131178,0.386663,
        1.570796,2.131223,-1.174311,
        1.570796,2.134269,-1.948042,
        1.570796,2.134684,0.907065,
        1.570796,2.137615,2.988806,
        1.570796,2.145635,-1.327708,
        1.570796,2.146467,2.055081,
        1.570796,2.146582,-1.793788,
        1.570796,2.149210,-1.482321,
        1.570796,2.152901,-1.636788,
        1.570796,2.156337,1.151558,
        1.570796,2.156474,-2.912065,
        1.570796,2.157194,-0.792563,
        1.570796,2.166724,-0.401638,
        1.570796,2.166980,-0.134879,
        1.570796,2.173644,-2.403429,
        1.570796,2.175711,-2.655523,
        1.570796,2.177132,2.600419,
        1.570796,2.178879,0.767722,
        1.570796,2.180752,1.908485,
        1.570796,2.181338,2.313721,
        1.570796,2.185580,1.287673,
        1.570796,2.195400,-2.217546,
        1.570796,2.200418,-3.056442,
        1.570796,2.201829,-0.001735,
        1.570796,2.203804,1.756176,
        1.570796,2.211602,1.443930,
        1.570796,2.212686,-0.544809,
        1.570796,2.213937,0.621892,
        1.570796,2.214075,1.600691,
        1.570796,2.216256,1.020029,
        1.570796,2.219060,2.886631,
        1.570796,2.223375,-0.906185,
        1.570796,2.226002,-1.060248,
        1.570796,2.226413,-2.064034,
        1.570796,2.230554,2.171334,
        1.570796,2.232413,2.735777,
        1.570796,2.233709,3.070830,
        1.570796,2.234383,-2.792065,
        1.570796,2.237183,0.468799,
        1.570796,2.237695,-0.271388,
        1.570796,2.239419,0.151693,
        1.570796,2.243347,-2.521049,
        1.570796,2.244035,-1.229456,
        1.570796,2.244355,0.309899,
        1.570796,2.245175,2.460236,
        1.570796,2.245628,-1.899216,
        1.570796,2.245973,-0.702234,
        1.570796,2.255601,-1.728131,
        1.570796,2.256702,-1.401367,
        1.570796,2.261466,-1.557500,
        1.570796,2.267390,0.878499,
        1.570796,2.272099,2.018594,
        1.570796,2.278618,1.170386,
        1.570796,2.283304,-2.345584,
        1.570796,2.285175,-2.943090,
        1.570796,2.293000,-0.124451,
        1.570796,2.294058,-0.421070,
        1.570796,2.303283,1.855938,
        1.570796,2.304968,2.608978,
        1.570796,2.305347,2.311346,
        1.570796,2.307188,-2.656134,
        1.570796,2.307301,1.336200,
        1.570796,2.309655,0.726020,
        1.570796,2.317634,-2.185225,
        1.570796,2.321510,1.685664,
        1.570796,2.324376,-3.111006,
        1.570796,2.324731,1.511186,
        1.570796,2.329587,0.035498,
        1.570796,2.330158,2.977264,
        1.570796,2.335096,-0.588468,
        1.570796,2.336046,-1.115799,
        1.570796,2.338060,2.803613,
        1.570796,2.338083,-0.943400,
        1.570796,2.340210,0.562145,
        1.570796,2.341931,1.023926,
        1.570796,2.343265,-0.779061,
        1.570796,2.343434,-2.013692,
        1.570796,2.353029,-1.310529,
        1.570796,2.354868,0.211543,
        1.570796,2.355120,0.388593,
        1.570796,2.355165,-1.833173,
        1.570796,2.355462,2.151441,
        1.570796,2.362948,-2.487477,
        1.570796,2.363101,-2.810456,
        1.570796,2.363394,-0.272918,
        1.570796,2.364509,-1.644533,
        1.570796,2.370465,2.466824,
        1.570796,2.386110,-1.481351,
        1.570796,2.394960,0.862133,
        1.570796,2.395716,1.977184,
        1.570796,2.396339,1.200900,
        1.570796,2.407941,-2.316753,
        1.570796,2.409450,-2.988432,
        1.570796,2.417202,-0.103363,
        1.570796,2.417386,-0.443097,
        1.570796,2.417950,2.666662,
        1.570796,2.423190,1.789783,
        1.570796,2.423692,1.392696,
        1.570796,2.431122,3.086379,
        1.570796,2.431406,-2.647820,
        1.570796,2.431584,2.307631,
        1.570796,2.434090,1.592818,
        1.570796,2.435993,0.683844,
        1.570796,2.440898,-2.133300,
        1.570796,2.442222,-1.195837,
        1.570796,2.443112,2.883761,
        1.570796,2.449191,-0.633651,
        1.570796,2.450821,0.084741,
        1.570796,2.453499,-1.008861,
        1.570796,2.460010,-1.938011,
        1.570796,2.460199,-0.822008,
        1.570796,2.460902,0.490079,
        1.570796,2.461774,-1.734913,
        1.570796,2.463986,1.034407,
        1.570796,2.467647,0.287667,
        1.570796,2.479849,2.126925,
        1.570796,2.485907,-2.835471,
        1.570796,2.485945,-1.374853,
        1.570796,2.487747,2.505912,
        1.570796,2.487957,-0.274279,
        1.570796,2.489741,-2.468428,
        1.570796,2.512538,1.239900,
        1.570796,2.517528,1.923289,
        1.570796,2.518454,-1.565852,
        1.570796,2.519348,0.843503,
        1.570796,2.522766,-3.059231,
        1.570796,2.528686,2.737228,
        1.570796,2.534798,-2.266671,
        1.570796,2.536533,1.466274,
        1.570796,2.538504,1.699812,
        1.570796,2.539516,-0.476942,
        1.570796,2.539532,-0.074548,
        1.570796,2.543197,2.990252,
        1.570796,2.554022,2.320619,
        1.570796,2.556919,-2.655005,
        1.570796,2.558146,0.627191,
        1.570796,2.558223,-1.164189,
        1.570796,2.563452,-2.044028,
        1.570796,2.568462,0.150655,
        1.570796,2.569130,-0.700837,
        1.570796,2.570050,-1.804492,
        1.570796,2.572710,-0.934665,
        1.570796,2.575994,0.391718,
        1.570796,2.584774,1.046703,
        1.570796,2.601432,2.102535,
        1.570796,2.603559,2.555114,
        1.570796,2.604150,-2.885208,
        1.570796,2.608630,-1.378011,
        1.570796,2.611256,-0.279110,
        1.570796,2.617344,-2.439480,
        1.570796,2.627402,1.293552,
        1.570796,2.635396,1.848850,
        1.570796,2.635686,2.834950,
        1.570796,2.637087,3.132730,
        1.570796,2.642029,0.813274,
        1.570796,2.645260,1.570203,
        1.570796,2.646182,-1.614036,
        1.570796,2.660113,-2.186513,
        1.570796,2.660483,-0.034608,
        1.570796,2.661713,-0.529511,
        1.570796,2.672720,-1.110422,
        1.570796,2.676626,2.335353,
        1.570796,2.676638,0.540364,
        1.570796,2.676780,-1.901158,
        1.570796,2.681411,-2.667965,
        1.570796,2.682152,0.245751,
        1.570796,2.682859,-0.817071,
        1.570796,2.705107,1.062142,
        1.570796,2.717705,2.625995,
        1.570796,2.721001,2.060097,
        1.570796,2.722319,-2.954069,
        1.570796,2.729840,-1.368237,
        1.570796,2.734349,-0.289298,
        1.570796,2.738409,2.975671,
        1.570796,2.740366,1.374341,
        1.570796,2.744633,-2.391311,
        1.570796,2.747398,1.727166,
        1.570796,2.763799,0.760372,
        1.570796,2.768238,-1.679243,
        1.570796,2.779245,-2.046268,
        1.570796,2.780208,-0.626453,
        1.570796,2.780471,0.027732,
        1.570796,2.784696,-1.013128,
        1.570796,2.789817,0.395431,
        1.570796,2.801154,2.353143,
        1.570796,2.807091,-2.687477,
        1.570796,2.825762,1.085035,
        1.570796,2.830209,2.734943,
        1.570796,2.833517,-3.085508,
        1.570796,2.838301,1.980061,
        1.570796,2.849871,1.512012,
        1.570796,2.850089,-1.343899,
        1.570796,2.856905,-0.318441,
        1.570796,2.870445,-2.291001,
        1.570796,2.885017,0.655562,
        1.570796,2.886428,-1.793217,
        1.570796,2.890337,-0.825633,
        1.570796,2.898932,0.139973,
        1.570796,2.931927,-2.728768,
        1.570796,2.932096,2.377790,
        1.570796,2.937409,2.947245,
        1.570796,2.948048,1.133127,
        1.570796,2.950896,1.797889,
        1.570796,2.970019,-1.282382,
        1.570796,2.978333,-0.412090,
        1.570796,2.996437,-2.089121,
        1.570796,3.011274,0.434110,
        1.570796,3.050410,3.138492,
        1.570796,3.064492,1.647259,
        1.570796,3.088868,-0.989501,
        1.675516,0.053884,-2.408394,
        1.675516,0.077645,1.453433,
        1.675516,0.088408,-0.311160,
        1.675516,0.142314,2.590387,
        1.675516,0.147206,-1.385164,
        1.675516,0.163506,0.573203,
        1.675516,0.165485,-2.875396,
        1.675516,0.182564,-2.125821,
        1.675516,0.190569,1.929141,
        1.675516,0.201102,1.249082,
        1.675516,0.201397,-0.716646,
        1.675516,0.210568,-0.069559,
        1.675516,0.251672,2.848900,
        1.675516,0.266262,-1.678068,
        1.675516,0.269108,-1.215563,
        1.675516,0.273777,-2.506756,
        1.675516,0.279854,0.814581,
        1.675516,0.281664,-2.969147,
        1.675516,0.283625,2.405603,
        1.675516,0.284058,0.357350,
        1.675516,0.294107,1.593880,
        1.675516,0.302491,-0.400410,
        1.675516,0.308059,2.021491,
        1.675516,0.312219,-2.072057,
        1.675516,0.321905,-0.829121,
        1.675516,0.329984,1.201621,
        1.675516,0.339714,-0.013712,
        1.675516,0.368784,2.974974,
        1.675516,0.376428,-1.495532,
        1.675516,0.378103,2.646436,
        1.675516,0.385088,0.590355,
        1.675516,0.386248,-2.690943,
        1.675516,0.389851,-2.340031,
        1.675516,0.391502,-1.171374,
        1.675516,0.398249,-1.812694,
        1.675516,0.400542,-2.995729,
        1.675516,0.404273,0.918163,
        1.675516,0.405617,1.773533,
        1.675516,0.409009,-0.570243,
        1.675516,0.410700,0.270433,
        1.675516,0.415893,1.464890,
        1.675516,0.420680,2.088164,
        1.675516,0.424766,-0.258729,
        1.675516,0.426507,2.367872,
        1.675516,0.441292,-0.890727,
        1.675516,0.451839,-2.078193,
        1.675516,0.459270,1.191562,
        1.675516,0.470622,0.014002,
        1.675516,0.481292,-2.529098,
        1.675516,0.482460,2.849843,
        1.675516,0.492205,-1.397963,
        1.675516,0.493314,3.117264,
        1.675516,0.497870,0.727828,
        1.675516,0.499783,-1.654379,
        1.675516,0.502651,0.463333,
        1.675516,0.505463,2.594557,
        1.675516,0.506104,-2.874910,
        1.675516,0.513579,-1.147354,
        1.675516,0.518569,1.644237,
        1.675516,0.519001,1.900488,
        1.675516,0.519277,-0.673809,
        1.675516,0.524769,-0.421955,
        1.675516,0.525564,-2.288838,
        1.675516,0.532541,0.978737,
        1.675516,0.534610,2.154808,
        1.675516,0.535020,-1.891846,
        1.675516,0.540208,1.400497,
        1.675516,0.542006,0.228337,
        1.675516,0.550762,-0.180771,
        1.675516,0.560531,2.378474,
        1.675516,0.561629,-0.924576,
        1.675516,0.567465,-2.679392,
        1.675516,0.588527,1.181755,
        1.675516,0.591643,-2.098364,
        1.675516,0.594963,2.997575,
        1.675516,0.599643,-3.045235,
        1.675516,0.599962,2.773556,
        1.675516,0.602620,0.029270,
        1.675516,0.605711,0.598454,
        1.675516,0.607708,-1.541026,
        1.675516,0.609360,-2.464379,
        1.675516,0.610738,-1.326786,
        1.675516,0.616968,0.820489,
        1.675516,0.626379,1.773610,
        1.675516,0.628626,-1.751180,
        1.675516,0.630671,0.391949,
        1.675516,0.632622,-0.737642,
        1.675516,0.632708,2.566722,
        1.675516,0.632797,1.989858,
        1.675516,0.634455,-0.534790,
        1.675516,0.636205,1.561170,
        1.675516,0.637681,-1.123423,
        1.675516,0.641441,-2.842109,
        1.675516,0.644807,-0.327139,
        1.675516,0.652440,2.197129,
        1.675516,0.660897,-2.278593,
        1.675516,0.664008,1.360091,
        1.675516,0.670850,-1.945257,
        1.675516,0.671396,1.004859,
        1.675516,0.673985,0.203855,
        1.675516,0.680326,-0.133370,
        1.675516,0.685083,-0.942363,
        1.675516,0.688204,-2.632204,
        1.675516,0.689777,2.385011,
        1.675516,0.700065,3.115430,
        1.675516,0.704080,2.911636,
        1.675516,0.712883,1.174216,
        1.675516,0.715701,0.699906,
        1.675516,0.718851,-1.442215,
        1.675516,0.721019,2.721536,
        1.675516,0.725300,-2.975134,
        1.675516,0.727136,-1.633288,
        1.675516,0.727399,-2.120567,
        1.675516,0.730751,0.522488,
        1.675516,0.732892,-1.254330,
        1.675516,0.734333,-2.446184,
        1.675516,0.735125,0.037270,
        1.675516,0.737893,1.866463,
        1.675516,0.740650,1.681075,
        1.675516,0.745971,-0.781286,
        1.675516,0.749432,-0.442559,
        1.675516,0.750102,2.050636,
        1.675516,0.751046,0.863821,
        1.675516,0.757018,-0.608833,
        1.675516,0.757046,-1.816019,
        1.675516,0.757389,1.501184,
        1.675516,0.759800,2.546181,
        1.675516,0.759859,0.351923,
        1.675516,0.761141,-2.781102,
        1.675516,0.769035,-1.078574,
        1.675516,0.770166,-0.266604,
        1.675516,0.772218,2.227380,
        1.675516,0.788382,1.331051,
        1.675516,0.790269,-2.282409,
        1.675516,0.803410,-1.984567,
        1.675516,0.803813,0.191520,
        1.675516,0.806832,3.022155,
        1.675516,0.807476,-2.597675,
        1.675516,0.812512,-0.107335,
        1.675516,0.814161,1.010949,
        1.675516,0.817732,-3.084486,
        1.675516,0.817846,2.390688,
        1.675516,0.819323,2.845960,
        1.675516,0.819430,-0.915017,
        1.675516,0.829198,-1.532491,
        1.675516,0.829531,0.665695,
        1.675516,0.829816,1.167147,
        1.675516,0.833613,-1.357492,
        1.675516,0.842945,-2.907318,
        1.675516,0.844883,2.679708,
        1.675516,0.848189,-1.705003,
        1.675516,0.849888,1.767901,
        1.675516,0.851126,0.493370,
        1.675516,0.852911,1.933394,
        1.675516,0.857516,-2.432863,
        1.675516,0.858261,-1.191729,
        1.675516,0.858279,-2.138895,
        1.675516,0.859723,-0.734928,
        1.675516,0.859754,1.603292,
        1.675516,0.861655,-0.394821,
        1.675516,0.864300,0.043549,
        1.675516,0.866940,0.851229,
        1.675516,0.869748,2.095969,
        1.675516,0.877352,-0.574668,
        1.675516,0.879254,-2.733338,
        1.675516,0.882554,1.443757,
        1.675516,0.884060,0.334199,
        1.675516,0.885626,-1.866061,
        1.675516,0.889834,2.529957,
        1.675516,0.893140,2.251698,
        1.675516,0.897463,-0.240822,
        1.675516,0.898758,-1.031948,
        1.675516,0.913918,1.287618,
        1.675516,0.916555,-2.287173,
        1.675516,0.918305,3.107845,
        1.675516,0.919522,2.948277,
        1.675516,0.927138,-2.569686,
        1.675516,0.929031,0.185383,
        1.675516,0.933766,-3.015917,
        1.675516,0.934873,-2.010822,
        1.675516,0.935636,-1.445855,
        1.675516,0.936702,-0.866461,
        1.675516,0.937395,0.729108,
        1.675516,0.937910,2.787575,
        1.675516,0.939144,-0.092962,
        1.675516,0.941718,1.125085,
        1.675516,0.942873,-1.607680,
        1.675516,0.943701,0.980328,
        1.675516,0.947774,2.395256,
        1.675516,0.952433,0.584347,
        1.675516,0.954456,-1.287293,
        1.675516,0.958731,-2.852552,
        1.675516,0.961838,1.839284,
        1.675516,0.961898,-0.469425,
        1.675516,0.965426,1.686615,
        1.675516,0.970108,1.990771,
        1.675516,0.972820,-1.762780,
        1.675516,0.973500,-0.711881,
        1.675516,0.975253,2.641082,
        1.675516,0.977824,0.441928,
        1.675516,0.981485,-2.424766,
        1.675516,0.981877,1.537100,
        1.675516,0.982609,-1.136920,
        1.675516,0.984265,-2.156073,
        1.675516,0.989150,0.045493,
        1.675516,0.993052,2.137478,
        1.675516,0.997766,-2.693871,
        1.675516,0.999847,-0.334738,
        1.675516,1.001920,0.848054,
        1.675516,1.006146,1.389734,
        1.675516,1.012180,0.304066,
        1.675516,1.016938,2.270674,
        1.675516,1.017371,-0.979790,
        1.675516,1.019495,-1.894487,
        1.675516,1.023796,-0.583691,
        1.675516,1.026525,2.506945,
        1.675516,1.027004,-0.203653,
        1.675516,1.028027,3.036573,
        1.675516,1.031435,1.241021,
        1.675516,1.032316,2.884221,
        1.675516,1.035863,-3.110475,
        1.675516,1.042085,-2.292950,
        1.675516,1.043272,-1.522986,
        1.675516,1.046265,-2.958847,
        1.675516,1.049272,-1.376344,
        1.675516,1.050039,-2.552086,
        1.675516,1.053682,0.689863,
        1.675516,1.055458,0.170394,
        1.675516,1.056622,-0.831866,
        1.675516,1.057515,1.097998,
        1.675516,1.058706,-2.036438,
        1.675516,1.061389,-1.672248,
        1.675516,1.063758,2.743093,
        1.675516,1.070157,-0.073422,
        1.675516,1.072154,1.761506,
        1.675516,1.074029,1.909193,
        1.675516,1.074204,0.545276,
        1.675516,1.074714,-2.805147,
        1.675516,1.076262,0.966104,
        1.675516,1.077642,-1.227073,
        1.675516,1.084151,1.618525,
        1.675516,1.085085,2.377673,
        1.675516,1.086822,-0.441112,
        1.675516,1.087142,2.058815,
        1.675516,1.098383,-0.696430,
        1.675516,1.099138,-1.084033,
        1.675516,1.103208,1.477773,
        1.675516,1.103783,0.408254,
        1.675516,1.106275,-2.420800,
        1.675516,1.108079,2.612943,
        1.675516,1.109345,-2.172718,
        1.675516,1.115957,-1.787597,
        1.675516,1.118930,0.045983,
        1.675516,1.119777,0.834702,
        1.675516,1.121943,-2.669042,
        1.675516,1.124930,2.219917,
        1.675516,1.125217,1.336294,
        1.675516,1.130958,2.972931,
        1.675516,1.131400,-0.306629,
        1.675516,1.137135,-0.936514,
        1.675516,1.139125,-1.924528,
        1.675516,1.139877,0.276382,
        1.675516,1.142003,-3.058022,
        1.675516,1.147363,-0.558448,
        1.675516,1.150348,-0.176418,
        1.675516,1.151198,1.197016,
        1.675516,1.153930,-1.597104,
        1.675516,1.155389,2.837148,
        1.675516,1.155458,-1.460623,
        1.675516,1.157331,-2.906362,
        1.675516,1.158594,2.487618,
        1.675516,1.162103,-1.322553,
        1.675516,1.165407,3.092861,
        1.675516,1.168025,0.717075,
        1.675516,1.168406,-2.301519,
        1.675516,1.174782,-2.538755,
        1.675516,1.177998,1.061256,
        1.675516,1.178898,1.827305,
        1.675516,1.181331,-2.060330,
        1.675516,1.182540,-0.798952,
        1.675516,1.184569,0.147517,
        1.675516,1.184734,1.968919,
        1.675516,1.184804,0.593386,
        1.675516,1.188278,1.690808,
        1.675516,1.194360,2.110440,
        1.675516,1.194776,2.710455,
        1.675516,1.197745,-1.173705,
        1.675516,1.199898,-2.777257,
        1.675516,1.201164,2.355931,
        1.675516,1.203760,1.556404,
        1.675516,1.205845,-0.415993,
        1.675516,1.206585,0.933648,
        1.675516,1.206950,-0.037248,
        1.675516,1.209341,0.471162,
        1.675516,1.213483,-1.039417,
        1.675516,1.218146,-0.672643,
        1.675516,1.221608,1.421669,
        1.675516,1.226043,-1.693533,
        1.675516,1.232870,-1.822773,
        1.675516,1.234375,-2.421194,
        1.675516,1.234531,-2.190148,
        1.675516,1.239414,0.348959,
        1.675516,1.240183,2.587255,
        1.675516,1.242312,2.941234,
        1.675516,1.244075,1.286644,
        1.675516,1.245961,-2.997461,
        1.675516,1.246495,-2.646872,
        1.675516,1.247736,2.242810,
        1.675516,1.253604,0.812222,
        1.675516,1.257816,-0.894317,
        1.675516,1.258247,-1.953634,
        1.675516,1.258993,-0.284099,
        1.675516,1.262380,-0.157609,
        1.675516,1.262791,-1.549229,
        1.675516,1.263772,-0.533073,
        1.675516,1.266731,-1.412461,
        1.675516,1.268591,-3.122146,
        1.675516,1.269614,-1.275246,
        1.675516,1.270745,1.152533,
        1.675516,1.273606,0.224545,
        1.675516,1.279716,2.812101,
        1.675516,1.283786,2.462124,
        1.675516,1.284798,0.078209,
        1.675516,1.285402,-2.882049,
        1.675516,1.286936,1.886053,
        1.675516,1.289408,0.683122,
        1.675516,1.294736,1.756184,
        1.675516,1.296978,-2.311511,
        1.675516,1.298546,2.020601,
        1.675516,1.300924,1.022478,
        1.675516,1.303305,3.043152,
        1.675516,1.304210,-2.084225,
        1.675516,1.305241,-2.530249,
        1.675516,1.307513,1.628003,
        1.675516,1.310722,0.558267,
        1.675516,1.314142,2.139329,
        1.675516,1.316087,-0.777971,
        1.675516,1.317907,-1.121023,
        1.675516,1.321019,-2.749776,
        1.675516,1.321609,2.687717,
        1.675516,1.322568,1.499415,
        1.675516,1.322972,-0.395699,
        1.675516,1.325286,-0.994192,
        1.675516,1.329146,-0.648513,
        1.675516,1.329795,-0.042570,
        1.675516,1.331651,2.344497,
        1.675516,1.336522,0.435098,
        1.675516,1.339410,0.900087,
        1.675516,1.339550,1.370353,
        1.675516,1.340441,-1.731580,
        1.675516,1.351855,-1.855234,
        1.675516,1.360161,-3.018404,
        1.675516,1.361031,-1.613971,
        1.675516,1.361196,2.914780,
        1.675516,1.361998,-2.207254,
        1.675516,1.362302,1.239384,
        1.675516,1.365228,2.563808,
        1.675516,1.365452,0.310522,
        1.675516,1.367028,-2.420041,
        1.675516,1.373422,-1.484281,
        1.675516,1.374245,-1.226417,
        1.675516,1.375020,-0.170393,
        1.675516,1.376560,-1.355481,
        1.675516,1.377070,-1.980984,
        1.675516,1.377578,-2.636339,
        1.675516,1.378096,0.774626,
        1.675516,1.379086,-0.508192,
        1.675516,1.382562,0.177141,
        1.675516,1.388849,2.233847,
        1.675516,1.389268,-0.886349,
        1.675516,1.391324,1.109396,
        1.675516,1.391699,-0.288725,
        1.675516,1.392869,3.136903,
        1.675516,1.393612,1.952016,
        1.675516,1.395545,-2.850807,
        1.675516,1.401926,1.822930,
        1.675516,1.402149,2.788898,
        1.675516,1.409552,0.647980,
        1.675516,1.410840,0.056219,
        1.675516,1.411944,2.444377,
        1.675516,1.413639,1.697982,
        1.675516,1.426479,1.573427,
        1.675516,1.426725,2.102242,
        1.675516,1.427868,0.984637,
        1.675516,1.429580,-2.107232,
        1.675516,1.431375,-2.313269,
        1.675516,1.431907,3.015410,
        1.675516,1.432325,-1.029054,
        1.675516,1.433341,0.524268,
        1.675516,1.435181,-0.739771,
        1.675516,1.437208,-2.529085,
        1.675516,1.439041,-0.617966,
        1.675516,1.441473,1.448544,
        1.675516,1.444527,2.664744,
        1.675516,1.450513,-0.400625,
        1.675516,1.451907,-0.066823,
        1.675516,1.452754,-2.738631,
        1.675516,1.454061,-2.950416,
        1.675516,1.456736,1.323154,
        1.675516,1.458370,0.400433,
        1.675516,1.459314,-1.765003,
        1.675516,1.459902,-1.145835,
        1.675516,1.464952,2.329245,
        1.675516,1.466050,0.862104,
        1.675516,1.471330,-1.886160,
        1.675516,1.475119,-1.648255,
        1.675516,1.480878,1.194576,
        1.675516,1.480942,0.273235,
        1.675516,1.481086,2.890269,
        1.675516,1.484118,-1.529156,
        1.675516,1.484741,-1.279219,
        1.675516,1.487681,-1.406069,
        1.675516,1.487886,-3.066037,
        1.675516,1.490555,2.543284,
        1.675516,1.492424,-0.190932,
        1.675516,1.495002,-0.839810,
        1.675516,1.495297,2.007188,
        1.675516,1.496414,-2.009267,
        1.675516,1.498261,-2.425270,
        1.675516,1.498907,-2.207327,
        1.675516,1.500204,0.737638,
        1.675516,1.502469,0.151156,
        1.675516,1.507448,-0.514782,
        1.675516,1.509287,2.204647,
        1.675516,1.510518,1.884920,
        1.675516,1.511948,-2.632360,
        1.675516,1.515235,1.068565,
        1.675516,1.522237,3.102018,
        1.675516,1.522373,1.764991,
        1.675516,1.522416,-0.954956,
        1.675516,1.523102,2.765920,
        1.675516,1.524811,-2.844718,
        1.675516,1.527261,-0.304916,
        1.675516,1.529788,0.613208,
        1.675516,1.533648,1.644808,
        1.675516,1.535018,0.028869,
        1.675516,1.541082,2.424608,
        1.675516,1.546010,1.523462,
        1.675516,1.551135,-0.642994,
        1.675516,1.553895,0.489607,
        1.675516,1.555041,0.947813,
        1.675516,1.559671,-1.074934,
        1.675516,1.560345,2.985899,
        1.675516,1.560487,-2.323400,
        1.675516,1.561127,1.401111,
        1.675516,1.567193,2.643172,
        1.675516,1.568892,2.098362,
        1.675516,1.571702,-2.098435,
        1.675516,1.572701,-2.528730,
        1.675516,1.573572,1.278593,
        1.675516,1.574849,-0.091505,
        1.675516,1.576334,0.365236,
        1.675516,1.577029,-2.962888,
        1.675516,1.577688,-1.814941,
        1.675516,1.579542,-0.755917,
        1.675516,1.579624,-1.199759,
        1.675516,1.581515,-0.419414,
        1.675516,1.584581,-2.738702,
        1.675516,1.586948,-1.696296,
        1.675516,1.588901,2.301496,
        1.675516,1.589936,0.825020,
        1.675516,1.590018,-1.941652,
        1.675516,1.596968,-1.328358,
        1.675516,1.597178,-1.575607,
        1.675516,1.600470,-1.453196,
        1.675516,1.601408,0.242409,
        1.675516,1.602259,1.152656,
        1.675516,1.603064,2.865774,
        1.675516,1.612085,1.968802,
        1.675516,1.612720,-3.080498,
        1.675516,1.616061,2.521531,
        1.675516,1.616900,-0.207457,
        1.675516,1.620988,-0.875920,
        1.675516,1.622114,0.701509,
        1.675516,1.622749,0.121426,
        1.675516,1.622993,-2.221629,
        1.675516,1.629346,-0.542669,
        1.675516,1.630007,1.844552,
        1.675516,1.634747,-2.426198,
        1.675516,1.642346,1.722617,
        1.675516,1.643925,1.034629,
        1.675516,1.644848,2.743774,
        1.675516,1.645258,2.188109,
        1.675516,1.645417,-2.634725,
        1.675516,1.647395,3.085437,
        1.675516,1.648538,-2.855044,
        1.675516,1.649932,0.576808,
        1.675516,1.652224,-0.996105,
        1.675516,1.652871,1.600681,
        1.675516,1.657997,-0.325127,
        1.675516,1.661936,0.001037,
        1.675516,1.665187,1.477402,
        1.675516,1.665496,2.397872,
        1.675516,1.673267,-0.665453,
        1.675516,1.673557,0.452526,
        1.675516,1.679277,-1.122421,
        1.675516,1.679696,0.912482,
        1.675516,1.682420,1.353922,
        1.675516,1.685031,-2.009038,
        1.675516,1.685196,2.967046,
        1.675516,1.690107,2.620910,
        1.675516,1.690453,-2.123076,
        1.675516,1.691693,-1.871572,
        1.675516,1.691817,2.061688,
        1.675516,1.692644,1.237670,
        1.675516,1.696386,0.328123,
        1.675516,1.697162,-1.249026,
        1.675516,1.697488,-2.322785,
        1.675516,1.698726,-2.973576,
        1.675516,1.699172,-1.749717,
        1.675516,1.704478,-0.106295,
        1.675516,1.707244,-0.444770,
        1.675516,1.707500,-2.531497,
        1.675516,1.707714,-0.787321,
        1.675516,1.708430,-1.627552,
        1.675516,1.710583,-1.376890,
        1.675516,1.712415,-2.747921,
        1.675516,1.713125,-1.503073,
        1.675516,1.713452,0.789419,
        1.675516,1.720612,2.279888,
        1.675516,1.724852,0.202831,
        1.675516,1.724868,2.846246,
        1.675516,1.728940,1.931535,
        1.675516,1.732838,1.127584,
        1.675516,1.735719,-3.092485,
        1.675516,1.739676,2.496017,
        1.675516,1.739971,-0.228876,
        1.675516,1.743328,0.663347,
        1.675516,1.744927,-0.911892,
        1.675516,1.745952,0.084733,
        1.675516,1.748077,1.804354,
        1.675516,1.752959,-0.567548,
        1.675516,1.760169,1.678132,
        1.675516,1.763887,-2.218450,
        1.675516,1.766737,2.722752,
        1.675516,1.767999,1.004737,
        1.675516,1.769149,1.550298,
        1.675516,1.770119,-2.427175,
        1.675516,1.770460,2.154208,
        1.675516,1.770514,3.070982,
        1.675516,1.770735,0.537245,
        1.675516,1.772673,-2.864760,
        1.675516,1.773527,-1.039582,
        1.675516,1.776419,-2.642421,
        1.675516,1.780769,1.416330,
        1.675516,1.786634,-0.346247,
        1.675516,1.788829,-1.947553,
        1.675516,1.793737,0.410270,
        1.675516,1.794553,2.374618,
        1.675516,1.794699,-0.694061,
        1.675516,1.797469,-1.169105,
        1.675516,1.803705,0.881076,
        1.675516,1.804124,1.264249,
        1.675516,1.807825,-2.090455,
        1.675516,1.808822,-1.817375,
        1.675516,1.809020,2.949741,
        1.675516,1.812466,2.597595,
        1.675516,1.813159,2.025395,
        1.675516,1.813816,-1.298435,
        1.675516,1.815016,0.278951,
        1.675516,1.816510,-0.008631,
        1.675516,1.817772,-1.689625,
        1.675516,1.821360,-2.984352,
        1.675516,1.822312,-0.127211,
        1.675516,1.823386,-1.429807,
        1.675516,1.823834,-1.560483,
        1.675516,1.832060,-0.821627,
        1.675516,1.832690,-0.468440,
        1.675516,1.833945,-2.320278,
        1.675516,1.834787,0.751968,
        1.675516,1.840579,-2.536445,
        1.675516,1.841033,-2.756932,
        1.675516,1.846350,1.893983,
        1.675516,1.846644,1.113766,
        1.675516,1.846833,2.826543,
        1.675516,1.847325,2.249051,
        1.675516,1.854119,0.134161,
        1.675516,1.858863,-3.106022,
        1.675516,1.865046,1.763384,
        1.675516,1.865619,-0.952036,
        1.675516,1.865912,0.624453,
        1.675516,1.866694,2.473686,
        1.675516,1.867524,-0.240083,
        1.675516,1.875561,1.632831,
        1.675516,1.876049,-0.594378,
        1.675516,1.880509,1.500204,
        1.675516,1.883504,1.360457,
        1.675516,1.886971,-2.196665,
        1.675516,1.888979,0.982922,
        1.675516,1.889815,2.701570,
        1.675516,1.892144,-2.005340,
        1.675516,1.892348,0.496522,
        1.675516,1.893281,-1.083976,
        1.675516,1.893828,3.053453,
        1.675516,1.895117,2.121107,
        1.675516,1.898741,-2.875735,
        1.675516,1.904429,-2.428029,
        1.675516,1.908408,-2.648903,
        1.675516,1.911828,-0.364489,
        1.675516,1.912224,0.366565,
        1.675516,1.914854,1.214468,
        1.675516,1.915467,-1.883143,
        1.675516,1.915763,-0.724679,
        1.675516,1.915790,-1.216366,
        1.675516,1.922256,2.347806,
        1.675516,1.922616,0.844633,
        1.675516,1.925185,0.231170,
        1.675516,1.927474,-1.754616,
        1.675516,1.928342,-1.351174,
        1.675516,1.933328,-1.623328,
        1.675516,1.933878,0.034150,
        1.675516,1.934585,-1.487979,
        1.675516,1.934802,2.927634,
        1.675516,1.934862,1.989776,
        1.675516,1.936327,-0.099350,
        1.675516,1.938088,2.578507,
        1.675516,1.945677,-2.999805,
        1.675516,1.954572,-0.859208,
        1.675516,1.957868,-0.490784,
        1.675516,1.958989,0.712772,
        1.675516,1.960212,-2.307561,
        1.675516,1.964774,1.855507,
        1.675516,1.964971,-2.100533,
        1.675516,1.969112,1.088273,
        1.675516,1.971438,-2.765415,
        1.675516,1.972312,2.805134,
        1.675516,1.974950,2.220596,
        1.675516,1.975135,-2.539127,
        1.675516,1.982107,1.720829,
        1.675516,1.983195,-3.127984,
        1.675516,1.983318,-0.247182,
        1.675516,1.986720,-0.993220,
        1.675516,1.990381,1.586005,
        1.675516,1.990470,0.580283,
        1.675516,1.990523,1.450326,
        1.675516,1.993909,1.312897,
        1.675516,1.994694,2.452813,
        1.675516,1.999226,-0.620717,
        1.675516,2.005145,0.941012,
        1.675516,2.007301,0.142337,
        1.675516,2.014149,0.444138,
        1.675516,2.014338,-1.127392,
        1.675516,2.018131,3.021778,
        1.675516,2.019676,-1.980944,
        1.675516,2.019912,2.686612,
        1.675516,2.020545,2.088672,
        1.675516,2.026074,0.300945,
        1.675516,2.026196,-2.889057,
        1.675516,2.031959,-1.268178,
        1.675516,2.032917,-2.199917,
        1.675516,2.033572,-1.842884,
        1.675516,2.034712,-2.419505,
        1.675516,2.035818,-0.379479,
        1.675516,2.036004,-0.757062,
        1.675516,2.040501,-1.411488,
        1.675516,2.042874,-1.701367,
        1.675516,2.043060,-2.653746,
        1.675516,2.043729,-1.557938,
        1.675516,2.048538,0.806623,
        1.675516,2.049006,1.199930,
        1.675516,2.049502,-0.137989,
        1.675516,2.052181,2.326959,
        1.675516,2.052794,-0.008526,
        1.675516,2.057608,1.951726,
        1.675516,2.065568,2.567730,
        1.675516,2.070371,2.879808,
        1.675516,2.071256,-3.018961,
        1.675516,2.079746,-0.897699,
        1.675516,2.082784,1.042403,
        1.675516,2.083752,-0.511086,
        1.675516,2.084401,1.810502,
        1.675516,2.085731,0.671081,
        1.675516,2.097496,1.373172,
        1.675516,2.099185,1.667006,
        1.675516,2.100482,-2.301084,
        1.675516,2.103118,2.194745,
        1.675516,2.103223,1.521685,
        1.675516,2.103695,-2.777707,
        1.675516,2.107523,3.129581,
        1.675516,2.108277,-2.095907,
        1.675516,2.108891,-2.532166,
        1.675516,2.111003,-0.268632,
        1.675516,2.112276,-1.027246,
        1.675516,2.114380,0.531457,
        1.675516,2.119166,0.100552,
        1.675516,2.119564,2.754818,
        1.675516,2.122474,-0.648314,
        1.675516,2.125183,2.445598,
        1.675516,2.129501,0.238290,
        1.675516,2.131178,0.386663,
        1.675516,2.131223,-1.174311,
        1.675516,2.134269,-1.948042,
        1.675516,2.134684,0.907065,
        1.675516,2.137615,2.988806,
        1.675516,2.145635,-1.327708,
        1.675516,2.146467,2.055081,
        1.675516,2.146582,-1.793788,
        1.675516,2.149210,-1.482321,
        1.675516,2.152901,-1.636788,
        1.675516,2.156337,1.151558,
        1.675516,2.156474,-2.912065,
        1.675516,2.157194,-0.792563,
        1.675516,2.166724,-0.401638,
        1.675516,2.166980,-0.134879,
        1.675516,2.173644,-2.403429,
        1.675516,2.175711,-2.655523,
        1.675516,2.177132,2.600419,
        1.675516,2.178879,0.767722,
        1.675516,2.180752,1.908485,
        1.675516,2.181338,2.313721,
        1.675516,2.185580,1.287673,
        1.675516,2.195400,-2.217546,
        1.675516,2.200418,-3.056442,
        1.675516,2.201829,-0.001735,
        1.675516,2.203804,1.756176,
        1.675516,2.211602,1.443930,
        1.675516,2.212686,-0.544809,
        1.675516,2.213937,0.621892,
        1.675516,2.214075,1.600691,
        1.675516,2.216256,1.020029,
        1.675516,2.219060,2.886631,
        1.675516,2.223375,-0.906185,
        1.675516,2.226002,-1.060248,
        1.675516,2.226413,-2.064034,
        1.675516,2.230554,2.171334,
        1.675516,2.232413,2.735777,
        1.675516,2.233709,3.070830,
        1.675516,2.234383,-2.792065,
        1.675516,2.237183,0.468799,
        1.675516,2.237695,-0.271388,
        1.675516,2.239419,0.151693,
        1.675516,2.243347,-2.521049,
        1.675516,2.244035,-1.229456,
        1.675516,2.244355,0.309899,
        1.675516,2.245175,2.460236,
        1.675516,2.245628,-1.899216,
        1.675516,2.245973,-0.702234,
        1.675516,2.255601,-1.728131,
        1.675516,2.256702,-1.401367,
        1.675516,2.261466,-1.557500,
        1.675516,2.267390,0.878499,
        1.675516,2.272099,2.018594,
        1.675516,2.278618,1.170386,
        1.675516,2.283304,-2.345584,
        1.675516,2.285175,-2.943090,
        1.675516,2.293000,-0.124451,
        1.675516,2.294058,-0.421070,
        1.675516,2.303283,1.855938,
        1.675516,2.304968,2.608978,
        1.675516,2.305347,2.311346,
        1.675516,2.307188,-2.656134,
        1.675516,2.307301,1.336200,
        1.675516,2.309655,0.726020,
        1.675516,2.317634,-2.185225,
        1.675516,2.321510,1.685664,
        1.675516,2.324376,-3.111006,
        1.675516,2.324731,1.511186,
        1.675516,2.329587,0.035498,
        1.675516,2.330158,2.977264,
        1.675516,2.335096,-0.588468,
        1.675516,2.336046,-1.115799,
        1.675516,2.338060,2.803613,
        1.675516,2.338083,-0.943400,
        1.675516,2.340210,0.562145,
        1.675516,2.341931,1.023926,
        1.675516,2.343265,-0.779061,
        1.675516,2.343434,-2.013692,
        1.675516,2.353029,-1.310529,
        1.675516,2.354868,0.211543,
        1.675516,2.355120,0.388593,
        1.675516,2.355165,-1.833173,
        1.675516,2.355462,2.151441,
        1.675516,2.362948,-2.487477,
        1.675516,2.363101,-2.810456,
        1.675516,2.363394,-0.272918,
        1.675516,2.364509,-1.644533,
        1.675516,2.370465,2.466824,
        1.675516,2.386110,-1.481351,
        1.675516,2.394960,0.862133,
        1.675516,2.395716,1.977184,
        1.675516,2.396339,1.200900,
        1.675516,2.407941,-2.316753,
        1.675516,2.409450,-2.988432,
        1.675516,2.417202,-0.103363,
        1.675516,2.417386,-0.443097,
        1.675516,2.417950,2.666662,
        1.675516,2.423190,1.789783,
        1.675516,2.423692,1.392696,
        1.675516,2.431122,3.086379,
        1.675516,2.431406,-2.647820,
        1.675516,2.431584,2.307631,
        1.675516,2.434090,1.592818,
        1.675516,2.435993,0.683844,
        1.675516,2.440898,-2.133300,
        1.675516,2.442222,-1.195837,
        1.675516,2.443112,2.883761,
        1.675516,2.449191,-0.633651,
        1.675516,2.450821,0.084741,
        1.675516,2.453499,-1.008861,
        1.675516,2.460010,-1.938011,
        1.675516,2.460199,-0.822008,
        1.675516,2.460902,0.490079,
        1.675516,2.461774,-1.734913,
        1.675516,2.463986,1.034407,
        1.675516,2.467647,0.287667,
        1.675516,2.479849,2.126925,
        1.675516,2.485907,-2.835471,
        1.675516,2.485945,-1.374853,
        1.675516,2.487747,2.505912,
        1.675516,2.487957,-0.274279,
        1.675516,2.489741,-2.468428,
        1.675516,2.512538,1.239900,
        1.675516,2.517528,1.923289,
        1.675516,2.518454,-1.565852,
        1.675516,2.519348,0.843503,
        1.675516,2.522766,-3.059231,
        1.675516,2.528686,2.737228,
        1.675516,2.534798,-2.266671,
        1.675516,2.536533,1.466274,
        1.675516,2.538504,1.699812,
        1.675516,2.539516,-0.476942,
        1.675516,2.539532,-0.074548,
        1.675516,2.543197,2.990252,
        1.675516,2.554022,2.320619,
        1.675516,2.556919,-2.655005,
        1.675516,2.558146,0.627191,
        1.675516,2.558223,-1.164189,
        1.675516,2.563452,-2.044028,
        1.675516,2.568462,0.150655,
        1.675516,2.569130,-0.700837,
        1.675516,2.570050,-1.804492,
        1.675516,2.572710,-0.934665,
        1.675516,2.575994,0.391718,
        1.675516,2.584774,1.046703,
        1.675516,2.601432,2.102535,
        1.675516,2.603559,2.555114,
        1.675516,2.604150,-2.885208,
        1.675516,2.608630,-1.378011,
        1.675516,2.611256,-0.279110,
        1.675516,2.617344,-2.439480,
        1.675516,2.627402,1.293552,
        1.675516,2.635396,1.848850,
        1.675516,2.635686,2.834950,
        1.675516,2.637087,3.132730,
        1.675516,2.642029,0.813274,
        1.675516,2.645260,1.570203,
        1.675516,2.646182,-1.614036,
        1.675516,2.660113,-2.186513,
        1.675516,2.660483,-0.034608,
        1.675516,2.661713,-0.529511,
        1.675516,2.672720,-1.110422,
        1.675516,2.676626,2.335353,
        1.675516,2.676638,0.540364,
        1.675516,2.676780,-1.901158,
        1.675516,2.681411,-2.667965,
        1.675516,2.682152,0.245751,
        1.675516,2.682859,-0.817071,
        1.675516,2.705107,1.062142,
        1.675516,2.717705,2.625995,
        1.675516,2.721001,2.060097,
        1.675516,2.722319,-2.954069,
        1.675516,2.729840,-1.368237,
        1.675516,2.734349,-0.289298,
        1.675516,2.738409,2.975671,
        1.675516,2.740366,1.374341,
        1.675516,2.744633,-2.391311,
        1.675516,2.747398,1.727166,
        1.675516,2.763799,0.760372,
        1.675516,2.768238,-1.679243,
        1.675516,2.779245,-2.046268,
        1.675516,2.780208,-0.626453,
        1.675516,2.780471,0.027732,
        1.675516,2.784696,-1.013128,
        1.675516,2.789817,0.395431,
        1.675516,2.801154,2.353143,
        1.675516,2.807091,-2.687477,
        1.675516,2.825762,1.085035,
        1.675516,2.830209,2.734943,
        1.675516,2.833517,-3.085508,
        1.675516,2.838301,1.980061,
        1.675516,2.849871,1.512012,
        1.675516,2.850089,-1.343899,
        1.675516,2.856905,-0.318441,
        1.675516,2.870445,-2.291001,
        1.675516,2.885017,0.655562,
        1.675516,2.886428,-1.793217,
        1.675516,2.890337,-0.825633,
        1.675516,2.898932,0.139973,
        1.675516,2.931927,-2.728768,
        1.675516,2.932096,2.377790,
        1.675516,2.937409,2.947245,
        1.675516,2.948048,1.133127,
        1.675516,2.950896,1.797889,
        1.675516,2.970019,-1.282382,
        1.675516,2.978333,-0.412090,
        1.675516,2.996437,-2.089121,
        1.675516,3.011274,0.434110,
        1.675516,3.050410,3.138492,
        1.675516,3.064492,1.647259,
        1.675516,3.088868,-0.989501,
        1.780236,0.053884,-2.408394,
        1.780236,0.077645,1.453433,
        1.780236,0.088408,-0.311160,
        1.780236,0.142314,2.590387,
        1.780236,0.147206,-1.385164,
        1.780236,0.163506,0.573203,
        1.780236,0.165485,-2.875396,
        1.780236,0.182564,-2.125821,
        1.780236,0.190569,1.929141,
        1.780236,0.201102,1.249082,
        1.780236,0.201397,-0.716646,
        1.780236,0.210568,-0.069559,
        1.780236,0.251672,2.848900,
        1.780236,0.266262,-1.678068,
        1.780236,0.269108,-1.215563,
        1.780236,0.273777,-2.506756,
        1.780236,0.279854,0.814581,
        1.780236,0.281664,-2.969147,
        1.780236,0.283625,2.405603,
        1.780236,0.284058,0.357350,
        1.780236,0.294107,1.593880,
        1.780236,0.302491,-0.400410,
        1.780236,0.308059,2.021491,
        1.780236,0.312219,-2.072057,
        1.780236,0.321905,-0.829121,
        1.780236,0.329984,1.201621,
        1.780236,0.339714,-0.013712,
        1.780236,0.368784,2.974974,
        1.780236,0.376428,-1.495532,
        1.780236,0.378103,2.646436,
        1.780236,0.385088,0.590355,
        1.780236,0.386248,-2.690943,
        1.780236,0.389851,-2.340031,
        1.780236,0.391502,-1.171374,
        1.780236,0.398249,-1.812694,
        1.780236,0.400542,-2.995729,
        1.780236,0.404273,0.918163,
        1.780236,0.405617,1.773533,
        1.780236,0.409009,-0.570243,
        1.780236,0.410700,0.270433,
        1.780236,0.415893,1.464890,
        1.780236,0.420680,2.088164,
        1.780236,0.424766,-0.258729,
        1.780236,0.426507,2.367872,
        1.780236,0.441292,-0.890727,
        1.780236,0.451839,-2.078193,
        1.780236,0.459270,1.191562,
        1.780236,0.470622,0.014002,
        1.780236,0.481292,-2.529098,
        1.780236,0.482460,2.849843,
        1.780236,0.492205,-1.397963,
        1.780236,0.493314,3.117264,
        1.780236,0.497870,0.727828,
        1.780236,0.499783,-1.654379,
        1.780236,0.502651,0.463333,
        1.780236,0.505463,2.594557,
        1.780236,0.506104,-2.874910,
        1.780236,0.513579,-1.147354,
        1.780236,0.518569,1.644237,
        1.780236,0.519001,1.900488,
        1.780236,0.519277,-0.673809,
        1.780236,0.524769,-0.421955,
        1.780236,0.525564,-2.288838,
        1.780236,0.532541,0.978737,
        1.780236,0.534610,2.154808,
        1.780236,0.535020,-1.891846,
        1.780236,0.540208,1.400497,
        1.780236,0.542006,0.228337,
        1.780236,0.550762,-0.180771,
        1.780236,0.560531,2.378474,
        1.780236,0.561629,-0.924576,
        1.780236,0.567465,-2.679392,
        1.780236,0.588527,1.181755,
        1.780236,0.591643,-2.098364,
        1.780236,0.594963,2.997575,
        1.780236,0.599643,-3.045235,
        1.780236,0.599962,2.773556,
        1.780236,0.602620,0.029270,
        1.780236,0.605711,0.598454,
        1.780236,0.607708,-1.541026,
        1.780236,0.609360,-2.464379,
        1.780236,0.610738,-1.326786,
        1.780236,0.616968,0.820489,
        1.780236,0.626379,1.773610,
        1.780236,0.628626,-1.751180,
        1.780236,0.630671,0.391949,
        1.780236,0.632622,-0.737642,
        1.780236,0.632708,2.566722,
        1.780236,0.632797,1.989858,
        1.780236,0.634455,-0.534790,
        1.780236,0.636205,1.561170,
        1.780236,0.637681,-1.123423,
        1.780236,0.641441,-2.842109,
        1.780236,0.644807,-0.327139,
        1.780236,0.652440,2.197129,
        1.780236,0.660897,-2.278593,
        1.780236,0.664008,1.360091,
        1.780236,0.670850,-1.945257,
        1.780236,0.671396,1.004859,
        1.780236,0.673985,0.203855,
        1.780236,0.680326,-0.133370,
        1.780236,0.685083,-0.942363,
        1.780236,0.688204,-2.632204,
        1.780236,0.689777,2.385011,
        1.780236,0.700065,3.115430,
        1.780236,0.704080,2.911636,
        1.780236,0.712883,1.174216,
        1.780236,0.715701,0.699906,
        1.780236,0.718851,-1.442215,
        1.780236,0.721019,2.721536,
        1.780236,0.725300,-2.975134,
        1.780236,0.727136,-1.633288,
        1.780236,0.727399,-2.120567,
        1.780236,0.730751,0.522488,
        1.780236,0.732892,-1.254330,
        1.780236,0.734333,-2.446184,
        1.780236,0.735125,0.037270,
        1.780236,0.737893,1.866463,
        1.780236,0.740650,1.681075,
        1.780236,0.745971,-0.781286,
        1.780236,0.749432,-0.442559,
        1.780236,0.750102,2.050636,
        1.780236,0.751046,0.863821,
        1.780236,0.757018,-0.608833,
        1.780236,0.757046,-1.816019,
        1.780236,0.757389,1.501184,
        1.780236,0.759800,2.546181,
        1.780236,0.759859,0.351923,
        1.780236,0.761141,-2.781102,
        1.780236,0.769035,-1.078574,
        1.780236,0.770166,-0.266604,
        1.780236,0.772218,2.227380,
        1.780236,0.788382,1.331051,
        1.780236,0.790269,-2.282409,
        1.780236,0.803410,-1.984567,
        1.780236,0.803813,0.191520,
        1.780236,0.806832,3.022155,
        1.780236,0.807476,-2.597675,
        1.780236,0.812512,-0.107335,
        1.780236,0.814161,1.010949,
        1.780236,0.817732,-3.084486,
        1.780236,0.817846,2.390688,
        1.780236,0.819323,2.845960,
        1.780236,0.819430,-0.915017,
        1.780236,0.829198,-1.532491,
        1.780236,0.829531,0.665695,
        1.780236,0.829816,1.167147,
        1.780236,0.833613,-1.357492,
        1.780236,0.842945,-2.907318,
        1.780236,0.844883,2.679708,
        1.780236,0.848189,-1.705003,
        1.780236,0.849888,1.767901,
        1.780236,0.851126,0.493370,
        1.780236,0.852911,1.933394,
        1.780236,0.857516,-2.432863,
        1.780236,0.858261,-1.191729,
        1.780236,0.858279,-2.138895,
        1.780236,0.859723,-0.734928,
        1.780236,0.859754,1.603292,
        1.780236,0.861655,-0.394821,
        1.780236,0.864300,0.043549,
        1.780236,0.866940,0.851229,
        1.780236,0.869748,2.095969,
        1.780236,0.877352,-0.574668,
        1.780236,0.879254,-2.733338,
        1.780236,0.882554,1.443757,
        1.780236,0.884060,0.334199,
        1.780236,0.885626,-1.866061,
        1.780236,0.889834,2.529957,
        1.780236,0.893140,2.251698,
        1.780236,0.897463,-0.240822,
        1.780236,0.898758,-1.031948,
        1.780236,0.913918,1.287618,
        1.780236,0.916555,-2.287173,
        1.780236,0.918305,3.107845,
        1.780236,0.919522,2.948277,
        1.780236,0.927138,-2.569686,
        1.780236,0.929031,0.185383,
        1.780236,0.933766,-3.015917,
        1.780236,0.934873,-2.010822,
        1.780236,0.935636,-1.445855,
        1.780236,0.936702,-0.866461,
        1.780236,0.937395,0.729108,
        1.780236,0.937910,2.787575,
        1.780236,0.939144,-0.092962,
        1.780236,0.941718,1.125085,
        1.780236,0.942873,-1.607680,
        1.780236,0.943701,0.980328,
        1.780236,0.947774,2.395256,
        1.780236,0.952433,0.584347,
        1.780236,0.954456,-1.287293,
        1.780236,0.958731,-2.852552,
        1.780236,0.961838,1.839284,
        1.780236,0.961898,-0.469425,
        1.780236,0.965426,1.686615,
        1.780236,0.970108,1.990771,
        1.780236,0.972820,-1.762780,
        1.780236,0.973500,-0.711881,
        1.780236,0.975253,2.641082,
        1.780236,0.977824,0.441928,
        1.780236,0.981485,-2.424766,
        1.780236,0.981877,1.537100,
        1.780236,0.982609,-1.136920,
        1.780236,0.984265,-2.156073,
        1.780236,0.989150,0.045493,
        1.780236,0.993052,2.137478,
        1.780236,0.997766,-2.693871,
        1.780236,0.999847,-0.334738,
        1.780236,1.001920,0.848054,
        1.780236,1.006146,1.389734,
        1.780236,1.012180,0.304066,
        1.780236,1.016938,2.270674,
        1.780236,1.017371,-0.979790,
        1.780236,1.019495,-1.894487,
        1.780236,1.023796,-0.583691,
        1.780236,1.026525,2.506945,
        1.780236,1.027004,-0.203653,
        1.780236,1.028027,3.036573,
        1.780236,1.031435,1.241021,
        1.780236,1.032316,2.884221,
        1.780236,1.035863,-3.110475,
        1.780236,1.042085,-2.292950,
        1.780236,1.043272,-1.522986,
        1.780236,1.046265,-2.958847,
        1.780236,1.049272,-1.376344,
        1.780236,1.050039,-2.552086,
        1.780236,1.053682,0.689863,
        1.780236,1.055458,0.170394,
        1.780236,1.056622,-0.831866,
        1.780236,1.057515,1.097998,
        1.780236,1.058706,-2.036438,
        1.780236,1.061389,-1.672248,
        1.780236,1.063758,2.743093,
        1.780236,1.070157,-0.073422,
        1.780236,1.072154,1.761506,
        1.780236,1.074029,1.909193,
        1.780236,1.074204,0.545276,
        1.780236,1.074714,-2.805147,
        1.780236,1.076262,0.966104,
        1.780236,1.077642,-1.227073,
        1.780236,1.084151,1.618525,
        1.780236,1.085085,2.377673,
        1.780236,1.086822,-0.441112,
        1.780236,1.087142,2.058815,
        1.780236,1.098383,-0.696430,
        1.780236,1.099138,-1.084033,
        1.780236,1.103208,1.477773,
        1.780236,1.103783,0.408254,
        1.780236,1.106275,-2.420800,
        1.780236,1.108079,2.612943,
        1.780236,1.109345,-2.172718,
        1.780236,1.115957,-1.787597,
        1.780236,1.118930,0.045983,
        1.780236,1.119777,0.834702,
        1.780236,1.121943,-2.669042,
        1.780236,1.124930,2.219917,
        1.780236,1.125217,1.336294,
        1.780236,1.130958,2.972931,
        1.780236,1.131400,-0.306629,
        1.780236,1.137135,-0.936514,
        1.780236,1.139125,-1.924528,
        1.780236,1.139877,0.276382,
        1.780236,1.142003,-3.058022,
        1.780236,1.147363,-0.558448,
        1.780236,1.150348,-0.176418,
        1.780236,1.151198,1.197016,
        1.780236,1.153930,-1.597104,
        1.780236,1.155389,2.837148,
        1.780236,1.155458,-1.460623,
        1.780236,1.157331,-2.906362,
        1.780236,1.158594,2.487618,
        1.780236,1.162103,-1.322553,
        1.780236,1.165407,3.092861,
        1.780236,1.168025,0.717075,
        1.780236,1.168406,-2.301519,
        1.780236,1.174782,-2.538755,
        1.780236,1.177998,1.061256,
        1.780236,1.178898,1.827305,
        1.780236,1.181331,-2.060330,
        1.780236,1.182540,-0.798952,
        1.780236,1.184569,0.147517,
        1.780236,1.184734,1.968919,
        1.780236,1.184804,0.593386,
        1.780236,1.188278,1.690808,
        1.780236,1.194360,2.110440,
        1.780236,1.194776,2.710455,
        1.780236,1.197745,-1.173705,
        1.780236,1.199898,-2.777257,
        1.780236,1.201164,2.355931,
        1.780236,1.203760,1.556404,
        1.780236,1.205845,-0.415993,
        1.780236,1.206585,0.933648,
        1.780236,1.206950,-0.037248,
        1.780236,1.209341,0.471162,
        1.780236,1.213483,-1.039417,
        1.780236,1.218146,-0.672643,
        1.780236,1.221608,1.421669,
        1.780236,1.226043,-1.693533,
        1.780236,1.232870,-1.822773,
        1.780236,1.234375,-2.421194,
        1.780236,1.234531,-2.190148,
        1.780236,1.239414,0.348959,
        1.780236,1.240183,2.587255,
        1.780236,1.242312,2.941234,
        1.780236,1.244075,1.286644,
        1.780236,1.245961,-2.997461,
        1.780236,1.246495,-2.646872,
        1.780236,1.247736,2.242810,
        1.780236,1.253604,0.812222,
        1.780236,1.257816,-0.894317,
        1.780236,1.258247,-1.953634,
        1.780236,1.258993,-0.284099,
        1.780236,1.262380,-0.157609,
        1.780236,1.262791,-1.549229,
        1.780236,1.263772,-0.533073,
        1.780236,1.266731,-1.412461,
        1.780236,1.268591,-3.122146,
        1.780236,1.269614,-1.275246,
        1.780236,1.270745,1.152533,
        1.780236,1.273606,0.224545,
        1.780236,1.279716,2.812101,
        1.780236,1.283786,2.462124,
        1.780236,1.284798,0.078209,
        1.780236,1.285402,-2.882049,
        1.780236,1.286936,1.886053,
        1.780236,1.289408,0.683122,
        1.780236,1.294736,1.756184,
        1.780236,1.296978,-2.311511,
        1.780236,1.298546,2.020601,
        1.780236,1.300924,1.022478,
        1.780236,1.303305,3.043152,
        1.780236,1.304210,-2.084225,
        1.780236,1.305241,-2.530249,
        1.780236,1.307513,1.628003,
        1.780236,1.310722,0.558267,
        1.780236,1.314142,2.139329,
        1.780236,1.316087,-0.777971,
        1.780236,1.317907,-1.121023,
        1.780236,1.321019,-2.749776,
        1.780236,1.321609,2.687717,
        1.780236,1.322568,1.499415,
        1.780236,1.322972,-0.395699,
        1.780236,1.325286,-0.994192,
        1.780236,1.329146,-0.648513,
        1.780236,1.329795,-0.042570,
        1.780236,1.331651,2.344497,
        1.780236,1.336522,0.435098,
        1.780236,1.339410,0.900087,
        1.780236,1.339550,1.370353,
        1.780236,1.340441,-1.731580,
        1.780236,1.351855,-1.855234,
        1.780236,1.360161,-3.018404,
        1.780236,1.361031,-1.613971,
        1.780236,1.361196,2.914780,
        1.780236,1.361998,-2.207254,
        1.780236,1.362302,1.239384,
        1.780236,1.365228,2.563808,
        1.780236,1.365452,0.310522,
        1.780236,1.367028,-2.420041,
        1.780236,1.373422,-1.484281,
        1.780236,1.374245,-1.226417,
        1.780236,1.375020,-0.170393,
        1.780236,1.376560,-1.355481,
        1.780236,1.377070,-1.980984,
        1.780236,1.377578,-2.636339,
        1.780236,1.378096,0.774626,
        1.780236,1.379086,-0.508192,
        1.780236,1.382562,0.177141,
        1.780236,1.388849,2.233847,
        1.780236,1.389268,-0.886349,
        1.780236,1.391324,1.109396,
        1.780236,1.391699,-0.288725,
        1.780236,1.392869,3.136903,
        1.780236,1.393612,1.952016,
        1.780236,1.395545,-2.850807,
        1.780236,1.401926,1.822930,
        1.780236,1.402149,2.788898,
        1.780236,1.409552,0.647980,
        1.780236,1.410840,0.056219,
        1.780236,1.411944,2.444377,
        1.780236,1.413639,1.697982,
        1.780236,1.426479,1.573427,
        1.780236,1.426725,2.102242,
        1.780236,1.427868,0.984637,
        1.780236,1.429580,-2.107232,
        1.780236,1.431375,-2.313269,
        1.780236,1.431907,3.015410,
        1.780236,1.432325,-1.029054,
        1.780236,1.433341,0.524268,
        1.780236,1.435181,-0.739771,
        1.780236,1.437208,-2.529085,
        1.780236,1.439041,-0.617966,
        1.780236,1.441473,1.448544,
        1.780236,1.444527,2.664744,
        1.780236,1.450513,-0.400625,
        1.780236,1.451907,-0.066823,
        1.780236,1.452754,-2.738631,
        1.780236,1.454061,-2.950416,
        1.780236,1.456736,1.323154,
        1.780236,1.458370,0.400433,
        1.780236,1.459314,-1.765003,
        1.780236,1.459902,-1.145835,
        1.780236,1.464952,2.329245,
        1.780236,1.466050,0.862104,
        1.780236,1.471330,-1.886160,
        1.780236,1.475119,-1.648255,
        1.780236,1.480878,1.194576,
        1.780236,1.480942,0.273235,
        1.780236,1.481086,2.890269,
        1.780236,1.484118,-1.529156,
        1.780236,1.484741,-1.279219,
        1.780236,1.487681,-1.406069,
        1.780236,1.487886,-3.066037,
        1.780236,1.490555,2.543284,
        1.780236,1.492424,-0.190932,
        1.780236,1.495002,-0.839810,
        1.780236,1.495297,2.007188,
        1.780236,1.496414,-2.009267,
        1.780236,1.498261,-2.425270,
        1.780236,1.498907,-2.207327,
        1.780236,1.500204,0.737638,
        1.780236,1.502469,0.151156,
        1.780236,1.507448,-0.514782,
        1.780236,1.509287,2.204647,
        1.780236,1.510518,1.884920,
        1.780236,1.511948,-2.632360,
        1.780236,1.515235,1.068565,
        1.780236,1.522237,3.102018,
        1.780236,1.522373,1.764991,
        1.780236,1.522416,-0.954956,
        1.780236,1.523102,2.765920,
        1.780236,1.524811,-2.844718,
        1.780236,1.527261,-0.304916,
        1.780236,1.529788,0.613208,
        1.780236,1.533648,1.644808,
        1.780236,1.535018,0.028869,
        1.780236,1.541082,2.424608,
        1.780236,1.546010,1.523462,
        1.780236,1.551135,-0.642994,
        1.780236,1.553895,0.489607,
        1.780236,1.555041,0.947813,
        1.780236,1.559671,-1.074934,
        1.780236,1.560345,2.985899,
        1.780236,1.560487,-2.323400,
        1.780236,1.561127,1.401111,
        1.780236,1.567193,2.643172,
        1.780236,1.568892,2.098362,
        1.780236,1.571702,-2.098435,
        1.780236,1.572701,-2.528730,
        1.780236,1.573572,1.278593,
        1.780236,1.574849,-0.091505,
        1.780236,1.576334,0.365236,
        1.780236,1.577029,-2.962888,
        1.780236,1.577688,-1.814941,
        1.780236,1.579542,-0.755917,
        1.780236,1.579624,-1.199759,
        1.780236,1.581515,-0.419414,
        1.780236,1.584581,-2.738702,
        1.780236,1.586948,-1.696296,
        1.780236,1.588901,2.301496,
        1.780236,1.589936,0.825020,
        1.780236,1.590018,-1.941652,
        1.780236,1.596968,-1.328358,
        1.780236,1.597178,-1.575607,
        1.780236,1.600470,-1.453196,
        1.780236,1.601408,0.242409,
        1.780236,1.602259,1.152656,
        1.780236,1.603064,2.865774,
        1.780236,1.612085,1.968802,
        1.780236,1.612720,-3.080498,
        1.780236,1.616061,2.521531,
        1.780236,1.616900,-0.207457,
        1.780236,1.620988,-0.875920,
        1.780236,1.622114,0.701509,
        1.780236,1.622749,0.121426,
        1.780236,1.622993,-2.221629,
        1.780236,1.629346,-0.542669,
        1.780236,1.630007,1.844552,
        1.780236,1.634747,-2.426198,
        1.780236,1.642346,1.722617,
        1.780236,1.643925,1.034629,
        1.780236,1.644848,2.743774,
        1.780236,1.645258,2.188109,
        1.780236,1.645417,-2.634725,
        1.780236,1.647395,3.085437,
        1.780236,1.648538,-2.855044,
        1.780236,1.649932,0.576808,
        1.780236,1.652224,-0.996105,
        1.780236,1.652871,1.600681,
        1.780236,1.657997,-0.325127,
        1.780236,1.661936,0.001037,
        1.780236,1.665187,1.477402,
        1.780236,1.665496,2.397872,
        1.780236,1.673267,-0.665453,
        1.780236,1.673557,0.452526,
        1.780236,1.679277,-1.122421,
        1.780236,1.679696,0.912482,
        1.780236,1.682420,1.353922,
        1.780236,1.685031,-2.009038,
        1.780236,1.685196,2.967046,
        1.780236,1.690107,2.620910,
        1.780236,1.690453,-2.123076,
        1.780236,1.691693,-1.871572,
        1.780236,1.691817,2.061688,
        1.780236,1.692644,1.237670,
        1.780236,1.696386,0.328123,
        1.780236,1.697162,-1.249026,
        1.780236,1.697488,-2.322785,
        1.780236,1.698726,-2.973576,
        1.780236,1.699172,-1.749717,
        1.780236,1.704478,-0.106295,
        1.780236,1.707244,-0.444770,
        1.780236,1.707500,-2.531497,
        1.780236,1.707714,-0.787321,
        1.780236,1.708430,-1.627552,
        1.780236,1.710583,-1.376890,
        1.780236,1.712415,-2.747921,
        1.780236,1.713125,-1.503073,
        1.780236,1.713452,0.789419,
        1.780236,1.720612,2.279888,
        1.780236,1.724852,0.202831,
        1.780236,1.724868,2.846246,
        1.780236,1.728940,1.931535,
        1.780236,1.732838,1.127584,
        1.780236,1.735719,-3.092485,
        1.780236,1.739676,2.496017,
        1.780236,1.739971,-0.228876,
        1.780236,1.743328,0.663347,
        1.780236,1.744927,-0.911892,
        1.780236,1.745952,0.084733,
        1.780236,1.748077,1.804354,
        1.780236,1.752959,-0.567548,
        1.780236,1.760169,1.678132,
        1.780236,1.763887,-2.218450,
        1.780236,1.766737,2.722752,
        1.780236,1.767999,1.004737,
        1.780236,1.769149,1.550298,
        1.780236,1.770119,-2.427175,
        1.780236,1.770460,2.154208,
        1.780236,1.770514,3.070982,
        1.780236,1.770735,0.537245,
        1.780236,1.772673,-2.864760,
        1.780236,1.773527,-1.039582,
        1.780236,1.776419,-2.642421,
        1.780236,1.780769,1.416330,
        1.780236,1.786634,-0.346247,
        1.780236,1.788829,-1.947553,
        1.780236,1.793737,0.410270,
        1.780236,1.794553,2.374618,
        1.780236,1.794699,-0.694061,
        1.780236,1.797469,-1.169105,
        1.780236,1.803705,0.881076,
        1.780236,1.804124,1.264249,
        1.780236,1.807825,-2.090455,
        1.780236,1.808822,-1.817375,
        1.780236,1.809020,2.949741,
        1.780236,1.812466,2.597595,
        1.780236,1.813159,2.025395,
        1.780236,1.813816,-1.298435,
        1.780236,1.815016,0.278951,
        1.780236,1.816510,-0.008631,
        1.780236,1.817772,-1.689625,
        1.780236,1.821360,-2.984352,
        1.780236,1.822312,-0.127211,
        1.780236,1.823386,-1.429807,
        1.780236,1.823834,-1.560483,
        1.780236,1.832060,-0.821627,
        1.780236,1.832690,-0.468440,
        1.780236,1.833945,-2.320278,
        1.780236,1.834787,0.751968,
        1.780236,1.840579,-2.536445,
        1.780236,1.841033,-2.756932,
        1.780236,1.846350,1.893983,
        1.780236,1.846644,1.113766,
        1.780236,1.846833,2.826543,
        1.780236,1.847325,2.249051,
        1.780236,1.854119,0.134161,
        1.780236,1.858863,-3.106022,
        1.780236,1.865046,1.763384,
        1.780236,1.865619,-0.952036,
        1.780236,1.865912,0.624453,
        1.780236,1.866694,2.473686,
        1.780236,1.867524,-0.240083,
        1.780236,1.875561,1.632831,
        1.780236,1.876049,-0.594378,
        1.780236,1.880509,1.500204,
        1.780236,1.883504,1.360457,
        1.780236,1.886971,-2.196665,
        1.780236,1.888979,0.982922,
        1.780236,1.889815,2.701570,
        1.780236,1.892144,-2.005340,
        1.780236,1.892348,0.496522,
        1.780236,1.893281,-1.083976,
        1.780236,1.893828,3.053453,
        1.780236,1.895117,2.121107,
        1.780236,1.898741,-2.875735,
        1.780236,1.904429,-2.428029,
        1.780236,1.908408,-2.648903,
        1.780236,1.911828,-0.364489,
        1.780236,1.912224,0.366565,
        1.780236,1.914854,1.214468,
        1.780236,1.915467,-1.883143,
        1.780236,1.915763,-0.724679,
        1.780236,1.915790,-1.216366,
        1.780236,1.922256,2.347806,
        1.780236,1.922616,0.844633,
        1.780236,1.925185,0.231170,
        1.780236,1.927474,-1.754616,
        1.780236,1.928342,-1.351174,
        1.780236,1.933328,-1.623328,
        1.780236,1.933878,0.034150,
        1.780236,1.934585,-1.487979,
        1.780236,1.934802,2.927634,
        1.780236,1.934862,1.989776,
        1.780236,1.936327,-0.099350,
        1.780236,1.938088,2.578507,
        1.780236,1.945677,-2.999805,
        1.780236,1.954572,-0.859208,
        1.780236,1.957868,-0.490784,
        1.780236,1.958989,0.712772,
        1.780236,1.960212,-2.307561,
        1.780236,1.964774,1.855507,
        1.780236,1.964971,-2.100533,
        1.780236,1.969112,1.088273,
        1.780236,1.971438,-2.765415,
        1.780236,1.972312,2.805134,
        1.780236,1.974950,2.220596,
        1.780236,1.975135,-2.539127,
        1.780236,1.982107,1.720829,
        1.780236,1.983195,-3.127984,
        1.780236,1.983318,-0.247182,
        1.780236,1.986720,-0.993220,
        1.780236,1.990381,1.586005,
        1.780236,1.990470,0.580283,
        1.780236,1.990523,1.450326,
        1.780236,1.993909,1.312897,
        1.780236,1.994694,2.452813,
        1.780236,1.999226,-0.620717,
        1.780236,2.005145,0.941012,
        1.780236,2.007301,0.142337,
        1.780236,2.014149,0.444138,
        1.780236,2.014338,-1.127392,
        1.780236,2.018131,3.021778,
        1.780236,2.019676,-1.980944,
        1.780236,2.019912,2.686612,
        1.780236,2.020545,2.088672,
        1.780236,2.026074,0.300945,
        1.780236,2.026196,-2.889057,
        1.780236,2.031959,-1.268178,
        1.780236,2.032917,-2.199917,
        1.780236,2.033572,-1.842884,
        1.780236,2.034712,-2.419505,
        1.780236,2.035818,-0.379479,
        1.780236,2.036004,-0.757062,
        1.780236,2.040501,-1.411488,
        1.780236,2.042874,-1.701367,
        1.780236,2.043060,-2.653746,
        1.780236,2.043729,-1.557938,
        1.780236,2.048538,0.806623,
        1.780236,2.049006,1.199930,
        1.780236,2.049502,-0.137989,
        1.780236,2.052181,2.326959,
        1.780236,2.052794,-0.008526,
        1.780236,2.057608,1.951726,
        1.780236,2.065568,2.567730,
        1.780236,2.070371,2.879808,
        1.780236,2.071256,-3.018961,
        1.780236,2.079746,-0.897699,
        1.780236,2.082784,1.042403,
        1.780236,2.083752,-0.511086,
        1.780236,2.084401,1.810502,
        1.780236,2.085731,0.671081,
        1.780236,2.097496,1.373172,
        1.780236,2.099185,1.667006,
        1.780236,2.100482,-2.301084,
        1.780236,2.103118,2.194745,
        1.780236,2.103223,1.521685,
        1.780236,2.103695,-2.777707,
        1.780236,2.107523,3.129581,
        1.780236,2.108277,-2.095907,
        1.780236,2.108891,-2.532166,
        1.780236,2.111003,-0.268632,
        1.780236,2.112276,-1.027246,
        1.780236,2.114380,0.531457,
        1.780236,2.119166,0.100552,
        1.780236,2.119564,2.754818,
        1.780236,2.122474,-0.648314,
        1.780236,2.125183,2.445598,
        1.780236,2.129501,0.238290,
        1.780236,2.131178,0.386663,
        1.780236,2.131223,-1.174311,
        1.780236,2.134269,-1.948042,
        1.780236,2.134684,0.907065,
        1.780236,2.137615,2.988806,
        1.780236,2.145635,-1.327708,
        1.780236,2.146467,2.055081,
        1.780236,2.146582,-1.793788,
        1.780236,2.149210,-1.482321,
        1.780236,2.152901,-1.636788,
        1.780236,2.156337,1.151558,
        1.780236,2.156474,-2.912065,
        1.780236,2.157194,-0.792563,
        1.780236,2.166724,-0.401638,
        1.780236,2.166980,-0.134879,
        1.780236,2.173644,-2.403429,
        1.780236,2.175711,-2.655523,
        1.780236,2.177132,2.600419,
        1.780236,2.178879,0.767722,
        1.780236,2.180752,1.908485,
        1.780236,2.181338,2.313721,
        1.780236,2.185580,1.287673,
        1.780236,2.195400,-2.217546,
        1.780236,2.200418,-3.056442,
        1.780236,2.201829,-0.001735,
        1.780236,2.203804,1.756176,
        1.780236,2.211602,1.443930,
        1.780236,2.212686,-0.544809,
        1.780236,2.213937,0.621892,
        1.780236,2.214075,1.600691,
        1.780236,2.216256,1.020029,
        1.780236,2.219060,2.886631,
        1.780236,2.223375,-0.906185,
        1.780236,2.226002,-1.060248,
        1.780236,2.226413,-2.064034,
        1.780236,2.230554,2.171334,
        1.780236,2.232413,2.735777,
        1.780236,2.233709,3.070830,
        1.780236,2.234383,-2.792065,
        1.780236,2.237183,0.468799,
        1.780236,2.237695,-0.271388,
        1.780236,2.239419,0.151693,
        1.780236,2.243347,-2.521049,
        1.780236,2.244035,-1.229456,
        1.780236,2.244355,0.309899,
        1.780236,2.245175,2.460236,
        1.780236,2.245628,-1.899216,
        1.780236,2.245973,-0.702234,
        1.780236,2.255601,-1.728131,
        1.780236,2.256702,-1.401367,
        1.780236,2.261466,-1.557500,
        1.780236,2.267390,0.878499,
        1.780236,2.272099,2.018594,
        1.780236,2.278618,1.170386,
        1.780236,2.283304,-2.345584,
        1.780236,2.285175,-2.943090,
        1.780236,2.293000,-0.124451,
        1.780236,2.294058,-0.421070,
        1.780236,2.303283,1.855938,
        1.780236,2.304968,2.608978,
        1.780236,2.305347,2.311346,
        1.780236,2.307188,-2.656134,
        1.780236,2.307301,1.336200,
        1.780236,2.309655,0.726020,
        1.780236,2.317634,-2.185225,
        1.780236,2.321510,1.685664,
        1.780236,2.324376,-3.111006,
        1.780236,2.324731,1.511186,
        1.780236,2.329587,0.035498,
        1.780236,2.330158,2.977264,
        1.780236,2.335096,-0.588468,
        1.780236,2.336046,-1.115799,
        1.780236,2.338060,2.803613,
        1.780236,2.338083,-0.943400,
        1.780236,2.340210,0.562145,
        1.780236,2.341931,1.023926,
        1.780236,2.343265,-0.779061,
        1.780236,2.343434,-2.013692,
        1.780236,2.353029,-1.310529,
        1.780236,2.354868,0.211543,
        1.780236,2.355120,0.388593,
        1.780236,2.355165,-1.833173,
        1.780236,2.355462,2.151441,
        1.780236,2.362948,-2.487477,
        1.780236,2.363101,-2.810456,
        1.780236,2.363394,-0.272918,
        1.780236,2.364509,-1.644533,
        1.780236,2.370465,2.466824,
        1.780236,2.386110,-1.481351,
        1.780236,2.394960,0.862133,
        1.780236,2.395716,1.977184,
        1.780236,2.396339,1.200900,
        1.780236,2.407941,-2.316753,
        1.780236,2.409450,-2.988432,
        1.780236,2.417202,-0.103363,
        1.780236,2.417386,-0.443097,
        1.780236,2.417950,2.666662,
        1.780236,2.423190,1.789783,
        1.780236,2.423692,1.392696,
        1.780236,2.431122,3.086379,
        1.780236,2.431406,-2.647820,
        1.780236,2.431584,2.307631,
        1.780236,2.434090,1.592818,
        1.780236,2.435993,0.683844,
        1.780236,2.440898,-2.133300,
        1.780236,2.442222,-1.195837,
        1.780236,2.443112,2.883761,
        1.780236,2.449191,-0.633651,
        1.780236,2.450821,0.084741,
        1.780236,2.453499,-1.008861,
        1.780236,2.460010,-1.938011,
        1.780236,2.460199,-0.822008,
        1.780236,2.460902,0.490079,
        1.780236,2.461774,-1.734913,
        1.780236,2.463986,1.034407,
        1.780236,2.467647,0.287667,
        1.780236,2.479849,2.126925,
        1.780236,2.485907,-2.835471,
        1.780236,2.485945,-1.374853,
        1.780236,2.487747,2.505912,
        1.780236,2.487957,-0.274279,
        1.780236,2.489741,-2.468428,
        1.780236,2.512538,1.239900,
        1.780236,2.517528,1.923289,
        1.780236,2.518454,-1.565852,
        1.780236,2.519348,0.843503,
        1.780236,2.522766,-3.059231,
        1.780236,2.528686,2.737228,
        1.780236,2.534798,-2.266671,
        1.780236,2.536533,1.466274,
        1.780236,2.538504,1.699812,
        1.780236,2.539516,-0.476942,
        1.780236,2.539532,-0.074548,
        1.780236,2.543197,2.990252,
        1.780236,2.554022,2.320619,
        1.780236,2.556919,-2.655005,
        1.780236,2.558146,0.627191,
        1.780236,2.558223,-1.164189,
        1.780236,2.563452,-2.044028,
        1.780236,2.568462,0.150655,
        1.780236,2.569130,-0.700837,
        1.780236,2.570050,-1.804492,
        1.780236,2.572710,-0.934665,
        1.780236,2.575994,0.391718,
        1.780236,2.584774,1.046703,
        1.780236,2.601432,2.102535,
        1.780236,2.603559,2.555114,
        1.780236,2.604150,-2.885208,
        1.780236,2.608630,-1.378011,
        1.780236,2.611256,-0.279110,
        1.780236,2.617344,-2.439480,
        1.780236,2.627402,1.293552,
        1.780236,2.635396,1.848850,
        1.780236,2.635686,2.834950,
        1.780236,2.637087,3.132730,
        1.780236,2.642029,0.813274,
        1.780236,2.645260,1.570203,
        1.780236,2.646182,-1.614036,
        1.780236,2.660113,-2.186513,
        1.780236,2.660483,-0.034608,
        1.780236,2.661713,-0.529511,
        1.780236,2.672720,-1.110422,
        1.780236,2.676626,2.335353,
        1.780236,2.676638,0.540364,
        1.780236,2.676780,-1.901158,
        1.780236,2.681411,-2.667965,
        1.780236,2.682152,0.245751,
        1.780236,2.682859,-0.817071,
        1.780236,2.705107,1.062142,
        1.780236,2.717705,2.625995,
        1.780236,2.721001,2.060097,
        1.780236,2.722319,-2.954069,
        1.780236,2.729840,-1.368237,
        1.780236,2.734349,-0.289298,
        1.780236,2.738409,2.975671,
        1.780236,2.740366,1.374341,
        1.780236,2.744633,-2.391311,
        1.780236,2.747398,1.727166,
        1.780236,2.763799,0.760372,
        1.780236,2.768238,-1.679243,
        1.780236,2.779245,-2.046268,
        1.780236,2.780208,-0.626453,
        1.780236,2.780471,0.027732,
        1.780236,2.784696,-1.013128,
        1.780236,2.789817,0.395431,
        1.780236,2.801154,2.353143,
        1.780236,2.807091,-2.687477,
        1.780236,2.825762,1.085035,
        1.780236,2.830209,2.734943,
        1.780236,2.833517,-3.085508,
        1.780236,2.838301,1.980061,
        1.780236,2.849871,1.512012,
        1.780236,2.850089,-1.343899,
        1.780236,2.856905,-0.318441,
        1.780236,2.870445,-2.291001,
        1.780236,2.885017,0.655562,
        1.780236,2.886428,-1.793217,
        1.780236,2.890337,-0.825633,
        1.780236,2.898932,0.139973,
        1.780236,2.931927,-2.728768,
        1.780236,2.932096,2.377790,
        1.780236,2.937409,2.947245,
        1.780236,2.948048,1.133127,
        1.780236,2.950896,1.797889,
        1.780236,2.970019,-1.282382,
        1.780236,2.978333,-0.412090,
        1.780236,2.996437,-2.089121,
        1.780236,3.011274,0.434110,
        1.780236,3.050410,3.138492,
        1.780236,3.064492,1.647259,
        1.780236,3.088868,-0.989501,
        1.884956,0.053884,-2.408394,
        1.884956,0.077645,1.453433,
        1.884956,0.088408,-0.311160,
        1.884956,0.142314,2.590387,
        1.884956,0.147206,-1.385164,
        1.884956,0.163506,0.573203,
        1.884956,0.165485,-2.875396,
        1.884956,0.182564,-2.125821,
        1.884956,0.190569,1.929141,
        1.884956,0.201102,1.249082,
        1.884956,0.201397,-0.716646,
        1.884956,0.210568,-0.069559,
        1.884956,0.251672,2.848900,
        1.884956,0.266262,-1.678068,
        1.884956,0.269108,-1.215563,
        1.884956,0.273777,-2.506756,
        1.884956,0.279854,0.814581,
        1.884956,0.281664,-2.969147,
        1.884956,0.283625,2.405603,
        1.884956,0.284058,0.357350,
        1.884956,0.294107,1.593880,
        1.884956,0.302491,-0.400410,
        1.884956,0.308059,2.021491,
        1.884956,0.312219,-2.072057,
        1.884956,0.321905,-0.829121,
        1.884956,0.329984,1.201621,
        1.884956,0.339714,-0.013712,
        1.884956,0.368784,2.974974,
        1.884956,0.376428,-1.495532,
        1.884956,0.378103,2.646436,
        1.884956,0.385088,0.590355,
        1.884956,0.386248,-2.690943,
        1.884956,0.389851,-2.340031,
        1.884956,0.391502,-1.171374,
        1.884956,0.398249,-1.812694,
        1.884956,0.400542,-2.995729,
        1.884956,0.404273,0.918163,
        1.884956,0.405617,1.773533,
        1.884956,0.409009,-0.570243,
        1.884956,0.410700,0.270433,
        1.884956,0.415893,1.464890,
        1.884956,0.420680,2.088164,
        1.884956,0.424766,-0.258729,
        1.884956,0.426507,2.367872,
        1.884956,0.441292,-0.890727,
        1.884956,0.451839,-2.078193,
        1.884956,0.459270,1.191562,
        1.884956,0.470622,0.014002,
        1.884956,0.481292,-2.529098,
        1.884956,0.482460,2.849843,
        1.884956,0.492205,-1.397963,
        1.884956,0.493314,3.117264,
        1.884956,0.497870,0.727828,
        1.884956,0.499783,-1.654379,
        1.884956,0.502651,0.463333,
        1.884956,0.505463,2.594557,
        1.884956,0.506104,-2.874910,
        1.884956,0.513579,-1.147354,
        1.884956,0.518569,1.644237,
        1.884956,0.519001,1.900488,
        1.884956,0.519277,-0.673809,
        1.884956,0.524769,-0.421955,
        1.884956,0.525564,-2.288838,
        1.884956,0.532541,0.978737,
        1.884956,0.534610,2.154808,
        1.884956,0.535020,-1.891846,
        1.884956,0.540208,1.400497,
        1.884956,0.542006,0.228337,
        1.884956,0.550762,-0.180771,
        1.884956,0.560531,2.378474,
        1.884956,0.561629,-0.924576,
        1.884956,0.567465,-2.679392,
        1.884956,0.588527,1.181755,
        1.884956,0.591643,-2.098364,
        1.884956,0.594963,2.997575,
        1.884956,0.599643,-3.045235,
        1.884956,0.599962,2.773556,
        1.884956,0.602620,0.029270,
        1.884956,0.605711,0.598454,
        1.884956,0.607708,-1.541026,
        1.884956,0.609360,-2.464379,
        1.884956,0.610738,-1.326786,
        1.884956,0.616968,0.820489,
        1.884956,0.626379,1.773610,
        1.884956,0.628626,-1.751180,
        1.884956,0.630671,0.391949,
        1.884956,0.632622,-0.737642,
        1.884956,0.632708,2.566722,
        1.884956,0.632797,1.989858,
        1.884956,0.634455,-0.534790,
        1.884956,0.636205,1.561170,
        1.884956,0.637681,-1.123423,
        1.884956,0.641441,-2.842109,
        1.884956,0.644807,-0.327139,
        1.884956,0.652440,2.197129,
        1.884956,0.660897,-2.278593,
        1.884956,0.664008,1.360091,
        1.884956,0.670850,-1.945257,
        1.884956,0.671396,1.004859,
        1.884956,0.673985,0.203855,
        1.884956,0.680326,-0.133370,
        1.884956,0.685083,-0.942363,
        1.884956,0.688204,-2.632204,
        1.884956,0.689777,2.385011,
        1.884956,0.700065,3.115430,
        1.884956,0.704080,2.911636,
        1.884956,0.712883,1.174216,
        1.884956,0.715701,0.699906,
        1.884956,0.718851,-1.442215,
        1.884956,0.721019,2.721536,
        1.884956,0.725300,-2.975134,
        1.884956,0.727136,-1.633288,
        1.884956,0.727399,-2.120567,
        1.884956,0.730751,0.522488,
        1.884956,0.732892,-1.254330,
        1.884956,0.734333,-2.446184,
        1.884956,0.735125,0.037270,
        1.884956,0.737893,1.866463,
        1.884956,0.740650,1.681075,
        1.884956,0.745971,-0.781286,
        1.884956,0.749432,-0.442559,
        1.884956,0.750102,2.050636,
        1.884956,0.751046,0.863821,
        1.884956,0.757018,-0.608833,
        1.884956,0.757046,-1.816019,
        1.884956,0.757389,1.501184,
        1.884956,0.759800,2.546181,
        1.884956,0.759859,0.351923,
        1.884956,0.761141,-2.781102,
        1.884956,0.769035,-1.078574,
        1.884956,0.770166,-0.266604,
        1.884956,0.772218,2.227380,
        1.884956,0.788382,1.331051,
        1.884956,0.790269,-2.282409,
        1.884956,0.803410,-1.984567,
        1.884956,0.803813,0.191520,
        1.884956,0.806832,3.022155,
        1.884956,0.807476,-2.597675,
        1.884956,0.812512,-0.107335,
        1.884956,0.814161,1.010949,
        1.884956,0.817732,-3.084486,
        1.884956,0.817846,2.390688,
        1.884956,0.819323,2.845960,
        1.884956,0.819430,-0.915017,
        1.884956,0.829198,-1.532491,
        1.884956,0.829531,0.665695,
        1.884956,0.829816,1.167147,
        1.884956,0.833613,-1.357492,
        1.884956,0.842945,-2.907318,
        1.884956,0.844883,2.679708,
        1.884956,0.848189,-1.705003,
        1.884956,0.849888,1.767901,
        1.884956,0.851126,0.493370,
        1.884956,0.852911,1.933394,
        1.884956,0.857516,-2.432863,
        1.884956,0.858261,-1.191729,
        1.884956,0.858279,-2.138895,
        1.884956,0.859723,-0.734928,
        1.884956,0.859754,1.603292,
        1.884956,0.861655,-0.394821,
        1.884956,0.864300,0.043549,
        1.884956,0.866940,0.851229,
        1.884956,0.869748,2.095969,
        1.884956,0.877352,-0.574668,
        1.884956,0.879254,-2.733338,
        1.884956,0.882554,1.443757,
        1.884956,0.884060,0.334199,
        1.884956,0.885626,-1.866061,
        1.884956,0.889834,2.529957,
        1.884956,0.893140,2.251698,
        1.884956,0.897463,-0.240822,
        1.884956,0.898758,-1.031948,
        1.884956,0.913918,1.287618,
        1.884956,0.916555,-2.287173,
        1.884956,0.918305,3.107845,
        1.884956,0.919522,2.948277,
        1.884956,0.927138,-2.569686,
        1.884956,0.929031,0.185383,
        1.884956,0.933766,-3.015917,
        1.884956,0.934873,-2.010822,
        1.884956,0.935636,-1.445855,
        1.884956,0.936702,-0.866461,
        1.884956,0.937395,0.729108,
        1.884956,0.937910,2.787575,
        1.884956,0.939144,-0.092962,
        1.884956,0.941718,1.125085,
        1.884956,0.942873,-1.607680,
        1.884956,0.943701,0.980328,
        1.884956,0.947774,2.395256,
        1.884956,0.952433,0.584347,
        1.884956,0.954456,-1.287293,
        1.884956,0.958731,-2.852552,
        1.884956,0.961838,1.839284,
        1.884956,0.961898,-0.469425,
        1.884956,0.965426,1.686615,
        1.884956,0.970108,1.990771,
        1.884956,0.972820,-1.762780,
        1.884956,0.973500,-0.711881,
        1.884956,0.975253,2.641082,
        1.884956,0.977824,0.441928,
        1.884956,0.981485,-2.424766,
        1.884956,0.981877,1.537100,
        1.884956,0.982609,-1.136920,
        1.884956,0.984265,-2.156073,
        1.884956,0.989150,0.045493,
        1.884956,0.993052,2.137478,
        1.884956,0.997766,-2.693871,
        1.884956,0.999847,-0.334738,
        1.884956,1.001920,0.848054,
        1.884956,1.006146,1.389734,
        1.884956,1.012180,0.304066,
        1.884956,1.016938,2.270674,
        1.884956,1.017371,-0.979790,
        1.884956,1.019495,-1.894487,
        1.884956,1.023796,-0.583691,
        1.884956,1.026525,2.506945,
        1.884956,1.027004,-0.203653,
        1.884956,1.028027,3.036573,
        1.884956,1.031435,1.241021,
        1.884956,1.032316,2.884221,
        1.884956,1.035863,-3.110475,
        1.884956,1.042085,-2.292950,
        1.884956,1.043272,-1.522986,
        1.884956,1.046265,-2.958847,
        1.884956,1.049272,-1.376344,
        1.884956,1.050039,-2.552086,
        1.884956,1.053682,0.689863,
        1.884956,1.055458,0.170394,
        1.884956,1.056622,-0.831866,
        1.884956,1.057515,1.097998,
        1.884956,1.058706,-2.036438,
        1.884956,1.061389,-1.672248,
        1.884956,1.063758,2.743093,
        1.884956,1.070157,-0.073422,
        1.884956,1.072154,1.761506,
        1.884956,1.074029,1.909193,
        1.884956,1.074204,0.545276,
        1.884956,1.074714,-2.805147,
        1.884956,1.076262,0.966104,
        1.884956,1.077642,-1.227073,
        1.884956,1.084151,1.618525,
        1.884956,1.085085,2.377673,
        1.884956,1.086822,-0.441112,
        1.884956,1.087142,2.058815,
        1.884956,1.098383,-0.696430,
        1.884956,1.099138,-1.084033,
        1.884956,1.103208,1.477773,
        1.884956,1.103783,0.408254,
        1.884956,1.106275,-2.420800,
        1.884956,1.108079,2.612943,
        1.884956,1.109345,-2.172718,
        1.884956,1.115957,-1.787597,
        1.884956,1.118930,0.045983,
        1.884956,1.119777,0.834702,
        1.884956,1.121943,-2.669042,
        1.884956,1.124930,2.219917,
        1.884956,1.125217,1.336294,
        1.884956,1.130958,2.972931,
        1.884956,1.131400,-0.306629,
        1.884956,1.137135,-0.936514,
        1.884956,1.139125,-1.924528,
        1.884956,1.139877,0.276382,
        1.884956,1.142003,-3.058022,
        1.884956,1.147363,-0.558448,
        1.884956,1.150348,-0.176418,
        1.884956,1.151198,1.197016,
        1.884956,1.153930,-1.597104,
        1.884956,1.155389,2.837148,
        1.884956,1.155458,-1.460623,
        1.884956,1.157331,-2.906362,
        1.884956,1.158594,2.487618,
        1.884956,1.162103,-1.322553,
        1.884956,1.165407,3.092861,
        1.884956,1.168025,0.717075,
        1.884956,1.168406,-2.301519,
        1.884956,1.174782,-2.538755,
        1.884956,1.177998,1.061256,
        1.884956,1.178898,1.827305,
        1.884956,1.181331,-2.060330,
        1.884956,1.182540,-0.798952,
        1.884956,1.184569,0.147517,
        1.884956,1.184734,1.968919,
        1.884956,1.184804,0.593386,
        1.884956,1.188278,1.690808,
        1.884956,1.194360,2.110440,
        1.884956,1.194776,2.710455,
        1.884956,1.197745,-1.173705,
        1.884956,1.199898,-2.777257,
        1.884956,1.201164,2.355931,
        1.884956,1.203760,1.556404,
        1.884956,1.205845,-0.415993,
        1.884956,1.206585,0.933648,
        1.884956,1.206950,-0.037248,
        1.884956,1.209341,0.471162,
        1.884956,1.213483,-1.039417,
        1.884956,1.218146,-0.672643,
        1.884956,1.221608,1.421669,
        1.884956,1.226043,-1.693533,
        1.884956,1.232870,-1.822773,
        1.884956,1.234375,-2.421194,
        1.884956,1.234531,-2.190148,
        1.884956,1.239414,0.348959,
        1.884956,1.240183,2.587255,
        1.884956,1.242312,2.941234,
        1.884956,1.244075,1.286644,
        1.884956,1.245961,-2.997461,
        1.884956,1.246495,-2.646872,
        1.884956,1.247736,2.242810,
        1.884956,1.253604,0.812222,
        1.884956,1.257816,-0.894317,
        1.884956,1.258247,-1.953634,
        1.884956,1.258993,-0.284099,
        1.884956,1.262380,-0.157609,
        1.884956,1.262791,-1.549229,
        1.884956,1.263772,-0.533073,
        1.884956,1.266731,-1.412461,
        1.884956,1.268591,-3.122146,
        1.884956,1.269614,-1.275246,
        1.884956,1.270745,1.152533,
        1.884956,1.273606,0.224545,
        1.884956,1.279716,2.812101,
        1.884956,1.283786,2.462124,
        1.884956,1.284798,0.078209,
        1.884956,1.285402,-2.882049,
        1.884956,1.286936,1.886053,
        1.884956,1.289408,0.683122,
        1.884956,1.294736,1.756184,
        1.884956,1.296978,-2.311511,
        1.884956,1.298546,2.020601,
        1.884956,1.300924,1.022478,
        1.884956,1.303305,3.043152,
        1.884956,1.304210,-2.084225,
        1.884956,1.305241,-2.530249,
        1.884956,1.307513,1.628003,
        1.884956,1.310722,0.558267,
        1.884956,1.314142,2.139329,
        1.884956,1.316087,-0.777971,
        1.884956,1.317907,-1.121023,
        1.884956,1.321019,-2.749776,
        1.884956,1.321609,2.687717,
        1.884956,1.322568,1.499415,
        1.884956,1.322972,-0.395699,
        1.884956,1.325286,-0.994192,
        1.884956,1.329146,-0.648513,
        1.884956,1.329795,-0.042570,
        1.884956,1.331651,2.344497,
        1.884956,1.336522,0.435098,
        1.884956,1.339410,0.900087,
        1.884956,1.339550,1.370353,
        1.884956,1.340441,-1.731580,
        1.884956,1.351855,-1.855234,
        1.884956,1.360161,-3.018404,
        1.884956,1.361031,-1.613971,
        1.884956,1.361196,2.914780,
        1.884956,1.361998,-2.207254,
        1.884956,1.362302,1.239384,
        1.884956,1.365228,2.563808,
        1.884956,1.365452,0.310522,
        1.884956,1.367028,-2.420041,
        1.884956,1.373422,-1.484281,
        1.884956,1.374245,-1.226417,
        1.884956,1.375020,-0.170393,
        1.884956,1.376560,-1.355481,
        1.884956,1.377070,-1.980984,
        1.884956,1.377578,-2.636339,
        1.884956,1.378096,0.774626,
        1.884956,1.379086,-0.508192,
        1.884956,1.382562,0.177141,
        1.884956,1.388849,2.233847,
        1.884956,1.389268,-0.886349,
        1.884956,1.391324,1.109396,
        1.884956,1.391699,-0.288725,
        1.884956,1.392869,3.136903,
        1.884956,1.393612,1.952016,
        1.884956,1.395545,-2.850807,
        1.884956,1.401926,1.822930,
        1.884956,1.402149,2.788898,
        1.884956,1.409552,0.647980,
        1.884956,1.410840,0.056219,
        1.884956,1.411944,2.444377,
        1.884956,1.413639,1.697982,
        1.884956,1.426479,1.573427,
        1.884956,1.426725,2.102242,
        1.884956,1.427868,0.984637,
        1.884956,1.429580,-2.107232,
        1.884956,1.431375,-2.313269,
        1.884956,1.431907,3.015410,
        1.884956,1.432325,-1.029054,
        1.884956,1.433341,0.524268,
        1.884956,1.435181,-0.739771,
        1.884956,1.437208,-2.529085,
        1.884956,1.439041,-0.617966,
        1.884956,1.441473,1.448544,
        1.884956,1.444527,2.664744,
        1.884956,1.450513,-0.400625,
        1.884956,1.451907,-0.066823,
        1.884956,1.452754,-2.738631,
        1.884956,1.454061,-2.950416,
        1.884956,1.456736,1.323154,
        1.884956,1.458370,0.400433,
        1.884956,1.459314,-1.765003,
        1.884956,1.459902,-1.145835,
        1.884956,1.464952,2.329245,
        1.884956,1.466050,0.862104,
        1.884956,1.471330,-1.886160,
        1.884956,1.475119,-1.648255,
        1.884956,1.480878,1.194576,
        1.884956,1.480942,0.273235,
        1.884956,1.481086,2.890269,
        1.884956,1.484118,-1.529156,
        1.884956,1.484741,-1.279219,
        1.884956,1.487681,-1.406069,
        1.884956,1.487886,-3.066037,
        1.884956,1.490555,2.543284,
        1.884956,1.492424,-0.190932,
        1.884956,1.495002,-0.839810,
        1.884956,1.495297,2.007188,
        1.884956,1.496414,-2.009267,
        1.884956,1.498261,-2.425270,
        1.884956,1.498907,-2.207327,
        1.884956,1.500204,0.737638,
        1.884956,1.502469,0.151156,
        1.884956,1.507448,-0.514782,
        1.884956,1.509287,2.204647,
        1.884956,1.510518,1.884920,
        1.884956,1.511948,-2.632360,
        1.884956,1.515235,1.068565,
        1.884956,1.522237,3.102018,
        1.884956,1.522373,1.764991,
        1.884956,1.522416,-0.954956,
        1.884956,1.523102,2.765920,
        1.884956,1.524811,-2.844718,
        1.884956,1.527261,-0.304916,
        1.884956,1.529788,0.613208,
        1.884956,1.533648,1.644808,
        1.884956,1.535018,0.028869,
        1.884956,1.541082,2.424608,
        1.884956,1.546010,1.523462,
        1.884956,1.551135,-0.642994,
        1.884956,1.553895,0.489607,
        1.884956,1.555041,0.947813,
        1.884956,1.559671,-1.074934,
        1.884956,1.560345,2.985899,
        1.884956,1.560487,-2.323400,
        1.884956,1.561127,1.401111,
        1.884956,1.567193,2.643172,
        1.884956,1.568892,2.098362,
        1.884956,1.571702,-2.098435,
        1.884956,1.572701,-2.528730,
        1.884956,1.573572,1.278593,
        1.884956,1.574849,-0.091505,
        1.884956,1.576334,0.365236,
        1.884956,1.577029,-2.962888,
        1.884956,1.577688,-1.814941,
        1.884956,1.579542,-0.755917,
        1.884956,1.579624,-1.199759,
        1.884956,1.581515,-0.419414,
        1.884956,1.584581,-2.738702,
        1.884956,1.586948,-1.696296,
        1.884956,1.588901,2.301496,
        1.884956,1.589936,0.825020,
        1.884956,1.590018,-1.941652,
        1.884956,1.596968,-1.328358,
        1.884956,1.597178,-1.575607,
        1.884956,1.600470,-1.453196,
        1.884956,1.601408,0.242409,
        1.884956,1.602259,1.152656,
        1.884956,1.603064,2.865774,
        1.884956,1.612085,1.968802,
        1.884956,1.612720,-3.080498,
        1.884956,1.616061,2.521531,
        1.884956,1.616900,-0.207457,
        1.884956,1.620988,-0.875920,
        1.884956,1.622114,0.701509,
        1.884956,1.622749,0.121426,
        1.884956,1.622993,-2.221629,
        1.884956,1.629346,-0.542669,
        1.884956,1.630007,1.844552,
        1.884956,1.634747,-2.426198,
        1.884956,1.642346,1.722617,
        1.884956,1.643925,1.034629,
        1.884956,1.644848,2.743774,
        1.884956,1.645258,2.188109,
        1.884956,1.645417,-2.634725,
        1.884956,1.647395,3.085437,
        1.884956,1.648538,-2.855044,
        1.884956,1.649932,0.576808,
        1.884956,1.652224,-0.996105,
        1.884956,1.652871,1.600681,
        1.884956,1.657997,-0.325127,
        1.884956,1.661936,0.001037,
        1.884956,1.665187,1.477402,
        1.884956,1.665496,2.397872,
        1.884956,1.673267,-0.665453,
        1.884956,1.673557,0.452526,
        1.884956,1.679277,-1.122421,
        1.884956,1.679696,0.912482,
        1.884956,1.682420,1.353922,
        1.884956,1.685031,-2.009038,
        1.884956,1.685196,2.967046,
        1.884956,1.690107,2.620910,
        1.884956,1.690453,-2.123076,
        1.884956,1.691693,-1.871572,
        1.884956,1.691817,2.061688,
        1.884956,1.692644,1.237670,
        1.884956,1.696386,0.328123,
        1.884956,1.697162,-1.249026,
        1.884956,1.697488,-2.322785,
        1.884956,1.698726,-2.973576,
        1.884956,1.699172,-1.749717,
        1.884956,1.704478,-0.106295,
        1.884956,1.707244,-0.444770,
        1.884956,1.707500,-2.531497,
        1.884956,1.707714,-0.787321,
        1.884956,1.708430,-1.627552,
        1.884956,1.710583,-1.376890,
        1.884956,1.712415,-2.747921,
        1.884956,1.713125,-1.503073,
        1.884956,1.713452,0.789419,
        1.884956,1.720612,2.279888,
        1.884956,1.724852,0.202831,
        1.884956,1.724868,2.846246,
        1.884956,1.728940,1.931535,
        1.884956,1.732838,1.127584,
        1.884956,1.735719,-3.092485,
        1.884956,1.739676,2.496017,
        1.884956,1.739971,-0.228876,
        1.884956,1.743328,0.663347,
        1.884956,1.744927,-0.911892,
        1.884956,1.745952,0.084733,
        1.884956,1.748077,1.804354,
        1.884956,1.752959,-0.567548,
        1.884956,1.760169,1.678132,
        1.884956,1.763887,-2.218450,
        1.884956,1.766737,2.722752,
        1.884956,1.767999,1.004737,
        1.884956,1.769149,1.550298,
        1.884956,1.770119,-2.427175,
        1.884956,1.770460,2.154208,
        1.884956,1.770514,3.070982,
        1.884956,1.770735,0.537245,
        1.884956,1.772673,-2.864760,
        1.884956,1.773527,-1.039582,
        1.884956,1.776419,-2.642421,
        1.884956,1.780769,1.416330,
        1.884956,1.786634,-0.346247,
        1.884956,1.788829,-1.947553,
        1.884956,1.793737,0.410270,
        1.884956,1.794553,2.374618,
        1.884956,1.794699,-0.694061,
        1.884956,1.797469,-1.169105,
        1.884956,1.803705,0.881076,
        1.884956,1.804124,1.264249,
        1.884956,1.807825,-2.090455,
        1.884956,1.808822,-1.817375,
        1.884956,1.809020,2.949741,
        1.884956,1.812466,2.597595,
        1.884956,1.813159,2.025395,
        1.884956,1.813816,-1.298435,
        1.884956,1.815016,0.278951,
        1.884956,1.816510,-0.008631,
        1.884956,1.817772,-1.689625,
        1.884956,1.821360,-2.984352,
        1.884956,1.822312,-0.127211,
        1.884956,1.823386,-1.429807,
        1.884956,1.823834,-1.560483,
        1.884956,1.832060,-0.821627,
        1.884956,1.832690,-0.468440,
        1.884956,1.833945,-2.320278,
        1.884956,1.834787,0.751968,
        1.884956,1.840579,-2.536445,
        1.884956,1.841033,-2.756932,
        1.884956,1.846350,1.893983,
        1.884956,1.846644,1.113766,
        1.884956,1.846833,2.826543,
        1.884956,1.847325,2.249051,
        1.884956,1.854119,0.134161,
        1.884956,1.858863,-3.106022,
        1.884956,1.865046,1.763384,
        1.884956,1.865619,-0.952036,
        1.884956,1.865912,0.624453,
        1.884956,1.866694,2.473686,
        1.884956,1.867524,-0.240083,
        1.884956,1.875561,1.632831,
        1.884956,1.876049,-0.594378,
        1.884956,1.880509,1.500204,
        1.884956,1.883504,1.360457,
        1.884956,1.886971,-2.196665,
        1.884956,1.888979,0.982922,
        1.884956,1.889815,2.701570,
        1.884956,1.892144,-2.005340,
        1.884956,1.892348,0.496522,
        1.884956,1.893281,-1.083976,
        1.884956,1.893828,3.053453,
        1.884956,1.895117,2.121107,
        1.884956,1.898741,-2.875735,
        1.884956,1.904429,-2.428029,
        1.884956,1.908408,-2.648903,
        1.884956,1.911828,-0.364489,
        1.884956,1.912224,0.366565,
        1.884956,1.914854,1.214468,
        1.884956,1.915467,-1.883143,
        1.884956,1.915763,-0.724679,
        1.884956,1.915790,-1.216366,
        1.884956,1.922256,2.347806,
        1.884956,1.922616,0.844633,
        1.884956,1.925185,0.231170,
        1.884956,1.927474,-1.754616,
        1.884956,1.928342,-1.351174,
        1.884956,1.933328,-1.623328,
        1.884956,1.933878,0.034150,
        1.884956,1.934585,-1.487979,
        1.884956,1.934802,2.927634,
        1.884956,1.934862,1.989776,
        1.884956,1.936327,-0.099350,
        1.884956,1.938088,2.578507,
        1.884956,1.945677,-2.999805,
        1.884956,1.954572,-0.859208,
        1.884956,1.957868,-0.490784,
        1.884956,1.958989,0.712772,
        1.884956,1.960212,-2.307561,
        1.884956,1.964774,1.855507,
        1.884956,1.964971,-2.100533,
        1.884956,1.969112,1.088273,
        1.884956,1.971438,-2.765415,
        1.884956,1.972312,2.805134,
        1.884956,1.974950,2.220596,
        1.884956,1.975135,-2.539127,
        1.884956,1.982107,1.720829,
        1.884956,1.983195,-3.127984,
        1.884956,1.983318,-0.247182,
        1.884956,1.986720,-0.993220,
        1.884956,1.990381,1.586005,
        1.884956,1.990470,0.580283,
        1.884956,1.990523,1.450326,
        1.884956,1.993909,1.312897,
        1.884956,1.994694,2.452813,
        1.884956,1.999226,-0.620717,
        1.884956,2.005145,0.941012,
        1.884956,2.007301,0.142337,
        1.884956,2.014149,0.444138,
        1.884956,2.014338,-1.127392,
        1.884956,2.018131,3.021778,
        1.884956,2.019676,-1.980944,
        1.884956,2.019912,2.686612,
        1.884956,2.020545,2.088672,
        1.884956,2.026074,0.300945,
        1.884956,2.026196,-2.889057,
        1.884956,2.031959,-1.268178,
        1.884956,2.032917,-2.199917,
        1.884956,2.033572,-1.842884,
        1.884956,2.034712,-2.419505,
        1.884956,2.035818,-0.379479,
        1.884956,2.036004,-0.757062,
        1.884956,2.040501,-1.411488,
        1.884956,2.042874,-1.701367,
        1.884956,2.043060,-2.653746,
        1.884956,2.043729,-1.557938,
        1.884956,2.048538,0.806623,
        1.884956,2.049006,1.199930,
        1.884956,2.049502,-0.137989,
        1.884956,2.052181,2.326959,
        1.884956,2.052794,-0.008526,
        1.884956,2.057608,1.951726,
        1.884956,2.065568,2.567730,
        1.884956,2.070371,2.879808,
        1.884956,2.071256,-3.018961,
        1.884956,2.079746,-0.897699,
        1.884956,2.082784,1.042403,
        1.884956,2.083752,-0.511086,
        1.884956,2.084401,1.810502,
        1.884956,2.085731,0.671081,
        1.884956,2.097496,1.373172,
        1.884956,2.099185,1.667006,
        1.884956,2.100482,-2.301084,
        1.884956,2.103118,2.194745,
        1.884956,2.103223,1.521685,
        1.884956,2.103695,-2.777707,
        1.884956,2.107523,3.129581,
        1.884956,2.108277,-2.095907,
        1.884956,2.108891,-2.532166,
        1.884956,2.111003,-0.268632,
        1.884956,2.112276,-1.027246,
        1.884956,2.114380,0.531457,
        1.884956,2.119166,0.100552,
        1.884956,2.119564,2.754818,
        1.884956,2.122474,-0.648314,
        1.884956,2.125183,2.445598,
        1.884956,2.129501,0.238290,
        1.884956,2.131178,0.386663,
        1.884956,2.131223,-1.174311,
        1.884956,2.134269,-1.948042,
        1.884956,2.134684,0.907065,
        1.884956,2.137615,2.988806,
        1.884956,2.145635,-1.327708,
        1.884956,2.146467,2.055081,
        1.884956,2.146582,-1.793788,
        1.884956,2.149210,-1.482321,
        1.884956,2.152901,-1.636788,
        1.884956,2.156337,1.151558,
        1.884956,2.156474,-2.912065,
        1.884956,2.157194,-0.792563,
        1.884956,2.166724,-0.401638,
        1.884956,2.166980,-0.134879,
        1.884956,2.173644,-2.403429,
        1.884956,2.175711,-2.655523,
        1.884956,2.177132,2.600419,
        1.884956,2.178879,0.767722,
        1.884956,2.180752,1.908485,
        1.884956,2.181338,2.313721,
        1.884956,2.185580,1.287673,
        1.884956,2.195400,-2.217546,
        1.884956,2.200418,-3.056442,
        1.884956,2.201829,-0.001735,
        1.884956,2.203804,1.756176,
        1.884956,2.211602,1.443930,
        1.884956,2.212686,-0.544809,
        1.884956,2.213937,0.621892,
        1.884956,2.214075,1.600691,
        1.884956,2.216256,1.020029,
        1.884956,2.219060,2.886631,
        1.884956,2.223375,-0.906185,
        1.884956,2.226002,-1.060248,
        1.884956,2.226413,-2.064034,
        1.884956,2.230554,2.171334,
        1.884956,2.232413,2.735777,
        1.884956,2.233709,3.070830,
        1.884956,2.234383,-2.792065,
        1.884956,2.237183,0.468799,
        1.884956,2.237695,-0.271388,
        1.884956,2.239419,0.151693,
        1.884956,2.243347,-2.521049,
        1.884956,2.244035,-1.229456,
        1.884956,2.244355,0.309899,
        1.884956,2.245175,2.460236,
        1.884956,2.245628,-1.899216,
        1.884956,2.245973,-0.702234,
        1.884956,2.255601,-1.728131,
        1.884956,2.256702,-1.401367,
        1.884956,2.261466,-1.557500,
        1.884956,2.267390,0.878499,
        1.884956,2.272099,2.018594,
        1.884956,2.278618,1.170386,
        1.884956,2.283304,-2.345584,
        1.884956,2.285175,-2.943090,
        1.884956,2.293000,-0.124451,
        1.884956,2.294058,-0.421070,
        1.884956,2.303283,1.855938,
        1.884956,2.304968,2.608978,
        1.884956,2.305347,2.311346,
        1.884956,2.307188,-2.656134,
        1.884956,2.307301,1.336200,
        1.884956,2.309655,0.726020,
        1.884956,2.317634,-2.185225,
        1.884956,2.321510,1.685664,
        1.884956,2.324376,-3.111006,
        1.884956,2.324731,1.511186,
        1.884956,2.329587,0.035498,
        1.884956,2.330158,2.977264,
        1.884956,2.335096,-0.588468,
        1.884956,2.336046,-1.115799,
        1.884956,2.338060,2.803613,
        1.884956,2.338083,-0.943400,
        1.884956,2.340210,0.562145,
        1.884956,2.341931,1.023926,
        1.884956,2.343265,-0.779061,
        1.884956,2.343434,-2.013692,
        1.884956,2.353029,-1.310529,
        1.884956,2.354868,0.211543,
        1.884956,2.355120,0.388593,
        1.884956,2.355165,-1.833173,
        1.884956,2.355462,2.151441,
        1.884956,2.362948,-2.487477,
        1.884956,2.363101,-2.810456,
        1.884956,2.363394,-0.272918,
        1.884956,2.364509,-1.644533,
        1.884956,2.370465,2.466824,
        1.884956,2.386110,-1.481351,
        1.884956,2.394960,0.862133,
        1.884956,2.395716,1.977184,
        1.884956,2.396339,1.200900,
        1.884956,2.407941,-2.316753,
        1.884956,2.409450,-2.988432,
        1.884956,2.417202,-0.103363,
        1.884956,2.417386,-0.443097,
        1.884956,2.417950,2.666662,
        1.884956,2.423190,1.789783,
        1.884956,2.423692,1.392696,
        1.884956,2.431122,3.086379,
        1.884956,2.431406,-2.647820,
        1.884956,2.431584,2.307631,
        1.884956,2.434090,1.592818,
        1.884956,2.435993,0.683844,
        1.884956,2.440898,-2.133300,
        1.884956,2.442222,-1.195837,
        1.884956,2.443112,2.883761,
        1.884956,2.449191,-0.633651,
        1.884956,2.450821,0.084741,
        1.884956,2.453499,-1.008861,
        1.884956,2.460010,-1.938011,
        1.884956,2.460199,-0.822008,
        1.884956,2.460902,0.490079,
        1.884956,2.461774,-1.734913,
        1.884956,2.463986,1.034407,
        1.884956,2.467647,0.287667,
        1.884956,2.479849,2.126925,
        1.884956,2.485907,-2.835471,
        1.884956,2.485945,-1.374853,
        1.884956,2.487747,2.505912,
        1.884956,2.487957,-0.274279,
        1.884956,2.489741,-2.468428,
        1.884956,2.512538,1.239900,
        1.884956,2.517528,1.923289,
        1.884956,2.518454,-1.565852,
        1.884956,2.519348,0.843503,
        1.884956,2.522766,-3.059231,
        1.884956,2.528686,2.737228,
        1.884956,2.534798,-2.266671,
        1.884956,2.536533,1.466274,
        1.884956,2.538504,1.699812,
        1.884956,2.539516,-0.476942,
        1.884956,2.539532,-0.074548,
        1.884956,2.543197,2.990252,
        1.884956,2.554022,2.320619,
        1.884956,2.556919,-2.655005,
        1.884956,2.558146,0.627191,
        1.884956,2.558223,-1.164189,
        1.884956,2.563452,-2.044028,
        1.884956,2.568462,0.150655,
        1.884956,2.569130,-0.700837,
        1.884956,2.570050,-1.804492,
        1.884956,2.572710,-0.934665,
        1.884956,2.575994,0.391718,
        1.884956,2.584774,1.046703,
        1.884956,2.601432,2.102535,
        1.884956,2.603559,2.555114,
        1.884956,2.604150,-2.885208,
        1.884956,2.608630,-1.378011,
        1.884956,2.611256,-0.279110,
        1.884956,2.617344,-2.439480,
        1.884956,2.627402,1.293552,
        1.884956,2.635396,1.848850,
        1.884956,2.635686,2.834950,
        1.884956,2.637087,3.132730,
        1.884956,2.642029,0.813274,
        1.884956,2.645260,1.570203,
        1.884956,2.646182,-1.614036,
        1.884956,2.660113,-2.186513,
        1.884956,2.660483,-0.034608,
        1.884956,2.661713,-0.529511,
        1.884956,2.672720,-1.110422,
        1.884956,2.676626,2.335353,
        1.884956,2.676638,0.540364,
        1.884956,2.676780,-1.901158,
        1.884956,2.681411,-2.667965,
        1.884956,2.682152,0.245751,
        1.884956,2.682859,-0.817071,
        1.884956,2.705107,1.062142,
        1.884956,2.717705,2.625995,
        1.884956,2.721001,2.060097,
        1.884956,2.722319,-2.954069,
        1.884956,2.729840,-1.368237,
        1.884956,2.734349,-0.289298,
        1.884956,2.738409,2.975671,
        1.884956,2.740366,1.374341,
        1.884956,2.744633,-2.391311,
        1.884956,2.747398,1.727166,
        1.884956,2.763799,0.760372,
        1.884956,2.768238,-1.679243,
        1.884956,2.779245,-2.046268,
        1.884956,2.780208,-0.626453,
        1.884956,2.780471,0.027732,
        1.884956,2.784696,-1.013128,
        1.884956,2.789817,0.395431,
        1.884956,2.801154,2.353143,
        1.884956,2.807091,-2.687477,
        1.884956,2.825762,1.085035,
        1.884956,2.830209,2.734943,
        1.884956,2.833517,-3.085508,
        1.884956,2.838301,1.980061,
        1.884956,2.849871,1.512012,
        1.884956,2.850089,-1.343899,
        1.884956,2.856905,-0.318441,
        1.884956,2.870445,-2.291001,
        1.884956,2.885017,0.655562,
        1.884956,2.886428,-1.793217,
        1.884956,2.890337,-0.825633,
        1.884956,2.898932,0.139973,
        1.884956,2.931927,-2.728768,
        1.884956,2.932096,2.377790,
        1.884956,2.937409,2.947245,
        1.884956,2.948048,1.133127,
        1.884956,2.950896,1.797889,
        1.884956,2.970019,-1.282382,
        1.884956,2.978333,-0.412090,
        1.884956,2.996437,-2.089121,
        1.884956,3.011274,0.434110,
        1.884956,3.050410,3.138492,
        1.884956,3.064492,1.647259,
        1.884956,3.088868,-0.989501,
        1.989675,0.053884,-2.408394,
        1.989675,0.077645,1.453433,
        1.989675,0.088408,-0.311160,
        1.989675,0.142314,2.590387,
        1.989675,0.147206,-1.385164,
        1.989675,0.163506,0.573203,
        1.989675,0.165485,-2.875396,
        1.989675,0.182564,-2.125821,
        1.989675,0.190569,1.929141,
        1.989675,0.201102,1.249082,
        1.989675,0.201397,-0.716646,
        1.989675,0.210568,-0.069559,
        1.989675,0.251672,2.848900,
        1.989675,0.266262,-1.678068,
        1.989675,0.269108,-1.215563,
        1.989675,0.273777,-2.506756,
        1.989675,0.279854,0.814581,
        1.989675,0.281664,-2.969147,
        1.989675,0.283625,2.405603,
        1.989675,0.284058,0.357350,
        1.989675,0.294107,1.593880,
        1.989675,0.302491,-0.400410,
        1.989675,0.308059,2.021491,
        1.989675,0.312219,-2.072057,
        1.989675,0.321905,-0.829121,
        1.989675,0.329984,1.201621,
        1.989675,0.339714,-0.013712,
        1.989675,0.368784,2.974974,
        1.989675,0.376428,-1.495532,
        1.989675,0.378103,2.646436,
        1.989675,0.385088,0.590355,
        1.989675,0.386248,-2.690943,
        1.989675,0.389851,-2.340031,
        1.989675,0.391502,-1.171374,
        1.989675,0.398249,-1.812694,
        1.989675,0.400542,-2.995729,
        1.989675,0.404273,0.918163,
        1.989675,0.405617,1.773533,
        1.989675,0.409009,-0.570243,
        1.989675,0.410700,0.270433,
        1.989675,0.415893,1.464890,
        1.989675,0.420680,2.088164,
        1.989675,0.424766,-0.258729,
        1.989675,0.426507,2.367872,
        1.989675,0.441292,-0.890727,
        1.989675,0.451839,-2.078193,
        1.989675,0.459270,1.191562,
        1.989675,0.470622,0.014002,
        1.989675,0.481292,-2.529098,
        1.989675,0.482460,2.849843,
        1.989675,0.492205,-1.397963,
        1.989675,0.493314,3.117264,
        1.989675,0.497870,0.727828,
        1.989675,0.499783,-1.654379,
        1.989675,0.502651,0.463333,
        1.989675,0.505463,2.594557,
        1.989675,0.506104,-2.874910,
        1.989675,0.513579,-1.147354,
        1.989675,0.518569,1.644237,
        1.989675,0.519001,1.900488,
        1.989675,0.519277,-0.673809,
        1.989675,0.524769,-0.421955,
        1.989675,0.525564,-2.288838,
        1.989675,0.532541,0.978737,
        1.989675,0.534610,2.154808,
        1.989675,0.535020,-1.891846,
        1.989675,0.540208,1.400497,
        1.989675,0.542006,0.228337,
        1.989675,0.550762,-0.180771,
        1.989675,0.560531,2.378474,
        1.989675,0.561629,-0.924576,
        1.989675,0.567465,-2.679392,
        1.989675,0.588527,1.181755,
        1.989675,0.591643,-2.098364,
        1.989675,0.594963,2.997575,
        1.989675,0.599643,-3.045235,
        1.989675,0.599962,2.773556,
        1.989675,0.602620,0.029270,
        1.989675,0.605711,0.598454,
        1.989675,0.607708,-1.541026,
        1.989675,0.609360,-2.464379,
        1.989675,0.610738,-1.326786,
        1.989675,0.616968,0.820489,
        1.989675,0.626379,1.773610,
        1.989675,0.628626,-1.751180,
        1.989675,0.630671,0.391949,
        1.989675,0.632622,-0.737642,
        1.989675,0.632708,2.566722,
        1.989675,0.632797,1.989858,
        1.989675,0.634455,-0.534790,
        1.989675,0.636205,1.561170,
        1.989675,0.637681,-1.123423,
        1.989675,0.641441,-2.842109,
        1.989675,0.644807,-0.327139,
        1.989675,0.652440,2.197129,
        1.989675,0.660897,-2.278593,
        1.989675,0.664008,1.360091,
        1.989675,0.670850,-1.945257,
        1.989675,0.671396,1.004859,
        1.989675,0.673985,0.203855,
        1.989675,0.680326,-0.133370,
        1.989675,0.685083,-0.942363,
        1.989675,0.688204,-2.632204,
        1.989675,0.689777,2.385011,
        1.989675,0.700065,3.115430,
        1.989675,0.704080,2.911636,
        1.989675,0.712883,1.174216,
        1.989675,0.715701,0.699906,
        1.989675,0.718851,-1.442215,
        1.989675,0.721019,2.721536,
        1.989675,0.725300,-2.975134,
        1.989675,0.727136,-1.633288,
        1.989675,0.727399,-2.120567,
        1.989675,0.730751,0.522488,
        1.989675,0.732892,-1.254330,
        1.989675,0.734333,-2.446184,
        1.989675,0.735125,0.037270,
        1.989675,0.737893,1.866463,
        1.989675,0.740650,1.681075,
        1.989675,0.745971,-0.781286,
        1.989675,0.749432,-0.442559,
        1.989675,0.750102,2.050636,
        1.989675,0.751046,0.863821,
        1.989675,0.757018,-0.608833,
        1.989675,0.757046,-1.816019,
        1.989675,0.757389,1.501184,
        1.989675,0.759800,2.546181,
        1.989675,0.759859,0.351923,
        1.989675,0.761141,-2.781102,
        1.989675,0.769035,-1.078574,
        1.989675,0.770166,-0.266604,
        1.989675,0.772218,2.227380,
        1.989675,0.788382,1.331051,
        1.989675,0.790269,-2.282409,
        1.989675,0.803410,-1.984567,
        1.989675,0.803813,0.191520,
        1.989675,0.806832,3.022155,
        1.989675,0.807476,-2.597675,
        1.989675,0.812512,-0.107335,
        1.989675,0.814161,1.010949,
        1.989675,0.817732,-3.084486,
        1.989675,0.817846,2.390688,
        1.989675,0.819323,2.845960,
        1.989675,0.819430,-0.915017,
        1.989675,0.829198,-1.532491,
        1.989675,0.829531,0.665695,
        1.989675,0.829816,1.167147,
        1.989675,0.833613,-1.357492,
        1.989675,0.842945,-2.907318,
        1.989675,0.844883,2.679708,
        1.989675,0.848189,-1.705003,
        1.989675,0.849888,1.767901,
        1.989675,0.851126,0.493370,
        1.989675,0.852911,1.933394,
        1.989675,0.857516,-2.432863,
        1.989675,0.858261,-1.191729,
        1.989675,0.858279,-2.138895,
        1.989675,0.859723,-0.734928,
        1.989675,0.859754,1.603292,
        1.989675,0.861655,-0.394821,
        1.989675,0.864300,0.043549,
        1.989675,0.866940,0.851229,
        1.989675,0.869748,2.095969,
        1.989675,0.877352,-0.574668,
        1.989675,0.879254,-2.733338,
        1.989675,0.882554,1.443757,
        1.989675,0.884060,0.334199,
        1.989675,0.885626,-1.866061,
        1.989675,0.889834,2.529957,
        1.989675,0.893140,2.251698,
        1.989675,0.897463,-0.240822,
        1.989675,0.898758,-1.031948,
        1.989675,0.913918,1.287618,
        1.989675,0.916555,-2.287173,
        1.989675,0.918305,3.107845,
        1.989675,0.919522,2.948277,
        1.989675,0.927138,-2.569686,
        1.989675,0.929031,0.185383,
        1.989675,0.933766,-3.015917,
        1.989675,0.934873,-2.010822,
        1.989675,0.935636,-1.445855,
        1.989675,0.936702,-0.866461,
        1.989675,0.937395,0.729108,
        1.989675,0.937910,2.787575,
        1.989675,0.939144,-0.092962,
        1.989675,0.941718,1.125085,
        1.989675,0.942873,-1.607680,
        1.989675,0.943701,0.980328,
        1.989675,0.947774,2.395256,
        1.989675,0.952433,0.584347,
        1.989675,0.954456,-1.287293,
        1.989675,0.958731,-2.852552,
        1.989675,0.961838,1.839284,
        1.989675,0.961898,-0.469425,
        1.989675,0.965426,1.686615,
        1.989675,0.970108,1.990771,
        1.989675,0.972820,-1.762780,
        1.989675,0.973500,-0.711881,
        1.989675,0.975253,2.641082,
        1.989675,0.977824,0.441928,
        1.989675,0.981485,-2.424766,
        1.989675,0.981877,1.537100,
        1.989675,0.982609,-1.136920,
        1.989675,0.984265,-2.156073,
        1.989675,0.989150,0.045493,
        1.989675,0.993052,2.137478,
        1.989675,0.997766,-2.693871,
        1.989675,0.999847,-0.334738,
        1.989675,1.001920,0.848054,
        1.989675,1.006146,1.389734,
        1.989675,1.012180,0.304066,
        1.989675,1.016938,2.270674,
        1.989675,1.017371,-0.979790,
        1.989675,1.019495,-1.894487,
        1.989675,1.023796,-0.583691,
        1.989675,1.026525,2.506945,
        1.989675,1.027004,-0.203653,
        1.989675,1.028027,3.036573,
        1.989675,1.031435,1.241021,
        1.989675,1.032316,2.884221,
        1.989675,1.035863,-3.110475,
        1.989675,1.042085,-2.292950,
        1.989675,1.043272,-1.522986,
        1.989675,1.046265,-2.958847,
        1.989675,1.049272,-1.376344,
        1.989675,1.050039,-2.552086,
        1.989675,1.053682,0.689863,
        1.989675,1.055458,0.170394,
        1.989675,1.056622,-0.831866,
        1.989675,1.057515,1.097998,
        1.989675,1.058706,-2.036438,
        1.989675,1.061389,-1.672248,
        1.989675,1.063758,2.743093,
        1.989675,1.070157,-0.073422,
        1.989675,1.072154,1.761506,
        1.989675,1.074029,1.909193,
        1.989675,1.074204,0.545276,
        1.989675,1.074714,-2.805147,
        1.989675,1.076262,0.966104,
        1.989675,1.077642,-1.227073,
        1.989675,1.084151,1.618525,
        1.989675,1.085085,2.377673,
        1.989675,1.086822,-0.441112,
        1.989675,1.087142,2.058815,
        1.989675,1.098383,-0.696430,
        1.989675,1.099138,-1.084033,
        1.989675,1.103208,1.477773,
        1.989675,1.103783,0.408254,
        1.989675,1.106275,-2.420800,
        1.989675,1.108079,2.612943,
        1.989675,1.109345,-2.172718,
        1.989675,1.115957,-1.787597,
        1.989675,1.118930,0.045983,
        1.989675,1.119777,0.834702,
        1.989675,1.121943,-2.669042,
        1.989675,1.124930,2.219917,
        1.989675,1.125217,1.336294,
        1.989675,1.130958,2.972931,
        1.989675,1.131400,-0.306629,
        1.989675,1.137135,-0.936514,
        1.989675,1.139125,-1.924528,
        1.989675,1.139877,0.276382,
        1.989675,1.142003,-3.058022,
        1.989675,1.147363,-0.558448,
        1.989675,1.150348,-0.176418,
        1.989675,1.151198,1.197016,
        1.989675,1.153930,-1.597104,
        1.989675,1.155389,2.837148,
        1.989675,1.155458,-1.460623,
        1.989675,1.157331,-2.906362,
        1.989675,1.158594,2.487618,
        1.989675,1.162103,-1.322553,
        1.989675,1.165407,3.092861,
        1.989675,1.168025,0.717075,
        1.989675,1.168406,-2.301519,
        1.989675,1.174782,-2.538755,
        1.989675,1.177998,1.061256,
        1.989675,1.178898,1.827305,
        1.989675,1.181331,-2.060330,
        1.989675,1.182540,-0.798952,
        1.989675,1.184569,0.147517,
        1.989675,1.184734,1.968919,
        1.989675,1.184804,0.593386,
        1.989675,1.188278,1.690808,
        1.989675,1.194360,2.110440,
        1.989675,1.194776,2.710455,
        1.989675,1.197745,-1.173705,
        1.989675,1.199898,-2.777257,
        1.989675,1.201164,2.355931,
        1.989675,1.203760,1.556404,
        1.989675,1.205845,-0.415993,
        1.989675,1.206585,0.933648,
        1.989675,1.206950,-0.037248,
        1.989675,1.209341,0.471162,
        1.989675,1.213483,-1.039417,
        1.989675,1.218146,-0.672643,
        1.989675,1.221608,1.421669,
        1.989675,1.226043,-1.693533,
        1.989675,1.232870,-1.822773,
        1.989675,1.234375,-2.421194,
        1.989675,1.234531,-2.190148,
        1.989675,1.239414,0.348959,
        1.989675,1.240183,2.587255,
        1.989675,1.242312,2.941234,
        1.989675,1.244075,1.286644,
        1.989675,1.245961,-2.997461,
        1.989675,1.246495,-2.646872,
        1.989675,1.247736,2.242810,
        1.989675,1.253604,0.812222,
        1.989675,1.257816,-0.894317,
        1.989675,1.258247,-1.953634,
        1.989675,1.258993,-0.284099,
        1.989675,1.262380,-0.157609,
        1.989675,1.262791,-1.549229,
        1.989675,1.263772,-0.533073,
        1.989675,1.266731,-1.412461,
        1.989675,1.268591,-3.122146,
        1.989675,1.269614,-1.275246,
        1.989675,1.270745,1.152533,
        1.989675,1.273606,0.224545,
        1.989675,1.279716,2.812101,
        1.989675,1.283786,2.462124,
        1.989675,1.284798,0.078209,
        1.989675,1.285402,-2.882049,
        1.989675,1.286936,1.886053,
        1.989675,1.289408,0.683122,
        1.989675,1.294736,1.756184,
        1.989675,1.296978,-2.311511,
        1.989675,1.298546,2.020601,
        1.989675,1.300924,1.022478,
        1.989675,1.303305,3.043152,
        1.989675,1.304210,-2.084225,
        1.989675,1.305241,-2.530249,
        1.989675,1.307513,1.628003,
        1.989675,1.310722,0.558267,
        1.989675,1.314142,2.139329,
        1.989675,1.316087,-0.777971,
        1.989675,1.317907,-1.121023,
        1.989675,1.321019,-2.749776,
        1.989675,1.321609,2.687717,
        1.989675,1.322568,1.499415,
        1.989675,1.322972,-0.395699,
        1.989675,1.325286,-0.994192,
        1.989675,1.329146,-0.648513,
        1.989675,1.329795,-0.042570,
        1.989675,1.331651,2.344497,
        1.989675,1.336522,0.435098,
        1.989675,1.339410,0.900087,
        1.989675,1.339550,1.370353,
        1.989675,1.340441,-1.731580,
        1.989675,1.351855,-1.855234,
        1.989675,1.360161,-3.018404,
        1.989675,1.361031,-1.613971,
        1.989675,1.361196,2.914780,
        1.989675,1.361998,-2.207254,
        1.989675,1.362302,1.239384,
        1.989675,1.365228,2.563808,
        1.989675,1.365452,0.310522,
        1.989675,1.367028,-2.420041,
        1.989675,1.373422,-1.484281,
        1.989675,1.374245,-1.226417,
        1.989675,1.375020,-0.170393,
        1.989675,1.376560,-1.355481,
        1.989675,1.377070,-1.980984,
        1.989675,1.377578,-2.636339,
        1.989675,1.378096,0.774626,
        1.989675,1.379086,-0.508192,
        1.989675,1.382562,0.177141,
        1.989675,1.388849,2.233847,
        1.989675,1.389268,-0.886349,
        1.989675,1.391324,1.109396,
        1.989675,1.391699,-0.288725,
        1.989675,1.392869,3.136903,
        1.989675,1.393612,1.952016,
        1.989675,1.395545,-2.850807,
        1.989675,1.401926,1.822930,
        1.989675,1.402149,2.788898,
        1.989675,1.409552,0.647980,
        1.989675,1.410840,0.056219,
        1.989675,1.411944,2.444377,
        1.989675,1.413639,1.697982,
        1.989675,1.426479,1.573427,
        1.989675,1.426725,2.102242,
        1.989675,1.427868,0.984637,
        1.989675,1.429580,-2.107232,
        1.989675,1.431375,-2.313269,
        1.989675,1.431907,3.015410,
        1.989675,1.432325,-1.029054,
        1.989675,1.433341,0.524268,
        1.989675,1.435181,-0.739771,
        1.989675,1.437208,-2.529085,
        1.989675,1.439041,-0.617966,
        1.989675,1.441473,1.448544,
        1.989675,1.444527,2.664744,
        1.989675,1.450513,-0.400625,
        1.989675,1.451907,-0.066823,
        1.989675,1.452754,-2.738631,
        1.989675,1.454061,-2.950416,
        1.989675,1.456736,1.323154,
        1.989675,1.458370,0.400433,
        1.989675,1.459314,-1.765003,
        1.989675,1.459902,-1.145835,
        1.989675,1.464952,2.329245,
        1.989675,1.466050,0.862104,
        1.989675,1.471330,-1.886160,
        1.989675,1.475119,-1.648255,
        1.989675,1.480878,1.194576,
        1.989675,1.480942,0.273235,
        1.989675,1.481086,2.890269,
        1.989675,1.484118,-1.529156,
        1.989675,1.484741,-1.279219,
        1.989675,1.487681,-1.406069,
        1.989675,1.487886,-3.066037,
        1.989675,1.490555,2.543284,
        1.989675,1.492424,-0.190932,
        1.989675,1.495002,-0.839810,
        1.989675,1.495297,2.007188,
        1.989675,1.496414,-2.009267,
        1.989675,1.498261,-2.425270,
        1.989675,1.498907,-2.207327,
        1.989675,1.500204,0.737638,
        1.989675,1.502469,0.151156,
        1.989675,1.507448,-0.514782,
        1.989675,1.509287,2.204647,
        1.989675,1.510518,1.884920,
        1.989675,1.511948,-2.632360,
        1.989675,1.515235,1.068565,
        1.989675,1.522237,3.102018,
        1.989675,1.522373,1.764991,
        1.989675,1.522416,-0.954956,
        1.989675,1.523102,2.765920,
        1.989675,1.524811,-2.844718,
        1.989675,1.527261,-0.304916,
        1.989675,1.529788,0.613208,
        1.989675,1.533648,1.644808,
        1.989675,1.535018,0.028869,
        1.989675,1.541082,2.424608,
        1.989675,1.546010,1.523462,
        1.989675,1.551135,-0.642994,
        1.989675,1.553895,0.489607,
        1.989675,1.555041,0.947813,
        1.989675,1.559671,-1.074934,
        1.989675,1.560345,2.985899,
        1.989675,1.560487,-2.323400,
        1.989675,1.561127,1.401111,
        1.989675,1.567193,2.643172,
        1.989675,1.568892,2.098362,
        1.989675,1.571702,-2.098435,
        1.989675,1.572701,-2.528730,
        1.989675,1.573572,1.278593,
        1.989675,1.574849,-0.091505,
        1.989675,1.576334,0.365236,
        1.989675,1.577029,-2.962888,
        1.989675,1.577688,-1.814941,
        1.989675,1.579542,-0.755917,
        1.989675,1.579624,-1.199759,
        1.989675,1.581515,-0.419414,
        1.989675,1.584581,-2.738702,
        1.989675,1.586948,-1.696296,
        1.989675,1.588901,2.301496,
        1.989675,1.589936,0.825020,
        1.989675,1.590018,-1.941652,
        1.989675,1.596968,-1.328358,
        1.989675,1.597178,-1.575607,
        1.989675,1.600470,-1.453196,
        1.989675,1.601408,0.242409,
        1.989675,1.602259,1.152656,
        1.989675,1.603064,2.865774,
        1.989675,1.612085,1.968802,
        1.989675,1.612720,-3.080498,
        1.989675,1.616061,2.521531,
        1.989675,1.616900,-0.207457,
        1.989675,1.620988,-0.875920,
        1.989675,1.622114,0.701509,
        1.989675,1.622749,0.121426,
        1.989675,1.622993,-2.221629,
        1.989675,1.629346,-0.542669,
        1.989675,1.630007,1.844552,
        1.989675,1.634747,-2.426198,
        1.989675,1.642346,1.722617,
        1.989675,1.643925,1.034629,
        1.989675,1.644848,2.743774,
        1.989675,1.645258,2.188109,
        1.989675,1.645417,-2.634725,
        1.989675,1.647395,3.085437,
        1.989675,1.648538,-2.855044,
        1.989675,1.649932,0.576808,
        1.989675,1.652224,-0.996105,
        1.989675,1.652871,1.600681,
        1.989675,1.657997,-0.325127,
        1.989675,1.661936,0.001037,
        1.989675,1.665187,1.477402,
        1.989675,1.665496,2.397872,
        1.989675,1.673267,-0.665453,
        1.989675,1.673557,0.452526,
        1.989675,1.679277,-1.122421,
        1.989675,1.679696,0.912482,
        1.989675,1.682420,1.353922,
        1.989675,1.685031,-2.009038,
        1.989675,1.685196,2.967046,
        1.989675,1.690107,2.620910,
        1.989675,1.690453,-2.123076,
        1.989675,1.691693,-1.871572,
        1.989675,1.691817,2.061688,
        1.989675,1.692644,1.237670,
        1.989675,1.696386,0.328123,
        1.989675,1.697162,-1.249026,
        1.989675,1.697488,-2.322785,
        1.989675,1.698726,-2.973576,
        1.989675,1.699172,-1.749717,
        1.989675,1.704478,-0.106295,
        1.989675,1.707244,-0.444770,
        1.989675,1.707500,-2.531497,
        1.989675,1.707714,-0.787321,
        1.989675,1.708430,-1.627552,
        1.989675,1.710583,-1.376890,
        1.989675,1.712415,-2.747921,
        1.989675,1.713125,-1.503073,
        1.989675,1.713452,0.789419,
        1.989675,1.720612,2.279888,
        1.989675,1.724852,0.202831,
        1.989675,1.724868,2.846246,
        1.989675,1.728940,1.931535,
        1.989675,1.732838,1.127584,
        1.989675,1.735719,-3.092485,
        1.989675,1.739676,2.496017,
        1.989675,1.739971,-0.228876,
        1.989675,1.743328,0.663347,
        1.989675,1.744927,-0.911892,
        1.989675,1.745952,0.084733,
        1.989675,1.748077,1.804354,
        1.989675,1.752959,-0.567548,
        1.989675,1.760169,1.678132,
        1.989675,1.763887,-2.218450,
        1.989675,1.766737,2.722752,
        1.989675,1.767999,1.004737,
        1.989675,1.769149,1.550298,
        1.989675,1.770119,-2.427175,
        1.989675,1.770460,2.154208,
        1.989675,1.770514,3.070982,
        1.989675,1.770735,0.537245,
        1.989675,1.772673,-2.864760,
        1.989675,1.773527,-1.039582,
        1.989675,1.776419,-2.642421,
        1.989675,1.780769,1.416330,
        1.989675,1.786634,-0.346247,
        1.989675,1.788829,-1.947553,
        1.989675,1.793737,0.410270,
        1.989675,1.794553,2.374618,
        1.989675,1.794699,-0.694061,
        1.989675,1.797469,-1.169105,
        1.989675,1.803705,0.881076,
        1.989675,1.804124,1.264249,
        1.989675,1.807825,-2.090455,
        1.989675,1.808822,-1.817375,
        1.989675,1.809020,2.949741,
        1.989675,1.812466,2.597595,
        1.989675,1.813159,2.025395,
        1.989675,1.813816,-1.298435,
        1.989675,1.815016,0.278951,
        1.989675,1.816510,-0.008631,
        1.989675,1.817772,-1.689625,
        1.989675,1.821360,-2.984352,
        1.989675,1.822312,-0.127211,
        1.989675,1.823386,-1.429807,
        1.989675,1.823834,-1.560483,
        1.989675,1.832060,-0.821627,
        1.989675,1.832690,-0.468440,
        1.989675,1.833945,-2.320278,
        1.989675,1.834787,0.751968,
        1.989675,1.840579,-2.536445,
        1.989675,1.841033,-2.756932,
        1.989675,1.846350,1.893983,
        1.989675,1.846644,1.113766,
        1.989675,1.846833,2.826543,
        1.989675,1.847325,2.249051,
        1.989675,1.854119,0.134161,
        1.989675,1.858863,-3.106022,
        1.989675,1.865046,1.763384,
        1.989675,1.865619,-0.952036,
        1.989675,1.865912,0.624453,
        1.989675,1.866694,2.473686,
        1.989675,1.867524,-0.240083,
        1.989675,1.875561,1.632831,
        1.989675,1.876049,-0.594378,
        1.989675,1.880509,1.500204,
        1.989675,1.883504,1.360457,
        1.989675,1.886971,-2.196665,
        1.989675,1.888979,0.982922,
        1.989675,1.889815,2.701570,
        1.989675,1.892144,-2.005340,
        1.989675,1.892348,0.496522,
        1.989675,1.893281,-1.083976,
        1.989675,1.893828,3.053453,
        1.989675,1.895117,2.121107,
        1.989675,1.898741,-2.875735,
        1.989675,1.904429,-2.428029,
        1.989675,1.908408,-2.648903,
        1.989675,1.911828,-0.364489,
        1.989675,1.912224,0.366565,
        1.989675,1.914854,1.214468,
        1.989675,1.915467,-1.883143,
        1.989675,1.915763,-0.724679,
        1.989675,1.915790,-1.216366,
        1.989675,1.922256,2.347806,
        1.989675,1.922616,0.844633,
        1.989675,1.925185,0.231170,
        1.989675,1.927474,-1.754616,
        1.989675,1.928342,-1.351174,
        1.989675,1.933328,-1.623328,
        1.989675,1.933878,0.034150,
        1.989675,1.934585,-1.487979,
        1.989675,1.934802,2.927634,
        1.989675,1.934862,1.989776,
        1.989675,1.936327,-0.099350,
        1.989675,1.938088,2.578507,
        1.989675,1.945677,-2.999805,
        1.989675,1.954572,-0.859208,
        1.989675,1.957868,-0.490784,
        1.989675,1.958989,0.712772,
        1.989675,1.960212,-2.307561,
        1.989675,1.964774,1.855507,
        1.989675,1.964971,-2.100533,
        1.989675,1.969112,1.088273,
        1.989675,1.971438,-2.765415,
        1.989675,1.972312,2.805134,
        1.989675,1.974950,2.220596,
        1.989675,1.975135,-2.539127,
        1.989675,1.982107,1.720829,
        1.989675,1.983195,-3.127984,
        1.989675,1.983318,-0.247182,
        1.989675,1.986720,-0.993220,
        1.989675,1.990381,1.586005,
        1.989675,1.990470,0.580283,
        1.989675,1.990523,1.450326,
        1.989675,1.993909,1.312897,
        1.989675,1.994694,2.452813,
        1.989675,1.999226,-0.620717,
        1.989675,2.005145,0.941012,
        1.989675,2.007301,0.142337,
        1.989675,2.014149,0.444138,
        1.989675,2.014338,-1.127392,
        1.989675,2.018131,3.021778,
        1.989675,2.019676,-1.980944,
        1.989675,2.019912,2.686612,
        1.989675,2.020545,2.088672,
        1.989675,2.026074,0.300945,
        1.989675,2.026196,-2.889057,
        1.989675,2.031959,-1.268178,
        1.989675,2.032917,-2.199917,
        1.989675,2.033572,-1.842884,
        1.989675,2.034712,-2.419505,
        1.989675,2.035818,-0.379479,
        1.989675,2.036004,-0.757062,
        1.989675,2.040501,-1.411488,
        1.989675,2.042874,-1.701367,
        1.989675,2.043060,-2.653746,
        1.989675,2.043729,-1.557938,
        1.989675,2.048538,0.806623,
        1.989675,2.049006,1.199930,
        1.989675,2.049502,-0.137989,
        1.989675,2.052181,2.326959,
        1.989675,2.052794,-0.008526,
        1.989675,2.057608,1.951726,
        1.989675,2.065568,2.567730,
        1.989675,2.070371,2.879808,
        1.989675,2.071256,-3.018961,
        1.989675,2.079746,-0.897699,
        1.989675,2.082784,1.042403,
        1.989675,2.083752,-0.511086,
        1.989675,2.084401,1.810502,
        1.989675,2.085731,0.671081,
        1.989675,2.097496,1.373172,
        1.989675,2.099185,1.667006,
        1.989675,2.100482,-2.301084,
        1.989675,2.103118,2.194745,
        1.989675,2.103223,1.521685,
        1.989675,2.103695,-2.777707,
        1.989675,2.107523,3.129581,
        1.989675,2.108277,-2.095907,
        1.989675,2.108891,-2.532166,
        1.989675,2.111003,-0.268632,
        1.989675,2.112276,-1.027246,
        1.989675,2.114380,0.531457,
        1.989675,2.119166,0.100552,
        1.989675,2.119564,2.754818,
        1.989675,2.122474,-0.648314,
        1.989675,2.125183,2.445598,
        1.989675,2.129501,0.238290,
        1.989675,2.131178,0.386663,
        1.989675,2.131223,-1.174311,
        1.989675,2.134269,-1.948042,
        1.989675,2.134684,0.907065,
        1.989675,2.137615,2.988806,
        1.989675,2.145635,-1.327708,
        1.989675,2.146467,2.055081,
        1.989675,2.146582,-1.793788,
        1.989675,2.149210,-1.482321,
        1.989675,2.152901,-1.636788,
        1.989675,2.156337,1.151558,
        1.989675,2.156474,-2.912065,
        1.989675,2.157194,-0.792563,
        1.989675,2.166724,-0.401638,
        1.989675,2.166980,-0.134879,
        1.989675,2.173644,-2.403429,
        1.989675,2.175711,-2.655523,
        1.989675,2.177132,2.600419,
        1.989675,2.178879,0.767722,
        1.989675,2.180752,1.908485,
        1.989675,2.181338,2.313721,
        1.989675,2.185580,1.287673,
        1.989675,2.195400,-2.217546,
        1.989675,2.200418,-3.056442,
        1.989675,2.201829,-0.001735,
        1.989675,2.203804,1.756176,
        1.989675,2.211602,1.443930,
        1.989675,2.212686,-0.544809,
        1.989675,2.213937,0.621892,
        1.989675,2.214075,1.600691,
        1.989675,2.216256,1.020029,
        1.989675,2.219060,2.886631,
        1.989675,2.223375,-0.906185,
        1.989675,2.226002,-1.060248,
        1.989675,2.226413,-2.064034,
        1.989675,2.230554,2.171334,
        1.989675,2.232413,2.735777,
        1.989675,2.233709,3.070830,
        1.989675,2.234383,-2.792065,
        1.989675,2.237183,0.468799,
        1.989675,2.237695,-0.271388,
        1.989675,2.239419,0.151693,
        1.989675,2.243347,-2.521049,
        1.989675,2.244035,-1.229456,
        1.989675,2.244355,0.309899,
        1.989675,2.245175,2.460236,
        1.989675,2.245628,-1.899216,
        1.989675,2.245973,-0.702234,
        1.989675,2.255601,-1.728131,
        1.989675,2.256702,-1.401367,
        1.989675,2.261466,-1.557500,
        1.989675,2.267390,0.878499,
        1.989675,2.272099,2.018594,
        1.989675,2.278618,1.170386,
        1.989675,2.283304,-2.345584,
        1.989675,2.285175,-2.943090,
        1.989675,2.293000,-0.124451,
        1.989675,2.294058,-0.421070,
        1.989675,2.303283,1.855938,
        1.989675,2.304968,2.608978,
        1.989675,2.305347,2.311346,
        1.989675,2.307188,-2.656134,
        1.989675,2.307301,1.336200,
        1.989675,2.309655,0.726020,
        1.989675,2.317634,-2.185225,
        1.989675,2.321510,1.685664,
        1.989675,2.324376,-3.111006,
        1.989675,2.324731,1.511186,
        1.989675,2.329587,0.035498,
        1.989675,2.330158,2.977264,
        1.989675,2.335096,-0.588468,
        1.989675,2.336046,-1.115799,
        1.989675,2.338060,2.803613,
        1.989675,2.338083,-0.943400,
        1.989675,2.340210,0.562145,
        1.989675,2.341931,1.023926,
        1.989675,2.343265,-0.779061,
        1.989675,2.343434,-2.013692,
        1.989675,2.353029,-1.310529,
        1.989675,2.354868,0.211543,
        1.989675,2.355120,0.388593,
        1.989675,2.355165,-1.833173,
        1.989675,2.355462,2.151441,
        1.989675,2.362948,-2.487477,
        1.989675,2.363101,-2.810456,
        1.989675,2.363394,-0.272918,
        1.989675,2.364509,-1.644533,
        1.989675,2.370465,2.466824,
        1.989675,2.386110,-1.481351,
        1.989675,2.394960,0.862133,
        1.989675,2.395716,1.977184,
        1.989675,2.396339,1.200900,
        1.989675,2.407941,-2.316753,
        1.989675,2.409450,-2.988432,
        1.989675,2.417202,-0.103363,
        1.989675,2.417386,-0.443097,
        1.989675,2.417950,2.666662,
        1.989675,2.423190,1.789783,
        1.989675,2.423692,1.392696,
        1.989675,2.431122,3.086379,
        1.989675,2.431406,-2.647820,
        1.989675,2.431584,2.307631,
        1.989675,2.434090,1.592818,
        1.989675,2.435993,0.683844,
        1.989675,2.440898,-2.133300,
        1.989675,2.442222,-1.195837,
        1.989675,2.443112,2.883761,
        1.989675,2.449191,-0.633651,
        1.989675,2.450821,0.084741,
        1.989675,2.453499,-1.008861,
        1.989675,2.460010,-1.938011,
        1.989675,2.460199,-0.822008,
        1.989675,2.460902,0.490079,
        1.989675,2.461774,-1.734913,
        1.989675,2.463986,1.034407,
        1.989675,2.467647,0.287667,
        1.989675,2.479849,2.126925,
        1.989675,2.485907,-2.835471,
        1.989675,2.485945,-1.374853,
        1.989675,2.487747,2.505912,
        1.989675,2.487957,-0.274279,
        1.989675,2.489741,-2.468428,
        1.989675,2.512538,1.239900,
        1.989675,2.517528,1.923289,
        1.989675,2.518454,-1.565852,
        1.989675,2.519348,0.843503,
        1.989675,2.522766,-3.059231,
        1.989675,2.528686,2.737228,
        1.989675,2.534798,-2.266671,
        1.989675,2.536533,1.466274,
        1.989675,2.538504,1.699812,
        1.989675,2.539516,-0.476942,
        1.989675,2.539532,-0.074548,
        1.989675,2.543197,2.990252,
        1.989675,2.554022,2.320619,
        1.989675,2.556919,-2.655005,
        1.989675,2.558146,0.627191,
        1.989675,2.558223,-1.164189,
        1.989675,2.563452,-2.044028,
        1.989675,2.568462,0.150655,
        1.989675,2.569130,-0.700837,
        1.989675,2.570050,-1.804492,
        1.989675,2.572710,-0.934665,
        1.989675,2.575994,0.391718,
        1.989675,2.584774,1.046703,
        1.989675,2.601432,2.102535,
        1.989675,2.603559,2.555114,
        1.989675,2.604150,-2.885208,
        1.989675,2.608630,-1.378011,
        1.989675,2.611256,-0.279110,
        1.989675,2.617344,-2.439480,
        1.989675,2.627402,1.293552,
        1.989675,2.635396,1.848850,
        1.989675,2.635686,2.834950,
        1.989675,2.637087,3.132730,
        1.989675,2.642029,0.813274,
        1.989675,2.645260,1.570203,
        1.989675,2.646182,-1.614036,
        1.989675,2.660113,-2.186513,
        1.989675,2.660483,-0.034608,
        1.989675,2.661713,-0.529511,
        1.989675,2.672720,-1.110422,
        1.989675,2.676626,2.335353,
        1.989675,2.676638,0.540364,
        1.989675,2.676780,-1.901158,
        1.989675,2.681411,-2.667965,
        1.989675,2.682152,0.245751,
        1.989675,2.682859,-0.817071,
        1.989675,2.705107,1.062142,
        1.989675,2.717705,2.625995,
        1.989675,2.721001,2.060097,
        1.989675,2.722319,-2.954069,
        1.989675,2.729840,-1.368237,
        1.989675,2.734349,-0.289298,
        1.989675,2.738409,2.975671,
        1.989675,2.740366,1.374341,
        1.989675,2.744633,-2.391311,
        1.989675,2.747398,1.727166,
        1.989675,2.763799,0.760372,
        1.989675,2.768238,-1.679243,
        1.989675,2.779245,-2.046268,
        1.989675,2.780208,-0.626453,
        1.989675,2.780471,0.027732,
        1.989675,2.784696,-1.013128,
        1.989675,2.789817,0.395431,
        1.989675,2.801154,2.353143,
        1.989675,2.807091,-2.687477,
        1.989675,2.825762,1.085035,
        1.989675,2.830209,2.734943,
        1.989675,2.833517,-3.085508,
        1.989675,2.838301,1.980061,
        1.989675,2.849871,1.512012,
        1.989675,2.850089,-1.343899,
        1.989675,2.856905,-0.318441,
        1.989675,2.870445,-2.291001,
        1.989675,2.885017,0.655562,
        1.989675,2.886428,-1.793217,
        1.989675,2.890337,-0.825633,
        1.989675,2.898932,0.139973,
        1.989675,2.931927,-2.728768,
        1.989675,2.932096,2.377790,
        1.989675,2.937409,2.947245,
        1.989675,2.948048,1.133127,
        1.989675,2.950896,1.797889,
        1.989675,2.970019,-1.282382,
        1.989675,2.978333,-0.412090,
        1.989675,2.996437,-2.089121,
        1.989675,3.011274,0.434110,
        1.989675,3.050410,3.138492,
        1.989675,3.064492,1.647259,
        1.989675,3.088868,-0.989501,
        2.094395,0.053884,-2.408394,
        2.094395,0.077645,1.453433,
        2.094395,0.088408,-0.311160,
        2.094395,0.142314,2.590387,
        2.094395,0.147206,-1.385164,
        2.094395,0.163506,0.573203,
        2.094395,0.165485,-2.875396,
        2.094395,0.182564,-2.125821,
        2.094395,0.190569,1.929141,
        2.094395,0.201102,1.249082,
        2.094395,0.201397,-0.716646,
        2.094395,0.210568,-0.069559,
        2.094395,0.251672,2.848900,
        2.094395,0.266262,-1.678068,
        2.094395,0.269108,-1.215563,
        2.094395,0.273777,-2.506756,
        2.094395,0.279854,0.814581,
        2.094395,0.281664,-2.969147,
        2.094395,0.283625,2.405603,
        2.094395,0.284058,0.357350,
        2.094395,0.294107,1.593880,
        2.094395,0.302491,-0.400410,
        2.094395,0.308059,2.021491,
        2.094395,0.312219,-2.072057,
        2.094395,0.321905,-0.829121,
        2.094395,0.329984,1.201621,
        2.094395,0.339714,-0.013712,
        2.094395,0.368784,2.974974,
        2.094395,0.376428,-1.495532,
        2.094395,0.378103,2.646436,
        2.094395,0.385088,0.590355,
        2.094395,0.386248,-2.690943,
        2.094395,0.389851,-2.340031,
        2.094395,0.391502,-1.171374,
        2.094395,0.398249,-1.812694,
        2.094395,0.400542,-2.995729,
        2.094395,0.404273,0.918163,
        2.094395,0.405617,1.773533,
        2.094395,0.409009,-0.570243,
        2.094395,0.410700,0.270433,
        2.094395,0.415893,1.464890,
        2.094395,0.420680,2.088164,
        2.094395,0.424766,-0.258729,
        2.094395,0.426507,2.367872,
        2.094395,0.441292,-0.890727,
        2.094395,0.451839,-2.078193,
        2.094395,0.459270,1.191562,
        2.094395,0.470622,0.014002,
        2.094395,0.481292,-2.529098,
        2.094395,0.482460,2.849843,
        2.094395,0.492205,-1.397963,
        2.094395,0.493314,3.117264,
        2.094395,0.497870,0.727828,
        2.094395,0.499783,-1.654379,
        2.094395,0.502651,0.463333,
        2.094395,0.505463,2.594557,
        2.094395,0.506104,-2.874910,
        2.094395,0.513579,-1.147354,
        2.094395,0.518569,1.644237,
        2.094395,0.519001,1.900488,
        2.094395,0.519277,-0.673809,
        2.094395,0.524769,-0.421955,
        2.094395,0.525564,-2.288838,
        2.094395,0.532541,0.978737,
        2.094395,0.534610,2.154808,
        2.094395,0.535020,-1.891846,
        2.094395,0.540208,1.400497,
        2.094395,0.542006,0.228337,
        2.094395,0.550762,-0.180771,
        2.094395,0.560531,2.378474,
        2.094395,0.561629,-0.924576,
        2.094395,0.567465,-2.679392,
        2.094395,0.588527,1.181755,
        2.094395,0.591643,-2.098364,
        2.094395,0.594963,2.997575,
        2.094395,0.599643,-3.045235,
        2.094395,0.599962,2.773556,
        2.094395,0.602620,0.029270,
        2.094395,0.605711,0.598454,
        2.094395,0.607708,-1.541026,
        2.094395,0.609360,-2.464379,
        2.094395,0.610738,-1.326786,
        2.094395,0.616968,0.820489,
        2.094395,0.626379,1.773610,
        2.094395,0.628626,-1.751180,
        2.094395,0.630671,0.391949,
        2.094395,0.632622,-0.737642,
        2.094395,0.632708,2.566722,
        2.094395,0.632797,1.989858,
        2.094395,0.634455,-0.534790,
        2.094395,0.636205,1.561170,
        2.094395,0.637681,-1.123423,
        2.094395,0.641441,-2.842109,
        2.094395,0.644807,-0.327139,
        2.094395,0.652440,2.197129,
        2.094395,0.660897,-2.278593,
        2.094395,0.664008,1.360091,
        2.094395,0.670850,-1.945257,
        2.094395,0.671396,1.004859,
        2.094395,0.673985,0.203855,
        2.094395,0.680326,-0.133370,
        2.094395,0.685083,-0.942363,
        2.094395,0.688204,-2.632204,
        2.094395,0.689777,2.385011,
        2.094395,0.700065,3.115430,
        2.094395,0.704080,2.911636,
        2.094395,0.712883,1.174216,
        2.094395,0.715701,0.699906,
        2.094395,0.718851,-1.442215,
        2.094395,0.721019,2.721536,
        2.094395,0.725300,-2.975134,
        2.094395,0.727136,-1.633288,
        2.094395,0.727399,-2.120567,
        2.094395,0.730751,0.522488,
        2.094395,0.732892,-1.254330,
        2.094395,0.734333,-2.446184,
        2.094395,0.735125,0.037270,
        2.094395,0.737893,1.866463,
        2.094395,0.740650,1.681075,
        2.094395,0.745971,-0.781286,
        2.094395,0.749432,-0.442559,
        2.094395,0.750102,2.050636,
        2.094395,0.751046,0.863821,
        2.094395,0.757018,-0.608833,
        2.094395,0.757046,-1.816019,
        2.094395,0.757389,1.501184,
        2.094395,0.759800,2.546181,
        2.094395,0.759859,0.351923,
        2.094395,0.761141,-2.781102,
        2.094395,0.769035,-1.078574,
        2.094395,0.770166,-0.266604,
        2.094395,0.772218,2.227380,
        2.094395,0.788382,1.331051,
        2.094395,0.790269,-2.282409,
        2.094395,0.803410,-1.984567,
        2.094395,0.803813,0.191520,
        2.094395,0.806832,3.022155,
        2.094395,0.807476,-2.597675,
        2.094395,0.812512,-0.107335,
        2.094395,0.814161,1.010949,
        2.094395,0.817732,-3.084486,
        2.094395,0.817846,2.390688,
        2.094395,0.819323,2.845960,
        2.094395,0.819430,-0.915017,
        2.094395,0.829198,-1.532491,
        2.094395,0.829531,0.665695,
        2.094395,0.829816,1.167147,
        2.094395,0.833613,-1.357492,
        2.094395,0.842945,-2.907318,
        2.094395,0.844883,2.679708,
        2.094395,0.848189,-1.705003,
        2.094395,0.849888,1.767901,
        2.094395,0.851126,0.493370,
        2.094395,0.852911,1.933394,
        2.094395,0.857516,-2.432863,
        2.094395,0.858261,-1.191729,
        2.094395,0.858279,-2.138895,
        2.094395,0.859723,-0.734928,
        2.094395,0.859754,1.603292,
        2.094395,0.861655,-0.394821,
        2.094395,0.864300,0.043549,
        2.094395,0.866940,0.851229,
        2.094395,0.869748,2.095969,
        2.094395,0.877352,-0.574668,
        2.094395,0.879254,-2.733338,
        2.094395,0.882554,1.443757,
        2.094395,0.884060,0.334199,
        2.094395,0.885626,-1.866061,
        2.094395,0.889834,2.529957,
        2.094395,0.893140,2.251698,
        2.094395,0.897463,-0.240822,
        2.094395,0.898758,-1.031948,
        2.094395,0.913918,1.287618,
        2.094395,0.916555,-2.287173,
        2.094395,0.918305,3.107845,
        2.094395,0.919522,2.948277,
        2.094395,0.927138,-2.569686,
        2.094395,0.929031,0.185383,
        2.094395,0.933766,-3.015917,
        2.094395,0.934873,-2.010822,
        2.094395,0.935636,-1.445855,
        2.094395,0.936702,-0.866461,
        2.094395,0.937395,0.729108,
        2.094395,0.937910,2.787575,
        2.094395,0.939144,-0.092962,
        2.094395,0.941718,1.125085,
        2.094395,0.942873,-1.607680,
        2.094395,0.943701,0.980328,
        2.094395,0.947774,2.395256,
        2.094395,0.952433,0.584347,
        2.094395,0.954456,-1.287293,
        2.094395,0.958731,-2.852552,
        2.094395,0.961838,1.839284,
        2.094395,0.961898,-0.469425,
        2.094395,0.965426,1.686615,
        2.094395,0.970108,1.990771,
        2.094395,0.972820,-1.762780,
        2.094395,0.973500,-0.711881,
        2.094395,0.975253,2.641082,
        2.094395,0.977824,0.441928,
        2.094395,0.981485,-2.424766,
        2.094395,0.981877,1.537100,
        2.094395,0.982609,-1.136920,
        2.094395,0.984265,-2.156073,
        2.094395,0.989150,0.045493,
        2.094395,0.993052,2.137478,
        2.094395,0.997766,-2.693871,
        2.094395,0.999847,-0.334738,
        2.094395,1.001920,0.848054,
        2.094395,1.006146,1.389734,
        2.094395,1.012180,0.304066,
        2.094395,1.016938,2.270674,
        2.094395,1.017371,-0.979790,
        2.094395,1.019495,-1.894487,
        2.094395,1.023796,-0.583691,
        2.094395,1.026525,2.506945,
        2.094395,1.027004,-0.203653,
        2.094395,1.028027,3.036573,
        2.094395,1.031435,1.241021,
        2.094395,1.032316,2.884221,
        2.094395,1.035863,-3.110475,
        2.094395,1.042085,-2.292950,
        2.094395,1.043272,-1.522986,
        2.094395,1.046265,-2.958847,
        2.094395,1.049272,-1.376344,
        2.094395,1.050039,-2.552086,
        2.094395,1.053682,0.689863,
        2.094395,1.055458,0.170394,
        2.094395,1.056622,-0.831866,
        2.094395,1.057515,1.097998,
        2.094395,1.058706,-2.036438,
        2.094395,1.061389,-1.672248,
        2.094395,1.063758,2.743093,
        2.094395,1.070157,-0.073422,
        2.094395,1.072154,1.761506,
        2.094395,1.074029,1.909193,
        2.094395,1.074204,0.545276,
        2.094395,1.074714,-2.805147,
        2.094395,1.076262,0.966104,
        2.094395,1.077642,-1.227073,
        2.094395,1.084151,1.618525,
        2.094395,1.085085,2.377673,
        2.094395,1.086822,-0.441112,
        2.094395,1.087142,2.058815,
        2.094395,1.098383,-0.696430,
        2.094395,1.099138,-1.084033,
        2.094395,1.103208,1.477773,
        2.094395,1.103783,0.408254,
        2.094395,1.106275,-2.420800,
        2.094395,1.108079,2.612943,
        2.094395,1.109345,-2.172718,
        2.094395,1.115957,-1.787597,
        2.094395,1.118930,0.045983,
        2.094395,1.119777,0.834702,
        2.094395,1.121943,-2.669042,
        2.094395,1.124930,2.219917,
        2.094395,1.125217,1.336294,
        2.094395,1.130958,2.972931,
        2.094395,1.131400,-0.306629,
        2.094395,1.137135,-0.936514,
        2.094395,1.139125,-1.924528,
        2.094395,1.139877,0.276382,
        2.094395,1.142003,-3.058022,
        2.094395,1.147363,-0.558448,
        2.094395,1.150348,-0.176418,
        2.094395,1.151198,1.197016,
        2.094395,1.153930,-1.597104,
        2.094395,1.155389,2.837148,
        2.094395,1.155458,-1.460623,
        2.094395,1.157331,-2.906362,
        2.094395,1.158594,2.487618,
        2.094395,1.162103,-1.322553,
        2.094395,1.165407,3.092861,
        2.094395,1.168025,0.717075,
        2.094395,1.168406,-2.301519,
        2.094395,1.174782,-2.538755,
        2.094395,1.177998,1.061256,
        2.094395,1.178898,1.827305,
        2.094395,1.181331,-2.060330,
        2.094395,1.182540,-0.798952,
        2.094395,1.184569,0.147517,
        2.094395,1.184734,1.968919,
        2.094395,1.184804,0.593386,
        2.094395,1.188278,1.690808,
        2.094395,1.194360,2.110440,
        2.094395,1.194776,2.710455,
        2.094395,1.197745,-1.173705,
        2.094395,1.199898,-2.777257,
        2.094395,1.201164,2.355931,
        2.094395,1.203760,1.556404,
        2.094395,1.205845,-0.415993,
        2.094395,1.206585,0.933648,
        2.094395,1.206950,-0.037248,
        2.094395,1.209341,0.471162,
        2.094395,1.213483,-1.039417,
        2.094395,1.218146,-0.672643,
        2.094395,1.221608,1.421669,
        2.094395,1.226043,-1.693533,
        2.094395,1.232870,-1.822773,
        2.094395,1.234375,-2.421194,
        2.094395,1.234531,-2.190148,
        2.094395,1.239414,0.348959,
        2.094395,1.240183,2.587255,
        2.094395,1.242312,2.941234,
        2.094395,1.244075,1.286644,
        2.094395,1.245961,-2.997461,
        2.094395,1.246495,-2.646872,
        2.094395,1.247736,2.242810,
        2.094395,1.253604,0.812222,
        2.094395,1.257816,-0.894317,
        2.094395,1.258247,-1.953634,
        2.094395,1.258993,-0.284099,
        2.094395,1.262380,-0.157609,
        2.094395,1.262791,-1.549229,
        2.094395,1.263772,-0.533073,
        2.094395,1.266731,-1.412461,
        2.094395,1.268591,-3.122146,
        2.094395,1.269614,-1.275246,
        2.094395,1.270745,1.152533,
        2.094395,1.273606,0.224545,
        2.094395,1.279716,2.812101,
        2.094395,1.283786,2.462124,
        2.094395,1.284798,0.078209,
        2.094395,1.285402,-2.882049,
        2.094395,1.286936,1.886053,
        2.094395,1.289408,0.683122,
        2.094395,1.294736,1.756184,
        2.094395,1.296978,-2.311511,
        2.094395,1.298546,2.020601,
        2.094395,1.300924,1.022478,
        2.094395,1.303305,3.043152,
        2.094395,1.304210,-2.084225,
        2.094395,1.305241,-2.530249,
        2.094395,1.307513,1.628003,
        2.094395,1.310722,0.558267,
        2.094395,1.314142,2.139329,
        2.094395,1.316087,-0.777971,
        2.094395,1.317907,-1.121023,
        2.094395,1.321019,-2.749776,
        2.094395,1.321609,2.687717,
        2.094395,1.322568,1.499415,
        2.094395,1.322972,-0.395699,
        2.094395,1.325286,-0.994192,
        2.094395,1.329146,-0.648513,
        2.094395,1.329795,-0.042570,
        2.094395,1.331651,2.344497,
        2.094395,1.336522,0.435098,
        2.094395,1.339410,0.900087,
        2.094395,1.339550,1.370353,
        2.094395,1.340441,-1.731580,
        2.094395,1.351855,-1.855234,
        2.094395,1.360161,-3.018404,
        2.094395,1.361031,-1.613971,
        2.094395,1.361196,2.914780,
        2.094395,1.361998,-2.207254,
        2.094395,1.362302,1.239384,
        2.094395,1.365228,2.563808,
        2.094395,1.365452,0.310522,
        2.094395,1.367028,-2.420041,
        2.094395,1.373422,-1.484281,
        2.094395,1.374245,-1.226417,
        2.094395,1.375020,-0.170393,
        2.094395,1.376560,-1.355481,
        2.094395,1.377070,-1.980984,
        2.094395,1.377578,-2.636339,
        2.094395,1.378096,0.774626,
        2.094395,1.379086,-0.508192,
        2.094395,1.382562,0.177141,
        2.094395,1.388849,2.233847,
        2.094395,1.389268,-0.886349,
        2.094395,1.391324,1.109396,
        2.094395,1.391699,-0.288725,
        2.094395,1.392869,3.136903,
        2.094395,1.393612,1.952016,
        2.094395,1.395545,-2.850807,
        2.094395,1.401926,1.822930,
        2.094395,1.402149,2.788898,
        2.094395,1.409552,0.647980,
        2.094395,1.410840,0.056219,
        2.094395,1.411944,2.444377,
        2.094395,1.413639,1.697982,
        2.094395,1.426479,1.573427,
        2.094395,1.426725,2.102242,
        2.094395,1.427868,0.984637,
        2.094395,1.429580,-2.107232,
        2.094395,1.431375,-2.313269,
        2.094395,1.431907,3.015410,
        2.094395,1.432325,-1.029054,
        2.094395,1.433341,0.524268,
        2.094395,1.435181,-0.739771,
        2.094395,1.437208,-2.529085,
        2.094395,1.439041,-0.617966,
        2.094395,1.441473,1.448544,
        2.094395,1.444527,2.664744,
        2.094395,1.450513,-0.400625,
        2.094395,1.451907,-0.066823,
        2.094395,1.452754,-2.738631,
        2.094395,1.454061,-2.950416,
        2.094395,1.456736,1.323154,
        2.094395,1.458370,0.400433,
        2.094395,1.459314,-1.765003,
        2.094395,1.459902,-1.145835,
        2.094395,1.464952,2.329245,
        2.094395,1.466050,0.862104,
        2.094395,1.471330,-1.886160,
        2.094395,1.475119,-1.648255,
        2.094395,1.480878,1.194576,
        2.094395,1.480942,0.273235,
        2.094395,1.481086,2.890269,
        2.094395,1.484118,-1.529156,
        2.094395,1.484741,-1.279219,
        2.094395,1.487681,-1.406069,
        2.094395,1.487886,-3.066037,
        2.094395,1.490555,2.543284,
        2.094395,1.492424,-0.190932,
        2.094395,1.495002,-0.839810,
        2.094395,1.495297,2.007188,
        2.094395,1.496414,-2.009267,
        2.094395,1.498261,-2.425270,
        2.094395,1.498907,-2.207327,
        2.094395,1.500204,0.737638,
        2.094395,1.502469,0.151156,
        2.094395,1.507448,-0.514782,
        2.094395,1.509287,2.204647,
        2.094395,1.510518,1.884920,
        2.094395,1.511948,-2.632360,
        2.094395,1.515235,1.068565,
        2.094395,1.522237,3.102018,
        2.094395,1.522373,1.764991,
        2.094395,1.522416,-0.954956,
        2.094395,1.523102,2.765920,
        2.094395,1.524811,-2.844718,
        2.094395,1.527261,-0.304916,
        2.094395,1.529788,0.613208,
        2.094395,1.533648,1.644808,
        2.094395,1.535018,0.028869,
        2.094395,1.541082,2.424608,
        2.094395,1.546010,1.523462,
        2.094395,1.551135,-0.642994,
        2.094395,1.553895,0.489607,
        2.094395,1.555041,0.947813,
        2.094395,1.559671,-1.074934,
        2.094395,1.560345,2.985899,
        2.094395,1.560487,-2.323400,
        2.094395,1.561127,1.401111,
        2.094395,1.567193,2.643172,
        2.094395,1.568892,2.098362,
        2.094395,1.571702,-2.098435,
        2.094395,1.572701,-2.528730,
        2.094395,1.573572,1.278593,
        2.094395,1.574849,-0.091505,
        2.094395,1.576334,0.365236,
        2.094395,1.577029,-2.962888,
        2.094395,1.577688,-1.814941,
        2.094395,1.579542,-0.755917,
        2.094395,1.579624,-1.199759,
        2.094395,1.581515,-0.419414,
        2.094395,1.584581,-2.738702,
        2.094395,1.586948,-1.696296,
        2.094395,1.588901,2.301496,
        2.094395,1.589936,0.825020,
        2.094395,1.590018,-1.941652,
        2.094395,1.596968,-1.328358,
        2.094395,1.597178,-1.575607,
        2.094395,1.600470,-1.453196,
        2.094395,1.601408,0.242409,
        2.094395,1.602259,1.152656,
        2.094395,1.603064,2.865774,
        2.094395,1.612085,1.968802,
        2.094395,1.612720,-3.080498,
        2.094395,1.616061,2.521531,
        2.094395,1.616900,-0.207457,
        2.094395,1.620988,-0.875920,
        2.094395,1.622114,0.701509,
        2.094395,1.622749,0.121426,
        2.094395,1.622993,-2.221629,
        2.094395,1.629346,-0.542669,
        2.094395,1.630007,1.844552,
        2.094395,1.634747,-2.426198,
        2.094395,1.642346,1.722617,
        2.094395,1.643925,1.034629,
        2.094395,1.644848,2.743774,
        2.094395,1.645258,2.188109,
        2.094395,1.645417,-2.634725,
        2.094395,1.647395,3.085437,
        2.094395,1.648538,-2.855044,
        2.094395,1.649932,0.576808,
        2.094395,1.652224,-0.996105,
        2.094395,1.652871,1.600681,
        2.094395,1.657997,-0.325127,
        2.094395,1.661936,0.001037,
        2.094395,1.665187,1.477402,
        2.094395,1.665496,2.397872,
        2.094395,1.673267,-0.665453,
        2.094395,1.673557,0.452526,
        2.094395,1.679277,-1.122421,
        2.094395,1.679696,0.912482,
        2.094395,1.682420,1.353922,
        2.094395,1.685031,-2.009038,
        2.094395,1.685196,2.967046,
        2.094395,1.690107,2.620910,
        2.094395,1.690453,-2.123076,
        2.094395,1.691693,-1.871572,
        2.094395,1.691817,2.061688,
        2.094395,1.692644,1.237670,
        2.094395,1.696386,0.328123,
        2.094395,1.697162,-1.249026,
        2.094395,1.697488,-2.322785,
        2.094395,1.698726,-2.973576,
        2.094395,1.699172,-1.749717,
        2.094395,1.704478,-0.106295,
        2.094395,1.707244,-0.444770,
        2.094395,1.707500,-2.531497,
        2.094395,1.707714,-0.787321,
        2.094395,1.708430,-1.627552,
        2.094395,1.710583,-1.376890,
        2.094395,1.712415,-2.747921,
        2.094395,1.713125,-1.503073,
        2.094395,1.713452,0.789419,
        2.094395,1.720612,2.279888,
        2.094395,1.724852,0.202831,
        2.094395,1.724868,2.846246,
        2.094395,1.728940,1.931535,
        2.094395,1.732838,1.127584,
        2.094395,1.735719,-3.092485,
        2.094395,1.739676,2.496017,
        2.094395,1.739971,-0.228876,
        2.094395,1.743328,0.663347,
        2.094395,1.744927,-0.911892,
        2.094395,1.745952,0.084733,
        2.094395,1.748077,1.804354,
        2.094395,1.752959,-0.567548,
        2.094395,1.760169,1.678132,
        2.094395,1.763887,-2.218450,
        2.094395,1.766737,2.722752,
        2.094395,1.767999,1.004737,
        2.094395,1.769149,1.550298,
        2.094395,1.770119,-2.427175,
        2.094395,1.770460,2.154208,
        2.094395,1.770514,3.070982,
        2.094395,1.770735,0.537245,
        2.094395,1.772673,-2.864760,
        2.094395,1.773527,-1.039582,
        2.094395,1.776419,-2.642421,
        2.094395,1.780769,1.416330,
        2.094395,1.786634,-0.346247,
        2.094395,1.788829,-1.947553,
        2.094395,1.793737,0.410270,
        2.094395,1.794553,2.374618,
        2.094395,1.794699,-0.694061,
        2.094395,1.797469,-1.169105,
        2.094395,1.803705,0.881076,
        2.094395,1.804124,1.264249,
        2.094395,1.807825,-2.090455,
        2.094395,1.808822,-1.817375,
        2.094395,1.809020,2.949741,
        2.094395,1.812466,2.597595,
        2.094395,1.813159,2.025395,
        2.094395,1.813816,-1.298435,
        2.094395,1.815016,0.278951,
        2.094395,1.816510,-0.008631,
        2.094395,1.817772,-1.689625,
        2.094395,1.821360,-2.984352,
        2.094395,1.822312,-0.127211,
        2.094395,1.823386,-1.429807,
        2.094395,1.823834,-1.560483,
        2.094395,1.832060,-0.821627,
        2.094395,1.832690,-0.468440,
        2.094395,1.833945,-2.320278,
        2.094395,1.834787,0.751968,
        2.094395,1.840579,-2.536445,
        2.094395,1.841033,-2.756932,
        2.094395,1.846350,1.893983,
        2.094395,1.846644,1.113766,
        2.094395,1.846833,2.826543,
        2.094395,1.847325,2.249051,
        2.094395,1.854119,0.134161,
        2.094395,1.858863,-3.106022,
        2.094395,1.865046,1.763384,
        2.094395,1.865619,-0.952036,
        2.094395,1.865912,0.624453,
        2.094395,1.866694,2.473686,
        2.094395,1.867524,-0.240083,
        2.094395,1.875561,1.632831,
        2.094395,1.876049,-0.594378,
        2.094395,1.880509,1.500204,
        2.094395,1.883504,1.360457,
        2.094395,1.886971,-2.196665,
        2.094395,1.888979,0.982922,
        2.094395,1.889815,2.701570,
        2.094395,1.892144,-2.005340,
        2.094395,1.892348,0.496522,
        2.094395,1.893281,-1.083976,
        2.094395,1.893828,3.053453,
        2.094395,1.895117,2.121107,
        2.094395,1.898741,-2.875735,
        2.094395,1.904429,-2.428029,
        2.094395,1.908408,-2.648903,
        2.094395,1.911828,-0.364489,
        2.094395,1.912224,0.366565,
        2.094395,1.914854,1.214468,
        2.094395,1.915467,-1.883143,
        2.094395,1.915763,-0.724679,
        2.094395,1.915790,-1.216366,
        2.094395,1.922256,2.347806,
        2.094395,1.922616,0.844633,
        2.094395,1.925185,0.231170,
        2.094395,1.927474,-1.754616,
        2.094395,1.928342,-1.351174,
        2.094395,1.933328,-1.623328,
        2.094395,1.933878,0.034150,
        2.094395,1.934585,-1.487979,
        2.094395,1.934802,2.927634,
        2.094395,1.934862,1.989776,
        2.094395,1.936327,-0.099350,
        2.094395,1.938088,2.578507,
        2.094395,1.945677,-2.999805,
        2.094395,1.954572,-0.859208,
        2.094395,1.957868,-0.490784,
        2.094395,1.958989,0.712772,
        2.094395,1.960212,-2.307561,
        2.094395,1.964774,1.855507,
        2.094395,1.964971,-2.100533,
        2.094395,1.969112,1.088273,
        2.094395,1.971438,-2.765415,
        2.094395,1.972312,2.805134,
        2.094395,1.974950,2.220596,
        2.094395,1.975135,-2.539127,
        2.094395,1.982107,1.720829,
        2.094395,1.983195,-3.127984,
        2.094395,1.983318,-0.247182,
        2.094395,1.986720,-0.993220,
        2.094395,1.990381,1.586005,
        2.094395,1.990470,0.580283,
        2.094395,1.990523,1.450326,
        2.094395,1.993909,1.312897,
        2.094395,1.994694,2.452813,
        2.094395,1.999226,-0.620717,
        2.094395,2.005145,0.941012,
        2.094395,2.007301,0.142337,
        2.094395,2.014149,0.444138,
        2.094395,2.014338,-1.127392,
        2.094395,2.018131,3.021778,
        2.094395,2.019676,-1.980944,
        2.094395,2.019912,2.686612,
        2.094395,2.020545,2.088672,
        2.094395,2.026074,0.300945,
        2.094395,2.026196,-2.889057,
        2.094395,2.031959,-1.268178,
        2.094395,2.032917,-2.199917,
        2.094395,2.033572,-1.842884,
        2.094395,2.034712,-2.419505,
        2.094395,2.035818,-0.379479,
        2.094395,2.036004,-0.757062,
        2.094395,2.040501,-1.411488,
        2.094395,2.042874,-1.701367,
        2.094395,2.043060,-2.653746,
        2.094395,2.043729,-1.557938,
        2.094395,2.048538,0.806623,
        2.094395,2.049006,1.199930,
        2.094395,2.049502,-0.137989,
        2.094395,2.052181,2.326959,
        2.094395,2.052794,-0.008526,
        2.094395,2.057608,1.951726,
        2.094395,2.065568,2.567730,
        2.094395,2.070371,2.879808,
        2.094395,2.071256,-3.018961,
        2.094395,2.079746,-0.897699,
        2.094395,2.082784,1.042403,
        2.094395,2.083752,-0.511086,
        2.094395,2.084401,1.810502,
        2.094395,2.085731,0.671081,
        2.094395,2.097496,1.373172,
        2.094395,2.099185,1.667006,
        2.094395,2.100482,-2.301084,
        2.094395,2.103118,2.194745,
        2.094395,2.103223,1.521685,
        2.094395,2.103695,-2.777707,
        2.094395,2.107523,3.129581,
        2.094395,2.108277,-2.095907,
        2.094395,2.108891,-2.532166,
        2.094395,2.111003,-0.268632,
        2.094395,2.112276,-1.027246,
        2.094395,2.114380,0.531457,
        2.094395,2.119166,0.100552,
        2.094395,2.119564,2.754818,
        2.094395,2.122474,-0.648314,
        2.094395,2.125183,2.445598,
        2.094395,2.129501,0.238290,
        2.094395,2.131178,0.386663,
        2.094395,2.131223,-1.174311,
        2.094395,2.134269,-1.948042,
        2.094395,2.134684,0.907065,
        2.094395,2.137615,2.988806,
        2.094395,2.145635,-1.327708,
        2.094395,2.146467,2.055081,
        2.094395,2.146582,-1.793788,
        2.094395,2.149210,-1.482321,
        2.094395,2.152901,-1.636788,
        2.094395,2.156337,1.151558,
        2.094395,2.156474,-2.912065,
        2.094395,2.157194,-0.792563,
        2.094395,2.166724,-0.401638,
        2.094395,2.166980,-0.134879,
        2.094395,2.173644,-2.403429,
        2.094395,2.175711,-2.655523,
        2.094395,2.177132,2.600419,
        2.094395,2.178879,0.767722,
        2.094395,2.180752,1.908485,
        2.094395,2.181338,2.313721,
        2.094395,2.185580,1.287673,
        2.094395,2.195400,-2.217546,
        2.094395,2.200418,-3.056442,
        2.094395,2.201829,-0.001735,
        2.094395,2.203804,1.756176,
        2.094395,2.211602,1.443930,
        2.094395,2.212686,-0.544809,
        2.094395,2.213937,0.621892,
        2.094395,2.214075,1.600691,
        2.094395,2.216256,1.020029,
        2.094395,2.219060,2.886631,
        2.094395,2.223375,-0.906185,
        2.094395,2.226002,-1.060248,
        2.094395,2.226413,-2.064034,
        2.094395,2.230554,2.171334,
        2.094395,2.232413,2.735777,
        2.094395,2.233709,3.070830,
        2.094395,2.234383,-2.792065,
        2.094395,2.237183,0.468799,
        2.094395,2.237695,-0.271388,
        2.094395,2.239419,0.151693,
        2.094395,2.243347,-2.521049,
        2.094395,2.244035,-1.229456,
        2.094395,2.244355,0.309899,
        2.094395,2.245175,2.460236,
        2.094395,2.245628,-1.899216,
        2.094395,2.245973,-0.702234,
        2.094395,2.255601,-1.728131,
        2.094395,2.256702,-1.401367,
        2.094395,2.261466,-1.557500,
        2.094395,2.267390,0.878499,
        2.094395,2.272099,2.018594,
        2.094395,2.278618,1.170386,
        2.094395,2.283304,-2.345584,
        2.094395,2.285175,-2.943090,
        2.094395,2.293000,-0.124451,
        2.094395,2.294058,-0.421070,
        2.094395,2.303283,1.855938,
        2.094395,2.304968,2.608978,
        2.094395,2.305347,2.311346,
        2.094395,2.307188,-2.656134,
        2.094395,2.307301,1.336200,
        2.094395,2.309655,0.726020,
        2.094395,2.317634,-2.185225,
        2.094395,2.321510,1.685664,
        2.094395,2.324376,-3.111006,
        2.094395,2.324731,1.511186,
        2.094395,2.329587,0.035498,
        2.094395,2.330158,2.977264,
        2.094395,2.335096,-0.588468,
        2.094395,2.336046,-1.115799,
        2.094395,2.338060,2.803613,
        2.094395,2.338083,-0.943400,
        2.094395,2.340210,0.562145,
        2.094395,2.341931,1.023926,
        2.094395,2.343265,-0.779061,
        2.094395,2.343434,-2.013692,
        2.094395,2.353029,-1.310529,
        2.094395,2.354868,0.211543,
        2.094395,2.355120,0.388593,
        2.094395,2.355165,-1.833173,
        2.094395,2.355462,2.151441,
        2.094395,2.362948,-2.487477,
        2.094395,2.363101,-2.810456,
        2.094395,2.363394,-0.272918,
        2.094395,2.364509,-1.644533,
        2.094395,2.370465,2.466824,
        2.094395,2.386110,-1.481351,
        2.094395,2.394960,0.862133,
        2.094395,2.395716,1.977184,
        2.094395,2.396339,1.200900,
        2.094395,2.407941,-2.316753,
        2.094395,2.409450,-2.988432,
        2.094395,2.417202,-0.103363,
        2.094395,2.417386,-0.443097,
        2.094395,2.417950,2.666662,
        2.094395,2.423190,1.789783,
        2.094395,2.423692,1.392696,
        2.094395,2.431122,3.086379,
        2.094395,2.431406,-2.647820,
        2.094395,2.431584,2.307631,
        2.094395,2.434090,1.592818,
        2.094395,2.435993,0.683844,
        2.094395,2.440898,-2.133300,
        2.094395,2.442222,-1.195837,
        2.094395,2.443112,2.883761,
        2.094395,2.449191,-0.633651,
        2.094395,2.450821,0.084741,
        2.094395,2.453499,-1.008861,
        2.094395,2.460010,-1.938011,
        2.094395,2.460199,-0.822008,
        2.094395,2.460902,0.490079,
        2.094395,2.461774,-1.734913,
        2.094395,2.463986,1.034407,
        2.094395,2.467647,0.287667,
        2.094395,2.479849,2.126925,
        2.094395,2.485907,-2.835471,
        2.094395,2.485945,-1.374853,
        2.094395,2.487747,2.505912,
        2.094395,2.487957,-0.274279,
        2.094395,2.489741,-2.468428,
        2.094395,2.512538,1.239900,
        2.094395,2.517528,1.923289,
        2.094395,2.518454,-1.565852,
        2.094395,2.519348,0.843503,
        2.094395,2.522766,-3.059231,
        2.094395,2.528686,2.737228,
        2.094395,2.534798,-2.266671,
        2.094395,2.536533,1.466274,
        2.094395,2.538504,1.699812,
        2.094395,2.539516,-0.476942,
        2.094395,2.539532,-0.074548,
        2.094395,2.543197,2.990252,
        2.094395,2.554022,2.320619,
        2.094395,2.556919,-2.655005,
        2.094395,2.558146,0.627191,
        2.094395,2.558223,-1.164189,
        2.094395,2.563452,-2.044028,
        2.094395,2.568462,0.150655,
        2.094395,2.569130,-0.700837,
        2.094395,2.570050,-1.804492,
        2.094395,2.572710,-0.934665,
        2.094395,2.575994,0.391718,
        2.094395,2.584774,1.046703,
        2.094395,2.601432,2.102535,
        2.094395,2.603559,2.555114,
        2.094395,2.604150,-2.885208,
        2.094395,2.608630,-1.378011,
        2.094395,2.611256,-0.279110,
        2.094395,2.617344,-2.439480,
        2.094395,2.627402,1.293552,
        2.094395,2.635396,1.848850,
        2.094395,2.635686,2.834950,
        2.094395,2.637087,3.132730,
        2.094395,2.642029,0.813274,
        2.094395,2.645260,1.570203,
        2.094395,2.646182,-1.614036,
        2.094395,2.660113,-2.186513,
        2.094395,2.660483,-0.034608,
        2.094395,2.661713,-0.529511,
        2.094395,2.672720,-1.110422,
        2.094395,2.676626,2.335353,
        2.094395,2.676638,0.540364,
        2.094395,2.676780,-1.901158,
        2.094395,2.681411,-2.667965,
        2.094395,2.682152,0.245751,
        2.094395,2.682859,-0.817071,
        2.094395,2.705107,1.062142,
        2.094395,2.717705,2.625995,
        2.094395,2.721001,2.060097,
        2.094395,2.722319,-2.954069,
        2.094395,2.729840,-1.368237,
        2.094395,2.734349,-0.289298,
        2.094395,2.738409,2.975671,
        2.094395,2.740366,1.374341,
        2.094395,2.744633,-2.391311,
        2.094395,2.747398,1.727166,
        2.094395,2.763799,0.760372,
        2.094395,2.768238,-1.679243,
        2.094395,2.779245,-2.046268,
        2.094395,2.780208,-0.626453,
        2.094395,2.780471,0.027732,
        2.094395,2.784696,-1.013128,
        2.094395,2.789817,0.395431,
        2.094395,2.801154,2.353143,
        2.094395,2.807091,-2.687477,
        2.094395,2.825762,1.085035,
        2.094395,2.830209,2.734943,
        2.094395,2.833517,-3.085508,
        2.094395,2.838301,1.980061,
        2.094395,2.849871,1.512012,
        2.094395,2.850089,-1.343899,
        2.094395,2.856905,-0.318441,
        2.094395,2.870445,-2.291001,
        2.094395,2.885017,0.655562,
        2.094395,2.886428,-1.793217,
        2.094395,2.890337,-0.825633,
        2.094395,2.898932,0.139973,
        2.094395,2.931927,-2.728768,
        2.094395,2.932096,2.377790,
        2.094395,2.937409,2.947245,
        2.094395,2.948048,1.133127,
        2.094395,2.950896,1.797889,
        2.094395,2.970019,-1.282382,
        2.094395,2.978333,-0.412090,
        2.094395,2.996437,-2.089121,
        2.094395,3.011274,0.434110,
        2.094395,3.050410,3.138492,
        2.094395,3.064492,1.647259,
        2.094395,3.088868,-0.989501,
        2.199115,0.053884,-2.408394,
        2.199115,0.077645,1.453433,
        2.199115,0.088408,-0.311160,
        2.199115,0.142314,2.590387,
        2.199115,0.147206,-1.385164,
        2.199115,0.163506,0.573203,
        2.199115,0.165485,-2.875396,
        2.199115,0.182564,-2.125821,
        2.199115,0.190569,1.929141,
        2.199115,0.201102,1.249082,
        2.199115,0.201397,-0.716646,
        2.199115,0.210568,-0.069559,
        2.199115,0.251672,2.848900,
        2.199115,0.266262,-1.678068,
        2.199115,0.269108,-1.215563,
        2.199115,0.273777,-2.506756,
        2.199115,0.279854,0.814581,
        2.199115,0.281664,-2.969147,
        2.199115,0.283625,2.405603,
        2.199115,0.284058,0.357350,
        2.199115,0.294107,1.593880,
        2.199115,0.302491,-0.400410,
        2.199115,0.308059,2.021491,
        2.199115,0.312219,-2.072057,
        2.199115,0.321905,-0.829121,
        2.199115,0.329984,1.201621,
        2.199115,0.339714,-0.013712,
        2.199115,0.368784,2.974974,
        2.199115,0.376428,-1.495532,
        2.199115,0.378103,2.646436,
        2.199115,0.385088,0.590355,
        2.199115,0.386248,-2.690943,
        2.199115,0.389851,-2.340031,
        2.199115,0.391502,-1.171374,
        2.199115,0.398249,-1.812694,
        2.199115,0.400542,-2.995729,
        2.199115,0.404273,0.918163,
        2.199115,0.405617,1.773533,
        2.199115,0.409009,-0.570243,
        2.199115,0.410700,0.270433,
        2.199115,0.415893,1.464890,
        2.199115,0.420680,2.088164,
        2.199115,0.424766,-0.258729,
        2.199115,0.426507,2.367872,
        2.199115,0.441292,-0.890727,
        2.199115,0.451839,-2.078193,
        2.199115,0.459270,1.191562,
        2.199115,0.470622,0.014002,
        2.199115,0.481292,-2.529098,
        2.199115,0.482460,2.849843,
        2.199115,0.492205,-1.397963,
        2.199115,0.493314,3.117264,
        2.199115,0.497870,0.727828,
        2.199115,0.499783,-1.654379,
        2.199115,0.502651,0.463333,
        2.199115,0.505463,2.594557,
        2.199115,0.506104,-2.874910,
        2.199115,0.513579,-1.147354,
        2.199115,0.518569,1.644237,
        2.199115,0.519001,1.900488,
        2.199115,0.519277,-0.673809,
        2.199115,0.524769,-0.421955,
        2.199115,0.525564,-2.288838,
        2.199115,0.532541,0.978737,
        2.199115,0.534610,2.154808,
        2.199115,0.535020,-1.891846,
        2.199115,0.540208,1.400497,
        2.199115,0.542006,0.228337,
        2.199115,0.550762,-0.180771,
        2.199115,0.560531,2.378474,
        2.199115,0.561629,-0.924576,
        2.199115,0.567465,-2.679392,
        2.199115,0.588527,1.181755,
        2.199115,0.591643,-2.098364,
        2.199115,0.594963,2.997575,
        2.199115,0.599643,-3.045235,
        2.199115,0.599962,2.773556,
        2.199115,0.602620,0.029270,
        2.199115,0.605711,0.598454,
        2.199115,0.607708,-1.541026,
        2.199115,0.609360,-2.464379,
        2.199115,0.610738,-1.326786,
        2.199115,0.616968,0.820489,
        2.199115,0.626379,1.773610,
        2.199115,0.628626,-1.751180,
        2.199115,0.630671,0.391949,
        2.199115,0.632622,-0.737642,
        2.199115,0.632708,2.566722,
        2.199115,0.632797,1.989858,
        2.199115,0.634455,-0.534790,
        2.199115,0.636205,1.561170,
        2.199115,0.637681,-1.123423,
        2.199115,0.641441,-2.842109,
        2.199115,0.644807,-0.327139,
        2.199115,0.652440,2.197129,
        2.199115,0.660897,-2.278593,
        2.199115,0.664008,1.360091,
        2.199115,0.670850,-1.945257,
        2.199115,0.671396,1.004859,
        2.199115,0.673985,0.203855,
        2.199115,0.680326,-0.133370,
        2.199115,0.685083,-0.942363,
        2.199115,0.688204,-2.632204,
        2.199115,0.689777,2.385011,
        2.199115,0.700065,3.115430,
        2.199115,0.704080,2.911636,
        2.199115,0.712883,1.174216,
        2.199115,0.715701,0.699906,
        2.199115,0.718851,-1.442215,
        2.199115,0.721019,2.721536,
        2.199115,0.725300,-2.975134,
        2.199115,0.727136,-1.633288,
        2.199115,0.727399,-2.120567,
        2.199115,0.730751,0.522488,
        2.199115,0.732892,-1.254330,
        2.199115,0.734333,-2.446184,
        2.199115,0.735125,0.037270,
        2.199115,0.737893,1.866463,
        2.199115,0.740650,1.681075,
        2.199115,0.745971,-0.781286,
        2.199115,0.749432,-0.442559,
        2.199115,0.750102,2.050636,
        2.199115,0.751046,0.863821,
        2.199115,0.757018,-0.608833,
        2.199115,0.757046,-1.816019,
        2.199115,0.757389,1.501184,
        2.199115,0.759800,2.546181,
        2.199115,0.759859,0.351923,
        2.199115,0.761141,-2.781102,
        2.199115,0.769035,-1.078574,
        2.199115,0.770166,-0.266604,
        2.199115,0.772218,2.227380,
        2.199115,0.788382,1.331051,
        2.199115,0.790269,-2.282409,
        2.199115,0.803410,-1.984567,
        2.199115,0.803813,0.191520,
        2.199115,0.806832,3.022155,
        2.199115,0.807476,-2.597675,
        2.199115,0.812512,-0.107335,
        2.199115,0.814161,1.010949,
        2.199115,0.817732,-3.084486,
        2.199115,0.817846,2.390688,
        2.199115,0.819323,2.845960,
        2.199115,0.819430,-0.915017,
        2.199115,0.829198,-1.532491,
        2.199115,0.829531,0.665695,
        2.199115,0.829816,1.167147,
        2.199115,0.833613,-1.357492,
        2.199115,0.842945,-2.907318,
        2.199115,0.844883,2.679708,
        2.199115,0.848189,-1.705003,
        2.199115,0.849888,1.767901,
        2.199115,0.851126,0.493370,
        2.199115,0.852911,1.933394,
        2.199115,0.857516,-2.432863,
        2.199115,0.858261,-1.191729,
        2.199115,0.858279,-2.138895,
        2.199115,0.859723,-0.734928,
        2.199115,0.859754,1.603292,
        2.199115,0.861655,-0.394821,
        2.199115,0.864300,0.043549,
        2.199115,0.866940,0.851229,
        2.199115,0.869748,2.095969,
        2.199115,0.877352,-0.574668,
        2.199115,0.879254,-2.733338,
        2.199115,0.882554,1.443757,
        2.199115,0.884060,0.334199,
        2.199115,0.885626,-1.866061,
        2.199115,0.889834,2.529957,
        2.199115,0.893140,2.251698,
        2.199115,0.897463,-0.240822,
        2.199115,0.898758,-1.031948,
        2.199115,0.913918,1.287618,
        2.199115,0.916555,-2.287173,
        2.199115,0.918305,3.107845,
        2.199115,0.919522,2.948277,
        2.199115,0.927138,-2.569686,
        2.199115,0.929031,0.185383,
        2.199115,0.933766,-3.015917,
        2.199115,0.934873,-2.010822,
        2.199115,0.935636,-1.445855,
        2.199115,0.936702,-0.866461,
        2.199115,0.937395,0.729108,
        2.199115,0.937910,2.787575,
        2.199115,0.939144,-0.092962,
        2.199115,0.941718,1.125085,
        2.199115,0.942873,-1.607680,
        2.199115,0.943701,0.980328,
        2.199115,0.947774,2.395256,
        2.199115,0.952433,0.584347,
        2.199115,0.954456,-1.287293,
        2.199115,0.958731,-2.852552,
        2.199115,0.961838,1.839284,
        2.199115,0.961898,-0.469425,
        2.199115,0.965426,1.686615,
        2.199115,0.970108,1.990771,
        2.199115,0.972820,-1.762780,
        2.199115,0.973500,-0.711881,
        2.199115,0.975253,2.641082,
        2.199115,0.977824,0.441928,
        2.199115,0.981485,-2.424766,
        2.199115,0.981877,1.537100,
        2.199115,0.982609,-1.136920,
        2.199115,0.984265,-2.156073,
        2.199115,0.989150,0.045493,
        2.199115,0.993052,2.137478,
        2.199115,0.997766,-2.693871,
        2.199115,0.999847,-0.334738,
        2.199115,1.001920,0.848054,
        2.199115,1.006146,1.389734,
        2.199115,1.012180,0.304066,
        2.199115,1.016938,2.270674,
        2.199115,1.017371,-0.979790,
        2.199115,1.019495,-1.894487,
        2.199115,1.023796,-0.583691,
        2.199115,1.026525,2.506945,
        2.199115,1.027004,-0.203653,
        2.199115,1.028027,3.036573,
        2.199115,1.031435,1.241021,
        2.199115,1.032316,2.884221,
        2.199115,1.035863,-3.110475,
        2.199115,1.042085,-2.292950,
        2.199115,1.043272,-1.522986,
        2.199115,1.046265,-2.958847,
        2.199115,1.049272,-1.376344,
        2.199115,1.050039,-2.552086,
        2.199115,1.053682,0.689863,
        2.199115,1.055458,0.170394,
        2.199115,1.056622,-0.831866,
        2.199115,1.057515,1.097998,
        2.199115,1.058706,-2.036438,
        2.199115,1.061389,-1.672248,
        2.199115,1.063758,2.743093,
        2.199115,1.070157,-0.073422,
        2.199115,1.072154,1.761506,
        2.199115,1.074029,1.909193,
        2.199115,1.074204,0.545276,
        2.199115,1.074714,-2.805147,
        2.199115,1.076262,0.966104,
        2.199115,1.077642,-1.227073,
        2.199115,1.084151,1.618525,
        2.199115,1.085085,2.377673,
        2.199115,1.086822,-0.441112,
        2.199115,1.087142,2.058815,
        2.199115,1.098383,-0.696430,
        2.199115,1.099138,-1.084033,
        2.199115,1.103208,1.477773,
        2.199115,1.103783,0.408254,
        2.199115,1.106275,-2.420800,
        2.199115,1.108079,2.612943,
        2.199115,1.109345,-2.172718,
        2.199115,1.115957,-1.787597,
        2.199115,1.118930,0.045983,
        2.199115,1.119777,0.834702,
        2.199115,1.121943,-2.669042,
        2.199115,1.124930,2.219917,
        2.199115,1.125217,1.336294,
        2.199115,1.130958,2.972931,
        2.199115,1.131400,-0.306629,
        2.199115,1.137135,-0.936514,
        2.199115,1.139125,-1.924528,
        2.199115,1.139877,0.276382,
        2.199115,1.142003,-3.058022,
        2.199115,1.147363,-0.558448,
        2.199115,1.150348,-0.176418,
        2.199115,1.151198,1.197016,
        2.199115,1.153930,-1.597104,
        2.199115,1.155389,2.837148,
        2.199115,1.155458,-1.460623,
        2.199115,1.157331,-2.906362,
        2.199115,1.158594,2.487618,
        2.199115,1.162103,-1.322553,
        2.199115,1.165407,3.092861,
        2.199115,1.168025,0.717075,
        2.199115,1.168406,-2.301519,
        2.199115,1.174782,-2.538755,
        2.199115,1.177998,1.061256,
        2.199115,1.178898,1.827305,
        2.199115,1.181331,-2.060330,
        2.199115,1.182540,-0.798952,
        2.199115,1.184569,0.147517,
        2.199115,1.184734,1.968919,
        2.199115,1.184804,0.593386,
        2.199115,1.188278,1.690808,
        2.199115,1.194360,2.110440,
        2.199115,1.194776,2.710455,
        2.199115,1.197745,-1.173705,
        2.199115,1.199898,-2.777257,
        2.199115,1.201164,2.355931,
        2.199115,1.203760,1.556404,
        2.199115,1.205845,-0.415993,
        2.199115,1.206585,0.933648,
        2.199115,1.206950,-0.037248,
        2.199115,1.209341,0.471162,
        2.199115,1.213483,-1.039417,
        2.199115,1.218146,-0.672643,
        2.199115,1.221608,1.421669,
        2.199115,1.226043,-1.693533,
        2.199115,1.232870,-1.822773,
        2.199115,1.234375,-2.421194,
        2.199115,1.234531,-2.190148,
        2.199115,1.239414,0.348959,
        2.199115,1.240183,2.587255,
        2.199115,1.242312,2.941234,
        2.199115,1.244075,1.286644,
        2.199115,1.245961,-2.997461,
        2.199115,1.246495,-2.646872,
        2.199115,1.247736,2.242810,
        2.199115,1.253604,0.812222,
        2.199115,1.257816,-0.894317,
        2.199115,1.258247,-1.953634,
        2.199115,1.258993,-0.284099,
        2.199115,1.262380,-0.157609,
        2.199115,1.262791,-1.549229,
        2.199115,1.263772,-0.533073,
        2.199115,1.266731,-1.412461,
        2.199115,1.268591,-3.122146,
        2.199115,1.269614,-1.275246,
        2.199115,1.270745,1.152533,
        2.199115,1.273606,0.224545,
        2.199115,1.279716,2.812101,
        2.199115,1.283786,2.462124,
        2.199115,1.284798,0.078209,
        2.199115,1.285402,-2.882049,
        2.199115,1.286936,1.886053,
        2.199115,1.289408,0.683122,
        2.199115,1.294736,1.756184,
        2.199115,1.296978,-2.311511,
        2.199115,1.298546,2.020601,
        2.199115,1.300924,1.022478,
        2.199115,1.303305,3.043152,
        2.199115,1.304210,-2.084225,
        2.199115,1.305241,-2.530249,
        2.199115,1.307513,1.628003,
        2.199115,1.310722,0.558267,
        2.199115,1.314142,2.139329,
        2.199115,1.316087,-0.777971,
        2.199115,1.317907,-1.121023,
        2.199115,1.321019,-2.749776,
        2.199115,1.321609,2.687717,
        2.199115,1.322568,1.499415,
        2.199115,1.322972,-0.395699,
        2.199115,1.325286,-0.994192,
        2.199115,1.329146,-0.648513,
        2.199115,1.329795,-0.042570,
        2.199115,1.331651,2.344497,
        2.199115,1.336522,0.435098,
        2.199115,1.339410,0.900087,
        2.199115,1.339550,1.370353,
        2.199115,1.340441,-1.731580,
        2.199115,1.351855,-1.855234,
        2.199115,1.360161,-3.018404,
        2.199115,1.361031,-1.613971,
        2.199115,1.361196,2.914780,
        2.199115,1.361998,-2.207254,
        2.199115,1.362302,1.239384,
        2.199115,1.365228,2.563808,
        2.199115,1.365452,0.310522,
        2.199115,1.367028,-2.420041,
        2.199115,1.373422,-1.484281,
        2.199115,1.374245,-1.226417,
        2.199115,1.375020,-0.170393,
        2.199115,1.376560,-1.355481,
        2.199115,1.377070,-1.980984,
        2.199115,1.377578,-2.636339,
        2.199115,1.378096,0.774626,
        2.199115,1.379086,-0.508192,
        2.199115,1.382562,0.177141,
        2.199115,1.388849,2.233847,
        2.199115,1.389268,-0.886349,
        2.199115,1.391324,1.109396,
        2.199115,1.391699,-0.288725,
        2.199115,1.392869,3.136903,
        2.199115,1.393612,1.952016,
        2.199115,1.395545,-2.850807,
        2.199115,1.401926,1.822930,
        2.199115,1.402149,2.788898,
        2.199115,1.409552,0.647980,
        2.199115,1.410840,0.056219,
        2.199115,1.411944,2.444377,
        2.199115,1.413639,1.697982,
        2.199115,1.426479,1.573427,
        2.199115,1.426725,2.102242,
        2.199115,1.427868,0.984637,
        2.199115,1.429580,-2.107232,
        2.199115,1.431375,-2.313269,
        2.199115,1.431907,3.015410,
        2.199115,1.432325,-1.029054,
        2.199115,1.433341,0.524268,
        2.199115,1.435181,-0.739771,
        2.199115,1.437208,-2.529085,
        2.199115,1.439041,-0.617966,
        2.199115,1.441473,1.448544,
        2.199115,1.444527,2.664744,
        2.199115,1.450513,-0.400625,
        2.199115,1.451907,-0.066823,
        2.199115,1.452754,-2.738631,
        2.199115,1.454061,-2.950416,
        2.199115,1.456736,1.323154,
        2.199115,1.458370,0.400433,
        2.199115,1.459314,-1.765003,
        2.199115,1.459902,-1.145835,
        2.199115,1.464952,2.329245,
        2.199115,1.466050,0.862104,
        2.199115,1.471330,-1.886160,
        2.199115,1.475119,-1.648255,
        2.199115,1.480878,1.194576,
        2.199115,1.480942,0.273235,
        2.199115,1.481086,2.890269,
        2.199115,1.484118,-1.529156,
        2.199115,1.484741,-1.279219,
        2.199115,1.487681,-1.406069,
        2.199115,1.487886,-3.066037,
        2.199115,1.490555,2.543284,
        2.199115,1.492424,-0.190932,
        2.199115,1.495002,-0.839810,
        2.199115,1.495297,2.007188,
        2.199115,1.496414,-2.009267,
        2.199115,1.498261,-2.425270,
        2.199115,1.498907,-2.207327,
        2.199115,1.500204,0.737638,
        2.199115,1.502469,0.151156,
        2.199115,1.507448,-0.514782,
        2.199115,1.509287,2.204647,
        2.199115,1.510518,1.884920,
        2.199115,1.511948,-2.632360,
        2.199115,1.515235,1.068565,
        2.199115,1.522237,3.102018,
        2.199115,1.522373,1.764991,
        2.199115,1.522416,-0.954956,
        2.199115,1.523102,2.765920,
        2.199115,1.524811,-2.844718,
        2.199115,1.527261,-0.304916,
        2.199115,1.529788,0.613208,
        2.199115,1.533648,1.644808,
        2.199115,1.535018,0.028869,
        2.199115,1.541082,2.424608,
        2.199115,1.546010,1.523462,
        2.199115,1.551135,-0.642994,
        2.199115,1.553895,0.489607,
        2.199115,1.555041,0.947813,
        2.199115,1.559671,-1.074934,
        2.199115,1.560345,2.985899,
        2.199115,1.560487,-2.323400,
        2.199115,1.561127,1.401111,
        2.199115,1.567193,2.643172,
        2.199115,1.568892,2.098362,
        2.199115,1.571702,-2.098435,
        2.199115,1.572701,-2.528730,
        2.199115,1.573572,1.278593,
        2.199115,1.574849,-0.091505,
        2.199115,1.576334,0.365236,
        2.199115,1.577029,-2.962888,
        2.199115,1.577688,-1.814941,
        2.199115,1.579542,-0.755917,
        2.199115,1.579624,-1.199759,
        2.199115,1.581515,-0.419414,
        2.199115,1.584581,-2.738702,
        2.199115,1.586948,-1.696296,
        2.199115,1.588901,2.301496,
        2.199115,1.589936,0.825020,
        2.199115,1.590018,-1.941652,
        2.199115,1.596968,-1.328358,
        2.199115,1.597178,-1.575607,
        2.199115,1.600470,-1.453196,
        2.199115,1.601408,0.242409,
        2.199115,1.602259,1.152656,
        2.199115,1.603064,2.865774,
        2.199115,1.612085,1.968802,
        2.199115,1.612720,-3.080498,
        2.199115,1.616061,2.521531,
        2.199115,1.616900,-0.207457,
        2.199115,1.620988,-0.875920,
        2.199115,1.622114,0.701509,
        2.199115,1.622749,0.121426,
        2.199115,1.622993,-2.221629,
        2.199115,1.629346,-0.542669,
        2.199115,1.630007,1.844552,
        2.199115,1.634747,-2.426198,
        2.199115,1.642346,1.722617,
        2.199115,1.643925,1.034629,
        2.199115,1.644848,2.743774,
        2.199115,1.645258,2.188109,
        2.199115,1.645417,-2.634725,
        2.199115,1.647395,3.085437,
        2.199115,1.648538,-2.855044,
        2.199115,1.649932,0.576808,
        2.199115,1.652224,-0.996105,
        2.199115,1.652871,1.600681,
        2.199115,1.657997,-0.325127,
        2.199115,1.661936,0.001037,
        2.199115,1.665187,1.477402,
        2.199115,1.665496,2.397872,
        2.199115,1.673267,-0.665453,
        2.199115,1.673557,0.452526,
        2.199115,1.679277,-1.122421,
        2.199115,1.679696,0.912482,
        2.199115,1.682420,1.353922,
        2.199115,1.685031,-2.009038,
        2.199115,1.685196,2.967046,
        2.199115,1.690107,2.620910,
        2.199115,1.690453,-2.123076,
        2.199115,1.691693,-1.871572,
        2.199115,1.691817,2.061688,
        2.199115,1.692644,1.237670,
        2.199115,1.696386,0.328123,
        2.199115,1.697162,-1.249026,
        2.199115,1.697488,-2.322785,
        2.199115,1.698726,-2.973576,
        2.199115,1.699172,-1.749717,
        2.199115,1.704478,-0.106295,
        2.199115,1.707244,-0.444770,
        2.199115,1.707500,-2.531497,
        2.199115,1.707714,-0.787321,
        2.199115,1.708430,-1.627552,
        2.199115,1.710583,-1.376890,
        2.199115,1.712415,-2.747921,
        2.199115,1.713125,-1.503073,
        2.199115,1.713452,0.789419,
        2.199115,1.720612,2.279888,
        2.199115,1.724852,0.202831,
        2.199115,1.724868,2.846246,
        2.199115,1.728940,1.931535,
        2.199115,1.732838,1.127584,
        2.199115,1.735719,-3.092485,
        2.199115,1.739676,2.496017,
        2.199115,1.739971,-0.228876,
        2.199115,1.743328,0.663347,
        2.199115,1.744927,-0.911892,
        2.199115,1.745952,0.084733,
        2.199115,1.748077,1.804354,
        2.199115,1.752959,-0.567548,
        2.199115,1.760169,1.678132,
        2.199115,1.763887,-2.218450,
        2.199115,1.766737,2.722752,
        2.199115,1.767999,1.004737,
        2.199115,1.769149,1.550298,
        2.199115,1.770119,-2.427175,
        2.199115,1.770460,2.154208,
        2.199115,1.770514,3.070982,
        2.199115,1.770735,0.537245,
        2.199115,1.772673,-2.864760,
        2.199115,1.773527,-1.039582,
        2.199115,1.776419,-2.642421,
        2.199115,1.780769,1.416330,
        2.199115,1.786634,-0.346247,
        2.199115,1.788829,-1.947553,
        2.199115,1.793737,0.410270,
        2.199115,1.794553,2.374618,
        2.199115,1.794699,-0.694061,
        2.199115,1.797469,-1.169105,
        2.199115,1.803705,0.881076,
        2.199115,1.804124,1.264249,
        2.199115,1.807825,-2.090455,
        2.199115,1.808822,-1.817375,
        2.199115,1.809020,2.949741,
        2.199115,1.812466,2.597595,
        2.199115,1.813159,2.025395,
        2.199115,1.813816,-1.298435,
        2.199115,1.815016,0.278951,
        2.199115,1.816510,-0.008631,
        2.199115,1.817772,-1.689625,
        2.199115,1.821360,-2.984352,
        2.199115,1.822312,-0.127211,
        2.199115,1.823386,-1.429807,
        2.199115,1.823834,-1.560483,
        2.199115,1.832060,-0.821627,
        2.199115,1.832690,-0.468440,
        2.199115,1.833945,-2.320278,
        2.199115,1.834787,0.751968,
        2.199115,1.840579,-2.536445,
        2.199115,1.841033,-2.756932,
        2.199115,1.846350,1.893983,
        2.199115,1.846644,1.113766,
        2.199115,1.846833,2.826543,
        2.199115,1.847325,2.249051,
        2.199115,1.854119,0.134161,
        2.199115,1.858863,-3.106022,
        2.199115,1.865046,1.763384,
        2.199115,1.865619,-0.952036,
        2.199115,1.865912,0.624453,
        2.199115,1.866694,2.473686,
        2.199115,1.867524,-0.240083,
        2.199115,1.875561,1.632831,
        2.199115,1.876049,-0.594378,
        2.199115,1.880509,1.500204,
        2.199115,1.883504,1.360457,
        2.199115,1.886971,-2.196665,
        2.199115,1.888979,0.982922,
        2.199115,1.889815,2.701570,
        2.199115,1.892144,-2.005340,
        2.199115,1.892348,0.496522,
        2.199115,1.893281,-1.083976,
        2.199115,1.893828,3.053453,
        2.199115,1.895117,2.121107,
        2.199115,1.898741,-2.875735,
        2.199115,1.904429,-2.428029,
        2.199115,1.908408,-2.648903,
        2.199115,1.911828,-0.364489,
        2.199115,1.912224,0.366565,
        2.199115,1.914854,1.214468,
        2.199115,1.915467,-1.883143,
        2.199115,1.915763,-0.724679,
        2.199115,1.915790,-1.216366,
        2.199115,1.922256,2.347806,
        2.199115,1.922616,0.844633,
        2.199115,1.925185,0.231170,
        2.199115,1.927474,-1.754616,
        2.199115,1.928342,-1.351174,
        2.199115,1.933328,-1.623328,
        2.199115,1.933878,0.034150,
        2.199115,1.934585,-1.487979,
        2.199115,1.934802,2.927634,
        2.199115,1.934862,1.989776,
        2.199115,1.936327,-0.099350,
        2.199115,1.938088,2.578507,
        2.199115,1.945677,-2.999805,
        2.199115,1.954572,-0.859208,
        2.199115,1.957868,-0.490784,
        2.199115,1.958989,0.712772,
        2.199115,1.960212,-2.307561,
        2.199115,1.964774,1.855507,
        2.199115,1.964971,-2.100533,
        2.199115,1.969112,1.088273,
        2.199115,1.971438,-2.765415,
        2.199115,1.972312,2.805134,
        2.199115,1.974950,2.220596,
        2.199115,1.975135,-2.539127,
        2.199115,1.982107,1.720829,
        2.199115,1.983195,-3.127984,
        2.199115,1.983318,-0.247182,
        2.199115,1.986720,-0.993220,
        2.199115,1.990381,1.586005,
        2.199115,1.990470,0.580283,
        2.199115,1.990523,1.450326,
        2.199115,1.993909,1.312897,
        2.199115,1.994694,2.452813,
        2.199115,1.999226,-0.620717,
        2.199115,2.005145,0.941012,
        2.199115,2.007301,0.142337,
        2.199115,2.014149,0.444138,
        2.199115,2.014338,-1.127392,
        2.199115,2.018131,3.021778,
        2.199115,2.019676,-1.980944,
        2.199115,2.019912,2.686612,
        2.199115,2.020545,2.088672,
        2.199115,2.026074,0.300945,
        2.199115,2.026196,-2.889057,
        2.199115,2.031959,-1.268178,
        2.199115,2.032917,-2.199917,
        2.199115,2.033572,-1.842884,
        2.199115,2.034712,-2.419505,
        2.199115,2.035818,-0.379479,
        2.199115,2.036004,-0.757062,
        2.199115,2.040501,-1.411488,
        2.199115,2.042874,-1.701367,
        2.199115,2.043060,-2.653746,
        2.199115,2.043729,-1.557938,
        2.199115,2.048538,0.806623,
        2.199115,2.049006,1.199930,
        2.199115,2.049502,-0.137989,
        2.199115,2.052181,2.326959,
        2.199115,2.052794,-0.008526,
        2.199115,2.057608,1.951726,
        2.199115,2.065568,2.567730,
        2.199115,2.070371,2.879808,
        2.199115,2.071256,-3.018961,
        2.199115,2.079746,-0.897699,
        2.199115,2.082784,1.042403,
        2.199115,2.083752,-0.511086,
        2.199115,2.084401,1.810502,
        2.199115,2.085731,0.671081,
        2.199115,2.097496,1.373172,
        2.199115,2.099185,1.667006,
        2.199115,2.100482,-2.301084,
        2.199115,2.103118,2.194745,
        2.199115,2.103223,1.521685,
        2.199115,2.103695,-2.777707,
        2.199115,2.107523,3.129581,
        2.199115,2.108277,-2.095907,
        2.199115,2.108891,-2.532166,
        2.199115,2.111003,-0.268632,
        2.199115,2.112276,-1.027246,
        2.199115,2.114380,0.531457,
        2.199115,2.119166,0.100552,
        2.199115,2.119564,2.754818,
        2.199115,2.122474,-0.648314,
        2.199115,2.125183,2.445598,
        2.199115,2.129501,0.238290,
        2.199115,2.131178,0.386663,
        2.199115,2.131223,-1.174311,
        2.199115,2.134269,-1.948042,
        2.199115,2.134684,0.907065,
        2.199115,2.137615,2.988806,
        2.199115,2.145635,-1.327708,
        2.199115,2.146467,2.055081,
        2.199115,2.146582,-1.793788,
        2.199115,2.149210,-1.482321,
        2.199115,2.152901,-1.636788,
        2.199115,2.156337,1.151558,
        2.199115,2.156474,-2.912065,
        2.199115,2.157194,-0.792563,
        2.199115,2.166724,-0.401638,
        2.199115,2.166980,-0.134879,
        2.199115,2.173644,-2.403429,
        2.199115,2.175711,-2.655523,
        2.199115,2.177132,2.600419,
        2.199115,2.178879,0.767722,
        2.199115,2.180752,1.908485,
        2.199115,2.181338,2.313721,
        2.199115,2.185580,1.287673,
        2.199115,2.195400,-2.217546,
        2.199115,2.200418,-3.056442,
        2.199115,2.201829,-0.001735,
        2.199115,2.203804,1.756176,
        2.199115,2.211602,1.443930,
        2.199115,2.212686,-0.544809,
        2.199115,2.213937,0.621892,
        2.199115,2.214075,1.600691,
        2.199115,2.216256,1.020029,
        2.199115,2.219060,2.886631,
        2.199115,2.223375,-0.906185,
        2.199115,2.226002,-1.060248,
        2.199115,2.226413,-2.064034,
        2.199115,2.230554,2.171334,
        2.199115,2.232413,2.735777,
        2.199115,2.233709,3.070830,
        2.199115,2.234383,-2.792065,
        2.199115,2.237183,0.468799,
        2.199115,2.237695,-0.271388,
        2.199115,2.239419,0.151693,
        2.199115,2.243347,-2.521049,
        2.199115,2.244035,-1.229456,
        2.199115,2.244355,0.309899,
        2.199115,2.245175,2.460236,
        2.199115,2.245628,-1.899216,
        2.199115,2.245973,-0.702234,
        2.199115,2.255601,-1.728131,
        2.199115,2.256702,-1.401367,
        2.199115,2.261466,-1.557500,
        2.199115,2.267390,0.878499,
        2.199115,2.272099,2.018594,
        2.199115,2.278618,1.170386,
        2.199115,2.283304,-2.345584,
        2.199115,2.285175,-2.943090,
        2.199115,2.293000,-0.124451,
        2.199115,2.294058,-0.421070,
        2.199115,2.303283,1.855938,
        2.199115,2.304968,2.608978,
        2.199115,2.305347,2.311346,
        2.199115,2.307188,-2.656134,
        2.199115,2.307301,1.336200,
        2.199115,2.309655,0.726020,
        2.199115,2.317634,-2.185225,
        2.199115,2.321510,1.685664,
        2.199115,2.324376,-3.111006,
        2.199115,2.324731,1.511186,
        2.199115,2.329587,0.035498,
        2.199115,2.330158,2.977264,
        2.199115,2.335096,-0.588468,
        2.199115,2.336046,-1.115799,
        2.199115,2.338060,2.803613,
        2.199115,2.338083,-0.943400,
        2.199115,2.340210,0.562145,
        2.199115,2.341931,1.023926,
        2.199115,2.343265,-0.779061,
        2.199115,2.343434,-2.013692,
        2.199115,2.353029,-1.310529,
        2.199115,2.354868,0.211543,
        2.199115,2.355120,0.388593,
        2.199115,2.355165,-1.833173,
        2.199115,2.355462,2.151441,
        2.199115,2.362948,-2.487477,
        2.199115,2.363101,-2.810456,
        2.199115,2.363394,-0.272918,
        2.199115,2.364509,-1.644533,
        2.199115,2.370465,2.466824,
        2.199115,2.386110,-1.481351,
        2.199115,2.394960,0.862133,
        2.199115,2.395716,1.977184,
        2.199115,2.396339,1.200900,
        2.199115,2.407941,-2.316753,
        2.199115,2.409450,-2.988432,
        2.199115,2.417202,-0.103363,
        2.199115,2.417386,-0.443097,
        2.199115,2.417950,2.666662,
        2.199115,2.423190,1.789783,
        2.199115,2.423692,1.392696,
        2.199115,2.431122,3.086379,
        2.199115,2.431406,-2.647820,
        2.199115,2.431584,2.307631,
        2.199115,2.434090,1.592818,
        2.199115,2.435993,0.683844,
        2.199115,2.440898,-2.133300,
        2.199115,2.442222,-1.195837,
        2.199115,2.443112,2.883761,
        2.199115,2.449191,-0.633651,
        2.199115,2.450821,0.084741,
        2.199115,2.453499,-1.008861,
        2.199115,2.460010,-1.938011,
        2.199115,2.460199,-0.822008,
        2.199115,2.460902,0.490079,
        2.199115,2.461774,-1.734913,
        2.199115,2.463986,1.034407,
        2.199115,2.467647,0.287667,
        2.199115,2.479849,2.126925,
        2.199115,2.485907,-2.835471,
        2.199115,2.485945,-1.374853,
        2.199115,2.487747,2.505912,
        2.199115,2.487957,-0.274279,
        2.199115,2.489741,-2.468428,
        2.199115,2.512538,1.239900,
        2.199115,2.517528,1.923289,
        2.199115,2.518454,-1.565852,
        2.199115,2.519348,0.843503,
        2.199115,2.522766,-3.059231,
        2.199115,2.528686,2.737228,
        2.199115,2.534798,-2.266671,
        2.199115,2.536533,1.466274,
        2.199115,2.538504,1.699812,
        2.199115,2.539516,-0.476942,
        2.199115,2.539532,-0.074548,
        2.199115,2.543197,2.990252,
        2.199115,2.554022,2.320619,
        2.199115,2.556919,-2.655005,
        2.199115,2.558146,0.627191,
        2.199115,2.558223,-1.164189,
        2.199115,2.563452,-2.044028,
        2.199115,2.568462,0.150655,
        2.199115,2.569130,-0.700837,
        2.199115,2.570050,-1.804492,
        2.199115,2.572710,-0.934665,
        2.199115,2.575994,0.391718,
        2.199115,2.584774,1.046703,
        2.199115,2.601432,2.102535,
        2.199115,2.603559,2.555114,
        2.199115,2.604150,-2.885208,
        2.199115,2.608630,-1.378011,
        2.199115,2.611256,-0.279110,
        2.199115,2.617344,-2.439480,
        2.199115,2.627402,1.293552,
        2.199115,2.635396,1.848850,
        2.199115,2.635686,2.834950,
        2.199115,2.637087,3.132730,
        2.199115,2.642029,0.813274,
        2.199115,2.645260,1.570203,
        2.199115,2.646182,-1.614036,
        2.199115,2.660113,-2.186513,
        2.199115,2.660483,-0.034608,
        2.199115,2.661713,-0.529511,
        2.199115,2.672720,-1.110422,
        2.199115,2.676626,2.335353,
        2.199115,2.676638,0.540364,
        2.199115,2.676780,-1.901158,
        2.199115,2.681411,-2.667965,
        2.199115,2.682152,0.245751,
        2.199115,2.682859,-0.817071,
        2.199115,2.705107,1.062142,
        2.199115,2.717705,2.625995,
        2.199115,2.721001,2.060097,
        2.199115,2.722319,-2.954069,
        2.199115,2.729840,-1.368237,
        2.199115,2.734349,-0.289298,
        2.199115,2.738409,2.975671,
        2.199115,2.740366,1.374341,
        2.199115,2.744633,-2.391311,
        2.199115,2.747398,1.727166,
        2.199115,2.763799,0.760372,
        2.199115,2.768238,-1.679243,
        2.199115,2.779245,-2.046268,
        2.199115,2.780208,-0.626453,
        2.199115,2.780471,0.027732,
        2.199115,2.784696,-1.013128,
        2.199115,2.789817,0.395431,
        2.199115,2.801154,2.353143,
        2.199115,2.807091,-2.687477,
        2.199115,2.825762,1.085035,
        2.199115,2.830209,2.734943,
        2.199115,2.833517,-3.085508,
        2.199115,2.838301,1.980061,
        2.199115,2.849871,1.512012,
        2.199115,2.850089,-1.343899,
        2.199115,2.856905,-0.318441,
        2.199115,2.870445,-2.291001,
        2.199115,2.885017,0.655562,
        2.199115,2.886428,-1.793217,
        2.199115,2.890337,-0.825633,
        2.199115,2.898932,0.139973,
        2.199115,2.931927,-2.728768,
        2.199115,2.932096,2.377790,
        2.199115,2.937409,2.947245,
        2.199115,2.948048,1.133127,
        2.199115,2.950896,1.797889,
        2.199115,2.970019,-1.282382,
        2.199115,2.978333,-0.412090,
        2.199115,2.996437,-2.089121,
        2.199115,3.011274,0.434110,
        2.199115,3.050410,3.138492,
        2.199115,3.064492,1.647259,
        2.199115,3.088868,-0.989501,
        2.303835,0.053884,-2.408394,
        2.303835,0.077645,1.453433,
        2.303835,0.088408,-0.311160,
        2.303835,0.142314,2.590387,
        2.303835,0.147206,-1.385164,
        2.303835,0.163506,0.573203,
        2.303835,0.165485,-2.875396,
        2.303835,0.182564,-2.125821,
        2.303835,0.190569,1.929141,
        2.303835,0.201102,1.249082,
        2.303835,0.201397,-0.716646,
        2.303835,0.210568,-0.069559,
        2.303835,0.251672,2.848900,
        2.303835,0.266262,-1.678068,
        2.303835,0.269108,-1.215563,
        2.303835,0.273777,-2.506756,
        2.303835,0.279854,0.814581,
        2.303835,0.281664,-2.969147,
        2.303835,0.283625,2.405603,
        2.303835,0.284058,0.357350,
        2.303835,0.294107,1.593880,
        2.303835,0.302491,-0.400410,
        2.303835,0.308059,2.021491,
        2.303835,0.312219,-2.072057,
        2.303835,0.321905,-0.829121,
        2.303835,0.329984,1.201621,
        2.303835,0.339714,-0.013712,
        2.303835,0.368784,2.974974,
        2.303835,0.376428,-1.495532,
        2.303835,0.378103,2.646436,
        2.303835,0.385088,0.590355,
        2.303835,0.386248,-2.690943,
        2.303835,0.389851,-2.340031,
        2.303835,0.391502,-1.171374,
        2.303835,0.398249,-1.812694,
        2.303835,0.400542,-2.995729,
        2.303835,0.404273,0.918163,
        2.303835,0.405617,1.773533,
        2.303835,0.409009,-0.570243,
        2.303835,0.410700,0.270433,
        2.303835,0.415893,1.464890,
        2.303835,0.420680,2.088164,
        2.303835,0.424766,-0.258729,
        2.303835,0.426507,2.367872,
        2.303835,0.441292,-0.890727,
        2.303835,0.451839,-2.078193,
        2.303835,0.459270,1.191562,
        2.303835,0.470622,0.014002,
        2.303835,0.481292,-2.529098,
        2.303835,0.482460,2.849843,
        2.303835,0.492205,-1.397963,
        2.303835,0.493314,3.117264,
        2.303835,0.497870,0.727828,
        2.303835,0.499783,-1.654379,
        2.303835,0.502651,0.463333,
        2.303835,0.505463,2.594557,
        2.303835,0.506104,-2.874910,
        2.303835,0.513579,-1.147354,
        2.303835,0.518569,1.644237,
        2.303835,0.519001,1.900488,
        2.303835,0.519277,-0.673809,
        2.303835,0.524769,-0.421955,
        2.303835,0.525564,-2.288838,
        2.303835,0.532541,0.978737,
        2.303835,0.534610,2.154808,
        2.303835,0.535020,-1.891846,
        2.303835,0.540208,1.400497,
        2.303835,0.542006,0.228337,
        2.303835,0.550762,-0.180771,
        2.303835,0.560531,2.378474,
        2.303835,0.561629,-0.924576,
        2.303835,0.567465,-2.679392,
        2.303835,0.588527,1.181755,
        2.303835,0.591643,-2.098364,
        2.303835,0.594963,2.997575,
        2.303835,0.599643,-3.045235,
        2.303835,0.599962,2.773556,
        2.303835,0.602620,0.029270,
        2.303835,0.605711,0.598454,
        2.303835,0.607708,-1.541026,
        2.303835,0.609360,-2.464379,
        2.303835,0.610738,-1.326786,
        2.303835,0.616968,0.820489,
        2.303835,0.626379,1.773610,
        2.303835,0.628626,-1.751180,
        2.303835,0.630671,0.391949,
        2.303835,0.632622,-0.737642,
        2.303835,0.632708,2.566722,
        2.303835,0.632797,1.989858,
        2.303835,0.634455,-0.534790,
        2.303835,0.636205,1.561170,
        2.303835,0.637681,-1.123423,
        2.303835,0.641441,-2.842109,
        2.303835,0.644807,-0.327139,
        2.303835,0.652440,2.197129,
        2.303835,0.660897,-2.278593,
        2.303835,0.664008,1.360091,
        2.303835,0.670850,-1.945257,
        2.303835,0.671396,1.004859,
        2.303835,0.673985,0.203855,
        2.303835,0.680326,-0.133370,
        2.303835,0.685083,-0.942363,
        2.303835,0.688204,-2.632204,
        2.303835,0.689777,2.385011,
        2.303835,0.700065,3.115430,
        2.303835,0.704080,2.911636,
        2.303835,0.712883,1.174216,
        2.303835,0.715701,0.699906,
        2.303835,0.718851,-1.442215,
        2.303835,0.721019,2.721536,
        2.303835,0.725300,-2.975134,
        2.303835,0.727136,-1.633288,
        2.303835,0.727399,-2.120567,
        2.303835,0.730751,0.522488,
        2.303835,0.732892,-1.254330,
        2.303835,0.734333,-2.446184,
        2.303835,0.735125,0.037270,
        2.303835,0.737893,1.866463,
        2.303835,0.740650,1.681075,
        2.303835,0.745971,-0.781286,
        2.303835,0.749432,-0.442559,
        2.303835,0.750102,2.050636,
        2.303835,0.751046,0.863821,
        2.303835,0.757018,-0.608833,
        2.303835,0.757046,-1.816019,
        2.303835,0.757389,1.501184,
        2.303835,0.759800,2.546181,
        2.303835,0.759859,0.351923,
        2.303835,0.761141,-2.781102,
        2.303835,0.769035,-1.078574,
        2.303835,0.770166,-0.266604,
        2.303835,0.772218,2.227380,
        2.303835,0.788382,1.331051,
        2.303835,0.790269,-2.282409,
        2.303835,0.803410,-1.984567,
        2.303835,0.803813,0.191520,
        2.303835,0.806832,3.022155,
        2.303835,0.807476,-2.597675,
        2.303835,0.812512,-0.107335,
        2.303835,0.814161,1.010949,
        2.303835,0.817732,-3.084486,
        2.303835,0.817846,2.390688,
        2.303835,0.819323,2.845960,
        2.303835,0.819430,-0.915017,
        2.303835,0.829198,-1.532491,
        2.303835,0.829531,0.665695,
        2.303835,0.829816,1.167147,
        2.303835,0.833613,-1.357492,
        2.303835,0.842945,-2.907318,
        2.303835,0.844883,2.679708,
        2.303835,0.848189,-1.705003,
        2.303835,0.849888,1.767901,
        2.303835,0.851126,0.493370,
        2.303835,0.852911,1.933394,
        2.303835,0.857516,-2.432863,
        2.303835,0.858261,-1.191729,
        2.303835,0.858279,-2.138895,
        2.303835,0.859723,-0.734928,
        2.303835,0.859754,1.603292,
        2.303835,0.861655,-0.394821,
        2.303835,0.864300,0.043549,
        2.303835,0.866940,0.851229,
        2.303835,0.869748,2.095969,
        2.303835,0.877352,-0.574668,
        2.303835,0.879254,-2.733338,
        2.303835,0.882554,1.443757,
        2.303835,0.884060,0.334199,
        2.303835,0.885626,-1.866061,
        2.303835,0.889834,2.529957,
        2.303835,0.893140,2.251698,
        2.303835,0.897463,-0.240822,
        2.303835,0.898758,-1.031948,
        2.303835,0.913918,1.287618,
        2.303835,0.916555,-2.287173,
        2.303835,0.918305,3.107845,
        2.303835,0.919522,2.948277,
        2.303835,0.927138,-2.569686,
        2.303835,0.929031,0.185383,
        2.303835,0.933766,-3.015917,
        2.303835,0.934873,-2.010822,
        2.303835,0.935636,-1.445855,
        2.303835,0.936702,-0.866461,
        2.303835,0.937395,0.729108,
        2.303835,0.937910,2.787575,
        2.303835,0.939144,-0.092962,
        2.303835,0.941718,1.125085,
        2.303835,0.942873,-1.607680,
        2.303835,0.943701,0.980328,
        2.303835,0.947774,2.395256,
        2.303835,0.952433,0.584347,
        2.303835,0.954456,-1.287293,
        2.303835,0.958731,-2.852552,
        2.303835,0.961838,1.839284,
        2.303835,0.961898,-0.469425,
        2.303835,0.965426,1.686615,
        2.303835,0.970108,1.990771,
        2.303835,0.972820,-1.762780,
        2.303835,0.973500,-0.711881,
        2.303835,0.975253,2.641082,
        2.303835,0.977824,0.441928,
        2.303835,0.981485,-2.424766,
        2.303835,0.981877,1.537100,
        2.303835,0.982609,-1.136920,
        2.303835,0.984265,-2.156073,
        2.303835,0.989150,0.045493,
        2.303835,0.993052,2.137478,
        2.303835,0.997766,-2.693871,
        2.303835,0.999847,-0.334738,
        2.303835,1.001920,0.848054,
        2.303835,1.006146,1.389734,
        2.303835,1.012180,0.304066,
        2.303835,1.016938,2.270674,
        2.303835,1.017371,-0.979790,
        2.303835,1.019495,-1.894487,
        2.303835,1.023796,-0.583691,
        2.303835,1.026525,2.506945,
        2.303835,1.027004,-0.203653,
        2.303835,1.028027,3.036573,
        2.303835,1.031435,1.241021,
        2.303835,1.032316,2.884221,
        2.303835,1.035863,-3.110475,
        2.303835,1.042085,-2.292950,
        2.303835,1.043272,-1.522986,
        2.303835,1.046265,-2.958847,
        2.303835,1.049272,-1.376344,
        2.303835,1.050039,-2.552086,
        2.303835,1.053682,0.689863,
        2.303835,1.055458,0.170394,
        2.303835,1.056622,-0.831866,
        2.303835,1.057515,1.097998,
        2.303835,1.058706,-2.036438,
        2.303835,1.061389,-1.672248,
        2.303835,1.063758,2.743093,
        2.303835,1.070157,-0.073422,
        2.303835,1.072154,1.761506,
        2.303835,1.074029,1.909193,
        2.303835,1.074204,0.545276,
        2.303835,1.074714,-2.805147,
        2.303835,1.076262,0.966104,
        2.303835,1.077642,-1.227073,
        2.303835,1.084151,1.618525,
        2.303835,1.085085,2.377673,
        2.303835,1.086822,-0.441112,
        2.303835,1.087142,2.058815,
        2.303835,1.098383,-0.696430,
        2.303835,1.099138,-1.084033,
        2.303835,1.103208,1.477773,
        2.303835,1.103783,0.408254,
        2.303835,1.106275,-2.420800,
        2.303835,1.108079,2.612943,
        2.303835,1.109345,-2.172718,
        2.303835,1.115957,-1.787597,
        2.303835,1.118930,0.045983,
        2.303835,1.119777,0.834702,
        2.303835,1.121943,-2.669042,
        2.303835,1.124930,2.219917,
        2.303835,1.125217,1.336294,
        2.303835,1.130958,2.972931,
        2.303835,1.131400,-0.306629,
        2.303835,1.137135,-0.936514,
        2.303835,1.139125,-1.924528,
        2.303835,1.139877,0.276382,
        2.303835,1.142003,-3.058022,
        2.303835,1.147363,-0.558448,
        2.303835,1.150348,-0.176418,
        2.303835,1.151198,1.197016,
        2.303835,1.153930,-1.597104,
        2.303835,1.155389,2.837148,
        2.303835,1.155458,-1.460623,
        2.303835,1.157331,-2.906362,
        2.303835,1.158594,2.487618,
        2.303835,1.162103,-1.322553,
        2.303835,1.165407,3.092861,
        2.303835,1.168025,0.717075,
        2.303835,1.168406,-2.301519,
        2.303835,1.174782,-2.538755,
        2.303835,1.177998,1.061256,
        2.303835,1.178898,1.827305,
        2.303835,1.181331,-2.060330,
        2.303835,1.182540,-0.798952,
        2.303835,1.184569,0.147517,
        2.303835,1.184734,1.968919,
        2.303835,1.184804,0.593386,
        2.303835,1.188278,1.690808,
        2.303835,1.194360,2.110440,
        2.303835,1.194776,2.710455,
        2.303835,1.197745,-1.173705,
        2.303835,1.199898,-2.777257,
        2.303835,1.201164,2.355931,
        2.303835,1.203760,1.556404,
        2.303835,1.205845,-0.415993,
        2.303835,1.206585,0.933648,
        2.303835,1.206950,-0.037248,
        2.303835,1.209341,0.471162,
        2.303835,1.213483,-1.039417,
        2.303835,1.218146,-0.672643,
        2.303835,1.221608,1.421669,
        2.303835,1.226043,-1.693533,
        2.303835,1.232870,-1.822773,
        2.303835,1.234375,-2.421194,
        2.303835,1.234531,-2.190148,
        2.303835,1.239414,0.348959,
        2.303835,1.240183,2.587255,
        2.303835,1.242312,2.941234,
        2.303835,1.244075,1.286644,
        2.303835,1.245961,-2.997461,
        2.303835,1.246495,-2.646872,
        2.303835,1.247736,2.242810,
        2.303835,1.253604,0.812222,
        2.303835,1.257816,-0.894317,
        2.303835,1.258247,-1.953634,
        2.303835,1.258993,-0.284099,
        2.303835,1.262380,-0.157609,
        2.303835,1.262791,-1.549229,
        2.303835,1.263772,-0.533073,
        2.303835,1.266731,-1.412461,
        2.303835,1.268591,-3.122146,
        2.303835,1.269614,-1.275246,
        2.303835,1.270745,1.152533,
        2.303835,1.273606,0.224545,
        2.303835,1.279716,2.812101,
        2.303835,1.283786,2.462124,
        2.303835,1.284798,0.078209,
        2.303835,1.285402,-2.882049,
        2.303835,1.286936,1.886053,
        2.303835,1.289408,0.683122,
        2.303835,1.294736,1.756184,
        2.303835,1.296978,-2.311511,
        2.303835,1.298546,2.020601,
        2.303835,1.300924,1.022478,
        2.303835,1.303305,3.043152,
        2.303835,1.304210,-2.084225,
        2.303835,1.305241,-2.530249,
        2.303835,1.307513,1.628003,
        2.303835,1.310722,0.558267,
        2.303835,1.314142,2.139329,
        2.303835,1.316087,-0.777971,
        2.303835,1.317907,-1.121023,
        2.303835,1.321019,-2.749776,
        2.303835,1.321609,2.687717,
        2.303835,1.322568,1.499415,
        2.303835,1.322972,-0.395699,
        2.303835,1.325286,-0.994192,
        2.303835,1.329146,-0.648513,
        2.303835,1.329795,-0.042570,
        2.303835,1.331651,2.344497,
        2.303835,1.336522,0.435098,
        2.303835,1.339410,0.900087,
        2.303835,1.339550,1.370353,
        2.303835,1.340441,-1.731580,
        2.303835,1.351855,-1.855234,
        2.303835,1.360161,-3.018404,
        2.303835,1.361031,-1.613971,
        2.303835,1.361196,2.914780,
        2.303835,1.361998,-2.207254,
        2.303835,1.362302,1.239384,
        2.303835,1.365228,2.563808,
        2.303835,1.365452,0.310522,
        2.303835,1.367028,-2.420041,
        2.303835,1.373422,-1.484281,
        2.303835,1.374245,-1.226417,
        2.303835,1.375020,-0.170393,
        2.303835,1.376560,-1.355481,
        2.303835,1.377070,-1.980984,
        2.303835,1.377578,-2.636339,
        2.303835,1.378096,0.774626,
        2.303835,1.379086,-0.508192,
        2.303835,1.382562,0.177141,
        2.303835,1.388849,2.233847,
        2.303835,1.389268,-0.886349,
        2.303835,1.391324,1.109396,
        2.303835,1.391699,-0.288725,
        2.303835,1.392869,3.136903,
        2.303835,1.393612,1.952016,
        2.303835,1.395545,-2.850807,
        2.303835,1.401926,1.822930,
        2.303835,1.402149,2.788898,
        2.303835,1.409552,0.647980,
        2.303835,1.410840,0.056219,
        2.303835,1.411944,2.444377,
        2.303835,1.413639,1.697982,
        2.303835,1.426479,1.573427,
        2.303835,1.426725,2.102242,
        2.303835,1.427868,0.984637,
        2.303835,1.429580,-2.107232,
        2.303835,1.431375,-2.313269,
        2.303835,1.431907,3.015410,
        2.303835,1.432325,-1.029054,
        2.303835,1.433341,0.524268,
        2.303835,1.435181,-0.739771,
        2.303835,1.437208,-2.529085,
        2.303835,1.439041,-0.617966,
        2.303835,1.441473,1.448544,
        2.303835,1.444527,2.664744,
        2.303835,1.450513,-0.400625,
        2.303835,1.451907,-0.066823,
        2.303835,1.452754,-2.738631,
        2.303835,1.454061,-2.950416,
        2.303835,1.456736,1.323154,
        2.303835,1.458370,0.400433,
        2.303835,1.459314,-1.765003,
        2.303835,1.459902,-1.145835,
        2.303835,1.464952,2.329245,
        2.303835,1.466050,0.862104,
        2.303835,1.471330,-1.886160,
        2.303835,1.475119,-1.648255,
        2.303835,1.480878,1.194576,
        2.303835,1.480942,0.273235,
        2.303835,1.481086,2.890269,
        2.303835,1.484118,-1.529156,
        2.303835,1.484741,-1.279219,
        2.303835,1.487681,-1.406069,
        2.303835,1.487886,-3.066037,
        2.303835,1.490555,2.543284,
        2.303835,1.492424,-0.190932,
        2.303835,1.495002,-0.839810,
        2.303835,1.495297,2.007188,
        2.303835,1.496414,-2.009267,
        2.303835,1.498261,-2.425270,
        2.303835,1.498907,-2.207327,
        2.303835,1.500204,0.737638,
        2.303835,1.502469,0.151156,
        2.303835,1.507448,-0.514782,
        2.303835,1.509287,2.204647,
        2.303835,1.510518,1.884920,
        2.303835,1.511948,-2.632360,
        2.303835,1.515235,1.068565,
        2.303835,1.522237,3.102018,
        2.303835,1.522373,1.764991,
        2.303835,1.522416,-0.954956,
        2.303835,1.523102,2.765920,
        2.303835,1.524811,-2.844718,
        2.303835,1.527261,-0.304916,
        2.303835,1.529788,0.613208,
        2.303835,1.533648,1.644808,
        2.303835,1.535018,0.028869,
        2.303835,1.541082,2.424608,
        2.303835,1.546010,1.523462,
        2.303835,1.551135,-0.642994,
        2.303835,1.553895,0.489607,
        2.303835,1.555041,0.947813,
        2.303835,1.559671,-1.074934,
        2.303835,1.560345,2.985899,
        2.303835,1.560487,-2.323400,
        2.303835,1.561127,1.401111,
        2.303835,1.567193,2.643172,
        2.303835,1.568892,2.098362,
        2.303835,1.571702,-2.098435,
        2.303835,1.572701,-2.528730,
        2.303835,1.573572,1.278593,
        2.303835,1.574849,-0.091505,
        2.303835,1.576334,0.365236,
        2.303835,1.577029,-2.962888,
        2.303835,1.577688,-1.814941,
        2.303835,1.579542,-0.755917,
        2.303835,1.579624,-1.199759,
        2.303835,1.581515,-0.419414,
        2.303835,1.584581,-2.738702,
        2.303835,1.586948,-1.696296,
        2.303835,1.588901,2.301496,
        2.303835,1.589936,0.825020,
        2.303835,1.590018,-1.941652,
        2.303835,1.596968,-1.328358,
        2.303835,1.597178,-1.575607,
        2.303835,1.600470,-1.453196,
        2.303835,1.601408,0.242409,
        2.303835,1.602259,1.152656,
        2.303835,1.603064,2.865774,
        2.303835,1.612085,1.968802,
        2.303835,1.612720,-3.080498,
        2.303835,1.616061,2.521531,
        2.303835,1.616900,-0.207457,
        2.303835,1.620988,-0.875920,
        2.303835,1.622114,0.701509,
        2.303835,1.622749,0.121426,
        2.303835,1.622993,-2.221629,
        2.303835,1.629346,-0.542669,
        2.303835,1.630007,1.844552,
        2.303835,1.634747,-2.426198,
        2.303835,1.642346,1.722617,
        2.303835,1.643925,1.034629,
        2.303835,1.644848,2.743774,
        2.303835,1.645258,2.188109,
        2.303835,1.645417,-2.634725,
        2.303835,1.647395,3.085437,
        2.303835,1.648538,-2.855044,
        2.303835,1.649932,0.576808,
        2.303835,1.652224,-0.996105,
        2.303835,1.652871,1.600681,
        2.303835,1.657997,-0.325127,
        2.303835,1.661936,0.001037,
        2.303835,1.665187,1.477402,
        2.303835,1.665496,2.397872,
        2.303835,1.673267,-0.665453,
        2.303835,1.673557,0.452526,
        2.303835,1.679277,-1.122421,
        2.303835,1.679696,0.912482,
        2.303835,1.682420,1.353922,
        2.303835,1.685031,-2.009038,
        2.303835,1.685196,2.967046,
        2.303835,1.690107,2.620910,
        2.303835,1.690453,-2.123076,
        2.303835,1.691693,-1.871572,
        2.303835,1.691817,2.061688,
        2.303835,1.692644,1.237670,
        2.303835,1.696386,0.328123,
        2.303835,1.697162,-1.249026,
        2.303835,1.697488,-2.322785,
        2.303835,1.698726,-2.973576,
        2.303835,1.699172,-1.749717,
        2.303835,1.704478,-0.106295,
        2.303835,1.707244,-0.444770,
        2.303835,1.707500,-2.531497,
        2.303835,1.707714,-0.787321,
        2.303835,1.708430,-1.627552,
        2.303835,1.710583,-1.376890,
        2.303835,1.712415,-2.747921,
        2.303835,1.713125,-1.503073,
        2.303835,1.713452,0.789419,
        2.303835,1.720612,2.279888,
        2.303835,1.724852,0.202831,
        2.303835,1.724868,2.846246,
        2.303835,1.728940,1.931535,
        2.303835,1.732838,1.127584,
        2.303835,1.735719,-3.092485,
        2.303835,1.739676,2.496017,
        2.303835,1.739971,-0.228876,
        2.303835,1.743328,0.663347,
        2.303835,1.744927,-0.911892,
        2.303835,1.745952,0.084733,
        2.303835,1.748077,1.804354,
        2.303835,1.752959,-0.567548,
        2.303835,1.760169,1.678132,
        2.303835,1.763887,-2.218450,
        2.303835,1.766737,2.722752,
        2.303835,1.767999,1.004737,
        2.303835,1.769149,1.550298,
        2.303835,1.770119,-2.427175,
        2.303835,1.770460,2.154208,
        2.303835,1.770514,3.070982,
        2.303835,1.770735,0.537245,
        2.303835,1.772673,-2.864760,
        2.303835,1.773527,-1.039582,
        2.303835,1.776419,-2.642421,
        2.303835,1.780769,1.416330,
        2.303835,1.786634,-0.346247,
        2.303835,1.788829,-1.947553,
        2.303835,1.793737,0.410270,
        2.303835,1.794553,2.374618,
        2.303835,1.794699,-0.694061,
        2.303835,1.797469,-1.169105,
        2.303835,1.803705,0.881076,
        2.303835,1.804124,1.264249,
        2.303835,1.807825,-2.090455,
        2.303835,1.808822,-1.817375,
        2.303835,1.809020,2.949741,
        2.303835,1.812466,2.597595,
        2.303835,1.813159,2.025395,
        2.303835,1.813816,-1.298435,
        2.303835,1.815016,0.278951,
        2.303835,1.816510,-0.008631,
        2.303835,1.817772,-1.689625,
        2.303835,1.821360,-2.984352,
        2.303835,1.822312,-0.127211,
        2.303835,1.823386,-1.429807,
        2.303835,1.823834,-1.560483,
        2.303835,1.832060,-0.821627,
        2.303835,1.832690,-0.468440,
        2.303835,1.833945,-2.320278,
        2.303835,1.834787,0.751968,
        2.303835,1.840579,-2.536445,
        2.303835,1.841033,-2.756932,
        2.303835,1.846350,1.893983,
        2.303835,1.846644,1.113766,
        2.303835,1.846833,2.826543,
        2.303835,1.847325,2.249051,
        2.303835,1.854119,0.134161,
        2.303835,1.858863,-3.106022,
        2.303835,1.865046,1.763384,
        2.303835,1.865619,-0.952036,
        2.303835,1.865912,0.624453,
        2.303835,1.866694,2.473686,
        2.303835,1.867524,-0.240083,
        2.303835,1.875561,1.632831,
        2.303835,1.876049,-0.594378,
        2.303835,1.880509,1.500204,
        2.303835,1.883504,1.360457,
        2.303835,1.886971,-2.196665,
        2.303835,1.888979,0.982922,
        2.303835,1.889815,2.701570,
        2.303835,1.892144,-2.005340,
        2.303835,1.892348,0.496522,
        2.303835,1.893281,-1.083976,
        2.303835,1.893828,3.053453,
        2.303835,1.895117,2.121107,
        2.303835,1.898741,-2.875735,
        2.303835,1.904429,-2.428029,
        2.303835,1.908408,-2.648903,
        2.303835,1.911828,-0.364489,
        2.303835,1.912224,0.366565,
        2.303835,1.914854,1.214468,
        2.303835,1.915467,-1.883143,
        2.303835,1.915763,-0.724679,
        2.303835,1.915790,-1.216366,
        2.303835,1.922256,2.347806,
        2.303835,1.922616,0.844633,
        2.303835,1.925185,0.231170,
        2.303835,1.927474,-1.754616,
        2.303835,1.928342,-1.351174,
        2.303835,1.933328,-1.623328,
        2.303835,1.933878,0.034150,
        2.303835,1.934585,-1.487979,
        2.303835,1.934802,2.927634,
        2.303835,1.934862,1.989776,
        2.303835,1.936327,-0.099350,
        2.303835,1.938088,2.578507,
        2.303835,1.945677,-2.999805,
        2.303835,1.954572,-0.859208,
        2.303835,1.957868,-0.490784,
        2.303835,1.958989,0.712772,
        2.303835,1.960212,-2.307561,
        2.303835,1.964774,1.855507,
        2.303835,1.964971,-2.100533,
        2.303835,1.969112,1.088273,
        2.303835,1.971438,-2.765415,
        2.303835,1.972312,2.805134,
        2.303835,1.974950,2.220596,
        2.303835,1.975135,-2.539127,
        2.303835,1.982107,1.720829,
        2.303835,1.983195,-3.127984,
        2.303835,1.983318,-0.247182,
        2.303835,1.986720,-0.993220,
        2.303835,1.990381,1.586005,
        2.303835,1.990470,0.580283,
        2.303835,1.990523,1.450326,
        2.303835,1.993909,1.312897,
        2.303835,1.994694,2.452813,
        2.303835,1.999226,-0.620717,
        2.303835,2.005145,0.941012,
        2.303835,2.007301,0.142337,
        2.303835,2.014149,0.444138,
        2.303835,2.014338,-1.127392,
        2.303835,2.018131,3.021778,
        2.303835,2.019676,-1.980944,
        2.303835,2.019912,2.686612,
        2.303835,2.020545,2.088672,
        2.303835,2.026074,0.300945,
        2.303835,2.026196,-2.889057,
        2.303835,2.031959,-1.268178,
        2.303835,2.032917,-2.199917,
        2.303835,2.033572,-1.842884,
        2.303835,2.034712,-2.419505,
        2.303835,2.035818,-0.379479,
        2.303835,2.036004,-0.757062,
        2.303835,2.040501,-1.411488,
        2.303835,2.042874,-1.701367,
        2.303835,2.043060,-2.653746,
        2.303835,2.043729,-1.557938,
        2.303835,2.048538,0.806623,
        2.303835,2.049006,1.199930,
        2.303835,2.049502,-0.137989,
        2.303835,2.052181,2.326959,
        2.303835,2.052794,-0.008526,
        2.303835,2.057608,1.951726,
        2.303835,2.065568,2.567730,
        2.303835,2.070371,2.879808,
        2.303835,2.071256,-3.018961,
        2.303835,2.079746,-0.897699,
        2.303835,2.082784,1.042403,
        2.303835,2.083752,-0.511086,
        2.303835,2.084401,1.810502,
        2.303835,2.085731,0.671081,
        2.303835,2.097496,1.373172,
        2.303835,2.099185,1.667006,
        2.303835,2.100482,-2.301084,
        2.303835,2.103118,2.194745,
        2.303835,2.103223,1.521685,
        2.303835,2.103695,-2.777707,
        2.303835,2.107523,3.129581,
        2.303835,2.108277,-2.095907,
        2.303835,2.108891,-2.532166,
        2.303835,2.111003,-0.268632,
        2.303835,2.112276,-1.027246,
        2.303835,2.114380,0.531457,
        2.303835,2.119166,0.100552,
        2.303835,2.119564,2.754818,
        2.303835,2.122474,-0.648314,
        2.303835,2.125183,2.445598,
        2.303835,2.129501,0.238290,
        2.303835,2.131178,0.386663,
        2.303835,2.131223,-1.174311,
        2.303835,2.134269,-1.948042,
        2.303835,2.134684,0.907065,
        2.303835,2.137615,2.988806,
        2.303835,2.145635,-1.327708,
        2.303835,2.146467,2.055081,
        2.303835,2.146582,-1.793788,
        2.303835,2.149210,-1.482321,
        2.303835,2.152901,-1.636788,
        2.303835,2.156337,1.151558,
        2.303835,2.156474,-2.912065,
        2.303835,2.157194,-0.792563,
        2.303835,2.166724,-0.401638,
        2.303835,2.166980,-0.134879,
        2.303835,2.173644,-2.403429,
        2.303835,2.175711,-2.655523,
        2.303835,2.177132,2.600419,
        2.303835,2.178879,0.767722,
        2.303835,2.180752,1.908485,
        2.303835,2.181338,2.313721,
        2.303835,2.185580,1.287673,
        2.303835,2.195400,-2.217546,
        2.303835,2.200418,-3.056442,
        2.303835,2.201829,-0.001735,
        2.303835,2.203804,1.756176,
        2.303835,2.211602,1.443930,
        2.303835,2.212686,-0.544809,
        2.303835,2.213937,0.621892,
        2.303835,2.214075,1.600691,
        2.303835,2.216256,1.020029,
        2.303835,2.219060,2.886631,
        2.303835,2.223375,-0.906185,
        2.303835,2.226002,-1.060248,
        2.303835,2.226413,-2.064034,
        2.303835,2.230554,2.171334,
        2.303835,2.232413,2.735777,
        2.303835,2.233709,3.070830,
        2.303835,2.234383,-2.792065,
        2.303835,2.237183,0.468799,
        2.303835,2.237695,-0.271388,
        2.303835,2.239419,0.151693,
        2.303835,2.243347,-2.521049,
        2.303835,2.244035,-1.229456,
        2.303835,2.244355,0.309899,
        2.303835,2.245175,2.460236,
        2.303835,2.245628,-1.899216,
        2.303835,2.245973,-0.702234,
        2.303835,2.255601,-1.728131,
        2.303835,2.256702,-1.401367,
        2.303835,2.261466,-1.557500,
        2.303835,2.267390,0.878499,
        2.303835,2.272099,2.018594,
        2.303835,2.278618,1.170386,
        2.303835,2.283304,-2.345584,
        2.303835,2.285175,-2.943090,
        2.303835,2.293000,-0.124451,
        2.303835,2.294058,-0.421070,
        2.303835,2.303283,1.855938,
        2.303835,2.304968,2.608978,
        2.303835,2.305347,2.311346,
        2.303835,2.307188,-2.656134,
        2.303835,2.307301,1.336200,
        2.303835,2.309655,0.726020,
        2.303835,2.317634,-2.185225,
        2.303835,2.321510,1.685664,
        2.303835,2.324376,-3.111006,
        2.303835,2.324731,1.511186,
        2.303835,2.329587,0.035498,
        2.303835,2.330158,2.977264,
        2.303835,2.335096,-0.588468,
        2.303835,2.336046,-1.115799,
        2.303835,2.338060,2.803613,
        2.303835,2.338083,-0.943400,
        2.303835,2.340210,0.562145,
        2.303835,2.341931,1.023926,
        2.303835,2.343265,-0.779061,
        2.303835,2.343434,-2.013692,
        2.303835,2.353029,-1.310529,
        2.303835,2.354868,0.211543,
        2.303835,2.355120,0.388593,
        2.303835,2.355165,-1.833173,
        2.303835,2.355462,2.151441,
        2.303835,2.362948,-2.487477,
        2.303835,2.363101,-2.810456,
        2.303835,2.363394,-0.272918,
        2.303835,2.364509,-1.644533,
        2.303835,2.370465,2.466824,
        2.303835,2.386110,-1.481351,
        2.303835,2.394960,0.862133,
        2.303835,2.395716,1.977184,
        2.303835,2.396339,1.200900,
        2.303835,2.407941,-2.316753,
        2.303835,2.409450,-2.988432,
        2.303835,2.417202,-0.103363,
        2.303835,2.417386,-0.443097,
        2.303835,2.417950,2.666662,
        2.303835,2.423190,1.789783,
        2.303835,2.423692,1.392696,
        2.303835,2.431122,3.086379,
        2.303835,2.431406,-2.647820,
        2.303835,2.431584,2.307631,
        2.303835,2.434090,1.592818,
        2.303835,2.435993,0.683844,
        2.303835,2.440898,-2.133300,
        2.303835,2.442222,-1.195837,
        2.303835,2.443112,2.883761,
        2.303835,2.449191,-0.633651,
        2.303835,2.450821,0.084741,
        2.303835,2.453499,-1.008861,
        2.303835,2.460010,-1.938011,
        2.303835,2.460199,-0.822008,
        2.303835,2.460902,0.490079,
        2.303835,2.461774,-1.734913,
        2.303835,2.463986,1.034407,
        2.303835,2.467647,0.287667,
        2.303835,2.479849,2.126925,
        2.303835,2.485907,-2.835471,
        2.303835,2.485945,-1.374853,
        2.303835,2.487747,2.505912,
        2.303835,2.487957,-0.274279,
        2.303835,2.489741,-2.468428,
        2.303835,2.512538,1.239900,
        2.303835,2.517528,1.923289,
        2.303835,2.518454,-1.565852,
        2.303835,2.519348,0.843503,
        2.303835,2.522766,-3.059231,
        2.303835,2.528686,2.737228,
        2.303835,2.534798,-2.266671,
        2.303835,2.536533,1.466274,
        2.303835,2.538504,1.699812,
        2.303835,2.539516,-0.476942,
        2.303835,2.539532,-0.074548,
        2.303835,2.543197,2.990252,
        2.303835,2.554022,2.320619,
        2.303835,2.556919,-2.655005,
        2.303835,2.558146,0.627191,
        2.303835,2.558223,-1.164189,
        2.303835,2.563452,-2.044028,
        2.303835,2.568462,0.150655,
        2.303835,2.569130,-0.700837,
        2.303835,2.570050,-1.804492,
        2.303835,2.572710,-0.934665,
        2.303835,2.575994,0.391718,
        2.303835,2.584774,1.046703,
        2.303835,2.601432,2.102535,
        2.303835,2.603559,2.555114,
        2.303835,2.604150,-2.885208,
        2.303835,2.608630,-1.378011,
        2.303835,2.611256,-0.279110,
        2.303835,2.617344,-2.439480,
        2.303835,2.627402,1.293552,
        2.303835,2.635396,1.848850,
        2.303835,2.635686,2.834950,
        2.303835,2.637087,3.132730,
        2.303835,2.642029,0.813274,
        2.303835,2.645260,1.570203,
        2.303835,2.646182,-1.614036,
        2.303835,2.660113,-2.186513,
        2.303835,2.660483,-0.034608,
        2.303835,2.661713,-0.529511,
        2.303835,2.672720,-1.110422,
        2.303835,2.676626,2.335353,
        2.303835,2.676638,0.540364,
        2.303835,2.676780,-1.901158,
        2.303835,2.681411,-2.667965,
        2.303835,2.682152,0.245751,
        2.303835,2.682859,-0.817071,
        2.303835,2.705107,1.062142,
        2.303835,2.717705,2.625995,
        2.303835,2.721001,2.060097,
        2.303835,2.722319,-2.954069,
        2.303835,2.729840,-1.368237,
        2.303835,2.734349,-0.289298,
        2.303835,2.738409,2.975671,
        2.303835,2.740366,1.374341,
        2.303835,2.744633,-2.391311,
        2.303835,2.747398,1.727166,
        2.303835,2.763799,0.760372,
        2.303835,2.768238,-1.679243,
        2.303835,2.779245,-2.046268,
        2.303835,2.780208,-0.626453,
        2.303835,2.780471,0.027732,
        2.303835,2.784696,-1.013128,
        2.303835,2.789817,0.395431,
        2.303835,2.801154,2.353143,
        2.303835,2.807091,-2.687477,
        2.303835,2.825762,1.085035,
        2.303835,2.830209,2.734943,
        2.303835,2.833517,-3.085508,
        2.303835,2.838301,1.980061,
        2.303835,2.849871,1.512012,
        2.303835,2.850089,-1.343899,
        2.303835,2.856905,-0.318441,
        2.303835,2.870445,-2.291001,
        2.303835,2.885017,0.655562,
        2.303835,2.886428,-1.793217,
        2.303835,2.890337,-0.825633,
        2.303835,2.898932,0.139973,
        2.303835,2.931927,-2.728768,
        2.303835,2.932096,2.377790,
        2.303835,2.937409,2.947245,
        2.303835,2.948048,1.133127,
        2.303835,2.950896,1.797889,
        2.303835,2.970019,-1.282382,
        2.303835,2.978333,-0.412090,
        2.303835,2.996437,-2.089121,
        2.303835,3.011274,0.434110,
        2.303835,3.050410,3.138492,
        2.303835,3.064492,1.647259,
        2.303835,3.088868,-0.989501,
        2.408554,0.053884,-2.408394,
        2.408554,0.077645,1.453433,
        2.408554,0.088408,-0.311160,
        2.408554,0.142314,2.590387,
        2.408554,0.147206,-1.385164,
        2.408554,0.163506,0.573203,
        2.408554,0.165485,-2.875396,
        2.408554,0.182564,-2.125821,
        2.408554,0.190569,1.929141,
        2.408554,0.201102,1.249082,
        2.408554,0.201397,-0.716646,
        2.408554,0.210568,-0.069559,
        2.408554,0.251672,2.848900,
        2.408554,0.266262,-1.678068,
        2.408554,0.269108,-1.215563,
        2.408554,0.273777,-2.506756,
        2.408554,0.279854,0.814581,
        2.408554,0.281664,-2.969147,
        2.408554,0.283625,2.405603,
        2.408554,0.284058,0.357350,
        2.408554,0.294107,1.593880,
        2.408554,0.302491,-0.400410,
        2.408554,0.308059,2.021491,
        2.408554,0.312219,-2.072057,
        2.408554,0.321905,-0.829121,
        2.408554,0.329984,1.201621,
        2.408554,0.339714,-0.013712,
        2.408554,0.368784,2.974974,
        2.408554,0.376428,-1.495532,
        2.408554,0.378103,2.646436,
        2.408554,0.385088,0.590355,
        2.408554,0.386248,-2.690943,
        2.408554,0.389851,-2.340031,
        2.408554,0.391502,-1.171374,
        2.408554,0.398249,-1.812694,
        2.408554,0.400542,-2.995729,
        2.408554,0.404273,0.918163,
        2.408554,0.405617,1.773533,
        2.408554,0.409009,-0.570243,
        2.408554,0.410700,0.270433,
        2.408554,0.415893,1.464890,
        2.408554,0.420680,2.088164,
        2.408554,0.424766,-0.258729,
        2.408554,0.426507,2.367872,
        2.408554,0.441292,-0.890727,
        2.408554,0.451839,-2.078193,
        2.408554,0.459270,1.191562,
        2.408554,0.470622,0.014002,
        2.408554,0.481292,-2.529098,
        2.408554,0.482460,2.849843,
        2.408554,0.492205,-1.397963,
        2.408554,0.493314,3.117264,
        2.408554,0.497870,0.727828,
        2.408554,0.499783,-1.654379,
        2.408554,0.502651,0.463333,
        2.408554,0.505463,2.594557,
        2.408554,0.506104,-2.874910,
        2.408554,0.513579,-1.147354,
        2.408554,0.518569,1.644237,
        2.408554,0.519001,1.900488,
        2.408554,0.519277,-0.673809,
        2.408554,0.524769,-0.421955,
        2.408554,0.525564,-2.288838,
        2.408554,0.532541,0.978737,
        2.408554,0.534610,2.154808,
        2.408554,0.535020,-1.891846,
        2.408554,0.540208,1.400497,
        2.408554,0.542006,0.228337,
        2.408554,0.550762,-0.180771,
        2.408554,0.560531,2.378474,
        2.408554,0.561629,-0.924576,
        2.408554,0.567465,-2.679392,
        2.408554,0.588527,1.181755,
        2.408554,0.591643,-2.098364,
        2.408554,0.594963,2.997575,
        2.408554,0.599643,-3.045235,
        2.408554,0.599962,2.773556,
        2.408554,0.602620,0.029270,
        2.408554,0.605711,0.598454,
        2.408554,0.607708,-1.541026,
        2.408554,0.609360,-2.464379,
        2.408554,0.610738,-1.326786,
        2.408554,0.616968,0.820489,
        2.408554,0.626379,1.773610,
        2.408554,0.628626,-1.751180,
        2.408554,0.630671,0.391949,
        2.408554,0.632622,-0.737642,
        2.408554,0.632708,2.566722,
        2.408554,0.632797,1.989858,
        2.408554,0.634455,-0.534790,
        2.408554,0.636205,1.561170,
        2.408554,0.637681,-1.123423,
        2.408554,0.641441,-2.842109,
        2.408554,0.644807,-0.327139,
        2.408554,0.652440,2.197129,
        2.408554,0.660897,-2.278593,
        2.408554,0.664008,1.360091,
        2.408554,0.670850,-1.945257,
        2.408554,0.671396,1.004859,
        2.408554,0.673985,0.203855,
        2.408554,0.680326,-0.133370,
        2.408554,0.685083,-0.942363,
        2.408554,0.688204,-2.632204,
        2.408554,0.689777,2.385011,
        2.408554,0.700065,3.115430,
        2.408554,0.704080,2.911636,
        2.408554,0.712883,1.174216,
        2.408554,0.715701,0.699906,
        2.408554,0.718851,-1.442215,
        2.408554,0.721019,2.721536,
        2.408554,0.725300,-2.975134,
        2.408554,0.727136,-1.633288,
        2.408554,0.727399,-2.120567,
        2.408554,0.730751,0.522488,
        2.408554,0.732892,-1.254330,
        2.408554,0.734333,-2.446184,
        2.408554,0.735125,0.037270,
        2.408554,0.737893,1.866463,
        2.408554,0.740650,1.681075,
        2.408554,0.745971,-0.781286,
        2.408554,0.749432,-0.442559,
        2.408554,0.750102,2.050636,
        2.408554,0.751046,0.863821,
        2.408554,0.757018,-0.608833,
        2.408554,0.757046,-1.816019,
        2.408554,0.757389,1.501184,
        2.408554,0.759800,2.546181,
        2.408554,0.759859,0.351923,
        2.408554,0.761141,-2.781102,
        2.408554,0.769035,-1.078574,
        2.408554,0.770166,-0.266604,
        2.408554,0.772218,2.227380,
        2.408554,0.788382,1.331051,
        2.408554,0.790269,-2.282409,
        2.408554,0.803410,-1.984567,
        2.408554,0.803813,0.191520,
        2.408554,0.806832,3.022155,
        2.408554,0.807476,-2.597675,
        2.408554,0.812512,-0.107335,
        2.408554,0.814161,1.010949,
        2.408554,0.817732,-3.084486,
        2.408554,0.817846,2.390688,
        2.408554,0.819323,2.845960,
        2.408554,0.819430,-0.915017,
        2.408554,0.829198,-1.532491,
        2.408554,0.829531,0.665695,
        2.408554,0.829816,1.167147,
        2.408554,0.833613,-1.357492,
        2.408554,0.842945,-2.907318,
        2.408554,0.844883,2.679708,
        2.408554,0.848189,-1.705003,
        2.408554,0.849888,1.767901,
        2.408554,0.851126,0.493370,
        2.408554,0.852911,1.933394,
        2.408554,0.857516,-2.432863,
        2.408554,0.858261,-1.191729,
        2.408554,0.858279,-2.138895,
        2.408554,0.859723,-0.734928,
        2.408554,0.859754,1.603292,
        2.408554,0.861655,-0.394821,
        2.408554,0.864300,0.043549,
        2.408554,0.866940,0.851229,
        2.408554,0.869748,2.095969,
        2.408554,0.877352,-0.574668,
        2.408554,0.879254,-2.733338,
        2.408554,0.882554,1.443757,
        2.408554,0.884060,0.334199,
        2.408554,0.885626,-1.866061,
        2.408554,0.889834,2.529957,
        2.408554,0.893140,2.251698,
        2.408554,0.897463,-0.240822,
        2.408554,0.898758,-1.031948,
        2.408554,0.913918,1.287618,
        2.408554,0.916555,-2.287173,
        2.408554,0.918305,3.107845,
        2.408554,0.919522,2.948277,
        2.408554,0.927138,-2.569686,
        2.408554,0.929031,0.185383,
        2.408554,0.933766,-3.015917,
        2.408554,0.934873,-2.010822,
        2.408554,0.935636,-1.445855,
        2.408554,0.936702,-0.866461,
        2.408554,0.937395,0.729108,
        2.408554,0.937910,2.787575,
        2.408554,0.939144,-0.092962,
        2.408554,0.941718,1.125085,
        2.408554,0.942873,-1.607680,
        2.408554,0.943701,0.980328,
        2.408554,0.947774,2.395256,
        2.408554,0.952433,0.584347,
        2.408554,0.954456,-1.287293,
        2.408554,0.958731,-2.852552,
        2.408554,0.961838,1.839284,
        2.408554,0.961898,-0.469425,
        2.408554,0.965426,1.686615,
        2.408554,0.970108,1.990771,
        2.408554,0.972820,-1.762780,
        2.408554,0.973500,-0.711881,
        2.408554,0.975253,2.641082,
        2.408554,0.977824,0.441928,
        2.408554,0.981485,-2.424766,
        2.408554,0.981877,1.537100,
        2.408554,0.982609,-1.136920,
        2.408554,0.984265,-2.156073,
        2.408554,0.989150,0.045493,
        2.408554,0.993052,2.137478,
        2.408554,0.997766,-2.693871,
        2.408554,0.999847,-0.334738,
        2.408554,1.001920,0.848054,
        2.408554,1.006146,1.389734,
        2.408554,1.012180,0.304066,
        2.408554,1.016938,2.270674,
        2.408554,1.017371,-0.979790,
        2.408554,1.019495,-1.894487,
        2.408554,1.023796,-0.583691,
        2.408554,1.026525,2.506945,
        2.408554,1.027004,-0.203653,
        2.408554,1.028027,3.036573,
        2.408554,1.031435,1.241021,
        2.408554,1.032316,2.884221,
        2.408554,1.035863,-3.110475,
        2.408554,1.042085,-2.292950,
        2.408554,1.043272,-1.522986,
        2.408554,1.046265,-2.958847,
        2.408554,1.049272,-1.376344,
        2.408554,1.050039,-2.552086,
        2.408554,1.053682,0.689863,
        2.408554,1.055458,0.170394,
        2.408554,1.056622,-0.831866,
        2.408554,1.057515,1.097998,
        2.408554,1.058706,-2.036438,
        2.408554,1.061389,-1.672248,
        2.408554,1.063758,2.743093,
        2.408554,1.070157,-0.073422,
        2.408554,1.072154,1.761506,
        2.408554,1.074029,1.909193,
        2.408554,1.074204,0.545276,
        2.408554,1.074714,-2.805147,
        2.408554,1.076262,0.966104,
        2.408554,1.077642,-1.227073,
        2.408554,1.084151,1.618525,
        2.408554,1.085085,2.377673,
        2.408554,1.086822,-0.441112,
        2.408554,1.087142,2.058815,
        2.408554,1.098383,-0.696430,
        2.408554,1.099138,-1.084033,
        2.408554,1.103208,1.477773,
        2.408554,1.103783,0.408254,
        2.408554,1.106275,-2.420800,
        2.408554,1.108079,2.612943,
        2.408554,1.109345,-2.172718,
        2.408554,1.115957,-1.787597,
        2.408554,1.118930,0.045983,
        2.408554,1.119777,0.834702,
        2.408554,1.121943,-2.669042,
        2.408554,1.124930,2.219917,
        2.408554,1.125217,1.336294,
        2.408554,1.130958,2.972931,
        2.408554,1.131400,-0.306629,
        2.408554,1.137135,-0.936514,
        2.408554,1.139125,-1.924528,
        2.408554,1.139877,0.276382,
        2.408554,1.142003,-3.058022,
        2.408554,1.147363,-0.558448,
        2.408554,1.150348,-0.176418,
        2.408554,1.151198,1.197016,
        2.408554,1.153930,-1.597104,
        2.408554,1.155389,2.837148,
        2.408554,1.155458,-1.460623,
        2.408554,1.157331,-2.906362,
        2.408554,1.158594,2.487618,
        2.408554,1.162103,-1.322553,
        2.408554,1.165407,3.092861,
        2.408554,1.168025,0.717075,
        2.408554,1.168406,-2.301519,
        2.408554,1.174782,-2.538755,
        2.408554,1.177998,1.061256,
        2.408554,1.178898,1.827305,
        2.408554,1.181331,-2.060330,
        2.408554,1.182540,-0.798952,
        2.408554,1.184569,0.147517,
        2.408554,1.184734,1.968919,
        2.408554,1.184804,0.593386,
        2.408554,1.188278,1.690808,
        2.408554,1.194360,2.110440,
        2.408554,1.194776,2.710455,
        2.408554,1.197745,-1.173705,
        2.408554,1.199898,-2.777257,
        2.408554,1.201164,2.355931,
        2.408554,1.203760,1.556404,
        2.408554,1.205845,-0.415993,
        2.408554,1.206585,0.933648,
        2.408554,1.206950,-0.037248,
        2.408554,1.209341,0.471162,
        2.408554,1.213483,-1.039417,
        2.408554,1.218146,-0.672643,
        2.408554,1.221608,1.421669,
        2.408554,1.226043,-1.693533,
        2.408554,1.232870,-1.822773,
        2.408554,1.234375,-2.421194,
        2.408554,1.234531,-2.190148,
        2.408554,1.239414,0.348959,
        2.408554,1.240183,2.587255,
        2.408554,1.242312,2.941234,
        2.408554,1.244075,1.286644,
        2.408554,1.245961,-2.997461,
        2.408554,1.246495,-2.646872,
        2.408554,1.247736,2.242810,
        2.408554,1.253604,0.812222,
        2.408554,1.257816,-0.894317,
        2.408554,1.258247,-1.953634,
        2.408554,1.258993,-0.284099,
        2.408554,1.262380,-0.157609,
        2.408554,1.262791,-1.549229,
        2.408554,1.263772,-0.533073,
        2.408554,1.266731,-1.412461,
        2.408554,1.268591,-3.122146,
        2.408554,1.269614,-1.275246,
        2.408554,1.270745,1.152533,
        2.408554,1.273606,0.224545,
        2.408554,1.279716,2.812101,
        2.408554,1.283786,2.462124,
        2.408554,1.284798,0.078209,
        2.408554,1.285402,-2.882049,
        2.408554,1.286936,1.886053,
        2.408554,1.289408,0.683122,
        2.408554,1.294736,1.756184,
        2.408554,1.296978,-2.311511,
        2.408554,1.298546,2.020601,
        2.408554,1.300924,1.022478,
        2.408554,1.303305,3.043152,
        2.408554,1.304210,-2.084225,
        2.408554,1.305241,-2.530249,
        2.408554,1.307513,1.628003,
        2.408554,1.310722,0.558267,
        2.408554,1.314142,2.139329,
        2.408554,1.316087,-0.777971,
        2.408554,1.317907,-1.121023,
        2.408554,1.321019,-2.749776,
        2.408554,1.321609,2.687717,
        2.408554,1.322568,1.499415,
        2.408554,1.322972,-0.395699,
        2.408554,1.325286,-0.994192,
        2.408554,1.329146,-0.648513,
        2.408554,1.329795,-0.042570,
        2.408554,1.331651,2.344497,
        2.408554,1.336522,0.435098,
        2.408554,1.339410,0.900087,
        2.408554,1.339550,1.370353,
        2.408554,1.340441,-1.731580,
        2.408554,1.351855,-1.855234,
        2.408554,1.360161,-3.018404,
        2.408554,1.361031,-1.613971,
        2.408554,1.361196,2.914780,
        2.408554,1.361998,-2.207254,
        2.408554,1.362302,1.239384,
        2.408554,1.365228,2.563808,
        2.408554,1.365452,0.310522,
        2.408554,1.367028,-2.420041,
        2.408554,1.373422,-1.484281,
        2.408554,1.374245,-1.226417,
        2.408554,1.375020,-0.170393,
        2.408554,1.376560,-1.355481,
        2.408554,1.377070,-1.980984,
        2.408554,1.377578,-2.636339,
        2.408554,1.378096,0.774626,
        2.408554,1.379086,-0.508192,
        2.408554,1.382562,0.177141,
        2.408554,1.388849,2.233847,
        2.408554,1.389268,-0.886349,
        2.408554,1.391324,1.109396,
        2.408554,1.391699,-0.288725,
        2.408554,1.392869,3.136903,
        2.408554,1.393612,1.952016,
        2.408554,1.395545,-2.850807,
        2.408554,1.401926,1.822930,
        2.408554,1.402149,2.788898,
        2.408554,1.409552,0.647980,
        2.408554,1.410840,0.056219,
        2.408554,1.411944,2.444377,
        2.408554,1.413639,1.697982,
        2.408554,1.426479,1.573427,
        2.408554,1.426725,2.102242,
        2.408554,1.427868,0.984637,
        2.408554,1.429580,-2.107232,
        2.408554,1.431375,-2.313269,
        2.408554,1.431907,3.015410,
        2.408554,1.432325,-1.029054,
        2.408554,1.433341,0.524268,
        2.408554,1.435181,-0.739771,
        2.408554,1.437208,-2.529085,
        2.408554,1.439041,-0.617966,
        2.408554,1.441473,1.448544,
        2.408554,1.444527,2.664744,
        2.408554,1.450513,-0.400625,
        2.408554,1.451907,-0.066823,
        2.408554,1.452754,-2.738631,
        2.408554,1.454061,-2.950416,
        2.408554,1.456736,1.323154,
        2.408554,1.458370,0.400433,
        2.408554,1.459314,-1.765003,
        2.408554,1.459902,-1.145835,
        2.408554,1.464952,2.329245,
        2.408554,1.466050,0.862104,
        2.408554,1.471330,-1.886160,
        2.408554,1.475119,-1.648255,
        2.408554,1.480878,1.194576,
        2.408554,1.480942,0.273235,
        2.408554,1.481086,2.890269,
        2.408554,1.484118,-1.529156,
        2.408554,1.484741,-1.279219,
        2.408554,1.487681,-1.406069,
        2.408554,1.487886,-3.066037,
        2.408554,1.490555,2.543284,
        2.408554,1.492424,-0.190932,
        2.408554,1.495002,-0.839810,
        2.408554,1.495297,2.007188,
        2.408554,1.496414,-2.009267,
        2.408554,1.498261,-2.425270,
        2.408554,1.498907,-2.207327,
        2.408554,1.500204,0.737638,
        2.408554,1.502469,0.151156,
        2.408554,1.507448,-0.514782,
        2.408554,1.509287,2.204647,
        2.408554,1.510518,1.884920,
        2.408554,1.511948,-2.632360,
        2.408554,1.515235,1.068565,
        2.408554,1.522237,3.102018,
        2.408554,1.522373,1.764991,
        2.408554,1.522416,-0.954956,
        2.408554,1.523102,2.765920,
        2.408554,1.524811,-2.844718,
        2.408554,1.527261,-0.304916,
        2.408554,1.529788,0.613208,
        2.408554,1.533648,1.644808,
        2.408554,1.535018,0.028869,
        2.408554,1.541082,2.424608,
        2.408554,1.546010,1.523462,
        2.408554,1.551135,-0.642994,
        2.408554,1.553895,0.489607,
        2.408554,1.555041,0.947813,
        2.408554,1.559671,-1.074934,
        2.408554,1.560345,2.985899,
        2.408554,1.560487,-2.323400,
        2.408554,1.561127,1.401111,
        2.408554,1.567193,2.643172,
        2.408554,1.568892,2.098362,
        2.408554,1.571702,-2.098435,
        2.408554,1.572701,-2.528730,
        2.408554,1.573572,1.278593,
        2.408554,1.574849,-0.091505,
        2.408554,1.576334,0.365236,
        2.408554,1.577029,-2.962888,
        2.408554,1.577688,-1.814941,
        2.408554,1.579542,-0.755917,
        2.408554,1.579624,-1.199759,
        2.408554,1.581515,-0.419414,
        2.408554,1.584581,-2.738702,
        2.408554,1.586948,-1.696296,
        2.408554,1.588901,2.301496,
        2.408554,1.589936,0.825020,
        2.408554,1.590018,-1.941652,
        2.408554,1.596968,-1.328358,
        2.408554,1.597178,-1.575607,
        2.408554,1.600470,-1.453196,
        2.408554,1.601408,0.242409,
        2.408554,1.602259,1.152656,
        2.408554,1.603064,2.865774,
        2.408554,1.612085,1.968802,
        2.408554,1.612720,-3.080498,
        2.408554,1.616061,2.521531,
        2.408554,1.616900,-0.207457,
        2.408554,1.620988,-0.875920,
        2.408554,1.622114,0.701509,
        2.408554,1.622749,0.121426,
        2.408554,1.622993,-2.221629,
        2.408554,1.629346,-0.542669,
        2.408554,1.630007,1.844552,
        2.408554,1.634747,-2.426198,
        2.408554,1.642346,1.722617,
        2.408554,1.643925,1.034629,
        2.408554,1.644848,2.743774,
        2.408554,1.645258,2.188109,
        2.408554,1.645417,-2.634725,
        2.408554,1.647395,3.085437,
        2.408554,1.648538,-2.855044,
        2.408554,1.649932,0.576808,
        2.408554,1.652224,-0.996105,
        2.408554,1.652871,1.600681,
        2.408554,1.657997,-0.325127,
        2.408554,1.661936,0.001037,
        2.408554,1.665187,1.477402,
        2.408554,1.665496,2.397872,
        2.408554,1.673267,-0.665453,
        2.408554,1.673557,0.452526,
        2.408554,1.679277,-1.122421,
        2.408554,1.679696,0.912482,
        2.408554,1.682420,1.353922,
        2.408554,1.685031,-2.009038,
        2.408554,1.685196,2.967046,
        2.408554,1.690107,2.620910,
        2.408554,1.690453,-2.123076,
        2.408554,1.691693,-1.871572,
        2.408554,1.691817,2.061688,
        2.408554,1.692644,1.237670,
        2.408554,1.696386,0.328123,
        2.408554,1.697162,-1.249026,
        2.408554,1.697488,-2.322785,
        2.408554,1.698726,-2.973576,
        2.408554,1.699172,-1.749717,
        2.408554,1.704478,-0.106295,
        2.408554,1.707244,-0.444770,
        2.408554,1.707500,-2.531497,
        2.408554,1.707714,-0.787321,
        2.408554,1.708430,-1.627552,
        2.408554,1.710583,-1.376890,
        2.408554,1.712415,-2.747921,
        2.408554,1.713125,-1.503073,
        2.408554,1.713452,0.789419,
        2.408554,1.720612,2.279888,
        2.408554,1.724852,0.202831,
        2.408554,1.724868,2.846246,
        2.408554,1.728940,1.931535,
        2.408554,1.732838,1.127584,
        2.408554,1.735719,-3.092485,
        2.408554,1.739676,2.496017,
        2.408554,1.739971,-0.228876,
        2.408554,1.743328,0.663347,
        2.408554,1.744927,-0.911892,
        2.408554,1.745952,0.084733,
        2.408554,1.748077,1.804354,
        2.408554,1.752959,-0.567548,
        2.408554,1.760169,1.678132,
        2.408554,1.763887,-2.218450,
        2.408554,1.766737,2.722752,
        2.408554,1.767999,1.004737,
        2.408554,1.769149,1.550298,
        2.408554,1.770119,-2.427175,
        2.408554,1.770460,2.154208,
        2.408554,1.770514,3.070982,
        2.408554,1.770735,0.537245,
        2.408554,1.772673,-2.864760,
        2.408554,1.773527,-1.039582,
        2.408554,1.776419,-2.642421,
        2.408554,1.780769,1.416330,
        2.408554,1.786634,-0.346247,
        2.408554,1.788829,-1.947553,
        2.408554,1.793737,0.410270,
        2.408554,1.794553,2.374618,
        2.408554,1.794699,-0.694061,
        2.408554,1.797469,-1.169105,
        2.408554,1.803705,0.881076,
        2.408554,1.804124,1.264249,
        2.408554,1.807825,-2.090455,
        2.408554,1.808822,-1.817375,
        2.408554,1.809020,2.949741,
        2.408554,1.812466,2.597595,
        2.408554,1.813159,2.025395,
        2.408554,1.813816,-1.298435,
        2.408554,1.815016,0.278951,
        2.408554,1.816510,-0.008631,
        2.408554,1.817772,-1.689625,
        2.408554,1.821360,-2.984352,
        2.408554,1.822312,-0.127211,
        2.408554,1.823386,-1.429807,
        2.408554,1.823834,-1.560483,
        2.408554,1.832060,-0.821627,
        2.408554,1.832690,-0.468440,
        2.408554,1.833945,-2.320278,
        2.408554,1.834787,0.751968,
        2.408554,1.840579,-2.536445,
        2.408554,1.841033,-2.756932,
        2.408554,1.846350,1.893983,
        2.408554,1.846644,1.113766,
        2.408554,1.846833,2.826543,
        2.408554,1.847325,2.249051,
        2.408554,1.854119,0.134161,
        2.408554,1.858863,-3.106022,
        2.408554,1.865046,1.763384,
        2.408554,1.865619,-0.952036,
        2.408554,1.865912,0.624453,
        2.408554,1.866694,2.473686,
        2.408554,1.867524,-0.240083,
        2.408554,1.875561,1.632831,
        2.408554,1.876049,-0.594378,
        2.408554,1.880509,1.500204,
        2.408554,1.883504,1.360457,
        2.408554,1.886971,-2.196665,
        2.408554,1.888979,0.982922,
        2.408554,1.889815,2.701570,
        2.408554,1.892144,-2.005340,
        2.408554,1.892348,0.496522,
        2.408554,1.893281,-1.083976,
        2.408554,1.893828,3.053453,
        2.408554,1.895117,2.121107,
        2.408554,1.898741,-2.875735,
        2.408554,1.904429,-2.428029,
        2.408554,1.908408,-2.648903,
        2.408554,1.911828,-0.364489,
        2.408554,1.912224,0.366565,
        2.408554,1.914854,1.214468,
        2.408554,1.915467,-1.883143,
        2.408554,1.915763,-0.724679,
        2.408554,1.915790,-1.216366,
        2.408554,1.922256,2.347806,
        2.408554,1.922616,0.844633,
        2.408554,1.925185,0.231170,
        2.408554,1.927474,-1.754616,
        2.408554,1.928342,-1.351174,
        2.408554,1.933328,-1.623328,
        2.408554,1.933878,0.034150,
        2.408554,1.934585,-1.487979,
        2.408554,1.934802,2.927634,
        2.408554,1.934862,1.989776,
        2.408554,1.936327,-0.099350,
        2.408554,1.938088,2.578507,
        2.408554,1.945677,-2.999805,
        2.408554,1.954572,-0.859208,
        2.408554,1.957868,-0.490784,
        2.408554,1.958989,0.712772,
        2.408554,1.960212,-2.307561,
        2.408554,1.964774,1.855507,
        2.408554,1.964971,-2.100533,
        2.408554,1.969112,1.088273,
        2.408554,1.971438,-2.765415,
        2.408554,1.972312,2.805134,
        2.408554,1.974950,2.220596,
        2.408554,1.975135,-2.539127,
        2.408554,1.982107,1.720829,
        2.408554,1.983195,-3.127984,
        2.408554,1.983318,-0.247182,
        2.408554,1.986720,-0.993220,
        2.408554,1.990381,1.586005,
        2.408554,1.990470,0.580283,
        2.408554,1.990523,1.450326,
        2.408554,1.993909,1.312897,
        2.408554,1.994694,2.452813,
        2.408554,1.999226,-0.620717,
        2.408554,2.005145,0.941012,
        2.408554,2.007301,0.142337,
        2.408554,2.014149,0.444138,
        2.408554,2.014338,-1.127392,
        2.408554,2.018131,3.021778,
        2.408554,2.019676,-1.980944,
        2.408554,2.019912,2.686612,
        2.408554,2.020545,2.088672,
        2.408554,2.026074,0.300945,
        2.408554,2.026196,-2.889057,
        2.408554,2.031959,-1.268178,
        2.408554,2.032917,-2.199917,
        2.408554,2.033572,-1.842884,
        2.408554,2.034712,-2.419505,
        2.408554,2.035818,-0.379479,
        2.408554,2.036004,-0.757062,
        2.408554,2.040501,-1.411488,
        2.408554,2.042874,-1.701367,
        2.408554,2.043060,-2.653746,
        2.408554,2.043729,-1.557938,
        2.408554,2.048538,0.806623,
        2.408554,2.049006,1.199930,
        2.408554,2.049502,-0.137989,
        2.408554,2.052181,2.326959,
        2.408554,2.052794,-0.008526,
        2.408554,2.057608,1.951726,
        2.408554,2.065568,2.567730,
        2.408554,2.070371,2.879808,
        2.408554,2.071256,-3.018961,
        2.408554,2.079746,-0.897699,
        2.408554,2.082784,1.042403,
        2.408554,2.083752,-0.511086,
        2.408554,2.084401,1.810502,
        2.408554,2.085731,0.671081,
        2.408554,2.097496,1.373172,
        2.408554,2.099185,1.667006,
        2.408554,2.100482,-2.301084,
        2.408554,2.103118,2.194745,
        2.408554,2.103223,1.521685,
        2.408554,2.103695,-2.777707,
        2.408554,2.107523,3.129581,
        2.408554,2.108277,-2.095907,
        2.408554,2.108891,-2.532166,
        2.408554,2.111003,-0.268632,
        2.408554,2.112276,-1.027246,
        2.408554,2.114380,0.531457,
        2.408554,2.119166,0.100552,
        2.408554,2.119564,2.754818,
        2.408554,2.122474,-0.648314,
        2.408554,2.125183,2.445598,
        2.408554,2.129501,0.238290,
        2.408554,2.131178,0.386663,
        2.408554,2.131223,-1.174311,
        2.408554,2.134269,-1.948042,
        2.408554,2.134684,0.907065,
        2.408554,2.137615,2.988806,
        2.408554,2.145635,-1.327708,
        2.408554,2.146467,2.055081,
        2.408554,2.146582,-1.793788,
        2.408554,2.149210,-1.482321,
        2.408554,2.152901,-1.636788,
        2.408554,2.156337,1.151558,
        2.408554,2.156474,-2.912065,
        2.408554,2.157194,-0.792563,
        2.408554,2.166724,-0.401638,
        2.408554,2.166980,-0.134879,
        2.408554,2.173644,-2.403429,
        2.408554,2.175711,-2.655523,
        2.408554,2.177132,2.600419,
        2.408554,2.178879,0.767722,
        2.408554,2.180752,1.908485,
        2.408554,2.181338,2.313721,
        2.408554,2.185580,1.287673,
        2.408554,2.195400,-2.217546,
        2.408554,2.200418,-3.056442,
        2.408554,2.201829,-0.001735,
        2.408554,2.203804,1.756176,
        2.408554,2.211602,1.443930,
        2.408554,2.212686,-0.544809,
        2.408554,2.213937,0.621892,
        2.408554,2.214075,1.600691,
        2.408554,2.216256,1.020029,
        2.408554,2.219060,2.886631,
        2.408554,2.223375,-0.906185,
        2.408554,2.226002,-1.060248,
        2.408554,2.226413,-2.064034,
        2.408554,2.230554,2.171334,
        2.408554,2.232413,2.735777,
        2.408554,2.233709,3.070830,
        2.408554,2.234383,-2.792065,
        2.408554,2.237183,0.468799,
        2.408554,2.237695,-0.271388,
        2.408554,2.239419,0.151693,
        2.408554,2.243347,-2.521049,
        2.408554,2.244035,-1.229456,
        2.408554,2.244355,0.309899,
        2.408554,2.245175,2.460236,
        2.408554,2.245628,-1.899216,
        2.408554,2.245973,-0.702234,
        2.408554,2.255601,-1.728131,
        2.408554,2.256702,-1.401367,
        2.408554,2.261466,-1.557500,
        2.408554,2.267390,0.878499,
        2.408554,2.272099,2.018594,
        2.408554,2.278618,1.170386,
        2.408554,2.283304,-2.345584,
        2.408554,2.285175,-2.943090,
        2.408554,2.293000,-0.124451,
        2.408554,2.294058,-0.421070,
        2.408554,2.303283,1.855938,
        2.408554,2.304968,2.608978,
        2.408554,2.305347,2.311346,
        2.408554,2.307188,-2.656134,
        2.408554,2.307301,1.336200,
        2.408554,2.309655,0.726020,
        2.408554,2.317634,-2.185225,
        2.408554,2.321510,1.685664,
        2.408554,2.324376,-3.111006,
        2.408554,2.324731,1.511186,
        2.408554,2.329587,0.035498,
        2.408554,2.330158,2.977264,
        2.408554,2.335096,-0.588468,
        2.408554,2.336046,-1.115799,
        2.408554,2.338060,2.803613,
        2.408554,2.338083,-0.943400,
        2.408554,2.340210,0.562145,
        2.408554,2.341931,1.023926,
        2.408554,2.343265,-0.779061,
        2.408554,2.343434,-2.013692,
        2.408554,2.353029,-1.310529,
        2.408554,2.354868,0.211543,
        2.408554,2.355120,0.388593,
        2.408554,2.355165,-1.833173,
        2.408554,2.355462,2.151441,
        2.408554,2.362948,-2.487477,
        2.408554,2.363101,-2.810456,
        2.408554,2.363394,-0.272918,
        2.408554,2.364509,-1.644533,
        2.408554,2.370465,2.466824,
        2.408554,2.386110,-1.481351,
        2.408554,2.394960,0.862133,
        2.408554,2.395716,1.977184,
        2.408554,2.396339,1.200900,
        2.408554,2.407941,-2.316753,
        2.408554,2.409450,-2.988432,
        2.408554,2.417202,-0.103363,
        2.408554,2.417386,-0.443097,
        2.408554,2.417950,2.666662,
        2.408554,2.423190,1.789783,
        2.408554,2.423692,1.392696,
        2.408554,2.431122,3.086379,
        2.408554,2.431406,-2.647820,
        2.408554,2.431584,2.307631,
        2.408554,2.434090,1.592818,
        2.408554,2.435993,0.683844,
        2.408554,2.440898,-2.133300,
        2.408554,2.442222,-1.195837,
        2.408554,2.443112,2.883761,
        2.408554,2.449191,-0.633651,
        2.408554,2.450821,0.084741,
        2.408554,2.453499,-1.008861,
        2.408554,2.460010,-1.938011,
        2.408554,2.460199,-0.822008,
        2.408554,2.460902,0.490079,
        2.408554,2.461774,-1.734913,
        2.408554,2.463986,1.034407,
        2.408554,2.467647,0.287667,
        2.408554,2.479849,2.126925,
        2.408554,2.485907,-2.835471,
        2.408554,2.485945,-1.374853,
        2.408554,2.487747,2.505912,
        2.408554,2.487957,-0.274279,
        2.408554,2.489741,-2.468428,
        2.408554,2.512538,1.239900,
        2.408554,2.517528,1.923289,
        2.408554,2.518454,-1.565852,
        2.408554,2.519348,0.843503,
        2.408554,2.522766,-3.059231,
        2.408554,2.528686,2.737228,
        2.408554,2.534798,-2.266671,
        2.408554,2.536533,1.466274,
        2.408554,2.538504,1.699812,
        2.408554,2.539516,-0.476942,
        2.408554,2.539532,-0.074548,
        2.408554,2.543197,2.990252,
        2.408554,2.554022,2.320619,
        2.408554,2.556919,-2.655005,
        2.408554,2.558146,0.627191,
        2.408554,2.558223,-1.164189,
        2.408554,2.563452,-2.044028,
        2.408554,2.568462,0.150655,
        2.408554,2.569130,-0.700837,
        2.408554,2.570050,-1.804492,
        2.408554,2.572710,-0.934665,
        2.408554,2.575994,0.391718,
        2.408554,2.584774,1.046703,
        2.408554,2.601432,2.102535,
        2.408554,2.603559,2.555114,
        2.408554,2.604150,-2.885208,
        2.408554,2.608630,-1.378011,
        2.408554,2.611256,-0.279110,
        2.408554,2.617344,-2.439480,
        2.408554,2.627402,1.293552,
        2.408554,2.635396,1.848850,
        2.408554,2.635686,2.834950,
        2.408554,2.637087,3.132730,
        2.408554,2.642029,0.813274,
        2.408554,2.645260,1.570203,
        2.408554,2.646182,-1.614036,
        2.408554,2.660113,-2.186513,
        2.408554,2.660483,-0.034608,
        2.408554,2.661713,-0.529511,
        2.408554,2.672720,-1.110422,
        2.408554,2.676626,2.335353,
        2.408554,2.676638,0.540364,
        2.408554,2.676780,-1.901158,
        2.408554,2.681411,-2.667965,
        2.408554,2.682152,0.245751,
        2.408554,2.682859,-0.817071,
        2.408554,2.705107,1.062142,
        2.408554,2.717705,2.625995,
        2.408554,2.721001,2.060097,
        2.408554,2.722319,-2.954069,
        2.408554,2.729840,-1.368237,
        2.408554,2.734349,-0.289298,
        2.408554,2.738409,2.975671,
        2.408554,2.740366,1.374341,
        2.408554,2.744633,-2.391311,
        2.408554,2.747398,1.727166,
        2.408554,2.763799,0.760372,
        2.408554,2.768238,-1.679243,
        2.408554,2.779245,-2.046268,
        2.408554,2.780208,-0.626453,
        2.408554,2.780471,0.027732,
        2.408554,2.784696,-1.013128,
        2.408554,2.789817,0.395431,
        2.408554,2.801154,2.353143,
        2.408554,2.807091,-2.687477,
        2.408554,2.825762,1.085035,
        2.408554,2.830209,2.734943,
        2.408554,2.833517,-3.085508,
        2.408554,2.838301,1.980061,
        2.408554,2.849871,1.512012,
        2.408554,2.850089,-1.343899,
        2.408554,2.856905,-0.318441,
        2.408554,2.870445,-2.291001,
        2.408554,2.885017,0.655562,
        2.408554,2.886428,-1.793217,
        2.408554,2.890337,-0.825633,
        2.408554,2.898932,0.139973,
        2.408554,2.931927,-2.728768,
        2.408554,2.932096,2.377790,
        2.408554,2.937409,2.947245,
        2.408554,2.948048,1.133127,
        2.408554,2.950896,1.797889,
        2.408554,2.970019,-1.282382,
        2.408554,2.978333,-0.412090,
        2.408554,2.996437,-2.089121,
        2.408554,3.011274,0.434110,
        2.408554,3.050410,3.138492,
        2.408554,3.064492,1.647259,
        2.408554,3.088868,-0.989501,
        2.513274,0.053884,-2.408394,
        2.513274,0.077645,1.453433,
        2.513274,0.088408,-0.311160,
        2.513274,0.142314,2.590387,
        2.513274,0.147206,-1.385164,
        2.513274,0.163506,0.573203,
        2.513274,0.165485,-2.875396,
        2.513274,0.182564,-2.125821,
        2.513274,0.190569,1.929141,
        2.513274,0.201102,1.249082,
        2.513274,0.201397,-0.716646,
        2.513274,0.210568,-0.069559,
        2.513274,0.251672,2.848900,
        2.513274,0.266262,-1.678068,
        2.513274,0.269108,-1.215563,
        2.513274,0.273777,-2.506756,
        2.513274,0.279854,0.814581,
        2.513274,0.281664,-2.969147,
        2.513274,0.283625,2.405603,
        2.513274,0.284058,0.357350,
        2.513274,0.294107,1.593880,
        2.513274,0.302491,-0.400410,
        2.513274,0.308059,2.021491,
        2.513274,0.312219,-2.072057,
        2.513274,0.321905,-0.829121,
        2.513274,0.329984,1.201621,
        2.513274,0.339714,-0.013712,
        2.513274,0.368784,2.974974,
        2.513274,0.376428,-1.495532,
        2.513274,0.378103,2.646436,
        2.513274,0.385088,0.590355,
        2.513274,0.386248,-2.690943,
        2.513274,0.389851,-2.340031,
        2.513274,0.391502,-1.171374,
        2.513274,0.398249,-1.812694,
        2.513274,0.400542,-2.995729,
        2.513274,0.404273,0.918163,
        2.513274,0.405617,1.773533,
        2.513274,0.409009,-0.570243,
        2.513274,0.410700,0.270433,
        2.513274,0.415893,1.464890,
        2.513274,0.420680,2.088164,
        2.513274,0.424766,-0.258729,
        2.513274,0.426507,2.367872,
        2.513274,0.441292,-0.890727,
        2.513274,0.451839,-2.078193,
        2.513274,0.459270,1.191562,
        2.513274,0.470622,0.014002,
        2.513274,0.481292,-2.529098,
        2.513274,0.482460,2.849843,
        2.513274,0.492205,-1.397963,
        2.513274,0.493314,3.117264,
        2.513274,0.497870,0.727828,
        2.513274,0.499783,-1.654379,
        2.513274,0.502651,0.463333,
        2.513274,0.505463,2.594557,
        2.513274,0.506104,-2.874910,
        2.513274,0.513579,-1.147354,
        2.513274,0.518569,1.644237,
        2.513274,0.519001,1.900488,
        2.513274,0.519277,-0.673809,
        2.513274,0.524769,-0.421955,
        2.513274,0.525564,-2.288838,
        2.513274,0.532541,0.978737,
        2.513274,0.534610,2.154808,
        2.513274,0.535020,-1.891846,
        2.513274,0.540208,1.400497,
        2.513274,0.542006,0.228337,
        2.513274,0.550762,-0.180771,
        2.513274,0.560531,2.378474,
        2.513274,0.561629,-0.924576,
        2.513274,0.567465,-2.679392,
        2.513274,0.588527,1.181755,
        2.513274,0.591643,-2.098364,
        2.513274,0.594963,2.997575,
        2.513274,0.599643,-3.045235,
        2.513274,0.599962,2.773556,
        2.513274,0.602620,0.029270,
        2.513274,0.605711,0.598454,
        2.513274,0.607708,-1.541026,
        2.513274,0.609360,-2.464379,
        2.513274,0.610738,-1.326786,
        2.513274,0.616968,0.820489,
        2.513274,0.626379,1.773610,
        2.513274,0.628626,-1.751180,
        2.513274,0.630671,0.391949,
        2.513274,0.632622,-0.737642,
        2.513274,0.632708,2.566722,
        2.513274,0.632797,1.989858,
        2.513274,0.634455,-0.534790,
        2.513274,0.636205,1.561170,
        2.513274,0.637681,-1.123423,
        2.513274,0.641441,-2.842109,
        2.513274,0.644807,-0.327139,
        2.513274,0.652440,2.197129,
        2.513274,0.660897,-2.278593,
        2.513274,0.664008,1.360091,
        2.513274,0.670850,-1.945257,
        2.513274,0.671396,1.004859,
        2.513274,0.673985,0.203855,
        2.513274,0.680326,-0.133370,
        2.513274,0.685083,-0.942363,
        2.513274,0.688204,-2.632204,
        2.513274,0.689777,2.385011,
        2.513274,0.700065,3.115430,
        2.513274,0.704080,2.911636,
        2.513274,0.712883,1.174216,
        2.513274,0.715701,0.699906,
        2.513274,0.718851,-1.442215,
        2.513274,0.721019,2.721536,
        2.513274,0.725300,-2.975134,
        2.513274,0.727136,-1.633288,
        2.513274,0.727399,-2.120567,
        2.513274,0.730751,0.522488,
        2.513274,0.732892,-1.254330,
        2.513274,0.734333,-2.446184,
        2.513274,0.735125,0.037270,
        2.513274,0.737893,1.866463,
        2.513274,0.740650,1.681075,
        2.513274,0.745971,-0.781286,
        2.513274,0.749432,-0.442559,
        2.513274,0.750102,2.050636,
        2.513274,0.751046,0.863821,
        2.513274,0.757018,-0.608833,
        2.513274,0.757046,-1.816019,
        2.513274,0.757389,1.501184,
        2.513274,0.759800,2.546181,
        2.513274,0.759859,0.351923,
        2.513274,0.761141,-2.781102,
        2.513274,0.769035,-1.078574,
        2.513274,0.770166,-0.266604,
        2.513274,0.772218,2.227380,
        2.513274,0.788382,1.331051,
        2.513274,0.790269,-2.282409,
        2.513274,0.803410,-1.984567,
        2.513274,0.803813,0.191520,
        2.513274,0.806832,3.022155,
        2.513274,0.807476,-2.597675,
        2.513274,0.812512,-0.107335,
        2.513274,0.814161,1.010949,
        2.513274,0.817732,-3.084486,
        2.513274,0.817846,2.390688,
        2.513274,0.819323,2.845960,
        2.513274,0.819430,-0.915017,
        2.513274,0.829198,-1.532491,
        2.513274,0.829531,0.665695,
        2.513274,0.829816,1.167147,
        2.513274,0.833613,-1.357492,
        2.513274,0.842945,-2.907318,
        2.513274,0.844883,2.679708,
        2.513274,0.848189,-1.705003,
        2.513274,0.849888,1.767901,
        2.513274,0.851126,0.493370,
        2.513274,0.852911,1.933394,
        2.513274,0.857516,-2.432863,
        2.513274,0.858261,-1.191729,
        2.513274,0.858279,-2.138895,
        2.513274,0.859723,-0.734928,
        2.513274,0.859754,1.603292,
        2.513274,0.861655,-0.394821,
        2.513274,0.864300,0.043549,
        2.513274,0.866940,0.851229,
        2.513274,0.869748,2.095969,
        2.513274,0.877352,-0.574668,
        2.513274,0.879254,-2.733338,
        2.513274,0.882554,1.443757,
        2.513274,0.884060,0.334199,
        2.513274,0.885626,-1.866061,
        2.513274,0.889834,2.529957,
        2.513274,0.893140,2.251698,
        2.513274,0.897463,-0.240822,
        2.513274,0.898758,-1.031948,
        2.513274,0.913918,1.287618,
        2.513274,0.916555,-2.287173,
        2.513274,0.918305,3.107845,
        2.513274,0.919522,2.948277,
        2.513274,0.927138,-2.569686,
        2.513274,0.929031,0.185383,
        2.513274,0.933766,-3.015917,
        2.513274,0.934873,-2.010822,
        2.513274,0.935636,-1.445855,
        2.513274,0.936702,-0.866461,
        2.513274,0.937395,0.729108,
        2.513274,0.937910,2.787575,
        2.513274,0.939144,-0.092962,
        2.513274,0.941718,1.125085,
        2.513274,0.942873,-1.607680,
        2.513274,0.943701,0.980328,
        2.513274,0.947774,2.395256,
        2.513274,0.952433,0.584347,
        2.513274,0.954456,-1.287293,
        2.513274,0.958731,-2.852552,
        2.513274,0.961838,1.839284,
        2.513274,0.961898,-0.469425,
        2.513274,0.965426,1.686615,
        2.513274,0.970108,1.990771,
        2.513274,0.972820,-1.762780,
        2.513274,0.973500,-0.711881,
        2.513274,0.975253,2.641082,
        2.513274,0.977824,0.441928,
        2.513274,0.981485,-2.424766,
        2.513274,0.981877,1.537100,
        2.513274,0.982609,-1.136920,
        2.513274,0.984265,-2.156073,
        2.513274,0.989150,0.045493,
        2.513274,0.993052,2.137478,
        2.513274,0.997766,-2.693871,
        2.513274,0.999847,-0.334738,
        2.513274,1.001920,0.848054,
        2.513274,1.006146,1.389734,
        2.513274,1.012180,0.304066,
        2.513274,1.016938,2.270674,
        2.513274,1.017371,-0.979790,
        2.513274,1.019495,-1.894487,
        2.513274,1.023796,-0.583691,
        2.513274,1.026525,2.506945,
        2.513274,1.027004,-0.203653,
        2.513274,1.028027,3.036573,
        2.513274,1.031435,1.241021,
        2.513274,1.032316,2.884221,
        2.513274,1.035863,-3.110475,
        2.513274,1.042085,-2.292950,
        2.513274,1.043272,-1.522986,
        2.513274,1.046265,-2.958847,
        2.513274,1.049272,-1.376344,
        2.513274,1.050039,-2.552086,
        2.513274,1.053682,0.689863,
        2.513274,1.055458,0.170394,
        2.513274,1.056622,-0.831866,
        2.513274,1.057515,1.097998,
        2.513274,1.058706,-2.036438,
        2.513274,1.061389,-1.672248,
        2.513274,1.063758,2.743093,
        2.513274,1.070157,-0.073422,
        2.513274,1.072154,1.761506,
        2.513274,1.074029,1.909193,
        2.513274,1.074204,0.545276,
        2.513274,1.074714,-2.805147,
        2.513274,1.076262,0.966104,
        2.513274,1.077642,-1.227073,
        2.513274,1.084151,1.618525,
        2.513274,1.085085,2.377673,
        2.513274,1.086822,-0.441112,
        2.513274,1.087142,2.058815,
        2.513274,1.098383,-0.696430,
        2.513274,1.099138,-1.084033,
        2.513274,1.103208,1.477773,
        2.513274,1.103783,0.408254,
        2.513274,1.106275,-2.420800,
        2.513274,1.108079,2.612943,
        2.513274,1.109345,-2.172718,
        2.513274,1.115957,-1.787597,
        2.513274,1.118930,0.045983,
        2.513274,1.119777,0.834702,
        2.513274,1.121943,-2.669042,
        2.513274,1.124930,2.219917,
        2.513274,1.125217,1.336294,
        2.513274,1.130958,2.972931,
        2.513274,1.131400,-0.306629,
        2.513274,1.137135,-0.936514,
        2.513274,1.139125,-1.924528,
        2.513274,1.139877,0.276382,
        2.513274,1.142003,-3.058022,
        2.513274,1.147363,-0.558448,
        2.513274,1.150348,-0.176418,
        2.513274,1.151198,1.197016,
        2.513274,1.153930,-1.597104,
        2.513274,1.155389,2.837148,
        2.513274,1.155458,-1.460623,
        2.513274,1.157331,-2.906362,
        2.513274,1.158594,2.487618,
        2.513274,1.162103,-1.322553,
        2.513274,1.165407,3.092861,
        2.513274,1.168025,0.717075,
        2.513274,1.168406,-2.301519,
        2.513274,1.174782,-2.538755,
        2.513274,1.177998,1.061256,
        2.513274,1.178898,1.827305,
        2.513274,1.181331,-2.060330,
        2.513274,1.182540,-0.798952,
        2.513274,1.184569,0.147517,
        2.513274,1.184734,1.968919,
        2.513274,1.184804,0.593386,
        2.513274,1.188278,1.690808,
        2.513274,1.194360,2.110440,
        2.513274,1.194776,2.710455,
        2.513274,1.197745,-1.173705,
        2.513274,1.199898,-2.777257,
        2.513274,1.201164,2.355931,
        2.513274,1.203760,1.556404,
        2.513274,1.205845,-0.415993,
        2.513274,1.206585,0.933648,
        2.513274,1.206950,-0.037248,
        2.513274,1.209341,0.471162,
        2.513274,1.213483,-1.039417,
        2.513274,1.218146,-0.672643,
        2.513274,1.221608,1.421669,
        2.513274,1.226043,-1.693533,
        2.513274,1.232870,-1.822773,
        2.513274,1.234375,-2.421194,
        2.513274,1.234531,-2.190148,
        2.513274,1.239414,0.348959,
        2.513274,1.240183,2.587255,
        2.513274,1.242312,2.941234,
        2.513274,1.244075,1.286644,
        2.513274,1.245961,-2.997461,
        2.513274,1.246495,-2.646872,
        2.513274,1.247736,2.242810,
        2.513274,1.253604,0.812222,
        2.513274,1.257816,-0.894317,
        2.513274,1.258247,-1.953634,
        2.513274,1.258993,-0.284099,
        2.513274,1.262380,-0.157609,
        2.513274,1.262791,-1.549229,
        2.513274,1.263772,-0.533073,
        2.513274,1.266731,-1.412461,
        2.513274,1.268591,-3.122146,
        2.513274,1.269614,-1.275246,
        2.513274,1.270745,1.152533,
        2.513274,1.273606,0.224545,
        2.513274,1.279716,2.812101,
        2.513274,1.283786,2.462124,
        2.513274,1.284798,0.078209,
        2.513274,1.285402,-2.882049,
        2.513274,1.286936,1.886053,
        2.513274,1.289408,0.683122,
        2.513274,1.294736,1.756184,
        2.513274,1.296978,-2.311511,
        2.513274,1.298546,2.020601,
        2.513274,1.300924,1.022478,
        2.513274,1.303305,3.043152,
        2.513274,1.304210,-2.084225,
        2.513274,1.305241,-2.530249,
        2.513274,1.307513,1.628003,
        2.513274,1.310722,0.558267,
        2.513274,1.314142,2.139329,
        2.513274,1.316087,-0.777971,
        2.513274,1.317907,-1.121023,
        2.513274,1.321019,-2.749776,
        2.513274,1.321609,2.687717,
        2.513274,1.322568,1.499415,
        2.513274,1.322972,-0.395699,
        2.513274,1.325286,-0.994192,
        2.513274,1.329146,-0.648513,
        2.513274,1.329795,-0.042570,
        2.513274,1.331651,2.344497,
        2.513274,1.336522,0.435098,
        2.513274,1.339410,0.900087,
        2.513274,1.339550,1.370353,
        2.513274,1.340441,-1.731580,
        2.513274,1.351855,-1.855234,
        2.513274,1.360161,-3.018404,
        2.513274,1.361031,-1.613971,
        2.513274,1.361196,2.914780,
        2.513274,1.361998,-2.207254,
        2.513274,1.362302,1.239384,
        2.513274,1.365228,2.563808,
        2.513274,1.365452,0.310522,
        2.513274,1.367028,-2.420041,
        2.513274,1.373422,-1.484281,
        2.513274,1.374245,-1.226417,
        2.513274,1.375020,-0.170393,
        2.513274,1.376560,-1.355481,
        2.513274,1.377070,-1.980984,
        2.513274,1.377578,-2.636339,
        2.513274,1.378096,0.774626,
        2.513274,1.379086,-0.508192,
        2.513274,1.382562,0.177141,
        2.513274,1.388849,2.233847,
        2.513274,1.389268,-0.886349,
        2.513274,1.391324,1.109396,
        2.513274,1.391699,-0.288725,
        2.513274,1.392869,3.136903,
        2.513274,1.393612,1.952016,
        2.513274,1.395545,-2.850807,
        2.513274,1.401926,1.822930,
        2.513274,1.402149,2.788898,
        2.513274,1.409552,0.647980,
        2.513274,1.410840,0.056219,
        2.513274,1.411944,2.444377,
        2.513274,1.413639,1.697982,
        2.513274,1.426479,1.573427,
        2.513274,1.426725,2.102242,
        2.513274,1.427868,0.984637,
        2.513274,1.429580,-2.107232,
        2.513274,1.431375,-2.313269,
        2.513274,1.431907,3.015410,
        2.513274,1.432325,-1.029054,
        2.513274,1.433341,0.524268,
        2.513274,1.435181,-0.739771,
        2.513274,1.437208,-2.529085,
        2.513274,1.439041,-0.617966,
        2.513274,1.441473,1.448544,
        2.513274,1.444527,2.664744,
        2.513274,1.450513,-0.400625,
        2.513274,1.451907,-0.066823,
        2.513274,1.452754,-2.738631,
        2.513274,1.454061,-2.950416,
        2.513274,1.456736,1.323154,
        2.513274,1.458370,0.400433,
        2.513274,1.459314,-1.765003,
        2.513274,1.459902,-1.145835,
        2.513274,1.464952,2.329245,
        2.513274,1.466050,0.862104,
        2.513274,1.471330,-1.886160,
        2.513274,1.475119,-1.648255,
        2.513274,1.480878,1.194576,
        2.513274,1.480942,0.273235,
        2.513274,1.481086,2.890269,
        2.513274,1.484118,-1.529156,
        2.513274,1.484741,-1.279219,
        2.513274,1.487681,-1.406069,
        2.513274,1.487886,-3.066037,
        2.513274,1.490555,2.543284,
        2.513274,1.492424,-0.190932,
        2.513274,1.495002,-0.839810,
        2.513274,1.495297,2.007188,
        2.513274,1.496414,-2.009267,
        2.513274,1.498261,-2.425270,
        2.513274,1.498907,-2.207327,
        2.513274,1.500204,0.737638,
        2.513274,1.502469,0.151156,
        2.513274,1.507448,-0.514782,
        2.513274,1.509287,2.204647,
        2.513274,1.510518,1.884920,
        2.513274,1.511948,-2.632360,
        2.513274,1.515235,1.068565,
        2.513274,1.522237,3.102018,
        2.513274,1.522373,1.764991,
        2.513274,1.522416,-0.954956,
        2.513274,1.523102,2.765920,
        2.513274,1.524811,-2.844718,
        2.513274,1.527261,-0.304916,
        2.513274,1.529788,0.613208,
        2.513274,1.533648,1.644808,
        2.513274,1.535018,0.028869,
        2.513274,1.541082,2.424608,
        2.513274,1.546010,1.523462,
        2.513274,1.551135,-0.642994,
        2.513274,1.553895,0.489607,
        2.513274,1.555041,0.947813,
        2.513274,1.559671,-1.074934,
        2.513274,1.560345,2.985899,
        2.513274,1.560487,-2.323400,
        2.513274,1.561127,1.401111,
        2.513274,1.567193,2.643172,
        2.513274,1.568892,2.098362,
        2.513274,1.571702,-2.098435,
        2.513274,1.572701,-2.528730,
        2.513274,1.573572,1.278593,
        2.513274,1.574849,-0.091505,
        2.513274,1.576334,0.365236,
        2.513274,1.577029,-2.962888,
        2.513274,1.577688,-1.814941,
        2.513274,1.579542,-0.755917,
        2.513274,1.579624,-1.199759,
        2.513274,1.581515,-0.419414,
        2.513274,1.584581,-2.738702,
        2.513274,1.586948,-1.696296,
        2.513274,1.588901,2.301496,
        2.513274,1.589936,0.825020,
        2.513274,1.590018,-1.941652,
        2.513274,1.596968,-1.328358,
        2.513274,1.597178,-1.575607,
        2.513274,1.600470,-1.453196,
        2.513274,1.601408,0.242409,
        2.513274,1.602259,1.152656,
        2.513274,1.603064,2.865774,
        2.513274,1.612085,1.968802,
        2.513274,1.612720,-3.080498,
        2.513274,1.616061,2.521531,
        2.513274,1.616900,-0.207457,
        2.513274,1.620988,-0.875920,
        2.513274,1.622114,0.701509,
        2.513274,1.622749,0.121426,
        2.513274,1.622993,-2.221629,
        2.513274,1.629346,-0.542669,
        2.513274,1.630007,1.844552,
        2.513274,1.634747,-2.426198,
        2.513274,1.642346,1.722617,
        2.513274,1.643925,1.034629,
        2.513274,1.644848,2.743774,
        2.513274,1.645258,2.188109,
        2.513274,1.645417,-2.634725,
        2.513274,1.647395,3.085437,
        2.513274,1.648538,-2.855044,
        2.513274,1.649932,0.576808,
        2.513274,1.652224,-0.996105,
        2.513274,1.652871,1.600681,
        2.513274,1.657997,-0.325127,
        2.513274,1.661936,0.001037,
        2.513274,1.665187,1.477402,
        2.513274,1.665496,2.397872,
        2.513274,1.673267,-0.665453,
        2.513274,1.673557,0.452526,
        2.513274,1.679277,-1.122421,
        2.513274,1.679696,0.912482,
        2.513274,1.682420,1.353922,
        2.513274,1.685031,-2.009038,
        2.513274,1.685196,2.967046,
        2.513274,1.690107,2.620910,
        2.513274,1.690453,-2.123076,
        2.513274,1.691693,-1.871572,
        2.513274,1.691817,2.061688,
        2.513274,1.692644,1.237670,
        2.513274,1.696386,0.328123,
        2.513274,1.697162,-1.249026,
        2.513274,1.697488,-2.322785,
        2.513274,1.698726,-2.973576,
        2.513274,1.699172,-1.749717,
        2.513274,1.704478,-0.106295,
        2.513274,1.707244,-0.444770,
        2.513274,1.707500,-2.531497,
        2.513274,1.707714,-0.787321,
        2.513274,1.708430,-1.627552,
        2.513274,1.710583,-1.376890,
        2.513274,1.712415,-2.747921,
        2.513274,1.713125,-1.503073,
        2.513274,1.713452,0.789419,
        2.513274,1.720612,2.279888,
        2.513274,1.724852,0.202831,
        2.513274,1.724868,2.846246,
        2.513274,1.728940,1.931535,
        2.513274,1.732838,1.127584,
        2.513274,1.735719,-3.092485,
        2.513274,1.739676,2.496017,
        2.513274,1.739971,-0.228876,
        2.513274,1.743328,0.663347,
        2.513274,1.744927,-0.911892,
        2.513274,1.745952,0.084733,
        2.513274,1.748077,1.804354,
        2.513274,1.752959,-0.567548,
        2.513274,1.760169,1.678132,
        2.513274,1.763887,-2.218450,
        2.513274,1.766737,2.722752,
        2.513274,1.767999,1.004737,
        2.513274,1.769149,1.550298,
        2.513274,1.770119,-2.427175,
        2.513274,1.770460,2.154208,
        2.513274,1.770514,3.070982,
        2.513274,1.770735,0.537245,
        2.513274,1.772673,-2.864760,
        2.513274,1.773527,-1.039582,
        2.513274,1.776419,-2.642421,
        2.513274,1.780769,1.416330,
        2.513274,1.786634,-0.346247,
        2.513274,1.788829,-1.947553,
        2.513274,1.793737,0.410270,
        2.513274,1.794553,2.374618,
        2.513274,1.794699,-0.694061,
        2.513274,1.797469,-1.169105,
        2.513274,1.803705,0.881076,
        2.513274,1.804124,1.264249,
        2.513274,1.807825,-2.090455,
        2.513274,1.808822,-1.817375,
        2.513274,1.809020,2.949741,
        2.513274,1.812466,2.597595,
        2.513274,1.813159,2.025395,
        2.513274,1.813816,-1.298435,
        2.513274,1.815016,0.278951,
        2.513274,1.816510,-0.008631,
        2.513274,1.817772,-1.689625,
        2.513274,1.821360,-2.984352,
        2.513274,1.822312,-0.127211,
        2.513274,1.823386,-1.429807,
        2.513274,1.823834,-1.560483,
        2.513274,1.832060,-0.821627,
        2.513274,1.832690,-0.468440,
        2.513274,1.833945,-2.320278,
        2.513274,1.834787,0.751968,
        2.513274,1.840579,-2.536445,
        2.513274,1.841033,-2.756932,
        2.513274,1.846350,1.893983,
        2.513274,1.846644,1.113766,
        2.513274,1.846833,2.826543,
        2.513274,1.847325,2.249051,
        2.513274,1.854119,0.134161,
        2.513274,1.858863,-3.106022,
        2.513274,1.865046,1.763384,
        2.513274,1.865619,-0.952036,
        2.513274,1.865912,0.624453,
        2.513274,1.866694,2.473686,
        2.513274,1.867524,-0.240083,
        2.513274,1.875561,1.632831,
        2.513274,1.876049,-0.594378,
        2.513274,1.880509,1.500204,
        2.513274,1.883504,1.360457,
        2.513274,1.886971,-2.196665,
        2.513274,1.888979,0.982922,
        2.513274,1.889815,2.701570,
        2.513274,1.892144,-2.005340,
        2.513274,1.892348,0.496522,
        2.513274,1.893281,-1.083976,
        2.513274,1.893828,3.053453,
        2.513274,1.895117,2.121107,
        2.513274,1.898741,-2.875735,
        2.513274,1.904429,-2.428029,
        2.513274,1.908408,-2.648903,
        2.513274,1.911828,-0.364489,
        2.513274,1.912224,0.366565,
        2.513274,1.914854,1.214468,
        2.513274,1.915467,-1.883143,
        2.513274,1.915763,-0.724679,
        2.513274,1.915790,-1.216366,
        2.513274,1.922256,2.347806,
        2.513274,1.922616,0.844633,
        2.513274,1.925185,0.231170,
        2.513274,1.927474,-1.754616,
        2.513274,1.928342,-1.351174,
        2.513274,1.933328,-1.623328,
        2.513274,1.933878,0.034150,
        2.513274,1.934585,-1.487979,
        2.513274,1.934802,2.927634,
        2.513274,1.934862,1.989776,
        2.513274,1.936327,-0.099350,
        2.513274,1.938088,2.578507,
        2.513274,1.945677,-2.999805,
        2.513274,1.954572,-0.859208,
        2.513274,1.957868,-0.490784,
        2.513274,1.958989,0.712772,
        2.513274,1.960212,-2.307561,
        2.513274,1.964774,1.855507,
        2.513274,1.964971,-2.100533,
        2.513274,1.969112,1.088273,
        2.513274,1.971438,-2.765415,
        2.513274,1.972312,2.805134,
        2.513274,1.974950,2.220596,
        2.513274,1.975135,-2.539127,
        2.513274,1.982107,1.720829,
        2.513274,1.983195,-3.127984,
        2.513274,1.983318,-0.247182,
        2.513274,1.986720,-0.993220,
        2.513274,1.990381,1.586005,
        2.513274,1.990470,0.580283,
        2.513274,1.990523,1.450326,
        2.513274,1.993909,1.312897,
        2.513274,1.994694,2.452813,
        2.513274,1.999226,-0.620717,
        2.513274,2.005145,0.941012,
        2.513274,2.007301,0.142337,
        2.513274,2.014149,0.444138,
        2.513274,2.014338,-1.127392,
        2.513274,2.018131,3.021778,
        2.513274,2.019676,-1.980944,
        2.513274,2.019912,2.686612,
        2.513274,2.020545,2.088672,
        2.513274,2.026074,0.300945,
        2.513274,2.026196,-2.889057,
        2.513274,2.031959,-1.268178,
        2.513274,2.032917,-2.199917,
        2.513274,2.033572,-1.842884,
        2.513274,2.034712,-2.419505,
        2.513274,2.035818,-0.379479,
        2.513274,2.036004,-0.757062,
        2.513274,2.040501,-1.411488,
        2.513274,2.042874,-1.701367,
        2.513274,2.043060,-2.653746,
        2.513274,2.043729,-1.557938,
        2.513274,2.048538,0.806623,
        2.513274,2.049006,1.199930,
        2.513274,2.049502,-0.137989,
        2.513274,2.052181,2.326959,
        2.513274,2.052794,-0.008526,
        2.513274,2.057608,1.951726,
        2.513274,2.065568,2.567730,
        2.513274,2.070371,2.879808,
        2.513274,2.071256,-3.018961,
        2.513274,2.079746,-0.897699,
        2.513274,2.082784,1.042403,
        2.513274,2.083752,-0.511086,
        2.513274,2.084401,1.810502,
        2.513274,2.085731,0.671081,
        2.513274,2.097496,1.373172,
        2.513274,2.099185,1.667006,
        2.513274,2.100482,-2.301084,
        2.513274,2.103118,2.194745,
        2.513274,2.103223,1.521685,
        2.513274,2.103695,-2.777707,
        2.513274,2.107523,3.129581,
        2.513274,2.108277,-2.095907,
        2.513274,2.108891,-2.532166,
        2.513274,2.111003,-0.268632,
        2.513274,2.112276,-1.027246,
        2.513274,2.114380,0.531457,
        2.513274,2.119166,0.100552,
        2.513274,2.119564,2.754818,
        2.513274,2.122474,-0.648314,
        2.513274,2.125183,2.445598,
        2.513274,2.129501,0.238290,
        2.513274,2.131178,0.386663,
        2.513274,2.131223,-1.174311,
        2.513274,2.134269,-1.948042,
        2.513274,2.134684,0.907065,
        2.513274,2.137615,2.988806,
        2.513274,2.145635,-1.327708,
        2.513274,2.146467,2.055081,
        2.513274,2.146582,-1.793788,
        2.513274,2.149210,-1.482321,
        2.513274,2.152901,-1.636788,
        2.513274,2.156337,1.151558,
        2.513274,2.156474,-2.912065,
        2.513274,2.157194,-0.792563,
        2.513274,2.166724,-0.401638,
        2.513274,2.166980,-0.134879,
        2.513274,2.173644,-2.403429,
        2.513274,2.175711,-2.655523,
        2.513274,2.177132,2.600419,
        2.513274,2.178879,0.767722,
        2.513274,2.180752,1.908485,
        2.513274,2.181338,2.313721,
        2.513274,2.185580,1.287673,
        2.513274,2.195400,-2.217546,
        2.513274,2.200418,-3.056442,
        2.513274,2.201829,-0.001735,
        2.513274,2.203804,1.756176,
        2.513274,2.211602,1.443930,
        2.513274,2.212686,-0.544809,
        2.513274,2.213937,0.621892,
        2.513274,2.214075,1.600691,
        2.513274,2.216256,1.020029,
        2.513274,2.219060,2.886631,
        2.513274,2.223375,-0.906185,
        2.513274,2.226002,-1.060248,
        2.513274,2.226413,-2.064034,
        2.513274,2.230554,2.171334,
        2.513274,2.232413,2.735777,
        2.513274,2.233709,3.070830,
        2.513274,2.234383,-2.792065,
        2.513274,2.237183,0.468799,
        2.513274,2.237695,-0.271388,
        2.513274,2.239419,0.151693,
        2.513274,2.243347,-2.521049,
        2.513274,2.244035,-1.229456,
        2.513274,2.244355,0.309899,
        2.513274,2.245175,2.460236,
        2.513274,2.245628,-1.899216,
        2.513274,2.245973,-0.702234,
        2.513274,2.255601,-1.728131,
        2.513274,2.256702,-1.401367,
        2.513274,2.261466,-1.557500,
        2.513274,2.267390,0.878499,
        2.513274,2.272099,2.018594,
        2.513274,2.278618,1.170386,
        2.513274,2.283304,-2.345584,
        2.513274,2.285175,-2.943090,
        2.513274,2.293000,-0.124451,
        2.513274,2.294058,-0.421070,
        2.513274,2.303283,1.855938,
        2.513274,2.304968,2.608978,
        2.513274,2.305347,2.311346,
        2.513274,2.307188,-2.656134,
        2.513274,2.307301,1.336200,
        2.513274,2.309655,0.726020,
        2.513274,2.317634,-2.185225,
        2.513274,2.321510,1.685664,
        2.513274,2.324376,-3.111006,
        2.513274,2.324731,1.511186,
        2.513274,2.329587,0.035498,
        2.513274,2.330158,2.977264,
        2.513274,2.335096,-0.588468,
        2.513274,2.336046,-1.115799,
        2.513274,2.338060,2.803613,
        2.513274,2.338083,-0.943400,
        2.513274,2.340210,0.562145,
        2.513274,2.341931,1.023926,
        2.513274,2.343265,-0.779061,
        2.513274,2.343434,-2.013692,
        2.513274,2.353029,-1.310529,
        2.513274,2.354868,0.211543,
        2.513274,2.355120,0.388593,
        2.513274,2.355165,-1.833173,
        2.513274,2.355462,2.151441,
        2.513274,2.362948,-2.487477,
        2.513274,2.363101,-2.810456,
        2.513274,2.363394,-0.272918,
        2.513274,2.364509,-1.644533,
        2.513274,2.370465,2.466824,
        2.513274,2.386110,-1.481351,
        2.513274,2.394960,0.862133,
        2.513274,2.395716,1.977184,
        2.513274,2.396339,1.200900,
        2.513274,2.407941,-2.316753,
        2.513274,2.409450,-2.988432,
        2.513274,2.417202,-0.103363,
        2.513274,2.417386,-0.443097,
        2.513274,2.417950,2.666662,
        2.513274,2.423190,1.789783,
        2.513274,2.423692,1.392696,
        2.513274,2.431122,3.086379,
        2.513274,2.431406,-2.647820,
        2.513274,2.431584,2.307631,
        2.513274,2.434090,1.592818,
        2.513274,2.435993,0.683844,
        2.513274,2.440898,-2.133300,
        2.513274,2.442222,-1.195837,
        2.513274,2.443112,2.883761,
        2.513274,2.449191,-0.633651,
        2.513274,2.450821,0.084741,
        2.513274,2.453499,-1.008861,
        2.513274,2.460010,-1.938011,
        2.513274,2.460199,-0.822008,
        2.513274,2.460902,0.490079,
        2.513274,2.461774,-1.734913,
        2.513274,2.463986,1.034407,
        2.513274,2.467647,0.287667,
        2.513274,2.479849,2.126925,
        2.513274,2.485907,-2.835471,
        2.513274,2.485945,-1.374853,
        2.513274,2.487747,2.505912,
        2.513274,2.487957,-0.274279,
        2.513274,2.489741,-2.468428,
        2.513274,2.512538,1.239900,
        2.513274,2.517528,1.923289,
        2.513274,2.518454,-1.565852,
        2.513274,2.519348,0.843503,
        2.513274,2.522766,-3.059231,
        2.513274,2.528686,2.737228,
        2.513274,2.534798,-2.266671,
        2.513274,2.536533,1.466274,
        2.513274,2.538504,1.699812,
        2.513274,2.539516,-0.476942,
        2.513274,2.539532,-0.074548,
        2.513274,2.543197,2.990252,
        2.513274,2.554022,2.320619,
        2.513274,2.556919,-2.655005,
        2.513274,2.558146,0.627191,
        2.513274,2.558223,-1.164189,
        2.513274,2.563452,-2.044028,
        2.513274,2.568462,0.150655,
        2.513274,2.569130,-0.700837,
        2.513274,2.570050,-1.804492,
        2.513274,2.572710,-0.934665,
        2.513274,2.575994,0.391718,
        2.513274,2.584774,1.046703,
        2.513274,2.601432,2.102535,
        2.513274,2.603559,2.555114,
        2.513274,2.604150,-2.885208,
        2.513274,2.608630,-1.378011,
        2.513274,2.611256,-0.279110,
        2.513274,2.617344,-2.439480,
        2.513274,2.627402,1.293552,
        2.513274,2.635396,1.848850,
        2.513274,2.635686,2.834950,
        2.513274,2.637087,3.132730,
        2.513274,2.642029,0.813274,
        2.513274,2.645260,1.570203,
        2.513274,2.646182,-1.614036,
        2.513274,2.660113,-2.186513,
        2.513274,2.660483,-0.034608,
        2.513274,2.661713,-0.529511,
        2.513274,2.672720,-1.110422,
        2.513274,2.676626,2.335353,
        2.513274,2.676638,0.540364,
        2.513274,2.676780,-1.901158,
        2.513274,2.681411,-2.667965,
        2.513274,2.682152,0.245751,
        2.513274,2.682859,-0.817071,
        2.513274,2.705107,1.062142,
        2.513274,2.717705,2.625995,
        2.513274,2.721001,2.060097,
        2.513274,2.722319,-2.954069,
        2.513274,2.729840,-1.368237,
        2.513274,2.734349,-0.289298,
        2.513274,2.738409,2.975671,
        2.513274,2.740366,1.374341,
        2.513274,2.744633,-2.391311,
        2.513274,2.747398,1.727166,
        2.513274,2.763799,0.760372,
        2.513274,2.768238,-1.679243,
        2.513274,2.779245,-2.046268,
        2.513274,2.780208,-0.626453,
        2.513274,2.780471,0.027732,
        2.513274,2.784696,-1.013128,
        2.513274,2.789817,0.395431,
        2.513274,2.801154,2.353143,
        2.513274,2.807091,-2.687477,
        2.513274,2.825762,1.085035,
        2.513274,2.830209,2.734943,
        2.513274,2.833517,-3.085508,
        2.513274,2.838301,1.980061,
        2.513274,2.849871,1.512012,
        2.513274,2.850089,-1.343899,
        2.513274,2.856905,-0.318441,
        2.513274,2.870445,-2.291001,
        2.513274,2.885017,0.655562,
        2.513274,2.886428,-1.793217,
        2.513274,2.890337,-0.825633,
        2.513274,2.898932,0.139973,
        2.513274,2.931927,-2.728768,
        2.513274,2.932096,2.377790,
        2.513274,2.937409,2.947245,
        2.513274,2.948048,1.133127,
        2.513274,2.950896,1.797889,
        2.513274,2.970019,-1.282382,
        2.513274,2.978333,-0.412090,
        2.513274,2.996437,-2.089121,
        2.513274,3.011274,0.434110,
        2.513274,3.050410,3.138492,
        2.513274,3.064492,1.647259,
        2.513274,3.088868,-0.989501,
        2.617994,0.053884,-2.408394,
        2.617994,0.077645,1.453433,
        2.617994,0.088408,-0.311160,
        2.617994,0.142314,2.590387,
        2.617994,0.147206,-1.385164,
        2.617994,0.163506,0.573203,
        2.617994,0.165485,-2.875396,
        2.617994,0.182564,-2.125821,
        2.617994,0.190569,1.929141,
        2.617994,0.201102,1.249082,
        2.617994,0.201397,-0.716646,
        2.617994,0.210568,-0.069559,
        2.617994,0.251672,2.848900,
        2.617994,0.266262,-1.678068,
        2.617994,0.269108,-1.215563,
        2.617994,0.273777,-2.506756,
        2.617994,0.279854,0.814581,
        2.617994,0.281664,-2.969147,
        2.617994,0.283625,2.405603,
        2.617994,0.284058,0.357350,
        2.617994,0.294107,1.593880,
        2.617994,0.302491,-0.400410,
        2.617994,0.308059,2.021491,
        2.617994,0.312219,-2.072057,
        2.617994,0.321905,-0.829121,
        2.617994,0.329984,1.201621,
        2.617994,0.339714,-0.013712,
        2.617994,0.368784,2.974974,
        2.617994,0.376428,-1.495532,
        2.617994,0.378103,2.646436,
        2.617994,0.385088,0.590355,
        2.617994,0.386248,-2.690943,
        2.617994,0.389851,-2.340031,
        2.617994,0.391502,-1.171374,
        2.617994,0.398249,-1.812694,
        2.617994,0.400542,-2.995729,
        2.617994,0.404273,0.918163,
        2.617994,0.405617,1.773533,
        2.617994,0.409009,-0.570243,
        2.617994,0.410700,0.270433,
        2.617994,0.415893,1.464890,
        2.617994,0.420680,2.088164,
        2.617994,0.424766,-0.258729,
        2.617994,0.426507,2.367872,
        2.617994,0.441292,-0.890727,
        2.617994,0.451839,-2.078193,
        2.617994,0.459270,1.191562,
        2.617994,0.470622,0.014002,
        2.617994,0.481292,-2.529098,
        2.617994,0.482460,2.849843,
        2.617994,0.492205,-1.397963,
        2.617994,0.493314,3.117264,
        2.617994,0.497870,0.727828,
        2.617994,0.499783,-1.654379,
        2.617994,0.502651,0.463333,
        2.617994,0.505463,2.594557,
        2.617994,0.506104,-2.874910,
        2.617994,0.513579,-1.147354,
        2.617994,0.518569,1.644237,
        2.617994,0.519001,1.900488,
        2.617994,0.519277,-0.673809,
        2.617994,0.524769,-0.421955,
        2.617994,0.525564,-2.288838,
        2.617994,0.532541,0.978737,
        2.617994,0.534610,2.154808,
        2.617994,0.535020,-1.891846,
        2.617994,0.540208,1.400497,
        2.617994,0.542006,0.228337,
        2.617994,0.550762,-0.180771,
        2.617994,0.560531,2.378474,
        2.617994,0.561629,-0.924576,
        2.617994,0.567465,-2.679392,
        2.617994,0.588527,1.181755,
        2.617994,0.591643,-2.098364,
        2.617994,0.594963,2.997575,
        2.617994,0.599643,-3.045235,
        2.617994,0.599962,2.773556,
        2.617994,0.602620,0.029270,
        2.617994,0.605711,0.598454,
        2.617994,0.607708,-1.541026,
        2.617994,0.609360,-2.464379,
        2.617994,0.610738,-1.326786,
        2.617994,0.616968,0.820489,
        2.617994,0.626379,1.773610,
        2.617994,0.628626,-1.751180,
        2.617994,0.630671,0.391949,
        2.617994,0.632622,-0.737642,
        2.617994,0.632708,2.566722,
        2.617994,0.632797,1.989858,
        2.617994,0.634455,-0.534790,
        2.617994,0.636205,1.561170,
        2.617994,0.637681,-1.123423,
        2.617994,0.641441,-2.842109,
        2.617994,0.644807,-0.327139,
        2.617994,0.652440,2.197129,
        2.617994,0.660897,-2.278593,
        2.617994,0.664008,1.360091,
        2.617994,0.670850,-1.945257,
        2.617994,0.671396,1.004859,
        2.617994,0.673985,0.203855,
        2.617994,0.680326,-0.133370,
        2.617994,0.685083,-0.942363,
        2.617994,0.688204,-2.632204,
        2.617994,0.689777,2.385011,
        2.617994,0.700065,3.115430,
        2.617994,0.704080,2.911636,
        2.617994,0.712883,1.174216,
        2.617994,0.715701,0.699906,
        2.617994,0.718851,-1.442215,
        2.617994,0.721019,2.721536,
        2.617994,0.725300,-2.975134,
        2.617994,0.727136,-1.633288,
        2.617994,0.727399,-2.120567,
        2.617994,0.730751,0.522488,
        2.617994,0.732892,-1.254330,
        2.617994,0.734333,-2.446184,
        2.617994,0.735125,0.037270,
        2.617994,0.737893,1.866463,
        2.617994,0.740650,1.681075,
        2.617994,0.745971,-0.781286,
        2.617994,0.749432,-0.442559,
        2.617994,0.750102,2.050636,
        2.617994,0.751046,0.863821,
        2.617994,0.757018,-0.608833,
        2.617994,0.757046,-1.816019,
        2.617994,0.757389,1.501184,
        2.617994,0.759800,2.546181,
        2.617994,0.759859,0.351923,
        2.617994,0.761141,-2.781102,
        2.617994,0.769035,-1.078574,
        2.617994,0.770166,-0.266604,
        2.617994,0.772218,2.227380,
        2.617994,0.788382,1.331051,
        2.617994,0.790269,-2.282409,
        2.617994,0.803410,-1.984567,
        2.617994,0.803813,0.191520,
        2.617994,0.806832,3.022155,
        2.617994,0.807476,-2.597675,
        2.617994,0.812512,-0.107335,
        2.617994,0.814161,1.010949,
        2.617994,0.817732,-3.084486,
        2.617994,0.817846,2.390688,
        2.617994,0.819323,2.845960,
        2.617994,0.819430,-0.915017,
        2.617994,0.829198,-1.532491,
        2.617994,0.829531,0.665695,
        2.617994,0.829816,1.167147,
        2.617994,0.833613,-1.357492,
        2.617994,0.842945,-2.907318,
        2.617994,0.844883,2.679708,
        2.617994,0.848189,-1.705003,
        2.617994,0.849888,1.767901,
        2.617994,0.851126,0.493370,
        2.617994,0.852911,1.933394,
        2.617994,0.857516,-2.432863,
        2.617994,0.858261,-1.191729,
        2.617994,0.858279,-2.138895,
        2.617994,0.859723,-0.734928,
        2.617994,0.859754,1.603292,
        2.617994,0.861655,-0.394821,
        2.617994,0.864300,0.043549,
        2.617994,0.866940,0.851229,
        2.617994,0.869748,2.095969,
        2.617994,0.877352,-0.574668,
        2.617994,0.879254,-2.733338,
        2.617994,0.882554,1.443757,
        2.617994,0.884060,0.334199,
        2.617994,0.885626,-1.866061,
        2.617994,0.889834,2.529957,
        2.617994,0.893140,2.251698,
        2.617994,0.897463,-0.240822,
        2.617994,0.898758,-1.031948,
        2.617994,0.913918,1.287618,
        2.617994,0.916555,-2.287173,
        2.617994,0.918305,3.107845,
        2.617994,0.919522,2.948277,
        2.617994,0.927138,-2.569686,
        2.617994,0.929031,0.185383,
        2.617994,0.933766,-3.015917,
        2.617994,0.934873,-2.010822,
        2.617994,0.935636,-1.445855,
        2.617994,0.936702,-0.866461,
        2.617994,0.937395,0.729108,
        2.617994,0.937910,2.787575,
        2.617994,0.939144,-0.092962,
        2.617994,0.941718,1.125085,
        2.617994,0.942873,-1.607680,
        2.617994,0.943701,0.980328,
        2.617994,0.947774,2.395256,
        2.617994,0.952433,0.584347,
        2.617994,0.954456,-1.287293,
        2.617994,0.958731,-2.852552,
        2.617994,0.961838,1.839284,
        2.617994,0.961898,-0.469425,
        2.617994,0.965426,1.686615,
        2.617994,0.970108,1.990771,
        2.617994,0.972820,-1.762780,
        2.617994,0.973500,-0.711881,
        2.617994,0.975253,2.641082,
        2.617994,0.977824,0.441928,
        2.617994,0.981485,-2.424766,
        2.617994,0.981877,1.537100,
        2.617994,0.982609,-1.136920,
        2.617994,0.984265,-2.156073,
        2.617994,0.989150,0.045493,
        2.617994,0.993052,2.137478,
        2.617994,0.997766,-2.693871,
        2.617994,0.999847,-0.334738,
        2.617994,1.001920,0.848054,
        2.617994,1.006146,1.389734,
        2.617994,1.012180,0.304066,
        2.617994,1.016938,2.270674,
        2.617994,1.017371,-0.979790,
        2.617994,1.019495,-1.894487,
        2.617994,1.023796,-0.583691,
        2.617994,1.026525,2.506945,
        2.617994,1.027004,-0.203653,
        2.617994,1.028027,3.036573,
        2.617994,1.031435,1.241021,
        2.617994,1.032316,2.884221,
        2.617994,1.035863,-3.110475,
        2.617994,1.042085,-2.292950,
        2.617994,1.043272,-1.522986,
        2.617994,1.046265,-2.958847,
        2.617994,1.049272,-1.376344,
        2.617994,1.050039,-2.552086,
        2.617994,1.053682,0.689863,
        2.617994,1.055458,0.170394,
        2.617994,1.056622,-0.831866,
        2.617994,1.057515,1.097998,
        2.617994,1.058706,-2.036438,
        2.617994,1.061389,-1.672248,
        2.617994,1.063758,2.743093,
        2.617994,1.070157,-0.073422,
        2.617994,1.072154,1.761506,
        2.617994,1.074029,1.909193,
        2.617994,1.074204,0.545276,
        2.617994,1.074714,-2.805147,
        2.617994,1.076262,0.966104,
        2.617994,1.077642,-1.227073,
        2.617994,1.084151,1.618525,
        2.617994,1.085085,2.377673,
        2.617994,1.086822,-0.441112,
        2.617994,1.087142,2.058815,
        2.617994,1.098383,-0.696430,
        2.617994,1.099138,-1.084033,
        2.617994,1.103208,1.477773,
        2.617994,1.103783,0.408254,
        2.617994,1.106275,-2.420800,
        2.617994,1.108079,2.612943,
        2.617994,1.109345,-2.172718,
        2.617994,1.115957,-1.787597,
        2.617994,1.118930,0.045983,
        2.617994,1.119777,0.834702,
        2.617994,1.121943,-2.669042,
        2.617994,1.124930,2.219917,
        2.617994,1.125217,1.336294,
        2.617994,1.130958,2.972931,
        2.617994,1.131400,-0.306629,
        2.617994,1.137135,-0.936514,
        2.617994,1.139125,-1.924528,
        2.617994,1.139877,0.276382,
        2.617994,1.142003,-3.058022,
        2.617994,1.147363,-0.558448,
        2.617994,1.150348,-0.176418,
        2.617994,1.151198,1.197016,
        2.617994,1.153930,-1.597104,
        2.617994,1.155389,2.837148,
        2.617994,1.155458,-1.460623,
        2.617994,1.157331,-2.906362,
        2.617994,1.158594,2.487618,
        2.617994,1.162103,-1.322553,
        2.617994,1.165407,3.092861,
        2.617994,1.168025,0.717075,
        2.617994,1.168406,-2.301519,
        2.617994,1.174782,-2.538755,
        2.617994,1.177998,1.061256,
        2.617994,1.178898,1.827305,
        2.617994,1.181331,-2.060330,
        2.617994,1.182540,-0.798952,
        2.617994,1.184569,0.147517,
        2.617994,1.184734,1.968919,
        2.617994,1.184804,0.593386,
        2.617994,1.188278,1.690808,
        2.617994,1.194360,2.110440,
        2.617994,1.194776,2.710455,
        2.617994,1.197745,-1.173705,
        2.617994,1.199898,-2.777257,
        2.617994,1.201164,2.355931,
        2.617994,1.203760,1.556404,
        2.617994,1.205845,-0.415993,
        2.617994,1.206585,0.933648,
        2.617994,1.206950,-0.037248,
        2.617994,1.209341,0.471162,
        2.617994,1.213483,-1.039417,
        2.617994,1.218146,-0.672643,
        2.617994,1.221608,1.421669,
        2.617994,1.226043,-1.693533,
        2.617994,1.232870,-1.822773,
        2.617994,1.234375,-2.421194,
        2.617994,1.234531,-2.190148,
        2.617994,1.239414,0.348959,
        2.617994,1.240183,2.587255,
        2.617994,1.242312,2.941234,
        2.617994,1.244075,1.286644,
        2.617994,1.245961,-2.997461,
        2.617994,1.246495,-2.646872,
        2.617994,1.247736,2.242810,
        2.617994,1.253604,0.812222,
        2.617994,1.257816,-0.894317,
        2.617994,1.258247,-1.953634,
        2.617994,1.258993,-0.284099,
        2.617994,1.262380,-0.157609,
        2.617994,1.262791,-1.549229,
        2.617994,1.263772,-0.533073,
        2.617994,1.266731,-1.412461,
        2.617994,1.268591,-3.122146,
        2.617994,1.269614,-1.275246,
        2.617994,1.270745,1.152533,
        2.617994,1.273606,0.224545,
        2.617994,1.279716,2.812101,
        2.617994,1.283786,2.462124,
        2.617994,1.284798,0.078209,
        2.617994,1.285402,-2.882049,
        2.617994,1.286936,1.886053,
        2.617994,1.289408,0.683122,
        2.617994,1.294736,1.756184,
        2.617994,1.296978,-2.311511,
        2.617994,1.298546,2.020601,
        2.617994,1.300924,1.022478,
        2.617994,1.303305,3.043152,
        2.617994,1.304210,-2.084225,
        2.617994,1.305241,-2.530249,
        2.617994,1.307513,1.628003,
        2.617994,1.310722,0.558267,
        2.617994,1.314142,2.139329,
        2.617994,1.316087,-0.777971,
        2.617994,1.317907,-1.121023,
        2.617994,1.321019,-2.749776,
        2.617994,1.321609,2.687717,
        2.617994,1.322568,1.499415,
        2.617994,1.322972,-0.395699,
        2.617994,1.325286,-0.994192,
        2.617994,1.329146,-0.648513,
        2.617994,1.329795,-0.042570,
        2.617994,1.331651,2.344497,
        2.617994,1.336522,0.435098,
        2.617994,1.339410,0.900087,
        2.617994,1.339550,1.370353,
        2.617994,1.340441,-1.731580,
        2.617994,1.351855,-1.855234,
        2.617994,1.360161,-3.018404,
        2.617994,1.361031,-1.613971,
        2.617994,1.361196,2.914780,
        2.617994,1.361998,-2.207254,
        2.617994,1.362302,1.239384,
        2.617994,1.365228,2.563808,
        2.617994,1.365452,0.310522,
        2.617994,1.367028,-2.420041,
        2.617994,1.373422,-1.484281,
        2.617994,1.374245,-1.226417,
        2.617994,1.375020,-0.170393,
        2.617994,1.376560,-1.355481,
        2.617994,1.377070,-1.980984,
        2.617994,1.377578,-2.636339,
        2.617994,1.378096,0.774626,
        2.617994,1.379086,-0.508192,
        2.617994,1.382562,0.177141,
        2.617994,1.388849,2.233847,
        2.617994,1.389268,-0.886349,
        2.617994,1.391324,1.109396,
        2.617994,1.391699,-0.288725,
        2.617994,1.392869,3.136903,
        2.617994,1.393612,1.952016,
        2.617994,1.395545,-2.850807,
        2.617994,1.401926,1.822930,
        2.617994,1.402149,2.788898,
        2.617994,1.409552,0.647980,
        2.617994,1.410840,0.056219,
        2.617994,1.411944,2.444377,
        2.617994,1.413639,1.697982,
        2.617994,1.426479,1.573427,
        2.617994,1.426725,2.102242,
        2.617994,1.427868,0.984637,
        2.617994,1.429580,-2.107232,
        2.617994,1.431375,-2.313269,
        2.617994,1.431907,3.015410,
        2.617994,1.432325,-1.029054,
        2.617994,1.433341,0.524268,
        2.617994,1.435181,-0.739771,
        2.617994,1.437208,-2.529085,
        2.617994,1.439041,-0.617966,
        2.617994,1.441473,1.448544,
        2.617994,1.444527,2.664744,
        2.617994,1.450513,-0.400625,
        2.617994,1.451907,-0.066823,
        2.617994,1.452754,-2.738631,
        2.617994,1.454061,-2.950416,
        2.617994,1.456736,1.323154,
        2.617994,1.458370,0.400433,
        2.617994,1.459314,-1.765003,
        2.617994,1.459902,-1.145835,
        2.617994,1.464952,2.329245,
        2.617994,1.466050,0.862104,
        2.617994,1.471330,-1.886160,
        2.617994,1.475119,-1.648255,
        2.617994,1.480878,1.194576,
        2.617994,1.480942,0.273235,
        2.617994,1.481086,2.890269,
        2.617994,1.484118,-1.529156,
        2.617994,1.484741,-1.279219,
        2.617994,1.487681,-1.406069,
        2.617994,1.487886,-3.066037,
        2.617994,1.490555,2.543284,
        2.617994,1.492424,-0.190932,
        2.617994,1.495002,-0.839810,
        2.617994,1.495297,2.007188,
        2.617994,1.496414,-2.009267,
        2.617994,1.498261,-2.425270,
        2.617994,1.498907,-2.207327,
        2.617994,1.500204,0.737638,
        2.617994,1.502469,0.151156,
        2.617994,1.507448,-0.514782,
        2.617994,1.509287,2.204647,
        2.617994,1.510518,1.884920,
        2.617994,1.511948,-2.632360,
        2.617994,1.515235,1.068565,
        2.617994,1.522237,3.102018,
        2.617994,1.522373,1.764991,
        2.617994,1.522416,-0.954956,
        2.617994,1.523102,2.765920,
        2.617994,1.524811,-2.844718,
        2.617994,1.527261,-0.304916,
        2.617994,1.529788,0.613208,
        2.617994,1.533648,1.644808,
        2.617994,1.535018,0.028869,
        2.617994,1.541082,2.424608,
        2.617994,1.546010,1.523462,
        2.617994,1.551135,-0.642994,
        2.617994,1.553895,0.489607,
        2.617994,1.555041,0.947813,
        2.617994,1.559671,-1.074934,
        2.617994,1.560345,2.985899,
        2.617994,1.560487,-2.323400,
        2.617994,1.561127,1.401111,
        2.617994,1.567193,2.643172,
        2.617994,1.568892,2.098362,
        2.617994,1.571702,-2.098435,
        2.617994,1.572701,-2.528730,
        2.617994,1.573572,1.278593,
        2.617994,1.574849,-0.091505,
        2.617994,1.576334,0.365236,
        2.617994,1.577029,-2.962888,
        2.617994,1.577688,-1.814941,
        2.617994,1.579542,-0.755917,
        2.617994,1.579624,-1.199759,
        2.617994,1.581515,-0.419414,
        2.617994,1.584581,-2.738702,
        2.617994,1.586948,-1.696296,
        2.617994,1.588901,2.301496,
        2.617994,1.589936,0.825020,
        2.617994,1.590018,-1.941652,
        2.617994,1.596968,-1.328358,
        2.617994,1.597178,-1.575607,
        2.617994,1.600470,-1.453196,
        2.617994,1.601408,0.242409,
        2.617994,1.602259,1.152656,
        2.617994,1.603064,2.865774,
        2.617994,1.612085,1.968802,
        2.617994,1.612720,-3.080498,
        2.617994,1.616061,2.521531,
        2.617994,1.616900,-0.207457,
        2.617994,1.620988,-0.875920,
        2.617994,1.622114,0.701509,
        2.617994,1.622749,0.121426,
        2.617994,1.622993,-2.221629,
        2.617994,1.629346,-0.542669,
        2.617994,1.630007,1.844552,
        2.617994,1.634747,-2.426198,
        2.617994,1.642346,1.722617,
        2.617994,1.643925,1.034629,
        2.617994,1.644848,2.743774,
        2.617994,1.645258,2.188109,
        2.617994,1.645417,-2.634725,
        2.617994,1.647395,3.085437,
        2.617994,1.648538,-2.855044,
        2.617994,1.649932,0.576808,
        2.617994,1.652224,-0.996105,
        2.617994,1.652871,1.600681,
        2.617994,1.657997,-0.325127,
        2.617994,1.661936,0.001037,
        2.617994,1.665187,1.477402,
        2.617994,1.665496,2.397872,
        2.617994,1.673267,-0.665453,
        2.617994,1.673557,0.452526,
        2.617994,1.679277,-1.122421,
        2.617994,1.679696,0.912482,
        2.617994,1.682420,1.353922,
        2.617994,1.685031,-2.009038,
        2.617994,1.685196,2.967046,
        2.617994,1.690107,2.620910,
        2.617994,1.690453,-2.123076,
        2.617994,1.691693,-1.871572,
        2.617994,1.691817,2.061688,
        2.617994,1.692644,1.237670,
        2.617994,1.696386,0.328123,
        2.617994,1.697162,-1.249026,
        2.617994,1.697488,-2.322785,
        2.617994,1.698726,-2.973576,
        2.617994,1.699172,-1.749717,
        2.617994,1.704478,-0.106295,
        2.617994,1.707244,-0.444770,
        2.617994,1.707500,-2.531497,
        2.617994,1.707714,-0.787321,
        2.617994,1.708430,-1.627552,
        2.617994,1.710583,-1.376890,
        2.617994,1.712415,-2.747921,
        2.617994,1.713125,-1.503073,
        2.617994,1.713452,0.789419,
        2.617994,1.720612,2.279888,
        2.617994,1.724852,0.202831,
        2.617994,1.724868,2.846246,
        2.617994,1.728940,1.931535,
        2.617994,1.732838,1.127584,
        2.617994,1.735719,-3.092485,
        2.617994,1.739676,2.496017,
        2.617994,1.739971,-0.228876,
        2.617994,1.743328,0.663347,
        2.617994,1.744927,-0.911892,
        2.617994,1.745952,0.084733,
        2.617994,1.748077,1.804354,
        2.617994,1.752959,-0.567548,
        2.617994,1.760169,1.678132,
        2.617994,1.763887,-2.218450,
        2.617994,1.766737,2.722752,
        2.617994,1.767999,1.004737,
        2.617994,1.769149,1.550298,
        2.617994,1.770119,-2.427175,
        2.617994,1.770460,2.154208,
        2.617994,1.770514,3.070982,
        2.617994,1.770735,0.537245,
        2.617994,1.772673,-2.864760,
        2.617994,1.773527,-1.039582,
        2.617994,1.776419,-2.642421,
        2.617994,1.780769,1.416330,
        2.617994,1.786634,-0.346247,
        2.617994,1.788829,-1.947553,
        2.617994,1.793737,0.410270,
        2.617994,1.794553,2.374618,
        2.617994,1.794699,-0.694061,
        2.617994,1.797469,-1.169105,
        2.617994,1.803705,0.881076,
        2.617994,1.804124,1.264249,
        2.617994,1.807825,-2.090455,
        2.617994,1.808822,-1.817375,
        2.617994,1.809020,2.949741,
        2.617994,1.812466,2.597595,
        2.617994,1.813159,2.025395,
        2.617994,1.813816,-1.298435,
        2.617994,1.815016,0.278951,
        2.617994,1.816510,-0.008631,
        2.617994,1.817772,-1.689625,
        2.617994,1.821360,-2.984352,
        2.617994,1.822312,-0.127211,
        2.617994,1.823386,-1.429807,
        2.617994,1.823834,-1.560483,
        2.617994,1.832060,-0.821627,
        2.617994,1.832690,-0.468440,
        2.617994,1.833945,-2.320278,
        2.617994,1.834787,0.751968,
        2.617994,1.840579,-2.536445,
        2.617994,1.841033,-2.756932,
        2.617994,1.846350,1.893983,
        2.617994,1.846644,1.113766,
        2.617994,1.846833,2.826543,
        2.617994,1.847325,2.249051,
        2.617994,1.854119,0.134161,
        2.617994,1.858863,-3.106022,
        2.617994,1.865046,1.763384,
        2.617994,1.865619,-0.952036,
        2.617994,1.865912,0.624453,
        2.617994,1.866694,2.473686,
        2.617994,1.867524,-0.240083,
        2.617994,1.875561,1.632831,
        2.617994,1.876049,-0.594378,
        2.617994,1.880509,1.500204,
        2.617994,1.883504,1.360457,
        2.617994,1.886971,-2.196665,
        2.617994,1.888979,0.982922,
        2.617994,1.889815,2.701570,
        2.617994,1.892144,-2.005340,
        2.617994,1.892348,0.496522,
        2.617994,1.893281,-1.083976,
        2.617994,1.893828,3.053453,
        2.617994,1.895117,2.121107,
        2.617994,1.898741,-2.875735,
        2.617994,1.904429,-2.428029,
        2.617994,1.908408,-2.648903,
        2.617994,1.911828,-0.364489,
        2.617994,1.912224,0.366565,
        2.617994,1.914854,1.214468,
        2.617994,1.915467,-1.883143,
        2.617994,1.915763,-0.724679,
        2.617994,1.915790,-1.216366,
        2.617994,1.922256,2.347806,
        2.617994,1.922616,0.844633,
        2.617994,1.925185,0.231170,
        2.617994,1.927474,-1.754616,
        2.617994,1.928342,-1.351174,
        2.617994,1.933328,-1.623328,
        2.617994,1.933878,0.034150,
        2.617994,1.934585,-1.487979,
        2.617994,1.934802,2.927634,
        2.617994,1.934862,1.989776,
        2.617994,1.936327,-0.099350,
        2.617994,1.938088,2.578507,
        2.617994,1.945677,-2.999805,
        2.617994,1.954572,-0.859208,
        2.617994,1.957868,-0.490784,
        2.617994,1.958989,0.712772,
        2.617994,1.960212,-2.307561,
        2.617994,1.964774,1.855507,
        2.617994,1.964971,-2.100533,
        2.617994,1.969112,1.088273,
        2.617994,1.971438,-2.765415,
        2.617994,1.972312,2.805134,
        2.617994,1.974950,2.220596,
        2.617994,1.975135,-2.539127,
        2.617994,1.982107,1.720829,
        2.617994,1.983195,-3.127984,
        2.617994,1.983318,-0.247182,
        2.617994,1.986720,-0.993220,
        2.617994,1.990381,1.586005,
        2.617994,1.990470,0.580283,
        2.617994,1.990523,1.450326,
        2.617994,1.993909,1.312897,
        2.617994,1.994694,2.452813,
        2.617994,1.999226,-0.620717,
        2.617994,2.005145,0.941012,
        2.617994,2.007301,0.142337,
        2.617994,2.014149,0.444138,
        2.617994,2.014338,-1.127392,
        2.617994,2.018131,3.021778,
        2.617994,2.019676,-1.980944,
        2.617994,2.019912,2.686612,
        2.617994,2.020545,2.088672,
        2.617994,2.026074,0.300945,
        2.617994,2.026196,-2.889057,
        2.617994,2.031959,-1.268178,
        2.617994,2.032917,-2.199917,
        2.617994,2.033572,-1.842884,
        2.617994,2.034712,-2.419505,
        2.617994,2.035818,-0.379479,
        2.617994,2.036004,-0.757062,
        2.617994,2.040501,-1.411488,
        2.617994,2.042874,-1.701367,
        2.617994,2.043060,-2.653746,
        2.617994,2.043729,-1.557938,
        2.617994,2.048538,0.806623,
        2.617994,2.049006,1.199930,
        2.617994,2.049502,-0.137989,
        2.617994,2.052181,2.326959,
        2.617994,2.052794,-0.008526,
        2.617994,2.057608,1.951726,
        2.617994,2.065568,2.567730,
        2.617994,2.070371,2.879808,
        2.617994,2.071256,-3.018961,
        2.617994,2.079746,-0.897699,
        2.617994,2.082784,1.042403,
        2.617994,2.083752,-0.511086,
        2.617994,2.084401,1.810502,
        2.617994,2.085731,0.671081,
        2.617994,2.097496,1.373172,
        2.617994,2.099185,1.667006,
        2.617994,2.100482,-2.301084,
        2.617994,2.103118,2.194745,
        2.617994,2.103223,1.521685,
        2.617994,2.103695,-2.777707,
        2.617994,2.107523,3.129581,
        2.617994,2.108277,-2.095907,
        2.617994,2.108891,-2.532166,
        2.617994,2.111003,-0.268632,
        2.617994,2.112276,-1.027246,
        2.617994,2.114380,0.531457,
        2.617994,2.119166,0.100552,
        2.617994,2.119564,2.754818,
        2.617994,2.122474,-0.648314,
        2.617994,2.125183,2.445598,
        2.617994,2.129501,0.238290,
        2.617994,2.131178,0.386663,
        2.617994,2.131223,-1.174311,
        2.617994,2.134269,-1.948042,
        2.617994,2.134684,0.907065,
        2.617994,2.137615,2.988806,
        2.617994,2.145635,-1.327708,
        2.617994,2.146467,2.055081,
        2.617994,2.146582,-1.793788,
        2.617994,2.149210,-1.482321,
        2.617994,2.152901,-1.636788,
        2.617994,2.156337,1.151558,
        2.617994,2.156474,-2.912065,
        2.617994,2.157194,-0.792563,
        2.617994,2.166724,-0.401638,
        2.617994,2.166980,-0.134879,
        2.617994,2.173644,-2.403429,
        2.617994,2.175711,-2.655523,
        2.617994,2.177132,2.600419,
        2.617994,2.178879,0.767722,
        2.617994,2.180752,1.908485,
        2.617994,2.181338,2.313721,
        2.617994,2.185580,1.287673,
        2.617994,2.195400,-2.217546,
        2.617994,2.200418,-3.056442,
        2.617994,2.201829,-0.001735,
        2.617994,2.203804,1.756176,
        2.617994,2.211602,1.443930,
        2.617994,2.212686,-0.544809,
        2.617994,2.213937,0.621892,
        2.617994,2.214075,1.600691,
        2.617994,2.216256,1.020029,
        2.617994,2.219060,2.886631,
        2.617994,2.223375,-0.906185,
        2.617994,2.226002,-1.060248,
        2.617994,2.226413,-2.064034,
        2.617994,2.230554,2.171334,
        2.617994,2.232413,2.735777,
        2.617994,2.233709,3.070830,
        2.617994,2.234383,-2.792065,
        2.617994,2.237183,0.468799,
        2.617994,2.237695,-0.271388,
        2.617994,2.239419,0.151693,
        2.617994,2.243347,-2.521049,
        2.617994,2.244035,-1.229456,
        2.617994,2.244355,0.309899,
        2.617994,2.245175,2.460236,
        2.617994,2.245628,-1.899216,
        2.617994,2.245973,-0.702234,
        2.617994,2.255601,-1.728131,
        2.617994,2.256702,-1.401367,
        2.617994,2.261466,-1.557500,
        2.617994,2.267390,0.878499,
        2.617994,2.272099,2.018594,
        2.617994,2.278618,1.170386,
        2.617994,2.283304,-2.345584,
        2.617994,2.285175,-2.943090,
        2.617994,2.293000,-0.124451,
        2.617994,2.294058,-0.421070,
        2.617994,2.303283,1.855938,
        2.617994,2.304968,2.608978,
        2.617994,2.305347,2.311346,
        2.617994,2.307188,-2.656134,
        2.617994,2.307301,1.336200,
        2.617994,2.309655,0.726020,
        2.617994,2.317634,-2.185225,
        2.617994,2.321510,1.685664,
        2.617994,2.324376,-3.111006,
        2.617994,2.324731,1.511186,
        2.617994,2.329587,0.035498,
        2.617994,2.330158,2.977264,
        2.617994,2.335096,-0.588468,
        2.617994,2.336046,-1.115799,
        2.617994,2.338060,2.803613,
        2.617994,2.338083,-0.943400,
        2.617994,2.340210,0.562145,
        2.617994,2.341931,1.023926,
        2.617994,2.343265,-0.779061,
        2.617994,2.343434,-2.013692,
        2.617994,2.353029,-1.310529,
        2.617994,2.354868,0.211543,
        2.617994,2.355120,0.388593,
        2.617994,2.355165,-1.833173,
        2.617994,2.355462,2.151441,
        2.617994,2.362948,-2.487477,
        2.617994,2.363101,-2.810456,
        2.617994,2.363394,-0.272918,
        2.617994,2.364509,-1.644533,
        2.617994,2.370465,2.466824,
        2.617994,2.386110,-1.481351,
        2.617994,2.394960,0.862133,
        2.617994,2.395716,1.977184,
        2.617994,2.396339,1.200900,
        2.617994,2.407941,-2.316753,
        2.617994,2.409450,-2.988432,
        2.617994,2.417202,-0.103363,
        2.617994,2.417386,-0.443097,
        2.617994,2.417950,2.666662,
        2.617994,2.423190,1.789783,
        2.617994,2.423692,1.392696,
        2.617994,2.431122,3.086379,
        2.617994,2.431406,-2.647820,
        2.617994,2.431584,2.307631,
        2.617994,2.434090,1.592818,
        2.617994,2.435993,0.683844,
        2.617994,2.440898,-2.133300,
        2.617994,2.442222,-1.195837,
        2.617994,2.443112,2.883761,
        2.617994,2.449191,-0.633651,
        2.617994,2.450821,0.084741,
        2.617994,2.453499,-1.008861,
        2.617994,2.460010,-1.938011,
        2.617994,2.460199,-0.822008,
        2.617994,2.460902,0.490079,
        2.617994,2.461774,-1.734913,
        2.617994,2.463986,1.034407,
        2.617994,2.467647,0.287667,
        2.617994,2.479849,2.126925,
        2.617994,2.485907,-2.835471,
        2.617994,2.485945,-1.374853,
        2.617994,2.487747,2.505912,
        2.617994,2.487957,-0.274279,
        2.617994,2.489741,-2.468428,
        2.617994,2.512538,1.239900,
        2.617994,2.517528,1.923289,
        2.617994,2.518454,-1.565852,
        2.617994,2.519348,0.843503,
        2.617994,2.522766,-3.059231,
        2.617994,2.528686,2.737228,
        2.617994,2.534798,-2.266671,
        2.617994,2.536533,1.466274,
        2.617994,2.538504,1.699812,
        2.617994,2.539516,-0.476942,
        2.617994,2.539532,-0.074548,
        2.617994,2.543197,2.990252,
        2.617994,2.554022,2.320619,
        2.617994,2.556919,-2.655005,
        2.617994,2.558146,0.627191,
        2.617994,2.558223,-1.164189,
        2.617994,2.563452,-2.044028,
        2.617994,2.568462,0.150655,
        2.617994,2.569130,-0.700837,
        2.617994,2.570050,-1.804492,
        2.617994,2.572710,-0.934665,
        2.617994,2.575994,0.391718,
        2.617994,2.584774,1.046703,
        2.617994,2.601432,2.102535,
        2.617994,2.603559,2.555114,
        2.617994,2.604150,-2.885208,
        2.617994,2.608630,-1.378011,
        2.617994,2.611256,-0.279110,
        2.617994,2.617344,-2.439480,
        2.617994,2.627402,1.293552,
        2.617994,2.635396,1.848850,
        2.617994,2.635686,2.834950,
        2.617994,2.637087,3.132730,
        2.617994,2.642029,0.813274,
        2.617994,2.645260,1.570203,
        2.617994,2.646182,-1.614036,
        2.617994,2.660113,-2.186513,
        2.617994,2.660483,-0.034608,
        2.617994,2.661713,-0.529511,
        2.617994,2.672720,-1.110422,
        2.617994,2.676626,2.335353,
        2.617994,2.676638,0.540364,
        2.617994,2.676780,-1.901158,
        2.617994,2.681411,-2.667965,
        2.617994,2.682152,0.245751,
        2.617994,2.682859,-0.817071,
        2.617994,2.705107,1.062142,
        2.617994,2.717705,2.625995,
        2.617994,2.721001,2.060097,
        2.617994,2.722319,-2.954069,
        2.617994,2.729840,-1.368237,
        2.617994,2.734349,-0.289298,
        2.617994,2.738409,2.975671,
        2.617994,2.740366,1.374341,
        2.617994,2.744633,-2.391311,
        2.617994,2.747398,1.727166,
        2.617994,2.763799,0.760372,
        2.617994,2.768238,-1.679243,
        2.617994,2.779245,-2.046268,
        2.617994,2.780208,-0.626453,
        2.617994,2.780471,0.027732,
        2.617994,2.784696,-1.013128,
        2.617994,2.789817,0.395431,
        2.617994,2.801154,2.353143,
        2.617994,2.807091,-2.687477,
        2.617994,2.825762,1.085035,
        2.617994,2.830209,2.734943,
        2.617994,2.833517,-3.085508,
        2.617994,2.838301,1.980061,
        2.617994,2.849871,1.512012,
        2.617994,2.850089,-1.343899,
        2.617994,2.856905,-0.318441,
        2.617994,2.870445,-2.291001,
        2.617994,2.885017,0.655562,
        2.617994,2.886428,-1.793217,
        2.617994,2.890337,-0.825633,
        2.617994,2.898932,0.139973,
        2.617994,2.931927,-2.728768,
        2.617994,2.932096,2.377790,
        2.617994,2.937409,2.947245,
        2.617994,2.948048,1.133127,
        2.617994,2.950896,1.797889,
        2.617994,2.970019,-1.282382,
        2.617994,2.978333,-0.412090,
        2.617994,2.996437,-2.089121,
        2.617994,3.011274,0.434110,
        2.617994,3.050410,3.138492,
        2.617994,3.064492,1.647259,
        2.617994,3.088868,-0.989501,
        2.722714,0.053884,-2.408394,
        2.722714,0.077645,1.453433,
        2.722714,0.088408,-0.311160,
        2.722714,0.142314,2.590387,
        2.722714,0.147206,-1.385164,
        2.722714,0.163506,0.573203,
        2.722714,0.165485,-2.875396,
        2.722714,0.182564,-2.125821,
        2.722714,0.190569,1.929141,
        2.722714,0.201102,1.249082,
        2.722714,0.201397,-0.716646,
        2.722714,0.210568,-0.069559,
        2.722714,0.251672,2.848900,
        2.722714,0.266262,-1.678068,
        2.722714,0.269108,-1.215563,
        2.722714,0.273777,-2.506756,
        2.722714,0.279854,0.814581,
        2.722714,0.281664,-2.969147,
        2.722714,0.283625,2.405603,
        2.722714,0.284058,0.357350,
        2.722714,0.294107,1.593880,
        2.722714,0.302491,-0.400410,
        2.722714,0.308059,2.021491,
        2.722714,0.312219,-2.072057,
        2.722714,0.321905,-0.829121,
        2.722714,0.329984,1.201621,
        2.722714,0.339714,-0.013712,
        2.722714,0.368784,2.974974,
        2.722714,0.376428,-1.495532,
        2.722714,0.378103,2.646436,
        2.722714,0.385088,0.590355,
        2.722714,0.386248,-2.690943,
        2.722714,0.389851,-2.340031,
        2.722714,0.391502,-1.171374,
        2.722714,0.398249,-1.812694,
        2.722714,0.400542,-2.995729,
        2.722714,0.404273,0.918163,
        2.722714,0.405617,1.773533,
        2.722714,0.409009,-0.570243,
        2.722714,0.410700,0.270433,
        2.722714,0.415893,1.464890,
        2.722714,0.420680,2.088164,
        2.722714,0.424766,-0.258729,
        2.722714,0.426507,2.367872,
        2.722714,0.441292,-0.890727,
        2.722714,0.451839,-2.078193,
        2.722714,0.459270,1.191562,
        2.722714,0.470622,0.014002,
        2.722714,0.481292,-2.529098,
        2.722714,0.482460,2.849843,
        2.722714,0.492205,-1.397963,
        2.722714,0.493314,3.117264,
        2.722714,0.497870,0.727828,
        2.722714,0.499783,-1.654379,
        2.722714,0.502651,0.463333,
        2.722714,0.505463,2.594557,
        2.722714,0.506104,-2.874910,
        2.722714,0.513579,-1.147354,
        2.722714,0.518569,1.644237,
        2.722714,0.519001,1.900488,
        2.722714,0.519277,-0.673809,
        2.722714,0.524769,-0.421955,
        2.722714,0.525564,-2.288838,
        2.722714,0.532541,0.978737,
        2.722714,0.534610,2.154808,
        2.722714,0.535020,-1.891846,
        2.722714,0.540208,1.400497,
        2.722714,0.542006,0.228337,
        2.722714,0.550762,-0.180771,
        2.722714,0.560531,2.378474,
        2.722714,0.561629,-0.924576,
        2.722714,0.567465,-2.679392,
        2.722714,0.588527,1.181755,
        2.722714,0.591643,-2.098364,
        2.722714,0.594963,2.997575,
        2.722714,0.599643,-3.045235,
        2.722714,0.599962,2.773556,
        2.722714,0.602620,0.029270,
        2.722714,0.605711,0.598454,
        2.722714,0.607708,-1.541026,
        2.722714,0.609360,-2.464379,
        2.722714,0.610738,-1.326786,
        2.722714,0.616968,0.820489,
        2.722714,0.626379,1.773610,
        2.722714,0.628626,-1.751180,
        2.722714,0.630671,0.391949,
        2.722714,0.632622,-0.737642,
        2.722714,0.632708,2.566722,
        2.722714,0.632797,1.989858,
        2.722714,0.634455,-0.534790,
        2.722714,0.636205,1.561170,
        2.722714,0.637681,-1.123423,
        2.722714,0.641441,-2.842109,
        2.722714,0.644807,-0.327139,
        2.722714,0.652440,2.197129,
        2.722714,0.660897,-2.278593,
        2.722714,0.664008,1.360091,
        2.722714,0.670850,-1.945257,
        2.722714,0.671396,1.004859,
        2.722714,0.673985,0.203855,
        2.722714,0.680326,-0.133370,
        2.722714,0.685083,-0.942363,
        2.722714,0.688204,-2.632204,
        2.722714,0.689777,2.385011,
        2.722714,0.700065,3.115430,
        2.722714,0.704080,2.911636,
        2.722714,0.712883,1.174216,
        2.722714,0.715701,0.699906,
        2.722714,0.718851,-1.442215,
        2.722714,0.721019,2.721536,
        2.722714,0.725300,-2.975134,
        2.722714,0.727136,-1.633288,
        2.722714,0.727399,-2.120567,
        2.722714,0.730751,0.522488,
        2.722714,0.732892,-1.254330,
        2.722714,0.734333,-2.446184,
        2.722714,0.735125,0.037270,
        2.722714,0.737893,1.866463,
        2.722714,0.740650,1.681075,
        2.722714,0.745971,-0.781286,
        2.722714,0.749432,-0.442559,
        2.722714,0.750102,2.050636,
        2.722714,0.751046,0.863821,
        2.722714,0.757018,-0.608833,
        2.722714,0.757046,-1.816019,
        2.722714,0.757389,1.501184,
        2.722714,0.759800,2.546181,
        2.722714,0.759859,0.351923,
        2.722714,0.761141,-2.781102,
        2.722714,0.769035,-1.078574,
        2.722714,0.770166,-0.266604,
        2.722714,0.772218,2.227380,
        2.722714,0.788382,1.331051,
        2.722714,0.790269,-2.282409,
        2.722714,0.803410,-1.984567,
        2.722714,0.803813,0.191520,
        2.722714,0.806832,3.022155,
        2.722714,0.807476,-2.597675,
        2.722714,0.812512,-0.107335,
        2.722714,0.814161,1.010949,
        2.722714,0.817732,-3.084486,
        2.722714,0.817846,2.390688,
        2.722714,0.819323,2.845960,
        2.722714,0.819430,-0.915017,
        2.722714,0.829198,-1.532491,
        2.722714,0.829531,0.665695,
        2.722714,0.829816,1.167147,
        2.722714,0.833613,-1.357492,
        2.722714,0.842945,-2.907318,
        2.722714,0.844883,2.679708,
        2.722714,0.848189,-1.705003,
        2.722714,0.849888,1.767901,
        2.722714,0.851126,0.493370,
        2.722714,0.852911,1.933394,
        2.722714,0.857516,-2.432863,
        2.722714,0.858261,-1.191729,
        2.722714,0.858279,-2.138895,
        2.722714,0.859723,-0.734928,
        2.722714,0.859754,1.603292,
        2.722714,0.861655,-0.394821,
        2.722714,0.864300,0.043549,
        2.722714,0.866940,0.851229,
        2.722714,0.869748,2.095969,
        2.722714,0.877352,-0.574668,
        2.722714,0.879254,-2.733338,
        2.722714,0.882554,1.443757,
        2.722714,0.884060,0.334199,
        2.722714,0.885626,-1.866061,
        2.722714,0.889834,2.529957,
        2.722714,0.893140,2.251698,
        2.722714,0.897463,-0.240822,
        2.722714,0.898758,-1.031948,
        2.722714,0.913918,1.287618,
        2.722714,0.916555,-2.287173,
        2.722714,0.918305,3.107845,
        2.722714,0.919522,2.948277,
        2.722714,0.927138,-2.569686,
        2.722714,0.929031,0.185383,
        2.722714,0.933766,-3.015917,
        2.722714,0.934873,-2.010822,
        2.722714,0.935636,-1.445855,
        2.722714,0.936702,-0.866461,
        2.722714,0.937395,0.729108,
        2.722714,0.937910,2.787575,
        2.722714,0.939144,-0.092962,
        2.722714,0.941718,1.125085,
        2.722714,0.942873,-1.607680,
        2.722714,0.943701,0.980328,
        2.722714,0.947774,2.395256,
        2.722714,0.952433,0.584347,
        2.722714,0.954456,-1.287293,
        2.722714,0.958731,-2.852552,
        2.722714,0.961838,1.839284,
        2.722714,0.961898,-0.469425,
        2.722714,0.965426,1.686615,
        2.722714,0.970108,1.990771,
        2.722714,0.972820,-1.762780,
        2.722714,0.973500,-0.711881,
        2.722714,0.975253,2.641082,
        2.722714,0.977824,0.441928,
        2.722714,0.981485,-2.424766,
        2.722714,0.981877,1.537100,
        2.722714,0.982609,-1.136920,
        2.722714,0.984265,-2.156073,
        2.722714,0.989150,0.045493,
        2.722714,0.993052,2.137478,
        2.722714,0.997766,-2.693871,
        2.722714,0.999847,-0.334738,
        2.722714,1.001920,0.848054,
        2.722714,1.006146,1.389734,
        2.722714,1.012180,0.304066,
        2.722714,1.016938,2.270674,
        2.722714,1.017371,-0.979790,
        2.722714,1.019495,-1.894487,
        2.722714,1.023796,-0.583691,
        2.722714,1.026525,2.506945,
        2.722714,1.027004,-0.203653,
        2.722714,1.028027,3.036573,
        2.722714,1.031435,1.241021,
        2.722714,1.032316,2.884221,
        2.722714,1.035863,-3.110475,
        2.722714,1.042085,-2.292950,
        2.722714,1.043272,-1.522986,
        2.722714,1.046265,-2.958847,
        2.722714,1.049272,-1.376344,
        2.722714,1.050039,-2.552086,
        2.722714,1.053682,0.689863,
        2.722714,1.055458,0.170394,
        2.722714,1.056622,-0.831866,
        2.722714,1.057515,1.097998,
        2.722714,1.058706,-2.036438,
        2.722714,1.061389,-1.672248,
        2.722714,1.063758,2.743093,
        2.722714,1.070157,-0.073422,
        2.722714,1.072154,1.761506,
        2.722714,1.074029,1.909193,
        2.722714,1.074204,0.545276,
        2.722714,1.074714,-2.805147,
        2.722714,1.076262,0.966104,
        2.722714,1.077642,-1.227073,
        2.722714,1.084151,1.618525,
        2.722714,1.085085,2.377673,
        2.722714,1.086822,-0.441112,
        2.722714,1.087142,2.058815,
        2.722714,1.098383,-0.696430,
        2.722714,1.099138,-1.084033,
        2.722714,1.103208,1.477773,
        2.722714,1.103783,0.408254,
        2.722714,1.106275,-2.420800,
        2.722714,1.108079,2.612943,
        2.722714,1.109345,-2.172718,
        2.722714,1.115957,-1.787597,
        2.722714,1.118930,0.045983,
        2.722714,1.119777,0.834702,
        2.722714,1.121943,-2.669042,
        2.722714,1.124930,2.219917,
        2.722714,1.125217,1.336294,
        2.722714,1.130958,2.972931,
        2.722714,1.131400,-0.306629,
        2.722714,1.137135,-0.936514,
        2.722714,1.139125,-1.924528,
        2.722714,1.139877,0.276382,
        2.722714,1.142003,-3.058022,
        2.722714,1.147363,-0.558448,
        2.722714,1.150348,-0.176418,
        2.722714,1.151198,1.197016,
        2.722714,1.153930,-1.597104,
        2.722714,1.155389,2.837148,
        2.722714,1.155458,-1.460623,
        2.722714,1.157331,-2.906362,
        2.722714,1.158594,2.487618,
        2.722714,1.162103,-1.322553,
        2.722714,1.165407,3.092861,
        2.722714,1.168025,0.717075,
        2.722714,1.168406,-2.301519,
        2.722714,1.174782,-2.538755,
        2.722714,1.177998,1.061256,
        2.722714,1.178898,1.827305,
        2.722714,1.181331,-2.060330,
        2.722714,1.182540,-0.798952,
        2.722714,1.184569,0.147517,
        2.722714,1.184734,1.968919,
        2.722714,1.184804,0.593386,
        2.722714,1.188278,1.690808,
        2.722714,1.194360,2.110440,
        2.722714,1.194776,2.710455,
        2.722714,1.197745,-1.173705,
        2.722714,1.199898,-2.777257,
        2.722714,1.201164,2.355931,
        2.722714,1.203760,1.556404,
        2.722714,1.205845,-0.415993,
        2.722714,1.206585,0.933648,
        2.722714,1.206950,-0.037248,
        2.722714,1.209341,0.471162,
        2.722714,1.213483,-1.039417,
        2.722714,1.218146,-0.672643,
        2.722714,1.221608,1.421669,
        2.722714,1.226043,-1.693533,
        2.722714,1.232870,-1.822773,
        2.722714,1.234375,-2.421194,
        2.722714,1.234531,-2.190148,
        2.722714,1.239414,0.348959,
        2.722714,1.240183,2.587255,
        2.722714,1.242312,2.941234,
        2.722714,1.244075,1.286644,
        2.722714,1.245961,-2.997461,
        2.722714,1.246495,-2.646872,
        2.722714,1.247736,2.242810,
        2.722714,1.253604,0.812222,
        2.722714,1.257816,-0.894317,
        2.722714,1.258247,-1.953634,
        2.722714,1.258993,-0.284099,
        2.722714,1.262380,-0.157609,
        2.722714,1.262791,-1.549229,
        2.722714,1.263772,-0.533073,
        2.722714,1.266731,-1.412461,
        2.722714,1.268591,-3.122146,
        2.722714,1.269614,-1.275246,
        2.722714,1.270745,1.152533,
        2.722714,1.273606,0.224545,
        2.722714,1.279716,2.812101,
        2.722714,1.283786,2.462124,
        2.722714,1.284798,0.078209,
        2.722714,1.285402,-2.882049,
        2.722714,1.286936,1.886053,
        2.722714,1.289408,0.683122,
        2.722714,1.294736,1.756184,
        2.722714,1.296978,-2.311511,
        2.722714,1.298546,2.020601,
        2.722714,1.300924,1.022478,
        2.722714,1.303305,3.043152,
        2.722714,1.304210,-2.084225,
        2.722714,1.305241,-2.530249,
        2.722714,1.307513,1.628003,
        2.722714,1.310722,0.558267,
        2.722714,1.314142,2.139329,
        2.722714,1.316087,-0.777971,
        2.722714,1.317907,-1.121023,
        2.722714,1.321019,-2.749776,
        2.722714,1.321609,2.687717,
        2.722714,1.322568,1.499415,
        2.722714,1.322972,-0.395699,
        2.722714,1.325286,-0.994192,
        2.722714,1.329146,-0.648513,
        2.722714,1.329795,-0.042570,
        2.722714,1.331651,2.344497,
        2.722714,1.336522,0.435098,
        2.722714,1.339410,0.900087,
        2.722714,1.339550,1.370353,
        2.722714,1.340441,-1.731580,
        2.722714,1.351855,-1.855234,
        2.722714,1.360161,-3.018404,
        2.722714,1.361031,-1.613971,
        2.722714,1.361196,2.914780,
        2.722714,1.361998,-2.207254,
        2.722714,1.362302,1.239384,
        2.722714,1.365228,2.563808,
        2.722714,1.365452,0.310522,
        2.722714,1.367028,-2.420041,
        2.722714,1.373422,-1.484281,
        2.722714,1.374245,-1.226417,
        2.722714,1.375020,-0.170393,
        2.722714,1.376560,-1.355481,
        2.722714,1.377070,-1.980984,
        2.722714,1.377578,-2.636339,
        2.722714,1.378096,0.774626,
        2.722714,1.379086,-0.508192,
        2.722714,1.382562,0.177141,
        2.722714,1.388849,2.233847,
        2.722714,1.389268,-0.886349,
        2.722714,1.391324,1.109396,
        2.722714,1.391699,-0.288725,
        2.722714,1.392869,3.136903,
        2.722714,1.393612,1.952016,
        2.722714,1.395545,-2.850807,
        2.722714,1.401926,1.822930,
        2.722714,1.402149,2.788898,
        2.722714,1.409552,0.647980,
        2.722714,1.410840,0.056219,
        2.722714,1.411944,2.444377,
        2.722714,1.413639,1.697982,
        2.722714,1.426479,1.573427,
        2.722714,1.426725,2.102242,
        2.722714,1.427868,0.984637,
        2.722714,1.429580,-2.107232,
        2.722714,1.431375,-2.313269,
        2.722714,1.431907,3.015410,
        2.722714,1.432325,-1.029054,
        2.722714,1.433341,0.524268,
        2.722714,1.435181,-0.739771,
        2.722714,1.437208,-2.529085,
        2.722714,1.439041,-0.617966,
        2.722714,1.441473,1.448544,
        2.722714,1.444527,2.664744,
        2.722714,1.450513,-0.400625,
        2.722714,1.451907,-0.066823,
        2.722714,1.452754,-2.738631,
        2.722714,1.454061,-2.950416,
        2.722714,1.456736,1.323154,
        2.722714,1.458370,0.400433,
        2.722714,1.459314,-1.765003,
        2.722714,1.459902,-1.145835,
        2.722714,1.464952,2.329245,
        2.722714,1.466050,0.862104,
        2.722714,1.471330,-1.886160,
        2.722714,1.475119,-1.648255,
        2.722714,1.480878,1.194576,
        2.722714,1.480942,0.273235,
        2.722714,1.481086,2.890269,
        2.722714,1.484118,-1.529156,
        2.722714,1.484741,-1.279219,
        2.722714,1.487681,-1.406069,
        2.722714,1.487886,-3.066037,
        2.722714,1.490555,2.543284,
        2.722714,1.492424,-0.190932,
        2.722714,1.495002,-0.839810,
        2.722714,1.495297,2.007188,
        2.722714,1.496414,-2.009267,
        2.722714,1.498261,-2.425270,
        2.722714,1.498907,-2.207327,
        2.722714,1.500204,0.737638,
        2.722714,1.502469,0.151156,
        2.722714,1.507448,-0.514782,
        2.722714,1.509287,2.204647,
        2.722714,1.510518,1.884920,
        2.722714,1.511948,-2.632360,
        2.722714,1.515235,1.068565,
        2.722714,1.522237,3.102018,
        2.722714,1.522373,1.764991,
        2.722714,1.522416,-0.954956,
        2.722714,1.523102,2.765920,
        2.722714,1.524811,-2.844718,
        2.722714,1.527261,-0.304916,
        2.722714,1.529788,0.613208,
        2.722714,1.533648,1.644808,
        2.722714,1.535018,0.028869,
        2.722714,1.541082,2.424608,
        2.722714,1.546010,1.523462,
        2.722714,1.551135,-0.642994,
        2.722714,1.553895,0.489607,
        2.722714,1.555041,0.947813,
        2.722714,1.559671,-1.074934,
        2.722714,1.560345,2.985899,
        2.722714,1.560487,-2.323400,
        2.722714,1.561127,1.401111,
        2.722714,1.567193,2.643172,
        2.722714,1.568892,2.098362,
        2.722714,1.571702,-2.098435,
        2.722714,1.572701,-2.528730,
        2.722714,1.573572,1.278593,
        2.722714,1.574849,-0.091505,
        2.722714,1.576334,0.365236,
        2.722714,1.577029,-2.962888,
        2.722714,1.577688,-1.814941,
        2.722714,1.579542,-0.755917,
        2.722714,1.579624,-1.199759,
        2.722714,1.581515,-0.419414,
        2.722714,1.584581,-2.738702,
        2.722714,1.586948,-1.696296,
        2.722714,1.588901,2.301496,
        2.722714,1.589936,0.825020,
        2.722714,1.590018,-1.941652,
        2.722714,1.596968,-1.328358,
        2.722714,1.597178,-1.575607,
        2.722714,1.600470,-1.453196,
        2.722714,1.601408,0.242409,
        2.722714,1.602259,1.152656,
        2.722714,1.603064,2.865774,
        2.722714,1.612085,1.968802,
        2.722714,1.612720,-3.080498,
        2.722714,1.616061,2.521531,
        2.722714,1.616900,-0.207457,
        2.722714,1.620988,-0.875920,
        2.722714,1.622114,0.701509,
        2.722714,1.622749,0.121426,
        2.722714,1.622993,-2.221629,
        2.722714,1.629346,-0.542669,
        2.722714,1.630007,1.844552,
        2.722714,1.634747,-2.426198,
        2.722714,1.642346,1.722617,
        2.722714,1.643925,1.034629,
        2.722714,1.644848,2.743774,
        2.722714,1.645258,2.188109,
        2.722714,1.645417,-2.634725,
        2.722714,1.647395,3.085437,
        2.722714,1.648538,-2.855044,
        2.722714,1.649932,0.576808,
        2.722714,1.652224,-0.996105,
        2.722714,1.652871,1.600681,
        2.722714,1.657997,-0.325127,
        2.722714,1.661936,0.001037,
        2.722714,1.665187,1.477402,
        2.722714,1.665496,2.397872,
        2.722714,1.673267,-0.665453,
        2.722714,1.673557,0.452526,
        2.722714,1.679277,-1.122421,
        2.722714,1.679696,0.912482,
        2.722714,1.682420,1.353922,
        2.722714,1.685031,-2.009038,
        2.722714,1.685196,2.967046,
        2.722714,1.690107,2.620910,
        2.722714,1.690453,-2.123076,
        2.722714,1.691693,-1.871572,
        2.722714,1.691817,2.061688,
        2.722714,1.692644,1.237670,
        2.722714,1.696386,0.328123,
        2.722714,1.697162,-1.249026,
        2.722714,1.697488,-2.322785,
        2.722714,1.698726,-2.973576,
        2.722714,1.699172,-1.749717,
        2.722714,1.704478,-0.106295,
        2.722714,1.707244,-0.444770,
        2.722714,1.707500,-2.531497,
        2.722714,1.707714,-0.787321,
        2.722714,1.708430,-1.627552,
        2.722714,1.710583,-1.376890,
        2.722714,1.712415,-2.747921,
        2.722714,1.713125,-1.503073,
        2.722714,1.713452,0.789419,
        2.722714,1.720612,2.279888,
        2.722714,1.724852,0.202831,
        2.722714,1.724868,2.846246,
        2.722714,1.728940,1.931535,
        2.722714,1.732838,1.127584,
        2.722714,1.735719,-3.092485,
        2.722714,1.739676,2.496017,
        2.722714,1.739971,-0.228876,
        2.722714,1.743328,0.663347,
        2.722714,1.744927,-0.911892,
        2.722714,1.745952,0.084733,
        2.722714,1.748077,1.804354,
        2.722714,1.752959,-0.567548,
        2.722714,1.760169,1.678132,
        2.722714,1.763887,-2.218450,
        2.722714,1.766737,2.722752,
        2.722714,1.767999,1.004737,
        2.722714,1.769149,1.550298,
        2.722714,1.770119,-2.427175,
        2.722714,1.770460,2.154208,
        2.722714,1.770514,3.070982,
        2.722714,1.770735,0.537245,
        2.722714,1.772673,-2.864760,
        2.722714,1.773527,-1.039582,
        2.722714,1.776419,-2.642421,
        2.722714,1.780769,1.416330,
        2.722714,1.786634,-0.346247,
        2.722714,1.788829,-1.947553,
        2.722714,1.793737,0.410270,
        2.722714,1.794553,2.374618,
        2.722714,1.794699,-0.694061,
        2.722714,1.797469,-1.169105,
        2.722714,1.803705,0.881076,
        2.722714,1.804124,1.264249,
        2.722714,1.807825,-2.090455,
        2.722714,1.808822,-1.817375,
        2.722714,1.809020,2.949741,
        2.722714,1.812466,2.597595,
        2.722714,1.813159,2.025395,
        2.722714,1.813816,-1.298435,
        2.722714,1.815016,0.278951,
        2.722714,1.816510,-0.008631,
        2.722714,1.817772,-1.689625,
        2.722714,1.821360,-2.984352,
        2.722714,1.822312,-0.127211,
        2.722714,1.823386,-1.429807,
        2.722714,1.823834,-1.560483,
        2.722714,1.832060,-0.821627,
        2.722714,1.832690,-0.468440,
        2.722714,1.833945,-2.320278,
        2.722714,1.834787,0.751968,
        2.722714,1.840579,-2.536445,
        2.722714,1.841033,-2.756932,
        2.722714,1.846350,1.893983,
        2.722714,1.846644,1.113766,
        2.722714,1.846833,2.826543,
        2.722714,1.847325,2.249051,
        2.722714,1.854119,0.134161,
        2.722714,1.858863,-3.106022,
        2.722714,1.865046,1.763384,
        2.722714,1.865619,-0.952036,
        2.722714,1.865912,0.624453,
        2.722714,1.866694,2.473686,
        2.722714,1.867524,-0.240083,
        2.722714,1.875561,1.632831,
        2.722714,1.876049,-0.594378,
        2.722714,1.880509,1.500204,
        2.722714,1.883504,1.360457,
        2.722714,1.886971,-2.196665,
        2.722714,1.888979,0.982922,
        2.722714,1.889815,2.701570,
        2.722714,1.892144,-2.005340,
        2.722714,1.892348,0.496522,
        2.722714,1.893281,-1.083976,
        2.722714,1.893828,3.053453,
        2.722714,1.895117,2.121107,
        2.722714,1.898741,-2.875735,
        2.722714,1.904429,-2.428029,
        2.722714,1.908408,-2.648903,
        2.722714,1.911828,-0.364489,
        2.722714,1.912224,0.366565,
        2.722714,1.914854,1.214468,
        2.722714,1.915467,-1.883143,
        2.722714,1.915763,-0.724679,
        2.722714,1.915790,-1.216366,
        2.722714,1.922256,2.347806,
        2.722714,1.922616,0.844633,
        2.722714,1.925185,0.231170,
        2.722714,1.927474,-1.754616,
        2.722714,1.928342,-1.351174,
        2.722714,1.933328,-1.623328,
        2.722714,1.933878,0.034150,
        2.722714,1.934585,-1.487979,
        2.722714,1.934802,2.927634,
        2.722714,1.934862,1.989776,
        2.722714,1.936327,-0.099350,
        2.722714,1.938088,2.578507,
        2.722714,1.945677,-2.999805,
        2.722714,1.954572,-0.859208,
        2.722714,1.957868,-0.490784,
        2.722714,1.958989,0.712772,
        2.722714,1.960212,-2.307561,
        2.722714,1.964774,1.855507,
        2.722714,1.964971,-2.100533,
        2.722714,1.969112,1.088273,
        2.722714,1.971438,-2.765415,
        2.722714,1.972312,2.805134,
        2.722714,1.974950,2.220596,
        2.722714,1.975135,-2.539127,
        2.722714,1.982107,1.720829,
        2.722714,1.983195,-3.127984,
        2.722714,1.983318,-0.247182,
        2.722714,1.986720,-0.993220,
        2.722714,1.990381,1.586005,
        2.722714,1.990470,0.580283,
        2.722714,1.990523,1.450326,
        2.722714,1.993909,1.312897,
        2.722714,1.994694,2.452813,
        2.722714,1.999226,-0.620717,
        2.722714,2.005145,0.941012,
        2.722714,2.007301,0.142337,
        2.722714,2.014149,0.444138,
        2.722714,2.014338,-1.127392,
        2.722714,2.018131,3.021778,
        2.722714,2.019676,-1.980944,
        2.722714,2.019912,2.686612,
        2.722714,2.020545,2.088672,
        2.722714,2.026074,0.300945,
        2.722714,2.026196,-2.889057,
        2.722714,2.031959,-1.268178,
        2.722714,2.032917,-2.199917,
        2.722714,2.033572,-1.842884,
        2.722714,2.034712,-2.419505,
        2.722714,2.035818,-0.379479,
        2.722714,2.036004,-0.757062,
        2.722714,2.040501,-1.411488,
        2.722714,2.042874,-1.701367,
        2.722714,2.043060,-2.653746,
        2.722714,2.043729,-1.557938,
        2.722714,2.048538,0.806623,
        2.722714,2.049006,1.199930,
        2.722714,2.049502,-0.137989,
        2.722714,2.052181,2.326959,
        2.722714,2.052794,-0.008526,
        2.722714,2.057608,1.951726,
        2.722714,2.065568,2.567730,
        2.722714,2.070371,2.879808,
        2.722714,2.071256,-3.018961,
        2.722714,2.079746,-0.897699,
        2.722714,2.082784,1.042403,
        2.722714,2.083752,-0.511086,
        2.722714,2.084401,1.810502,
        2.722714,2.085731,0.671081,
        2.722714,2.097496,1.373172,
        2.722714,2.099185,1.667006,
        2.722714,2.100482,-2.301084,
        2.722714,2.103118,2.194745,
        2.722714,2.103223,1.521685,
        2.722714,2.103695,-2.777707,
        2.722714,2.107523,3.129581,
        2.722714,2.108277,-2.095907,
        2.722714,2.108891,-2.532166,
        2.722714,2.111003,-0.268632,
        2.722714,2.112276,-1.027246,
        2.722714,2.114380,0.531457,
        2.722714,2.119166,0.100552,
        2.722714,2.119564,2.754818,
        2.722714,2.122474,-0.648314,
        2.722714,2.125183,2.445598,
        2.722714,2.129501,0.238290,
        2.722714,2.131178,0.386663,
        2.722714,2.131223,-1.174311,
        2.722714,2.134269,-1.948042,
        2.722714,2.134684,0.907065,
        2.722714,2.137615,2.988806,
        2.722714,2.145635,-1.327708,
        2.722714,2.146467,2.055081,
        2.722714,2.146582,-1.793788,
        2.722714,2.149210,-1.482321,
        2.722714,2.152901,-1.636788,
        2.722714,2.156337,1.151558,
        2.722714,2.156474,-2.912065,
        2.722714,2.157194,-0.792563,
        2.722714,2.166724,-0.401638,
        2.722714,2.166980,-0.134879,
        2.722714,2.173644,-2.403429,
        2.722714,2.175711,-2.655523,
        2.722714,2.177132,2.600419,
        2.722714,2.178879,0.767722,
        2.722714,2.180752,1.908485,
        2.722714,2.181338,2.313721,
        2.722714,2.185580,1.287673,
        2.722714,2.195400,-2.217546,
        2.722714,2.200418,-3.056442,
        2.722714,2.201829,-0.001735,
        2.722714,2.203804,1.756176,
        2.722714,2.211602,1.443930,
        2.722714,2.212686,-0.544809,
        2.722714,2.213937,0.621892,
        2.722714,2.214075,1.600691,
        2.722714,2.216256,1.020029,
        2.722714,2.219060,2.886631,
        2.722714,2.223375,-0.906185,
        2.722714,2.226002,-1.060248,
        2.722714,2.226413,-2.064034,
        2.722714,2.230554,2.171334,
        2.722714,2.232413,2.735777,
        2.722714,2.233709,3.070830,
        2.722714,2.234383,-2.792065,
        2.722714,2.237183,0.468799,
        2.722714,2.237695,-0.271388,
        2.722714,2.239419,0.151693,
        2.722714,2.243347,-2.521049,
        2.722714,2.244035,-1.229456,
        2.722714,2.244355,0.309899,
        2.722714,2.245175,2.460236,
        2.722714,2.245628,-1.899216,
        2.722714,2.245973,-0.702234,
        2.722714,2.255601,-1.728131,
        2.722714,2.256702,-1.401367,
        2.722714,2.261466,-1.557500,
        2.722714,2.267390,0.878499,
        2.722714,2.272099,2.018594,
        2.722714,2.278618,1.170386,
        2.722714,2.283304,-2.345584,
        2.722714,2.285175,-2.943090,
        2.722714,2.293000,-0.124451,
        2.722714,2.294058,-0.421070,
        2.722714,2.303283,1.855938,
        2.722714,2.304968,2.608978,
        2.722714,2.305347,2.311346,
        2.722714,2.307188,-2.656134,
        2.722714,2.307301,1.336200,
        2.722714,2.309655,0.726020,
        2.722714,2.317634,-2.185225,
        2.722714,2.321510,1.685664,
        2.722714,2.324376,-3.111006,
        2.722714,2.324731,1.511186,
        2.722714,2.329587,0.035498,
        2.722714,2.330158,2.977264,
        2.722714,2.335096,-0.588468,
        2.722714,2.336046,-1.115799,
        2.722714,2.338060,2.803613,
        2.722714,2.338083,-0.943400,
        2.722714,2.340210,0.562145,
        2.722714,2.341931,1.023926,
        2.722714,2.343265,-0.779061,
        2.722714,2.343434,-2.013692,
        2.722714,2.353029,-1.310529,
        2.722714,2.354868,0.211543,
        2.722714,2.355120,0.388593,
        2.722714,2.355165,-1.833173,
        2.722714,2.355462,2.151441,
        2.722714,2.362948,-2.487477,
        2.722714,2.363101,-2.810456,
        2.722714,2.363394,-0.272918,
        2.722714,2.364509,-1.644533,
        2.722714,2.370465,2.466824,
        2.722714,2.386110,-1.481351,
        2.722714,2.394960,0.862133,
        2.722714,2.395716,1.977184,
        2.722714,2.396339,1.200900,
        2.722714,2.407941,-2.316753,
        2.722714,2.409450,-2.988432,
        2.722714,2.417202,-0.103363,
        2.722714,2.417386,-0.443097,
        2.722714,2.417950,2.666662,
        2.722714,2.423190,1.789783,
        2.722714,2.423692,1.392696,
        2.722714,2.431122,3.086379,
        2.722714,2.431406,-2.647820,
        2.722714,2.431584,2.307631,
        2.722714,2.434090,1.592818,
        2.722714,2.435993,0.683844,
        2.722714,2.440898,-2.133300,
        2.722714,2.442222,-1.195837,
        2.722714,2.443112,2.883761,
        2.722714,2.449191,-0.633651,
        2.722714,2.450821,0.084741,
        2.722714,2.453499,-1.008861,
        2.722714,2.460010,-1.938011,
        2.722714,2.460199,-0.822008,
        2.722714,2.460902,0.490079,
        2.722714,2.461774,-1.734913,
        2.722714,2.463986,1.034407,
        2.722714,2.467647,0.287667,
        2.722714,2.479849,2.126925,
        2.722714,2.485907,-2.835471,
        2.722714,2.485945,-1.374853,
        2.722714,2.487747,2.505912,
        2.722714,2.487957,-0.274279,
        2.722714,2.489741,-2.468428,
        2.722714,2.512538,1.239900,
        2.722714,2.517528,1.923289,
        2.722714,2.518454,-1.565852,
        2.722714,2.519348,0.843503,
        2.722714,2.522766,-3.059231,
        2.722714,2.528686,2.737228,
        2.722714,2.534798,-2.266671,
        2.722714,2.536533,1.466274,
        2.722714,2.538504,1.699812,
        2.722714,2.539516,-0.476942,
        2.722714,2.539532,-0.074548,
        2.722714,2.543197,2.990252,
        2.722714,2.554022,2.320619,
        2.722714,2.556919,-2.655005,
        2.722714,2.558146,0.627191,
        2.722714,2.558223,-1.164189,
        2.722714,2.563452,-2.044028,
        2.722714,2.568462,0.150655,
        2.722714,2.569130,-0.700837,
        2.722714,2.570050,-1.804492,
        2.722714,2.572710,-0.934665,
        2.722714,2.575994,0.391718,
        2.722714,2.584774,1.046703,
        2.722714,2.601432,2.102535,
        2.722714,2.603559,2.555114,
        2.722714,2.604150,-2.885208,
        2.722714,2.608630,-1.378011,
        2.722714,2.611256,-0.279110,
        2.722714,2.617344,-2.439480,
        2.722714,2.627402,1.293552,
        2.722714,2.635396,1.848850,
        2.722714,2.635686,2.834950,
        2.722714,2.637087,3.132730,
        2.722714,2.642029,0.813274,
        2.722714,2.645260,1.570203,
        2.722714,2.646182,-1.614036,
        2.722714,2.660113,-2.186513,
        2.722714,2.660483,-0.034608,
        2.722714,2.661713,-0.529511,
        2.722714,2.672720,-1.110422,
        2.722714,2.676626,2.335353,
        2.722714,2.676638,0.540364,
        2.722714,2.676780,-1.901158,
        2.722714,2.681411,-2.667965,
        2.722714,2.682152,0.245751,
        2.722714,2.682859,-0.817071,
        2.722714,2.705107,1.062142,
        2.722714,2.717705,2.625995,
        2.722714,2.721001,2.060097,
        2.722714,2.722319,-2.954069,
        2.722714,2.729840,-1.368237,
        2.722714,2.734349,-0.289298,
        2.722714,2.738409,2.975671,
        2.722714,2.740366,1.374341,
        2.722714,2.744633,-2.391311,
        2.722714,2.747398,1.727166,
        2.722714,2.763799,0.760372,
        2.722714,2.768238,-1.679243,
        2.722714,2.779245,-2.046268,
        2.722714,2.780208,-0.626453,
        2.722714,2.780471,0.027732,
        2.722714,2.784696,-1.013128,
        2.722714,2.789817,0.395431,
        2.722714,2.801154,2.353143,
        2.722714,2.807091,-2.687477,
        2.722714,2.825762,1.085035,
        2.722714,2.830209,2.734943,
        2.722714,2.833517,-3.085508,
        2.722714,2.838301,1.980061,
        2.722714,2.849871,1.512012,
        2.722714,2.850089,-1.343899,
        2.722714,2.856905,-0.318441,
        2.722714,2.870445,-2.291001,
        2.722714,2.885017,0.655562,
        2.722714,2.886428,-1.793217,
        2.722714,2.890337,-0.825633,
        2.722714,2.898932,0.139973,
        2.722714,2.931927,-2.728768,
        2.722714,2.932096,2.377790,
        2.722714,2.937409,2.947245,
        2.722714,2.948048,1.133127,
        2.722714,2.950896,1.797889,
        2.722714,2.970019,-1.282382,
        2.722714,2.978333,-0.412090,
        2.722714,2.996437,-2.089121,
        2.722714,3.011274,0.434110,
        2.722714,3.050410,3.138492,
        2.722714,3.064492,1.647259,
        2.722714,3.088868,-0.989501,
        2.827433,0.053884,-2.408394,
        2.827433,0.077645,1.453433,
        2.827433,0.088408,-0.311160,
        2.827433,0.142314,2.590387,
        2.827433,0.147206,-1.385164,
        2.827433,0.163506,0.573203,
        2.827433,0.165485,-2.875396,
        2.827433,0.182564,-2.125821,
        2.827433,0.190569,1.929141,
        2.827433,0.201102,1.249082,
        2.827433,0.201397,-0.716646,
        2.827433,0.210568,-0.069559,
        2.827433,0.251672,2.848900,
        2.827433,0.266262,-1.678068,
        2.827433,0.269108,-1.215563,
        2.827433,0.273777,-2.506756,
        2.827433,0.279854,0.814581,
        2.827433,0.281664,-2.969147,
        2.827433,0.283625,2.405603,
        2.827433,0.284058,0.357350,
        2.827433,0.294107,1.593880,
        2.827433,0.302491,-0.400410,
        2.827433,0.308059,2.021491,
        2.827433,0.312219,-2.072057,
        2.827433,0.321905,-0.829121,
        2.827433,0.329984,1.201621,
        2.827433,0.339714,-0.013712,
        2.827433,0.368784,2.974974,
        2.827433,0.376428,-1.495532,
        2.827433,0.378103,2.646436,
        2.827433,0.385088,0.590355,
        2.827433,0.386248,-2.690943,
        2.827433,0.389851,-2.340031,
        2.827433,0.391502,-1.171374,
        2.827433,0.398249,-1.812694,
        2.827433,0.400542,-2.995729,
        2.827433,0.404273,0.918163,
        2.827433,0.405617,1.773533,
        2.827433,0.409009,-0.570243,
        2.827433,0.410700,0.270433,
        2.827433,0.415893,1.464890,
        2.827433,0.420680,2.088164,
        2.827433,0.424766,-0.258729,
        2.827433,0.426507,2.367872,
        2.827433,0.441292,-0.890727,
        2.827433,0.451839,-2.078193,
        2.827433,0.459270,1.191562,
        2.827433,0.470622,0.014002,
        2.827433,0.481292,-2.529098,
        2.827433,0.482460,2.849843,
        2.827433,0.492205,-1.397963,
        2.827433,0.493314,3.117264,
        2.827433,0.497870,0.727828,
        2.827433,0.499783,-1.654379,
        2.827433,0.502651,0.463333,
        2.827433,0.505463,2.594557,
        2.827433,0.506104,-2.874910,
        2.827433,0.513579,-1.147354,
        2.827433,0.518569,1.644237,
        2.827433,0.519001,1.900488,
        2.827433,0.519277,-0.673809,
        2.827433,0.524769,-0.421955,
        2.827433,0.525564,-2.288838,
        2.827433,0.532541,0.978737,
        2.827433,0.534610,2.154808,
        2.827433,0.535020,-1.891846,
        2.827433,0.540208,1.400497,
        2.827433,0.542006,0.228337,
        2.827433,0.550762,-0.180771,
        2.827433,0.560531,2.378474,
        2.827433,0.561629,-0.924576,
        2.827433,0.567465,-2.679392,
        2.827433,0.588527,1.181755,
        2.827433,0.591643,-2.098364,
        2.827433,0.594963,2.997575,
        2.827433,0.599643,-3.045235,
        2.827433,0.599962,2.773556,
        2.827433,0.602620,0.029270,
        2.827433,0.605711,0.598454,
        2.827433,0.607708,-1.541026,
        2.827433,0.609360,-2.464379,
        2.827433,0.610738,-1.326786,
        2.827433,0.616968,0.820489,
        2.827433,0.626379,1.773610,
        2.827433,0.628626,-1.751180,
        2.827433,0.630671,0.391949,
        2.827433,0.632622,-0.737642,
        2.827433,0.632708,2.566722,
        2.827433,0.632797,1.989858,
        2.827433,0.634455,-0.534790,
        2.827433,0.636205,1.561170,
        2.827433,0.637681,-1.123423,
        2.827433,0.641441,-2.842109,
        2.827433,0.644807,-0.327139,
        2.827433,0.652440,2.197129,
        2.827433,0.660897,-2.278593,
        2.827433,0.664008,1.360091,
        2.827433,0.670850,-1.945257,
        2.827433,0.671396,1.004859,
        2.827433,0.673985,0.203855,
        2.827433,0.680326,-0.133370,
        2.827433,0.685083,-0.942363,
        2.827433,0.688204,-2.632204,
        2.827433,0.689777,2.385011,
        2.827433,0.700065,3.115430,
        2.827433,0.704080,2.911636,
        2.827433,0.712883,1.174216,
        2.827433,0.715701,0.699906,
        2.827433,0.718851,-1.442215,
        2.827433,0.721019,2.721536,
        2.827433,0.725300,-2.975134,
        2.827433,0.727136,-1.633288,
        2.827433,0.727399,-2.120567,
        2.827433,0.730751,0.522488,
        2.827433,0.732892,-1.254330,
        2.827433,0.734333,-2.446184,
        2.827433,0.735125,0.037270,
        2.827433,0.737893,1.866463,
        2.827433,0.740650,1.681075,
        2.827433,0.745971,-0.781286,
        2.827433,0.749432,-0.442559,
        2.827433,0.750102,2.050636,
        2.827433,0.751046,0.863821,
        2.827433,0.757018,-0.608833,
        2.827433,0.757046,-1.816019,
        2.827433,0.757389,1.501184,
        2.827433,0.759800,2.546181,
        2.827433,0.759859,0.351923,
        2.827433,0.761141,-2.781102,
        2.827433,0.769035,-1.078574,
        2.827433,0.770166,-0.266604,
        2.827433,0.772218,2.227380,
        2.827433,0.788382,1.331051,
        2.827433,0.790269,-2.282409,
        2.827433,0.803410,-1.984567,
        2.827433,0.803813,0.191520,
        2.827433,0.806832,3.022155,
        2.827433,0.807476,-2.597675,
        2.827433,0.812512,-0.107335,
        2.827433,0.814161,1.010949,
        2.827433,0.817732,-3.084486,
        2.827433,0.817846,2.390688,
        2.827433,0.819323,2.845960,
        2.827433,0.819430,-0.915017,
        2.827433,0.829198,-1.532491,
        2.827433,0.829531,0.665695,
        2.827433,0.829816,1.167147,
        2.827433,0.833613,-1.357492,
        2.827433,0.842945,-2.907318,
        2.827433,0.844883,2.679708,
        2.827433,0.848189,-1.705003,
        2.827433,0.849888,1.767901,
        2.827433,0.851126,0.493370,
        2.827433,0.852911,1.933394,
        2.827433,0.857516,-2.432863,
        2.827433,0.858261,-1.191729,
        2.827433,0.858279,-2.138895,
        2.827433,0.859723,-0.734928,
        2.827433,0.859754,1.603292,
        2.827433,0.861655,-0.394821,
        2.827433,0.864300,0.043549,
        2.827433,0.866940,0.851229,
        2.827433,0.869748,2.095969,
        2.827433,0.877352,-0.574668,
        2.827433,0.879254,-2.733338,
        2.827433,0.882554,1.443757,
        2.827433,0.884060,0.334199,
        2.827433,0.885626,-1.866061,
        2.827433,0.889834,2.529957,
        2.827433,0.893140,2.251698,
        2.827433,0.897463,-0.240822,
        2.827433,0.898758,-1.031948,
        2.827433,0.913918,1.287618,
        2.827433,0.916555,-2.287173,
        2.827433,0.918305,3.107845,
        2.827433,0.919522,2.948277,
        2.827433,0.927138,-2.569686,
        2.827433,0.929031,0.185383,
        2.827433,0.933766,-3.015917,
        2.827433,0.934873,-2.010822,
        2.827433,0.935636,-1.445855,
        2.827433,0.936702,-0.866461,
        2.827433,0.937395,0.729108,
        2.827433,0.937910,2.787575,
        2.827433,0.939144,-0.092962,
        2.827433,0.941718,1.125085,
        2.827433,0.942873,-1.607680,
        2.827433,0.943701,0.980328,
        2.827433,0.947774,2.395256,
        2.827433,0.952433,0.584347,
        2.827433,0.954456,-1.287293,
        2.827433,0.958731,-2.852552,
        2.827433,0.961838,1.839284,
        2.827433,0.961898,-0.469425,
        2.827433,0.965426,1.686615,
        2.827433,0.970108,1.990771,
        2.827433,0.972820,-1.762780,
        2.827433,0.973500,-0.711881,
        2.827433,0.975253,2.641082,
        2.827433,0.977824,0.441928,
        2.827433,0.981485,-2.424766,
        2.827433,0.981877,1.537100,
        2.827433,0.982609,-1.136920,
        2.827433,0.984265,-2.156073,
        2.827433,0.989150,0.045493,
        2.827433,0.993052,2.137478,
        2.827433,0.997766,-2.693871,
        2.827433,0.999847,-0.334738,
        2.827433,1.001920,0.848054,
        2.827433,1.006146,1.389734,
        2.827433,1.012180,0.304066,
        2.827433,1.016938,2.270674,
        2.827433,1.017371,-0.979790,
        2.827433,1.019495,-1.894487,
        2.827433,1.023796,-0.583691,
        2.827433,1.026525,2.506945,
        2.827433,1.027004,-0.203653,
        2.827433,1.028027,3.036573,
        2.827433,1.031435,1.241021,
        2.827433,1.032316,2.884221,
        2.827433,1.035863,-3.110475,
        2.827433,1.042085,-2.292950,
        2.827433,1.043272,-1.522986,
        2.827433,1.046265,-2.958847,
        2.827433,1.049272,-1.376344,
        2.827433,1.050039,-2.552086,
        2.827433,1.053682,0.689863,
        2.827433,1.055458,0.170394,
        2.827433,1.056622,-0.831866,
        2.827433,1.057515,1.097998,
        2.827433,1.058706,-2.036438,
        2.827433,1.061389,-1.672248,
        2.827433,1.063758,2.743093,
        2.827433,1.070157,-0.073422,
        2.827433,1.072154,1.761506,
        2.827433,1.074029,1.909193,
        2.827433,1.074204,0.545276,
        2.827433,1.074714,-2.805147,
        2.827433,1.076262,0.966104,
        2.827433,1.077642,-1.227073,
        2.827433,1.084151,1.618525,
        2.827433,1.085085,2.377673,
        2.827433,1.086822,-0.441112,
        2.827433,1.087142,2.058815,
        2.827433,1.098383,-0.696430,
        2.827433,1.099138,-1.084033,
        2.827433,1.103208,1.477773,
        2.827433,1.103783,0.408254,
        2.827433,1.106275,-2.420800,
        2.827433,1.108079,2.612943,
        2.827433,1.109345,-2.172718,
        2.827433,1.115957,-1.787597,
        2.827433,1.118930,0.045983,
        2.827433,1.119777,0.834702,
        2.827433,1.121943,-2.669042,
        2.827433,1.124930,2.219917,
        2.827433,1.125217,1.336294,
        2.827433,1.130958,2.972931,
        2.827433,1.131400,-0.306629,
        2.827433,1.137135,-0.936514,
        2.827433,1.139125,-1.924528,
        2.827433,1.139877,0.276382,
        2.827433,1.142003,-3.058022,
        2.827433,1.147363,-0.558448,
        2.827433,1.150348,-0.176418,
        2.827433,1.151198,1.197016,
        2.827433,1.153930,-1.597104,
        2.827433,1.155389,2.837148,
        2.827433,1.155458,-1.460623,
        2.827433,1.157331,-2.906362,
        2.827433,1.158594,2.487618,
        2.827433,1.162103,-1.322553,
        2.827433,1.165407,3.092861,
        2.827433,1.168025,0.717075,
        2.827433,1.168406,-2.301519,
        2.827433,1.174782,-2.538755,
        2.827433,1.177998,1.061256,
        2.827433,1.178898,1.827305,
        2.827433,1.181331,-2.060330,
        2.827433,1.182540,-0.798952,
        2.827433,1.184569,0.147517,
        2.827433,1.184734,1.968919,
        2.827433,1.184804,0.593386,
        2.827433,1.188278,1.690808,
        2.827433,1.194360,2.110440,
        2.827433,1.194776,2.710455,
        2.827433,1.197745,-1.173705,
        2.827433,1.199898,-2.777257,
        2.827433,1.201164,2.355931,
        2.827433,1.203760,1.556404,
        2.827433,1.205845,-0.415993,
        2.827433,1.206585,0.933648,
        2.827433,1.206950,-0.037248,
        2.827433,1.209341,0.471162,
        2.827433,1.213483,-1.039417,
        2.827433,1.218146,-0.672643,
        2.827433,1.221608,1.421669,
        2.827433,1.226043,-1.693533,
        2.827433,1.232870,-1.822773,
        2.827433,1.234375,-2.421194,
        2.827433,1.234531,-2.190148,
        2.827433,1.239414,0.348959,
        2.827433,1.240183,2.587255,
        2.827433,1.242312,2.941234,
        2.827433,1.244075,1.286644,
        2.827433,1.245961,-2.997461,
        2.827433,1.246495,-2.646872,
        2.827433,1.247736,2.242810,
        2.827433,1.253604,0.812222,
        2.827433,1.257816,-0.894317,
        2.827433,1.258247,-1.953634,
        2.827433,1.258993,-0.284099,
        2.827433,1.262380,-0.157609,
        2.827433,1.262791,-1.549229,
        2.827433,1.263772,-0.533073,
        2.827433,1.266731,-1.412461,
        2.827433,1.268591,-3.122146,
        2.827433,1.269614,-1.275246,
        2.827433,1.270745,1.152533,
        2.827433,1.273606,0.224545,
        2.827433,1.279716,2.812101,
        2.827433,1.283786,2.462124,
        2.827433,1.284798,0.078209,
        2.827433,1.285402,-2.882049,
        2.827433,1.286936,1.886053,
        2.827433,1.289408,0.683122,
        2.827433,1.294736,1.756184,
        2.827433,1.296978,-2.311511,
        2.827433,1.298546,2.020601,
        2.827433,1.300924,1.022478,
        2.827433,1.303305,3.043152,
        2.827433,1.304210,-2.084225,
        2.827433,1.305241,-2.530249,
        2.827433,1.307513,1.628003,
        2.827433,1.310722,0.558267,
        2.827433,1.314142,2.139329,
        2.827433,1.316087,-0.777971,
        2.827433,1.317907,-1.121023,
        2.827433,1.321019,-2.749776,
        2.827433,1.321609,2.687717,
        2.827433,1.322568,1.499415,
        2.827433,1.322972,-0.395699,
        2.827433,1.325286,-0.994192,
        2.827433,1.329146,-0.648513,
        2.827433,1.329795,-0.042570,
        2.827433,1.331651,2.344497,
        2.827433,1.336522,0.435098,
        2.827433,1.339410,0.900087,
        2.827433,1.339550,1.370353,
        2.827433,1.340441,-1.731580,
        2.827433,1.351855,-1.855234,
        2.827433,1.360161,-3.018404,
        2.827433,1.361031,-1.613971,
        2.827433,1.361196,2.914780,
        2.827433,1.361998,-2.207254,
        2.827433,1.362302,1.239384,
        2.827433,1.365228,2.563808,
        2.827433,1.365452,0.310522,
        2.827433,1.367028,-2.420041,
        2.827433,1.373422,-1.484281,
        2.827433,1.374245,-1.226417,
        2.827433,1.375020,-0.170393,
        2.827433,1.376560,-1.355481,
        2.827433,1.377070,-1.980984,
        2.827433,1.377578,-2.636339,
        2.827433,1.378096,0.774626,
        2.827433,1.379086,-0.508192,
        2.827433,1.382562,0.177141,
        2.827433,1.388849,2.233847,
        2.827433,1.389268,-0.886349,
        2.827433,1.391324,1.109396,
        2.827433,1.391699,-0.288725,
        2.827433,1.392869,3.136903,
        2.827433,1.393612,1.952016,
        2.827433,1.395545,-2.850807,
        2.827433,1.401926,1.822930,
        2.827433,1.402149,2.788898,
        2.827433,1.409552,0.647980,
        2.827433,1.410840,0.056219,
        2.827433,1.411944,2.444377,
        2.827433,1.413639,1.697982,
        2.827433,1.426479,1.573427,
        2.827433,1.426725,2.102242,
        2.827433,1.427868,0.984637,
        2.827433,1.429580,-2.107232,
        2.827433,1.431375,-2.313269,
        2.827433,1.431907,3.015410,
        2.827433,1.432325,-1.029054,
        2.827433,1.433341,0.524268,
        2.827433,1.435181,-0.739771,
        2.827433,1.437208,-2.529085,
        2.827433,1.439041,-0.617966,
        2.827433,1.441473,1.448544,
        2.827433,1.444527,2.664744,
        2.827433,1.450513,-0.400625,
        2.827433,1.451907,-0.066823,
        2.827433,1.452754,-2.738631,
        2.827433,1.454061,-2.950416,
        2.827433,1.456736,1.323154,
        2.827433,1.458370,0.400433,
        2.827433,1.459314,-1.765003,
        2.827433,1.459902,-1.145835,
        2.827433,1.464952,2.329245,
        2.827433,1.466050,0.862104,
        2.827433,1.471330,-1.886160,
        2.827433,1.475119,-1.648255,
        2.827433,1.480878,1.194576,
        2.827433,1.480942,0.273235,
        2.827433,1.481086,2.890269,
        2.827433,1.484118,-1.529156,
        2.827433,1.484741,-1.279219,
        2.827433,1.487681,-1.406069,
        2.827433,1.487886,-3.066037,
        2.827433,1.490555,2.543284,
        2.827433,1.492424,-0.190932,
        2.827433,1.495002,-0.839810,
        2.827433,1.495297,2.007188,
        2.827433,1.496414,-2.009267,
        2.827433,1.498261,-2.425270,
        2.827433,1.498907,-2.207327,
        2.827433,1.500204,0.737638,
        2.827433,1.502469,0.151156,
        2.827433,1.507448,-0.514782,
        2.827433,1.509287,2.204647,
        2.827433,1.510518,1.884920,
        2.827433,1.511948,-2.632360,
        2.827433,1.515235,1.068565,
        2.827433,1.522237,3.102018,
        2.827433,1.522373,1.764991,
        2.827433,1.522416,-0.954956,
        2.827433,1.523102,2.765920,
        2.827433,1.524811,-2.844718,
        2.827433,1.527261,-0.304916,
        2.827433,1.529788,0.613208,
        2.827433,1.533648,1.644808,
        2.827433,1.535018,0.028869,
        2.827433,1.541082,2.424608,
        2.827433,1.546010,1.523462,
        2.827433,1.551135,-0.642994,
        2.827433,1.553895,0.489607,
        2.827433,1.555041,0.947813,
        2.827433,1.559671,-1.074934,
        2.827433,1.560345,2.985899,
        2.827433,1.560487,-2.323400,
        2.827433,1.561127,1.401111,
        2.827433,1.567193,2.643172,
        2.827433,1.568892,2.098362,
        2.827433,1.571702,-2.098435,
        2.827433,1.572701,-2.528730,
        2.827433,1.573572,1.278593,
        2.827433,1.574849,-0.091505,
        2.827433,1.576334,0.365236,
        2.827433,1.577029,-2.962888,
        2.827433,1.577688,-1.814941,
        2.827433,1.579542,-0.755917,
        2.827433,1.579624,-1.199759,
        2.827433,1.581515,-0.419414,
        2.827433,1.584581,-2.738702,
        2.827433,1.586948,-1.696296,
        2.827433,1.588901,2.301496,
        2.827433,1.589936,0.825020,
        2.827433,1.590018,-1.941652,
        2.827433,1.596968,-1.328358,
        2.827433,1.597178,-1.575607,
        2.827433,1.600470,-1.453196,
        2.827433,1.601408,0.242409,
        2.827433,1.602259,1.152656,
        2.827433,1.603064,2.865774,
        2.827433,1.612085,1.968802,
        2.827433,1.612720,-3.080498,
        2.827433,1.616061,2.521531,
        2.827433,1.616900,-0.207457,
        2.827433,1.620988,-0.875920,
        2.827433,1.622114,0.701509,
        2.827433,1.622749,0.121426,
        2.827433,1.622993,-2.221629,
        2.827433,1.629346,-0.542669,
        2.827433,1.630007,1.844552,
        2.827433,1.634747,-2.426198,
        2.827433,1.642346,1.722617,
        2.827433,1.643925,1.034629,
        2.827433,1.644848,2.743774,
        2.827433,1.645258,2.188109,
        2.827433,1.645417,-2.634725,
        2.827433,1.647395,3.085437,
        2.827433,1.648538,-2.855044,
        2.827433,1.649932,0.576808,
        2.827433,1.652224,-0.996105,
        2.827433,1.652871,1.600681,
        2.827433,1.657997,-0.325127,
        2.827433,1.661936,0.001037,
        2.827433,1.665187,1.477402,
        2.827433,1.665496,2.397872,
        2.827433,1.673267,-0.665453,
        2.827433,1.673557,0.452526,
        2.827433,1.679277,-1.122421,
        2.827433,1.679696,0.912482,
        2.827433,1.682420,1.353922,
        2.827433,1.685031,-2.009038,
        2.827433,1.685196,2.967046,
        2.827433,1.690107,2.620910,
        2.827433,1.690453,-2.123076,
        2.827433,1.691693,-1.871572,
        2.827433,1.691817,2.061688,
        2.827433,1.692644,1.237670,
        2.827433,1.696386,0.328123,
        2.827433,1.697162,-1.249026,
        2.827433,1.697488,-2.322785,
        2.827433,1.698726,-2.973576,
        2.827433,1.699172,-1.749717,
        2.827433,1.704478,-0.106295,
        2.827433,1.707244,-0.444770,
        2.827433,1.707500,-2.531497,
        2.827433,1.707714,-0.787321,
        2.827433,1.708430,-1.627552,
        2.827433,1.710583,-1.376890,
        2.827433,1.712415,-2.747921,
        2.827433,1.713125,-1.503073,
        2.827433,1.713452,0.789419,
        2.827433,1.720612,2.279888,
        2.827433,1.724852,0.202831,
        2.827433,1.724868,2.846246,
        2.827433,1.728940,1.931535,
        2.827433,1.732838,1.127584,
        2.827433,1.735719,-3.092485,
        2.827433,1.739676,2.496017,
        2.827433,1.739971,-0.228876,
        2.827433,1.743328,0.663347,
        2.827433,1.744927,-0.911892,
        2.827433,1.745952,0.084733,
        2.827433,1.748077,1.804354,
        2.827433,1.752959,-0.567548,
        2.827433,1.760169,1.678132,
        2.827433,1.763887,-2.218450,
        2.827433,1.766737,2.722752,
        2.827433,1.767999,1.004737,
        2.827433,1.769149,1.550298,
        2.827433,1.770119,-2.427175,
        2.827433,1.770460,2.154208,
        2.827433,1.770514,3.070982,
        2.827433,1.770735,0.537245,
        2.827433,1.772673,-2.864760,
        2.827433,1.773527,-1.039582,
        2.827433,1.776419,-2.642421,
        2.827433,1.780769,1.416330,
        2.827433,1.786634,-0.346247,
        2.827433,1.788829,-1.947553,
        2.827433,1.793737,0.410270,
        2.827433,1.794553,2.374618,
        2.827433,1.794699,-0.694061,
        2.827433,1.797469,-1.169105,
        2.827433,1.803705,0.881076,
        2.827433,1.804124,1.264249,
        2.827433,1.807825,-2.090455,
        2.827433,1.808822,-1.817375,
        2.827433,1.809020,2.949741,
        2.827433,1.812466,2.597595,
        2.827433,1.813159,2.025395,
        2.827433,1.813816,-1.298435,
        2.827433,1.815016,0.278951,
        2.827433,1.816510,-0.008631,
        2.827433,1.817772,-1.689625,
        2.827433,1.821360,-2.984352,
        2.827433,1.822312,-0.127211,
        2.827433,1.823386,-1.429807,
        2.827433,1.823834,-1.560483,
        2.827433,1.832060,-0.821627,
        2.827433,1.832690,-0.468440,
        2.827433,1.833945,-2.320278,
        2.827433,1.834787,0.751968,
        2.827433,1.840579,-2.536445,
        2.827433,1.841033,-2.756932,
        2.827433,1.846350,1.893983,
        2.827433,1.846644,1.113766,
        2.827433,1.846833,2.826543,
        2.827433,1.847325,2.249051,
        2.827433,1.854119,0.134161,
        2.827433,1.858863,-3.106022,
        2.827433,1.865046,1.763384,
        2.827433,1.865619,-0.952036,
        2.827433,1.865912,0.624453,
        2.827433,1.866694,2.473686,
        2.827433,1.867524,-0.240083,
        2.827433,1.875561,1.632831,
        2.827433,1.876049,-0.594378,
        2.827433,1.880509,1.500204,
        2.827433,1.883504,1.360457,
        2.827433,1.886971,-2.196665,
        2.827433,1.888979,0.982922,
        2.827433,1.889815,2.701570,
        2.827433,1.892144,-2.005340,
        2.827433,1.892348,0.496522,
        2.827433,1.893281,-1.083976,
        2.827433,1.893828,3.053453,
        2.827433,1.895117,2.121107,
        2.827433,1.898741,-2.875735,
        2.827433,1.904429,-2.428029,
        2.827433,1.908408,-2.648903,
        2.827433,1.911828,-0.364489,
        2.827433,1.912224,0.366565,
        2.827433,1.914854,1.214468,
        2.827433,1.915467,-1.883143,
        2.827433,1.915763,-0.724679,
        2.827433,1.915790,-1.216366,
        2.827433,1.922256,2.347806,
        2.827433,1.922616,0.844633,
        2.827433,1.925185,0.231170,
        2.827433,1.927474,-1.754616,
        2.827433,1.928342,-1.351174,
        2.827433,1.933328,-1.623328,
        2.827433,1.933878,0.034150,
        2.827433,1.934585,-1.487979,
        2.827433,1.934802,2.927634,
        2.827433,1.934862,1.989776,
        2.827433,1.936327,-0.099350,
        2.827433,1.938088,2.578507,
        2.827433,1.945677,-2.999805,
        2.827433,1.954572,-0.859208,
        2.827433,1.957868,-0.490784,
        2.827433,1.958989,0.712772,
        2.827433,1.960212,-2.307561,
        2.827433,1.964774,1.855507,
        2.827433,1.964971,-2.100533,
        2.827433,1.969112,1.088273,
        2.827433,1.971438,-2.765415,
        2.827433,1.972312,2.805134,
        2.827433,1.974950,2.220596,
        2.827433,1.975135,-2.539127,
        2.827433,1.982107,1.720829,
        2.827433,1.983195,-3.127984,
        2.827433,1.983318,-0.247182,
        2.827433,1.986720,-0.993220,
        2.827433,1.990381,1.586005,
        2.827433,1.990470,0.580283,
        2.827433,1.990523,1.450326,
        2.827433,1.993909,1.312897,
        2.827433,1.994694,2.452813,
        2.827433,1.999226,-0.620717,
        2.827433,2.005145,0.941012,
        2.827433,2.007301,0.142337,
        2.827433,2.014149,0.444138,
        2.827433,2.014338,-1.127392,
        2.827433,2.018131,3.021778,
        2.827433,2.019676,-1.980944,
        2.827433,2.019912,2.686612,
        2.827433,2.020545,2.088672,
        2.827433,2.026074,0.300945,
        2.827433,2.026196,-2.889057,
        2.827433,2.031959,-1.268178,
        2.827433,2.032917,-2.199917,
        2.827433,2.033572,-1.842884,
        2.827433,2.034712,-2.419505,
        2.827433,2.035818,-0.379479,
        2.827433,2.036004,-0.757062,
        2.827433,2.040501,-1.411488,
        2.827433,2.042874,-1.701367,
        2.827433,2.043060,-2.653746,
        2.827433,2.043729,-1.557938,
        2.827433,2.048538,0.806623,
        2.827433,2.049006,1.199930,
        2.827433,2.049502,-0.137989,
        2.827433,2.052181,2.326959,
        2.827433,2.052794,-0.008526,
        2.827433,2.057608,1.951726,
        2.827433,2.065568,2.567730,
        2.827433,2.070371,2.879808,
        2.827433,2.071256,-3.018961,
        2.827433,2.079746,-0.897699,
        2.827433,2.082784,1.042403,
        2.827433,2.083752,-0.511086,
        2.827433,2.084401,1.810502,
        2.827433,2.085731,0.671081,
        2.827433,2.097496,1.373172,
        2.827433,2.099185,1.667006,
        2.827433,2.100482,-2.301084,
        2.827433,2.103118,2.194745,
        2.827433,2.103223,1.521685,
        2.827433,2.103695,-2.777707,
        2.827433,2.107523,3.129581,
        2.827433,2.108277,-2.095907,
        2.827433,2.108891,-2.532166,
        2.827433,2.111003,-0.268632,
        2.827433,2.112276,-1.027246,
        2.827433,2.114380,0.531457,
        2.827433,2.119166,0.100552,
        2.827433,2.119564,2.754818,
        2.827433,2.122474,-0.648314,
        2.827433,2.125183,2.445598,
        2.827433,2.129501,0.238290,
        2.827433,2.131178,0.386663,
        2.827433,2.131223,-1.174311,
        2.827433,2.134269,-1.948042,
        2.827433,2.134684,0.907065,
        2.827433,2.137615,2.988806,
        2.827433,2.145635,-1.327708,
        2.827433,2.146467,2.055081,
        2.827433,2.146582,-1.793788,
        2.827433,2.149210,-1.482321,
        2.827433,2.152901,-1.636788,
        2.827433,2.156337,1.151558,
        2.827433,2.156474,-2.912065,
        2.827433,2.157194,-0.792563,
        2.827433,2.166724,-0.401638,
        2.827433,2.166980,-0.134879,
        2.827433,2.173644,-2.403429,
        2.827433,2.175711,-2.655523,
        2.827433,2.177132,2.600419,
        2.827433,2.178879,0.767722,
        2.827433,2.180752,1.908485,
        2.827433,2.181338,2.313721,
        2.827433,2.185580,1.287673,
        2.827433,2.195400,-2.217546,
        2.827433,2.200418,-3.056442,
        2.827433,2.201829,-0.001735,
        2.827433,2.203804,1.756176,
        2.827433,2.211602,1.443930,
        2.827433,2.212686,-0.544809,
        2.827433,2.213937,0.621892,
        2.827433,2.214075,1.600691,
        2.827433,2.216256,1.020029,
        2.827433,2.219060,2.886631,
        2.827433,2.223375,-0.906185,
        2.827433,2.226002,-1.060248,
        2.827433,2.226413,-2.064034,
        2.827433,2.230554,2.171334,
        2.827433,2.232413,2.735777,
        2.827433,2.233709,3.070830,
        2.827433,2.234383,-2.792065,
        2.827433,2.237183,0.468799,
        2.827433,2.237695,-0.271388,
        2.827433,2.239419,0.151693,
        2.827433,2.243347,-2.521049,
        2.827433,2.244035,-1.229456,
        2.827433,2.244355,0.309899,
        2.827433,2.245175,2.460236,
        2.827433,2.245628,-1.899216,
        2.827433,2.245973,-0.702234,
        2.827433,2.255601,-1.728131,
        2.827433,2.256702,-1.401367,
        2.827433,2.261466,-1.557500,
        2.827433,2.267390,0.878499,
        2.827433,2.272099,2.018594,
        2.827433,2.278618,1.170386,
        2.827433,2.283304,-2.345584,
        2.827433,2.285175,-2.943090,
        2.827433,2.293000,-0.124451,
        2.827433,2.294058,-0.421070,
        2.827433,2.303283,1.855938,
        2.827433,2.304968,2.608978,
        2.827433,2.305347,2.311346,
        2.827433,2.307188,-2.656134,
        2.827433,2.307301,1.336200,
        2.827433,2.309655,0.726020,
        2.827433,2.317634,-2.185225,
        2.827433,2.321510,1.685664,
        2.827433,2.324376,-3.111006,
        2.827433,2.324731,1.511186,
        2.827433,2.329587,0.035498,
        2.827433,2.330158,2.977264,
        2.827433,2.335096,-0.588468,
        2.827433,2.336046,-1.115799,
        2.827433,2.338060,2.803613,
        2.827433,2.338083,-0.943400,
        2.827433,2.340210,0.562145,
        2.827433,2.341931,1.023926,
        2.827433,2.343265,-0.779061,
        2.827433,2.343434,-2.013692,
        2.827433,2.353029,-1.310529,
        2.827433,2.354868,0.211543,
        2.827433,2.355120,0.388593,
        2.827433,2.355165,-1.833173,
        2.827433,2.355462,2.151441,
        2.827433,2.362948,-2.487477,
        2.827433,2.363101,-2.810456,
        2.827433,2.363394,-0.272918,
        2.827433,2.364509,-1.644533,
        2.827433,2.370465,2.466824,
        2.827433,2.386110,-1.481351,
        2.827433,2.394960,0.862133,
        2.827433,2.395716,1.977184,
        2.827433,2.396339,1.200900,
        2.827433,2.407941,-2.316753,
        2.827433,2.409450,-2.988432,
        2.827433,2.417202,-0.103363,
        2.827433,2.417386,-0.443097,
        2.827433,2.417950,2.666662,
        2.827433,2.423190,1.789783,
        2.827433,2.423692,1.392696,
        2.827433,2.431122,3.086379,
        2.827433,2.431406,-2.647820,
        2.827433,2.431584,2.307631,
        2.827433,2.434090,1.592818,
        2.827433,2.435993,0.683844,
        2.827433,2.440898,-2.133300,
        2.827433,2.442222,-1.195837,
        2.827433,2.443112,2.883761,
        2.827433,2.449191,-0.633651,
        2.827433,2.450821,0.084741,
        2.827433,2.453499,-1.008861,
        2.827433,2.460010,-1.938011,
        2.827433,2.460199,-0.822008,
        2.827433,2.460902,0.490079,
        2.827433,2.461774,-1.734913,
        2.827433,2.463986,1.034407,
        2.827433,2.467647,0.287667,
        2.827433,2.479849,2.126925,
        2.827433,2.485907,-2.835471,
        2.827433,2.485945,-1.374853,
        2.827433,2.487747,2.505912,
        2.827433,2.487957,-0.274279,
        2.827433,2.489741,-2.468428,
        2.827433,2.512538,1.239900,
        2.827433,2.517528,1.923289,
        2.827433,2.518454,-1.565852,
        2.827433,2.519348,0.843503,
        2.827433,2.522766,-3.059231,
        2.827433,2.528686,2.737228,
        2.827433,2.534798,-2.266671,
        2.827433,2.536533,1.466274,
        2.827433,2.538504,1.699812,
        2.827433,2.539516,-0.476942,
        2.827433,2.539532,-0.074548,
        2.827433,2.543197,2.990252,
        2.827433,2.554022,2.320619,
        2.827433,2.556919,-2.655005,
        2.827433,2.558146,0.627191,
        2.827433,2.558223,-1.164189,
        2.827433,2.563452,-2.044028,
        2.827433,2.568462,0.150655,
        2.827433,2.569130,-0.700837,
        2.827433,2.570050,-1.804492,
        2.827433,2.572710,-0.934665,
        2.827433,2.575994,0.391718,
        2.827433,2.584774,1.046703,
        2.827433,2.601432,2.102535,
        2.827433,2.603559,2.555114,
        2.827433,2.604150,-2.885208,
        2.827433,2.608630,-1.378011,
        2.827433,2.611256,-0.279110,
        2.827433,2.617344,-2.439480,
        2.827433,2.627402,1.293552,
        2.827433,2.635396,1.848850,
        2.827433,2.635686,2.834950,
        2.827433,2.637087,3.132730,
        2.827433,2.642029,0.813274,
        2.827433,2.645260,1.570203,
        2.827433,2.646182,-1.614036,
        2.827433,2.660113,-2.186513,
        2.827433,2.660483,-0.034608,
        2.827433,2.661713,-0.529511,
        2.827433,2.672720,-1.110422,
        2.827433,2.676626,2.335353,
        2.827433,2.676638,0.540364,
        2.827433,2.676780,-1.901158,
        2.827433,2.681411,-2.667965,
        2.827433,2.682152,0.245751,
        2.827433,2.682859,-0.817071,
        2.827433,2.705107,1.062142,
        2.827433,2.717705,2.625995,
        2.827433,2.721001,2.060097,
        2.827433,2.722319,-2.954069,
        2.827433,2.729840,-1.368237,
        2.827433,2.734349,-0.289298,
        2.827433,2.738409,2.975671,
        2.827433,2.740366,1.374341,
        2.827433,2.744633,-2.391311,
        2.827433,2.747398,1.727166,
        2.827433,2.763799,0.760372,
        2.827433,2.768238,-1.679243,
        2.827433,2.779245,-2.046268,
        2.827433,2.780208,-0.626453,
        2.827433,2.780471,0.027732,
        2.827433,2.784696,-1.013128,
        2.827433,2.789817,0.395431,
        2.827433,2.801154,2.353143,
        2.827433,2.807091,-2.687477,
        2.827433,2.825762,1.085035,
        2.827433,2.830209,2.734943,
        2.827433,2.833517,-3.085508,
        2.827433,2.838301,1.980061,
        2.827433,2.849871,1.512012,
        2.827433,2.850089,-1.343899,
        2.827433,2.856905,-0.318441,
        2.827433,2.870445,-2.291001,
        2.827433,2.885017,0.655562,
        2.827433,2.886428,-1.793217,
        2.827433,2.890337,-0.825633,
        2.827433,2.898932,0.139973,
        2.827433,2.931927,-2.728768,
        2.827433,2.932096,2.377790,
        2.827433,2.937409,2.947245,
        2.827433,2.948048,1.133127,
        2.827433,2.950896,1.797889,
        2.827433,2.970019,-1.282382,
        2.827433,2.978333,-0.412090,
        2.827433,2.996437,-2.089121,
        2.827433,3.011274,0.434110,
        2.827433,3.050410,3.138492,
        2.827433,3.064492,1.647259,
        2.827433,3.088868,-0.989501,
        2.932153,0.053884,-2.408394,
        2.932153,0.077645,1.453433,
        2.932153,0.088408,-0.311160,
        2.932153,0.142314,2.590387,
        2.932153,0.147206,-1.385164,
        2.932153,0.163506,0.573203,
        2.932153,0.165485,-2.875396,
        2.932153,0.182564,-2.125821,
        2.932153,0.190569,1.929141,
        2.932153,0.201102,1.249082,
        2.932153,0.201397,-0.716646,
        2.932153,0.210568,-0.069559,
        2.932153,0.251672,2.848900,
        2.932153,0.266262,-1.678068,
        2.932153,0.269108,-1.215563,
        2.932153,0.273777,-2.506756,
        2.932153,0.279854,0.814581,
        2.932153,0.281664,-2.969147,
        2.932153,0.283625,2.405603,
        2.932153,0.284058,0.357350,
        2.932153,0.294107,1.593880,
        2.932153,0.302491,-0.400410,
        2.932153,0.308059,2.021491,
        2.932153,0.312219,-2.072057,
        2.932153,0.321905,-0.829121,
        2.932153,0.329984,1.201621,
        2.932153,0.339714,-0.013712,
        2.932153,0.368784,2.974974,
        2.932153,0.376428,-1.495532,
        2.932153,0.378103,2.646436,
        2.932153,0.385088,0.590355,
        2.932153,0.386248,-2.690943,
        2.932153,0.389851,-2.340031,
        2.932153,0.391502,-1.171374,
        2.932153,0.398249,-1.812694,
        2.932153,0.400542,-2.995729,
        2.932153,0.404273,0.918163,
        2.932153,0.405617,1.773533,
        2.932153,0.409009,-0.570243,
        2.932153,0.410700,0.270433,
        2.932153,0.415893,1.464890,
        2.932153,0.420680,2.088164,
        2.932153,0.424766,-0.258729,
        2.932153,0.426507,2.367872,
        2.932153,0.441292,-0.890727,
        2.932153,0.451839,-2.078193,
        2.932153,0.459270,1.191562,
        2.932153,0.470622,0.014002,
        2.932153,0.481292,-2.529098,
        2.932153,0.482460,2.849843,
        2.932153,0.492205,-1.397963,
        2.932153,0.493314,3.117264,
        2.932153,0.497870,0.727828,
        2.932153,0.499783,-1.654379,
        2.932153,0.502651,0.463333,
        2.932153,0.505463,2.594557,
        2.932153,0.506104,-2.874910,
        2.932153,0.513579,-1.147354,
        2.932153,0.518569,1.644237,
        2.932153,0.519001,1.900488,
        2.932153,0.519277,-0.673809,
        2.932153,0.524769,-0.421955,
        2.932153,0.525564,-2.288838,
        2.932153,0.532541,0.978737,
        2.932153,0.534610,2.154808,
        2.932153,0.535020,-1.891846,
        2.932153,0.540208,1.400497,
        2.932153,0.542006,0.228337,
        2.932153,0.550762,-0.180771,
        2.932153,0.560531,2.378474,
        2.932153,0.561629,-0.924576,
        2.932153,0.567465,-2.679392,
        2.932153,0.588527,1.181755,
        2.932153,0.591643,-2.098364,
        2.932153,0.594963,2.997575,
        2.932153,0.599643,-3.045235,
        2.932153,0.599962,2.773556,
        2.932153,0.602620,0.029270,
        2.932153,0.605711,0.598454,
        2.932153,0.607708,-1.541026,
        2.932153,0.609360,-2.464379,
        2.932153,0.610738,-1.326786,
        2.932153,0.616968,0.820489,
        2.932153,0.626379,1.773610,
        2.932153,0.628626,-1.751180,
        2.932153,0.630671,0.391949,
        2.932153,0.632622,-0.737642,
        2.932153,0.632708,2.566722,
        2.932153,0.632797,1.989858,
        2.932153,0.634455,-0.534790,
        2.932153,0.636205,1.561170,
        2.932153,0.637681,-1.123423,
        2.932153,0.641441,-2.842109,
        2.932153,0.644807,-0.327139,
        2.932153,0.652440,2.197129,
        2.932153,0.660897,-2.278593,
        2.932153,0.664008,1.360091,
        2.932153,0.670850,-1.945257,
        2.932153,0.671396,1.004859,
        2.932153,0.673985,0.203855,
        2.932153,0.680326,-0.133370,
        2.932153,0.685083,-0.942363,
        2.932153,0.688204,-2.632204,
        2.932153,0.689777,2.385011,
        2.932153,0.700065,3.115430,
        2.932153,0.704080,2.911636,
        2.932153,0.712883,1.174216,
        2.932153,0.715701,0.699906,
        2.932153,0.718851,-1.442215,
        2.932153,0.721019,2.721536,
        2.932153,0.725300,-2.975134,
        2.932153,0.727136,-1.633288,
        2.932153,0.727399,-2.120567,
        2.932153,0.730751,0.522488,
        2.932153,0.732892,-1.254330,
        2.932153,0.734333,-2.446184,
        2.932153,0.735125,0.037270,
        2.932153,0.737893,1.866463,
        2.932153,0.740650,1.681075,
        2.932153,0.745971,-0.781286,
        2.932153,0.749432,-0.442559,
        2.932153,0.750102,2.050636,
        2.932153,0.751046,0.863821,
        2.932153,0.757018,-0.608833,
        2.932153,0.757046,-1.816019,
        2.932153,0.757389,1.501184,
        2.932153,0.759800,2.546181,
        2.932153,0.759859,0.351923,
        2.932153,0.761141,-2.781102,
        2.932153,0.769035,-1.078574,
        2.932153,0.770166,-0.266604,
        2.932153,0.772218,2.227380,
        2.932153,0.788382,1.331051,
        2.932153,0.790269,-2.282409,
        2.932153,0.803410,-1.984567,
        2.932153,0.803813,0.191520,
        2.932153,0.806832,3.022155,
        2.932153,0.807476,-2.597675,
        2.932153,0.812512,-0.107335,
        2.932153,0.814161,1.010949,
        2.932153,0.817732,-3.084486,
        2.932153,0.817846,2.390688,
        2.932153,0.819323,2.845960,
        2.932153,0.819430,-0.915017,
        2.932153,0.829198,-1.532491,
        2.932153,0.829531,0.665695,
        2.932153,0.829816,1.167147,
        2.932153,0.833613,-1.357492,
        2.932153,0.842945,-2.907318,
        2.932153,0.844883,2.679708,
        2.932153,0.848189,-1.705003,
        2.932153,0.849888,1.767901,
        2.932153,0.851126,0.493370,
        2.932153,0.852911,1.933394,
        2.932153,0.857516,-2.432863,
        2.932153,0.858261,-1.191729,
        2.932153,0.858279,-2.138895,
        2.932153,0.859723,-0.734928,
        2.932153,0.859754,1.603292,
        2.932153,0.861655,-0.394821,
        2.932153,0.864300,0.043549,
        2.932153,0.866940,0.851229,
        2.932153,0.869748,2.095969,
        2.932153,0.877352,-0.574668,
        2.932153,0.879254,-2.733338,
        2.932153,0.882554,1.443757,
        2.932153,0.884060,0.334199,
        2.932153,0.885626,-1.866061,
        2.932153,0.889834,2.529957,
        2.932153,0.893140,2.251698,
        2.932153,0.897463,-0.240822,
        2.932153,0.898758,-1.031948,
        2.932153,0.913918,1.287618,
        2.932153,0.916555,-2.287173,
        2.932153,0.918305,3.107845,
        2.932153,0.919522,2.948277,
        2.932153,0.927138,-2.569686,
        2.932153,0.929031,0.185383,
        2.932153,0.933766,-3.015917,
        2.932153,0.934873,-2.010822,
        2.932153,0.935636,-1.445855,
        2.932153,0.936702,-0.866461,
        2.932153,0.937395,0.729108,
        2.932153,0.937910,2.787575,
        2.932153,0.939144,-0.092962,
        2.932153,0.941718,1.125085,
        2.932153,0.942873,-1.607680,
        2.932153,0.943701,0.980328,
        2.932153,0.947774,2.395256,
        2.932153,0.952433,0.584347,
        2.932153,0.954456,-1.287293,
        2.932153,0.958731,-2.852552,
        2.932153,0.961838,1.839284,
        2.932153,0.961898,-0.469425,
        2.932153,0.965426,1.686615,
        2.932153,0.970108,1.990771,
        2.932153,0.972820,-1.762780,
        2.932153,0.973500,-0.711881,
        2.932153,0.975253,2.641082,
        2.932153,0.977824,0.441928,
        2.932153,0.981485,-2.424766,
        2.932153,0.981877,1.537100,
        2.932153,0.982609,-1.136920,
        2.932153,0.984265,-2.156073,
        2.932153,0.989150,0.045493,
        2.932153,0.993052,2.137478,
        2.932153,0.997766,-2.693871,
        2.932153,0.999847,-0.334738,
        2.932153,1.001920,0.848054,
        2.932153,1.006146,1.389734,
        2.932153,1.012180,0.304066,
        2.932153,1.016938,2.270674,
        2.932153,1.017371,-0.979790,
        2.932153,1.019495,-1.894487,
        2.932153,1.023796,-0.583691,
        2.932153,1.026525,2.506945,
        2.932153,1.027004,-0.203653,
        2.932153,1.028027,3.036573,
        2.932153,1.031435,1.241021,
        2.932153,1.032316,2.884221,
        2.932153,1.035863,-3.110475,
        2.932153,1.042085,-2.292950,
        2.932153,1.043272,-1.522986,
        2.932153,1.046265,-2.958847,
        2.932153,1.049272,-1.376344,
        2.932153,1.050039,-2.552086,
        2.932153,1.053682,0.689863,
        2.932153,1.055458,0.170394,
        2.932153,1.056622,-0.831866,
        2.932153,1.057515,1.097998,
        2.932153,1.058706,-2.036438,
        2.932153,1.061389,-1.672248,
        2.932153,1.063758,2.743093,
        2.932153,1.070157,-0.073422,
        2.932153,1.072154,1.761506,
        2.932153,1.074029,1.909193,
        2.932153,1.074204,0.545276,
        2.932153,1.074714,-2.805147,
        2.932153,1.076262,0.966104,
        2.932153,1.077642,-1.227073,
        2.932153,1.084151,1.618525,
        2.932153,1.085085,2.377673,
        2.932153,1.086822,-0.441112,
        2.932153,1.087142,2.058815,
        2.932153,1.098383,-0.696430,
        2.932153,1.099138,-1.084033,
        2.932153,1.103208,1.477773,
        2.932153,1.103783,0.408254,
        2.932153,1.106275,-2.420800,
        2.932153,1.108079,2.612943,
        2.932153,1.109345,-2.172718,
        2.932153,1.115957,-1.787597,
        2.932153,1.118930,0.045983,
        2.932153,1.119777,0.834702,
        2.932153,1.121943,-2.669042,
        2.932153,1.124930,2.219917,
        2.932153,1.125217,1.336294,
        2.932153,1.130958,2.972931,
        2.932153,1.131400,-0.306629,
        2.932153,1.137135,-0.936514,
        2.932153,1.139125,-1.924528,
        2.932153,1.139877,0.276382,
        2.932153,1.142003,-3.058022,
        2.932153,1.147363,-0.558448,
        2.932153,1.150348,-0.176418,
        2.932153,1.151198,1.197016,
        2.932153,1.153930,-1.597104,
        2.932153,1.155389,2.837148,
        2.932153,1.155458,-1.460623,
        2.932153,1.157331,-2.906362,
        2.932153,1.158594,2.487618,
        2.932153,1.162103,-1.322553,
        2.932153,1.165407,3.092861,
        2.932153,1.168025,0.717075,
        2.932153,1.168406,-2.301519,
        2.932153,1.174782,-2.538755,
        2.932153,1.177998,1.061256,
        2.932153,1.178898,1.827305,
        2.932153,1.181331,-2.060330,
        2.932153,1.182540,-0.798952,
        2.932153,1.184569,0.147517,
        2.932153,1.184734,1.968919,
        2.932153,1.184804,0.593386,
        2.932153,1.188278,1.690808,
        2.932153,1.194360,2.110440,
        2.932153,1.194776,2.710455,
        2.932153,1.197745,-1.173705,
        2.932153,1.199898,-2.777257,
        2.932153,1.201164,2.355931,
        2.932153,1.203760,1.556404,
        2.932153,1.205845,-0.415993,
        2.932153,1.206585,0.933648,
        2.932153,1.206950,-0.037248,
        2.932153,1.209341,0.471162,
        2.932153,1.213483,-1.039417,
        2.932153,1.218146,-0.672643,
        2.932153,1.221608,1.421669,
        2.932153,1.226043,-1.693533,
        2.932153,1.232870,-1.822773,
        2.932153,1.234375,-2.421194,
        2.932153,1.234531,-2.190148,
        2.932153,1.239414,0.348959,
        2.932153,1.240183,2.587255,
        2.932153,1.242312,2.941234,
        2.932153,1.244075,1.286644,
        2.932153,1.245961,-2.997461,
        2.932153,1.246495,-2.646872,
        2.932153,1.247736,2.242810,
        2.932153,1.253604,0.812222,
        2.932153,1.257816,-0.894317,
        2.932153,1.258247,-1.953634,
        2.932153,1.258993,-0.284099,
        2.932153,1.262380,-0.157609,
        2.932153,1.262791,-1.549229,
        2.932153,1.263772,-0.533073,
        2.932153,1.266731,-1.412461,
        2.932153,1.268591,-3.122146,
        2.932153,1.269614,-1.275246,
        2.932153,1.270745,1.152533,
        2.932153,1.273606,0.224545,
        2.932153,1.279716,2.812101,
        2.932153,1.283786,2.462124,
        2.932153,1.284798,0.078209,
        2.932153,1.285402,-2.882049,
        2.932153,1.286936,1.886053,
        2.932153,1.289408,0.683122,
        2.932153,1.294736,1.756184,
        2.932153,1.296978,-2.311511,
        2.932153,1.298546,2.020601,
        2.932153,1.300924,1.022478,
        2.932153,1.303305,3.043152,
        2.932153,1.304210,-2.084225,
        2.932153,1.305241,-2.530249,
        2.932153,1.307513,1.628003,
        2.932153,1.310722,0.558267,
        2.932153,1.314142,2.139329,
        2.932153,1.316087,-0.777971,
        2.932153,1.317907,-1.121023,
        2.932153,1.321019,-2.749776,
        2.932153,1.321609,2.687717,
        2.932153,1.322568,1.499415,
        2.932153,1.322972,-0.395699,
        2.932153,1.325286,-0.994192,
        2.932153,1.329146,-0.648513,
        2.932153,1.329795,-0.042570,
        2.932153,1.331651,2.344497,
        2.932153,1.336522,0.435098,
        2.932153,1.339410,0.900087,
        2.932153,1.339550,1.370353,
        2.932153,1.340441,-1.731580,
        2.932153,1.351855,-1.855234,
        2.932153,1.360161,-3.018404,
        2.932153,1.361031,-1.613971,
        2.932153,1.361196,2.914780,
        2.932153,1.361998,-2.207254,
        2.932153,1.362302,1.239384,
        2.932153,1.365228,2.563808,
        2.932153,1.365452,0.310522,
        2.932153,1.367028,-2.420041,
        2.932153,1.373422,-1.484281,
        2.932153,1.374245,-1.226417,
        2.932153,1.375020,-0.170393,
        2.932153,1.376560,-1.355481,
        2.932153,1.377070,-1.980984,
        2.932153,1.377578,-2.636339,
        2.932153,1.378096,0.774626,
        2.932153,1.379086,-0.508192,
        2.932153,1.382562,0.177141,
        2.932153,1.388849,2.233847,
        2.932153,1.389268,-0.886349,
        2.932153,1.391324,1.109396,
        2.932153,1.391699,-0.288725,
        2.932153,1.392869,3.136903,
        2.932153,1.393612,1.952016,
        2.932153,1.395545,-2.850807,
        2.932153,1.401926,1.822930,
        2.932153,1.402149,2.788898,
        2.932153,1.409552,0.647980,
        2.932153,1.410840,0.056219,
        2.932153,1.411944,2.444377,
        2.932153,1.413639,1.697982,
        2.932153,1.426479,1.573427,
        2.932153,1.426725,2.102242,
        2.932153,1.427868,0.984637,
        2.932153,1.429580,-2.107232,
        2.932153,1.431375,-2.313269,
        2.932153,1.431907,3.015410,
        2.932153,1.432325,-1.029054,
        2.932153,1.433341,0.524268,
        2.932153,1.435181,-0.739771,
        2.932153,1.437208,-2.529085,
        2.932153,1.439041,-0.617966,
        2.932153,1.441473,1.448544,
        2.932153,1.444527,2.664744,
        2.932153,1.450513,-0.400625,
        2.932153,1.451907,-0.066823,
        2.932153,1.452754,-2.738631,
        2.932153,1.454061,-2.950416,
        2.932153,1.456736,1.323154,
        2.932153,1.458370,0.400433,
        2.932153,1.459314,-1.765003,
        2.932153,1.459902,-1.145835,
        2.932153,1.464952,2.329245,
        2.932153,1.466050,0.862104,
        2.932153,1.471330,-1.886160,
        2.932153,1.475119,-1.648255,
        2.932153,1.480878,1.194576,
        2.932153,1.480942,0.273235,
        2.932153,1.481086,2.890269,
        2.932153,1.484118,-1.529156,
        2.932153,1.484741,-1.279219,
        2.932153,1.487681,-1.406069,
        2.932153,1.487886,-3.066037,
        2.932153,1.490555,2.543284,
        2.932153,1.492424,-0.190932,
        2.932153,1.495002,-0.839810,
        2.932153,1.495297,2.007188,
        2.932153,1.496414,-2.009267,
        2.932153,1.498261,-2.425270,
        2.932153,1.498907,-2.207327,
        2.932153,1.500204,0.737638,
        2.932153,1.502469,0.151156,
        2.932153,1.507448,-0.514782,
        2.932153,1.509287,2.204647,
        2.932153,1.510518,1.884920,
        2.932153,1.511948,-2.632360,
        2.932153,1.515235,1.068565,
        2.932153,1.522237,3.102018,
        2.932153,1.522373,1.764991,
        2.932153,1.522416,-0.954956,
        2.932153,1.523102,2.765920,
        2.932153,1.524811,-2.844718,
        2.932153,1.527261,-0.304916,
        2.932153,1.529788,0.613208,
        2.932153,1.533648,1.644808,
        2.932153,1.535018,0.028869,
        2.932153,1.541082,2.424608,
        2.932153,1.546010,1.523462,
        2.932153,1.551135,-0.642994,
        2.932153,1.553895,0.489607,
        2.932153,1.555041,0.947813,
        2.932153,1.559671,-1.074934,
        2.932153,1.560345,2.985899,
        2.932153,1.560487,-2.323400,
        2.932153,1.561127,1.401111,
        2.932153,1.567193,2.643172,
        2.932153,1.568892,2.098362,
        2.932153,1.571702,-2.098435,
        2.932153,1.572701,-2.528730,
        2.932153,1.573572,1.278593,
        2.932153,1.574849,-0.091505,
        2.932153,1.576334,0.365236,
        2.932153,1.577029,-2.962888,
        2.932153,1.577688,-1.814941,
        2.932153,1.579542,-0.755917,
        2.932153,1.579624,-1.199759,
        2.932153,1.581515,-0.419414,
        2.932153,1.584581,-2.738702,
        2.932153,1.586948,-1.696296,
        2.932153,1.588901,2.301496,
        2.932153,1.589936,0.825020,
        2.932153,1.590018,-1.941652,
        2.932153,1.596968,-1.328358,
        2.932153,1.597178,-1.575607,
        2.932153,1.600470,-1.453196,
        2.932153,1.601408,0.242409,
        2.932153,1.602259,1.152656,
        2.932153,1.603064,2.865774,
        2.932153,1.612085,1.968802,
        2.932153,1.612720,-3.080498,
        2.932153,1.616061,2.521531,
        2.932153,1.616900,-0.207457,
        2.932153,1.620988,-0.875920,
        2.932153,1.622114,0.701509,
        2.932153,1.622749,0.121426,
        2.932153,1.622993,-2.221629,
        2.932153,1.629346,-0.542669,
        2.932153,1.630007,1.844552,
        2.932153,1.634747,-2.426198,
        2.932153,1.642346,1.722617,
        2.932153,1.643925,1.034629,
        2.932153,1.644848,2.743774,
        2.932153,1.645258,2.188109,
        2.932153,1.645417,-2.634725,
        2.932153,1.647395,3.085437,
        2.932153,1.648538,-2.855044,
        2.932153,1.649932,0.576808,
        2.932153,1.652224,-0.996105,
        2.932153,1.652871,1.600681,
        2.932153,1.657997,-0.325127,
        2.932153,1.661936,0.001037,
        2.932153,1.665187,1.477402,
        2.932153,1.665496,2.397872,
        2.932153,1.673267,-0.665453,
        2.932153,1.673557,0.452526,
        2.932153,1.679277,-1.122421,
        2.932153,1.679696,0.912482,
        2.932153,1.682420,1.353922,
        2.932153,1.685031,-2.009038,
        2.932153,1.685196,2.967046,
        2.932153,1.690107,2.620910,
        2.932153,1.690453,-2.123076,
        2.932153,1.691693,-1.871572,
        2.932153,1.691817,2.061688,
        2.932153,1.692644,1.237670,
        2.932153,1.696386,0.328123,
        2.932153,1.697162,-1.249026,
        2.932153,1.697488,-2.322785,
        2.932153,1.698726,-2.973576,
        2.932153,1.699172,-1.749717,
        2.932153,1.704478,-0.106295,
        2.932153,1.707244,-0.444770,
        2.932153,1.707500,-2.531497,
        2.932153,1.707714,-0.787321,
        2.932153,1.708430,-1.627552,
        2.932153,1.710583,-1.376890,
        2.932153,1.712415,-2.747921,
        2.932153,1.713125,-1.503073,
        2.932153,1.713452,0.789419,
        2.932153,1.720612,2.279888,
        2.932153,1.724852,0.202831,
        2.932153,1.724868,2.846246,
        2.932153,1.728940,1.931535,
        2.932153,1.732838,1.127584,
        2.932153,1.735719,-3.092485,
        2.932153,1.739676,2.496017,
        2.932153,1.739971,-0.228876,
        2.932153,1.743328,0.663347,
        2.932153,1.744927,-0.911892,
        2.932153,1.745952,0.084733,
        2.932153,1.748077,1.804354,
        2.932153,1.752959,-0.567548,
        2.932153,1.760169,1.678132,
        2.932153,1.763887,-2.218450,
        2.932153,1.766737,2.722752,
        2.932153,1.767999,1.004737,
        2.932153,1.769149,1.550298,
        2.932153,1.770119,-2.427175,
        2.932153,1.770460,2.154208,
        2.932153,1.770514,3.070982,
        2.932153,1.770735,0.537245,
        2.932153,1.772673,-2.864760,
        2.932153,1.773527,-1.039582,
        2.932153,1.776419,-2.642421,
        2.932153,1.780769,1.416330,
        2.932153,1.786634,-0.346247,
        2.932153,1.788829,-1.947553,
        2.932153,1.793737,0.410270,
        2.932153,1.794553,2.374618,
        2.932153,1.794699,-0.694061,
        2.932153,1.797469,-1.169105,
        2.932153,1.803705,0.881076,
        2.932153,1.804124,1.264249,
        2.932153,1.807825,-2.090455,
        2.932153,1.808822,-1.817375,
        2.932153,1.809020,2.949741,
        2.932153,1.812466,2.597595,
        2.932153,1.813159,2.025395,
        2.932153,1.813816,-1.298435,
        2.932153,1.815016,0.278951,
        2.932153,1.816510,-0.008631,
        2.932153,1.817772,-1.689625,
        2.932153,1.821360,-2.984352,
        2.932153,1.822312,-0.127211,
        2.932153,1.823386,-1.429807,
        2.932153,1.823834,-1.560483,
        2.932153,1.832060,-0.821627,
        2.932153,1.832690,-0.468440,
        2.932153,1.833945,-2.320278,
        2.932153,1.834787,0.751968,
        2.932153,1.840579,-2.536445,
        2.932153,1.841033,-2.756932,
        2.932153,1.846350,1.893983,
        2.932153,1.846644,1.113766,
        2.932153,1.846833,2.826543,
        2.932153,1.847325,2.249051,
        2.932153,1.854119,0.134161,
        2.932153,1.858863,-3.106022,
        2.932153,1.865046,1.763384,
        2.932153,1.865619,-0.952036,
        2.932153,1.865912,0.624453,
        2.932153,1.866694,2.473686,
        2.932153,1.867524,-0.240083,
        2.932153,1.875561,1.632831,
        2.932153,1.876049,-0.594378,
        2.932153,1.880509,1.500204,
        2.932153,1.883504,1.360457,
        2.932153,1.886971,-2.196665,
        2.932153,1.888979,0.982922,
        2.932153,1.889815,2.701570,
        2.932153,1.892144,-2.005340,
        2.932153,1.892348,0.496522,
        2.932153,1.893281,-1.083976,
        2.932153,1.893828,3.053453,
        2.932153,1.895117,2.121107,
        2.932153,1.898741,-2.875735,
        2.932153,1.904429,-2.428029,
        2.932153,1.908408,-2.648903,
        2.932153,1.911828,-0.364489,
        2.932153,1.912224,0.366565,
        2.932153,1.914854,1.214468,
        2.932153,1.915467,-1.883143,
        2.932153,1.915763,-0.724679,
        2.932153,1.915790,-1.216366,
        2.932153,1.922256,2.347806,
        2.932153,1.922616,0.844633,
        2.932153,1.925185,0.231170,
        2.932153,1.927474,-1.754616,
        2.932153,1.928342,-1.351174,
        2.932153,1.933328,-1.623328,
        2.932153,1.933878,0.034150,
        2.932153,1.934585,-1.487979,
        2.932153,1.934802,2.927634,
        2.932153,1.934862,1.989776,
        2.932153,1.936327,-0.099350,
        2.932153,1.938088,2.578507,
        2.932153,1.945677,-2.999805,
        2.932153,1.954572,-0.859208,
        2.932153,1.957868,-0.490784,
        2.932153,1.958989,0.712772,
        2.932153,1.960212,-2.307561,
        2.932153,1.964774,1.855507,
        2.932153,1.964971,-2.100533,
        2.932153,1.969112,1.088273,
        2.932153,1.971438,-2.765415,
        2.932153,1.972312,2.805134,
        2.932153,1.974950,2.220596,
        2.932153,1.975135,-2.539127,
        2.932153,1.982107,1.720829,
        2.932153,1.983195,-3.127984,
        2.932153,1.983318,-0.247182,
        2.932153,1.986720,-0.993220,
        2.932153,1.990381,1.586005,
        2.932153,1.990470,0.580283,
        2.932153,1.990523,1.450326,
        2.932153,1.993909,1.312897,
        2.932153,1.994694,2.452813,
        2.932153,1.999226,-0.620717,
        2.932153,2.005145,0.941012,
        2.932153,2.007301,0.142337,
        2.932153,2.014149,0.444138,
        2.932153,2.014338,-1.127392,
        2.932153,2.018131,3.021778,
        2.932153,2.019676,-1.980944,
        2.932153,2.019912,2.686612,
        2.932153,2.020545,2.088672,
        2.932153,2.026074,0.300945,
        2.932153,2.026196,-2.889057,
        2.932153,2.031959,-1.268178,
        2.932153,2.032917,-2.199917,
        2.932153,2.033572,-1.842884,
        2.932153,2.034712,-2.419505,
        2.932153,2.035818,-0.379479,
        2.932153,2.036004,-0.757062,
        2.932153,2.040501,-1.411488,
        2.932153,2.042874,-1.701367,
        2.932153,2.043060,-2.653746,
        2.932153,2.043729,-1.557938,
        2.932153,2.048538,0.806623,
        2.932153,2.049006,1.199930,
        2.932153,2.049502,-0.137989,
        2.932153,2.052181,2.326959,
        2.932153,2.052794,-0.008526,
        2.932153,2.057608,1.951726,
        2.932153,2.065568,2.567730,
        2.932153,2.070371,2.879808,
        2.932153,2.071256,-3.018961,
        2.932153,2.079746,-0.897699,
        2.932153,2.082784,1.042403,
        2.932153,2.083752,-0.511086,
        2.932153,2.084401,1.810502,
        2.932153,2.085731,0.671081,
        2.932153,2.097496,1.373172,
        2.932153,2.099185,1.667006,
        2.932153,2.100482,-2.301084,
        2.932153,2.103118,2.194745,
        2.932153,2.103223,1.521685,
        2.932153,2.103695,-2.777707,
        2.932153,2.107523,3.129581,
        2.932153,2.108277,-2.095907,
        2.932153,2.108891,-2.532166,
        2.932153,2.111003,-0.268632,
        2.932153,2.112276,-1.027246,
        2.932153,2.114380,0.531457,
        2.932153,2.119166,0.100552,
        2.932153,2.119564,2.754818,
        2.932153,2.122474,-0.648314,
        2.932153,2.125183,2.445598,
        2.932153,2.129501,0.238290,
        2.932153,2.131178,0.386663,
        2.932153,2.131223,-1.174311,
        2.932153,2.134269,-1.948042,
        2.932153,2.134684,0.907065,
        2.932153,2.137615,2.988806,
        2.932153,2.145635,-1.327708,
        2.932153,2.146467,2.055081,
        2.932153,2.146582,-1.793788,
        2.932153,2.149210,-1.482321,
        2.932153,2.152901,-1.636788,
        2.932153,2.156337,1.151558,
        2.932153,2.156474,-2.912065,
        2.932153,2.157194,-0.792563,
        2.932153,2.166724,-0.401638,
        2.932153,2.166980,-0.134879,
        2.932153,2.173644,-2.403429,
        2.932153,2.175711,-2.655523,
        2.932153,2.177132,2.600419,
        2.932153,2.178879,0.767722,
        2.932153,2.180752,1.908485,
        2.932153,2.181338,2.313721,
        2.932153,2.185580,1.287673,
        2.932153,2.195400,-2.217546,
        2.932153,2.200418,-3.056442,
        2.932153,2.201829,-0.001735,
        2.932153,2.203804,1.756176,
        2.932153,2.211602,1.443930,
        2.932153,2.212686,-0.544809,
        2.932153,2.213937,0.621892,
        2.932153,2.214075,1.600691,
        2.932153,2.216256,1.020029,
        2.932153,2.219060,2.886631,
        2.932153,2.223375,-0.906185,
        2.932153,2.226002,-1.060248,
        2.932153,2.226413,-2.064034,
        2.932153,2.230554,2.171334,
        2.932153,2.232413,2.735777,
        2.932153,2.233709,3.070830,
        2.932153,2.234383,-2.792065,
        2.932153,2.237183,0.468799,
        2.932153,2.237695,-0.271388,
        2.932153,2.239419,0.151693,
        2.932153,2.243347,-2.521049,
        2.932153,2.244035,-1.229456,
        2.932153,2.244355,0.309899,
        2.932153,2.245175,2.460236,
        2.932153,2.245628,-1.899216,
        2.932153,2.245973,-0.702234,
        2.932153,2.255601,-1.728131,
        2.932153,2.256702,-1.401367,
        2.932153,2.261466,-1.557500,
        2.932153,2.267390,0.878499,
        2.932153,2.272099,2.018594,
        2.932153,2.278618,1.170386,
        2.932153,2.283304,-2.345584,
        2.932153,2.285175,-2.943090,
        2.932153,2.293000,-0.124451,
        2.932153,2.294058,-0.421070,
        2.932153,2.303283,1.855938,
        2.932153,2.304968,2.608978,
        2.932153,2.305347,2.311346,
        2.932153,2.307188,-2.656134,
        2.932153,2.307301,1.336200,
        2.932153,2.309655,0.726020,
        2.932153,2.317634,-2.185225,
        2.932153,2.321510,1.685664,
        2.932153,2.324376,-3.111006,
        2.932153,2.324731,1.511186,
        2.932153,2.329587,0.035498,
        2.932153,2.330158,2.977264,
        2.932153,2.335096,-0.588468,
        2.932153,2.336046,-1.115799,
        2.932153,2.338060,2.803613,
        2.932153,2.338083,-0.943400,
        2.932153,2.340210,0.562145,
        2.932153,2.341931,1.023926,
        2.932153,2.343265,-0.779061,
        2.932153,2.343434,-2.013692,
        2.932153,2.353029,-1.310529,
        2.932153,2.354868,0.211543,
        2.932153,2.355120,0.388593,
        2.932153,2.355165,-1.833173,
        2.932153,2.355462,2.151441,
        2.932153,2.362948,-2.487477,
        2.932153,2.363101,-2.810456,
        2.932153,2.363394,-0.272918,
        2.932153,2.364509,-1.644533,
        2.932153,2.370465,2.466824,
        2.932153,2.386110,-1.481351,
        2.932153,2.394960,0.862133,
        2.932153,2.395716,1.977184,
        2.932153,2.396339,1.200900,
        2.932153,2.407941,-2.316753,
        2.932153,2.409450,-2.988432,
        2.932153,2.417202,-0.103363,
        2.932153,2.417386,-0.443097,
        2.932153,2.417950,2.666662,
        2.932153,2.423190,1.789783,
        2.932153,2.423692,1.392696,
        2.932153,2.431122,3.086379,
        2.932153,2.431406,-2.647820,
        2.932153,2.431584,2.307631,
        2.932153,2.434090,1.592818,
        2.932153,2.435993,0.683844,
        2.932153,2.440898,-2.133300,
        2.932153,2.442222,-1.195837,
        2.932153,2.443112,2.883761,
        2.932153,2.449191,-0.633651,
        2.932153,2.450821,0.084741,
        2.932153,2.453499,-1.008861,
        2.932153,2.460010,-1.938011,
        2.932153,2.460199,-0.822008,
        2.932153,2.460902,0.490079,
        2.932153,2.461774,-1.734913,
        2.932153,2.463986,1.034407,
        2.932153,2.467647,0.287667,
        2.932153,2.479849,2.126925,
        2.932153,2.485907,-2.835471,
        2.932153,2.485945,-1.374853,
        2.932153,2.487747,2.505912,
        2.932153,2.487957,-0.274279,
        2.932153,2.489741,-2.468428,
        2.932153,2.512538,1.239900,
        2.932153,2.517528,1.923289,
        2.932153,2.518454,-1.565852,
        2.932153,2.519348,0.843503,
        2.932153,2.522766,-3.059231,
        2.932153,2.528686,2.737228,
        2.932153,2.534798,-2.266671,
        2.932153,2.536533,1.466274,
        2.932153,2.538504,1.699812,
        2.932153,2.539516,-0.476942,
        2.932153,2.539532,-0.074548,
        2.932153,2.543197,2.990252,
        2.932153,2.554022,2.320619,
        2.932153,2.556919,-2.655005,
        2.932153,2.558146,0.627191,
        2.932153,2.558223,-1.164189,
        2.932153,2.563452,-2.044028,
        2.932153,2.568462,0.150655,
        2.932153,2.569130,-0.700837,
        2.932153,2.570050,-1.804492,
        2.932153,2.572710,-0.934665,
        2.932153,2.575994,0.391718,
        2.932153,2.584774,1.046703,
        2.932153,2.601432,2.102535,
        2.932153,2.603559,2.555114,
        2.932153,2.604150,-2.885208,
        2.932153,2.608630,-1.378011,
        2.932153,2.611256,-0.279110,
        2.932153,2.617344,-2.439480,
        2.932153,2.627402,1.293552,
        2.932153,2.635396,1.848850,
        2.932153,2.635686,2.834950,
        2.932153,2.637087,3.132730,
        2.932153,2.642029,0.813274,
        2.932153,2.645260,1.570203,
        2.932153,2.646182,-1.614036,
        2.932153,2.660113,-2.186513,
        2.932153,2.660483,-0.034608,
        2.932153,2.661713,-0.529511,
        2.932153,2.672720,-1.110422,
        2.932153,2.676626,2.335353,
        2.932153,2.676638,0.540364,
        2.932153,2.676780,-1.901158,
        2.932153,2.681411,-2.667965,
        2.932153,2.682152,0.245751,
        2.932153,2.682859,-0.817071,
        2.932153,2.705107,1.062142,
        2.932153,2.717705,2.625995,
        2.932153,2.721001,2.060097,
        2.932153,2.722319,-2.954069,
        2.932153,2.729840,-1.368237,
        2.932153,2.734349,-0.289298,
        2.932153,2.738409,2.975671,
        2.932153,2.740366,1.374341,
        2.932153,2.744633,-2.391311,
        2.932153,2.747398,1.727166,
        2.932153,2.763799,0.760372,
        2.932153,2.768238,-1.679243,
        2.932153,2.779245,-2.046268,
        2.932153,2.780208,-0.626453,
        2.932153,2.780471,0.027732,
        2.932153,2.784696,-1.013128,
        2.932153,2.789817,0.395431,
        2.932153,2.801154,2.353143,
        2.932153,2.807091,-2.687477,
        2.932153,2.825762,1.085035,
        2.932153,2.830209,2.734943,
        2.932153,2.833517,-3.085508,
        2.932153,2.838301,1.980061,
        2.932153,2.849871,1.512012,
        2.932153,2.850089,-1.343899,
        2.932153,2.856905,-0.318441,
        2.932153,2.870445,-2.291001,
        2.932153,2.885017,0.655562,
        2.932153,2.886428,-1.793217,
        2.932153,2.890337,-0.825633,
        2.932153,2.898932,0.139973,
        2.932153,2.931927,-2.728768,
        2.932153,2.932096,2.377790,
        2.932153,2.937409,2.947245,
        2.932153,2.948048,1.133127,
        2.932153,2.950896,1.797889,
        2.932153,2.970019,-1.282382,
        2.932153,2.978333,-0.412090,
        2.932153,2.996437,-2.089121,
        2.932153,3.011274,0.434110,
        2.932153,3.050410,3.138492,
        2.932153,3.064492,1.647259,
        2.932153,3.088868,-0.989501,
        3.036873,0.053884,-2.408394,
        3.036873,0.077645,1.453433,
        3.036873,0.088408,-0.311160,
        3.036873,0.142314,2.590387,
        3.036873,0.147206,-1.385164,
        3.036873,0.163506,0.573203,
        3.036873,0.165485,-2.875396,
        3.036873,0.182564,-2.125821,
        3.036873,0.190569,1.929141,
        3.036873,0.201102,1.249082,
        3.036873,0.201397,-0.716646,
        3.036873,0.210568,-0.069559,
        3.036873,0.251672,2.848900,
        3.036873,0.266262,-1.678068,
        3.036873,0.269108,-1.215563,
        3.036873,0.273777,-2.506756,
        3.036873,0.279854,0.814581,
        3.036873,0.281664,-2.969147,
        3.036873,0.283625,2.405603,
        3.036873,0.284058,0.357350,
        3.036873,0.294107,1.593880,
        3.036873,0.302491,-0.400410,
        3.036873,0.308059,2.021491,
        3.036873,0.312219,-2.072057,
        3.036873,0.321905,-0.829121,
        3.036873,0.329984,1.201621,
        3.036873,0.339714,-0.013712,
        3.036873,0.368784,2.974974,
        3.036873,0.376428,-1.495532,
        3.036873,0.378103,2.646436,
        3.036873,0.385088,0.590355,
        3.036873,0.386248,-2.690943,
        3.036873,0.389851,-2.340031,
        3.036873,0.391502,-1.171374,
        3.036873,0.398249,-1.812694,
        3.036873,0.400542,-2.995729,
        3.036873,0.404273,0.918163,
        3.036873,0.405617,1.773533,
        3.036873,0.409009,-0.570243,
        3.036873,0.410700,0.270433,
        3.036873,0.415893,1.464890,
        3.036873,0.420680,2.088164,
        3.036873,0.424766,-0.258729,
        3.036873,0.426507,2.367872,
        3.036873,0.441292,-0.890727,
        3.036873,0.451839,-2.078193,
        3.036873,0.459270,1.191562,
        3.036873,0.470622,0.014002,
        3.036873,0.481292,-2.529098,
        3.036873,0.482460,2.849843,
        3.036873,0.492205,-1.397963,
        3.036873,0.493314,3.117264,
        3.036873,0.497870,0.727828,
        3.036873,0.499783,-1.654379,
        3.036873,0.502651,0.463333,
        3.036873,0.505463,2.594557,
        3.036873,0.506104,-2.874910,
        3.036873,0.513579,-1.147354,
        3.036873,0.518569,1.644237,
        3.036873,0.519001,1.900488,
        3.036873,0.519277,-0.673809,
        3.036873,0.524769,-0.421955,
        3.036873,0.525564,-2.288838,
        3.036873,0.532541,0.978737,
        3.036873,0.534610,2.154808,
        3.036873,0.535020,-1.891846,
        3.036873,0.540208,1.400497,
        3.036873,0.542006,0.228337,
        3.036873,0.550762,-0.180771,
        3.036873,0.560531,2.378474,
        3.036873,0.561629,-0.924576,
        3.036873,0.567465,-2.679392,
        3.036873,0.588527,1.181755,
        3.036873,0.591643,-2.098364,
        3.036873,0.594963,2.997575,
        3.036873,0.599643,-3.045235,
        3.036873,0.599962,2.773556,
        3.036873,0.602620,0.029270,
        3.036873,0.605711,0.598454,
        3.036873,0.607708,-1.541026,
        3.036873,0.609360,-2.464379,
        3.036873,0.610738,-1.326786,
        3.036873,0.616968,0.820489,
        3.036873,0.626379,1.773610,
        3.036873,0.628626,-1.751180,
        3.036873,0.630671,0.391949,
        3.036873,0.632622,-0.737642,
        3.036873,0.632708,2.566722,
        3.036873,0.632797,1.989858,
        3.036873,0.634455,-0.534790,
        3.036873,0.636205,1.561170,
        3.036873,0.637681,-1.123423,
        3.036873,0.641441,-2.842109,
        3.036873,0.644807,-0.327139,
        3.036873,0.652440,2.197129,
        3.036873,0.660897,-2.278593,
        3.036873,0.664008,1.360091,
        3.036873,0.670850,-1.945257,
        3.036873,0.671396,1.004859,
        3.036873,0.673985,0.203855,
        3.036873,0.680326,-0.133370,
        3.036873,0.685083,-0.942363,
        3.036873,0.688204,-2.632204,
        3.036873,0.689777,2.385011,
        3.036873,0.700065,3.115430,
        3.036873,0.704080,2.911636,
        3.036873,0.712883,1.174216,
        3.036873,0.715701,0.699906,
        3.036873,0.718851,-1.442215,
        3.036873,0.721019,2.721536,
        3.036873,0.725300,-2.975134,
        3.036873,0.727136,-1.633288,
        3.036873,0.727399,-2.120567,
        3.036873,0.730751,0.522488,
        3.036873,0.732892,-1.254330,
        3.036873,0.734333,-2.446184,
        3.036873,0.735125,0.037270,
        3.036873,0.737893,1.866463,
        3.036873,0.740650,1.681075,
        3.036873,0.745971,-0.781286,
        3.036873,0.749432,-0.442559,
        3.036873,0.750102,2.050636,
        3.036873,0.751046,0.863821,
        3.036873,0.757018,-0.608833,
        3.036873,0.757046,-1.816019,
        3.036873,0.757389,1.501184,
        3.036873,0.759800,2.546181,
        3.036873,0.759859,0.351923,
        3.036873,0.761141,-2.781102,
        3.036873,0.769035,-1.078574,
        3.036873,0.770166,-0.266604,
        3.036873,0.772218,2.227380,
        3.036873,0.788382,1.331051,
        3.036873,0.790269,-2.282409,
        3.036873,0.803410,-1.984567,
        3.036873,0.803813,0.191520,
        3.036873,0.806832,3.022155,
        3.036873,0.807476,-2.597675,
        3.036873,0.812512,-0.107335,
        3.036873,0.814161,1.010949,
        3.036873,0.817732,-3.084486,
        3.036873,0.817846,2.390688,
        3.036873,0.819323,2.845960,
        3.036873,0.819430,-0.915017,
        3.036873,0.829198,-1.532491,
        3.036873,0.829531,0.665695,
        3.036873,0.829816,1.167147,
        3.036873,0.833613,-1.357492,
        3.036873,0.842945,-2.907318,
        3.036873,0.844883,2.679708,
        3.036873,0.848189,-1.705003,
        3.036873,0.849888,1.767901,
        3.036873,0.851126,0.493370,
        3.036873,0.852911,1.933394,
        3.036873,0.857516,-2.432863,
        3.036873,0.858261,-1.191729,
        3.036873,0.858279,-2.138895,
        3.036873,0.859723,-0.734928,
        3.036873,0.859754,1.603292,
        3.036873,0.861655,-0.394821,
        3.036873,0.864300,0.043549,
        3.036873,0.866940,0.851229,
        3.036873,0.869748,2.095969,
        3.036873,0.877352,-0.574668,
        3.036873,0.879254,-2.733338,
        3.036873,0.882554,1.443757,
        3.036873,0.884060,0.334199,
        3.036873,0.885626,-1.866061,
        3.036873,0.889834,2.529957,
        3.036873,0.893140,2.251698,
        3.036873,0.897463,-0.240822,
        3.036873,0.898758,-1.031948,
        3.036873,0.913918,1.287618,
        3.036873,0.916555,-2.287173,
        3.036873,0.918305,3.107845,
        3.036873,0.919522,2.948277,
        3.036873,0.927138,-2.569686,
        3.036873,0.929031,0.185383,
        3.036873,0.933766,-3.015917,
        3.036873,0.934873,-2.010822,
        3.036873,0.935636,-1.445855,
        3.036873,0.936702,-0.866461,
        3.036873,0.937395,0.729108,
        3.036873,0.937910,2.787575,
        3.036873,0.939144,-0.092962,
        3.036873,0.941718,1.125085,
        3.036873,0.942873,-1.607680,
        3.036873,0.943701,0.980328,
        3.036873,0.947774,2.395256,
        3.036873,0.952433,0.584347,
        3.036873,0.954456,-1.287293,
        3.036873,0.958731,-2.852552,
        3.036873,0.961838,1.839284,
        3.036873,0.961898,-0.469425,
        3.036873,0.965426,1.686615,
        3.036873,0.970108,1.990771,
        3.036873,0.972820,-1.762780,
        3.036873,0.973500,-0.711881,
        3.036873,0.975253,2.641082,
        3.036873,0.977824,0.441928,
        3.036873,0.981485,-2.424766,
        3.036873,0.981877,1.537100,
        3.036873,0.982609,-1.136920,
        3.036873,0.984265,-2.156073,
        3.036873,0.989150,0.045493,
        3.036873,0.993052,2.137478,
        3.036873,0.997766,-2.693871,
        3.036873,0.999847,-0.334738,
        3.036873,1.001920,0.848054,
        3.036873,1.006146,1.389734,
        3.036873,1.012180,0.304066,
        3.036873,1.016938,2.270674,
        3.036873,1.017371,-0.979790,
        3.036873,1.019495,-1.894487,
        3.036873,1.023796,-0.583691,
        3.036873,1.026525,2.506945,
        3.036873,1.027004,-0.203653,
        3.036873,1.028027,3.036573,
        3.036873,1.031435,1.241021,
        3.036873,1.032316,2.884221,
        3.036873,1.035863,-3.110475,
        3.036873,1.042085,-2.292950,
        3.036873,1.043272,-1.522986,
        3.036873,1.046265,-2.958847,
        3.036873,1.049272,-1.376344,
        3.036873,1.050039,-2.552086,
        3.036873,1.053682,0.689863,
        3.036873,1.055458,0.170394,
        3.036873,1.056622,-0.831866,
        3.036873,1.057515,1.097998,
        3.036873,1.058706,-2.036438,
        3.036873,1.061389,-1.672248,
        3.036873,1.063758,2.743093,
        3.036873,1.070157,-0.073422,
        3.036873,1.072154,1.761506,
        3.036873,1.074029,1.909193,
        3.036873,1.074204,0.545276,
        3.036873,1.074714,-2.805147,
        3.036873,1.076262,0.966104,
        3.036873,1.077642,-1.227073,
        3.036873,1.084151,1.618525,
        3.036873,1.085085,2.377673,
        3.036873,1.086822,-0.441112,
        3.036873,1.087142,2.058815,
        3.036873,1.098383,-0.696430,
        3.036873,1.099138,-1.084033,
        3.036873,1.103208,1.477773,
        3.036873,1.103783,0.408254,
        3.036873,1.106275,-2.420800,
        3.036873,1.108079,2.612943,
        3.036873,1.109345,-2.172718,
        3.036873,1.115957,-1.787597,
        3.036873,1.118930,0.045983,
        3.036873,1.119777,0.834702,
        3.036873,1.121943,-2.669042,
        3.036873,1.124930,2.219917,
        3.036873,1.125217,1.336294,
        3.036873,1.130958,2.972931,
        3.036873,1.131400,-0.306629,
        3.036873,1.137135,-0.936514,
        3.036873,1.139125,-1.924528,
        3.036873,1.139877,0.276382,
        3.036873,1.142003,-3.058022,
        3.036873,1.147363,-0.558448,
        3.036873,1.150348,-0.176418,
        3.036873,1.151198,1.197016,
        3.036873,1.153930,-1.597104,
        3.036873,1.155389,2.837148,
        3.036873,1.155458,-1.460623,
        3.036873,1.157331,-2.906362,
        3.036873,1.158594,2.487618,
        3.036873,1.162103,-1.322553,
        3.036873,1.165407,3.092861,
        3.036873,1.168025,0.717075,
        3.036873,1.168406,-2.301519,
        3.036873,1.174782,-2.538755,
        3.036873,1.177998,1.061256,
        3.036873,1.178898,1.827305,
        3.036873,1.181331,-2.060330,
        3.036873,1.182540,-0.798952,
        3.036873,1.184569,0.147517,
        3.036873,1.184734,1.968919,
        3.036873,1.184804,0.593386,
        3.036873,1.188278,1.690808,
        3.036873,1.194360,2.110440,
        3.036873,1.194776,2.710455,
        3.036873,1.197745,-1.173705,
        3.036873,1.199898,-2.777257,
        3.036873,1.201164,2.355931,
        3.036873,1.203760,1.556404,
        3.036873,1.205845,-0.415993,
        3.036873,1.206585,0.933648,
        3.036873,1.206950,-0.037248,
        3.036873,1.209341,0.471162,
        3.036873,1.213483,-1.039417,
        3.036873,1.218146,-0.672643,
        3.036873,1.221608,1.421669,
        3.036873,1.226043,-1.693533,
        3.036873,1.232870,-1.822773,
        3.036873,1.234375,-2.421194,
        3.036873,1.234531,-2.190148,
        3.036873,1.239414,0.348959,
        3.036873,1.240183,2.587255,
        3.036873,1.242312,2.941234,
        3.036873,1.244075,1.286644,
        3.036873,1.245961,-2.997461,
        3.036873,1.246495,-2.646872,
        3.036873,1.247736,2.242810,
        3.036873,1.253604,0.812222,
        3.036873,1.257816,-0.894317,
        3.036873,1.258247,-1.953634,
        3.036873,1.258993,-0.284099,
        3.036873,1.262380,-0.157609,
        3.036873,1.262791,-1.549229,
        3.036873,1.263772,-0.533073,
        3.036873,1.266731,-1.412461,
        3.036873,1.268591,-3.122146,
        3.036873,1.269614,-1.275246,
        3.036873,1.270745,1.152533,
        3.036873,1.273606,0.224545,
        3.036873,1.279716,2.812101,
        3.036873,1.283786,2.462124,
        3.036873,1.284798,0.078209,
        3.036873,1.285402,-2.882049,
        3.036873,1.286936,1.886053,
        3.036873,1.289408,0.683122,
        3.036873,1.294736,1.756184,
        3.036873,1.296978,-2.311511,
        3.036873,1.298546,2.020601,
        3.036873,1.300924,1.022478,
        3.036873,1.303305,3.043152,
        3.036873,1.304210,-2.084225,
        3.036873,1.305241,-2.530249,
        3.036873,1.307513,1.628003,
        3.036873,1.310722,0.558267,
        3.036873,1.314142,2.139329,
        3.036873,1.316087,-0.777971,
        3.036873,1.317907,-1.121023,
        3.036873,1.321019,-2.749776,
        3.036873,1.321609,2.687717,
        3.036873,1.322568,1.499415,
        3.036873,1.322972,-0.395699,
        3.036873,1.325286,-0.994192,
        3.036873,1.329146,-0.648513,
        3.036873,1.329795,-0.042570,
        3.036873,1.331651,2.344497,
        3.036873,1.336522,0.435098,
        3.036873,1.339410,0.900087,
        3.036873,1.339550,1.370353,
        3.036873,1.340441,-1.731580,
        3.036873,1.351855,-1.855234,
        3.036873,1.360161,-3.018404,
        3.036873,1.361031,-1.613971,
        3.036873,1.361196,2.914780,
        3.036873,1.361998,-2.207254,
        3.036873,1.362302,1.239384,
        3.036873,1.365228,2.563808,
        3.036873,1.365452,0.310522,
        3.036873,1.367028,-2.420041,
        3.036873,1.373422,-1.484281,
        3.036873,1.374245,-1.226417,
        3.036873,1.375020,-0.170393,
        3.036873,1.376560,-1.355481,
        3.036873,1.377070,-1.980984,
        3.036873,1.377578,-2.636339,
        3.036873,1.378096,0.774626,
        3.036873,1.379086,-0.508192,
        3.036873,1.382562,0.177141,
        3.036873,1.388849,2.233847,
        3.036873,1.389268,-0.886349,
        3.036873,1.391324,1.109396,
        3.036873,1.391699,-0.288725,
        3.036873,1.392869,3.136903,
        3.036873,1.393612,1.952016,
        3.036873,1.395545,-2.850807,
        3.036873,1.401926,1.822930,
        3.036873,1.402149,2.788898,
        3.036873,1.409552,0.647980,
        3.036873,1.410840,0.056219,
        3.036873,1.411944,2.444377,
        3.036873,1.413639,1.697982,
        3.036873,1.426479,1.573427,
        3.036873,1.426725,2.102242,
        3.036873,1.427868,0.984637,
        3.036873,1.429580,-2.107232,
        3.036873,1.431375,-2.313269,
        3.036873,1.431907,3.015410,
        3.036873,1.432325,-1.029054,
        3.036873,1.433341,0.524268,
        3.036873,1.435181,-0.739771,
        3.036873,1.437208,-2.529085,
        3.036873,1.439041,-0.617966,
        3.036873,1.441473,1.448544,
        3.036873,1.444527,2.664744,
        3.036873,1.450513,-0.400625,
        3.036873,1.451907,-0.066823,
        3.036873,1.452754,-2.738631,
        3.036873,1.454061,-2.950416,
        3.036873,1.456736,1.323154,
        3.036873,1.458370,0.400433,
        3.036873,1.459314,-1.765003,
        3.036873,1.459902,-1.145835,
        3.036873,1.464952,2.329245,
        3.036873,1.466050,0.862104,
        3.036873,1.471330,-1.886160,
        3.036873,1.475119,-1.648255,
        3.036873,1.480878,1.194576,
        3.036873,1.480942,0.273235,
        3.036873,1.481086,2.890269,
        3.036873,1.484118,-1.529156,
        3.036873,1.484741,-1.279219,
        3.036873,1.487681,-1.406069,
        3.036873,1.487886,-3.066037,
        3.036873,1.490555,2.543284,
        3.036873,1.492424,-0.190932,
        3.036873,1.495002,-0.839810,
        3.036873,1.495297,2.007188,
        3.036873,1.496414,-2.009267,
        3.036873,1.498261,-2.425270,
        3.036873,1.498907,-2.207327,
        3.036873,1.500204,0.737638,
        3.036873,1.502469,0.151156,
        3.036873,1.507448,-0.514782,
        3.036873,1.509287,2.204647,
        3.036873,1.510518,1.884920,
        3.036873,1.511948,-2.632360,
        3.036873,1.515235,1.068565,
        3.036873,1.522237,3.102018,
        3.036873,1.522373,1.764991,
        3.036873,1.522416,-0.954956,
        3.036873,1.523102,2.765920,
        3.036873,1.524811,-2.844718,
        3.036873,1.527261,-0.304916,
        3.036873,1.529788,0.613208,
        3.036873,1.533648,1.644808,
        3.036873,1.535018,0.028869,
        3.036873,1.541082,2.424608,
        3.036873,1.546010,1.523462,
        3.036873,1.551135,-0.642994,
        3.036873,1.553895,0.489607,
        3.036873,1.555041,0.947813,
        3.036873,1.559671,-1.074934,
        3.036873,1.560345,2.985899,
        3.036873,1.560487,-2.323400,
        3.036873,1.561127,1.401111,
        3.036873,1.567193,2.643172,
        3.036873,1.568892,2.098362,
        3.036873,1.571702,-2.098435,
        3.036873,1.572701,-2.528730,
        3.036873,1.573572,1.278593,
        3.036873,1.574849,-0.091505,
        3.036873,1.576334,0.365236,
        3.036873,1.577029,-2.962888,
        3.036873,1.577688,-1.814941,
        3.036873,1.579542,-0.755917,
        3.036873,1.579624,-1.199759,
        3.036873,1.581515,-0.419414,
        3.036873,1.584581,-2.738702,
        3.036873,1.586948,-1.696296,
        3.036873,1.588901,2.301496,
        3.036873,1.589936,0.825020,
        3.036873,1.590018,-1.941652,
        3.036873,1.596968,-1.328358,
        3.036873,1.597178,-1.575607,
        3.036873,1.600470,-1.453196,
        3.036873,1.601408,0.242409,
        3.036873,1.602259,1.152656,
        3.036873,1.603064,2.865774,
        3.036873,1.612085,1.968802,
        3.036873,1.612720,-3.080498,
        3.036873,1.616061,2.521531,
        3.036873,1.616900,-0.207457,
        3.036873,1.620988,-0.875920,
        3.036873,1.622114,0.701509,
        3.036873,1.622749,0.121426,
        3.036873,1.622993,-2.221629,
        3.036873,1.629346,-0.542669,
        3.036873,1.630007,1.844552,
        3.036873,1.634747,-2.426198,
        3.036873,1.642346,1.722617,
        3.036873,1.643925,1.034629,
        3.036873,1.644848,2.743774,
        3.036873,1.645258,2.188109,
        3.036873,1.645417,-2.634725,
        3.036873,1.647395,3.085437,
        3.036873,1.648538,-2.855044,
        3.036873,1.649932,0.576808,
        3.036873,1.652224,-0.996105,
        3.036873,1.652871,1.600681,
        3.036873,1.657997,-0.325127,
        3.036873,1.661936,0.001037,
        3.036873,1.665187,1.477402,
        3.036873,1.665496,2.397872,
        3.036873,1.673267,-0.665453,
        3.036873,1.673557,0.452526,
        3.036873,1.679277,-1.122421,
        3.036873,1.679696,0.912482,
        3.036873,1.682420,1.353922,
        3.036873,1.685031,-2.009038,
        3.036873,1.685196,2.967046,
        3.036873,1.690107,2.620910,
        3.036873,1.690453,-2.123076,
        3.036873,1.691693,-1.871572,
        3.036873,1.691817,2.061688,
        3.036873,1.692644,1.237670,
        3.036873,1.696386,0.328123,
        3.036873,1.697162,-1.249026,
        3.036873,1.697488,-2.322785,
        3.036873,1.698726,-2.973576,
        3.036873,1.699172,-1.749717,
        3.036873,1.704478,-0.106295,
        3.036873,1.707244,-0.444770,
        3.036873,1.707500,-2.531497,
        3.036873,1.707714,-0.787321,
        3.036873,1.708430,-1.627552,
        3.036873,1.710583,-1.376890,
        3.036873,1.712415,-2.747921,
        3.036873,1.713125,-1.503073,
        3.036873,1.713452,0.789419,
        3.036873,1.720612,2.279888,
        3.036873,1.724852,0.202831,
        3.036873,1.724868,2.846246,
        3.036873,1.728940,1.931535,
        3.036873,1.732838,1.127584,
        3.036873,1.735719,-3.092485,
        3.036873,1.739676,2.496017,
        3.036873,1.739971,-0.228876,
        3.036873,1.743328,0.663347,
        3.036873,1.744927,-0.911892,
        3.036873,1.745952,0.084733,
        3.036873,1.748077,1.804354,
        3.036873,1.752959,-0.567548,
        3.036873,1.760169,1.678132,
        3.036873,1.763887,-2.218450,
        3.036873,1.766737,2.722752,
        3.036873,1.767999,1.004737,
        3.036873,1.769149,1.550298,
        3.036873,1.770119,-2.427175,
        3.036873,1.770460,2.154208,
        3.036873,1.770514,3.070982,
        3.036873,1.770735,0.537245,
        3.036873,1.772673,-2.864760,
        3.036873,1.773527,-1.039582,
        3.036873,1.776419,-2.642421,
        3.036873,1.780769,1.416330,
        3.036873,1.786634,-0.346247,
        3.036873,1.788829,-1.947553,
        3.036873,1.793737,0.410270,
        3.036873,1.794553,2.374618,
        3.036873,1.794699,-0.694061,
        3.036873,1.797469,-1.169105,
        3.036873,1.803705,0.881076,
        3.036873,1.804124,1.264249,
        3.036873,1.807825,-2.090455,
        3.036873,1.808822,-1.817375,
        3.036873,1.809020,2.949741,
        3.036873,1.812466,2.597595,
        3.036873,1.813159,2.025395,
        3.036873,1.813816,-1.298435,
        3.036873,1.815016,0.278951,
        3.036873,1.816510,-0.008631,
        3.036873,1.817772,-1.689625,
        3.036873,1.821360,-2.984352,
        3.036873,1.822312,-0.127211,
        3.036873,1.823386,-1.429807,
        3.036873,1.823834,-1.560483,
        3.036873,1.832060,-0.821627,
        3.036873,1.832690,-0.468440,
        3.036873,1.833945,-2.320278,
        3.036873,1.834787,0.751968,
        3.036873,1.840579,-2.536445,
        3.036873,1.841033,-2.756932,
        3.036873,1.846350,1.893983,
        3.036873,1.846644,1.113766,
        3.036873,1.846833,2.826543,
        3.036873,1.847325,2.249051,
        3.036873,1.854119,0.134161,
        3.036873,1.858863,-3.106022,
        3.036873,1.865046,1.763384,
        3.036873,1.865619,-0.952036,
        3.036873,1.865912,0.624453,
        3.036873,1.866694,2.473686,
        3.036873,1.867524,-0.240083,
        3.036873,1.875561,1.632831,
        3.036873,1.876049,-0.594378,
        3.036873,1.880509,1.500204,
        3.036873,1.883504,1.360457,
        3.036873,1.886971,-2.196665,
        3.036873,1.888979,0.982922,
        3.036873,1.889815,2.701570,
        3.036873,1.892144,-2.005340,
        3.036873,1.892348,0.496522,
        3.036873,1.893281,-1.083976,
        3.036873,1.893828,3.053453,
        3.036873,1.895117,2.121107,
        3.036873,1.898741,-2.875735,
        3.036873,1.904429,-2.428029,
        3.036873,1.908408,-2.648903,
        3.036873,1.911828,-0.364489,
        3.036873,1.912224,0.366565,
        3.036873,1.914854,1.214468,
        3.036873,1.915467,-1.883143,
        3.036873,1.915763,-0.724679,
        3.036873,1.915790,-1.216366,
        3.036873,1.922256,2.347806,
        3.036873,1.922616,0.844633,
        3.036873,1.925185,0.231170,
        3.036873,1.927474,-1.754616,
        3.036873,1.928342,-1.351174,
        3.036873,1.933328,-1.623328,
        3.036873,1.933878,0.034150,
        3.036873,1.934585,-1.487979,
        3.036873,1.934802,2.927634,
        3.036873,1.934862,1.989776,
        3.036873,1.936327,-0.099350,
        3.036873,1.938088,2.578507,
        3.036873,1.945677,-2.999805,
        3.036873,1.954572,-0.859208,
        3.036873,1.957868,-0.490784,
        3.036873,1.958989,0.712772,
        3.036873,1.960212,-2.307561,
        3.036873,1.964774,1.855507,
        3.036873,1.964971,-2.100533,
        3.036873,1.969112,1.088273,
        3.036873,1.971438,-2.765415,
        3.036873,1.972312,2.805134,
        3.036873,1.974950,2.220596,
        3.036873,1.975135,-2.539127,
        3.036873,1.982107,1.720829,
        3.036873,1.983195,-3.127984,
        3.036873,1.983318,-0.247182,
        3.036873,1.986720,-0.993220,
        3.036873,1.990381,1.586005,
        3.036873,1.990470,0.580283,
        3.036873,1.990523,1.450326,
        3.036873,1.993909,1.312897,
        3.036873,1.994694,2.452813,
        3.036873,1.999226,-0.620717,
        3.036873,2.005145,0.941012,
        3.036873,2.007301,0.142337,
        3.036873,2.014149,0.444138,
        3.036873,2.014338,-1.127392,
        3.036873,2.018131,3.021778,
        3.036873,2.019676,-1.980944,
        3.036873,2.019912,2.686612,
        3.036873,2.020545,2.088672,
        3.036873,2.026074,0.300945,
        3.036873,2.026196,-2.889057,
        3.036873,2.031959,-1.268178,
        3.036873,2.032917,-2.199917,
        3.036873,2.033572,-1.842884,
        3.036873,2.034712,-2.419505,
        3.036873,2.035818,-0.379479,
        3.036873,2.036004,-0.757062,
        3.036873,2.040501,-1.411488,
        3.036873,2.042874,-1.701367,
        3.036873,2.043060,-2.653746,
        3.036873,2.043729,-1.557938,
        3.036873,2.048538,0.806623,
        3.036873,2.049006,1.199930,
        3.036873,2.049502,-0.137989,
        3.036873,2.052181,2.326959,
        3.036873,2.052794,-0.008526,
        3.036873,2.057608,1.951726,
        3.036873,2.065568,2.567730,
        3.036873,2.070371,2.879808,
        3.036873,2.071256,-3.018961,
        3.036873,2.079746,-0.897699,
        3.036873,2.082784,1.042403,
        3.036873,2.083752,-0.511086,
        3.036873,2.084401,1.810502,
        3.036873,2.085731,0.671081,
        3.036873,2.097496,1.373172,
        3.036873,2.099185,1.667006,
        3.036873,2.100482,-2.301084,
        3.036873,2.103118,2.194745,
        3.036873,2.103223,1.521685,
        3.036873,2.103695,-2.777707,
        3.036873,2.107523,3.129581,
        3.036873,2.108277,-2.095907,
        3.036873,2.108891,-2.532166,
        3.036873,2.111003,-0.268632,
        3.036873,2.112276,-1.027246,
        3.036873,2.114380,0.531457,
        3.036873,2.119166,0.100552,
        3.036873,2.119564,2.754818,
        3.036873,2.122474,-0.648314,
        3.036873,2.125183,2.445598,
        3.036873,2.129501,0.238290,
        3.036873,2.131178,0.386663,
        3.036873,2.131223,-1.174311,
        3.036873,2.134269,-1.948042,
        3.036873,2.134684,0.907065,
        3.036873,2.137615,2.988806,
        3.036873,2.145635,-1.327708,
        3.036873,2.146467,2.055081,
        3.036873,2.146582,-1.793788,
        3.036873,2.149210,-1.482321,
        3.036873,2.152901,-1.636788,
        3.036873,2.156337,1.151558,
        3.036873,2.156474,-2.912065,
        3.036873,2.157194,-0.792563,
        3.036873,2.166724,-0.401638,
        3.036873,2.166980,-0.134879,
        3.036873,2.173644,-2.403429,
        3.036873,2.175711,-2.655523,
        3.036873,2.177132,2.600419,
        3.036873,2.178879,0.767722,
        3.036873,2.180752,1.908485,
        3.036873,2.181338,2.313721,
        3.036873,2.185580,1.287673,
        3.036873,2.195400,-2.217546,
        3.036873,2.200418,-3.056442,
        3.036873,2.201829,-0.001735,
        3.036873,2.203804,1.756176,
        3.036873,2.211602,1.443930,
        3.036873,2.212686,-0.544809,
        3.036873,2.213937,0.621892,
        3.036873,2.214075,1.600691,
        3.036873,2.216256,1.020029,
        3.036873,2.219060,2.886631,
        3.036873,2.223375,-0.906185,
        3.036873,2.226002,-1.060248,
        3.036873,2.226413,-2.064034,
        3.036873,2.230554,2.171334,
        3.036873,2.232413,2.735777,
        3.036873,2.233709,3.070830,
        3.036873,2.234383,-2.792065,
        3.036873,2.237183,0.468799,
        3.036873,2.237695,-0.271388,
        3.036873,2.239419,0.151693,
        3.036873,2.243347,-2.521049,
        3.036873,2.244035,-1.229456,
        3.036873,2.244355,0.309899,
        3.036873,2.245175,2.460236,
        3.036873,2.245628,-1.899216,
        3.036873,2.245973,-0.702234,
        3.036873,2.255601,-1.728131,
        3.036873,2.256702,-1.401367,
        3.036873,2.261466,-1.557500,
        3.036873,2.267390,0.878499,
        3.036873,2.272099,2.018594,
        3.036873,2.278618,1.170386,
        3.036873,2.283304,-2.345584,
        3.036873,2.285175,-2.943090,
        3.036873,2.293000,-0.124451,
        3.036873,2.294058,-0.421070,
        3.036873,2.303283,1.855938,
        3.036873,2.304968,2.608978,
        3.036873,2.305347,2.311346,
        3.036873,2.307188,-2.656134,
        3.036873,2.307301,1.336200,
        3.036873,2.309655,0.726020,
        3.036873,2.317634,-2.185225,
        3.036873,2.321510,1.685664,
        3.036873,2.324376,-3.111006,
        3.036873,2.324731,1.511186,
        3.036873,2.329587,0.035498,
        3.036873,2.330158,2.977264,
        3.036873,2.335096,-0.588468,
        3.036873,2.336046,-1.115799,
        3.036873,2.338060,2.803613,
        3.036873,2.338083,-0.943400,
        3.036873,2.340210,0.562145,
        3.036873,2.341931,1.023926,
        3.036873,2.343265,-0.779061,
        3.036873,2.343434,-2.013692,
        3.036873,2.353029,-1.310529,
        3.036873,2.354868,0.211543,
        3.036873,2.355120,0.388593,
        3.036873,2.355165,-1.833173,
        3.036873,2.355462,2.151441,
        3.036873,2.362948,-2.487477,
        3.036873,2.363101,-2.810456,
        3.036873,2.363394,-0.272918,
        3.036873,2.364509,-1.644533,
        3.036873,2.370465,2.466824,
        3.036873,2.386110,-1.481351,
        3.036873,2.394960,0.862133,
        3.036873,2.395716,1.977184,
        3.036873,2.396339,1.200900,
        3.036873,2.407941,-2.316753,
        3.036873,2.409450,-2.988432,
        3.036873,2.417202,-0.103363,
        3.036873,2.417386,-0.443097,
        3.036873,2.417950,2.666662,
        3.036873,2.423190,1.789783,
        3.036873,2.423692,1.392696,
        3.036873,2.431122,3.086379,
        3.036873,2.431406,-2.647820,
        3.036873,2.431584,2.307631,
        3.036873,2.434090,1.592818,
        3.036873,2.435993,0.683844,
        3.036873,2.440898,-2.133300,
        3.036873,2.442222,-1.195837,
        3.036873,2.443112,2.883761,
        3.036873,2.449191,-0.633651,
        3.036873,2.450821,0.084741,
        3.036873,2.453499,-1.008861,
        3.036873,2.460010,-1.938011,
        3.036873,2.460199,-0.822008,
        3.036873,2.460902,0.490079,
        3.036873,2.461774,-1.734913,
        3.036873,2.463986,1.034407,
        3.036873,2.467647,0.287667,
        3.036873,2.479849,2.126925,
        3.036873,2.485907,-2.835471,
        3.036873,2.485945,-1.374853,
        3.036873,2.487747,2.505912,
        3.036873,2.487957,-0.274279,
        3.036873,2.489741,-2.468428,
        3.036873,2.512538,1.239900,
        3.036873,2.517528,1.923289,
        3.036873,2.518454,-1.565852,
        3.036873,2.519348,0.843503,
        3.036873,2.522766,-3.059231,
        3.036873,2.528686,2.737228,
        3.036873,2.534798,-2.266671,
        3.036873,2.536533,1.466274,
        3.036873,2.538504,1.699812,
        3.036873,2.539516,-0.476942,
        3.036873,2.539532,-0.074548,
        3.036873,2.543197,2.990252,
        3.036873,2.554022,2.320619,
        3.036873,2.556919,-2.655005,
        3.036873,2.558146,0.627191,
        3.036873,2.558223,-1.164189,
        3.036873,2.563452,-2.044028,
        3.036873,2.568462,0.150655,
        3.036873,2.569130,-0.700837,
        3.036873,2.570050,-1.804492,
        3.036873,2.572710,-0.934665,
        3.036873,2.575994,0.391718,
        3.036873,2.584774,1.046703,
        3.036873,2.601432,2.102535,
        3.036873,2.603559,2.555114,
        3.036873,2.604150,-2.885208,
        3.036873,2.608630,-1.378011,
        3.036873,2.611256,-0.279110,
        3.036873,2.617344,-2.439480,
        3.036873,2.627402,1.293552,
        3.036873,2.635396,1.848850,
        3.036873,2.635686,2.834950,
        3.036873,2.637087,3.132730,
        3.036873,2.642029,0.813274,
        3.036873,2.645260,1.570203,
        3.036873,2.646182,-1.614036,
        3.036873,2.660113,-2.186513,
        3.036873,2.660483,-0.034608,
        3.036873,2.661713,-0.529511,
        3.036873,2.672720,-1.110422,
        3.036873,2.676626,2.335353,
        3.036873,2.676638,0.540364,
        3.036873,2.676780,-1.901158,
        3.036873,2.681411,-2.667965,
        3.036873,2.682152,0.245751,
        3.036873,2.682859,-0.817071,
        3.036873,2.705107,1.062142,
        3.036873,2.717705,2.625995,
        3.036873,2.721001,2.060097,
        3.036873,2.722319,-2.954069,
        3.036873,2.729840,-1.368237,
        3.036873,2.734349,-0.289298,
        3.036873,2.738409,2.975671,
        3.036873,2.740366,1.374341,
        3.036873,2.744633,-2.391311,
        3.036873,2.747398,1.727166,
        3.036873,2.763799,0.760372,
        3.036873,2.768238,-1.679243,
        3.036873,2.779245,-2.046268,
        3.036873,2.780208,-0.626453,
        3.036873,2.780471,0.027732,
        3.036873,2.784696,-1.013128,
        3.036873,2.789817,0.395431,
        3.036873,2.801154,2.353143,
        3.036873,2.807091,-2.687477,
        3.036873,2.825762,1.085035,
        3.036873,2.830209,2.734943,
        3.036873,2.833517,-3.085508,
        3.036873,2.838301,1.980061,
        3.036873,2.849871,1.512012,
        3.036873,2.850089,-1.343899,
        3.036873,2.856905,-0.318441,
        3.036873,2.870445,-2.291001,
        3.036873,2.885017,0.655562,
        3.036873,2.886428,-1.793217,
        3.036873,2.890337,-0.825633,
        3.036873,2.898932,0.139973,
        3.036873,2.931927,-2.728768,
        3.036873,2.932096,2.377790,
        3.036873,2.937409,2.947245,
        3.036873,2.948048,1.133127,
        3.036873,2.950896,1.797889,
        3.036873,2.970019,-1.282382,
        3.036873,2.978333,-0.412090,
        3.036873,2.996437,-2.089121,
        3.036873,3.011274,0.434110,
        3.036873,3.050410,3.138492,
        3.036873,3.064492,1.647259,
        3.036873,3.088868,-0.989501
    };

    _Num_rot_angles  = sizeof(Zangle)/sizeof(Zangle[0])/3;
    _Zangle   = new float[_Num_rot_angles*3];

    for( int i = 0 ; i < _Num_rot_angles*3 ; i++ ) {
        _Zangle[i] = Zangle[i];
    }
    //*/
    return;
}
